/* $Id: agentapi.h,v 1.27 2001/09/30 15:02:54 jm Exp $
 * API support routines for agents.
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifndef AGENTAPI_H
#define AGENTAPI_H

#include "../config.h"

#include <sys/socket.h>

#ifdef DYN_TARGET_LINUX
#include <sys/un.h>
typedef struct sockaddr_un dyn_api_sockaddr;
#endif
#ifdef DYN_TARGET_WINDOWS
typedef struct sockaddr_in dyn_api_sockaddr;
#endif

#include "dyn_api.h"

#define API_DATA_SIZE 16000
#define SOCK_BUFFER_SIZE 1024*1500

/* function codes */

enum api_functions {
	API_DUMMY,
	API_GET_TUNNELS,
	API_GET_TUNNEL_INFO,
	API_DESTROY_TUNNEL,
	API_ENABLE_MOBILE,
	API_GET_CAREOF_ADDR,
	API_GET_TUNNELING_MODE,
	API_GET_STATUS,
	API_CONNECT,
	API_DISCONNECT,
	API_UPDATE_LOCATION,
	API_CONFIRM,
	API_CANCEL,
	API_FORCE_FA,
	API_UPDATE_LOCATION_BLOCK,
	API_GET_FA_LIST,
	API_GET_FA_INFO,
	API_IW_SET_CH,
	API_IW_GET_CH,
	API_RESCAN,
	API_ATTACH_INFO_SOCKET,
	API_DETACH_INFO_SOCKET,
	API_REGISTER_DEV_INFO_SOCKET,
	API_POLICY_OFF,
	API_POLICY_ON,
	API_GET_POLICY,
	API_GET_MON_CONF,
	API_GET_MON_CONF_VAR,
	API_SET_MON_CONF_VAR,
	API_RELOAD_CONFIG,
	API_SEND_ADV,
	API_GET_BINDINGS	
};

/* message types */

#define API_CALL_MSG  0
#define API_REPLY_MSG 1

/* api message */

struct api_msg {
	__u16 type;			/* type of message */
	__s16 code;			/* function or return value */
        __u8  mac[6];                   /* MAC Address of Client */
        __u32 ip;                       /* IP Address of Client */
	__u16 length;			/* length of params */
	__u8 params[API_DATA_SIZE];	/* additional parameter data */
};


#define API_MSG_FIXED_LEN ((int)&((struct api_msg *) 0)->params)


int api_open_socket(char *path, char *group, char *owner, int permissions);
void api_close_socket(int s);
int api_receive(int s, dyn_api_sockaddr *addr, socklen_t *addrlen,
		struct api_msg *msg);
int api_send(int s, dyn_api_sockaddr *addr, socklen_t addrlen,
	     struct api_msg *msg);
int api_send_reply(int s, dyn_api_sockaddr *addr, socklen_t addrlen, int code,
                   unsigned char *data, int datalen);
int api_open_socket(char *, char *, char *, int);

#endif
