#include <time.h>
#include <string.h>
#include <arpa/inet.h>

#include <net/if.h>
#include <util/util.h>
#include <util/list.h>
#include <util/hash.h>
#include "ha.h"
#include "fa.h"
#include "proxy_types.h"
#include "proxy_util.h"
#include "proxy_papi.h"
#include <amapi/amapi.h>
#include <amapi/amapi_response.h>
#include <amapi/amapi_common.h>
#include "binding.h"
#include <dyn_api.h>

static int faMode_changefa(sList *, char *);
static int faMode_changeha(sList *, char *);
static int faMode_changemobgbl(sList *, char *);
static int faConfig_foreignagent(sList *, char *);
static int faConfig_homeagent(sList *, char *);
static int faConfig_securemobile(sList *, char *);
static int faConfig_secureforeign(sList *, char *);
static int faDelete_secureforeign(sList *, char *);
static int faConfig_foreignsecure(sList *, char *);
static int faDelete_foreignsecure(sList *, char *);
static int faShow_foreignagent(sList *, char *);
static int faShow_homeagent(sList *, char *);
static int faShow_habindings(sList *, char *);


static int faEnConfig_securemobile(AMAPICommandObject *, char *);
static int faEnConfig_secureforeign(AMAPICommandObject *, char *);
static int faEnConfig_foreignsecure(AMAPICommandObject *, char *);
static int faShow_secure(AMAPICommandObject *, char *);
static int faShow_hasecure(AMAPICommandObject *, char *);

/* Helper Function */
static int faShow_config(sList *, char *);
//static int faShow_secure(sList *, char *);
static int faShow_status(sList *, char *);
static int faShow_pending(sList *, char *);
static int faCheck_unconfirmed(struct node *node, void *data);
extern char *arubaCliActionEncrypt(int, const char *);

//static int faShow_fapending(sList *, char *);


#define FA_CREATE_NAMEVALUEHDR(b, t, tn) { \
  start_table(b, tn, 2); \
  strcpy(t[0], "Name"); \
  strcpy(t[1], "Value"); \
}


#define FA_CREATE_REGSTATHDR(b, t, n) { \
  start_table(b, "Registration Rejection Statistics", n); \
  strcpy(t[0], "Unknown Extension"); \
  strcpy(t[2], "Unknown Vendor Extension"); \
  strcpy(t[1], "Malformed Packet"); \
 }



u_int8_t bInitConf = FALSE;
extern struct fa_config *config;


typedef int (*cli_handler_t)(sList *list, char *);
typedef int (*encr_cli_handler_t)(AMAPICommandObject *, char *);
static char *buffer;

cli_handler_t cmd_table[] = {
	/*  0 */  faMode_changefa,
	/*  1 */  faMode_changeha, 
	/*  2 */  faMode_changemobgbl, 
	/*  3 */  faConfig_homeagent,
	/*  4 */  faConfig_securemobile,
	/*  5 */  NULL,
	/*  6 */  faDelete_secureforeign,
	/*  7 */  faConfig_secureforeign,
	/*  8 */  NULL,
	/*  9 */  NULL,
	/*  10 */ faConfig_foreignagent,
	/*  11 */ faConfig_foreignsecure,
	/*  12 */ faDelete_foreignsecure,
	/*  13 */ NULL,
	/*  14 */ NULL,
	/*  15 */ NULL,
	/*  16 */ NULL,
	/*  17 */ NULL,
	/*  18 */ NULL,
	/*  18 */ NULL,
	/*  20 */ NULL,
	/*  21 */ faShow_homeagent,
	/*  22 */ faShow_habindings,
	/*  23 */ NULL,
	/*  24 */ NULL,
	/*  25 */ NULL,
	/*  26 */ NULL,
	/*  27 */ NULL,
	/*  28 */ NULL,
	/*  29 */ NULL,
	/*  30 */ NULL,
	/*  31 */ faShow_foreignagent
};

encr_cli_handler_t encr_cmdtable[] = {
	/*  0 */  NULL,
	/*  1 */  NULL, 
	/*  2 */  NULL, 
	/*  3 */  NULL,
	/*  4 */  faEnConfig_securemobile,
	/*  5 */  NULL,
	/*  6 */  NULL,
	/*  7 */  faEnConfig_secureforeign,
	/*  8 */  NULL,
	/*  9 */  NULL,
	/*  10 */ NULL,
	/*  11 */ NULL,
	/*  12 */ NULL,
	/*  13 */ NULL,
	/*  14 */ NULL,
	/*  15 */ NULL,
	/*  16 */ faEnConfig_secureforeign,
	/*  17 */ faShow_hasecure,
	/*  18 */ faShow_secure,
	/*  19 */ faEnConfig_foreignsecure,
	/*  20 */ faEnConfig_securemobile
};




#define NUM_CMDS	sizeof(cmd_table)/sizeof(cli_handler_t)

#ifdef APP_SAVE_CONFIG
#define FA_SAVE_CMD(fc, tc) { \
 list_init_node(&((fc)->node)); \
  if(list_is_empty(&config->facmdbuf)) \
    list_add_head(&config->facmdbuf, &((fc)->node)); \
  else \
    list_add_tail(&config->facmdbuf, &((fc)->node)); \
  sprintf((fc)->cmd, "%s", (tc)); \
}

static int
faRemoveCmd(int k)
{
  FACmdNode_t *f = NULL;
  
  if(list_is_empty(&config->facmdbuf))
    return OK;
  
  f = (FACmdNode_t *)list_get_first(&config->facmdbuf);
  while(f) {
    if(f->key == k) {
      list_remove(&f->node);
      mmFree_memory(__FUNCTION__, f);
      return OK;
    }
    f = (FACmdNode_t *)list_get_next(&f->node);
  }
  return OK;
}
#endif

static int
faMode_changefa(sList *kvalues, char *buffer)
{
  sprintf(buffer, "PROMPT=(config-fa) ");
  return OK;
}


static int
faMode_changeha(sList *kvalues, char *buffer)
{
  sprintf(buffer, "PROMPT=(config-ha) ");
  return OK;
}


static int
faMode_changemobgbl(sList *kvalues, char *buffer)
{
  sprintf(buffer, "PROMPT=(config-mobagent) ");
  return OK;
}

static int
faConfig_homeagent(sList *kvalues, char *buffer)
{
  return haConfig_homeagent(kvalues, buffer);
}

static int
faConfig_securemobile(sList *kvalues, char *buffer)
{
  return haConfig_securemobile(kvalues, buffer);
  
}


static int
faEnConfig_securemobile(AMAPICommandObject *obj, char *buffer)
{
  int rc;
  rc= haEnConfig_securemobile((Opaque)obj, buffer);
  return rc;
}

static int
faConfig_secureforeign(sList *kvalues, char *buffer)
{

  return haConfig_secureforeign(kvalues, buffer);
}

static int
faEnConfig_secureforeign(AMAPICommandObject *obj, char *buffer)
{
  return haEnConfig_secureforeign((Opaque)obj, buffer);
}

static int
faDelete_secureforeign(sList *kvalues, char *buffer)
{

  return haDelete_secureforeign(kvalues, buffer);
}


static int
faConfig_foreignagent(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;

  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("registration_life" , kv->key)) 
      config->fa_default_tunnel_lifetime = atoi(kv->value);
    else if(0 == strcasecmp("BINDINGS", kv->name)) 
      config->max_bindings = atoi(kv->value);
    else if(0 == strcasecmp("PENDING", kv->name)) 
      config->max_pending = atoi(kv->value);
    else if(0 == strcmp("PENDING-TIME", kv->name)) 
      config->delete_pending_after = atoi(kv->value);
  }

  return OK;
}

static int
faConfig_foreignsecure(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spi = 0, ip = 0;
  uchar shared_secret[MAXSHAREDSECRETLEN];
  FACmdNode_t *fc = NULL;
#ifdef APP_SAVE_CONFIG
  char tcmd[MAXLINE];
#endif
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi", kv->name))
      spi = atoi(kv->value);
    else if(0 == strcasecmp("shared_secret", kv->key))
      memcpy((char *)shared_secret, kv->value, MAXSHAREDSECRETLEN);
    else if(0 == strcasecmp("ip", kv->key))
      ip = str_to_ip(kv->value);
    else {
      sprintf(buffer, "Unkown command parameter %s", kv->key);
      return ERROR;
    }
  }

  ASSERT(spi && ip && strlen(shared_secret));

  if(mmfa_spilist_add(spi, ip, shared_secret)) {
    sprintf(buffer, "Error adding security association with HA");
    return ERROR;
  }
  
#ifdef APP_SAVE_CONFIG
  fc = mmGet_memory(__FUNCTION__, sizeof(FACmdNode_t));
  sprintf(tcmd, "secure %s %d %s\n", ip_to_str(ip), spi, shared_secret);
  FA_SAVE_CMD(fc, tcmd);
  fc->key = spi;
#endif
  return OK;
}



static int
faEnConfig_foreignsecure(AMAPICommandObject *obj, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spi = 0, ip = 0;
  int rc = OK;
  uchar shared_secret[MAXSHAREDSECRETLEN];
  sList *kvalues = obj->keyValues;
  FACmdNode_t *fc = NULL;
#ifdef APP_SAVE_CONFIG
  char tcmd[MAXLINE];
#endif
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi", kv->name))
      spi = atoi(kv->value);
    else if(0 == strcasecmp("shared_secret", kv->key)) {
      if(amapiIsConfigEncrypted()) {
	char s[STR_LEN_LARGE];
	if(proxyDecryptSecret(kv->value, s)) {
	  LOG2(LOG_ERR, "Error decrypting shared secret");
	  sprintf(buffer, "Error decrypting shared secret");
	  return ERROR; 
	}
	if(strlen(s) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	}
        strlcpy(shared_secret, s, sizeof(shared_secret)); 
      }
      else {
	DEBUG(DEBUG_FLAG, "This is an unencrypted secret");
	if(strlen(kv->value) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	} 
	strlcpy((char *)shared_secret, kv->value, sizeof(shared_secret));
      }
    }
    else if(0 == strcasecmp("ip", kv->key))
      ip = str_to_ip(kv->value);
    else {
      sprintf(buffer, "Unkown command parameter %s", kv->key);
      return ERROR;
    }
  }

  ASSERT(spi && ip && strlen(shared_secret));

  if((rc = mmfa_spilist_add(spi, ip, shared_secret))) {
    if(ERROR == rc) {
      sprintf(buffer, "Error adding security association with HA");
      return ERROR;
    }
    return OK;
  }
  
#ifdef APP_SAVE_CONFIG
  fc = (FACmdNode_t *)mmGet_memory(__FUNCTION__, sizeof(FACmdNode_t));
  sprintf(tcmd, "secure host %s spi %d shared-secret %s\n", ip_to_str(ip), spi, shared_secret);
  FA_SAVE_CMD(fc, tcmd);
  fc->key = spi;
#endif
  return OK;
}



static int
faDelete_foreignsecure(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spi = 0;
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key))
      spi = atoi(kv->value);
    else {
      sprintf(buffer, "Unkown command parameter %s", kv->key);
      return ERROR;
    }
  }

  ASSERT(spi);

  if(mmfa_spilist_del(spi)) {
    sprintf(buffer, "Error deleting security association with HA");
    return ERROR;
  }
  
#ifdef APP_SAVE_CONFIG
  faRemoveCmd(spi);
#endif
  return OK;
}



static int
faShow_homeagent(sList *kvalues, char *buffer)
{

  return haShow_homeagent(kvalues, buffer);

}

static int
faShow_hasecure(AMAPICommandObject *obj, char *buffer)
{
  DEBUG(DEBUG_FLAG, "In %s", __FUNCTION__);
  return haShow_secure((Opaque)obj, buffer);
}

static int
faShow_habindings(sList *kvalues, char *buffer)
{

  return haShow_bindings(kvalues, buffer);

}




static int
faShow_foreignagent(sList *kvalues, char *buffer)
{
  int pc = 0;
  
  start_result(buffer,MAXMESG);
  faShow_config(kvalues, buffer);
  pc = faShow_status(kvalues, buffer);
  if(pc)
    faShow_pending(kvalues, buffer);
  end_result(buffer);

  return OK;
}

static int
faShow_config(sList *kvalues, char *buffer)
{
  char tunmode[MAXLINE];
  char        *  tablerow[STR_LEN_LARGE]; 
  int i = 0;

  for(i = 0; i < 2; i++) {
    if(NULL == (tablerow[i] = (char *)mmGet_memory(__FUNCTION__,STR_LEN_LARGE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return ERROR;
    }
  }

  FA_CREATE_NAMEVALUEHDR(buffer, tablerow, "FA Configuration");
  add_header(buffer, (u_int8_t **)&tablerow);
    
  if((config->enable_reverse_tunneling) && (config->enable_triangle_tunneling))
    sprintf(tunmode, "Reverse Triangle");
  else if(config->enable_reverse_tunneling)
    sprintf(tunmode, "Reverse");
  else if (config->enable_triangle_tunneling)
    sprintf(tunmode, "Triangle");
  else
    sprintf(tunmode, "Unknown");
  

  sprintf(tablerow[0], "Switch IP Address");
  sprintf(tablerow[1], "%s", inet_ntoa(config->highest_fa_addr));
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "UDP Port");
  sprintf(tablerow[1], "%d", config->udp_port);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Tunnel Lifetime (secs)");
  sprintf(tablerow[1], "%d", config->fa_default_tunnel_lifetime);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Tunnel Mode");
  sprintf(tablerow[1], "%s", tunmode);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Required");
  sprintf(tablerow[1], "%s", (config->reg_required)? "TRUE":"FALSE");
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Bindings Limit");
  sprintf(tablerow[1], "%d", config->max_bindings);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Pending Limit");
  sprintf(tablerow[1], "%d", config->max_pending);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Pending Lifetime Limit (secs)");
  sprintf(tablerow[1], "%d", config->delete_pending_after);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);
  add_data(buffer, "NewLine", "", FALSE);
  for(i = 0; i < 2; i++) 
    mmFree_memory(__FUNCTION__,tablerow[i]);
  return OK;
}


static int
faShow_secure(AMAPICommandObject *obj, char *buffer)
{
  char temp[MAXLINE];
  struct fa_spi_entry *sa = NULL;
  
  start_result(buffer,MAXMESG);
  sprintf(temp, "Authentication Alogorithm: %s", mmAuth_Algo(FA_DEFAULT_AUTH_ALGO));
  add_data(buffer, " ", temp, FALSE);
  
  for (sa = (struct fa_spi_entry *)list_get_first(&config->fa_spi_list); sa != NULL;
       sa = (struct fa_spi_entry *)list_get_next(&sa->node)) {  
    char tmp[MAXSHAREDSECRETLEN];
    sprintf(temp, "SPI: %d\tHome Agent: %s\tShared Secret: ",sa->spi, inet_ntoa(sa->addr));
    if(TRUE == amapiGetEncryptFlag(obj)) {
      memset(tmp, '*', strlen(sa->shared_secret));
      tmp[strlen(sa->shared_secret)] = '\0';
    }
    else 
      strcpy(tmp, sa->shared_secret);
    strcat(temp, tmp);
    add_data(buffer, " ", temp, FALSE);
  }
 
  end_result(buffer);
  return OK;
}


static int
faShow_status(sList *kvalues, char *buffer)
{
  char        *  tablerow[STR_LEN_LARGE]; 
  extern struct dynamics_fa_status stats;
  extern struct binding_counters bcounters;
  int i = 0;
  
  for(i = 0; i < 5; i++) {
    if(NULL == (tablerow[i] = (char *)mmGet_memory(__FUNCTION__,STR_LEN_LARGE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return ERROR;
    }
  }
  
  FA_CREATE_NAMEVALUEHDR(buffer, tablerow, "FA Status");
  add_header(buffer, (u_int8_t **)&tablerow);

   sprintf(tablerow[0], "Total Tunnels");
  sprintf(tablerow[1], "%d", stats.tunnel_count);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Total Bindings");
  sprintf(tablerow[1], "%d", bcounters.bindingcount);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Total Pending");
  sprintf(tablerow[1], "%d", stats.pending_count);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Requests Accepted");
  sprintf(tablerow[1], "%ld", stats.req_accepted);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Requests Rejected");
  sprintf(tablerow[1], "%ld", stats.req_rejected);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Replies Accepted");
  sprintf(tablerow[1], "%ld", stats.rep_accepted);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Replies Rejected");
  sprintf(tablerow[1], "%ld", stats.rep_rejected);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);

  add_data(buffer, "NewLine", "", FALSE);

  FA_CREATE_REGSTATHDR(buffer, tablerow, 3);
  add_header(buffer, (u_int8_t **)&tablerow);
  sprintf(tablerow[0], "%ld", stats.discarded_unknown_ext);
  sprintf(tablerow[1], "%ld", stats.discarded_malformed_msg);
  sprintf(tablerow[2], "%ld", stats.discarded_vendor_ext);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);
  
  for(i = 0; i < 5; i++) 
    mmFree_memory(__FUNCTION__,tablerow[i]);

  return stats.pending_count;
}



static int
faShow_pending(sList *kvalues, char *buffer)
{
  extern struct bindingtable *bindings_hash;
  char *tablerow[STR_LEN_LARGE]; 
  int inti = 0;
  
  for(inti = 0; inti < 4; inti++) {
    if(NULL == (tablerow[inti] = (char *)mmGet_memory(__FUNCTION__,MAXLINE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return -1;
    }
  }
  
  start_table(buffer, "Pending Requests", 4); 
  strcpy(tablerow[0], "ID"); 
  strcpy(tablerow[1], "MN"); 
  strcpy(tablerow[2], "HA"); 
  strcpy(tablerow[3], "Elapsed(secs)"); 

  add_header(buffer, (u_int8_t **)&tablerow);
  
  binding_iterator(bindings_hash, faCheck_unconfirmed, (char **)&tablerow);
  
  end_table(buffer);
  
  for(inti = 0; inti < 4; inti++) {
    mmFree_memory(__FUNCTION__,tablerow[inti]);
  }
 
  return 0;
}


static int
faCheck_unconfirmed(struct node *node, void *data) 
{
  struct tunnel_data *tunnel = (struct tunnel_data *)((struct bindingentry *)node)->data;
  struct unconfirmed_request *unc;
  char **tablerow = (char **)data;
  // char *tablerow[4]; 

  if(tunnel->confirmed)
    return 0;

 /*  for(inti = 0; inti < 4; inti++) { */
/*     if(NULL == (tablerow[inti] = (char *)mmGet_memory(__FUNCTION__,MAXLINE))) { */
/*       LOG2(LOG_ERR, "Unable to allocate memory"); */
/*       return -1; */
/*     } */
/*   } */

  unc = tunnel->unc_req;
  while (unc != NULL) {
    sprintf(*tablerow, "%d%d", unc->id[0], unc->id[1]);
    strcpy(((*tablerow)+1), inet_ntoa(unc->req.home_addr));
    strcpy(((*tablerow)+2), inet_ntoa(unc->ha_addr)); 
    sprintf(((*tablerow)+3), "%g", difftime(time(0),unc->created));
    
    add_row(buffer, (u_int8_t **)tablerow);
    unc = unc->next;
  }

 /*  for(inti = 0; inti < 4; inti++) { */
/*     mmFree_memory(__FUNCTION__,tablerow[inti]); */
/*   } */
  
  return 0;
}


#ifndef APP_SAVE_CONFIG
static AMAPIKeyValue *
faCreate_amapikv(AMAPIKeyValue *okv)
{
  AMAPIKeyValue * kv;
  
  kv = (AMAPIKeyValue *)mmGet_memory(__FUNCTION__,sizeof(AMAPIKeyValue));
  if(!kv) {
    LOG2(LOG_CRIT, "Error allocating memory for amapikv object");
    return NULL;
  }
  strcpy(kv->name,  okv->name);
  strcpy(kv->key,   okv->key);
  strcpy(kv->value, okv->value);
  if(okv->prevword!=NULL)
	strcpy(kv->prevword, okv->prevword);
  
  return(kv);
}

static AMAPICommandMode *
faCreate_amapicm(AMAPICommandMode *ocm)
{
  AMAPICommandMode * cm;
  
  cm = (AMAPICommandMode *)mmGet_memory(__FUNCTION__,sizeof(AMAPICommandMode));
  if(!cm) {
    LOG2(LOG_CRIT, "Error allocating memory for amapicm object");
    return NULL;
  }
  strcpy(cm->modeName,  ocm->modeName);
  strcpy(ocm->modeInstance, ocm->modeInstance);
  
  return(cm);
}


static AMAPICommandObject *
faCreate_amapiobj (AMAPICommandObject *ocmd)
{
  AMAPICommandObject * cmd = NULL;
  AMAPICommandMode *cm = NULL;
  AMAPIKeyValue *kv = NULL;

  cmd = (AMAPICommandObject *) mmGet_memory(__FUNCTION__,sizeof(AMAPICommandObject));
  if(!cmd) {
    LOG2(LOG_CRIT, "Error allocating memory for Amapi Command Object");
    return NULL;
  }

  bzero((char *)cmd, sizeof (AMAPICommandObject));
  cmd->appId  = ocmd->appId;
  cmd->objectType   = ocmd->objectType;
  cmd->optype       = AMAPI_SET;
  cmd->bIsLocalCommand = ocmd->bIsLocalCommand;
  cmd->majorVer = ocmd->majorVer;
  cmd->minorVer = cmd->minorVer;
  strncpy(cmd->instance, ocmd->instance, MAX_VALUE_LEN);
  cmd->bIsLocalCommand = ocmd->bIsLocalCommand;
  cmd->bIsRemoteCommand = ocmd->bIsRemoteCommand ;
  strncpy(cmd->szRemoteIpAddr, ocmd->szRemoteIpAddr, MAX_IP_ADDR_LEN);
  cmd->keyValues    = ListCreate();
  cmd->commandModes = ListCreate();

  ListGet (ocmd->keyValues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (ocmd->keyValues, GET_NEXT))
	 != NULL) {
    ListInsert(cmd->keyValues, faCreate_amapikv(kv));
  }
  ListGet (ocmd->commandModes, RESET);
  while ((cm = (AMAPICommandMode *) ListGet (ocmd->commandModes, GET_NEXT))
	 != NULL) {
    ListInsert(cmd->commandModes, faCreate_amapicm(cm));
  }

  return (cmd);
}
#endif


extern __u8 *
executeCommandObject (AMAPICommandObject * command, AMAPIRetMessage * msg)
{
  int objtype = command->objectType;
  AMAPIKeyValue *kv = NULL;
  
  if (msg != NULL)
    msg->errNo = AMAPI_SUCCESS;
  
  DEBUG(DEBUG_INFO, " KEY-VALUE PAIRS FOR OBJTYPE %d:\n", 
	command->objectType);
  ListGet (command->keyValues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (command->keyValues, GET_NEXT))
	 != NULL) {
    DEBUG(DEBUG_INFO, "  Name: %s Key: %s, Value: %s\n", 
	  kv->name, kv->key, kv->value);
  }

  buffer = (char *)mmGet_memory(__FUNCTION__,5 * MAXMESG);
  
  if (objtype < 0 || objtype >= NUM_CMDS) {
    DEBUG (DEBUG_INFO, "Wrong Object type %d", objtype);
    if (msg != NULL)
       msg->errNo = AMAPI_GENERAL_FAILURE;
    return NULL;
  }
  else {
    if((objtype > 15 && objtype <= 20) || objtype == 4 || objtype == 7) {
      if((*encr_cmdtable[objtype])(command, buffer)) {
        if (msg != NULL) {
	  msg->errNo = AMAPI_GENERAL_FAILURE;
	}
	return NULL;
      }
    }
    else {
      if((*cmd_table[objtype])(command->keyValues, buffer)) {
        if (msg != NULL) {
	  msg->errNo = AMAPI_GENERAL_FAILURE;
	}
	return NULL;
      }
    }
  }

  return buffer;
  
}


	
extern u_int16_t
initializeCommandObjects (sList * command_list)
{
  AMAPICommandObject * command;

  bInitConf = TRUE;
  ListGet (command_list, RESET);
  while ((command = (AMAPICommandObject *) ListGet (command_list, GET_NEXT))
	 != NULL) {
    executeCommandObject (command, NULL);
    proxyFree_clibuf();
  }
  bInitConf = FALSE;
  return (AMAPI_SUCCESS);
}

extern u_int16_t
reinitializeCommandObjects (sList * commandObjs)
{
  LOG2(LOG_INFO, "Received reinitializeCommandObjects command");
  return (initializeCommandObjects (commandObjs));
}


#ifdef APP_SAVE_CONFIG
extern sList * 
getCommandObjects (AMAPIGetCommandObjectAction nAction,AMAPIConfigType nLocal)
{
  int szBuf = 0, szHa = 0, maxsz = (MAXMESG*5)-3;
  char *pFabuf = NULL, temp[STR_LEN_LARGE];
  int bEncrypt = amapiShouldEncryptSecrets(nAction);

  if(bEncrypt) // saving some buffer space to accomodate bigger size of encrypted secrets
    maxsz -= 300;

  szBuf = haGetgblcmdstrs(config->cmdbuf, bEncrypt); 
  pFabuf = config->cmdbuf + szBuf;

  DEBUG(DEBUG_CONFIG, "%s", config->cmdbuf);
  DEBUG(DEBUG_FLAG, "szBuf = %d", szBuf);
  sprintf(temp, "  foreign-agent parameters %d bindings %d pending %d pending-time %d\n", 
	  config->fa_default_tunnel_lifetime, config->max_bindings, config->max_pending, 
	  config->delete_pending_after);
  if(maxsz > (szBuf + strlen(temp))) {
    strcat(pFabuf, temp);
    szBuf += strlen(temp);
  }
  strcat(pFabuf, "!\n");
  szBuf += 2;

  pFabuf += strlen(temp)+2;
  szHa +=  haGetcmdstrs(pFabuf, maxsz - szBuf, bEncrypt); 
  DEBUG(DEBUG_FLAG, "szHa = %d", szHa);
  pFabuf +=  szHa;
  szBuf += szHa;
  //  DEBUG(DEBUG_CONFIG, "%s", config->cmdbuf);
  DEBUG(DEBUG_FLAG, "szBuf = %d", szBuf);

  if(!list_is_empty(&config->facmdbuf) && maxsz > (szBuf + strlen(temp))) {
    sprintf(pFabuf, "foreign-agent\n");
    FACmdNode_t *f = (FACmdNode_t *)list_get_first(&config->facmdbuf);
    while(f && maxsz > (szBuf + strlen(f->cmd))) {
      char *pStr = NULL, cmd[STR_LEN_LARGE];
      int szSecret = 0;
      sprintf(temp, "  ");
      strcpy(cmd, f->cmd);
      if(bEncrypt && (pStr = strstr(cmd, "shared-secret"))) {
	    char *estr= NULL;
	    char *shared_secret = pStr + strlen("shared-secret") + 1; 
	    estr= arubaCliActionEncrypt(bEncrypt, shared_secret);
	    //DEBUG(DEBUG_INFO, "Clear Text = %s, Encrypted = %s", shared_secret, estr);
	    *shared_secret = '\0';
	    strcat(temp, cmd);
	    strcat(temp, estr);
	    szSecret = strlen(estr)/2;
      }
      else 
	    strcat(temp, cmd);
      strcat(temp, "\n");
      DEBUG(DEBUG_FLAG, "Command = %s", temp);
      szBuf += strlen(temp)+szSecret;
      strcat(pFabuf, temp);
      f = (FACmdNode_t *)list_get_next(&(f->node));
    }



#if 0
    for(inti = 0; inti < config->indCurCmd && szBuf+strlen(config->facmdbuf[inti])+3 < maxsz; 
	inti++) {
      char *pStr = NULL, cmd[STR_LEN_LARGE];
      int  szSecret = 0;
      strcat(temp, "  ");
      strcpy(cmd, config->facmdbuf[inti]);
      if(bEncrypt && (pStr = strstr(cmd, "shared-secret"))) {
	char estr[STR_LEN_LARGE];
	char *shared_secret = pStr + strlen("shared-secret") + 1; 
	proxyEncryptSecret(shared_secret, estr);
	*shared_secret = '\0';
	strcat(temp, cmd);
	strcat(temp, estr);
	szSecret = strlen(estr);
      }
      else
	strcat(temp, config->facmdbuf[inti]);
      strcat(temp, "\n");
      DEBUG(DEBUG_CONFIG, "FA secure command = %s", temp);
      //  pFabuf += strlen(config->facmdbuf[inti])+2;     
      //  temp += strlen(config->facmdbuf[inti])+2; 
    }
#endif
    strcat(pFabuf, "!\n");
    //    DEBUG(DEBUG_FLAG, "Temp = %s, pFabuf = %p", temp, pFabuf);
    // sprintf(pFabuf, "%s", temp);
    DEBUG(DEBUG_CONFIG, "FA Config = %s", pFabuf);
  }

  DEBUG(DEBUG_CONFIG, "%s", config->cmdbuf);
  
  return (sList *)config->cmdbuf;
}
#else
extern sList * 
getCommandObjects (AMAPIGetCommandObjectAction nAction,AMAPIConfigType nLocal)
{
  AMAPIKeyValue *kv = NULL;
  AMAPICommandObject *command;
  sList *cList = ListCreate();
  
  DEBUG(DEBUG_CONFIG, "Returning current configuration");
  ListGet ((sList *)config->cmdlist, RESET);
  while ((command = (AMAPICommandObject *) ListGet ((sList *)config->cmdlist, GET_NEXT))
	 != NULL) {
    DEBUG(DEBUG_INFO, " KEY-VALUE PAIRS FOR OBJTYPE %d:\n", 
	  command->objectType);
    ListGet (command->keyValues, RESET);
    while ((kv = (AMAPIKeyValue *) ListGet (command->keyValues, GET_NEXT))
	   != NULL) {
      DEBUG(DEBUG_INFO, "  Name: %s Key: %s, Value: %s\n", 
	    kv->name, kv->key, kv->value);
    }
    ListInsert(cList, faCreate_amapiobj(command));
  }

  return cList;
}
#endif

void
proxyFree_clibuf()
{
  if(buffer) {
    mmFree_memory(__FUNCTION__,buffer);
    buffer = NULL;
  }
}

