/* $Id: ha.h,v 1.13 2001/02/17 18:19:09 jm Exp $
 * Header file for Home Agent
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifndef HA_H
#define HA_H

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "owntypes.h"
#include "proxy_types.h"
#include "list.h"
#include "util/list.h"

#define HA_CONF_FILE         "arubaha.conf"
#define HA_GLOBAL_CONF_FILE  SYSCONFDIR "/" HA_CONF_FILE 

/* The process id of the daemon is saved here to make it easier to
   kill the correct daemon when necessary. */
#define HA_PID_FILE          PIDDIR "/arubaha.pid"

#ifndef ETH_ALEN
#define ETH_ALEN 6
#endif

enum { ENCAPS_IPIP, ENCAPS_MINIMAL, ENCAPS_GRE };
enum { AUTH_RFC2002, AUTH_RFC2002BIS };

extern u_int8_t bHaInit;

struct interface_entry {
	struct node node;

	/* data from conf file */
	char dev[IFNAMSIZ + 1];
        int  type;
	int ha_disc; /* allow dynamics HA discovery */
	int agentadv; /* whether to send agent advertisements or not */
	int interval; /* interval to send advertisements */
	struct in_addr force_addr; /* which source address to use if multiple
				    * addresses available for interface */

	/* other data */
	int if_index;
	struct in_addr addr; /* local address to be used for this interface */
	struct in_addr bcaddr; /* broadcast address of this interfaced; used
				* used with HA discovery */
	struct timeval last_adv;
	int icmp_sock, udp_sock, udp_bc_sock, udp_bc_sock2, udp_packet;
};


struct ha_tunnel_data {
	struct in_addr lower_saddr; /* source address of the request */
	__u32 nonce; /* last sent nonce */
	__u32 prev_nonce; /* last sent nonce */
	unsigned char auth_type; /* AUTH_RFC2002 for RFC 2002bis MAC
				  * calculation (i.e., include the SPI field);
				  * AUTH_RFC2002BIS for RFC 2002 MAC
				  * calculation */
	int reverse_tunnel; /* 0 = triangle, 1 = reverse */
	int encapsulation; /* encapsulation type */
	char arp_if[IFNAMSIZ]; /* interface used for proxy ARP entry */
	__u32 arp_subnet;/* interface used for proxy ARP entry */
	char mac[ETH_ALEN];
	__u16 curvlan; /* Vlan on which the MN is currently associated  on the FA */
	__u32 cursubnet; /* Subnet on which the MN is currently associated on the FA */
        __s32 stationloc; /* Building Floor location of AP with which the station is associated */
        int   masquerade_timer; /* Timer to send Gratuitous ARP every 1 min if station masquerade enabled */
	time_t last_failure_time; /* time of the last failure reply */
};

/* ha.c */
extern int    ha_init(u_int32_t);
extern void   haCleanup();
extern int    haHandle_api(int, int);
extern int    haHandle_regmsg(struct interface_entry *, int, char *, int, struct sockaddr_in);
extern void   haTimeout_handler();
extern void   haCheck_bindings();
extern int    haReload_config(int);
extern int    haSet_exprtimer(struct timeval *);
extern int    haTunnel_delete(u_int32_t);

/* ha_cli.c */
extern int haConfig_homeagent(sList *, char *);
extern int haShow_homeagent(sList *, char *);
extern int haShow_bindings(sList *, char *);
extern int haShow_secure(Opaque, char *);
extern int haConfig_securemobile(sList *, char *);
extern int haConfig_secureforeign(sList *, char *);
extern int haEnConfig_securemobile(Opaque, char *);
extern int haEnConfig_secureforeign(Opaque, char *);
extern int haDelete_secureforeign(sList *, char *);
extern int haGetcmdstrs(char *, int, int);
extern int haGetgblcmdstrs(char *, int);
extern struct bindingtable * haGet_BindingTable();
extern uint8_t ha_check_binding_present(uint32_t home_addr);
#endif /* HA_H */
