#include <time.h>
#include <string.h>
#include <arpa/inet.h>
#include <net/if.h>
#include "ha.h"
#include "ha_config.h"
#include "proxy_types.h"
#include "proxy_util.h"
#include "proxy_papi.h"
#include <util/util.h>
#include <util/list.h>
#include <util/hash.h>
#include <amapi/amapi.h>
#include <amapi/amapi_response.h>
#include <amapi/amapi_common.h>
#include <dyn_api.h>
#include <binding.h>


static int haShow_status(sList *, char *);
static int haShow_config(sList *, char *);

extern struct    ha_config haconfig;
extern u_int8_t  bInitConf;
extern char *arubaCliActionEncrypt(int, const char *);

#define HA_CREATE_BINDHDR(b, t, n) { \
  start_table(b, "Mobile Bindings", n); \
  strcpy(t[0], "Id"); \
  strcpy(t[1], "MAC Address"); \
  strcpy(t[2], "IP"); \
  strcpy(t[3], "Home VLAN"); \
  strcpy(t[4], "Care-Of Address"); \
  strcpy(t[5], "Current Location"); \
 }

#define HA_CREATE_REGSTATHDR(b, t, n) { \
  start_table(b, "Registration Rejection Statistics", n); \
  strcpy(t[0], "Unknown Extension"); \
  strcpy(t[2], "Unknown Vendor Extension"); \
  strcpy(t[1], "Malformed Packet"); \
  strcpy(t[3], "Wrong Type"); \
  strcpy(t[4], "ID Mismatch"); \
 }


#define HA_CREATE_NAMEVALUEHDR(b, t, tn) { \
  start_table(b, tn, 2); \
  strcpy(t[0], "Name"); \
  strcpy(t[1], "Value"); \
}


#ifdef APP_SAVE_CONFIG
#define HA_SAVE_CMD(hc, tc) { \
 list_init_node(&((hc)->node)); \
 if(list_is_empty(&haconfig.cmdbuf)) \
  list_add_head(&haconfig.cmdbuf, &((hc)->node)); \
 else \
  list_add_tail(&haconfig.cmdbuf, &((hc)->node)); \
 sprintf((hc)->cmd, "%s", tc); \
}

static int
haRemoveCmd(int k)
{
  HACmdNode_t *h = NULL;
  
  if(list_is_empty(&haconfig.cmdbuf))
    return OK;
  
  h = (HACmdNode_t *)list_get_first(&haconfig.cmdbuf);
  while(h) {
    if(h->key == k) {
      list_remove(&h->node);
      mmFree_memory(__FUNCTION__, h);
      return OK;
    }
    h = (HACmdNode_t *)list_get_next(&h->node);
  }
  return OK;
}
#endif


int
haConfig_homeagent(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;

  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcmp("registration_life" , kv->key)) 
      haconfig.ha_default_tunnel_lifetime = atoi(kv->value);
    else if(0 == strcmp("max_bindings", kv->key)) 
      haconfig.max_bindings = atoi(kv->value);
  }
  
  return OK;
}


int
haConfig_securemobile(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spistart = 0;
  uchar shared_secret[MAXSHAREDSECRETLEN];
 
#ifndef WITH_DP
  return OK;
#endif
  
  DEBUG(DEBUG_FLAG, "In %s", __FUNCTION__);
  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key)) 
      spistart = atoi(kv->value);
    else if(0 == strcmp("shared_secret", kv->key))
      memcpy((char *)shared_secret, kv->value, MAXSHAREDSECRETLEN);
    //    else if(0 == strcmp("binding_life", kv->key)) 
    //      maxlife = atoi(kv->value);
    else {
      sprintf(buffer, "Unknown command parameter %s", kv->key);
      return ERROR;
    }
  }
    
  if(mmha_authlistadd(spistart)) {
    sprintf(buffer, "Error adding security association with MN");
    return ERROR;
  }
  
  DEBUG(DEBUG_INFO, "Sending maxlife = %d", haconfig.ha_default_tunnel_lifetime);
  if(mmha_mnspiadd(spistart, haconfig.ha_default_tunnel_lifetime, shared_secret)) {
    sprintf(buffer, "Error adding security association with MN");
    return ERROR;
  }

  return OK;
}


int
haEnConfig_securemobile(Opaque o, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spistart = 0;
  uchar shared_secret[MAXSHAREDSECRETLEN];
  AMAPICommandObject *obj = (AMAPICommandObject *)o;
  sList *kvalues = obj->keyValues;
#ifndef WITH_DP
  return OK;
#endif

  DEBUG(DEBUG_FLAG, "In %s", __FUNCTION__);
  bzero(shared_secret, MAXSHAREDSECRETLEN);
  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key)) 
      spistart = atoi(kv->value);
    else if(0 == strcasecmp("shared_secret", kv->key)) {
     DEBUG(DEBUG_FLAG, "Encrypt Flag = %d", obj->encryptFlag);
     printf("%d %d\n", strlen(kv->value),  MAXSHAREDSECRETLEN);
     if(amapiIsConfigEncrypted()) {
        char s[STR_LEN_LARGE];
	if(proxyDecryptSecret(kv->value, s))  {
	  LOG2(LOG_ERR, "Error decrypting shared secret");
	  sprintf(buffer, "Error decrypting shared secret");
	  return ERROR; 
	}
        if(strlen(s) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	}
        strlcpy(shared_secret, s, sizeof(shared_secret)); 
     }
     else {
	DEBUG(DEBUG_FLAG, "This is an unencrypted secret");
        if(strlen(kv->value) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	} 
	strlcpy((char *)shared_secret, kv->value, sizeof(shared_secret));
     }
     DEBUG(DEBUG_FLAG, "Decrypted secret = %s", shared_secret);
    }
    else {
      sprintf(buffer, "Unknown command parameter %s", kv->key);
      return ERROR;
    }
  }
    
  if(mmha_authlistadd(spistart)) {
    sprintf(buffer, "Error adding security association with MN");
    return ERROR;
  }
 
  DEBUG(DEBUG_INFO, "Sending maxlife = %d", haconfig.ha_default_tunnel_lifetime);
  if(mmha_mnspiadd(spistart, haconfig.ha_default_tunnel_lifetime, shared_secret)) {
    sprintf(buffer, "Error adding security association with MN");
    return ERROR;
  }
  printf("Hello\n");
   
  return OK;
}



int
haConfig_secureforeign(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spistart = 0, network=0, netmask=0xffffffff;
  uchar shared_secret[MAXSHAREDSECRETLEN];
  HACmdNode_t *hc = NULL;
#ifdef APP_SAVE_CONFIG
  char tcmd[MAXLINE];
#endif
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key)) 
      spistart = atoi(kv->value);
    else if(0 == strcmp("shared_secret", kv->key)) 
      memcpy((char *)shared_secret, kv->value, MAXSHAREDSECRETLEN);
    else if(0 == strcmp("ip", kv->key)) {
      network = str_to_ip(kv->value);
      netmask = 0xffffffff;
    }
    else {
      sprintf(buffer, "Unkown command parameter %s", kv->key);
      return ERROR;
    }
  }

  if(mmha_faspiadd(spistart, network, shared_secret)) {
    sprintf(buffer, "Error adding security association with FA");
    return ERROR;
  }

#ifdef APP_SAVE_CONFIG
  hc = mmGet_memory(__FUNCTION__, sizeof(HACmdNode_t)); 
  sprintf(tcmd, "secure-foreign spi %d %s %s\n", spistart, ip_to_str(network), shared_secret);
  HA_SAVE_CMD(hc, tcmd);
  hc->key = spistart;
#endif

  return OK;
}




int
haEnConfig_secureforeign(Opaque o, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spistart = 0, network=0, netmask=0xffffffff;
  int rc = OK;
  uchar shared_secret[MAXSHAREDSECRETLEN];
  AMAPICommandObject *obj = (AMAPICommandObject *)o;
  sList *kvalues = obj->keyValues;
  HACmdNode_t *hc = NULL;
#ifdef APP_SAVE_CONFIG
  char tcmd[MAXLINE];
#endif
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key)) 
      spistart = atoi(kv->value);
    else if(0 == strcmp("shared_secret", kv->key)) {
      LOG2(LOG_INFO, "Getting secret from %s", kv->value);
  
      if(amapiIsConfigEncrypted()) {
	char s[STR_LEN_LARGE];
	if(proxyDecryptSecret(kv->value, s))  {
	  sprintf(buffer, "Error decrypting shared secret");
	  LOG2(LOG_ERR, "Error decrypting shared secret");
	  return ERROR; 
	}
        if(strlen(s) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	}
        strlcpy(shared_secret, s, sizeof(shared_secret)); 
      }
      else {
	DEBUG(DEBUG_FLAG, "This is an unencrypted secret");
	if(strlen(kv->value) > MAXSHAREDSECRETLEN) {
	  LOG2(LOG_ERR, "Shared secret too long. Should  be <= 32 bytes");
	  sprintf(buffer, "Shared secret too long. Should  be <= 32 bytes");
	  return ERROR;
	} 
	strlcpy((char *)shared_secret, kv->value, sizeof(shared_secret));
      }
      
      DEBUG(DEBUG_FLAG, "Secret = %s", shared_secret);
    }
    else if(0 == strcmp("ip", kv->key)) {
      network = str_to_ip(kv->value);
      netmask = 0xffffffff;
    }
    else {
      sprintf(buffer, "Unknown command parameter %s", kv->key);
      return ERROR;
    }
  }

  if((rc = mmha_faspiadd(spistart, network, shared_secret))) {
    if(ERROR == rc) {
      sprintf(buffer, "Error adding security association with FA");
      return ERROR;
    }
    return OK;
  }
  

#ifdef APP_SAVE_CONFIG
  hc =  mmGet_memory(__FUNCTION__, sizeof(HACmdNode_t)); 
  sprintf(tcmd, "secure-foreign spi %d host %s shared-secret %s", 
	  spistart, ip_to_str(network), shared_secret);
  HA_SAVE_CMD(hc, tcmd);
  hc->key = spistart;
#endif

  return OK;
}





int
haDelete_secureforeign(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  u_int32_t spistart = 0;
  //#ifdef APP_SAVE_CONFIG
  // char tcmd[MAXLINE];
  //#endif
 
#ifndef WITH_DP
  return OK;
#endif


  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("spi_value", kv->key)) 
      spistart = atoi(kv->value);
    else {
      sprintf(buffer, "Unkown command parameter %s", kv->key);
      return ERROR;
    }
  }

  if(mmha_faspidel(spistart)) {
    sprintf(buffer, "Error deleting security association with FA");
    return ERROR;
  }

#ifdef APP_SAVE_CONFIG
  //  sprintf(tcmd, "secure-foreign delete %d", spistart);
  haRemoveCmd(spistart);
#endif

  return OK;
}





int
haShow_homeagent(sList *kvalues, char *buffer)
{
  start_result(buffer,MAXMESG);

  haShow_config(kvalues, buffer);
  haShow_status(kvalues, buffer);
  
  end_result(buffer);
  
  return OK;
}


static int
haShow_config(sList *kvalues, char *buffer)
{
  char temp[MAXLINE], tunmode[MAXLINE];
  char        *  tablerow[STR_LEN_LARGE]; 
  int i = 0;

  for(i = 0; i < 2; i++) {
    if(NULL == (tablerow[i] = (char *)mmGet_memory(__FUNCTION__,STR_LEN_LARGE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return ERROR;
    }
  }

  HA_CREATE_NAMEVALUEHDR(buffer, tablerow, "HA Configuration");
  add_header(buffer, (u_int8_t **)&tablerow);

  if((haconfig.enable_reverse_tunneling) && (haconfig.enable_triangle_tunneling))
    sprintf(tunmode, "Reverse Triangle");
  else if(haconfig.enable_reverse_tunneling)
    sprintf(tunmode, "Reverse");
  else if (haconfig.enable_triangle_tunneling)
    sprintf(tunmode, "Triangle");
  else
    sprintf(tunmode, "Unknown");
  
  sprintf(tablerow[0], "Tunnel Lifetime (secs)");
  sprintf(tablerow[1], "%d", haconfig.ha_default_tunnel_lifetime);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Tunnel Mode");
  sprintf(tablerow[1], "%s", tunmode);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Bindings Limit");
  sprintf(tablerow[1], "%d", haconfig.max_bindings);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Station Masquerade");
  sprintf(tablerow[1], "%d", haconfig.bStationMasquerade);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);
  add_data(buffer, "NewLine", "", FALSE);

  for(i = 0; i < 2; i++) 
    mmFree_memory(__FUNCTION__,tablerow[i]);
  
  return OK;
}


static int
haShow_status(sList *kvalues, char *buffer)
{
  char temp[MAXLINE];
  char        *  tablerow[STR_LEN_LARGE]; 
  extern struct dynamics_ha_status stats;
  extern int bindingcount;
  int i = 0;
  
  for(i = 0; i < 5; i++) {
    if(NULL == (tablerow[i] = (char *)mmGet_memory(__FUNCTION__,STR_LEN_LARGE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return ERROR;
    }
  }
  
  HA_CREATE_NAMEVALUEHDR(buffer, tablerow, "HA Status");
  add_header(buffer, (u_int8_t **)&tablerow);

  sprintf(tablerow[0], "Total Tunnels");
  sprintf(tablerow[1], "%d", stats.tunnel_count);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Total Bindings");
  sprintf(tablerow[1], "%d", bindingcount);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Requests Accepted");
  sprintf(tablerow[1], "%ld", stats.req_accepted);
  add_row(buffer, (u_int8_t **)tablerow);
  sprintf(tablerow[0], "Registration Requests Rejected");
  sprintf(tablerow[1], "%ld", stats.req_rejected);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);

  add_data(buffer, "NewLine", "", FALSE);

  HA_CREATE_REGSTATHDR(buffer, tablerow, 5);
  add_header(buffer, (u_int8_t **)&tablerow);
  sprintf(tablerow[0], "%ld", stats.discarded_unknown_ext);
  sprintf(tablerow[1], "%ld", stats.discarded_malformed_msg);
  sprintf(tablerow[2], "%ld", stats.discarded_vendor_ext);
  sprintf(tablerow[3], "%ld", stats.discarded_not_request);
  sprintf(tablerow[4], "%ld", stats.discarded_id_mismatch);
  add_row(buffer, (u_int8_t **)tablerow);
  end_table(buffer);
  
  for(i = 0; i < 5; i++) 
    mmFree_memory(__FUNCTION__,tablerow[i]);
  return OK;
}


int
haShow_bindings(sList *kvalues, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  struct bindingentry *be = NULL;
  struct bindingtable *bindings = haGet_BindingTable();
  char        *  tablerow[STR_LEN_LARGE]; 
  int            i = 0, n= 1;
  u_int8_t    *  b = NULL;

  if(!bindings || !bindings->hashtable || bindings->hashtablesize > MAX_BINDING_HASH) {
    DEBUG(DEBUG_INFO, "Binding Table Error");
    return ERROR;
  }
  
  for(i = 0; i < 6; i++) {
    if(NULL == (tablerow[i] = (char *)mmGet_memory(__FUNCTION__,STR_LEN_LARGE))) {
      LOG2(LOG_ERR, "Unable to allocate memory");
      return ERROR;
    }
  }

  start_result(buffer,5*MAXMESG);
  ListGet (kvalues, RESET);
  kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT);
  kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT);
  if(kv && 0 == strcasecmp(kv->key, "ip")){
    struct bindingkey key;
    struct ha_tunnel_data *tdata = NULL;
    char   timestr[STR_LEN_MED];
    
    DEBUG(DEBUG_API, "%s: %s : %s", __FUNCTION__, kv->key, kv->value);
    key.mn_addr.s_addr = str_to_ip(kv->value);
    key.ha_addr.s_addr = haconfig.haip.s_addr;
    key.priv_ha = 0;

    be = (struct bindingentry *)binding_fetch((struct bindingtable *)bindings, &key);
    if(!be)
      goto done;
    DEBUG(DEBUG_FLAG, "Got binding entry");

    tdata = (struct ha_tunnel_data *)be->data;
    HA_CREATE_NAMEVALUEHDR(buffer, tablerow, "Binding Entry");
    add_header(buffer, (u_int8_t **)&tablerow);
    sprintf(tablerow[0], "MAC Address");
    sprintf(tablerow[1], "%s", mac_to_str(tdata->mac));
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "IP Address");
    sprintf(tablerow[1], "%s", ip_to_str(be->mn_addr.s_addr));
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Home VLAN");
    sprintf(tablerow[1], "%d", *(u_int16_t *)tdata->arp_if);
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Care-Of Address");
    sprintf(tablerow[1], "%s",  ip_to_str(be->lower_addr.s_addr));
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Current Location");
    b = (u_int8_t *)&tdata->stationloc;
    DEBUG(DEBUG_INFO, "Location = %d.%d.%d",*b, *(b+1), *(u_int16_t *)(b+2));
    sprintf(tablerow[1], "%d.%d.%d", *b, *(b+1), *(u_int16_t *)(b+2));
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Tunnel ID");
    sprintf(tablerow[1], "%s", be->tun_dev);
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Tunnel Type");
    sprintf(tablerow[1], "%s",(tdata->reverse_tunnel)? "Reverse":"Triangle");
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Tunnel Lifetime (secs)");
    sprintf(tablerow[1], "%d", be->timeout);
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Created");
    sprintf(timestr, "%s", ctime(&be->create_time));
    timestr[strlen(timestr)-1] = '\0';
    sprintf(tablerow[1], "%s", timestr);
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Expires");
    sprintf(timestr, "%s", ctime(&be->exp_time));
    timestr[strlen(timestr)-1] = '\0';
    sprintf(tablerow[1], "%s", timestr);
    add_row(buffer, (u_int8_t **)tablerow);
    sprintf(tablerow[0], "Last Refreshed");
    sprintf(timestr, "%s", ctime(&be->mod_time));
    timestr[strlen(timestr)-1] = '\0';
    sprintf(tablerow[1], "%s", timestr);
    add_row(buffer, (u_int8_t **)tablerow);
    if(tdata->last_failure_time) {
       sprintf(tablerow[0], "Last Registration Failure");
       sprintf(timestr, "%s", ctime(&tdata->last_failure_time));
       timestr[strlen(timestr)-1] = '\0';
       sprintf(tablerow[1], "%s", timestr);
       add_row(buffer, (u_int8_t **)tablerow);
    }
    end_table(buffer);
  }
  else {     
    HA_CREATE_BINDHDR(buffer, tablerow, 6);
    add_header(buffer, (u_int8_t **)&tablerow);
    DEBUG(DEBUG_FLAG, "Binding tablesize = %d", bindings->hashtable->hashtablesize);

    for (i = 0; i < bindings->hashtable->hashtablesize; i++) {
      be = (struct bindingentry *)list_get_first(&bindings->hashtable->hasharray[i]);
      while(be) { 
	struct ha_tunnel_data *tdata = (struct ha_tunnel_data *)be->data;
	
	if(strlen(buffer)+300 > 5*MAXMESG) {
	  LOG2(LOG_WARNING, "Buffer pverflow: Current len %d, max capacity %d. retruning ony %d entries",
	       strlen(buffer), 5*MAXMESG, n-1);
	  end_table(buffer);
	  goto done;
	}
	  
	sprintf(tablerow[0], "%d", n++);
	sprintf(tablerow[1], "%s", mac_to_str(tdata->mac));
	sprintf(tablerow[2], "%s", ip_to_str(be->mn_addr.s_addr));
	sprintf(tablerow[3], "%d", *(u_int16_t *)tdata->arp_if);
	sprintf(tablerow[4], "%s",  ip_to_str(be->lower_addr.s_addr));
	b = (u_int8_t *)&tdata->stationloc;
	DEBUG(DEBUG_INFO, "Location = %d.%d.%d",*b, *(b+1), *(u_int16_t *)(b+2));
	sprintf(tablerow[5], "%d.%d.%d", *b, *(b+1), *(u_int16_t *)(b+2));
	add_row(buffer, (u_int8_t **)tablerow);
	be = (struct bindingentry *)list_get_next(&be->hashnode);
      }
    }
    
    end_table(buffer);
  }
  
 done:
  for(i = 0; i < 6; i++) 
    mmFree_memory(__FUNCTION__,tablerow[i]);
  
  end_result(buffer);
  
  return OK;
}




int
haShow_secure(Opaque o, char *buffer)
{
  AMAPIKeyValue *kv = NULL;
  char temp[MAXLINE];
  //ipstr[IP_STRLEN];
  u_int8_t bMobile = FALSE;
  AMAPICommandObject *obj = (AMAPICommandObject *)o;
  sList *kvalues = obj->keyValues;

  DEBUG(DEBUG_FLAG, "In %s, kvalues = %p", __FUNCTION__, kvalues);
  ListGet (kvalues, RESET);
  while ((kv = (AMAPIKeyValue *) ListGet (kvalues, GET_NEXT))
	 != NULL) {
    if(0 == strcasecmp("MOBILE-NODE", kv->name))
      bMobile = TRUE;
  }
  start_result(buffer,MAXMESG);
  
  if(bMobile) {
    struct spi_entry *sa = NULL;
    //    struct authorized_entry *auth = NULL;

    DEBUG(DEBUG_INFO, "Getting HA to mobile security association");
    sprintf(temp, "Authentication Algorithm: %s\tReplay Method: %s",
	    mmAuth_Algo(HA_DEFAULT_AUTH_ALGO), mmReplay(HA_DEFAULT_REPLAY_METHOD));
    add_data(buffer, "", temp, FALSE);
    
    //  add_data(buffer, "", "SPI Information", FALSE);
    for (sa = (struct spi_entry *)list_get_first(&haconfig.spi_list); sa != NULL;
	 sa = (struct spi_entry *)list_get_next(&sa->node)) {
      char tmp[MAXSHAREDSECRETLEN];
      sprintf(temp, "SPI: %d Binding Life: %d Shared Secret: ", sa->spi, sa->max_lifetime);
      if(TRUE == amapiGetEncryptFlag(obj)) {
	memset(tmp, '*', strlen(sa->shared_secret));
	tmp[strlen(sa->shared_secret)] = '\0';
      }
      else 
	strcpy(tmp, sa->shared_secret);
      strcat(temp, tmp);
      add_data(buffer, "", temp, FALSE);
    }
    DEBUG(DEBUG_INFO, "%s", buffer);
#if 0
    add_data(buffer, "", "Authorization Information", FALSE);
    for (auth = (struct authorized_entry *)list_get_first(&haconfig.authorized_list); auth != NULL;
	 auth = (struct authorized_entry *)list_get_next(&auth->node)) {
      sprintf(ipstr, "%s", inet_ntoa(auth->netmask));
      sprintf(temp, "Mobile Node: %s/%s\tSpi: %d", inet_ntoa(auth->network), ipstr, 
	      auth->spi_low);
      if(auth->spi_high != auth->spi_low) {
	char tmp[MAXLINE];
	sprintf(tmp, " - %d", auth->spi_high);
	strcat(temp, tmp);
      }
      add_data(buffer, "", temp, FALSE);
    }
#endif	    
  }
  else  {
    struct fa_spi_entry *sa = NULL;
    
    for (sa = (struct fa_spi_entry *)list_get_first(&haconfig.fa_spi_list); sa != NULL;
	 sa = (struct fa_spi_entry *)list_get_next(&sa->node)) {
      char tmp[MAXSHAREDSECRETLEN];
      sprintf(temp, "SPI: %d\tFA: %s\tShared Secret: ", sa->spi,inet_ntoa(sa->addr));
      if(TRUE == amapiGetEncryptFlag(obj)) {
	memset(tmp, '*', strlen(sa->shared_secret));
	tmp[strlen(sa->shared_secret)] = '\0';
      }
      else 
	strcpy(tmp, sa->shared_secret);
       strcat(temp, tmp);
      add_data(buffer, "", temp, FALSE);
    }
  }
 
  end_result(buffer);
  return OK;
}

#ifdef APP_SAVE_CONFIG
int
haGetcmdstrs(char *pHabuf, int maxsz, int bEncrypt)
{
  int szBuf = 0;
  char temp[MAXMESG*5];
  HACmdNode_t *h = NULL;
  char *estr= NULL;

  if(list_is_empty(&haconfig.cmdbuf))
    return szBuf;

  if(szBuf + strlen("home-agent\n") < maxsz) {
    sprintf(temp, "home-agent\n");
    szBuf += strlen("home-agent\n");
  }
  
  h = (HACmdNode_t *)list_get_first(&haconfig.cmdbuf);
  while(h) {
    char *pStr = NULL, cmd[STR_LEN_LARGE];
    int  szSecret = 0;
    strcat(temp, "  ");
    strcpy(cmd, h->cmd);
    if(bEncrypt && (pStr = strstr(cmd, "shared-secret"))) {
     char *shared_secret = pStr + strlen("shared-secret") + 1;
     estr= arubaCliActionEncrypt(bEncrypt, shared_secret);
     //DEBUG(DEBUG_INFO, "Clear Text = %s, Encrypted = %s", shared_secret, estr);
     *shared_secret = '\0';
     strcat(temp, cmd);
     strcat(temp, estr);
     szSecret = strlen(estr);
   }
   else
     strcat(temp, cmd);
    strcat(temp, "\n");
    DEBUG(DEBUG_FLAG, "Command = %s", temp);
    szBuf += strlen(cmd)+szSecret+3;
    h = (HACmdNode_t *)list_get_next(&h->node);
  }

#if 0
  for(inti = 0; inti < haconfig.indCurCmd && szBuf+strlen(haconfig.cmdbuf[inti])+3 < maxsz; 
      inti++) {
    char *pStr = NULL, cmd[STR_LEN_LARGE];
    int  szSecret = 0;
    strcat(temp, "  ");
    strcpy(cmd, haconfig.cmdbuf[inti]);
    if(bEncrypt && (pStr = strstr(cmd, "shared-secret"))) {
      char estr[STR_LEN_LARGE];
      char *shared_secret = pStr + strlen("shared-secret") + 1; 
      proxyEncryptSecret(shared_secret, estr);
      DEBUG(DEBUG_INFO, "Clear Text = %s, Encrypted = %s", shared_secret, estr);
      *shared_secret = '\0';
      strcat(temp, cmd);
      strcat(temp, estr);
      szSecret = strlen(estr);
    }
    else 
      strcat(temp, haconfig.cmdbuf[inti]);
    strcat(temp, "\n");
    DEBUG(DEBUG_FLAG, "Command = %s", temp);
    szBuf += strlen(cmd)+szSecret+3; 
  }
#endif

  strcat(temp, "!\n");
  szBuf += strlen("!\n");

  sprintf(pHabuf, "%s", temp);
  DEBUG(DEBUG_CONFIG, "HA Local Config = %s", pHabuf);
  return szBuf;
}


int
haGetgblcmdstrs(char *pHabuf, int bEncrypt)
{
  int szBuf = 0;
  char temp[STR_LEN_LARGE];
  struct spi_entry *spi = NULL;

  sprintf(pHabuf, "mobagent\n");
  szBuf += strlen("mobagent\n");
  sprintf(temp, "  home-agent parameters %d bindings %d\n", haconfig.ha_default_tunnel_lifetime,
	  haconfig.max_bindings);
  szBuf += strlen(temp);
  strcat(pHabuf, temp);
  spi = (struct spi_entry *)list_get_first(&haconfig.spi_list);
  if(spi) {
    sprintf(temp, "  secure-mobile spi %d ", spi->spi);
    strcat(temp, arubaCliActionEncrypt(bEncrypt, spi->shared_secret));
    strcat(temp, "\n");	
    szBuf += strlen(temp);
    strcat(pHabuf, temp);
  }

  return szBuf;
}
#endif
