/* $Id: dyn_ip.h,v 1.46 2001/09/26 18:25:24 jm Exp $
 * Tunnel and route interface to kernel
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */


#ifndef DYN_IP_H
#define DYN_IP_H

#include <netinet/in.h>

#ifndef IFNAMSIZ
#define	IFNAMSIZ 16
#endif

struct idxmap
{
	struct idxmap *next;
	int index;
	char name[IFNAMSIZ];
};


struct current_route_info {
	int known; /* whether the current situation is known */
	struct in_addr via; /* 0.0.0.0 = no via address */
	struct in_addr prefsrc; /* 0.0.0.0 = no prefsrc address */
	int ifindex; /* default route interface index */
	char ifname[IFNAMSIZ]; /* default route interface name */

	struct in_addr net; /* network address: net/prefix_len */
	int prefix_len;
	int ifindex_net; /* net route interface index
			  * -1 = remove net route entry */
	char ifname_net[IFNAMSIZ]; /* net route interface name */
};


int dyn_ip_tunnel_add(const char *dev, struct in_addr remote, 
		      struct in_addr local);
int dyn_ip_tunnel_add_gre(const char *dev, struct in_addr remote,
			  struct in_addr local, int key);
int dyn_ip_tunnel_del(const char *dev);
int dyn_ip_link_set_dev_up(const char *dev);
int dyn_ip_link_set_dev_down(const char *dev);
int dyn_ip_addr_add(const char *dev, struct in_addr addr);
int dyn_ip_addr_del(const char *dev, struct in_addr addr);
int dyn_ip_route_add_table(const char *dev, int table, 
			   const struct in_addr *dst);
int dyn_ip_route_del_table(const char *dev, int table, 
			   const struct in_addr *dst);
int dyn_ip_route_add_net(const char *dev, const struct in_addr *dst, int plen);
int dyn_ip_route_replace_net(const char *dev, const struct in_addr *dst,
			     int plen);
int dyn_ip_route_del_net(const char *dev, const struct in_addr *dst, int plen);
int dyn_ip_route_add_blackhole(int table);
int dyn_ip_route_del_blackhole(int table);
int dyn_ip_rule_add_table(struct in_addr *src, struct in_addr *dst, int table,
			  const char *device);
int dyn_ip_rule_del_table(struct in_addr *src, struct in_addr *dst, int table,
			  const char *device);
int ip_rule_get_table(struct in_addr *src, struct in_addr *dst,const char *device);
int dyn_ip_route_get(struct in_addr dst, char *dev, int max_len);
int dyn_ip_route_to_host(struct in_addr dst);
int dyn_ip_get_ifaddr(const char *dev, struct in_addr *addr);
int dyn_ip_get_bcaddr(const char *dev, struct in_addr *addr);
int dyn_ip_get_local_addr(const struct in_addr *remote, struct in_addr *local);
int dyn_ip_route_replace(struct in_addr dst_addr, const char *dev, const char *mnmac);
int dyn_ip_route_replace_table(struct in_addr dst_addr, const char *dev,
			       int table);
int dyn_ip_route_add(struct in_addr dst_addr, const char *dev);
int dyn_ip_route_del(struct in_addr dst_addr, const char *dev);
int dyn_ip_get_saved_ifindex(void);
int dyn_ip_route_save_default(struct in_addr *forced_gateway);
int dyn_ip_route_restore_default(const char *force_dev);
int dyn_ip_route_replace_default(const char *dev, const struct in_addr *via,
				 const struct in_addr *prefsrc);
int dyn_ip_check_adv_routing(void);
int dyn_ip_check_ipip(void);
int dyn_ip_check_gre(void);
int dyn_ip_check_netlink(void);
struct idxmap* dyn_ip_get_interface_map(void);
void dyn_ip_free_interface_map(struct idxmap *map);
int dyn_ip_get_ifindex(const char *dev);
int dyn_ip_get_ifname(int ifindex, char *dev);
int dyn_ip_route_cache_flush(void);
int dyn_ip_get_ifflags(const char *dev);
int dyn_ip_route_exists(const struct in_addr *addr, int prefix_len);
int dyn_ip_monitor_open(void);
int dyn_ip_monitor_get(int fd, const struct current_route_info *cur);
int dyn_ip_iface_address(int ifindex, const struct in_addr *addr);
int dyn_ip_get_hwaddr(const char *dev, char *hwaddr, int max_len);
int dyn_ip_mask_to_prefix_len(unsigned int mask);
unsigned int dyn_ip_prefix_len_to_mask(int plen);
int dyn_ip_route_delete_net_routes(const struct in_addr *addr, int prefix_len);

#ifdef DYN_TARGET_WINDOWS
int dyn_ip_set_ttl(int ttl);
int dyn_ip_route_set_net(const char *dev, const struct in_addr *dst, int plen,
			 const struct in_addr *via, int metric);
#endif /* DYN_TARGET_WINDOWS */

#endif
