/*
 * Layer Two Tunnelling Protocol Daemon
 * Copyright (C) 1998 Adtran, Inc.
 *
 * Mark Spencer
 *
 * This software is distributed under the terms
 * of the GPL, which you should have received
 * along with this source.
 *
 * Main Daemon source.
 *
 * CHANGES:
 * -------
 *   09/13/02 - csellin - Fix synchronization issue between pppd and l2tpd
 */

#include <stdlib.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#if (__GLIBC__ < 2)
# if defined(FREEBSD)
#  include <sys/signal.h>
# elif defined(LINUX)
#  include <bsd/signal.h>
# elif defined(SOLARIS)
#  include <signal.h>
# endif
#else
# include <signal.h>
#endif
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>

#include "l2tp.h"
#include "l2tp_cli.h"

#include <vpn/ike_ipc.h>

#include <syslog.h>
#include <syslog/syslogutil.h>
#include <syslog/arubalog.h>
#include <vpn/shared_cli.h>
#include <vpn/l2tp_ipc.h>
#include <msgHandler/msgHandler.h>
#include <papi/MHApi.h>
#include "misc.h"
#include <include/mgmtcommon.h>
#include <assert.h>

#define IPTOSTR(a) (inet_ntoa(*(struct in_addr *)(&(a))))

/* sigh...should use BSD list...*/
struct tunnel_list tunnels;
STAILQ_HEAD(, tunnel) free_tunnels;
struct tunnel      tuns[8192];
#define MAX_TUNS (sizeof(tuns)/sizeof(tuns[0]))
struct utsname uts;
int ppd = 1;                    /* Packet processing delay */
int control_fd;                 /* descriptor of control area */
char *args;

char *dial_no_tmp;              /* jz: Dialnumber for Outgoing Call */
int switch_io = 0;              /* jz: Switch for Incoming or Outgoing Call */
int ipc_fd;
struct sockaddr_in ipc_sockaddr;

static int  l2tpd_ipc_init();
static int  l2tpd_ipc_session_up(struct call *c);
#if 0
static void l2tpd_ipc_close();
#endif

#define MAX_FDS 16000

static int
init_tunnels()
{
    int i;
    memset(tuns, 0, sizeof(tuns));
    /* insert into free list */
    STAILQ_INIT(&free_tunnels);
    /* we skip index 0...*/
    for (i=1; i<(sizeof(tuns)/sizeof(tuns[0])); i++) {
        tuns[i].ourtid = i;
        STAILQ_INSERT_TAIL(&free_tunnels, &tuns[i], stq);
    }
    return 0;
}

static struct tunnel *
alloc_tunnel()
{
    struct tunnel *tmp = STAILQ_FIRST(&free_tunnels);
    if (tmp) {
        STAILQ_REMOVE_HEAD(&free_tunnels, stq);
    }
    return tmp;
}

static void
free_tunnel(struct tunnel *tun)
{
    if (tun) {
        assert(tun->ourtid == (tun - &tuns[0]));
        STAILQ_INSERT_TAIL(&free_tunnels, tun, stq);
    }
}

static int
mydup(int fd)
{
    static int g_fdstart = MAX_FDS-1;
    int ret;
    if (fd < 0)
        return -1;
        
    ret = dup2(fd, g_fdstart);
    //arubaLog(LOG_DEBUG, "dup2 from %d->%d, ret=%d\n", fd, g_fdstart, ret);
    if (ret > 0) {
        safe_close(&fd);
        g_fdstart--;
        if (g_fdstart <= 1024) {
            g_fdstart = MAX_FDS-1;
        }
        return ret;
    } else
        return fd;
}


void init_tunnel_list (struct tunnel_list *t)
{
    t->head = NULL;
    t->count = 0;
    t->calls = 0;
}

void show_status (int fd)
{
#if 0
    struct schedule_entry *se;
    struct tunnel *t;
    struct call *c;
    struct lns *tlns;
    struct lac *tlac;
    struct host *h;
    int s = 0;
    int fd2 = dup (fd);
    FILE *f = fdopen (fd2, "a");
    if (!f)
    {
        log (LOG_WARN, "show_status: fdopen() failed on fd %d\n", fd);
        return;
    }
    fprintf (f, "====== l2tpd statistics ========\n");
    fprintf (f, " Scheduler entries:\n");
    se = events;
    while (se)
    {
        s++;
        t = (struct tunnel *) se->data;
        tlac = (struct lac *) se->data;
        c = (struct call *) se->data;
        if (se->func == &hello)
        {
            fprintf (f, "%d: HELLO to %d\n", s, t->tid);
        }
        else if (se->func == &magic_lac_dial)
        {
            fprintf (f, "%d: Magic dial on %s\n", s, tlac->entname);
        }
        else if (se->func == &send_zlb)
        {
            fprintf (f, "%d: Send payload ZLB on call %d:%d\n", s,
                     c->container->tid, c->cid);
        }
        else if (se->func == &dethrottle)
        {
            fprintf (f, "%d: Dethrottle call %d:%d\n", s, c->container->tid,
                     c->cid);
        }
        else
            fprintf (f, "%d: Unknown event\n", s);
        se = se->next;
    };
    fprintf (f, "Total Events scheduled: %d\n", s);
    fprintf (f, "Number of tunnels open: %d\n", tunnels.count);
    fprintf (f, "Highest file descriptor: %d\n", fd2);
    t = tunnels.head;
    while (t)
    {
        fprintf (f, "Tunnel %s, ID = %d (local), %d (remote) to %s:%d\n"
                 "   control_seq_num = %d, control_rec_seq_num = %d,\n"
                 "   cLr = %d\n",
                 (t->lac ? t->lac->entname : (t->lns ? t->lns->entname : "")),
                 t->ourtid, t->tid, IPADDY (t->peer.sin_addr),
                 ntohs (t->peer.sin_port), t->control_seq_num,
                 t->control_rec_seq_num, t->cLr);
        c = t->call_head;
        while (c)
        {
            fprintf (f,
                     "Call %s, ID = %d (local), %d (remote), serno = %d,\n"
                     "      data_seq_num = %d, data_rec_seq_num = %d,\n"
                     "      pLr = %d, tx = %d bytes (%d), rx= %d bytes (%d)\n",
                     (c->lac ? c->lac->
                      entname : (c->lns ? c->lns->entname : "")), c->ourcid,
                     c->cid, c->serno, c->data_seq_num, c->data_rec_seq_num,
                     c->pLr, c->tx_bytes, c->tx_pkts, c->rx_bytes, c->rx_pkts);
            c = c->next;
        }
        t = t->next;
    }
    fprintf (f, "==========Config File===========\n");
    tlns = lnslist;
    while (tlns)
    {
        fprintf (f, "LNS entry %s\n",
                 tlns->entname[0] ? tlns->entname : "(unnamed)");
        tlns = tlns->next;
    };
    tlac = laclist;
    while (tlac)
    {
        fprintf (f, "LAC entry %s, LNS is/are:",
                 tlac->entname[0] ? tlac->entname : "(unnamed)");
        h = tlac->lns;
        if (h)
        {
            while (h)
            {
                fprintf (f, " %s", h->hostname);
                h = h->next;
            }
        }
        else
            fprintf (f, " [none]");
        fprintf (f, "\n");
        tlac = tlac->next;
    };
    fprintf (f, "================================\n");
    fclose (f);
    safe_close (&fd2);
#endif
}

void null_handler(int sig)
{
       /* FIXME 
        * A sighup is received when a call is terminated, unknown origine .. 
        * I catch it and ll looks good, but .. 
	*/
}

void status_handler (int sig)
{
    show_status (1);
}

void child_handler (int signal)
{
    /*
     * Oops, somebody we launched was killed.
     * It's time to reap them and close that call.
     * But first, we have to find out what PID died.
     * unfortunately, pppd will 
     */
    struct tunnel *t;
    pid_t pid;
    int status;
    t = tunnels.head;
    pid = waitpid (-1, &status, WNOHANG);
    if (pid < 1)
    {
        /*
         * Oh well, nobody there.  Maybe we reaped it
         * somewhere else already
         */
        return;
    }
#if 0 // NOT CLEANING UP SINCE WE ARE KILLING PPPD IMMEDIATELY !
    while (t)
    {
      struct call *c;

        c = t->call_head;
        while (c)
        {
            if (c->pppd == pid)
            {
                log (LOG_DEBUG, "%s : pppd died for call %d\n", __FUNCTION__,
                     c->cid);
                c->needclose = -1;
                /* 
                 * OK...pppd died, we can go ahead and close the pty for
                 * it
                 */

                close (c->fd);
                return;
            }
            c = c->next;
        }
        t = t->next;
    }
#endif
}

void death_handler (int signal)
{
#if 0
    /*
       * If we get here, somebody terminated us with a kill or a control-c.
       * we call call_close on each tunnel twice to get a StopCCN out
       * for each one (we can't pause to make sure it's received.
       * Then we close the connections
     */
    struct tunnel *st, *st2;
    int sec;
    log (LOG_CRIT, "%s: Fatal signal %d received\n", __FUNCTION__, signal);
    st = tunnels.head;
    while (st)
    {
        st2 = st->next;
        strcpy (st->self->errormsg, "Server closing");
        sec = st->self->closing;
        if (st->lac)
            st->lac->redial = 0;
        call_close (st->self);
        if (!sec)
        {
            st->self->closing = -1;
            call_close (st->self);
        }
        st = st2;
    }
#endif

    /* erase pid file */
	unlink (gconfig.pidfile);

    exit (1);
}

// #include <stropts.h> /* for I_PUSH */
/* can't find system include that contains these functions */
extern int grantpt (int fd);
extern int unlockpt (int fd);

static int
my_openpty(int *master, int *slave, char *name, void *unused1, void *unused2)
{
     int fd_master, fd_slave;
     char slave_name[32];
     int ptyno;
     slave_name[0] = '\0';
     
     fd_master = open("/dev/ptmx", O_RDWR);
     if (fd_master < 0) {
         log(LOG_ERR, "L2TPD: Failed to allocate pty");
         return -1;
     }

     if (ioctl(fd_master, TIOCGPTN, &ptyno) == 0) {
         sprintf(slave_name, "/dev/pts/%d", ptyno);
     } else {
         log(LOG_ERR, "%s", sys_errlist[errno]);
     }
     
     log(LOG_DEBUG, "L2TPD: pty name is %s", slave_name);
     grantpt(fd_master);
     unlockpt(fd_master);
     if (name) {
         strcpy(name, slave_name);
     }
     fd_slave = open(slave_name, O_RDWR);
     if (fd_slave < 0) {
         log(LOG_ERR, "L2TPD: Failed to allocate slave pty %s, master pty is %d", slave_name, fd_master);
         return -1;
     }

//     ioctl(fd_slave, I_PUSH, "ptem");
//     ioctl(fd_slave, I_PUSH, "ldterm");
     if (master)
         *master = fd_master;
     if (slave)
         *slave = fd_slave;
     log(LOG_DEBUG, "L2TPD: Allocated pty/tty pair (%d,%d)",
	 fd_master, fd_slave);
     return 0;
}

#define NUM_STROPT 128
static char g_stropt[NUM_STROPT][256];

#define SEND_TTYNAME 1

int start_pppd (struct call *c, struct ppp_opts *opts)
{
    //char a, b;
    char tty[80];
    struct ppp_opts *p;
    char *stropt[NUM_STROPT];
    
#ifdef USE_KERNEL
    struct l2tp_call_opts co;
#endif
#if SEND_TTYNAME
    int pos = 3; /* start from pos 3, 1 & 2 are reserved for l2tpdevname */
#else
    int pos = 1;
#endif /* SEND_TTYNAME */
    int fd2;
    struct termios ptyconf;
    char *str;
    int i;


    for (i=0; i<NUM_STROPT; i++)
        stropt[i] = g_stropt[i];
    
    p = opts;
    strlcpy(g_stropt[0], PPPD, sizeof(g_stropt[0]));
    while (p && pos < NUM_STROPT)
    {
        strlcpy(g_stropt[pos], p->option, sizeof(g_stropt[pos]));
        pos++;
        p = p->next;
    }
    if (c->pppd > 0)
    {
        log (LOG_WARN, "%s: PPP already started on call!\n", __FUNCTION__);
        return -EINVAL;
    }
    if (c->fd > -1)
    {
        log (LOG_WARN, "%s: file descriptor already assigned!\n",
             __FUNCTION__);
        return -EINVAL;
    }
#ifdef USE_KERNEL
    if (kernel_support)
    {
        abort();
    }
    else
    {
#endif
#if 0
        if ((c->fd = getPtyMaster (&a, &b)) < 0)
        {
            log (LOG_WARN, "%s: unable to allocate pty, abandoning!\n",
                 __FUNCTION__);
            return -EINVAL;
        }
#endif
        if (my_openpty(&c->fd, &fd2, tty, NULL, NULL) < 0) {
            log (LOG_WARN, "%s: unable to allocate pty, abandoning!\n",
                 __FUNCTION__);
            return -EINVAL;
        }
        
        /* set fd opened above to not echo so we don't see read our own packets
           back of the file descriptor that we just wrote them to */
        tcgetattr (c->fd, &ptyconf);
        //*(c->oldptyconf) = ptyconf;
        ptyconf.c_cflag &= ~(ICANON | ECHO);
        tcsetattr (c->fd, TCSANOW, &ptyconf);

        /* set to async */
        {
            long arg = 0;
            fcntl(c->fd, F_GETFL);
            arg |= O_NONBLOCK;
            fcntl(c->fd, F_SETFL, arg);
        }
        //snprintf (tty, sizeof (tty), "/dev/tty%c%c", a, b);
        //fd2 = open (tty, O_RDWR);

#ifdef USE_KERNEL
    }
#endif

    /* Allocate space for the option name "orig_ip" and value (16B)       */
    strlcpy(g_stropt[pos], "orig_ip", sizeof(g_stropt[pos]));
    pos++;

    strlcpy(g_stropt[pos] , IPADDY(c->container->peer.sin_addr), sizeof(g_stropt[pos]));
    pos++;

    /* Passing the external IP address to the ppp daemon for log purposes */
    str = g_stropt[0];
#if SEND_TTYNAME
    strlcpy(g_stropt[1], "l2tpdevname", sizeof(g_stropt[1]));
    strlcpy(g_stropt[2], ttyname(fd2), sizeof(g_stropt[2]));
#endif /* SEND_TTYNAME */
    
    /* Allocate space for the option name "l2tptunid" and value (16B)       */
    strlcpy(g_stropt[pos], "l2tptunid", sizeof(g_stropt[pos]));
    pos++;

    sprintf (g_stropt[pos], "%d", c->container->ourtid);
    pos++;

    /* Allocate space for the option name "l2tpcallid" and value (16B)       */
    strlcpy(g_stropt[pos], "l2tpcallid", sizeof(g_stropt[pos]));
    pos++;

    sprintf (g_stropt[pos], "%d", c->ourcid);
    pos++;

    /* Opening 2 ppp devices so that we can do the cleanup in L2TP and not PPP */
    c->ppp_dev_fd = open("/dev/ppp", O_RDWR);
    if (c->ppp_dev_fd < 0)
    {
      log(LOG_CRIT, "could not open /dev/ppp");
      exit(1); //TO DO MUST CHANGE EXIT !!
    }
    
    /* Allocate space for the option name "orig_ip" and value (16B)       */
    strlcpy(g_stropt[pos], "ppp_dev_fd", sizeof(g_stropt[pos]));
    pos++;

    sprintf (g_stropt[pos], "%d", c->ppp_dev_fd);
    pos++;

    c->ppp_dev_fd_2 = open("/dev/ppp", O_RDWR);
    if (c->ppp_dev_fd_2 < 0)
    {
      log(LOG_CRIT, "could not open /dev/ppp");
      exit(1); //TO DO MUST CHANGE EXIT !!
    }

    /* Allocate space for the option name "orig_ip" and value (16B)       */
    strlcpy(g_stropt[pos], "ppp_dev_fd_2", sizeof(g_stropt[pos]));
    pos++;

    sprintf (g_stropt[pos], "%d", c->ppp_dev_fd_2);
    pos++;


    // debug level
    /* Allocate space for the option name "orig_ip" and value (16B)       */
    strlcpy(g_stropt[pos], "dbglevel", sizeof(g_stropt[pos]));
    pos++;

    {
        // sigh...pass in the ppp flags for debugging...
        int l, dbglevel = 0;
        for (l=0; l<lgClientInfo.numDebugEntries; l++) {
            if (lgClientInfo.dEnable[l].dId == DEBUG_PPP_PACKETS &&
                lgClientInfo.dEnable[l].enable) {
                dbglevel = 1;
            }
        }
        sprintf (g_stropt[pos], "%d", dbglevel);
        pos++;
    }
    
    
    stropt[pos] = NULL;

    c->pppd = fork ();
    if (c->pppd < 0)
    {
        arubaLog (LOG_CRIT,
                  "%s: unable to fork(), abandoning!\n", __FUNCTION__);
        return -EINVAL;
    }
    else if (!c->pppd)
    {
        struct call *sc;
        struct tunnel *st;

#if 0
        close (0);
        close (1);
        close (2);
#endif
#ifdef USE_KERNEL
        if (!kernel_support && (fd2 < 0))
#else
        if (fd2 < 0)
#endif
        {
            log (LOG_WARN, "%s: Unable to open %s errno %d to launch pppd!\n",
                 __FUNCTION__, tty, errno);
            exit (1);
        }
        dup2 (fd2, 0);
        dup2 (fd2, 1);
        /* close all the calls pty fds */
        st = tunnels.head;
        while (st)
        {
            sc = st->call_head;
            while (sc)
            {
                if (sc->fd > 0)
                    safe_close (&sc->fd);
                sc->fd = -1;
                sc = sc->next;
            }
            st = st->next;
        }

        /* close the UDP socket fd */
        close (server_socket);

        /* close the control pipe fd */
        close (control_fd);
        close (ipc_fd);
        close (papi_fd);

        execv (PPPD, stropt);
        log (LOG_WARN, "%s: Exec of %s failed!\n", __FUNCTION__, PPPD);
        exit (1);
    };

    // using this fd for synchronization, make sure the close cleans it up !!
    // close (fd2);
    c->pppdsync = fd2;
    // toschedule for the master & slave pts to be closed in 120 seconds...
#if 0
    if (c->toclose) {
        deschedule(c->toclose);
    }
    {
        struct timeval tv;
        tv.tv_sec = 30;
        tv.tv_usec = 0;
        //syslog(LOG_DEBUG, "Scheduling close_ptys...\n");
        c->toclose = schedule(tv, close_ptys, c);
    }
#endif
    gettimeofday (&c->toclose, NULL);
    c->toclose.tv_sec += 30;

    return 0;
}

void destroy_tunnel (struct tunnel *t)
{
    /*
     * Immediately destroy a tunnel (and all its calls)
     * and free its resources.  This may be called
     * by the tunnel itself,so it needs to be
     * "suicide safe"
     */

    struct call *c, *me, *c2;
    struct tunnel *p;
    struct timeval tv;
    if (!t)
        return;

    /*
     * Save ourselves until the very
     * end, since we might be calling this ourselves.
     * We must divorce ourself from the tunnel
     * structure, however, to avoid recursion
     * because of the logic of the destroy_call
     */
    me = t->self;

    /*
     * Destroy all the member calls
     */
    
    c = t->call_head;
    while (c) {
        c2 = c->next;
        destroy_call (c);
        c = c2;
    };
    /*
     * Remove ourselves from the list of tunnels
     */

    if (tunnels.head == t)
    {
        tunnels.head = t->next;
        tunnels.count--;
    }
    else
    {
        p = tunnels.head;
        if (p)
        {
            while (p->next && (p->next != t))
                p = p->next;
            if (p->next)
            {
                p->next = t->next;
                tunnels.count--;
            }
            else
            {
                log (LOG_WARN,
                     "%s: unable to locate tunnel in tunnel list\n",
                     __FUNCTION__);
            }
        }
        else
        {
            log (LOG_WARN, "%s: tunnel list is empty!\n", __FUNCTION__);
        }
    }
    if (t->lac)
    {
#if 0
        t->lac->t = NULL;
        if (t->lac->redial && (t->lac->rtimeout > 0) && !t->lac->rsched &&
            t->lac->active)
        {
            log (LOG_LOG, "%s: Will redial in %d seconds\n", __FUNCTION__,
                 t->lac->rtimeout);
            tv.tv_sec = t->lac->rtimeout;
            tv.tv_usec = 0;
            t->lac->rsched = schedule (tv, magic_lac_dial, t->lac);
        }
#endif
    }
    /* XXX L2TP/IPSec: remove relevant SAs here?  NTB 20011010
     * XXX But what if another tunnel is using same SA?
     */
    if (t->lns)
        t->lns->t = NULL;

#if 0
    // ARUBA: USED TO BE A LEAK !!!
    if (t->chal_them.vector)
    {
      L2TPFREE (__FUNCTION__, t->chal_them.vector);
      t->chal_them.vector = NULL;
    }
#endif

#if 0
    if (me->oldptyconf)
    {
      L2TPFREE (__FUNCTION__, me->oldptyconf);      
      me->oldptyconf = NULL;
    }
#endif

    t->indp = 0;

    free_call(me);
    
    if (t->isdestroyed) {
        arubaLog(LOG_CRIT,
                 "Software error, multiple memory locations overwritten, memory area is 0x%x\n", t->isdestroyed);
    } else {
        t->isdestroyed = 0x12345678;
        free_tunnel(t);
    }
    g_tunnfreed    = 1;
        
}

struct tunnel *
get_tunnel(int tunnel, unsigned int addr, int port)
{
    struct tunnel *tmp;
    if (tunnel <= 0 || (tunnel >= MAX_TUNS))
        return NULL;
    tmp = &tuns[tunnel];
    if (tmp->isdestroyed)
        return NULL;
    else
        return tmp;
}

struct call *
get_call_from_tunnel (int tunnel, unsigned int addr, int port)
{
    struct tunnel *tmp = get_tunnel(tunnel, addr, port);
    if (tmp) {
        return tmp->self;
    } else
        return NULL;
}


struct tunnel *l2tp_call (char *host, int port, struct lac *lac,
                          struct lns *lns)
{
    /*
     * Establish a tunnel from us to host
     * on port port
     */
    struct call *tmp = NULL;
    struct hostent *hp;
    unsigned int addr;
    port = htons (port);
    hp = gethostbyname (host);
    if (!hp)
    {
        return NULL;
    }
    bcopy (hp->h_addr, &addr, hp->h_length);
    /* Force creation of a new tunnel
       and set it's tid to 0 to cause
       negotiation to occur */
    /* XXX L2TP/IPSec: Set up SA to addr:port here?  NTB 20011010
     */
    tmp = get_call (0, 0, addr, port);
    if (!tmp)
    {
        return NULL;
    }
    tmp->container->tid = 0;
    tmp->container->lac = lac;
    tmp->container->lns = lns;
    tmp->lac = lac;
    tmp->lns = lns;
    if (lac)
        lac->t = tmp->container;
    if (lns)
        lns->t = tmp->container;
    /*
     * Since our state is 0, we will establish a tunnel now
     */
    arubaDebugLog (DEBUG_VPDN_PACKETS, LOG_DEBUG,
                   "%s:Connecting to host %s, port %d\n",
                   __FUNCTION__, host, ntohs (port));
    control_finish (tmp->container, tmp);
    return tmp->container;
}

void magic_lac_tunnel (void *data)
{
    struct lac *lac;
    lac = (struct lac *) data;
    if (!lac)
    {
        log (LOG_WARN, "%s: magic_lac_tunnel: called on NULL lac!\n",
             __FUNCTION__);
        return;
    }
    if (lac->lns)
    {
        /* FIXME: I should try different LNS's if I get failures */
        l2tp_call (lac->lns->hostname, lac->lns->port, lac, NULL);
        return;
    }
    else if (deflac && deflac->lns)
    {
        l2tp_call (deflac->lns->hostname, deflac->lns->port, lac, NULL);
        return;
    }
    else
    {
        log (LOG_WARN, "%s: Unable to find hostname to dial for '%s'\n",
             __FUNCTION__, lac->entname);
        return;
    }
}

struct call *lac_call (int tid, struct lac *lac, struct lns *lns)
{
#if 0
    struct tunnel *t = tunnels.head;
    struct call *tmp;
    while (t)
    {
        if (t->ourtid == tid)
        {
            tmp = new_call (t);
            if (!tmp)
            {
                log (LOG_WARN, "%s: unable to create new call\n",
                     __FUNCTION__);
                return NULL;
            }
            tmp->next = t->call_head;
            t->call_head = tmp;
            t->count++;
            tmp->cid = 0;
            tmp->lac = lac;
            tmp->lns = lns;
            if (lac)
                lac->c = tmp;
            log (LOG_LOG, "%s: Calling on tunnel %d\n", __FUNCTION__, tid);
            strcpy (tmp->dial_no, dial_no_tmp); /*  jz: copy dialnumber to tmp->dial_no  */
            control_finish (t, tmp);
            return tmp;
        }
        t = t->next;
    };
    log (LOG_DEBUG, "%s: No such tunnel %d to generate call.\n", __FUNCTION__,
         tid);
#endif
    return NULL;
}

void magic_lac_dial (void *data)
{
    struct lac *lac;
    lac = (struct lac *) data;
    if (!lac->active)
    {
        log (LOG_DEBUG, "%s: LAC %s not active", __FUNCTION__, lac->entname);
        return;
    }
    lac->rsched = NULL;
    lac->rtries++;
    if (lac->rmax && (lac->rtries > lac->rmax))
    {
        log (LOG_LOG, "%s: maximum retries exceeded.\n", __FUNCTION__);
        return;
    }
    if (!lac)
    {
        log (LOG_WARN, "%s : called on NULL lac!\n", __FUNCTION__);
        return;
    }
    if (!lac->t)
    {
#ifdef DEGUG_MAGIC
        log (LOG_DEBUG, "%s : tunnel not up!  Connecting!\n", __FUNCTION__);
#endif
        magic_lac_tunnel (lac);
        return;
    }
    lac_call (lac->t->ourtid, lac, NULL);
}

void lac_hangup (int cid)
{
    struct tunnel *t = tunnels.head;
    struct call *tmp;
    while (t)
    {
        tmp = t->call_head;
        while (tmp)
        {
            if (tmp->ourcid == cid)
            {
                arubaDebugLog (DEBUG_VPDN_PACKETS, LOG_DEBUG,
                     "Hanging up call #%d, Local call id: %d, Remote call id: %d\n",
                     tmp->serno, tmp->ourcid, tmp->cid);
                strcpy (tmp->errormsg, "Goodbye!");
/*				tmp->needclose = -1; */
                kill (tmp->pppd, SIGTERM);
                return;
            }
            tmp = tmp->next;
        }
        t = t->next;
    };
    log (LOG_DEBUG, "%s : No such call %d to hang up.\n", __FUNCTION__, cid);
    return;
}

void lac_disconnect (int tid)
{
    struct tunnel *t = tunnels.head;
    while (t)
    {
        if (t->ourtid == tid)
        {
            arubaDebugLog (DEBUG_VPDN_PACKETS, LOG_DEBUG,
                 "Disconnecting from %s, Local: %d, Remote: %d\n",
                 IPADDY (t->peer.sin_addr), t->ourtid, t->tid);
            t->self->needclose = -1;
            strcpy (t->self->errormsg, "Goodbye!");
            call_close (t->self);
            return;
        }
        t = t->next;
    };
    log (LOG_DEBUG, "%s: No such tunnel %d to hang up.\n", __FUNCTION__, tid);
    return;
}


struct tunnel *new_tunnel ()
{
    struct tunnel *tmp = alloc_tunnel();
    if (!tmp)
        return NULL;

    tmp->control_seq_num = 0;
    tmp->control_rec_seq_num = 0;
    tmp->cLr = 0;
    tmp->call_head = NULL;
    tmp->next = NULL;
    tmp->debug = -1;
    tmp->tid = -1;
    tmp->indp = 0;
    /* hard code to make it easier to search */
    assert(tmp->ourtid == (tmp - &tuns[0]));

    tmp->nego = 0;
    tmp->count = 0;
    tmp->state = 0;             /* Nothing */
    tmp->peer.sin_family = AF_INET;
    tmp->peer.sin_port = 0;
    bzero (&(tmp->peer.sin_addr), sizeof (tmp->peer.sin_addr));
    tmp->sanity = -1;
    tmp->qtid = -1;
    tmp->ourfc = ASYNC_FRAMING | SYNC_FRAMING;
    tmp->ourbc = 0;
    tmp->ourtb = (((_u64) rand ()) << 32) | ((_u64) rand ());
    tmp->fc = -1;               /* These really need to be specified by the peer */
    tmp->bc = -1;               /* And we want to know if they forgot */
    tmp->hostname[0] = 0;
    tmp->vendor[0] = 0;
    //tmp->secret[0] = 0;
    if (!(tmp->self = new_call (tmp)))
    {
        free_tunnel(tmp);
        return NULL;
    };
    tmp->ourrws = DEFAULT_RWS_SIZE;
    tmp->self->ourfbit = FBIT;
    tmp->lac = NULL;
    tmp->lns = NULL;
#if 0
    tmp->chal_us.state = 0;
    tmp->chal_us.secret[0] = 0;
    memset (tmp->chal_us.reply, 0, MD_SIG_SIZE);
    tmp->chal_them.state = 0;
    tmp->chal_them.secret[0] = 0;
    memset (tmp->chal_them.reply, 0, MD_SIG_SIZE);
    tmp->chal_them.vector = (unsigned char *) L2TPMALLOC (__FUNCTION__, VECTOR_SIZE);
    tmp->chal_us.vector = NULL;
#endif
    tmp->hbit = 0;
    // for debugging multiple frees...
    tmp->isdestroyed = 0;

    gettimeofday(&tmp->created, NULL);
    
    return tmp;
}

void do_control ()
{
    char buf[1024];
    char *host;
    char *tunstr;
    char *callstr;

    char *sub_str;              /* jz: use by the strtok function */
    char *tmp_ptr;              /* jz: use by the strtok function */
    struct lac *lac;
    int call;
    int tunl;
    int cnt = -1;
    while (cnt)
    {
        cnt = read (control_fd, buf, sizeof (buf)-1);
        // make sure we null terminate...
        buf[sizeof(buf)-1] = '\0';
        
        if (cnt < 0)
            break;
        
        if (cnt > 0)
        {
            // make sure we null terminate, this is ok since we
            // "read" up to sizeof(buf) - 1
            buf[cnt] = '\0';
            
            if (buf[cnt - 1] == '\n')
                buf[--cnt] = 0;
            switch (buf[0])
            {
            case 't':
                host = strchr (buf, ' ') + 1;
                
                if (!host)
                    break;
                l2tp_call (host, UDP_LISTEN_PORT, NULL, NULL);
                break;
            case 'c':

                switch_io = 1;  /* jz: Switch for Incoming - Outgoing Calls */

                tunstr = strchr (buf, ' ') + 1;
                if (!tunstr)
                    break;
                lac = laclist;
                while (lac)
                {
                    if (!strcasecmp (lac->entname, tunstr))
                    {
                        lac->active = -1;
                        lac->rtries = 0;
                        if (!lac->c)
                            magic_lac_dial (lac);
                        break;
                    }
                    lac = lac->next;
                }
                if (lac)
                    break;
                tunl = atoi (tunstr);
                if (!tunl)
                {
                    log (LOG_DEBUG, "%s: No such tunnel '%s'\n", __FUNCTION__,
                         tunstr);
                    break;
                }
#ifdef DEBUG_CONTROL
                log (LOG_DEBUG, "%s: Attempting to call on tunnel %d\n",
                     __FUNCTION__, tunl);
#endif
                lac_call (tunl, NULL, NULL);
                break;

            case 'o':          /* jz: option 'o' for doing a outgoing call */

                switch_io = 0;  /* jz: Switch for incoming - outgoing Calls */

                sub_str = strchr (buf, ' ') + 1;
                if (!sub_str)
                    break;

                tunstr = strtok (sub_str, " "); /* jz: using strtok function to get */
                if (!tunstr)
                    break;
                
                tmp_ptr = strtok (NULL, " ");   /*     params out of the pipe       */
                if (!tmp_ptr)
                    break;
                
                strcpy (dial_no_tmp, tmp_ptr);

                lac = laclist;
                while (lac)
                {
                    if (!strcasecmp (lac->entname, tunstr))
                    {
                        lac->active = -1;
                        lac->rtries = 0;
                        if (!lac->c)
                            magic_lac_dial (lac);
                        else
                            log (LOG_DEBUG,
                                 "%s: Session '%s' already active!\n",
                                 __FUNCTION__, lac->entname);
                        break;
                    }
                    lac = lac->next;
                }
                if (lac)
                    break;
                tunl = atoi (tunstr);
                if (!tunl)
                {
                    log (LOG_DEBUG, "%s: No such tunnel '%s'\n", __FUNCTION__,
                         tunstr);
                    break;
                }
#ifdef DEBUG_CONTROL
                log (LOG_DEBUG, "%s: Attempting to call on tunnel %d\n",
                     __FUNCTION__, tunl);
#endif
                lac_call (tunl, NULL, NULL);
                break;

            case 'h':
                callstr = strchr (buf, ' ') + 1;
                if (!callstr)
                    break;
                
                call = atoi (callstr);
#ifdef DEBUG_CONTROL
                log (LOG_DEBUG, "%s: Attempting to call %d\n", __FUNCTION__,
                     call);
#endif
                lac_hangup (call);
                break;
            case 'd':
                tunstr = strchr (buf, ' ') + 1;
                if (!tunstr)
                    break;
                lac = laclist;
                while (lac)
                {
                    if (!strcasecmp (lac->entname, tunstr))
                    {
                        lac->active = 0;
                        lac->rtries = 0;
                        if (lac->t)
                            lac_disconnect (lac->t->ourtid);
                        else
                            log (LOG_DEBUG, "%s: Session '%s' not up\n",
                                 __FUNCTION__, lac->entname);
                        break;
                    }
                    lac = lac->next;
                }
                if (lac)
                    break;
                tunl = atoi (tunstr);
                if (!tunl)
                {
                    log (LOG_DEBUG, "%s: No such tunnel '%s'\n", __FUNCTION__,
                         tunstr);
                    break;
                }
#ifdef DEBUG_CONTROL
                log (LOG_DEBUG, "%s: Attempting to disconnect tunnel %d\n",
                     __FUNCTION__, tunl);
#endif
                lac_disconnect (tunl);
                break;
            case 's':
                show_status (1);
                break;
            default:
                log (LOG_DEBUG, "%s: Unknown command %c\n", __FUNCTION__,
                     buf[0]);
            }
        }
    }
    /* Otherwise select goes nuts */
    safe_close (&control_fd);
    control_fd = open (CONTROL_PIPE, O_RDONLY | O_NONBLOCK, 0600);
}

void usage(void) {
    printf("Usage: l2tpd -D -c [config file] -s [secret file] -p [pid file]\n");
    printf("\n");
    exit(1);
}

void init_args(int argc, char *argv[]) {
    int i=0;
    gconfig.daemon=1;
    memset(gconfig.altauthfile,0,STRLEN);
    memset(gconfig.altconfigfile,0,STRLEN);
    memset(gconfig.authfile,0,STRLEN);
    memset(gconfig.configfile,0,STRLEN);
    memset(gconfig.pidfile,0,STRLEN);
    strncpy(gconfig.altauthfile,ALT_DEFAULT_AUTH_FILE,
            sizeof(gconfig.altauthfile) - 1);
    strncpy(gconfig.altconfigfile,ALT_DEFAULT_CONFIG_FILE,
            sizeof(gconfig.altconfigfile) - 1);
    strncpy(gconfig.authfile,DEFAULT_AUTH_FILE,
            sizeof(gconfig.authfile) - 1);
    strncpy(gconfig.configfile,DEFAULT_CONFIG_FILE,
            sizeof(gconfig.configfile) - 1);
    strncpy(gconfig.pidfile,DEFAULT_PID_FILE,
            sizeof(gconfig.pidfile) - 1);
    gconfig.daemon = 0;
    for (i = 1; i < argc; i++) {
        if(! strncmp(argv[i],"-c",2)) {
            if(++i == argc)
                usage();
            else
                strncpy(gconfig.configfile,argv[i],
                        sizeof(gconfig.configfile) - 1);
        }
        else if (! strncmp(argv[i],"-D",2)) {
            gconfig.daemon=0;
        }
        else if (! strncmp(argv[i],"-s",2)) {
            if(++i == argc)
                usage();
            else
                strncpy(gconfig.authfile,argv[i],
                        sizeof(gconfig.authfile) - 1);
        }
        else if (! strncmp(argv[i],"-p",2)) {
            if(++i == argc)
                usage();
            else
                strncpy(gconfig.pidfile,argv[i],
                        sizeof(gconfig.pidfile) - 1);
        }
        else {
            usage();
        }
    }
}


void daemonize() {
    int pid=0;
    int i,l;
    char buf[STRLEN];
    int pidfilewritten=0;

    if((pid = fork()) < 0) {
        log(LOG_LOG, "%s: Unable to fork ()\n",__FUNCTION__);
        close(server_socket);
        exit(1);
    }
    else if (pid)
        exit(0);

    close(0);
    close(1);
    close(2);

    /* Read previous pid file. */
    if ((i = open(gconfig.pidfile,O_RDONLY)) > 0) {
        l=read(i,buf,sizeof(buf)-1);
        if (i < 0) {
            log(LOG_LOG, "%s: Unable to read pid file [%s]\n",
                    __FUNCTION__, gconfig.pidfile);
        }
        buf[i] = '\0';
        pid = atoi(buf);

        /* If the previous server process is not still running,
           write a new pid file immediately. */
        if (pid && (pid == getpid () || kill (pid, 0) < 0)) {
            unlink (gconfig.pidfile);
            if ((i = open (gconfig.pidfile, O_WRONLY | O_CREAT, 0640)) >= 0)
            {
                snprintf (buf, sizeof(buf), "%d\n", (int)getpid());
                write (i, buf, strlen(buf));
                close (i);
                pidfilewritten = 1;
            }
        }
        else
        {
            log(LOG_LOG, "%s: There's already a l2tpd server running.\n",
                    __FUNCTION__);
            close(server_socket);
            exit(1);
        }
    }

    pid = setsid();

    if(! pidfilewritten) {
        unlink(gconfig.pidfile);
        if ((i = open (gconfig.pidfile, O_WRONLY | O_CREAT, 0640)) >= 0) {
            snprintf (buf, strlen(buf), "%d\n", (int)getpid());
            write (i, buf, strlen(buf));
            close (i);
            pidfilewritten = 1;
        }
    }
}

void init (int argc,char *argv[])
{
    struct lac *lac;
    int         n;
    
    init_args (argc,argv);
    init_addr ();
    if (init_config ())
    {
        log (LOG_CRIT, "%s: Unable to load config file\n", __FUNCTION__);
        exit (1);
    }
    if (uname (&uts))
    {
        log (LOG_CRIT, "%s : Unable to determine host system\n",
             __FUNCTION__);
        exit (1);
    }
    init_tunnel_list (&tunnels);
    if (init_network ())
      exit (1);
    if (gconfig.daemon)
	daemonize ();

#if 0
	for (n = 1; n < _NSIG; n++)
		signal(n, SIG_IGN);
#endif
    signal (SIGPIPE, SIG_IGN);
    
    signal (SIGTERM, &death_handler);
    signal (SIGINT, &death_handler);
    signal (SIGCHLD, &child_handler);
    signal (SIGUSR1, &status_handler);
    signal (SIGHUP, &null_handler);
    //init_scheduler ();
    mkfifo (CONTROL_PIPE, 0600);
    control_fd = open (CONTROL_PIPE, O_RDONLY | O_NONBLOCK, 0600);
    if (control_fd < 0)
    {
        log (LOG_CRIT, "%s: Unable to open " CONTROL_PIPE " for reading.",
             __FUNCTION__);
        exit (1);
    }

    lac = laclist;
    while (lac)
    {
        if (lac->autodial)
        {
            lac->active = -1;
            switch_io = 1;      /* If we're a LAC, autodials will be ICRQ's */
            magic_lac_dial (lac);
        }
        lac = lac->next;
    }
}

#include <sys/resource.h>

//static int FDS[1500];

int main (int argc, char *argv[])
{
    /* increase number of file descriptors to MAX_FDS */
    struct rlimit rl;

    memset(&rl, 0, sizeof(rl));
    if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
        fprintf(stderr, "getrlimit NOFILE cur=%d, max=%d\n",
                (int)rl.rlim_cur, (int)rl.rlim_max);
        rl.rlim_cur = rl.rlim_max = MAX_FDS;
        if (setrlimit(RLIMIT_NOFILE, &rl) == 0) {
            fprintf(stderr, "setrlimit NOFILE cur=%d, max=%d\n",
                    (int)rl.rlim_cur, (int)rl.rlim_max);
        } else {
            fprintf(stderr, "setrlimit NOFILE cur=%d, max=%d FAILED\n",
                    (int)rl.rlim_cur, (int)rl.rlim_max);
        }
    } else {
        fprintf(stderr, "getrlimit failed\n");
    }
    
    if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
        fprintf(stderr, "configured with cur=%d, max=%d\n",
                (int)rl.rlim_cur, (int)rl.rlim_max);
    } else
        fprintf(stderr, "getrlimit failed\n");

#if 0
    fprintf(stderr, "trying to open 1500 FDs\n");
    for (i=0; i<1500; i++) {
        if ((FDS[i] = socket(PF_INET, SOCK_DGRAM, 0)) < 0) {
            fprintf(stderr, "socket failed at index %d\n", i);
            break;
        }
    }
    fprintf(stderr, "Opened %d sockets\n", i);
    for (i=i-1; i>=100; i--) {
        close(FDS[i]);
    }
#endif
    init_tunnels();
    
    init(argc,argv);
    //dial_no_tmp = calloc (128, sizeof (char));
    dial_no_tmp = vpnmalloc(__FUNCTION__, 128);
    memset(dial_no_tmp, 0, 128);
    if ( l2tpd_ipc_init() )
      return (-1);
    network_thread ();
    return 0;
}

static int l2tpd_ipc_init()
{
  int rc = 0;

  ipc_fd = socket(AF_INET, SOCK_DGRAM, 0);
  
  if (ipc_fd < 0)
  {
    log (LOG_CRIT, "%s: Failed in socket()\n", __FUNCTION__);
    rc = -1;
    goto exit;
  }

  ipc_sockaddr.sin_family = AF_INET;
  if ( inet_pton(AF_INET,IKE_DAEMON_RAW_IP, &ipc_sockaddr.sin_addr) <= 0)
  {
    log (LOG_CRIT, "%s: Failed in inet_pton() errno %d\n",__FUNCTION__, errno);
    rc = -1;
    goto exit;
  }
  ipc_sockaddr.sin_port = htons(L2TPD_DAEMON_RAW_PORT);

  rc = bind (ipc_fd, (struct sockaddr *) &ipc_sockaddr, sizeof(ipc_sockaddr));
  if (rc < 0)
  {
    log(LOG_CRIT, "Cannot bind socket for raw messaging errno %d", errno);
    rc = -1;
    goto exit;
  }

  /* Prepare the sockaddr for the sendto */
  ipc_sockaddr.sin_port   = htons(IKE_DAEMON_RAW_PORT);

 exit:
  return(rc);
}


void l2tp_shared_ipc_handler(char *buf, int buflen)
{
  struct call *call;
  char innerip[16];
  struct in_addr ip;
  fpappsMsg_t *fpappsMsg;

  if ( ((msgHndl_Hdr *)buf)->SrcPortNum == ntohs (FASTPATH_SERVER))
  {
    fpappsMsg = (fpappsMsg_t *) (buf + sizeof (msgHndl_Hdr));
    if (fpappsMsg->msgType == FPAPPS_RESP_SWITCH_IP)
    {
      listen_addy = fpappsMsg->msg.switchIPResp.ipaddr;

      if (listen_addy)
	      mswitch_addy = listen_addy;
      
      if ( init_network() )
	exit(1);//TODO : LOG
    }
    else
      log(LOG_ERR, "Invalid message from fpapps: %d.", fpappsMsg->msgType);
  }
  else
  {
      enum L2TP_MSG *type;

      if (buflen <= (sizeof(msgHndl_Hdr) + 4))
          return;
      type = (enum L2TP_MSG *)((msgHndl_Hdr *)buf + 1);

      switch(*type) {
      case L2TP_IPUP: {
          l2tp_ipc_ppp_msg *msg = (l2tp_ipc_ppp_msg *)(type);

          if (buflen != (sizeof(msgHndl_Hdr)+msg->m_size)) {
              arubaLog(LOG_ERR, "Size mismatch in L2TP_IPUP %d vs %d, hdlsize=%d",
                       buflen, sizeof(msgHndl_Hdr)+msg->m_size,
                       sizeof(msgHndl_Hdr));
              return;
          }
          ip.s_addr = msg->radiusip;
          strcpy(innerip, inet_ntoa(ip));
          
          log(LOG_DEBUG, "IP UP from PPPD: TID %d, CID %d, Inner ip %s\n",
              msg->ourtid, msg->ourcid, innerip);
          
          // 0 represents the IP and PORT, it is not needed when doing a search
          call = get_call (msg->ourtid, msg->ourcid, 0, 0);
          
          if (!call)
              log(LOG_ERR, "IP UP msg, could not find the corresponding tunnel");
          else {
              if (call->addr && call->addr != msg->radiusip) {
                  shared_cli_free_addr(call->addr);
                  call->radius_ip = 1;
              }
              //DO NOT bring the session down this is called on session up !!
              
              // Update the inner IP to be the RADIUS allocated IP.
              call->addr = msg->radiusip;
              
              if (l2tpd_ipc_session_up(call))
                  goto exit; // TODO: LOG
          }
      }
          break;
      case L2TP_IPDOWN: {
          // msg from ike
          l2tp_ipdown_msg *msg = (l2tp_ipdown_msg *)(type);
          if (buflen != (sizeof(msgHndl_Hdr)+msg->m_size)) {
              arubaLog(LOG_ERR, "Size mismatch in L2TP_IPDOWN %d vs %d",
                       buflen, sizeof(msgHndl_Hdr)+msg->m_size);
              return;
          }
          ip.s_addr = msg->innerip;
          strcpy(innerip, inet_ntoa(ip));
          shared_cli_free_addr(msg->innerip);
      }
          break;
      case L2TP_TUNNDOWN: {
          // TODO: shutdown L2TP tunnel and call
          // msg from ike
          l2tp_ipdown_msg *msg = (l2tp_ipdown_msg *)(type);
          if (buflen != (sizeof(msgHndl_Hdr)+msg->m_size)) {
              arubaLog(LOG_ERR, "Size mismatch in L2TP_IPDOWN %d vs %d",
                       buflen, sizeof(msgHndl_Hdr)+msg->m_size);
              return;
          }
          ip.s_addr = msg->innerip;
          strcpy(innerip, inet_ntoa(ip));
          arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG,
                        "Tunnel tear down from initiated from IKE, innerip=%s\n",
                        innerip);
          
          tunndown(msg->innerip);
      }
          break;
      case L2TP_IKEREKEY: {
          l2tp_ikerekey_msg *msg = (l2tp_ikerekey_msg *)(type);
          struct tunnel *t;
          if (buflen != (sizeof(msgHndl_Hdr)+msg->m_size)) {
              arubaLog(LOG_ERR, "Size mismatch in L2TP_IKEREKEY %d vs %d",
                       buflen, sizeof(msgHndl_Hdr)+msg->m_size);
              return;
          }
          arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG,
                        "IKE rekey, tunnel=%d, outerip=%s\n",
                        msg->ourtid, IPTOSTR(msg->outerip));
          
          t = get_tunnel(msg->ourtid, 0, 0);
          if (t) {
              t->inSPI = msg->inSPI;
              t->outSPI = msg->outSPI;
          } else {
              /* not a fatal problem, so leave under debug */
              arubaLog(LOG_DEBUG, "Rekeying tunnel %d, but tunnel not found, clientIP=%s", msg->ourtid, IPTOSTR(msg->outerip));
          }
      }
          break;
      }
  }
 exit:
  // If the IP is set free it !!
  return;

}


static int l2tpd_ipc_session_up(struct call *c)
{
  ike_ipc_msg msg;
  int rc = 0;
  int len;
  struct in_addr ip;
  char outip[16];
  char inip[16];

  if (!c)
    goto exit;

  msg.msgtype           = IKE_IPC_CREATE_L2TP_SESSION;
  msg.l2tp.tid          = c->container->tid;
  msg.l2tp.ourtid       = c->container->ourtid;
  msg.l2tp.cid          = c->cid;
  msg.l2tp.ourcid       = c->ourcid;
  msg.l2tp.l2tphello    = shared_cli_conf.hello_time;
  msg.outer_ip          = c->container->peer.sin_addr.s_addr;
  msg.inner_ip          = c->addr;
  msg.inSPI             = c->container->inSPI;
  msg.outSPI            = c->container->outSPI;

  c->container->indp    = 1;
  
  ip.s_addr = msg.outer_ip;
  strcpy(outip, inet_ntoa(ip));

  ip.s_addr = msg.inner_ip;
  strcpy(inip, inet_ntoa(ip));

  arubaLog(LOG_INFO, "Creating L2TP Tunnel from %s(innerip=%s)", outip, inip);
  arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG,
                "Msg type: %d; Peer tunnel id %d, our tunnel id %d, peer call id %d, our call id %d", msg.msgtype,
      msg.l2tp.tid, msg.l2tp.ourtid, msg.l2tp.cid,
      msg.l2tp.ourcid);

 resend:
  len = sendto(ipc_fd, &msg, sizeof(msg), 0, (struct sockaddr *)&ipc_sockaddr,
	       sizeof(ipc_sockaddr));

  if ( len != sizeof(msg) )
  {
    if (errno == EINTR)
      goto resend;

    log (LOG_CRIT, "%s: Failed in sendto() errno %d\n", __FUNCTION__, errno);
    rc = -1;
    goto exit;
  }

 exit:
  return (rc);

}

int l2tpd_ipc_session_down(struct call *c)
{
  ike_ipc_msg msg;
  int rc = 0;
  int len;
  struct in_addr ip;
  char outip[16];
  char inip[16];

  if (!c)
    goto exit;

  msg.msgtype           = IKE_IPC_DESTROY_L2TP_SESSION;
  msg.l2tp.tid          = c->container->tid;
  msg.l2tp.ourtid       = c->container->ourtid;
  msg.l2tp.cid          = c->cid;
  msg.l2tp.ourcid       = c->ourcid;
  msg.outer_ip          = c->container->peer.sin_addr.s_addr;
  msg.inner_ip          = c->addr;
  msg.inSPI             = c->container->inSPI;
  msg.outSPI            = c->container->outSPI;

  ip.s_addr = msg.outer_ip;
  strcpy(outip, inet_ntoa(ip));

  ip.s_addr = msg.inner_ip;
  strcpy(inip, inet_ntoa(ip));

  arubaLog(LOG_INFO, "Deleting L2TP Tunnel from %s(innerip=%s)", outip, inip);
  arubaDebugLog(DEBUG_VPDN_PACKETS, LOG_DEBUG,
                "Msgtype: %d; Peer tunnel id %d, our tunnel id %d, peer call id %d, our call id %d", msg.msgtype,
      msg.l2tp.tid, msg.l2tp.ourtid, msg.l2tp.cid,
      msg.l2tp.ourcid);
  
 resend:
  len = sendto(ipc_fd, &msg, sizeof(msg), 0, (struct sockaddr *)&ipc_sockaddr,
	       sizeof(ipc_sockaddr));

  if ( len != sizeof(msg) )
  {
    if (errno == EINTR)
      goto resend;

    log (LOG_CRIT, "%s: Failed in sendto() errno %d\n", __FUNCTION__, errno);
    rc = -1;
    goto exit;
  }

 exit:
  return (rc);
}

#if 0
static void l2tpd_ipc_close()
{
  if (ipc_fd)
    close(ipc_fd);

  return;
}
#endif


int
get_ipc_fd()
{
    return ipc_fd;
}


int
get_control_fd()
{
    return control_fd;
}


