/*
 * Layer Two Tunnelling Protocol Daemon
 * Copyright (C) 1998 Adtran, Inc.
 * Copyright (C) 2002 Jeff McAdams
 *
 * Mark Spencer
 *
 * This software is distributed under the terms
 * of the GPL, which you should have received
 * along with this source.
 *
 * Pseudo-pty allocation routines...  Concepts and code borrowed
 * from pty-redir by Magosanyi Arpad.
 *
 */

#include <stdio.h>
#include "l2tp.h"
#include <fcntl.h>
#define __USE_XOPEN
#define __USE_GNU
#include <stdlib.h>
#undef __USE_XOPEN
#undef __USE_GNU
#include <unistd.h>
#include <string.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <errno.h>
#ifdef SOFTAP
#include <stropts.h>
#endif
#ifdef SOLARIS
#define PTY00 "/dev/ptyXX"
#define PTY10 "pqrstuvwxyz"
#define PTY01 "0123456789abcdef"
#endif

#define PTY00 "/dev/ptyXX"
#define PTY10 "pqrstuvwxyzabcde"
#define PTY01 "0123456789abcdef"

#ifdef FREEBSD

#define PTY00 "/dev/ptyXX"
#define PTY10 "p"
#define PTY01 "0123456789abcdefghijklmnopqrstuv"
#endif

#if 0
int getPtyMaster (char *tty10, char *tty01)
{
    char *p10;
    char *p01;
    static char dev[] = PTY00;
    int fd;

    for (p10 = PTY10; *p10; p10++)
    {
        dev[8] = *p10;
        for (p01 = PTY01; *p01; p01++)
        {
            dev[9] = *p01;
            fd = open (dev, O_RDWR | O_NONBLOCK);
            if (fd >= 0)
            {
                *tty10 = *p10;
                *tty01 = *p01;
                return fd;
            }
        }
    }
    l2_log (LOG_CRIT, "%s: No more free pseudo-tty's\n", __FUNCTION__);
    return -1;
}
#else

int
my_openpty(int *master, int *slave, char *name)
{
     int fd_master, fd_slave;
     char slave_name[32];
     int ptyno;
     slave_name[0] = '\0';
     
     fd_master = open("/dev/ptmx", O_RDWR);
     if (fd_master < 0) {
         l2_log(LOG_CRIT, "L2TPD: Failed to allocate pty");
         return -1;
     }

     if (ioctl(fd_master, TIOCGPTN, &ptyno) == 0) {
         sprintf(slave_name, "/dev/pts/%d", ptyno);
     } else {
         l2_log(LOG_CRIT, "TIOCGPTN Failed on fd_master");
     }
     
     l2_log(LOG_DEBUG, "L2TPD: pty name is %s", slave_name);
     grantpt(fd_master);
     unlockpt(fd_master);
     if (name) {
         strcpy(name, slave_name);
     }
     fd_slave = open(slave_name, O_RDWR);
     if (fd_slave < 0) {
         l2_log(LOG_CRIT, "L2TPD: Failed to allocate slave pty %s, master pty is %d", slave_name, fd_master);
         return -1;
     }
#ifdef SOFTAP // Fails in MIPS since no STREAMS support
     if (ioctl(fd_slave, I_PUSH, "ptem"))
       l2_log(LOG_CRIT, "I_PUSH: ptem %s", strerror(errno)) ;
     if (ioctl(fd_slave, I_PUSH, "ldterm"))
       l2_log(LOG_CRIT, "I_PUSH: ldterm %s", strerror(errno)) ;
#endif
     if (master)
         *master = fd_master;
     if (slave)
         *slave = fd_slave;
     l2_log(LOG_DEBUG, "L2TPD: Allocated pty/tty pair (%d,%d)",
	 fd_master, fd_slave);
     return 0;
}

int l2tp_get_pty (int *fd2, char *tty)
{
  int p,i ;

  p = open("/dev/ptmx", 2);
  if (p > 0){
    grantpt(p);
    unlockpt(p);
    if (isatty(p) && (ioctl(p, TIOCGPTN, &i) == 0)){
      sprintf(tty, "/dev/pts/%d", i);
    }
    l2_log(LOG_LOG, "l2tp_get_pty: slave TYY is %s\n", tty) ;
    *fd2 = open(tty, O_RDWR) ;
  }
  return p ;	   
}
#endif
