/*
 *  linux/arch/ppc/kernel/cabernet_exp.c
 *
 *  Derived from "arch/ppc/kernel/process.c"
 *    Copyright (C) 1995  Linus Torvalds
 *
 *  Updated and modified by Sameer Dekate (sdekate@arubanetworks.com)
 *
 * When a PANIC occurs this program saves the info into the RTC NVRAM 
 * for analysis after the reboot.
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/user.h>
#include <linux/elf.h>
#include <linux/init.h>
#include <linux/prctl.h>
#include <linux/trace.h>
#include <linux/compile.h>
#include <asm/pgtable.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/processor.h>
#include <asm/mmu.h>
#include <asm/prom.h>
#include <asm/cabernet_panic.h>

#include "../../../../cmn/util/nvram.h"
#include "../../../../cmn/util/reboot.h"
void cab_flash_write(char *flash_addr, char *mem_addr, int len);
extern ulong todc_get_rtc_time(void);
extern unsigned long _get_SP(void);

unsigned int findEmptySpotInNvram(void)
{
	PanicRecord *myPtr;
	int oldestTime = -1;
	int oldestLoc = 0;
	int i;

	for(i=0;i<NVRAM_PANICS;i++){
		myPtr = (PanicRecord*) ((unsigned int)NVRAM_BASE_ADDRESS +
				        (unsigned int)NVRAM_PANIC_OFFSET(i));

		if(myPtr->panicHeader.Magic != MAGIC_VAL){
			printk("No panic at 0x%08x using it\n", (int) myPtr);
			return ((unsigned int) myPtr);
		} else {
			/* find the one with least seconds (since Jan 1st 1970) */
			if(oldestTime > myPtr->panicHeader.CurrentTime) {
				oldestTime = myPtr->panicHeader.CurrentTime;
				oldestLoc  = i;
			}
		}
	}
	/* no empty spot found - erase the oldest */
	myPtr = (PanicRecord*) ((unsigned int)NVRAM_BASE_ADDRESS +
			        (unsigned int)NVRAM_PANIC_OFFSET(oldestLoc));
	myPtr->panicHeader.Magic = 0;
	printk("ARUBA: Empty spot %d addr 0x%x.\n", oldestLoc, (unsigned int)myPtr );
	return ((unsigned int)myPtr);
}

PanicRecord rec;

extern void do_gettimeofday(struct timeval *tv);
void SaveRegisters(int panic, char* panicString, struct pt_regs * regs)
{
	unsigned int nvramLoc, i;
	unsigned char rebootCause;
	unsigned char version[128];
	struct timeval tv;

	nvramLoc = findEmptySpotInNvram();
	if(nvramLoc == -1) {
		printk("ARUBA: No place to save panic info.\n");
		return;
	}

	printk("ARUBA: Saving registers into NVRAM at 0x%x (0x%x)\n", nvramLoc, MAGIC_VAL);

	sprintf(version,"%s by %s", ARUBA_VERSION, LINUX_COMPILE_BY );

	/* fill in rec */
	memset(&rec, 0, sizeof(rec));
	rec.panicHeader.Magic        = MAGIC_VAL;
	rec.panicHeader.PanicInfoVer = PANIC_INFO_VER;
	do_gettimeofday(&tv);
	rec.panicHeader.CurrentTime  = tv.tv_sec;
	rec.panicHeader.LastSyscall  = current->thread.last_syscall;
	strncpy(rec.panicHeader.taskCommand, current->comm, sizeof(rec.panicHeader.taskCommand));
	strncpy(rec.panicHeader.KernelVersion, version, sizeof(rec.panicHeader.KernelVersion));
	if(panic){ 
		int size = (sizeof(rec.u.panicInfo.PanicString) < strlen(panicString)) ? 
			sizeof(rec.u.panicInfo.PanicString) : strlen(panicString);
		rec.panicHeader.IsPanic = 1;            /* got a panic */
		strncpy(rec.u.panicInfo.PanicString, panicString, size);
		printk("SP = 0x%08X\n", _get_SP());
		rec.u.panicInfo.sp = _get_SP();
		memcpy(rec.u.panicInfo.Stack, (void *)_get_SP(), sizeof(rec.u.panicInfo.Stack));
	} else {
		rec.panicHeader.IsPanic = 0;            /* not a panic - exception */
		memcpy(&rec.u.regAndStack.regs, regs, sizeof(rec.u.regAndStack.regs));
		memcpy(&rec.u.regAndStack.Stack, (void *)regs->gpr[1], sizeof(rec.u.regAndStack.Stack));
	}

	syr_flash_write(nvramLoc, &rec, sizeof(rec));

	/* save cause of reboot */
	rebootCause = INTENTIONAL_REBOOT;
	syr_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_INTENT_OFFSET, 
			&rebootCause, 
			1);

	rebootCause = PANIC_REBOOT;
	syr_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_CAUSE_OFFSET, 
			&rebootCause, 
			1);

	printk("done\n");

	return;
}
