/*
 * arch/ppc/platforms/cabernet.h
 * 
 * Definitions for Aruba Cabernet Platform
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 *         Sameer M. Dekate
 *         sdekate@arubanetworks.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * Cabernet uses the CHRP map (Map B).
 */

#ifndef __PPC_PLATFORMS_CABERNET_H
#define __PPC_PLATFORMS_CABERNET_H

#define CABERNET_EUMB_BASE     0xFC000000

/*
 * The cabernet boards have processor modules that have an 8241, which has
 * an MPC107 host bridge.  These bridges have an IDSEL line that allows
 * them to respond to PCI transactions as if they were a normal PCI devices.
 * However, the processor on the processor side of the bridge can not reach
 * out onto the PCI bus and then select the bridge or bad things will happen
 * (documented in the 8240 and 107 manuals).
 * Because of this, we always skip the bridge PCI device when accessing the
 * PCI bus.  The PCI slot that the bridge occupies is defined by the macro
 * below.
 */
#define CABERNET_HOST_BRIDGE_IDSEL     12

void cabernet_find_bridges(void);

/* Cabernet base addresses */
#define CABERNET_COMPACT_FLASH_BASE		0x70000000
#define CABERNET_FPGA_BASE			0x70010000
#define CABERNET_CPLD_BASE			0xFF000000

/* CPLD definitions */
/* CPLD registers */
#define CABERNET_CPLD_FPGA_INTERFACE_REG(x)	((x) + 0x0)
#define CABERNET_CPLD_TI_POE_REG(x)		((x) + 0x1)
#define CABERNET_CPLD_RESET_ENABLE_REG(x)	((x) + 0x2)
#define CABERNET_CPLD_RESET_REG(x)		((x) + 0x3)
#define CABERNET_CPLD_RESERVED_0_REG(x)		((x) + 0x4)
#define CABERNET_CPLD_RESERVED_1_REG(x)		((x) + 0x5)
#define CABERNET_CPLD_JTAG_REG(x)		((x) + 0x6)
#define CABERNET_CPLD_RESERVED_2_REG(x)		((x) + 0x7)
#define CABERNET_CPLD_VER_REV_REG(x)		((x) + 0x8)

/* Bit definitions of CPLD registers */

/* register settings for CABERNET_CPLD_FPGA_INTERFACE_REG */
#define CABERNET_CPLD_FPGA_PRO			0x01
#define CABERNET_CPLD_FPGA_INIT			0x02
#define CABERNET_CPLD_FPGA_DONE			0x04
#define CABERNET_CPLD_FPGA_DLL_RST		0x08
#define CABERNET_CPLD_FPGA_PROGRAM_ENABLE	0x80

/* register settings for CABERNET_CPLD_TI_PEO_REG */
#define CABERNET_CPLD_TI_POE_ENABLE		0x01

/* register settings for CABERNET_CPLD_RESET_ENABLE_REG */
#define CABERNET_CPLD_RESET_ENABLE		0x05

/* register settings for CABERNET_CPLD_RESET_REG */
#define CABERNET_CPLD_RESET_BCM1125		0x01
#define CABERNET_CPLD_RESET_COMPACT_FLASH	0x02
#define CABERNET_CPLD_RESET_FPGA		0x04
#define CABERNET_CPLD_RESET_CRYPTO		0x08
#define CABERNET_CPLD_RESET_ROBOSWITCH		0x10
#define CABERNET_CPLD_RESET_TI_POE		0x20
#define CABERNET_CPLD_RESET_PHYTER		0x40
#define CABERNET_CPLD_RESET_MASTER		0x80

/* register settings for CABERNET_CPLD_JTAG_REG */
#define CABERNET_CPLD_JTAG_TDO			0x01
#define CABERNET_CPLD_JTAG_TDI			0x02
#define CABERNET_CPLD_JTAG_TMS			0x04
#define CABERNET_CPLD_JTAG_CLK			0x08

#endif /* __PPC_PLATFORMS_CABERNET_H */
