/*
 * arch/ppc/platforms/malbec_pci.c
 * 
 * PCI setup routines for the Malbec Platform
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 *         Sameer M. Dekate (Malbec mods)
 *         sdekate@arubanetworks.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/mpc10x.h>

#include "malbec.h"

/*
 * Malbec interrupt routing.
 */
static inline int
malbec_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{
	static char pci_irq_table[][4] =
	/*
	 *	PCI IDSEL/INTPIN->INTLINE 
	 * 	   A   B   C   D
	 */
	{
		{ 16, 16, 16, 16 },/* BCM1125= IDSEL 20, devNo 20, irq 16 = int0 on 8241 */
		{ 17, 17, 17, 17 },/* CRYPTO = IDSEL 21, devNo 24, irq 17 = int1 on 8241 */
	};

	/* 
	 * Compact Flash is at no IDSEL (not a pci device), no devNo, irq 2 on 8241
	 * FPGA          is at no IDSEL (not a pci device), no devNo, irq 3 on 8241
	 * LM75          is at no IDSEL (not a pci device), no devNo, irq 4 on 8241
	 */
	const long min_idsel = 20, max_idsel = 21, irqs_per_slot = 4;
	return PCI_IRQ_TABLE_LOOKUP;
}

static int                     
malbec_exclude_device(u_char bus, u_char devfn)
{
	if ((bus == 0) && (PCI_SLOT(devfn) == MALBEC_HOST_BRIDGE_IDSEL)) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	else {
		return PCIBIOS_SUCCESSFUL;
	}
}

void __init
malbec_find_bridges(void)
{
	struct pci_controller	*hose;

	hose = pcibios_alloc_controller();

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno = 0xff;

	if (mpc10x_bridge_init(hose,
			       MPC10X_MEM_MAP_B,
			       MPC10X_MEM_MAP_B,
			       MALBEC_EUMB_BASE) == 0) {

		hose->mem_space.start = 0xf9000000;
		hose->mem_space.end = 0xfbffffff;

		ppc_md.pci_exclude_device = malbec_exclude_device;
		hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);

		ppc_md.pcibios_fixup = NULL;
		ppc_md.pcibios_fixup_bus = NULL;
		ppc_md.pci_swizzle = common_swizzle;
		ppc_md.pci_map_irq = malbec_map_irq;
	}
	else {
		if (ppc_md.progress)
			ppc_md.progress("Bridge init failed", 0x100);
		printk("Host bridge init failed\n");
	}

	return;
}
