/*
 * arch/ppc/platforms/mswitch.h
 * 
 * Definitions for Motorola SPS Mswitch Test Platform
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */

/*
 * Mswitch uses the CHRP map (Map B).
 */

#ifndef __PPC_PLATFORMS_MSWITCH_H
#define __PPC_PLATFORMS_MSWITCH_H

#define MSWITCH_EUMB_BASE     0xFC000000

/*
 * The mswitch boards have processor modules that have an 8245, which has
 * an MPC107 host bridge.  These bridges have an IDSEL line that allows
 * them to respond to PCI transactions as if they were a normal PCI devices.
 * However, the processor on the processor side of the bridge can not reach
 * out onto the PCI bus and then select the bridge or bad things will happen
 * (documented in the 8240 and 107 manuals).
 * Because of this, we always skip the bridge PCI device when accessing the
 * PCI bus.  The PCI slot that the bridge occupies is defined by the macro
 * below.
 */
#define MSWITCH_HOST_BRIDGE_IDSEL     12

void mswitch_find_bridges(void);

/* CPLD definitions */

/* CPLD registers */
#define MSWITCH_CPLD_BASE			0x70000000
#define MSWITCH_FPGA_BASE			0x70001000
#define MSWITCH_CPLD_SIZE			0x70002000
#define MSWITCH_CPLD_INTERFACE_REG(x)		((x) + 0x0)
#define MSWITCH_CPLD_ACTIVE_CONTROL_REG(x)	((x) + 0x1)
#define MSWITCH_CPLD_RESET_ENABLE_REG(x)	((x) + 0x2)
#define MSWITCH_CPLD_LOCAL_RESET_REG(x)		((x) + 0x3)
#define MSWITCH_CPLD_REMOTE_RESET_REG(x)	((x) + 0x4)
#define MSWITCH_CPLD_JTAG_SEL_CNTRL_REG(x)	((x) + 0x5)
#define MSWITCH_CPLD_VER_REV_REG(x)		((x) + 0x6)

/* Bit definitions of CPLD registers */

/* register settings for MSWITCH_CPLD_RESET_ENABLE_REG */
#define MSWITCH_CPLD_RESET_ENABLE	0x55

/* register settings for MSWITCH_CPLD_LOCAL_RESET_REG */
#define MSWITCH_CPLD_LOCAL_RESET_BCM1250	0x01
#define MSWITCH_CPLD_LOCAL_RESET_I82559		0x02
#define MSWITCH_CPLD_LOCAL_RESET_TI1520		0x04
#define MSWITCH_CPLD_LOCAL_RESET_PM8354		0x08
#define MSWITCH_CPLD_LOCAL_RESET_FPGA		0x01
#define MSWITCH_CPLD_LOCAL_RESET_CN1320		0x02
#define MSWITCH_CPLD_LOCAL_RESET_CRYPT_FPGA	0x04
#define MSWITCH_CPLD_LOCAL_RESET_MASTER		0x08


/* register settings for MSWITCH_CPLD_REMOTE_RESET_REG */
#define MSWITCH_CPLD_REMOTE_RESET_LC0		0x01
#define MSWITCH_CPLD_REMOTE_RESET_LC1		0x02
#define MSWITCH_CPLD_REMOTE_RESET_LC2		0x04
#define MSWITCH_CPLD_REMOTE_RESET_PEER_SC	0x08

#endif /* __PPC_PLATFORMS_MSWITCH_H */
