/*
 * arch/ppc/platforms/mswitch_pci.c
 * 
 * PCI setup routines for the Mswitch Platform
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/mpc10x.h>

#include "mswitch.h"

/*
 * Mswitch interrupt routing.
 */
static inline int
mswitch_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{
	static char pci_irq_table[][4] =
	/*
	 *	PCI IDSEL/INTPIN->INTLINE 
	 * 	   A   B   C   D
	 */
	{
		{ 16, 16, 16, 16 },/* BCM1250    = IDSEL 20, devNo 20, irq 16 = int0 on 8245 */
		{ 18, 18, 18, 18 },/* i82559     = IDSEL 22, devNo 22, irq 18 = int2 on 8245 */
		{ 17, 17, 17, 17 },/* TI PCI1520 = IDSEL 21, devNo 21, irq 17 = int1 on 8245 */
	};

	/* 
	 * SCCI ARUBA FPGA is at NO IDSEL (not a pci device), no devNo, irq 4
	 * ARUBA Daughter card is at IDSEL 23&24, devNo 23&24, irq 5
	 */
	const long min_idsel = 20, max_idsel = 22, irqs_per_slot = 4;
	return PCI_IRQ_TABLE_LOOKUP;
}

static int                     
mswitch_exclude_device(u_char bus, u_char devfn)
{
	if ((bus == 0) && (PCI_SLOT(devfn) == MSWITCH_HOST_BRIDGE_IDSEL)) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	else {
		return PCIBIOS_SUCCESSFUL;
	}
}

void __init
mswitch_find_bridges(void)
{
	struct pci_controller	*hose;

	hose = pcibios_alloc_controller();

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno = 0xff;

	if (mpc10x_bridge_init(hose,
			       MPC10X_MEM_MAP_B,
			       MPC10X_MEM_MAP_B,
			       MSWITCH_EUMB_BASE) == 0) {

		ppc_md.pci_exclude_device = mswitch_exclude_device;
		hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);

		ppc_md.pcibios_fixup = NULL;
		ppc_md.pcibios_fixup_bus = NULL;
		ppc_md.pci_swizzle = common_swizzle;
		ppc_md.pci_map_irq = mswitch_map_irq;
	}
	else {
		if (ppc_md.progress)
			ppc_md.progress("Bridge init failed", 0x100);
		printk("Host bridge init failed\n");
	}

	return;
}
