/*
 * arch/ppc/platforms/sap_pci.c
 * 
 * PCI setup routines for the Accton SAP Platform
 *
 * Author: Mark A. Greer
 *         mgreer@mvista.com
 *
 * Copyright 2000, 2001 MontaVista Software Inc.
 *
 * This program is free software; you can redistribute  it and/or modify it
 * under  the terms of  the GNU General  Public License as published by the
 * Free Software Foundation;  either version 2 of the  License, or (at your
 * option) any later version.
 */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/slab.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/machdep.h>
#include <asm/pci-bridge.h>
#include <asm/mpc10x.h>

//fadi #include "mswitch.h"
#include "sap.h"

/*
 * sap(??) interrupt routing.
 */
static inline int
sap_map_irq(struct pci_dev *dev, unsigned char idsel, unsigned char pin)
{
    int ret = 0;
    static char pci_irq_table[][4] =
	/*
	 *	PCI IDSEL/INTPIN->INTLINE 
	 * 	   A   B   C   D
	 */
      {
        { 19, 19, 19, 19 },	/* IDSEL 18 - ASAP PCI slot 1 (Intersil) uses IRQ19 */
        { 18, 18, 18, 18 },	/* IDSEL 19 - ASAP PCI slot 3 (ADMtek) uses IRQ18 */
        { 17, 17, 17, 17 },	/* IDSEL 20 - ASAP PCI slot 2 (Atheros) uses IRQ17 */
        { 0,  0,  0,  0  },	/* IDSEL 21 */
        { 0,  0,  0,  0  },	/* IDSEL 22 */
        { 0,  0,  0,  0  },	/* IDSEL 23 */
        { 0,  0,  0,  0  },	/* IDSEL 24 */
        { 0,  0,  0,  0  },	/* IDSEL 25 */
        { 0,  0,  0,  0  },	/* IDSEL 26 */
        { 0,  0,  0,  0  },	/* IDSEL 27 */
        { 19, 19, 19, 19  },	/* IDSEL 28 - Airespace PCI slot 1 (Atheros) uses IRQ19 */
        { 18, 18, 18, 18 },	/* IDSEL 29 - Airespace PCI slot 3 (NatSemi) uses IRQ18 */
        { 20, 20, 20, 20 },	/* IDSEL 30 - Airespace PCI slot 2 (Atheros) uses IRQ20 */
      };

    const long min_idsel = 18, max_idsel = 30, irqs_per_slot = 1;
    ret = PCI_IRQ_TABLE_LOOKUP; // in include/asm/pci_bridge.h
    return ret;
}

static int                     
sap_exclude_device(u_char bus, u_char devfn)
{
	if ((bus == 0) && (PCI_SLOT(devfn) == MSWITCH_HOST_BRIDGE_IDSEL)) {
		return PCIBIOS_DEVICE_NOT_FOUND;
	}
	else {
		return PCIBIOS_SUCCESSFUL;
	}
}

void __init
sap_find_bridges(void)
{
	struct pci_controller	*hose;

    hose = pcibios_alloc_controller();

	if (!hose)
		return;

	hose->first_busno = 0;
	hose->last_busno = 0xff;

    if (mpc10x_bridge_init(hose,
			       MPC10X_MEM_MAP_B,
			       MPC10X_MEM_MAP_B,
			       SAP_EUMB_BASE) == 0) {

        ppc_md.pci_exclude_device = sap_exclude_device;
		hose->last_busno = pciauto_bus_scan(hose, hose->first_busno);

		ppc_md.pcibios_fixup = NULL;
		ppc_md.pcibios_fixup_bus = NULL;
		ppc_md.pci_swizzle = common_swizzle;
		ppc_md.pci_map_irq = sap_map_irq;
	}
	else {
		if (ppc_md.progress)
			ppc_md.progress("Bridge init failed", 0x100);
    }

	return;
}
