#define	Z_WAKE

/*
 *  linux/drivers/char/cabernet.c
 *
 * This file contains the driver for the cabernet switch
 * that contain serial ports on ethernet ports.
 *
 * Based upon cyclades.c
 *
 * Much of the design and some of the code came from serial.c
 * which was copyright (C) 1991, 1992  Linus Torvalds.  It was
 * extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92,
 * and then fixed as suggested by Michael K. Johnson 12/12/92.
 */

/* If you need to install more boards than NR_CARDS, change the constant
   in the definition below. No other change is necessary to support up to
   eight boards. Beyond that you'll have to extend cab_isa_addresses. */

#define NR_CARDS        1 /* not to be confused with line cards */
                          /* there is only one FPGA, and hence 1 card */

/*
   If the total number of ports is larger than NR_PORTS, change this
   constant in the definition below. No other change is necessary to
   support more boards/ports. */
#define NR_PORTS        CAB_SOE_PORTS

#define	SERIAL_PARANOIA_CHECK
#undef	CAB_DEBUG_OPEN
#undef	CAB_DEBUG_THROTTLE
#undef	CAB_DEBUG_OTHER
#undef	CAB_DEBUG_IO
#undef	CAB_DEBUG_IOCTL
#undef	CAB_DEBUG_COUNT
#undef	CAB_DEBUG_DTR
#undef	CAB_DEBUG_WAIT_UNTIL_SENT
#undef	CAB_DEBUG_INTERRUPTS
#undef	CAB_DEBUG_NVRAM
#undef	CAB_ENABLE_MONITORING

/*
 * Include section 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/cabernet.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#define	CAB_LOCK(info,flags) \
    do { \
	spin_lock_irqsave(&cab_card[info->card].card_lock, flags); \
    } while (0)
		
#define	CAB_UNLOCK(info,flags) \
    do { \
	spin_unlock_irqrestore(&cab_card[info->card].card_lock, flags); \
    } while (0)

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/version.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include "../../../cmn/util/cab_cpld.h"
#include "../../../cmn/util/cab_fpga.h"
#include "../../../cmn/util/robo.h"
#include "../../../cmn/util/util_arp.h"

#define PHY_TIMEOUT    10000

static void cab_soe_read(ulong base_addr, int port, ulong *data)
{
    CAB_REG_WRITE(base_addr, CAB_SOE_CONTROL, 
		  (CAB_SOE_CONTROL_RX |
		   (port << CAB_SOE_CONTROL_ADDR_SHIFT)));
    do {
        CAB_REG_READ(base_addr, CAB_SOE_CONTROL, *data);
    } while (!(*data & CAB_SOE_CONTROL_RD_DONE));
}

static void cab_soe_write(ulong base_addr, int port, ulong data)
{
#if 0
    ulong status;
    do {
        CAB_REG_READ(base_addr, CAB_SOE_TX_FIFO_NOTFULL, &status);
    } while (!(status & CAB_SOE_PORT(port)));
#endif

    CAB_REG_WRITE(base_addr, CAB_SOE_CONTROL, 
		  ((port << CAB_SOE_CONTROL_ADDR_SHIFT) |
		   (data << CAB_SOE_CONTROL_DATA_SHIFT)));
}

static ulong robo_write(ulong base, ulong addr, unsigned char *data)
{
    ulong error;
    ROBO_REG_WRITE(base, addr, data, error);
    return error;
}

static ulong robo_read(ulong base, ulong addr, unsigned char *data)
{
    ulong error;
    ROBO_REG_READ(base, addr, data, error);
    return error;
}

static ulong robo_write_phy(ulong base, ulong device, ulong reg, ulong val)
{
    unsigned char data[2]; /* all phy regs 2 bytes wide */
    data[0] = val >> 8;
    data[1] = val & 0xff;

    return robo_write(base, ROBO_REG(ROBO_PAGE_MII(device), reg, 2), data);
}

static ulong robo_read_phy(ulong base, ulong device, ulong reg, ulong *val)
{
    unsigned char data[2]; /* all phy regs 2 bytes wide */
    ulong error;
    error = robo_read(base, ROBO_REG(ROBO_PAGE_MII(device), reg, 2), data);
    if (!error) {
      *val = data[0] << 8 | data[1];
    }
    return error;
}

#define cab_put_user	put_user

static unsigned long 
cab_get_user(unsigned long *addr)
{
    unsigned long result = 0;
    int error = get_user (result, addr);
    if (error)
        printk ("cabernet: cab_get_user: error == %d\n", error);
    return result;
}

#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef SERIAL_XMIT_SIZE
#define	SERIAL_XMIT_SIZE	(MIN(PAGE_SIZE, 4096))
#endif
#define WAKEUP_CHARS		256
#define STD_COM_FLAGS (0)

#define	JIFFIES_DIFF(n, j)	((j) - (n))

static DECLARE_TASK_QUEUE(tq_cabernet);

static struct tty_driver cab_driver, cab_soe_driver;
static int serial_refcount;


/* This is the per-card data structure containing address, irq, number of
   channels, etc. This driver supports a maximum of NR_CARDS cards.
*/
static struct cabernet_card cab_card[NR_CARDS];

/* This is the per-channel data structure containing pointers, flags
 and variables for the port. This driver supports a maximum of NR_PORTS.
*/
static struct cabernet_port cab_port[NR_PORTS];
static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the copy_from_user blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.  This buffer is
 * allocated when the first cab_open occurs.
 */
static unsigned char *tmp_buf;
DECLARE_MUTEX(tmp_buf_sem);

static unsigned short	cab_nboard;
static void cab_start(struct tty_struct *);
static void set_line_char(struct cabernet_port *);

static int cabernet_get_proc_info(char *, char **, off_t , int , int *, void *);

static inline int
serial_paranoia_check(struct cabernet_port *info,
                        kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
    static const char *badmagic =
        "mxc: Warning: bad magic number for serial struct (%s) in %s\n";
    static const char *badinfo =
        "mxc: Warning: null cabernet_port for (%s) in %s\n";
    static const char *badrange =
        "mxc: Warning: cabernet_port out of range for (%s) in %s\n";

    if (!info) {
        printk(badinfo, kdevname(device), routine);
        return 1;
    }

    if( (long)info < (long)(&cab_port[0])
    || (long)(&cab_port[NR_PORTS]) < (long)info ){
        printk(badrange, kdevname(device), routine);
        return 1;
    }

    if (info->magic != CABERNET_MAGIC) {
        printk(badmagic, kdevname(device), routine);
        return 1;
    }
#endif
        return 0;
} /* serial_paranoia_check */

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver
 * (also known as the "bottom half").  This can be called any
 * number of times for any channel without harm.
 */
static inline void
cab_sched_event(struct cabernet_port *info, int event)
{
    info->event |= 1 << event; /* remember what kind of event and who */
    queue_task(&info->tqueue, &tq_cabernet); /* it belongs to */
    mark_bh(CYCLADES_BH);                   /* then trigger event */
} /* cab_sched_event */


/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * mx#/_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using cab_sched_event(), and they get done here.
 *
 * This is done through one level of indirection--the task queue.
 * When a hardware interrupt service routine wants service by the
 * driver's bottom half, it enqueues the appropriate tq_struct (one
 * per port) to the tq_cabernet work queue and sets a request flag
 * via mark_bh for processing that queue.  When the time is right,
 * do_cabernet_bh is called (because of the mark_bh) and it requests
 * that the work queue be processed.
 *
 * Although this may seem unwieldy, it gives the system a way to
 * pass an argument (in this case the pointer to the cabernet_port
 * structure) to the bottom half of the driver.  Previous kernels
 * had to poll every port to see if that port needed servicing.
 */
static void
do_cabernet_bh(void)
{
    run_task_queue(&tq_cabernet);
} /* do_cabernet_bh */

static void
do_softint(void *private_)
{
  struct cabernet_port *info = (struct cabernet_port *) private_;
  struct tty_struct    *tty;

    tty = info->tty;
    if (!tty)
        return;

    if (test_and_clear_bit(CAB_EVENT_HANGUP, &info->event)) {
        tty_hangup(info->tty);
        wake_up_interruptible(&info->open_wait);
        info->flags &= ~(ASYNC_NORMAL_ACTIVE|
                             ASYNC_CALLOUT_ACTIVE);
    }
    if (test_and_clear_bit(CAB_EVENT_OPEN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->open_wait);
    }
    if (test_and_clear_bit(CAB_EVENT_WRITE_WAKEUP, &info->event)) {
        if((tty->flags & (1<< TTY_DO_WRITE_WAKEUP))
        && tty->ldisc.write_wakeup){
            (tty->ldisc.write_wakeup)(tty);
        }
        wake_up_interruptible(&tty->write_wait);
    }
#ifdef Z_WAKE
    if (test_and_clear_bit(CAB_EVENT_SHUTDOWN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->shutdown_wait);
    }
#endif
} /* do_softint */


/* The real interrupt service routine is called
   whenever the card wants its hand held--chars
   received, out buffer empty, modem change, etc.
 */
static void
cab_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct cabernet_card *cinfo;
    struct cabernet_port *info;
    struct tty_struct *tty;
    ulong base_addr;
    ulong status;
    ulong data;
    ulong char_count;
    int outch;
    int i;
    ulong chan;

    if ((cinfo = (struct cabernet_card *)dev_id) == 0) {
#ifdef CAB_DEBUG_INTERRUPTS
	printk("cab_interrupt: spurious interrupt %d\n\r", irq);
#endif
        return; /* spurious interrupt */
    }

    /* This loop checks all chips in the card.  Make a note whenever
       _any_ chip had some work to do, as this is considered an
       indication that there will be more to do.  Only when no chip
       has any work does this outermost loop exit.
     */
    base_addr = cinfo->base_addr;
    while (1) {
	CAB_REG_READ(base_addr, CAB_SOE_INTR, status);
	if (!(status & CAB_SOE_INTR_RDY)) break;
#ifdef CAB_DEBUG_INTERRUPTS
	printk("mxc: soe int %lx\n\r", status);
#endif
	chan = (status & CAB_SOE_INTR_PORT_MASK) >> CAB_SOE_INTR_PORT_SHIFT;
	i = chan + cinfo->first_line;
	info = &cab_port[i];
	if (status & CAB_SOE_INTR_RX) { /* RX */
#ifdef CAB_DEBUG_INTERRUPTS
	    printk("mxc: rcvd intr\n\r");
#endif
#if 0
	    if (status & CAB_SOE_INTR_RX_FERR) {
	        /* RX errors */
	        continue;
	    }
#endif
	    /* read char */
	    info->last_active = jiffies;
	    /* load # chars available from the chip */
	    char_count = (status & CAB_SOE_INTR_CNT_MASK) >> CAB_SOE_INTR_CNT_SHIFT;
#ifdef CAB_DEBUG_INTERRUPTS
	    printk("mxc: rcvd intr, chars %ld\n\r", char_count);
#endif

	    /* if there is nowhere to put the data, discard it */
	    if (info->tty == 0) {
	        while (char_count--) {
		    cab_soe_read(base_addr, chan, &data);
		}
	    } else { /* there is an open port for this data */
	        /* normal character reception */
	        tty = info->tty;
#ifdef CAB_ENABLE_MONITORING
		++info->mon.int_count;
		info->mon.char_count += char_count;
		if (char_count > info->mon.char_max)
		    info->mon.char_max = char_count;
		info->mon.char_last = char_count;
#endif
		info->idle_stats.recv_bytes += char_count;
		info->idle_stats.recv_idle   = jiffies;
		while (char_count--) {
		    if (tty->flip.count >= TTY_FLIPBUF_SIZE){
		        break;
		    }
		    cab_soe_read(base_addr, chan, &data);
		    tty->flip.count++;
		    *tty->flip.flag_buf_ptr++ = TTY_NORMAL;
		    *tty->flip.char_buf_ptr++ = data;
		    info->icount.rx++;
		}
		queue_task(&tty->flip.tqueue, &tq_timer);
	    }
	} else if (status & CAB_SOE_INTR_TX_FE) { /* TX interrupt */
	    /* Since we only get here when the transmit buffer
	       is empty, we know we can always stuff 16
	       characters. */
#ifdef CAB_DEBUG_INTERRUPTS
	    printk("mxc: xmit intrn\r");
#endif
	    info->last_active = jiffies;
	    if (info->tty == 0) {
	    disable_txint:
	        CAB_REG_READ(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
		data &= ~CAB_SOE_PORT(chan);
	        CAB_REG_WRITE(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
		goto txdone;
	    }

#ifdef CAB_DEBUG_INTERRUPTS
	    printk("mxc: xmit intr, cnt = %d\n\r", info->xmit_cnt);
#endif
	    /* load the on-chip space for outbound data */
	    char_count = info->xmit_fifo_size;
	    while (char_count-- > 0){
	        if (!info->xmit_cnt){
#ifdef CAB_DEBUG_INTERRUPTS
		    printk("mxc: xmit intr, !cnt\n\r");
#endif
		    goto disable_txint;
		}
		if ((info->xmit_buf == 0) ||
		    info->tty->stopped || info->tty->hw_stopped) {
#ifdef CAB_DEBUG_INTERRUPTS
		    printk("mxc: xmit intr, stopped\n\r");
#endif
		    goto disable_txint;
		}
		outch = info->xmit_buf[info->xmit_tail];
		info->xmit_cnt--;
		info->xmit_tail = ((info->xmit_tail + 1) &
				   (SERIAL_XMIT_SIZE - 1));
		cab_soe_write(base_addr, chan, outch);
		info->icount.tx++;
	    }
	txdone:
	    if (info->xmit_cnt < WAKEUP_CHARS) {
	        cab_sched_event(info, CAB_EVENT_WRITE_WAKEUP);
	    }
	}
    }
} /* cab_interrupt */

/* This is called whenever a port becomes active;
   interrupts are enabled and DTR & RTS are turned on.
 */
static int
startup(struct cabernet_port * info)
{
    int retval = 0;
    ulong base_addr;
    int card,channel;
    unsigned long page;
    int iport, chan;
    unsigned long val;

    card = info->card;
    channel = (info->line) - (cab_card[card].first_line);

    page = get_free_page(GFP_KERNEL);
    if (!page)
	return -ENOMEM;

    if (info->flags & ASYNC_INITIALIZED){
	free_page(page);
	goto errout;
    }

    if (!info->type){
        if (info->tty){
            set_bit(TTY_IO_ERROR, &info->tty->flags);
        }
	free_page(page);
	goto errout;
    }

    if (info->xmit_buf)
	free_page(page);
    else
	info->xmit_buf = (unsigned char *) page;

    set_line_char(info);

    base_addr = cab_card[card].base_addr;

#ifdef CAB_DEBUG_OPEN
    printk("mxc: startup card %d, channel %d, base_addr %lx\n",
	   card, channel, base_addr);
#endif

    iport = channel;
    chan = CAB_SOE_CHAN(iport);

#ifdef CAB_DEBUG_OPEN
    printk(" mxc: startup port %d\n", chan);
#endif

    /* reset tx and rx fifos */
    CAB_REG_WRITE(base_addr, CAB_SOE_RX_FIFO_RESET, CAB_SOE_PORT(chan));
    udelay(2000);
    CAB_REG_WRITE(base_addr, CAB_SOE_RX_FIFO_RESET, 0);
    /* enable inband flow control (Xon/Xoff) */
    CAB_REG_READ(base_addr, CAB_SOE_FC_CONTROL, val);
    val |= CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_FC_CONTROL, val);
    /* catch rx notempty interrupts */
    CAB_REG_READ(base_addr, CAB_SOE_RX_FIFO_NOTEMPTY_MASK,val);
    val |= CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_RX_FIFO_NOTEMPTY_MASK, val);

    info->flags |= ASYNC_INITIALIZED;

    if (info->tty) {
        clear_bit(TTY_IO_ERROR, &info->tty->flags);
    }
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    info->breakon = info->breakoff = 0;
    memset((char *)&info->idle_stats, 0, sizeof(info->idle_stats));
    info->idle_stats.in_use    =
      info->idle_stats.recv_idle =
      info->idle_stats.xmit_idle = jiffies;
    
#ifdef CAB_DEBUG_OPEN
    printk(" mxc: startup done\n");
#endif
    return 0;
	
errout:
    return retval;
} /* startup */


static void
start_xmit( struct cabernet_port *info )
{
    ulong base_addr;
    ulong card,iport,channel,chan,data;

    card = info->card;
    channel = (info->line) - (cab_card[card].first_line);
    iport = channel;
    chan = CAB_SOE_CHAN(iport);
    base_addr = cab_card[card].base_addr;

#ifdef CAB_DEBUG_OTHER
    printk("mxc: start xmit port %ld\n", chan);
#endif

    CAB_REG_READ(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
    data |= CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
} /* start_xmit */

/*
 * This routine shuts down a serial port; interrupts are disabled,
 * and DTR is dropped if the hangup on close termio flag is on.
 */
static void
shutdown(struct cabernet_port * info)
{
    unsigned long base_addr;
    int card,channel,chan,iport;
    ulong data;

    if (!(info->flags & ASYNC_INITIALIZED)){
        return;
    }

    card = info->card;
    channel = info->line - cab_card[card].first_line;
    base_addr = cab_card[card].base_addr;
    iport = channel;
    chan = CAB_SOE_CHAN(iport);

#ifdef CAB_DEBUG_OPEN
    printk("mxc: shutdown card %d, channel %d, base_addr %lx\n",
	   card, channel, (long)base_addr);
#endif

    if (info->xmit_buf){
        unsigned char * temp;
	temp = info->xmit_buf;
	info->xmit_buf = 0;
	free_page((unsigned long) temp);
    }
    /* disable tx empty interrupts */
    CAB_REG_READ(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
    data &= ~CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_TX_FIFO_EMPTY_MASK, data);
    /* disable rx notempty interrupts */
    CAB_REG_READ(base_addr, CAB_SOE_RX_FIFO_NOTEMPTY_MASK, data);
    data &= ~CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_RX_FIFO_NOTEMPTY_MASK, data);
    /* disable inband flow control (Xon/Xoff) */
    CAB_REG_READ(base_addr, CAB_SOE_FC_CONTROL, data);
    data &= ~CAB_SOE_PORT(chan);
    CAB_REG_WRITE(base_addr, CAB_SOE_FC_CONTROL, data);

    if (info->tty){
        set_bit(TTY_IO_ERROR, &info->tty->flags);
    }
    info->flags &= ~ASYNC_INITIALIZED;

#ifdef CAB_DEBUG_OPEN
    printk(" mxc: shutdown done\n");
#endif
    return;
} /* shutdown */


/*
 * ------------------------------------------------------------
 * cab_open() and friends
 * ------------------------------------------------------------
 */

static int
block_til_ready(struct tty_struct *tty, struct file * filp,
                           struct cabernet_port *info)
{
    DECLARE_WAITQUEUE(wait, current);
    struct cabernet_card *cinfo;
    int channel;
    int retval;
    unsigned long base_addr;

    cinfo = &cab_card[info->card];
    channel = info->line - cinfo->first_line;

    /*
     * If the device is in the middle of being closed, then block
     * until it's done, and then try again.
     */
    if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
	if (info->flags & ASYNC_CLOSING) {
            interruptible_sleep_on(&info->close_wait);
	}
        return ((info->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    /*
     * If this is a callout device, then just make sure the normal
     * device isn't being used.
     */
    if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
        if (info->flags & ASYNC_NORMAL_ACTIVE){
            return -EBUSY;
        }
        if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
            (info->flags & ASYNC_SESSION_LOCKOUT) &&
            (info->session != current->session)){
            return -EBUSY;
        }
        if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
            (info->flags & ASYNC_PGRP_LOCKOUT) &&
            (info->pgrp != current->pgrp)){
            return -EBUSY;
        }
        info->flags |= ASYNC_CALLOUT_ACTIVE;
        return 0;
    }

    /*
     * If non-blocking mode is set, then make the check up front
     * and then exit.
     */
    if ((filp->f_flags & O_NONBLOCK) ||
	(tty->flags & (1 << TTY_IO_ERROR))) {
        if (info->flags & ASYNC_CALLOUT_ACTIVE){
            return -EBUSY;
        }
        info->flags |= ASYNC_NORMAL_ACTIVE;
        return 0;
    }

    /*
     * Block waiting for the carrier detect and the line to become
     * free (i.e., not in use by the callout).  While we are in
     * this loop, info->count is dropped by one, so that
     * cab_close() knows when to free things.  We restore it upon
     * exit, either normal or abnormal.
     */
    retval = 0;
    add_wait_queue(&info->open_wait, &wait);
#ifdef CAB_DEBUG_OPEN
    printk("mxc: block_til_ready before block: mxc%d, count = %d\n",
           info->line, info->count);/**/
#endif
    if (!tty_hung_up_p(filp))
	info->count--;
#ifdef CAB_DEBUG_COUNT
    printk("mxc: block_til_ready: (%d): decrementing count to %d\n",
        current->pid, info->count);
#endif
    info->blocked_open++;

    channel &= 0x03;
    base_addr = cinfo->base_addr;

    while (1) {
	set_current_state(TASK_INTERRUPTIBLE);
	if (tty_hung_up_p(filp) ||
	    !(info->flags & ASYNC_INITIALIZED)) {
	    retval = ((info->flags & ASYNC_HUP_NOTIFY) ? 
		      -EAGAIN : -ERESTARTSYS);
	    break;
	}

	if (signal_pending(current)) {
	    retval = -ERESTARTSYS;
	    break;
	}
#ifdef CAB_DEBUG_OPEN
	printk("mxc: block_til_ready blocking: mxc%d, count = %d\n",
	       info->line, info->count);/**/
#endif
	schedule();
    }

    current->state = TASK_RUNNING;
    remove_wait_queue(&info->open_wait, &wait);
    if (!tty_hung_up_p(filp)){
	info->count++;
#ifdef CAB_DEBUG_COUNT
	printk("mxc: block_til_ready (%d): incrementing count to %d\n",
	    current->pid, info->count);
#endif
    }
    info->blocked_open--;
#ifdef CAB_DEBUG_OPEN
    printk("mxc: block_til_ready after blocking: mxc%d, count = %d\n",
	   info->line, info->count);/**/
#endif
    if (retval)
	return retval;
    info->flags |= ASYNC_NORMAL_ACTIVE;
    return 0;
} /* block_til_ready */

/*
 * This routine is called whenever a serial port is opened.  It
 * performs the serial-specific initialization for the tty structure.
 */
static int
cab_open(struct tty_struct *tty, struct file * filp)
{
    struct cabernet_port  *info;
    int retval, line;
    ulong page;
    ulong val;

    MOD_INC_USE_COUNT;
    line = MINOR(tty->device) - tty->driver.minor_start;
    if ((line < 0) || (NR_PORTS <= line)){
	MOD_DEC_USE_COUNT;
        return -ENODEV;
    }
    info = &cab_port[line];
    if (info->line < 0){
	MOD_DEC_USE_COUNT;
        return -ENODEV;
    }

    if ((MAJOR(tty->device) == CABERNET_MAJOR) &&
	!strcmp(tty->driver.name, cab_driver.name)) {
#ifdef CAB_DEBUG_OPEN
        printk("mxc: cab_open mxc %d %s\n", CABERNET_MAJOR, tty->driver.name);
#endif
	tty->driver_data = info;
	return 0;
    }

    tty->driver_data = info;
    info->tty = tty;
    if (serial_paranoia_check(info, tty->device, "cab_open")) {
        return -ENODEV;
    }

    /* enable SOE interrupts on CAB */
    CAB_REG_READ(cab_card[info->card].base_addr, CAB_INTR_MASK, val);
    val |= (CAB_INTR_SOE_TX_EMPTY |
	    CAB_INTR_SOE_RX_NOTEMPTY |
	    0 /*CAB_INTR_SOE_RX_ERR*/);
    CAB_REG_WRITE(cab_card[info->card].base_addr, CAB_INTR_MASK, val);

#ifdef CAB_DEBUG_OPEN
    printk("mxc: cab_open mxc%d, count = %d\n",
	   info->line, info->count);/**/
#endif
    info->count++;
#ifdef CAB_DEBUG_COUNT
    printk("mxc: cab_open (%d): incrementing count to %d\n",
	   current->pid, info->count);
#endif
    if (!tmp_buf) {
	page = get_free_page(GFP_KERNEL);
	if (!page)
	    return -ENOMEM;
	if (tmp_buf)
	    free_page(page);
	else
	    tmp_buf = (unsigned char *) page;
    }

    /*
     * If the port is the middle of closing, bail out now
     */
    if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
	if (info->flags & ASYNC_CLOSING)
	    interruptible_sleep_on(&info->close_wait);
	return ((info->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    /*
     * Start up serial port
     */
    retval = startup(info);
    if (retval){
        return retval;
    }

    retval = block_til_ready(tty, filp, info);
    if (retval) {
#ifdef CAB_DEBUG_OPEN
        printk("mxc: cab_open returning after block_til_ready with %d\n",
               retval);
#endif
        return retval;
    }

    if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
        *tty->termios = info->normal_termios;
    }

    info->session = current->session;
    info->pgrp = current->pgrp;

#ifdef CAB_DEBUG_OPEN
    printk(" mxc: cab_open done\n");/**/
#endif

    return 0;
} /* cab_open */


/*
 * cab_wait_until_sent() --- wait until the transmitter is empty
 */
static void 
cab_wait_until_sent(struct tty_struct *tty, int timeout)
{
    struct cabernet_port * info = (struct cabernet_port *)tty->driver_data;
    unsigned long orig_jiffies, char_time;
	
    if (serial_paranoia_check(info, tty->device, "cab_wait_until_sent"))
	return;

    if (info->xmit_fifo_size == 0)
	return; /* Just in case.... */


    orig_jiffies = jiffies;
    /*
     * Set the check interval to be 1/5 of the estimated time to
     * send a single character, and make it at least 1.  The check
     * interval should also be less than the timeout.
     * 
     * Note: we have to use pretty tight timings here to satisfy
     * the NIST-PCTS.
     */
    char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
    char_time = char_time / 5;
    if (char_time == 0)
	char_time = 1;
    if (timeout < 0)
	timeout = 0;
    if (timeout)
	char_time = MIN(char_time, timeout);
    /*
     * If the transmitter hasn't cleared in twice the approximate
     * amount of time to send the entire FIFO, it probably won't
     * ever clear.  This assumes the UART isn't doing flow
     * control, which is currently the case.  Hence, if it ever
     * takes longer than info->timeout, this is probably due to a
     * UART bug of some kind.  So, we clamp the timeout parameter at
     * 2*info->timeout.
     */
    if (!timeout || timeout > 2*info->timeout)
	timeout = 2*info->timeout;
#ifdef CAB_DEBUG_WAIT_UNTIL_SENT
    printk("In cab_wait_until_sent(%d) check=%lu...", timeout, char_time);
    printk("jiff=%lu...", jiffies);
#endif

    /* Run one more char cycle */
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(char_time * 5);
    current->state = TASK_RUNNING;
#ifdef CAB_DEBUG_WAIT_UNTIL_SENT
    printk("Clean (jiff=%lu)...done\n", jiffies);
#endif
}

/*
 * This routine is called when a particular tty device is closed.
 */
static void
cab_close(struct tty_struct *tty, struct file *filp)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;

#ifdef CAB_DEBUG_OTHER
    printk("mxc: cab_close mxc%d\n", info->line);
#endif

    if ((MAJOR(tty->device) == CABERNET_MAJOR) &&
	!strcmp(tty->driver.name, cab_driver.name)) {
	MOD_DEC_USE_COUNT;
        return;
    }

    if (!info || serial_paranoia_check(info, tty->device, "cab_close")){
        return;
    }

    /* If the TTY is being hung up, nothing to do */
    if (tty_hung_up_p(filp)) {
	MOD_DEC_USE_COUNT;
        return;
    }
        
#ifdef CAB_DEBUG_OPEN
    printk("mxc: cab_close mxc%d, count = %d\n", info->line, info->count);
#endif
    if ((tty->count == 1) && (info->count != 1)) {
        /*
         * Uh, oh.  tty->count is 1, which means that the tty
         * structure will be freed.  Info->count should always
         * be one in these conditions.  If it's greater than
         * one, we've got real problems, since it means the
         * serial port won't be shutdown.
         */
        printk("mxc: cab_close: bad serial port count; tty->count is 1, "
           "info->count is %d\n", info->count);
        info->count = 1;
    }
#ifdef CAB_DEBUG_COUNT
    printk("mxc: cab_close at (%d): decrementing count to %d\n",
        current->pid, info->count - 1);
#endif
    if (--info->count < 0) {
#ifdef CAB_DEBUG_COUNT
        printk("mxc: mxc_close setting count to 0\n");
#endif
        info->count = 0;
    }
    if (info->count) {
	MOD_DEC_USE_COUNT;
        return;
    }
    info->flags |= ASYNC_CLOSING;
    /*
     * Save the termios structure, since this port may have
     * separate termios for callout and dialin.
     */
    if (info->flags & ASYNC_NORMAL_ACTIVE)
        info->normal_termios = *tty->termios;

    /*
    * Now we wait for the transmit buffer to clear; and we notify
    * the line discipline to only process XON/XOFF characters.
    */
    tty->closing = 1;
    if (info->closing_wait != CAB_CLOSING_WAIT_NONE) {
	tty_wait_until_sent(tty, info->closing_wait);
    }
    {
	if (info->flags & ASYNC_INITIALIZED) {
	    /* Waiting for on-board buffers to be empty before closing 
	       the port */
	    cab_wait_until_sent(tty, info->timeout);
	}
    }

    shutdown(info);
    if (tty->driver.flush_buffer)
        tty->driver.flush_buffer(tty);
    if (tty->ldisc.flush_buffer)
        tty->ldisc.flush_buffer(tty);

    tty->closing = 0;
    info->event = 0;
    info->tty = 0;
    if (info->blocked_open) {
        if (info->close_delay) {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(info->close_delay);
        }
        wake_up_interruptible(&info->open_wait);
    }
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
                     ASYNC_CLOSING);
    wake_up_interruptible(&info->close_wait);

#ifdef CAB_DEBUG_OTHER
    printk(" mxc: cab_close done\n");
#endif

    MOD_DEC_USE_COUNT;
    return;
} /* cab_close */


/* This routine gets called when tty_write has put something into
 * the write_queue.  The characters may come from user space or
 * kernel space.
 *
 * This routine will return the number of characters actually
 * accepted for writing.
 *
 * If the port is not already transmitting stuff, start it off by
 * enabling interrupts.  The interrupt service routine will then
 * ensure that the characters are sent.
 * If the port is already active, there is no need to kick it.
 *
 */
static int
cab_write(struct tty_struct * tty, int from_user,
	 const unsigned char *buf, int count)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;
    int c, ret = 0;

#ifdef CAB_DEBUG_IO
    printk("mxc: cab_write mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "cab_write")){
        return 0;
    }
        
    if (!tty || !info->xmit_buf || !tmp_buf){
        return 0;
    }

    if (from_user) {
	down(&tmp_buf_sem);
	while (1) {
	    int c1;
	    
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
				SERIAL_XMIT_SIZE - info->xmit_head));
	    if (c <= 0)
		break;

	    c -= copy_from_user(tmp_buf, buf, c);
	    if (!c) {
		if (!ret) {
		    ret = -EFAULT;
		}
		break;
	    }
	    c1 = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			SERIAL_XMIT_SIZE - info->xmit_head));
			
	    if (c1 < c)
	    	c = c1;
	    memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
	    info->xmit_head = ((info->xmit_head + c) & (SERIAL_XMIT_SIZE-1));
	    info->xmit_cnt += c;
	    buf += c;
	    count -= c;
	    ret += c;
	}
	up(&tmp_buf_sem);
    } else {
	while (1) {
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1, 
			SERIAL_XMIT_SIZE - info->xmit_head));
	        
	    if (c <= 0)
		break;

	    memcpy(info->xmit_buf + info->xmit_head, buf, c);
	    info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE-1);
	    info->xmit_cnt += c;
	    buf += c;
	    count -= c;
	    ret += c;
	}
    }

    info->idle_stats.xmit_bytes += ret;
    info->idle_stats.xmit_idle   = jiffies;

    if (info->xmit_cnt && !tty->stopped && !tty->hw_stopped) {
        start_xmit(info);
    }
    return ret;
} /* cab_write */


/*
 * This routine is called by the kernel to write a single
 * character to the tty device.  If the kernel uses this routine,
 * it must call the flush_chars() routine (if defined) when it is
 * done stuffing characters into the driver.  If there is no room
 * in the queue, the character is ignored.
 */
static void
cab_put_char(struct tty_struct *tty, unsigned char ch)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;

#ifdef CAB_DEBUG_IO
    printk("mxc: cab_put_char mxc%d\n", info->line);
#endif

    if (serial_paranoia_check(info, tty->device, "cab_put_char"))
        return;

    if (!tty || !info->xmit_buf)
        return;

    if (info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) {
	return;
    }

    info->xmit_buf[info->xmit_head++] = ch;
    info->xmit_head &= SERIAL_XMIT_SIZE - 1;
    info->xmit_cnt++;
    info->idle_stats.xmit_bytes++;
    info->idle_stats.xmit_idle = jiffies;
} /* cab_put_char */


/*
 * This routine is called by the kernel after it has written a
 * series of characters to the tty device using put_char().  
 */
static void
cab_flush_chars(struct tty_struct *tty)
{
  struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;
                                
#ifdef CAB_DEBUG_IO
    printk("mxc: cab_flush_chars mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "cab_flush_chars"))
        return;

    if (info->xmit_cnt <= 0 || tty->stopped||
	tty->hw_stopped || !info->xmit_buf)
        return;

    start_xmit(info);
} /* cab_flush_chars */


/*
 * This routine returns the numbers of characters the tty driver
 * will accept for queuing to be written.  This number is subject
 * to change as output buffers get emptied, or if the output flow
 * control is activated.
 */
static int
cab_write_room(struct tty_struct *tty)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;
    int   ret;
                                
#ifdef CAB_DEBUG_IO
    printk("mxc: cab_write_room mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "cab_write_room"))
        return 0;

    ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
    if (ret < 0)
        ret = 0;

    return ret;
} /* cab_write_room */


static int
cab_chars_in_buffer(struct tty_struct *tty)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;
    int card, channel;
                                
    if (serial_paranoia_check(info, tty->device, "cab_chars_in_buffer"))
        return 0;

    card = info->card;
    channel = (info->line) - (cab_card[card].first_line);


#ifdef CAB_DEBUG_IO
    printk("mxc: cab_chars_in_buffer mxc%d %d\n",
	   info->line, info->xmit_cnt); /* */
#endif
    return info->xmit_cnt;
} /* cab_chars_in_buffer */


/*
 * ------------------------------------------------------------
 * cab_ioctl() and friends
 * ------------------------------------------------------------
 */

/*
 * This routine finds or computes the various line characteristics.
 * It used to be called config_setup
 */
static void
set_line_char(struct cabernet_port * info)
{
    unsigned long base_addr;
    int card,channel,iport,chan;

    if (!info->tty || !info->tty->termios){
        return;
    }
    if (info->line == -1){
        return;
    }
    card = info->card;
    base_addr = cab_card[card].base_addr;
    channel = (info->line) - (cab_card[card].first_line);
    iport = channel;
    chan = CAB_SOE_CHAN(iport);

    /* XXX hard-coded to 9600 8N1 */

} /* set_line_char */


static int
get_serial_info(struct cabernet_port * info,
		struct serial_struct * retinfo)
{
  struct serial_struct tmp;
  struct cabernet_card *cinfo = &cab_card[info->card];

    if (!retinfo)
            return -EFAULT;
    memset(&tmp, 0, sizeof(tmp));
    tmp.type = info->type;
    tmp.line = info->line;
    tmp.port = info->card * 0x100 + info->line - cinfo->first_line;
    tmp.irq = cinfo->irq;
    tmp.flags = info->flags;
    tmp.close_delay = info->close_delay;
    tmp.baud_base = info->baud;
    tmp.custom_divisor = info->custom_divisor;
    tmp.hub6 = 0;               /*!!!*/
    return copy_to_user(retinfo,&tmp,sizeof(*retinfo))?-EFAULT:0;
} /* get_serial_info */


static int
set_serial_info(struct cabernet_port * info,
		struct serial_struct * new_info)
{
   struct serial_struct new_serial;
   struct cabernet_port old_info;

    if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
	return -EFAULT;
    old_info = *info;

    if (!capable(CAP_SYS_ADMIN)) {
            if ((new_serial.close_delay != info->close_delay) ||
		(new_serial.baud_base != info->baud) ||
		((new_serial.flags & ASYNC_FLAGS & ~ASYNC_USR_MASK) !=
		 (info->flags & ASYNC_FLAGS & ~ASYNC_USR_MASK)))
                    return -EPERM;
            info->flags = ((info->flags & ~ASYNC_USR_MASK) |
                           (new_serial.flags & ASYNC_USR_MASK));
            info->baud = new_serial.baud_base;
	    info->custom_divisor = new_serial.custom_divisor;
            goto check_and_exit;
    }


    /*
     * OK, past this point, all the error checking has been done.
     * At this point, we start making changes.....
     */

    info->baud = new_serial.baud_base;
    info->custom_divisor = new_serial.custom_divisor;
    info->flags = ((info->flags & ~ASYNC_FLAGS) |
                    (new_serial.flags & ASYNC_FLAGS));
    info->close_delay = new_serial.close_delay * HZ/100;
    info->closing_wait = new_serial.closing_wait * HZ/100;

check_and_exit:
    if (info->flags & ASYNC_INITIALIZED){
        set_line_char(info);
        return 0;
    }else{
        return startup(info);
    }
} /* set_serial_info */


/*
 * cab_break() --- routine which turns the break handling on or off
 */
static void
cab_break(struct tty_struct *tty, int break_state)
{
    struct cabernet_port * info = (struct cabernet_port *)tty->driver_data;

    if (serial_paranoia_check(info, tty->device, "cab_break"))
	return;

        /* Let the transmit ISR take care of this (since it
	   requires stuffing characters into the output stream).
        */
	if (break_state == -1) {
	    if (!info->breakon) {
		info->breakon = 1;
		if (!info->xmit_cnt) {
		    start_xmit(info);
		}
	    }
	} else {
	    if (!info->breakoff) {
		info->breakoff = 1;
		if (!info->xmit_cnt) {
		    start_xmit(info);
		}
	    }
	}
} /* cab_break */

static int
get_mon_info(struct cabernet_port * info, struct cabernet_monitor * mon)
{

    if(copy_to_user(mon, &info->mon, sizeof(struct cabernet_monitor)))
    	return -EFAULT;
    info->mon.int_count  = 0;
    info->mon.char_count = 0;
    info->mon.char_max   = 0;
    info->mon.char_last  = 0;
    return 0;
}/* get_mon_info */

static int
cab_cab_read(struct cabernet_port * info,
	    cab_ioctl_t *u)
{
    cab_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    CAB_REG_READ(cab_card[info->card].base_addr, k.reg, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
cab_cab_write(struct cabernet_port * info,
	     cab_ioctl_t *u)
{
    cab_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    CAB_REG_WRITE(cab_card[info->card].base_addr, k.reg, k.data);
    return 0;
}

static int
cab_robo_read(struct cabernet_port * info,
	     robo_ioctl_t *u)
{
    robo_ioctl_t k;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (robo_read(cab_card[info->card].base_addr, k.reg, k.data))
        rc = -ENODEV;
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
cab_robo_write(struct cabernet_port * info,
	     robo_ioctl_t *u)
{
    robo_ioctl_t k;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (robo_write(cab_card[info->card].base_addr, k.reg, k.data))
        rc = -ENODEV;
    return rc;
}

static int
cab_phy_read(struct cabernet_port * info,
	    robo_phy_ioctl_t *u)
{
    robo_phy_ioctl_t k;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (robo_read_phy(cab_card[info->card].base_addr,
		     k.dev, k.reg, &k.data))
        rc = -ENODEV;
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
cab_phy_write(struct cabernet_port * info,
	     robo_phy_ioctl_t *u)
{
    robo_phy_ioctl_t k;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (robo_write_phy(cab_card[info->card].base_addr,
		      k.dev, k.reg, k.data))
        rc = -ENODEV;
    return rc;
}

static int
cab_cpld_read(struct cabernet_port * info,
	     cab_cpld_ioctl_t *u)
{
    cab_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    CAB_CPLD_REG_READ(cab_card[info->card].ctl_addr, k.reg, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
cab_cpld_write(struct cabernet_port * info,
	      cab_cpld_ioctl_t *u)
{
    cab_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, k.reg, k.data);
    return 0;
}

#define JTAG_WRITE(val) { \
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_JTAG_OPERATION, \
		      val); \
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_JTAG_OPERATION, \
		      val | CAB_CPLD_JTAG_OPERATION_CLK); \
}

#define JTAG_READ(val, rc) { \
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_JTAG_OPERATION, \
		      val | CAB_CPLD_JTAG_OPERATION_CLK); \
    CAB_CPLD_REG_READ(cab_card[info->card].ctl_addr, CAB_CPLD_JTAG_OPERATION, \
		     rc); \
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_JTAG_OPERATION, \
		      val); \
}

static int
cab_cab_prog(struct cabernet_port * info, int dev, __u8 *data, int len)
{
    __u8 *ptr;
    ulong val;
    char buf[128];
    int bit;
    int i;
    int rc = 0;

    /* enable JTAG programming */
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_FPGA_INTERFACE,
		      dev | CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 5; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Enter Run-Test-Idle */
    JTAG_WRITE(0);
    /* 3. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 4. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    /* 5. Load CFG-IN */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
    }
    /* 6. Load last bit of CFG-IN */
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    /* 7. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 8. Enter SHIFT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }

    /* 9. Send data */
    ptr = buf;
    for (i = 0; i < len-1; i++, ptr++, data++) {
        if ((i % sizeof(buf)) == 0) {
	    if (copy_from_user(buf, data, MIN(sizeof(buf), (len-i)))) {
	        rc = -EFAULT;
		goto error;
	    }
	    ptr = buf;
	}
        for (bit = 7; bit >= 0; bit--) {
	    val = ((*ptr >> bit) & 0x1) << 1;
	    JTAG_WRITE(val);
	}
    }
    /* 10. Send last byte and set TMS for last bit */
    if ((i % sizeof(buf)) == 0) {
        if (copy_from_user(buf, data, MIN(sizeof(buf), (len-i)))) {
	    rc = -EFAULT;
	    goto error;
	}
	ptr = buf;
    }
    for (bit = 7; bit > 0; bit--) {
        val = ((*ptr >> bit) & 0x1) << 1;
	JTAG_WRITE(val);
    }
    /* last bit */
    val = (*ptr & 0x1) << 1;
    val |= CAB_CPLD_JTAG_OPERATION_TMS;
    JTAG_WRITE(val);

    /* 11. Enter UPDATE-DR */
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    /* 12. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 13. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    /* 14. Load JSTART */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
    }
    /* 15. Load last bit of JSTART */
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    /* 16. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 17. Enter SHIFT-DR and clock STARTUP */
    for (i = 0; i < 14; i++) {
        JTAG_WRITE(0);
    }
    /* 18. Enter UPDATE-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* Enter Run-Test-Idle */
    JTAG_WRITE(0);
    
    /* disable JTAG programming */
 error:
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr,
		      CAB_CPLD_FPGA_INTERFACE, 0);
    return rc;
}

static int
cab_fpga_erase(struct cabernet_port * info)
{
    ulong status;
    int i;
    
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_FPGA_INTERFACE,
		      CAB_CPLD_FPGA_INTERFACE_PRO);
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_FPGA_INTERFACE,
		       CAB_CPLD_FPGA_INTERFACE_PRO|
		       CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    udelay(1); /* wait at least one us */
    CAB_CPLD_REG_WRITE(cab_card[info->card].ctl_addr, CAB_CPLD_FPGA_INTERFACE,
		       CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    i = 0;
    while (1) {
        udelay(1); /* wait at least one us */
        CAB_CPLD_REG_READ(cab_card[info->card].ctl_addr, CAB_CPLD_FPGA_INTERFACE,
			 status);
	if (status & CAB_CPLD_FPGA_INTERFACE_INIT_L) break;
	if (i++ > 100000) {
	    return 1;
	}
    }
    return 0;
}

static int
cab_fpga_prog(struct cabernet_port * info,
	     fpga_ioctl_t *u)
{
    fpga_ioctl_t k;
    int rc = 0;
    int dev = 0;

    if (copy_from_user(&k, u, sizeof(k)))
	return -EFAULT;

    /* erase FPGA */
    if (cab_fpga_erase(info)) 
        return -EFAULT;
    /* program FPGA */
    if (cab_cab_prog(info, dev, (__u8 *)k.addr, k.len))
        rc = -EFAULT;

    return rc;
}

static int
cab_arp_add(struct cabernet_port * info, arp_entry_t *u)
{   
    return (arp_entry_add(u));
}

static int
cab_arp_del(struct cabernet_port * info, arp_entry_t *u)
{   
    return (arp_entry_del(u));
}

static int
cab_arp_clear_all(struct cabernet_port * info, arp_entry_t *u)
{   
    return (arp_clear_all());
}

static int
cab_ap_led_read(struct cabernet_port * info,
		mxc_apled_t *u)
{
    mxc_apled_t k;
    ulong data;
    ulong subport;
    ulong reg;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.port > CAB_SOE_PORTS) 
        return -ENODEV;
    reg = CAB_LED_AP;
    CAB_REG_READ(cab_card[info->card].base_addr, reg, data);
    subport = k.port & 0x7; /* unneeded */
    k.color = (data & CAB_LED_AP_COLOR_MASK(subport)) >> CAB_LED_AP_COLOR_SHIFT(subport);
    k.state = (data & CAB_LED_AP_STATE_MASK(subport)) >> CAB_LED_AP_STATE_SHIFT(subport);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
cab_ap_led_write(struct cabernet_port * info,
		 mxc_apled_t *u)
{
    mxc_apled_t k;
    ulong data;
    ulong subport;
    ulong reg;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if ((k.port >= CAB_SOE_PORTS) ||
	(k.color > MXC_AP_LED_COLOR_YELLOW) ||
	(k.state > MXC_AP_LED_STATE_BLINK))
        return -ENODEV;
    reg = CAB_LED_AP;
    CAB_REG_READ(cab_card[info->card].base_addr, reg, data);
    subport = k.port & 0x7; /* unneeded */
    data &= ~(CAB_LED_AP_COLOR_MASK(subport)|CAB_LED_AP_STATE_MASK(subport));
    data |= ((k.color << CAB_LED_AP_COLOR_SHIFT(subport)) |
	     (k.state << CAB_LED_AP_STATE_SHIFT(subport)));
    CAB_REG_WRITE(cab_card[info->card].base_addr, reg, data);
    return rc;
}

static int
cab_sw_version(struct cabernet_port * info,
	       mxc_string_t *u)
{
    mxc_string_t k;
    if (strlen(ARUBA_VERSION)+1 >= sizeof(k.string))
        return -EFAULT;
    sprintf(k.string, "%s", ARUBA_VERSION);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_sc_id(struct cabernet_port * info,
	  mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    CAB_REG_READ(cab_card[info->card].base_addr, CAB_ID, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_lc_id(struct cabernet_port * info,
	  mxc_lc_t *u)
{
    mxc_lc_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    /* use sc id */
    CAB_REG_READ(cab_card[info->card].base_addr, CAB_ID, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_lc_switch_rev(struct cabernet_port * info,
                  mxc_lc_t *u)
{
    mxc_lc_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    if (robo_read(cab_card[info->card].base_addr, ROBO_REG_REVISION_ID,
                  (unsigned char *)&k.data))
        return -ENODEV;
    k.data >>= 24;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_lc_status(struct cabernet_port * info,
	      mxc_lc_t *u)
{
    mxc_lc_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    k.data = (MXC_LC_PRES |
	      MXC_LC_SPOE_PRES |
	      MXC_LC_SC_PRES(0) |
	      MXC_LC_SC_ACTIVE(0));
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_fan_status(struct cabernet_port * info,
	       mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    int i;
    k.data = 0;
    for (i = 0; i < MXC_MAX_FANS; i++) {
       k.data |= MXC_FAN_PRES(i) | MXC_FAN_OK(i);
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_ps_status(struct cabernet_port * info,
	      mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    int i;
    k.data = 0;
    for (i = 0; i < MXC_MAX_PSS; i++) {
        k.data |= MXC_PS_PRES(i) | MXC_PS_OK(i);
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_temperature(struct cabernet_port * info,
		mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0; /* XXX i2c */
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_sc_slot(struct cabernet_port * info,
	    mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_wd_reset(struct cabernet_port * info)
{
    /* XXX i2c */
    return 0;
}

#ifdef CAB_DEBUG_NVRAM
static void 
cab_print_sector(volatile char *buffer)
{
	int i;
	for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) {
		if((i%16)==0) printk("\n0x%08x: ", (unsigned int)&buffer[i]);
		printk("%02x ", (unsigned char)buffer[i]);
	}
	printk("\n");
}
#endif

static int
cab_nvram_read(struct cabernet_port * info,
	       mxc_nvram_t *u)
{
    mxc_nvram_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset + k.length > NVRAM_SIZE)
	return -EFAULT;
    if (copy_to_user(k.data, (char *)(NVRAM_BASE_ADDRESS + k.offset), k.length))
        return -EFAULT;
    return 0;
}

void
cab_flash_write(char *flash_addr, char *mem_addr, int len)
{
    int i;
    char buffer[MXC_FLASH_SECTOR_SIZE];
    volatile char *flash_sect = (char *)((long)flash_addr & ~(MXC_FLASH_SECTOR_SIZE-1));
    volatile char *flash_base = (char *)MXC_FLASH_BASE_ADDR;
    int copy_len;

    do {
	copy_len = MIN(len, MXC_FLASH_SECTOR_SIZE - (flash_addr - flash_sect));
	if (copy_len < MXC_FLASH_SECTOR_SIZE) {
	    /* read entire flash sector if we are only updating some of it */
	    for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) buffer[i]=flash_sect[i];
	}

#ifdef CAB_DEBUG_NVRAM
	 printk("READ VALUES:\n"); cab_print_sector(flash_sect);
#endif
        /* update buffer with write data */
	memcpy(&buffer[flash_addr - flash_sect], mem_addr, copy_len);

	/* tell flash we want to program one sector */
	flash_base[0x5555] = 0xaa;
	flash_base[0x2aaa] = 0x55;
	flash_base[0x5555] = 0xa0;
        /* program the sector */
	for(i=0;i<MXC_FLASH_SECTOR_SIZE;i++) {
		flash_sect[i]=buffer[i];
	}
	/* wait for the Write Cycle time before moving to next sector */
	udelay(20000);

#ifdef CAB_DEBUG_NVRAM
	printk("AFTER WRITE: VALUES ARE:\n"); cab_print_sector(flash_sect);
#endif

        /* advance ptrs for additional sectors if any */
	len -= copy_len;
	flash_addr += copy_len;
	flash_sect += MXC_FLASH_SECTOR_SIZE;
	mem_addr += copy_len;
    } while (len);
}

static int
cab_nvram_write(struct cabernet_port * info,
		mxc_nvram_t *u)
{
    mxc_nvram_t k;
    char buf[256];
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset + k.length > NVRAM_SIZE)
	return -EFAULT;
    if (k.length > sizeof(buf))
	return -EFAULT;
    if (copy_from_user(buf, k.data, k.length))
	return -EFAULT;
    cab_flash_write((char *)NVRAM_BASE_ADDRESS + k.offset,
		    buf,
		    k.length);
    return 0;
}

static int
cab_panic_read(struct cabernet_port * info,
	      mxc_panic_t *u)
{
    mxc_panic_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    if (copy_to_user(k.record,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
	   NVRAM_PANIC_SIZE))
        return -EFAULT;
    return 0;
}

static int
cab_panic_clear(struct cabernet_port * info,
	       mxc_panic_t *u)
{
    mxc_panic_t k;
    char buf[NVRAM_PANIC_SIZE];
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    memset(buf, 0, NVRAM_PANIC_SIZE);
    cab_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
		    buf,
		    NVRAM_PANIC_SIZE);
    return 0;
}

static int
cab_reboot_str_read(struct cabernet_port * info,
		   mxc_string_t *u)
{
    mxc_string_t k;
    memcpy(k.string,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
	   NVRAM_REBOOT_STRING_SIZE);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
cab_reboot_str_write(struct cabernet_port * info,
		    mxc_string_t *u)
{
    mxc_string_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    cab_flash_write((char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
		    k.string,
		    NVRAM_REBOOT_STRING_SIZE);
    return 0;
}

static int
cab_scratch_read(struct cabernet_port * info,
		 mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    CAB_REG_READ(cab_card[info->card].base_addr, CAB_SCRATCHPAD, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
cab_scratch_write(struct cabernet_port * info,
		  mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    CAB_REG_WRITE(cab_card[info->card].base_addr, CAB_SCRATCHPAD, k.data);
    return 0;
}

/*
 * This routine allows the tty driver to implement device-
 * specific ioctl's.  If the ioctl number passed in cmd is
 * not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int
cab_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
    struct cabernet_port * info = (struct cabernet_port *)tty->driver_data;
    int ret_val = 0;

    if (serial_paranoia_check(info, tty->device, "cab_ioctl"))
	return -ENODEV;

#ifdef CAB_DEBUG_IOCTL
    printk("mxc: cab_ioctl mxc%d, cmd = %x arg = %lx\n",
	   info->line, cmd, arg); /* */
#endif

    switch (cmd) {
    case MXGETMON:
        ret_val = get_mon_info(info, (struct cabernet_monitor *)arg);
	break;
    case MXGETCARDINFO:
        if (copy_to_user((void *)arg, (void *)&cab_card[info->card], 
			 sizeof (struct cabernet_card))) {
	    ret_val = -EFAULT;
	    break;
	}
	ret_val = 0;
	break;
    case MXSETWAIT:
        info->closing_wait = (unsigned short)arg * HZ/100;
	ret_val = 0;
	break;
    case MXGETWAIT:
        ret_val = info->closing_wait / (HZ/100);
	break;
    case MXCABREAD:
        ret_val = cab_cab_read(info, (cab_ioctl_t *) arg);
        break;
    case MXCABWRITE:
        ret_val = cab_cab_write(info, (cab_ioctl_t *) arg);
        break;
    case MXROBOREAD:
        ret_val = cab_robo_read(info, (robo_ioctl_t *) arg);
        break;
    case MXROBOWRITE:
        ret_val = cab_robo_write(info, (robo_ioctl_t *) arg);
        break;
    case MXPHYREAD:
        ret_val = cab_phy_read(info, (robo_phy_ioctl_t *) arg);
        break;
    case MXPHYWRITE:
        ret_val = cab_phy_write(info, (robo_phy_ioctl_t *) arg);
        break;
    case MXCPLDREAD:
        ret_val = cab_cpld_read(info, (cab_cpld_ioctl_t *) arg);
        break;
    case MXCPLDWRITE:
        ret_val = cab_cpld_write(info, (cab_cpld_ioctl_t *) arg);
        break;
    case MXFPGAPROG:
        ret_val = cab_fpga_prog(info, (fpga_ioctl_t *) arg);
        break;

	/* arp ioctls */
    case MXARPENTRYADD:
        ret_val = cab_arp_add(info, (arp_entry_t *) arg);
        break;
    case MXARPENTRYDEL:
        ret_val = cab_arp_del(info, (arp_entry_t *) arg);
        break;
    case MXARPCLEARALL:
        ret_val = cab_arp_clear_all(info, (arp_entry_t *) arg);
        break;

	/* AP led ioctls */
    case MXAPLEDREAD:
        ret_val = cab_ap_led_read(info, (mxc_apled_t *) arg);
	break;
    case MXAPLEDWRITE:
        ret_val = cab_ap_led_write(info, (mxc_apled_t *) arg);
	break;
 
	/* status ioctls */
    case MXSWVERSION:
        ret_val = cab_sw_version(info, (mxc_string_t *) arg);
	break;
    case MXSCID:
        ret_val = cab_sc_id(info, (mxc_ioctl_t *) arg);
	break;
    case MXFANSTATUS:
        ret_val = cab_fan_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXPSSTATUS:
        ret_val = cab_ps_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXTEMPERATURE:
        ret_val = cab_temperature(info, (mxc_ioctl_t *) arg);
	break;
    case MXSCSLOT:
        ret_val = cab_sc_slot(info, (mxc_ioctl_t *) arg);
	break;
    case MXLCID:
        ret_val = cab_lc_id(info, (mxc_lc_t *) arg);
	break;
    case MXLCSWITCHREV:
        ret_val = cab_lc_switch_rev(info, (mxc_lc_t *) arg);
	break;
    case MXLCSTATUS:
        ret_val = cab_lc_status(info, (mxc_lc_t *) arg);
	break;

	/* reboot ioctls */
    case MXWDRESET:
        ret_val = cab_wd_reset(info);
        break;
    case MXPANIC:
        panic("test panic");
        break;
    case MXNULLPTR:
        *(char *)0x0 = 1;
	printk("NULLPTR ptr val = 0x%x\n", *(char *)0x0);
        break;
    case MXHANG:
        while(1);
        break;

	/* "NVRAM" access ioctls */
    case MXNVRAMREAD:
        ret_val = cab_nvram_read(info, (mxc_nvram_t *) arg);
        break;
    case MXNVRAMWRITE:
        ret_val = cab_nvram_write(info, (mxc_nvram_t *) arg);
        break;
    case MXPANICREAD:
        ret_val = cab_panic_read(info, (mxc_panic_t *) arg);
        break;
    case MXPANICCLEAR:
        ret_val = cab_panic_clear(info, (mxc_panic_t *) arg);
        break;
    case MXREBOOTSTRREAD:
        ret_val = cab_reboot_str_read(info, (mxc_string_t *) arg);
        break;
    case MXREBOOTSTRWRITE:
        ret_val = cab_reboot_str_write(info, (mxc_string_t *) arg);
        break;

	/* scratch ioctls */
    case MXSCRATCHREAD:
        ret_val = cab_scratch_read(info, (mxc_ioctl_t *) arg);
        break;
    case MXSCRATCHWRITE:
        ret_val = cab_scratch_write(info, (mxc_ioctl_t *) arg);
	break;

	/* serial ioctls */
    case TIOCGSERIAL:
        ret_val = get_serial_info(info, (struct serial_struct *) arg);
	break;
    case TIOCSSERIAL:
        ret_val = set_serial_info(info, (struct serial_struct *) arg);
	break;
    default:
        ret_val = -ENOIOCTLCMD;
    }

#ifdef CAB_DEBUG_IOCTL
    printk(" mxc: cab_ioctl done %x\n", ret_val);
#endif

    return ret_val;
} /* cab_ioctl */


/*
 * This routine allows the tty driver to be notified when
 * device's termios settings have changed.  Note that a
 * well-designed tty driver should be prepared to accept the case
 * where old == NULL, and try to do something rational.
 */
static void
cab_set_termios(struct tty_struct *tty, struct termios * old_termios)
{
    struct cabernet_port *info = (struct cabernet_port *)tty->driver_data;

#ifdef CAB_DEBUG_OTHER
    printk("mxc: cab_set_termios mxc%d\n", info->line);
#endif

    if ((tty->termios->c_cflag == old_termios->c_cflag) &&
	((tty->termios->c_iflag & (IXON|IXANY)) == 
	 (old_termios->c_iflag & (IXON|IXANY))))
        return;
    set_line_char(info);

    if ((old_termios->c_cflag & CRTSCTS) &&
        !(tty->termios->c_cflag & CRTSCTS)) {
            tty->hw_stopped = 0;
            cab_start(tty);
    }

    return;
} /* cab_set_termios */

/* This routine is called by the upper-layer tty layer to signal
   that incoming characters should be throttled because the input
   buffers are close to full.
 */
static void
cab_throttle(struct tty_struct * tty)
{
} /* cab_throttle */


/*
 * This routine notifies the tty driver that it should signal
 * that characters can now be sent to the tty without fear of
 * overrunning the input buffers of the line disciplines.
 */
static void
cab_unthrottle(struct tty_struct * tty)
{
} /* cab_unthrottle */


/* cab_start and cab_stop provide software output flow control as a
   function of XON/XOFF, software CTS, and other such stuff.
*/
static void
cab_stop(struct tty_struct *tty)
{
} /* cab_stop */


static void
cab_start(struct tty_struct *tty)
{
} /* cab_start */


static void
cab_flush_buffer(struct tty_struct *tty)
{
} /* cab_flush_buffer */


/*
 * cab_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void
cab_hangup(struct tty_struct *tty)
{
    struct cabernet_port * info = (struct cabernet_port *)tty->driver_data;
        
#ifdef CAB_DEBUG_OTHER
    printk("mxc: cab_hangup mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "cab_hangup"))
        return;

    cab_flush_buffer(tty);
    shutdown(info);
    info->event = 0;
    info->count = 0;
#ifdef CAB_DEBUG_COUNT
    printk("mxc: cab_hangup (%d): setting count to 0\n", current->pid);
#endif
    info->tty = 0;
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
    wake_up_interruptible(&info->open_wait);
} /* cab_hangup */


/*
 * This routine prints out the appropriate serial driver version number
 * and identifies which options were configured into this driver.
 */
static inline void
show_version(void)
{
    printk("Cabernet control driver, built %s %s\n",
	   __DATE__, __TIME__);
} /* show_version */

static int 
cabernet_get_proc_info(char *buf, char **start, off_t offset, int length,
		       int *eof, void *data)
{
    struct cabernet_port  *info;
    int i;
    int len=0;
    off_t begin=0;
    off_t pos=0;
    int size;
    __u32 cur_jifs = jiffies;

    size = sprintf(buf, "Dev TimeOpen   BytesOut  IdleOut    BytesIn   IdleIn  Overruns  Ldisc\n");

    pos += size;
    len += size;

    /* Output one line for each known port */
    for (i = 0; i < NR_PORTS && cab_port[i].line >= 0; i++) {
	info = &cab_port[i];

	if (info->count)
	    size = sprintf(buf+len,
			"%3d %8lu %10lu %8lu %10lu %8lu %9lu %6ld\n",
			info->line,
			JIFFIES_DIFF(info->idle_stats.in_use, cur_jifs) / HZ,
			info->idle_stats.xmit_bytes,
			JIFFIES_DIFF(info->idle_stats.xmit_idle, cur_jifs) / HZ,
			info->idle_stats.recv_bytes,
			JIFFIES_DIFF(info->idle_stats.recv_idle, cur_jifs) / HZ,
			info->idle_stats.overruns,
			(long) info->tty->ldisc.num);
	else
	    size = sprintf(buf+len,
			"%3d %8lu %10lu %8lu %10lu %8lu %9lu %6ld\n",
			info->line, 0L, 0L, 0L, 0L, 0L, 0L, 0L);
	len += size;
	pos = begin + len;

	if (pos < offset) {
	    len   = 0;
	    begin = pos;
	}
	if (pos > offset + length)
	    goto done;
    }
    *eof = 1;
done:
    *start = buf + (offset - begin);	/* Start of wanted data */
    len -= (offset - begin);		/* Start slop */
    if (len > length)
	len = length;			/* Ending slop */
    if (len < 0)
	len = 0;
    return len;
}

/* The serial driver boot-time initialization code!
    Hardware I/O ports are mapped to character special devices on a
    first found, first allocated manner.  That is, this code searches
    for Mxclom cards in the system.  As each is found, it is probed
    to discover how many chips (and thus how many ports) are present.
    These ports are mapped to the tty ports 32 and upward in monotonic
    fashion.  If an 8-port card is replaced with a 16-port card, the
    port mapping on a following card will shift.

    This approach is different from what is used in the other serial
    device driver because the Mxclom is more properly a multiplexer,
    not just an aggregation of serial ports on one card.

    If there are more cards with more ports than have been
    statically allocated above, a warning is printed and the
    extra ports are ignored.
 */

int __init
cab_init(void)
{
    struct cabernet_port  *info;
    struct cabernet_card *cinfo;
    int board,port,i;
    int nports;

    init_bh(CYCLADES_BH, do_cabernet_bh);

#ifdef CAB_DEBUG_OPEN
    show_version();
#endif

    /* Initialize the tty_driver structure */
    
    memset(&cab_driver, 0, sizeof(struct tty_driver));
    cab_driver.magic = TTY_DRIVER_MAGIC;
    cab_driver.driver_name = "cabernet";
    cab_driver.name = "mxc";
    cab_driver.major = CABERNET_MAJOR;
    cab_driver.minor_start = 0;
    cab_driver.num = NR_PORTS;
    cab_driver.type = TTY_DRIVER_TYPE_SERIAL;
    cab_driver.subtype = SERIAL_TYPE_NORMAL;
    cab_driver.init_termios = tty_std_termios;
    cab_driver.init_termios.c_cflag =
            B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    cab_driver.flags = TTY_DRIVER_REAL_RAW;
    cab_driver.refcount = &serial_refcount;
    cab_driver.table = serial_table;
    cab_driver.termios = serial_termios;
    cab_driver.termios_locked = serial_termios_locked;

    cab_driver.open = cab_open;
    cab_driver.close = cab_close;
    cab_driver.write = cab_write;
    cab_driver.put_char = cab_put_char;
    cab_driver.flush_chars = cab_flush_chars;
    cab_driver.write_room = cab_write_room;
    cab_driver.chars_in_buffer = cab_chars_in_buffer;
    cab_driver.flush_buffer = cab_flush_buffer;
    cab_driver.ioctl = cab_ioctl;
    cab_driver.throttle = cab_throttle;
    cab_driver.unthrottle = cab_unthrottle;
    cab_driver.set_termios = cab_set_termios;
    cab_driver.stop = cab_stop;
    cab_driver.start = cab_start;
    cab_driver.hangup = cab_hangup;
    cab_driver.break_ctl = cab_break;
    cab_driver.wait_until_sent = cab_wait_until_sent;
    cab_driver.read_proc = cabernet_get_proc_info;

    /* SOE driver */
    cab_soe_driver = cab_driver;
    cab_soe_driver.name = "ttySOE";
    cab_soe_driver.major = CABERNETSPOE_MAJOR;


    if (tty_register_driver(&cab_driver))
            panic("Couldn't register Cabernet serial driver\n");
    if (tty_register_driver(&cab_soe_driver))
            panic("Couldn't register Cabernet SOE driver\n");

    for (i = 0; i < NR_CARDS; i++) {
            /* base_addr=0 indicates board not found */
            cab_card[i].base_addr = 0;
    }

    /* the code below is responsible to find the boards. Each different
       type of board has its own detection routine. If a board is found,
       the next cab_card structure available is set by the detection
       routine. These functions are responsible for checking the
       availability of cab_card and cab_port data structures */

    /* assume it always exists */
    cab_card[0].ctl_phys = CAB_CPLD_BASE_ADDR;
    cab_card[0].ctl_addr = (ulong)ioremap(cab_card[0].ctl_phys, 0x100);
    cab_card[0].base_phys = CAB_BASE_ADDR;
    cab_card[0].base_addr = (ulong)ioremap(cab_card[0].base_phys, 0x100);
    cab_card[0].irq = 19; /* cabernet_setup.c */
    cab_card[0].num_chips = 1;
    cab_card[0].first_line = 0;
    if(request_irq(cab_card[0].irq, cab_interrupt, SA_SHIRQ, "Cabernet", &cab_card[0])) {
        printk("mxc: could not allocate IRQ#%d\n", cab_card[0].irq);
	return 1;
    }

#ifdef ARUBA_DEBUG
    printk("mxc: Cabernet Control Driver, IRQ %d\n", cab_card[0].irq);
#endif

    cab_nboard = 1;

    /* invalidate remaining cab_card structures */
    for (i = 0 ; i < NR_CARDS ; i++) {
        if (cab_card[i].base_addr == 0) {
                cab_card[i].first_line = -1;
                cab_card[i].ctl_addr = 0;
                cab_card[i].irq = 0;
                cab_card[i].first_line = 0;
                cab_card[i].num_chips = 0;
        }
    }
    /* invalidate remaining cab_port structures */
    for (i = 0; i < NR_PORTS ; i++) {
        cab_port[i].line = -1;
        cab_port[i].magic = -1;
    }

    nports = NR_PORTS;
    /* initialize per-port data structures for each valid board found */
    for (board = 0; board < cab_nboard; board++) {
        cinfo = &cab_card[board];
	cinfo->nports = nports;
	spin_lock_init(&cinfo->card_lock);
	for (port = cinfo->first_line ;
	     port < cinfo->first_line + nports;
	     port++) {
	    info = &cab_port[port];
	    info->magic = CABERNET_MAGIC;
	    info->type = PORT_PHILIPS;
	    info->card = board;
	    info->line = port;
	    info->flags = STD_COM_FLAGS;
	    info->tty = 0;
	    info->xmit_fifo_size = CAB_FIFO_SIZE;
	    info->custom_divisor = 0;
	    info->close_delay = 5*HZ/10;
	    info->closing_wait = CLOSING_WAIT_DELAY;
	    info->icount.cts = info->icount.dsr = 
	      info->icount.rng = info->icount.dcd = 0;
	    info->icount.rx = info->icount.tx = 0;
	    info->icount.frame = info->icount.parity = 0;
	    info->icount.overrun = info->icount.brk = 0;
	    info->event = 0;
	    info->count = 0;
	    info->blocked_open = 0;
	    info->tqueue.routine = do_softint;
	    info->tqueue.data = info;
	    info->normal_termios =
	      cab_driver.init_termios;
	    init_waitqueue_head(&info->open_wait);
	    init_waitqueue_head(&info->close_wait);
	    init_waitqueue_head(&info->shutdown_wait);
	    /* info->session */
	    /* info->pgrp */
	    /* info->timeout */
	    /* Bentson's vars */
	    info->jiffies[0] = 0;
	    info->jiffies[1] = 0;
	    info->jiffies[2] = 0;
	}
    }

    return 0;
    
} /* cab_init */

#ifdef MODULE
void
cab_cleanup_module(void)
{
    int i;
    int e;
    unsigned long flags;

    save_flags(flags); cli();
    remove_bh(CABERNET_BH);

    if ((e = tty_unregister_driver(&cab_driver)))
            printk("mxc: failed to unregister Cabernet driver(%d)\n",
		   e);
    if ((e = tty_unregister_driver(&cab_soe_driver)))
            printk("mxc: failed to unregister Cabernet SOE driver (%d)\n", 
		   e);

    restore_flags(flags);

    for (i = 0; i < NR_CARDS; i++) {
        if (cab_card[i].base_addr != 0) {
	    iounmap((void *)cab_card[i].base_addr);
	    if (cab_card[i].ctl_addr != 0)
		iounmap((void *)cab_card[i].ctl_addr);
	    if (cab_card[i].irq)
		free_irq(cab_card[i].irq, &cab_card[i]);
        }
    }
    if (tmp_buf) {
	free_page((unsigned long) tmp_buf);
	tmp_buf = NULL;
    }
} /* cab_cleanup_module */

/* Module entry-points */
module_init(cab_init);
module_exit(cab_cleanup_module);

#else /* MODULE */
/* called by linux/init/main.c to parse command line options */
void
cab_setup(char *str, int *ints)
{
} /* cab_setup */
#endif /* MODULE */

MODULE_LICENSE("GPL");
