#define	Z_WAKE

/*
 *  linux/drivers/char/mswitch.c
 *
 * This file contains the driver for the mswitch async linecard
 * that contain serial ports.
 *
 * Based upon cyclades.c
 *
 * Much of the design and some of the code came from serial.c
 * which was copyright (C) 1991, 1992  Linus Torvalds.  It was
 * extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92,
 * and then fixed as suggested by Michael K. Johnson 12/12/92.
 */

/* If you need to install more boards than NR_CARDS, change the constant
   in the definition below. No other change is necessary to support up to
   eight boards. Beyond that you'll have to extend mx_isa_addresses. */

#define NR_CARDS        1 /* not to be confused with line cards */
                          /* there is only one SCCI, and hence 1 card */

/*
   If the total number of ports is larger than NR_PORTS, change this
   constant in the definition below. No other change is necessary to
   support more boards/ports. */
#define NR_LC_PORTS     (LCCI_SOE_PORTS + LCCI_POE_PORTS)
#define NR_PORTS        (SCCI_LCS * NR_LC_PORTS)

#define	SERIAL_PARANOIA_CHECK
#undef	MX_DEBUG_OPEN
#undef	MX_DEBUG_THROTTLE
#undef	MX_DEBUG_OTHER
#undef	MX_DEBUG_IO
#undef	MX_DEBUG_IOCTL
#undef	MX_DEBUG_COUNT
#undef	MX_DEBUG_DTR
#undef	MX_DEBUG_WAIT_UNTIL_SENT
#undef	MX_DEBUG_INTERRUPTS
#undef	MX_ENABLE_MONITORING

/*
 * Include section 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/mswitch.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>

#define	MX_LOCK(info,flags) \
    do { \
	spin_lock_irqsave(&mx_card[info->card].card_lock, flags); \
    } while (0)
		
#define	MX_UNLOCK(info,flags) \
    do { \
	spin_unlock_irqrestore(&mx_card[info->card].card_lock, flags); \
    } while (0)

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/version.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include "../../../cmn/util/sc_cpld.h"
#include "../../../cmn/util/scci.h"
#include "../../../cmn/util/lcci.h"
#include "../../../cmn/util/sc28.h"
#include "../../../cmn/util/cxe.h"
#include "../../../cmn/util/util_arp.h"

#define PHY_TIMEOUT    10000

char __mxc_platid[10];

static ulong lcci_write(ulong base, ulong lc, ulong reg, ulong val)
{
    ulong error;
    LCCI_REG_WRITE(base, lc, reg, val, error);
    return error;
}

static ulong lcci_read(ulong base, ulong lc, ulong reg, ulong *data)
{
    ulong error, val;
    LCCI_REG_READ(base, lc, reg, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong lcci_write_spoe(ulong base, ulong lc, ulong dev, ulong reg, ulong val)
{
    ulong error;
    LCCI_SPOE_REG_WRITE(base, lc, dev, reg, val, error);
    return error;
}

static ulong lcci_read_spoe(ulong base, ulong lc, ulong dev, ulong reg, ulong *data)
{
    ulong error, val;
    LCCI_SPOE_REG_READ(base, lc, dev, reg, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong lcci_read_i2c(ulong base, ulong lc, ulong dev, ulong reg, ulong *data)
{
    ulong error, val;
    LCCI_I2C_REG_READ(base, lc, dev, reg, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong lcci_exists(ulong base_addr, ulong lc)
{
    ulong val;
    if (lc < SCCI_LCS) {
        SCCI_REG_READ(base_addr, SCCI_MODULE_STATUS, val);
	return ((val & SCCI_MODULE_LC_PRES(lc))? 1: 0);
    }
    return 0;
}

static ulong lcci_spoe_exists(ulong base_addr, ulong lc)
{
    ulong val;
    lcci_read(base_addr, lc, LCCI_MODULE_STATUS, &val);
    return (val & LCCI_SPOE_PRES);
}

static ulong cxe_write(ulong base, ulong lc, ulong addr, ulong data)
{
    ulong error;
    CXE_REG_WRITE(base, lc, 0, addr, data, error);
    return error;
}

static ulong cxe_read(ulong base, ulong lc, ulong addr, ulong *data)
{
    ulong val, error;
    CXE_REG_READ(base, lc, 0, addr, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong cxe_write_phy(ulong base, ulong lc, ulong device, ulong reg, ulong val)
{
    ulong tries, error;
    cxe_write(base, lc, CXE_GC_MDIO_ACCESS,
	      (CXE_GC_MDIO_ACCESS_WRITE |
	       (device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT) |
	       (val << CXE_GC_MDIO_ACCESS_DATA_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(base, lc, CXE_GC_MDIO_ACCESS, &val);
	if (error) {
	    return error;
	}
	if (val & CXE_GC_MDIO_ACCESS_DONE) {
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

static ulong cxe_read_phy(ulong base, ulong lc, ulong device, ulong reg, ulong *val)
{
    ulong tries, error;
    cxe_write(base, lc, CXE_GC_MDIO_ACCESS,
	      ((device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(base, lc, CXE_GC_MDIO_ACCESS, val);
	if (error) {
	    break;
	}
	if (*val & CXE_GC_MDIO_ACCESS_DONE) {
	    *val &= 0xffff;
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

#define mx_put_user	put_user

static unsigned long 
mx_get_user(unsigned long *addr)
{
    unsigned long result = 0;
    int error = get_user (result, addr);
    if (error)
        printk ("mswitch: mx_get_user: error == %d\n", error);
    return result;
}

#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef SERIAL_XMIT_SIZE
#define	SERIAL_XMIT_SIZE	(MIN(PAGE_SIZE, 4096))
#endif
#define WAKEUP_CHARS		256
#define STD_COM_FLAGS (0)

#define	JIFFIES_DIFF(n, j)	((j) - (n))

static DECLARE_TASK_QUEUE(tq_mswitch);

static struct tty_driver mx_driver, mx_soe_driver, mx_poe_driver;
static int serial_refcount;


/* This is the per-card data structure containing address, irq, number of
   channels, etc. This driver supports a maximum of NR_CARDS cards.
*/
static struct mswitch_card mx_card[NR_CARDS];

/* This is the per-channel data structure containing pointers, flags
 and variables for the port. This driver supports a maximum of NR_PORTS.
*/
static struct mswitch_port mx_port[NR_PORTS];
static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];

/*
 * tmp_buf is used as a temporary buffer by serial_write.  We need to
 * lock it in case the copy_from_user blocks while swapping in a page,
 * and some other program tries to do a serial write at the same time.
 * Since the lock will only come under contention when the system is
 * swapping and available memory is low, it makes sense to share one
 * buffer across all the serial ports, since it significantly saves
 * memory if large numbers of serial ports are open.  This buffer is
 * allocated when the first mx_open occurs.
 */
static unsigned char *tmp_buf;
DECLARE_MUTEX(tmp_buf_sem);

static unsigned short	mx_nboard;
static void mx_start(struct tty_struct *);
static void set_line_char(struct mswitch_port *);

static int mswitch_get_proc_info(char *, char **, off_t , int , int *, void *);

static inline int
serial_paranoia_check(struct mswitch_port *info,
                        kdev_t device, const char *routine)
{
#ifdef SERIAL_PARANOIA_CHECK
    static const char *badmagic =
        "mxc: Warning: bad magic number for serial struct (%s) in %s\n";
    static const char *badinfo =
        "mxc: Warning: null mswitch_port for (%s) in %s\n";
    static const char *badrange =
        "mxc: Warning: mswitch_port out of range for (%s) in %s\n";

    if (!info) {
        printk(badinfo, kdevname(device), routine);
        return 1;
    }

    if( (long)info < (long)(&mx_port[0])
    || (long)(&mx_port[NR_PORTS]) < (long)info ){
        printk(badrange, kdevname(device), routine);
        return 1;
    }

    if (info->magic != MSWITCH_MAGIC) {
        printk(badmagic, kdevname(device), routine);
        return 1;
    }
#endif
        return 0;
} /* serial_paranoia_check */

/*
 * This routine is used by the interrupt handler to schedule
 * processing in the software interrupt portion of the driver
 * (also known as the "bottom half").  This can be called any
 * number of times for any channel without harm.
 */
static inline void
mx_sched_event(struct mswitch_port *info, int event)
{
    info->event |= 1 << event; /* remember what kind of event and who */
    queue_task(&info->tqueue, &tq_mswitch); /* it belongs to */
    mark_bh(CYCLADES_BH);                   /* then trigger event */
} /* mx_sched_event */


/*
 * This routine is used to handle the "bottom half" processing for the
 * serial driver, known also the "software interrupt" processing.
 * This processing is done at the kernel interrupt level, after the
 * mx#/_interrupt() has returned, BUT WITH INTERRUPTS TURNED ON.  This
 * is where time-consuming activities which can not be done in the
 * interrupt driver proper are done; the interrupt driver schedules
 * them using mx_sched_event(), and they get done here.
 *
 * This is done through one level of indirection--the task queue.
 * When a hardware interrupt service routine wants service by the
 * driver's bottom half, it enqueues the appropriate tq_struct (one
 * per port) to the tq_mswitch work queue and sets a request flag
 * via mark_bh for processing that queue.  When the time is right,
 * do_mswitch_bh is called (because of the mark_bh) and it requests
 * that the work queue be processed.
 *
 * Although this may seem unwieldy, it gives the system a way to
 * pass an argument (in this case the pointer to the mswitch_port
 * structure) to the bottom half of the driver.  Previous kernels
 * had to poll every port to see if that port needed servicing.
 */
static void
do_mswitch_bh(void)
{
    run_task_queue(&tq_mswitch);
} /* do_mswitch_bh */

static void
do_softint(void *private_)
{
  struct mswitch_port *info = (struct mswitch_port *) private_;
  struct tty_struct    *tty;

    tty = info->tty;
    if (!tty)
        return;

    if (test_and_clear_bit(Mx_EVENT_HANGUP, &info->event)) {
        tty_hangup(info->tty);
        wake_up_interruptible(&info->open_wait);
        info->flags &= ~(ASYNC_NORMAL_ACTIVE|
                             ASYNC_CALLOUT_ACTIVE);
    }
    if (test_and_clear_bit(Mx_EVENT_OPEN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->open_wait);
    }
    if (test_and_clear_bit(Mx_EVENT_WRITE_WAKEUP, &info->event)) {
        if((tty->flags & (1<< TTY_DO_WRITE_WAKEUP))
        && tty->ldisc.write_wakeup){
            (tty->ldisc.write_wakeup)(tty);
        }
        wake_up_interruptible(&tty->write_wait);
    }
#ifdef Z_WAKE
    if (test_and_clear_bit(Mx_EVENT_SHUTDOWN_WAKEUP, &info->event)) {
        wake_up_interruptible(&info->shutdown_wait);
    }
#endif
} /* do_softint */


/* The real interrupt service routine is called
   whenever the card wants its hand held--chars
   received, out buffer empty, modem change, etc.
 */
static void
mx_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    struct mswitch_card *cinfo;
    struct mswitch_port *info;
    struct tty_struct *tty;
    ulong base_addr;
    int chip;
    ulong status;
    ulong data;
    ulong char_count;
    int outch;
    int i;
    int had_work;
    ulong scci_intrs;
    ulong lcci_intrs;
    ulong lc, chan;

    if ((cinfo = (struct mswitch_card *)dev_id) == 0) {
#ifdef MX_DEBUG_INTERRUPTS
	printk("mx_interrupt: spurious interrupt %d\n\r", irq);
#endif
        return; /* spurious interrupt */
    }

    /* This loop checks all chips in the card.  Make a note whenever
       _any_ chip had some work to do, as this is considered an
       indication that there will be more to do.  Only when no chip
       has any work does this outermost loop exit.
     */
    base_addr = cinfo->base_addr;
    do {
        scci_intrs = 0;
	SCCI_REG_READ(base_addr, SCCI_INTR_STATUS, scci_intrs);
#ifdef MX_DEBUG_INTERRUPTS
	printk("mxc: scci int %lx\n\r", scci_intrs);
#endif
	if (!scci_intrs) break;
        had_work = 0;
	for (lc = 0; lc < SCCI_LCS; lc++) {
	    if (!(scci_intrs & SCCI_INTR_LC_INTR(lc))) continue;
	    lcci_intrs = 0;
	    spin_lock(&cinfo->card_lock);
	    lcci_read(base_addr, lc, LCCI_INTR_STATUS, &lcci_intrs);
	    spin_unlock(&cinfo->card_lock);
#ifdef MX_DEBUG_INTERRUPTS
	    printk("mxc: lcci %ld int %lx\n\r", lc, lcci_intrs);
#endif
	    for (chip = 0; chip < LCCI_SOE_CHIPS+LCCI_POE_CHIPS; chip++) {
	        if (!(lcci_intrs & LCCI_INTR_SOE_INT(chip))) continue;
		spin_lock(&cinfo->card_lock);
		lcci_write_spoe(base_addr, lc, chip, SC28_UCIR, 0);
		lcci_read_spoe(base_addr, lc, chip, SC28_CIR, &status);
		spin_unlock(&cinfo->card_lock);
#ifdef MX_DEBUG_INTERRUPTS
		printk("mxc: spoe %d int %lx\n\r", chip, status);
#endif
		if (!status) continue;
		had_work = 1;
		chan = status & 0x7;
		i = chan + (chip * 8) + (lc * NR_LC_PORTS) + cinfo->first_line;
		info = &mx_port[i];
		if (status & SC28_CIR_Rx) { /* RX */
#ifdef MX_DEBUG_INTERRUPTS
		    printk("mxc: rcvd intr, chip %d\n\r", chip);
#endif
		    spin_lock(&cinfo->card_lock);
		    if (status & 0x40) {
		        /* RX errors */
                        data = SC28_CR_CMD_RESET_ERROR|SC28_CR_TxRxLOCK;
			lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), data);
		    }
                    /* read char */
                    info->last_active = jiffies;
		    /* load # chars available from the chip */
		    lcci_read_spoe(base_addr, lc, chip, SC28_GIBCR, &char_count);
		    char_count++;
#ifdef MX_DEBUG_INTERRUPTS
		    printk("mxc: rcvd intr, chars %ld\n\r", char_count);
#endif

                    /* if there is nowhere to put the data, discard it */
                    if (info->tty == 0) {
			while (char_count--) {
			    lcci_read_spoe(base_addr, lc, chip, SC28_RxFIFO(chan), &data);
			}
                    } else { /* there is an open port for this data */
			/* normal character reception */
		        tty = info->tty;
#ifdef MX_ENABLE_MONITORING
			++info->mon.int_count;
			info->mon.char_count += char_count;
			if (char_count > info->mon.char_max)
			    info->mon.char_max = char_count;
			info->mon.char_last = char_count;
#endif
			info->idle_stats.recv_bytes += char_count;
			info->idle_stats.recv_idle   = jiffies;
			while (char_count--) {
			    if (tty->flip.count >= TTY_FLIPBUF_SIZE){
			        break;
			    }
			    lcci_read_spoe(base_addr, lc, chip, SC28_RxFIFO(chan), &data);
			    tty->flip.count++;
			    *tty->flip.flag_buf_ptr++ = TTY_NORMAL;
			    *tty->flip.char_buf_ptr++ = data;
			    info->icount.rx++;
			}
                        queue_task(&tty->flip.tqueue, &tq_timer);
                    }
		    spin_unlock(&cinfo->card_lock);
		} else if (status & SC28_CIR_Tx) { /* TX interrupt */
                    /* Since we only get here when the transmit buffer
                       is empty, we know we can always stuff a dozen
                       characters. */
#ifdef MX_DEBUG_INTERRUPTS
		    printk("mxc: xmit intr, chip %d\n\r", chip);
#endif
		    spin_lock(&cinfo->card_lock);
                    info->last_active = jiffies;
                    if (info->tty == 0) {
		        lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan),
					SC28_IMR_RxRDY);
                        goto txdone;
                    }

#ifdef MX_DEBUG_INTERRUPTS
		    printk("mxc: xmit intr, cnt = %d\n\r", info->xmit_cnt);
#endif
                    /* load the on-chip space for outbound data */
                    char_count = info->xmit_fifo_size;
                    while (char_count-- > 0){
			if (!info->xmit_cnt){
#ifdef MX_DEBUG_INTERRUPTS
			    printk("mxc: xmit intr, !cnt\n\r");
#endif
			    lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan),
					    SC28_IMR_RxRDY);
			    goto txdone;
			}
			if ((info->xmit_buf == 0) ||
			    info->tty->stopped || info->tty->hw_stopped) {
#ifdef MX_DEBUG_INTERRUPTS
			    printk("mxc: xmit intr, stopped\n\r");
#endif
			    lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan),
					    SC28_IMR_RxRDY);
			    goto txdone;
			}
                        outch = info->xmit_buf[info->xmit_tail];
			info->xmit_cnt--;
			info->xmit_tail = ((info->xmit_tail + 1) &
					   (SERIAL_XMIT_SIZE - 1));
			lcci_write_spoe(base_addr, lc, chip, SC28_TxFIFO(chan), outch);
			info->icount.tx++;
                    }
		txdone:
                    if (info->xmit_cnt < WAKEUP_CHARS) {
                        mx_sched_event(info, Mx_EVENT_WRITE_WAKEUP);
                    }
		    spin_unlock(&cinfo->card_lock);
                } else { /* other interrupts */
		    spin_lock(&cinfo->card_lock);
		    switch (status & SC28_CIR_OTHER_MASK) {
		    case SC28_CIR_OTHER_XonXoff:
#ifdef MX_DEBUG_INTERRUPTS
		        printk("mxc: other intr, XonXoff\n\r");
#endif
			lcci_read_spoe(base_addr, lc, chip, SC28_XISR(chan), &data);
		        break;
		    case SC28_CIR_OTHER_BREAK:
#ifdef MX_DEBUG_INTERRUPTS
		        printk("mxc: other intr, break\n\r");
#endif
			data = SC28_CR_CMD_RESET_BREAK|SC28_CR_TxRxLOCK;
			lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), data);
			break;
		    default:
#ifdef MX_DEBUG_INTERRUPTS
		        printk("mxc: other intr, %lx\n\r", status);
#endif
		        break;
		    }
		    spin_unlock(&cinfo->card_lock);
		}
            }
        }
    } while(had_work);
} /* mx_interrupt */

static void
mswitch_soe_ssr_set(ulong base_addr, int lc, int iport, int on)
{
    static ulong soe_on[SCCI_LCS] = {0};
    int i;

    if (iport >= LCCI_SOE_PORTS) {
      return;
    }

    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x0); /* bit 1 is SCK,  0 is RCK */
    lcci_write_spoe(base_addr, lc, 1, 0x97, 0x2); /* bit 1 is EN_L, 0 is SHIFT_IN */
    lcci_write_spoe(base_addr, lc, 2, 0x97, 0x0); /* bit 1 is NC,   0 is CLR_L */
    
    /* tell UARTs to use 0x97 contents to output to GOUT pins */
    lcci_write_spoe(base_addr, lc, 0, 0x87, 0xef);
    lcci_write_spoe(base_addr, lc, 1, 0x87, 0xef);
    lcci_write_spoe(base_addr, lc, 2, 0x87, 0xef);

    /* generate a few clocks */
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x3);
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x0);
    lcci_write_spoe(base_addr, lc, 2, 0x97, 0x1);
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x3);
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x0);

    /* clock shift chain */
    soe_on[lc] &= ~(0x1 << iport);
    soe_on[lc] |= (on << iport);
    for (i = LCCI_SOE_PORTS - 1; i >= 0; i--) {
        int port = (soe_on[lc] & (0x1 << i))? 1: 0;
        lcci_write_spoe(base_addr, lc, 1, 0x97, 0x2 | port);
        lcci_write_spoe(base_addr, lc, 0, 0x97, 0x2);
        lcci_write_spoe(base_addr, lc, 0, 0x97, 0x0);
    }
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x1);
    lcci_write_spoe(base_addr, lc, 0, 0x97, 0x0);
    lcci_write_spoe(base_addr, lc, 1, 0x97, 0x0);
}

static void
mswitch_soe_chan_init(ulong base_addr, int lc, int chip, int chan)
{
    lcci_write_spoe(base_addr, lc, chip, SC28_GCCR, 0);
    if (chip < LCCI_SOE_CHIPS) {
        /* set default 9600, 8N1 */
        lcci_write_spoe(base_addr, lc, chip, SC28_MR1(chan), SC28_MR1_8N);
        lcci_write_spoe(base_addr, lc, chip, SC28_MR2(chan), 0);
        lcci_write_spoe(base_addr, lc, chip, SC28_TxCSR(chan), SC28_CSR_9600);
        lcci_write_spoe(base_addr, lc, chip, SC28_RxCSR(chan), SC28_CSR_9600);
        /* enable in-band FC */
        lcci_write_spoe(base_addr, lc, chip, SC28_MR0(chan),
			SC28_MR0_IGFC|SC28_MR0_TXFC|SC28_MR0_RXFC);
    } else {
        /* set default 19200, 8N1 */
        lcci_write_spoe(base_addr, lc, chip, SC28_MR1(chan), SC28_MR1_8N);
        lcci_write_spoe(base_addr, lc, chip, SC28_MR2(chan), 0);
        lcci_write_spoe(base_addr, lc, chip, SC28_TxCSR(chan), SC28_CSR_19200);
        lcci_write_spoe(base_addr, lc, chip, SC28_RxCSR(chan), SC28_CSR_19200);
        /* no flow control for POE */
        lcci_write_spoe(base_addr, lc, chip, SC28_MR0(chan), 0);
    }
    lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), SC28_CR_CMD_RxRST);
    lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), SC28_CR_CMD_TxRST);
    lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan), SC28_IMR_RxRDY);
    lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), SC28_CR_TxEN|SC28_CR_RxEN);
}

/* This is called whenever a port becomes active;
   interrupts are enabled and DTR & RTS are turned on.
 */
static int
startup(struct mswitch_port * info)
{
    unsigned long flags;
    int retval = 0;
    ulong base_addr;
    int card,channel;
    unsigned long page;
    unsigned long xisr;
    int lc, iport, chan, chip;

    card = info->card;
    channel = (info->line) - (mx_card[card].first_line);

    page = get_free_page(GFP_KERNEL);
    if (!page)
	return -ENOMEM;

    MX_LOCK(info, flags);

    if (info->flags & ASYNC_INITIALIZED){
	free_page(page);
	goto errout;
    }

    if (!info->type){
        if (info->tty){
            set_bit(TTY_IO_ERROR, &info->tty->flags);
        }
	free_page(page);
	goto errout;
    }

    if (info->xmit_buf)
	free_page(page);
    else
	info->xmit_buf = (unsigned char *) page;

    MX_UNLOCK(info, flags);

    set_line_char(info);

    base_addr = mx_card[card].base_addr;

#ifdef MX_DEBUG_OPEN
    printk("mxc: startup card %d, channel %d, base_addr %lx\n",
	   card, channel, base_addr);
#endif

    MX_LOCK(info, flags);

    lc = channel / NR_LC_PORTS;
    iport = channel % NR_LC_PORTS;
    chip = SC28_LC_CHIP(iport);
    chan = SC28_LC_CHAN(iport);

#ifdef MX_DEBUG_OPEN
    printk(" mxc: startup lc %d, chip %d, port %d\n",
	   lc, chip, chan);
#endif

    lcci_read_spoe(base_addr, lc, chip, SC28_XISR(chan), &xisr);
    if ((xisr & SC28_XISR_TXFLOW_STATUS_MASK) == SC28_XISR_TXFLOW_STATUS_DIS) {
        /* Stray Xoff, we have to reset CHIP, by setting chan 0 COMMAND reg */
        struct mswitch_port * rinfo;
        int ri;
        lcci_write_spoe(base_addr, lc, chip, SC28_CR(0), SC28_CR_CMD_STOP_RESET_CHIP);
        /* re-init any open channels */
        rinfo = &mx_port[(lc * NR_LC_PORTS) + (chip * 8)];
        for (ri = 0; ri < 8; ri++, rinfo++) {
            if (rinfo->flags & ASYNC_INITIALIZED) {
                mswitch_soe_chan_init(base_addr, lc, chip, ri);
            }
        }
    }

    /* initialize hardware for channel */
    mswitch_soe_chan_init(base_addr, lc, chip, chan);
    /* enable channel in shift register chain */
    mswitch_soe_ssr_set(base_addr, lc, iport, 1);

    info->flags |= ASYNC_INITIALIZED;

    if (info->tty) {
        clear_bit(TTY_IO_ERROR, &info->tty->flags);
    }
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    info->breakon = info->breakoff = 0;
    memset((char *)&info->idle_stats, 0, sizeof(info->idle_stats));
    info->idle_stats.in_use    =
      info->idle_stats.recv_idle =
      info->idle_stats.xmit_idle = jiffies;
    
    MX_UNLOCK(info, flags);

#ifdef MX_DEBUG_OPEN
    printk(" mxc: startup done\n");
#endif
    return 0;
	
errout:
    MX_UNLOCK(info, flags);
    return retval;
} /* startup */


static void
start_xmit( struct mswitch_port *info )
{
    ulong flags;
    ulong base_addr;
    ulong card,lc,iport,chip,channel,chan;

    card = info->card;
    channel = (info->line) - (mx_card[card].first_line);
    lc = channel / NR_LC_PORTS;
    iport = channel % NR_LC_PORTS;
    chip = SC28_LC_CHIP(iport);
    chan = SC28_LC_CHAN(iport);
    base_addr = mx_card[card].base_addr;

#ifdef MX_DEBUG_OTHER
    printk("mxc: start xmit lc %ld, chip %ld, port %ld\n",
	   lc, chip, chan);
#endif

    MX_LOCK(info, flags);
    lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan),
		    SC28_IMR_TxRDY|SC28_IMR_RxRDY);
    MX_UNLOCK(info, flags);
} /* start_xmit */

/*
 * This routine shuts down a serial port; interrupts are disabled,
 * and DTR is dropped if the hangup on close termio flag is on.
 */
static void
shutdown(struct mswitch_port * info)
{
    unsigned long flags;
    unsigned long base_addr;
    int card,chip,channel,chan,iport,lc;

    if (!(info->flags & ASYNC_INITIALIZED)){
        return;
    }

    card = info->card;
    channel = info->line - mx_card[card].first_line;
    base_addr = mx_card[card].base_addr;
    lc = channel / NR_LC_PORTS;
    iport = channel % NR_LC_PORTS;
    chip = SC28_LC_CHIP(iport);
    chan = SC28_LC_CHAN(iport);

#ifdef MX_DEBUG_OPEN
    printk("mxc: shutdown card %d, channel %d, base_addr %lx\n",
	   card, channel, (long)base_addr);
    printk(" mxc: shutdown lc %d, chip %d, port %d\n",
	   lc, chip, chan);
#endif

    MX_LOCK(info, flags);

    if (info->xmit_buf){
        unsigned char * temp;
	temp = info->xmit_buf;
	info->xmit_buf = 0;
	free_page((unsigned long) temp);
    }
    lcci_write_spoe(base_addr, lc, chip, SC28_IMR(chan), 0);
    lcci_write_spoe(base_addr, lc, chip, SC28_CR(chan), 0);
    lcci_write_spoe(base_addr, lc, chip, SC28_MR0(chan), 0); /* disable FC */
    /* disable channel in shift register chain */
    mswitch_soe_ssr_set(base_addr, lc, iport, 0);

    if (info->tty){
        set_bit(TTY_IO_ERROR, &info->tty->flags);
    }
    info->flags &= ~ASYNC_INITIALIZED;
    MX_UNLOCK(info, flags);

#ifdef MX_DEBUG_OPEN
    printk(" mxc: shutdown done\n");
#endif
    return;
} /* shutdown */


/*
 * ------------------------------------------------------------
 * mx_open() and friends
 * ------------------------------------------------------------
 */

static int
block_til_ready(struct tty_struct *tty, struct file * filp,
                           struct mswitch_port *info)
{
    DECLARE_WAITQUEUE(wait, current);
    struct mswitch_card *cinfo;
    unsigned long flags;
    int chip, channel;
    int retval;
    unsigned long base_addr;

    cinfo = &mx_card[info->card];
    channel = info->line - cinfo->first_line;

    /*
     * If the device is in the middle of being closed, then block
     * until it's done, and then try again.
     */
    if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
	if (info->flags & ASYNC_CLOSING) {
            interruptible_sleep_on(&info->close_wait);
	}
        return ((info->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    /*
     * If this is a callout device, then just make sure the normal
     * device isn't being used.
     */
    if (tty->driver.subtype == SERIAL_TYPE_CALLOUT) {
        if (info->flags & ASYNC_NORMAL_ACTIVE){
            return -EBUSY;
        }
        if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
            (info->flags & ASYNC_SESSION_LOCKOUT) &&
            (info->session != current->session)){
            return -EBUSY;
        }
        if ((info->flags & ASYNC_CALLOUT_ACTIVE) &&
            (info->flags & ASYNC_PGRP_LOCKOUT) &&
            (info->pgrp != current->pgrp)){
            return -EBUSY;
        }
        info->flags |= ASYNC_CALLOUT_ACTIVE;
        return 0;
    }

    /*
     * If non-blocking mode is set, then make the check up front
     * and then exit.
     */
    if ((filp->f_flags & O_NONBLOCK) ||
	(tty->flags & (1 << TTY_IO_ERROR))) {
        if (info->flags & ASYNC_CALLOUT_ACTIVE){
            return -EBUSY;
        }
        info->flags |= ASYNC_NORMAL_ACTIVE;
        return 0;
    }

    /*
     * Block waiting for the carrier detect and the line to become
     * free (i.e., not in use by the callout).  While we are in
     * this loop, info->count is dropped by one, so that
     * mx_close() knows when to free things.  We restore it upon
     * exit, either normal or abnormal.
     */
    retval = 0;
    add_wait_queue(&info->open_wait, &wait);
#ifdef MX_DEBUG_OPEN
    printk("mxc: block_til_ready before block: mxc%d, count = %d\n",
           info->line, info->count);/**/
#endif
    MX_LOCK(info, flags);
    if (!tty_hung_up_p(filp))
	info->count--;
    MX_UNLOCK(info, flags);
#ifdef MX_DEBUG_COUNT
    printk("mxc: block_til_ready: (%d): decrementing count to %d\n",
        current->pid, info->count);
#endif
    info->blocked_open++;

    chip = channel>>2;
    channel &= 0x03;
    base_addr = cinfo->base_addr;

    while (1) {
	set_current_state(TASK_INTERRUPTIBLE);
	if (tty_hung_up_p(filp) ||
	    !(info->flags & ASYNC_INITIALIZED)) {
	    retval = ((info->flags & ASYNC_HUP_NOTIFY) ? 
		      -EAGAIN : -ERESTARTSYS);
	    break;
	}

	if (signal_pending(current)) {
	    retval = -ERESTARTSYS;
	    break;
	}
#ifdef MX_DEBUG_OPEN
	printk("mxc: block_til_ready blocking: mxc%d, count = %d\n",
	       info->line, info->count);/**/
#endif
	schedule();
    }

    current->state = TASK_RUNNING;
    remove_wait_queue(&info->open_wait, &wait);
    if (!tty_hung_up_p(filp)){
	info->count++;
#ifdef MX_DEBUG_COUNT
	printk("mxc: block_til_ready (%d): incrementing count to %d\n",
	    current->pid, info->count);
#endif
    }
    info->blocked_open--;
#ifdef MX_DEBUG_OPEN
    printk("mxc: block_til_ready after blocking: mxc%d, count = %d\n",
	   info->line, info->count);/**/
#endif
    if (retval)
	return retval;
    info->flags |= ASYNC_NORMAL_ACTIVE;
    return 0;
} /* block_til_ready */

/*
 * This routine is called whenever a serial port is opened.  It
 * performs the serial-specific initialization for the tty structure.
 */
static int
mx_open(struct tty_struct *tty, struct file * filp)
{
    struct mswitch_port  *info;
    int retval, line;
    ulong flags;
    ulong page;
    ulong val;
    ulong lc;

    MOD_INC_USE_COUNT;
    line = MINOR(tty->device) - tty->driver.minor_start;
    if ((line < 0) || (NR_PORTS <= line)){
	MOD_DEC_USE_COUNT;
        return -ENODEV;
    }
    info = &mx_port[line];
    if (info->line < 0){
	MOD_DEC_USE_COUNT;
        return -ENODEV;
    }

    if ((MAJOR(tty->device) == MSWITCH_MAJOR) &&
	!strcmp(tty->driver.name, mx_driver.name)) {
#ifdef MX_DEBUG_OPEN
        printk("mxc: mx_open mxc %d %s\n", MSWITCH_MAJOR, tty->driver.name);
#endif
	tty->driver_data = info;
	return 0;
    }

    lc = line / NR_LC_PORTS;
    if (!lcci_exists(mx_card[info->card].base_addr, lc) || 
	!lcci_spoe_exists(mx_card[info->card].base_addr, lc)) {
#ifdef MX_DEBUG_OPEN
        printk("mxc: spoe %ld not present\n", lc);
#endif
	MOD_DEC_USE_COUNT;
        return -ENODEV;
    }

    tty->driver_data = info;
    info->tty = tty;
    if (serial_paranoia_check(info, tty->device, "mx_open")) {
        return -ENODEV;
    }

    MX_LOCK(info, flags);
    /* Enable LCCI interrupts */
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_INTR_MASK, val);
    val |= SCCI_INTR_LC_INTR(lc);
    SCCI_REG_WRITE(mx_card[info->card].base_addr, SCCI_INTR_MASK, val);

// mlord; can't be done here anymore.  performed in fpapps hapi layer as needed.
#if 0
    /* Enable LCCI access */
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_SCB_CONTROL, val);
    val |= SCCI_SCBC_LC_ENA(lc);
    SCCI_REG_WRITE(mx_card[info->card].base_addr, SCCI_SCB_CONTROL, val);
#endif

    /* enable SOE interrupts on LCCI */
    lcci_read(mx_card[info->card].base_addr, lc, LCCI_INTR_MASK, &val);
    val |= LCCI_INTR_SOE_INT((line % NR_LC_PORTS) / 8);
    lcci_write(mx_card[info->card].base_addr, lc, LCCI_INTR_MASK, val);
    MX_UNLOCK(info, flags);

#ifdef MX_DEBUG_OPEN
    printk("mxc: mx_open mxc%d, count = %d\n",
	   info->line, info->count);/**/
#endif
    info->count++;
#ifdef MX_DEBUG_COUNT
    printk("mxc: mx_open (%d): incrementing count to %d\n",
	   current->pid, info->count);
#endif
    if (!tmp_buf) {
	page = get_free_page(GFP_KERNEL);
	if (!page)
	    return -ENOMEM;
	if (tmp_buf)
	    free_page(page);
	else
	    tmp_buf = (unsigned char *) page;
    }

    /*
     * If the port is the middle of closing, bail out now
     */
    if (tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
	if (info->flags & ASYNC_CLOSING)
	    interruptible_sleep_on(&info->close_wait);
	return ((info->flags & ASYNC_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

    /*
     * Start up serial port
     */
    retval = startup(info);
    if (retval){
        return retval;
    }

    retval = block_til_ready(tty, filp, info);
    if (retval) {
#ifdef MX_DEBUG_OPEN
        printk("mxc: mx_open returning after block_til_ready with %d\n",
               retval);
#endif
        return retval;
    }

    if ((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
        *tty->termios = info->normal_termios;
    }

    info->session = current->session;
    info->pgrp = current->pgrp;

#ifdef MX_DEBUG_OPEN
    printk(" mxc: mx_open done\n");/**/
#endif

    return 0;
} /* mx_open */


/*
 * mx_wait_until_sent() --- wait until the transmitter is empty
 */
static void 
mx_wait_until_sent(struct tty_struct *tty, int timeout)
{
    struct mswitch_port * info = (struct mswitch_port *)tty->driver_data;
    unsigned long orig_jiffies, char_time;
	
    if (serial_paranoia_check(info, tty->device, "mx_wait_until_sent"))
	return;

    if (info->xmit_fifo_size == 0)
	return; /* Just in case.... */


    orig_jiffies = jiffies;
    /*
     * Set the check interval to be 1/5 of the estimated time to
     * send a single character, and make it at least 1.  The check
     * interval should also be less than the timeout.
     * 
     * Note: we have to use pretty tight timings here to satisfy
     * the NIST-PCTS.
     */
    char_time = (info->timeout - HZ/50) / info->xmit_fifo_size;
    char_time = char_time / 5;
    if (char_time == 0)
	char_time = 1;
    if (timeout < 0)
	timeout = 0;
    if (timeout)
	char_time = MIN(char_time, timeout);
    /*
     * If the transmitter hasn't cleared in twice the approximate
     * amount of time to send the entire FIFO, it probably won't
     * ever clear.  This assumes the UART isn't doing flow
     * control, which is currently the case.  Hence, if it ever
     * takes longer than info->timeout, this is probably due to a
     * UART bug of some kind.  So, we clamp the timeout parameter at
     * 2*info->timeout.
     */
    if (!timeout || timeout > 2*info->timeout)
	timeout = 2*info->timeout;
#ifdef MX_DEBUG_WAIT_UNTIL_SENT
    printk("In mx_wait_until_sent(%d) check=%lu...", timeout, char_time);
    printk("jiff=%lu...", jiffies);
#endif

    /* Run one more char cycle */
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(char_time * 5);
    current->state = TASK_RUNNING;
#ifdef MX_DEBUG_WAIT_UNTIL_SENT
    printk("Clean (jiff=%lu)...done\n", jiffies);
#endif
}

/*
 * This routine is called when a particular tty device is closed.
 */
static void
mx_close(struct tty_struct *tty, struct file *filp)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
    unsigned long flags;

#ifdef MX_DEBUG_OTHER
    printk("mxc: mx_close mxc%d\n", info->line);
#endif

    if ((MAJOR(tty->device) == MSWITCH_MAJOR) &&
	!strcmp(tty->driver.name, mx_driver.name)) {
	MOD_DEC_USE_COUNT;
        return;
    }

    if (!info || serial_paranoia_check(info, tty->device, "mx_close")){
        return;
    }

    MX_LOCK(info, flags);
    /* If the TTY is being hung up, nothing to do */
    if (tty_hung_up_p(filp)) {
	MOD_DEC_USE_COUNT;
	MX_UNLOCK(info, flags);
        return;
    }
        
#ifdef MX_DEBUG_OPEN
    printk("mxc: mx_close mxc%d, count = %d\n", info->line, info->count);
#endif
    if ((tty->count == 1) && (info->count != 1)) {
        /*
         * Uh, oh.  tty->count is 1, which means that the tty
         * structure will be freed.  Info->count should always
         * be one in these conditions.  If it's greater than
         * one, we've got real problems, since it means the
         * serial port won't be shutdown.
         */
        printk("mxc: mx_close: bad serial port count; tty->count is 1, "
           "info->count is %d\n", info->count);
        info->count = 1;
    }
#ifdef MX_DEBUG_COUNT
    printk("mxc: mx_close at (%d): decrementing count to %d\n",
        current->pid, info->count - 1);
#endif
    if (--info->count < 0) {
#ifdef MX_DEBUG_COUNT
        printk("mxc: mxc_close setting count to 0\n");
#endif
        info->count = 0;
    }
    if (info->count) {
	MOD_DEC_USE_COUNT;
	MX_UNLOCK(info, flags);
        return;
    }
    info->flags |= ASYNC_CLOSING;
    /*
     * Save the termios structure, since this port may have
     * separate termios for callout and dialin.
     */
    if (info->flags & ASYNC_NORMAL_ACTIVE)
        info->normal_termios = *tty->termios;

    /*
    * Now we wait for the transmit buffer to clear; and we notify
    * the line discipline to only process XON/XOFF characters.
    */
    tty->closing = 1;
    MX_UNLOCK(info, flags);
    if (info->closing_wait != MX_CLOSING_WAIT_NONE) {
	tty_wait_until_sent(tty, info->closing_wait);
    }
    MX_LOCK(info, flags);
    {
	if (info->flags & ASYNC_INITIALIZED) {
	    /* Waiting for on-board buffers to be empty before closing 
	       the port */
	    MX_UNLOCK(info, flags);
	    mx_wait_until_sent(tty, info->timeout);
	    MX_LOCK(info, flags);
	}
    }

    MX_UNLOCK(info, flags);
    shutdown(info);
    if (tty->driver.flush_buffer)
        tty->driver.flush_buffer(tty);
    if (tty->ldisc.flush_buffer)
        tty->ldisc.flush_buffer(tty);
    MX_LOCK(info, flags);

    tty->closing = 0;
    info->event = 0;
    info->tty = 0;
    if (info->blocked_open) {
	MX_UNLOCK(info, flags);
        if (info->close_delay) {
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(info->close_delay);
        }
        wake_up_interruptible(&info->open_wait);
	MX_LOCK(info, flags);
    }
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|
                     ASYNC_CLOSING);
    wake_up_interruptible(&info->close_wait);

#ifdef MX_DEBUG_OTHER
    printk(" mxc: mx_close done\n");
#endif

    MOD_DEC_USE_COUNT;
    MX_UNLOCK(info, flags);
    return;
} /* mx_close */


/* This routine gets called when tty_write has put something into
 * the write_queue.  The characters may come from user space or
 * kernel space.
 *
 * This routine will return the number of characters actually
 * accepted for writing.
 *
 * If the port is not already transmitting stuff, start it off by
 * enabling interrupts.  The interrupt service routine will then
 * ensure that the characters are sent.
 * If the port is already active, there is no need to kick it.
 *
 */
static int
mx_write(struct tty_struct * tty, int from_user,
           const unsigned char *buf, int count)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
    unsigned long flags;
    int c, ret = 0;

#ifdef MX_DEBUG_IO
    printk("mxc: mx_write mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "mx_write")){
        return 0;
    }
        
    if (!tty || !info->xmit_buf || !tmp_buf){
        return 0;
    }

    if (from_user) {
	down(&tmp_buf_sem);
	while (1) {
	    int c1;
	    
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
				SERIAL_XMIT_SIZE - info->xmit_head));
	    if (c <= 0)
		break;

	    c -= copy_from_user(tmp_buf, buf, c);
	    if (!c) {
		if (!ret) {
		    ret = -EFAULT;
		}
		break;
	    }
	    MX_LOCK(info, flags);
	    c1 = MIN(c, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
			SERIAL_XMIT_SIZE - info->xmit_head));
			
	    if (c1 < c)
	    	c = c1;
	    memcpy(info->xmit_buf + info->xmit_head, tmp_buf, c);
	    info->xmit_head = ((info->xmit_head + c) & (SERIAL_XMIT_SIZE-1));
	    info->xmit_cnt += c;
            MX_UNLOCK(info, flags);
	    buf += c;
	    count -= c;
	    ret += c;
	}
	up(&tmp_buf_sem);
    } else {
	MX_LOCK(info, flags);
	while (1) {
	    c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1, 
			SERIAL_XMIT_SIZE - info->xmit_head));
	        
	    if (c <= 0)
		break;

	    memcpy(info->xmit_buf + info->xmit_head, buf, c);
	    info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE-1);
	    info->xmit_cnt += c;
	    buf += c;
	    count -= c;
	    ret += c;
	}
        MX_UNLOCK(info, flags);
    }

    info->idle_stats.xmit_bytes += ret;
    info->idle_stats.xmit_idle   = jiffies;

    if (info->xmit_cnt && !tty->stopped && !tty->hw_stopped) {
        start_xmit(info);
    }
    return ret;
} /* mx_write */


/*
 * This routine is called by the kernel to write a single
 * character to the tty device.  If the kernel uses this routine,
 * it must call the flush_chars() routine (if defined) when it is
 * done stuffing characters into the driver.  If there is no room
 * in the queue, the character is ignored.
 */
static void
mx_put_char(struct tty_struct *tty, unsigned char ch)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
    unsigned long flags;

#ifdef MX_DEBUG_IO
    printk("mxc: mx_put_char mxc%d\n", info->line);
#endif

    if (serial_paranoia_check(info, tty->device, "mx_put_char"))
        return;

    if (!tty || !info->xmit_buf)
        return;

    MX_LOCK(info, flags);
    if (info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) {
        MX_UNLOCK(info, flags);
	return;
    }

    info->xmit_buf[info->xmit_head++] = ch;
    info->xmit_head &= SERIAL_XMIT_SIZE - 1;
    info->xmit_cnt++;
    info->idle_stats.xmit_bytes++;
    info->idle_stats.xmit_idle = jiffies;
    MX_UNLOCK(info, flags);
} /* mx_put_char */


/*
 * This routine is called by the kernel after it has written a
 * series of characters to the tty device using put_char().  
 */
static void
mx_flush_chars(struct tty_struct *tty)
{
  struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
                                
#ifdef MX_DEBUG_IO
    printk("mxc: mx_flush_chars mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "mx_flush_chars"))
        return;

    if (info->xmit_cnt <= 0 || tty->stopped||
	tty->hw_stopped || !info->xmit_buf)
        return;

    start_xmit(info);
} /* mx_flush_chars */


/*
 * This routine returns the numbers of characters the tty driver
 * will accept for queuing to be written.  This number is subject
 * to change as output buffers get emptied, or if the output flow
 * control is activated.
 */
static int
mx_write_room(struct tty_struct *tty)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
    int   ret;
                                
#ifdef MX_DEBUG_IO
    printk("mxc: mx_write_room mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "mx_write_room"))
        return 0;

    ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
    if (ret < 0)
        ret = 0;

    return ret;
} /* mx_write_room */


static int
mx_chars_in_buffer(struct tty_struct *tty)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;
    int card, channel;
                                
    if (serial_paranoia_check(info, tty->device, "mx_chars_in_buffer"))
        return 0;

    card = info->card;
    channel = (info->line) - (mx_card[card].first_line);


#ifdef MX_DEBUG_IO
    printk("mxc: mx_chars_in_buffer mxc%d %d\n",
	   info->line, info->xmit_cnt); /* */
#endif
    return info->xmit_cnt;
} /* mx_chars_in_buffer */


/*
 * ------------------------------------------------------------
 * mx_ioctl() and friends
 * ------------------------------------------------------------
 */

/*
 * This routine finds or computes the various line characteristics.
 * It used to be called config_setup
 */
static void
set_line_char(struct mswitch_port * info)
{
    unsigned long flags;
    unsigned long base_addr;
    int card,chip,lc,channel,iport,chan;
    unsigned cflag, iflag;
    int baud;
    ulong csr, mr1, mr2;

    if (!info->tty || !info->tty->termios){
        return;
    }
    if (info->line == -1){
        return;
    }
    cflag = info->tty->termios->c_cflag;
    iflag = info->tty->termios->c_iflag; /* Xon/Xoff XXX */

    card = info->card;
    base_addr = mx_card[card].base_addr;
    channel = (info->line) - (mx_card[card].first_line);
    lc = channel / NR_LC_PORTS;
    iport = channel % NR_LC_PORTS;
    chip = SC28_LC_CHIP(iport);
    chan = SC28_LC_CHAN(iport);

    /* baud rate */
    baud = tty_get_baud_rate(info->tty);
    switch(baud) {
    case 2400:
        csr = SC28_CSR_2400;
	break;
    case 3600:
        csr = SC28_CSR_3600;
	break;
    case 4800:
        csr = SC28_CSR_4800;
	break;
    case 7200:
        csr = SC28_CSR_7200;
	break;
    case 9600:
        csr = SC28_CSR_9600;
	break;
    case 14400:
        csr = SC28_CSR_14400;
	break;
    case 19200:
        csr = SC28_CSR_19200;
	break;
    case 28800:
        csr = SC28_CSR_28800;
	break;
    case 38400:
        csr = SC28_CSR_38400;
	break;
    case 57600:
        csr = SC28_CSR_57600;
	break;
    case 115200:
        csr = SC28_CSR_115200;
	break;
    case 230400:
        csr = SC28_CSR_230400;
	break;
    default:
        csr = SC28_CSR_9600;
    }

    /* bits per character */
    switch(cflag & CSIZE){
    case CS5:
        mr1 = 0;
	break;
    case CS6:
        mr1 = SC28_MR1_6;
	break;
    case CS7:
        mr1 = SC28_MR1_7;
	break;
    case CS8:
    default:
        mr1 = SC28_MR1_8;
	break;
    }

    /* parity */
    if (cflag & PARENB) {
        if (cflag & PARODD) {
	    mr1 |= SC28_MR1_ODD;
	}
    } else {
        mr1 |= SC28_MR1_NONE;
    }

    /* stop bits */
    if (cflag & CSTOPB) {
        mr2 = SC28_MR2_2;
    } else {
        mr2 = 0;
    }

    if (chip == LCCI_POE_UART) {
        /* POE only supports 19200 8N1 */
        csr = SC28_CSR_19200;
	mr1 = SC28_MR1_8N; /* 8 bits, no parity */
	mr2 = 0; /* 1 stop bit */
    }

    /* set registers */
    MX_LOCK(info, flags);
    lcci_write_spoe(base_addr, lc, chip, SC28_MR1(chan), mr1);
    lcci_write_spoe(base_addr, lc, chip, SC28_MR2(chan), mr2);
    lcci_write_spoe(base_addr, lc, chip, SC28_TxCSR(chan), csr);
    lcci_write_spoe(base_addr, lc, chip, SC28_RxCSR(chan), csr);
    MX_UNLOCK(info, flags);
} /* set_line_char */


static int
get_serial_info(struct mswitch_port * info,
		struct serial_struct * retinfo)
{
  struct serial_struct tmp;
  struct mswitch_card *cinfo = &mx_card[info->card];

    if (!retinfo)
            return -EFAULT;
    memset(&tmp, 0, sizeof(tmp));
    tmp.type = info->type;
    tmp.line = info->line;
    tmp.port = info->card * 0x100 + info->line - cinfo->first_line;
    tmp.irq = cinfo->irq;
    tmp.flags = info->flags;
    tmp.close_delay = info->close_delay;
    tmp.baud_base = info->baud;
    tmp.custom_divisor = info->custom_divisor;
    tmp.hub6 = 0;               /*!!!*/
    return copy_to_user(retinfo,&tmp,sizeof(*retinfo))?-EFAULT:0;
} /* get_serial_info */


static int
set_serial_info(struct mswitch_port * info,
		struct serial_struct * new_info)
{
   struct serial_struct new_serial;
   struct mswitch_port old_info;

    if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
	return -EFAULT;
    old_info = *info;

    if (!capable(CAP_SYS_ADMIN)) {
            if ((new_serial.close_delay != info->close_delay) ||
		(new_serial.baud_base != info->baud) ||
		((new_serial.flags & ASYNC_FLAGS & ~ASYNC_USR_MASK) !=
		 (info->flags & ASYNC_FLAGS & ~ASYNC_USR_MASK)))
                    return -EPERM;
            info->flags = ((info->flags & ~ASYNC_USR_MASK) |
                           (new_serial.flags & ASYNC_USR_MASK));
            info->baud = new_serial.baud_base;
	    info->custom_divisor = new_serial.custom_divisor;
            goto check_and_exit;
    }


    /*
     * OK, past this point, all the error checking has been done.
     * At this point, we start making changes.....
     */

    info->baud = new_serial.baud_base;
    info->custom_divisor = new_serial.custom_divisor;
    info->flags = ((info->flags & ~ASYNC_FLAGS) |
                    (new_serial.flags & ASYNC_FLAGS));
    info->close_delay = new_serial.close_delay * HZ/100;
    info->closing_wait = new_serial.closing_wait * HZ/100;

check_and_exit:
    if (info->flags & ASYNC_INITIALIZED){
        set_line_char(info);
        return 0;
    }else{
        return startup(info);
    }
} /* set_serial_info */


/*
 * mx_break() --- routine which turns the break handling on or off
 */
static void
mx_break(struct tty_struct *tty, int break_state)
{
    struct mswitch_port * info = (struct mswitch_port *)tty->driver_data;
    unsigned long flags;

    if (serial_paranoia_check(info, tty->device, "mx_break"))
	return;

    MX_LOCK(info, flags);
        /* Let the transmit ISR take care of this (since it
	   requires stuffing characters into the output stream).
        */
	if (break_state == -1) {
	    if (!info->breakon) {
		info->breakon = 1;
		if (!info->xmit_cnt) {
		    MX_UNLOCK(info, flags);
		    start_xmit(info);
		    MX_LOCK(info, flags);
		}
	    }
	} else {
	    if (!info->breakoff) {
		info->breakoff = 1;
		if (!info->xmit_cnt) {
		    MX_UNLOCK(info, flags);
		    start_xmit(info);
		    MX_LOCK(info, flags);
		}
	    }
	}
    MX_UNLOCK(info, flags);
} /* mx_break */

static int
get_mon_info(struct mswitch_port * info, struct mswitch_monitor * mon)
{

    if(copy_to_user(mon, &info->mon, sizeof(struct mswitch_monitor)))
    	return -EFAULT;
    info->mon.int_count  = 0;
    info->mon.char_count = 0;
    info->mon.char_max   = 0;
    info->mon.char_last  = 0;
    return 0;
}/* get_mon_info */

static int
mx_scci_read(struct mswitch_port * info,
	     scci_ioctl_t *u)
{
    scci_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    SCCI_REG_READ(mx_card[info->card].base_addr, k.reg, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mx_scci_write(struct mswitch_port * info,
	      scci_ioctl_t *u)
{
    scci_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    SCCI_REG_WRITE(mx_card[info->card].base_addr, k.reg, k.data);
    return 0;
}

static int
mx_lcci_read(struct mswitch_port * info,
	     lcci_ioctl_t *u)
{
    lcci_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	lcci_read(mx_card[info->card].base_addr, k.lc, k.reg, &k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
mx_lcci_write(struct mswitch_port * info,
	      lcci_ioctl_t *u)
{
    lcci_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	lcci_write(mx_card[info->card].base_addr, k.lc, k.reg, k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    return rc;
}

static int
mx_lcci_i2c_read(struct mswitch_port * info,
                 lcci_spoe_ioctl_t *u)
{
    lcci_spoe_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	lcci_read_i2c(mx_card[info->card].base_addr, k.lc,
                      k.dev, k.reg, &k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}


static int
mx_ap_led_read(struct mswitch_port * info,
               mxc_apled_t *u)
{
    mxc_apled_t k;
    ulong flags;
    ulong data;
    ulong subport;
    ulong reg;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    reg = LCCI_AP_LED_0 + k.port / 8;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	(k.port > LCCI_SOE_PORTS) ||
	lcci_read(mx_card[info->card].base_addr, k.lc, reg, &data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    subport = k.port & 0x7;
    k.color = (data & LCCI_LED_COLOR_MASK(subport)) >> LCCI_LED_COLOR_SHIFT(subport);
    k.state = (data & LCCI_LED_STATE_MASK(subport)) >> LCCI_LED_STATE_SHIFT(subport);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static char lc_type[SCCI_LCS];

static int
mx_ap_led_write(struct mswitch_port * info,
		mxc_apled_t *u)
{
    mxc_apled_t k;
    ulong flags;
    ulong data;
    ulong subport;
    ulong reg;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    reg = LCCI_AP_LED_0 + k.port / 8;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
        ((lc_type[k.lc] != '0') && (lc_type[k.lc] != '1')) ||
	(k.port >= LCCI_SOE_PORTS) ||
	(k.color > MXC_AP_LED_COLOR_YELLOW) ||
	(k.state > MXC_AP_LED_STATE_BLINK) ||
	lcci_read(mx_card[info->card].base_addr, k.lc, reg, &data)) {
        rc = -ENODEV;
    } else {
        subport = k.port & 0x7;
	data &= ~(LCCI_LED_COLOR_MASK(subport)|LCCI_LED_STATE_MASK(subport));
	data |= ((k.color << LCCI_LED_COLOR_SHIFT(subport)) |
		 (k.state << LCCI_LED_STATE_SHIFT(subport)));
	lcci_write(mx_card[info->card].base_addr, k.lc, reg, data);
    }
    MX_UNLOCK(info, flags);
    return rc;
}

static int
mx_spoe_read(struct mswitch_port * info,
	     lcci_spoe_ioctl_t *u)
{
    lcci_spoe_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	!lcci_spoe_exists(mx_card[info->card].base_addr, k.lc) ||
	lcci_read_spoe(mx_card[info->card].base_addr, k.lc,
		       k.dev, k.reg, &k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
mx_spoe_write(struct mswitch_port * info,
	      lcci_spoe_ioctl_t *u)
{
    lcci_spoe_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	!lcci_spoe_exists(mx_card[info->card].base_addr, k.lc) ||
	lcci_write_spoe(mx_card[info->card].base_addr, k.lc,
			k.dev, k.reg, k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    return rc;
}

static int
mx_cxe_read(struct mswitch_port * info,
	    cxe_ioctl_t *u)
{
    cxe_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	cxe_read(mx_card[info->card].base_addr, k.lc, k.reg, &k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
mx_cxe_write(struct mswitch_port * info,
	     cxe_ioctl_t *u)
{
    cxe_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	cxe_write(mx_card[info->card].base_addr, k.lc, k.reg, k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    return rc;
}

static int
mx_phy_read(struct mswitch_port * info,
	    cxe_phy_ioctl_t *u)
{
    cxe_phy_ioctl_t k;
    ulong flags;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	cxe_read_phy(mx_card[info->card].base_addr, k.lc,
		     k.dev, k.reg, &k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;
    return rc;
}

static int
mx_phy_write(struct mswitch_port * info,
	     cxe_phy_ioctl_t *u)
{
    cxe_phy_ioctl_t k;
    int rc = 0;
    ulong flags;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	cxe_write_phy(mx_card[info->card].base_addr, k.lc,
		      k.dev, k.reg, k.data))
        rc = -ENODEV;
    MX_UNLOCK(info, flags);
    return rc;
}

static int
mx_cpld_read(struct mswitch_port * info,
	     sc_cpld_ioctl_t *u)
{
    sc_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    SC_CPLD_REG_READ(mx_card[info->card].ctl_addr, k.reg, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mx_cpld_write(struct mswitch_port * info,
	      sc_cpld_ioctl_t *u)
{
    sc_cpld_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    /* no semaphore needed */
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, k.reg, k.data);
    return 0;
}

#define JTAG_WRITE(val) { \
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_OPERATION, \
		      val); \
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_OPERATION, \
		      val | SC_CPLD_JTAG_OPERATION_CLK); \
}

#define JTAG_READ(val, rc) { \
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_OPERATION, \
		      val | SC_CPLD_JTAG_OPERATION_CLK); \
    SC_CPLD_REG_READ(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_OPERATION, \
		     rc); \
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_OPERATION, \
		      val); \
}

static int
mx_fpga_prog(struct mswitch_port * info, int dev, int fpga,
	     __u8 *data, int len)
{
    __u8 *ptr;
    ulong val;
    char buf[128];
    int bit;
    int i;
    int rc = 0;

    /* enable JTAG programming */
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_SELECTION,
		      dev | SC_CPLD_JTAG_SELECTION_ENABLE);

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 5; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Enter Run-Test-Idle */
    JTAG_WRITE(0);
    /* 3. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 4. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    if (dev != SC_CPLD_JTAG_SELECTION_LOCAL) { /* LCCI */
        /* 5. Load CFG-IN */
        for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    JTAG_WRITE(0);
	}
	/* 6. Load last bit of CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    } else if (fpga) {
        /* 5. Load CFG-IN */
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(0);
	/* Load bypass to other FPGA */
	for (i = 0; i < 4; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 6. Load last bit of bypass */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
    } else { /* SCCI */
        /* Load bypass to other FPGA */
        for (i = 0; i < 5; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 5. Load CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	/* 6. Load last bit of CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 7. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 8. Enter SHIFT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    if (fpga) {
        /* must shift in 31 zeros (32 - device) first */
        for (i = 0; i < 31; i++) {
	    JTAG_WRITE(0);
	}
    }

    /* 9. Send data */
    ptr = buf;
    for (i = 0; i < len-1; i++, ptr++, data++) {
        if ((i % sizeof(buf)) == 0) {
	    if (copy_from_user(buf, data, MIN(sizeof(buf), (len-i)))) {
	        rc = -EFAULT;
		goto error;
	    }
	    ptr = buf;
	}
        for (bit = 7; bit >= 0; bit--) {
	    val = ((*ptr >> bit) & 0x1) << 1;
	    JTAG_WRITE(val);
	}
    }
    /* 10. Send last byte and set TMS for last bit */
    if ((i % sizeof(buf)) == 0) {
        if (copy_from_user(buf, data, MIN(sizeof(buf), (len-i)))) {
	    rc = -EFAULT;
	    goto error;
	}
	ptr = buf;
    }
    for (bit = 7; bit > 0; bit--) {
        val = ((*ptr >> bit) & 0x1) << 1;
	JTAG_WRITE(val);
    }
    /* last bit */
    val = (*ptr & 0x1) << 1;
    if (!fpga) {
        val |= SC_CPLD_JTAG_OPERATION_TMS;
    }
    JTAG_WRITE(val);

    if (fpga) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
    }

    /* 11. Enter UPDATE-DR */
    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    /* 12. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 13. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    /* 14. Load JSTART */
    if (dev != SC_CPLD_JTAG_SELECTION_LOCAL) { /* LCCI */
        for (i = 0; i < 2; i++) {
	    JTAG_WRITE(0);
	}
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 15. Load last bit of JSTART */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    } else if (fpga) { /* Crypto */
        /* 14. Load JSTART */
        JTAG_WRITE(0);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	/* Load bypass to other FPGA */
	for (i = 0; i < 4; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 15. Load last bit of bypass */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
    } else { /* SCCI */
        /* Load bypass to other FPGA */
        for (i = 0; i < 5; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 14. Load JSTART */
	JTAG_WRITE(0);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	/* 15. Load last bit of JSTART */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 16. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 17. Enter SHIFT-DR and clock STARTUP */
    for (i = 0; i < 14; i++) {
        JTAG_WRITE(0);
    }
    /* 18. Enter UPDATE-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* Enter Run-Test-Idle */
    JTAG_WRITE(0);
    
    /* disable JTAG programming */
 error:
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr,
		      SC_CPLD_JTAG_SELECTION, 0);
    return rc;
}

static int
mx_fpga_id(struct mswitch_port * info, int dev, int magic, ulong *id)
{
    ulong val;
    ulong tmp;
    int i;

    /* enable JTAG programming */
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_JTAG_SELECTION,
		      dev | SC_CPLD_JTAG_SELECTION_ENABLE);

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 12; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Send 12 1's */
    for (i = 0; i < 12; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS|SC_CPLD_JTAG_OPERATION_TDI);
    }
    /* 3. Send magic in reverse order */
    for (i = 0; i < 8; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS|
		   ((magic & (1 << i))? SC_CPLD_JTAG_OPERATION_TDI:0));
    }
    /* 4. Read ID */
    tmp = 0;
    for (i = 0; i < 8; i++) {
        JTAG_READ(SC_CPLD_JTAG_OPERATION_TMS, val);
	tmp |= ((val & SC_CPLD_JTAG_OPERATION_TDO)? 1:0) << i;
    }

    /* disable JTAG programming */
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr,
		      SC_CPLD_JTAG_SELECTION, 0);
    *id = tmp;
    return 0;
}

static int
mx_scci_erase(struct mswitch_port * info)
{
    ulong status;
    int i;
    
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_FPGA_INTERFACE,
		      SC_CPLD_FPGA_INTERFACE_PRO);
    i = 0;
    while (1) {
        udelay(1); /* wait at least one us */
        SC_CPLD_REG_READ(mx_card[info->card].ctl_addr, SC_CPLD_FPGA_INTERFACE,
			 status);
	if (status & SC_CPLD_FPGA_INTERFACE_INIT_L) break;
	if (i++ > 100000) {
	    return 1;
	}
    }
    return 0;
}

static int
mx_crypto_erase(struct mswitch_port * info)
{
    ulong status;
    int i;
    
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_FPGA_INTERFACE,
		      SC_CPLD_FPGA_INTERFACE_CRYPTO_PRO);
    i = 0;
    while (1) {
        udelay(1); /* wait at least one us */
        SC_CPLD_REG_READ(mx_card[info->card].ctl_addr, SC_CPLD_FPGA_INTERFACE,
			 status);
	if (status & SC_CPLD_FPGA_INTERFACE_INIT_L) break;
	if (i++ > 100000) {
	    return 1;
	}
    }
    return 0;
}

static int
mx_scci_prog(struct mswitch_port * info,
	     scci_ioctl_t *u)
{
    scci_ioctl_t k;
    int rc = 0;
    int dev;

    if (copy_from_user(&k, u, sizeof(k)))
	return -EFAULT;

    /* erase FPGA */
    if (mx_scci_erase(info)) 
        return -EFAULT;
    /* program FPGA */
    dev = SC_CPLD_JTAG_SELECTION_LOCAL;
    if (mx_fpga_prog(info, dev, 0, (__u8 *)k.reg, k.data))
        rc = -EFAULT;

    return rc;
}

static int
mx_crypto_prog(struct mswitch_port * info,
	       scci_ioctl_t *u)
{
    scci_ioctl_t k;
    int rc = 0;
    int dev;

    if (copy_from_user(&k, u, sizeof(k)))
	return -EFAULT;

    /* erase FPGA */
    if (mx_crypto_erase(info)) 
        return -EFAULT;
    /* program FPGA */
    dev = SC_CPLD_JTAG_SELECTION_LOCAL;
    if (mx_fpga_prog(info, dev, 1, (__u8 *)k.reg, k.data))
        rc = -EFAULT;

    return rc;
}

static int
mx_arp_add(struct mswitch_port * info, arp_entry_t *u)
{   
    return (arp_entry_add(u));
}

static int
mx_arp_del(struct mswitch_port * info, arp_entry_t *u)
{   
    return (arp_entry_del(u));
}

static int
mx_arp_clear_all(struct mswitch_port * info, arp_entry_t *u)
{   
    return (arp_clear_all());
}

static int
mx_lcci_prog(struct mswitch_port * info,
	     lcci_ioctl_t *u)
{
    lcci_ioctl_t k;
    int rc = 0;
    int dev;
    ulong val;

    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;

    dev = k.lc + SC_CPLD_JTAG_SELECTION_LC0;
    /* erase FPGA */
    if (mx_fpga_id(info, dev, 0xa8, &val))
        return -EFAULT;
    /* program FPGA */
    if (mx_fpga_prog(info, dev, 0, (__u8 *)k.reg, k.data))
        rc = -EFAULT;

    return rc;
}

static int
mx_lcci_id(struct mswitch_port * info,
	   lcci_ioctl_t *u)
{
    lcci_ioctl_t k;
    int rc = 0;
    int dev;

    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;

    dev = k.lc + SC_CPLD_JTAG_SELECTION_LC0;
    if (mx_fpga_id(info, dev, 0x55, &k.data))
        rc = -EFAULT;

    if (copy_to_user(u, &k, sizeof(k)))
        rc = -EFAULT;

    return rc;
}

static int
mx_sw_version(struct mswitch_port * info,
	      mxc_string_t *u)
{
    mxc_string_t k;
    if (strlen(ARUBA_VERSION)+1 >= sizeof(k.string))
        return -EFAULT;
    sprintf(k.string, "%s", ARUBA_VERSION);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_platid_get(struct mswitch_port * info,
	      mxc_string_t *u)
{
    mxc_string_t k;
    if (strlen(__mxc_platid)+1 >= sizeof(k.string))
        return -EFAULT;
    sprintf(k.string, "%s", __mxc_platid);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_platid_set(struct mswitch_port * info,
	      mxc_string_t *u)
{
    mxc_string_t k;
    if (copy_from_user(&k, u, sizeof(k)))
	return -EFAULT;
    snprintf(__mxc_platid, sizeof(__mxc_platid), "%s", k.string);
    return 0;
}

static int
mx_sc_id(struct mswitch_port * info,
	 mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_ID, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_lc_id(struct mswitch_port * info,
	 mxc_lc_t *u)
{
    mxc_lc_t k;
    ulong sc_status;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS,
		  sc_status);
    if (!(SCCI_MODULE_LC_PRES(k.lc) & sc_status))
        return -ENODEV;
    lcci_read(mx_card[info->card].base_addr, k.lc, LCCI_ID, &k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_lc_type(struct mswitch_port * info,
           mxc_lc_t *u)
{
    mxc_lc_t k;
    ulong sc_status;
    int reg;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS,
		  sc_status);
    if (!(SCCI_MODULE_LC_PRES(k.lc) & sc_status))
        return -ENODEV;
    if (lcci_read_i2c(mx_card[info->card].base_addr, k.lc,
                      0x50, 1, &k.data))
        return -EFAULT;
    /*
     * Due to inconsistency in the way Assembly numbers are programmed by 
     * HW and OP's, we have to use a little different logic than before.
     * If the value of the 1st char is '2', then it is programmed by OP.
     * If the value of the 1st char is '0' then it is (W) programmed by HW.
     * If it is by OP read the 7th Char. If it is by HW read the 5 th char.
     */
    if (k.data == '2')
      reg = 7;
    else
      reg = 5;
    if (lcci_read_i2c(mx_card[info->card].base_addr, k.lc,
                      0x50, reg, &k.data))
        return -EFAULT;
    lc_type[k.lc] = (char) k.data;
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_lc_switch_rev(struct mswitch_port * info,
                 mxc_lc_t *u)
{
    ulong flags;
    mxc_lc_t k;
    int rc = 0;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    MX_LOCK(info, flags);
    if (!lcci_exists(mx_card[info->card].base_addr, k.lc) ||
	cxe_read(mx_card[info->card].base_addr, k.lc, CXE_CG_DEVICE_ID, &k.data))
        rc = -ENODEV;
    k.data = CXE_CG_DEVICE_ID_REV(k.data);
    MX_UNLOCK(info, flags);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_lc_status(struct mswitch_port * info,
	     mxc_lc_t *u)
{
    mxc_lc_t k;
    ulong sc_status;
    ulong lc_status = 0;
    int sc;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.lc > MXC_MAX_LCS)
	return -ENODEV;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS,
		  sc_status);
    k.data = 0;
    if (SCCI_MODULE_LC_PRES(k.lc) & sc_status) {
        k.data |= MXC_LC_PRES;
	lcci_read(mx_card[info->card].base_addr, k.lc, LCCI_MODULE_STATUS,
		  &lc_status);
	if (lc_status & LCCI_MEZZ_PRES) {
	    k.data |= MXC_LC_MEZZ_PRES;
	}
	if (lc_status & LCCI_SPOE_PRES) {
	    k.data |= MXC_LC_SPOE_PRES;
	}
	for (sc = 0; sc < MXC_MAX_SCS; sc++) {
	    if (lc_status & LCCI_SC_PRES(sc)) {
	        k.data |= MXC_LC_SC_PRES(sc);
	    }
	    if (lc_status & LCCI_SC_ACTIVE(sc)) {
	        k.data |= MXC_LC_SC_ACTIVE(sc);
	    }
	}
    } 
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_fan_status(struct mswitch_port * info,
	      mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    ulong sc_status;
    int i;
    k.data = 0;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS, sc_status);
    if (SCCI_MODULE_FAN_PRES & sc_status) {
        for (i = 0; i < MXC_MAX_FANS; i++) {
	    k.data |= MXC_FAN_PRES(i);
	    if (SCCI_MODULE_FAN_OK(i) & sc_status) {
	        k.data |= MXC_FAN_OK(i);
	    }
	    if (SCCI_MODULE_FAN_FAST & sc_status) {
	        k.data |= MXC_FAN_FAST(i);
	    }
	}
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_ps_status(struct mswitch_port * info,
	     mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    ulong sc_status;
    int i;
    k.data = 0;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS, sc_status);
    if (SCCI_MODULE_PS_TYPE & sc_status) {
        k.data |= MXC_PS_TYPE; /* "1" for Power One Power supplies else XP "0"*/
    }
    for (i = 0; i < MXC_MAX_PSS; i++) {
        if (SCCI_MODULE_PS_PRES(i) & sc_status) {
	    k.data |= MXC_PS_PRES(i);
	}
        if (SCCI_MODULE_PS_OK(i) & sc_status) {
	    k.data |= MXC_PS_OK(i);
	}
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_temperature(struct mswitch_port * info,
	       mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    k.data = 0; /* XXX */
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_sc_slot(struct mswitch_port * info,
	   mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    ulong sc_status;
    k.data = MXC_SC_PEER_SUPP;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_MODULE_STATUS, sc_status);
    k.data |= SCCI_MODULE_ID(sc_status) & 0x1;
    if (sc_status & SCCI_MODULE_PSC_PRES) {
        k.data |= MXC_SC_PEER_PRES;
    }
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_wd_reset(struct mswitch_port * info)
{
    SC_CPLD_REG_WRITE(mx_card[info->card].ctl_addr, SC_CPLD_WATCHDOG, 0x5);
    return 0;
}

static int
mx_nvram_read(struct mswitch_port * info,
	      mxc_nvram_t *u)
{
    mxc_nvram_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset >= NVRAM_SIZE)
	return -EFAULT;
    if (copy_to_user(k.data, (char *)NVRAM_BASE_ADDRESS + k.offset, k.length))
        return -EFAULT;
    return 0;
}

static int
mx_nvram_write(struct mswitch_port * info,
	      mxc_nvram_t *u)
{
    mxc_nvram_t k;
    volatile char *ptr; 
    char *buffer;
    int i;

    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.offset + k.length >= NVRAM_SIZE)
	return -EFAULT;

    buffer = kmalloc(k.length,GFP_KERNEL);
    if(!buffer){
	    return -EFAULT;
    }
    if (copy_from_user(buffer, k.data, k.length))
        return -EFAULT;

    ptr = (volatile char *)(NVRAM_BASE_ADDRESS + k.offset);
    for(i=0;i<k.length;i++){
	    ptr[i] = buffer[i];
    }

    kfree(buffer);

    return 0;
}

static int
mx_panic_read(struct mswitch_port * info,
	      mxc_panic_t *u)
{
    mxc_panic_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    memcpy(k.record,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
	   NVRAM_PANIC_SIZE);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mx_panic_clear(struct mswitch_port * info,
	       mxc_panic_t *u)
{
    mxc_panic_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    if (k.block >= NVRAM_PANICS)
	return -EFAULT;
    memset((char *)NVRAM_BASE_ADDRESS + NVRAM_PANIC_OFFSET(k.block),
	   0, NVRAM_PANIC_SIZE);
    return 0;
}

static int
mx_reboot_str_read(struct mswitch_port * info,
		   mxc_string_t *u)
{
    mxc_string_t k;
    memcpy(k.string,
	   (char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
	   NVRAM_REBOOT_STRING_SIZE);
    if (copy_to_user(u, &k, sizeof(k)))
        return -EFAULT;
    return 0;
}

static int
mx_reboot_str_write(struct mswitch_port * info,
		    mxc_string_t *u)
{
    mxc_string_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    memcpy((char *)NVRAM_BASE_ADDRESS + NVRAM_REBOOT_STRING_OFFSET,
	   k.string,
	   NVRAM_REBOOT_STRING_SIZE);	   
    return 0;
}

static int
mx_scratch_read(struct mswitch_port * info,
		mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    SCCI_REG_READ(mx_card[info->card].base_addr, SCCI_SCRATCHPAD, k.data);
    if (copy_to_user(u, &k, sizeof(k)))
	return -EFAULT;
    return 0;
}

static int
mx_scratch_write(struct mswitch_port * info,
		 mxc_ioctl_t *u)
{
    mxc_ioctl_t k;
    if (copy_from_user(&k, u,sizeof(k)))
	return -EFAULT;
    SCCI_REG_WRITE(mx_card[info->card].base_addr, SCCI_SCRATCHPAD, k.data);
    return 0;
}

/*
 * This routine allows the tty driver to implement device-
 * specific ioctl's.  If the ioctl number passed in cmd is
 * not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int
mx_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
    struct mswitch_port * info = (struct mswitch_port *)tty->driver_data;
    int ret_val = 0;

    if (serial_paranoia_check(info, tty->device, "mx_ioctl"))
	return -ENODEV;

#ifdef MX_DEBUG_IOCTL
    printk("mxc: mx_ioctl mxc%d, cmd = %x arg = %lx\n",
	   info->line, cmd, arg); /* */
#endif

    switch (cmd) {
    case MXGETMON:
        ret_val = get_mon_info(info, (struct mswitch_monitor *)arg);
	break;
    case MXGETCARDINFO:
        if (copy_to_user((void *)arg, (void *)&mx_card[info->card], 
			 sizeof (struct mswitch_card))) {
	    ret_val = -EFAULT;
	    break;
	}
	ret_val = 0;
	break;
    case MXSETWAIT:
        info->closing_wait = (unsigned short)arg * HZ/100;
	ret_val = 0;
	break;
    case MXGETWAIT:
        ret_val = info->closing_wait / (HZ/100);
	break;
    case MXSCCIREAD:
        ret_val = mx_scci_read(info, (scci_ioctl_t *) arg);
        break;
    case MXSCCIWRITE:
        ret_val = mx_scci_write(info, (scci_ioctl_t *) arg);
        break;
    case MXLCCIREAD:
        ret_val = mx_lcci_read(info, (lcci_ioctl_t *) arg);
        break;
    case MXLCCIWRITE:
        ret_val = mx_lcci_write(info, (lcci_ioctl_t *) arg);
        break;
    case MXSPOEREAD:
        ret_val = mx_spoe_read(info, (lcci_spoe_ioctl_t *) arg);
        break;
    case MXSPOEWRITE:
        ret_val = mx_spoe_write(info, (lcci_spoe_ioctl_t *) arg);
        break;
    case MXCXEREAD:
        ret_val = mx_cxe_read(info, (cxe_ioctl_t *) arg);
        break;
    case MXCXEWRITE:
        ret_val = mx_cxe_write(info, (cxe_ioctl_t *) arg);
        break;
    case MXLCCII2CREAD:
        ret_val = mx_lcci_i2c_read(info, (lcci_spoe_ioctl_t *) arg);
        break;
    case MXPHYREAD:
        ret_val = mx_phy_read(info, (cxe_phy_ioctl_t *) arg);
        break;
    case MXPHYWRITE:
        ret_val = mx_phy_write(info, (cxe_phy_ioctl_t *) arg);
        break;
    case MXCPLDREAD:
        ret_val = mx_cpld_read(info, (sc_cpld_ioctl_t *) arg);
        break;
    case MXCPLDWRITE:
        ret_val = mx_cpld_write(info, (sc_cpld_ioctl_t *) arg);
        break;
    case MXSCCIPROG:
        ret_val = mx_scci_prog(info, (scci_ioctl_t *) arg);
        break;
    case MXLCCIID:
        ret_val = mx_lcci_id(info, (lcci_ioctl_t *) arg);
        break;
    case MXLCCIPROG:
        ret_val = mx_lcci_prog(info, (lcci_ioctl_t *) arg);
        break;
    case MXCRYPTOPROG:
        ret_val = mx_crypto_prog(info, (scci_ioctl_t *) arg);
        break;
    case MXARPENTRYADD:
        ret_val = mx_arp_add(info, (arp_entry_t *) arg);
        break;
    case MXARPENTRYDEL:
        ret_val = mx_arp_del(info, (arp_entry_t *) arg);
        break;
    case MXARPCLEARALL:
        ret_val = mx_arp_clear_all(info, (arp_entry_t *) arg);
        break;
    case MXAPLEDREAD:
        ret_val = mx_ap_led_read(info, (mxc_apled_t *) arg);
	break;
    case MXAPLEDWRITE:
        ret_val = mx_ap_led_write(info, (mxc_apled_t *) arg);
	break;
 
	/* status ioctls */
    case MXSWVERSION:
        ret_val = mx_sw_version(info, (mxc_string_t *) arg);
	break;
    case MXSCID:
        ret_val = mx_sc_id(info, (mxc_ioctl_t *) arg);
	break;
    case MXFANSTATUS:
        ret_val = mx_fan_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXPSSTATUS:
        ret_val = mx_ps_status(info, (mxc_ioctl_t *) arg);
	break;
    case MXTEMPERATURE:
        ret_val = mx_temperature(info, (mxc_ioctl_t *) arg);
	break;
    case MXSCSLOT:
        ret_val = mx_sc_slot(info, (mxc_ioctl_t *) arg);
	break;
    case MXLCID:
        ret_val = mx_lc_id(info, (mxc_lc_t *) arg);
	break;
    case MXLCSWITCHREV:
        ret_val = mx_lc_switch_rev(info, (mxc_lc_t *) arg);
	break;
    case MXLCTYPE:
        ret_val = mx_lc_type(info, (mxc_lc_t *) arg);
	break;
    case MXLCSTATUS:
        ret_val = mx_lc_status(info, (mxc_lc_t *) arg);
	break;

    case MXPLATIDGET:
        ret_val = mx_platid_get(info, (mxc_string_t *) arg);
	break;
    case MXPLATIDSET:
        ret_val = mx_platid_set(info, (mxc_string_t *) arg);
	break;

	/* reboot ioctls */
    case MXWDRESET:
        ret_val = mx_wd_reset(info);
        break;
    case MXPANIC:
        panic("test panic");
        break;
    case MXNULLPTR:
        *(char *)0x0 = 1;
	printk("NULLPTR ptr val = 0x%x\n", *(char *)0x0);
        break;
    case MXHANG:
        while(1);
        break;

	/* "NVRAM" access ioctls */
    case MXNVRAMREAD:
        ret_val = mx_nvram_read(info, (mxc_nvram_t *) arg);
        break;
    case MXNVRAMWRITE:
        ret_val = mx_nvram_write(info, (mxc_nvram_t *) arg);
        break;
    case MXPANICREAD:
        ret_val = mx_panic_read(info, (mxc_panic_t *) arg);
        break;
    case MXPANICCLEAR:
        ret_val = mx_panic_clear(info, (mxc_panic_t *) arg);
        break;
    case MXREBOOTSTRREAD:
        ret_val = mx_reboot_str_read(info, (mxc_string_t *) arg);
        break;
    case MXREBOOTSTRWRITE:
        ret_val = mx_reboot_str_write(info, (mxc_string_t *) arg);
        break;

	/* scratch ioctls */
    case MXSCRATCHREAD:
        ret_val = mx_scratch_read(info, (mxc_ioctl_t *) arg);
        break;
    case MXSCRATCHWRITE:
        ret_val = mx_scratch_write(info, (mxc_ioctl_t *) arg);
	break;

	/* serial ioctls */
    case TIOCGSERIAL:
        ret_val = get_serial_info(info, (struct serial_struct *) arg);
	break;
    case TIOCSSERIAL:
        ret_val = set_serial_info(info, (struct serial_struct *) arg);
	break;
    default:
        ret_val = -ENOIOCTLCMD;
    }

#ifdef MX_DEBUG_IOCTL
    printk(" mxc: mx_ioctl done %x\n", ret_val);
#endif

    return ret_val;
} /* mx_ioctl */


/*
 * This routine allows the tty driver to be notified when
 * device's termios settings have changed.  Note that a
 * well-designed tty driver should be prepared to accept the case
 * where old == NULL, and try to do something rational.
 */
static void
mx_set_termios(struct tty_struct *tty, struct termios * old_termios)
{
    struct mswitch_port *info = (struct mswitch_port *)tty->driver_data;

#ifdef MX_DEBUG_OTHER
    printk("mxc: mx_set_termios mxc%d\n", info->line);
#endif

    if ((tty->termios->c_cflag == old_termios->c_cflag) &&
	((tty->termios->c_iflag & (IXON|IXANY)) == 
	 (old_termios->c_iflag & (IXON|IXANY))))
        return;
    set_line_char(info);

    if ((old_termios->c_cflag & CRTSCTS) &&
        !(tty->termios->c_cflag & CRTSCTS)) {
            tty->hw_stopped = 0;
            mx_start(tty);
    }

    return;
} /* mx_set_termios */

/* This routine is called by the upper-layer tty layer to signal
   that incoming characters should be throttled because the input
   buffers are close to full.
 */
static void
mx_throttle(struct tty_struct * tty)
{
} /* mx_throttle */


/*
 * This routine notifies the tty driver that it should signal
 * that characters can now be sent to the tty without fear of
 * overrunning the input buffers of the line disciplines.
 */
static void
mx_unthrottle(struct tty_struct * tty)
{
} /* mx_unthrottle */


/* mx_start and mx_stop provide software output flow control as a
   function of XON/XOFF, software CTS, and other such stuff.
*/
static void
mx_stop(struct tty_struct *tty)
{
} /* mx_stop */


static void
mx_start(struct tty_struct *tty)
{
} /* mx_start */


static void
mx_flush_buffer(struct tty_struct *tty)
{
} /* mx_flush_buffer */


/*
 * mx_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void
mx_hangup(struct tty_struct *tty)
{
    struct mswitch_port * info = (struct mswitch_port *)tty->driver_data;
        
#ifdef MX_DEBUG_OTHER
    printk("mxc: mx_hangup mxc%d\n", info->line); /* */
#endif

    if (serial_paranoia_check(info, tty->device, "mx_hangup"))
        return;

    mx_flush_buffer(tty);
    shutdown(info);
    info->event = 0;
    info->count = 0;
#ifdef MX_DEBUG_COUNT
    printk("mxc: mx_hangup (%d): setting count to 0\n", current->pid);
#endif
    info->tty = 0;
    info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE);
    wake_up_interruptible(&info->open_wait);
} /* mx_hangup */


/*
 * This routine prints out the appropriate serial driver version number
 * and identifies which options were configured into this driver.
 */
static inline void
show_version(void)
{
    printk("Mswitch control driver, built %s %s\n",
	   __DATE__, __TIME__);
} /* show_version */

static int 
mswitch_get_proc_info(char *buf, char **start, off_t offset, int length,
		       int *eof, void *data)
{
    struct mswitch_port  *info;
    int i;
    int len=0;
    off_t begin=0;
    off_t pos=0;
    int size;
    __u32 cur_jifs = jiffies;

    size = sprintf(buf, "Dev TimeOpen   BytesOut  IdleOut    BytesIn   IdleIn  Overruns  Ldisc\n");

    pos += size;
    len += size;

    /* Output one line for each known port */
    for (i = 0; i < NR_PORTS && mx_port[i].line >= 0; i++) {
	info = &mx_port[i];

	if (info->count)
	    size = sprintf(buf+len,
			"%3d %8lu %10lu %8lu %10lu %8lu %9lu %6ld\n",
			info->line,
			JIFFIES_DIFF(info->idle_stats.in_use, cur_jifs) / HZ,
			info->idle_stats.xmit_bytes,
			JIFFIES_DIFF(info->idle_stats.xmit_idle, cur_jifs) / HZ,
			info->idle_stats.recv_bytes,
			JIFFIES_DIFF(info->idle_stats.recv_idle, cur_jifs) / HZ,
			info->idle_stats.overruns,
			(long) info->tty->ldisc.num);
	else
	    size = sprintf(buf+len,
			"%3d %8lu %10lu %8lu %10lu %8lu %9lu %6ld\n",
			info->line, 0L, 0L, 0L, 0L, 0L, 0L, 0L);
	len += size;
	pos = begin + len;

	if (pos < offset) {
	    len   = 0;
	    begin = pos;
	}
	if (pos > offset + length)
	    goto done;
    }
    *eof = 1;
done:
    *start = buf + (offset - begin);	/* Start of wanted data */
    len -= (offset - begin);		/* Start slop */
    if (len > length)
	len = length;			/* Ending slop */
    if (len < 0)
	len = 0;
    return len;
}

/* The serial driver boot-time initialization code!
    Hardware I/O ports are mapped to character special devices on a
    first found, first allocated manner.  That is, this code searches
    for Mxclom cards in the system.  As each is found, it is probed
    to discover how many chips (and thus how many ports) are present.
    These ports are mapped to the tty ports 32 and upward in monotonic
    fashion.  If an 8-port card is replaced with a 16-port card, the
    port mapping on a following card will shift.

    This approach is different from what is used in the other serial
    device driver because the Mxclom is more properly a multiplexer,
    not just an aggregation of serial ports on one card.

    If there are more cards with more ports than have been
    statically allocated above, a warning is printed and the
    extra ports are ignored.
 */

int __init
mx_init(void)
{
    struct mswitch_port  *info;
    struct mswitch_card *cinfo;
    int board,port,i;
    int nports;

    init_bh(CYCLADES_BH, do_mswitch_bh);

#ifdef MX_DEBUG_OPEN
    show_version();
#endif

    /* Initialize the tty_driver structure */
    
    memset(&mx_driver, 0, sizeof(struct tty_driver));
    mx_driver.magic = TTY_DRIVER_MAGIC;
    mx_driver.driver_name = "mswitch";
    mx_driver.name = "mxc";
    mx_driver.major = MSWITCH_MAJOR;
    mx_driver.minor_start = 0;
    mx_driver.num = NR_PORTS;
    mx_driver.type = TTY_DRIVER_TYPE_SERIAL;
    mx_driver.subtype = SERIAL_TYPE_NORMAL;
    mx_driver.init_termios = tty_std_termios;
    mx_driver.init_termios.c_cflag =
            B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    mx_driver.flags = TTY_DRIVER_REAL_RAW;
    mx_driver.refcount = &serial_refcount;
    mx_driver.table = serial_table;
    mx_driver.termios = serial_termios;
    mx_driver.termios_locked = serial_termios_locked;

    mx_driver.open = mx_open;
    mx_driver.close = mx_close;
    mx_driver.write = mx_write;
    mx_driver.put_char = mx_put_char;
    mx_driver.flush_chars = mx_flush_chars;
    mx_driver.write_room = mx_write_room;
    mx_driver.chars_in_buffer = mx_chars_in_buffer;
    mx_driver.flush_buffer = mx_flush_buffer;
    mx_driver.ioctl = mx_ioctl;
    mx_driver.throttle = mx_throttle;
    mx_driver.unthrottle = mx_unthrottle;
    mx_driver.set_termios = mx_set_termios;
    mx_driver.stop = mx_stop;
    mx_driver.start = mx_start;
    mx_driver.hangup = mx_hangup;
    mx_driver.break_ctl = mx_break;
    mx_driver.wait_until_sent = mx_wait_until_sent;
    mx_driver.read_proc = mswitch_get_proc_info;

    /* POE driver */
    mx_poe_driver = mx_driver;
    mx_poe_driver.name = "ttyPOE";
    mx_poe_driver.major = MSWITCHSPOE_MAJOR;

    /* SOE driver */
    mx_soe_driver = mx_driver;
    mx_soe_driver.name = "ttySOE";
    mx_soe_driver.major = MSWITCHSPOE_MAJOR;


    if (tty_register_driver(&mx_driver))
            panic("Couldn't register Mswitch serial driver\n");
    if (tty_register_driver(&mx_poe_driver))
            panic("Couldn't register Mswitch POE driver\n");
    if (tty_register_driver(&mx_soe_driver))
            panic("Couldn't register Mswitch SOE driver\n");

    for (i = 0; i < NR_CARDS; i++) {
            /* base_addr=0 indicates board not found */
            mx_card[i].base_addr = 0;
    }

    /* the code below is responsible to find the boards. Each different
       type of board has its own detection routine. If a board is found,
       the next mx_card structure available is set by the detection
       routine. These functions are responsible for checking the
       availability of mx_card and mx_port data structures */

    /* assume it always exists */
    mx_card[0].ctl_phys = SC_CPLD_BASE_ADDR;
    mx_card[0].ctl_addr = (ulong)ioremap(mx_card[0].ctl_phys, 0x100);
    mx_card[0].base_phys = SCCI_BASE_ADDR;
    mx_card[0].base_addr = (ulong)ioremap(mx_card[0].base_phys, 0x100);
    mx_card[0].irq = 19; /* mswitch_setup.c */
    mx_card[0].num_chips = 9; /* 3 octals per lc */
    mx_card[0].first_line = 0;
    if(request_irq(mx_card[0].irq, mx_interrupt, SA_SHIRQ, "Mswitch", &mx_card[0])) {
        printk("mxc: could not allocate IRQ#%d\n", mx_card[0].irq);
	return 1;
    }

#ifdef ARUBA_DEBUG
    printk("mxc: Mswitch Control Driver, IRQ %d\n", mx_card[0].irq);
#endif

    mx_nboard = 1;

    /* invalidate remaining mx_card structures */
    for (i = 0 ; i < NR_CARDS ; i++) {
        if (mx_card[i].base_addr == 0) {
                mx_card[i].first_line = -1;
                mx_card[i].ctl_addr = 0;
                mx_card[i].irq = 0;
                mx_card[i].first_line = 0;
                mx_card[i].num_chips = 0;
        }
    }
    /* invalidate remaining mx_port structures */
    for (i = 0; i < NR_PORTS ; i++) {
        mx_port[i].line = -1;
        mx_port[i].magic = -1;
    }

    nports = NR_PORTS;
    /* initialize per-port data structures for each valid board found */
    for (board = 0; board < mx_nboard; board++) {
        cinfo = &mx_card[board];
	cinfo->nports = nports;
	spin_lock_init(&cinfo->card_lock);
	for (port = cinfo->first_line ;
	     port < cinfo->first_line + nports;
	     port++) {
	    info = &mx_port[port];
	    info->magic = MSWITCH_MAGIC;
	    info->type = PORT_PHILIPS;
	    info->card = board;
	    info->line = port;
	    info->flags = STD_COM_FLAGS;
	    info->tty = 0;
	    info->xmit_fifo_size = MX_FIFO_SIZE;
	    info->custom_divisor = 0;
	    info->close_delay = 5*HZ/10;
	    info->closing_wait = CLOSING_WAIT_DELAY;
	    info->icount.cts = info->icount.dsr = 
	      info->icount.rng = info->icount.dcd = 0;
	    info->icount.rx = info->icount.tx = 0;
	    info->icount.frame = info->icount.parity = 0;
	    info->icount.overrun = info->icount.brk = 0;
	    info->event = 0;
	    info->count = 0;
	    info->blocked_open = 0;
	    info->tqueue.routine = do_softint;
	    info->tqueue.data = info;
	    info->normal_termios =
	      mx_driver.init_termios;
	    init_waitqueue_head(&info->open_wait);
	    init_waitqueue_head(&info->close_wait);
	    init_waitqueue_head(&info->shutdown_wait);
	    /* info->session */
	    /* info->pgrp */
	    /* info->timeout */
	    /* Bentson's vars */
	    info->jiffies[0] = 0;
	    info->jiffies[1] = 0;
	    info->jiffies[2] = 0;
	}
    }

    strcpy(__mxc_platid, "Aruba5000");

    return 0;
    
} /* mx_init */

#ifdef MODULE
void
mx_cleanup_module(void)
{
    int i;
    int e;
    unsigned long flags;

    save_flags(flags); cli();
    remove_bh(MSWITCH_BH);

    if ((e = tty_unregister_driver(&mx_driver)))
            printk("mxc: failed to unregister Mswitch driver(%d)\n",
		   e);
    if ((e = tty_unregister_driver(&mx_poe_driver)))
            printk("mxc: failed to unregister Mswitch POE driver (%d)\n", 
		   e2);
    if ((e = tty_unregister_driver(&mx_soe_driver)))
            printk("mxc: failed to unregister Mswitch SOE driver (%d)\n", 
		   e);

    restore_flags(flags);

    for (i = 0; i < NR_CARDS; i++) {
        if (mx_card[i].base_addr != 0) {
	    iounmap((void *)mx_card[i].base_addr);
	    if (mx_card[i].ctl_addr != 0)
		iounmap((void *)mx_card[i].ctl_addr);
	    if (mx_card[i].irq)
		free_irq(mx_card[i].irq, &mx_card[i]);
        }
    }
    if (tmp_buf) {
	free_page((unsigned long) tmp_buf);
	tmp_buf = NULL;
    }
} /* mx_cleanup_module */

/* Module entry-points */
module_init(mx_init);
module_exit(mx_cleanup_module);

#else /* MODULE */
/* called by linux/init/main.c to parse command line options */
void
mx_setup(char *str, int *ints)
{
} /* mx_setup */
#endif /* MODULE */

MODULE_LICENSE("GPL");
