#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>

#include <linux/aruba_sap.h>

// The timer needs to be reset within 16 seconds

#define SAP_WDT 203

#define WDT_TIMER 4*HZ

// refresh for 60 seconds and no more
#define WDT_COUNT 15

#define WDT_ADDR (volatile __u8 *)SAP_CPLD_WDG_BASE
#define LED_ADDR (volatile __u8 *)SAP_CPLD_LED_BASE

static struct timer_list wdt_timer;
static int wdt_count = WDT_COUNT;

void wdt_accton_send_pulse(void){
  static int toggle = 1;

  volatile __u8 *wdt = WDT_ADDR;

  if (toggle == 0){
    *wdt = 0xff;
    toggle = 1;
  } else {
    *wdt = 0x7f;
    toggle = 0;
  }

}


// until the user process is ready, keep clearing the timer within the kernel

static void wdt_accton_timer_fn(unsigned long data){

  wdt_accton_send_pulse();
  if (--wdt_count >= 0)
    mod_timer(&wdt_timer, jiffies+WDT_TIMER);

}

static int wdt_accton_setup_timer(void){

  init_timer(&wdt_timer);
  wdt_timer.function = wdt_accton_timer_fn;
  wdt_timer.data = 0;
  wdt_timer.expires = jiffies + HZ;
  add_timer(&wdt_timer);
  return 0;
}


static ssize_t wdt_accton_read(struct file *file, char *buf, 
			       size_t count, loff_t *pos){

  return 0;

}

static ssize_t wdt_accton_write(struct file *file, const char *buf, 
				size_t count, loff_t *pos){

  wdt_accton_send_pulse();
  return count;

}

static int wdt_accton_open(struct inode *inode, struct file *file){

  // no need to delete the timer since we will stop after 60 seconds
  //  del_timer(&wdt_timer);
  return 0;

}

static int wdt_accton_close(struct inode *inode, struct file *file){

  return 0;
}

static struct file_operations wdt_ops = {
  owner: THIS_MODULE,
  read:  wdt_accton_read,
  write: wdt_accton_write,
  open:  wdt_accton_open,
  release: wdt_accton_close
};

static struct miscdevice wdt_accton_dev = {
  SAP_WDT,
  "net/wdt",
  &wdt_ops
};


int __init wdt_accton_init(void){

  //  printk("Accton WDT initialization\n");
  if (misc_register(&wdt_accton_dev)){
    printk("Error registering wdt device\n");
    return -1;
  }

  wdt_accton_setup_timer();

  //  Turn off the LEDs now
  *(LED_ADDR) = 0;
  
  wdt_accton_send_pulse();

  return 0;

}

void wdt_accton_cleanup(void){
  misc_deregister(&wdt_accton_dev);
}


