#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>

#include <linux/aruba_sap.h>

/* The timer needs to be reset every second */

#define SAP_WDT       203
#define WDT_FREQ_DIV  4 /* refresh every 1/x second */
#define WDT_APP_SECS  8 /* x second app wdog */

static struct timer_list wdt_timer;
static int wdt_app;
static int wdt_count;

/* Until the user process is ready, refresh forever.
 * Once user process is ready, only refresh if process
 * refreshed within WDT_APP_SECS seconds
 */

static void wdt_airespace_timer_fn(unsigned long data)
{
  WATCHDOG_RESET();
  if (wdt_count < WDT_APP_SECS*WDT_FREQ_DIV) {
      mod_timer(&wdt_timer, jiffies + HZ/WDT_FREQ_DIV);
  }
}

static int wdt_airespace_setup_timer(void)
{
  init_timer(&wdt_timer);
  wdt_timer.function = wdt_airespace_timer_fn;
  wdt_timer.data = 0;
  wdt_timer.expires = jiffies + HZ/WDT_FREQ_DIV;
  if (wdt_app) {
    wdt_count++; /* read mod write race IS possible */
  }
  add_timer(&wdt_timer);
  return 0;
}


static ssize_t wdt_airespace_read(struct file *file, char *buf, 
                                  size_t count, loff_t *pos)
{
  wdt_count = 0; /* refresh app wdog */
  return 0;
}

static ssize_t wdt_airespace_write(struct file *file, const char *buf, 
                                   size_t count, loff_t *pos)
{
  wdt_count = 0; /* refresh app wdog */
  return count;
}

static int wdt_airespace_open(struct inode *inode, struct file *file)
{
  wdt_app = 1; /* app is refresing once a second */
  return 0;
}

static int wdt_airespace_close(struct inode *inode, struct file *file)
{
  return 0;
}

static struct file_operations wdt_ops = {
  owner: THIS_MODULE,
  read:  wdt_airespace_read,
  write: wdt_airespace_write,
  open:  wdt_airespace_open,
  release: wdt_airespace_close
};

static struct miscdevice wdt_airespace_dev = {
  SAP_WDT,
  "net/wdt",
  &wdt_ops
};

int __init wdt_airespace_init(void)
{
  //  printk("Airespace WDT initialization\n");
  if (misc_register(&wdt_airespace_dev)){
    printk("Error registering wdt device\n");
    return -1;
  }

  /* refresh watchdog automatically for 60 seconds, no more,
   * to wait for app to start refreshing
   */
  wdt_airespace_setup_timer();
  WATCHDOG_RESET();

  return 0;
}

void wdt_airespace_cleanup(void)
{
  misc_deregister(&wdt_airespace_dev);
}


