/***********************************************************************
 *
 * (C) Copyright 2000
 * DENX Software Engineering
 * Wolfgang Denk, wd@denx.de
 * All rights reserved.
 ***********************************************************************/

/* ------------------------------------------------------------------------- */
/* i2c-aruba.c I2C Interface driver module for the ARUBA board               */
/* ------------------------------------------------------------------------- */
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.		     */
/* ------------------------------------------------------------------------- */

#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/init.h>
#include <asm/irq.h>
#include <asm/io.h>
#include <asm/mpc8240.h>

#include <linux/i2c.h>
#include <linux/i2c-algo-8240.h>

#ifdef CONFIG_MSWITCH
#include <platforms/mswitch.h>
#define ARUBA_EUMB_BASE MSWITCH_EUMB_BASE
#elif defined(CONFIG_CABERNET)
#include <platforms/cabernet.h>
#define ARUBA_EUMB_BASE CABERNET_EUMB_BASE
#elif defined(CONFIG_MALBEC)
#include <platforms/malbec.h>
#define ARUBA_EUMB_BASE MALBEC_EUMB_BASE
#elif defined(CONFIG_SYRAH)
#include <platforms/syrah.h>
#define ARUBA_EUMB_BASE SYRAH_EUMB_BASE
#elif defined(CONFIG_ASAP)
#include <platforms/sap.h>
#include <linux/aruba_sap.h>
#define ARUBA_EUMB_BASE SAP_EUMB_BASE
#endif


static int aruba_i2c_reg(struct i2c_client *client)
{
	return 0;
}


static int aruba_i2c_unreg(struct i2c_client *client)
{
	return 0;
}

static void aruba_i2c_inc_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_INC_USE_COUNT;
#endif
}

static void aruba_i2c_dec_use(struct i2c_adapter *adap)
{
#ifdef MODULE
	MOD_DEC_USE_COUNT;
#endif
}


static struct i2c_algo_8240_data aruba_i2c_data = {
	NULL
};

static struct i2c_adapter aruba_i2c_ops = {
	"ARUBA I2C Interface",
	I2C_HW_MPC8240_ARUBA,
	NULL,
	&aruba_i2c_data,
	aruba_i2c_inc_use,
	aruba_i2c_dec_use,
	aruba_i2c_reg,
	aruba_i2c_unreg,
};

int __init i2c_aruba_init(void)
{
	int res = -ENODEV;

#ifdef ARUBA_DEBUG
	printk("i2c-aruba.o: i2c ARUBA module\n");
#endif

	aruba_i2c_data.i2c = ioremap(ARUBA_EUMB_BASE + M8240_I2C_OFFSET,
	                             M8240_I2C_SIZE);

	if (i2c_8240_add_bus(&aruba_i2c_ops) < 0)
	{
		goto Done;
	}

#ifdef ARUBA_DEBUG
	printk("i2c-aruba.o: found device at 0x%08x.\n",
	       (unsigned int)aruba_i2c_data.i2c);
#endif

	res = 0;
Done:
	return res;
}

int i2c_aruba_write(u16 addr, u8 command, char data)
{
  union i2c_smbus_data temp;
  temp.byte = data;
  return i2c_smbus_xfer(&aruba_i2c_ops, addr, 0, I2C_SMBUS_WRITE,
                        command, I2C_SMBUS_BYTE_DATA, &temp);
}

unsigned int i2c_aruba_read(u16 addr, u8 command)
{
  union i2c_smbus_data temp;
  if(i2c_smbus_xfer(&aruba_i2c_ops, addr, 0, I2C_SMBUS_READ,
                    command, I2C_SMBUS_BYTE_DATA, &temp))
    return -1;
  else
    return 0xFF & temp.byte;
}

EXPORT_SYMBOL(i2c_aruba_write);
EXPORT_SYMBOL(i2c_aruba_read);
/* EXPORT_NO_SYMBOLS; */

#ifdef MODULE
MODULE_AUTHOR("Sameer Dekate <wd@denx.de>");
MODULE_DESCRIPTION("I2C-Bus driver routines for the ARUBA board");


int init_module(void)
{
	return i2c_aruba_init();
}

void cleanup_module(void)
{
	i2c_8240_del_bus(&aruba_i2c_ops);
}

#endif
