/*
 *  ae5311End.c
 *
 *  Atheros ethernet driver for AR5311 chipset.  Based on InSilicon IPR.
 *
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/ae5311End.c#1 $"

/* Defines for development and test */

#ifdef DEBUG
#define DRV_DEBUG 1
#endif
#define PHY_POLLED      1   /* Poll for link changes and counter updates.*/
                            /*   Normally 1.                             */
#define TX_TWO_BUF      1   /* Use split tx dma buffers to align most on */
                            /*   16 byte bndry. Normally 1.              */
#define TX_OSF          0   /* 0 = SF, !OSF; 1 = !SF, !OSF; 2 = !SF, OSF */

#ifdef AR5311_ES1
#define DESC_BUG_RESET  1   /* Enables workaround for rev1 hw bug #3001. */
#else
#define DESC_BUG_RESET  0   /* Disable workaround for rev1 hw bug #3001. */
#endif
#define KEEP_BAD_FRAME  0   /* Used for debug only, normally 0.          */

#define PHYPOLL_PRI     50      /* tPhyPoll task priority (shared) */
#define PHYPOLL_STACK   2048    /* tPhyPoll task stack size */

/* Set number of rx descriptors to process before rescheduling */
#define AE_RX_RESCHED   68      /* define as 0 to remove this fcn */

#ifdef PRJ_BUILD
#include "prjComps.h"       /* Bring in INCLUDE_BRIDGE */
#endif

#include "end.h"
#include "ae5311End.h"
#include "ae5311reg.h"
#include "ar5311reg.h"
#include "config.h"
#include "histograms.h"

#ifdef BUILD_AP                 /* no histogram for bootrom */
#include "ui.h"
#endif

#include "vxworks.h"
#include "stdlib.h"
#include "cacheLib.h"
#include "intLib.h"
#include "endLib.h"
#include "lstLib.h"
#include "tickLib.h"
#include "wdLib.h"
#include "iv.h"
#include "semLib.h"
#include "etherLib.h"
#include "logLib.h"
#include "netLib.h"
#include "stdio.h"
#include "sysLib.h"
#include "errno.h"
#include "errnoLib.h"
#include "memLib.h"
#include "iosLib.h"
#include "assert.h"

#undef  ETHER_MAP_IP_MULTICAST
#include "etherMultiLib.h"

#include "net/mbuf.h"
#include "net/unixLib.h"
#include "net/protosw.h"
#include "net/systm.h"
#include "net/if_subr.h"
#include "net/route.h"

#include "sys/socket.h"
#include "sys/ioctl.h"
#include "sys/times.h"

#ifdef DRV_DEBUG
#include "netShow.h"
extern BOOT_PARAMS  sysBootParams;      /* parameters from boot line */
#endif

#define SWCOHERENCY 1
#define A_DATA_CACHE_INVAL(addr, len) \
        cacheDataInvalidate(addr, len)


/* END Debug Macros */
#ifdef  DRV_DEBUG

#define DRV_DEBUG_OFF           0x000000
#define DRV_DEBUG_RX            0x000001
#define DRV_DEBUG_TX            0x000002
#define DRV_DEBUG_INT           0x000004
#define DRV_DEBUG_POLL_RX       0x000008
#define DRV_DEBUG_POLL_TX       0x000010
#define DRV_DEBUG_POLL          (DRV_DEBUG_POLL_RX | DRV_DEBUG_POLL_TX)
#define DRV_DEBUG_LOAD          0x000020
#define DRV_DEBUG_IOCTL         0x000040
#define DRV_DEBUG_POLL_MODE     0x000080
#define DRV_DEBUG_PHYSET        0x000100
#define DRV_DEBUG_RXH           0x000200
#define DRV_DEBUG_TXM           0x000400
#define DRV_DEBUG_RXD           0x000800
#define DRV_DEBUG_ORN           0x001000
#define DRV_DEBUG_RXF           0x002000
#define DRV_DEBUG_ENDFLAG       0x004000
#define DRV_DEBUG_POLL_REDIR    0x010000
#define DRV_DEBUG_LOG_NVRAM     0x020000
#define DRV_DEBUG_LOG_ALL       0x080FFF
#define DRV_DEBUG_MSG           0x100000
#define DRV_DEBUG_NOUNIT        0x200000

#define DEBUG_FLAGS_0       0

#define DEBUG_FLAGS_1       (DRV_DEBUG_PHYSET \
                            | DRV_DEBUG_INT \
                            | DRV_DEBUG_TX \
                            | DRV_DEBUG_RXH \
                            | DRV_DEBUG_TXM \
                            | DRV_DEBUG_MSG)

#define DEBUG_FLAGS_2       (DRV_DEBUG_MSG | DRV_DEBUG_PHYSET)

#define DEBUG_FLAGS_3       (DRV_DEBUG_MSG | DRV_DEBUG_PHYSET | DRV_DEBUG_RXD)

#define DRV_DEBUG_FLAGS     (DEBUG_FLAGS_0);

int ae5311Debug = 0;

#define DRV_LOG(FLG, X0, X1, X2, X3, X4, X5, X6)    \
{                                                   \
    if (ae5311Debug & (FLG)) {                      \
        logMsg(X0, X1, X2, X3, X4, X5, X6);         \
    }                                               \
}

#define DRV_MSG(x,a,b,c,d,e,f)                      \
    logMsg(x,a,b,c,d,e,f)

#define DRV_PRINT(FLG, X)                           \
{                                                   \
    if (ae5311Debug & (FLG)) {                      \
        printf X;                                   \
    }                                               \
}

/* Requires low level uart driver output stub */
#define DRV_POLL_PRINT(FLG, X0, X1, X2, X3, X4, X5, X6)

typedef struct ae_dbg
{
    UINT32 ae5311regIsr;    /* snapshot of regIsr                           */
    UINT32 ae5311Ints;      /* count of ISR entries                         */
    UINT32 ae5311RxInts;    /* count of ISR entries with DmaIntRxCompleted  */
    UINT32 dbgRcvCnt;       /* count of ae5311Recv entries                  */
    UINT32 dbgRxCnt;        /* count of rxDesc processed                    */
    UINT32 dbgRxGoodCnt;    /* count of good rx packets sent up from driver */
    UINT32 dbgRxNoMb;       /* count of rx pkt dropped with no mblk resource*/
    UINT32 dbgRxErrPkt;     /* count of rx pkt dropped with error           */
    UINT32 dbgTxNot;        /* count of tx not attempted due to             */
                            /*          aeProcessRst | AE5311_POLLING       */
    UINT32 dbgTxLenReq;     /* count of tx not attempted due to len too long*/
    UINT32 dbgTxNoDesc;     /* count of tx not attempted due to no desc     */
                            /*          available.                          */
    UINT32 dbgTxNoMblk;     /* count of tx not attempted due to no mblk     */
                            /*          available.                          */
    UINT32 dbgTxErrPkt;     /* count of prev tx attempt with error          */
    UINT32 dbgTxSentPkt;    /* count of tx attempts without error           */
    UINT32 dbgTxmNext;      /* count of tx with mNext not NULL              */
    UINT32 intAbnRxDesc;    /* count of abnormal rx no desc available       */
                            /*          interrupts detected                 */
    UINT32 intAbnRxStop;    /* count of abnormal rx stopped interrupts      */
                            /*          detected                            */
    UINT32 intAbnTx;        /* count of abnormal tx interrupts detected     */
    UINT32 intBusError;     /* count of fatal bus error conditions detected */
#if DESC_BUG_RESET == 1
    UINT32 dmaBugReset;     /* count of eth resets due to goofy desc bug    */
#endif
} AE_DBG;

AE_DBG aeDbg[AE_MAX_ETH_UNITS];

#define DEBUG_STAT_INC(_dev, _stat)             (++(aeDbg[_dev].##_stat))
#define DEBUG_STAT_ASSIGN(_dev, _stat, _val)    (aeDbg[_dev].##_stat = (_val))

NET_POOL_ID aep0;
NET_POOL_ID aep1;

void dumpPkt(char *, int);

#else /* not DRV_DEBUG */

#define DRV_LOG(DBG_SW, X0, X1, X2, X3, X4, X5, X6)
#define DRV_MSG(x,a,b,c,d,e,f)
#define DRV_PRINT(DBG_SW,X)
#define DRV_POLL_PRINT(FLG, X0, X1, X2, X3, X4, X5, X6)

#define DEBUG_STAT_INC(_dev, _stat)
#define DEBUG_STAT_ASSIGN(_dev, _stat, _val)

#endif  /* not DRV_DEBUG */


/* A shortcut for getting the hardware address from the MIB II stuff. */
#define END_HADDR(pEnd)     ((pEnd)->mib2Tbl.ifPhysAddress.phyAddress)

#define END_HADDR_LEN(pEnd) ((pEnd)->mib2Tbl.ifPhysAddress.addrLength)

/* data types */

typedef volatile UINT8 * VIRT_ADDR;     /* CPU virtual address */

/* Driver private descriptor macros */
#define AE_QUEUE_ELE_NEXT(q)    \
                q->curDescAddr = AE_QUEUE_ELE_NEXT_GET(q, q->curDescAddr)

#define AE_QUEUE_ELE_NEXT_GET(q, descAddr)                          \
        ((descAddr) == (q)->lastDescAddr) ? (q)->firstDescAddr :    \
        (VIRT_ADDR)((UINT32)(descAddr) + AE_QUEUE_ELE_SIZE)

/* Cache */
#ifdef SWCOHERENCY
#define END_CACHED_DATA 1
/* FOO Only CACHED_DATA is supported at this time. */
/*
#define END_CACHED_DESC 1
*/
#endif


/*
 * Macros to get/set the values of descriptor fields with
 * appropriate address and byte order translations, if needed.
 */

/* Macros to get/set the values of the dma descriptor fields */

#define AE_DESC_STATUS_GET(ptr)       *(volatile UINT32 *)((UINT32)ptr + AE_DESC_STATUS)
#define AE_DESC_STATUS_SET(ptr, val)  AE_DESC_STATUS_GET(ptr) = (val)
#define AE_DESC_CTRLEN_GET(ptr)       *(volatile UINT32 *)((UINT32)ptr + AE_DESC_CTRLEN)
#define AE_DESC_CTRLEN_SET(ptr, val)  AE_DESC_CTRLEN_GET(ptr) = (val)
#define AE_DESC_BUFPTR_GET(ptr)       *(volatile UINT32 *)((UINT32)ptr + AE_DESC_BUFPTR)
#define AE_DESC_BUFPTR_SET(ptr,val)   AE_DESC_BUFPTR_GET(ptr) = (val)
#define AE_DESC_LNKBUF_GET(ptr)       *(volatile UINT32 *)((UINT32)ptr + AE_DESC_LNKBUF)
#define AE_DESC_LNKBUF_SET(ptr, val)  AE_DESC_LNKBUF_GET(ptr) = (val)

#define AE_DESC_MBLKPTR_GET(ptr)    (M_BLK_ID)(*(volatile UINT32 *) \
                                            ((UINT32)ptr + AE_DESC_MBLKPTR))
#define AE_DESC_MBLKPTR_SET(ptr,val)    AE_DESC_MBLKPTR_GET(ptr) = (val)


/* Determine Broadcasts */
/* FOO should share with other places */
/* FOO is this being checked more than once by bridge too? */
typedef unsigned char MAC_ADDR[6];
#define IS_BROADCAST(A) ((bcmp((A), bcastMacAddr, sizeof(MAC_ADDR)) == 0) ? TRUE : FALSE)
LOCAL MAC_ADDR bcastMacAddr = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };


/* The definition of the driver control structures */

/* Descriptor queue */
typedef struct ae5311_queue
{
    VIRT_ADDR   firstDescAddr;  /* descriptor array address */
    VIRT_ADDR   lastDescAddr;   /* last descriptor address */
    VIRT_ADDR   curDescAddr;    /* current descriptor address */
    VIRT_ADDR   reapDescAddr;   /* current tail of tx descriptors reaped */
    UINT16      count;          /* number of elements */
} AE5311_QUEUE;

/* END device */
typedef struct end_device
{
    END_OBJ         end;            /* The class we inherit from. */
    int             unit;           /* unit number */
    int             ivec;           /* interrupt vector */
    int             ilevel;         /* interrupt level */
    UINT32          macBase;        /* base address of MAC registers */
    UINT32          dmaBase;        /* base address of DMA registers  */
    UINT32          phyAddr;        /* PHY device address on MII interface  */
    char*           pDmaBuf;        /* ptr to device  memory */
    char*           pMclMem;        /* ptr to MBLK/CL_BLK memory */
    int             dmaBufSize;     /* size of user specified memory */
    UINT16          txDescCount;    /* Transmit descriptor count */
    UINT16          rxDescCount;    /* Receive descriptor count */
    UINT16          mBlkCount;      /* MBLK count */
    UINT16          clCount;        /* Extra Cluster count */
    AE5311_QUEUE    txQueue;        /* Transmit descriptor queue */
    AE5311_QUEUE    rxQueue;        /* Receive descriptor queue */
    long            flags;          /* Our local flags. */
    UCHAR           enetAddr[6];    /* ethernet address */
    CACHE_FUNCS     cacheFuncs;     /* cache function pointers */
    bool            aeStart;        /* flag to indicate device has been started */
    bool            aeProcessRcv;   /* flag to indicate processing pkt recv */
    bool            aeProcessRst;   /* flag to indicate processing chip reset */
    M_BLK_ID        pollMblk;       /* Used in polling mode only*/
} END_DEVICE;


#if defined (DRV_DEBUG) || defined (PHY_POLLED)
LOCAL END_DEVICE *pCtrlAe[AE_MAX_ETH_UNITS] = {(END_DEVICE *)0,(END_DEVICE *)0,};
    /* FOO need proper global initialization */
#endif

#if defined (PHY_POLLED)
int idPhyPollTask;
int refPhyPollTask = 0; /* FOO need proper global initialization */
UINT32 aeRxMissedFrameFO[AE_MAX_ETH_UNITS];
UINT32 aeRxMissedFrameND[AE_MAX_ETH_UNITS];
int aeOvrnFlag[AE_MAX_ETH_UNITS];
int aeTimeout[AE_MAX_ETH_UNITS] = {0,0,};
#endif



/* from END template */
#define END_SPEED_10M       10000000    /* 10Mbs */
#define END_SPEED_100M      100000000   /* 100Mbs */
#define END_SPEED           END_SPEED_100M


/* Driver private routines */
LOCAL STATUS    ae5311Start(VOID *);
LOCAL STATUS    ae5311Stop(VOID *);
LOCAL STATUS    ae5311Unload(VOID *);
LOCAL int       ae5311Ioctl(VOID *, int, caddr_t);
LOCAL STATUS    ae5311Send(VOID *, M_BLK_ID);
LOCAL STATUS    ae5311PollSend(VOID *, M_BLK_ID);
LOCAL STATUS    ae5311PollRecv(VOID *, M_BLK_ID);
LOCAL STATUS    ae5311PollStart(VOID *);
LOCAL STATUS    ae5311PollStop(VOID *);
LOCAL STATUS    ae5311Parse();
LOCAL STATUS    ae5311MemInit(END_DEVICE *);
LOCAL STATUS    ae5311MCastAdd(VOID *, char *);
LOCAL STATUS    ae5311MCastDel(VOID *, char *);
LOCAL STATUS    ae5311MCastGet(VOID *, MULTI_TABLE *);

LOCAL STATUS    ae5311QueueCreate(END_DEVICE *, AE5311_QUEUE *, char *, int, int);
LOCAL STATUS    ae5311QueueDelete(struct ae5311_queue *);
LOCAL void      ae5311MacAddressSet(END_DEVICE *);
LOCAL void      ae5311MacAddressGet(END_DEVICE *);
LOCAL STATUS    ae5311Config(END_DEVICE *);
LOCAL void      ae5311Int(END_DEVICE  *);
LOCAL STATUS    ae5311Recv(END_DEVICE *,int);
LOCAL void      ae5311RxFilterConfig(END_DEVICE *);
LOCAL VIRT_ADDR ae5311TxDescGet(END_DEVICE *);
LOCAL VIRT_ADDR ae5311RxDescGet(END_DEVICE *);
LOCAL STATUS    ae_PhySetup(END_DEVICE *);
LOCAL STATUS    ae_SetMacFromPhy(END_DEVICE *);
LOCAL STATUS    sysAe5311ParamGet(END_DEVICE *);
#ifdef PHY_POLLED
LOCAL void ae_PhyPollTask(void);
#endif
LOCAL void ae5311MiiWrite(END_DEVICE *, unsigned char, unsigned short);
LOCAL unsigned short ae5311MiiRead(END_DEVICE *, unsigned char);
LOCAL unsigned int ae5311ReadMacReg(END_DEVICE *, unsigned int);
LOCAL void ae5311WriteMacReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL void ae5311SetMacReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL void ae5311ClearMacReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL unsigned int ae5311ReadDmaReg(END_DEVICE *, unsigned int);
LOCAL void ae5311WriteDmaReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL void ae5311DmaRxStart(END_DEVICE *);
LOCAL void ae5311DmaTxStart(END_DEVICE *);
LOCAL void ae5311DmaIntEnable(END_DEVICE *);
LOCAL void ae5311DmaIntDisable(END_DEVICE *);
LOCAL void aeResetOnError(END_DEVICE *);
LOCAL void ae5311DmaIntClear(END_DEVICE *);
LOCAL void ae5311SetDmaReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL void ae5311ClearDmaReg(END_DEVICE *, unsigned int, unsigned int);
LOCAL void ae5311DmaTxStop(END_DEVICE *);
LOCAL void ae5311DmaRxStop(END_DEVICE *);
LOCAL void ae5311DmaReset(END_DEVICE *);
LOCAL void ethStrayInt(int);
#if defined (PHY_POLLED) || defined (DRV_DEBUG)
LOCAL int updateMissedFrameCounters(END_DEVICE *, UINT32 *, UINT32 *);
#endif

/*
 * Declare our END function table.  This is static across all driver
 * instances.
 */
LOCAL NET_FUNCS ae5311FuncTable =
{
    (FUNCPTR)ae5311Start,           /* Function to start the device. */
    (FUNCPTR)ae5311Stop,            /* Function to stop the device. */
    (FUNCPTR)ae5311Unload,          /* Unloading function for the driver. */
    (FUNCPTR)ae5311Ioctl,           /* Ioctl function for the driver. */
    (FUNCPTR)ae5311Send,            /* Send function for the driver. */
    (FUNCPTR)ae5311MCastAdd,        /* Multicast add function for the driver. */
    (FUNCPTR)ae5311MCastDel,        /* Multicast delete function for the driver. */
    (FUNCPTR)ae5311MCastGet,        /* Multicast retrieve function for the driver.*/
    (FUNCPTR)ae5311PollSend,        /* Polling send function */
    (FUNCPTR)ae5311PollRecv,        /* Polling receive function */
    endEtherAddressForm,            /* put address info into a NET_BUFFER */
    (FUNCPTR)endEtherPacketDataGet, /* get pointer to data in NET_BUFFER */
    (FUNCPTR)endEtherPacketAddrGet  /* Get packet addresses. */
};


/* Get size of Rx data from desc, in bytes */
#define AE_DESC_STATUS_RX_SIZE(x)   (((x)&DescFrameLengthMask)>>DescFrameLengthShift)

/* Size of desc buffer1 len in bytes */
#define AE_DESC_CTRLEN_BUF1_LEN(x)  ((x)&DescSize1Mask)


HISTOGRAM_DEFINE_BINS(aeRx, 2, AE_RX_DESC_COUNT_DEFAULT, 2);

/*****************************************************************************
* ae5311EndLoad - initialize END interface for atheros ethernet controller
*
* This routine initializes the driver and the device to the operational state.
* All device specific parameters are obtained from sysAe5311ParamGet() and
* driver specific parameters are passed in the <initString>.
*
* The <initString> is a colon-separated string of parameters with the following
* syntax,
*
*  "unit:dmaBufAdr:dmaBufSize:txDescCount:rxDescCount:clCount"
*
* <unit>        Device unit number, is prefixed by automatically by the MUX
* <dmaBufAdr>   Address of the DMA buffer area
* <dmaBufSize>  Size of DMA buffer area
* <txDescCount> Transmit descriptor count
* <rxDescCount> Receive descriptor count
* <clCount>     Cluster (extra to be loaned) count
*
* If <dmaBufAdr> is specified as NULL (0) or NONE (-1), the driver allocates
* the required DMA buffer area using cacheDmaMalloc.  Otherwise, <dmaBufAdr>
* should contain the address of a cache-safe memory location, whose size is
* specified by <dmaBufSize>.
*
* The parameters <txDescCount> and <rxDescCount> specify the number of
* transmit and recive descriptors.  NULL (0) in these fields are used to
* specify the default values, AE_TX_DESC_DEFAULT and AE_RX_DESC_DEFAULT.
* 
* <clCount> specifies the number of extra clusters for loaning.  NULL (0)
* in this field is used to specify the default value, AE_CL_EXTRA_DEFAULT.
* If <clCount> is specified it should be at least AE_CL_EXTRA_MIN
* otherwise the driver will set <clCount> to this value.
*
* RETURNS: An END object pointer or NULL on error.
*/

END_OBJ * ae5311EndLoad
(
    char *initString,   /* String to be parsed by the driver. */
    void *p             /* unused */
)
{
    END_DEVICE  *pDrvCtrl;

    #ifdef DRV_DEBUG
    ae5311Debug = DRV_DEBUG_FLAGS; /* FOO should change to be unique to multi devices */
    #endif

    if (initString == NULL)
    {
        DRV_LOG (DRV_DEBUG_NOUNIT,
            "ae5311EndLoad: Null initString\n",
                1, 2, 3, 4, 5, 6);
        return (NULL);
    }

    /* First time ae5311EndLoad is called, with empty initString, */
    /* return name of device in initString, regardless of unit number. */
    if (initString[0] == EOS)
    {
        DRV_LOG (DRV_DEBUG_LOAD,"ae5311EndLoad: First pass\n", 1, 2, 3, 4, 5, 6);
        bcopy ((char *) AE_DEV_NAME, initString, strlen(AE_DEV_NAME)+1);
        return (0);
    }

    DRV_LOG (DRV_DEBUG_LOAD,"ae5311EndLoad: Second pass, loading device ...\n", 1, 2, 3, 4, 5, 6);

    /* allocate the device structure */

    pDrvCtrl = (END_DEVICE *) calloc (sizeof (END_DEVICE), 1);
    if (pDrvCtrl == NULL)
        goto errorExit;

    /* parse the init string, filling in the device structure */

    if (ae5311Parse (pDrvCtrl, initString) == ERROR)
        goto errorExit;

    pDrvCtrl->aeStart = 0;
    pDrvCtrl->aeProcessRcv = 0;
    pDrvCtrl->aeProcessRst = 0;
    pDrvCtrl->pollMblk = NULL;

    DRV_LOG (DRV_DEBUG_LOAD,
        "ae5311EndLoad: unit=%d, pDmaBuf=0x%x, dmaBufSize=%d, txDescCount=%d, rxDescCount=%d, clCount=%d\n",
            pDrvCtrl->unit,
            (UINT32)pDrvCtrl->pDmaBuf,
            pDrvCtrl->dmaBufSize,
            pDrvCtrl->txDescCount,
            pDrvCtrl->rxDescCount,
            pDrvCtrl->clCount);

    #if defined (DRV_DEBUG) || (PHY_POLLED)
    pCtrlAe[pDrvCtrl->unit] = pDrvCtrl;
    #endif
    #if defined (PHY_POLLED)
    aeRxMissedFrameFO[pDrvCtrl->unit] = 0;
    aeRxMissedFrameND[pDrvCtrl->unit] = 0;
    aeOvrnFlag[pDrvCtrl->unit] = 0;
    #endif
    #ifdef DRV_DEBUG
    memset(&aeDbg[pDrvCtrl->unit], 0, sizeof(AE_DBG));
    #endif

    /* Get parameters from the BSP and fill in the device structure */
    if (sysAe5311ParamGet (pDrvCtrl) == ERROR)
        goto errorExit;

    DRV_LOG (DRV_DEBUG_LOAD,
        "ae5311EndLoad: System param: mac=%x, dma=%x, phy=%x, ivec=%x, ilev=%x\n",
            pDrvCtrl->macBase,
            pDrvCtrl->dmaBase,
            pDrvCtrl->phyAddr,
            pDrvCtrl->ivec,
            pDrvCtrl->ilevel,
            6);

    /* Get the ethernet address from the FLASH */
    ae5311MacAddressGet (pDrvCtrl);

    DRV_LOG (DRV_DEBUG_LOAD, "ae5311EndLoad: Flash ea = %02x:%02x:%02x:%02x:%02x:%02x\n",
            pDrvCtrl->enetAddr[0], pDrvCtrl->enetAddr[1],
            pDrvCtrl->enetAddr[2], pDrvCtrl->enetAddr[3],
            pDrvCtrl->enetAddr[4], pDrvCtrl->enetAddr[5]);

    /* initialize the END and MIB2 parts of the structure */

    if ((END_OBJ_INIT (&pDrvCtrl->end, (DEV_OBJ *)pDrvCtrl, AE_DEV_NAME,
            pDrvCtrl->unit, &ae5311FuncTable,
            AE_DRV_NAME) == ERROR) ||
            (END_MIB_INIT (&pDrvCtrl->end, M2_ifType_ethernet_csmacd,
            &pDrvCtrl->enetAddr[0], 6, ETHERMTU,
            END_SPEED) == ERROR))
    {
        goto errorExit;
    }

    /* Perform memory allocation/distribution */

    if (ae5311MemInit (pDrvCtrl) == ERROR)
        goto errorExit;

    /* Configure the device */

    if (ae5311Config (pDrvCtrl) == ERROR)
    {
        printf("Error to configure the ae5311\n");
        goto errorExit;
    }

    /* Set the flags to indicate readiness */

    END_OBJ_READY (&pDrvCtrl->end,
                IFF_UP | IFF_RUNNING | IFF_NOTRAILERS | IFF_SIMPLEX
                | IFF_BROADCAST | IFF_ALLMULTI /*| IFF_PROMISC */);

    DRV_LOG (DRV_DEBUG_LOAD, "ae5311EndLoad: Done loading, pDrvCtrl=%p txQ=%p rxQ=%p\n",
            (int)pDrvCtrl,
            (int) &pDrvCtrl->txQueue,
            (int) &pDrvCtrl->rxQueue,
            4, 5, 6);

    DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
        pDrvCtrl->unit, pDrvCtrl->end.flags, 100,4,5,6);

    #ifdef PHY_POLLED
    /* Only start one instance of phy polling task */
    if (0 == refPhyPollTask)
    {
        idPhyPollTask = taskSpawn("tPhyPoll", PHYPOLL_PRI, 0, PHYPOLL_STACK,
                                (FUNCPTR)ae_PhyPollTask,
                                0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
        if (ERROR == idPhyPollTask)
        {
            DRV_MSG ("ae5311EndLoad: unit %d, unable to start phy polling task!\n",
                pDrvCtrl->unit, 2, 3, 4, 5, 6);
        }
    }
    refPhyPollTask++;
    #endif

    return (&pDrvCtrl->end);

errorExit:
    if (pDrvCtrl != NULL)
        free ((char *) pDrvCtrl);

    DRV_MSG ("ae5311EndLoad: errorExit!\n", 1, 2, 3, 4, 5, 6);

    return NULL;
}


/*****************************************************************************
*
* ae5311Parse - parse the init string
*
* Parse the input string.  Fill in values in the driver control structure.
*
* The initialization string format is:
* "unit:pDmaBuf:dmaBufSize:txDescCount:rxDescCount:clCount"
*
* unit          Device unit number, a small integer
* pDmaBuf       Shared memory address
* dmaBufSize    Size of memory reserved
* txDescCount   Transmit descriptor count
* rxDescCount   Receive descriptor count
* clCount       Extra (loaned) Cluster count
*
* RETURNS: OK or ERROR for invalid arguments.
*/

LOCAL STATUS ae5311Parse
(
    END_DEVICE  *pDrvCtrl,      /* device pointer */
    char        *initString     /* information string */
)
{
    char *tok;
    char *pHolder = NULL;

    /* Parse the initString */

    /* Unit number. */
    tok = strtok_r (initString, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->unit = strtoul (tok, NULL, (int)NULL);

    /* Device DMA buffer address. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->pDmaBuf = (char*) strtoul (tok, NULL, (int)NULL);

    /* DMA buffer size. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->dmaBufSize = strtoul (tok, NULL, (int)NULL);

    /* Transmit descriptor count. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->txDescCount = strtoul (tok, NULL, (int)NULL);

    if (pDrvCtrl->txDescCount == 0)
        pDrvCtrl->txDescCount = AE_TX_DESC_COUNT_DEFAULT;

    /* Receive descriptor count. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->rxDescCount = strtoul (tok, NULL, (int)NULL);

    if (pDrvCtrl->rxDescCount == 0)
        pDrvCtrl->rxDescCount = AE_RX_DESC_COUNT_DEFAULT;

    /* Cluster count and mBlk count. */
    tok = strtok_r (NULL, ":", &pHolder);
    if (tok == NULL)
        return ERROR;
    pDrvCtrl->clCount =  strtoul (tok, NULL, (int)NULL);

    if (pDrvCtrl->clCount == 0) {
        pDrvCtrl->clCount = AE_CL_EXTRA_DEFAULT;
    }

    /* the seemingly wasteful equals comparision below is to
     * keep the compiler happy when AE_CL_EXTRA_MIN is 0
     */
    if (pDrvCtrl->clCount <= AE_CL_EXTRA_MIN) {
        pDrvCtrl->clCount = AE_CL_EXTRA_MIN;
    }

    /* above number + rxDescCount for the receive ring */
    pDrvCtrl->clCount += pDrvCtrl->rxDescCount;
    pDrvCtrl->mBlkCount = pDrvCtrl->clCount;

    return OK;
}


/******************************************************************************
*
*
* ae5311MemInit - initialize queues and netpool for the device
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311MemInit
(
    END_DEVICE * pDrvCtrl   /* device to be initialized */
)
{
    M_CL_CONFIG mclBlkConfig;
    CL_DESC     clDescTbl[1];
    char        *pMemBuf = NULL;
    UINT32      txQMemSize;
    UINT32      rxQMemSize;
    UINT32      totalSize;

    /* Allocate memory for the net pool */
    if ((pDrvCtrl->end.pNetPool = malloc (sizeof(NET_POOL))) == NULL)
    {
        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: Could not allocate net pool memory\n",
                1, 2, 3, 4, 5, 6);
        goto aeMemInitFail;
    }

    /* Allocate memory for mBlks and clBlks */
    mclBlkConfig.mBlkNum  = pDrvCtrl->mBlkCount;
    mclBlkConfig.clBlkNum = pDrvCtrl->clCount;

    /*
    * Because every mBlk and cluster is prepended by a 4-byte header
    * (which is why the size calculations for clusters and mBlk structures
    * contained an extra sizeof(long).
    */
    mclBlkConfig.memSize = (mclBlkConfig.mBlkNum * (M_BLK_SZ + sizeof(long))) +
                            (mclBlkConfig.clBlkNum * (CL_BLK_SZ + sizeof(long)));
    mclBlkConfig.memArea = pDrvCtrl->pMclMem =
                            memalign (sizeof (long), mclBlkConfig.memSize);

    if (mclBlkConfig.memArea == NULL)
    {
        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: Could not allocate mclBlk memory\n",
                1, 2, 3, 4, 5, 6);
        goto aeMemInitFail;
    }

    /* Allocate memory for clusters, Tx and Rx Queues */

    /* Get the total memory size needed */
    #ifdef END_CACHED_DATA
    clDescTbl[0].clSize  = A_ROUNDUP(AE_BUF_SIZE,CACHE_LINE_SIZE) - sizeof(long);
    #else
    clDescTbl[0].clSize  = AE_BUF_SIZE;
    #endif
    clDescTbl[0].clNum   = pDrvCtrl->clCount;
    clDescTbl[0].memSize = ((clDescTbl[0].clNum *
                            (clDescTbl[0].clSize + ( 2 * sizeof (long)))) +
                            sizeof(int));

    txQMemSize = AE_QUEUE_ELE_SIZE * pDrvCtrl->txDescCount;
    rxQMemSize = AE_QUEUE_ELE_SIZE * pDrvCtrl->rxDescCount;

    #ifdef END_CACHED_DATA
    /* add extra CACHE_LINE_SIZE bytes for buffer cache line align */
    totalSize = clDescTbl[0].memSize + CACHE_LINE_SIZE;
    #else
    totalSize = clDescTbl[0].memSize
                + txQMemSize
                + rxQMemSize
                + AE_ALIGN;
    #endif

    /* If user has specified DevMem, validate; else allocate memory */

    /* User has specified DevMem, assume pool is cache coherent. */
    if ((pDrvCtrl->pDmaBuf != NULL) && (pDrvCtrl->pDmaBuf != (char*)NONE))
    {
        if (pDrvCtrl->dmaBufSize < totalSize)
        {
            DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: not enough dma memory provided, "
                    "need %ul got %d\n",
                    totalSize,pDrvCtrl->dmaBufSize, 3, 4, 5, 6);
            goto aeMemInitFail;
        }

        /* assume pool is cache coherent, copy null structure */
        pDrvCtrl->cacheFuncs = cacheNullFuncs;
    }
    else
    {
    #ifndef END_CACHED_DATA
        /* Because the structures that are shared between the device
         * and the driver may share cache lines, the possibility exists
         * that the driver could flush a cache line for a structure and
         * wipe out an asynchronous change by the device to a neighboring
         * structure. Therefore, this driver cannot operate with memory
         * that is not write coherent.  We check for the availability of
         * such memory here, and abort if the system did not give us what
         * we need.
         */

        pDrvCtrl->pDmaBuf = (char*) cacheDmaMalloc (totalSize);
    #else
        pDrvCtrl->pDmaBuf = (char *)malloc(totalSize);
    #endif

        if (pDrvCtrl->pDmaBuf == NULL)
            goto aeMemInitFail;

        /* pDrvCtrl->cacheFuncs = cacheDmaFuncs; */

        pDrvCtrl->flags |= AE5311_BUF_LOCAL;

        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: allocated, pDmaBuf=0x%x, buffend = %x\n",
                (UINT32)pDrvCtrl->pDmaBuf,
                (UINT32)pDrvCtrl->pDmaBuf + clDescTbl[0].memSize,
                3, 4, 5, 6);
    }

    /* Clear and align memory */

    memset (pDrvCtrl->pDmaBuf, 0, totalSize);
    #ifndef END_CACHED_DATA
    pMemBuf = (char*)(((UINT32)pDrvCtrl->pDmaBuf + AE_ALIGN) & ~(AE_ALIGN - 1));
    #else
    pMemBuf = (char *)A_ROUNDUP((UINT32)pDrvCtrl->pDmaBuf + CACHE_LINE_SIZE - 1,
                        CACHE_LINE_SIZE);
    #endif

    /* Initialize the memory pool. */

    clDescTbl[0].memArea = pMemBuf;

    #ifndef END_CACHED_DATA
    pMemBuf = pMemBuf + clDescTbl[0].memSize;
    pMemBuf = (char*)(((UINT32)pMemBuf + AE_ALIGN) & ~(AE_ALIGN - 1));
    #else
    pMemBuf = (char*) cacheDmaMalloc (txQMemSize + rxQMemSize);
    if (pMemBuf == NULL)
        goto aeMemInitFail;
    #endif

    if (netPoolInit (pDrvCtrl->end.pNetPool, &mclBlkConfig, &clDescTbl[0], 1, NULL)
            == ERROR)
    {
        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: Failed to initialize memory pool\n",
                1, 2, 3, 4, 5, 6);
        goto aeMemInitFail;
    }

    #ifdef DRV_DEBUG
    if (0 == pDrvCtrl->unit)
        aep0 = pDrvCtrl->end.pNetPool;
    if (1 == pDrvCtrl->unit)
        aep1 = pDrvCtrl->end.pNetPool;
    #endif

    /* Create Tx and Rx queues */
    #ifdef END_CACHED_DATA
    memset (pMemBuf, 0, txQMemSize + rxQMemSize);
    pMemBuf = (char*)(((UINT32)pMemBuf + AE_ALIGN) & ~(AE_ALIGN - 1));
    #endif

    if (ae5311QueueCreate (pDrvCtrl, &pDrvCtrl->txQueue, pMemBuf,
                            pDrvCtrl->txDescCount,
                            AE_QUEUE_TYPE_TX) != OK)
    {
        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: eth%d Failed to create tx queue\n",
                pDrvCtrl->unit, 2,3,4,5,6);
        goto aeMemInitFail;
    }

    pMemBuf += txQMemSize;

    if (ae5311QueueCreate (pDrvCtrl, &pDrvCtrl->rxQueue, pMemBuf,
                            pDrvCtrl->rxDescCount,
                            AE_QUEUE_TYPE_RX) != OK)
    {
        DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: eth%d Failed to create rx queue\n",
                pDrvCtrl->unit,2,3,4,5,6);
        goto aeMemInitFail;
    }

    DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: Memory setup complete.\n",
        1, 2, 3, 4, 5, 6);

    return OK;

aeMemInitFail:

    /* free memory allocated for mBlk-clBlks */

    if (mclBlkConfig.memArea != NULL)
    {
        free(mclBlkConfig.memArea);
    }

    /* free memory allocated for clusters and Tx/Rx Queues */

    if ((pDrvCtrl->flags & AE5311_BUF_LOCAL) && (pDrvCtrl->pDmaBuf != NULL))
    {
    #ifndef END_CACHED_DATA
        cacheDmaFree (pDrvCtrl->pDmaBuf);
    #else
        free (pDrvCtrl->pDmaBuf);
    }


    if ( pMemBuf != NULL )
    {
        cacheDmaFree (pMemBuf);
    #endif
    }

    /* free memory allocated for net pool */
    if (pDrvCtrl->end.pNetPool != NULL)
    {
        free (pDrvCtrl->end.pNetPool);
    }

    DRV_LOG (DRV_DEBUG_LOAD, "ae5311MemInit: Failed!\n",
        1, 2, 3, 4, 5, 6);

    return ERROR;
}


/*****************************************************************************
*
* ae5311Start - start the device
*
* This function calls BSP functions to connect interrupts and start the
* device running in interrupt mode.
*
* RETURNS: OK or ERROR
*
*/

LOCAL STATUS ae5311Start( VOID * pObj )     /* device ID */
{
    END_DEVICE *pDrvCtrl;   /* device to be started */

    pDrvCtrl = (END_DEVICE *)pObj;

    /* Install interrupt vector */

    (void)intConnect(INUM_TO_IVEC(pDrvCtrl->ivec), ae5311Int, (int)pDrvCtrl);

    DRV_LOG (DRV_DEBUG_LOAD,
        "ae5311Start: Interrupt connected %X %X %X %X.\n",
            pDrvCtrl->unit,
            pDrvCtrl->ivec,
            (int)pDrvCtrl,
            4, 5, 6);

    /* Turn on device interrupts -- enable most errors */
    ae5311DmaIntClear(pDrvCtrl);    /* clear interrupts */
    ae5311DmaIntEnable(pDrvCtrl);   /* enable interrupts */

    /* Enable Tx/Rx */
    ae5311SetMacReg(pDrvCtrl,MacControl,(MacRxEnable|MacTxEnable));
    ae5311DmaRxStart(pDrvCtrl);     /* start receiver  */
    ae5311DmaTxStart(pDrvCtrl);     /* start transmitter, it must go to suspend immediately  */

    DRV_LOG (DRV_DEBUG_LOAD, "ae5311Start: complete. im = %X, is = %X.\n",
        ae5311ReadDmaReg(pDrvCtrl,DmaInterrupt),
        ae5311ReadDmaReg(pDrvCtrl,DmaStatus),
        3, 4, 5, 6);

    intEnable(pDrvCtrl->ilevel);

    pDrvCtrl->aeStart = 1;
    END_FLAGS_SET (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
    DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
        pDrvCtrl->unit, pDrvCtrl->end.flags, 101, 4,5,6);

    return (OK);
}


/*******************************************************************************
*
* ae5311Int - handle controller interrupt
*
* This routine is called at interrupt level in response to an interrupt from
* the controller.
*
* RETURNS: N/A.
*/

LOCAL void ae5311Int( END_DEVICE  *pDrvCtrl )   /* interrupting device */
{
    UINT32 regIsr;

    DEBUG_STAT_INC(pDrvCtrl->unit, ae5311Ints);

    if (pDrvCtrl->aeProcessRst != 0)
        return;

    regIsr = ae5311ReadDmaReg(pDrvCtrl,DmaStatus);

    DRV_LOG (DRV_DEBUG_INT, "eth%d: intStatus=0x%x \n",
            pDrvCtrl->unit, regIsr, 3, 4, 5, 6);

    DEBUG_STAT_ASSIGN(pDrvCtrl->unit, ae5311regIsr, regIsr);

    /* FOO need to fill in error handler */
    /* Abnormal Rx Error */
    if (regIsr & DmaIntRxNoBuffer )
    {
        /* Throttle these errors */
        if (aeTimeout[pDrvCtrl->unit] == 0)
        {
            aeTimeout[pDrvCtrl->unit] = 1;

            /* mask this int */
            ae5311ClearDmaReg(pDrvCtrl, DmaInterrupt, DmaIeRxNoBuffer);

            DEBUG_STAT_INC(pDrvCtrl->unit, intAbnRxDesc);

            DRV_LOG (DRV_DEBUG_MSG,
                "eth%d: DmaIntRxNoDesc 0x%8.8x; intAbnRxDesc = %d\n",
                    pDrvCtrl->unit,
                    regIsr,
                    aeDbg[pDrvCtrl->unit].intAbnRxDesc,
                    0,0,0);

            /* FOO Special workaround 020312 */
            if (pDrvCtrl->aeProcessRcv == 0)
            {
                if (netJobAdd ((FUNCPTR)ae5311Recv, (int)pDrvCtrl, 2,0,0,0) == OK)
                {
                    pDrvCtrl->aeProcessRcv = 1;
                }
                else
                {
                    DRV_MSG ("Int %d: netJobAdd(ae5311Recv 2) failed!\n",
                        pDrvCtrl->unit,2,3,4,5,6);
                }
            }
            /* */
        }

        /* Clear this interrupt */
        ae5311WriteDmaReg(pDrvCtrl,DmaStatus,DmaIntRxNoBuffer);
    }

    /* FOO need to fill in error handler */
    /* Abnormal Rx Error */
    if (regIsr & DmaIntRxStopped )
    {
        /* Throttle error message. */

        DRV_LOG (DRV_DEBUG_MSG,"eth%d: DmaIntRxStopped 0x%8.8x\n",
            pDrvCtrl->unit, regIsr, 0,0,0,0);

        DEBUG_STAT_INC(pDrvCtrl->unit, intAbnRxStop);

        /* Clear this interrupt */
        ae5311WriteDmaReg(pDrvCtrl,DmaStatus,DmaIntRxStopped);
    }

    /* FOO need to fill in error handler */
    /* Abnormal Tx Error */
    if (regIsr & (DmaIntTxStopped | DmaIntTxJabber | DmaIntTxUnderflow))
    {
        DRV_LOG (DRV_DEBUG_MSG, "eth%d: DmaIntTxError 0x%8.8x\n",
            pDrvCtrl->unit, regIsr, 0,0,0,0);

        DEBUG_STAT_INC(pDrvCtrl->unit, intAbnTx);

        /* Clear this interrupt */
        ae5311WriteDmaReg(pDrvCtrl,DmaStatus,
            (DmaIntTxStopped | DmaIntTxJabber | DmaIntTxUnderflow));
    }

    /* FOO need to fill in error handler */
    /* Fatal Bus Error */
    if (regIsr & DmaIntBusError)
    {
        DRV_LOG (DRV_DEBUG_MSG, "eth%d: DmaIntBusError 0x%8.8x\n",
            pDrvCtrl->unit, regIsr, 0,0,0,0);

        DEBUG_STAT_INC(pDrvCtrl->unit, intBusError);

        /* Clear this interrupt */
        ae5311WriteDmaReg(pDrvCtrl,DmaStatus, DmaIntBusError);

        /* Reset the chip */
        if (pDrvCtrl->aeProcessRst != 0)
            return;

        if (netJobAdd ((FUNCPTR)aeResetOnError, (int)pDrvCtrl, 0,0,0,0) == OK)
        {
            /* Disable the interrupt */
            ae5311DmaIntDisable(pDrvCtrl);

            /* Disable tx and rx */
            ae5311DmaTxStop(pDrvCtrl);
            ae5311DmaRxStop(pDrvCtrl);
            ae5311ClearMacReg(pDrvCtrl,MacControl,(MacRxEnable|MacTxEnable));

            /*Set the reset flag*/
            pDrvCtrl->aeProcessRst = 1;

            /*Clear the running flag*/
            END_FLAGS_CLR (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
            DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
                pDrvCtrl->unit, pDrvCtrl->end.flags, 102,4,5,6);

            return;
        }
        else
        {
            DRV_MSG ("Int %d: netJobAdd(aeResetOnError) failed!\n",
                pDrvCtrl->unit,2,3,4,5,6);
        }
    }

    /* Rx Complete */
    if (regIsr & DmaIntRxCompleted)
    {
        DEBUG_STAT_INC(pDrvCtrl->unit, ae5311RxInts);

        if (pDrvCtrl->aeProcessRcv == 0)
        {

            if (netJobAdd ((FUNCPTR)ae5311Recv, (int)pDrvCtrl, 0,0,0,0) == OK)
            {
                /* mask rx int */
                ae5311ClearDmaReg(pDrvCtrl, DmaInterrupt, DmaIeRxCompleted);
                pDrvCtrl->aeProcessRcv = 1;
                /* Clear this interrupt */
                ae5311WriteDmaReg(pDrvCtrl,DmaStatus,DmaIntRxCompleted);
            }
            else
            {
                DRV_MSG ("Int %d: netJobAdd(ae5311Recv 0) failed!\n",
                    pDrvCtrl->unit,2,3,4,5,6);
            }
        }
    }

    /* Clear any unhandled intr causes. */
    ae5311WriteDmaReg(pDrvCtrl,DmaStatus,UnhandledIntrMask);
}


/*******************************************************************************
*
* ae5311Recv - process the next incoming packet
*
* Handle one incoming packet.  The packet is checked for errors and sent
* to the upper layer.
*
* RETURNS: OK on success or ERROR.
*/

LOCAL STATUS ae5311Recv(END_DEVICE *pDrvCtrl, int tag)  /* device structure */
{
    UINT32      cmdsts;
    int         len;
    VIRT_ADDR   rxDesc;
    M_BLK_ID    pNewMblk = NULL;
    M_BLK_ID    pRxMblk = NULL;
#if AE_RX_RESCHED > 0
    int         rxCountdown = AE_RX_RESCHED;
#endif

    HISTOGRAM_START_COUNTER(aeRx);

    /* Make sure we are not in polling mode */
    if (pDrvCtrl->flags & AE5311_POLLING) {
        /*
         * Should perhaps panic(). This should never happen.
         */
        DRV_LOG (DRV_DEBUG_RX,
            "eth%d error: ae5311Recv called in POLLING mode\n", pDrvCtrl->unit,2,3,4,5,6);
        return ERROR;
    }

#ifdef DRV_DEBUG
    if (tag == 1)
        DRV_LOG (DRV_DEBUG_RXF, "eth%d Recv 1\n", pDrvCtrl->unit,2,3,4,5,6);
    if (tag == 2)
        DRV_LOG (DRV_DEBUG_RXF, "eth%d Recv 2\n", pDrvCtrl->unit,2,3,4,5,6);
#endif

    DEBUG_STAT_INC(pDrvCtrl->unit, dbgRcvCnt);
    DRV_LOG (DRV_DEBUG_RX,
        "ae5311Recv: i=%d, ir=%d, r=%d\n",
            aeDbg[pDrvCtrl->unit].ae5311Ints,
            aeDbg[pDrvCtrl->unit].ae5311RxInts,
            aeDbg[pDrvCtrl->unit].dbgRcvCnt,
            4,5,6);

    while ((rxDesc = pDrvCtrl->rxQueue.curDescAddr)) {

        cmdsts = AE_DESC_STATUS_GET(rxDesc);
        if (cmdsts & DescOwnByDma) {
            break;
        }

#if AE_RX_RESCHED > 0
        /*
        *  RX loop to relinquish and reschedule after processing
        *  AE_RX_RESCHED descriptors.
        */
        if (0 == rxCountdown--) {
            if (netJobAdd ((FUNCPTR)ae5311Recv, (int)pDrvCtrl, 1,0,0,0) != OK) {
                DRV_MSG ("ae%d Rx thr: netJobAdd(ae5311Recv 1) failed!\n",
                        pDrvCtrl->unit,2,3,4,5,6);
                break;
            }
            HISTOGRAM_BIN(aeRx, pDrvCtrl->unit);
            return OK;
        }
#endif

        DEBUG_STAT_INC(pDrvCtrl->unit, dbgRxCnt);
        DRV_LOG ((DRV_DEBUG_RX | DRV_DEBUG_RXH),
            "ae Rx: Desc=0x%8.8x, S=0x%8.8x, B=0x%8.8x, M=0x%8.8x, d=0x%8.8x, c=%d\n",
                (UINT32)rxDesc,
                AE_DESC_STATUS_GET(rxDesc),
                AE_DESC_BUFPTR_GET(rxDesc),
                (UINT32)(AE_DESC_MBLKPTR_GET(rxDesc)),
                (UINT32)((AE_DESC_MBLKPTR_GET(rxDesc))->mBlkHdr.mData),
                aeDbg[pDrvCtrl->unit].dbgRxCnt);

        /*
         *  Process this packet if no errors.
         */

        len = AE_DESC_STATUS_RX_SIZE(cmdsts) - ETH_CRC_LEN;

        if ( (cmdsts & (DescRxFirst|DescRxLast|AE_DESC_STATUS_RX_ERRORS)) == (DescRxFirst | DescRxLast) ) {

            /* Allocate a new replacement mblock tuple (mblk-clblk-cluster) */
            pNewMblk = netTupleGet (pDrvCtrl->end.pNetPool, AE_BUF_SIZE,
                                        M_DONTWAIT, MT_DATA, FALSE);
            if (!pNewMblk) {
                /* others aren't returning my buffers fast enough; go do
                 * some transmit reaping for them and come back to this
                 */
                DRV_LOG (DRV_DEBUG_RX, "eth%d Rx: No mblk resources\n", pDrvCtrl->unit, 2,3,4,5,6);
                DEBUG_STAT_INC(pDrvCtrl->unit, dbgRxNoMb);
                if (netJobAdd ((FUNCPTR)ae5311Recv, (int)pDrvCtrl, 1,0,0,0) != OK) {
                    DRV_MSG ("ae%d Rx thr: netJobAdd(ae5311Recv 1) failed!\n",
                            pDrvCtrl->unit,2,3,4,5,6);
                    break;
                }
                HISTOGRAM_BIN(aeRx, pDrvCtrl->unit);
                return OK;
            }

            AE_QUEUE_ELE_NEXT((&pDrvCtrl->rxQueue));
            HISTOGRAM_COUNT(aeRx);

            pRxMblk = AE_DESC_MBLKPTR_GET (rxDesc);

#ifdef END_CACHED_DATA
            /* Make the data coherent */
            A_DATA_CACHE_INVAL(pRxMblk->mBlkHdr.mData, len );

            DRV_LOG (DRV_DEBUG_RX,
                "ae A_DATA_CACHE_INVAL %2.2x %2.2x %2.2x %2.2x %2.2x %2.2x\n",
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData)),
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData+1)),
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData+2)),
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData+3)),
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData+4)),
                (UINT32)(*((unsigned char *)pRxMblk->mBlkHdr.mData+5)) );
#endif  /* END_CACHED_DATA */

            /* Update the received Mblk */

            pRxMblk->mBlkHdr.mLen    = len;
            pRxMblk->mBlkHdr.mFlags |= M_PKTHDR;
            pRxMblk->mBlkPktHdr.len  = len;

            DEBUG_STAT_INC(pDrvCtrl->unit, dbgRxGoodCnt);

            DRV_LOG (DRV_DEBUG_RX,
                "RxDesc <%4d> [ISR:%8.8X] %8.8X :",
                aeDbg[pDrvCtrl->unit].dbgRxCnt,
                aeDbg[pDrvCtrl->unit].ae5311regIsr,
                (unsigned int)rxDesc,
                4,5,6);
            DRV_LOG (DRV_DEBUG_RX,
                "%8.8X %8.8X %8.8X %8.8X %8.8X",
                *((unsigned int *)rxDesc),
                *((unsigned int *)(rxDesc+4)),
                *((unsigned int *)(rxDesc+8)),
                *((unsigned int *)(rxDesc+12)),
                *((unsigned int *)(rxDesc+16)),
                6);

            /* Increment the packet count */
            if (cmdsts & DescRxMulticast) {
                if (IS_BROADCAST(pRxMblk->mBlkHdr.mData)) {
                    pRxMblk->mBlkHdr.mFlags |= M_BCAST;
                } else {
                    pRxMblk->mBlkHdr.mFlags |= M_MCAST;
                }
            } else {
                END_ERR_ADD (&pDrvCtrl->end, MIB2_IN_UCAST, +1);
            }


            /* Send data to the stack */
            DRV_LOG (DRV_DEBUG_RX, "eth%d Calling upper layer! %d\n",
                pDrvCtrl->unit,
                aeDbg[pDrvCtrl->unit].dbgRxGoodCnt,
                3, 4, 5, 6);
            END_RCV_RTN_CALL(&pDrvCtrl->end, pRxMblk);


            /* Replace the data buffer in the rxDesc */

            /* Rx DMA buffer on odd half word alignment */
            (UINT32)pNewMblk->mBlkHdr.mData += 2 + FASTPATH_HEADROOM;

            AE_DESC_BUFPTR_SET (rxDesc, (UINT32)pNewMblk->mBlkHdr.mData);
            AE_DESC_MBLKPTR_SET (rxDesc, pNewMblk);
            AE_DESC_STATUS_SET (rxDesc, DescOwnByDma);
            sysWbFlush();

            continue;
        }

        AE_QUEUE_ELE_NEXT((&pDrvCtrl->rxQueue));
        HISTOGRAM_COUNT(aeRx);

        if ( (cmdsts & (DescRxFirst|DescRxLast)) != (DescRxFirst | DescRxLast) ) {
            /*
            *   Goofy descriptor bug encountered.
            */

            /* Increment error statistics */
            END_ERR_ADD (&pDrvCtrl->end, MIB2_IN_ERRS, +1); /* bad packet */
            DRV_LOG (DRV_DEBUG_MSG, "ae%d Rx: Descr bug 0x%8.8x <0x%8.8x> 0x%8.8x [0x%8.8x]\n",
                pDrvCtrl->unit,
                cmdsts,
                AE_DESC_STATUS_RX_ERRORS,
                (int)rxDesc,
                ae5311ReadDmaReg(pDrvCtrl,DmaStatus),
                6);
            DEBUG_STAT_INC(pDrvCtrl->unit, dbgRxErrPkt);

#if DESC_BUG_RESET == 1
            /* Reset the controller when this bug is encountered */
            pDrvCtrl->aeProcessRst = 1;
            END_FLAGS_CLR (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
            DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
                    pDrvCtrl->unit, pDrvCtrl->end.flags, 103,4,5,6);

            DEBUG_STAT_INC(pDrvCtrl->unit, dmaBugReset);

            ae5311DmaReset(pDrvCtrl);

            return OK;
#endif
        } else if ( (cmdsts & (DescRxFirst|DescRxLast|AE_DESC_STATUS_RX_ERRORS)) ) {
            /*
            *   Bad Packet.
            */

            /* Increment error statistics */
            END_ERR_ADD (&pDrvCtrl->end, MIB2_IN_ERRS, +1); /* bad packet */
            DRV_LOG (DRV_DEBUG_MSG, "ae%d Rx: Bad packet 0x%8.8x <0x%8.8x> 0x%8.8x [0x%8.8x]\n",
                pDrvCtrl->unit,
                cmdsts,
                AE_DESC_STATUS_RX_ERRORS,
                (int)rxDesc,
                ae5311ReadDmaReg(pDrvCtrl,DmaStatus),
                6);
            DEBUG_STAT_INC(pDrvCtrl->unit, dbgRxErrPkt);
        } else {
            /*
            *   unknown value for cmdsts.
            */

            logMsg("SW error in ae5311Recv - invalid cmdsts! %8.8x\n",
                    cmdsts,
                    2,3,4,5,6);
        }

        /* Cleanup the descriptor and make available for reception */
        AE_DESC_STATUS_SET (rxDesc, DescOwnByDma);
        sysWbFlush();

    } /* end loop */

    HISTOGRAM_BIN(aeRx, pDrvCtrl->unit);

    /* Enable rx interrupt */
    pDrvCtrl->aeProcessRcv = 0;
    ae5311SetDmaReg(pDrvCtrl, DmaInterrupt, DmaIeRxCompleted);

    return OK;
}


/******************************************************************************
*
* ae5311MacAddressSet - Set the ethernet address
*
* Sets the ethernet address from the given address field
*
* RETURNS: void
*/

LOCAL void ae5311MacAddressSet( END_DEVICE *pDrvCtrl )
{
    unsigned int    data;
    unsigned char   *macAddr;

    macAddr = pDrvCtrl->enetAddr;

    /* set our MAC address  */
    data = (macAddr[5]<<8) | macAddr[4];
    ae5311WriteMacReg(pDrvCtrl, MacAddrHigh, data );

    data = (macAddr[3]<<24) | (macAddr[2]<<16) | (macAddr[1]<<8) | macAddr[0];
    ae5311WriteMacReg(pDrvCtrl, MacAddrLow, data );

    DRV_LOG (DRV_DEBUG_LOAD, "ae%d Verify MAC address %8.8X %8.8X \n",
                         pDrvCtrl->unit,
                         ae5311ReadMacReg(pDrvCtrl, MacAddrLow),
                         ae5311ReadMacReg(pDrvCtrl, MacAddrHigh),
                         4, 5, 6);
    DRV_LOG (DRV_DEBUG_LOAD, "  sb = %2.2X %2.2X %2.2X %2.2X %2.2X %2.2X\n",
                        0xff&macAddr[0],
                        0xff&macAddr[1],
                        0xff&macAddr[2],
                        0xff&macAddr[3],
                        0xff&macAddr[4],
                        0xff&macAddr[5]);
}


/******************************************************************************
*
* ae5311MacAddressGet - get the ethernet address
*
* Gets the ethernet address from FLASH and writes into the given address field.
*
* RETURNS: void
*/

LOCAL void ae5311MacAddressGet( END_DEVICE *pDrvCtrl )
{
    sysLanEnetAddrGet(pDrvCtrl->unit, &pDrvCtrl->enetAddr[0]);
}


/******************************************************************************
*
* ae5311QueueCreate - create a circular queue of descriptors
*
* This routine allocates a descriptor buffer array aligned  on a word
* boundary, initializes and links the array to make a circular queue.
*
* RETURNS: OK or ERROR
*/

LOCAL STATUS ae5311QueueCreate
(
    END_DEVICE      *pDrvCtrl,
    AE5311_QUEUE    *q,
    char            *pMem,
    int             count,
    int             qtype
)
{
    int         i;
    VIRT_ADDR   descAddr;
    M_BLK_ID    pMblk = NULL;   /* pointer to the mBlk */

    /* Initialize queue data */
    q->firstDescAddr = pMem;
    q->lastDescAddr = (VIRT_ADDR)((UINT32)q->firstDescAddr + (count - 1) * AE_QUEUE_ELE_SIZE);
    q->curDescAddr = q->firstDescAddr;
    q->reapDescAddr = q->lastDescAddr;
    q->count = count;

    /* Initialize each buffer descriptor, and link them into circular queue */

    for (i=0, descAddr=q->firstDescAddr;
            i<count;
            i++, descAddr=(VIRT_ADDR)((UINT32)descAddr + AE_QUEUE_ELE_SIZE))
    {
        /* Update the size, BUFPTR, and MBLKPTR fields */

        if (qtype == AE_QUEUE_TYPE_RX)
        {
            /* get a mBlk/clBlk/cluster tuple */

            pMblk = netTupleGet (pDrvCtrl->end.pNetPool, 0, M_DONTWAIT, MT_DATA, FALSE);

            if (pMblk == NULL)
            {
                DRV_LOG (DRV_DEBUG_LOAD, "ae%d queue:Error to join pMblk \n",
                    pDrvCtrl->unit,
                    2,3,4,5,6);
                ae5311QueueDelete (q);
                return (ERROR);
            }

            AE_DESC_STATUS_SET (descAddr, DescOwnByDma);
            AE_DESC_CTRLEN_SET (descAddr,
                    AE_DESC_CTRLEN_BUF1_LEN(AE_BUF_SIZE - FASTPATH_HEADROOM));

            /* Rx DMA buffer on odd half word alignment */
            (UINT32)pMblk->mBlkHdr.mData += 2 + FASTPATH_HEADROOM;

            AE_DESC_BUFPTR_SET (descAddr, (UINT32)pMblk->mBlkHdr.mData);
            AE_DESC_LNKBUF_SET (descAddr, (UINT32)0);
            AE_DESC_MBLKPTR_SET (descAddr, pMblk);
        }

        else if (qtype == AE_QUEUE_TYPE_TX)
        {
            AE_DESC_STATUS_SET (descAddr, 0);
            AE_DESC_CTRLEN_SET (descAddr, 0);

            AE_DESC_BUFPTR_SET (descAddr, (UINT32)0);
            AE_DESC_LNKBUF_SET (descAddr, (UINT32)0);
            AE_DESC_MBLKPTR_SET (descAddr, (M_BLK_ID)0);
        }
        else
        {
            DRV_LOG (DRV_DEBUG_LOAD, "ae5311QueueCreate %d: invalid queue type %d\n",
                        pDrvCtrl->unit,
                        qtype,
                        3,4,5,6);
            ae5311QueueDelete (q);
            return (ERROR);
        }

    } /* for each desc */

    /* Make the queue circular */
    AE_DESC_CTRLEN_SET (q->lastDescAddr, DescEndOfRing|AE_DESC_CTRLEN_GET(q->lastDescAddr));

    DRV_LOG (DRV_DEBUG_LOAD, "ae%d qt = %d, buf begin = %x, buf end = %x\n",
            pDrvCtrl->unit,
            qtype,
            AE_DESC_BUFPTR_GET(q->firstDescAddr),
            AE_DESC_BUFPTR_GET(q->lastDescAddr),
            5,6);

    DRV_LOG (DRV_DEBUG_LOAD, "ae%d qt = %d, drsc begin = %x, desc end = %x\n",
            pDrvCtrl->unit,
            qtype,
            (UINT32)q->firstDescAddr,
            (UINT32)q->lastDescAddr,
            5,6);

    return OK;
}


/******************************************************************************
*
* ae5311QueueDelete - frees an allocated descriptor queue
*
* This routine frees all allocated memory for a descriptor queue.
*
* RETURNS: OK
*/

LOCAL STATUS ae5311QueueDelete
(
    struct ae5311_queue *q
)
{
    int i;
    VIRT_ADDR descAddr;
    M_BLK_ID pMblk = NULL;  /* pointer to the mBlk */

    /* Free all mBlk/clBlk/cluster tuples in the queue */

    for (i=0, descAddr=q->firstDescAddr;
            (i < q->count);
            i++, descAddr += AE_QUEUE_ELE_SIZE)
    {
        pMblk = (M_BLK_ID) AE_DESC_MBLKPTR_GET (descAddr);
        if (pMblk != NULL)
            netMblkClFree (pMblk);
    }

    return (OK);
}


/*******************************************************************************
*
* ae5311Ioctl - the driver I/O control routine
*
* Process an ioctl request.
*
* RETURNS: A command specific response, usually OK or ERROR.
*/

LOCAL int ae5311Ioctl
(
    VOID    *pObj,  /* device receiving command */
    int     cmd,    /* ioctl command code */
    caddr_t data    /* command argument */
)
{
    END_DEVICE *pDrvCtrl = pObj;    /* device receiving command */
    int error = 0;
    long value;
    UINT32 savedFlags;

    DRV_LOG (DRV_DEBUG_IOCTL, "ioctl AE_CMD=0x%x\n", cmd, 2, 3, 4, 5, 6);

    switch (cmd)
    {

    /* Set device address */
    case EIOCSADDR:
        if (data == NULL)
            return (EINVAL);
        bcopy ((char *)data, (char *)END_HADDR(&pDrvCtrl->end),
                END_HADDR_LEN(&pDrvCtrl->end));
        bcopy((char*)data, (char*)&pDrvCtrl->enetAddr[0], 6);
        break;

    /* Get device address */
    case EIOCGADDR:
        if (data == NULL)
            return (EINVAL);
        bcopy ((char *)END_HADDR(&pDrvCtrl->end), (char *)data,
                END_HADDR_LEN(&pDrvCtrl->end));
        break;

    /* Set device flag */
    case EIOCSFLAGS:
        savedFlags = END_FLAGS_GET(&pDrvCtrl->end);
        value = (long)data;
        if (value < 0)
        {
            END_FLAGS_CLR (&pDrvCtrl->end, -value);
        }
        else
        {
            END_FLAGS_SET (&pDrvCtrl->end, value);
        }

        if ((END_FLAGS_GET(&pDrvCtrl->end)!= savedFlags) &&
            (END_FLAGS_GET (&pDrvCtrl->end) & IFF_UP))
        {
            END_FLAGS_CLR (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
            ae5311RxFilterConfig (pDrvCtrl);
            END_FLAGS_SET (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
        }

        DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
                pDrvCtrl->unit, pDrvCtrl->end.flags, 104,4,5,6);
        break;

    /* Get device flags */
    case EIOCGFLAGS:
        *(int *)data = END_FLAGS_GET(&pDrvCtrl->end);
        break;

    /* Begin polled operation */
    case EIOCPOLLSTART:
        ae5311PollStart (pDrvCtrl);
        break;

    /* End polled operation */
    case EIOCPOLLSTOP:
        ae5311PollStop (pDrvCtrl);
        break;

    /* return MIB information */
    case EIOCGMIB2:
        if (data == NULL)
            return (EINVAL);
        bcopy((char *)&pDrvCtrl->end.mib2Tbl, (char *)data,
                sizeof(pDrvCtrl->end.mib2Tbl));
        break;

    /* return device name */
    case EIOCGNAME:
        strcpy((char *)data, AE_DEV_NAME);
        break;

    /* return minimum First Buffer for chaining */
    case EIOCGFBUF:
        if (data == NULL)
            return (EINVAL);
        *(int *)data = AE_BUF_SIZE;

        DRV_LOG (DRV_DEBUG_IOCTL, "ae5311Ioctl EIOCGFBUF %x %x\n",
            (int)data, (int)AE_BUF_SIZE, 3, 4, 5, 6);

        break;

    default:
        error = EINVAL;
    }

    return (error);
}


/*******************************************************************************
*
* ae5311TxReap - the driver tx completion routine
*
* This routine reaps M_Blks which have been transmitted
* Assumes caller hold device lock
*
*/

LOCAL void ae5311TxReap(VOID *pObj)
{
    END_DEVICE      *pDrvCtrl = pObj;
    AE5311_QUEUE    *txq = &pDrvCtrl->txQueue;
    M_BLK_ID        pTxMblk;
    VIRT_ADDR       txDesc;
    UINT32          cmdsts;

    while (1) {
        txDesc = AE_QUEUE_ELE_NEXT_GET(txq, txq->reapDescAddr);
        if (txDesc == txq->curDescAddr) {
            break;
        }

        cmdsts = AE_DESC_STATUS_GET(txDesc);
        if (cmdsts & DescOwnByDma) {
            break;
        }

        /* Release mblk chain of the transmit */
        pTxMblk = AE_DESC_MBLKPTR_GET (txDesc);
        if (pTxMblk) {
            netMblkClChainFree(pTxMblk);
        }

        /* Update errors of the transmit */
        if (cmdsts & AE_DESC_STATUS_TX_ERRORS) {
            END_ERR_ADD (&pDrvCtrl->end, MIB2_OUT_ERRS, +1);
            DRV_LOG (DRV_DEBUG_MSG,
                "ae%d Tx: Err prev pkt 0x%8.8x <0x%8.8x> 0x%8.8x\n",
                pDrvCtrl->unit,
                cmdsts,
                AE_DESC_STATUS_TX_ERRORS,
                (int)txDesc,
                5,6);
            DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxErrPkt);
        }

        txq->reapDescAddr = txDesc;
    }
}

/*******************************************************************************
*
* ae5311Send - the driver send routine
*
* This routine takes a M_BLK_ID and sends off the data in the M_BLK_ID.
* The buffer must already have the addressing information properly installed
* in it.  This is done by a higher layer.
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311Send
(
    VOID        *pObj,  /* device ptr */
    M_BLK_ID    pMblk   /* data to send */
)
{
    END_DEVICE  *pDrvCtrl = pObj;   /* device ptr */
    VIRT_ADDR   txDesc;
    UINT32      ctrlen;
    char*       buf2 = NULL;
    int         len, lenb1 = 0;
    UINT16      tmpword;

    /* Make sure OK to attempt transmission */
    if (pDrvCtrl->flags & AE5311_POLLING) {
        DRV_MSG ("ae%d Tx: In Polling mode - drop frame\n",
            pDrvCtrl->unit,
            2,3,4,5,6);
        DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNot);

            /* toss packet */
        goto dropFrameOnError;
    }

    END_TX_SEM_TAKE (&pDrvCtrl->end, WAIT_FOREVER);

    /* Check if this port is up, else toss packet */
    if (!pDrvCtrl->aeStart) {
        DRV_LOG (DRV_DEBUG_TXM,
            "ae%d Tx Down, dropping mBlk=0x%8.8x, N=0x%8.8x, d=0x%8.8x, l=0x%8.8x, T=0x%8.8x\n",
            pDrvCtrl->unit,
            (UINT32)&pMblk,
            (UINT32)pMblk->mBlkHdr.mNext,
            (UINT32)pMblk->mBlkHdr.mData,
            (UINT32)pMblk->mBlkHdr.mLen,
            (UINT32)pMblk->mBlkPktHdr.len);

        goto dropFrameOnError;
    }

    if (pDrvCtrl->aeProcessRst) {
        DRV_MSG ("ae%d Tx: In Chip reset - drop frame\n",
            pDrvCtrl->unit,
            2,3,4,5,6);
        DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNot);

        /* toss packet */
        goto dropFrameOnError;
    }

    /* Check if we can transport this packet */
    len = max (ETHERSMALL, pMblk->mBlkPktHdr.len);  /* total length */
    if (len > ETH_MAX_PKT_SIZE) {
        DRV_MSG ("ae%d Tx: len %d too long\n",
            pDrvCtrl->unit, len,
            3,4,5,6);
        DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxLenReq);

        /* toss packet */
        goto dropFrameOnError;
    }

#ifdef DRV_DEBUG
    if (ae5311Debug & DRV_DEBUG_TXM) {
        M_BLK_ID    tmpMblkPtr;
        int n;
        for (tmpMblkPtr = pMblk, n = 0; tmpMblkPtr != 0; tmpMblkPtr = tmpMblkPtr->mBlkHdr.mNext, n++) {
            logMsg("Tx: mBlk=0x%8.8x, N=0x%8.8x, d=0x%8.8x, l=0x%8.8x, T=0x%8.8x, n=%d\n",
                (UINT32)&pMblk,
                (UINT32)tmpMblkPtr->mBlkHdr.mNext,
                (UINT32)tmpMblkPtr->mBlkHdr.mData,
                (UINT32)tmpMblkPtr->mBlkHdr.mLen,
                (UINT32)tmpMblkPtr->mBlkPktHdr.len,
                n);
            logMsg("  : len=0x%8.8x, clSize=0x%8.8x, clrefCnt=0x%8.8x\n",
                len,
                (UINT32)tmpMblkPtr->pClBlk->clSize,
                (UINT32)tmpMblkPtr->pClBlk->clRefCnt,
                4,5,6);
        }
    }
#endif

    /* Do some tx reaping */
    ae5311TxReap(pObj);

    txDesc = pDrvCtrl->txQueue.curDescAddr;
    if (txDesc == pDrvCtrl->txQueue.reapDescAddr) {
        END_ERR_ADD (&pDrvCtrl->end, MIB2_OUT_ERRS, +1);
        DRV_LOG (DRV_DEBUG_TX,
            "ae%d Tx: cannot get txDesc\n",
            pDrvCtrl->unit,
            2,3,4,5,6);
        DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNoDesc);

        /* toss packet */
        goto dropFrameOnError;
    }

    /* gather data spread across 3 or more buffers in software via
     * bcopy; DMA gathering 2 buffers is easy; TODO: do it for the
     * 3+ case also
     */
    if (pMblk->mBlkHdr.mNext) {
        if (pMblk->mBlkHdr.mNext->mBlkHdr.mNext) {
            M_BLK_ID pTxMblk;

            pTxMblk = netTupleGet(pDrvCtrl->end.pNetPool,
                        AE_BUF_SIZE, M_DONTWAIT, MT_DATA, FALSE);
            if (pTxMblk == NULL) {
                END_ERR_ADD (&pDrvCtrl->end, MIB2_OUT_ERRS, +1);
                DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNoMblk);
                DRV_LOG (DRV_DEBUG_TX,
                    "ae%d Tx: cannot get mblk\n",
                    pDrvCtrl->unit,
                    2,3,4,5,6);

                /* toss packet */
                goto dropFrameOnError;
            }

            pTxMblk->mBlkHdr.mFlags |= M_PKTHDR;
            pTxMblk->mBlkHdr.mLen = len;
            pTxMblk->mBlkPktHdr.len = len;
            netMblkToBufCopy (pMblk, pTxMblk->mBlkHdr.mData, NULL);

            netMblkClChainFree (pMblk);
            pMblk = pTxMblk;
        } else {
            lenb1 = pMblk->mBlkHdr.mLen;
            buf2 = pMblk->mBlkHdr.mNext->mBlkHdr.mData;
        }
    }

    /* we won't fail now; so we can consume this descriptor */
    AE_QUEUE_ELE_NEXT((&pDrvCtrl->txQueue));

#if TX_TWO_BUF == 1
    if (!buf2) {
        /* break tx dma into two blocks, the second being sdram burst aligned */
        lenb1 = 0x10 - ((UINT32)pMblk->mBlkHdr.mData & 0x0f);
        buf2 = (char *)pMblk->mBlkHdr.mData + lenb1;
    }
#endif

    /* Update the descriptor */
    AE_DESC_BUFPTR_SET (txDesc, (UINT32)pMblk->mBlkHdr.mData);
    AE_DESC_MBLKPTR_SET (txDesc, pMblk);
    ctrlen = AE_DESC_CTRLEN_GET (txDesc);
    ctrlen = (ctrlen & (DescEndOfRing))
            | DescTxFirst
            | DescTxLast
            | DescTxIntEnable;

    if (buf2) {
        AE_DESC_LNKBUF_SET (txDesc, (UINT32)buf2);
        ctrlen |= ((lenb1 << DescSize1Shift) & DescSize1Mask)
                | (((len - lenb1) << DescSize2Shift) & DescSize2Mask);
    } else {
        ctrlen |= ((len << DescSize1Shift) & DescSize1Mask);
    }

    AE_DESC_CTRLEN_SET (txDesc, ctrlen);
    AE_DESC_STATUS_SET (txDesc, DescOwnByDma);
    sysWbFlush();

    DRV_LOG (DRV_DEBUG_TX,
        "ae%d Tx: Desc=0x%8.8x, L=0x%8.8x, D=0x%8.8x, d=0x%8.8x, len=0x%8.8x\n",
        pDrvCtrl->unit,
        (UINT32)txDesc,
        AE_DESC_CTRLEN_GET(txDesc),
        (UINT32)pMblk->mBlkHdr.mData,
        AE_DESC_LNKBUF_GET(txDesc),
        len);

    /* DMA Tx resume */
    ae5311WriteDmaReg(pDrvCtrl, DmaTxPollDemand, 0);
    sysWbFlush();

    /* Count packet */
    DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxSentPkt);
    tmpword = (UINT16)(*((unsigned short *)pMblk->mBlkHdr.mData));
    if (tmpword & 0x0100) {
        pDrvCtrl->end.mib2Tbl.ifOutNUcastPkts += 1;
    } else {
        END_ERR_ADD (&pDrvCtrl->end, MIB2_OUT_UCAST, +1);
    }

    /* Release the TX semaphore */
    END_TX_SEM_GIVE (&pDrvCtrl->end);

    return OK;

dropFrameOnError:

    END_TX_SEM_GIVE (&pDrvCtrl->end);
    netMblkClChainFree (pMblk);
    return ERROR;
}


/******************************************************************************
*
* ae5311Config - reconfigure the interface under us.
*
* Install the Tx and Rx Queues and the Interrupt vector.
* Reconfigure the interface setting promiscuous mode, and changing the
* multicast interface list.
*
* This routine can be called only once after the device has been initialized.
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311Config
(
    END_DEVICE *pDrvCtrl    /* device to be re-configured */
)
{
    unsigned int    regtmp;
    unsigned int    mask;
    #if KEEP_BAD_FRAME == 1
    unsigned int    mac;
    #endif

    /*
        Bring MAC and PHY out of reset
    */
    mask = (0 == pDrvCtrl->unit) ?
        (RESET_ENET0|RESET_EPHY0) : (RESET_ENET1|RESET_EPHY1);

    regtmp = sysRegRead(AR5311_RESET);
    sysRegWrite(AR5311_RESET, regtmp | mask);
    sysMsDelay(15);
    regtmp = sysRegRead(AR5311_RESET);
    sysRegWrite(AR5311_RESET, regtmp & ~mask);
    sysUDelay(25);

    /* Setup phy capabilities */
    if (ae_PhySetup (pDrvCtrl) != OK)
    {
        return ERROR;
    }

    /*
    *   Init AE5311 DMA Engine
    */

    /* Reset DMA engine  */
    ae5311WriteDmaReg(pDrvCtrl, DmaBusMode, DmaResetOn);

    /* Delay 2 usec */
    sysUDelay( 2 );

    /* Set init register values  */
    ae5311WriteDmaReg(pDrvCtrl, DmaBusMode, DmaBusModeInit);

    /* Install the first Tx and Rx queues on the device */
    ae5311WriteDmaReg(pDrvCtrl, DmaRxBaseAddr,
                (UINT32)pDrvCtrl->rxQueue.firstDescAddr);
    ae5311WriteDmaReg(pDrvCtrl, DmaTxBaseAddr,
                (UINT32)pDrvCtrl->txQueue.firstDescAddr);

#if TX_OSF == 2
    ae5311WriteDmaReg(pDrvCtrl, DmaControl, DmaTxSecondFrame | DmaTxThreshCtl256);
#elif TX_OSF == 1
    ae5311WriteDmaReg(pDrvCtrl, DmaControl, DmaTxThreshCtl256);
#else
    ae5311WriteDmaReg(pDrvCtrl, DmaControl, DmaStoreAndForward);
#endif

    ae5311WriteDmaReg(pDrvCtrl, DmaInterrupt, DmaIntDisable);

    /*
    *   Init AE5311 MAC
    */

#if 0   /* FOO fix Hash table */
    int n;

    /* clear hash table  */
    ae5311WriteMacReg(pDrvCtrl, MacHashHigh, 0);
    ae5311WriteMacReg(pDrvCtrl, MacHashLow, 0);

    /* set the broadcast address  */
    n = Hash( Broadcast);

    if( n>=32 )
        ae5311SetMacReg(pDrvCtrl, MacHashHigh, 1<<(n-32));
    else
        ae5311SetMacReg(pDrvCtrl, MacHashLow, 1<<n);
#endif

    /*
        Set Mac and Phy modes.
    */

    /* default values */
    pDrvCtrl->flags |= (AE5311_FULL_DUPLEX | AE5311_SPEED_100);
    ae5311WriteMacReg(pDrvCtrl, MacControl, MacControlInitFdx);

    ae_SetMacFromPhy(pDrvCtrl);

    DRV_PRINT(DRV_DEBUG_PHYSET,("\nae %d AN complete; flags 0x%x\n",
        pDrvCtrl->unit,(int)pDrvCtrl->flags));

    #if KEEP_BAD_FRAME == 1
    mac = ae5311ReadMacReg(pDrvCtrl,MacControl);
    mac |= MacBadFramesEnable;  /* for debug */
    ae5311WriteMacReg(pDrvCtrl,MacControl, mac);
    #endif

    ae5311DmaIntClear(pDrvCtrl);        /* clear interrupt requests  */

    DRV_LOG (DRV_DEBUG_LOAD,"ae%d: setting TXDP=0x%x RXDP=0x%x\n",
            pDrvCtrl->unit,
            (UINT32)(pDrvCtrl->txQueue.firstDescAddr),
            (UINT32)(pDrvCtrl->rxQueue.firstDescAddr),
            4, 5, 6);

    /* VLAN tag aware, so as to increase max rx frame size from 1518 to 1522. */
    ae5311WriteMacReg(pDrvCtrl,MacVlan1,0x8100);

    /* Setup the ethernet address */
    ae5311MacAddressSet (pDrvCtrl);

    /* Receive filtering setup */
    ae5311RxFilterConfig (pDrvCtrl);

    return OK;
}


/*****************************************************************************
*
* ae5311MCastAdd - add a multicast address for the device
*
* This routine adds a multicast address to whatever the driver
* is already listening for.  It then resets the address filter.
*
* RETURNS: OK or ERROR.
*/
/* FOO -- unsupported */
LOCAL STATUS ae5311MCastAdd
(
    VOID    *pObj,      /* device pointer */
    char    *pAddress   /* new address to add */
)
{
    END_DEVICE *pDrvCtrl = pObj;    /* device pointer */
    int error;

    if ((error = etherMultiAdd (&pDrvCtrl->end.multiList,
                pAddress)) == ENETRESET)
        ae5311RxFilterConfig (pDrvCtrl);

    return (OK);
}


/*****************************************************************************
*
* ae5311MCastDel - delete a multicast address for the device
*
* This routine removes a multicast address from whatever the driver
* is listening for.  It then resets the address filter.
*
* RETURNS: OK or ERROR.
*/
/* FOO -- unsupported */
LOCAL STATUS ae5311MCastDel
(
    VOID    *pObj,      /* device pointer */
    char    *pAddress   /* address to be deleted */
)
{
    END_DEVICE *pDrvCtrl = pObj;    /* device pointer */
    int error;

    if ((error = etherMultiDel (&pDrvCtrl->end.multiList,
                (char *)pAddress)) == ENETRESET)
        ae5311RxFilterConfig (pDrvCtrl);

    return (OK);
}


/*****************************************************************************
*
* ae5311MCastGet - get the multicast address list for the device
*
* This routine gets the multicast list of whatever the driver
* is already listening for.
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311MCastGet
(
    void        *pObj,      /* device pointer */
    MULTI_TABLE *pTable     /* address table to be filled in */
)
{
    END_DEVICE *pDrvCtrl;   /* device pointer */
    pDrvCtrl = pObj;

    return (etherMultiGet (&pDrvCtrl->end.multiList, pTable));
}


/*******************************************************************************
*
* ae5311Stop - stop the device
*
* This function calls BSP functions to disconnect interrupts and stop
* the device from operating in interrupt mode.
*
* RETURNS: OK or ERROR.
*/
LOCAL STATUS ae5311Stop( VOID *pObj ) /* device to be stopped */
{
    END_DEVICE  *pDrvCtrl;  /* device to be stopped */
    STATUS      result;

    pDrvCtrl = (END_DEVICE *)pObj;  /* device to be stopped */
    result = OK;

    /* Disable device interrupts prior to any errors during stop */
    ae5311WriteDmaReg(pDrvCtrl,DmaInterrupt,0);
    intDisable(pDrvCtrl->ilevel);

    /* Disable tx and tx */
    ae5311DmaRxStop(pDrvCtrl);
    ae5311DmaTxStop(pDrvCtrl);
    ae5311ClearMacReg(pDrvCtrl,MacControl,(MacRxEnable|MacTxEnable));

    /* Disconnect interrupt routine */
    /* SYS_INT_DISCONNECT */
    (void)intConnect(INUM_TO_IVEC(pDrvCtrl->ivec), ethStrayInt, pDrvCtrl->ivec);

    pDrvCtrl->aeStart = 0;
    END_FLAGS_CLR (&pDrvCtrl->end, IFF_UP | IFF_RUNNING);
    DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
        pDrvCtrl->unit, pDrvCtrl->end.flags, 105,4,5,6);


    return (result);
}


/******************************************************************************
*
* ae5311Unload - unload a driver from the system
*
* This function first brings down the device, and then frees any
* stuff that was allocated by the driver in the load function.
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311Unload
(
    void *pObj  /* device to be unloaded */
)
{
    END_DEVICE *pDrvCtrl = pObj;    /* device to be unloaded */

    /* Bring down the device */

    END_OBJECT_UNLOAD (&pDrvCtrl->end);

    /* free memory allocated for clusters and Tx/Rx Queues */

    if ((pDrvCtrl->flags & AE5311_BUF_LOCAL) &&
        (pDrvCtrl->pDmaBuf != NULL))
    {
        cacheDmaFree (pDrvCtrl->pDmaBuf);
    }

    /* free memory allocated for mBlks and cluster blocks */

    if (pDrvCtrl->pMclMem)
        free (pDrvCtrl->pMclMem);

    /* free memory allocated for net pool */
    if (pDrvCtrl->end.pNetPool != NULL)
        free (pDrvCtrl->end.pNetPool);

    #if defined (DRV_DEBUG) || (PHY_POLLED)
    pCtrlAe[pDrvCtrl->unit] = (END_DEVICE *)0;
    refPhyPollTask--;
    /* FOO when the last enet device is unloaded, stop ae_PhyPollTask */
    #endif

    /* free memory allocated for device structure */
    if (pDrvCtrl != NULL)
        free ((char*) pDrvCtrl);

    return (OK);
}


/*******************************************************************************
*
* ae5311PollStart - start polled mode operations
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311PollStart
(
    void *pObj  /* device to be polled */
)
{
    END_DEVICE  *pDrvCtrl = pObj;   /* device to be started */
    UINT        oldLevel;

    if(pDrvCtrl->aeProcessRst != 0)
        return ERROR;

    /*
     * In polled mode the send always copies into the
     * the device's pollMblk and then sends it. This is done
     * so that we don't do any mblk alloc/free in poll send.
     * This is important because poll mode is used by debugger and
     * called in system mode. We always want to have an mblk otherwise
     * debugging would not work in case we were out of mblk/buffers.
     * Therefore we allocate it once on the first call to pollStart.
     * NOTE: In the current implementation pollMblk is never freed
     * so yes we'll lose one buffer. But this is less of concern than
     * not being able to debug. So please do not free pollMblk.
     * In short, each Ethernet device, that is in polling mode,
     * will always have one mblk alloced at pollStart time and from then
     * on it will use that mblk to copy into during ae5311PollSend
     */

     if(pDrvCtrl->pollMblk == NULL) {
         /*
          * First call to poll start so allocate a buffer.
          * This will never be freed.
          */
         pDrvCtrl->pollMblk = netTupleGet(pDrvCtrl->end.pNetPool,
             AE_BUF_SIZE, M_DONTWAIT, MT_DATA, FALSE);
         if (pDrvCtrl->pollMblk == NULL) {
             DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNoMblk);
             DRV_LOG (DRV_DEBUG_TX, "ae%d Poll Start: cannot get mblk\n",
                    pDrvCtrl->unit,2,3,4,5,6);
             return ERROR;
            }
     }

    oldLevel = intLock();

    pDrvCtrl->flags |= AE5311_POLLING;
    ae5311WriteDmaReg(pDrvCtrl, DmaInterrupt, 0);
    sysWbFlush();

    intUnlock (oldLevel);

    ae5311TxReap(pDrvCtrl);

    DRV_POLL_PRINT(DRV_DEBUG_POLL_MODE, "ae%d PollStart ",
        pDrvCtrl->unit,
        2, 3, 4, 5, 6);

    return (OK);
}


/*******************************************************************************
*
* ae5311PollStop - stop polled mode operations
*
* This function terminates polled mode operation.  The device returns to
* interrupt mode.
*
* The device interrupts are enabled, the current mode flag is switched
* to indicate interrupt mode and the device is then reconfigured for
* interrupt operation.
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae5311PollStop
(
    void *pObj  /* device to be polled */
)
{
    END_DEVICE  *pDrvCtrl = pObj;   /* device to be started */
    UINT        oldLevel;

    if(pDrvCtrl->aeProcessRst != 0)
        return ERROR;

    /* Turn on device interrupts */
    oldLevel = intLock();
    pDrvCtrl->flags &= ~AE5311_POLLING;

    /* Turn on device interrupts -- enable most errors */
    ae5311WriteDmaReg(pDrvCtrl, DmaStatus, ae5311ReadDmaReg(pDrvCtrl, DmaStatus));  /* clear all interrupt requests */
    ae5311WriteDmaReg(pDrvCtrl, DmaInterrupt, DmaIntEnable);
    sysWbFlush();

    intUnlock (oldLevel);

    DRV_POLL_PRINT(DRV_DEBUG_POLL_MODE, "ae%d PollStop ",
        pDrvCtrl->unit,
        2, 3, 4, 5, 6);

    return (OK);
}


/*******************************************************************************
*
* ae5311PollRecv - routine to receive a packet in polled mode.
*
* This routine is called by a user to try and get a packet from the
* device.
*
* RETURNS: OK upon success.  EAGAIN is returned when no packet is available.
*/

LOCAL STATUS ae5311PollRecv
(
    void        *pObj,  /* device to be polled */
    M_BLK_ID    pMblk   /* ptr to buffer */
)
{
    END_DEVICE  *pDrvCtrl = (END_DEVICE *) pObj;
    UINT32      cmdsts = 0;
    int         len;
    VIRT_ADDR   rxDesc;
    M_BLK_ID    pRxMblk;
    UINT        regIsr;

    if(pDrvCtrl->aeProcessRst != 0)
        return ERROR;

    /* Read the Status */

    regIsr = ae5311ReadDmaReg(pDrvCtrl, DmaStatus);
    if (regIsr)
        DRV_POLL_PRINT (DRV_DEBUG_POLL_RX, "ae%d Isr=0x%x ",
                        pDrvCtrl->unit,
                        regIsr,
                        3, 4, 5, 6);


    rxDesc = ae5311RxDescGet (pDrvCtrl);
    if (rxDesc == NULL)
        return EAGAIN;

        DRV_POLL_PRINT (DRV_DEBUG_POLL_RX, "ae%d Poll rxDesc = 0x%8.8X ",
                pDrvCtrl->unit,,
                (UINT32)rxDesc,
                4, 5, 6);

    cmdsts = AE_DESC_STATUS_GET (rxDesc);

    if ((cmdsts & AE_DESC_STATUS_RX_ERRORS) != 0) {
        /*  Cleanup and make available to receive */
        AE_DESC_STATUS_SET (rxDesc, DescOwnByDma);
        sysWbFlush();
        ae5311DmaRxStart(pDrvCtrl);
        DRV_POLL_PRINT (DRV_DEBUG_POLL_RX, "ae%d Poll Recv dropped ",
                        pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return EAGAIN;
    }


    /* Process the packet */

    len = AE_DESC_STATUS_RX_SIZE(cmdsts) - ETH_CRC_LEN;

    if (!pMblk ||
        (pMblk->mBlkHdr.mLen < len) ||
        (!(pMblk->mBlkHdr.mFlags & M_EXT))) {

        /* Cleanup and make available to receive */
        AE_DESC_STATUS_SET (rxDesc, DescOwnByDma);
        sysWbFlush();
        ae5311DmaRxStart(pDrvCtrl);

        DRV_POLL_PRINT (DRV_DEBUG_POLL_RX, "ae%d Poll Recv Bad Mblk \n\r",
                        pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return EAGAIN;
    }

    pRxMblk = (M_BLK_ID) AE_DESC_MBLKPTR_GET (rxDesc);
    A_DATA_CACHE_INVAL (pRxMblk->mBlkHdr.mData, len);

    /* Copy received data and update the MBLK */
    bcopy (pRxMblk->mBlkHdr.mData, pMblk->mBlkHdr.mData, len);
    pMblk->mBlkHdr.mLen    = len;
    pMblk->mBlkHdr.mFlags |= M_PKTHDR;
    pMblk->mBlkPktHdr.len  = len;

    if (cmdsts & DescRxMulticast)
    {
        if (IS_BROADCAST(pRxMblk->mBlkHdr.mData))
        {
            pMblk->mBlkHdr.mFlags |= M_BCAST;
        }
        else
        {
            pMblk->mBlkHdr.mFlags |= M_MCAST;
        }
    }

    /* Cleanup the descriptor and make available for reception */
    AE_DESC_STATUS_SET (rxDesc, DescOwnByDma);
    sysWbFlush();

    ae5311DmaRxStart(pDrvCtrl);

    DRV_POLL_PRINT (DRV_DEBUG_POLL_RX, "ae%d Poll RX OK\r\n",
                    pDrvCtrl->unit, 2, 3, 4, 5, 6);

    return OK;
}


/*******************************************************************************
*
* ae5311PollSend - routine to send a packet in polled mode.
*
* This routine sends data in polled mode and assumes the caller
* has already called ae5311PollStart. It allows for one send at a time
* only. No queuing is allowed.  Caller's buffer is not freed. This is important
* because that is what the WDB agent (prime client of this
* routine) expects.  This routine always copies caller's data into
* the devices pollMblk (see ae5311PollStart). This allows us to
* send single or chained mblks without requiring mblk alloc/free in this
* routine. pollMblk is a single mblk alloced once for this device when
* ae5311PollStart was called for the first time.
*
* RETURNS: OK upon success.  EAGAIN if device is busy.
*/

LOCAL STATUS ae5311PollSend
(
    void        *pObj,  /* device to be polled */
     M_BLK_ID   pMblk   /* packet to send */
)
{
    END_DEVICE  *pDrvCtrl = pObj;   /* device ptr */
    M_BLK_ID     pollMblk = pDrvCtrl->pollMblk;
    VIRT_ADDR   txDesc;
    UINT32      ctrlen;
    int         len, tries;

#define MAX_TRIES 1000000

    /* Make sure OK to attempt transmission */
    if(pDrvCtrl->aeProcessRst != 0)
    {
        DRV_POLL_PRINT (DRV_DEBUG_POLL_TX, "ae%d Poll TX: Chip reset\n",
            pDrvCtrl->unit,
            2,3,4,5,6);
        DEBUG_STAT_INC(pDrvCtrl->unit, dbgTxNot);
        return EAGAIN;
    }

    if ((pDrvCtrl->flags & AE5311_POLLING) == 0)
    {
        DRV_POLL_PRINT (DRV_DEBUG_POLL_TX, "ae%d Poll TX: ErrorNotPolling\r\n",
                        pDrvCtrl->unit,
                        2, 3, 4, 5, 6);
        return ERROR;
    }

    if (pollMblk == NULL) {
        DRV_POLL_PRINT (DRV_DEBUG_POLL_TX,
            "ae%d Poll TX Error: No pollMblk \r\n",
            pDrvCtrl->unit, 2, 3, 4, 5, 6);
        return ERROR;
    }

    len = max (ETHERSMALL, pMblk->mBlkPktHdr.len);  /* total length */
    if (len > ETH_MAX_PKT_SIZE) {
        #ifdef DRV_DEBUG
        aeDbg[pDrvCtrl->unit].dbgTxLenReq++;
        #endif
        return ERROR;
    }

    /*
     * Always copy into our pollMblk. Polling is not a performance
     * critical data-path.
     */

    netMblkToBufCopy (pMblk, pollMblk->mBlkHdr.mData, NULL);

    txDesc = ae5311TxDescGet (pDrvCtrl);
    if (txDesc == NULL)
    {
        DRV_POLL_PRINT (DRV_DEBUG_POLL_TX, "ae%d Poll TX: No txDesc\r\n",
                        pDrvCtrl->unit,
                        2, 3, 4, 5, 6);
        return EAGAIN;
    }

    /*
     * Set the buffer pointer to be the pollMblk buffer.
     * This is a poll routine and we can only have one
     * TX outstanding. No queuing. Don't care about the
     * mblk pointer. Next reap will free the original
     * mblk at this desc.
     */
    AE_DESC_BUFPTR_SET (txDesc,(UINT32)pMblk->mBlkHdr.mData);

    ctrlen = AE_DESC_CTRLEN_GET (txDesc);
    ctrlen &= (DescEndOfRing);
    ctrlen |= ((len << DescSize1Shift) & DescSize1Mask)
            | DescTxFirst
            | DescTxLast;
    AE_DESC_CTRLEN_SET (txDesc, ctrlen);
    AE_DESC_STATUS_SET (txDesc, DescOwnByDma);
    sysWbFlush();


    /* DMA Tx resume */
    ae5311WriteDmaReg(pDrvCtrl,DmaTxPollDemand, 0);
    sysWbFlush();


    /* Wait until the packet is transmitted */
    tries = MAX_TRIES;
    while (--tries) {
        if ((AE_DESC_STATUS_GET (txDesc) & DescOwnByDma) == 0) {
            break;
        }
        DRV_POLL_PRINT (DRV_DEBUG_POLL_TX, "ae%d +",
            pDrvCtrl->unit,
            2, 3, 4, 5, 6);
    }

    if (0 == tries) {
        printf("ae%d: max tries exceeded in poll send\n", pDrvCtrl->unit);
        return ERROR;
    }

    return OK;
}


/******************************************************************************
*
* ae5311RxFilterConfig - configure rx packet filtering
*
* Enables appropritate receive filtering on the device.
*
* RETURNS: N/A
*/
/* FOO -- not finished. */
LOCAL void ae5311RxFilterConfig
(
    END_DEVICE *pDrvCtrl
)
{
    UINT32  mac;
    UINT32  devFlags = END_FLAGS_GET (&pDrvCtrl->end);

    DRV_LOG (DRV_DEBUG_ENDFLAG,"eth%d: RxFilterConfig devFlags=0x%8.8x\n",
            pDrvCtrl->unit, devFlags, 3, 4, 5, 6);

    /* default mode: normal unicast, perfect filtering, broadcast */
    mac = ae5311ReadMacReg(pDrvCtrl,MacControl);

    /* Broadcast */
    if (devFlags & IFF_BROADCAST)
    {
        mac &= ~MacBroadcastDisable;
    }
    else
    {
        mac |= MacBroadcastDisable;
    }

    /* Promiscuous */
    if (devFlags & IFF_PROMISC)
    {
        mac |= MacPromiscuousModeOn;
    }
    else
    {
        mac &= ~MacPromiscuousModeOn;
    }

    /* All multicast */
    if (devFlags & IFF_ALLMULTI)
    {
        mac |= MacMulticastFilterOff;   /* filter off means accept all m/c */
    }
    else
    {
        mac &= ~MacMulticastFilterOff;

        /* Setup to receive programmed multicast packets */

        /* clear hash table */
        ae5311WriteMacReg(pDrvCtrl, MacHashHigh, 0);
        ae5311WriteMacReg(pDrvCtrl, MacHashLow, 0);

        /* install the hash table */
    }

    ae5311WriteMacReg(pDrvCtrl,MacControl,mac);
    DRV_LOG (DRV_DEBUG_LOAD,"ae5311RxFilterConfig: MacControl = %x\n",
        mac,2,3,4,5,6);

    return;
}


/******************************************************************************
*
* ae5311TxDescGet - get a valid transmit descriptor
*
* This routine returns the current descriptor from the txQueue if driver is
* the owner, else returns NULL
*
* RETURNS: TxDesc or NULL
*
*/

LOCAL VIRT_ADDR ae5311TxDescGet
(
    END_DEVICE *pDrvCtrl
)
{
    AE5311_QUEUE    *q;
    VIRT_ADDR       descAddr = NULL;

    q = &pDrvCtrl->txQueue;

    /* Check if we own the descriptor */

    if (((AE_DESC_STATUS_GET(q->curDescAddr)) & DescOwnByDma) == 0)
    {
        descAddr = q->curDescAddr;
        AE_QUEUE_ELE_NEXT (q);                 /* Move to the next element */
    }

    return descAddr;
}


/******************************************************************************
*
* ae5311RxDescGet - get a valid receive descriptor
*
* This routine returns the current descriptor from the rxQueue if driver is
* the owner, else returns NULL
*
* RETURNS: RxDesc or NULL
*
*/

LOCAL VIRT_ADDR ae5311RxDescGet
(
    END_DEVICE *pDrvCtrl
)
{
    AE5311_QUEUE    *q;
    VIRT_ADDR       descAddr = NULL;

    q = &pDrvCtrl->rxQueue;

    /* Check if we own the descriptor */

    if ( 0 == ((AE_DESC_STATUS_GET (q->curDescAddr)) & DescOwnByDma) )
    {
        descAddr = q->curDescAddr;
        AE_QUEUE_ELE_NEXT(q);                 /* Move to the next element */
    }

    return descAddr;
}


/******************************************************************************
*
* ae_PhySetup - reset and setup the PHY device
*
* resets the PHY chip
*
* RETURNS: OK or ERROR.
*/

LOCAL STATUS ae_PhySetup
(
    END_DEVICE * pDrvCtrl
)
{
    UINT16  phySts = 0;
    UINT16  timeout;

    /* Reset phy */
    ae5311MiiWrite(pDrvCtrl,GEN_ctl,PHY_SW_RST|AUTONEGENA);

    sysMsDelay(1500);


    /* Wait 6s */
    for (timeout=40; timeout; timeout--)
    {
        /* wait for AN to complete */
        phySts = (UINT16)ae5311MiiRead(pDrvCtrl,GEN_sts);
        if (AUTOCMPLT & phySts)
            break;

        sysMsDelay(150);
    }
    if (timeout == 0) {
        DRV_LOG (DRV_DEBUG_LOAD, "ae%d Phy Autonegotiate Failed\n",
            pDrvCtrl->unit,
            2,3,4,5,6);
        printf("Please check the ethernet cable on port %d!\n", pDrvCtrl->unit);
    }


    /* Check for link */
    if (phySts & LINK)
        pDrvCtrl->flags |= AE5311_LINK_UP;
    else
        pDrvCtrl->flags &= ~AE5311_LINK_UP;

    DRV_LOG (DRV_DEBUG_LOAD,
            "eth%d: duplex %d, link %d, S=%4.4x\n",
                pDrvCtrl->unit,
                (pDrvCtrl->flags) & AE5311_FULL_DUPLEX ? 1 : 0,
                (pDrvCtrl->flags) & AE5311_LINK_UP ? 1 : 0,
                phySts,
                5,6);

    return (OK);
}


/******************************************************************************
*
* aeResetOnError - reset the AE if fatal error happens
*
* RETURNS: N/A
*/
LOCAL void aeResetOnError( END_DEVICE* pDrvCtrl )
{

    DRV_LOG(DRV_DEBUG_MSG, "aeResetOnError: reset ae5311 %d\n",
        pDrvCtrl->unit,2,3,4,5,6);

    ae5311DmaReset(pDrvCtrl);
}


/******************************************************************************
*
* ae_SetMacFromPhy - read Phy settings and update Mac and dev->flags
*                    with current duplex and speed.
*
* RETURNS:
*/

LOCAL STATUS ae_SetMacFromPhy(END_DEVICE *pDev)
{
    UINT16  phyCtl;
    UINT32  macCtl;
    UINT16  phyLpa;
    unsigned int oldLevel;


    oldLevel = intLock();

    /* Get duplex mode from Phy */

    phyCtl = (UINT16)ae5311MiiRead(pDev,GEN_ctl);

    /* Flag is set for full duplex mode, else cleared */
    macCtl = ae5311ReadMacReg(pDev, MacControl);

    if (phyCtl & DUPLEX)
    {
        pDev->flags |= AE5311_FULL_DUPLEX;

        /* set values of control registers */
        macCtl &= ~MacDisableRxOwn;
        macCtl |= MacFullDuplex;
        ae5311WriteMacReg(pDev, MacControl, macCtl);
        ae5311WriteMacReg(pDev, MacFlowControl, MacFlowControlInitFdx);
    }
    else
    {
        pDev->flags &= ~AE5311_FULL_DUPLEX;

        /* set values of control registers */
        ae5311WriteMacReg(pDev, MacFlowControl, MacFlowControlInitHdx);
        macCtl |= MacDisableRxOwn;
        macCtl &= ~MacFullDuplex;
        ae5311WriteMacReg(pDev, MacControl, macCtl);
    }


    /* Get speed from Phy. */

    phyLpa = (UINT16)ae5311MiiRead(pDev,AN_lpa);

    /* Flag is set for 100 Mbs mode, else cleared */
    if (phyLpa & (LPA_TXFD | LPA_TX))
        pDev->flags |= AE5311_SPEED_100;
    else
        pDev->flags &= ~AE5311_SPEED_100;

    intUnlock (oldLevel);

    return OK;
}


/******************************************************************************
*
* ae5311ReadMacReg - read AE MAC register
*
* RETURNS: register value
*/

LOCAL UINT32 ae5311ReadMacReg(END_DEVICE *pDev, UINT32 reg)
{
    UINT32 data;

    data = RegRead(pDev->macBase + reg);
    return data;
}


/******************************************************************************
*
* ae5311WriteMacReg - write AE MAC register
*
* RETURNS: N/A
*/

LOCAL void ae5311WriteMacReg(END_DEVICE *pDev, UINT32 reg, UINT32 data)
{
    RegWrite(data,pDev->macBase+reg);
}


/******************************************************************************
*
* ae5311SetMacReg - set bits in AE MAC register
*
* RETURNS: N/A
*/

LOCAL void ae5311SetMacReg(END_DEVICE *pDev, UINT32 reg, UINT32 val)
{
    UINT32 addr = pDev->macBase+reg;
    UINT32 data = RegRead(addr);

    data |= val;
    RegWrite(data,addr);
}

/******************************************************************************
*
* ae5311ClearMacReg - clear bits in AE MAC register
*
* RETURNS: N/A
*/

LOCAL void ae5311ClearMacReg(END_DEVICE *pDev, UINT32 reg, UINT32 val)
{
    UINT32 addr = pDev->macBase+reg;
    UINT32 data = RegRead(addr);

    data &= ~val;
    RegWrite(data,addr);
}


/******************************************************************************
*
* ae5311ReadDmaReg - read AE DMA register
*
* RETURNS: register value
*/

LOCAL UINT32 ae5311ReadDmaReg(END_DEVICE *pDev,UINT32 reg)
{
    UINT32 data = RegRead(pDev->dmaBase+reg);
    return data;
}


/******************************************************************************
*
* ae5311WriteDmaReg - write AE DMA register
*
* RETURNS: N/A
*/

LOCAL void ae5311WriteDmaReg(END_DEVICE *pDev, UINT32 reg, UINT32 data)
{
    RegWrite(data,pDev->dmaBase+reg);
}


/******************************************************************************
*
* ae5311SetDmaReg - set bits in an AE DMA register
*
* RETURNS: N/A
*/

LOCAL void ae5311SetDmaReg(END_DEVICE *pDev, UINT32 reg, UINT32 val)
{
    UINT32 addr = pDev->dmaBase+reg;
    UINT32 data = RegRead(addr);
    data |= val;
    RegWrite(data,addr);
}


/******************************************************************************
*
* ae5311ClearDmaReg - clear bits in an AE DMA register
*
* RETURNS: N/A
*/

LOCAL void ae5311ClearDmaReg(END_DEVICE *pDev, UINT32 reg, UINT32 val)
{
    UINT32 addr = pDev->dmaBase+reg;
    UINT32 data = RegRead(addr);
    data &= ~val;
    RegWrite(data,addr);
}


/******************************************************************************
*
* ae5311MiiRead - read AE Mii register
*
* RETURNS: register value
*/

LOCAL UINT16 ae5311MiiRead(END_DEVICE *pDev, UINT8 reg)
{
    UINT32 addr;
    UINT16 data;

    addr = ((pDev->phyAddr << MiiDevShift) & MiiDevMask) | ((reg << MiiRegShift) & MiiRegMask);

    ae5311WriteMacReg(pDev, MacMiiAddr, addr );
    do{}
        while( (ae5311ReadMacReg(pDev, MacMiiAddr ) & MiiBusy) == MiiBusy );

    data = ae5311ReadMacReg(pDev, MacMiiData ) & 0xFFFF;

    return data;
}


/******************************************************************************
*
* ae5311MiiWrite - write AE Mii register
*
* RETURNS: N/A
*/

LOCAL void ae5311MiiWrite(END_DEVICE *pDev, UINT8 reg, UINT16 data)
{
    UINT32 addr;

    ae5311WriteMacReg(pDev, MacMiiData, data );

    addr = ((pDev->phyAddr << MiiDevShift) & MiiDevMask) | ((reg << MiiRegShift) & MiiRegMask) | MiiWrite;
    ae5311WriteMacReg(pDev, MacMiiAddr, addr );

    do{}
        while( (ae5311ReadMacReg(pDev, MacMiiAddr ) & MiiBusy) == MiiBusy );
}


/******************************************************************************
*
* ae5311DmaRxStart - Start Rx
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaRxStart(END_DEVICE *pDev)
{
    ae5311SetDmaReg(pDev,DmaControl,DmaRxStart);
    sysWbFlush();
}


/******************************************************************************
*
* ae5311DmaRxStop - Stop Rx
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaRxStop(END_DEVICE *pDev)
{
    ae5311ClearDmaReg(pDev,DmaControl,DmaRxStart);
}


/******************************************************************************
*
* ae5311DmaTxStart - Start Tx
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaTxStart(END_DEVICE *pDev)
{
    ae5311SetDmaReg(pDev,DmaControl,DmaTxStart);
}


/******************************************************************************
*
* ae5311DmaTxStop - Stop Tx
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaTxStop(END_DEVICE *pDev)
{
    ae5311ClearDmaReg(pDev,DmaControl,DmaTxStart);
}


/******************************************************************************
*
* ae5311DmaIntEnable - Enable DMA interrupts
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaIntEnable(END_DEVICE *pDev)
{
    ae5311WriteDmaReg(pDev, DmaInterrupt, DmaIntEnable);    /* enable interrupts */
}


/******************************************************************************
*
* ae5311DmaIntDisable - Disable DMA interrupts
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaIntDisable(END_DEVICE *pDev)
{
    ae5311WriteDmaReg(pDev, DmaInterrupt, DmaIntDisable);   /* disable interrupts */
}


/******************************************************************************
*
* ae5311DmaIntClear - Clear DMA interrupts
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaIntClear(END_DEVICE *pDev)
{
    ae5311WriteDmaReg(pDev, DmaStatus, ae5311ReadDmaReg(pDev, DmaStatus));  /* clear all interrupt requests */
}


/******************************************************************************
*
* ae5311DmaReset - Reset DMA and TLI controllers
*
* RETURNS: N/A
*/

LOCAL void ae5311DmaReset(END_DEVICE *pDev)
{
    int     i;
    UINT32  descAddr;


    END_TX_SEM_TAKE (&pDev->end, WAIT_FOREVER);

    /* Disable device interrupts prior to any errors during stop */
    intDisable(pDev->ilevel);

    /* Disable MAC tx and tx */
    ae5311ClearMacReg(pDev,MacControl,(MacRxEnable|MacTxEnable));

    /* Reset dma controller */
    ae5311WriteDmaReg(pDev,DmaBusMode,DmaResetOn);

    /* Delay 2 usec */
    sysUDelay( 2 );

    /* Flush the rx queue */
    descAddr = (UINT32)pDev->rxQueue.firstDescAddr;
    pDev->rxQueue.curDescAddr = pDev->rxQueue.firstDescAddr;
    for (i=0; i<(pDev->rxDescCount); i++, descAddr += AE_QUEUE_ELE_SIZE)
        AE_DESC_STATUS_SET (descAddr, DescOwnByDma);

    /* Flush the tx queue */
    descAddr = (UINT32)pDev->txQueue.firstDescAddr;
    pDev->txQueue.curDescAddr = pDev->txQueue.firstDescAddr;
    pDev->txQueue.reapDescAddr = pDev->txQueue.lastDescAddr;
    for (i=0; i<(pDev->txDescCount); i++, descAddr += AE_QUEUE_ELE_SIZE)
        AE_DESC_STATUS_SET (descAddr, 0);

    /* Set init register values  */
    ae5311WriteDmaReg(pDev, DmaBusMode, DmaBusModeInit);

    /* Install the first Tx and Rx queues on the device */
    ae5311WriteDmaReg(pDev, DmaRxBaseAddr,
                (UINT32)pDev->rxQueue.firstDescAddr);
    ae5311WriteDmaReg(pDev, DmaTxBaseAddr,
                (UINT32)pDev->txQueue.firstDescAddr);

#if TX_OSF == 2
    ae5311WriteDmaReg(pDev, DmaControl, DmaTxSecondFrame | DmaTxThreshCtl256);
#elif TX_OSF == 1
    ae5311WriteDmaReg(pDev, DmaControl, DmaTxThreshCtl256);
#else
    ae5311WriteDmaReg(pDev, DmaControl, DmaStoreAndForward);
#endif

    ae5311WriteDmaReg(pDev, DmaInterrupt, DmaIntDisable);


    DRV_LOG (DRV_DEBUG_MSG, "eth%d: DMA RESET!\n",
        pDev->unit,2,3,4,5,6);

    /* reset processing flag */
    pDev->aeProcessRcv = 0;

    /* Turn on device interrupts -- enable most errors */
    ae5311DmaIntClear(pDev);    /* clear interrupt requests  */
    ae5311DmaIntEnable(pDev);   /* enable interrupts */

    /* Enable Tx/Rx */
    ae5311SetMacReg(pDev,MacControl,(MacRxEnable|MacTxEnable));
    ae5311DmaRxStart(pDev);     /* start receiver  */
    ae5311DmaTxStart(pDev);     /* start transmitter, it must go to suspend immediately  */

    /* Set the running flag */
    END_FLAGS_SET (&pDev->end, IFF_UP | IFF_RUNNING);
    DRV_LOG (DRV_DEBUG_ENDFLAG, "ae%d end flag = 0x%8.8x %d\n",
        pDev->unit, pDev->end.flags, 106,4,5,6);

    pDev->aeProcessRst = 0;

    intEnable(pDev->ilevel);

    END_TX_SEM_GIVE (&pDev->end);

}


/*
 *  "handle" an unexpected ethernet interrupt.
 */
LOCAL void
ethStrayInt(int vec)
{
    asm_puts("ethStrayInt(");
    asm_puthex(vec);
    asm_puts(")\r\n");

    return;
}


/*******************************************************************************
* sysAe5311ParamGet - Get the device parameters of a device.
*
*
* RETURNS: OK on success and ERROR otherwise.
*/

LOCAL STATUS sysAe5311ParamGet(END_DEVICE *pDev)
{
    /* Initialize device parameters */
    if (pDev->unit == 0)
    {
        pDev->macBase = (UINT32) (PHYS_TO_K1(AR5311_ENET0)+AE_MAC_OFFSET);
        pDev->dmaBase = (UINT32) (PHYS_TO_K1(AR5311_ENET0)+AE_DMA_OFFSET);
        pDev->phyAddr = ETH0_PHY_ADDR;
        pDev->ivec   = (int) IV_ENET0_VEC;
        pDev->ilevel = (int) AR5311_INTR_ENET0;
    }
    else if (pDev->unit == 1)
    {
        pDev->macBase = (UINT32) (PHYS_TO_K1(AR5311_ENET1)+AE_MAC_OFFSET);
        pDev->dmaBase = (UINT32) (PHYS_TO_K1(AR5311_ENET1)+AE_DMA_OFFSET);
        pDev->phyAddr = ETH1_PHY_ADDR;
        pDev->ivec   = (int) IV_ENET1_VEC;
        pDev->ilevel = (int) AR5311_INTR_ENET1;
    }
    else
    {
        return ERROR; /* invalid AE5311 unit number */
    }

    return OK;
}


#if defined (PHY_POLLED)
/*****************************************************************************
*
* ae_PhyPollTask - poll the phys of any active ethernet port to determine
* link status changes.
* At this time, share the poll interval to also check for ethernet DMA
* rx fifo overflows.
*
* RETURNS: never
*/

LOCAL void ae_PhyPollTask(void)
{
    int             i;
    UINT16          phySts = 0;
    bool            wait[AE_MAX_ETH_UNITS] = {FALSE,FALSE,};
    unsigned int    oldLevel;

    while (1)
    {
        taskDelay(sysClkRateGet() * PHY_POLL_PERIOD_SECONDS);

        for ( i = 0; i < AE_MAX_ETH_UNITS; i++ )
        {
            /* Check for port present and running in system */
            if( ((END_DEVICE *)0 == pCtrlAe[i]) ||
                (pCtrlAe[i]->aeProcessRst != 0) )
            {
                continue;
            }

            /* In debug and instrumented builds, check rx ovrn */
            aeOvrnFlag[i] |= updateMissedFrameCounters(pCtrlAe[i], &aeRxMissedFrameFO[i], &aeRxMissedFrameND[i]);
            if (aeOvrnFlag[i] & 2)
            {
                /* lost some frames due to rx fifo overflow */
                /* Overrun message is throttled to polling freq */
                DRV_LOG (DRV_DEBUG_ORN, "eth%d: DmaRxMissFrameFO 0x%8.8x\n",
                    pCtrlAe[i]->unit, aeRxMissedFrameFO[i], 0,0,0,0);
            }

            if (aeOvrnFlag[i] & 1)
            {
                /* lost some frames due to rx no descriptor available */
                /* Overrun message is throttled to polling freq */
                DRV_LOG (DRV_DEBUG_ORN, "eth%d: DmaRxMissFrameND 0x%8.8x\n",
                    pCtrlAe[i]->unit, aeRxMissedFrameND[i], 0,0,0,0);
            }

            aeOvrnFlag[i] = 0;

            /* Restore throttled int mask */
            oldLevel = intLock();
            if (aeTimeout[i])
                ae5311SetDmaReg(pCtrlAe[i], DmaInterrupt, DmaIeRxNoBuffer);
            intUnlock (oldLevel);

            /* Check link status */
            phySts = (UINT16)ae5311MiiRead(pCtrlAe[i],GEN_sts);

            /* LinkStatus is latched low if link failed */
            if ( (FALSE == wait[i]) && (0 == (LINK & phySts)) )
            {
                DRV_PRINT(DRV_DEBUG_PHYSET,("\nae%d Link down\n",pCtrlAe[i]->unit));

                wait[i] = TRUE;

                /* stop device */
                if (0 == (pCtrlAe[i]->flags & AE5311_POLLING))
                    ae5311Stop (pCtrlAe[i]);

                continue;
            }

            /* Check for AN complete */
            if ((TRUE == wait[i]) && (AUTOCMPLT & phySts))
            {
                DRV_PRINT(DRV_DEBUG_PHYSET,("\nae%d AN complete; flags 0x%x --",pCtrlAe[i]->unit,(int)pCtrlAe[i]->flags));

                wait[i] = FALSE;

                /* Check for link */
                if (phySts & LINK)
                    pCtrlAe[i]->flags |= AE5311_LINK_UP;
                else
                    pCtrlAe[i]->flags &= ~AE5311_LINK_UP;

                /* update mac state */
                ae_SetMacFromPhy(pCtrlAe[i]);

                /* re-start device */
                if (0 == (pCtrlAe[i]->flags & AE5311_POLLING))
                    ae5311Start (pCtrlAe[i]);

                DRV_PRINT(DRV_DEBUG_PHYSET,(" 0x%x\n",(int)pCtrlAe[i]->flags));
            } /* end check if AN complete */

        } /* end for loop each unit */

        aeTimeout[0] = 0;
        aeTimeout[1] = 0;

    } /* end while(1) */

}   /* function never returns */
#endif


#if defined (PHY_POLLED) || defined (DRV_DEBUG)
/*****************************************************************************
*
* updateMissedFrameCounters -
*
* Update the counts of frames missed due to Rx fifo overrun and
* Rx descriptor unavailable.
*
* RETURNS: bitmap
*          0 if no change
*          1 if no desc counter changed
*          2 if fifo overrun counter changed
*/

LOCAL int updateMissedFrameCounters(
    END_DEVICE *pDev,
    UINT32 *pRxOvrnCtr,
    UINT32 *pNoDescCtr)
{
    UINT32          thisCountReg;
    UINT32          thisCount;
    unsigned int    oldLevel;
    int             retval = 0;


    oldLevel = intLock();

    /* this register clears on read */
    thisCountReg = ae5311ReadDmaReg(pDev, DmaOverflowCnt);

    thisCount = DmaMissedFrameCountMask & (thisCountReg >> DmaMissedFrameShift);
    if (thisCount > 0)
    {
        /* lost some frames due to rx fifo overflow */
        *pRxOvrnCtr += thisCount;

        retval |= 2;
    }

    thisCount = DmaMissedFrameCountMask & thisCountReg;
    if (thisCount > 0)
    {
        /* lost some frames due to rx no desc available */
        *pNoDescCtr += thisCount;

        retval |= 1;
    }

    intUnlock (oldLevel);

    return retval;
}
#endif

#ifdef DRV_DEBUG
/*****************************************************************************
*
* ethP - force ethernet port into promiscuous mode
*
* RETURNS:
*/

int ethP(int unit)
{
    ae5311Ioctl(pCtrlAe[unit], EIOCSFLAGS, (caddr_t)IFF_PROMISC);
    return unit;
}


/*****************************************************************************
*
* ethU - force ethernet port out of promiscuous mode
*
* RETURNS:
*/

int ethU(int unit)
{
    ae5311Ioctl(pCtrlAe[unit], EIOCSFLAGS, (caddr_t)(-IFF_PROMISC));
    return unit;
}


/*****************************************************************************
*
* ethDump - display info about the ethernet interface
*
* RETURNS:
*/

int ethDump(int unit)
{
    END_DEVICE      eDev;
    AE_DBG  aDbg;
    unsigned int    oldLevel;

    if ((unit < 0) || (unit >= AE_MAX_ETH_UNITS))
        unit = 0;

    oldLevel = intLock();
    bcopy((const char *)pCtrlAe[unit], (char *)&eDev, sizeof(eDev));
    #ifdef DRV_DEBUG
    bcopy((const char *)&aeDbg[unit], (char *)&aDbg, sizeof(aDbg));
    #endif
    intUnlock (oldLevel);

    printf("\n Dumping ethernet port %d\n", unit);

    printf("\n End device at 0x%8.8x, 0x%8.8x bytes ", (int)pCtrlAe[unit], sizeof(eDev));
    printf("\n flags 0x%8.8lx", pCtrlAe[unit]->flags);

    printf("\n   End object at 0x%8.8x", (int)&pCtrlAe[unit]->end);
    printf("\n   attached %d", pCtrlAe[unit]->end.attached);
    printf("\n   flags 0x%8.8lx", pCtrlAe[unit]->end.flags);
    printf("\n     Mib2Tbl");
    printf("\n     ifIndex %d", eDev.end.mib2Tbl.ifIndex);
    printf("\n     ifType  %ld", eDev.end.mib2Tbl.ifType);
    printf("\n     ifMtu   %ld", eDev.end.mib2Tbl.ifMtu);
    printf("\n     ifSpeed %ld", pCtrlAe[unit]->end.mib2Tbl.ifSpeed);
    printf("\n     ifAdminStatus   %ld", eDev.end.mib2Tbl.ifAdminStatus);
    printf("\n     ifOperStatus    %ld", eDev.end.mib2Tbl.ifOperStatus);
    printf("\n     ifLastChange    %ld", eDev.end.mib2Tbl.ifLastChange);
    printf("\n     ifInOctets      %10ld", eDev.end.mib2Tbl.ifInOctets);
    printf("\n     ifInUcastPkts   %10ld", eDev.end.mib2Tbl.ifInUcastPkts);
    printf("\n     ifInNUcastPkts  %10ld", eDev.end.mib2Tbl.ifInNUcastPkts);
    printf("\n     ifInDiscards    %10ld", eDev.end.mib2Tbl.ifInDiscards);
    printf("\n     ifInErrors      %10ld", eDev.end.mib2Tbl.ifInErrors);
    printf("\n     ifInUnkProtos   %10ld", eDev.end.mib2Tbl.ifInUnknownProtos);
    printf("\n     ifOutOctets     %10ld", eDev.end.mib2Tbl.ifOutOctets);
    printf("\n     ifOutUcastPkts  %10ld", eDev.end.mib2Tbl.ifOutUcastPkts);
    printf("\n     ifOutNUcastPkts %10ld", eDev.end.mib2Tbl.ifOutNUcastPkts);
    printf("\n     ifOutDiscards   %10ld", eDev.end.mib2Tbl.ifOutDiscards);
    printf("\n     ifOutErrors     %10ld", eDev.end.mib2Tbl.ifOutErrors);
    printf("\n     ifOutQLen       %10ld", eDev.end.mib2Tbl.ifOutQLen);

    printf("\n     dbgRxCnt        %10d", aDbg.dbgRxCnt);
    printf("\n     dbgRxGoodCnt    %10d", aDbg.dbgRxGoodCnt);
    printf("\n     dbgRxNoMb       %10d", aDbg.dbgRxNoMb);
    printf("\n     dbgRxErrPkt     %10d", aDbg.dbgRxErrPkt);
    printf("\n     dbgTxNot        %10d", aDbg.dbgTxNot);
    printf("\n     dbgTxLenReq     %10d", aDbg.dbgTxLenReq);
    printf("\n     dbgTxNoDesc     %10d", aDbg.dbgTxNoDesc);
    printf("\n     dbgTxNoMblk     %10d", aDbg.dbgTxNoMblk);
    printf("\n     dbgTxErrPkt     %10d", aDbg.dbgTxErrPkt);
    printf("\n     dbgTxSentPkt    %10d", aDbg.dbgTxSentPkt);
    printf("\n     dbgTxmNext      %10d", aDbg.dbgTxmNext);
    printf("\n     intAbnRxDesc    %10d", aDbg.intAbnRxDesc);
    printf("\n     intAbnRxStop    %10d", aDbg.intAbnRxStop);
    printf("\n     intAbnTx        %10d", aDbg.intAbnTx);
    printf("\n     intBusError     %10d", aDbg.intBusError);

    printf("\n     Rx Missed Frame FifoOverrun  %8d", aeRxMissedFrameFO[unit]);
    printf("\n     Rx Missed Frame NoDescriptor %8d", aeRxMissedFrameND[unit]);
#if DESC_BUG_RESET == 1
    printf("\n     Descriptor Bug Resets        %8d", aDbg.dmaBugReset);
#endif
    printf("\n");

    HISTOGRAM_DUMP("ethernet rx processing", aeRx, unit);

    return unit;
}


/*****************************************************************************
*
* ethClear - clear info about the ethernet interface
*
* RETURNS:
*/

int ethClear(int unit)
{
    unsigned int oldLevel;

    if ((unit < 0) || (unit >= AE_MAX_ETH_UNITS))
        unit = 0;
    printf("\n Clearing info on ethernet port %d\n", unit);

    oldLevel = intLock();
    pCtrlAe[unit]->end.mib2Tbl.ifInOctets = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifInUcastPkts = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifInNUcastPkts = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifInDiscards = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifInErrors = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifInUnknownProtos = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutOctets = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutUcastPkts = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutNUcastPkts = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutDiscards = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutErrors = 0;
    pCtrlAe[unit]->end.mib2Tbl.ifOutQLen = 0;
    #ifdef DRV_DEBUG
    aeDbg[unit].dbgRxCnt = 0;
    aeDbg[unit].dbgRxGoodCnt = 0;
    aeDbg[unit].dbgRxNoMb = 0;
    aeDbg[unit].dbgRxErrPkt = 0;
    aeDbg[unit].dbgTxNot = 0;
    aeDbg[unit].dbgTxLenReq = 0;
    aeDbg[unit].dbgTxNoDesc = 0;
    aeDbg[unit].dbgTxNoMblk = 0;
    aeDbg[unit].dbgTxErrPkt = 0;
    aeDbg[unit].dbgTxSentPkt = 0;
    aeDbg[unit].dbgTxmNext = 0;
    aeDbg[unit].intAbnRxDesc = 0;
    aeDbg[unit].intAbnRxStop = 0;
    aeDbg[unit].intAbnTx = 0;
    aeDbg[unit].intBusError = 0;
    aeRxMissedFrameFO[unit] = 0;
    aeRxMissedFrameND[unit] = 0;
#if DESC_BUG_RESET == 1
    aeDbg[unit].dmaBugReset = 0;
#endif
    #endif
    intUnlock (oldLevel);

    return unit;
}
#endif

#ifdef DRV_DEBUG
/*****************************************************************************
*
* descShow - display the contents of a descriptor ring
*
*            descShow <unit:0,1>,<q:TX=1,RX=2>
*
* RETURNS: N/A
*/

int descShow(int unit, int qt)
{
    UINT32 *pQue;
    UINT16 cnt;
    int i;

    printf ("\nDesc ring unit %d for ", unit);

    if (qt == 1)
    {
        printf("TX: start 0x%8.8x, end 0x%8.8x, curr 0x%8.8x\n",
        (UINT32)pCtrlAe[unit]->txQueue.firstDescAddr,
        (UINT32)pCtrlAe[unit]->txQueue.lastDescAddr,
        (UINT32)pCtrlAe[unit]->txQueue.curDescAddr);

        pQue = (UINT32 *)pCtrlAe[unit]->txQueue.firstDescAddr;
        cnt = pCtrlAe[unit]->txQueue.count;
    }
    else
    {
        printf("RX: start 0x%8.8x, end 0x%8.8x, curr 0x%8.8x\n",
        (UINT32)pCtrlAe[unit]->rxQueue.firstDescAddr,
        (UINT32)pCtrlAe[unit]->rxQueue.lastDescAddr,
        (UINT32)pCtrlAe[unit]->rxQueue.curDescAddr);

        pQue = (UINT32 *)pCtrlAe[unit]->rxQueue.firstDescAddr;
        cnt = pCtrlAe[unit]->rxQueue.count;
    }

    for (i=0; i < cnt; i++, pQue+=5)
    {
        printf("  0x%8.8x: 0x%8.8x, 0x%8.8x, 0x%8.8x, 0x%8.8x, 0x%8.8x\n",
            (UINT32)pQue, *pQue, *(pQue+1), *(pQue+2), *(pQue+3), *(pQue+4));
    }

    return cnt;
}


/*****************************************************************************
*
* dumpPkt - display the contents of an ethernet packet
*
* RETURNS: N/A
*/

void dumpPkt(char *pData, int len)
{
    int i;

    while (len)
    {
        printf("\n%8.8x: ", (UINT32)pData);

        for (i = 0; len && (i < 16); i++, len--, pData++)
            printf ("%02x ", (UINT8) *pData);
    }
}
#endif


#ifdef DRV_DEBUG
/*****************************************************************************
*
* pokePhy - reset the phy
*           pokePhy <unit:0,1>
*
* RETURNS: N/A
*/

void pokePhy(int unit)
{
    ae_PhySetup(pCtrlAe[unit]);
}


/*****************************************************************************
*
* pokeDma - soft reset the dma
*           pokeDma <unit:0,1>
*
* RETURNS: N/A
*/

void pokeDma(int unit)
{

    pCtrlAe[unit]->aeProcessRst = 1;
    END_FLAGS_CLR (&pCtrlAe[unit]->end, IFF_UP | IFF_RUNNING);

    ae5311DmaReset(pCtrlAe[unit]);
}
#endif


/*****************************************************************************
*
* getmacstate - get mac state
*           getmacstate <unit:0,1>
*
* RETURNS: N/A
*/

int getmacstate(int unit)
{
    printf("  aeStart      = %d\n", pCtrlAe[unit]->aeStart);
    printf("  aeProcessRcv = %d\n", pCtrlAe[unit]->aeProcessRcv);
    printf("\n");
    printf("  DmaStatus  = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaStatus));
    printf("  DmaBusMode = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaBusMode));
    printf("  DmaRxBase  = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaRxBaseAddr));
    printf("  DmaTxBase  = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaTxBaseAddr));
    printf("  DmaControl = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaControl));
    printf("  DmaIntr    = 0x%8.8x\n", ae5311ReadDmaReg(pCtrlAe[unit],DmaInterrupt));

    printf("  MacControl = 0x%8.8x\n", ae5311ReadMacReg(pCtrlAe[unit],MacControl));
    printf("  MacAddrHi  = 0x%8.8x\n", ae5311ReadMacReg(pCtrlAe[unit],MacAddrHigh));
    printf("  MacAddrLo  = 0x%8.8x\n", ae5311ReadMacReg(pCtrlAe[unit],MacAddrLow));
    printf("  MacVlan1   = 0x%8.8x\n", ae5311ReadMacReg(pCtrlAe[unit],MacVlan1));
    printf("  MacVlan2   = 0x%8.8x\n", ae5311ReadMacReg(pCtrlAe[unit],MacVlan2));
    printf("\n");
    printf("Rx Queue b=0x%8.8x, e=0x%8.8x, c=0x%8.8x, s@c=0x%8.8x\n",
        (UINT32)pCtrlAe[unit]->rxQueue.firstDescAddr,
        (UINT32)pCtrlAe[unit]->rxQueue.lastDescAddr,
        (UINT32)pCtrlAe[unit]->rxQueue.curDescAddr,
        *((UINT32 *)pCtrlAe[unit]->rxQueue.curDescAddr));
    printf("Current Rx buffer = 0x%8.8x\n",
        ae5311ReadDmaReg(pCtrlAe[unit],DmaRxCurrAddr));

    printf("Tx Queue b=0x%8.8x, e=0x%8.8x, c=0x%8.8x, s@c=0x%8.8x\n",
        (UINT32)pCtrlAe[unit]->txQueue.firstDescAddr,
        (UINT32)pCtrlAe[unit]->txQueue.lastDescAddr,
        (UINT32)pCtrlAe[unit]->txQueue.curDescAddr,
        *((UINT32 *)pCtrlAe[unit]->txQueue.curDescAddr));
    printf("Current Tx buffer = 0x%8.8x\n",
        ae5311ReadDmaReg(pCtrlAe[unit],DmaTxCurrAddr));

    return unit;
}

#ifdef DEBUG

#ifndef WDB_COMM_END_ENTRY
#define WDB_COMM_END_ENTRY 1
#endif

extern STATUS bridgePortRemove(char *, int);
extern STATUS bridgePortAdd(char *, int);
LOCAL bool debuggerEnabled = FALSE;
LOCAL int ethDbgPort = WDB_COMM_END_ENTRY;

void
ethDebuggerGetMacAddr()
{
    MAC_ADDR mac;

    ae5311Ioctl(pCtrlAe[ethDbgPort],EIOCGADDR,mac);
    printf("Debug port MAC address: %.2x-%.2x-%.2x-%.2x-%.2x-%.2x\n",
        mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
    printf("You may use the above MAC address to set a static ARP entry\n");
}

void
ethDebuggerEnable ()
{
    printf("Enabling the debugger on ae%d.\n", ethDbgPort);

    if(debuggerEnabled == TRUE) {
        printf("Debugger is already enabled on ae%d\n", ethDbgPort);
        return;
    }

    if (bridgePortRemove("ae",ethDbgPort) == OK) {
        printf("ae%d removed from the bridge. It is now a DEBUG only port\n",
        ethDbgPort);
    } else {
        printf("Failed to remove ae%d from the bridge. Cannot enable DEBUG\n",
        ethDbgPort);
        return;
    }

    /*
     * Make non promisc
     */
    ethU(ethDbgPort);

    ethDebuggerGetMacAddr();

    debuggerEnabled = TRUE;
    printf("Debugger enabled on ae%d\n",ethDbgPort);
}

void
ethDebuggerDisable ()
{
    printf("Disabling the debugger on ae%d\n", ethDbgPort);

    if (debuggerEnabled == FALSE) {
        printf("Debugger is already disabled. Returning without any action\n");
        return;
    }

    bridgePortAdd("ae", ethDbgPort);
    printf("ae%d is back in the bridge\n", ethDbgPort);

    /*
     * Set it to promisc back.
     */
    ethP(ethDbgPort);
    debuggerEnabled = FALSE;
}

#endif
