/*
 *  ae5311end.h
 *
 *  Header file for Atheros ethernet driver for AR5311 chipset.
 *
 *  Copyright  2001 Atheros Communications, Inc.,  All Rights Reserved.
 */

#ifndef _AE5311END_H_
#define _AE5311END_H_

#ident "ACI $Header: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/ae5311End.h#1 $"

#define A_ROUNDUP(x, y)   ((((x)+((y)-1))/(y))*(y))

END_OBJ *ae5311EndLoad(char *, void *);

/*
    Values specified here are:
    <mem loc> <mem size> <num tx desc> <num rx desc> <num clusters>
    <unit> is pre-pended by init code
    we need a bunch of extra clusters here for wlan powersave operation
*/
#ifdef PRJ_BUILD                /* AP and MDK */
#define AE_LOAD_STRING_BOOT	"0x0:0x0:0x0:0x0:0x0"
#else                           /* bootrom is small */
#define AE_LOAD_STRING_BOOT	"0x0:0x0:25:50:50"
#endif

/* Definitions for the END DEVICE flags field */
#define AE5311_PROMISCUOUS	0x0001
#define AE5311_POLLING		0x0002
#define	AE5311_BUF_LOCAL	0x0008
#define	AE5311_FULL_DUPLEX	0x0010
#define	AE5311_SPEED_100	0x0020
#define	AE5311_LINK_UP		0x0040
#define	AE5311_START		0x0200

/* Default phy status polling period */
#define	PHY_POLL_PERIOD_SECONDS	2	/* FOO make adjustable parameter? */

/* Queue types -- qtype */
#define AE_QUEUE_TYPE_TX	1	/* Transmit queue */
#define AE_QUEUE_TYPE_RX	2	/* Receive queue */

#define AE_DEV_NAME         "ae"
#define AE_DRV_NAME         "Atheros Spirit AE5311 Ethernet"
#define AE_MAX_ETH_UNITS    2   /* maximum number of AE5311 Ethernet cores present */ 

/* Default Driver Buffering Values - the extra default
 * should ideally be very close to the rx depth to
 * throttle the receive processing appropriately (see
 * comment in ar5hwcEnd.h)
 */
#define AE_RX_DESC_COUNT_DEFAULT	128     /* Receive descriptors */
#define	AE_TX_DESC_COUNT_DEFAULT	128     /* Transmit descriptors */
#define AE_CL_EXTRA_MIN				 16     /* Extra clusters - for loaning outside */
#define AE_CL_EXTRA_DEFAULT			AE_RX_DESC_COUNT_DEFAULT


/* Descriptor Layout */
#define AE_DESC_SIZE			0x10 /* 4 words, 16 bytes */
#define AE_QUEUE_ELE_SIZE		0x14 /* with drvr extension */

#define AE_DESC_STATUS			0x00 /* Status offset */
#define AE_DESC_CTRLEN			0x04 /* Control and Length offset */	
#define AE_DESC_BUFPTR			0x08 /* Buffer pointer offset */
#define AE_DESC_LNKBUF			0x0c /* Link field offset, or ptr to 2nd buf */

#define AE_DESC_MBLKPTR			0x10 /* MBLK pointer offset */

#define AE_DESC_STATUS_TX_ERRORS	(DescTxTimeout		| \
/* FOO quite a few DescTxLostCarrier	DescTxLostCarrier	| */ \
/* FOO quite a few DescTxNoCarrier	DescTxNoCarrier		| */ \
									DescTxLateCollision	| \
									DescTxExcCollisions	| \
									DescTxExcDeferral	| \
									DescTxUnderflow)

#define AE_DESC_STATUS_RX_ERRORS	(DescRxTruncated	| \
/* FOO lots of runt fragments */	DescRxRunt			| \
/* FOO vlan h/w doesnt seem to honor DescRxLongFrame		| */ \
									DescRxLateColl		| \
									DescRxMiiError		| \
									DescRxCrc)

/* fastpath enables building hdr on wlan side in place */
#define FASTPATH_HEADROOM   64

#define AE_ALIGN            4
#define VLAN_HDR_SIZE       4
#define ETH_CRC_LEN         4
#define ETH_MAX_PKT_SIZE    (ETHERMTU + ENET_HDR_REAL_SIZ + VLAN_HDR_SIZE + ETH_CRC_LEN)

/* align to long, then an extra for dma shift and fastpath headroom */
#define AE_BUF_SIZE         A_ROUNDUP((ETH_MAX_PKT_SIZE + FASTPATH_HEADROOM + 4),4)


#define CACHE_LINE_SIZE	_CACHE_ALIGN_SIZE

/* Interface to VxWorks BSP */

typedef struct ae5311_param
{
	UINT32	iobase;			/* Device base address */
	UINT32	macBase;		/* MAC subsection base address */
	UINT32	dmaBase;		/* DMA subsection base address */
	int		ivec;			/* Device interrupt vector */
	int		ilevel;			/* Device interrupt level */
} AE5311_PARAM;


#define RegRead(addr)	\
	(*(volatile unsigned int *)(addr))

#define RegWrite(val,addr)	\
	((*(volatile unsigned int *)(addr)) = (val))

#endif /* _AE5311END_H_ */
