/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved 
 *
 * apdefs.h Access Point specific declarations 
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/apdefs.h#2 $ 
 */

#ifndef _APDEFS_H_
#define _APDEFS_H_

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifdef VXWORKS
#include "vxWorks.h"
#include "msgQLib.h"
#include "semLib.h"
#include "muxLib.h"
#include "inetLib.h"
#endif

#include "linuxdrv.h"

#include "wlantype.h"
#include "wlanproto.h"
#include "wlandrv.h"
#include "wlanchannel.h"

#include "apvlan.h"

/*
 * Define task priorities for vxWorks: 
 *   - tNetTask takes most of the CPU. 
 *   - other tasks run at the same priority as tNetTask. 
 *   - round robin scheduling is used. 
 *   - tasks other than tNetTask should try to be polite. 
 *   - some radio tasks require higher priority. 
 *   - some vxworks tasks are ok at lower priority. 
 * 
 * The vxWorks shell is moved down from the default of 1 
 * to 50.  This makes some of it's debugging features less 
 * useful, but it allows slow debugging dump routines to 
 * not trip-up the tTrapMon task. 
 * 
 * Rate control in background mode will take what it can 
 * get, and will assert itself into the foreground and at 
 * high priority once a second. 
 * 
 * The calibration and radar throttle tasks require higher 
 * priority so they can control access to the wlan device. 
 * 
 * The dosfs buffer cache (tDcacheUpdate) and tffs polling 
 * (tTffsPTask) tasks can remain at low priority.  They 
 * are typically needed during configuration and update 
 * which has it's own mechanisms to ensure these operations 
 * complete. 
 */ 
#define TRAPMON_TASK_PRIORITY           1               /* highest! */
#define CALIBRATION_TASK_PRIORITY       10
#define RADAR_THROTTLE_TASK_PRIORITY    11
#define RC_TASK_FG_MODE_PRIORITY        12
#define CLI_TASK_PRIORITY               50 
#define VXWORKS_SHELL_PRIORITY          50              /* defaults to 1 */ 
#define TELNET_TASK_PRIORITY            50 
#define RLI_TASK_PRIORITY               50 
#define PING_TASK_PRIORITY              50 
#define DOT1X_TASK_PRIORITY             50 
#define CFGFILE_TASK_PRIORITY           50 
#define BRIDGE_AGER_PRIORITY            50 
#define AP_HOUSEKEEPING_TASK_PRIORITY   50 
#define RADAR_TASK_PRIORITY             50 
#define RC_TASK_BG_MODE_PRIORITY        255 
 
/* High level states for AP */ 
#define AP_RESET                0 
#define AP_INIT                 1 
#define AP_IDLE                 2 
 
/* Initialization states */ 
#define AP_INIT_NO_BSS          11 
#define AP_INIT_BSS             12 
 
#ifdef BUILD_AP 
 
/* Rate table */ 
#define RATE_6                  11 
#define RATE_9                  15 
#define RATE_12                 10 
#define RATE_18                 14 
#define RATE_24                 9 
#define RATE_36                 13 
#define RATE_48                 8 
#define RATE_54                 12 
 
#define RATE_TABLE      (RATE_6, RATE_9, RATE_12, RATE_18, \
                        RATE_24, RATE_36, RATE_48, RATE_54) 
 
/* 
 * The followings are the default factory settings 
 */ 
#ifdef  DEBUG 
#define DEFAULT_USERNAME                "" 
#define DEFAULT_PASSWORD                "" 
#else 
#define DEFAULT_USERNAME                "Admin" 
#define DEFAULT_PASSWORD                "5up" 
#endif /* DEBUG */ 
 
/* 
 * Change default to use high power channel instead of low power channel 
 */ 
#define DEFAULT_AIFS                    2 
#define DEFAULT_CWMIN                   0xf 
#define DEFAULT_DATA_RATE               "best" 
#define DEFAULT_SSID                    "aruba-ap"
#define DEFAULT_SSID_LEN                9
#define DEFAULT_SYSNAME                 "" 
#define DEFAULT_AUTH_TYPE               0               /* Open-system */ 
#define DEFAULT_AUTO_CHAN_SELECT_MODE   TRUE 
#define DEFAULT_BEACON_PERIOD           100 
#define DEFAULT_CALIBRATION_TIME        60 
#define DEFAULT_DTIM                    1 
#define DEFAULT_IP_ADDR                 "192.168.1.20" 
#define DEFAULT_SUBNET_MASK             "255.255.255.0" 
#define DEFAULT_FRAG_THRESHOLD          2346 
#define DEFAULT_RTS_THRESHOLD           2346 
#undef  DEFAULT_KEY_SIZE 
#define DEFAULT_QOS_MODE                FALSE 
#define DEFAULT_SNMP_COMMUNITY          "public" 
#define DEFAULT_TXPOWER                 0               /* full */ 
 
#define DEFAULT_SET_SNMP_COMMUNITY      "netman" 
#define DEFAULT_TURBO_MODE              FALSE 
#define DEFAULT_WDG_MODE                TRUE 
#define DEFAULT_WEP_MODE                FALSE 
#define DEFAULT_CIPHER_SELECTION        CIPHER_SELECTION_WEP 
#define DEFAULT_WEP_KEY                 0 
#define DEFAULT_RATECTRL_MODE           TRUE 
#define DEFAULT_SW_RETRY_MODE           TRUE 
#define DEFAULT_LIMIT_TXRETRIES         8 
#define DEFAULT_SWRETRY_MAXRETRIES      3 
#define DEFAULT_ACL_MODE                1 
#define DEFAULT_PS_QDEPTH               50 
#define DEFAULT_MAX_RECEIVE_LIFETIME            512 
#define DEFAULT_MAX_TRANSMIT_MSDU_LIFETIME      512 
 
#define DEFAULT_UPDATE_HOSTNAME         "" 
#define DEFAULT_UPDATE_USERNAME         "" 
#define DEFAULT_UPDATE_PASSWORD         "" 
#define DEFAULT_UPDATE_IMAGEPATH        "" 
#define DEFAULT_UPDATE_IMAGENAME        "" 
#define DEFAULT_UPDATE_LOCALFILE        "" 
#define DEFAULT_UPDATE_FLAG             0 
 
#define DEFAULT_SCRIPT_HOSTNAME         "" 
#define DEFAULT_SCRIPT_USERNAME         "" 
#define DEFAULT_SCRIPT_PASSWORD         "" 
#define DEFAULT_SCRIPT_IMAGEPATH        "" 
#define DEFAULT_SCRIPT_IMAGENAME        "" 
#define DEFAULT_SCRIPT_LOCALFILE        "_stemp_" 
#define DEFAULT_SCRIPT_FLAG             0 
#define DEFAULT_DOT1X_MODE              0 /* all keys from flash...802.1x disabled */ 
#define DEFAULT_RADIUSADDR              "" 
#define DEFAULT_RADIUSPORT              1812 
#define DEFAULT_RADIUSSECRET            "" 
#define DEFAULT_DNSADDR                 "" 
#define DEFAULT_DOMAINSUFFIX            "" 
 
#define MAX_SCRIPT_ALLOWED              (32768)                 /* in bytes */ 
#define DEFAULT_SCRIPT_RESULTFILE       "_rtemp_" 
#define MAX_FTP_RESULTSTRING            4096                    /* in bytes */ 
 
#define AP_IMG1_FILENAME                "apimg1" 
#define MAX_IMG1_FILENAME               (sizeof(AP_IMG1_FILENAME)+1) 
#define IMG1_FILENAME_EXIST             0x01 
#define AP_IMG1_HDRFILE                 "apimg1.hdr" 
#define MAX_IMG1_HDRFILE                (sizeof(AP_IMG1_HDRFILE)+1) 
#define IMG1_HDRFILE_EXIST              0x10 
#define AP_IMG2_FILENAME                "apimg2" 
#define MAX_IMG2_FILENAME               (sizeof(AP_IMG2_FILENAME)+1) 
#define IMG2_FILENAME_EXIST             0x02 
#define AP_IMG2_HDRFILE                 "apimg2.hdr" 
#define MAX_IMG2_HDRFILE                (sizeof(AP_IMG2_HDRFILE)+1) 
#define IMG2_HDRFILE_EXIST              0x20 
#define AP_IMG_TMPFILE                  "_tempi_" 
#define MAX_IMG_TMPFILE                 (sizeof(AP_IMG_TMPFILE)+1) 
#define IMG_TMPFILE_EXIST               0x800 
#define AP_CFG_FILESAVE                 "apcfg.sav" 
 
#define AP_FLASH_FS                     "/fl" 
#define AP_CFG_FILENAME                 "apcfg" 
#define MAX_CFG_FILENAME                (sizeof(AP_CFG_FILENAME)+1) 
 
#define CFG_CUR_REV                     2 
#define CFG_IEEE_REV                    2               /* Use IEEE Channel Number */ 
#define CFG_MAGIC_WORD                  "Ar52xxAP" 
 
#define CFG_OK                          OK 
#define CFG_ERROR                       ERROR 
 
#define CFG_INITIALIZED                 1 
 
#define CFG_MIN_KEY                     1 
#define CFG_MIN_SHARED_KEY              1 
#define CFG_MAX_SHARED_KEY              4 
#define CFG_MAX_KEY                     2048 
#define CFG_MIN_ACL                     4 
#define CFG_MAX_ACL                     4096 
#define INVALID_SHARED_KEY              (CFG_MAX_SHARED_KEY + 1 - CFG_MIN_KEY) 
 
#define CFG_MAX_SW_RETRY                DEFAULT_SW_RETRY 
 
#define CFG_MAX_SNMP_COMMUNITY          128 
#define CFG_MAX_USERNAME                32 
#define CFG_MAX_PASSWORD                32 
#define CFG_MAX_SSID                    32 
#define CFG_MAX_SYSNAME                 32 
#define CFG_MAX_KEYSTR                  32 
#define CFG_MAX_IPADDR                  20 
#define CFG_TIME_ZONE_LEN               8 
#define CFG_MAX_RADIUSNAME              64 
#define CFG_MAX_SECRETLEN               128 
#define CFG_MAX_DOMAINNMLEN             64 
 
#define TASK_STATUS_SUSPEND             "SUSPEND" 
#define CLI_TASK_NAME                   "tCliTask" 
#define CLI_TELNET_NAME                 "tCliTelnet" 
 
/* MACACL_ENTRY w/o the pNext + reference bit */ 
typedef 
struct cfg_macacl_entr_s { 
    WLAN_MACADDR    mac; 
    A_UINT8         flags; 
    A_UINT8         used; 
    A_UINT16        key; 
} CFG_MACACL_ENTRY; 
 
#define MAX_BRIDGE_ENTRIES 16 
typedef 
struct cfg_bridge_entr_s { 
    WLAN_MACADDR    bridgeApMac; 
    WLAN_MACADDR    hostMac; 
} CFG_BRIDGE_ENTRY; 
 
#define MAX_VLAN_ENTRIES    6 
typedef 
struct cfg_vlan_entr_s { 
    VLAN_PRIVACY    priv; 
    VLAN_TAG        tag; 
} CFG_VLAN_ENTRY; 
 
/* FTP Parameters */ 
#define CFG_MAX_FTP_HOSTNAME            33 
#define CFG_MAX_FTP_USERNAME            33 
#define CFG_MAX_FTP_PASSWORD            34 
#define CFG_MAX_FTP_IMAGEPATH           33 
#define CFG_MAX_FTP_IMAGENAME           33 
#define CFG_MAX_FTP_LOCALFILE           33 
 
typedef struct cfg_ftpparam_s { 
    A_INT8    hostname[CFG_MAX_FTP_HOSTNAME]; 
    A_INT8    username[CFG_MAX_FTP_USERNAME]; 
    A_INT8    password[CFG_MAX_FTP_PASSWORD]; 
    A_INT8    imagepath[CFG_MAX_FTP_IMAGEPATH]; 
    A_INT8    imagename[CFG_MAX_FTP_IMAGENAME]; 
    A_INT8    localfile[CFG_MAX_FTP_LOCALFILE]; 
    A_UINT16  flag; 
} CFG_FTPPARAM_ENTRY; 
 
#define MAX_FTP_STATUS_STRING   132 
 
struct ap_cfg_s { 
    A_UINT16                rev; 
    A_UINT16                length; 
    A_UINT8                 magicWord[sizeof(CFG_MAGIC_WORD)]; 
    A_UINT32                lastChecksum; 
    A_UINT32                checksum; 
    A_UINT32                calibrationTime; 
    A_UINT8                 authType; 
    A_UINT16                channel; 
    A_UINT8                 freqSpec; 
    A_UINT8                 rate[8]; 
    A_UINT8                 ssid[CFG_MAX_SSID + 1]; 
    A_UINT8                 sysName[CFG_MAX_SYSNAME + 1]; 
    A_UINT16                beaconInterval; 
    A_UINT16                dtim; 
    A_UINT16                fragThreshold; 
    A_UINT16                rtsThreshold; 
    SEM_ID                  cfgFileSemId; 
    BOOT_PARAMS             boot; 
    struct in_addr          ipAddr; 
    struct in_addr          ipMask; 
    A_UINT32                keyTableSize; 
    A_UINT32                keyLast; 
    WLAN_PRIV_RECORD        *keyTable; 
    A_UINT32                aclTableSize; 
    CFG_MACACL_ENTRY        *aclTable; 
    A_UINT8                 defaultWepKey; 
    A_UINT8                 defaultAesKey; 
    A_UINT8                 limitTxRetries;     /* # xmits before giving up */ 
    A_UINT8                 snmpCommunity[CFG_MAX_SNMP_COMMUNITY + 1]; 
    A_UINT8                 setSnmpCommunity[CFG_MAX_SNMP_COMMUNITY + 1]; 
    A_UINT8                 sntpServer[CFG_MAX_IPADDR + 1]; 
    A_UINT8                 username[CFG_MAX_USERNAME + 1]; 
    A_UINT8                 password[CFG_MAX_PASSWORD + 1]; 
    A_UINT8                 nameSrv[CFG_MAX_IPADDR + 1]; 
    A_UINT8                 nameDomain[CFG_MAX_DOMAINNMLEN + 1]; 
    A_UINT8                 radiusSrv[CFG_MAX_RADIUSNAME + 1]; 
    A_UINT8                 radiusSecret[CFG_MAX_SECRETLEN + 1]; 
    A_UINT16                radiusPort; 
    A_BOOL                  calibration; 
    A_BOOL                  RebootInProgress; 
    A_BOOL                  forceUpdate; 
    A_BOOL                  rateCtrlEnabled; 
    A_BOOL                  securityCheck; 
    A_BOOL                  snmpEnabled; 
    A_BOOL                  swretryEnabled; 
    A_BOOL                  turbo; 
    A_BOOL                  wdg; 
    A_BOOL                  wep; 
    A_UINT8                 cipher; 
    A_BOOL                  unencryptFrame; 
    A_BOOL                  aclMode; 
    A_BOOL                  noFS;               /* cannot save config */ 
    A_BOOL                  initialized; 
    A_BOOL                  cfgFileDone; 
    A_BOOL                  autoChanSelect; 
    A_BOOL                  qos; 
    A_BOOL                  bootParamUpdate; 
    A_BOOL                  firstTime; 
    A_BOOL                  telnetEnabled; 
    A_BOOL                  outdoorChanEnabled; 
    A_UINT8                 txPower; 
    A_UINT8                 cwmin; 
    A_UINT32                aifs; 
    A_UINT16                cfgFileUpdate; 
    A_UINT32                fwc; 
    A_UINT32                tzSecs; 
    A_UINT8                 timezone[CFG_TIME_ZONE_LEN]; 
    A_UINT16                countryCode; 
    CFG_BRIDGE_ENTRY        bridgeEntry[MAX_BRIDGE_ENTRIES]; 
    A_INT32                 numBridgeEntries; 
    A_UINT32                agingInterval; 
    A_UINT32                keySrc; 
    CFG_VLAN_ENTRY          vlanEntry[MAX_VLAN_ENTRIES]; 
    A_INT32                 numVlanEntries; 
    CFG_FTPPARAM_ENTRY      ftpDefParam;        /* Manufacturer defaults for fw update */ 
    CFG_FTPPARAM_ENTRY      ftpParam;           /* User settings for firmware update */ 
    CFG_FTPPARAM_ENTRY      ftpScript;          /* User settings for script file */ 
}; 
 
struct parse_token_s { 
    char    *pCmd; 
    char    *pHelpText; 
    FUNCPTR fHandler; 
    struct  parse_token_s *pNxtParseTbl; 
    char    *pProtected; 
    int     abbrevMatchCnt; 
}; 
 
/* define beacon related constants */ 
#define MAXBEACONINTERVAL 1000          /* 1000TU = 1002.4ms */ 
#define MINBEACONINTERVAL 20            /* 20TU = 20.48 ms */ 
#define BEACON_INTERVAL   100           /* 100TU = 102.4ms */ 
#define DTIM_PERIOD       1 
 
/* task related */ 
extern int      tNetTaskId;             /* ID of tNetTask for Q semaphores */ 
 
#else /* BUILD_AP */ /* #else section added by Fadi */ 
/* 
 * The followings are the default factory settings 
 */ 
#ifdef  DEBUG 
#define DEFAULT_USERNAME                "" 
#define DEFAULT_PASSWORD                "" 
#else 
#define DEFAULT_USERNAME                "Admin" 
#define DEFAULT_PASSWORD                "5up" 
#endif /* DEBUG */ 
 
/* 
 * Change default to use high power channel instead of low power channel 
 */ 
#define DEFAULT_AIFS                    2 
#define DEFAULT_CWMIN                   0xf 
#define DEFAULT_DATA_RATE               "best" 
#define DEFAULT_SSID                    "aruba-ap"
#define DEFAULT_SSID_LEN                9
#define DEFAULT_SYSNAME                 "" 
#define DEFAULT_AUTH_TYPE               0               /* Open-system */ 
#define DEFAULT_AUTO_CHAN_SELECT_MODE   TRUE 
#define DEFAULT_BEACON_PERIOD           100 
#define DEFAULT_CALIBRATION_TIME        60 
#define DEFAULT_DTIM                    1 
#define DEFAULT_IP_ADDR                 "192.168.1.20" 
#define DEFAULT_SUBNET_MASK             "255.255.255.0" 
#define DEFAULT_FRAG_THRESHOLD          2346 
#define DEFAULT_RTS_THRESHOLD           2346 
#undef  DEFAULT_KEY_SIZE 
#define DEFAULT_QOS_MODE                FALSE 
#define DEFAULT_SNMP_COMMUNITY          "public" 
#define DEFAULT_TXPOWER                 0               /* full */ 
 
#define DEFAULT_SET_SNMP_COMMUNITY      "netman" 
#define DEFAULT_TURBO_MODE              FALSE 
#define DEFAULT_WDG_MODE                TRUE 
#define DEFAULT_WEP_MODE                FALSE 
#define DEFAULT_CIPHER_SELECTION        CIPHER_SELECTION_WEP 
#define DEFAULT_WEP_KEY                 0 
#define DEFAULT_RATECTRL_MODE           TRUE 
#define DEFAULT_SW_RETRY_MODE           TRUE 
#define DEFAULT_LIMIT_TXRETRIES         8 
#define DEFAULT_SWRETRY_MAXRETRIES      3 
#define DEFAULT_ACL_MODE                1 
#define DEFAULT_PS_QDEPTH               50 
#define DEFAULT_MAX_RECEIVE_LIFETIME            512 
#define DEFAULT_MAX_TRANSMIT_MSDU_LIFETIME      512 
 
#define DEFAULT_UPDATE_HOSTNAME         "" 
#define DEFAULT_UPDATE_USERNAME         "" 
#define DEFAULT_UPDATE_PASSWORD         "" 
#define DEFAULT_UPDATE_IMAGEPATH        "" 
#define DEFAULT_UPDATE_IMAGENAME        "" 
#define DEFAULT_UPDATE_LOCALFILE        "" 
#define DEFAULT_UPDATE_FLAG             0 
 
#define DEFAULT_SCRIPT_HOSTNAME         "" 
#define DEFAULT_SCRIPT_USERNAME         "" 
#define DEFAULT_SCRIPT_PASSWORD         "" 
#define DEFAULT_SCRIPT_IMAGEPATH        "" 
#define DEFAULT_SCRIPT_IMAGENAME        "" 
#define DEFAULT_SCRIPT_LOCALFILE        "_stemp_" 
#define DEFAULT_SCRIPT_FLAG             0 
#define DEFAULT_DOT1X_MODE              0 /* all keys from flash...802.1x disabled */ 
#define DEFAULT_RADIUSADDR              "" 
#define DEFAULT_RADIUSPORT              1812 
#define DEFAULT_RADIUSSECRET            "" 
#define DEFAULT_DNSADDR                 "" 
#define DEFAULT_DOMAINSUFFIX            "" 
 
#define MAX_SCRIPT_ALLOWED              (32768)                 /* in bytes */ 
#define DEFAULT_SCRIPT_RESULTFILE       "_rtemp_" 
#define MAX_FTP_RESULTSTRING            4096                    /* in bytes */ 
 
#define AP_IMG1_FILENAME                "apimg1" 
#define MAX_IMG1_FILENAME               (sizeof(AP_IMG1_FILENAME)+1) 
#define IMG1_FILENAME_EXIST             0x01 
#define AP_IMG1_HDRFILE                 "apimg1.hdr" 
#define MAX_IMG1_HDRFILE                (sizeof(AP_IMG1_HDRFILE)+1) 
#define IMG1_HDRFILE_EXIST              0x10 
#define AP_IMG2_FILENAME                "apimg2" 
#define MAX_IMG2_FILENAME               (sizeof(AP_IMG2_FILENAME)+1) 
#define IMG2_FILENAME_EXIST             0x02 
#define AP_IMG2_HDRFILE                 "apimg2.hdr" 
#define MAX_IMG2_HDRFILE                (sizeof(AP_IMG2_HDRFILE)+1) 
#define IMG2_HDRFILE_EXIST              0x20 
#define AP_IMG_TMPFILE                  "_tempi_" 
#define MAX_IMG_TMPFILE                 (sizeof(AP_IMG_TMPFILE)+1) 
#define IMG_TMPFILE_EXIST               0x800 
#define AP_CFG_FILESAVE                 "apcfg.sav" 
 
#define AP_FLASH_FS                     "/fl" 
#define AP_CFG_FILENAME                 "apcfg" 
#define MAX_CFG_FILENAME                (sizeof(AP_CFG_FILENAME)+1) 
 
#define CFG_CUR_REV                     2 
#define CFG_IEEE_REV                    2               /* Use IEEE Channel Number */ 
#define CFG_MAGIC_WORD                  "Ar52xxAP" 
 
#define CFG_OK                          OK 
#define CFG_ERROR                       ERROR 
 
#define CFG_INITIALIZED                 1 
 
#define CFG_MIN_KEY                     1 
#define CFG_MIN_SHARED_KEY              1 
#define CFG_MAX_SHARED_KEY              4 
#define CFG_MAX_KEY                     2048 
#define CFG_MIN_ACL                     4 
#define CFG_MAX_ACL                     4096 
#define INVALID_SHARED_KEY              (CFG_MAX_SHARED_KEY + 1 - CFG_MIN_KEY) 
 
#define CFG_MAX_SW_RETRY                DEFAULT_SW_RETRY 
 
#define CFG_MAX_SNMP_COMMUNITY          128 
#define CFG_MAX_USERNAME                32 
#define CFG_MAX_PASSWORD                32 
#define CFG_MAX_SSID                    32 
#define CFG_MAX_SYSNAME                 32 
#define CFG_MAX_KEYSTR                  32 
#define CFG_MAX_IPADDR                  20 
#define CFG_TIME_ZONE_LEN               8 
#define CFG_MAX_RADIUSNAME              64 
#define CFG_MAX_SECRETLEN               128 
#define CFG_MAX_DOMAINNMLEN             64 
 
#endif /* BUILD_AP */ 
 
/* Items thought to not be worth tuning */ 
#define TXURN_COUNT_ALLOW  25 
#define TXURN_PERIOD_ALLOW 1000         /* ms */ 
#define RXEOL_COUNT_ALLOW  10 
#define RXEOL_PERIOD_ALLOW 5000         /* ms */ 
#define RESET_COUNT_ALLOW  256          /* absolute not timed */ 
 
//fadi #ifdef VXWORKS 
/* driver hook up pointers */ 
extern void *planCookie[]; 
extern void *pwlanCookie[]; 
extern WLAN_DEV_INFO *pwlanDevInfo; 
//fadi #endif 
 
/* AP init routines */ 
A_STATUS apInit(void); 
 
void smeInit(void); 
 
/* task main routines */ 
void apHouseKeepingTask(void); 
void apTrapMonTask(void); 
 
int strtoWlanMAC(char *macstr, WLAN_MACADDR *mac); 
 
struct ioctlKeyDesc_s { 
    WLAN_MACADDR *macAddr; 
    A_UINT8  *secret; 
    A_UINT32 usage; 
    A_UINT32 keyIndex; 
    A_UINT16 keyType; 
    A_UINT32 keyLength; // this version is in bits 
}; 
#define KEYDESC_USAGE_MULTICAST 0x1 
#define KEYDESC_USAGE_UNICAST   0x2 
 
 
/* Atheros specific IOCTLs */ 
#define EIOCSUHASH _IOW('e', 130, void *) /* Set unicast hash for the given MAC addr */ 
#define EIOCCUHASH _IOW('e', 131, void *) /* Clear unicast hash for the given MAC addr */ 
 
#define EIOCSKEY   _IOW('e', 132, void *) /* Set this key into s/w and h/w */ 
 
void apPanic(char *format, ...) __ATTRIB_PRINTF __ATTRIB_NORETURN; 
void apPanicDebug(void); 
 
#define STR_EQ(a, b) (strcmp((a), (b)) == 0) 
 
#ifdef __cplusplus 
} 
#endif /* __cplusplus */ 
 
#endif /* _APDEFS_H_ */ 
