/*
 *  Copyright  2000-2001 Atheros Communications, Inc.,  All Rights Reserved.
 *
 *  functions for memory management header file for vxworks
 *
 *  $Id$
 */


/**************************************************************************
 * memAllocateDescriptor - allocate a descriptor from the descriptor pool
 *
 * This routine will get a single ATHEROS descriptor from the descriptor pool
 *
 * RETURNS: N/A
 */
ATHEROS_DESC *
memAllocateDescriptor(WLAN_DEV_INFO *pdevInfo);

/**************************************************************************
* memFreeDescriptor - free a descriptor
*
* This routine will free a descriptor that was already created by
* memAllocateDescriptor()
*
* RETURNS: N/A
*/
void
memFreeDescriptor(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

/**************************************************************************
* hwAllocatePhysBuffer - allocate a physically contiguous buffer
*
* This routine will create a physically contiguous buffer and get
* the virtual and physical address of it.
*
* RETURNS: A_OK if got buffer, A_ERROR otherwise
*/
A_STATUS
hwAllocatePhysBuffer(WLAN_DEV_INFO *pdevInfo, A_UINT16 buffSize,
    A_UINT32 *pPhysAddress, void **ppVirtAddress, void **ppOSDescPtr);

/**************************************************************************
* hwFreePhysBuffer - free the physically contiguous buffer
*
* This routine will free the physically contiguous buffer previously
* allocated by hwAllocatePhysBuffer()
*
* RETURNS: N/A
*/
void
hwFreePhysBuffer(WLAN_DEV_INFO *pdevInfo, void *pVirtAddress, void *pOSDescPtr);

/**************************************************************************
* createHeaderBuffandDesc - create a physically contiguous buffer
*                           and descriptor for 802.11 header
*
* Create a physically contiguous data buffer and a descriptor
* to go with it.  An ATHEROS_DESC gets returned, pointer to data buffer
* can be assertained from the descriptor
*
* RETURNS: ATHEROS_DESC if successful, NULL otherwise
*/
ATHEROS_DESC *
createHeaderBuffandDesc(WLAN_DEV_INFO *pdevInfo, A_UINT16 buffSize);

/**************************************************************************
* memCopyToPSBuffers - since power save frames are going to be a while
*   its important to release their buffers back to the sending device
*   and copy the data over to buffers reserved for power save; the wlan
*   driver can reserve buffers for power save based on the amount of
*   memory available on the system
*
* RETURNS: A_OK if successful, A_ERROR otherwise
*/
ATHEROS_DESC *
memCopyToPSBuffers(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

/**************************************************************************
* createBuffandDesc - create a physically contiguous buffer and descriptor
*
* Create a physically contiguous data buffer and create a descriptor
* to go with it.  An ATHEROS_DESC gets returned, pointer to data buffer
* can be assertained from the descriptor
*
* RETURNS: A_OK if successful, A_ERROR otherwise
*/
A_STATUS
createBuffandDesc(WLAN_DEV_INFO *pdevInfo, A_UINT16 buffSize, ATHEROS_DESC **ppDesc);

/**************************************************************************
* freeBuffandDesc - free buffer and descriptor
*
* Free the buffer and descriptor, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void
freeBuffandDesc(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

/**************************************************************************
* freeBuffandDescChain - free buffer and descriptor chain
*
* Free the buffer and descriptor chain, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void
freeBuffandDescChain(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

/**************************************************************************
* descInfoDup - duplicate os dependent fields in descriptor chain
*
* read os dependent fields in the original descriptor and copy it over to
* to be duplicated descriptor
*
* RETURNS: N/A
*/
void
descInfoDup(ATHEROS_DESC *pOrigDesc, ATHEROS_DESC *pDupDesc);

/**************************************************************************
* bufferRefUpdate - update the reference info for the buffer
*
* for vxWorks this is increase the the reference counter for the buffer
*
* for the NDIS this could be set flag
*
* RETURNS: N/A
*/
void
bufferRefUpdate(ATHEROS_DESC *pDesc);

/**************************************************************************
* addDescToQueueTail - add the descriptor to the tail of queue struct
*
* Adds the given descriptor the tail of the queue structure
*
* RETURNS: N/A
*/
void
addDescToQueueTail(ATHEROS_DESC *pDesc, ATHEROS_DESC *pTail, QUEUE_DETAILS *pQueue);

/**************************************************************************
* getPktFromQueueStart - get a packet from the start of queue
*
* Removes the packet's descriptor chain from the head of the queue.
* Sets the new head of the queue to the next packet's descriptor chain
*
* RETURNS: Pointer to the descriptor chain of the packet if successful
*          NULL if failed
*/
ATHEROS_DESC *
getPktFromQueueStart(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue);

/**************************************************************************
* agingQueue - aging out the buffered traffic from PS queue
*
* RETURNS: N/A
*/
void
ageQueue(WLAN_DEV_INFO *pdevInfo, QUEUE_DETAILS *pQueue, A_UINT32 agingInterval,
    A_UINT32 now);


/**************************************************************************
* Resource usage tracking macros
*/

#define LOG_DROP_FRAME(_p, _r)          (++((_p)->_r##FramesDropped))
