/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Chips specific device attachment and device info collection
 *  Connects Init Reg Vectors, EEPROM Data, and device Functions to pDev
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211Attach.c#5 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "halUtil.h"

                       
/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Attach.h"
#include "ar5211.h"

/* Header files exporting vector mapped functions */
#include "ar5211Reset.h"
#include "ar5211Misc.h"
#include "ar5211KeyCache.h"
#include "ar5211Power.h"
#include "ar5211Beacon.h"
#include "ar5211Transmit.h"
#include "ar5211Receive.h"
#include "ar5211Interrupts.h"

#if defined(AR5311)
#include "ar5311reg.h"
#endif

/* Forward Declarations */
static A_STATUS
ar5211Detach(WLAN_DEV_INFO *pDev);

A_STATUS
ar5211VerifyEepromChecksum(WLAN_DEV_INFO *pDev);

static A_STATUS
ar5211AllocateProm(EEP_MAP *pEepMap);

static A_STATUS
ar5211ReadEepromIntoDataset(WLAN_DEV_INFO *pDev);

static A_STATUS
ar5211ReadHeaderInfo(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo);

static void
getPcdacInterceptsFromPcdacMinMax(A_UINT16 version, A_UINT16 pcdacMin,
    A_UINT16 pcdacMax, A_UINT16 *pPcdacValues);

/* Constants */
static const HW_FUNCS ar5211Funcs = {
    ar5211Detach,
    /* Reset Functions */
    ar5211Reset,
    ar5211PhyDisable,
    ar5211Disable,
    ar5211ResetDma,
    ar5211PerCalibration,

    /* TX Functions */
    ar5211GetTxDP,
    ar5211SetTxDP,
    ar5211UpdateTxTrigLevel,
    ar5211StartDmaTransmit,
    ar5211IsTxQueueStopped,
    ar5211ClearTransmitFilter,
    ar5211GetTransmitFilterIndex,
    ar5211GetShortRetryLimit,
    ar5211GetLongRetryLimit,
    ar5211SetTxDescHdrLen,
    ar5211SetTxDescLongPkt,
    ar5211SetTxDescAntModeXmit,
    ar5211SetTxDescPktType,
    ar5211GetTxDescEncryptKeyIndex,
    ar5211SetTxDescEncryptKeyIndex,
    ar5211SetTxDescRtsDuration,
    ar5211SetTxDescNoAck,
    ar5211SetTxDescVeol,
    ar5211DebugPrintTxDesc,

    /* RX Functions */
    ar5211GetRxDP,
    ar5211SetRxDP,
    ar5211EnableReceive,
    ar5211StopDmaReceive,
    ar5211StartPcuReceive,
    ar5211StopPcuReceive,
    ar5211SetMulticastFilter,
    ar5211MulticastFilterIndex,
    ar5211RxFilter,
    ar5211GetRxDescRxAntenna,
    ar5211GetRxDescRxTimeStamp,
    ar5211GetRxDescKeyIndex,
    ar5211GetRxDescKeyCacheMiss,
    ar5211EnableRxZlf,

    /* Misc Functions */
    ar5211SetRegulatoryDomain,
    ar5211SetLedState,
    ar5211WriteAssocid,
    ar5211GpioGet,
    ar5211Gpio0SetIntr,
    ar5211SetStaBeaconTimers,
    ar5211GetTsf,
    ar5211ResetTsf,
    ar5211SetAdhocMode,
    ar5211GetRegDmn,
    ar5211GetRfKill,
    ar5211GetTurboDisable,
    ar5211GetChannelEdges,
    ar5211GetMaxTurboRate,
    ar5211GetRandomSeed,
    ar5211DetectCardPresent,
    ar5211UpdateMibCounters,
    ar5211IsHwAesSupported,
    ar5211IsSwAesSupported,
    ar5211EnableHwEncryption,
    ar5211DisableHwEncryption,
    ar5211EnableRadarDetection,
    ar5211GetCurRssi,
    ar5211GetDefAntenna,
    ar5211SetDefAntenna,


    /* Key Cache Functions */
    NULL, //ar5211GetKeyCacheSize,
    NULL, //ar5211ResetKeyCacheEntry,
    NULL, //ar5211SetKeyCacheEntry,

    /* Power Management Functions */
    ar5211SetPowerModeAuto,
    ar5211SetPowerModeAwake,
    ar5211SetPowerModeSleep,
    ar5211QueryPSPollSupport,
    ar5211InitPSPoll,
    ar5211EnablePSPoll,
    ar5211DisablePSPoll,

    /* Beacon Functions */
    ar5211WaitForBeaconDone,
    ar5211UpdateBeaconDesc,
    ar5211QueueBeaconAndCab,
    ar5211SetupBeaconDesc,
    ar5211BeaconInit,

    /* Interrupt Functions */
    ar5211IsInterruptPending,
    ar5211GetInterrupts,
    ar5211EnableInterrupts,
    ar5211DisableInterrupts
};

/*
 * TODO: Need to talk to Praveen about this, these are
 * not valid 2.4 channels, either we change these
 * or I need to change the beanie coding to accept these
 */
static const A_UINT16 channels2_4[] = {2412, 2447, 2484};

/**************************************************************
 * ar5211Attach
 *
 * ar5211Attach performs the device specific functions for halAttach()
 */
A_STATUS
ar5211Attach(WLAN_DEV_INFO *pDev, A_UINT16 deviceID)
{
    A_UINT16 i;
    EEP_MAP  *pEMap;
    A_UINT32 revId, promSize;
    A_STATUS status;
    A_UINT16 eepVersion;
    A_UINT32 macRev;
    A_UINT32 addr, loop, wrData, rdData=0, pattern;
    A_UINT32 regAddr[2] = {MAC_STA_ID0, PHY_BASE+(8 << 2)};
    A_UINT32 regHold[2];
    A_UINT32 patternData[4] = {0x55555555, 0xaaaaaaaa, 0x66666666, 0x99999999};

    aPrintk("Entering ar5211Attach.......\n");
    
    ASSERT(pDev);

    /* Pull Chip out of Reset */
    status = ar5211ChipReset(pDev, 0);
    if (status != A_OK) {
        aPrintk("ATTACH: In Attach...chip reset failed!\n");
        return status;
    }
    if (pDev->pciInfo.DeviceID == DEV_AR5211_FPGA11B) {
        // set it back to OFDM mode to be able to read analog rev id.
        writePlatformReg(pDev, PHY_5211_MODE, PHY_5211_MODE_OFDM);
        writePlatformReg(pDev, PHY_PLL_CTL, PHY_PLL_CTL_44);
        udelay(1000);
    }

    /* Read Revisions from Chips */
#if defined(PCI_INTERFACE)
    macRev           = readPlatformReg(pDev, MAC_SREV) & MAC_SREV_ID_M;
    pDev->macVersion = (A_UINT16) (macRev >> MAC_SREV_ID_S);
    pDev->macRev     = (A_UINT16) (macRev & MAC_SREV_REVISION_M);
#elif defined(AR5311)
    macRev           = sysRegRead(AR5311_REV);
    pDev->macVersion = (macRev & REV_WMAC_MAJ) >> REV_WMAC_MAJ_S;
    pDev->macRev     = (macRev & REV_WMAC_MIN) >> REV_WMAC_MIN_S;
#endif

#ifdef DEBUG
      aPrintk("macRev read=0x%x[pDev->macVer=0x%x][pDev->macRev=0x%x]\n", 
             (unsigned int)macRev, pDev->macVersion, pDev->macRev);
#endif

    if ((pDev->macVersion < MAC_SREV_VERSION_MAUI_2) ||
        (pDev->macVersion > MAC_SREV_VERSION_OAHU))
    {
        aPrintk("ar5211Attach: Mac Chip Rev 0x%02X is not supported by this driver\n",
                 pDev->macVersion);
        return A_HARDWARE;
    }

    pDev->phyRev = (A_UINT16)readPlatformReg(pDev, PHY_CHIP_ID);

#ifdef DEBUG
      aPrintk("phyRev read=0x%x\n", pDev->phyRev);
#endif


    // Test PHY & MAC registers
    for (i = 0; i < 2; i++) {
        addr       = regAddr[i];
        regHold[i] = readPlatformReg(pDev, addr);

        for (loop = 0; loop < 0x100; loop++) {

            wrData = (loop << 16) | loop;
            writePlatformReg(pDev, addr, wrData);
            rdData = readPlatformReg(pDev, addr);
            if (rdData != wrData) {
                aPrintk("ar5211Attach: address test failed addr: 0x%08lx - \
                       wr:0x%08lx != rd:0x%08lx\n", (unsigned long)addr, 
                       (unsigned long)wrData, (unsigned long)rdData);
                return A_HARDWARE;
            }
        }
        for (pattern = 0; pattern < 4; pattern++) {
            wrData = patternData[pattern];
            writePlatformReg(pDev, addr, wrData);
            rdData = readPlatformReg(pDev, addr);
            if (wrData != rdData) {
                aPrintk("ar5211Attach: address test failed addr: 0x%08lx - \
                       wr:0x%08lx != rd:0x%08lx\n",(unsigned long)addr, 
                       (unsigned long)wrData, (unsigned long)rdData);
                return A_HARDWARE;
            }
        }
        writePlatformReg(pDev, regAddr[i], regHold[i]);
    }
    udelay(100);

    /* Set correct Baseband to analog shift setting to access analog chips. */
    if (pDev->macVersion >= MAC_SREV_VERSION_OAHU) {
        writePlatformReg(pDev, PHY_BASE, 0x00000007);
    } else {
        writePlatformReg(pDev, PHY_BASE, 0x00000047);
    }
    udelay(2000);

    /* Read Radio Chip Rev Extract */
    writePlatformReg(pDev, (PHY_BASE + (0x34 << 2)), 0x00001c16);
    for (i = 0; i < 8; i++) {
        writePlatformReg(pDev, (PHY_BASE + (0x20 << 2)), 0x00010000);
    }
    revId = (readPlatformReg(pDev, PHY_BASE + (256 << 2)) >> 24) & 0xff;
    revId = ((revId & 0xf0) >> 4) | ((revId & 0x0f) << 4);
    pDev->analog5GhzRev = (A_UINT16)reverseBits(revId, 8);
    if ((pDev->analog5GhzRev & 0xF0) != RAD5_SREV_MAJOR) {
        aPrintk("ar5211Attach: 5G Radio Chip Rev 0x%02X is not supported by this driver\n",
            pDev->analog5GhzRev);
        return A_HARDWARE;
    }

    /* EEPROM version and size checking */
    status = ar5211EepromRead(pDev, ATHEROS_EEPROM_OFFSET + 1, &eepVersion);
    if (status != A_OK) {
        goto ee_error;
    }

#if defined(PCI_INTERFACE)
    promSize = (readPlatformReg(pDev, MAC_PCICFG) & MAC_PCICFG_EEPROM_SIZE_M) >>
               MAC_PCICFG_EEPROM_SIZE_S;
#else
    promSize = MAC_PCICFG_EEPROM_SIZE_16K;
#endif

    // Allocate EEPROM Map
    pEMap = (struct eepMap *)kmalloc(sizeof(struct eepMap),GFP_KERNEL);
    if (!pEMap) {
        aPrintk("ar5211Attach: Could not allocate memory for version 3 EEPROM\n");
        return A_NO_MEMORY;
    }
    pDev->pHalInfo->pEepData = pEMap;

    if ( (eepVersion >= EEPROM_VER3) && (promSize == MAC_PCICFG_EEPROM_SIZE_16K) ) {
        status = ar5211VerifyEepromChecksum(pDev);
        if (status != A_OK) {
            goto ee_error;
        }
        status = ar5211AllocateProm(pEMap);
        if (status != A_OK) {
            goto ee_error;
        }

        /* Set Version So EEPROM Header can be correctly interpreted */
        pEMap->version = eepVersion;

        status = ar5211ReadEepromIntoDataset(pDev);
        if (status != A_OK) {
            goto ee_error;
        }

        status = ar5211EepromRead(pDev, EEPROM_PROTECT_OFFSET, &pEMap->protect);
        if (status != A_OK) {
            goto ee_error;
        }

        /* If Bmode and AR5211, verify 2.4 analog exists */
        if ((pDev->macVersion >= MAC_SREV_VERSION_OAHU) &&
            pEMap->pEepHeader->Bmode)
        {
            /* Set correct Baseband to analog shift setting to access analog chips. */
            writePlatformReg(pDev, PHY_BASE, 0x00004007);
            udelay(2000);
            writePlatformReg(pDev, (PHY_BASE + (0x34 << 2)), 0x00001c16);
            for (i = 0; i < 8; i++) {   
                writePlatformReg(pDev, (PHY_BASE + (0x20 << 2)), 0x00010000);
            }
            revId = (readPlatformReg(pDev, PHY_BASE + (256 << 2)) >> 24) & 0xff;   
            revId = ((revId & 0xf0) >> 4) | ((revId & 0x0f) << 4);
            pDev->analog2GhzRev = (A_UINT16)reverseBits(revId, 8);

            /* Set baseband for 5GHz chip */
            writePlatformReg(pDev, PHY_BASE, 0x00000007);
            udelay(2000);
            if ((pDev->analog2GhzRev & 0xF0) != RAD2_SREV_MAJOR) {
                aPrintk("ar5211Attach: 2G Radio Chip Rev 0x%02X is not supported by this driver\n",
                         pDev->analog2GhzRev);
                goto ee_error;
            }
        } else {
            pEMap->pEepHeader->Bmode = 0;
        }

        status = ar5211EepromRead(pDev, REGULATORY_DOMAIN_OFFSET, &pEMap->regDmnEnum);
        if (status != A_OK) {
            goto ee_error;
        }
        pEMap->rfKill  = pEMap->pEepHeader->rfKill;

        /* TODO: Remove this hack to fix bad reg domains */
        if (pEMap->regDmnEnum == 0x0001) {
            aPrintk("*- Fixing this device's Reg Domain of 1 -*\n");
            ar5211SetRegulatoryDomain(pDev, 0x10);
        }
    } else {
        aPrintk("ar5211Attach: A Rev 3 16k calibrated EEPROM is required for correct operation\n");
        goto ee_error;
    }

    /*
     * It's okay for the address to be assigned by software, so we ignore the
     * following error value if returned.
     */
    (void) ar5211GetMacAddr(pDev, &pDev->staConfig.macPermAddr);

    if (deviceID == DEV_AR5211_FPGA11B) {
        /* Map data Q's to QCU/DCU values */
        pDev->dataQueueNum   = 0;
        pDev->psPollQueueNum = 1;
        pDev->cabQueueNum    = 1;
        pDev->beaconQueueNum = 1;
    } else {
        /* Map data Q's to QCU/DCU values */
        pDev->dataQueueNum   = 0;
        pDev->psPollQueueNum = 1;
        pDev->cabQueueNum    = 8;
        pDev->beaconQueueNum = 9;
    }

    /* Initialize active Q mask with only data enabled */
    /* TODO: Move this to a data queue enable routine?? */
    pDev->activeQueueMask = 1 << pDev->dataQueueNum;

#ifdef DEBUG
    aPrintk("wlan revisions: mac %d.%d phy %d.%d analog %d.%d\n",
             pDev->macVersion, pDev->macRev,
             pDev->phyRev >> 4, pDev->phyRev & 0xf,
             pDev->analog5GhzRev >> 4, pDev->analog5GhzRev & 0xf);
#endif

    /* Attach device specific functions to the hal */
    pDev->pHwFunc = &ar5211Funcs;

	aPrintk("....Done with attach\n");

    return A_OK;

ee_error:
    aPrintk("ar5211Attach: EEPROM-related failure\n");
    ar5211Detach(pDev);
    return A_HARDWARE;
}

/**************************************************************
 * ar5211Detach
 *
 * Detach removes all hwLayer allocated data structures and places the
 * device into reset.
 */
A_STATUS
ar5211Detach(WLAN_DEV_INFO *pDev)
{
    HAL_INFO       *pInfo;
    struct eepMap  *pData;
    
    ASSERT(pDev && pDev->pHalInfo);

    pInfo = pDev->pHalInfo;
    pData = pInfo->pEepData;
    
    if (pData) {
        ASSERT(pData->version >= EEPROM_VER3);

        if (pData->pEepHeader) {
            A_DRIVER_FREE(pData->pEepHeader, sizeof(EEP_HEADER_INFO), 0);
        }

        if (pData->pPcdacInfo) {
            A_DRIVER_FREE(pData->pPcdacInfo, sizeof(PCDACS_ALL_MODES), 0);
        }

        if (pData->pTrgtPowerInfo) {
            A_DRIVER_FREE(pData->pTrgtPowerInfo, sizeof(TRGT_POWER_ALL_MODES), 0);
        }

        if (pData->pRdEdgesPower) {
            A_DRIVER_FREE(pData->pRdEdgesPower, 
                          sizeof(RD_EDGES_POWER) * NUM_EDGES * NUM_CTLS, 0);
        }

        A_DRIVER_FREE(pData, sizeof(struct eepMap), 0);
        pInfo->pEepData = NULL;
    }

    return A_OK;
}

/************** EEPROM REV 3 ATTACH FUNCTIONS *****************/

/**************************************************************
 * ar5211VerifyEepromChecksum
 *
 * Allocates memory for the EEPROM structures
 */
A_STATUS
ar5211VerifyEepromChecksum(WLAN_DEV_INFO *pDev)
{
    A_STATUS    status = A_OK;
    A_UINT16    data;
    A_UINT32    addr;
    A_UINT16    chkSum = 0;

    /* EEPROM version and size checking */

    for (addr = ATHEROS_EEPROM_OFFSET; addr < 0x400; addr++) {
        status = ar5211EepromRead(pDev, addr, &data);
        if (status != A_OK) {
            goto ee_error;
        }
        chkSum ^= (data & 0xffff);
    }

    chkSum &= 0xffff;
    if (chkSum != 0xffff) {
        aPrintk("ar5211VerifyEepromChecksum: Checksum %04x != 0xffff\n", chkSum);
        return A_HARDWARE;
    }

    return status;

ee_error:
    aPrintk("ar5211Attach: EEPROM read failed\n");
    return A_HARDWARE;
}

/**************************************************************
 * ar5211AllocateProm
 *
 * Allocates memory for the EEPROM structures
 */
A_STATUS
ar5211AllocateProm(EEP_MAP *pEepMap)
{
    /* TODO: only allocate 2.4 info if configured */
    int i;

    /* allocate the struct to hold the header info */
    pEepMap->pEepHeader = (EEP_HEADER_INFO *)kmalloc(sizeof(EEP_HEADER_INFO), GFP_KERNEL);
    if (!pEepMap->pEepHeader) {
        aPrintk("Unable to allocate eeprom structure for header info\n");
        return A_NO_MEMORY;
    }
    A_MEM_ZERO(pEepMap->pEepHeader, sizeof(EEP_HEADER_INFO));

    /* allocate the struct to hold the pcdac/power info */
    pEepMap->pPcdacInfo = (PCDACS_ALL_MODES *)kmalloc(sizeof(PCDACS_ALL_MODES), GFP_KERNEL);
    if (!pEepMap->pPcdacInfo) {
        aPrintk("Unable to allocate eeprom structure for pcdac/power info\n");
        A_DRIVER_FREE(pEepMap->pEepHeader, sizeof(EEP_HEADER_INFO), 0);
        return A_NO_MEMORY;
    }
    A_MEM_ZERO(pEepMap->pPcdacInfo, sizeof(PCDACS_ALL_MODES));

    pEepMap->pPcdacInfo->numChannels11a = NUM_11A_EEPROM_CHANNELS;
    pEepMap->pPcdacInfo->numChannels2_4 = NUM_2_4_EEPROM_CHANNELS;

    for (i = 0; i < NUM_11A_EEPROM_CHANNELS; i ++) {
        pEepMap->pPcdacInfo->DataPerChannel11a[i].numPcdacValues = NUM_PCDAC_VALUES;
    }

    /* the channel list for 2.4 is fixed, fill this in here */
    for (i = 0; i < NUM_2_4_EEPROM_CHANNELS; i++) {
        pEepMap->pPcdacInfo->Channels2_4[i] = channels2_4[i];
        pEepMap->pPcdacInfo->DataPerChannel11b[i].numPcdacValues = NUM_PCDAC_VALUES;
        pEepMap->pPcdacInfo->DataPerChannel11g[i].numPcdacValues = NUM_PCDAC_VALUES;
    }

    /* allocate the structure to hold target power info */
    pEepMap->pTrgtPowerInfo = (TRGT_POWER_ALL_MODES *)kmalloc(sizeof(TRGT_POWER_ALL_MODES), GFP_KERNEL);
    if (!pEepMap->pTrgtPowerInfo) {
        aPrintk("Unable to allocate eeprom structure for target power info\n");
        A_DRIVER_FREE(pEepMap->pEepHeader, sizeof(EEP_HEADER_INFO), 0);
        A_DRIVER_FREE(pEepMap->pPcdacInfo, sizeof(PCDACS_ALL_MODES), 0);
        return A_NO_MEMORY;
    }
    A_MEM_ZERO(pEepMap->pTrgtPowerInfo, sizeof(TRGT_POWER_ALL_MODES));

    /* allocate structure for RD edges */
    pEepMap->pRdEdgesPower = (RD_EDGES_POWER *)kmalloc((sizeof(RD_EDGES_POWER) * NUM_EDGES * NUM_CTLS), GFP_KERNEL);
    if (!pEepMap->pRdEdgesPower) {
        aPrintk("Unable to allocate eeprom structure for RD edges info\n");
        A_DRIVER_FREE(pEepMap->pEepHeader, sizeof(EEP_HEADER_INFO), 0);
        A_DRIVER_FREE(pEepMap->pPcdacInfo, sizeof(PCDACS_ALL_MODES), 0);
        A_DRIVER_FREE(pEepMap->pTrgtPowerInfo, sizeof(TRGT_POWER_ALL_MODES), 0);
        return A_NO_MEMORY;
    }
    A_MEM_ZERO(pEepMap->pRdEdgesPower, sizeof(RD_EDGES_POWER) * NUM_EDGES * NUM_CTLS);
    return A_OK;
}

/**************************************************************
 * ar5211ReadEepromIntoDataset
 *
 * Now verify and copy EEPROM contents into the allocated space
 */
A_STATUS
ar5211ReadEepromIntoDataset(WLAN_DEV_INFO *pDev)
{
    A_STATUS             status = A_OK;
    A_UINT16             tempValue;
    A_UINT32             i, j;
    A_UINT32             offset = 0;
    A_UINT16             mode;
    A_UINT16             enable24;
    A_UINT16             numChannels = 0;
    DATA_PER_CHANNEL     *pChannelData = NULL;
    PCDACS_ALL_MODES     *pEepromData;
    RD_EDGES_POWER       *pRdEdgePwrInfo;
    struct eepMap        *pMap;

    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);

    pMap           = pDev->pHalInfo->pEepData;
    pEepromData    = pMap->pPcdacInfo;
    pRdEdgePwrInfo = pMap->pRdEdgesPower;
    
    /* Read the header information here */
    ar5211ReadHeaderInfo(pDev, pMap->pEepHeader);

    enable24 = pMap->pEepHeader->Bmode;

    /*
     * Group 1: frequency pier locations readback
     * check that the structure has been populated with enough space to hold the channels
     *
     * NOTE: Group 1 contains the 5 GHz channel numbers that have dBm->pcdac calibrated information
     */
    status = ar5211EepromRead(pDev, 0x100, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }

    pEepromData->Channels11a[0] = (A_UINT16)(( tempValue >> 9 ) & FREQ_MASK);
    pEepromData->Channels11a[1] = (A_UINT16)(( tempValue >> 2 ) & FREQ_MASK);
    pEepromData->Channels11a[2] = (A_UINT16)(( tempValue << 5 ) & FREQ_MASK);

    status = ar5211EepromRead(pDev, 0x101, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pEepromData->Channels11a[2] = (A_UINT16)((( tempValue >> 11 ) & 0x1f) | pEepromData->Channels11a[2]);
    pEepromData->Channels11a[3] = (A_UINT16)(( tempValue >> 4 ) & FREQ_MASK);
    pEepromData->Channels11a[4] = (A_UINT16)(( tempValue << 3 ) & FREQ_MASK);

    status = ar5211EepromRead(pDev, 0x102, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pEepromData->Channels11a[4] = (A_UINT16)((( tempValue >> 13 ) & 0x7) | pEepromData->Channels11a[4]);
    pEepromData->Channels11a[5] = (A_UINT16)(( tempValue >> 6 ) & FREQ_MASK);
    pEepromData->Channels11a[6] =  (A_UINT16)(( tempValue << 1 ) & FREQ_MASK);

    status = ar5211EepromRead(pDev, 0x103, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pEepromData->Channels11a[6] = (A_UINT16)((( tempValue >> 15 ) & 0x1) | pEepromData->Channels11a[6]);
    pEepromData->Channels11a[7] = (A_UINT16)(( tempValue >> 8 ) & FREQ_MASK);
    pEepromData->Channels11a[8] =  (A_UINT16)(( tempValue >> 1 ) & FREQ_MASK);
    pEepromData->Channels11a[9] =  (A_UINT16)(( tempValue << 6 ) & FREQ_MASK);

    status = ar5211EepromRead(pDev, 0x104, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pEepromData->Channels11a[9] = (A_UINT16)((( tempValue >> 10 ) & 0x3f) | pEepromData->Channels11a[9]);

    for (i = 0; i < pEepromData->numChannels11a; i++ ) {
        pEepromData->Channels11a[i] = FBIN2FREQ(pEepromData->Channels11a[i]);
    }

    /*
     * Group 2:  readback data for all frequency piers
     *
     * NOTE: Group 2 contains the raw power calibration information for each of the channels
     * that we recorded above
     */
    for (mode = 0; mode <= MAX_MODE; mode++) {
        A_UINT16 *pChannels = NULL;

        switch (mode) {
        case MODE_11A:
            offset       = 0x105;
            numChannels  = pEepromData->numChannels11a;
            pChannelData = pEepromData->DataPerChannel11a;
            pChannels    = pEepromData->Channels11a;
            break;

        case MODE_OFDM_24:
            if (!enable24) {
                continue;
            }
            offset       = 0x146;
            numChannels  = pEepromData->numChannels2_4;
            pChannelData = pEepromData->DataPerChannel11g;
            pChannels    = pEepromData->Channels2_4;
            break;

        case MODE_11B:
            if (!enable24) {
                continue;
            }
            offset       = 0x137;
            numChannels  = pEepromData->numChannels2_4;
            pChannelData = pEepromData->DataPerChannel11b;
            pChannels    = pEepromData->Channels2_4;

            break;
        default:
            ASSERT(0);
        }

        for (i = 0; i < numChannels; i++) {
            pChannelData->channelValue = pChannels[i];

            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pChannelData->pcdacMax     = (A_UINT16)(( tempValue >> 10 ) & PCDAC_MASK);
            pChannelData->pcdacMin     = (A_UINT16)(( tempValue >> 4  ) & PCDAC_MASK);
            pChannelData->PwrValues[0] = (A_UINT16)(( tempValue << 2  ) & POWER_MASK);

            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pChannelData->PwrValues[0] = (A_UINT16)(((tempValue >> 14 ) & 0x3 ) | pChannelData->PwrValues[0]);
            pChannelData->PwrValues[1] = (A_UINT16)((tempValue >> 8 ) & POWER_MASK);
            pChannelData->PwrValues[2] = (A_UINT16)((tempValue >> 2 ) & POWER_MASK);
            pChannelData->PwrValues[3] = (A_UINT16)((tempValue << 4 ) & POWER_MASK);

            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pChannelData->PwrValues[3] = (A_UINT16)(( ( tempValue >> 12 ) & 0xf ) | pChannelData->PwrValues[3]);
            pChannelData->PwrValues[4] = (A_UINT16)(( tempValue >> 6 ) & POWER_MASK);
            pChannelData->PwrValues[5] = (A_UINT16)(tempValue  & POWER_MASK);

            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pChannelData->PwrValues[6] = (A_UINT16)(( tempValue >> 10 ) & POWER_MASK);
            pChannelData->PwrValues[7] = (A_UINT16)(( tempValue >> 4  ) & POWER_MASK);
            pChannelData->PwrValues[8] = (A_UINT16)(( tempValue << 2  ) & POWER_MASK);

            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pChannelData->PwrValues[8]  = (A_UINT16)(( ( tempValue >> 14 ) & 0x3 ) | pChannelData->PwrValues[8]);
            pChannelData->PwrValues[9]  = (A_UINT16)(( tempValue >> 8 ) & POWER_MASK);
            pChannelData->PwrValues[10] = (A_UINT16)(( tempValue >> 2 ) & POWER_MASK);

            getPcdacInterceptsFromPcdacMinMax(pMap->version, pChannelData->pcdacMin,
                        pChannelData->pcdacMax, pChannelData->PcdacValues) ;

            for (j = 0; j < pChannelData->numPcdacValues; j++) {
                pChannelData->PwrValues[j] = (A_UINT16)(PWR_STEP * pChannelData->PwrValues[j]);
                /* Note these values are scaled up. */
            }
            pChannelData++;
        }
    }

    /*
     * Group 5: target power values per rate
     *
     * NOTE: Group 5 contains the recorded maximum power in dB that can be attained
     * for the given rate.
     */
    /* Read the power per rate info for test channels */
    for (mode = 0; mode <= MAX_MODE; mode++) {
        TRGT_POWER_INFO     *pPowerInfo = NULL;
        A_UINT16            *pNumTrgtChannels = NULL;

        switch (mode) {
        case MODE_11A:
            offset           = 0x155;
            numChannels      = NUM_TEST_FREQUENCIES;
            pPowerInfo       = pMap->pTrgtPowerInfo->trgtPwr_11a;
            pNumTrgtChannels = &pMap->pTrgtPowerInfo->numTargetPwr_11a;
            break;

        case MODE_OFDM_24:
            if (!enable24) {
                continue;
            }
            offset           = 0x169;
            numChannels      = 3;
            pPowerInfo       = pMap->pTrgtPowerInfo->trgtPwr_11g;
            pNumTrgtChannels = &pMap->pTrgtPowerInfo->numTargetPwr_11g;
            break;

        case MODE_11B:
            if (!enable24) {
                continue;
            }
            offset           = 0x165;
            numChannels      = 2;
            pPowerInfo       = pMap->pTrgtPowerInfo->trgtPwr_11b;
            pNumTrgtChannels = &pMap->pTrgtPowerInfo->numTargetPwr_11b;
            break;
        default:
            ASSERT(0);
        } //end switch

        *pNumTrgtChannels = 0;
        for (i = 0; i < numChannels; i++) {
            status = ar5211EepromRead(pDev, offset++, &tempValue);
            if (status != A_OK) {
               goto ee_error;
            }
            pPowerInfo->testChannel = (A_UINT16)(( tempValue >> 9 ) & 0x7f);

            if (pPowerInfo->testChannel != 0) {
                /* get the channel value and read reset of info */
                if (mode == MODE_11A) {
                    pPowerInfo->testChannel = FBIN2FREQ(pPowerInfo->testChannel);
                } else {
                    pPowerInfo->testChannel = FBIN2FREQ2_4(pPowerInfo->testChannel);
                }

                pPowerInfo->twicePwr6_24 = (A_UINT16)(( tempValue >> 3 ) & POWER_MASK);
                pPowerInfo->twicePwr36   = (A_UINT16)(( tempValue << 3 ) & POWER_MASK);

                status = ar5211EepromRead(pDev, offset++, &tempValue);
                if (status != A_OK) {
                   goto ee_error;
                }
                pPowerInfo->twicePwr36 =
                    (A_UINT16)(( ( tempValue >> 13 ) & 0x7 ) | pPowerInfo->twicePwr36);
                pPowerInfo->twicePwr48 = (A_UINT16)(( tempValue >> 7 ) & POWER_MASK);
                pPowerInfo->twicePwr54 = (A_UINT16)(( tempValue >> 1 ) & POWER_MASK);
                (*pNumTrgtChannels)++;
            }
            pPowerInfo++;
        }
    }

    /*
     * Group 8: Conformance Test Limits information
     *
     * NOTE: Group 8 contains the values to limit the maximum transmit power
     * value based on any band edge violations.
     */
    /* Read the RD edge power limits */
    offset = 0x16f;
    for (i = 0; i < NUM_CTLS; i++) {
        if (pMap->pEepHeader->testGroups[i] == 0) {
            continue;
        }

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[0].rdEdge = (A_UINT16)(( tempValue >> 9 ) & FREQ_MASK);
        pRdEdgePwrInfo[1].rdEdge = (A_UINT16)(( tempValue >> 2 ) & FREQ_MASK);
        pRdEdgePwrInfo[2].rdEdge = (A_UINT16)(( tempValue << 5 ) & FREQ_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[2].rdEdge = (A_UINT16)((( tempValue >> 11 ) & 0x1f) | pRdEdgePwrInfo[2].rdEdge);
        pRdEdgePwrInfo[3].rdEdge = (A_UINT16)(( tempValue >> 4 ) & FREQ_MASK);
        pRdEdgePwrInfo[4].rdEdge = (A_UINT16)(( tempValue << 3 ) & FREQ_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[4].rdEdge = (A_UINT16)((( tempValue >> 13 ) & 0x7) | pRdEdgePwrInfo[4].rdEdge);
        pRdEdgePwrInfo[5].rdEdge = (A_UINT16)(( tempValue >> 6 ) & FREQ_MASK);
        pRdEdgePwrInfo[6].rdEdge = (A_UINT16)(( tempValue << 1 ) & FREQ_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[6].rdEdge = (A_UINT16)((( tempValue >> 15 ) & 0x1) | pRdEdgePwrInfo[6].rdEdge);
        pRdEdgePwrInfo[7].rdEdge = (A_UINT16)(( tempValue >> 8 ) & FREQ_MASK);

        pRdEdgePwrInfo[0].twice_rdEdgePower = (A_UINT16)(( tempValue >> 2 ) & POWER_MASK);
        pRdEdgePwrInfo[1].twice_rdEdgePower = (A_UINT16)(( tempValue << 4 ) & POWER_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[1].twice_rdEdgePower = (A_UINT16)((( tempValue >> 12 ) & 0xf) | pRdEdgePwrInfo[1].twice_rdEdgePower);
        pRdEdgePwrInfo[2].twice_rdEdgePower = (A_UINT16)(( tempValue >> 6 ) & POWER_MASK);
        pRdEdgePwrInfo[3].twice_rdEdgePower = (A_UINT16)(tempValue & POWER_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[4].twice_rdEdgePower = (A_UINT16)(( tempValue >> 10 ) & POWER_MASK);
        pRdEdgePwrInfo[5].twice_rdEdgePower = (A_UINT16)(( tempValue >> 4 ) & POWER_MASK);
        pRdEdgePwrInfo[6].twice_rdEdgePower = (A_UINT16)(( tempValue << 2 ) & POWER_MASK);

        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pRdEdgePwrInfo[6].twice_rdEdgePower = (A_UINT16)((( tempValue >> 14 ) & 0x3) | pRdEdgePwrInfo[6].twice_rdEdgePower);
        pRdEdgePwrInfo[7].twice_rdEdgePower = (A_UINT16)(( tempValue >> 8 ) & POWER_MASK);

        for (j = 0; j < pEepromData->numChannels11a; j++ ) {
            if (((pMap->pEepHeader->testGroups[i] & 0x3) == 0) ||
                ((pMap->pEepHeader->testGroups[i] & 0x3) == 3))
            {
                pRdEdgePwrInfo[j].rdEdge = FBIN2FREQ(pRdEdgePwrInfo[j].rdEdge);
            }
            else {
                pRdEdgePwrInfo[j].rdEdge = FBIN2FREQ2_4(pRdEdgePwrInfo[j].rdEdge);
            }
        }
        pRdEdgePwrInfo += NUM_EDGES;
    }

    return status;

ee_error:
    aPrintk("ar5211Attach: EEPROM read failed\n");
    return A_HARDWARE;
}

/**************************************************************
 * ar5211ReadHeaderInfo
 *
 * Read the individual header fields for a Rev 3 EEPROM
 */
A_STATUS
ar5211ReadHeaderInfo(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo)
{
    /* TODO: only read 2.4 info if configured */
    A_STATUS status = A_OK;
    A_UINT16 tempValue;
    A_UINT32 offset;
    A_UINT16 i;

    status = ar5211EepromRead(pDev, 0xc2, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->turboDisable = (A_UINT8)((tempValue >> 15) & 0x01);
    pHeaderInfo->rfKill = (A_UINT8)((tempValue >> 14) & 0x01);
    pHeaderInfo->deviceType = (A_UINT8)((tempValue >> 11) & 0x07);
    pHeaderInfo->turbo2WMaxPower = (A_UINT8)((tempValue >> 4) & 0x7F);
    pHeaderInfo->xtnd5GSupport = (A_UINT8)((tempValue >> 3) & 0x01);
    pHeaderInfo->Gmode = (A_UINT8)((tempValue >> 2) & 0x01);
    pHeaderInfo->Bmode = (A_UINT8)((tempValue >> 1) & 0x01);
    pHeaderInfo->Amode = (A_UINT8)(tempValue & 0x01);

    status = ar5211EepromRead(pDev, 0xc4, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaGainMax[0] = (A_INT8)((tempValue >> 8) & 0xFF);
    pHeaderInfo->antennaGainMax[1] = (A_INT8)(tempValue & 0xFF);


    status = ar5211EepromRead(pDev, 0xc5, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->switchSettling[0] = (A_UINT16)((tempValue >> 8) & 0x7f);
    pHeaderInfo->txrxAtten[0] = (A_UINT16)((tempValue >> 2) & 0x3f);
    pHeaderInfo->antennaControl[0][0] = (A_UINT16)((tempValue << 4) & 0x3f);

    status = ar5211EepromRead(pDev, 0xc6, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[0][0] = (A_UINT16)(((tempValue >> 12) & 0x0f) | pHeaderInfo->antennaControl[0][0]);
    pHeaderInfo->antennaControl[1][0] = (A_UINT16)((tempValue >> 6) & 0x3f);
    pHeaderInfo->antennaControl[2][0] = (A_UINT16)(tempValue  & 0x3f);

    status = ar5211EepromRead(pDev, 0xc7, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[3][0] = (A_UINT16)((tempValue >> 10)  & 0x3f);
    pHeaderInfo->antennaControl[4][0] = (A_UINT16)((tempValue >> 4)  & 0x3f);
    pHeaderInfo->antennaControl[5][0] = (A_UINT16)((tempValue << 2)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xc8, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[5][0] = (A_UINT16)(((tempValue >> 14)  & 0x03) | pHeaderInfo->antennaControl[5][0]);
    pHeaderInfo->antennaControl[6][0] = (A_UINT16)((tempValue >> 8)  & 0x3f);
    pHeaderInfo->antennaControl[7][0] = (A_UINT16)((tempValue >> 2)  & 0x3f);
    pHeaderInfo->antennaControl[8][0] = (A_UINT16)((tempValue << 4)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xc9, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[8][0] = (A_UINT16)(((tempValue >> 12)  & 0x0f) | pHeaderInfo->antennaControl[8][0]);
    pHeaderInfo->antennaControl[9][0] = (A_UINT16)((tempValue >> 6)  & 0x3f);
    pHeaderInfo->antennaControl[10][0] = (A_UINT16)(tempValue & 0x3f);

    status = ar5211EepromRead(pDev, 0xca, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->adcDesiredSize[0] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->ob4 = (A_UINT16)((tempValue >> 5)  & 0x07);
    pHeaderInfo->db4 = (A_UINT16)((tempValue >> 2)  & 0x07);
    pHeaderInfo->ob3 = (A_UINT16)((tempValue << 1)  & 0x07);

    status = ar5211EepromRead(pDev, 0xcb, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->ob3 = (A_UINT16)(((tempValue >> 15)  & 0x01) | pHeaderInfo->ob3);
    pHeaderInfo->db3 = (A_UINT16)((tempValue >> 12)  & 0x07);
    pHeaderInfo->ob2 = (A_UINT16)((tempValue >> 9)  & 0x07);
    pHeaderInfo->db2 = (A_UINT16)((tempValue >> 6)  & 0x07);
    pHeaderInfo->ob1 = (A_UINT16)((tempValue >> 3)  & 0x07);
    pHeaderInfo->db1 = (A_UINT16)(tempValue & 0x07);

    status = ar5211EepromRead(pDev, 0xcc, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXLNAOn[0] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->thresh62[0] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xcd, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXPAOff[0] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->txFrameToXPAOn[0] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xce, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->pgaDesiredSize[0] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->noiseFloorThresh[0] = (tempValue  & 0xff);
    if (pHeaderInfo->noiseFloorThresh[0] & 0x80) {
        pHeaderInfo->noiseFloorThresh[0] = 0 - ((pHeaderInfo->noiseFloorThresh[0] ^ 0xff) + 1);
    }


    status = ar5211EepromRead(pDev, 0xcf, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->xlnaGain[0] = (A_UINT16)((tempValue >> 5)  & 0xff);
    pHeaderInfo->xgain[0] = (A_UINT16)((tempValue >> 1)  & 0x0f);
    pHeaderInfo->xpd[0] = (A_UINT8)(tempValue  & 0x01);

    /* 11b params */
    status = ar5211EepromRead(pDev, 0xd0, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->switchSettling[1] = (A_UINT16)((tempValue >> 8) & 0x7f);
    pHeaderInfo->txrxAtten[1] = (A_UINT16)((tempValue >> 2) & 0xff);
    pHeaderInfo->antennaControl[0][1] = (A_UINT16)((tempValue << 4) & 0x3f);

    status = ar5211EepromRead(pDev, 0xd1, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[0][1] = (A_UINT16)(((tempValue >> 12) & 0x0f) | pHeaderInfo->antennaControl[0][1]);
    pHeaderInfo->antennaControl[1][1] = (A_UINT16)((tempValue >> 6) & 0x3f);
    pHeaderInfo->antennaControl[2][1] = (A_UINT16)(tempValue  & 0x3f);

    status = ar5211EepromRead(pDev, 0xd2, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[3][1] = (A_UINT16)((tempValue >> 10)  & 0x3f);
    pHeaderInfo->antennaControl[4][1] = (A_UINT16)((tempValue >> 4)  & 0x3f);
    pHeaderInfo->antennaControl[5][1] = (A_UINT16)((tempValue << 2)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xd3, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[5][1] = (A_UINT16)(((tempValue >> 14)  & 0x03) | pHeaderInfo->antennaControl[5][1]);
    pHeaderInfo->antennaControl[6][1] = (A_UINT16)((tempValue >> 8)  & 0x3f);
    pHeaderInfo->antennaControl[7][1] = (A_UINT16)((tempValue >> 2)  & 0x3f);
    pHeaderInfo->antennaControl[8][1] = (A_UINT16)((tempValue << 4)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xd4, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[8][1] = (A_UINT16)(((tempValue >> 12)  & 0x0f) | pHeaderInfo->antennaControl[8][1]);
    pHeaderInfo->antennaControl[9][1] = (A_UINT16)((tempValue >> 6)  & 0x3f);
    pHeaderInfo->antennaControl[10][1] = (A_UINT16)(tempValue & 0x3f);

    status = ar5211EepromRead(pDev, 0xd5, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->adcDesiredSize[1] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->obFor24 = (A_UINT16)((tempValue >> 4)  & 0x07);
    pHeaderInfo->dbFor24 = (A_UINT16)(tempValue & 0x07);

    status = ar5211EepromRead(pDev, 0xd6, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXLNAOn[1] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->thresh62[1] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xd7, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXPAOff[1] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->txFrameToXPAOn[1] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xd8, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->pgaDesiredSize[1] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->noiseFloorThresh[1] = (A_INT16)(tempValue  & 0xff);
    if (pHeaderInfo->noiseFloorThresh[1] & 0x80) {
        pHeaderInfo->noiseFloorThresh[1] = 0 - ((pHeaderInfo->noiseFloorThresh[1] ^ 0xff) + 1);
    }

    status = ar5211EepromRead(pDev, 0xd9, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->xlnaGain[1] = (A_UINT16)((tempValue >> 5)  & 0xff);
    pHeaderInfo->xgain[1] = (A_UINT16)((tempValue >> 1)  & 0x0f);
    pHeaderInfo->xpd[1] = (A_UINT8)(tempValue  & 0x01);

    /* 11g params */
    status = ar5211EepromRead(pDev, 0xda, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->switchSettling[2] = (A_UINT16)((tempValue >> 8) & 0x7f);
    pHeaderInfo->txrxAtten[2] = (A_UINT16)((tempValue >> 2) & 0xff);
    pHeaderInfo->antennaControl[0][2] = (A_UINT16)((tempValue << 4) & 0x3f);

    status = ar5211EepromRead(pDev, 0xdb, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[0][2] = (A_UINT16)(((tempValue >> 12) & 0x0f) | pHeaderInfo->antennaControl[0][2]);
    pHeaderInfo->antennaControl[1][2] = (A_UINT16)((tempValue >> 6) & 0x3f);
    pHeaderInfo->antennaControl[2][2] = (A_UINT16)(tempValue  & 0x3f);

    status = ar5211EepromRead(pDev, 0xdc, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[3][2] = (A_UINT16)((tempValue >> 10)  & 0x3f);
    pHeaderInfo->antennaControl[4][2] = (A_UINT16)((tempValue >> 4)  & 0x3f);
    pHeaderInfo->antennaControl[5][2] = (A_UINT16)((tempValue << 2)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xdd, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[5][2] = (A_UINT16)(((tempValue >> 14)  & 0x03) | pHeaderInfo->antennaControl[5][2]);
    pHeaderInfo->antennaControl[6][2] = (A_UINT16)((tempValue >> 8)  & 0x3f);
    pHeaderInfo->antennaControl[7][2] = (A_UINT16)((tempValue >> 2)  & 0x3f);
    pHeaderInfo->antennaControl[8][2] = (A_UINT16)((tempValue << 4)  & 0x3f);

    status = ar5211EepromRead(pDev, 0xde, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->antennaControl[8][2] = (A_UINT16)(((tempValue >> 12)  & 0x0f) | pHeaderInfo->antennaControl[8][2]);
    pHeaderInfo->antennaControl[9][2] = (A_UINT16)((tempValue >> 6)  & 0x3f);
    pHeaderInfo->antennaControl[10][2] = (A_UINT16)(tempValue & 0x3f);

    status = ar5211EepromRead(pDev, 0xdf, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->adcDesiredSize[2] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->obFor24g = (A_UINT16)((tempValue >> 4)  & 0x07);
    pHeaderInfo->dbFor24g = (A_UINT16)(tempValue & 0x07);

    status = ar5211EepromRead(pDev, 0xe0, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXLNAOn[2] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->thresh62[2] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xe1, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->txEndToXPAOff[2] = (A_UINT16)((tempValue >> 8)  & 0xff);
    pHeaderInfo->txFrameToXPAOn[2] = (A_UINT16)(tempValue  & 0xff);

    status = ar5211EepromRead(pDev, 0xe2, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->pgaDesiredSize[2] = (A_INT8)((tempValue >> 8)  & 0xff);
    pHeaderInfo->noiseFloorThresh[2] = (A_INT16)(tempValue  & 0xff);
    if (pHeaderInfo->noiseFloorThresh[2] & 0x80) {
        pHeaderInfo->noiseFloorThresh[2] = 0 - ((pHeaderInfo->noiseFloorThresh[2] ^ 0xff) + 1);
    }

    status = ar5211EepromRead(pDev, 0xe3, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->xlnaGain[2] = (A_UINT16)((tempValue >> 5)  & 0xff);
    pHeaderInfo->xgain[2] = (A_UINT16)((tempValue >> 1)  & 0x0f);
    pHeaderInfo->xpd[2] = (A_UINT8)(tempValue  & 0x01);

    /* WAR for incorrectly programmed EEPROM's before version 3.2 */
    if (pDev->pHalInfo->pEepData->version < EEPROM_VER3_2) {
        for (i = 0; i < 3; i++) {
            pHeaderInfo->noiseFloorThresh[i] = -70;
        }
    }

    /* WAR to turn override thresh62 for better 2.4 and 5 operation */
    if (pDev->pHalInfo->pEepData->version <= EEPROM_VER3_2) {
        pHeaderInfo->thresh62[0] = 15; // 11A
        pHeaderInfo->thresh62[1] = 28; // 11B
        pHeaderInfo->thresh62[2] = 28; // 11G
    }

    /*
     * Read the test group values
     * These are used to match regulatory domain testing needs with
     * the tests that have been calibrated in the EEPROM.
     */
    offset = 0xe4;
    for (i = 0; i < NUM_CTLS; i += 2) {
        status = ar5211EepromRead(pDev, offset++, &tempValue);
        if (status != A_OK) {
           goto ee_error;
        }
        pHeaderInfo->testGroups[i] = (A_UINT16)((tempValue >> 8) & 0xff);
        pHeaderInfo->testGroups[i+1] = (A_UINT16)(tempValue & 0xff);
    }

    /* Version 3.1+ specific parameters */
    status = ar5211EepromRead(pDev, 0xec, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->ob2GHz[0] = tempValue & 0x7;
    pHeaderInfo->db2GHz[0] = (tempValue >> 3) & 0x7;

    status = ar5211EepromRead(pDev, 0xed, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->ob2GHz[1] = tempValue & 0x7;
    pHeaderInfo->db2GHz[1] = (tempValue >> 3) & 0x7;

    /* read corner cal info */
    status = ar5211EepromRead(pDev, 0xee, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->cornerCal[0].gSel = (A_UINT16)(tempValue & 0x1);
    pHeaderInfo->cornerCal[0].pd84 = (A_UINT16)((tempValue >> 1) & 0x1);
    pHeaderInfo->cornerCal[0].pd90 = (A_UINT16)((tempValue >> 2) & 0x1);
    pHeaderInfo->cornerCal[0].clip = (A_UINT16)((tempValue >> 3) & 0x7);
    pHeaderInfo->cornerCal[1].gSel = (A_UINT16)((tempValue >> 6) & 0x1);
    pHeaderInfo->cornerCal[1].pd84 = (A_UINT16)((tempValue >> 7) & 0x1);
    pHeaderInfo->cornerCal[1].pd90 = (A_UINT16)((tempValue >> 8) & 0x1);
    pHeaderInfo->cornerCal[1].clip = (A_UINT16)((tempValue >> 9) & 0x7);
    pHeaderInfo->cornerCal[2].gSel = (A_UINT16)((tempValue >> 12) & 0x1);
    pHeaderInfo->cornerCal[2].pd84 = (A_UINT16)((tempValue >> 13) & 0x1);
    pHeaderInfo->cornerCal[2].pd90 = (A_UINT16)((tempValue >> 14) & 0x1);
    pHeaderInfo->cornerCal[2].clip = (A_UINT16)((tempValue >> 15) & 0x1);
    
    status = ar5211EepromRead(pDev, 0xef, &tempValue);
    if (status != A_OK) {
       goto ee_error;
    }
    pHeaderInfo->cornerCal[2].clip = (A_UINT16)(((tempValue & 0x3) << 1) | pHeaderInfo->cornerCal[2].clip);
    pHeaderInfo->cornerCal[3].gSel = (A_UINT16)((tempValue >> 2) & 0x1);
    pHeaderInfo->cornerCal[3].pd84 = (A_UINT16)((tempValue >> 3) & 0x1);
    pHeaderInfo->cornerCal[3].pd90 = (A_UINT16)((tempValue >> 4) & 0x1);
    pHeaderInfo->cornerCal[3].clip = (A_UINT16)((tempValue >> 5) & 0x7);

    return A_OK;

ee_error:
    aPrintk("ar5211Attach: EEPROM read failed\n");
    return A_HARDWARE;
}

void
getPcdacInterceptsFromPcdacMinMax(A_UINT16 version, A_UINT16 pcdacMin,
    A_UINT16 pcdacMax, A_UINT16 *pPcdacValues)
{
    const static A_UINT16 intercepts3[] = {0,5,10,20,30,50,70,85,90,95,100};
    const static A_UINT16 intercepts3_2[] = {0,10,20,30,40,50,60,70,80,90,100};

    A_UINT16 i;
    const A_UINT16 *pIntercepts;

    if (version >= EEPROM_VER3_2) {
        pIntercepts = intercepts3_2;
    } else {
        pIntercepts = intercepts3;
    }

    /* loop for the percentages in steps or 5 */
    for (i = 0; i < NUM_INTERCEPTS; i++ ) {
        *pPcdacValues =  (A_UINT16)((pIntercepts[i] * pcdacMax + (100 - pIntercepts[i]) * pcdacMin) / 100);
        pPcdacValues++;
    }
}

void printWlanErrorType (int errCode)
{
    if (errCode == A_ERROR) aPrintk ("Generic error return\n");
    if (errCode == A_DEVICE_NOT_FOUND) aPrintk ("not able to find PCI device \n");
    if (errCode == A_NO_MEMORY) aPrintk ("not able to allocate memory, not available\n");
    if (errCode == A_MEMORY_NOT_AVAIL) aPrintk ("memory region is not free for mapping\n");
    if (errCode == A_NO_FREE_DESC) aPrintk ("no free descriptors available\n");
    if (errCode == A_BAD_ADDRESS) aPrintk ("address does not match descriptor\n");
    if (errCode == A_WIN_DRIVER_ERROR) aPrintk ("used in NT_HW version, if problem at init\n");
    if (errCode == A_REGS_NOT_MAPPED) aPrintk ("registers not correctly mapped\n");
    if (errCode == A_EPERM) aPrintk ("Not superuser\n");
    if (errCode == A_EACCES) aPrintk ("Access denied\n");
    if (errCode == A_ENOENT) aPrintk ("No such entry, search failed, etc.\n");
    if (errCode == A_EEXIST) aPrintk ("The object already exists (can't create)\n");
    if (errCode == A_EFAULT) aPrintk ("Bad address fault\n");
    if (errCode == A_EBUSY) aPrintk ("Object is busy\n");
    if (errCode == A_EINVAL) aPrintk ("Invalid parameter\n");
    if (errCode == A_ECANCELED) aPrintk ("Operation canceled\n");
    if (errCode == A_ENOTSUP) aPrintk ("Operation not supported\n");
    if (errCode == A_ECOMM) aPrintk ("Communication error on send\n");
    if (errCode == A_EPROTO) aPrintk ("Protocol error\n");
    if (errCode == A_ENODEV) aPrintk ("No such device\n");
    if (errCode == A_NO_RESOURCE) aPrintk ("No resources for requested operation\n");
    if (errCode == A_HARDWARE) aPrintk ("Hardware failure\n");
    if (errCode == A_PENDING) aPrintk ("Asynchronous routine; will send up results later (typically in callback)\n");
    if (errCode == A_EBADCHANNEL) aPrintk ("The channel cannot be used\n");
    if (errCode == A_DECRYPT_ERROR) aPrintk ("Decryption errror\n");
    return;
}


#endif /* #ifdef BUILD_AR5211 */
