/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * TODO: comment
 *
 * $Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Interrupts.c#2 $
 */

#ifdef BUILD_AR5211

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "ui.h"
#include "halApi.h"
#include "hal.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Interrupts.h"
#include "ar5211.h"


/**************************************************************
 * ar5211IsInteruptPending
 *
 * Checks to see if an interrupt is pending on our NIC
 *
 * Returns: TRUE    if an interrupt is pending
 *          FALSE   if not
 */
A_BOOL
ar5211IsInterruptPending(WLAN_DEV_INFO *pDev)
{
#if defined(PCI_INTERFACE)
    return (readPlatformReg(pDev, MAC_INTPEND) != 0);
#elif defined(AR5311)
    return TRUE;
#endif
}

/**************************************************************
 * halGetInterrupts
 *
 * Reads the Interrupt Status Register value from the NIC, thus deasserting
 * the interrupt line, and returns both the masked and unmasked mapped ISR
 * values.  The value returned is mapped to abstract the hw-specific bit
 * locations in the Interrupt Status Register.
 *
 * Returns: A hardware-abstracted bitmap of all non-masked-out
 *          interrupts pending, as well as an unmasked value
 */
#include <linux/pci.h>
extern struct pci_dev * gPciDev;
void dumpDevBufsAndDescs (WLAN_DEV_INFO* pDev);

HAL_INT_TYPE
ar5211GetInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE *pUnmaskedValue)
{
    A_UINT32        isr, maskedIsr;
    HAL_INT_TYPE    ints = 0, maskedInts = 0;

    isr = readPlatformReg(pDev, MAC_ISR_RAC);

    if (isr == 0xffffffff) {
        return HAL_INT_NOCARD;
    }

    maskedIsr = isr & pDev->MaskReg;
    ASSERT(maskedIsr);

    /* Mask out non-common interrupts.  These will be added in below. */
    maskedInts = maskedIsr & HAL_INT_COMMON;

#ifdef PCI_INTERFACE
    if (maskedIsr & MAC_ISR_HIUERR) {
        A_UINT32 temp = readPlatformReg(pDev, MAC_ISR_S2_S);


A_UINT16 tmp16;
pci_read_config_word(gPciDev, PCI_STATUS, (A_UINT16 *)&tmp16);
printk("\n");
if (tmp16 & PCI_STATUS_REC_MASTER_ABORT) 
    printk(".............PCI_STATUS_REC_MASTER_ABORT\n");
if (tmp16 & PCI_STATUS_REC_TARGET_ABORT) 
    printk(".............PCI_STATUS_REC_TARGET_ABORT\n");
if (tmp16 & PCI_STATUS_SIG_TARGET_ABORT) 
    printk(".............PCI_STATUS_SIG_TARGET_ABORT\n");
if (tmp16 & PCI_STATUS_SIG_SYSTEM_ERROR) 
    printk(".............PCI_STATUS_SIG_SYSTEM_ERROR\n");
if (tmp16 & PCI_STATUS_DETECTED_PARITY) 
    printk(".............PCI_STATUS_DETECTED_PARITY\n");
if (tmp16 & PCI_STATUS_CAP_LIST) 
    printk(".............PCI_STATUS_CAP_LIST\n");
if (tmp16 & PCI_STATUS_66MHZ) 
    printk(".............PCI_STATUS_66MHZ\n");
if (tmp16 & PCI_STATUS_FAST_BACK) 
    printk(".............PCI_STATUS_FAST_BACK\n");
if (tmp16 & PCI_STATUS_PARITY) 
    printk(".............PCI_STATUS_PARITY\n");
if (tmp16 & PCI_STATUS_DEVSEL_MASK) 
    printk(".............PCI_STATUS_DEVSEL_MASK\n");


        if (temp & MAC_ISR_S2_MCABT) {
            isrPrintf("hwGetInterrupts: Bus Master Cycle Abort Error!\n");
//fadi            dumpDevBufsAndDescs (pDev);
        }
        if (temp & MAC_ISR_S2_SSERR) {
            isrPrintf("hwGetInterrupts: Bus Signalled System Error!\n");
        }
        if (temp & MAC_ISR_S2_DPERR) {
            isrPrintf("hwGetInterrupts: Bus Parity Error!\n");
        }
        maskedInts |= HAL_INT_FATAL;
    }
#endif

    if (maskedIsr & (MAC_ISR_RXOK | MAC_ISR_RXERR)) {
        maskedInts |= HAL_INT_RX;
    }
    if (maskedIsr & (MAC_ISR_TXOK | MAC_ISR_TXERR)) {
        maskedInts |= HAL_INT_TX;
    }

    /* Receive overrun is usually non-fatal on Oahu/Spirit.
     *
     * BUT early silicon had a bug which causes the rx to fail and the chip
     * must be reset.
     *
     * AND even AR5311 v3.1 and AR5211 v4.2 silicon seems to have a problem
     * with RXORN which hangs the receive path and requires a chip reset
     * to proceed (see bug 3996). So for now, we force a hardware reset in
     * all cases.
     */
    if (isr & MAC_ISR_RXORN) {
        maskedInts |= HAL_INT_FATAL;
    }

    /*
     * Now we take care of the unmasked interrupts.  We need to return certain
     * unmasked interrupts or we'll "lose" them after reading the ISR register.
     */
    ints = isr & HAL_INT_COMMON;

    if (isr & (MAC_ISR_RXOK | MAC_ISR_RXERR)) {
        ints |= HAL_INT_RX;
    }
    if (isr & (MAC_ISR_TXOK | MAC_ISR_TXERR)) {
        ints |= HAL_INT_TX;
    }

    /* Return the unmasked value. */
    *pUnmaskedValue = ints;

    /* Return the masked value. */
    return maskedInts;
}

/**************************************************************
 * ar5211EnableInterrupts
 *
 * Atomically enables NIC interrupts.  Interrupts are passed in
 * via the enumerated bitmask in ints.
 */
void
ar5211EnableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints)
{
    A_UINT32    mask;
    A_BOOL      intsEnabled;

    if (ints & HAL_INT_GLOBAL) {
        /* Assert that no other ints are combined with the global bit. */
        ASSERT((ints & ~HAL_INT_GLOBAL) == 0);

        ASSERT(pDev->pHalInfo->globIntRefCount!=0);

        if(pDev->pHalInfo->globIntRefCount > 0) {
            pDev->pHalInfo->globIntRefCount--;
        }
        if (pDev->pHalInfo->globIntRefCount == 0) {
            writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
        }
        return;
    }

    /*
     * We have to disable interrupts here before reading & modifying
     * the mask so that the ISR does not modify the mask out from under us.
     */
    intsEnabled = (readPlatformReg(pDev, MAC_IER) == MAC_IER_ENABLE);
    if (intsEnabled) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_DISABLE);
    }

    mask = pDev->MaskReg;

    /* Take care of interrupt sources that differ from HW to HW */
    
    if (ints & HAL_INT_TX) {
        A_UINT32 txMask;

        txMask = (1 << pDev->dataQueueNum) | (1 << pDev->cabQueueNum);
        writePlatformReg(pDev, MAC_IMR_S0, txMask); /* TXOK, not TXDESC */
        writePlatformReg(pDev, MAC_IMR_S1, txMask); /* TXERR, not TXEOL */
        txMask |= MAC_IMR_S2_MCABT | MAC_IMR_S2_SSERR | MAC_IMR_S2_DPERR;


        writePlatformReg(pDev, MAC_IMR_S2, txMask); /* TXURN, et al */
        mask |= MAC_IMR_TXOK | MAC_IMR_TXERR;
    }

    if (ints & HAL_INT_RX) {
        mask |= MAC_IMR_RXOK | MAC_IMR_RXERR;
    }

    /* Take care of common interrupt sources */

    ints &= HAL_INT_COMMON;
    mask |= ints;

    /* Write the new IMR and store off our SW copy. */
    writePlatformReg(pDev, MAC_IMR, mask);
    pDev->MaskReg = mask;

    /* Re-enable interrupts if they were enabled before. */
    if (intsEnabled) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
    }
}

/**************************************************************
 * ar5211DisableInterrupts
 *
 * Atomically disables NIC interrupts.  Interrupts are passed in
 * via the enumerated bitmask in "ints".
 */
void
ar5211DisableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints)
{
    A_UINT32    mask;
    A_BOOL      intsEnabled;

    if (ints & HAL_INT_GLOBAL) {
        /* Assert that no other ints are combined with the global bit. */
        ASSERT((ints & ~HAL_INT_GLOBAL) == 0);

        pDev->pHalInfo->globIntRefCount++;
        writePlatformReg(pDev, MAC_IER, MAC_IER_DISABLE);
        return;
    }

    /*
     * We have to disable interrupts here before reading & modifying
     * the mask so that the ISR does not modify the mask out from under us.
     */
    intsEnabled = (readPlatformReg(pDev, MAC_IER) == MAC_IER_ENABLE);
    if (intsEnabled) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_DISABLE);
    }

    mask = pDev->MaskReg;

    /* Take care of interrupt sources that differ from HW to HW */

    if (ints & HAL_INT_TX) {
        writePlatformReg(pDev, MAC_IMR_S0, 0); /* TXOK */
        writePlatformReg(pDev, MAC_IMR_S1, 0); /* TXERR */
        mask &= ~(MAC_IMR_TXOK | MAC_IMR_TXERR);
    }

    if (ints & HAL_INT_RX) {
        mask &= ~(MAC_IMR_RXOK | MAC_IMR_RXERR);
    }

    /* Take care of common interrupt sources */

    ints &= HAL_INT_COMMON;
    mask &= ~ints;

    /* Write the new IMR and store off our SW copy. */
    writePlatformReg(pDev, MAC_IMR, mask);
    pDev->MaskReg = mask;

    /* Re-enable interrupts if they were enabled before. */
    if (intsEnabled) {
        writePlatformReg(pDev, MAC_IER, MAC_IER_ENABLE);
    }
}

void dumpDevBufsAndDescs (WLAN_DEV_INFO* pdevInfo)
{
    int i=0;
    ATHEROS_DESC *pDesc;
    QUEUE_DETAILS *pQueue;

    printk("\nWelcome to dumpDevBufsAndDescs...\n");
    printk("==================================\n\n");
    printk("freeDescCount=%d, txHdrDescCount=%d, psQueueDepth=%d, rxQueueDepth=%d\n", 
           pdevInfo->freeDescCount, pdevInfo->txHeaderDescCount,
           pdevInfo->psQueueDepth, pdevInfo->rxQueueDepth);
    printk("AR_DESC_SIZE=%d, AR_HEADER_SIZE=%d, AR_BUF_SIZE=%d\n", 
           AR_DESC_SIZE, AR_HEADER_SIZE, AR_BUF_SIZE);
    printk("pdevInfo->pDmaBuf = %p[dma_size=%d][pMclMem=%p]\n", 
           pdevInfo->pDmaBuf, pdevInfo->dmaBufSize, pdevInfo->pMclMem);
    printk("\n");
    printk("pdevInfo->emptyDescQueue:[head should be at (pDmaBuf + clMemSize)]\n");
    printk("\tpdevInfo->emptyDescQueue.pDescQueueHead = %p\n", pdevInfo->emptyDescQueue.pDescQueueHead);
    printk("\tpdevInfo->emptyDescQueue.physQueueHead  = 0x%x\n", pdevInfo->emptyDescQueue.physQueueHead);
    printk("\tpdevInfo->emptyDescQueue.pDescQueueTail = %p\n", pdevInfo->emptyDescQueue.pDescQueueTail);

    pQueue = &pdevInfo->emptyDescQueue;
    for (i = 0, pDesc = pQueue->pDescQueueHead; 
          i < pdevInfo->freeDescCount; ++i, pDesc = pDesc->pNextVirtPtr) {
        printk("empty[%d]: next %p, pnext %p, this %p, free %p, bufphys %p, pbufvirt %p, pos %p\n",
               i, pDesc->nextPhysPtr, pDesc->pNextVirtPtr, pDesc->thisPhysPtr,
               pDesc->freeQueue, pDesc->bufferPhysPtr, pDesc->pBufferVirtPtr.ptr,
               pDesc->pOSDescPtr);
#if 0
        printk("\tpDesc->nextPhysPtr  = 0x%x\n", pDesc->nextPhysPtr);
        printk("\tpDesc->pNextVirtPtr = %p\n", pDesc->pNextVirtPtr);
        printk("\tpDesc->thisPhysPtr  = 0x%x\n", pDesc->thisPhysPtr);
        printk("\tpDesc->freeQueue    = %p\n", pDesc->freeQueue);
        printk("\tpDesc->bufferPhysPtr = 0x%x\n", pDesc->bufferPhysPtr);
        printk("\tpDesc->pBufferVirtPtr.ptr = %p\n", pDesc->pBufferVirtPtr.ptr);
        printk("\tpDesc->pOrigBufferVirtPtr = %p\n", pDesc->pOrigBufferVirtPtr);
        printk("\tpDesc->pOSDescPtr = %p\n", pDesc->pOSDescPtr);
#endif
    }
    printk("\tpQueue->pDescQueueTail->nextPhysPtr  = 0x%x\n", pQueue->pDescQueueTail->nextPhysPtr);
    printk("\tpQueue->pDescQueueTail->pNextVirtPtr = %p\n", pQueue->pDescQueueTail->pNextVirtPtr);

    printk("\n");
    printk("pdevInfo->txHeaderQueue:[head should be at (emptyDescQueue + freeDescCnt)]\n");
    printk("\tpdevInfo->txHeaderQueue.pDescQueueHead = %p\n", pdevInfo->txHeaderQueue.pDescQueueHead);
    printk("\tpdevInfo->txHeaderQueue.physQueueHead  = 0x%x\n", pdevInfo->txHeaderQueue.physQueueHead);
    printk("\tpdevInfo->txHeaderQueue.pDescQueueTail = %p\n", pdevInfo->txHeaderQueue.pDescQueueTail);
    
    pQueue = &pdevInfo->txHeaderQueue;
    for (i = 0, pDesc = pQueue->pDescQueueHead; 
          i < pdevInfo->txHeaderDescCount; ++i, pDesc = pDesc->pNextVirtPtr) {
            printk("txhdr[%d]: next %p, pnext %p, this %p, free %p, bufphys %p, pbufvirt %p, pos %p\n",
               i, pDesc->nextPhysPtr, pDesc->pNextVirtPtr, pDesc->thisPhysPtr,
               pDesc->freeQueue, pDesc->bufferPhysPtr, pDesc->pBufferVirtPtr.ptr,
               pDesc->pOSDescPtr);
#if 0
        printk("\tpDesc->nextPhysPtr  = 0x%x\n", pDesc->nextPhysPtr);
        printk("\tpDesc->pNextVirtPtr = %p\n", pDesc->pNextVirtPtr);
        printk("\tpDesc->thisPhysPtr  = 0x%x\n", pDesc->thisPhysPtr);
        printk("\tpDesc->freeQueue    = %p\n", pDesc->freeQueue);
        printk("\tpDesc->bufferPhysPtr = 0x%x\n", pDesc->bufferPhysPtr);
        printk("\tpDesc->pBufferVirtPtr.ptr = %p\n", pDesc->pBufferVirtPtr.ptr);
        printk("\tpDesc->pOrigBufferVirtPtr = %p\n", pDesc->pOrigBufferVirtPtr);
        printk("\tpDesc->pOSDescPtr = %p\n", pDesc->pOSDescPtr);
#endif
    }
    printk("\tpQueue->pDescQueueTail->nextPhysPtr  = 0x%x\n", pQueue->pDescQueueTail->nextPhysPtr);
    printk("\tpQueue->pDescQueueTail->pNextVirtPtr = %p\n", pQueue->pDescQueueTail->pNextVirtPtr);
#if FADI
    printk("\n");
    printk("pdevInfo->psBufferQueue:[head should be at (txHeaderQueue + txHeaderDescCount)]\n");
    printk("\tpdevInfo->psBufferQueue.pDescQueueHead = %p\n", pdevInfo->psBufferQueue.pDescQueueHead);
    printk("\tpdevInfo->psBufferQueue.physQueueHead  = 0x%x\n", pdevInfo->psBufferQueue.physQueueHead);
    printk("\tpdevInfo->psBufferQueue.pDescQueueTail = %p\n", pdevInfo->psBufferQueue.pDescQueueTail);
    
    pQueue = &pdevInfo->psBufferQueue;
    for (i = 0, pDesc = pQueue->pDescQueueHead; 
          i < count; ++i, pDesc = pDesc->pNextVirtPtr) {
        printk("\tpDesc->nextPhysPtr  = 0x%x\n", pDesc->nextPhysPtr);
        printk("\tpDesc->pNextVirtPtr = %p\n", pDesc->pNextVirtPtr);
        printk("\tpDesc->thisPhysPtr  = 0x%x\n", pDesc->thisPhysPtr);
        printk("\tpDesc->freeQueue    = %p\n", pDesc->freeQueue);
        printk("\tpDesc->bufferPhysPtr = 0x%x\n", pDesc->bufferPhysPtr);
        printk("\tpDesc->pBufferVirtPtr.ptr = %p\n", pDesc->pBufferVirtPtr.ptr);
        printk("\tpDesc->pOrigBufferVirtPtr = %p\n", pDesc->pOrigBufferVirtPtr);
        printk("\tpDesc->pOSDescPtr = %p\n", pDesc->pOSDescPtr);
    }
    printk("\tpQueue->pDescQueueTail->nextPhysPtr  = 0x%x\n", pQueue->pDescQueueTail->nextPhysPtr);
    printk("\tpQueue->pDescQueueTail->pNextVirtPtr = %p\n", pQueue->pDescQueueTail->pNextVirtPtr);
#endif

    printk("\n");
    printk("pdevInfo->rxQueue:[head should be at (psBufferQueue + psQueueDepth)]\n");
    printk("\tpdevInfo->rxQueue.pDescQueueHead = %p\n", pdevInfo->rxQueue.pDescQueueHead);
    printk("\tpdevInfo->rxQueue.physQueueHead  = 0x%x\n", pdevInfo->rxQueue.physQueueHead);
    printk("\tpdevInfo->rxQueue.pDescQueueTail = %p\n", pdevInfo->rxQueue.pDescQueueTail);
    
    pQueue = &pdevInfo->rxQueue;
    for (i = 0, pDesc = pQueue->pDescQueueHead; 
          i < pdevInfo->rxQueueDepth; ++i, pDesc = pDesc->pNextVirtPtr) {
        printk("rxQueue[%d]: next %p, pnext %p, this %p, free %p, bufphys %p, pbufvirt %p, pos %p\n",
               i, pDesc->nextPhysPtr, pDesc->pNextVirtPtr, pDesc->thisPhysPtr,
               pDesc->freeQueue, pDesc->bufferPhysPtr, pDesc->pBufferVirtPtr.ptr,
               pDesc->pOSDescPtr);
#if 0
        printk("\tpDesc->nextPhysPtr  = 0x%x\n", pDesc->nextPhysPtr);
        printk("\tpDesc->pNextVirtPtr = %p\n", pDesc->pNextVirtPtr);
        printk("\tpDesc->thisPhysPtr  = 0x%x\n", pDesc->thisPhysPtr);
        printk("\tpDesc->freeQueue    = %p\n", pDesc->freeQueue);
        printk("\tpDesc->bufferPhysPtr = 0x%x\n", pDesc->bufferPhysPtr);
        printk("\tpDesc->pBufferVirtPtr.ptr = %p\n", pDesc->pBufferVirtPtr.ptr);
        printk("\tpDesc->pOrigBufferVirtPtr = %p\n", pDesc->pOrigBufferVirtPtr);
        printk("\tpDesc->pOSDescPtr = %p\n", pDesc->pOSDescPtr);
#endif
    }
    printk("\tpQueue->pDescQueueTail->nextPhysPtr  = 0x%x\n", pQueue->pDescQueueTail->nextPhysPtr);
    printk("\tpQueue->pDescQueueTail->pNextVirtPtr = %p\n", pQueue->pDescQueueTail->pNextVirtPtr);
    
    printk("\n");
    printk("pdevInfo->txQueue.pDescQueueHead  = %p\n", pdevInfo->txQueue.pDescQueueHead);
    printk("pdevInfo->txQueue.pDescQueueTail  = %p\n", pdevInfo->txQueue.pDescQueueTail);
    printk("pdevInfo->altQueue.pDescQueueHead = %p\n", pdevInfo->altQueue.pDescQueueHead);
    printk("pdevInfo->altQueue.pDescQueueTail = %p\n", pdevInfo->altQueue.pDescQueueTail);
    printk("pdevInfo->lastTxFrameDesc         = %0\n", pdevInfo->lastTxFrameDesc);
    printk("==================================\n\n");
}

#endif /* BUILD_AR5211 */
