/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 *  Transmit HAL functions
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Receive.c#1 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Transmit.h"
#include "ar5211.h"

/**************************************************************
 * ar5211GetRxDP
 *
 * Get the RXDP.
 */
A_UINT32
ar5211GetRxDP(WLAN_DEV_INFO *pDev)
{
    //printk("Inside ar5211GetRxDP.....\n");
    return readPlatformReg(pDev, MAC_RXDP);
}

/**************************************************************
 * ar5211SetRxDP
 *
 * Set the RxDP.
 */
void
ar5211SetRxDP(WLAN_DEV_INFO *pDev, A_UINT32 rxdp)
{
    //printk("Inside ar5211SetRxDP.....\n");
    writePlatformReg(pDev, MAC_RXDP, rxdp);
    ASSERT(readPlatformReg(pDev, MAC_RXDP) == rxdp);
}


/**************************************************************
 * ar5211EnableReceive
 *
 * Set Receive Enable bits.
 */
void
ar5211EnableReceive(WLAN_DEV_INFO *pDev)
{
    writePlatformReg(pDev, MAC_CR, MAC_CR_RXE);
}

/**************************************************************
 * ar5211StopDmaReceive
 *
 * Stop Receive at the DMA engine
 */
A_STATUS
ar5211StopDmaReceive(WLAN_DEV_INFO *pDev)
{
    A_UINT32    regVal, i;

    printk("STOPPPPPPPP ar5211StopDmaReceive.....\n");

    /* Set receive disable bit */
    writePlatformReg(pDev, MAC_CR, MAC_CR_RXD);

    /* Wait up to 10ms */
    for (i=0; i<1000; i++) {

        regVal = readPlatformReg(pDev, MAC_CR);

        if (!(regVal & MAC_CR_RXE)) {
            return A_OK;
        }

        udelay(1000); // used to be 10
    }

    ASSERT(0);
    return A_HARDWARE;
}

/**************************************************************
 * ar5211StartPcuReceive
 *
 * Start Transmit at the PCU engine (unpause receive)
 */
void
ar5211StartPcuReceive(WLAN_DEV_INFO *pDev)
{
    writePlatformReg(pDev, MAC_DIAG_SW,
        readPlatformReg(pDev, MAC_DIAG_SW) & ~(MAC_DIAG_RX_DIS));
}

/**************************************************************
 * ar5211StopPcuReceive
 *
 * Stop Transmit at the PCU engine (pause receive)
 */
void
ar5211StopPcuReceive(WLAN_DEV_INFO *pDev)
{
    writePlatformReg(pDev, MAC_DIAG_SW, readPlatformReg(pDev, MAC_DIAG_SW) | MAC_DIAG_RX_DIS);
}

/**************************************************************
 * ar5211SetMulticastFilter
 *
 * Set multicast filter 0 (lower 32-bits)
 *               filter 1 (upper 32-bits)
 */
void
ar5211SetMulticastFilter(WLAN_DEV_INFO *pDev, A_UINT32 filter0, A_UINT32 filter1)
{
    writePlatformReg(pDev, MAC_MCAST_FIL0, filter0);
    writePlatformReg(pDev, MAC_MCAST_FIL1, filter1);
}

/**************************************************************
 * ar5211MulticastFilterIndex
 *
 * Clear/Set multicast filter by index
 */
void
ar5211MulticastFilterIndex(WLAN_DEV_INFO *pDev, A_UINT32 index, A_BOOL bSet)
{
    A_UINT32 filterReg;
    A_UINT32 filterIndex;
    A_UINT32 value;

    ASSERT(index < 64);

    if (index < 32) {
        filterReg   = MAC_MCAST_FIL0;
        filterIndex = index;
    } else {
        filterReg   = MAC_MCAST_FIL1;
        filterIndex = index - 32;
    }

    value = readPlatformReg(pDev, filterReg);

    if (bSet) {
        value |= (1 << filterIndex);
    } else {
        value &= ~(1 << filterIndex);
    }

    writePlatformReg(pDev, filterReg, value);
}

/**************************************************************************
 * ar5211RxFilter
 *
 * Turn off/on bits in the receive packet filter.  Save it for resets, etc.
 *
 * Bits can be MAC_RX_UCAST, MAC_RX_MCAST, MAC_RX_BCAST, MAC_RX_BEACON,
 * MAC_RX_PROM, MAC_RX_CONTROL
 *
 * oper can be HAL_RX_FILTER_INIT, HAL_RX_FILTER_SET, HAL_RX_FILTER_CLEAR,
 * HAL_RX_FILTER_TEMP, HAL_RX_FILTER_RESTORE.
 *
 * NOTE: the bits parameter is not used for operation type HAL_RX_FILTER_RESTORE.
 *
 */
void
ar5211RxFilter(WLAN_DEV_INFO *pDev, A_UINT32 bits, HAL_RX_FILTER_OPER oper)
{
    A_UINT32 value;

    //printk("INSIDE ar5211RxFilter.....COMMAND IS: ");

    switch (oper) {

    case HAL_RX_FILTER_INIT:

        //printk("[HAL_RX_FILTER_INIT]\n");
        /* Init hardware register and software copy */

        pDev->rxFilterReg = bits;
        writePlatformReg(pDev, MAC_RX_FILTER, pDev->rxFilterReg);
        break;


    case HAL_RX_FILTER_SET:

        //printk("[HAL_RX_FILTER_SET]\n");
        /* Set frame type bits */

        value = readPlatformReg(pDev, MAC_RX_FILTER);
        ASSERT(pDev->rxFilterReg == value);

        pDev->rxFilterReg |= bits;
        writePlatformReg(pDev, MAC_RX_FILTER, pDev->rxFilterReg);
        break;

    case HAL_RX_FILTER_CLEAR:

        //printk("[HAL_RX_FILTER_CLEAR]\n");
        /* Clear frame type bits */

        value = readPlatformReg(pDev, MAC_RX_FILTER);
        ASSERT(pDev->rxFilterReg == value);

        pDev->rxFilterReg &= ~bits;
        writePlatformReg(pDev, MAC_RX_FILTER, pDev->rxFilterReg);
        break;

    case HAL_RX_FILTER_TEMP:

        //printk("[HAL_RX_FILTER_TEMP]\n");
        /* Temporarily set hardware register ONLY. */
        writePlatformReg(pDev, MAC_RX_FILTER, bits);
        break;

    case HAL_RX_FILTER_RESTORE:

        //printk("[HAL_RX_FILTER_RESTORE]\n");
        /* Restore hardware register with software copy */
        writePlatformReg(pDev, MAC_RX_FILTER, pDev->rxFilterReg);
        break;

    default:
        // Unsupported operation
        ASSERT(0);
    }
}

A_UINT16
ar5211GetRxDescRxAntenna(ATHEROS_DESC *pDesc)
{
    AR5211_RX_STATUS *pRxStatus= (AR5211_RX_STATUS *) &pDesc->hwStatus.rx;

    return (A_UINT16)pRxStatus->rxAntenna;
}

A_UINT16
ar5211GetRxDescRxTimeStamp(ATHEROS_DESC *pDesc)
{
    AR5211_RX_STATUS *pRxStatus= (AR5211_RX_STATUS *) &pDesc->hwStatus.rx;

    return (A_UINT16)pRxStatus->rxTimestamp;
}

A_UINT16
ar5211GetRxDescKeyIndex(ATHEROS_DESC *pDesc)
{
    AR5211_RX_STATUS *pRxStatus= (AR5211_RX_STATUS *) &pDesc->hwStatus.rx;

    return (A_UINT16)pRxStatus->keyIndex;
}

A_UINT16
ar5211GetRxDescKeyCacheMiss(ATHEROS_DESC *pDesc)
{
    AR5211_RX_STATUS *pRxStatus= (AR5211_RX_STATUS *) &pDesc->hwStatus.rx;

    return (A_UINT16)pRxStatus->keyCacheMiss;
}

/**************************************************************
 * ar5211EnableRxZlf
 *
 * Allow reception of zero-length frames from DMA
 */
void
ar5211EnableRxZlf(WLAN_DEV_INFO *pDev)
{
    writePlatformReg(pDev, MAC_RXCFG, readPlatformReg(pDev, MAC_RXCFG) | MAC_RXCFG_ZLFDMA);
}

#endif // #ifdef BUILD_AR5211
