/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * Chips specific device attachment and device info collection
 * Connects Init Reg Vectors, EEPROM Data, and device Functions to pDev
 */

#ifdef BUILD_AR5211

#ident "$Id: //depot/sw/releases/2.0.1/src/hal/ar5211/ar5211Reset.c#9 $"

/* Standard HAL Headers */
#include "wlantype.h"
#include "wlandrv.h"
#include "wlanchannel.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "halUtil.h"

/* Headers for HW private items */
#include "ar5hwcEnd.h"

#include "ar5211Reg.h"
#include "ar5211Misc.h"
#include "ar5211Reset.h"
#include "ar5211Power.h"
#include "ar5211Receive.h"
#include "ar5211.h"

#if defined(AR5311)
#include "ar5311reg.h"  /* ar5311 system definitions */
#endif

/* Add static register initialization vectors */
#include "boss.ini"

// Structure to hold 11b tuning information for Beanie/Sombrero
// 16 MHz mode, divider ratio = 198 = NP+S. N=16, S=4 or 6, P=12
typedef struct chan11b_info
{
    A_UINT32    index;          // index 1 thru 14
    A_UINT32    freq11bMhz;     // center frequency in MHz
    A_UINT32    freqBeanieMhz;  // Beanie frequency setting
    A_UINT32    refClkSel;      // reference clock select, 1 for 16 MHz
    A_UINT32    channelSelect;  // P[7:4]S[3:0] bits
    A_UINT32    sombFreqMhz;    // sombrero frequency MHz
    A_UINT16    sombChannel;    // 11a channel for Sombrero
} CHAN11B_INFO;

#define MAX_11B_CHANNELS    26  // 1 thru 26

CHAN11B_INFO    chan11bData[MAX_11B_CHANNELS] = {
    { 1,  2412, 3168, 1, 0x46, 5580, 116 },
    { 2,  2417, 3168, 1, 0x46, 5585, 117 },
    { 3,  2422, 3168, 1, 0x46, 5590, 118 },
    { 4,  2427, 3168, 1, 0x46, 5595, 119 },
    { 5,  2432, 3168, 1, 0x46, 5600, 120 },
    { 6,  2437, 3168, 1, 0x46, 5605, 121 },
    { 7,  2442, 3168, 1, 0x46, 5610, 122 },
    { 8,  2447, 3168, 1, 0x46, 5615, 123 },
    { 9,  2452, 3168, 1, 0x46, 5620, 124 },
    { 10, 2457, 3168, 1, 0x46, 5625, 125 },
    { 11, 2462, 3168, 1, 0x46, 5630, 126 },
    { 12, 2467, 3168, 1, 0x46, 5635, 127 },
    { 13, 2472, 3168, 1, 0x46, 5640, 128 },
    { 14, 2484, 3168, 1, 0x44, 5620, 125 },
    { 15, 2512, 3168, 1, 0x46, 5680, 136 },
    { 16, 2532, 3168, 1, 0x46, 5700, 140 },
    { 17, 2552, 3168, 1, 0x46, 5720, 144 },
    { 18, 2572, 3168, 1, 0x46, 5740, 148 },
    { 19, 2592, 3168, 1, 0x46, 5760, 152 },
    { 20, 2612, 3168, 1, 0x46, 5780, 156 },
    { 21, 2632, 3168, 1, 0x46, 5800, 160 },
    { 22, 2652, 3168, 1, 0x46, 5820, 164 },
    { 23, 2672, 3168, 1, 0x46, 5840, 168 },
    { 24, 2692, 3168, 1, 0x46, 5860, 172 },
    { 25, 2712, 3168, 1, 0x46, 5880, 176 },
    { 26, 2732, 3168, 1, 0x46, 5900, 180 }
};

// Power timeouts
#define POWER_UP_TIME           2000    // usecs to wait for chip to come out of sleep state

static void
ar5211SetTxRegs(WLAN_DEV_INFO *pdev, CHAN_VALUES *pChval);

static A_INT16
ar5211RunNoiseFloor(WLAN_DEV_INFO *pDev, A_UINT8 runTime, A_INT16 startingNF);

static A_BOOL
ar5211IsNfGood(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

static A_STATUS
ar5211CalNoiseFloor(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

static A_STATUS
ar5211SetResetReg(WLAN_DEV_INFO *pDev, A_UINT32 resetMask);

A_STATUS
ar5211SetChannel(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

static void
ar5211SetRf6and7(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo, CHAN_VALUES *pChval);

static void
ar5211SetBoardValues(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo, CHAN_VALUES *pChval);

static A_STATUS
ar5211SetTransmitPower(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

static void
ar5211SetPowerTable(WLAN_DEV_INFO *pDev, PCDACS_EEPROM *pSrcStruct, A_UINT16 channel);

static void
ar5211SetRateTable(WLAN_DEV_INFO *pDev, RD_EDGES_POWER *pRdEdgesPower,
                   TRGT_POWER_INFO *pPowerInfo, A_UINT16 numChannels,
                   CHAN_VALUES *pChval, A_UINT16 tpcScaleReduction,
                   struct eepMap *pData);

A_UINT16
ar5211GetScaledPower(A_UINT16 channel, A_UINT16 pcdacValue, PCDACS_EEPROM *pSrcStruct);

static A_BOOL
ar5211FindValueInList(A_UINT16 channel, A_UINT16 pcdacValue, PCDACS_EEPROM *pSrcStruct, A_UINT16 *powerValue);

A_UINT16
ar5211GetInterpolatedValue(A_UINT16 target, A_UINT16 srcLeft, A_UINT16 srcRight,
                           A_UINT16 targetLeft, A_UINT16 targetRight, A_BOOL scaleUp);

static void
ar5211GetLowerUpperValues(A_UINT16 value, A_UINT16 *pList, A_UINT16 listSize,
                          A_UINT16 *pLowerValue, A_UINT16 *pUpperValue);

static void
ar5211GetLowerUpperPcdacs(A_UINT16 pcdac, A_UINT16 channel, PCDACS_EEPROM *pSrcStruct,
                    A_UINT16 *pLowerPcdac, A_UINT16 *pUpperPcdac);

/**************************************************************
 * ar5211Reset
 *
 * Places the device in and out of reset and then places sane
 * values in the registers based on EEPROM config, initialization
 * vectors (as determined by the mode), and station configuration
 *
 * bChannelChange is used to preserve DMA/PCU registers across
 * a HW Reset during channel change.
 */
A_STATUS
ar5211Reset(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval, A_BOOL bChannelChange)
{
    A_UINT32        ulAddressLow, ulAddressHigh;
    A_UINT32        i;
    A_UINT32        reg, data, synthDelay;
    A_STATUS        status;
    A_BOOL          isAp;
    A_BOOL          is11b = IS_CHAN_CCK(pChval->channelFlags);
    A_UINT16        arrayEntries;
    A_UINT16        modesIndex = 0, freqIndex = 0;
    A_UINT32        saveFrameSeqCount[MAC_NUM_DCU];
    A_UINT32        saveTsfLow = 0, saveTsfHigh = 0;
    struct eepMap   *pEepData;
    A_UINT32	    saveDefAntenna;

    ASSERT(((pChval->channelFlags & CHANNEL_2GHZ) || 0) ^
        ((pChval->channelFlags & CHANNEL_5GHZ) || 0));
    ASSERT(((pChval->channelFlags & CHANNEL_OFDM) || 0) ^
        ((pChval->channelFlags & CHANNEL_CCK) || 0));

    pEepData = pDev->pHalInfo->pEepData;
    ASSERT(pEepData->version >= EEPROM_VER3);

    /* Preserve  certain DMA hardware registers on a channel change */
    if (bChannelChange) {
        /* AR5211 WAR 
         * Need to save/restore the TSF because of a hardware bug
         * that accelerates the TSF during a chip reset.
         *
         * We could use system timer routines to more
         * accurately restore the TSF, but 
         * 1. Timer routines on certain platforms are
         *    not accurate enough (e.g. 1 ms resolution).
         * 2. It would still not be accurate.
         *
         * The most important aspect of this workaround,
         * is that, after reset, the TSF is behind
         * other STAs TSFs (see bug 4127).  This will
         * allow the STA to properly resynchronize its
         * TSF in adhoc mode.
         */
        saveTsfLow  = readPlatformReg(pDev, MAC_TSF_L32);
        saveTsfHigh = readPlatformReg(pDev, MAC_TSF_U32);

        /* Read frame sequence count */
        if (pDev->macVersion >= MAC_SREV_VERSION_OAHU) {
            saveFrameSeqCount[0] = readPlatformReg(pDev, MAC_D0_SEQNUM);
        } else {
            for (i=0; i < MAC_NUM_DCU; i++) {
                saveFrameSeqCount[i] = readPlatformReg(pDev, MAC_D0_SEQNUM + (i*sizeof(A_UINT32)));
            }
        }
    }

    /*
     * Preserve the antenna on a channel change
     */
    saveDefAntenna = readPlatformReg(pDev, MAC_DEF_ANTENNA);
    if (saveDefAntenna == 0) {
        saveDefAntenna = 1;
    }

    status = ar5211ChipReset(pDev, pChval->channelFlags);

    // Initialize global Interrupt Enable Reference Count.  Assume interrupt was disabled.  
    pDev->pHalInfo->globIntRefCount=1;

    if (status != A_OK) {
        return status;
    }

    /* Setup the indices for the next set of register array writes */
    switch (pChval->channelFlags & CHANNEL_ALL) {
    case CHANNEL_A:
        modesIndex = 1;
        freqIndex = 1;
        break;
    case CHANNEL_T:
        modesIndex = 2;
        freqIndex = 1;
        break;
    case CHANNEL_B:
        modesIndex = 3;
        freqIndex = 2;
        break;
    case CHANNEL_PUREG:
        modesIndex = 4;
        freqIndex = 2;
        break;
    default:
        /* Ah, a new wireless mode */
        ASSERT(0);
        break;
    }

    /* Write parameters specific to AR5211 */
    if (pDev->macVersion >= MAC_SREV_VERSION_OAHU) {
        if (pChval->channelFlags & CHANNEL_2GHZ) {
            if (pDev->pHalInfo->pEepData->version >= EEPROM_VER3_1) {
                A_UINT32 ob2GHz, db2GHz;

                if (pChval->channelFlags & CHANNEL_CCK) {
                    ob2GHz = pEepData->pEepHeader->ob2GHz[0];
                    db2GHz = pEepData->pEepHeader->db2GHz[0];
                } else {
                    ob2GHz = pEepData->pEepHeader->ob2GHz[1];
                    db2GHz = pEepData->pEepHeader->db2GHz[1];
                }
                ob2GHz = reverseBits(ob2GHz, 3);
                db2GHz = reverseBits(db2GHz, 3);
                ar5211Mode2_4[25][freqIndex] = (ar5211Mode2_4[25][freqIndex] & ~0xC0) |
                    ((ob2GHz << 6) & 0xC0);
                ar5211Mode2_4[26][freqIndex] = (ar5211Mode2_4[26][freqIndex] & ~0x0F) |
                    ( ((ob2GHz >> 2) & 0x1) | ((db2GHz << 1) & 0x0E) );
            }
        }
        arrayEntries = sizeof(ar5211Mode2_4) / (sizeof(A_UINT32) * 3);
        for (i = 0; i < arrayEntries; i++) {
            writePlatformReg(pDev, ar5211Mode2_4[i][0], ar5211Mode2_4[i][freqIndex]);
        }
    }

    /* Write the analog registers 6 and 7 before other config */
    ar5211SetRf6and7(pDev, pEepData->pEepHeader, pChval);

    /* Write registers that vary across all modes */
    arrayEntries = sizeof(ar5211Modes) / (sizeof(A_UINT32) * 5);
    for (i = 0; i < arrayEntries; i++) {
        writePlatformReg(pDev, ar5211Modes[i][0], ar5211Modes[i][modesIndex]);
    }

    arrayEntries = sizeof(ar5211BB_RfGain) / (sizeof(A_UINT32) * 3);
    /* Write RFGain Parameters that differ between 2.4 and 5 GHz */
    for (i = 0; i < arrayEntries; i++) {
        writePlatformReg(pDev, ar5211BB_RfGain[i][0], ar5211BB_RfGain[i][freqIndex]);
    }

    arrayEntries = sizeof(ar5211Common) / (sizeof(A_UINT32) * 2);
    /* Write Common Array Parameters */
    for (i = 0; i < arrayEntries; i++) {

        reg = ar5211Common[i][0];

        /* On channel change, don't reset the PCU registers */
        if ( !(bChannelChange && (reg >= 0x8000 && reg < 0x9000)) ) {
            writePlatformReg(pDev, reg, ar5211Common[i][1]);
        }
    }

    /* Fix pre-AR5211 register values, this includes AR5311s. */
    if (pDev->macVersion < MAC_SREV_VERSION_OAHU) {
        /*
         * The TX and RX latency values have changed locations within the
         * USEC register in AR5211.  Since they're set via the .ini, for both
         * AR5211 and AR5311, they are written properly here for AR5311.
         */
        data = readPlatformReg(pDev, MAC_USEC);
        ASSERT((data & 0x00700000) == 0); // Must be 0 for proper write in AR5311
        writePlatformReg(pDev, MAC_USEC,
            (data & (MAC_USEC_M | MAC_USEC_32_M | MAC_5311_USEC_TX_LAT_M)) |
            ((29 << MAC_5311_USEC_RX_LAT_S) & MAC_5311_USEC_RX_LAT_M));

        /* The following registers exist only on AR5311. */
        writePlatformReg(pDev, MAC_5311_QDCLKGATE, 0);

        /* Set proper ADC & DAC delays for AR5311. */
        writePlatformReg(pDev, 0x00009878, 0x00000008);

        /* Enable the PCU FIFO corruption ECO on AR5311. */
        writePlatformReg(pDev, MAC_DIAG_SW,
                         readPlatformReg(pDev, MAC_DIAG_SW) | MAC_5311_DIAG_USE_ECO);
    }

    /* Restore certain DMA hardware registers on a channel change */
    if (bChannelChange) {
        /* AR5211 WAR - Restore TSF */
        writePlatformReg(pDev, MAC_TSF_L32, saveTsfLow);
        writePlatformReg(pDev, MAC_TSF_U32, saveTsfHigh);

        if (pDev->macVersion >= MAC_SREV_VERSION_OAHU) {
            writePlatformReg(pDev, MAC_D0_SEQNUM, saveFrameSeqCount[0]);
        } else {
            for (i = 0; i < MAC_NUM_DCU; i++) {
                writePlatformReg(pDev, MAC_D0_SEQNUM + (i * sizeof(A_UINT32)),
                                 saveFrameSeqCount[i]);
            }
        }
    }

    ulAddressLow  = cpu2le32(pDev->staConfig.macAddr.st.word),
    ulAddressHigh = cpu2le16(pDev->staConfig.macAddr.st.half);

    writePlatformReg(pDev, MAC_STA_ID0, ulAddressLow);

    isAp = A_MACADDR_COMP(&pDev->staConfig.macAddr, &pDev->staConfig.bssId) == 0;

    if (isAp) {
        writePlatformReg(pDev, MAC_STA_ID1, ulAddressHigh |
                         MAC_STA_ID1_STA_AP |
                         MAC_STA_ID1_RTS_USE_DEF);
    } else if (pDev->staConfig.bssType == INDEPENDENT_BSS) {
        /*
         * Set the IBSS mode only if we were configured to be so. If bssType is
         * ANY_BSS then the scan will determine what mode we finally end up in.
         */
        writePlatformReg(pDev, MAC_STA_ID1, ulAddressHigh |
                         MAC_STA_ID1_AD_HOC |
                         MAC_STA_ID1_RTS_USE_DEF);
    } else if (pDev->staConfig.bssType == INFRASTRUCTURE_BSS) { // Infrastructure Mode
        writePlatformReg(pDev, MAC_STA_ID1, ulAddressHigh |
                         MAC_STA_ID1_USE_DEFANT |
                         //MAC_STA_ID1_DEFANT_UPDATE |
			             0
		);
    } else {
        uiPrintf("ar5211Reset: Reset operation mode (AP/Adhoc/Infra) does not match a known value\n");
        return A_EINVAL;
    }

    /* Restore previous antenna */
    writePlatformReg(pDev, MAC_DEF_ANTENNA, saveDefAntenna);

    /* then our BSSID */
    ulAddressLow  = pDev->staConfig.bssId.st.word;
    ulAddressHigh = pDev->staConfig.bssId.st.half;

    writePlatformReg(pDev, MAC_BSS_ID0, cpu2le32(ulAddressLow));
    ulAddressHigh = 0x0000ffff & (cpu2le32(ulAddressHigh));
    writePlatformReg(pDev, MAC_BSS_ID1, ulAddressHigh);

    writePlatformReg(pDev, MAC_ISR, 0xFFFFFFFF); // cleared on write

    // Set CFG, little endian for register
#ifdef BIG_ENDIAN
    writePlatformReg(pDev, MAC_CFG, MAC_CFG_SWTD | MAC_CFG_SWRD);
#else
    writePlatformReg(pDev, MAC_CFG, INIT_CONFIG_STATUS);
#endif

    /* AR5211 WAR - pre-Production Oahu only.
     * Disable clock gating in all DMA blocks. This is a workaround for
     * a hardware problem seen with 11B and AES.
     * This workaround will result in higher power consumption.
     */
    if ((pDev->macVersion == MAC_SREV_VERSION_OAHU) && 
        (pDev->macRev < MAC_SREV_OAHU_PROD))
    {
        writePlatformReg(pDev, MAC_CFG,
                    readPlatformReg(pDev, MAC_CFG) | MAC_5211_CFG_CLK_GATE_DIS);
    }

    /* AR5211 CB22 WAR - suspected 32 MHz spur causes false detects in OFDM,
     * causing Tx Hangs.  Decrease weak signal sensitivity for this card.
     */
    if (pDev->pciInfo.SubVendorDeviceID == 0x1022 &&
        pChval->channelFlags & CHANNEL_OFDM)
    {
        writePlatformReg(pDev, 0x9924, 
                         (readPlatformReg(pDev, 0x9924) & 0xffffff01) | (8 << 1));
    }

    writePlatformReg(pDev, MAC_RSSI_THR, INIT_RSSI_THR);

    // Setup the transmit power values
    if (ar5211SetTransmitPower(pDev, pChval) != A_OK) {
        return A_ERROR;
    }

    // Setup board specific options for EEPROM vers 3
    ar5211SetBoardValues(pDev, pEepData->pEepHeader, pChval);

    status = ar5211SetChannel(pDev, pChval);
    if (status != A_OK) {
        return status;
    }

    // activate D2
    if ( pDev->pciInfo.DeviceID == DEV_AR5211_FPGA11B && (pChval->channelFlags & CHANNEL_2GHZ) ) {
        writePlatformReg(pDev, 0xd808, 0x502); // required for FPGA
    }
    writePlatformReg(pDev, PHY_ACTIVE, PHY_ACTIVE_EN);

    // Wait for the channel synth to settle (Synth goes on via PHY_ACTIVE_EN)
    // Read the phy active delay register. Value is in 100ns increments.
    data = readPlatformReg(pDev, PHY_RX_DELAY) & PHY_RX_DELAY_M;

    if (pChval->channelFlags & CHANNEL_CCK) {
        synthDelay = (4 * data) / 22;
    } else {
        synthDelay = data / 10;
    }

    // Delay 
    udelay(synthDelay);

    /*
     * TODO: this delay is here as sometimes (mostly in G mode) the synth
     * will not settle in just the 1ms above
     */
    udelay(2000);

    // calibrate it and poll the bit going to 0 for completion
    writePlatformReg(pDev, PHY_AGC_CONTROL,
                     readPlatformReg(pDev, PHY_AGC_CONTROL) | PHY_AGC_CONTROL_CAL);

    for (i = 0; i < 10000; i++) {
        if ( (readPlatformReg(pDev, PHY_AGC_CONTROL) &
              PHY_AGC_CONTROL_CAL) == 0 ) {
            break;
        }
        udelay(1000); // used to be 50
    }

    // Perform noise floor and set status
    status = ar5211CalNoiseFloor(pDev, pChval);
    if (status != A_OK) {
        pChval->channelFlags |= (is11b ? 0 : CHANNEL_CW_INT);
        return status;
    }

    // Start IQ calibration w/ 2^(INIT_IQCAL_LOG_COUNT_MAX+1) samples
    if (pDev->staConfig.calibrationTime != 0) {
        writePlatformReg(pDev, PHY_TIMING_CTRL4,
            PHY_TIMING_CTRL4_DO_IQCAL | (INIT_IQCAL_LOG_COUNT_MAX << PHY_TIMING_CTRL4_IQCAL_LOG_COUNT_MAX_S));
        pDev->bIQCalibration = TRUE;
    }

    ar5211SetTxRegs(pDev, pChval);

    /* Initialize active Q mask with only data enabled */
    pDev->activeQueueMask = 1 << pDev->dataQueueNum;

    // Setup QCU0 transmit interrupt masks (TX_ERR, TX_OK, TX_DESC, TX_URN)
    writePlatformReg(pDev, MAC_IMR_S0,
                     (MAC_IMR_S0_QCU_TXOK_M & MAC_QCU_0) |
                     (MAC_IMR_S0_QCU_TXDESC_M & (MAC_QCU_0<<MAC_IMR_S0_QCU_TXDESC_S)));
    writePlatformReg(pDev, MAC_IMR_S1, (MAC_IMR_S1_QCU_TXERR_M & MAC_QCU_0));

    writePlatformReg(pDev, MAC_IMR_S2, (MAC_IMR_S2_QCU_TXURN_M & MAC_QCU_0));

    if (pDev->macVersion < MAC_SREV_VERSION_OAHU) {
        // Configure DCU 0 (data queue) to use the global sequence count.
        // The beacon DCU will also use the global sequence count.
        writePlatformReg(pDev, MAC_D0_MISC,
                         readPlatformReg(pDev, MAC_D0_MISC) | MAC_5311_D_MISC_SEQ_NUM_CONTROL);
    }

    /* Set QCU to DCU Mask for active Q's */
    /* Set early termination for all Q's */
    for (i = 0; i < MAC_NUM_DCU; i++) {
        writePlatformReg(pDev, MAC_D0_QCUMASK + (i * sizeof(A_UINT32)), 1 << i);
        writePlatformReg(pDev, MAC_Q0_MISC + (i * sizeof(A_UINT32)), MAC_Q_MISC_DCU_EARLY_TERM_REQ);
    }
    // DPD Added 4/19 from mainline
    /* AR5211 WAR - GBL_EIFS must always be written after writing to any QCUMASK register */
    writePlatformReg(pDev, MAC_D_GBL_IFS_EIFS, readPlatformReg(pDev, MAC_D_GBL_IFS_EIFS));

    // Now set up the Interrupt Mask Register and save it for future use
    writePlatformReg(pDev, MAC_IMR, INIT_INTERRUPT_MASK);
    pDev->MaskReg = INIT_INTERRUPT_MASK;

    // Enable bus error interrupts
    writePlatformReg(pDev, MAC_IMR_S2, readPlatformReg(pDev, MAC_IMR_S2) |
                     MAC_IMR_S2_MCABT | MAC_IMR_S2_SSERR | MAC_IMR_S2_DPERR);

    // Enable interrupts specific to AP/client
    if (isAp) {
        // Not taking key cache miss or MIB interrupts on the client
        writePlatformReg(pDev, MAC_IMR, readPlatformReg(pDev, MAC_IMR) |
                         MAC_IMR_MIB);
        pDev->MaskReg |= MAC_IMR_MIB;
    }

    if (pEepData->rfKill) {
        ar5211EnableRfKill(pDev);
    }

    /* Restore the Rx Filter with the software copy.
     * NOTE: Do not reset the software copy of the Rx Filter.
     */
    ar5211RxFilter(pDev, 0, HAL_RX_FILTER_RESTORE);

    /*
     * Writing to MAC_BEACON will start timers. Hence it should be the last
     * register to be written. Do not reset tsf, do not enable beacons at this point,
     * but preserve other values like beaconInterval.
     */
    writePlatformReg(pDev, MAC_BEACON,
        (readPlatformReg(pDev, MAC_BEACON) & ~(MAC_BEACON_EN|MAC_BEACON_RESET_TSF)));


    return A_OK;
}

/**************************************************************
 * ar5211PhyDisable
 *
 * Places the PHY and Radio chips into reset.  A full reset
 * must be called to leave this state.  The PCI/MAC/PCU are
 * not placed into reset as we must receive interrupt to
 * re-enable the hardware.
 */
A_STATUS
ar5211PhyDisable(WLAN_DEV_INFO *pDev)
{
    return ar5211SetResetReg(pDev, MAC_RC_BB);
}

/**************************************************************
 * ar5211Disable
 *
 * Places all of hardware into reset
 */
A_STATUS
ar5211Disable(WLAN_DEV_INFO *pDev)
{
    A_STATUS status;

    status = ar5211SetPowerModeAwake(pDev, TRUE);
    if (status != A_OK) {
        return status;
    }

    // Reset the HW - PCI must be reset after the rest of the device has been reset
    status = ar5211SetResetReg(pDev, MAC_RC_MAC | MAC_RC_BB | MAC_RC_PCI);
    if (status != A_OK) {
        return status;
    }

    udelay(2100);       // 8245 @ 96Mhz hangs with 2000us.

    return status;
}

/**************************************************************
 * ar5211ResetDMA
 *
 * Places the hardware into reset and then pulls it out of reset
 */
A_STATUS
ar5211ResetDma(WLAN_DEV_INFO *pDev)
{
    uiPrintf("no resetDMA for AR5211!\n");
    ASSERT(0);

    return A_ERROR;
}

/**************************************************************
 * ar5211ChipReset
 *
 * Places the hardware into reset and then pulls it out of reset
 *
 * Only write the PLL if we're changing to or from CCK mode
 *
 * Attach calls with channelFlags = 0, as the coldreset should have
 * us in the correct mode and we cannot check the hwchannel flags.
 */
A_STATUS
ar5211ChipReset(WLAN_DEV_INFO *pDev, A_UINT16 channelFlags)
{
    A_STATUS status;

    status = ar5211SetPowerModeAwake(pDev, TRUE);
    if (status != A_OK) {
        return status;
    }


    /* Set CCK and Turbo modes correctly */
    if (channelFlags & CHANNEL_2GHZ) {
        if (channelFlags & CHANNEL_CCK ) {
            /* 2.4 GHZ CCK Mode */
            writePlatformReg(pDev, PHY_TURBO, 0);
            writePlatformReg(pDev, PHY_5211_MODE, PHY_5211_MODE_CCK | PHY_5211_MODE_RF2GHZ);
            writePlatformReg(pDev, PHY_PLL_CTL, PHY_PLL_CTL_44);
            /* Wait for the PLL to settle */
            udelay(1000);
        } else {
            /* 2.4 GHZ OFDM Mode */
            writePlatformReg(pDev, PHY_TURBO, 0);
            if (pDev->pciInfo.DeviceID == DEV_AR5212) {
                writePlatformReg(pDev, PHY_PLL_CTL, PHY_PLL_CTL_40);
                udelay(1000);
                writePlatformReg(pDev, PHY_5211_MODE, PHY_5211_MODE_OFDM | PHY_5211_MODE_RF2GHZ);
            }
        }
    } else if (channelFlags & CHANNEL_5GHZ) {
        if (channelFlags & CHANNEL_TURBO) {
            /* 5 GHZ OFDM Turbo Mode */
            writePlatformReg(pDev, PHY_TURBO, PHY_FC_TURBO_MODE | PHY_FC_TURBO_SHORT);
            if (pDev->pciInfo.DeviceID == DEV_AR5212) {
                writePlatformReg(pDev, PHY_PLL_CTL, PHY_PLL_CTL_40);
                udelay(1000);
                writePlatformReg(pDev, PHY_5211_MODE, PHY_5211_MODE_OFDM | PHY_5211_MODE_RF5GHZ);
            }
        } else {
            /* 5 GHZ OFDM Mode */
            writePlatformReg(pDev, PHY_TURBO, 0);
            if (pDev->pciInfo.DeviceID == DEV_AR5212) {
                writePlatformReg(pDev, PHY_PLL_CTL, PHY_PLL_CTL_40);
                udelay(1000);
                writePlatformReg(pDev, PHY_5211_MODE, PHY_5211_MODE_OFDM | PHY_5211_MODE_RF5GHZ);
            }
        }
    }
    /* else no flags set - must be attach calling - do nothing */

    // Reset the HW - PCI must be reset after the rest of the device has been reset
    status = ar5211SetResetReg(pDev, MAC_RC_MAC | MAC_RC_BB | MAC_RC_PCI);
    if (status != A_OK) {
        printk("RESET: FAILED TO SetResetReg\n");
        return status;
    }

    udelay(2100);       // 8245 @ 96Mhz hangs with 2000us.

    // Bring out of sleep mode (AGAIN)
    status = ar5211SetPowerModeAwake(pDev, TRUE);
    if (status != A_OK) {
        aPrintk("ar5211ChipReset: FAILED TO SETPOWERMODEAWAKE\n");
        return status;
    }

    // Clear warm reset reg
    status = ar5211SetResetReg(pDev, 0);

    return status;
}


/**************************************************************************
 * ar5211PerCalibration - Periodic calibration of PHY
 *
 * Recalibrate the lower PHY chips to account for temperature/environment
 * changes.
 */
A_STATUS
ar5211PerCalibration(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    A_BOOL   is11b  = IS_CHAN_CCK(pDev->staConfig.pChannel->channelFlags);
    A_STATUS status = A_OK;
    A_INT32  qCoff;
    A_INT32  qCoffDenom;
    A_UINT32 data;
    A_INT32  iqCorrMeas;
    A_INT32  iCoff;
    A_INT32  iCoffDenom;
    A_UINT32 powerMeasQ;
    A_UINT32 powerMeasI;

    // IQ calibration in progress => check to see if it has finished
    if (pDev->bIQCalibration &&
        !(readPlatformReg(pDev, PHY_TIMING_CTRL4) & PHY_TIMING_CTRL4_DO_IQCAL)) {
        // IQ Calibration has finished.
        pDev->bIQCalibration = FALSE;

        // Read calibration results
        powerMeasI = readPlatformReg(pDev, PHY_IQCAL_RES_PWR_MEAS_I);
        powerMeasQ = readPlatformReg(pDev, PHY_IQCAL_RES_PWR_MEAS_Q);
        iqCorrMeas = readPlatformReg(pDev, PHY_IQCAL_RES_IQ_CORR_MEAS);

        /*
         * Prescale these values to remove 64-bit operation requirement at the loss
         * of a little precision.
         */
        iCoffDenom = (powerMeasI / 2 + powerMeasQ / 2) / 128;
        qCoffDenom = powerMeasQ / 64;

        // Protect against div by 0
        if ( (iCoffDenom != 0) && (qCoffDenom != 0) ) {
            iCoff = (-iqCorrMeas) / iCoffDenom;
            // IQCORR_Q_I_COFF is a signed 6 bit number
            iCoff = iCoff & 0x3f;

            qCoff = ((A_INT32)powerMeasI / qCoffDenom) - 64;
            // IQCORR_Q_Q_COFF is a signed 5 bit number
            qCoff = qCoff & 0x1f;

#ifdef CALIBRATION_DEBUG
            uiPrintf("powerMeasI = 0x%08x\n", powerMeasI);
            uiPrintf("powerMeasQ = 0x%08x\n", powerMeasQ);
            uiPrintf("iqCorrMeas = 0x%08x\n", iqCorrMeas);
            uiPrintf("iCoff      = %d\n", iCoff);
            uiPrintf("qCoff      = %d\n", qCoff);
#endif

            // Write IQ
            data  = readPlatformReg(pDev, PHY_TIMING_CTRL4) |
                PHY_TIMING_CTRL4_IQCORR_ENABLE |
                (((A_UINT32)iCoff) << PHY_TIMING_CTRL4_IQCORR_Q_I_COFF_S) |
                ((A_UINT32)qCoff);
            writePlatformReg(pDev, PHY_TIMING_CTRL4, data);
        }
    }

    // Perform noise floor and set status
    if (pDev->macVersion >= MAC_SREV_VERSION_OAHU) {
        if (ar5211IsNfGood(pDev, pChval) == FALSE) {
            pChval->channelFlags |= (is11b ? 0 : CHANNEL_CW_INT);
            return A_EBADCHANNEL;
        }
    }

    status = ar5211CalNoiseFloor(pDev, pChval);
    if (status != A_OK) {
        // Delay 5ms before retrying the noise floor - just to make sure
        // We're in an error condition here
        udelay(5000);
        status = ar5211CalNoiseFloor(pDev, pChval);
    }

    if (status == A_EBADCHANNEL && !is11b) {
        pChval->channelFlags |= CHANNEL_CW_INT;
    }

    return status;
}

/**************************************************************
 * ar5211SetResetReg
 *
 * Writes the given reset bit mask into the reset register
 */
A_STATUS
ar5211SetResetReg(WLAN_DEV_INFO *pDev, A_UINT32 resetMask)
{
#if defined(PCI_INTERFACE)
    A_UINT32    mask = resetMask ? resetMask : ~0;
    int         i;

    /* flush any pending MMR writes */
    readPlatformReg(pDev, MAC_RXDP);

    writePlatformReg(pDev, MAC_RC, resetMask);

    /* need to wait at least 128 clocks when reseting PCI before read */
    udelay(2100); // used to be 15

    resetMask  &= MAC_RC_MAC | MAC_RC_BB;
    mask &= MAC_RC_MAC | MAC_RC_BB;

    for (i = 0; i < 10000; i++) {
        if ((readPlatformReg(pDev, MAC_RC) & mask) == resetMask) {
            return A_OK;
        }
        udelay(2100); // used to be 50
    }

    aPrintk("RESETTTTTT: ar5211SetResetReg: resetHW failed set %x != %lx\n", 
             (unsigned long)resetMask, (unsigned long)readPlatformReg(pDev, MAC_RC));

    return A_HARDWARE;
#elif defined(AR5311)
    A_UINT32 reg = sysRegRead(AR5311_RESET);    /* read before */

    if (resetMask == MAC_RC_BB) {
        /* Put baseband in reset */
        reg |= RESET_WARM_WBB;
        sysRegWrite(AR5311_RESET, reg);
        (void)sysRegRead(AR5311_RESET);         /* read after */
        return A_OK;
    }

    if (resetMask) {
        /*
         * Reset the MAC and baseband.  This is a bit different than
         * the PCI version, but holding in reset causes problems.
         */
        reg &= ~(RESET_WLAN | RESET_WARM_WLAN | RESET_WARM_WBB);
        reg |= RESET_WLAN | RESET_WARM_WLAN;
        sysRegWrite(AR5311_RESET, reg);
        (void)sysRegRead(AR5311_RESET);         /* read after */
        udelay(1000); // used to be 100
    }

    /* Bring MAC and baseband out of reset */
    reg &= ~(RESET_WLAN | RESET_WARM_WLAN | RESET_WARM_WBB);
    (void)sysRegRead(AR5311_RESET);             /* read before */
    sysRegWrite(AR5311_RESET, reg);
    (void)sysRegRead(AR5311_RESET);             /* read after */

    return A_OK;
#endif
}

/**************************************************************
 * ar5211SetChannel
 *
 * Takes the MHz channel value and sets the Channel value
 *
 * ASSUMES: Writes enabled to analog bus before AGC is active
 *   or by disabling the AGC.
 */
A_STATUS
ar5211SetChannel(WLAN_DEV_INFO *pDev,  CHAN_VALUES *pChval)
{
    A_UINT32    refClk, data, tmpData;
    A_UINT16    lowChannel, highChannel;
    A_UINT32    beanieData = 0;
    A_UINT16    channelIEEE;

    A_UINT16      chanb;

    if (pChval->channelFlags & CHANNEL_2GHZ) {
        // 2.4 GHZ spectrum. Compute beanie value and pass it on for Sombrero programming
        ar5211GetChannelEdges(pDev, pChval->channelFlags, &lowChannel, &highChannel);
        if ((pChval->channel < lowChannel) || (pChval->channel > highChannel))
        {
            uiPrintf("ar5211SetChannel: Invalid Channel Number %4d MHz\n", pChval->channel);
            return A_EINVAL;
        }
        chanb = wlanConvertGHztoCh(pChval->channel, CHANNEL_2GHZ); // beanie channel
        channelIEEE = chan11bData[chanb-1].sombChannel; // sombrero channel
        tmpData = reverseBits(chan11bData[chanb-1].channelSelect, 8) & 0xff;
        beanieData = (tmpData << 5) | (chan11bData[chanb-1].refClkSel << 4);
    } else {
        ar5211GetChannelEdges(pDev, pChval->channelFlags, &lowChannel, &highChannel);
        if ((pChval->channel < lowChannel) || (pChval->channel > highChannel))
        {
            uiPrintf("ar5211SetChannel: Invalid Channel Number %4d MHz\n", pChval->channel);
            return A_EINVAL;
        }
        channelIEEE = MHZ2IEEE_CH(pChval->channel);
    }

    // Rest of the code is common for 5 GHz and 2.4 GHz.

    if (channelIEEE >= 145 || (channelIEEE & 0x1)) {
        tmpData = reverseBits((channelIEEE - 24), 8);
        refClk = 1;
    } else {
        tmpData = reverseBits(((channelIEEE - 24)/2), 8);
        refClk = 0;
    }

    data = (tmpData << 2) | (refClk << 1) | (1 << 10) | 0x1;
    writePlatformReg(pDev, PHY_BASE+(0x27<<2),
                        (((beanieData & 0xff) << 8) | (data & 0xff)));
    data = (data >> 8) & 0xff;
    writePlatformReg(pDev, PHY_BASE+(0x34<<2),
                     ((beanieData & 0x0000ff00) | data));

    pDev->staConfig.phwChannel = pChval;
    return A_OK;
}

/**************************************************************************
 * Peform the noisefloor calibration for the length of time set
 * in runTime (valid values 1 to 7)
 *
 * Returns: The NF value at the end of the given time (or 0 for failure)
 */
A_INT16
ar5211RunNoiseFloor(WLAN_DEV_INFO *pDev, A_UINT8 runTime, A_INT16 startingNF)
{
    A_INT16 nf, i;
    int searchTime;

    ASSERT(runTime <= 7);

    /* Setup  noise floor run time and starting value */
    writePlatformReg(pDev, PHY_BASE+(25<<2), (readPlatformReg(pDev, PHY_BASE+(25<<2)) & ~0xFFF) |
        ((runTime << 9) & 0xE00) | (startingNF & 0x1FF));

    /* Calibrate the noise floor */
    writePlatformReg(pDev, PHY_AGC_CONTROL, readPlatformReg(pDev, PHY_AGC_CONTROL) |
        PHY_AGC_CONTROL_NF);

    /* Compute the required amount of searchTime needed to finish NF */
    if (runTime == 0) {
        /* 8 search windows * 6.4us each */
        searchTime = 8  * 7;
    } else {
        /* 512 * runtime search windows * 6.4us each */
        searchTime = (runTime * 512)  * 7;
    }

    /* Do not read noise floor until it has been updated */
    /* 
     *  As a guesstimate - we may only get 1/60th the time on the air to see search windows
     *  in a heavily congested network (40 us every 2400 us of time)
     */
    for (i = 0; ((i < 60) && (readPlatformReg(pDev, PHY_AGC_CONTROL) & PHY_AGC_CONTROL_NF)); i++) {
        udelay(searchTime);
    }
    if (i >= 60) {
#ifdef DEBUG
        uiPrintf("NF with runTime %d failed to end on channel %d\n", runTime, pDev->staConfig.phwChannel->channel);
        uiPrintf("  PHY NF Reg state:     0x%lx\n", (unsigned long)readPlatformReg(pDev, PHY_AGC_CONTROL));
        uiPrintf("  PHY Active Reg state: 0x%lx\n", (unsigned long)readPlatformReg(pDev, PHY_ACTIVE));
#endif
        return 0;
    }
    nf = (A_INT16)(readPlatformReg(pDev, PHY_BASE+(25<<2)) >> 19) & 0x1ff;
    if (nf & 0x100) {
        nf = 0 - ((nf ^ 0x1ff) + 1);
    }
    return nf;
}

/**************************************************************************
 * Read the NF and check it against the noise floor threshhold
 *
 * Returns: TRUE if the NF is good
 */
A_BOOL
ar5211IsNfGood(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    A_INT16 nf, nfThresh;
    int      arrayMode = 0;

    switch (pChval->channelFlags & (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ)) {
    case (CHANNEL_OFDM | CHANNEL_5GHZ):
        arrayMode = 0;
        break;
    case (CHANNEL_CCK | CHANNEL_2GHZ):
        arrayMode = 1;
        break;
    case (CHANNEL_OFDM | CHANNEL_2GHZ):
        arrayMode = 2;
        break;
    default:
        ASSERT(0);
        break;
    }
    if (readPlatformReg(pDev, PHY_AGC_CONTROL) & PHY_AGC_CONTROL_NF) {
        uiPrintf("NF failed to complete in calibration window\n");
    }
    nfThresh = pDev->pHalInfo->pEepData->pEepHeader->noiseFloorThresh[arrayMode];
    nf = (A_INT16)(readPlatformReg(pDev, PHY_BASE+(25<<2)) >> 19) & 0x1ff;
    if (nf & 0x100) {
        nf = 0 - ((nf ^ 0x1ff) + 1);
    }
    if (nf > nfThresh) {
        uiPrintf("Noise floor failed detected %d higher than thresh %d\n", nf, nfThresh);
        return FALSE;
    }
    return TRUE;
}

/**************************************************************************
 * ar5211CalNoiseFloor
 *
 * Peform the noisefloor calibration and check for any constant channel
 * interference.
 *
 * NOTE: preAR5211 have a lengthy carrier wave detection process - hence
 * it is if'ed for MKK regulatory domain only.
 *
 * Returns: TRUE for a successful noise floor calibration; else FALSE
 */
A_STATUS
ar5211CalNoiseFloor(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    A_INT16 nf, arrayMode = 0;
    A_INT16 nfThresh;

    /* Check for Carrier Wave interference in MKK regulatory zone */
    if ((pDev->macVersion < MAC_SREV_VERSION_OAHU) &&
        wlanIsNfCheckRequired(pDev))
    {
        switch (pChval->channelFlags & (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ)) {
        case (CHANNEL_OFDM | CHANNEL_5GHZ):
            arrayMode = 0;
            break;
        case (CHANNEL_CCK | CHANNEL_2GHZ):
            arrayMode = 1;
            break;
        case (CHANNEL_OFDM | CHANNEL_2GHZ):
            arrayMode = 2;
            break;
        default:
            ASSERT(0);
            break;
        }
        nfThresh = pDev->pHalInfo->pEepData->pEepHeader->noiseFloorThresh[arrayMode];

        /* Run a quick noise floor that will hopefully complete (decrease delay time) */
        nf = ar5211RunNoiseFloor(pDev, 0, 0);
        if (nf > nfThresh) {
            uiPrintf("First run failed with value %d higher than thresh %d\n", nf, nfThresh);
            nf = ar5211RunNoiseFloor(pDev, 2, nf);
            if (nf > nfThresh) {
                uiPrintf("Second run failed with value %d higher than thresh %d\n", nf, nfThresh);
                nf = ar5211RunNoiseFloor(pDev, 7, nf);
                if (nf > nfThresh) {
                    uiPrintf("Full run failed with value %d higher than thresh %d\n", nf, nfThresh);
                    return A_HARDWARE;
                }
            }
        }
    } else {
        /* Calibrate the noise floor */
        writePlatformReg(pDev, PHY_AGC_CONTROL, readPlatformReg(pDev, PHY_AGC_CONTROL) |
            PHY_AGC_CONTROL_NF);
    }
    return A_OK;
}



/**************************************************************************
 * ar5211SetTxRegs
 *
 * Set the retry, aifs and cwmin regs for each data queue
 *
 * TODO: do this per enabled Data Queue.  Requires adding Q structs and QOS.
 */
void
ar5211SetTxRegs(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    A_UINT32 retryReg, cwminVal, dcuId;
    A_UINT32 hwRetryVal;

    /*
     * Make sure that the CWmin is of the form (2^n - 1); and then
     * set the retry limits
     */
    if (!pDev->staConfig.QosEnabled) {

        if ((pChval->channelFlags & CHANNEL_B) == CHANNEL_B) {
            pDev->staConfig.cwmin = INIT_CWMIN_11B;
        } else {
            pDev->staConfig.cwmin = INIT_CWMIN;
        }
    }
    cwminVal = 1;
    while (cwminVal < pDev->staConfig.cwmin) {
        cwminVal = (cwminVal << 1) | 1;
    }

    retryReg = ((INIT_SSH_RETRY << MAC_D_RETRY_LIMIT_STA_SH_S) & MAC_D_RETRY_LIMIT_STA_SH_M) |
        ((INIT_SLG_RETRY << MAC_D_RETRY_LIMIT_STA_LG_S) & MAC_D_RETRY_LIMIT_STA_LG_M);

    if (pDev->staConfig.swretryEnabled) {
#ifdef BUILD_AP  /* TODO - AP/STA should sync up on how to set this... */
        hwRetryVal = pDev->staConfig.limitTxRetries / 2;
#else
        hwRetryVal = pDev->staConfig.limitTxRetries;
#endif
        if (hwRetryVal > MAC_D_RETRY_LIMIT_FR_SH_M) {
            hwRetryVal = MAC_D_RETRY_LIMIT_FR_SH_M;
        }
        retryReg |= ((hwRetryVal << MAC_D_RETRY_LIMIT_FR_LG_S) & MAC_D_RETRY_LIMIT_FR_LG_M) |
            (hwRetryVal & MAC_D_RETRY_LIMIT_FR_SH_M);
    } else {
        retryReg |= ((INIT_LG_RETRY << MAC_D_RETRY_LIMIT_FR_LG_S) & MAC_D_RETRY_LIMIT_FR_LG_M) |
            (INIT_SH_RETRY & MAC_D_RETRY_LIMIT_FR_SH_M);
    }

#ifndef BUILD_AP
    if (!pDev->staConfig.QosEnabled) {
        pDev->staConfig.aifs = 2;
    }
#endif /* BUILD_AP */

    dcuId = pDev->dataQueueNum;
    writePlatformReg(pDev, MAC_D0_LCL_IFS + (dcuId * 4),
        (cwminVal & MAC_D_LCL_IFS_CWMIN_M) |
        ((INIT_CWMAX << MAC_D_LCL_IFS_CWMAX_S) & MAC_D_LCL_IFS_CWMAX_M) |
        ((pDev->staConfig.aifs << MAC_D_LCL_IFS_AIFS_S) & MAC_D_LCL_IFS_AIFS_M));

    /* Set Retries for each data queue */
    writePlatformReg(pDev, MAC_D0_RETRY_LIMIT + (dcuId * 4), retryReg);
}

/**************************************************************
 * ar5211ar5211GetCornerCalIndex
 *
 * Returns the index for the Corner Calibration static mapping
 * of channels.  5GHz only.
 */
A_UINT16
ar5211GetCornerCalIndex(A_UINT16 channel) {
    ASSERT(channel > 3000);
    
    if (channel <= 5250) {
        return 0;
    } else if (channel <= 5480) {
        return 1;
    } else if (channel <= 5732) {
        return 2;
    } else {
        return 3;
    }
}

/**************************************************************
 * ar5211SetRf6and7
 *
 * Reads EEPROM header info from device structure and programs
 * analog registers 6 and 7
 *
 * REQUIRES: Access to the analog device
 */
static void
ar5211SetRf6and7(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo, CHAN_VALUES *pChval)
{
    A_UINT16    rfXpdGain, rfPloSel, rfPwdXpd;
    A_UINT16    tempOB     = 0;
    A_UINT16    tempDB     = 0;
    A_UINT16    freqIndex;
    A_BOOL      arrayMode = 0;
    int         arrayEntries, i;

    freqIndex = (pChval->channelFlags & CHANNEL_2GHZ) ? 2 : 1;

    /* 
     * TODO: This array mode correspondes with the index used during the read
     * For readability, this should be changed to an enum or #define
     */
    switch (pChval->channelFlags &
        (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ))
    {
    case (CHANNEL_OFDM | CHANNEL_5GHZ):
        if ((pChval->channel > 4000) && (pChval->channel < 5260)) {
            tempOB = pHeaderInfo->ob1;
            tempDB = pHeaderInfo->db1;
        } else if ((pChval->channel >= 5260) && (pChval->channel < 5500)) {
            tempOB = pHeaderInfo->ob2;
            tempDB = pHeaderInfo->db2;
        } else if ((pChval->channel >= 5500) && (pChval->channel < 5725)) {
            tempOB = pHeaderInfo->ob3;
            tempDB = pHeaderInfo->db3;
        }
        else if (pChval->channel >= 5725) {
            tempOB = pHeaderInfo->ob4;
            tempDB = pHeaderInfo->db4;
        }
        arrayMode = 0;
        if (pDev->pHalInfo->pEepData->version >= EEPROM_VER3_2) {
            A_UINT32 rfPwd84, rfPwd90, rfGainSel, cornerIndex;

            cornerIndex = ar5211GetCornerCalIndex(pChval->channel);
            rfPwd84 = pHeaderInfo->cornerCal[cornerIndex].pd84;
            rfPwd90 = pHeaderInfo->cornerCal[cornerIndex].pd90;
            rfGainSel = pHeaderInfo->cornerCal[cornerIndex].gSel;
            ar5211Rf6n7[5][freqIndex] =  (ar5211Rf6n7[5][freqIndex] & ~0x10000000) |
                (rfPwd84 << 28);
            ar5211Rf6n7[6][freqIndex] =  (ar5211Rf6n7[6][freqIndex] & ~0x04000000) |
                (rfPwd90 << 26);
            ar5211Rf6n7[21][freqIndex] =  (ar5211Rf6n7[21][freqIndex] & ~0x08) |
                (rfGainSel << 3);
        }
        break;
    case (CHANNEL_CCK | CHANNEL_2GHZ):
        tempOB = pHeaderInfo->obFor24;
        tempDB = pHeaderInfo->dbFor24;
        arrayMode = 1;
        break;
    case (CHANNEL_OFDM | CHANNEL_2GHZ):
        tempOB = pHeaderInfo->obFor24g;
        tempDB = pHeaderInfo->dbFor24g;
        arrayMode = 2;
        break;
    default:
        ASSERT(0);
        break;
    }

    rfXpdGain = pHeaderInfo->xgain[arrayMode];
    rfPloSel = pHeaderInfo->xpd[arrayMode];
    rfPwdXpd = !pHeaderInfo->xpd[arrayMode];

    if ( (tempOB < 1) || (tempOB > 5) ) {
        uiPrintf("ar5211SetBoardValues: OB should be between 1 and 5 inclusively, not %d", tempOB);
    }
    if ( (tempDB < 1) || (tempDB > 5) ) {
        uiPrintf("ar5211SetBoardValues: DB should be between 1 and 5 inclusively, not %d", tempDB);
    }

    /* Set rfXpdGain and rfPwdXpd */
    ar5211Rf6n7[11][freqIndex] =  (ar5211Rf6n7[11][freqIndex] & ~0xC0) |
        (((reverseBits(rfXpdGain, 4) << 7) | (rfPwdXpd << 6)) & 0xC0);
    ar5211Rf6n7[12][freqIndex] =  (ar5211Rf6n7[12][freqIndex] & ~0x07) |
        ((reverseBits(rfXpdGain, 4) >> 1) & 0x07);

    /* Set OB */
    ar5211Rf6n7[12][freqIndex] =  (ar5211Rf6n7[12][freqIndex] & ~0x80) |
        ((reverseBits(tempOB, 3) << 7) & 0x80);
    ar5211Rf6n7[13][freqIndex] =  (ar5211Rf6n7[13][freqIndex] & ~0x03) |
        ((reverseBits(tempOB, 3) >> 1) & 0x03);

    /* Set DB */
    ar5211Rf6n7[13][freqIndex] =  (ar5211Rf6n7[13][freqIndex] & ~0x1C) |
        ((reverseBits(tempDB, 3) << 2) & 0x1C);

    /* Set rfPloSel */
    ar5211Rf6n7[17][freqIndex] =  (ar5211Rf6n7[17][freqIndex] & ~0x08) |
        ((rfPloSel << 3) & 0x08);

    /* Write the Rf registers 6 & 7 */
    arrayEntries = sizeof(ar5211Rf6n7) / (sizeof(A_UINT32) * 3);
    for (i = 0; i < arrayEntries; i++) {
        writePlatformReg(pDev, ar5211Rf6n7[i][0], ar5211Rf6n7[i][freqIndex]);
    }

}


/**************************************************************
 * ar5211SetBoardValues
 *
 * Reads EEPROM header info and programs the device for correct operation
 * given the channel value
 */
void
ar5211SetBoardValues(WLAN_DEV_INFO *pDev, EEP_HEADER_INFO *pHeaderInfo, CHAN_VALUES *pChval)
{
    int      arrayMode = 0;

    switch (pChval->channelFlags &
        (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ))
    {
    case (CHANNEL_OFDM | CHANNEL_5GHZ):
        arrayMode = 0;
        if (pDev->pHalInfo->pEepData->version >= EEPROM_VER3_2) {
            A_UINT32 phyTxClip, cornerIndex;

            cornerIndex = ar5211GetCornerCalIndex(pChval->channel);
            phyTxClip = pHeaderInfo->cornerCal[cornerIndex].clip;
            writePlatformReg(pDev, PHY_BASE + (81 << 2),
                (readPlatformReg(pDev, PHY_BASE + (81 << 2)) & 0xFFFFFFC7) |
                (phyTxClip << 3));
        }
        break;
    case (CHANNEL_CCK | CHANNEL_2GHZ):
        arrayMode = 1;
        break;
    case (CHANNEL_OFDM | CHANNEL_2GHZ):
        arrayMode = 2;
        break;
    default:
        ASSERT(0);
        break;
    }

    /* Set the antenna register(s) correctly for the chip revision */
    if (pDev->macVersion < MAC_SREV_VERSION_OAHU) {
        writePlatformReg(pDev, PHY_BASE + (68 << 2),
            (readPlatformReg(pDev, PHY_BASE + (68 << 2)) & 0xFFFFFFFC) | 0x3);
    } else {
        writePlatformReg(pDev, PHY_BASE + (68 << 2),
            (readPlatformReg(pDev, PHY_BASE + (68 << 2)) & 0xFFFFFC06) |
            (pHeaderInfo->antennaControl[0][arrayMode] << 4) | 0x1);
        writePlatformReg(pDev, PHY_BASE + (88 << 2), pHeaderInfo->antennaControl[1][arrayMode] |
            (pHeaderInfo->antennaControl[2][arrayMode] << 6) |
            (pHeaderInfo->antennaControl[3][arrayMode] << 12) |
            (pHeaderInfo->antennaControl[4][arrayMode] << 18) |
            (pHeaderInfo->antennaControl[5][arrayMode] << 24));
        writePlatformReg(pDev, PHY_BASE + (89 << 2), pHeaderInfo->antennaControl[6][arrayMode] |
            (pHeaderInfo->antennaControl[7][arrayMode] << 6) |
            (pHeaderInfo->antennaControl[8][arrayMode] << 12) |
            (pHeaderInfo->antennaControl[9][arrayMode] << 18) |
            (pHeaderInfo->antennaControl[10][arrayMode] << 24));

        /* Set the Noise Floor Thresh on ar5211 devices */
        writePlatformReg(pDev, PHY_BASE + (90 << 2), (pHeaderInfo->noiseFloorThresh[arrayMode] & 0x1FF) |
            (1 << 9));
    }
    writePlatformReg(pDev, PHY_BASE + (17 << 2),
        (readPlatformReg(pDev, PHY_BASE + (17 << 2)) & 0xFFFFC07F) |
        ((pHeaderInfo->switchSettling[arrayMode] << 7) & 0x3F80));
    writePlatformReg(pDev, PHY_BASE + (18 << 2),
        (readPlatformReg(pDev, PHY_BASE + (18 << 2)) & 0xFFFC0FFF) |
        ((pHeaderInfo->txrxAtten[arrayMode] << 12) & 0x3F000));
    writePlatformReg(pDev, PHY_BASE + (20 << 2),
        (readPlatformReg(pDev, PHY_BASE + (20 << 2)) & 0xFFFF0000) |
        ((pHeaderInfo->pgaDesiredSize[arrayMode] << 8) & 0xFF00) |
        (pHeaderInfo->adcDesiredSize[arrayMode] & 0x00FF));
    writePlatformReg(pDev, PHY_BASE + (13 << 2),
        (pHeaderInfo->txEndToXPAOff[arrayMode] << 24) |
        (pHeaderInfo->txEndToXPAOff[arrayMode] << 16) |
        (pHeaderInfo->txFrameToXPAOn[arrayMode] << 8) |
        pHeaderInfo->txFrameToXPAOn[arrayMode]);
    writePlatformReg(pDev, PHY_BASE + (10 << 2),
        (readPlatformReg(pDev, PHY_BASE + (10 << 2)) & 0xFFFF00FF) |
        (pHeaderInfo->txEndToXLNAOn[arrayMode] << 8));
    writePlatformReg(pDev, PHY_BASE + (25 << 2),
        (readPlatformReg(pDev, PHY_BASE + (25 << 2)) & 0xFFF80FFF) |
        ((pHeaderInfo->thresh62[arrayMode] << 12) & 0x7F000));
}

static const A_UINT16 tpcScaleReductionTable[5] = {0, 3, 6, 9, 63};
/**************************************************************
 * ar5211SetTransmitPower
 *
 * Sets the transmit power in the baseband for the given
 * operating channel and mode.
 */
A_STATUS
ar5211SetTransmitPower(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    A_STATUS        status = A_OK;
    A_UINT16        numChannels = 0;
    TRGT_POWER_INFO *pPowerInfo = NULL;
    RD_EDGES_POWER  *pRdEdgePower = NULL;
    PCDACS_EEPROM   eepromPcdacs;
    struct eepMap   *pData;
    A_UINT8         cfgCtl;
    int             i;

    ASSERT(pDev->pHalInfo && pDev->pHalInfo->pEepData);

    pData = pDev->pHalInfo->pEepData;

    /* setup the pcdac struct to point to the correct info, based on mode */
    switch (pChval->channelFlags &
        (CHANNEL_OFDM | CHANNEL_CCK | CHANNEL_5GHZ | CHANNEL_2GHZ))
    {
    case (CHANNEL_OFDM | CHANNEL_5GHZ):
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels11a;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels11a;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11a;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11a;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11a;
        break;

    case (CHANNEL_OFDM | CHANNEL_2GHZ):
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels2_4;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels2_4;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11g;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11g;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11g;
        break;

    case (CHANNEL_CCK | CHANNEL_2GHZ):
        eepromPcdacs.numChannels     = pData->pPcdacInfo->numChannels2_4;
        eepromPcdacs.pChannelList    = pData->pPcdacInfo->Channels2_4;
        eepromPcdacs.pDataPerChannel = pData->pPcdacInfo->DataPerChannel11b;
        numChannels = pData->pTrgtPowerInfo->numTargetPwr_11b;
        pPowerInfo  = pData->pTrgtPowerInfo->trgtPwr_11b;
        break;
    default:
        ASSERT(0);
        break;
    }

    ar5211SetPowerTable(pDev, &eepromPcdacs, pChval->channel);

    /* Match CTL to EEPROM value */
    cfgCtl = wlanGetCtl(pDev, pChval);

    for (i = 0; i < NUM_CTLS; i++) {
        if (pData->pEepHeader->testGroups[i] == 0) {
            continue;
        } else if (pData->pEepHeader->testGroups[i] == cfgCtl) {
            pRdEdgePower = &(pData->pRdEdgesPower[i * NUM_EDGES]);
            break;
        }
    }

    ar5211SetRateTable(pDev, pRdEdgePower,
        pPowerInfo, numChannels, pChval,
        tpcScaleReductionTable[pDev->staConfig.tpScale],
        pData);

    return status;
}

/**************************************************************
 * ar5211SetPowerTable
 *
 * Read the transmit power levels from the structures taken from EEPROM
 * Interpolate read transmit power values for this channel
 * Organize the transmit power values into a table for writing into the hardware
 */
void
ar5211SetPowerTable(WLAN_DEV_INFO *pDev, PCDACS_EEPROM *pSrcStruct, A_UINT16 channel)
{
    static FULL_PCDAC_STRUCT pcdacStruct;
    static A_UINT16          pcdacTable[PWR_TABLE_SIZE];
    A_UINT16     i, j;
    A_UINT16     *pPcdacValues;
    A_INT16      *pScaledUpDbm;
    A_INT16      minScaledPwr;
    A_INT16      maxScaledPwr;
    A_INT16      pwr;
    A_UINT16     pcdacMin = 0;
    A_UINT16     pcdacMax = 63;
    A_UINT16     pcdacTableIndex;
    A_UINT16     scaledPcdac;
    A_UINT32     addr;
    A_UINT32     temp32;

    A_MEM_ZERO(&pcdacStruct, sizeof(FULL_PCDAC_STRUCT));
    A_MEM_ZERO(pcdacTable, sizeof(A_UINT16) * PWR_TABLE_SIZE);
    pPcdacValues = pcdacStruct.PcdacValues;
    pScaledUpDbm = pcdacStruct.PwrValues;

    /* Initialize the pcdacs to dBM structs pcdacs to be 1 to 63 */
    for (i = PCDAC_START, j = 0; i <= PCDAC_STOP; i+= PCDAC_STEP, j++) {
        pPcdacValues[j] = i;
    }

    pcdacStruct.numPcdacValues = j;
    pcdacStruct.pcdacMin = PCDAC_START;
    pcdacStruct.pcdacMax = PCDAC_STOP;

    /* Fill out the power values for this channel */
    for (j = 0; j < pcdacStruct.numPcdacValues; j++ ) {
        pScaledUpDbm[j] = ar5211GetScaledPower(channel, pPcdacValues[j], pSrcStruct);
    }

    /* Now scale the pcdac values to fit in the 64 entry power table */
    minScaledPwr = pScaledUpDbm[0];
    maxScaledPwr = pScaledUpDbm[pcdacStruct.numPcdacValues - 1];

    /* find minimum and make monotonic */
    for (j = 0; j < pcdacStruct.numPcdacValues; j++) {
        if (minScaledPwr >= pScaledUpDbm[j]) {
            minScaledPwr = pScaledUpDbm[j];
            pcdacMin = j;
        }
        /*
         * Make the full_hsh monotonically increasing otherwise interpolation algorithm will get fooled
         * gotta start working from the top, hence i = 63 - j.
         */
        i = (A_UINT16)(pcdacStruct.numPcdacValues - 1 - j);
        if (i == 0) {
            break;
        }
        if (pScaledUpDbm[i-1] > pScaledUpDbm[i]) {
            //it could be a glitch, so make the power for this pcdac,
            //the same as the power from the next highest pcdac
            pScaledUpDbm[i - 1] = pScaledUpDbm[i];
        }
    }

    for (j = 0; j < pcdacStruct.numPcdacValues; j++) {
        if (maxScaledPwr < pScaledUpDbm[j]) {
            maxScaledPwr = pScaledUpDbm[j];
            pcdacMax = j;
        }
    }

    /* Find the first power level with a pcdac */
    pwr = (A_UINT16)(PWR_STEP * ((minScaledPwr - PWR_MIN + PWR_STEP / 2) / PWR_STEP)  + PWR_MIN);

    /* Write all the first pcdac entries based off the pcdacMin */
    pcdacTableIndex = 0;
    for (i = 0; i < (2 * (pwr - PWR_MIN) / EEP_SCALE + 1); i++) {
        pcdacTable[pcdacTableIndex++] = pcdacMin;
    }

    i = 0;
    while (pwr < pScaledUpDbm[pcdacStruct.numPcdacValues - 1]) {
        pwr += PWR_STEP;
        /* stop if dbM > max_power_possible */
        while ((pwr < pScaledUpDbm[pcdacStruct.numPcdacValues - 1]) &&
               ((pwr - pScaledUpDbm[i])*(pwr - pScaledUpDbm[i+1]) > 0))
        {
            i++;
        }
        /* scale by 2 and add 1 to enable round up or down as needed */
        scaledPcdac = (A_UINT16)(ar5211GetInterpolatedValue(pwr, pScaledUpDbm[i],
            pScaledUpDbm[i+1], (A_UINT16)(pPcdacValues[i]*2),
            (A_UINT16)(pPcdacValues[i+1]*2), 0 ) + 1);

        pcdacTable[pcdacTableIndex] = scaledPcdac / 2;
        if (pcdacTable[pcdacTableIndex] > pcdacMax) {
            pcdacTable[pcdacTableIndex] = pcdacMax;
        }
        pcdacTableIndex++;
    }

    /* Write all the last pcdac entries based off the last valid pcdac */
    while (pcdacTableIndex < PWR_TABLE_SIZE) {
        pcdacTable[pcdacTableIndex] = pcdacTable[pcdacTableIndex - 1];
        pcdacTableIndex++;
    }

    /* Finally, write the power values into the baseband power table */
    addr = PHY_BASE + (608 << 2);
    for (i = 0; i < 32; i++) {
        temp32 = 0xffff & ((pcdacTable[2 * i + 1] << 8) | 0xff);
        temp32 = (temp32 << 16) | (0xffff & ((pcdacTable[2 * i] << 8) | 0xff));
        writePlatformReg(pDev, addr, temp32);
        addr += 4;
    }

}

/**************************************************************
 * ar5211SetRateTable
 *
 * Sets the transmit power in the baseband for the given
 * operating channel and mode.
 */
void
ar5211SetRateTable(WLAN_DEV_INFO *pDev, RD_EDGES_POWER *pRdEdgesPower,
                   TRGT_POWER_INFO *pPowerInfo, A_UINT16 numChannels,
                   CHAN_VALUES *pChval, A_UINT16 tpcScaleReduction,
                   struct eepMap *pData)
{
    static A_UINT16 ratesArray[NUM_RATES];
    A_UINT16    *pRatesPower;
    A_UINT16    lowerChannel, lowerIndex=0, lowerPower=0;
    A_UINT16    upperChannel, upperIndex=0, upperPower=0;
    A_UINT16    twiceMaxEdgePower=63;
    A_UINT16    twicePower = 0;
    A_UINT16    i;
    A_UINT16    tempChannelList[NUM_EDGES]; /* temp array for holding edge channels */
    A_UINT16    twiceMaxRDPower;
    A_INT16     scaledPower = 0;        /* for gcc -O2 */
    A_UINT16    mask = 0x3f;
    A_UINT32    regOffset;
    A_UINT32    reg32;
    A_BOOL      paPreDEnable = 0;
    A_INT8      twiceAntennaGain, twiceAntennaReduction = 0;

    pRatesPower = ratesArray;
    twiceMaxRDPower = wlanGetChannelPower(pDev, pChval) * 2;

    if (pChval->channelFlags & CHANNEL_5GHZ) {
        twiceAntennaGain = pData->pEepHeader->antennaGainMax[0];
    } else {
        twiceAntennaGain = pData->pEepHeader->antennaGainMax[1];
    }

    twiceAntennaReduction = wlanGetAntennaReduction(pDev, pChval, twiceAntennaGain);

    if (pRdEdgesPower) {
        /* Get the edge power */
        for (i = 0; i < NUM_EDGES; i++) {
            tempChannelList[i] = pRdEdgesPower[i].rdEdge;
        }

        ar5211GetLowerUpperValues(pChval->channel, tempChannelList, NUM_EDGES, &lowerChannel, &upperChannel);

        if (lowerChannel == upperChannel) {
            /* Get the index for this channel */
            for (i = 0; i < NUM_EDGES; i++) {
                if (lowerChannel == tempChannelList[i]) {
                    break;
                }
            }
            twiceMaxEdgePower = pRdEdgesPower[i].twice_rdEdgePower;
        }
    }

    /* extrapolate the power values for the test Groups */
    for (i = 0; i < numChannels; i++) {
        tempChannelList[i] = pPowerInfo[i].testChannel;
    }

    ar5211GetLowerUpperValues(pChval->channel, tempChannelList, numChannels, &lowerChannel, &upperChannel);

    /* get the index for the channel */
    for (i = 0; i < numChannels; i++) {
        if (lowerChannel == tempChannelList[i]) {
            lowerIndex = i;
        }

        if (upperChannel == tempChannelList[i]) {
            upperIndex = i;
            break;
        }
    }

    for (i = 0; i < NUM_RATES; i++) {
        if (pChval->channelFlags & CHANNEL_OFDM) {
            /* power for rates 6,9,12,18,24 is all the same */
            if (i < 5) {
                lowerPower = pPowerInfo[lowerIndex].twicePwr6_24;
                upperPower = pPowerInfo[upperIndex].twicePwr6_24;
            } else if (i == 5) {
                lowerPower = pPowerInfo[lowerIndex].twicePwr36;
                upperPower = pPowerInfo[upperIndex].twicePwr36;
            } else if (i == 6) {
                lowerPower = pPowerInfo[lowerIndex].twicePwr48;
                upperPower = pPowerInfo[upperIndex].twicePwr48;
            } else if (i == 7) {
                lowerPower = pPowerInfo[lowerIndex].twicePwr54;
                upperPower = pPowerInfo[upperIndex].twicePwr54;
            }
        } else {
            switch (i) {
            case 0:
            case 1:
                lowerPower = pPowerInfo[lowerIndex].twicePwr6_24;
                upperPower = pPowerInfo[upperIndex].twicePwr6_24;
                break;

            case 2:
            case 3:
                lowerPower = pPowerInfo[lowerIndex].twicePwr36;
                upperPower = pPowerInfo[upperIndex].twicePwr36;
                break;

            case 4:
            case 5:
                lowerPower = pPowerInfo[lowerIndex].twicePwr48;
                upperPower = pPowerInfo[upperIndex].twicePwr48;
                break;

            case 6:
            case 7:
                lowerPower = pPowerInfo[lowerIndex].twicePwr54;
                upperPower = pPowerInfo[upperIndex].twicePwr54;
                break;
            }

        }

        twicePower = (A_UINT16)ar5211GetInterpolatedValue(pChval->channel, lowerChannel,
            upperChannel, lowerPower, upperPower, 0);

        /* Reduce power by band edge restrictions */
        twicePower = A_MIN(twicePower, twiceMaxEdgePower); 

        if (pChval->channelFlags & CHANNEL_TURBO) {
            /* If turbo is set, reduce power to keep power consumption under 2 Watts */
            if (pData->version >= EEPROM_VER3_1) {
                twicePower = A_MIN(twicePower, pData->pEepHeader->turbo2WMaxPower);
            }
        }

        /* Reduce power by max regulatory domain allowed restrictions */
        pRatesPower[i] = A_MIN(twicePower, twiceMaxRDPower);

        /* Use 6 Mb power level for transmit power scaling reduction */
        /* We don't want to reduce higher rates if its not needed */
        if (i == 0) {
            scaledPower = (pRatesPower[0] - (tpcScaleReduction * 2)) - twiceAntennaReduction;
            if (scaledPower < 1) {
                scaledPower = 1;
            }
            /* Record txPower at Rate 6 for info gathering */
            pDev->tx6PowerInDbm = scaledPower / 2;
        }

        pRatesPower[i] = A_MIN(pRatesPower[i], scaledPower);
    }

#if 0
    uiPrintf("Final Output Power Setting for channel %d MHz\n", pChval->channel);
    uiPrintf("Final Power 6 Mb %d dBm, MaxRD: %d dBm, MaxEdge %d dBm, - TPC Scale %d dBm - Ant Red %d dBm\n",
        scaledPower / 2, twiceMaxRDPower / 2, twiceMaxEdgePower / 2, tpcScaleReduction, twiceAntennaReduction / 2);
    uiPrintf("Max Turbo %d dBm if turbo and version greater than 3001 : 0x%04X\n", pData->pEepHeader->turbo2WMaxPower,
        pData->version);
    for (i = 0; i < NUM_RATES; i++) {
        uiPrintf("  %d dBm |", pRatesPower[i] / 2);
    }
    uiPrintf("\n");
#endif

    /* Write the power table into the hardware */
    regOffset = 0x9934;
    reg32 = ( ((paPreDEnable & 1)<< 30) | ((pRatesPower[3] & mask) << 24) |
         ((paPreDEnable & 1)<< 22) | ((pRatesPower[2] & mask) << 16) |
         ((paPreDEnable & 1)<< 14) | ((pRatesPower[1] & mask) << 8) |
         ((paPreDEnable & 1)<< 6 ) |  (pRatesPower[0] & mask)  );
    writePlatformReg(pDev, regOffset, reg32);

    regOffset = 0x9938;
    reg32 = ( ((paPreDEnable & 1)<< 30) | ((pRatesPower[7] & mask) << 24) |
         ((paPreDEnable & 1)<< 22) | ((pRatesPower[6] & mask) << 16) |
         ((paPreDEnable & 1)<< 14) | ((pRatesPower[5] & mask) << 8) |
         ((paPreDEnable & 1)<< 6 ) |  (pRatesPower[4] & mask)  );
    writePlatformReg(pDev, regOffset, reg32);

    regOffset = 0x993c;

    /* set max power to the power value at rate 6 */
    writePlatformReg(pDev, regOffset, pRatesPower[0]);
}

/**************************************************************
 * ar5211GetScaledPower
 *
 * Gets or interpolates the pcdac value from the calibrated data
 */
A_UINT16
ar5211GetScaledPower(A_UINT16 channel, A_UINT16 pcdacValue, PCDACS_EEPROM *pSrcStruct)
{
    A_UINT16    powerValue;
    A_UINT16    lFreq, rFreq;           /* left and right frequency values */
    A_UINT16    llPcdac, ulPcdac;       /* lower and upper left pcdac values */
    A_UINT16    lrPcdac, urPcdac;       /* lower and upper right pcdac values */
    A_UINT16    lPwr, uPwr;             /* lower and upper temp pwr values */
    A_UINT16    lScaledPwr, rScaledPwr; /* left and right scaled power */

    if (ar5211FindValueInList(channel, pcdacValue, pSrcStruct, &powerValue)) {
        /* value was copied from srcStruct */
        return powerValue;
    }

    ar5211GetLowerUpperValues(channel, pSrcStruct->pChannelList, pSrcStruct->numChannels,
        &lFreq, &rFreq);
    ar5211GetLowerUpperPcdacs(pcdacValue, lFreq, pSrcStruct, &llPcdac, &ulPcdac);
    ar5211GetLowerUpperPcdacs(pcdacValue, rFreq, pSrcStruct, &lrPcdac, &urPcdac);

    /* get the power index for the pcdac value */
    ar5211FindValueInList(lFreq, llPcdac, pSrcStruct, &lPwr);
    ar5211FindValueInList(lFreq, ulPcdac, pSrcStruct, &uPwr);
    lScaledPwr = ar5211GetInterpolatedValue(pcdacValue, llPcdac, ulPcdac, lPwr, uPwr, 0);

    ar5211FindValueInList(rFreq, lrPcdac, pSrcStruct, &lPwr);
    ar5211FindValueInList(rFreq, urPcdac, pSrcStruct, &uPwr);
    rScaledPwr = ar5211GetInterpolatedValue(pcdacValue, lrPcdac, urPcdac, lPwr, uPwr, 0);

    return ar5211GetInterpolatedValue(channel, lFreq, rFreq, (A_UINT16)lScaledPwr, (A_UINT16)rScaledPwr, 0);
}

/**************************************************************
 * ar5211FindValueInList
 *
 * Find the value from the calibrated source data struct
 */
A_BOOL
ar5211FindValueInList(A_UINT16 channel, A_UINT16 pcdacValue, PCDACS_EEPROM *pSrcStruct, A_UINT16 *powerValue)
{
    DATA_PER_CHANNEL    *pChannelData;
    A_UINT16            *pPcdac;
    A_UINT16            i, j;

    pChannelData = pSrcStruct->pDataPerChannel;
    for (i = 0; i < pSrcStruct->numChannels; i++ ) {
        if (pChannelData->channelValue == channel) {
            pPcdac = pChannelData->PcdacValues;
            for (j = 0; j < pChannelData->numPcdacValues; j++ ) {
                if (*pPcdac == pcdacValue) {
                    *powerValue = pChannelData->PwrValues[j];
                    return TRUE;
                }
                pPcdac++;
            }
        }
        pChannelData++;
    }
    return FALSE;
}

/**************************************************************
 * ar5211GetInterpolatedValue
 *
 * Returns interpolated or the scaled up interpolated value
 */
A_UINT16
ar5211GetInterpolatedValue(A_UINT16 target, A_UINT16 srcLeft, A_UINT16 srcRight,
                           A_UINT16 targetLeft, A_UINT16 targetRight, A_BOOL scaleUp)
{
    A_UINT16 returnValue;
    A_UINT16 lRatio;
    A_UINT16 scaleValue = EEP_SCALE;

    /* to get an accurate ratio, always scale, if want to scale, then don't scale back down */
    if ((targetLeft * targetRight) == 0) {
        return 0;
    }

    if (scaleUp) {
        scaleValue = 1;
    }

    if (abs(srcRight - srcLeft) > 0) {
        /* note the ratio always need to be scaled, since it will be a fraction */
        lRatio = (A_UINT16)((target - srcLeft) * EEP_SCALE / (srcRight - srcLeft));
        returnValue = (A_UINT16)((lRatio * targetRight + (EEP_SCALE - lRatio) * targetLeft) / scaleValue);
    } else {
        returnValue = (A_UINT16)(targetLeft*(scaleUp ? EEP_SCALE : 1));
    }

    return returnValue;
}

/**************************************************************
 * ar5211GetLowerUpperValues
 *
 *  Look for value being within 0.1 of the search values
 *  however, NDIS can't do float calculations, so multiply everything
 *  up by EEP_SCALE so can do integer arithmatic
 *
 * INPUT  value       -value to search for
 * INPUT  pList       -ptr to the list to search
 * INPUT  listSize    -number of entries in list
 * OUTPUT pLowerValue -return the lower value
 * OUTPUT pUpperValue -return the upper value
 */
void
ar5211GetLowerUpperValues(A_UINT16 value, A_UINT16 *pList, A_UINT16 listSize,
                          A_UINT16 *pLowerValue, A_UINT16 *pUpperValue)
{
    A_UINT16    i;
    A_UINT16    listEndValue = *(pList + listSize - 1);
    A_UINT32    target       = value * EEP_SCALE;

    /*
     * See if value is lower than the first value in the list
     * if so return first value
     */
    if (target < (A_UINT32)(*pList * EEP_SCALE - EEP_DELTA)) {
        *pLowerValue = *pList;
        *pUpperValue = *pList;
        return;
    }

    /*
     * See if value is greater than last value in list
     * if so return last value
     */
    if (target > (A_UINT32)(listEndValue * EEP_SCALE + EEP_DELTA)) {
        *pLowerValue = listEndValue;
        *pUpperValue = listEndValue;
        return;
    }

    /* look for value being near or between 2 values in list */
    for (i = 0; i < listSize; i++) {
        /*
         * If value is close to the current value of the list
         * then target is not between values, it is one of the values
         */
        if (abs(pList[i] * EEP_SCALE - target) < EEP_DELTA) {
            *pLowerValue = pList[i];
            *pUpperValue = pList[i];
            return;
        }

        /*
         * Look for value being between current value and next value
         * if so return these 2 values
         */
        if (target < (A_UINT32)(pList[i + 1] * EEP_SCALE - EEP_DELTA)) {
            *pLowerValue = pList[i];
            *pUpperValue = pList[i + 1];
            return;
        }
    }
}

/**************************************************************
 * ar5211GetLowerUpperPcdacs
 *
 * Get the upper and lower pcdac given the channel and the pcdac
 * used in the search
 */
void
ar5211GetLowerUpperPcdacs(A_UINT16 pcdac, A_UINT16 channel, PCDACS_EEPROM *pSrcStruct,
                    A_UINT16 *pLowerPcdac, A_UINT16 *pUpperPcdac)
{
    DATA_PER_CHANNEL    *pChannelData;
    A_UINT16            i;

    /* Find the channel information */
    pChannelData = pSrcStruct->pDataPerChannel;
    for (i = 0; i < pSrcStruct->numChannels; i++) {
        if (pChannelData->channelValue == channel) {
            break;
        }
        pChannelData++;
    }

    ar5211GetLowerUpperValues(pcdac, pChannelData->PcdacValues, pChannelData->numPcdacValues,
                        pLowerPcdac, pUpperPcdac);
}

#endif // #ifdef BUILD_AR5211
