/* ar5311.h - ar5311 reference board header */

/* Copyright 1984-1999 Wind River Systems, Inc. */

/*
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * This file contains I/O addresses and related constants for the
 * ar5311 BSP.
 */

#ifndef	__ar5311_h
#define	__ar5311_h

#ident "ACI $Header: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/ar5311.h#1 $"

#define BUS		NONE		/* no off-board bus interface */
#define N_SIO_CHANNELS	2		/* 1 console, 1 high speed */

/* timer constants */
#define	SYS_CLK_RATE_MIN  	30	/* minimum system clock rate */
#define	SYS_CLK_RATE_MAX  	1000	/* maximum system clock rate */
#define	AUX_CLK_RATE_MIN  	30	/* minimum auxiliary clock rate */
#define	AUX_CLK_RATE_MAX  	5000	/* maximum auxiliary clock rate */

/* memory configuration */
#define LOCAL_MEM_LOCAL_ADRS	K0BASE	/* K0, but DMA ignores upper bits */

/*
 * Miscellaneous definitions go here. For example, macro definitions
 * for various devices.
 */

#undef	INCLUDE_EGL		/* configAll.h defines this for MIPS */

/* Directly connected MIPS interrupts -- usually starts @ 33 + 1 */
#define IV_WLAN_VEC		(IV_SWTRAP1_VEC + 1)
#define IV_ENET0_VEC		(IV_SWTRAP1_VEC + 2)
#define IV_ENET1_VEC		(IV_SWTRAP1_VEC + 3)
#define IV_MISCDMA_VEC		(IV_SWTRAP1_VEC + 4)
#define IV_MISCIO_VEC		(IV_SWTRAP1_VEC + 5)
#define IV_COMPARE_VEC		(IV_SWTRAP1_VEC + 6)

/* Misc IO vectored interrupts */
#define IV_TIMER_VEC		(IV_SWTRAP1_VEC + 7)
#define IV_AHBPROC_VEC		(IV_SWTRAP1_VEC + 8)
#define IV_AHBDMA_VEC		(IV_SWTRAP1_VEC + 9)
#define IV_GPIO_VEC		(IV_SWTRAP1_VEC + 10)
#define IV_UART1_VEC		(IV_SWTRAP1_VEC + 11)
#define IV_UART0_VEC		(IV_SWTRAP1_VEC + 12)
#define IV_WD_VEC		(IV_SWTRAP1_VEC + 13)

#define AR5311_SR		(SR_CU0 | SR_IE |      \
				 AR5311_INTR_MISCIO |  \
				 AR5311_INTR_COMPARE)

/* "NVRAM" in flash. */
#define NVRAM_INITIALIZED	"AR5311"
#define NVRAM_INIT_SIZE		(sizeof(NVRAM_INITIALIZED)-1)
#define NVRAM_BOOT_OFFSET	NVRAM_INIT_SIZE
#define NVRAM_SIZE		0x100		/* limited my memory map */

#undef INCLUDE_VME		/* no VME bus */
#undef INCLUDE_PCI		/* no PCI bus */
#undef INCLUDE_PCI_AUTOCONF	/* no PCI bus */
#undef INCLUDE_MMU_BASIC	/* no MMU support */
#undef INCLUDE_MMU_FULL		/* no MMU support */
#undef INCLUDE_MMU		/* no MMU support */
#undef INCLUDE_HW_FP		/* no FPU */

#undef INCLUDE_SCSI		/* no SCSI or related disk support */
#undef INCLUDE_SCSI_DMA
#undef INCLUDE_SCSI_BOOT
#undef INCLUDE_DOSFS
#undef INCLUDE_TAPEFS
#undef INCLUDE_CDROMFS
#undef SYS_SCSI_CONFIG
#undef INCLUDE_SCSI2

#define INCLUDE_SYS_HW_INIT_0
#define SYS_HW_INIT_0()	sysHwInit0()

#ifdef _ASMLANGUAGE

/*
 * Assembly helpers
 */
#define LEAF(name)	\
	.globl name;	\
	.ent name;	\
name:

#define XLEAF(name)	\
	.globl name;	\
name:

#define END(name)	\
	.end	name

#define ASM_PUTS(str)			\
	.data;				\
99:;					\
	.asciz str;			\
	.text;				\
	la	a0, 99b;		\
	jal	asm_puts

#define ROM_PUTS(str)			\
	b	99f;			\
98:;					\
	.asciz str;			\
	.align 2;			\
99:;					\
	la	a0, 98b;		\
	sub	a0, RAM_LOW_ADRS;	\
	or	a0, 0xbfc00000;		\
	bal	rom_puts

#else

/*
 * BSP specific prototypes
 */
unsigned int sysConfig1Get(void);
void sysWbFlush(void);

/* spirit performance counter helpers */
void sysPerfStart(int);                 /* start counters with mode */
#define SYSPERF_DC      0               /* watch data cache */
#define SYSPERF_IC      1               /* watch instruction cache */
#define SYSPERF_WB      2               /* watch write buffer */
#define SYSPERF_CH      3               /* watch cache hits */
#define SYSPERF_CM      4               /* watch cache misses */
void sysPerfUpdate(void);               /* sample + accumulate counters */
void sysPerfShow(void);                 /* display results */
void sysPerfStop(void);                 /* stop counters */

void asm_puts(char *);			/* low level console output */
void asm_puthex(UINT32);

void sysReset(void);                    /* hardware reset of the system */
void sysHwInit0(void);			/* very early init C code */

void sysUDelay(int);			/* micro second delay */
void sysMsDelay(int);			/* milli second delay */
void sysDelay(void);			/* delay to allow PIO to finish */

void sysToFactory(void);                /* set-up for factory defaults */

STATUS sysLanEnetAddrGet(int, char *);	/* get network MAC address */

int sysTffsMount(void);			/* mount flash filesystem */

void sysGpioCtrlOutput(int gpio);       /* set gpio for output */
void sysGpioCtrlInput(int gpio);        /* set gpio for input */
void sysGpioCtrlIntr(int gpio);         /* set gpio for interrupt */
void sysGpioConnect(int gpio, FUNCPTR cb, UINT32 arg);
void sysGpioSet(int gpio, int val);     /* set output gpio to val */
int  sysGpioGet(int gpio);              /* get input gpio val */

void cacheDataInvalidate(void *p, int len);
void cacheDescInvalidate(void *p);

void sysLedBlink(void);
void sysLed(int on);

/* hardware watchdog support */
#define SYSWATCHDOG
void sysWatchdogPeriodSet(int ticks);
void sysWatchdogUpdate(void);
void sysWatchdogEnable(void);
void sysWatchdogDisable(void);

/* configuration in flash support */
void sysFlashConfigInit(void *ptr, INT32, INT32, INT32);
void sysFlashConfigErase(int fcl);
UINT8 sysFlashConfigRead(int fcl , int offset);
void sysFlashConfigWrite(int fcl, int offset, UINT8 *data, int len);
int  sysRomSize(void);
#define NFLC		3		/* number of flash config segments */
#define FLC_BOOTLINE	0
#define FLC_BOARDDATA	1
#define FLC_RADIOCFG	2

void sysCacheEnable(void);              /* LATE_CACHE_ENABLE */

/* lost from dosFsLib.h in Tornado 2.1 */
STATUS dosFsMkfsOptionsSet(UINT options);

#define DELTA(a,b)	(((UINT32)(a)) - ((UINT32)(b)))

/*
 * Board configuration data to allow software to determine
 * configuration at run time.  Stored in a read only flash
 * configuration block.
 */
struct ar5311_boarddata {
    UINT32 magic;			/* board data is valid */
#define BD_MAGIC	0x35333131	/* "5311" */
    UINT16 cksum;                       /* checksum (starting with BD_REV 2) */
    UINT16 rev;                         /* revision of this struct */
#define BD_REV	2
    char   boardName[64];		/* Name of board */
    UINT16 major;			/* Board major number */
    UINT16 minor;			/* Board minor number */
    UINT32 config;			/* Board configuration */
#define BD_ENET0	0x00000001	/* ENET0 is stuffed */
#define BD_ENET1	0x00000002	/* ENET1 is stuffed */
#define BD_UART1	0x00000004	/* UART1 is stuffed */
#define BD_UART0	0x00000008	/* UART0 is stuffed (dma) */
#define BD_RSTFACTORY	0x00000010	/* Reset factory defaults stuffed */
#define BD_SYSLED	0x00000020	/* System LED stuffed */
#define BD_EXTUARTCLK	0x00000040	/* External UART clock */
#define BD_CPUFREQ	0x00000080	/* cpu freq is valid in nvram */
#define BD_SYSFREQ	0x00000100	/* sys freq is set in nvram */
#define BD_WLAN		0x00000200	/* WLAN exists (for emulation) */
#define BD_MEMCAP       0x00000400      /* CAP SDRAM @ memCap for testing */
#define BD_DISWATCHDOG  0x00000800      /* disable system watchdog */
#ifdef EMULATION
#define BD_BRINGUP
#endif
    UINT16 resetConfigGpio;	        /* Reset factory GPIO pin */
    UINT16 sysLedGpio;		        /* System LED GPIO pin */

    UINT32 cpuFreq;		        /* CPU core frequency in Hz */
    UINT32 sysFreq;		        /* System frequency in Hz */
    UINT32 cntFreq;		        /* Calculated C0_COUNT frequency */

    UINT8  wlanMac[6];
    UINT8  enet0Mac[6];
    UINT8  enet1Mac[6];

    UINT16 pciId;			/* Pseudo PCIID for common code */
    UINT16 memCap;                      /* cap bank1 in MB */
};
extern struct ar5311_boarddata sysBoardData;

int bdChange(void), bdShow(struct ar5311_boarddata *);
UINT16 sysBoardDataChecksum(struct ar5311_boarddata *bd);       /* for MDK */

#ifdef EMULATION
extern int emulation;
#endif

#ifdef INCLUDE_TELNET
/* need a place to insert prototypes for prjConfig.c */
STATUS cliTelnetHook(int, UINT32, int *, int *);
#endif

int loattach();				/* vxWorks missing prototype */

#endif

#endif	/* __ar5311_h */
