/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * ar5hwcEnd.h -- Atheros VxWorks WLAN Driver header file
 *
 * $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/atheros/drv/ar5hwcEnd.h#1 $
 */

#ifndef __AR5HWC_H__
#define __AR5HWC_H__

#include <asm/io.h>
#include <asm/byteorder.h>
/*
#ifdef __BIG_ENDIAN
#ifndef BIG_ENDIAN
#define BIG_ENDIAN __BIG_ENDIAN
#endif 
#endif
*/
#include <linux/delay.h>

#include "assert.h"
//#include "config.h"         /* BSP definitions */
#include "wlandrv.h"        /* WLAN_DEV_INFO */

#define AR_DEV_NAME        "ar"
#define AR_DRV_NAME        "Atheros wlan"

/* constants capturing the buffering requirements of the wlan
 * driver; the receive queue depth is generally determined by
 * the scheduling granularity of my rx interrupt processing;
 * it is also a tradeoff between the depth of the bursts we
 * want to tolerate, and the desirable max latency through a
 * box (assuming a bridging device - the kind we are); recv
 * buffers are taken off our queue and loaned to the transmit
 * side of (other) devices - thus requiring us to allocate
 * loaned buffers; the depth of the tx queue at a minimum
 * should be max num of buffers the other side's rx processing
 * is willing to loan outside at any given time - ideally one
 * wants to throttle a side's rx processing based on lack of
 * buffers i.e. when max buffers have crossed over to the other
 * side - thereby acheiveing the earliest time drop of buffers
 * (in NIC hw without any CPU time having being spent on them)
 * when falling behind in processing; again tx queue depth must
 * not be allowed to grow so large as to compromise latency goals
 * through the box; above explained is the ideal scenario which
 * converges to rx depth on one side = tx depth on other side,
 * and max loaned buffers count of rx queue depth; our tx path
 * absorbs from our rx (as well as SME) processing also..
 * moreover, the current ethernet may at times need a new tx
 * to release any tx completed buffers which might be stuck
 * there - transmit completion processing happens in the context
 * of transmit itself!! consequently, to keep our pumps primed
 * and not starve the wlan-to-wlan path we have some additional
 * buffers for our receive processing to loan; we also accomodate
 * a somewhat deeper (almost twice) tx queue to absorb local AP
 * stack traffic, SME traffic, second ethernet traffic (on Spirit),
 * to account for fragmentation (which could be a descriptor hog!),
 * as well as still appropriately throttle the ethernet receive
 * processing - in face of all of the above;
 * the amount of buffer resources allocated to power save could
 * possibily be determined by the amount of memory available on
 * the system - bears better understanding of latency impact etc.
 */
#if FADI  // for now, shrink the size of the static pool...maybe
// increase later...!???                                             
#define AR_TX_QUEUE_DEPTH          256          /* max nframes allowed in the transmit queue */
#define AR_TX_QUEUE_FRAG_DELTA     744          /* fragmentation descriptors */
#define AR_RX_QUEUE_DEPTH          128          /* the maintained receive queue size */
#define AR_MAX_LOANED_COUNT        144          /* see above */
#define AR_PS_QUEUE_DEPTH          256          /* max nframes allowed in the PS queues pending sta awake */
#else //fadi
#define AR_TX_QUEUE_DEPTH          300//256//96//32          
#define AR_TX_QUEUE_FRAG_DELTA     1          
#define AR_RX_QUEUE_DEPTH          128//48//16          
#define AR_MAX_LOANED_COUNT        1           
#define AR_PS_QUEUE_DEPTH          1//2 //32          
#endif //fadi

#define AR_HEADER_SIZE              64
#define AR_BUF_SIZE                 1632        /*used to be 1596, 32 * 50 - 4 i.e. 4 less than multiple of a cacheline size */
#define AR_DESC_SIZE                sizeof(ATHEROS_DESC)

#define WLAN_MAX_BODY               2312
#define ETH_MAX_PKT_SIZ             (2312 + ENET_HDR_REAL_SIZ + ETH_CRC_LEN)
#ifndef _CACHE_ALIGN_SIZE
#define _CACHE_ALIGN_SIZE           32
#endif /* _CACHE_ALIGN_SIZE */
#define CACHE_LINE_SIZE             _CACHE_ALIGN_SIZE   /* PPC 32, MIPS 16 */
#define DEFAULT_DOT11_RTS_THRESHOLD 512

/* Definitions for the flags field */
#define AR_PROMISCUOUS  0x1
#define AR_POLLING      0x2
#define AR_RCV          0x4
#define AR_BUF_LOCAL    0x8
#define AR_FULL_DUPLEX  0x10

/* IOCTL definitions */
#define EIOCBSSSTART    _IO('e', 128)


/* XXX This is defined for compiling, check it later */
#define END_SPEED       10000000         /* 10Mbs */

/*
 * SWAP_BUS_TO_CPU_XX and SWAP_CPU_TO_BUS_XX macros swap 16 and 32 bit values
 * between the PCI bus's little endian byte-order and the CPU's native
 * byte-order.
 */

#if (_BYTE_ORDER == _BIG_ENDIAN)
#define BUS_TO_CPU_SWAP_16(X)    swap16(X)
#define BUS_TO_CPU_SWAP_32(X)    swap32(X)
#define CPU_TO_BUS_SWAP_16(X)    swap16(X)
#define CPU_TO_BUS_SWAP_32(X)    swap32(X)
#define CPU_TO_NET_SWAP_16(X)    ((UINT16)(X))
#define CPU_TO_NET_SWAP_32(X)    ((UINT32)(X))
#define NET_TO_CPU_SWAP_16(X)    ((UINT16)(X))
#define NET_TO_CPU_SWAP_32(X)    ((UINT32)(X))
#else   /* (_BYTE_ORDER == _BIG_ENDIAN) */
#define BUS_TO_CPU_SWAP_32(X)    ((UINT32)(X))
#define BUS_TO_CPU_SWAP_16(X)    ((UINT16)(X))
#define CPU_TO_BUS_SWAP_32(X)    ((UINT32)(X))
#define CPU_TO_BUS_SWAP_16(X)    ((UINT16)(X))
#define CPU_TO_NET_SWAP_16(X)    swap16(X)
#define CPU_TO_NET_SWAP_32(X)    swap32(X)
#define NET_TO_CPU_SWAP_16(X)    swap16(X)
#define NET_TO_CPU_SWAP_32(X)    swap32(X)
#endif  /* (_BYTE_ORDER == _BIG_ENDIAN) */

/* Macros to read/write 32 bit data to/from AR5210 device registers */
#if defined(PPC8240)
#define writePlatformReg(pdevInfo, reg, val) \
        writel((val), (((pdevInfo)->baseAddress)+(reg)) )
//        sysPciOutLong((pdevInfo)->baseAddress+(reg), val)

#define readPlatformReg(pdevInfo, reg) \
        readl(((pdevInfo)->baseAddress)+(reg))
//        sysPciInLong((pdevInfo)->baseAddress+(reg))

#elif defined(AR5311)
#define writePlatformReg(pdevInfo, reg, val) \
    (*((volatile ULONG *)((pdevInfo)->baseAddress + (reg))) = (val))
#define readPlatformReg(pdevInfo, reg) \
    (*((volatile ULONG *)((pdevInfo)->baseAddress + (reg))))
#else
#define writePlatformReg(pdevInfo, reg, val)                            \
    IO_WRITE_32(((WLAN_DEV_INFO *)pdevInfo)->baseAddress+(reg),        \
                CPU_TO_BUS_SWAP_32(val))
#define readPlatformReg(pdevInfo, reg) \
    BUS_TO_CPU_SWAP_32(IO_READ_32(((WLAN_DEV_INFO *)pdevInfo)->baseAddress+(reg)))
#endif

/* A shortcut for getting the hardware address from the MIB II stuff. */
#define END_HADDR(pEnd) \
        ((pEnd)->mib2Tbl.ifPhysAddress.phyAddress)

#define END_HADDR_LEN(pEnd) \
        ((pEnd)->mib2Tbl.ifPhysAddress.addrLength)

//#define udelay(x)   sysUDelay(x)

typedef struct ar5_param {
    unsigned long   pciMemBase;     /* PCI memory base address */
    int             ivec;           /* Device interrupt vector */
    int             ilevel;         /* Device interrupt level */
    A_UINT16        pciVendorID;
    A_UINT16        pciDeviceID;
    A_UINT16        pciSubVendorID;
    A_UINT16        pciSubDeviceID;
    A_UINT8         pciRevisionID;
} AR5_PARAM;

extern AR5_PARAM arParam;

#if defined(PPC8240)
A_UINT16 swap16(A_UINT16 *);
A_UINT32 swap32(A_UINT32 *);
#endif

A_STATUS
beaconWdgStart(WLAN_DEV_INFO *pdevInfo);

A_STATUS
ar5hwcDrvSend(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL callerHoldsLock);

/*
 * This function is going to be called by MuxLoad through endDevTbl which is
 * defined in configNet.h. That header file also defines the init string.
 */
#ifdef VXWORKS
END_OBJ *
ar5hwcEndLoad(char *initString, void *);
#endif

void
processTxQueue(WLAN_DEV_INFO *pdevInfo);

#endif /* __AR5HWC_H__ */
