/* Host AP driver's support for Atheros AR5210/5211 PCI cards is based on
 * driver obtained from Atheros Inc. */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/delay.h>

#include <linux/ioport.h>
#include <linux/pci.h>

#include "linuxdrv.h"

MODULE_AUTHOR("Atheros Authors");
MODULE_DESCRIPTION("Support for Atheros521x-based 802.11A wireless LAN "
		   "PCI cards.");
MODULE_SUPPORTED_DEVICE("Atheros AR521x-based WLAN PCI cards");
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0))
/* PCI initialization uses Linux 2.4.x version and older kernels do not support
 * this */
#error Atheros521x PCI version requires at least Linux kernel version 2.4.0
#endif /* kernel < 2.4.0 */

// currently, path included since we're compiling in PRISM dir
// make sure to remove path eventually when files are in atheros dir
#ifdef CONFIG_ATHEROS_5210_PCI
#include "ar5210Reg.h"                                   
#else
#include "ar5211Reg.h"
#endif
#include "wlandrv.h"
#include "atheros_wlan.h"
#include "atheros_sap.h"

/* FIX: do we need mb/wmb/rmb with memory operations? */
extern void  beaconShow(WLAN_DEV_INFO *pwlan);

static struct pci_device_id atheros_pci_id_table[] __devinitdata = {
    /* Atheros AR5211 802.11A WLAN Controller */
	{ 0x168c, 0x0012, PCI_ANY_ID, PCI_ANY_ID },
    { 0x168c, 0x0007, PCI_ANY_ID, PCI_ANY_ID },
    { 0x168c, 0x0011, PCI_ANY_ID, PCI_ANY_ID }, /* PreProd AR5211 DevID */
    { 0x168c, 0x1112, PCI_ANY_ID, PCI_ANY_ID }, /* Default AR5211 DevID */
    { 0x168c, 0xF11B, PCI_ANY_ID, PCI_ANY_ID }, /* 11b emulation FPGA */
#ifdef CONFIG_ATHEROS_5210_PCI
    /* Atheros AR5210 802.11A WLAN Controller */
    { 0x168c, 0x0207, PCI_ANY_ID, PCI_ANY_ID }, 
    { 0x168c, DEV_LEGACY, PCI_ANY_ID, PCI_ANY_ID }, 
#endif // CONFIG_ATHEROS_5210_PCI
    { 0 }
};

                                          
WLAN_DEV_INFO *pwlanDevInfo;
struct pci_dev * gPciDev = NULL;
A_UINT32 g_pci_len = 0;

/* prototypes */
#ifndef FADI_TEMP
static int atheros_pci_probe(struct pci_dev *pdev,
			    const struct pci_device_id *id);
static void atheros_pci_remove(struct pci_dev *pdev);
#endif

MODULE_DEVICE_TABLE(pci, atheros_pci_id_table);

static struct pci_driver atheros_pci_drv_id = {
	name: "AR521x_pci",
	id_table: atheros_pci_id_table,
	probe: atheros_pci_probe,
	remove: atheros_pci_remove,
	save_state: NULL,
	suspend: NULL,
	resume: NULL,
	enable_wake: NULL
};


/* FIX: can jiffies be used this way from every place cor_sreset could be
 * called? (mainly, spin_lock_irq? hw interrupt handler?) */

static void atheros_pci_cor_sreset(atheros_local_info_t *local)
{
    aPrintk("Entering atheros_pci_cor_sreset...\n");
}


static struct atheros_helper_functions atheros_pci_funcs =
{
	card_present: NULL,
	cor_sreset: atheros_pci_cor_sreset,
	dev_open: NULL,
	dev_close: NULL
};


static int atheros_pci_probe(struct pci_dev *pdev,
			    const struct pci_device_id *id)
{
	unsigned long phymem;
	unsigned long mem = 0;
	atheros_local_info_t *local = NULL;
	struct net_device *dev;
	static int cards_found /* = 0 */;

    if (pci_enable_device(pdev))
		return -EIO;

    aPrintk("\tPCI ENABLED THE DEVICE!\n");

    phymem = pci_resource_start(pdev, 0);
    g_pci_len = pci_resource_len(pdev, 0);
    aPrintk("resource_len = %d\n", g_pci_len);

    aPrintk("\tPCI STARTED RESOURCE...phymem=0x%lx\n", phymem);

	if (!request_mem_region(phymem, g_pci_len, "Atheros521x")) {
		aPrintk(KERN_ERR "atheros: Cannot reserve PCI memory region\n");
		goto err_out_disable;
	}

    aPrintk("\trequest_mem_region with len=0x%lx\n", g_pci_len);

	mem = (unsigned long) ioremap(phymem, g_pci_len);
	if (!mem) {
		aPrintk(KERN_ERR "atheros: Cannot remap PCI memory region\n") ;
		goto fail;
	}

    aPrintk("\tioremapped the phymem to mem=0x%lx\n", mem);

#ifdef ATHEROS_BUS_MASTER
    // in setup_bus.c, enable_dev automatically enabled bus mastering
    aPrintk("\tSETTING PCI BUS MASTER STATUS.\n");
	pci_set_master(pdev);
#if 0
    A_UINT8 tmp8;
    A_UINT32 tmp32;
      
    pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, (A_UINT8 *)&tmp8);
	printk("\t\tCACHE LINE SIZE = %d(in 4Bytes)\n", tmp8);
                                               
    pci_read_config_byte(pdev, PCI_LATENCY_TIMER, (A_UINT8 *)&tmp8);
    printk("\t\tLATENCY_TIMER = %d(in cycles)\n", tmp8);
    pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 0x0);
    pci_read_config_byte(pdev, PCI_LATENCY_TIMER, (A_UINT8 *)&tmp8);
    printk("\t\tLATENCY_TIMER = %d(in cycles)\n", tmp8);


    pci_read_config_dword(pdev, 0x40, (A_UINT32 *)&tmp32);
    printk("\t\tPCI_RETRY_TIMER = 0x%8lx(in cycles)\n", tmp32);
    pci_write_config_dword(pdev, 0x40, (A_UINT32 *)0x0);
    pci_read_config_dword(pdev, 0x40, (A_UINT32 *)&tmp32);
    printk("\t\tPCI_RETRY_TIMER = 0x%8lx(in cycles)\n", tmp32);

#endif 


#endif /* ATHEROS_BUS_MASTER ... PRISM2_BUS_MASTER */

    local = atheros_init_local_data(&atheros_pci_funcs, cards_found, mem);
	if (local == NULL)
		goto fail;

    cards_found++;
    aPrintk("\tInit local data..done[cards_found=%d]\n", cards_found);

	dev = local->dev;
    dev->irq = pdev->irq;
    dev->mem_start = mem;
    dev->mem_end = mem + g_pci_len;
	dev->base_addr = mem; // I don't think it should be phys mem!! Fadi
    
    // init dev struct, registers device + forwarder initialization
	if (atheros_init_dev(local))
		goto fail;

    aPrintk("\tInit dev.w:\n");
    aPrintk("\tenabled? : %s\n", (local->dev_enabled)?"true":"false");
    aPrintk("\tnet_device: w\n");
      aPrintk("\t\tname: %s\n", local->dev->name);
      aPrintk("\t\tirq: %d\n", local->dev->irq);
      aPrintk("\t\tmtu: %d\n", local->dev->mtu);
	  aPrintk("\t\tdev_base_addr: 0x%lx\n", (A_UINT32)local->dev->base_addr);
      aPrintk("\t\tdev_mac_addr: %x:%x:%x:%x:%x:%x:%x:%x\n", 
			 local->dev->dev_addr[0],local->dev->dev_addr[1], 
			 local->dev->dev_addr[2],local->dev->dev_addr[3],
			 local->dev->dev_addr[4],local->dev->dev_addr[5], 
			 local->dev->dev_addr[6],local->dev->dev_addr[7]); //it's up to 8B
    aPrintk("\tinfofid=0x%x(mac buf id)\n", local->infofid);
    aPrintk("\tintransmitfid/txBufId=\n\t[0x%x][0x%x][0x%x][0x%x]\n\t[0x%x][0x%x][0x%x][0x%x]\n", 
           local->intransmitfid[0],local->intransmitfid[1],
           local->intransmitfid[2],local->intransmitfid[3],
           local->intransmitfid[4],local->intransmitfid[5],
           local->intransmitfid[6],local->intransmitfid[7]);
    aPrintk("\tessid=%s\n", local->essid);
    
    //fadi, halAttach in init_local_data has already done a halReset
    //atheros_pci_cor_sreset(local); 

	if (atheros_hw_config(dev, 1)) {
        aPrintk("hardware initialization failed\n");
		goto fail;
	}  

    aPrintk("*****    hw_config, done    *****\n");

	pci_set_drvdata(pdev, dev);

	if (request_irq(dev->irq, atheros_interrupt, SA_SHIRQ, dev->name,
			dev)) {
		aPrintk(KERN_WARNING "%s: request_irq failed\n", dev->name);
		goto fail;
	}

    gPciDev = pdev; // global PCI device
    
    aPrintk("******gPciDev = %p | NET_DEV=%p*******\n", gPciDev, dev);
#ifdef CONFIG_ATHEROS_5210_PCI
    printk(KERN_INFO "%s: Atheros AR5210 PCI: "
	       "mem=0x%lx, irq=%d\n", dev->name, phymem, dev->irq);
#else
	printk(KERN_INFO "%s: Atheros AR5211 PCI: "
	       "mem=0x%lx, irq=%d\n", dev->name, phymem, dev->irq);
#endif


	return 0;

 fail:
	atheros_free_local_data(local);

	if (mem)
		iounmap((void *) mem);

	release_mem_region(phymem, g_pci_len);

 err_out_disable:
	pci_disable_device(pdev);

	return -ENODEV;
}


static void atheros_pci_remove(struct pci_dev *pdev)
{
    pci_disable_device(pdev);
}

static int __init init_atheros_pci(void)
{
	aPrintk(KERN_INFO "%s: %s\n"
	       "%s: (c) Atheros Communications\n",
	       dev_info, version, dev_info);

	if (pci_register_driver(&atheros_pci_drv_id) <= 0) {
		printk("atheros_pci: No devices found, driver not "
		       "installed.\n");
		pci_unregister_driver(&atheros_pci_drv_id);
		return -ENODEV;
	}

	return 0;
}


static void __exit exit_atheros_pci(void)
{
	pci_unregister_driver(&atheros_pci_drv_id);
	printk(KERN_INFO "AR521x: Driver unloaded\n");
}


module_init(init_atheros_pci);
module_exit(exit_atheros_pci);
