#ifndef ATHEROS_WLAN_H
#define ATHEROS_WLAN_H

#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/netdevice.h>

#include "linuxdrv.h"
#include "wifi/asap_drv.h"

extern struct pci_dev * gPciDev;

#define BIT(x) (1 << (x))

/* netlink groups */
#define ATHEROS_MONITOR_GROUP BIT(0)

#ifndef ARPHRD_IEEE80211
#define ARPHRD_IEEE80211 801
#endif

typedef enum _monitor_type 
{
    ATHEROS_MONITOR_OFF, 
    ATHEROS_MONITOR_ON
} monitor_type;

extern monitor_type g_monitor_type;

struct ar5211_rx_frame {
	/* ATHEROS RX frame descriptor */
	u16 status; /* ATHEROS_RX_STATUS_ flags */
	u32 time; /* timestamp, 1 microsecond resolution */
	u8 silence; /* 27 .. 154; seems to be 0 */
	u8 signal; /* 27 .. 154 */
	u8 rate; /* 10, 20, 55, or 110 */
	u8 rxflow;
	u32 reserved;

	/* 802.11 */
	u16 frame_control;
	u16 duration_id;
	u8 addr1[6];
	u8 addr2[6];
	u8 addr3[6];
	u16 seq_ctrl;
	u8 addr4[6];
	u16 data_len;

	/* 802.3 */
	u8 dst_addr[6];
	u8 src_addr[6];
	u16 len;

	/* followed by frame data; max 2304 bytes */
} __attribute__ ((packed));


struct ar5211_tx_frame {
	/* ATHEROS TX frame descriptor */
	u16 status; /* ATHEROS_TX_STATUS_ flags */
	u16 reserved1;
	u16 reserved2;
	u32 sw_support;
	u8 retry_count; /* not yet implemented */
	u8 tx_rate; /* Host AP only; 0 = firmware, or 10, 20, 55, 110 */
	u16 tx_control; /* ATHEROS_TX_CTRL_ flags */

	/* 802.11 */
	u16 frame_control; /* parts not used */
	u16 duration_id;
	u8 addr1[6];
	u8 addr2[6]; /* filled by firmware */
	u8 addr3[6];
	u16 seq_ctrl; /* filled by firmware */
	u8 addr4[6];
	u16 data_len;

	/* 802.3 */
	u8 dst_addr[6];
	u8 src_addr[6];
	u16 len;

	/* followed by frame data; max 2304 bytes */
} __attribute__ ((packed));


struct ar5211_rid_hdr
{
	u16 len;
	u16 rid;
} __attribute__ ((packed));


struct ar5211_comp_ident
{
	u16 id;
	u16 variant;
	u16 major;
	u16 minor;
} __attribute__ ((packed));

#define ATHEROS_COMP_ID_PRI 0x15
#define ATHEROS_COMP_ID_STA 0x1f
#define ATHEROS_COMP_ID_FW_AP 0x14b

struct ar5211_sup_range
{
	u16 role;
	u16 id;
	u16 variant;
	u16 bottom;
	u16 top;
} __attribute__ ((packed));

struct ar5211_build_id
{
	u16 pri_seq;
	u16 sec_seq;
} __attribute__ ((packed));

/* BSS connection quality (RID FD43 range, RID FD51 dBm-normalized) */
struct ar5211_comms_quality {
	u16 comm_qual; /* 0 .. 92 */
	u16 signal_level; /* 27 .. 154 */
	u16 noise_level; /* 27 .. 154 */
} __attribute__ ((packed));

/* Macro for converting signal levels (range 27 .. 154) to wireless ext
 * dBm value with some accuracy */
#define ATHEROS_LEVEL_TO_dBm(v) 0x100 + (v) * 100 / 255 - 100

#define ATHEROS_SCAN_TXRATE_1 1
#define ATHEROS_SCAN_TXRATE_2 2
#define ATHEROS_SCAN_TXRATE_5_5 4
#define ATHEROS_SCAN_TXRATE_11 8

struct ar5211_scan_request {
	u16 channel_list;
	u16 txrate;
} __attribute__ ((packed));

struct ar5211_info_frame {
	u16 len;
	u16 type;
} __attribute__ ((packed));

struct ar5211_comm_tallies {
	u16 tx_unicast_frames;
	u16 tx_multicast_frames;
	u16 tx_fragments;
	u16 tx_unicast_octets;
	u16 tx_multicast_octets;
	u16 tx_deferred_transmissions;
	u16 tx_single_retry_frames;
	u16 tx_multiple_retry_frames;
	u16 tx_retry_limit_exceeded;
	u16 tx_discards;
	u16 rx_unicast_frames;
	u16 rx_multicast_frames;
	u16 rx_fragments;
	u16 rx_unicast_octets;
	u16 rx_multicast_octets;
	u16 rx_fcs_errors;
	u16 rx_discards_no_buffer;
	u16 tx_discards_wrong_sa;
	u16 rx_discards_wep_undecryptable;
	u16 rx_message_in_msg_fragments;
	u16 rx_message_in_bad_msg_fragments;
} __attribute__ ((packed));

struct ar5211_scan_result {
	u16 chid;
	u16 anl;
	u16 sl;
	u8 bssid[6];
	u16 beacon_interval;
	u16 capability;
	u8 ssid[34];
	u8 sup_rates[10];
	u16 rate;
} __attribute__ ((packed));



#if defined(ATHEROS_PCCARD) || defined(ATHEROS_PLX)
/* I/O ports for ATHEROS Controller access */
#define ATHEROS_CMD_OFF 0x00
#define ATHEROS_PARAM0_OFF 0x02
#define ATHEROS_PARAM1_OFF 0x04
#define ATHEROS_PARAM2_OFF 0x06
#define ATHEROS_STATUS_OFF 0x08
#define ATHEROS_RESP0_OFF 0x0A
#define ATHEROS_RESP1_OFF 0x0C
#define ATHEROS_RESP2_OFF 0x0E
#define ATHEROS_INFOFID_OFF 0x10
#define ATHEROS_CONTROL_OFF 0x14
#define ATHEROS_SELECT0_OFF 0x18
#define ATHEROS_SELECT1_OFF 0x1A
#define ATHEROS_OFFSET0_OFF 0x1C
#define ATHEROS_OFFSET1_OFF 0x1E
#define ATHEROS_RXFID_OFF 0x20
#define ATHEROS_ALLOCFID_OFF 0x22
#define ATHEROS_TXCOMPLFID_OFF 0x24
#define ATHEROS_SWSUPPORT0_OFF 0x28
#define ATHEROS_SWSUPPORT1_OFF 0x2A
#define ATHEROS_SWSUPPORT2_OFF 0x2C
#define ATHEROS_EVSTAT_OFF 0x30
#define ATHEROS_INTEN_OFF 0x32
#define ATHEROS_EVACK_OFF 0x34
#define ATHEROS_DATA0_OFF 0x36
#define ATHEROS_DATA1_OFF 0x38
#define ATHEROS_AUXPAGE_OFF 0x3A
#define ATHEROS_AUXOFFSET_OFF 0x3C
#define ATHEROS_AUXDATA_OFF 0x3E
#endif /* ATHEROS_PCCARD || ATHEROS_PLX */

#ifdef ATHEROS_PCI
/* Memory addresses for ISL3874 controller access */
#define ATHEROS_CMD_OFF 0x00
#define ATHEROS_PARAM0_OFF 0x04
#define ATHEROS_PARAM1_OFF 0x08
#define ATHEROS_PARAM2_OFF 0x0C
#define ATHEROS_STATUS_OFF 0x10
#define ATHEROS_RESP0_OFF 0x14
#define ATHEROS_RESP1_OFF 0x18
#define ATHEROS_RESP2_OFF 0x1C
#define ATHEROS_INFOFID_OFF 0x20
#define ATHEROS_CONTROL_OFF 0x28
#define ATHEROS_SELECT0_OFF 0x30
#define ATHEROS_SELECT1_OFF 0x34
#define ATHEROS_OFFSET0_OFF 0x38
#define ATHEROS_OFFSET1_OFF 0x3C
#define ATHEROS_RXFID_OFF 0x40
#define ATHEROS_ALLOCFID_OFF 0x44
#define ATHEROS_TXCOMPLFID_OFF 0x48
#define ATHEROS_PCICOR_OFF 0x4C
#define ATHEROS_SWSUPPORT0_OFF 0x50
#define ATHEROS_SWSUPPORT1_OFF 0x54
#define ATHEROS_SWSUPPORT2_OFF 0x58
#define ATHEROS_PCIHCR_OFF 0x5C
#define ATHEROS_EVSTAT_OFF 0x60
#define ATHEROS_INTEN_OFF 0x64
#define ATHEROS_EVACK_OFF 0x68
#define ATHEROS_DATA0_OFF 0x6C
#define ATHEROS_DATA1_OFF 0x70
#define ATHEROS_AUXPAGE_OFF 0x74
#define ATHEROS_AUXOFFSET_OFF 0x78
#define ATHEROS_AUXDATA_OFF 0x7C
#define ATHEROS_PCI_M0_ADDRH_OFF 0x80
#define ATHEROS_PCI_M0_ADDRL_OFF 0x84
#define ATHEROS_PCI_M0_LEN_OFF 0x88
#define ATHEROS_PCI_M0_CTL_OFF 0x8C
#define ATHEROS_PCI_STATUS_OFF 0x98
#define ATHEROS_PCI_M1_ADDRH_OFF 0xA0
#define ATHEROS_PCI_M1_ADDRL_OFF 0xA4
#define ATHEROS_PCI_M1_LEN_OFF 0xA8
#define ATHEROS_PCI_M1_CTL_OFF 0xAC

/* PCI bus master control bits (these are undocumented; based on guessing and
 * experimenting..) */
#define ATHEROS_PCI_CTL_FROM_BAP (BIT(5) | BIT(1) | BIT(0))
#define ATHEROS_PCI_CTL_TO_BAP (BIT(5) | BIT(0))

#endif /* ATHEROS_PCI */

#if defined(ATHEROS_PCCARD) || defined(ATHEROS_PLX)
#define ATHEROS_OUTB(v,a) outb((v), dev->base_addr + (a))
#define ATHEROS_INB(a) inb(dev->base_addr + (a))
#define ATHEROS_OUTW(v,a) outw((v), dev->base_addr + (a))
#define ATHEROS_INW(a) inw(dev->base_addr + (a))
#define ATHEROS_INSW(a, buf, wc) insw(dev->base_addr + (a), buf, wc)
#define ATHEROS_OUTSW(a, buf, wc) outsw(dev->base_addr + (a), buf, wc)
#endif /* ATHEROS_PCCARD || ATHEROS_PLX */

#ifdef ATHEROS_PCI
#define ATHEROS_OUTB(v,a) writeb((v), dev->mem_start + (a))
#define ATHEROS_INB(a) (u8) readb(dev->mem_start + (a))
#define ATHEROS_OUTW(v,a) writew((v), dev->mem_start + (a))
#define ATHEROS_INW(a) (u16) readw(dev->mem_start + (a))
#define ATHEROS_OUTW_DATA(v,a) writew(cpu_to_le16(v), dev->mem_start + (a))
#define ATHEROS_INW_DATA(a) (u16) le16_to_cpu(readw(dev->mem_start + (a)))
#endif /* ATHEROS_PCI */


/* Command codes for CMD reg. */
#define ATHEROS_CMDCODE_INIT 0x00
#define ATHEROS_CMDCODE_ENABLE 0x01
#define ATHEROS_CMDCODE_DISABLE 0x02
#define ATHEROS_CMDCODE_ALLOC 0x0A
#define ATHEROS_CMDCODE_TRANSMIT 0x0B
#define ATHEROS_CMDCODE_INQUIRE 0x11
#define ATHEROS_CMDCODE_ACCESS 0x21
#define ATHEROS_CMDCODE_ACCESS_WRITE (0x21 | BIT(8))
#define ATHEROS_CMDCODE_DOWNLOAD 0x22
#define ATHEROS_CMDCODE_READMIF 0x30
#define ATHEROS_CMDCODE_WRITEMIF 0x31
#define ATHEROS_CMDCODE_TEST 0x38

/* Test mode operations */
#define ATHEROS_TEST_MONITOR 0x0B
#define ATHEROS_TEST_STOP 0x0F
#define ATHEROS_TEST_CFG_BITS 0x15
#define ATHEROS_TEST_CFG_BIT_ALC BIT(3)

#define ATHEROS_CMD_BUSY BIT(15)

#define ATHEROS_CMD_TX_RECLAIM BIT(8)

#define ATHEROS_OFFSET_ERR BIT(14)
#define ATHEROS_OFFSET_BUSY BIT(15)


/* ProgMode for download command */
#define ATHEROS_PROGMODE_DISABLE 0
#define ATHEROS_PROGMODE_ENABLE_VOLATILE 1
#define ATHEROS_PROGMODE_ENABLE_NON_VOLATILE 2
#define ATHEROS_PROGMODE_PROGRAM_NON_VOLATILE 3

#define ATHEROS_AUX_MAGIC0 0xfe01
#define ATHEROS_AUX_MAGIC1 0xdc23
#define ATHEROS_AUX_MAGIC2 0xba45

#define ATHEROS_AUX_PORT_DISABLED 0
#define ATHEROS_AUX_PORT_DISABLE BIT(14)
#define ATHEROS_AUX_PORT_ENABLE BIT(15)
#define ATHEROS_AUX_PORT_ENABLED (BIT(14) | BIT(15))
#define ATHEROS_AUX_PORT_MASK (BIT(14) | BIT(15))

#define ATHEROS_PDA_SIZE 1024


/* Events; EvStat, Interrupt mask (IntEn), and acknowledge bits (EvAck) */
#define ATHEROS_EV_TICK BIT(15)
#define ATHEROS_EV_WTERR BIT(14)
#define ATHEROS_EV_INFDROP BIT(13)
#ifdef ATHEROS_PCI
#define ATHEROS_EV_PCI_M1 BIT(9)
#define ATHEROS_EV_PCI_M0 BIT(8)
#endif /* ATHEROS_PCI */
#define ATHEROS_EV_INFO BIT(7)
#define ATHEROS_EV_DTIM BIT(5)
#define ATHEROS_EV_CMD BIT(4)
#define ATHEROS_EV_ALLOC BIT(3)
#define ATHEROS_EV_TXEXC BIT(2)
#define ATHEROS_EV_TX BIT(1)
#define ATHEROS_EV_RX BIT(0)


/* ATHEROS Configuration RIDs */
#define ATHEROS_RID_CNFPORTTYPE 0xFC00
#define ATHEROS_RID_CNFOWNMACADDR 0xFC01
#define ATHEROS_RID_CNFDESIREDSSID 0xFC02
#define ATHEROS_RID_CNFOWNCHANNEL 0xFC03
#define ATHEROS_RID_CNFOWNSSID 0xFC04
#define ATHEROS_RID_CNFOWNATIMWINDOW 0xFC05
#define ATHEROS_RID_CNFSYSTEMSCALE 0xFC06
#define ATHEROS_RID_CNFMAXDATALEN 0xFC07
#define ATHEROS_RID_CNFWDSADDRESS 0xFC08
#define ATHEROS_RID_CNFPMENABLED 0xFC09
#define ATHEROS_RID_CNFPMEPS 0xFC0A
#define ATHEROS_RID_CNFMULTICASTRECEIVE 0xFC0B
#define ATHEROS_RID_CNFMAXSLEEPDURATION 0xFC0C
#define ATHEROS_RID_CNFPMHOLDOVERDURATION 0xFC0D
#define ATHEROS_RID_CNFOWNNAME 0xFC0E
#define ATHEROS_RID_CNFOWNDTIMPERIOD 0xFC10
#define ATHEROS_RID_CNFWDSADDRESS1 0xFC11
#define ATHEROS_RID_CNFWDSADDRESS2 0xFC12
#define ATHEROS_RID_CNFWDSADDRESS3 0xFC13
#define ATHEROS_RID_CNFWDSADDRESS4 0xFC14
#define ATHEROS_RID_CNFWDSADDRESS5 0xFC15
#define ATHEROS_RID_CNFWDSADDRESS6 0xFC16
#define ATHEROS_RID_CNFMULTICASTPMBUFFERING 0xFC17
#define ATHEROS_RID_UNKNOWN1 0xFC20
#define ATHEROS_RID_UNKNOWN2 0xFC21
#define ATHEROS_RID_CNFWEPDEFAULTKEYID 0xFC23
#define ATHEROS_RID_CNFDEFAULTKEY0 0xFC24
#define ATHEROS_RID_CNFDEFAULTKEY1 0xFC25
#define ATHEROS_RID_CNFDEFAULTKEY2 0xFC26
#define ATHEROS_RID_CNFDEFAULTKEY3 0xFC27
#define ATHEROS_RID_CNFWEPFLAGS 0xFC28
#define ATHEROS_RID_CNFWEPKEYMAPPINGTABLE 0xFC29
#define ATHEROS_RID_CNFAUTHENTICATION 0xFC2A
#define ATHEROS_RID_CNFMAXASSOCSTA 0xFC2B
#define ATHEROS_RID_CNFTXCONTROL 0xFC2C
#define ATHEROS_RID_CNFROAMINGMODE 0xFC2D
#define ATHEROS_RID_CNFHOSTAUTHENTICATION 0xFC2E
#define ATHEROS_RID_CNFRCVCRCERROR 0xFC30
#define ATHEROS_RID_CNFMMLIFE 0xFC31
#define ATHEROS_RID_CNFALTRETRYCOUNT 0xFC32
#define ATHEROS_RID_CNFBEACONINT 0xFC33
#define ATHEROS_RID_CNFAPPCFINFO 0xFC34
#define ATHEROS_RID_CNFSTAPCFINFO 0xFC35
#define ATHEROS_RID_CNFPRIORITYQUSAGE 0xFC37
#define ATHEROS_RID_CNFTIMCTRL 0xFC40
#define ATHEROS_RID_UNKNOWN3 0xFC41 /* added in STA f/w 0.7.x */
#define ATHEROS_RID_CNFTHIRTY2TALLY 0xFC42 /* added in STA f/w 0.8.0 */
#define ATHEROS_RID_CNFENHSECURITY 0xFC43
#define ATHEROS_RID_CNFDBMADJUST 0xFC46 /* added in STA f/w 1.3.1 */
#define ATHEROS_RID_GROUPADDRESSES 0xFC80
#define ATHEROS_RID_CREATEIBSS 0xFC81
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD 0xFC82
#define ATHEROS_RID_RTSTHRESHOLD 0xFC83
#define ATHEROS_RID_TXRATECONTROL 0xFC84
#define ATHEROS_RID_PROMISCUOUSMODE 0xFC85
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD0 0xFC90 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD1 0xFC91 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD2 0xFC92 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD3 0xFC93 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD4 0xFC94 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD5 0xFC95 /* AP f/w only */
#define ATHEROS_RID_FRAGMENTATIONTHRESHOLD6 0xFC96 /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD0 0xFC97 /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD1 0xFC98 /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD2 0xFC99 /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD3 0xFC9A /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD4 0xFC9B /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD5 0xFC9C /* AP f/w only */
#define ATHEROS_RID_RTSTHRESHOLD6 0xFC9D /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL0 0xFC9E /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL1 0xFC9F /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL2 0xFCA0 /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL3 0xFCA1 /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL4 0xFCA2 /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL5 0xFCA3 /* AP f/w only */
#define ATHEROS_RID_TXRATECONTROL6 0xFCA4 /* AP f/w only */
#define ATHEROS_RID_CNFSHORTPREAMBLE 0xFCB0
#define ATHEROS_RID_CNFEXCLUDELONGPREAMBLE 0xFCB1
#define ATHEROS_RID_CNFAUTHENTICATIONRSPTO 0xFCB2
#define ATHEROS_RID_CNFBASICRATES 0xFCB3
#define ATHEROS_RID_CNFSUPPORTEDRATES 0xFCB4
#define ATHEROS_RID_UNKNOWN5 0xFCB5 /* added in STA f/w 1.3.1 */
#define ATHEROS_RID_WEPKEYDISABLE 0xFCB6 /* added in STA f/w 1.3.1 */
#define ATHEROS_RID_TICKTIME 0xFCE0
#define ATHEROS_RID_SCANREQUEST 0xFCE1
#define ATHEROS_RID_JOINREQUEST 0xFCE2
#define ATHEROS_RID_AUTHENTICATESTATION 0xFCE3 /* AP f/w only */
#define ATHEROS_RID_CHANNELINFOREQUEST 0xFCE4 /* AP f/w only */
#define ATHEROS_RID_HOSTSCAN 0xFCE5 /* added in STA f/w 1.3.1 */

/* ATHEROS Information RIDs */
#define ATHEROS_RID_MAXLOADTIME 0xFD00
#define ATHEROS_RID_DOWNLOADBUFFER 0xFD01
#define ATHEROS_RID_PRIID 0xFD02
#define ATHEROS_RID_PRISUPRANGE 0xFD03
#define ATHEROS_RID_CFIACTRANGES 0xFD04
#define ATHEROS_RID_NICSERNUM 0xFD0A
#define ATHEROS_RID_NICID 0xFD0B
#define ATHEROS_RID_MFISUPRANGE 0xFD0C
#define ATHEROS_RID_CFISUPRANGE 0xFD0D
#define ATHEROS_RID_CHANNELLIST 0xFD10
#define ATHEROS_RID_REGULATORYDOMAINS 0xFD11
#define ATHEROS_RID_TEMPTYPE 0xFD12
#define ATHEROS_RID_CIS 0xFD13
#define ATHEROS_RID_STAID 0xFD20
#define ATHEROS_RID_STASUPRANGE 0xFD21
#define ATHEROS_RID_MFIACTRANGES 0xFD22
#define ATHEROS_RID_CFIACTRANGES2 0xFD23
#define ATHEROS_RID_PRODUCTNAME 0xFD24 /* added in STA f/w 1.3.1;
					* only Prism2.5(?) */
#define ATHEROS_RID_PORTSTATUS 0xFD40
#define ATHEROS_RID_CURRENTSSID 0xFD41
#define ATHEROS_RID_CURRENTBSSID 0xFD42
#define ATHEROS_RID_COMMSQUALITY 0xFD43
#define ATHEROS_RID_CURRENTTXRATE 0xFD44
#define ATHEROS_RID_CURRENTBEACONINTERVAL 0xFD45
#define ATHEROS_RID_CURRENTSCALETHRESHOLDS 0xFD46
#define ATHEROS_RID_PROTOCOLRSPTIME 0xFD47
#define ATHEROS_RID_SHORTRETRYLIMIT 0xFD48
#define ATHEROS_RID_LONGRETRYLIMIT 0xFD49
#define ATHEROS_RID_MAXTRANSMITLIFETIME 0xFD4A
#define ATHEROS_RID_MAXRECEIVELIFETIME 0xFD4B
#define ATHEROS_RID_CFPOLLABLE 0xFD4C
#define ATHEROS_RID_AUTHENTICATIONALGORITHMS 0xFD4D
#define ATHEROS_RID_PRIVACYOPTIONIMPLEMENTED 0xFD4F
#define ATHEROS_RID_DBMCOMMSQUALITY 0xFD51 /* added in STA f/w 1.3.1 */
#define ATHEROS_RID_CURRENTTXRATE1 0xFD80
#define ATHEROS_RID_CURRENTTXRATE2 0xFD81
#define ATHEROS_RID_CURRENTTXRATE3 0xFD82
#define ATHEROS_RID_CURRENTTXRATE4 0xFD83
#define ATHEROS_RID_CURRENTTXRATE5 0xFD84
#define ATHEROS_RID_CURRENTTXRATE6 0xFD85
#define ATHEROS_RID_OWNMACADDR 0xFD86
#define ATHEROS_RID_SCANRESULTSTABLE 0xFD88 /* added in STA f/w 0.8.3 */
#define ATHEROS_RID_HOSTSCANRESULTS 0xFD89 /* added in STA f/w 1.3.1 */
#define ATHEROS_RID_AUTHENTICATIONUSED 0xFD8A /* added in STA f/w 1.3.4 */
#define ATHEROS_RID_PHYTYPE 0xFDC0
#define ATHEROS_RID_CURRENTCHANNEL 0xFDC1
#define ATHEROS_RID_CURRENTPOWERSTATE 0xFDC2
#define ATHEROS_RID_CCAMODE 0xFDC3
#define ATHEROS_RID_SUPPORTEDDATARATES 0xFDC6
#define ATHEROS_RID_BUILDSEQ 0xFFFE
#define ATHEROS_RID_FWID 0xFFFF

/* ATHEROS Information frames */
#define ATHEROS_INFO_COMMTALLIES 0xF100
#define ATHEROS_INFO_SCANRESULTS 0xF101
#define ATHEROS_INFO_CHANNELINFORESULTS 0xF102 /* AP f/w only */
#define ATHEROS_INFO_HOSTSCANRESULTS 0xF103
#define ATHEROS_INFO_LINKSTATUS 0xF200

enum { ATHEROS_LINKSTATUS_CONNECTED = 1,
       ATHEROS_LINKSTATUS_DISCONNECTED = 2,
       ATHEROS_LINKSTATUS_AP_CHANGE = 3,
       ATHEROS_LINKSTATUS_AP_OUT_OF_RANGE = 4,
       ATHEROS_LINKSTATUS_AP_IN_RANGE = 5,
       ATHEROS_LINKSTATUS_ASSOC_FAILED = 6 };

enum { ATHEROS_PORTTYPE_BSS = 1, ATHEROS_PORTTYPE_WDS = 2,
       ATHEROS_PORTTYPE_PSEUDO_IBSS = 3, ATHEROS_PORTTYPE_IBSS = 0,
       ATHEROS_PORTTYPE_HOSTAP = 6 };

#define ATHEROS_RATES_1MBPS BIT(0)
#define ATHEROS_RATES_2MBPS BIT(1)
#define ATHEROS_RATES_5MBPS BIT(2)
#define ATHEROS_RATES_11MBPS BIT(3)

#define ATHEROS_WEPFLAGS_PRIVACYINVOKED BIT(0)
#define ATHEROS_WEPFLAGS_EXCLUDEUNENCRYPTED BIT(1)
#define ATHEROS_WEPFLAGS_HOSTENCRYPT BIT(4)
#define ATHEROS_WEPFLAGS_HOSTDECRYPT BIT(7)

#define ATHEROS_RX_STATUS_MSGTYPE (BIT(15) | BIT(14) | BIT(13))
#define ATHEROS_RX_STATUS_PCF BIT(12)
#define ATHEROS_RX_STATUS_MACPORT (BIT(10) | BIT(9) | BIT(8))
#define ATHEROS_RX_STATUS_UNDECR BIT(1)
#define ATHEROS_RX_STATUS_FCSERR BIT(0)

enum { ATHEROS_RX_MSGTYPE_NORMAL = 0, ATHEROS_RX_MSGTYPE_RFC1042 = 1,
       ATHEROS_RX_MSGTYPE_BRIDGETUNNEL = 2, ATHEROS_RX_MSGTYPE_MGMT = 4 };


#define ATHEROS_TX_CTRL_ALT_RTRY BIT(5)
#define ATHEROS_TX_CTRL_802_11 BIT(3)
#define ATHEROS_TX_CTRL_802_3 0
#define ATHEROS_TX_CTRL_TX_EX BIT(2)
#define ATHEROS_TX_CTRL_TX_OK BIT(1)

#define ATHEROS_TX_STATUS_RETRYERR BIT(0)
#define ATHEROS_TX_STATUS_AGEDERR BIT(1)
#define ATHEROS_TX_STATUS_DISCON BIT(2)
#define ATHEROS_TX_STATUS_FORMERR BIT(3)

/* HFA3861/3863 (BBP) Control Registers */
#define HFA386X_CR_A_D_TEST_MODES2 0x1A
#define HFA386X_CR_MANUAL_TX_POWER 0x3E

/* IEEE 802.11 defines */

#define WLAN_FC_PVER (BIT(1) | BIT(0))
#define WLAN_FC_TODS BIT(8)
#define WLAN_FC_FROMDS BIT(9)
#define WLAN_FC_MOREFRAG BIT(10)
#define WLAN_FC_RETRY BIT(11)
#define WLAN_FC_PWRMGT BIT(12)
#define WLAN_FC_MOREDATA BIT(13)
#define WLAN_FC_ISWEP BIT(14)
#define WLAN_FC_ORDER BIT(15)

#define WLAN_FC_GET_TYPE(fc) (((fc) & (BIT(3) | BIT(2))) >> 2)
#define WLAN_FC_GET_STYPE(fc) \
	(((fc) & (BIT(7) | BIT(6) | BIT(5) | BIT(4))) >> 4)

#define WLAN_GET_SEQ_FRAG(seq) ((seq) & (BIT(3) | BIT(2) | BIT(1) | BIT(0)))
#define WLAN_GET_SEQ_SEQ(seq) \
	(((seq) & (~(BIT(3) | BIT(2) | BIT(1) | BIT(0)))) >> 4)

#define WLAN_FC_TYPE_MGMT 0
#define WLAN_FC_TYPE_CTRL 1
#define WLAN_FC_TYPE_DATA 2

/* management */
#define WLAN_FC_STYPE_ASSOC_REQ 0
#define WLAN_FC_STYPE_ASSOC_RESP 1
#define WLAN_FC_STYPE_REASSOC_REQ 2
#define WLAN_FC_STYPE_REASSOC_RESP 3
#define WLAN_FC_STYPE_PROBE_REQ 4
#define WLAN_FC_STYPE_PROBE_RESP 5
#define WLAN_FC_STYPE_BEACON 8
#define WLAN_FC_STYPE_ATIM 9
#define WLAN_FC_STYPE_DISASSOC 10
#define WLAN_FC_STYPE_AUTH 11
#define WLAN_FC_STYPE_DEAUTH 12

/* control */
#define WLAN_FC_STYPE_PSPOLL 10
#define WLAN_FC_STYPE_RTS 11
#define WLAN_FC_STYPE_CTS 12
#define WLAN_FC_STYPE_ACK 13
#define WLAN_FC_STYPE_CFEND 14
#define WLAN_FC_STYPE_CFENDACK 15

/* data */
#define WLAN_FC_STYPE_DATA 0
#define WLAN_FC_STYPE_DATA_CFACK 1
#define WLAN_FC_STYPE_DATA_CFPOLL 2
#define WLAN_FC_STYPE_DATA_CFACKPOLL 3
#define WLAN_FC_STYPE_NULLFUNC 4
#define WLAN_FC_STYPE_CFACK 5
#define WLAN_FC_STYPE_CFPOLL 6
#define WLAN_FC_STYPE_CFACKPOLL 7

/* Authentication algorithms */
#define WLAN_AUTH_OPEN 0
#define WLAN_AUTH_SHARED_KEY 1

#define WLAN_AUTH_CHALLENGE_LEN 128

#define WLAN_CAPABILITY_ESS BIT(0)
#define WLAN_CAPABILITY_IBSS BIT(1)
#define WLAN_CAPABILITY_CF_POLLABLE BIT(2)
#define WLAN_CAPABILITY_CF_POLL_REQUEST BIT(3)
#define WLAN_CAPABILITY_PRIVACY BIT(4)

/* Status codes */
#define WLAN_STATUS_SUCCESS 0
#define WLAN_STATUS_UNSPECIFIED_FAILURE 1
#define WLAN_STATUS_CAPS_UNSUPPORTED 10
#define WLAN_STATUS_REASSOC_NO_ASSOC 11
#define WLAN_STATUS_ASSOC_DENIED_UNSPEC 12
#define WLAN_STATUS_NOT_SUPPORTED_AUTH_ALG 13
#define WLAN_STATUS_UNKNOWN_AUTH_TRANSACTION 14
#define WLAN_STATUS_CHALLENGE_FAIL 15
#define WLAN_STATUS_AUTH_TIMEOUT 16
#define WLAN_STATUS_AP_UNABLE_TO_HANDLE_NEW_STA 17
#define WLAN_STATUS_ASSOC_DENIED_RATES 18
/* 802.11b */
#define WLAN_STATUS_ASSOC_DENIED_NOSHORT 19
#define WLAN_STATUS_ASSOC_DENIED_NOPBCC 20
#define WLAN_STATUS_ASSOC_DENIED_NOAGILITY 21

/* Reason codes */
#define WLAN_REASON_UNSPECIFIED 1
#define WLAN_REASON_PREV_AUTH_NOT_VALID 2
#define WLAN_REASON_DEAUTH_LEAVING 3
#define WLAN_REASON_DISASSOC_DUE_TO_INACTIVITY 4
#define WLAN_REASON_DISASSOC_AP_BUSY 5
#define WLAN_REASON_CLASS2_FRAME_FROM_NONAUTH_STA 6
#define WLAN_REASON_CLASS3_FRAME_FROM_NONASSOC_STA 7
#define WLAN_REASON_DISASSOC_STA_HAS_LEFT 8
#define WLAN_REASON_STA_REQ_ASSOC_WITHOUT_AUTH 9


/* Information Element IDs */
#define WLAN_EID_SSID 0
#define WLAN_EID_SUPP_RATES 1
#define WLAN_EID_FH_PARAMS 2
#define WLAN_EID_DS_PARAMS 3
#define WLAN_EID_CF_PARAMS 4
#define WLAN_EID_TIM 5
#define WLAN_EID_IBSS_PARAMS 6
#define WLAN_EID_CHALLENGE 16


#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"

/* netdevice private ioctls (used, e.g., with iwpriv from user space) */

#define ATHEROS_IOCTL_MONITOR (SIOCDEVPRIVATE)
#define ATHEROS_IOCTL_ATHEROS_PARAM (SIOCDEVPRIVATE + 1)
#define ATHEROS_IOCTL_READMIF (SIOCDEVPRIVATE + 2)
#define ATHEROS_IOCTL_WRITEMIF (SIOCDEVPRIVATE + 3)
#define ATHEROS_IOCTL_RESET (SIOCDEVPRIVATE + 4)
#define ATHEROS_IOCTL_INQUIRE (SIOCDEVPRIVATE + 5)
#define ATHEROS_IOCTL_WDS_ADD (SIOCDEVPRIVATE + 6)
#define ATHEROS_IOCTL_WDS_DEL (SIOCDEVPRIVATE + 7)
#define ATHEROS_IOCTL_SET_RID_WORD (SIOCDEVPRIVATE + 8)
#define ATHEROS_IOCTL_MACCMD (SIOCDEVPRIVATE + 9)
#define ATHEROS_IOCTL_ADDMAC (SIOCDEVPRIVATE + 10)
#define ATHEROS_IOCTL_DELMAC (SIOCDEVPRIVATE + 11)
#define ATHEROS_IOCTL_KICKMAC (SIOCDEVPRIVATE + 12)
#define ATHEROS_IOCTL_DOWNLOAD (SIOCDEVPRIVATE + 13)
#define ATHEROS_IOCTL_TUNNEL_ENDPOINT (SIOCDEVPRIVATE + 14)
#define ATHEROS_IOCTL_TUN_PARAMS (SIOCDEVPRIVATE + 15)

/* ATHEROS_IOCTL_ATHEROS_PARAM ioctl() subtypes: */
enum {
	ATHEROS_PARAM_PTYPE = 1,
	ATHEROS_PARAM_TXRATECTRL = 2,
	ATHEROS_PARAM_BEACON_INT = 3,
	ATHEROS_PARAM_PSEUDO_IBSS = 4,
	ATHEROS_PARAM_ALC = 5,
	ATHEROS_PARAM_TXPOWER = 6,
	ATHEROS_PARAM_DUMP = 7,
	ATHEROS_PARAM_OTHER_AP_POLICY = 8,
	ATHEROS_PARAM_AP_MAX_INACTIVITY = 9,
	ATHEROS_PARAM_AP_BRIDGE_PACKETS = 10,
	ATHEROS_PARAM_DTIM_PERIOD = 11,
	ATHEROS_PARAM_AP_NULLFUNC_ACK = 12,
	ATHEROS_PARAM_MAX_WDS = 13,
	ATHEROS_PARAM_AP_AUTOM_AP_WDS = 14,
	ATHEROS_PARAM_AP_AUTH_ALGS = 15,
	ATHEROS_PARAM_MONITOR_ALLOW_FCSERR = 16,
	ATHEROS_PARAM_HOST_ENCRYPT = 17,
	ATHEROS_PARAM_HOST_DECRYPT = 18,
	ATHEROS_PARAM_BUS_MASTER_THRESHOLD_RX = 19,
	ATHEROS_PARAM_BUS_MASTER_THRESHOLD_TX = 20
};


struct atheros_download_param {
	u32 dl_cmd;
	u32 start_addr;
	u32 num_areas;
	struct atheros_download_area {
		u32 addr; /* wlan card address */
		u32 len;
		caddr_t ptr; /* pointer to data in user space */
	} data[0];
};

#define ATHEROS_MAX_DOWNLOAD_AREA_LEN 131072
#define ATHEROS_MAX_DOWNLOAD_LEN 262144

#ifdef __KERNEL__

#define ATHEROS_DATA_MAXLEN MAX_WLAN_BODY_SIZE //2304
#define MAX_SSID_LEN 32
#define MAX_NAME_LEN 32 /* this is assumed to be equal to MAX_SSID_LEN */

#define ATHEROS_DUMP_RX_HDR 1
#define ATHEROS_DUMP_TX_HDR 2

typedef struct atheros_local_info atheros_local_info_t;

struct atheros_helper_functions {
	int (*card_present)(atheros_local_info_t *local);
	void (*cor_sreset)(atheros_local_info_t *local);
	int (*dev_open)(atheros_local_info_t *local);
	int (*dev_close)(atheros_local_info_t *local);
  A_STATUS (*set_essid)(char *);

    struct asap_dev_itf asap;
};

struct atheros_local_info {
	int card_idx;
	int dev_enabled;
#if defined(ATHEROS_PLX) || defined(ATHEROS_PCI)
	unsigned long attr_mem;
	unsigned int cor_offset;
#endif /* ATHEROS_PLX || ATHEROS_PCI */
  WLAN_DEV_INFO *pWlanDev;
  struct net_device *dev;
  struct sk_buff_head ps_q;
  struct sk_buff_head tx_q;
	spinlock_t cmdlock, baplock, wdslock;
	struct ap_data *ap;
	int manfid; /* from CISTPL_MANFID */
	int prodid; /* from CISTPL_MANFID */
	int is_symbol; /* Symbol-based card */
	int is_lucent; /* Lucent-based card */

	char essid[MAX_SSID_LEN + 1];
	char name[MAX_NAME_LEN + 1];
	int name_set;
	u16 channel_mask;
	//fadi struct comm_tallies_sums comm_tallies;
	struct net_device_stats stats;
	struct proc_dir_entry *proc;
	int iw_mode; /* operating mode (IW_MODE_*) */
	int pseudo_adhoc; /* 0: IW_MODE_ADHOC is real 802.11 compliant IBSS
			   * 1: IW_MODE_ADHOC is "pseudo IBSS" */
	char bssid[ETH_ALEN];
	int channel;
	int beacon_int;
	int dtim_period;
	int frame_dump; /* dump RX/TX frame headers, ATHEROS_DUMP_ flags */
	int fw_tx_rate_control;
	u16 tx_rate_control;
	int hw_resetting;
	int hw_ready;
	int hw_reset_tries; /* how many times reset has been tried */

	enum {
		ATHEROS_TXPOWER_AUTO = 0, ATHEROS_TXPOWER_OFF,
		ATHEROS_TXPOWER_FIXED, ATHEROS_TXPOWER_UNKNOWN
	} txpower_type;
	int txpower; /* if txpower_type == ATHEROS_TXPOWER_FIXED */

	/* skb queue for packets to be send using dev_queue_xmit() after
	 * exiting hard IRQ handler (atheros_rx) */
	struct sk_buff_head bridge_list;
	struct tq_struct bridge_queue;

	int manual_retry_count; /* -1 = use f/w default; otherwise retry count
				 * to be used with all frames */

	//struct iw_statistics wstats;

	u32 wep_iv;
#define WEP_KEYS 4
#define WEP_KEY_LEN 13
	u8 wep_keys[WEP_KEYS][WEP_KEY_LEN + 1];
	u8 wep_key_len[WEP_KEYS];
	int wep_tx_key;
	int use_wep;
	int open_wep; /* allow unencrypted frames */
	int host_encrypt;
	int host_decrypt;

	struct atheros_helper_functions *func;

	u8 *pda;
	int fw_ap;
  void *config;
};


#if 1

#define DEBUG_FID BIT(0)
#define DEBUG_PS BIT(1)
#define DEBUG_FLOW BIT(2)
#define DEBUG_AP BIT(3)
#define DEBUG_HW BIT(4)
#define DEBUG_EXTRA BIT(5)
#define DEBUG_EXTRA2 BIT(6)
#define DEBUG_PS2 BIT(7)
#define DEBUG_MASK (DEBUG_PS | DEBUG_FLOW | DEBUG_AP | DEBUG_HW | \
	DEBUG_EXTRA)
#define PDEBUG(n, args...) \
do { if ((n) & DEBUG_MASK) printk(KERN_DEBUG args); } while (0)
#define PDEBUG2(n, args...) \
do { if ((n) & DEBUG_MASK) printk(args); } while (0)

#else

#define PDEBUG(n, args...)
#define PDEBUG2(n, args...)

#endif

int atheros_hw_config(struct net_device *dev, int initial);
void atheros_hw_shutdown(struct net_device *dev, int no_disable);
void atheros_hw_reset(struct net_device *dev);
int atheros_reset_port(struct net_device *dev);
atheros_local_info_t *atheros_init_local_data(struct atheros_helper_functions *funcs,
				     int card_idx, unsigned long mem_base);
int atheros_init_dev(atheros_local_info_t *local);
void atheros_free_local_data(atheros_local_info_t *local);

/* export these prototypes */
A_UINT16 atheros_get_channel (void);
A_STATUS atheros_set_channel (A_UINT16 channel);
A_UINT16 atheros_get_ammode  (void);
void     atheros_set_ammode  (monitor_type mode);

#endif /* __KERNEL__ */

#endif /* ATHEROS_WLAN_H */
