/*
 * Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * config.h - ar311 configuration header file
 *
 * $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/config.h#1 $
 */

/* Copyright 1984-1999 Wind River Systems, Inc. */

#ifndef INCconfigh
#define INCconfigh

#define BSP_VER_1_1     1       /* 1.2 is backward compatible with 1.1 */
#define BSP_VER_1_2     1
#undef BSP_VERSION
#define BSP_REV         ""      /* we do not need "/0" */

#include "configAll.h"          /* Set the VxWorks default configuration */

#if !defined(_ASMLANGUAGE) && defined(PRJ_BUILD)
/* work around for conflict with tffs w/o c++ included */
#include "private/cplusLibP.h"
#endif

#define DEFAULT_BOOT_LINE \
        "tffs:(0,0):/fl/APIMG1 f=0x00 e=192.168.1.20:0xffffff00 o=ae"

/* memory configuration */
#define USER_RESERVED_MEM       0               /* see sysMemTop() */
#define LOCAL_MEM_AUTOSIZE                      /* run-time memory sizing */

#define CPU_CLOCK_RATE          sysBoardData.cntFreq

/*
 * The constants ROM_TEXT_ADRS, ROM_SIZE, RAM_HIGH_ADRS, and RAM_LOW_ADRS
 * are defined in config.h and Makefile and the Tornado .prj files.
 * All definitions for these constants must be identical.
 */
#define ROM_BASE_ADRS           0x9fc00000      /* base address of ROM */
#define ROM_TEXT_ADRS           ROM_BASE_ADRS   /* with PC & SP */
#define ROM_WARM_ADRS           (ROM_TEXT_ADRS+8) /* warm reboot entry */
#define ROM_SIZE                sysRomSize()    /* calculate at run time */
#define RAM_HIGH_ADRS           0x80700000      /* RAM address for bootrom */
#define RAM_MID_ADRS            0x80480000      /* RAM address for compressed vxWorks image */
#define RAM_LOW_ADRS            0x80010000      /* RAM address for vxWorks */
#define RAM_EXC_STACK           (RAM_LOW_ADRS-8)/* Early exception stack */

/* Low memory map flags */
#define RAM_ICE_MAGIC_ADDR      0xa0000400      /* Magic left by ICE init */
#define RAM_ICE_MAGIC_VAL       0xabcdef01
#define SYS_NMI_CALLBACK        0xa0000410      /* NMI callback */


/* Serial port configuration */
#define INCLUDE_SERIAL
#undef  NUM_TTY
#define NUM_TTY                 N_SIO_CHANNELS  /* defined in ar5311.h */
#define CONSOLE_ADDR            AR5311_UART0    /* second serial port */
#undef  CONSOLE_TTY
#define CONSOLE_TTY             0
#undef  CONSOLE_BAUD_RATE
#define CONSOLE_BAUD_RATE       9600

/* Timer configuration */
#define INCLUDE_TIMESTAMP                       /* 4Kp count is used */

/* Cache support */
#define INCLUDE_CACHE_SUPPORT   /* enable cacheLib support */
#ifdef INCLUDE_CACHE_SUPPORT
#define USER_I_CACHE_ENABLE
#undef USER_I_CACHE_MODE
#define USER_I_CACHE_MODE       CACHE_WRITETHROUGH
#define USER_D_CACHE_ENABLE
#undef USER_D_CACHE_MODE
#define USER_D_CACHE_MODE       (CACHE_WRITETHROUGH|CACHE_NO_WRITEALLOCATE)
#undef  USER_B_CACHE_ENABLE
#else
#undef  USER_I_CACHE_ENABLE
#undef  USER_I_CACHE_MODE
#define USER_I_CACHE_MODE       CACHE_DISABLED
#undef  USER_D_CACHE_ENABLE
#undef  USER_D_CACHE_MODE
#define USER_D_CACHE_MODE       CACHE_DISABLED
#undef  USER_B_CACHE_ENABLE
#endif


#define INCLUDE_SHELL

/*
 * Default use ae1 as the debug port. Can be changed
 * to ae0 by changing the define below to 0.
 */
#ifdef DEBUG
#define WDB_COMM_END_ENTRY  1
#endif

#undef WDB_POOL_SIZE    /* Handle K0BASE vs 0 for various memory types */
#define WDB_POOL_SIZE   (((char *)PHYS_TO_K0(sysMemTop()) - FREE_RAM_ADRS) / 16)

/* MMU configuration */
#undef  INCLUDE_MMU_BASIC       /* no MMU support */
#undef  INCLUDE_MMU_FULL        /* no MMU support */

/* Flash and flash file system support */
#define INCLUDE_TFFS                    /* requires optiona TrueFFS package */
#ifdef INCLUDE_TFFS
#ifndef INCLUDE_DOSFS
#define INCLUDE_DOSFS
#endif
#undef INCLUDE_TFFS_BOOT_IMAGE          /* for boot block */
#define INCLUDE_TFFS_SHOW               /* for debugging */
#define INCLUDE_TL_FTL                  /* NOR based flash */
#define INCLUDE_MTD_CFISCS              /* base flash and cfi code */
#ifdef EMULATION
#define INCLUDE_CFI_SCS                 /* Intel part on emulation */
#else
#define INCLUDE_CFI_AMDFJ               /* AMD/FJ part for production */
#endif
#endif

/* Network driver configuration */
#define INCLUDE_NETWORK
#define INCLUDE_NET_INIT
#undef INCLUDE_EX
#undef INCLUDE_ENP
#undef INCLUDE_SM_NET

#define IP_MAX_UNITS    3       /* 2 enet + wlan */

/* Enhanced Network Driver (END) support */
#define INCLUDE_END             /* Enhanced Network Driver (see configNet.h) */
#undef INCLUDE_BSD              /* BSD 4.4 drivers */

#undef MIB2_SYS_DESCR
#undef MIB2_SYS_CONTACT
#undef MIB2_SYS_LOCATION
#undef MIB2_SYS_OBJID_LEN
#undef MIB2_SYS_OBJID
#define MIB2_SYS_DESCR                  "Atheros VxWorks SNMPv1/v2c Agent"
#define MIB2_SYS_CONTACT                "Atheros Communications Inc."
#define MIB2_SYS_LOCATION               "Sunnyvale, California, U.S.A"
#define MIB2_SYS_OBJID_LEN              8
#define MIB2_SYS_OBJID                  {1,3,6,1,4,1,9442,1}

/*
 * User application initialization
 *
 * USER_APPL_INIT must be a valid C statement or block.  It is
 * included in the usrRoot() routine only if INCLUDE_USER_APPL is
 * defined.  The code for USER_APPL_INIT is only an example.  The
 * user is expected to change it as needed.  The use of taskSpawn
 * is recommended over direct execution of the user routine.
 */

#undef  INCLUDE_USER_APPL
#define USER_APPL_INIT \
        { \
        IMPORT int myAppInit(); \
        taskSpawn ("myApp", 30, 0, 5120, \
                   myAppInit, 0x1, 0x2, 0x3, 0,0,0,0,0,0,0); \
        }

/* Configurable interrupt vectors/levels (static defs go in ar5311.h) */

#if     FALSE
#define INT_VEC_ABORT           xxx
#define INT_VEC_CLOCK           xxx
...
#define INT_LVL_ABORT           xxx
#define INT_LVL_CLOCK           xxx
...
#endif

#include "ar5311.h"             /* Verify user options, specify i/o addr etc */

#endif  /* INCconfigh */

#if defined(PRJ_BUILD)
#include "prjParams.h"
#endif

