/*
 *  Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/diags/display.c#1 $"

/*
 * DESCRIPTION
 * This contains functions to display descriptor and packets etc.
 */

#ifdef VXWORKS

#include "VxWorks.h"

#else

//#include <memory.h>
//#include <stdio.h>
//#include <stdlib.h>

#endif

#include "wlantype.h"
#include "wlanproto.h"
#include "wlansme.h"
#include "wlandrv.h"
#include "wlanext.h"
#include "display.h"
#include "ui.h"

/* Name strings for the frame types */

char *mgtTypeStr[] = {
    "ASSOC REQ",            /* 0 */
    "ASSOC RESP",           /* 1 */
    "REASSOC REQ",          /* 2 */
    "REASSOC RESP",         /* 3 */
    "PROBE REQ",            /* 4 */
    "PROBE RESP",           /* 5 */
    "",                     /* 6 */
    "",                     /* 7 */
    "BEACON",               /* 8 */
    "ATIM",                 /* 9 */
    "DISASSOC",             /* a */
    "AUTHENTICATE",         /* b */
    "DEAUTHENTICATE"        /* c */
};

char *ctrlTypeStr[] = {
    "", "", "", "", "",     /* 0 - 4 */
    "", "", "", "", "",     /* 5 - 9 */
    "PS POLL",              /* a */
    "RTS",                  /* b */
    "CTS",                  /* c */
    "ACK",                  /* d */
    "CF END",               /* e */
    "CF END+ACK"            /* f */
};

char    *dataTypeStr[] = {
    "DATA",                 /* 0 */
    "DATA+CF ACK",          /* 1 */
    "DATA+CF POLL",         /* 2 */
    "DATA+CF ACK+POLL",     /* 3 */
    "NULL",                 /* 4 */
    "NULL+CF ACK",          /* 5 */
    "NULL+CF POLL",         /* 6 */
    "NULL+CF ACK+POLL"      /* 7 */
};

/* Name strings for type of service */
char    *servTypeStr[] = {
    "STA",  /* Station WLAN_STA_SERVICE = 0 */
    "AP"    /* Access point WLAN_AP_SERVICE = 1 */
};


/**************************************************************************
* printMacAddress - Print a MAC address
*
*
* RETURNS: N/A
*/
void
printMacAddress(WLAN_MACADDR macAddress)
{
    A_UINT16 i;

    for (i = 0; i < WLAN_MAC_ADDR_SIZE; i++) {
        aPrintk("%02x%s", macAddress.octets[i], (i+1) < WLAN_MAC_ADDR_SIZE ? ":" : "" );
    }
}

/**************************************************************************
* displayFrameControl - Print the Frame control fields of header
*
*
* RETURNS: N/A
*/
void
printFrameControl(FRAME_CONTROL_FIELD *frameControl)
{
    aPrintk(PROTO_VER              "%01x  ", frameControl->protoVer);
    switch (frameControl->fType) {
    case FRAME_MGT:
        aPrintk(MGT_TYPE "  ");
        aPrintk(SUBTYPE                "%s\n", mgtTypeStr[frameControl->fSubtype]);
        break;
    case FRAME_CTRL:
        aPrintk(CTRL_TYPE "  ");
        aPrintk(SUBTYPE                "%s\n", ctrlTypeStr[frameControl->fSubtype]);
        break;
    case FRAME_DATA:
        aPrintk(DATA_TYPE "  ");
        aPrintk(SUBTYPE                "%s\n", dataTypeStr[frameControl->fSubtype]);
        break;
    default:
        aPrintk(TYPE                   "%01x  ", frameControl->fType);
        aPrintk(SUBTYPE                "%01x\n", frameControl->fSubtype);
    }

    aPrintk(TO_DS                  "%01x  ", frameControl->ToDS);
    aPrintk(FROM_DS                "%01x  ", frameControl->FromDS);
    aPrintk(MORE_FRAG              "%01x\n", frameControl->moreFrag);
    aPrintk(RETRY                  "%01x  ", frameControl->retry);
    aPrintk(PWR_MGT                "%01x  ", frameControl->pwrMgt);
    aPrintk(MORE_DATA              "%01x\n", frameControl->moreData);
    aPrintk(WEP                    "%01x  ", frameControl->wep);
    aPrintk(ORDER                  "%01x  ", frameControl->order);
    return;
}

/**************************************************************************
 * displayCapInfo - displays various fields of Capabilities Info structure
 *
 * Takes pointer to the CAP_INFO structure (which is only 2 bytes) and prints
 * the bit fields.
 *
 * RETURNS: N/A
 */
void
displayCapInfo(CAP_INFO *capInfo)
{
    aPrintk("Capability Information\n");
    aPrintk(CF_POLLABLE "%01x  ", capInfo->cfPollable);
    aPrintk(CF_POLLREQ  "%01x   ", capInfo->cfPollReq);
    aPrintk(ESS_CAP     "%01x   ", capInfo->ess);
    aPrintk(IBSS_CAP    "%01x\n", capInfo->ibss);
    aPrintk(WEP         "%01x\n", capInfo->privacy);
}

/**************************************************************************
 * displayInfoElement - displays Information elements in mgt frame
 *
 * Takes pointer to the INFO_ELEMENT structure and prints the element ID,
 * length and variable length value in hex dump.
 *
 * RETURNS: N/A
 */
void
displayInfoElement(INFO_ELEMENT *infoEle)
{
    int     i, j;

    aPrintk("INFO ELEMENT  ");
    if (infoEle == NULL) {
        aPrintk(" is NULL\n");
        return;
    }
    switch (infoEle->elementID) {
    case ELE_SSID:
        aPrintk(SSID_VALUE);
        break;

    case ELE_SUPPORTED_RATES:
        aPrintk(SUPP_RATES);
        break;

    case ELE_FH_PARAM_SET:
        aPrintk(FH_SET);
        break;

    case ELE_DS_PARAM_SET:
        aPrintk(DS_SET);
        break;

    case ELE_CF_PARAM_SET:
        aPrintk(CF_SET);
        break;

    case ELE_TIM:
        aPrintk(TIM_DISP);
        break;

    case ELE_IBSS_PARAM_SET:
        aPrintk(IBSS_DISP);
        break;

    case ELE_CHLG_TEXT:
        aPrintk(CHLG_DISP);
        break;

    case ELE_UCSE_SET:
        aPrintk(UCSE_DISP);
        break;

    case ELE_MCSE_SET:
        aPrintk(MCSE_DISP);
        break;

    case ELE_COUNTRY_INFO:
        aPrintk(COUNTRY_INFO_DISP);
        break;

    default:
        aPrintk("%15s %4d", UNKNOWN_DISP, infoEle->elementID);
        break;
    }
    aPrintk("Length  0x%02x bytes\n", infoEle->length);

    /* Now dump the value field */
    aPrintk("Value   ");
    i = 0;
    while(i < infoEle->length) {
        for (j = 0; i < infoEle->length && j < 16; j++, i++) {
            aPrintk("%02x ", infoEle->value[i]);
        }
        if (i < (infoEle->length - 1)) {
            aPrintk("\n        ");
        }
    }
    aPrintk("\n");
}

/**************************************************************************
 * displayInfoElementShort - displays Information elements (one per line)
 *
 * Takes pointer to the INFO_ELEMENT structure and prints the element ID,
 * length and variable length value in hex dump, all in one line
 *
 * RETURNS: N/A
 */
void
displayInfoElementShort(INFO_ELEMENT *infoEle)
{
    int     i;

    aPrintk("ELEM ");
    switch (infoEle->elementID) {
    case ELE_SSID:
        aPrintk(SSID_VALUE);
        break;

    case ELE_SUPPORTED_RATES:
        aPrintk(SUPP_RATES);
        break;

    case ELE_FH_PARAM_SET:
        aPrintk(FH_SET);
        break;

    case ELE_DS_PARAM_SET:
        aPrintk(DS_SET);
        break;

    case ELE_CF_PARAM_SET:
        aPrintk(CF_SET);
        break;

    case ELE_TIM:
        aPrintk(TIM_DISP);
        break;

    case ELE_IBSS_PARAM_SET:
        aPrintk(IBSS_DISP);
        break;

    case ELE_CHLG_TEXT:
        aPrintk(CHLG_DISP);
        break;

    default:
        aPrintk("%15s %4d", UNKNOWN_DISP, infoEle->elementID);
        break;
    }
    aPrintk("0x%02x = ", infoEle->length);

    /* Now dump the value field */
    i = 0;
    while(i < (infoEle->length) && i < 14) {
        aPrintk("%02x ", infoEle->value[i++]);
    }
    if (infoEle->length >= 14) {
        aPrintk("...");
    }
    aPrintk("\n");
}

/**************************************************************************
 * displayMgtFrameHdr - Display header of a management frame
 *
 * Displays only the frame header of a management frame.
 *
 * RETURNS: N/A
 */
void
displayMgtFrameHdr(WLAN_MGT_MAC_HEADER *pmacHdr)
{
    printFrameControl( &pmacHdr->frameControl );
    aPrintk(DUR_ID                 "%04x\n",
             pmacHdr->durationNav.duration);
    aPrintk(DEST_ADDR ); printMacAddress(pmacHdr->destAddr); aPrintk("   ");
    aPrintk(SRC_ADDR  ); printMacAddress(pmacHdr->srcAddr); aPrintk("\n");
    aPrintk(BSSID_ADDR); printMacAddress(pmacHdr->bssId); aPrintk("\n");
    aPrintk(SEQ_NUM                "%04x  ", pmacHdr->sequenceControl.seqNo);
    aPrintk(FRAG_NUM               "%01x  \n", pmacHdr->sequenceControl.fragNo);
}

/**************************************************************************
 * displayMgtFrame - Display a complete management frame
 *
 * Displays a complete management frame. It prints the standard mgt frame
 * header, followed by the known fixed size fields. Then it positions p1
 * at the beginning of the variable length INFO_ELEMENT field. A loop at the
 * end steps thru all the info elements and prints them.
 * PARAMETERS:
 * pmacHdr - pointer to frame
 * frameSize - length of the frame, not including fcs, but may include iv.
 *
 * RETURNS: N/A
 */
void
displayMgtFrame(WLAN_MGT_MAC_HEADER *pmacHdr, A_UINT16 frameSize)
{
    A_UINT16        subtype;
    A_UINT8         *p1;            /* Pointer to beginning of info element field */
    A_UINT16        len, elen;

    if (pmacHdr->frameControl.fType != FRAME_MGT) {
		aPrintk("FRAME IS NOT MGT FRAME!\n");
        return;
    }

    displayMgtFrameHdr(pmacHdr);

    subtype = (A_UINT16) pmacHdr->frameControl.fSubtype;
    switch (subtype) {
    case SUBT_ASSOC_REQ:
        {
            /* p2 is conveniently typecast pointer */
            WLAN_FRAME_ASSOC_REQ *p2 = (WLAN_FRAME_ASSOC_REQ *) pmacHdr;
            displayCapInfo(&(p2->capInfo));
            aPrintk(LISTEN_INT "%04x\n", p2->listenInterval);
            p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_REQ *)pmacHdr)->ssid);
        }
        break;

    case SUBT_ASSOC_RESP:
    case SUBT_REASSOC_RESP:
        {
            WLAN_FRAME_ASSOC_RESP *p2 = (WLAN_FRAME_ASSOC_RESP *) pmacHdr;
            displayCapInfo(&(p2->capInfo));
            aPrintk(STATUS_CODE "%04x  ", p2->statusCode);
            aPrintk(ASSOC_ID "%04x\n", p2->assocId & 0x3fff);
            p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_RESP *)pmacHdr)->rateSet);
        }
        break;

    case SUBT_REASSOC_REQ:
        {
            WLAN_FRAME_REASSOC_REQ *p2 = (WLAN_FRAME_REASSOC_REQ *) pmacHdr;
            displayCapInfo(&(p2->capInfo));
            aPrintk(LISTEN_INT "%04x", p2->listenInterval);
            aPrintk(CURAP_ADDR); printMacAddress(p2->curAPAddr); aPrintk("\n");
            p1 = (A_UINT8 *) &(((WLAN_FRAME_REASSOC_REQ *)pmacHdr)->ssid);
        }
        break;

    case SUBT_PROBE_REQ:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_REQ *)pmacHdr)->ssid);
        break;

    case SUBT_PROBE_RESP:
    case SUBT_BEACON:
        {
            WLAN_FRAME_PROBE_RESP *p2 = (WLAN_FRAME_PROBE_RESP *) pmacHdr;
            A_UINT32        val;
            A_DRIVER_BCOPY(&(p2->timestamp.high), &val, 4);
            aPrintk(TIMESTAMP_STR "%08x", val);
            A_DRIVER_BCOPY(&(p2->timestamp.low), &val, 4);
            aPrintk(TIMESTAMP_STR "%08x  ", val);
            aPrintk(BEACON_INTRV "%04x\n", p2->beaconInterval);
            displayCapInfo(&(p2->capInfo));
            p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_RESP *)pmacHdr)->buffer);
            break;
        }
    case SUBT_AUTH:
        {
            if (pmacHdr->frameControl.wep) {
                /* IV field in encrypted frame, use different frame */
                WLAN_FRAME_AUTH_ENCRYPT *p2 =
                    (WLAN_FRAME_AUTH_ENCRYPT *)pmacHdr;
                aPrintk(ALGO_NO "%04x ", p2->algNo);
                aPrintk(TRANS_SEQ "%04x ", p2->transSeqNo);
                aPrintk(STATUS_CODE "%04x\n", p2->statusCode);
                p1 = (A_UINT8 *) &(p2->challenge);
            } else {
                /* regular AUTH frame */
                WLAN_FRAME_AUTH_CHLG    *p2 = (WLAN_FRAME_AUTH_CHLG *)pmacHdr;
                aPrintk(ALGO_NO "%04x ", p2->algNo);
                aPrintk(TRANS_SEQ "%04x ", p2->transSeqNo);
                aPrintk(STATUS_CODE "%04x\n", p2->statusCode);
                p1 = (A_UINT8 *) &(p2->challenge);
            }
            break;
        }
    case SUBT_DEAUTH:
    case SUBT_DISASSOC:
        {
            WLAN_FRAME_DEAUTH       *p2 = (WLAN_FRAME_DEAUTH *)pmacHdr;
            aPrintk(REASON_CODE "%04x\n", p2->reasonCode);
        }
    default:
        return;
    }

    /* Now, p1 is the beginning of the variable size fields */
    /* len is the offset into the frame */
    len = (int) (p1 - (A_UINT8 *)pmacHdr);

    /* A valid element should have at least 3 bytes - elementID,
     * length and one value byte. If not, we stop. Last 4 bytes of frame are FCS.
     */
    while (len < (frameSize - 3)) {
        displayInfoElement((INFO_ELEMENT *) p1);
        /* Increment len by no. of bytes in this element.
         * Element Id and Length byte are added to the value
         * of length field. Even if length field is 0,
         * this will increment len by 2. Don't use INFO_ELEMENT_SIZE
         * macro, it will return 0 if length is 0.
         */
        elen = ((INFO_ELEMENT *)p1)->length +
            sizeof(A_UINT8) + sizeof(A_UINT8);
        len += elen;
        p1 += elen;
    }

    return;
}

/**************************************************************************
 * displayMgtFrameShort - Display a complete management frame, short display
 *
 * Displays a complete management frame but data is not expanded much
 *
 * RETURNS: N/A
 */

void
displayMgtFrameShort(WLAN_MGT_MAC_HEADER *pmacHdr, A_UINT16 frameSize)
{
    A_UINT16        subtype;
    A_UINT8         *p1;
    A_UINT16        len, elen;

    if (pmacHdr->frameControl.fType != FRAME_MGT)
        return;

    aPrintk("MGT ");
    aPrintk("%20s  ", mgtTypeStr[pmacHdr->frameControl.fSubtype]);
    aPrintk("Frame Control: " "%04x\n", *((A_UINT16 *) &(pmacHdr->frameControl)));
    aPrintk("DUR %04x ", pmacHdr->durationNav.duration);
    aPrintk("ADDR: "); printMacAddress(pmacHdr->destAddr);
    aPrintk(" "); printMacAddress(pmacHdr->srcAddr);
    aPrintk(" "); printMacAddress(pmacHdr->bssId);
    aPrintk(" SEQ/FRAG:" "%04x/%01x\n", pmacHdr->sequenceControl.seqNo,
             pmacHdr->sequenceControl.fragNo);

    subtype = (A_UINT16) pmacHdr->frameControl.fSubtype;
    switch (subtype) {
    case SUBT_ASSOC_REQ:
        {
            WLAN_FRAME_ASSOC_REQ *p2 = (WLAN_FRAME_ASSOC_REQ *) pmacHdr;
            aPrintk("CAP INFO: " "%04x", *((A_UINT16 *) &(p2->capInfo)));
            aPrintk(LISTEN_INT "%04x\n", p2->listenInterval);
            p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_REQ *)pmacHdr)->ssid);
        }
        break;

    case SUBT_ASSOC_RESP:
    case SUBT_REASSOC_RESP:
        {
            WLAN_FRAME_ASSOC_RESP *p2 = (WLAN_FRAME_ASSOC_RESP *) pmacHdr;
            aPrintk("CAP INFO: " "%04x", *((A_UINT16 *) &(p2->capInfo)));
            aPrintk(STATUS_CODE "%04x  ", p2->statusCode);
            aPrintk(ASSOC_ID "%04x\n", p2->assocId & 0x3fff);
            p1 = (A_UINT8 *) &(((WLAN_FRAME_ASSOC_RESP *)pmacHdr)->rateSet);
        }
        break;

    case SUBT_REASSOC_REQ:
        {
            WLAN_FRAME_REASSOC_REQ *p2 = (WLAN_FRAME_REASSOC_REQ *) pmacHdr;
            aPrintk("CAP INFO: " "%04x", *((A_UINT16 *) &(p2->capInfo)));
            aPrintk(LISTEN_INT "%04x", p2->listenInterval);
            aPrintk(CURAP_ADDR); printMacAddress(p2->curAPAddr); aPrintk("\n");
            p1 = (A_UINT8 *) &(((WLAN_FRAME_REASSOC_REQ *)pmacHdr)->ssid);
        }
        break;

    case SUBT_PROBE_REQ:
        p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_REQ *)pmacHdr)->ssid);
        break;

    case SUBT_PROBE_RESP:
    case SUBT_BEACON:
        {
            WLAN_FRAME_PROBE_RESP *p2 = (WLAN_FRAME_PROBE_RESP *) pmacHdr;
            A_UINT32        val;
            A_DRIVER_BCOPY(&(p2->timestamp.high), &val, 4);
            aPrintk(TIMESTAMP_STR "%08lx", val);
            A_DRIVER_BCOPY(&(p2->timestamp.low), &val, 4);
            aPrintk(":%08lx  ", val);
            aPrintk(BEACON_INTRV "%04x\n", p2->beaconInterval);
            aPrintk("CAP INFO: " "%04x\n", *((A_UINT16 *) &(p2->capInfo)));
            p1 = (A_UINT8 *) &(((WLAN_FRAME_PROBE_RESP *)pmacHdr)->buffer);
            break;
        }
    case SUBT_AUTH:
        {
            if (pmacHdr->frameControl.wep) {
                /* IV field in encrypted frame, use different frame */
                WLAN_FRAME_AUTH_ENCRYPT *p2 =
                    (WLAN_FRAME_AUTH_ENCRYPT *)pmacHdr;
                aPrintk(ALGO_NO "%04x ", p2->algNo);
                aPrintk(TRANS_SEQ "%04x ", p2->transSeqNo);
                aPrintk(STATUS_CODE "%04x\n", p2->statusCode);
                p1 = (A_UINT8 *) &(p2->challenge);

            } else {
                /* regulat AUTH frame */
                WLAN_FRAME_AUTH_CHLG    *p2 = (WLAN_FRAME_AUTH_CHLG *)pmacHdr;
                aPrintk(ALGO_NO "%04x ", p2->algNo);
                aPrintk(TRANS_SEQ "%04x ", p2->transSeqNo);
                aPrintk(STATUS_CODE "%04x\n", p2->statusCode);
                p1 = (A_UINT8 *) &(p2->challenge);

            }
            break;
        }
    case SUBT_DEAUTH:
    case SUBT_DISASSOC:
        {
            WLAN_FRAME_DEAUTH       *p2 = (WLAN_FRAME_DEAUTH *)pmacHdr;
            aPrintk(REASON_CODE "%04x\n", p2->reasonCode);
        }
    default:
        return;
    }

    /* Now, p1 is the beginning of the variable size fields */
    /* len is the offset into the frame */
    len = (int) (p1 - (A_UINT8 *)pmacHdr);

    /* A valid element should have at least 3 bytes - elementID,
     * length and one value byte. If not, we stop. Last 4 bytes of frame are FCS.
     */
    while(len < (frameSize - 3)) {
        displayInfoElementShort((INFO_ELEMENT *) p1);
        /* Increment len by no. of bytes in this element.
         * Element Id and Length byte are added to the value
         * of length field. Even if length field is 0,
         * this will increment len by 2. Don't use INFO_ELEMENT_SIZE
         * macro, it will return 0 if length is 0.
         */
        elen = ((INFO_ELEMENT *)p1)->length +
            sizeof(A_UINT8) + sizeof(A_UINT8);
        len += elen;
        p1 += elen;
    }

    return;
}

/**************************************************************************
 * displayCtrlFrame - Display a complete controlframe
 *
 * Displays a complete control frame.
 *
 * RETURNS: N/A
 */

void
displayCtrlFrame(WLAN_MGT_MAC_HEADER *pmacHdr)
{
    printFrameControl(&pmacHdr->frameControl);
    aPrintk(DUR_ID "%08x   ", pmacHdr->durationNav.duration);
    switch (pmacHdr->frameControl.fSubtype) {
    case SUBT_PSPOLL:
        {
            WLAN_FRAME_PSPOLL *p2 = (WLAN_FRAME_PSPOLL *)pmacHdr;
            aPrintk(BSSID_ADDR); printMacAddress(p2->bssId); aPrintk("  ");
            aPrintk(TRANS_ADDR  ); printMacAddress(p2->transAddr); aPrintk("\n");
            break;
        }
    case SUBT_RTS:
        {
            WLAN_FRAME_RTS *p2 = (WLAN_FRAME_RTS *)pmacHdr;
            aPrintk(REC_ADDR); printMacAddress(p2->recAddr); aPrintk("  ");
            aPrintk(TRANS_ADDR  ); printMacAddress(p2->transAddr); aPrintk("\n");
            break;
        }
    case SUBT_CTS:
    case SUBT_ACK:
        {
            WLAN_FRAME_CTS *p2 = (WLAN_FRAME_CTS *)pmacHdr;
            aPrintk(REC_ADDR  ); printMacAddress(p2->recAddr); aPrintk("\n");
            break;
        }
    case SUBT_CFEND:
    case SUBT_CFENDACK:
        {
            WLAN_FRAME_CFEND *p2 = (WLAN_FRAME_CFEND *)pmacHdr;
            aPrintk(REC_ADDR  ); printMacAddress(p2->recAddr); aPrintk("\n");
            aPrintk(BSSID_ADDR); printMacAddress(p2->bssId); aPrintk("  ");
            break;
        }
    }

}

/**************************************************************************
 * displayStates - Display a SIB Station state and transitional state
 *
 * RETURNS: N/A
 */
    void
displayStates(A_UINT32 sState, A_UINT32 tState)
{
    aPrintk("State:");
    if (sState & STATE_AUTH)
        aPrintk(" AUTH");

    if (sState & STATE_ASSOC)
        aPrintk(" ASSOC");

    if (sState & STATE_PS)
        aPrintk(" PS");

    if (sState & STATE_ADHOC)
        aPrintk(" ADHOC");

    if (sState & STATE_JOINED)
        aPrintk(" JOINED");

    if (sState & STATE_AP_UP)
        aPrintk(" AP-UP");

    if (sState & STATE_AP_STARTING)
        aPrintk(" AP-STARTING");

    if (sState & STATE_AP_DOWN)
        aPrintk(" AP-DOWN");

    if (sState & STATE_STA_LOCKED)
        aPrintk(" LOCKED");

    if (sState == 0)
        aPrintk(" None");

    aPrintk(", transState: 0x%08lx\n", tState);
}

/**************************************************************************
 * displaySib - Display a SIB Station Information Block
 *
 * Command exported to user to important fields of SIB , by supplying
 * the pointer.
 *
 * RETURNS: N/A
 */
    void
displaySib(SIB_ENTRY *pSib)
{
    aPrintk("SIB_ENTRY at %p, pNext %p, Index %d\n",
             pSib, pSib->pNext, pSib->staIndex);
    aPrintk(STA_ADDR); printMacAddress(pSib->macAddr); aPrintk(" ");
    aPrintk(BSSID_ADDR); printMacAddress(pSib->bssId); aPrintk("\n");
    aPrintk("Service %4s ",
             (pSib->serviceType == WLAN_STA_SERVICE) ? "STA" : "AP");
    displayStates(pSib->staState, pSib->transState);
}
