
/* Copyright 2002 Atheros Communications Inc */

/* $Id: //depot/sw/releases/2.0.1/src/include/dot1x.h#1 $ */
/* $Author: jrh $ */

#ifdef VXWORKS
#include "net/if_dl.h" // for struct sockaddr_dl
#endif

#define MAX_RADIOS 		1 // XXX 
#define MAX_DOT1X_SESSIONS 	256
#define RADIUS_PORT 		1812
#define MSG_AUTH_LEN 		16

#define PORTMODE_AUTO		1
#define PORTSTATUS_AUTH 	2
#define PORTSTATUS_UNAUTH 	3
#define PORTCONTROL_AUTO 	4

#define FORCE_AUTH		5
#define FORCE_UNAUTH		6
#define WAS_FORCED		7

#define EAP_SOCKPROTO 		0x0723

typedef enum {
    ST_INITIALIZE = 1,
    ST_HELD,
    ST_DISCONNECTED,
    ST_CONNECTING,
    ST_AUTHENTICATING,
    ST_AUTHENTICATED,
    ST_ABORTING,
    ST_FORCEAUTH,
    ST_FORCEUNAUTH,
    ST_IDLE,
    ST_RESPONSE,
    ST_REQUEST,
    ST_FAIL,
    ST_SUCCESS,
    ST_TIMEOUT,
    ST_NOKEYTX,
    ST_KEYTX,
    ST_INORBOTH,
    ST_REAUTHENTICATE,

} DOT1X_STATEVAR;

/*
 *  An EAPOL header
 */
typedef struct eapolHdr_s {
    A_UINT8	eapolProtoVers;
    A_UINT8	eapolType;
    A_UINT16	eapolBodyLength;
} __ATTRIB_PACK EAPOL_HEADER;

#define EAPOL_PROTOVERS			0x01
#define EAPOL_TYPE_PACKET		0x00
#define EAPOL_TYPE_START		0x01
#define	EAPOL_TYPE_LOGOFF		0x02
#define EAPOL_TYPE_KEYMSG		0x03
#define	EAPOL_TYPE_ASF_ALERT		0x04

/* These are control messages that will never go on the air */
/* they're just for the Authenticator's internals */
#define EAPOL_TYPE_PORTENABLED		0xFD
#define EAPOL_TYPE_PORTDISABLED		0xFE

/*
 *  An EAP header, used only if eapolType == EAP_TYPE_PACKET.
 *  Do NOT use sizeof on this...use the constants below
 */
typedef struct eap_s {
    A_UINT8		eapCode;
    A_UINT8		eapIdentifier;
    A_UINT16		eapLength; // inc code and ident fields
    A_UINT8		eapType;
    A_UINT8		eapData[1]; // variable size
} __ATTRIB_PACK EAP;

/*
 *  EAPOL-Key message
 */
typedef struct eapolKey_s {
    A_UINT8	keyType;
    A_UINT16	keyLength;
    A_UINT8	keyReplayCtr[8];
    A_UINT8	keyIV[16];
    A_UINT8	keyIndex;
#define KEY_USAGE_BROADCAST	0x00
#define KEY_USAGE_UNICAST	0x80
    A_UINT8	keySig[16];
    /* secret goes here */
} __ATTRIB_PACK EAPOL_KEY;

#define	EAPOL_KEY_TYPE_RC4		0x01

#define EAP_SUCCFAIL_LEN		0x0004
#define EAP_RQ_IDENTITY_LEN		0x0005

/* values for code field in EAP pkt */
#define EAP_CODE_REQUEST		0x01
#define EAP_CODE_RESPONSE		0x02
#define EAP_CODE_SUCCESS		0x03
#define EAP_CODE_FAILURE		0x04

/* values for type field in EAP pkt */
#define EAP_TYPE_IDENTITY		0x01
#define EAP_TYPE_NOTIFICATION		0x02
#define EAP_TYPE_NAK			0x03	
#define EAP_TYPE_MD5CHALLENGE		0x04
#define EAP_TYPE_EAPTLS			0x0D

#define EAP_MAX_LEN			0xffff
#define ASPROTO_RADIUS 			1

typedef enum {
    CONTROLDIRS_BOTH = 1,
    CONTROLDIRS_IN,
} CONTROLDIRS;

typedef struct dot1xSessionControlBlock_s {
    TAILQ_ENTRY(dot1xSessionControlBlock_s) link;
    A_UINT32		devUnit;   /* device index */
    void *		linkDevCookie; // back ptr to cookie for mux
    int			sockEAPfd; // one per radio
    int			sockASfd;
    WLAN_MACADDR	lAddr;		// local L2 address of radio
    WLAN_MACADDR	fAddr;		// foreign L2 address of EAP peer
    A_UINT8		asProtocol;
    A_UINT32		lastTick;	// in ticks
    A_UINT32		flags;

#define DOT1X_FLAGS_ISFREE	0x0 // currently not on session list
#define DOT1X_FLAGS_INUSE	0x1 // currently on session list
#define DOT1X_FLAGS_WANTFREE	0x2 // want off the session list

    /*
     *  We separate the variables into distinct state machine
     *  groups.  The constants are considered to be global
     *  across all sessions.
     */

    /* These variables are global to all SMs in this session */
    struct StateGlobal_s {
	A_BOOL		initialize;
	A_BOOL		authAbort;
	A_BOOL		authFail;
	A_BOOL		authStart;
	A_BOOL		authTimeout;
	A_BOOL		authSuccess;
	A_UINT8		currentId;
	A_UINT8		portControl;
	A_BOOL		portEnabled;
	A_UINT8		portStatus;
	A_BOOL		reAuthenticate;
	A_INT32		reAuthPeriod;
    } stateGlob;

    /* Authenticator State machine variables */
    struct StateAuth_s {
	DOT1X_STATEVAR	state;
	A_BOOL		eapLogoff;
	A_BOOL		eapStart;
	A_UINT8		portMode;
	A_UINT8		reAuthCount;
	A_UINT8		rxRespId;
    } stateAuth;

    /* Backend Auth state machine variables */
    struct StateAS_s {
	DOT1X_STATEVAR	state;
	A_UINT8		reqCount;
	A_BOOL		rxResp;
	A_BOOL		aSuccess;
	A_BOOL		aFail;
	A_BOOL		aReq;
    } stateAS;

    struct KeyTrasmit_s {
	DOT1X_STATEVAR	state;
	A_BOOL		keyAvailable;
    } stateKT;

    struct StateReauth_s {
	DOT1X_STATEVAR	state;
    } stateReauth;

    struct StateControlDirs_s {
	DOT1X_STATEVAR	state;
	A_BOOL		adminControlDirs;
	A_BOOL		openControlDirs;
    } stateControlDirs;

    /* Timers for Auth PAE SM, timers SM, backend SM and Reauth SM */
    /* All int32's so we can compare > 0...ie -1 is invalid */
    struct timer_s {
	A_INT32		aWhile; // used by AS SM
	A_INT32		quietWhile; // used by Auth SM
	A_INT32		reAuthWhen; // used by Reauth Timer SM
	A_INT32		txWhen;  // used by Auth SM
    } timer;

    /* Counters for Auth PAE SM */
    struct AuthCounter_s {
	A_UINT32	authEntersConnecting;
	A_UINT32	authEapLogoffsWhileConnecting;
	A_UINT32	authEntersAuthenticating;
	A_UINT32	authAuthSuccessesWhileAuthenticating;
	A_UINT32	authAuthTimeoutsWhileAuthenticating;
	A_UINT32	authAuthFailWhileAuthenticating;
	A_UINT32	authAuthReauthsWhileAuthenticating;
	A_UINT32	authAuthEapStartsWhileAuthenticating;
	A_UINT32	authAuthEapLogoffWhileAuthenticating;
	A_UINT32	authAuthReauthsWhileAuthenticated;
	A_UINT32	authAuthEapStartsWhileAuthenticated;
	A_UINT32	authAuthEapLogoffWhileAuthenticated;
    } AuthCounter;

    struct AuthServerCounter_s {
	A_UINT32	backendResponses;
	A_UINT32	backendAccessChallenges;
	A_UINT32	backendOtherRequestsToSupplicant;
	A_UINT32	backendNonNakResponsesFromSupplicant;
	A_UINT32	backendAuthSuccesses;
	A_UINT32	backendAuthFails;
    } AuthServerCounter;

    struct radiusInfo_s {
	/*
	 *  Copy of the RADIUS header Request Authenticator
	 *  to verify the Challenges
	 */
	A_UINT8			reqAuth[MSG_AUTH_LEN];

	/* Copy of the Message Authenticator for session */
	A_UINT8			msgAuth[MSG_AUTH_LEN];

	/* Copy of the EAP-Request/Identity which started the session */
	A_CHAR			identity[MAX_ETHER_BODY_SIZE]; // XXX 256?
	A_UINT16		identityLength;

	A_UCHAR			sendkey[64];
	A_UCHAR			recvkey[64];

	/*
	 *  State and Class may net be used by all servers, but
	 *  they are required when they are.
	 */
	A_UINT8			state[255];
	A_UINT8			stateLength; // 0 means no state seen yet
	A_UINT8			class[255];
	A_UINT8			classLength; // 0 means no class seen yet

    } radiusInfo;

    /* Copy of last frame in transaction */
    /* Be sure to erase this after auth done */
    struct data_s {
	A_UINT8		eapolBuf[MAX_ETHER_BODY_SIZE];
	A_UINT8		udpBuf[4096];
    } data;
} SCB;

/* These values are to be set by the MIB */
typedef struct AdminVar_s {
    A_UINT8		quietPeriod;
    A_UINT8		reAuthMax;
    A_UINT8		txPeriod;
    A_BOOL		keyTxEnabled;
    A_UINT8		suppTimeout;
    A_UINT8		serverTimeout;
    A_UINT8		maxReq;
    A_BOOL		reAuthEnabled;
    A_BOOL		initialize;
    A_UINT32		basePeriod;
    A_INT32		reAuthPeriod;
    char		*secret;
    A_UINT32		secretLen;
    struct in_addr	authServer; // address in host order
} ADMINVAR;

extern void sdlInit(struct sockaddr_dl *sdl, WLAN_MACADDR *macAddr, int unit);
extern void printKey(unsigned char *, int);

typedef struct arcfourContext_s {
    unsigned int x;
    unsigned int y;
    unsigned char state[256];
} ArcfourContext;

typedef struct mppeKey {
    A_UINT8	length;
    A_UINT8	secret[1];
} MPPE_KEY;

extern void *linkDevCookies[MAX_RADIOS];

#define DOT1X_STATE_DEBUG 		0x00000001 // see state transitions
#define DOT1X_SOCKET_DEBUG		0x00000002 // see socket operations
#define DOT1X_RADIUS_DEBUG		0x00000004 // see RADIUS transactions
#define DOT1X_KEYMSG_DEBUG		0x00000008 // see key messages flying
#define DOT1X_KEYVAL_DEBUG		0x00000010 // see key values
#define DOT1X_STATEVAR_DEBUG		0x00000020 // changes in state variables
#define DOT1X_IDENTITY_DEBUG		0x00000040 // see supplicant identities

#define DOT1X_DEBUG_CHATTER		0xFFFFFFFF

#define ST_ENTRY(x)         if (dot1xDebug & DOT1X_STATE_DEBUG) uiPrintf("ENTRY: %s\n", (x))
#define DOT1XPRINT          if (dot1xDebug & DOT1X_STATEVAR_DEBUG) uiPrintf
#define RCPRINT		    if (dot1xDebug & DOT1X_RADIUS_DEBUG) uiPrintf
