/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * Defines header objects used INTERNALLY by the HAL layer
 *
 * $Id: //depot/sw/releases/2.0.1/src/hal/hal.h#4 $
 */

#ifndef _HAL_H_
#define _HAL_H_

#include "wlantype.h"   /* A_UINTxx, etc. */

#ifdef VXWORKS
#include "ar5hwcEnd.h"
#define FIELD_OFFSET(type,field)    ((int)(&((type *)0)->field))
#else
//fadi #include "ndisdrvext.h"
#endif

#ifdef _cplusplus
extern "C" {
#endif

/* Storage for HAL-specific items */
typedef struct HalInfo {
    struct eepMap   *pEepData;          // Holds all info read from EEPROM on first reset
    A_UINT32        globIntRefCount;    // Reference count for global interrupt enable
} HAL_INFO;

/* The device specific function list for the HW layer */
typedef struct HwFuncs {
    A_STATUS  (*hwDetach)(WLAN_DEV_INFO *pDev);

    /* Reset Functions */
    A_STATUS  (*hwReset)(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval, A_BOOL bChannelChange);
    A_STATUS  (*hwPhyDisable)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwDisable)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwResetDma)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwPerCalibration)(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

    /* TX Functions */
    A_UINT32  (*hwGetTxDP)(WLAN_DEV_INFO *pDev);
    void      (*hwSetTxDP)(WLAN_DEV_INFO *pDev, A_UINT32 txdp);
    A_BOOL    (*hwUpdateTxTrigLevel)(WLAN_DEV_INFO *pDev, A_BOOL bIncTrigLevel);
    void      (*hwStartDmaTransmit)(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask);
    A_BOOL    (*hwIsTxQueueStopped)(WLAN_DEV_INFO *pDev, A_UINT16 queueMask);
    void      (*hwClearTransmitFilter)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwGetTransmitFilterIndex)(WLAN_DEV_INFO *pDev, A_UINT32 index);
    A_UINT32  (*hwGetShortRetryLimit)(WLAN_DEV_INFO *pDev);
    A_UINT32  (*hwGetLongRetryLimit)(WLAN_DEV_INFO *pDev);
    void      (*hwSetTxDescHdrLen)(ATHEROS_DESC *pDesc, A_UINT16 value);
    void      (*hwSetTxDescLongPkt)(ATHEROS_DESC *pDesc, A_UINT16 value);
    void      (*hwSetTxDescAntModeXmit)(ATHEROS_DESC *pDesc, A_UINT16 value);
    void      (*hwSetTxDescPktType)(ATHEROS_DESC *pDesc, HAL_DESC_PKT_TYPE value);
    A_UINT16  (*hwGetTxDescEncryptKeyIndex)(ATHEROS_DESC *pDesc);
    void      (*hwSetTxDescEncryptKeyIndex)(ATHEROS_DESC *pDesc, A_UINT16 value);
    void      (*hwSetTxDescRtsDuration)(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 rateIndex, A_UINT16 controlRate);
    void      (*hwSetTxDescNoAck)(ATHEROS_DESC *pDesc, A_UINT16 value);
    void      (*hwSetTxDescVeol)(ATHEROS_DESC *pDesc, A_UINT16 Lenvaluegth);
    void      (*hwDebugPrintTxDesc)(ATHEROS_DESC *pDesc);

    /* RX Functions */
    A_UINT32  (*hwGetRxDP)(WLAN_DEV_INFO *pDev);
    void      (*hwSetRxDP)(WLAN_DEV_INFO *pDev, A_UINT32 rxdp);
    void      (*hwEnableReceive)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwStopDmaReceive)(WLAN_DEV_INFO *pDev);
    void      (*hwStartPcuReceive)(WLAN_DEV_INFO *pDev);
    void      (*hwStopPcuReceive)(WLAN_DEV_INFO *pDev);
    void      (*hwSetMulticastFilter)(WLAN_DEV_INFO *pDev, A_UINT32 filter0, A_UINT32 filter1);
    void      (*hwMulticastFilterIndex)(WLAN_DEV_INFO *pDev, A_UINT32 index, A_BOOL bSet);
    void      (*hwRxFilter)(WLAN_DEV_INFO *pDev, A_UINT32 bits, HAL_RX_FILTER_OPER oper);
    A_UINT16  (*hwGetRxDescRxAntenna)(ATHEROS_DESC *pDesc);
    A_UINT16  (*hwGetRxDescRxTimeStamp)(ATHEROS_DESC *pDesc);
    A_UINT16  (*hwGetRxDescKeyIndex)(ATHEROS_DESC *pDesc);
    A_UINT16  (*hwGetRxDescKeyCacheMiss)(ATHEROS_DESC *pDesc);
    void      (*hwEnableRxZlf)(WLAN_DEV_INFO *pDev);

    /* Misc Functions */
    A_STATUS  (*hwSetRegulatoryDomain)(WLAN_DEV_INFO *pDev, A_UINT16 regDomain);
    void      (*hwSetLedState)(WLAN_DEV_INFO *pDev, A_BOOL bConnected);
    void      (*hwWriteAssocid)(WLAN_DEV_INFO *pDev, WLAN_MACADDR *bssid, A_UINT16 assocId, A_UINT16 timOffset);
    A_UINT32  (*hwGpioGet)(WLAN_DEV_INFO *pDev, A_UINT32 gpio);
    void      (*hwGpio0SetIntr)(WLAN_DEV_INFO *pDev, A_UINT32 ilevel);
    void      (*hwSetStaBeaconTimers)(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_UINT32 dtim_count, A_UINT32 cfp_count);
    void      (*hwGetTsf)(WLAN_DEV_INFO *pDev, WLAN_TIMESTAMP *tsf);
    void      (*hwResetTsf)(WLAN_DEV_INFO *pDev);
    void      (*hwSetAdhocMode)(WLAN_DEV_INFO *pDev);
    A_UINT16  (*hwGetRegDmn)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwGetRfKill)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwGetTurboDisable)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwGetChannelEdges)(WLAN_DEV_INFO *pDev, A_UINT16 channelFlags, A_UINT16 *pLowChannel, A_UINT16 *pHighChannel);
    A_UINT32  (*hwGetMaxTurboRate)(void);
    A_UINT32  (*hwGetRandomSeed)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwDetectCardPresent)(WLAN_DEV_INFO *pDev);
    void      (*hwUpdateMibCounters)(WLAN_DEV_INFO *pDev, A_BOOL bUpdateAll);
    A_BOOL    (*hwIsHwAesSupported)(WLAN_DEV_INFO *pDev);
    A_BOOL    (*hwIsSwAesSupported)(WLAN_DEV_INFO *pDev);
    void      (*hwEnableHwEncryption)(WLAN_DEV_INFO *pDev);
    void      (*hwDisableHwEncryption)(WLAN_DEV_INFO *pDev);
    void      (*hwEnableRadarDetection)(WLAN_DEV_INFO *pDev);
    A_UINT32  (*hwGetCurRssi)(WLAN_DEV_INFO *pDev);
    A_UINT32  (*hwGetDefAntenna) (WLAN_DEV_INFO *pDev);
    void      (*hwSetDefAntenna) (WLAN_DEV_INFO *pDev, A_UINT32 antenna);

    /* Key Cache Functions */
    A_UINT32  (*hwGetKeyCacheSize)(WLAN_DEV_INFO *pDev);
    void      (*hwResetKeyCacheEntry)(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex);
    void      (*hwSetKeyCacheEntry)(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex, WLAN_MACADDR *pMacAddr, WLAN_PRIV_RECORD *pWlanPrivRecord, A_BOOL bXorKey);

    /* Power Management Functions */
    void      (*hwSetPowerModeAuto)(WLAN_DEV_INFO *pDev, A_BOOL setChip, A_UINT16 sleepDuration);
    A_STATUS  (*hwSetPowerModeAwake)(WLAN_DEV_INFO *pDev, A_BOOL setChip);
    void      (*hwSetPowerModeSleep)(WLAN_DEV_INFO *pDev, A_BOOL setChip);
    A_BOOL    (*hwQueryPSPollSupport)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwInitPSPoll)(WLAN_DEV_INFO *pDev);
    A_STATUS  (*hwEnablePSPoll)(WLAN_DEV_INFO *pDev, WLAN_MACADDR *pBssid, A_UINT16 assocId);
    A_STATUS  (*hwDisablePSPoll)(WLAN_DEV_INFO *pDev);

    /* Beacon Functions */
    A_STATUS  (*hwWaitForBeaconDone)(WLAN_DEV_INFO *pDev);
    void      (*hwUpdateBeaconDesc)(ATHEROS_DESC *pDesc, A_UINT32 frameLen);
    void      (*hwQueueBeaconAndCab)(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pBeacon, ATHEROS_DESC *pCab);
    void      (*hwSetupBeaconDesc)(ATHEROS_DESC *pDesc, A_BOOL isAp);
    void      (*hwBeaconInit)(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_BOOL isAp);

    /* Interrupt Functions */
    A_BOOL    (*hwIsInterruptPending)(WLAN_DEV_INFO *pDev);
    HAL_INT_TYPE (*hwGetInterrupts)(WLAN_DEV_INFO *pDev, HAL_INT_TYPE *pUnmaskedValue);
    void      (*hwEnableInterrupts)(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints);
    void      (*hwDisableInterrupts)(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints);

} HW_FUNCS;

#ifdef _cplusplus
}
#endif

#endif /* _HAL_H */
