/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * Defines the MAC/PHY Hardware Abstraction Layer (HAL) API as well as commonly
 * used types and structs.
 *
 * $Id: //depot/sw/releases/2.0.1/src/hal/halApi.h#4 $
 */

#ifndef _HAL_API_H_
#define _HAL_API_H_

/*
 * Define isrPrintf() for printing from ISR context.  Things are different
 * for the AP and NDIS drivers, hence the following somewhat-less-than-elegant
 * solution until we have a common, comprehensive tracing facility.
 */
#include "ui.h"
#ifdef VXWORKS
#define __PROTOTYPE_5_0 1   /* use varargs logMsg() */
#include "logLib.h"
#define isrPrintf logMsg
#else /* VXWORKS */
#define isrPrintf uiPrintf
#endif /* VXWORKS */

#ifdef _cplusplus
extern "C" {
#endif

/* Attach Functions */
A_STATUS
halAttach(WLAN_DEV_INFO *pDev);

A_UINT32
halGetNumDevid(void);

A_UINT16
halGetDevids(A_UINT32 index);

A_STATUS
halDetach(WLAN_DEV_INFO *pDev);

/* Reset Functions */
A_STATUS
halReset(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval, A_BOOL bChannelChange);

A_STATUS
halPhyDisable(WLAN_DEV_INFO *pDev);

A_STATUS
halDisable(WLAN_DEV_INFO *pDev);

A_STATUS
halResetDma(WLAN_DEV_INFO *pDev);

A_STATUS
halPerCalibration(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval);

/* Transmit Functions */
A_UINT32
halGetTxDP(WLAN_DEV_INFO *pDev);

void
halSetTxDP(WLAN_DEV_INFO *pDev, A_UINT32 txdp);

A_BOOL
halUpdateTxTrigLevel(WLAN_DEV_INFO *pDev, A_BOOL bIncTrigLevel);

enum halTxQueueTypes {
    /* TX queue types for halStartDmaTransmit, halStopDmaTransmit  */
    HAL_TX_QUEUE_DATA        = 0x0001, // data queue
    HAL_TX_QUEUE_CAB         = 0x0002, // cab queue
    HAL_TX_QUEUE_BEACON      = 0x0004, // beacon queue
    HAL_TX_QUEUE_PS_POLL     = 0x0008, // PS Poll queue

    HAL_TX_QUEUE_ALL         = 0xFFFF  // all queue types
};

void
halStartDmaTransmit(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask);

A_BOOL
halIsTxQueueStopped(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask);

void
halClearTransmitFilter(WLAN_DEV_INFO *pDev);

A_BOOL
halGetTransmitFilterIndex(WLAN_DEV_INFO *pDev, A_UINT32 index);

A_UINT32
halGetShortRetryLimit(WLAN_DEV_INFO *pDev);

A_UINT32
halGetLongRetryLimit(WLAN_DEV_INFO *pDev);

void
halSetTxDescHdrLen(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halSetTxDescLongPkt(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halSetTxDescAntModeXmit(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halSetTxDescPktType(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, HAL_DESC_PKT_TYPE value);

A_UINT16
halGetTxDescEncryptKeyIndex(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

void
halSetTxDescEncryptKeyIndex(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halSetTxDescRtsDuration(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 rateIndex, A_UINT16 controlRate);

void
halSetTxDescNoAck(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halSetTxDescVeol(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value);

void
halDebugPrintTxDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

/* Receive Functions */

A_UINT32
halGetRxDP(WLAN_DEV_INFO *pDev);

void
halSetRxDP(WLAN_DEV_INFO *pDev, A_UINT32 rxdp);

void
halEnableReceive(WLAN_DEV_INFO *pDev);

A_STATUS
halStopDmaReceive(WLAN_DEV_INFO *pDev);

void
halStartPcuReceive(WLAN_DEV_INFO *pDev);

void
halStopPcuReceive(WLAN_DEV_INFO *pDev);

void
halSetMulticastFilter(WLAN_DEV_INFO *pDev, A_UINT32 filter0, A_UINT32 filter1);

void
halMulticastFilterIndex(WLAN_DEV_INFO *pDev, A_UINT32 index, A_BOOL bSet);


typedef enum halRxFilterOper {
    /* The Rx Filter Operations Types have the following affect on the
     * hardware register and the software copy of the hardware register.
     *
     *                          Hardware Register   Software Copy
     *
     * HAL_RX_FILTER_INIT       Updated             Updated
     * HAL_RX_FILTER_SET        Updated             Updated
     * HAL_RX_FILTER_CLEAR      Updated             Updated
     * HAL_RX_FILTER_TEMP       Updated             unmodified
     * HAL_RX_FILTER_RESTORE    Updated             unmodified
     */

    HAL_RX_FILTER_INIT,         // Init hardware register and software copy
    HAL_RX_FILTER_SET,          // Set frame type bits
    HAL_RX_FILTER_CLEAR,        // Clear frame type bits
    HAL_RX_FILTER_TEMP,         // Temporarily set hardware register ONLY.
    HAL_RX_FILTER_RESTORE       // Restore hardware register with software copy

} HAL_RX_FILTER_OPER;

enum halRxFilterFrameTypes {
    /* Rx Filter Frame Types */
    HAL_RX_UCAST        = 0x00000001, // Allow unicast frames
    HAL_RX_MCAST        = 0x00000002, // Allow multicast frames
    HAL_RX_BCAST        = 0x00000004, // Allow broadcast frames
    HAL_RX_CONTROL      = 0x00000008, // Allow control frames
    HAL_RX_BEACON       = 0x00000010, // Allow beacon frames
    HAL_RX_PROM         = 0x00000020, // Promiscuous mode, all packets
    HAL_RX_RADAR        = 0x00000080  // Allow only radar phy errors - noncommon mapping
};

void
halRxFilter(WLAN_DEV_INFO *pDev, A_UINT32 bits, HAL_RX_FILTER_OPER oper);

A_UINT16
halGetRxDescRxAntenna(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

A_UINT16
halGetRxDescRxTimeStamp(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

A_UINT16
halGetRxDescKeyIndex(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

A_UINT16
halGetRxDescKeyCacheMiss(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc);

void
halEnableRxZlf(WLAN_DEV_INFO *pDev);

/* Misc Functions */
A_STATUS
halSetRegulatoryDomain(WLAN_DEV_INFO *pDev, A_UINT16 regDomain);

void
halSetLedState(WLAN_DEV_INFO *pDev, A_BOOL bConnected);

void
halWriteAssocid(WLAN_DEV_INFO *pDev, WLAN_MACADDR *bssid, A_UINT16 assocId, A_UINT16 timOffset);

A_UINT32
halGpioGet(WLAN_DEV_INFO *pDev, A_UINT32 gpio);

void
halGpio0SetIntr(WLAN_DEV_INFO *pDev, A_UINT32 ilevel);

void
halSetStaBeaconTimers(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_UINT32 dtim_count, A_UINT32 cfp_count);

void
halGetTsf(WLAN_DEV_INFO *pDev, WLAN_TIMESTAMP *tsf);

void
halResetTsf(WLAN_DEV_INFO *pDev);

void
halSetAdhocMode(WLAN_DEV_INFO *pDev);

A_UINT16
halGetRegDmn(WLAN_DEV_INFO *pDev);

A_BOOL
halGetRfKill(WLAN_DEV_INFO *pDev);

A_BOOL
halGetTurboDisable(WLAN_DEV_INFO *pDev);

A_UINT16
halComputeTxTime(A_UINT32 bitNum, RATE_TABLE_STRUCT *pRateTable, A_UINT16 rateIndex, A_BOOL shortPreamble);

A_BOOL
halGetChannelEdges(WLAN_DEV_INFO *pDev, A_UINT16 channelFlags, A_UINT16 *pLowChannel, A_UINT16 *pHighChannel);

A_UINT32
halGetMaxTurboRate(WLAN_DEV_INFO *pDev);

A_UINT32
halGetRandomSeed(WLAN_DEV_INFO *pDev);

A_BOOL
halDetectCardPresent(WLAN_DEV_INFO *pDev);

void
halUpdateMibCounters(WLAN_DEV_INFO *pDev, A_BOOL bUpdateAll);

A_BOOL
halIsHwAesSupported(WLAN_DEV_INFO *pDev);

A_BOOL
halIsSwAesSupported(WLAN_DEV_INFO *pDev);

void
halEnableHwEncryption(WLAN_DEV_INFO *pDev);

void
halDisableHwEncryption(WLAN_DEV_INFO *pDev);

void
halEnableRadarDetection(WLAN_DEV_INFO *pDev);

A_UINT32
halGetCurRssi(WLAN_DEV_INFO *pDev);

A_UINT32
halGetDefAntenna(WLAN_DEV_INFO *pDev);

void
halSetDefAntenna(WLAN_DEV_INFO *pDev, A_UINT32 antenna);

/* Key Cache Functions */
A_UINT32
halGetKeyCacheSize(WLAN_DEV_INFO *pDev);

void
halResetKeyCache(WLAN_DEV_INFO *pDev);

void
halResetKeyCacheEntry(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex);

void
halSetKeyCacheEntry(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex,
                    WLAN_MACADDR *pMacAddr, WLAN_PRIV_RECORD *pWlanPrivRecord,
                    A_BOOL bXorKey);

/* Power Management Functions */
void
halSetPowerModeAuto(WLAN_DEV_INFO *pDev, A_BOOL setChip, A_UINT16 sleepDuration);

A_STATUS
halSetPowerModeAwake(WLAN_DEV_INFO *pDev, A_BOOL setChip);

void
halSetPowerModeSleep(WLAN_DEV_INFO *pDev, A_BOOL setChip);

A_BOOL
halQueryPSPollSupport(WLAN_DEV_INFO *pDev);

A_STATUS
halInitPSPoll(WLAN_DEV_INFO *pDev);

A_STATUS
halEnablePSPoll(WLAN_DEV_INFO *pDev, WLAN_MACADDR *pBssid, A_UINT16 assocId);

A_STATUS
halDisablePSPoll(WLAN_DEV_INFO *pDev);

/* Beacon functions */
A_STATUS
halWaitForBeaconDone(WLAN_DEV_INFO *pDev);

void
halUpdateBeaconDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT32 frameLen);

void
halQueueBeaconAndCab(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pBeacon, ATHEROS_DESC *pCab);

void
halSetupBeaconDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT32 frameLen, A_BOOL isAp);

void
halBeaconInit(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_BOOL isAp);

/* Interrupt functions */

/*
 * NOTE WELL:
 * These are mapped to take advantage of the common locations for many of
 * the bits on all of the currently supported MAC chips. This is to make the ISR
 * as efficient as possible, while still abstracting HW differences. When new
 * hardware breaks this commonality this enumerated type, as well as the HAL
 * functions using it, must be modified. All values are directly mapped unless
 * commented otherwise.
 */
typedef enum HalIntType {
    HAL_INT_RX      = 0x00000001,   /* Non-common mapping */
    HAL_INT_RXEOL   = 0x00000010,
    HAL_INT_RXORN   = 0x00000020,
    HAL_INT_TX      = 0x00000040,   /* Non-common mapping */
    HAL_INT_TXURN   = 0x00000800,
    HAL_INT_MIB     = 0x00001000,
    HAL_INT_RXPHY   = 0x00004000,
    HAL_INT_RXKCM   = 0x00008000,
    HAL_INT_SWBA    = 0x00010000,
    HAL_INT_BMISS   = 0x00040000,
    HAL_INT_BNR     = 0x00100000,   /* Non-common mapping */
    HAL_INT_GPIO    = 0x01000000,
    HAL_INT_FATAL   = 0x40000000,   /* Non-common mapping */
    HAL_INT_GLOBAL  = 0x80000000,   /* To set/clear IER. Don't combine with other ints */

    /* The following are special values which don't map directly to ISR bits */

    HAL_INT_COMMON  =  (HAL_INT_RXEOL | HAL_INT_RXORN | HAL_INT_TXURN | HAL_INT_MIB   |
                        HAL_INT_RXKCM | HAL_INT_SWBA  | HAL_INT_BMISS | HAL_INT_GPIO  |
                        HAL_INT_RXPHY),
    HAL_INT_INVALID = ~(HAL_INT_RX    | HAL_INT_RXEOL | HAL_INT_RXORN | HAL_INT_TX    |
                        HAL_INT_TXURN | HAL_INT_MIB   | HAL_INT_RXKCM | HAL_INT_SWBA  |
                        HAL_INT_BMISS | HAL_INT_BNR   | HAL_INT_GPIO  | HAL_INT_FATAL |
                        HAL_INT_RXPHY | HAL_INT_GLOBAL),
    HAL_INT_NOCARD  = 0xffffffff    /* To signal that the card has been removed */
} HAL_INT_TYPE;

A_BOOL
halIsInterruptPending(WLAN_DEV_INFO *pDev);

HAL_INT_TYPE
halGetInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE *pUnmaskedValue);

void
halEnableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints);

void
halDisableInterrupts(WLAN_DEV_INFO *pDev, HAL_INT_TYPE ints);


#ifdef _cplusplus
}
#endif

#endif /* _HAL_API_H */
