/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * TODO: comment
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halBeacon.c#1 $"

#include "wlantype.h"   /* A_UINTxx, etc. */
#include "wlandrv.h"    /* WLAN_DEV_INFO struct */
#include "halApi.h"     /* fn prototypes */
#include "hal.h"        /* HW_FUNCS struct */
#include "wlanchannel.h"


/**************************************************************
 * halWaitForBeaconDone
 *
 * Wait for the beacon queue(s) to complete.
 *
 * Returns: A_OK    if everything stopped successfully
 *          A_ERROR if something went wrong
 */
A_STATUS
halWaitForBeaconDone(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwWaitForBeaconDone);

    return pDev->pHwFunc->hwWaitForBeaconDone(pDev);
}

/**************************************************************
 * halUpdateBeaconDesc
 *
 * This routine is used by both the AP and client.  It constructs the beacon
 * descriptor based on HW requirements.
 *
 * Assumptions: frameLen does not include the FCS
 */
void
halUpdateBeaconDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT32 frameLen)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetupBeaconDesc);
    ASSERT(pDesc && frameLen);

    pDesc->hwControl.tx.frameLength = frameLen + FCS_FIELD_SIZE;

    pDev->pHwFunc->hwUpdateBeaconDesc(pDesc, frameLen);
}

/**************************************************************
 * halQueueBeaconAndCab
 *
 * Queues up the beacon to HW as well as any CAB (crap after beacon), in the
 * case of an AP with buffered BC/MC traffic or a client with DTIMs to send.
 * pBeacon must point to the beacon frame descriptor.  pCab may be left NULL
 * or may optionally point to a chain of frame descriptors.
 */
void
halQueueBeaconAndCab(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pBeacon, ATHEROS_DESC *pCab)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwQueueBeaconAndCab);

    pDev->pHwFunc->hwQueueBeaconAndCab(pDev, pBeacon, pCab);
}

/**************************************************************
 * halSetupBeaconDesc
 *
 * This routine is used by both the AP and client.  It constructs the beacon
 * descriptor based on HW requirements.
 *
 * Assumptions: frameLen does not include the FCS
 */
void
halSetupBeaconDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT32 frameLen, A_BOOL isAp)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetupBeaconDesc);
    ASSERT(pDesc && frameLen);

    pDesc->hwControl.tx.more         = 0;
    halSetTxDescAntModeXmit(pDev, pDesc, pDev->localSta->antTx);
    pDesc->hwControl.tx.transmitRate =
        pDev->staConfig.pRateTable->info[pDev->localSta->basicRateIndex].plcpRate;

    if (pDev->staConfig.pChannel->channelFlags & CHANNEL_CCK) {
        // 11b beacons always go out in long preamble
        pDesc->hwControl.tx.transmitRate &= ~CCK_SHORT_PREAMBLE_ENABLE;
    }

    /* Fill in the link pointers. */
    pDev->pHwFunc->hwSetupBeaconDesc(pDesc, isAp);

    /* Set the frame and buffer lengths. */
    halUpdateBeaconDesc(pDev, pDesc, frameLen);
}

/**************************************************************
 * halBeaconInit
 *
 * Initializes the HW registers required to send beacons.
 */
void
halBeaconInit(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_BOOL isAp)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwBeaconInit);

    pDev->pHwFunc->hwBeaconInit(pDev, tsf, isAp);
}

