/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * Defines the hardware descriptor data structures.
 *
 * Fields named hwSpecific are chipset dependent. Use the HAL descriptor
 * access functions to access these fields.
 * All other fields are common to all Atheros chipsets.
 *
 * $Id: //depot/sw/releases/2.0.1/src/hal/halDesc.h#1 $
 */

#ifndef _HAL_DESC_H_
#define _HAL_DESC_H_

#ifdef _cplusplus
extern "C" {
#endif

/* Common Hardware Descriptor */

#ifdef BIG_ENDIAN
/* structure for Tx Control dwords */
typedef struct halDescTxControl {
    A_UINT32    reserved1:2,        /* 31:30 reserved */
                interruptReq:1,     /* 29 set if want an interrupt on completion */
                hwSpecific0:4,      /* 28:25 hardware specific */
                clearDestMask:1,    /* 24 set if want to clear the Tx block bit */
                hwSpecific1:1,      /* 23 hardware specific */
                RTSCTSEnable:1,     /* 22 set to enable RTS/CTS */
                transmitRate:4,     /* 21:18 which rate to transmit at */
                hwSpecific2:6,      /* 17:12 hardware specific */
                frameLength:12;     /* 11:0 length of the 802.11 packet */

    A_UINT32    hwSpecific3:19,     /* 31:13 hardware specific  */
                more:1,             /* 12 set if packet continues in next descriptor */
                bufferLength:12;    /* 11:0 size of buffer for this descriptor */
} HAL_DESC_TX_CONTROL;
#else
/* structure for Tx Control dwords */
typedef struct halDescTxControl {
A_UINT32    frameLength:12,         /* 11:0 length of the 802.11 packet */
            hwSpecific2:6,          /* 17:12 hardware specific */
            transmitRate:4,         /* 21:18 which rate to transmit at */
            RTSCTSEnable:1,         /* 22 set to enable RTS/CTS */
            hwSpecific1:1,          /* 23 hardware specific */
            clearDestMask:1,        /* 24 set if want to clear the Tx block bit */
            hwSpecific0:4,          /* 28:25 hardware specific */
            interruptReq:1,         /* 29 set if want an interrupt on completion */
            reserved1:2;            /* 31:30 reserved */
A_UINT32    bufferLength:12,        /* 11:0 size of buffer for this descriptor */
            more:1,                 /* 12 set if packet continues in next descriptor */
            hwSpecific3:19;         /* 31:13 hardware specific */
} HAL_DESC_TX_CONTROL;
#endif //BIG_ENDIAN


#ifdef BIG_ENDIAN
/* structure for Rx control dwords */
typedef struct halDescRxControl {
    A_UINT32    notUsed1;           /* 31:0 reserved */
    A_UINT32    notUsed3:18,        /* 31:14 reserved */
                InterruptReq:1,     /* 13 specify mode of interrupt handling */
                notUsed2:1,         /* 12 reserved */
                bufferLength:12;    /* 11:0 length of the buffer */
} HAL_DESC_RX_CONTROL;
#else
/* structure for Rx control dwords */
typedef struct halDescRxControl {
    A_UINT32    notUsed1;           /* 31:0 reserved */
    A_UINT32    bufferLength:12,    /* 11:0 length of the buffer */
                notUsed2:1,         /* 12 reserved */
                InterruptReq:1,     /* 13 specify mode of interrupt handling */
                notUsed3:18;        /* 31:14 reserved */
} HAL_DESC_RX_CONTROL;
#endif

#ifdef BIG_ENDIAN
/* structure for Tx status dwords */
typedef volatile struct halDescTxStatus {
    A_UINT32    sendTimestamp:16,   /* 31:16 time packet was transmitted */
                hwSpecific0:4,      /* 15:12 hardware specific */
                longRetryCount:4,   /* 11:8 long retries attempted */
                shortRetryCount:4,  /* 7:4  short retries attempted */
                filtered:1,         /* 3 packet not sent due to transmit blocking */
                fifoUnderrun:1,     /* 2 set if there was a fifo overrun */
                excessiveRetries:1, /* 1 set if there were excessive retries */
                pktTransmitOK:1;    /* 0 set if packet transmitted OK */

    A_UINT32    notUsed:10,         /* 31:22 reserved*/
                hwSpecific1:1,      /* 21 hardware specific */
                ackSigStrength:8,   /* 20:13 signal strength ack was received at */
                seqNum:12,          /* 12:1 sequence number assigned by hardware */
                done:1;             /* 1 set by HW when completes descriptor update */
} HAL_DESC_TX_STATUS;
#else
/* structure for Tx status dwords */
typedef volatile struct halDescTxStatus {
    A_UINT32    pktTransmitOK:1,    /* 0 set if packet transmitted OK */
                excessiveRetries:1, /* 1 set if there were excessive retries */
                fifoUnderrun:1,     /* 2 set if there was a fifo overrun */
                filtered:1,         /* 3 packet not sent due to transmit blocking */
                shortRetryCount:4,  /* 7:4  short retries attempted */
                longRetryCount:4,   /* 11:8 long retries attempted */
                hwSpecific0:4,      /* 15:12 hardware specific */
                sendTimestamp:16;   /* 31:16 time packet was transmitted */
    A_UINT32    done:1,             /* 1 set by HW when completes descriptor update */
                seqNum:12,          /* 12:1 sequence number assigned by hardware */
                ackSigStrength:8,   /* 20:13 signal strength ack was received at */
                hwSpecific1:1,      /* 21 hardware specific */
                notUsed:10;         /* 31:22 reserved */
} HAL_DESC_TX_STATUS;
#endif


#ifdef BIG_ENDIAN
/* structure for Rx status dwords */
typedef volatile struct halDescRxStatus {
    A_UINT32    reserved0:2,        /* 31:30 reserved */
                hwSpecific0:3,      /* 29:27 hardware specific */
                rxSigStrength:8,    /* 26:19 strength packet was received at */
                rxRate:4,           /* 18:15 rate packet was received at */
                hwSpecific1:1,      /* 14 hardware specific */
                reserved1:1,        /* 13 reserved */
                more:1,             /* 12 set if packet continues in next descriptor */
                dataLength:12;      /* 11:0 number of bytes received into structure */
    A_UINT32    hwSpecific2:23,     /* 31:9 hardware specific */
                keyIndexValid:1,    /* 8 set if the key index is valid */
                phyError:3,         /* 7:5 set to indicate a phy error */
                decryptCRCError:1,  /* 4 set if there is an error on decryption */
                fifoOverrun:1,      /* 3 set if there is a fifo overrun */
                CRCError:1,         /* 2 set if there is a CRC error */
                pktReceivedOK:1,    /* 1 set if the packet was received OK */
                done:1;             /* 0 set when hardware is done updating the bit */
} HAL_DESC_RX_STATUS;
#else
/* structure for Rx status dwords */
typedef volatile struct halDescRxStatus {
    A_UINT32    dataLength:12,      /* 11:0 number of bytes received into structure */
                more:1,             /* 12 set if packet continues in next descriptor */
                reserved1:1,        /* 13 reserved */
                hwSpecific1:1,      /* 14 hardware specific */
                rxRate:4,           /* 18:15 rate packet was received at */
                rxSigStrength:8,    /* 26:19 strength packet was received at */
                hwSpecific0:3,      /* 29:27 hardware specific */
                reserved0:2;        /* 31:30 reserved */
    A_UINT32    done:1,             /* 0 set when hardware is done updating the bit */
                pktReceivedOK:1,    /* 1 set if the packet was received OK */
                CRCError:1,         /* 2 set if there is a CRC error */
                fifoOverrun:1,      /* 3 set if there is a fifo overrun */
                decryptCRCError:1,  /* 4 set if there is an error on decryption */
                phyError:3,         /* 7:5 set to indicate a phy error */
                keyIndexValid:1,    /* 8 set if the key index is valid */
                hwSpecific2:23;     /* 31:9 hardware specific */
} HAL_DESC_RX_STATUS;
#endif  //BIG_ENDIAN

/* Hardware Descriptor Packet Types */
typedef enum {
    HAL_DESC_PKT_TYPE_NORMAL,
    HAL_DESC_PKT_TYPE_ATIM,
    HAL_DESC_PKT_TYPE_PSPOLL,
    HAL_DESC_PKT_TYPE_BEACON
} HAL_DESC_PKT_TYPE;


#ifdef _cplusplus
}
#endif

#endif /* _HAL_DESC_H_ */
