/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halKeyCache.c#1 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"



/**************************************************************
 * halGetKeyCacheSize
 *
 * Returns size of hardware key cache.
 */
A_UINT32
halGetKeyCacheSize(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwGetKeyCacheSize );

    return pDev->pHwFunc->hwGetKeyCacheSize(pDev);
}

/**************************************************************
 * halResetKeyCache
 *
 * Clears ALL key cache entries
 */
void
halResetKeyCache(WLAN_DEV_INFO *pDev)
{
    A_UINT16 keyCacheIndex;

    ASSERT(pDev);

    for (keyCacheIndex = 0; keyCacheIndex < halGetKeyCacheSize(pDev); keyCacheIndex++) {
        halResetKeyCacheEntry(pDev, keyCacheIndex);
    }

}

/**************************************************************
 * halResetKeyCacheEntry
 *
 * Clears the specified key cache entry
 */
void
halResetKeyCacheEntry(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwResetKeyCacheEntry);

    pDev->pHwFunc->hwResetKeyCacheEntry(pDev, keyCacheIndex);
}

/**************************************************************
 * halSetKeyCacheEntry
 *
 * Sets the specified key cache entry
 */
void
halSetKeyCacheEntry(WLAN_DEV_INFO *pDev, A_UINT16 keyCacheIndex,
	                WLAN_MACADDR *pMacAddr, WLAN_PRIV_RECORD *pWlanPrivRecord,
                    A_BOOL bXorKey)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetKeyCacheEntry);

    pDev->pHwFunc->hwSetKeyCacheEntry(pDev, keyCacheIndex, pMacAddr, pWlanPrivRecord, bXorKey);
}



