/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halMisc.c#4 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"

/**************************************************************
 * halSetRegulatoryDomain
 *
 * Changes the Regulatory domain in the PROM.
 * Returns A_OK for a successful change.
 */
A_STATUS
halSetRegulatoryDomain(WLAN_DEV_INFO *pDev, A_UINT16 regDomain)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwSetRegulatoryDomain);
    return pDev->pHwFunc->hwSetRegulatoryDomain(pDev, regDomain);
}

/**************************************************************
 * halSetLedState
 *
 * Change LED state based on WLAN connection status
 */
void
halSetLedState(WLAN_DEV_INFO *pDev, A_BOOL bConnected)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwSetLedState);
    pDev->pHwFunc->hwSetLedState(pDev, bConnected);
}

/**************************************************************
 * halWriteAssocid
 *
 * Setup the association id (duh)
 */
void
halWriteAssocid(WLAN_DEV_INFO *pDev, WLAN_MACADDR *bssid, A_UINT16 assocId, A_UINT16 timOffset)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwWriteAssocid);
    pDev->pHwFunc->hwWriteAssocid(pDev, bssid, assocId, timOffset);
}

A_UINT32
halGpioGet(WLAN_DEV_INFO *pDev, A_UINT32 gpio)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGpioGet);
    return pDev->pHwFunc->hwGpioGet(pDev, gpio);
}

void
halGpio0SetIntr(WLAN_DEV_INFO *pDev, A_UINT32 ilevel)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGpio0SetIntr);
    pDev->pHwFunc->hwGpio0SetIntr(pDev, ilevel);
}

void
halSetStaBeaconTimers(WLAN_DEV_INFO *pDev, A_UINT32 tsf, A_UINT32 dtim_count, A_UINT32 cfp_count)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwSetStaBeaconTimers);
    pDev->pHwFunc->hwSetStaBeaconTimers(pDev, tsf, dtim_count, cfp_count);
}

void
halGetTsf(WLAN_DEV_INFO *pDev, WLAN_TIMESTAMP *tsf)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGetTsf);
    pDev->pHwFunc->hwGetTsf(pDev, tsf);
}

void
halResetTsf(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwResetTsf);
    pDev->pHwFunc->hwResetTsf(pDev);
}

void
halSetAdhocMode(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwSetAdhocMode);
    pDev->pHwFunc->hwSetAdhocMode(pDev);
}

A_UINT16
halGetRegDmn(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGetRegDmn);
    return pDev->pHwFunc->hwGetRegDmn(pDev);
}

A_BOOL
halGetRfKill(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGetRfKill);
    return pDev->pHwFunc->hwGetRfKill(pDev);
}

A_BOOL
halGetTurboDisable(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwGetTurboDisable);
    return pDev->pHwFunc->hwGetTurboDisable(pDev);
}

/**************************************************************
 * halGetChannelEdges
 *
 * Store the channel edges for the requested operational mode
 */
A_BOOL
halGetChannelEdges(WLAN_DEV_INFO *pDev, A_UINT16 channelFlags, A_UINT16 *pLowChannel, A_UINT16 *pHighChannel)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pLowChannel);
    ASSERT(pHighChannel);
    return pDev->pHwFunc->hwGetChannelEdges(pDev, channelFlags, pLowChannel, pHighChannel);
}

/**************************************************************
 * halGetMaxTurboRate
 *
 * Return the maximum rate for running turbo
 */
A_UINT32
halGetMaxTurboRate(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    return pDev->pHwFunc->hwGetMaxTurboRate();
}

/**************************************************************
 * halGetRandomSeed
 *
 * Grab a semi-random value from hardware registers - may not
 * change often
 */
A_UINT32
halGetRandomSeed(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    return pDev->pHwFunc->hwGetRandomSeed(pDev);
}

/**************************************************************
 * halDetectCardPresent
 *
 * Detect if our card is present
 */
A_BOOL
halDetectCardPresent(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwDetectCardPresent);
    return pDev->pHwFunc->hwDetectCardPresent(pDev);
}

/**************************************************************
 * halUpdateMibCounters
 *
 * Update MIB counters
 */
void
halUpdateMibCounters(WLAN_DEV_INFO *pDev, A_BOOL bUpdateAll)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwUpdateMibCounters);
    pDev->pHwFunc->hwUpdateMibCounters(pDev, bUpdateAll);
}

/**************************************************************
 * halIsHwAesSupported
 *
 * Detect if HW AES supported
 */
A_BOOL
halIsHwAesSupported(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwIsHwAesSupported);
    return pDev->pHwFunc->hwIsHwAesSupported(pDev);
}

/**************************************************************
 * halIsSwAesSupported
 *
 * Detect if SW AES supported, i.e. if HW engine stoppable
 */
A_BOOL
halIsSwAesSupported(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwIsSwAesSupported);
    return pDev->pHwFunc->hwIsSwAesSupported(pDev);
}

/**************************************************************
 * halEnableHwEncryption
 *
 * enable hw encryption/decryption engine
 */
void
halEnableHwEncryption(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwEnableHwEncryption);
    pDev->pHwFunc->hwEnableHwEncryption(pDev);
}

/**************************************************************
 * halDisableHwEncryption
 *
 * disable hw encryption/decryption engine
 */
void
halDisableHwEncryption(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwDisableHwEncryption);
    pDev->pHwFunc->hwDisableHwEncryption(pDev);
}

/**************************************************************
 * halEnableRadarDetection
 *
 * Enable radar detection in the hardware - may later be changed
 * to modify radar "sensitivity"
 */
void
halEnableRadarDetection(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwEnableRadarDetection);
    pDev->pHwFunc->hwEnableRadarDetection(pDev);
}

A_UINT32
halGetCurRssi(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    return pDev->pHwFunc->hwGetCurRssi(pDev);
}

A_UINT32
halGetDefAntenna(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    return pDev->pHwFunc->hwGetDefAntenna(pDev);
}

void
halSetDefAntenna(WLAN_DEV_INFO *pDev, A_UINT32 antenna)
{
    ASSERT(pDev);
    pDev->pHwFunc->hwSetDefAntenna(pDev, antenna);
}
