/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halPower.c#1 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"



/**************************************************************
 * halSetPowerModeAuto
 *
 * Set Power Mode to auto/normal.
 * Duration in units of 128us (1/8 TU)
 */
void
halSetPowerModeAuto(WLAN_DEV_INFO *pDev, A_BOOL setChip, A_UINT16 sleepDuration)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetPowerModeAuto );

    pDev->pHwFunc->hwSetPowerModeAuto(pDev, setChip, sleepDuration);
}

/**************************************************************
 * halSetPowerModeAwake
 *
 * Force chip to awake.
 * Returns A_OK if chip is successfully forced awake.
 */
A_STATUS
halSetPowerModeAwake(WLAN_DEV_INFO *pDev, A_BOOL setChip)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetPowerModeAwake );

    return pDev->pHwFunc->hwSetPowerModeAwake(pDev, setChip);
}

/**************************************************************
 * halSetPowerModeSleep
 *
 * Force chip to sleep.
 */
void
halSetPowerModeSleep(WLAN_DEV_INFO *pDev, A_BOOL setChip)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwSetPowerModeSleep);

    pDev->pHwFunc->hwSetPowerModeSleep(pDev, setChip);
}

/**************************************************************
 * halQueryPSPollSupport
 *
 * Query hardware PS-Poll support
 */
A_BOOL
halQueryPSPollSupport(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwQueryPSPollSupport);

    return pDev->pHwFunc->hwQueryPSPollSupport(pDev);
}


/**************************************************************
 * halInitPSPoll
 *
 * Initialize for PS-Polls
 */
A_STATUS
halInitPSPoll(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwInitPSPoll);

    return pDev->pHwFunc->hwInitPSPoll(pDev);
}

/**************************************************************
 * halEnablePSPoll
 *
 * Enable PS-Polls on the STA
 */
A_STATUS
halEnablePSPoll(WLAN_DEV_INFO *pDev, WLAN_MACADDR *pBssid, A_UINT16 assocId)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwEnablePSPoll);

    return pDev->pHwFunc->hwEnablePSPoll(pDev, pBssid, assocId);
}

/**************************************************************
 * halDisablePSPoll
 *
 * Disable PS-Polls on the STA
 */
A_STATUS
halDisablePSPoll(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev && pDev->pHwFunc && pDev->pHwFunc->hwDisablePSPoll);

    return pDev->pHwFunc->hwDisablePSPoll(pDev);
}
