/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halReset.c#1 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"

/**************************************************************
 * halPeriodicCal
 *
 * Readjusts hardware to environmental changes such as
 * temperature, noise, etc.  Performs channel "correctness" check.
 *
 */
A_STATUS
halPerCalibration(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwPerCalibration);
    return pDev->pHwFunc->hwPerCalibration(pDev, pChval);
}

/**************************************************************
 * halReset
 *
 * halReset requires a successful halInit.  halReset soft-resets
 * all chips present and then reloads their initialization
 * information based on the chips' selected operation mode.
 *
 */
A_STATUS
halReset(WLAN_DEV_INFO *pDev, CHAN_VALUES *pChval, A_BOOL bChannelChange)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwReset);
    return pDev->pHwFunc->hwReset(pDev, pChval, bChannelChange);
}

/**************************************************************
 * halDisable
 *
 * Disable the chip by putting it into reset
 */
A_STATUS
halDisable(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwDisable);
    return pDev->pHwFunc->hwDisable(pDev);
}

/**************************************************************
 * halPhyDisable
 *
 * Used by RFKIll to turn off all radio capabilities.
 */
A_STATUS
halPhyDisable(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwPhyDisable);
    return pDev->pHwFunc->hwPhyDisable(pDev);
}

/**************************************************************
 * halResetDma
 *
 * Resets DMA only - only available on ar5210
 */
A_STATUS
halResetDma(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);
    ASSERT(pDev->pHwFunc->hwResetDma);
    return pDev->pHwFunc->hwResetDma(pDev);
}
