/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/hal/halTransmit.c#1 $"

#include "wlantype.h"
#include "wlandrv.h"
#include "halApi.h"
#include "hal.h"
#include "ui.h"
#include "display.h"

/**************************************************************
 * halGetTxDP
 *
 * Get the TXDP for the "main" data queue.  Needs to be extended
 * for multiple Q functionality
 */
A_UINT32
halGetTxDP(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwGetTxDP(pDev);
}

/**************************************************************
 * halSetTxDP
 *
 * Set the TxDP for the "main" data queue.  Needs to be extended
 * for multiple Q functionality
 */
void
halSetTxDP(WLAN_DEV_INFO *pDev, A_UINT32 txdp)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    pDev->pHwFunc->hwSetTxDP(pDev, txdp);
}

/**************************************************************
 * halUpdateTxTrigLevel
 *
 * Update Tx FIFO trigger level.
 *
 * Set bIncTrigLevel to TRUE to increase the trigger level.
 * Set bIncTrigLevel to FALSE to decrease the trigger level.
 *
 * Returns TRUE if the trigger level was updated
 */
A_BOOL
halUpdateTxTrigLevel(WLAN_DEV_INFO *pDev, A_BOOL bIncTrigLevel)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwUpdateTxTrigLevel(pDev, bIncTrigLevel);

}

/**************************************************************
 * halStartDmaTransmit
 *
 * Set Transmit Enable bits for the specified queue types
 * (see enum halTxQueueTypes).
 */
void
halStartDmaTransmit(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    pDev->pHwFunc->hwStartDmaTransmit(pDev, queueTypeMask);
}


/**************************************************************
 * halIsTxQueueStopped
 *
 * See if the given queues are active or not
 */
A_BOOL
halIsTxQueueStopped(WLAN_DEV_INFO *pDev, A_UINT16 queueTypeMask)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwIsTxQueueStopped(pDev, queueTypeMask);
}

/**************************************************************
 * halClearTransmitFilter
 *
 * Clear transmit filter
 *
 */
void
halClearTransmitFilter(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    pDev->pHwFunc->hwClearTransmitFilter(pDev);
}

/**************************************************************
 * halGetTransmitFilterIndex
 *
 * Get transmit filter by index
 *
 * Returns TRUE if transmit filter bit is set
 */
A_BOOL
halGetTransmitFilterIndex(WLAN_DEV_INFO *pDev, A_UINT32 index)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwGetTransmitFilterIndex(pDev, index);
}


/**************************************************************
 * halGetShortRetryLimit
 *
 * Get short frame retry limit for the main data queue.
 * Needs to be extended for multiple Q functionality
 *
 */
A_UINT32
halGetShortRetryLimit(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwGetShortRetryLimit(pDev);
}

/**************************************************************
 * halGetLongRetryLimit
 *
 * Get long frame retry limit for the main data queue.
 * Needs to be extended for multiple Q functionality
 *
 */
A_UINT32
halGetLongRetryLimit(WLAN_DEV_INFO *pDev)
{
    ASSERT(pDev);
    ASSERT(pDev->pHwFunc);

    return pDev->pHwFunc->hwGetLongRetryLimit(pDev);
}


// Descriptor access functions
void
halSetTxDescHdrLen(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescHdrLen(pDesc, value);
}

void
halSetTxDescLongPkt(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescLongPkt(pDesc, value);
}

void
halSetTxDescAntModeXmit(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescAntModeXmit(pDesc, value);
}

void
halSetTxDescPktType(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, HAL_DESC_PKT_TYPE value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescPktType(pDesc, value);
}

A_UINT16
halGetTxDescEncryptKeyIndex(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    return pDev->pHwFunc->hwGetTxDescEncryptKeyIndex(pDesc);
}

/*
 * Set and validate HW specific key cache index
 */
void
halSetTxDescEncryptKeyIndex(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescEncryptKeyIndex(pDesc, value);
}

void
halSetTxDescRtsDuration(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 rateIndex, A_UINT16 controlRate)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescRtsDuration(pDev, pDesc, rateIndex, controlRate);
}

void
halSetTxDescNoAck(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescNoAck(pDesc, value);
}

void
halSetTxDescVeol(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc, A_UINT16 value)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    pDev->pHwFunc->hwSetTxDescVeol(pDesc, value);
}

void
halDebugPrintTxDesc(WLAN_DEV_INFO *pDev, ATHEROS_DESC *pDesc)
{
    ASSERT(pDev);
    ASSERT(pDesc);

    /* Display all 6 words of the descriptor */
    uiPrintf(DESC_ADDR		"%08lx  ", pDesc->thisPhysPtr);
    uiPrintf(NEXT_DESC		"%08lx  ", pDesc->nextPhysPtr);
    uiPrintf(BUFFER_PTR		"%08lx\n", pDesc->bufferPhysPtr);
    uiPrintf(ORIG_BUFVIRT_PTR"%p\n", pDesc->pOrigBufferVirtPtr);
    uiPrintf(CONTROL_1		"%08lx  ", pDesc->hwControl.word[0]);
    uiPrintf(CONTROL_2		"%08lx\n", pDesc->hwControl.word[1]);

    /* Display the hardware specific stuff */
    pDev->pHwFunc->hwDebugPrintTxDesc(pDesc);
}
