/*
 *  Copyright (c) 2000-2001 Atheros Communications, Inc., All Rights Reserved
 */

/* hwext.h - external declarations environment hardware access */

#ifndef __INChwexth
#define __INChwexth

#ident  "ACI $Header: //depot/sw/releases/2.0.1/src/include/diags/hwext.h#1 $"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* includes */

#include "wlandrv.h"


/**************************************************************************
* hwAllocatePhysBuffer - allocate a physically contiguous buffer
*
* This routine will create a physically contiguous buffer and get
* the virtual and physical address of it.  Note for test purposes
* this will just perform a malloc and return a dummy physical address
* pretent it is physically contiguous
*
* RETURNS: OK if got buffer, ERROR otherwise
*/
A_STATUS hwAllocatePhysBuffer
	(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	A_UINT16 buffSize,		 /* size of buffer to create */
	A_UINT32 *pPhysAddress,	 /* return the physical address of buffer */
	void     **ppVirtAddress,/* return the virtual address of buffer */
	void     **ppOSDescAddress /* return the virtual address of OS decriptor */
	);

/**************************************************************************
* hwFreePhysBuffer - free the physically contiguous buffer
*
* This routine will free the physically contiguous buffer previously
* allocated by hwAllocatePhysBuffer()
*
* RETURNS: N/A
*/
void hwFreePhysBuffer
	(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	void *pVirtAddress,				/* address of buffer to free */
	void *ppOSDescAddress			/* address of OS decriptor to free */
	);


/**************************************************************************
* hAllocateDescMemory - allocate memory for descriptor allocation
*
* This allocates some physically contiguous memory for descriptor 
* allocation. In the NT test environment, this malloc's an area 
* of memory that will be used for descriptors.
*
* RETURNS: N/A
*/
A_STATUS hwAllocateDescMemory
	(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	A_UINT16 size					/* size in bytes to allocate */
	);

/**************************************************************************
* hwFreeDescMemory - free descriptor memory area
*
* This routine will free the descriptor memory that was already created by 
* hwAllocateDescMemory()
*
* RETURNS: N/A
*/
void hwFreeDescMemory(
	struct wlanDevInfo	*pdevInfo	/* pointer to device info structure */
	);

/**************************************************************************
* memAllocateDescritor - allocate a descriptor from the descriptor pool
*
* This routine will get a hardware descriptor from the descriptor pool
* 
*
* RETURNS: N/A
*/
A_STATUS memAllocateDescriptor
	(
	WLAN_DEV_INFO	*pdevInfo,	/* pointer to device info structure */
	void    **ppVirtAddress		/* updated with virtual addr of desc. on return */
	);

/**************************************************************************
* memFreeDescritor - free a descriptor
*
* This routine will free a descriptor that was already created by 
* hwAllocateDescriptor()
*
* RETURNS: N/A
*/
void memFreeDescriptor
	(
	WLAN_DEV_INFO	*pdevInfo,	/* pointer to device info structure */
	void *pVirtAddress		/* address of descriptor to free */
	);

/**************************************************************************
* createBuffandDesc - create a physically contiguous buffer and descriptor
*
* Create a physically contiguous data buffer and create a descriptor
* to go with it.  An ATHEROS_DESC gets returned, pointer to data buffer
* can be assertained from the descriptor
*
* RETURNS: OK if successful, ERROR otherwise
*/
A_STATUS createBuffandDesc
	(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	A_UINT16 buffSize,
	void **ppVirtAddr		/* updated with virtual address of desc on return */
	);


/**************************************************************************
* freeBuffandDesc - free buffer and descriptor pair
*
* Free the buffer and descriptor, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void freeBuffandDesc
	(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	void *pVirtAddr		/* virtual address to start of descriptor */
	);

/**************************************************************************
* freeBuffandDescChain - free chain of buffer and descriptor pairs
*
* Free the buffer and descriptor, created with createBuffandDesc()
*
* RETURNS: N/A
*/
void freeBuffandDescChain(
	struct wlanDevInfo	*pdevInfo,	/* pointer to device info structure */
	void *pVirtAddr		/* virtual address to start of descriptor */
	);


/**************************************************************************
* descInfoDup - duplicate the os dependent info. field in the descriptor
*
* copy the os dependent info fields from the original descirptor to the
* to be duplicated descriptor
*
* RETURNS: N/A 
*/
extern void descInfoDup
	(
	ATHEROS_DESC *pOrigDesc,		/* pointer to the original descriptor */
	ATHEROS_DESC *pDupDesc			/* pointer to the to be duplicated descriptor */
	);

/**************************************************************************
* bufferRefUpdate - update the reference info for the buffer
*
* for vxWorks this is increase the the reference counter for the buffer of
* the input descriptor
*
* for the NDIS this could be set flag
*
* RETURNS: N/A 
*/
extern void bufferRefUpdate
	(
	ATHEROS_DESC *pDesc		/* pointer to the descriptor */
	);


#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __INChwexth */
