/*
 *  Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * linuxdrv.h - vxWorks specific function wrappers.
 *
 * "ACI $Id: //depot/sw/releases/2.0.1/src/include/vxdrv.h#1 $"
 */

#ifndef __INClinuxdrvh
#define __INClinuxdrvh

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

//#include "semLib.h"
#include <linux/cache.h>
//#include "tickLib.h"
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <asm/param.h>
#include <asm/io.h>
#include <asm/byteorder.h>
/* use Makefile for now!!
#ifdef __BIG_ENDIAN
#define BIG_ENDIAN __BIG_ENDIAN
#endif
*/
#include <linux/types.h>
#include <linux/ipc.h>
#include <linux/sem.h>

#include "wlantype.h"
#include "assert.h"
#include "apdefs.h"

#ifdef DEBUG    /* allow static functions to be exported for stack backtrace */
#undef LOCAL
#define LOCAL
#endif

#ifdef DEBUG_PRINT
#define aPrintk printk
#else  //DEBUG_PRINT
#define aPrintk(x,...)   
#endif //DEBUG_PRINT

#ifdef BUILD_AP //fadi
typedef END_OBJ OS_DEV_INFO;
#endif

#undef  WAIT_FOREVER
#define WAIT_FOREVER (-1)


#define A_MALLOC(a)                     (kmalloc(a, GFP_KERNEL))
#define A_FREE(a)                       (kfree(a))
#define A_DRIVER_MALLOC(a, b, c)        (kmalloc(a, GFP_KERNEL))
#define A_DRIVER_FREE(a, b, c)          (kfree(a))
#define A_MEM_SET(addr, value, size)    memset((char *)(addr), (int)(value), (unsigned long)(size))
#define A_MEM_ZERO(addr, size)          memset((char *)(addr), 0, (unsigned long)(size))
#define A_DRIVER_BCOPY(from, to, len)   memcpy((char *)(to), (char *)(from), (unsigned long)(len))
#define A_BCOPY(a, b, n)                memcpy((char *)(b), (char *)(a), (unsigned long)(n))
#define A_BCOMP(p1, p2, n)              memcmp((char *)(p1), (char *)(p2), (unsigned long)(n))

/*
 * Words (16b) should be aligned correctly on all platforms, so it
 * can be used even on platforms (mips) with strict alignment
 * requirements.
 */
#define A_MACADDR_COPY(from, to) \
    do { \
        (to)->words[0] = (from)->words[0]; \
        (to)->words[1] = (from)->words[1]; \
        (to)->words[2] = (from)->words[2]; \
    } while (0)

#define A_MACADDR_COMP(m1, m2) \
    ((((m1)->words[2] == (m2)->words[2]) && \
      ((m1)->words[1] == (m2)->words[1]) && \
      ((m1)->words[0] == (m2)->words[0])) == 0)

/*
 *  SIB entry and bridge aging period is 300 seconds
 */
#define DEFAULT_AGING_PERIOD            (60 * 5)

/*
 * Get the number of millisecond ticks since the system was started.
 *
 * For the AP, absolute accuracy is not all that important.  Use
 * an approximation of the conversion of ticks @ 60Hz to that does
 * not require any expensive math (multiply or divide).
 *
 * 60 ticks -> 997ms, or a 0.3% shortfall error, although there are
 * cases where the error is worse.
 *
 * vxAbsTicks.lower the same as getTick().
 */
#define A_MS_TICKGET()  ((jiffies*1000)/HZ)

#define A_ROUNDUP(x, y)         ((((x)+((y)-1))/(y))*(y))

#define ASSERT assert

/*
 * Cache flushing and virtual address translation support.  vxWorks macros
 * handle things cleanly but with some overhead, so optimize for supported
 * platforms.
 *   - A_DATA_() routines are for frame data buffers.
 *   - A_DESC_() macros are for DMA descriptors.  Currently
 *     hardware descriptors are cached as they share memory
 *     with the software frame description.
 *
 * SWCOHERENCY   - descriptors and data are cached
 * SWCOHERENCYWB - SWCOHERENCY with a write back cache
 */
#if defined(AR5311) || (USER_D_CACHE_MODE & CACHE_WRITETHROUGH)
/*
 * AR5311 uses a write back cache, with it's own lower overhead
 * cache invalidation routine.  V != P, but upper bits are
 * ignored by the DMA bus so a null V2P/P2V macros are used.
 */
#define SWCOHERENCY
#include "ar5311.h"
#define A_DATA_CACHE_INVAL(addr, len) \
    cacheDataInvalidate(addr.ptr, (len))
#define A_DATA_CACHE_FLUSH(addr, len)
#define A_DATA_V2P(addr)                ((A_UINT32)(addr))
#define A_DATA_P2V(addr)                (addr)
#define A_DESC_CACHE_INVAL(addr) \
    cacheDescInvalidate((void *)(&(addr)->hwStatus))
#define A_DESC_CACHE_FLUSH(addr)
#define A_DESC_V2P(addr)                printk("IN AR5311 A_DESC_V2P\n");((A_UINT32)(addr))
#define A_DESC_P2V(addr)                (addr)
#define A_PIPEFLUSH()                   sysWbFlush()

#elif defined(PPC8240) || (USER_D_CACHE_MODE & CACHE_SNOOP_ENABLE)
/*
 * MPC8245 is a fully coherent writeback cache with V==P.
 */

#define A_DATA_CACHE_INVAL(addr, len)
#define A_DATA_CACHE_FLUSH(addr, len)
#define A_DATA_V2P(addr)                ((A_UINT32)virt_to_phys((volatile void *)(addr)))
#define A_DATA_P2V(addr)                ((A_UINT32)phys_to_virt((A_UINT32)addr))
#define A_DESC_CACHE_INVAL(addr)
#define A_DESC_CACHE_FLUSH(addr)
#define A_DESC_V2P(addr)                ((A_UINT32)virt_to_phys((volatile void *)(addr)))
#define A_DESC_P2V(addr)                ((A_UINT32)phys_to_virt((A_UINT32)addr))
#define A_PIPEFLUSH()

#elif defined(CACHE_PROTO_COPYBACK) || (USER_D_CACHE_MODE & CACHE_COPYBACK)
/*
 * Sample writeback cache implementation keeping the desc/buffers in
 * the cached.  This code has not been tested, so it is not enabled
 * by default.
 */
#define SWCOHERENCY
#define SWCOHERENCYWB
#define A_DATA_CACHE_INVAL(addr, len)   CACHE_USER_INVALIDATE(addr, len)
#define A_DATA_CACHE_FLUSH(addr, len)   CACHE_USER_FLUSH(addr, len)
#define A_DATA_V2P(addr) \
    ((A_UINT32)CACHE_DRV_VIRT_TO_PHYS(&cacheUserFuncs, addr))
#define A_DATA_P2V(addr) \
    CACHE_DRV_PHYS_TO_VIRT(&cacheUserFuncs, addr)
#define A_DESC_CACHE_INVAL(addr) \
    A_DATA_CACHE_INVAL(addr, 32)
#define A_DESC_CACHE_FLUSH(addr) \
    A_DATA_CACHE_FLUSH(addr, 32)
#define A_DESC_V2P(addr)                printk("IN CACHE_PROTO_COPYBACK A_DESC_V2P\n");A_DATA_V2P(addr)
#define A_DESC_P2V(addr)                A_DATA_P2V(addr)
#define A_PIPEFLUSH()                   CACHE_PIPE_FLUSH()

#else

/*
 * Sample code to use cacheDmaAlloc() which is uncached on non-coherent
 * platforms.  This is the default case for non-coherent copyback
 * caches.  This code has not been extensively tested, but should
 * be safe on most platforms.
 */
#define VX_CACHE_DMA_MALLOC
#define SWCOHERENCY
#define A_DATA_CACHE_INVAL(addr, len) \
    CACHE_DRV_INVAL(&cacheDmaFuncs, addr, len)
#define A_DATA_CACHE_FLUSH(addr, len) \
    CACHE_DRV_FLUSH(&cacheDmaFuncs, addr, len)
#define A_DATA_V2P(addr) \
    ((A_UINT32)CACHE_DRV_VIRT_TO_PHYS(&cacheDmaFuncs, addr))
#define A_DATA_P2V(addr) \
    CACHE_DRV_PHYS_TO_VIRT(&cacheDmaFuncs, addr)
#define A_DESC_CACHE_INVAL(addr) \
    A_DATA_CACHE_INVAL(addr, 32)
#define A_DESC_CACHE_FLUSH(addr) \
    A_DATA_CACHE_FLUSH(addr, 32)
#define A_DESC_V2P(addr)                printk("IN ELSE A_DESC_V2P\n");A_DATA_V2P(addr)
#define A_DESC_P2V(addr)                A_DATA_P2V(addr)
#define A_PIPEFLUSH()                   CACHE_PIPE_FLUSH()

#endif /* PLATFORM */

/*
 * Locking macros
 */

#define CREATE_LOCKED   1
#define CREATE_UNLOCKED 0
#ifdef DEBUG

extern struct sembuf sem_lock;
extern struct sembuf sem_unlock;

#define A_SEM_INIT(sem_val, sem_key, make_locked) \
            do {} while (0)

#if FADI
          do {                                                    \
             printk("CALLING SEM_INI_T FILE %s, line %d\n", __FILE__, __LINE__);\
             union semun sem_arg;                                 \
             sem_arg.val = ((make_locked)?0:1);                   \
            (sem_val) = (sys_semget((key_t)sem_key, (1), (IPC_CREAT)));\
            printk("sem_val = %d[sem_arg.val=%d][key=%d]\n", sem_val, sem_arg.val, sem_key);\
            if(sys_semctl((sem_val), (0), (SETVAL), (sem_arg))==-1)\
                printk("FAILED TO INIT SEM File %s, line %d\n",   \
                       __FILE__, __LINE__);\
          }while(0)
#endif//FADI

#else

#define A_SEM_INIT(sem_val, sem_key, make_locked) \
              do {} while (0)

#endif /* not DEBUG */

#define A_SEM_LOCK(sem, timeout) \
              do {} while (0)

#define A_SEM_UNLOCK(sem) \
              do {} while (0)

#define A_SEM_DELETE(sem) \
              do {} while (0)

#define A_SIB_TAB_LOCK(sibTab)      printk("A_SIB_TAB_LOCK...file: %s, line:%d\n", __FILE__, __LINE__);

#define A_SIB_TAB_UNLOCK(sibTab)    printk("A_SIB_TAB_UNLOCK...file: %s, line:%d\n", __FILE__, __LINE__);

#define A_SIB_ENTRY_LOCK(sib)       printk("A_SIB_ENTRY_LOCK...file: %s, line:%d\n", __FILE__, __LINE__);

#define A_SIB_ENTRY_UNLOCK(sib)     printk("A_SIB_ENTRY_UNLOCK...file: %s, line:%d\n", __FILE__, __LINE__);

#define A_TIMEOUT(ptimer, func, period, param) \
    printk("A_TIMEOUT...file: %s, line:%d\n", __FILE__, __LINE__);
#define A_UNTIMEOUT(handle) \
    printk("A_UNTIMEOUT...file: %s, line:%d\n", __FILE__, __LINE__);

#ifdef __cplusplus
}
#endif

#endif /* not __INClinuxdrvh */
