/*
 * Copyright (c) 2001-2002 Atheros Communications, Inc., All Rights Reserved
 */

/* radardefs.h - DFS and radar detection data structures and definitions */

#ifndef		__INCradardefsh
#define		__INCradardefsh

#ident  "ACI $Header: //depot/sw/releases/2.0.1/src/include/radardefs.h#1 $"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "wlansme.h"
#include "wlandrv.h"

#define RADAR_ARRAY_SIZE	3000
#define MIL_1		1000000
#define PERCENTILE	5e-2
#define PERCENTILE_K	50	// PERCENTILE * 1000 integer value
#define DELTA_USEC	50		// 10 microseconds margin for timestamp
#define INFINITY	10000.0
#define PPS_ARRAY_SIZE	20

#define RADAR_DATA_LOCK		(A_SEM_LOCK(radarDataSemId, WAIT_FOREVER))
#define RADAR_DATA_UNLOCK	(A_SEM_UNLOCK(radarDataSemId))

typedef struct genStruct {
	A_INT32	arraySize;
	A_INT32	count;
	A_INT32	value[RADAR_ARRAY_SIZE];
} GEN_ARRAY;

// Structure to hold the event array. Event time deltas are stored in the
// circular array eStamp[].
typedef struct event_struct
{
	A_INT32	arraySize;		// max size of array
	A_INT32	count;			// current no of events.
	A_INT32	eFirst;			// first valid event
	A_INT32	eLast;			// last valid event (can wrap around)
	A_INT32	curStampLow;	// lower 32 bits of current timestamp
	A_INT32	curStampHigh;	// upper 32 bits of current timestamp
	A_INT32	totalDelta;		// time delta spanning across first to last
	A_INT32	eStamp[RADAR_ARRAY_SIZE]; // circular array for time deltas
} EVENT_ARRAY;

// Information about radar patterns
typedef struct ppsArray
{
	A_INT32	arraySize;
	A_INT32	count;
	struct ppsInfo
	{
		A_INT32		freq;		// value of pps
		A_INT32		totalNumPulses;	// burst size
		A_INT32		silentPeriod;	// scan rate in seconds
		A_INT32		minPulsesToDetect;	// minimum pulses required for +ve identification
		A_INT32		stepFrom;	// finer search from this pps
		A_INT32		stepIncr;	// with this increment
		A_INT32		stepTo;		// to this pps
		A_INT32		wideMargin;	// wide margin percentage * 1000 for first search, e.g. 5% = 50
		A_INT32		narrowMargin; // narrow margin percentage * 1000 for finer search  
	} info[PPS_ARRAY_SIZE];
} PPS_ARRAY;

#define NEXT_EVENT_INDEX(pArray, index) (((index+1) >= (pArray)->arraySize)? 0 : (index+1))

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif // __INCradardefsh
