/*
 *  Copyright  2000-2001 Atheros Communications, Inc.,  All Rights Reserved.
 */

/* stacserv.h - Connection services */

#ifndef     __INCstacservh
#define     __INCstacservh

#ident  "ACI $Header: //depot/sw/releases/2.0.1/src/include/stacserv.h#3 $"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Define policy and other mnemonics
 */

#define CSERV_JOIN_CONFIGBSS        0x1 // Join configured BSS only
#define CSERV_JOIN_FIRSTBSS         0x2 // Join first BSS found
#define CSERV_JOIN_BESTSIGNAL       0x3 // Join BSS with best signal strength
#define CSERV_JOIN_BESTRATE         0x4 // Join BSS with best rate match
#define CSERV_JOIN_CONFIGSSID       0x5 // Join configured SSID with best signal strength

#define CSERV_APCONN_MANUAL         0x1 // Manual connection thru OID's only
#define CSERV_APCONN_INITONLY       0x2 // connect to AP only at initialization
#define CSERV_APCONN_AUTOON         0x3 // Automatically reconnect to stay on.

#define CSERV_OP_SCAN               0x1
#define CSERV_OP_JOIN               0x2
#define CSERV_OP_AUTH               0x4
#define CSERV_OP_ASSOC              0x8

// state of what CS is doing right now
#define CSERV_STATE_IDLE            0x0
#define CSERV_STATE_BKSCANNING      0x1
#define CSERV_STATE_OTHER_CHANNEL   0x2
#define CSERV_STATE_CONNECTING      0x4
#define CSERV_STATE_ROAMING         0x8


#define CSERV_SYNC_EVENT            0x1
#define CSERV_SYNC_TIMER            0x2
#define CSERV_SYNC_QUERY_OID        0x3
#define CSERV_SYNC_SET_OID          0x4

// Connection services, power management, tx rate adaptation, etc. run
// run off of a polling routine staPollFunction(). Define timer values for
// all such services

// staPollTimer provides a logical clock for rest of the timers.
// CSERV_TIME_STAPOLL is period in milliseconds for staPollTimer. 100 ms works well.

#define CSERV_TIME_STAPOLL          100     // 100 ms

// rest of the timers are defined as multiples of CSERV_TIME_STAPOLL

#define CSERV_TIME_AUTOCONN         20      // 2 second period for checking connection status
#define CSERV_TIME_RATECHECK        10      // 1 second auto tx rate check period
#define CSERV_TIME_LONGSCAN         300     // 30 seconds to restart scan after giving up
#define CSERV_TIME_BKSCAN           3       // 300 msec to advance to next channel in bk scan

#define CSERV_TIME_SCANMIN_ACTIVE   50      // 50 ms for 1 probe response
#define CSERV_TIME_SCANMAX_ACTIVE   110     // 110 ms maximum wait in active normal scan.
#define CSERV_TIME_BKSCANMAX_ACTIVE 50      // 200 ms to stay on other channel during active bk scan
                                            // can be reduced to 30-50 ms if probe requests are enough
#define CSERV_TIME_SCANMIN_PASSIVE  200     // 200 ms for some chance of catching a beacon.
#define CSERV_TIME_SCANMAX_PASSIVE  300     // 300 ms maximum wait in passive normal scan
                                            // to collect slowest beacons
#define CSERV_TIME_BKSCANMAX_PASSIVE 150    // 150 ms to stay on other channel during bk scan
                                            // without disturbing upper layer protocols
#define CSERV_TIME_SCANLIMIT        3000    // upper limit for scanmax time & beacon period 3 seconds
#define CSERV_MAX_ATTEMPTS          6       // max 10 auth-assoc attempts before giving up
#define CSERV_TIME_AA_MAX           100     // max time for auth and assoc operations
#define CSERV_MIN_RATE_INDEX        4       // index for minimum rate (24) below which we'll kickoff bk scan
#define CSERV_MIN_RSSI_BKSCN        24      // minimum ack rssi below which we'll kick off bk scan
#define CSERV_VALID_SCAN_TIME       60000   // 1 minute. time to keep scan list valid in milliseconds
#define CSERV_DISCONNECT_TIMEOUT    100     // wait 10 seconds before reporting disconnect to OS
#define CSERV_JOIN_BEACON_CNT       5       // max number of beacon intervals to wait while joining.

typedef struct cservEvent {
#ifndef BUILD_AP
//fadi    NDIS_EVENT  *posEvent;
#endif  /* BUILD_AP */
    A_UINT32    resultCode;
    A_UINT32    syncType;       // synchronization type
} CSERV_SYNC;

typedef struct cservInfoStruct {
    A_UINT32    joinPolicy;
    A_UINT32    apConnPolicy;
    A_UINT32    apConnAttemptsMax;
    A_UINT32    apConnAttemptCur;
    A_UINT32    txRatePolicy;
    A_INT32     cservPollInterval;
    A_INT32     cservPollCur;
    A_INT32     noActivityInterval;
    A_INT32     noActivityCur;
    A_INT32     rateCheckInterval;
    A_INT32     rateCheckCur;
    A_INT32     longScanInterval;
    A_INT32     longScanCur;
    A_INT32     bkScanInterval;
    A_INT32     bkScanCur;
    A_INT32     bkPowerReq;
    A_INT32     bkScanMaxTime;
    A_UINT32    cservOp;        // Bits to indicate which CS operation to do next.
    A_UINT32    cservState;     // state of CS, bkscan, normal scan, switching AP
    CSERV_SYNC  *pSync; // wake up this event on completion
    struct athTimerStruct staPollTimer;
    A_UINT32    minChanTime; // min channel time for scan
    A_UINT32    maxChanTime; // max channel time for scan and join
    A_UINT32    opRetryScaleIndex;
    A_UINT32    scanInterval; // how many milliseconds between scans.
    A_UINT32    bkScanState;
    A_INT16     bkChannelIndex;
    CHAN_VALUES *bkChannel;
    A_BOOL      autoScanEnable;
    SCAN_INFO   scanInfo; // Beacons received during scan
    SSID        lastSsid;   // last ssid set or attempted
    SSID        desiredSsid;    // Ssid to be joined after scan.
    A_INT32     disconnectInterval; // how long to wait before indicating media disconnect
    A_INT32     disconnectCur;
    A_INT32     roamRssiA;  // rssi level below which, we'll try to roam
    A_INT32     roamRssiB;
    A_INT32     roamRateA;
    A_INT32     roamRateB;
    A_UINT32    scanValidTime;  // time to keep scan list valid in milliseconds
    A_BOOL      noActivity;         // used to keep track of the AP at the fringe
    A_UINT32    lastRxBeacons;      // used to keep track of the AP at the fringe
} CSERV_INFO;

/* Definition of AdhocBand registry field is copied from src\sta\ndis\stacfg\stacfg.h
 * Should go into a shared include file later.
 */

enum ATH_ADHOCBAND                  // Ad Hoc Network Band
{
    ATH_ADHOCBAND_BONLY,
    ATH_ADHOCBAND_AONLY,
    ATH_ADHOCBAND_TONLY,            // 11a Turbo
    ATH_ADHOCBAND_GONLY,            // 11b Pure G
    ATH_ADHOCBAND_END
};

extern A_UINT32 cservAssociate
    (
    WLAN_DEV_INFO   *pdevInfo,
    CSERV_SYNC      *pSync,
    A_UINT32        assocTimeout,
    A_UINT16        listenInterval
    );

extern A_UINT32 cservDisassociate
    (
    WLAN_DEV_INFO   *pdevInfo
    );

extern A_UINT32 cservAuthenticate
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync
        );

extern A_UINT32 cservDeauthenticate
        (
        WLAN_DEV_INFO   *pdevInfo
        );

extern A_UINT32 cservSendProbe
        (
        WLAN_DEV_INFO   *pdevInfo
        );

/* cservScan - routine to test scan operation */

extern A_UINT32 cservScan
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync,
        unsigned int    scanType
        );

extern A_UINT32 cservJoin
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync
        );

extern A_UINT32 cservStart
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync
        );

extern A_UINT32 cservStartAdhoc
        (
        WLAN_DEV_INFO   *pDevInfo,
        SSID            *pSsid
        );

/*******************************************************************
 * cservScanJoin - Scan for beacons and join the bss
 *
 * DESCRIPTION: This routine performs active or passive scan,
 * collects beacons, then it chooses the bss using joinPolicy and
 * performs a join operation. assocPolicy decideds the action to take
 * when link goes down.
 */
extern A_UINT32 cservScanJoin
        (
        WLAN_DEV_INFO   *pDevInfo,
        CSERV_SYNC      *pSync,
        SCANTYPE        scanType,
        A_UINT32        joinPolicy
        );

/*******************************************************************
 * cservAuthAssoc - authenticate and associate with bss
 *
 * DESCRIPTION: This routine performs authentication followed by
 * association. assocPolicy decideds the action to take
 * when current authentication gets terminated.
 */
extern A_UINT32 cservAuthAssoc
        (
        WLAN_DEV_INFO   *pDevInfo,
        CSERV_SYNC      *pSync
        );

/*******************************************************************
 * cservScanJoinAuthAssoc - Perform scan-join-auth-assoc operation
 *
 * DESCRIPTION: This routine performs active or passive scan,
 * collects beacons, then it chooses the bss using joinPolicy and
 * performs a join operation. Then it performs authentication followed
 * by association. This is an equivalent of bringing up the station
 * on the access point. assocPolicy decideds the action to take
 * when link goes down.
 */
extern A_UINT32 cservScanJoinAuthAssoc
        (
        WLAN_DEV_INFO   *pDevInfo,
        CSERV_SYNC      *pSync,
        SCANTYPE        scanType,
        A_UINT32        joinPolicy,
        A_UINT32        assocPolicy
        );

/*******************************************************************
 * cservDAuthAssoc - disassociate and deauthenticate with bss
 *
 * DESCRIPTION: This routine performs disassociation followed by
 * deauthentication. Stored value of assocPolicy is changed to manual.
 */
extern A_UINT32 cservDAuthAssoc
        (
        WLAN_DEV_INFO   *pDevInfo
        );

/*******************************************************************
 * cservChangeChannel - Change the channel and connect to AP
 *
 * DESCRIPTION: This routine stops the data traffic by calling
 * driver routine to mark the link down. Currently it tears down
 * the connection. Then it sets a new channel and performs
 * cservScanJoinAuthAssoc(). It can be enhanced to switch and scan
 * the other channel without tearing down the existing connection.
 */
extern A_UINT32 cservChangeChannel
        (
        WLAN_DEV_INFO   *pDevInfo,
        CSERV_SYNC      *pEvent,
        A_UINT32        channelNo,
        A_UINT32        channelPolicy
        );

/*******************************************************************
 * cservBeaconReception - Turn on/off beacon reception to application
 *
 * DESCRIPTION: This routine enables/disables beacon deliver to the
 * driver. If enabled, it sets up a callback to the given routine
 * for each beacon that comes in.
 */

extern A_UINT32 cservBeaconReception
        (
        WLAN_DEV_INFO   *pDevInfo,
        A_BOOL          enable,
        void            *beaconCallback(WLAN_FRAME_BEACON *pBeacon)
        );

/*******************************************************************
 * cservChangeRate - Change the rate settings
 *
 * DESCRIPTION: This routine changes the current operational rate
 * set and the current rate.
 *
 */
extern A_UINT32 cservChangeRate
        (
        WLAN_DEV_INFO   *pDevInfo,
        RATE_SET        *pOpRateSet,
        A_UCHAR         curRate
        );

extern A_UINT32 cservBkScan
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync,
        CHAN_VALUES     *pChval,
        unsigned int    scanType
        );

extern A_UINT32 cservBkScanAll
    (
    WLAN_DEV_INFO   *pDevInfo,
    CSERV_SYNC      *pSync,
    A_UINT32        bkScanMax,
    A_UINT32        bkScanInterval
    );

extern A_STATUS     cservInit
        (
        WLAN_DEV_INFO   *pDevInfo
        );

extern A_INT32  cservSelectBSS
        (
        WLAN_DEV_INFO   *pDevInfo,
        BSSDESCR_SET    *pbssSet,
        BSSTYPE         bssType,
        SSID            *pSsid,
        BSSID           *pBssid,
        A_INT32         minrssi,
        A_UINT16        cflags
        );

extern A_UINT32 cservSwitchAP
        (
        WLAN_DEV_INFO   *pdevInfo,
        BSSDESCR        *pbssDescr,
        CSERV_SYNC      *pSync
        );

#ifndef BUILD_AP
#ifndef OID_802_11_SSID
#ifdef WIN32 //fadi
#error ntddndis.h from Windows XP DDK required.
#endif
#endif

extern A_UINT32 cservSwitchSSID
        (
        WLAN_DEV_INFO   *pdevInfo,
        CSERV_SYNC      *pSync,
        SSID            *pSsid
        );

extern void cservStopBSS
        (
        WLAN_DEV_INFO   *pDevInfo
        );

extern void cservPollReset
    (
    WLAN_DEV_INFO   *pDevInfo
    );

extern void cservBmissNotify
    (
    WLAN_DEV_INFO   *pDevInfo
    );

extern void cservPollFunction
    (
    WLAN_DEV_INFO   *pDevInfo
    );

extern void cservTxrateNotify
    (
    WLAN_DEV_INFO   *pDevInfo,
    int             newRate,
    A_INT32         ackRssi
    );

extern void cservHwResetNotify
        (
        WLAN_DEV_INFO   *pDevInfo
        );

extern CHAN_VALUES  *cservSelectAdhocChannel
    (
    WLAN_DEV_INFO   *pDevInfo,
    CSERV_INFO      *pCsInfo
    );

extern void cservCreateAdhocBssid
    (
    WLAN_DEV_INFO   *pDevInfo,
    WLAN_MACADDR    *pMacaddr
    );

extern void staPollInit
    (
    WLAN_DEV_INFO *pDevInfo
    );
#endif   /* BUILD_AP */

extern void RateCtrlInit(void);

extern void RateCtrlTask(void);

extern void txRateFree(void);

extern void cservSyncComplete
        (
        WLAN_DEV_INFO   *pDevInfo,
        A_UINT32        resultCode
        );
extern void cservOperationCancel
        (
        WLAN_DEV_INFO   *pDevInfo
        );

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif      /* __INCstacservh */

