/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/* stamlme.c - Station Managment Entity */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/stamlme.c#4 $"

#ifdef VXWORKS
#include "vxWorks.h"
#include "sysLib.h"
#include "taskLib.h"
#include "stdlib.h"
#else
//#include <malloc.h>
//#include <memory.h>
//#include <string.h>
#endif

//#include <stdlib.h>
//#include <stdio.h>
//#include "semLib.h"

#include <linux/cache.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/param.h>
#include <linux/delay.h>
#include <asm/byteorder.h>

/*
#ifdef __BIG_ENDIAN
#define BIG_ENDIAN __BIG_ENDIAN
#endif
*/

#include "wlantype.h"
#include "wlanproto.h"
#include "wlansme.h"
#include "wlansmeext.h"
#include "wlandrv.h"
#include "wlanext.h"
#include "wlanbeacon.h"
#include "wlanchannel.h"
#include "ui.h"
#include "display.h"
#include "apmem.h"

#include "stacserv.h"
#include "halApi.h"


//fadi #ifdef    BUILD_AP
#include "apcfg.h"
#include "apvlan.h"
A_SEM_TYPE    aidSemId;
//fadi #endif    /* BUILD_AP */

#ifdef  DEBUG
int probeDebug = 1;
int mlmeDebug  = 0;
extern char *mgtTypeStr[];
#endif  /* DEBUG */

#if FADI

/*
 *  Note that the MLME layer expects to own the data it passes.  We must make
 *  copies of it here.
 */
MLME_RESULT
wlanMlmeInit(SME_INFO_STRUCT *psmeInfo)
{


    if (psmeInfo == NULL ||
        (psmeInfo->serviceType != WLAN_STA_SERVICE &&
         psmeInfo->serviceType != WLAN_AP_SERVICE )) {
        uiPrintf("wlanMlmeInit: ERROR: invalid service type\n");
        return MLME_INVALID_PARAM;
    }
    A_DRIVER_BCOPY(psmeInfo, &smeInfo, sizeof(SME_INFO_STRUCT));

//fadi #ifdef  BUILD_AP
    if (aidSemId == 0) {
        A_SEM_INIT(aidSemId, 0, CREATE_UNLOCKED);
    }
//fadi #endif  /* BUILD_AP */

    return MLME_SUCCESS;
}

/***************************/
/* Authentication routines */
/***************************/

/*
 * MLME Authentication primitives perform authentication with the
 * AP. There is no authentication in Ad-hoc.
 *
 * PARAMETERS: Parameters for all of the authentication routines are described
 *            here together.
 *
 * pPeerSTAAddress - MAC address of the other STA to be auth'ed with. It is
 *            the AP's STA address.
 * authType - OPEN_SYSTEM or SHARED_KEY.
 * authFailureTimeout - TU. Abort the operation if not completed within this
 *            time. Observed for the overall duration of frame exchange.
 *
 */

/* wlanMlmeAuthRequest - start authentication process
 *
 * Starts the authentication sequence with the peer STA. Returns
 * after initiating the operation. It sends out first frame.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_NOT_SUPPORETD
 */

MLME_RESULT
wlanMlmeAuthRequest(WLAN_DEV_INFO *pdevInfo,
                    WLAN_MACADDR  *pPeerSTAAddress,   /* Address of other STA */
                    A_UINT16      algNo,              /* open system or shared key */
                    A_UINT32      authFailureTimeout) /* timeout in TU */
{
    SIB_ENTRY       *pSib;
    SIB_ENTRY       *lSib;
    A_UINT16        frameLen;
    ATHEROS_DESC    *pDesc;
    WLAN_FRAME_AUTH *pFrame;    /* Pointer to assoc req frame */
    A_BOOL          faulted;

    /* Check that environment is up */

    /* Validate parameters */
    if (!pPeerSTAAddress) {
        return MLME_INVALID_PARAM;
    }

    /* Search sib table for pPeerSTAAddress, it should be an AP */
    if ((pSib = sibEntryFind(pPeerSTAAddress)) == NULL) {
        return MLME_INVALID_PARAM;
    }

    pSib->authenAlgo = algNo;

    if (!pSib->capInfo.ess || pSib->capInfo.ibss ||
        pSib->serviceType != WLAN_AP_SERVICE) {
        // Destination address is not an infrastructure access point.
        return MLME_INVALID_PARAM;
    }

    /* Ensure that WEP key for peer sta is in the hw cache. */
    if (pdevInfo->staConfig.privacyInvoked) {
        wlanKeyEnable(pdevInfo, pSib, pdevInfo->staConfig.defaultKey);
        wlanKeyCacheFault(pdevInfo, pSib, &faulted);
    }

    if (algNo == AUTH_SHARED_KEY) {
        /* shared key can take longer (esp with debug on) */
        authFailureTimeout *= 2;
    }

    lSib = pdevInfo->localSta;

     /* Lock sta */

    A_SIB_ENTRY_LOCK(lSib);

    /* check current state */
    if (lSib->transState != TSTATE_QUIET) {
        /* we are doing something. */
        A_SIB_ENTRY_UNLOCK(lSib);
        return MLME_REFUSED;
    }
    /* if already authenticated, allow preauthentication to other AP's. */
    if (!(lSib->staState & STATE_JOINED) ||
        ((lSib->staState & STATE_AUTH) &&
         (A_MACADDR_COMP(&lSib->bssId, pPeerSTAAddress) == 0)))
    {
        /* Trying to authenticate again with the same AP. */
        /* @@@@ We should actually send a message to connection services
         * force deauthentication now.
         */
        A_SIB_ENTRY_UNLOCK(lSib);
        return MLME_REFUSED;
    }

    /* modify sta state */

    /* Compute frame length. No challenge text */
    frameLen = sizeof(WLAN_FRAME_AUTH);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **)&pDesc)) != 0) {
        A_SIB_ENTRY_UNLOCK(lSib);
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.auth;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
            pPeerSTAAddress, SUBT_AUTH);

    pFrame->algNo      = algNo;    /* Open sys or shared key */
    pFrame->statusCode = 0;        /* reserved in frame 1 */
    pFrame->transSeqNo = 1;        /* frame 1 */

    /* send first frame */
    /* displayMgtFrame((WLAN_MGT_MAC_HEADER *)pFrame, frameLen); */
    pDesc->pDestSibEntry = pSib;
    lSib->transState     = TSTATE_AUTH_FRAME1_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (lSib->smePendDesc == NULL) {
            lSib->smePendDesc = (ATHEROS_DESC *)0xffffffff;
        }
        /* set timeout */
        A_TIMEOUT(&pdevInfo->smeTimer, authFailureTimeout,
                  (A_UINT32) pdevInfo, MLME_AUTH_TIMEOUT);
        A_SIB_ENTRY_UNLOCK(lSib);
        return MLME_SUCCESS;
    } else {
        lSib->transState = TSTATE_QUIET;
        A_SIB_ENTRY_UNLOCK(lSib);
        return MLME_DRIVER_ERROR;
    }
}

A_UINT16
mlmeAuthSta2Receive(WLAN_DEV_INFO        *pdevInfo,
                    WLAN_FRAME_AUTH_CHLG *authFrame,   /* Incoming frame pointer */
                    A_UINT16             len)          /* Total frame length */
{
    WLAN_FRAME_AUTH_ENCRYPT *pFrame;    /* return frame encrypted */
    SIB_ENTRY    *lSib;                 /* Local SIB */
    SIB_ENTRY    *pSib = NULL;          /* Peer (AP) SIB */
    ATHEROS_DESC *pDesc;
    A_UINT16     frameLen;

    /* validate parameters */
    lSib = pdevInfo->localSta;

    if (lSib->staState & STATE_AUTH ||
        lSib->transState != TSTATE_AUTH_FRAME1_SENT) {
        /* Wrong state */
        return MLME_REFUSED;
    }
    /* Ensure that it comes from desired AP (preauthenticate not supported yet) */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(authFrame->macHeader.bssId))) != 0) {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }

    /* This is the last frame in open system */
    if (authFrame->algNo == AUTH_OPEN_SYSTEM) {
        if (authFrame->statusCode == WLAN_SUCCESS) {
            /* Change the state in the SIB of the AP */
            if ((pSib = sibEntryFind(&authFrame->macHeader.srcAddr)) == NULL) {
                /* Program error, should never happen */
                return MLME_INVALID_PARAM;
            }
            pSib->staState |= STATE_AUTH;

            /* Now update local station's SIB */
            lSib->staState |= STATE_AUTH;
            lSib->transState = TSTATE_QUIET;
        } else {
            lSib->transState = TSTATE_QUIET;
        }

        A_UNTIMEOUT(&(pdevInfo->smeTimer));
        /* Need to call callback routine here. */
        wlanMlmeAuthConfirm(pdevInfo, &(authFrame->macHeader.srcAddr),
                    authFrame->algNo, (MLME_RESULT) authFrame->statusCode);
        return MLME_SUCCESS;
    }
    else if (authFrame->algNo != AUTH_SHARED_KEY) {
        lSib->transState = TSTATE_QUIET;

        A_UNTIMEOUT(&(pdevInfo->smeTimer));

        wlanMlmeAuthConfirm(pdevInfo, &(authFrame->macHeader.srcAddr),
                            authFrame->algNo, MLME_NOT_SUPPORTED);

        return MLME_NOT_SUPPORTED;
    }

    /* Our request may fail if oversubscribed, mac denyed, etc. */
    if (authFrame->statusCode != WLAN_SUCCESS) {
        lSib->transState = TSTATE_QUIET;
        A_UNTIMEOUT(&(pdevInfo->smeTimer));
        wlanMlmeAuthConfirm(pdevInfo, &(authFrame->macHeader.srcAddr),
                            authFrame->algNo,
                            (MLME_RESULT)authFrame->statusCode);

        return MLME_SUCCESS;
    }

    /* Build the response frame:
     * In shared key, need to create another frame with same text
     * and send it back with encryption.  Include challenge text
     * and iv in the frame length.
     */
    frameLen = sizeof(WLAN_FRAME_AUTH) +
        WEP_ICV_FIELD_SIZE +
        INFO_ELEMENT_SIZE(authFrame->challenge);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.authChalIv;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    &lSib->bssId, SUBT_AUTH);
    pFrame->macHeader.frameControl.wep = 1;     /* encrypted */
    pFrame->algNo      = authFrame->algNo;      /* Open sys or shared key */
    pFrame->statusCode = 0;                     /* reserved in frame 3 */
    pFrame->transSeqNo = 3;                     /* frame 1 */

    /* Copy the challenge text into the frame */
    mlmeElementCopy((INFO_ELEMENT *)&authFrame->challenge,
                    (INFO_ELEMENT *)&pFrame->challenge);

    /* send response frame */
    /* displayMgtFrame((WLAN_MGT_MAC_HEADER *)pFrame, frameLen); */
    pDesc->pDestSibEntry = pSib;
    lSib->transState = TSTATE_AUTH_FRAME3_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (lSib->smePendDesc == NULL) {
            lSib->smePendDesc = (ATHEROS_DESC *)0xffffffff;
        }
        return MLME_SUCCESS;
    } else {
        lSib->transState = TSTATE_QUIET;
        return MLME_DRIVER_ERROR;
    }
}

A_UINT16
mlmeAuthSta4Receive(WLAN_DEV_INFO   *pdevInfo,
                    WLAN_FRAME_AUTH *authFrame, /* Incoming frame pointer */
                    A_UINT16        len)        /* Total frame length */
{
    SIB_ENTRY    *lSib = pdevInfo->localSta;
    SIB_ENTRY    *pSib;

    /* validate parameters */
    if (lSib->staState & STATE_AUTH ||
        lSib->transState != TSTATE_AUTH_FRAME3_SENT) {
        /* Wrong state */
        return MLME_REFUSED;
    }

    /* Ensure that it comes from desired AP (preauthenticate not supported yet) */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(authFrame->macHeader.srcAddr))) != 0) {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }

    lSib->transState = TSTATE_QUIET;
    /* This is the last frame in shared key */
    if (authFrame->statusCode == WLAN_SUCCESS) {
        /* Change the state in the SIB of the AP */
        if ((pSib = sibEntryFind(&authFrame->macHeader.srcAddr)) == NULL) {
            /* Program error, should never happen */
            return MLME_INVALID_PARAM;
        }
        pSib->staState |= STATE_AUTH;
        lSib->staState |= STATE_AUTH;
    }
    // else wlanMlmeAuthConfirm() is responsible for any error reporting.

    A_UNTIMEOUT(&(pdevInfo->smeTimer));

    wlanMlmeAuthConfirm(pdevInfo, &(authFrame->macHeader.srcAddr),
                authFrame->algNo, (MLME_RESULT) authFrame->statusCode);

    return MLME_SUCCESS;
}

/* wlanMlmeAuthConfirm - confirm the result of AuthRequest operation.
 *
 * Calls the authCallback routine to deliver the result of authentication
 * operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_NOT_SUPPORTED,
 *            MLME_TOOMANY_REQ, MLME_TIMEOUT, MLME_REFUSED
 */

MLME_RESULT
wlanMlmeAuthConfirm(WLAN_DEV_INFO *pdevInfo,
                    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other MAC */
                    A_UINT16      authType,             /* open system or shared key */
                    MLME_RESULT   resultCode)           /* Result of operation */
{
    if (smeInfo.authCallback) {
        (*smeInfo.authCallback)(pdevInfo, pPeerSTAAddress,
                                authType, resultCode);
    }
    return MLME_SUCCESS;
}

/* wlanMlmeAuthIndication - indicate the result of AuthRequest operation in AP
 *
 * Calls the authCallback routine to deliver the result of authentication
 * operation.
 */

MLME_RESULT
wlanMlmeAuthIndication(WLAN_DEV_INFO *pdevInfo,
                       WLAN_MACADDR  *pPeerSTAAddress,  /* address of other MAC */
                       A_UINT16      authType)          /* open system or shared key */
{
    if (smeInfo.authIndicationCallback ) {
        (*smeInfo.authIndicationCallback)(pdevInfo, pPeerSTAAddress,
                                          authType);
    }
    return MLME_SUCCESS;
}

/* mlmeAuthAp1Receive - Process the received Auth frame #1 on AP
 *
 * This routine handles the incoming associate requests on AP. It checks the
 * state, assigns a new association id and sends back the response frame.
 * This routine is not part of MLME API.
 */

A_UINT16
mlmeAuthAp1Receive(WLAN_DEV_INFO   *pdevInfo,   /* device context */
                   WLAN_FRAME_AUTH *authFrame,  /* Incoming frame pointer */
                   A_UINT16        len)         /* Total frame length */
{
    SIB_ENTRY    *pSib;                 /* STA trying to associate */
    A_UINT16     statusCode;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    union {
        WLAN_FRAME_AUTH_CHLG *shared;   /* return frame w/text */
        WLAN_FRAME_AUTH      *open;     /* return frame regular */
    } pFrame;
    A_BOOL faulted;

    /* Check if I am up */
    if (!(pdevInfo->localSta->staState & STATE_AP_UP)) {
        return MLME_REFUSED;
    }

    if ((pSib = sibEntryFind(&authFrame->macHeader.srcAddr)) == NULL) {
        /* The sib should have been created while the frame
         * was received.  We should be disapatched before
         * sib can be aged and recycled.
         */
        return MLME_REFUSED;
    }

    if (pSib->assocId != 0) {
        /* Release the association ID */
        mlmeAssocIdRelease(pSib->assocId);
        pSib->assocId = 0;
    }

#ifdef    BUILD_AP
    if (pdevInfo->localSta->transState & TSTATE_SCANNING) {
/****        sibEntryDel(pSib); Bad idea ****/
        /* Scanning in progress, delete SIB entry and reject this request */
        return MLME_TOOMANY_REQ;
    }
#endif    /* BUILD_AP */

    /* Verify state */
    statusCode = WLAN_SUCCESS;

    if (pSib->staState & STATE_AUTH) {
        /* STA is already authenticated.  This is an error.
         * This can happen if an authenticated station crashes, boots up and
         * wants to authenticate again. AP has no way of knowing that STA crashed.
         * It should treat this as starting a new auth sequence. For now, we will
         * mark its state as clean. i.e. drop associations too. Sending disassociate
         * frame will be meaningless. Need to think about other possibilities.
         */
        if (pSib->staState & STATE_ASSOC && pSib->assocId != 0) {
            sibPowerStateClean(pSib);
            // cancel this association
            mlmeAssocIdRelease(pSib->assocId);
            pSib->assocId = 0;
            sibCseStateClean(pSib);
        }
        pSib->staState  &= ~(STATE_AUTH|STATE_ASSOC);
        pSib->transState = TSTATE_AUTH_FRAME1_RCVD;
    }

    /* We don't need to check transState because even if it is non-zero like
     * TSTATE_FRAME2_SENT, receiving frame 1 will reset it. This can happen
     * if remote sta didn't get an ack for previous frame 1 that it sent out.
     */
    pSib->authenAlgo = authFrame->algNo;

#ifdef BUILD_AP
    /*
     * Check the MAC address based permission table to see if this station
     * is allowed to associate.  Done at this level so we can give feedback
     * to the station via an error response instead of dropping the frame
     * and thrashing on the sib table.
     */
    if (macACLCheck(pdevInfo, &pSib->macAddr, pSib) == FALSE) {
        statusCode = WLAN_OUT_OF_SCOPE;

        /* Build an error response frame */
        frameLen = sizeof(WLAN_FRAME_AUTH);

        if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
            /* Failed to allocate buffer for frame and descriptor */
            return MLME_DRIVER_ERROR;
        }
        pFrame.open = pDesc->pBufferVirtPtr.auth;
        mgtFrameHdrFill(pdevInfo, &pFrame.open->macHeader, &pSib->macAddr,
                        SUBT_AUTH);
    }
    else
#endif
    if ((authFrame->algNo == AUTH_SHARED_KEY) &&
        (pdevInfo->staConfig.authAllowShared == TRUE)) {
        /* If !privacyInvoked, we return WLAN_UNSUPPORTED_ALG below */
        if (pSib->pChlg == NULL) {
            pSib->pChlg = (CHLG_TEXT *)A_DRIVER_MALLOC(sizeof(CHLG_TEXT), 0, 0);
            if (pSib->pChlg == NULL) {
                return MLME_DRIVER_ERROR;
            }
        }
        chlgTextFill(pSib->pChlg, AUTH_SHARED_KEY_CHLG_SIZE);

        /* Compute frame length. Include challenge text */
        frameLen = sizeof(WLAN_FRAME_AUTH) + INFO_ELEMENT_SIZE(*pSib->pChlg);

        /* Build the response frame */
        if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
            /* Failed to allocate buffer for frame and descriptor */
            A_DRIVER_FREE(pSib->pChlg, 0, 0);
            pSib->pChlg = NULL;
            return MLME_DRIVER_ERROR;
        }
        pFrame.shared = pDesc->pBufferVirtPtr.authChal;
        mgtFrameHdrFill(pdevInfo, &pFrame.shared->macHeader,
                        &pSib->macAddr, SUBT_AUTH);
        /* Copy the challenge text into the frame */
        mlmeElementCopy((INFO_ELEMENT *)pSib->pChlg,
                        (INFO_ELEMENT *)&pFrame.shared->challenge);
    }
    else if ((authFrame->algNo == AUTH_OPEN_SYSTEM) &&
             (pdevInfo->staConfig.authAllowOpen == TRUE)) {
        /* Simple open system auth.  No challenge text.
         * We may need to query some local database or table to validate
         * that this MAC address is allowed to enter the BSS network.
         */
        frameLen = sizeof(WLAN_FRAME_AUTH);

        /* Build the response frame */
        if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
            /* Failed to allocate buffer for frame and descriptor */
            return MLME_DRIVER_ERROR;
        }
        pFrame.open = pDesc->pBufferVirtPtr.auth;
        mgtFrameHdrFill(pdevInfo, &pFrame.open->macHeader, &pSib->macAddr,
                        SUBT_AUTH);
    } else {
        /*
         * Either shared or open auth can be configured to
         * be unsupported.
         */
        statusCode = WLAN_UNSUPPORTED_ALG;

        if ((authFrame->algNo != AUTH_SHARED_KEY) &&
            (authFrame->algNo != AUTH_OPEN_SYSTEM)) {
            uiPrintf("Unknown authentication service %d\n",
                     authFrame->algNo);
        }

        /* Build an error response frame */
        frameLen = sizeof(WLAN_FRAME_AUTH);

        if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
            /* Failed to allocate buffer for frame and descriptor */
            return MLME_DRIVER_ERROR;
        }
        pFrame.open = pDesc->pBufferVirtPtr.auth;
        mgtFrameHdrFill(pdevInfo, &pFrame.open->macHeader, &pSib->macAddr,
                        SUBT_AUTH);
    }

    if (pdevInfo->staConfig.privacyInvoked && (statusCode == WLAN_SUCCESS)) {
        /*
         * Fault in any unique key for this station for all
         * authentication types if WEP is enabled.  Done in
         * tNetTask context to avoid race conditions.
         */
        wlanKeyCacheFault(pdevInfo, pSib, &faulted);
    }

    pFrame.open->algNo      = authFrame->algNo; /* return same algno */
    pFrame.open->statusCode = statusCode;
    pFrame.open->transSeqNo = 2;                /* frame 2 */

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("\nmlmeAuthAp1Receive algNo = %d\n", authFrame->algNo);
        printMacAddress(pSib->macAddr);
        uiPrintf("  ");
        displayStates (pSib->staState, pSib->transState);
        if (mlmeDebug > 1) {
            displayMgtFrameShort(&pFrame.open->macHeader, frameLen);
        }
    }
#endif  /* DEBUG */

    /* send response frame */
    pDesc->pDestSibEntry = pSib;
    pSib->transState     = TSTATE_AUTH_FRAME2_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        pSib->stats.Authentications++;
        if (pSib->smePendDesc == NULL) {
            pSib->smePendDesc = (ATHEROS_DESC *)0xffffffff;
        }
        return MLME_SUCCESS;
    }
    pSib->transState = TSTATE_QUIET;
    return MLME_DRIVER_ERROR;
}

A_UINT16
mlmeAuthAp3Receive(WLAN_DEV_INFO           *pdevInfo,   /* device context */
                   WLAN_FRAME_AUTH_ENCRYPT *authFrame,  /* Incoming frame */
                   A_UINT16                len,         /* Total frame length */
                   A_BOOL                  decryptError)
{
    SIB_ENTRY       *pSib;      /* STA trying to associate */
    A_UINT16        statusCode;
    A_UINT16        frameLen;
    ATHEROS_DESC    *pDesc;
    WLAN_FRAME_AUTH *pFrame;    /* Auth #4 frame */

    /* Check if I am up */
    if (!(pdevInfo->localSta->staState & STATE_AP_UP)) {
        return MLME_REFUSED;
    }

    if ((pSib = sibEntryFind(&authFrame->macHeader.srcAddr)) == NULL) {
        uiPrintf("mlmeAuthAp3Receive: ERROR: failed to find sta in table\n");
        return MLME_REFUSED;
    }

    /* STA is already authenticated is not an error.
     * This can happen if the STA didn't get the frame #4 that
     * AP sent out last time and it retransmitted frame #3.
     *
     * Pick through things that could be wrong to return the
     * right errno.
     */
    if ((pSib->transState != TSTATE_AUTH_FRAME2_SENT) &&
        (pSib->transState != TSTATE_AUTH_FRAME4_SENT)) {
        statusCode = WLAN_OUT_OF_SEQUENCE;
    } else if (decryptError) {
        /* check before looking at the frame as it may be corrupt */
        statusCode = WLAN_CHLG_FAILURE;
    } else if (authFrame->algNo != AUTH_SHARED_KEY) {
        statusCode = WLAN_UNSUPPORTED_ALG;
    } else if (pSib->pChlg == NULL) {
        statusCode = WLAN_OUT_OF_SCOPE;
    } else if (authFrame->transSeqNo != 3) {
        statusCode = WLAN_OUT_OF_SEQUENCE;
    } else if ((authFrame->challenge.elementID != ELE_CHLG_TEXT) ||
               (authFrame->challenge.length != pSib->pChlg->length) ||
               (memcmp(&authFrame->challenge.chlgText, pSib->pChlg->chlgText,
                       pSib->pChlg->length) != 0)) {
        statusCode = WLAN_CHLG_FAILURE;
    } else {
        statusCode = WLAN_SUCCESS;
    }
    if (statusCode != WLAN_SUCCESS) {
        uiPrintf("mlmeAuthAp3Receive: shared key auth failure %d for ",
                 statusCode);
        printMacAddress(pSib->macAddr);
        uiPrintf("\n");
    }
    if (pSib->pChlg != NULL) {
        A_DRIVER_FREE(pSib->pChlg, 0, 0);
        pSib->pChlg = NULL;
    }

    /* Build the response frame - simple frame with no challenge. */
    frameLen = sizeof(WLAN_FRAME_AUTH);
    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.auth;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader, &pSib->macAddr, SUBT_AUTH);
    pFrame->algNo      = authFrame->algNo;      /* return algno */
    pFrame->statusCode = statusCode;
    pFrame->transSeqNo = 4;                     /* frame 4 */

    /* send response frame */
#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("\nmlmeAuthAp3Receive\n");
        if (mlmeDebug > 1) {
            displayMgtFrame(&pFrame->macHeader, frameLen);
        }
    }
#endif  /* DEBUG */
    pDesc->pDestSibEntry = pSib;
    pSib->transState     = TSTATE_AUTH_FRAME4_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (pSib->smePendDesc == NULL) {
            pSib->smePendDesc = (ATHEROS_DESC *) 0xffffffff;
        }
        return MLME_SUCCESS;
    } else {
        pSib->transState = TSTATE_QUIET;
        return MLME_DRIVER_ERROR;
    }

    /* According to the spec, sta is not considered authenticated until
     * AP transmits frame #4 successfully with an ACK from sta. It is done
     * by mlmeMgtTxCompletion().
     */
}

/* chlgTextFill - Generate challenge text for WEP authentication
 * DESCRIPTION
 * Fills the Challenge Text element with random byte values.
 * Spec section 8.1.2.2 dictates that it must use WEP PRNG.
 * We'll fill in this algorithm once it becomes available from
 * the MAC team.
 */

void
chlgTextFill(CHLG_TEXT *pChlg, A_UINT8 len)
{
    int i;

    pChlg->elementID = ELE_CHLG_TEXT;
    pChlg->length    = len;

#ifdef VXWORKS
    /* stir seed with time of this and last challenge text */
    srand(tickGet() ^ rand());
#endif

    for (i =0; i < len && i < ELE_CHLG_TEXT_SIZE; i++) {
#ifdef VXWORKS
        pChlg->chlgText[i] = rand() & 0xff;
#else
        pChlg->chlgText[i] = ~i;    /* not really random! */
#endif
    }
}

/*
 * wlanMlmeDeauthRequest - terminates current authentication.
 *
 * It is a noticification. There is no notion of success/failure.
 * It is called by SME on either STA or AP.
 * Spec 5.4.3.2 says that it must deauth and disassociate if an
 * association exists. We need to worry about operations in progress.
 * I think that any auth or assoc in progress must be cancelled if
 * they are eventually completed. Not implemented yet.
 *
 * RETURNS: MLME_SUCCESS or MLME_INVALID_PARAM
 */

MLME_RESULT
wlanMlmeDeauthRequest(WLAN_DEV_INFO *pdevInfo,
                      WLAN_MACADDR  *pPeerSTAAddress,   /* address of other MAC */
                      A_UINT16       reasonCode)        /* valid REASON_CODEs */
{
    SIB_ENTRY         *pSib;
    A_UINT16          frameLen;
    ATHEROS_DESC      *pDesc;
    WLAN_FRAME_DEAUTH *pFrame;    /* Pointer to deauth req frame */
    SIB_ENTRY         *lSib = pdevInfo->localSta;

    /* This can be done on either AP or STA. Don't assume it's
     * the local STA.
     */
    if (lSib->transState & (TSTATE_JOINING|TSTATE_SCANNING)) {
        /* Scan or join in progress */
        return MLME_TOOMANY_REQ;
    }

    // It's ok to not find the destination SIB.
    pSib = sibEntryFind(pPeerSTAAddress);

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("wlanMlmeDeauthRequest -- ");
        if (pSib) {
            printMacAddress(pSib->macAddr);
            uiPrintf("\n");
        } else {
            uiPrintf("pSib == NULL\n");
        }
    }
#endif  /* DEBUG */

    if (lSib->serviceType == WLAN_STA_SERVICE) {
        if ((lSib->staState & STATE_ASSOC) ||
            (pSib && (pSib->staState & STATE_ASSOC)))
        {
            // The receiving STA is associated too. Cancel that.
            wlanMlmeDisassocRequest(pdevInfo, pPeerSTAAddress,
                                    REASON_AUTH_LEAVING);
        }

        /* Since this is a notification, we must reflect the state change
         * immediately so that data transmit/receive will terminate
         * immediately.
         */
        lSib->staState  &= ~(STATE_AUTH|STATE_ASSOC);
        lSib->transState = TSTATE_QUIET;
    }

    if (pSib != NULL) {
        pSib->staState  &= ~(STATE_AUTH|STATE_ASSOC);
        pSib->transState = TSTATE_QUIET;
        pSib->stats.DeAuthentications++;
    }
    pdevInfo->localSta->stats.DeAuthentications++;

    /* Compute frame length. */
    frameLen = sizeof(WLAN_FRAME_DEAUTH);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.deAuth;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    pPeerSTAAddress, SUBT_DEAUTH);

    pFrame->reasonCode = reasonCode;    /* reason for deauth */

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    pDesc->pDestSibEntry = pSib;

    powerFrameTransmit(pdevInfo, pDesc); // return status doesn't matter

    return MLME_SUCCESS;
}


/* wlanMlmeDeauthConfirm - confirm a deauthentication operation.
 *
 * It will be called on transmission of deauth frame.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_TOOMANY_REQ
 *
 */

MLME_RESULT
wlanMlmeDeauthConfirm(WLAN_DEV_INFO *pdevInfo,
                      WLAN_MACADDR  *pPeerSTAAddress,   /* address of other STA */
                      MLME_RESULT   resultCode)         /* Result of operation */
{
    /* If callback routine exists, call it */
    if (smeInfo.deauthCallback) {
        (*smeInfo.deauthCallback)(pdevInfo, pPeerSTAAddress, resultCode);
    }
    return MLME_SUCCESS;
}

MLME_RESULT
wlanMlmeDeauthIndication(WLAN_DEV_INFO *pdevInfo,
                         WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
                         MLME_RESULT   resultCode)           /* Result of operation */
{
    /* If callback routine exists, call it */
    if (smeInfo.deauthIndicationCallback) {
        (*smeInfo.deauthIndicationCallback)(pdevInfo, pPeerSTAAddress, resultCode);
    }
    return MLME_SUCCESS;
}

MLME_RESULT
mlmeDeauthReceive(WLAN_DEV_INFO          *pdevInfo,
                  WLAN_FRAME_DEAUTH      *deauthFrame,  /* Incoming frame pointer */
                  A_UINT16                len)          /* Total frame length */
{
    SIB_ENTRY    *pSib;
    SIB_ENTRY    *lSib = pdevInfo->localSta;

#ifndef BUILD_AP
    /* Ensure that it comes from desired AP; accept deauth from authenticated bssid only */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(deauthFrame->macHeader.srcAddr))) != 0) {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }
#endif

    if ((pSib = sibEntryFind(&deauthFrame->macHeader.srcAddr)) == NULL) {
        uiPrintf("mlmeDeauthReceive: Warning: ");
        uiPrintf("Deauth frame from unknown source\n");
        // It's ok to proceed
    }

    if (lSib->serviceType == WLAN_STA_SERVICE) {
        if (lSib->staState & STATE_ASSOC) {
            // The receiving STA is associated too. Cancel that.
            wlanMlmeDisassocRequest(pdevInfo, &deauthFrame->macHeader.srcAddr,
                                    REASON_AUTH_LEAVING);
        }
        lSib->staState &= ~(STATE_ASSOC|STATE_AUTH);
        lSib->transState = TSTATE_QUIET;
    }

    if (pSib) {
        if (pSib->staState & STATE_ASSOC && pSib->assocId != 0) {
            /* The STA is associated too. Must force disassociation */
#ifdef BUILD_AP
            swretryQueueFlush(pdevInfo, pSib);
#endif
            sibPowerStateClean(pSib);
            mlmeAssocIdRelease(pSib->assocId);  /* Currently assigned ID */
            pSib->assocId = 0;
            sibCseStateClean(pSib);
            wlanMlmeDisassocIndication(pdevInfo, &deauthFrame->macHeader.srcAddr,
                                       deauthFrame->reasonCode);
        }

        /* STA is not authenticated. Error. ignore. */
        pSib->staState  &= ~(STATE_ASSOC|STATE_AUTH);
        pSib->transState = TSTATE_QUIET;
        pSib->stats.DeAuthentications++;
        pSib->stats.DeAuthReasonCode = deauthFrame->reasonCode;
    }

    wlanMlmeDeauthIndication(pdevInfo, &deauthFrame->macHeader.srcAddr,
                             (MLME_RESULT)deauthFrame->reasonCode);

    return MLME_SUCCESS;
}

MLME_RESULT
mlmeDeauthStaReceive(WLAN_DEV_INFO     *pdevInfo,
                     WLAN_FRAME_DEAUTH *deauthFrame,    /* Incoming frame pointer */
                     A_UINT16          len)             /* Total frame length */
{
    SIB_ENTRY    *pSib;
    SIB_ENTRY    *lSib;

    lSib = pdevInfo->localSta;

    if ((pSib = sibEntryFind( &deauthFrame->macHeader.srcAddr)) == NULL) {
        return MLME_INVALID_PARAM;
    }

    /* Verify state */

    lSib->transState = TSTATE_QUIET;
    if (lSib->staState & STATE_ASSOC) {
        /* The STA is associated too. Must force disassociation */
        lSib->assocId   = 0;
        lSib->staState &= ~STATE_ASSOC;
        lSib->stats.DeAuthentications++;
        lSib->stats.DeAuthReasonCode = deauthFrame->reasonCode;
        /* Call hardware routine to stop processing TIM from BSS */
        wlanMlmeDisassocIndication(pdevInfo, &deauthFrame->macHeader.srcAddr,
                       deauthFrame->reasonCode);
    }

    // since AP sent us a forced deauthenticate frame, we might as well consider that
    // as AP doesn't want this station. Cancel the JOINED state too. It will force
    // the station to look for another AP.
    // Clearing the JOINED state is required for DFS. AP will send forced deauth upon
    // detecting a radar interference. Stations should move away from this AP. There
    // is no special management frame used for this purpose yet.
    pSib->staState &= ~(STATE_AUTH|STATE_JOINED);
    pSib->transState = TSTATE_QUIET;
    pSib->stats.DeAuthentications++;
    pSib->stats.DeAuthReasonCode = deauthFrame->reasonCode;

    wlanMlmeDeauthIndication(pdevInfo, &deauthFrame->macHeader.srcAddr,
                 (MLME_RESULT) deauthFrame->reasonCode);

    return MLME_SUCCESS;
}

/************************/
/* Association routines */
/************************/

/* wlanMlmeAssocRequest - Request to associate with a AP
 *
 * This routine initiates the association operation with a AP.
 * It will fail if not authenticated with the requested AP or already
 * associated with some AP.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *            MLME_NOT_SUPPORTED
 */

MLME_RESULT
wlanMlmeAssocRequest(WLAN_DEV_INFO *pdevInfo,
                     WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
                     A_UINT32      assocFailureTimeout,  /* timeout in TU */
                     CAP_INFO      *capabilityInfo,      /* describe capabilities */
                     A_UINT16      listenInterval)       /* sleep interval */
{
    WLAN_FRAME_ASSOC_REQ *pFrame;       /* Pointer to assoc req frame */
    SIB_ENTRY    *pSib;
    SIB_ENTRY    *lSib;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    INFO_ELEMENT *p1;                   /* Temporary pointer */

    /* validate parameters */
    if (!pPeerSTAAddress) {
        return MLME_INVALID_PARAM;
    }

    if ((pSib = sibEntryFind(pPeerSTAAddress)) == NULL) {
        return MLME_INVALID_PARAM;
    }

    lSib = pdevInfo->localSta;

    /* check current state */
    if (!(lSib->staState & STATE_AUTH) ||
        (lSib->transState == TSTATE_ASSOC_REQ_SENT)) {
        /* Not authenticated yet, or in the middle of assoc. Refuse assoc */
        return MLME_REFUSED;
    }

    /* check security policy */
    if (!pdevInfo->staConfig.privacyInvoked && pSib->capInfo.privacy) {
        /* clear station no talking to encryption only AP,  Refuse assoc */
        return MLME_REFUSED;
    }

    if (pdevInfo->staConfig.privacyInvoked &&
        (pdevInfo->staConfig.encryptionAlg == CIPHER_SELECTION_AES) &&
        !VALID_UCSE_ELEMENT(&pSib->ucseSet))
    {
        /* strong security no talking to legacy AP,  Refuse assoc */
        return MLME_REFUSED;
    }

    initCipher(pdevInfo);

    /* Need to compare STA rate set with bss basic rates
     * before even starting assoc process. Also, match
     * AcpInfo fields too.
     */

    /* prepare associate frame */

    /* Compute length of the desired frame */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) +
        sizeof(pFrame->capInfo) +
        sizeof(pFrame->listenInterval) +
        INFO_ELEMENT_SIZE(pdevInfo->staConfig.ssid) +
        INFO_ELEMENT_SIZE(pdevInfo->localSta->operationalRateSet) +
        (VALID_UCSE_ELEMENT(&pSib->ucseSet) ?
         INFO_ELEMENT_SIZE(pdevInfo->localSta->ucseSet) : 0);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.assoc;

    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    pPeerSTAAddress, SUBT_ASSOC_REQ);

    pFrame->capInfo = pdevInfo->staConfig.capInfo;    /* ANSI C required. Copying 16 bit structure */

    // use short preamble if both of us support it, otherwise use long.
    pFrame->capInfo.shortPreamble = pdevInfo->bssDescr->capabilityInfo.shortPreamble &&
                                    pdevInfo->staConfig.shortPreamble;

    pFrame->listenInterval = listenInterval;

    /* Copy ssid and bump the pointer to next element */
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ssid,
                         (INFO_ELEMENT *)&pFrame->ssid);

    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet,
                         (INFO_ELEMENT *)p1);

    if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
        p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->localSta->ucseSet,
                             (INFO_ELEMENT *) p1);
    }

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    /* set timeout */

    /* send frame */
    pDesc->pDestSibEntry = pSib;
    lSib->transState = TSTATE_ASSOC_REQ_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (lSib->smePendDesc == NULL) {
            lSib->smePendDesc = (ATHEROS_DESC *)0xffffffff;
        }
        A_TIMEOUT(&pdevInfo->smeTimer, assocFailureTimeout,
                  (A_UINT32)pdevInfo, MLME_ASSOC_TIMEOUT);
        return MLME_SUCCESS;
    }

    /* else */
    lSib->transState = TSTATE_QUIET;
    return MLME_DRIVER_ERROR;
}

/* mlmeAssocIdNew - Search and assign association ID
 *
 * Called after receiving associate request. It will search thru
 * the bit array and assign a new one. Search is always from to low
 * to high order bits so that we assign the lowest available ID.
 * This reduces the possible size of TIM element in beacon.
 * Need to use locking mechanism to protect assocArray.
 */

A_UINT16
mlmeAssocIdNew(void)
{
    A_UINT8        bitIndex;
    A_UINT8        byteIndex;
    A_UINT8        value;

#ifdef  BUILD_AP
    A_SEM_LOCK (aidSemId, WAIT_FOREVER);
#endif  /* BUILD_AP */

    /* Assign the next available Association ID */
    for (byteIndex = 0; byteIndex < (ASSOCID_MAX+7)>>3 ; byteIndex++) {
        if (assocArray[byteIndex] == 0xff) {
            continue;
        }
        value = assocArray[byteIndex];
        for (bitIndex = 0; bitIndex < 8; bitIndex ++) {
            if ((assocArray[byteIndex] & (0x1 << bitIndex)) == 0) {
                /* Found a place */
                assocArray[byteIndex] |= (0x1 << bitIndex);
                /* Bits 14 and 15 must be 1 according to spec */
#ifdef    BUILD_AP
                A_SEM_UNLOCK(aidSemId);
#endif    /* BUILD_AP */
                return (A_UINT16)((byteIndex*8+bitIndex) | 0xc000);
            }
        }
    }

#ifdef    BUILD_AP
    A_SEM_UNLOCK(aidSemId);
#endif    /* BUILD_AP */

    return 0;
}

/* mlmeAssocIdRelease - Make assocId available again
 *
 * Called by disassociate primitive. It will clear the corresponding
 * bit and make the assocId available for others.
 * Need to use locking mechanism to protect assocArray.
 */

void
mlmeAssocIdRelease(A_UINT16 assocId)        /* Currently assigned ID */
{
    A_INT8      bitIndex;
    int         byteIndex;      /* gcc warns on char array index */

    assocId &= 0x3fff;          /* Clear top 2 bits */
    if (assocId > ASSOCID_MAX || assocId == 0) {
#ifdef DEBUG
        uiPrintf("mlmeAssocIdRelease: ERROR: Illegal AssocId %04x\n",
                 assocId);
#endif
        return;
    }

#ifdef  BUILD_AP
    A_SEM_LOCK(aidSemId, WAIT_FOREVER);
#endif  /* BUILD_AP */

    byteIndex = assocId >> 3;        /* 8 bits per byte */
    bitIndex  = assocId & 0x7;    /* Lowest 3 bits, modulo 8 */
    assocArray[byteIndex] = assocArray[byteIndex] & (~(0x1 << (assocId & 0x7)));

#ifdef    BUILD_AP
    A_SEM_UNLOCK(aidSemId);
#endif    /* BUILD_AP */

    return;
}

/* mlmeAssocReceive - Process the received Assoc Request frame on AP.
 *
 * This routine handles the incoming associate requests on AP. It checks the
 * state, assigns a new association id and sends back the response frame.
 * This routine is not part of MLME API.
 */

A_UINT16
mlmeAssocApReceive(WLAN_DEV_INFO        *pdevInfo,
                   WLAN_FRAME_ASSOC_REQ *assocFrame,    /* Incoming frame pointer */
                   A_UINT16             len)            /* Total frame length */
{
    WLAN_FRAME_ASSOC_RESP *pFrame;      /* Pointer to assoc req frame */
    SIB_ENTRY    *pSib;                 /* STA trying to associate */
    SIB_ENTRY    *lSib;
    A_UINT16     statusCode;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    RATE_SET     *pRate;
    INFO_ELEMENT *p1 = NULL;            /* Temporary pointer */
#ifdef BUILD_AP
    VLAN_TAG     tagClr, tagAes, tagWep;
#endif

    /* Check if I am up */
    if (!(pdevInfo->localSta->staState & STATE_AP_UP)) {
        return MLME_REFUSED;
    }

    lSib = pdevInfo->localSta;

    // Validate SSID in the request frame
    if (A_BCOMP(&assocFrame->ssid, &pdevInfo->bssDescr->ssid, assocFrame->ssid.length) != 0) {
        // sorry, wrong ssid
        return MLME_INVALID_PARAM;
    }

    if ((pSib = sibEntryFind(&assocFrame->macHeader.srcAddr)) == NULL) {
        /* 802.11 section 5.5 b) says that AP should send a deauth frame to sender */
        wlanMlmeDeauthRequest(pdevInfo, &assocFrame->macHeader.srcAddr, REASON_CLASS2);
        uiPrintf("mlmeAssocApReceive -- can't find SIB\n");
        pdevInfo->localSta->stats.DeAuthentications++;
        return MLME_INVALID_PARAM;
    }

    /* Verify state */

    if (!(pSib->staState & STATE_AUTH)) {
        /* STA is not authenticated. Error. Should send a deauth frame */
        /* 802.11 section 5.5 b) says that AP should send a deauth frame to sender */
        wlanMlmeDeauthRequest(pdevInfo, &assocFrame->macHeader.srcAddr,
                              REASON_CLASS2);
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeAssocApReceive -- Not Authenticated\n");
        }
#endif  /* DEBUG */
        pdevInfo->localSta->stats.DeAuthentications++;
        return MLME_REFUSED;
    }

    if (pSib->staState & STATE_ASSOC) {
        /* STA is already associated. Error. Should send a disassoc frame */
        /* 802.11 section 5.5 b) says that AP should send a disassoc frame to sender */
        wlanMlmeDisassocRequest(pdevInfo, &assocFrame->macHeader.srcAddr,
                                REASON_UNSPECIFIED);
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeAssocApReceive -- Already Associated\n");
        }
#endif  /* DEBUG */
        pdevInfo->localSta->stats.DisAssociations++;
        return MLME_REFUSED;
    }

    if (pdevInfo->localSta->transState & TSTATE_SCANNING) {
#if 0
        // Don't reply anything. AP function is either not up or it can be on a wrong
        // channel.
        /* STA is not authenticated. Error. Should send a deauth frame */
        /* 802.11 section 5.5 b) says that AP should send a deauth frame to sender */
        wlanMlmeDeauthRequest(pdevInfo, &assocFrame->macHeader.srcAddr,
                              REASON_CLASS2);
        pdevInfo->localSta->stats.DeAuthentications++;
#endif
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeAssocApReceive -- Not Authenticated\n");
        }
#endif  /* DEBUG */
        return MLME_REFUSED;
    }

    /* set the default value for statusCode */
    statusCode = WLAN_SUCCESS;

    /* Need to compare various bits in assoc frame capInfo field */
    if (assocFrame->capInfo.cfPollable && !(assocFrame->capInfo.cfPollReq)) {
        /* Spec section 7.3.1.4, table 16. We want to avoid the
         * case where STA is CF pollable and requesting to be
         * placed on the CF polling list.
         */
        statusCode = WLAN_UNSUPPORTED_CAP;
    }

    // if station is asking for more than what AP can do, reject it
    if (assocFrame->capInfo.shortPreamble == 1 && lSib->capInfo.shortPreamble == 0) {
        /* STA wants short preamble, but AP can do only long */
        statusCode = WLAN_UNSUPPORTED_CAP;
    }

    /* STA rate set compared with BSS basic rate set */
    pRate = (RATE_SET *)mgtElementGet(&assocFrame->macHeader,
                                      len, ELE_SUPPORTED_RATES);
    if (basicRateSubsetCheck(&pdevInfo->bssDescr->bssRateSet,
                             pRate) == FALSE) {
        statusCode = WLAN_UNSUPPORTED_RATE;
    }

    /* Update the working rate set to be used for communicating
     * to this sta. It's an intersection of the two sets.
     */
    xSectionRateSet(&pdevInfo->bssDescr->bssRateSet,
                    pRate, &pSib->workingRateSet);

    RateCtrlSibTxRateInit(pSib);

    /* get STA's cipher suites
     * make the element invalid first
     * need remove this when the mlmeElementCopy changed
     */
    pSib->ucseSet.length = 0;
    mlmeElementCopy((INFO_ELEMENT *)(mgtElementGet(&assocFrame->macHeader,
                                                   len, ELE_UCSE_SET)),
                    (INFO_ELEMENT *)&pSib->ucseSet);

    if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
        /* STA ucse set compared with BSS ucse set */
        if (ucseSubsetCheck(&pdevInfo->bssDescr->bssUcseSet,
                            &pSib->ucseSet) == FALSE) {
            statusCode = WLAN_OUT_OF_SCOPE;
        } else {
            /* Update the working ucse set to be used for communicating
             * to this sta. It's an intersection of the two sets.
             */
            xSectionUcseSet(&pdevInfo->bssDescr->bssUcseSet,
                            &pSib->ucseSet,
                            &pSib->workingUcseSet);

            /* set MCSE */
            if (VALID_MCSE_ELEMENT(&pdevInfo->bssDescr->bssMcseSet)) {
                mlmeElementCopy((INFO_ELEMENT *) &pdevInfo->bssDescr->bssMcseSet,
                                (INFO_ELEMENT *) &pSib->mcseSet);
            }

            if (pdevInfo->localSta->privacyInvoked &&
                wlanKeyUpdate(pdevInfo, pSib, CONTEXT_NONE) != A_OK) {
                wlanMlmeDeauthRequest(pdevInfo, &pSib->macAddr, REASON_UNSPECIFIED);
#ifdef  DEBUG
                if (mlmeDebug) {
                    uiPrintf ("mlmeAssocApReceive -- Associate failed-> AES could not init\n");
                }
#endif
                pdevInfo->localSta->stats.DeAuthentications++;
                return MLME_REFUSED;
            }
        }
    }

    /* update the privacy, preamble bits at the same time */
    pSib->capInfo = assocFrame->capInfo;

#ifdef BUILD_AP
    tagClr = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_NONE);
    tagWep = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_WEP);
    tagAes = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_AES);

    /* Place in clear vlan until the first data packet is seen. */
    pSib->privacyInvoked = FALSE;
    pSib->vlanTag = tagClr;

    /*
    * Special workaround for legacy sta w/o csn, 
    * if this STAs privacy is known, place in that vlan.
    */
    if ((VLAN_TAG_INVALID == tagClr) && 
        (tagWep != tagClr) && 
        ((tagWep == tagAes) || (VLAN_TAG_INVALID == tagAes)))
    {
        #ifdef DEBUG
        extern int vlanDebug;
        if (vlanDebug & 0x100) {
            uiPrintf("\nFFV W %02x%02x%02x%02x%02x%02x, %d\n",
                pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                tagWep);
        }
        #endif
        pSib->privacyInvoked = TRUE;
        pSib->vlanTag = tagWep;
    }
    if ((VLAN_TAG_INVALID == tagClr) && 
        (tagAes != tagClr) && 
        (VLAN_TAG_INVALID == tagWep))
    {
        #ifdef DEBUG
        extern int vlanDebug;
        if (vlanDebug & 0x100) {
            uiPrintf("\nFFV A %02x%02x%02x%02x%02x%02x, %d\n",
                pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                tagAes);
        }
        #endif
        pSib->privacyInvoked = TRUE;
        pSib->vlanTag = tagAes;
    }

    /* set up broadcast/multicast cipher */
    if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
        if ((pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB) ||
            (pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB_REVA)) {
            if (tagAes == tagClr) {
                pSib->mcseSet.mCipher[0] = CSE_NULL;
            } else if (tagAes != tagWep) {
                pSib->mcseSet.mCipher[0] = pSib->workingUcseSet.uCiphers[0];
            }
        }

        if (pSib->workingUcseSet.uCiphers[0] == CSE_BASIC_WEP) {
            if (tagWep == tagClr) {
                pSib->mcseSet.mCipher[0] = CSE_NULL;
            }
        }
    }
#endif

    /* Compute length of the desired frame */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) +
        sizeof(CAP_INFO) +
        sizeof(A_UINT16) + /* Association ID */
        sizeof(pFrame->statusCode) +
        INFO_ELEMENT_SIZE(pdevInfo->bssDescr->bssRateSet) +
        (VALID_UCSE_ELEMENT(&pSib->workingUcseSet) ?
         INFO_ELEMENT_SIZE(pSib->workingUcseSet) : 0) +  /* ucse set element */
        (VALID_MCSE_ELEMENT(&pSib->mcseSet) ?
         INFO_ELEMENT_SIZE(pSib->mcseSet) : 0);

    /* Allocate assoc response frame */
    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.assocResp;

    /* prepare assoc response frame */
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
            &pSib->macAddr, SUBT_ASSOC_RESP);
    pFrame->capInfo = pSib->capInfo;    /* ANSI C required. Copying 16 bit structure */

    if (pSib->assocId != 0) {
        /* Release the association ID */
        mlmeAssocIdRelease(pSib->assocId);
        pSib->assocId = 0;
    }

    if (statusCode == WLAN_SUCCESS) {
        if ((pFrame->assocId = mlmeAssocIdNew()) == 0) {
            /* Failed to get a new assoc id. Table must be full */
            statusCode = WLAN_TOOMANY_ASSOC;
        }
    } else {
        pFrame->assocId = 0;
    }
    pSib->assocId        = pFrame->assocId;
    pSib->listenInterval = assocFrame->listenInterval;
    pSib->stats.Associations++;
    pFrame->statusCode   = statusCode;

    mlmeElementCopy((INFO_ELEMENT *) &pdevInfo->bssDescr->bssRateSet,
                    (INFO_ELEMENT *) &pFrame->rateSet);

    if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
        p1 = mlmeElementCopy((INFO_ELEMENT *) &pSib->workingUcseSet,
                             (INFO_ELEMENT *) &pFrame->ucseSet);
    }

    if (VALID_MCSE_ELEMENT(&pSib->mcseSet)) {
        mlmeElementCopy((INFO_ELEMENT *)&pSib->mcseSet,
                        (INFO_ELEMENT *) p1);
    }
#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("\nmlmeAssocApReceive response\n");
        if (mlmeDebug > 1) {
            displayMgtFrame(&pFrame->macHeader, frameLen);
        }
    }
#endif  /* DEBUG */

    /* send frame */
    pDesc->pDestSibEntry = pSib;
    pSib->transState = TSTATE_ASSOC_RESP_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (pSib->smePendDesc == NULL) {
            pSib->smePendDesc = (ATHEROS_DESC *)0xffffffff;
        }
        return MLME_SUCCESS;
    }
    pSib->transState = TSTATE_QUIET;
    return MLME_DRIVER_ERROR;
}

A_UINT16
mlmeAssocStaReceive(WLAN_DEV_INFO         *pdevInfo,
                    WLAN_FRAME_ASSOC_RESP *assocFrame, /* Incoming assoc frame */
                    A_UINT16              len)         /* Total frame length */
{
    SIB_ENTRY    *lSib;
    SIB_ENTRY    *pSib;

    lSib = pdevInfo->localSta;

    /* Ensure that it comes from desired AP (preauthenticate not supported yet) */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(assocFrame->macHeader.srcAddr))) != 0) {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }
    if ((pSib = sibEntryFind(&assocFrame->macHeader.srcAddr)) == NULL) {
        /* Program error, should never happen */
        return MLME_INVALID_PARAM;
    }

    if (!(lSib->staState & STATE_AUTH)) {
        // our authentication got cancelled while assoc request was
        // in progress, then this frame arrived. We should treat is as fail.
        // This can happen when last assoc attempt takes too long, just when the
        // cows start coming home, this frame rushes in.
        // change statusCode so that it will escape out of rest of the logic.
        assocFrame->statusCode = WLAN_UNKNOWN_FAILURE;
        // we should actually send disassoc request frame to AP and explicitly
        // cancel association, but it may mess up connection services logic.
        // For right now we will let AP figure it out automatically by out of state requests.
    }
    /* check the status and record the results */
    if (assocFrame->statusCode == WLAN_SUCCESS) {
        /* Copy rate set, cipher set */
        mlmeElementCopy((INFO_ELEMENT *) &assocFrame->rateSet,
                        (INFO_ELEMENT *) &lSib->workingRateSet);
        mlmeElementCopy((INFO_ELEMENT *) &assocFrame->rateSet,
                        (INFO_ELEMENT *) &pSib->workingRateSet);
        pdevInfo->localSta->stats.MaxTxRateKb =
                maxTxRateCompute(&lSib->workingRateSet, pdevInfo->staConfig.pChannel->channelFlags);

        //TODO: long/short/rate table, etc.
        // mlmeStaRateAssign();
        RateCtrlSibTxRateInit(lSib);
        /* make the element invalid first */
        /* If it's not in the frame, mlmeElementCopy will skip */
        lSib->workingUcseSet.length = 0;
        mlmeElementCopy((INFO_ELEMENT *) mgtElementGet(&assocFrame->macHeader,
                                                       len, ELE_UCSE_SET),
                        (INFO_ELEMENT *) &lSib->workingUcseSet);
        pSib->workingUcseSet.length = 0;
        mlmeElementCopy((INFO_ELEMENT *) mgtElementGet(&assocFrame->macHeader,
                                                       len, ELE_UCSE_SET),
                        (INFO_ELEMENT *) &pSib->workingUcseSet);
        /* make the element invalid first */
        /* need remove this when the mlmeElementCopy changed */
        lSib->mcseSet.length = 0;
        mlmeElementCopy((INFO_ELEMENT *)mgtElementGet(&assocFrame->macHeader,
                                                      len, ELE_MCSE_SET),
                        (INFO_ELEMENT *) &lSib->mcseSet);
        pSib->mcseSet.length = 0;
        mlmeElementCopy((INFO_ELEMENT *)mgtElementGet(&assocFrame->macHeader,
                                                      len, ELE_MCSE_SET),
                        (INFO_ELEMENT *) &pSib->mcseSet);

        /* set key type */
        if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet) &&
            pdevInfo->localSta->privacyInvoked) {
            if (wlanKeyUpdate(pdevInfo, pSib, CONTEXT_NONE) != A_OK) {
                wlanMlmeDeauthRequest(pdevInfo, &lSib->bssId, REASON_UNSPECIFIED);
#ifdef DEBUG
                if (mlmeDebug) {
                    uiPrintf("mlmeAssocStaReceive -- Associate failed-> AES could not init\n");
                }
#endif
                pdevInfo->localSta->stats.DeAuthentications++;
                return MLME_REFUSED;
            } else {
                /* set up encryption */
                setupSwEncryption(pdevInfo);
            }
        }

        // TODO: Verify that shortpreamble bit is valid in AP's response according to spec.
        lSib->capInfo.shortPreamble = assocFrame->capInfo.shortPreamble;
        lSib->assocId = assocFrame->assocId;
        /* Based on the Assoc ID, tell the hardware, which TIM bit to look for */
        halWriteAssocid(pdevInfo, &(lSib->bssId), lSib->assocId, pdevInfo->bssDescr->timOffset);
        /* Does capInfo represent BSS cap info? Should we keep a copy of it? */
        lSib->staState |= STATE_ASSOC;
        uiPrintf("mlmeAssocStaReceive: Successful Associate, assocId = %x\n", lSib->assocId);
#ifndef BUILD_AP
        osIndicateResourcesAvailable(pdevInfo);
#endif
    } else {
        uiPrintf("mlmeAssocStaReceive: ERROR: Associate failed, code = %x\n", assocFrame->statusCode);
        if (assocFrame->statusCode == REASON_NOT_AUTH) {
            wlanMlmeDeauthRequest(pdevInfo, &(lSib->bssId), REASON_UNSPECIFIED);
        }
    }

    lSib->transState = TSTATE_QUIET;

    A_UNTIMEOUT(&(pdevInfo->smeTimer));

    wlanMlmeAssocConfirm(pdevInfo, (MLME_RESULT) assocFrame->statusCode);

    return MLME_SUCCESS;
}


/* wlanMlmeAssocConfirm - Confirm the result of assoc. request
 *
 * this routine calls back assocCallback() routine to deliver the result
 * of assoc request.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *            MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */

MLME_RESULT
wlanMlmeAssocConfirm(WLAN_DEV_INFO *pdevInfo, MLME_RESULT resultCode)
{
#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("wlanMlmeAssocConfirm: Successful Associate\n");
    }
#endif  /* DEBUG */

    /* If callback routine exists, call it */
    if (smeInfo.assocCallback) {
        (*smeInfo.assocCallback)(pdevInfo, resultCode);
    }
    return MLME_SUCCESS;
}

/* wlanMlmeAssocIndication - Inform upper layer of associtaion
 *
 * Currently unimplemented
 */

MLME_RESULT
wlanMlmeAssocIndication(WLAN_DEV_INFO *pdevInfo,
                        WLAN_MACADDR  *pPeerSTAAddress) /* address of other STA */
{

    /* If callback routine exists, call it */
    if (smeInfo.assocIndicationCallback) {
        (*smeInfo.assocIndicationCallback)(pdevInfo, pPeerSTAAddress);
    }
    return MLME_SUCCESS;
}

MLME_RESULT
wlanMlmeDisassocRequest(WLAN_DEV_INFO *pdevInfo,
                        WLAN_MACADDR  *pPeerSTAAddress, /* address of other STA */
                        A_UINT16      reasonCode)       /* reason for disassoc */
{
    WLAN_FRAME_DISASSOC *pFrame;
    SIB_ENTRY    *pSib = NULL;
    SIB_ENTRY    *lSib = pdevInfo->localSta;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;

    /* This can be done on either AP or STA. Don't assume it's
     * the local STA.
     */

    if (A_MACADDR_COMP(&beaconBroadcastAddr, pPeerSTAAddress) == 0) {
        pSib = NULL;
    } else {
        // It's ok to not find the destination SIB.
        pSib = sibEntryFind(pPeerSTAAddress);
    }

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("wlanMlmeDisassocRequest -- ");
        printMacAddress(pSib->macAddr);
        uiPrintf("\n");
    }
#endif  /* DEBUG */

    /* Since this is a notification, we must reflect the state change
     * immediately so that data transmit/receive will terminate
     * immediately.
     */

    if (lSib->serviceType == WLAN_STA_SERVICE) {
        // never refuse disassociate request
        lSib->staState &= ~STATE_ASSOC;
        lSib->transState = TSTATE_QUIET;

        // STA driver may call this function even when it knows that it's
        // not associated. This can happen when STA reboots but AP doesn't
        // know. AP still thinks STA is associated. So, we send this frame
        // without checking if STA is associated or not.
        if (lSib->assocId != 0) {
            /* Physically disassociate */
            lSib->assocId = 0;
            /* associd = 0 will disable ps-poll */
            halWriteAssocid(pdevInfo, &(pdevInfo->localSta->bssId), 0, 0);
            sibCseStateClean(lSib);
        }
    }

    if (lSib->serviceType == WLAN_AP_SERVICE) {
        // AP software may call this even when it knows that STA is not associated
        // in its association table. This can happen if AP reboots or has already
        // disassoced the STA while STA still thinks that it is associated.
        // Another case is when the L2 update frame is received on the DSM.

        // If actual association exists on AP, release it.
        /* Release the association ID for pSib. */
        if (pSib) {
            /*
             *  Free up the swretry state machine
             */
            if (pSib->staState & STATE_ASSOC) {
                swretryQueueFlush(pdevInfo, pSib);
            }
            if (pSib->assocId != 0) {
                sibPowerStateClean(pSib);
                /* Currently assigned ID */
                mlmeAssocIdRelease(pSib->assocId);
                pSib->assocId = 0;
            }
            sibCseStateClean(pSib);
            pSib->staState &= ~STATE_ASSOC;
            pSib->stats.DisAssociations++;
            pSib->stats.DisAssocReasonCode = reasonCode;
            pdevInfo->localSta->stats.DisAssociations++;
        }

        /* We do not want to send this frame out as the sta is probably gone */
        if (reasonCode == REASON_INACTIVE) {
            return MLME_SUCCESS;
        }
    }

    /* Compute frame length. */
    frameLen = sizeof(WLAN_FRAME_DISASSOC);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.disAssoc;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader, pPeerSTAAddress, SUBT_DISASSOC);

    pFrame->reasonCode   = reasonCode;  /* reason for disassoc */
    pDesc->pDestSibEntry = pSib;

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    powerFrameTransmit(pdevInfo, pDesc);        /* return status doesn't matter */
    return MLME_SUCCESS;
}

/* wlanMlmeDisassocConfirm - Confirm the disassoc operation.
 *
 * This routine calls back disassocCallback() to confirm the operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *            MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */

MLME_RESULT
wlanMlmeDisassocConfirm(WLAN_DEV_INFO *pdevInfo,
                        WLAN_MACADDR  *pmacAddr,
                        MLME_RESULT   resultCode)       /* Result of operation */
{
    /* If callback routine exists, call it */
    if (smeInfo.disassocCallback) {
        (*smeInfo.disassocCallback)(pdevInfo, pmacAddr, resultCode);
    }
    return MLME_SUCCESS;
}

/* wlanMlmeDisassocIndication - Inform upper layer of disassociation
 *
 * Currently unimplemented
 */

MLME_RESULT
wlanMlmeDisassocIndication(WLAN_DEV_INFO *pdevInfo,
                           WLAN_MACADDR  *pPeerSTAAddress,
                           MLME_RESULT   resultCode)    /* Result of operation */
{
    if (smeInfo.disassocIndicationCallback) {
        (*smeInfo.disassocIndicationCallback)(pdevInfo, pPeerSTAAddress, resultCode);
    }
    return MLME_SUCCESS;
}

MLME_RESULT
mlmeDisassocReceive(WLAN_DEV_INFO *pdevInfo,
                    WLAN_FRAME_DISASSOC *disassocFrame, /* Incoming frame pointer */
                    A_UINT16            len)            /* Total frame length */
{
    SIB_ENTRY    *pSib;
    SIB_ENTRY    *lSib = pdevInfo->localSta;

#ifndef BUILD_AP
    /* Ensure that it comes from desired AP; accept disassoc from our bssid only */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(disassocFrame->macHeader.srcAddr))) != 0)
    {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }
#endif

    /* search sta table */

    if ((pSib = sibEntryFind(&disassocFrame->macHeader.srcAddr)) == NULL)
    {
        // pSib can be NULL if AP rebooted and doesn't know anything about
        // the sending STA
        ;
    }

    if (lSib->serviceType == WLAN_STA_SERVICE) {

        lSib->staState  &= ~STATE_ASSOC;
        lSib->transState = TSTATE_QUIET;

        // STA driver may call this function even when it knows that it's
        // not associated. This can happen when STA reboots but AP doesn't
        // know. AP still thinks STA is associated. So, we send this frame
        // without checking if STA is associated or not. It could also be
        // a broadcast disassoc
        if (lSib->assocId != 0) {
            /* Physically disassociate */
            lSib->assocId = 0;
            /* associd = 0 will disable ps-poll */
            halWriteAssocid(pdevInfo, &(pdevInfo->localSta->bssId), 0, 0);
        }
        sibCseStateClean(lSib);
    }

    if (lSib->serviceType == WLAN_AP_SERVICE && pSib != NULL &&
        pSib->assocId != 0) {
#ifdef BUILD_AP
        swretryQueueFlush(pdevInfo, pSib);
#endif
        sibPowerStateClean(pSib);
        /* Release the association ID for pSib. */
        mlmeAssocIdRelease(pSib->assocId);
        pSib->assocId = 0;
        sibCseStateClean(pSib);
    }

    if (pSib != NULL) {
        pSib->staState  &= ~STATE_ASSOC;
        pSib->transState = TSTATE_QUIET;
        pSib->stats.DisAssociations++;
        pSib->stats.DisAssocReasonCode = disassocFrame->reasonCode;
    }

    wlanMlmeDisassocIndication(pdevInfo, &disassocFrame->macHeader.srcAddr,
                               disassocFrame->reasonCode);

    return MLME_SUCCESS;
}

ATHEROS_DESC *
buildBroadcastDisconnect(WLAN_DEV_INFO *pdevInfo)
{
    A_UINT16            frameLen;
    ATHEROS_DESC        *pDisassocDesc;  /* pointer to atheros descriptor */
    ATHEROS_DESC        *pDeauthDesc;    /* pointer to atheros descriptor */
    WLAN_FRAME_DISASSOC *pDisassocFrame; /* Pointer to assoc req frame */
    WLAN_FRAME_DEAUTH   *pDeauthFrame;   /* Pointer to assoc req frame */

    /* first build broadcast disassoc frame */
    /* Compute frame length. */
    frameLen = sizeof(WLAN_FRAME_DISASSOC);

    if ( (createBuffandDesc(pdevInfo, frameLen, (void **) &pDisassocDesc)) != 0)
    {
        /* Failed to allocate buffer for frame and descriptor */
        return NULL;
    }
    pDisassocFrame = pDisassocDesc->pBufferVirtPtr.disAssoc;
    mgtFrameHdrFill(pdevInfo, &pDisassocFrame->macHeader,
            &beaconBroadcastAddr, SUBT_DISASSOC);
    uiPrintf("First fill....\n");
    printFrameControl(&pDisassocFrame->macHeader.frameControl);
    pDisassocFrame->reasonCode = REASON_ASSOC_LEAVING;  /* reason for deauth */
    pDisassocDesc->pDestSibEntry = NULL;
//  pHwTxControl->clearDestMask = 1;
    pDisassocDesc->hwControl.tx.transmitRate = pdevInfo->staConfig.pRateTable->info[LOWEST_RATE_INDEX].plcpRate;
    pDisassocDesc->hwControl.tx.frameLength += FCS_FIELD_SIZE;
    halSetTxDescPktType(pdevInfo, pDisassocDesc, HAL_DESC_PKT_TYPE_NORMAL);
    pDisassocFrame->macHeader.durationNav.duration = 0;
    pDisassocFrame->macHeader.durationNav.clearBit = 0;
    pDisassocDesc->hwControl.tx.RTSCTSEnable = 0;
    halSetTxDescLongPkt(pdevInfo, pDisassocDesc, 0);
    pDisassocDesc->hwControl.tx.more = 0;
    pDisassocDesc->hwStatus.word[1] = 0;
    displayMgtFrame(&pDisassocFrame->macHeader, frameLen);
    uiPrintf("Lengths: Frame %d Buffer %d\n",
        pDisassocDesc->hwControl.tx.frameLength,
        pDisassocDesc->hwControl.tx.bufferLength);
    FRAME_CONVERT_CPUTOLE(pDisassocDesc->pBufferVirtPtr.header, pDisassocDesc->hwControl.tx.frameLength);
    uiPrintf("this phys %08x next phys %08x\n", pDisassocDesc->thisPhysPtr, pDisassocDesc->nextPhysPtr);

    // then build build broadcast deauth frame
    /* Compute frame length. */
    frameLen = sizeof(WLAN_FRAME_DEAUTH);

    if ( (createBuffandDesc(pdevInfo, frameLen, (void **) &pDeauthDesc)) != 0)
    {
        /* Failed to allocate buffer for frame and descriptor */
        return NULL;
    }
    pDeauthFrame = pDeauthDesc->pBufferVirtPtr.deAuth;
    mgtFrameHdrFill(pdevInfo, &pDeauthFrame->macHeader,
            &beaconBroadcastAddr, SUBT_DEAUTH);
    uiPrintf("Second fill....\n");
    printFrameControl(&pDeauthFrame->macHeader.frameControl);
    pDeauthFrame->reasonCode = REASON_AUTH_LEAVING; /* reason for deauth */
    pDeauthDesc->pDestSibEntry = NULL;
//  pHwTxControl->clearDestMask = 1;
    pDeauthDesc->hwControl.tx.transmitRate = pdevInfo->staConfig.pRateTable->info[LOWEST_RATE_INDEX].plcpRate;
    pDeauthDesc->hwControl.tx.frameLength += FCS_FIELD_SIZE;
    halSetTxDescPktType(pdevInfo, pDeauthDesc, HAL_DESC_PKT_TYPE_NORMAL);
    pDeauthFrame->macHeader.durationNav.duration = 0;
    pDisassocFrame->macHeader.durationNav.clearBit = 0;
    pDeauthDesc->hwControl.tx.RTSCTSEnable = 0;
    halSetTxDescLongPkt(pdevInfo, pDeauthDesc, 0);
    pDeauthDesc->hwControl.tx.more = 0;
    pDeauthDesc->hwStatus.word[1] = 0;
    pDeauthDesc->hwControl.tx.interruptReq = 1;
    displayMgtFrame(&pDeauthFrame->macHeader, frameLen);
    uiPrintf("Lengths: Frame %d buffer %d\n",
             pDeauthDesc->hwControl.tx.frameLength,
             pDeauthDesc->hwControl.tx.bufferLength);
    FRAME_CONVERT_CPUTOLE(pDeauthDesc->pBufferVirtPtr.header, pDeauthDesc->hwControl.tx.frameLength);

    // Now link them together

    pDisassocDesc->nextPhysPtr = pDeauthDesc->thisPhysPtr;
    pDisassocDesc->pNextVirtPtr = pDeauthDesc;
    pDeauthDesc->nextPhysPtr = 0;
    return pDisassocDesc;
}

/* wlanMlmeReassocRequest - Request to reassociate with another AP
 *
 * This routine initiates the reassociation operation with a AP.
 * It will fail if not authenticated with the requested AP or already
 * not associated with some AP.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *            MLME_NOT_SUPPORTED
 */

MLME_RESULT
wlanMlmeReassocRequest(WLAN_DEV_INFO *pdevInfo,
                       WLAN_MACADDR  *pPeerSTAAddress,  /* address of other STA */
                       A_UINT32      assocFailureTimeout,/* timeout in TU */
                       CAP_INFO      *capabilityInfo,   /* describe capabilities */
                       A_UINT16      listenInterval)    /* sleep interval */
{
    WLAN_FRAME_REASSOC_REQ *pFrame;
    SIB_ENTRY    *pSib;
    SIB_ENTRY    *lSib;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    INFO_ELEMENT *p1;           /* Temporary pointer */

    /* validate parameters */
    if (!pPeerSTAAddress) {
        return MLME_INVALID_PARAM;
    }

    if ((pSib = sibEntryFind(pPeerSTAAddress)) == NULL) {
        return MLME_INVALID_PARAM;
    }

    lSib = pdevInfo->localSta;

    /* check current state */
    if ((lSib->staState & (STATE_AUTH|STATE_ASSOC)) != (STATE_AUTH|STATE_ASSOC) ||
        (lSib->transState != TSTATE_QUIET)) {
        /* Not authenticated or associated with an existing AP yet. Refuse assoc */
        return MLME_REFUSED;
    }

    /* check security policy */
    if (!pdevInfo->staConfig.privacyInvoked && pSib->capInfo.privacy) {
        /* clear station no talking to encryption only AP,  Refuse reassoc */
        return MLME_REFUSED;
    }

    if (pdevInfo->staConfig.privacyInvoked &&
        (pdevInfo->staConfig.encryptionAlg == CIPHER_SELECTION_AES) &&
        !VALID_UCSE_ELEMENT(&pSib->ucseSet))
    {
        /* strong security no talking to legacy AP,  Refuse assoc */
        return MLME_REFUSED;
    }

    initCipher(pdevInfo);

    /* Need to compare STA rate set with bss basic rates
     * before even starting assoc process. Also, match
     * AcpInfo fields too.
     */

    /* prepare reassociate frame */

    /* Compute length of the desired frame */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) +
        sizeof(pFrame->capInfo) +
        sizeof(pFrame->listenInterval) +
        sizeof(WLAN_MACADDR) +
        INFO_ELEMENT_SIZE(pdevInfo->staConfig.ssid) +
        INFO_ELEMENT_SIZE(pdevInfo->localSta->operationalRateSet) +
        (VALID_UCSE_ELEMENT(&pSib->ucseSet) ?
         INFO_ELEMENT_SIZE(pdevInfo->localSta->ucseSet) : 0);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.reAssoc;

    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    pPeerSTAAddress, SUBT_REASSOC_REQ);

    pFrame->capInfo = pdevInfo->staConfig.capInfo;    /* ANSI C required. Copying 16 bit structure */

    pFrame->listenInterval = listenInterval;

    A_MACADDR_COPY(&pdevInfo->localSta->bssId, &pFrame->curAPAddr);

    /* Copy ssid and bump the pointer to next element */
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ssid,
                         (INFO_ELEMENT *)&pFrame->ssid);

    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet,
                         (INFO_ELEMENT *)p1);

    if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
        p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->localSta->ucseSet,
                             (INFO_ELEMENT *)p1);
    }

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    /* set timeout */

    /* send frame */
    pDesc->pDestSibEntry = pSib;
    lSib->transState = TSTATE_ASSOC_REQ_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (lSib->smePendDesc == NULL) {
            lSib->smePendDesc = (ATHEROS_DESC *) 0xffffffff;
        }
        A_TIMEOUT(&pdevInfo->smeTimer, assocFailureTimeout, (A_UINT32) pdevInfo, MLME_ASSOC_TIMEOUT);
        return MLME_SUCCESS;
    } else {
        lSib->transState = TSTATE_QUIET;
        return MLME_DRIVER_ERROR;
    }
}


/* mlmeReassocReceive - Process the received Reassoc Request frame on AP.
 *
 * This routine handles the incoming reassociate requests on AP. It checks the
 * state, assigns a new association id and sends back the response frame.
 * This routine is not part of MLME API.
 */

A_UINT16
mlmeReassocApReceive(WLAN_DEV_INFO          *pdevInfo,  /* device context */
                     WLAN_FRAME_REASSOC_REQ *reassocFrame,/* Incoming frame pointer */
                     A_UINT16               len)        /* Total frame length */
{
    WLAN_FRAME_REASSOC_RESP *pFrame;
    SIB_ENTRY    *pSib;         /* STA trying to associate */
    SIB_ENTRY    *lSib;
    A_UINT16     statusCode;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    RATE_SET     *pRate;
    INFO_ELEMENT *p1 = NULL;    /* Temporary pointer */
#ifdef BUILD_AP
    VLAN_TAG     tagClr, tagAes, tagWep;
#endif

    /* Check if I am up */
    if (!(pdevInfo->localSta->staState & STATE_AP_UP)) {
        return MLME_REFUSED;
    }

    // Validate SSID in the request frame
    if (A_BCOMP(&reassocFrame->ssid, &pdevInfo->bssDescr->ssid, reassocFrame->ssid.length) != 0) {
        // sorry, wrong ssid
        return MLME_INVALID_PARAM;
    }

    lSib = pdevInfo->localSta;

    if ((pSib = sibEntryFind(&reassocFrame->macHeader.srcAddr)) == NULL) {
        /* 802.11 section 5.5 b) says that AP should send a deauth frame to sender */
        wlanMlmeDeauthRequest(pdevInfo, &reassocFrame->macHeader.srcAddr, REASON_CLASS2);
        uiPrintf("mlmeReassocApReceive -- can't find SIB\n");
        pdevInfo->localSta->stats.DeAuthentications++;
        return MLME_INVALID_PARAM;
    }

    /* Verify state */

    if (!(pSib->staState & STATE_AUTH)) {
        /* STA is not authenticated. Error. Should send a deauth frame */
        /* 802.11 section 5.5 b) says that AP should send a deauth frame to sender */
        wlanMlmeDeauthRequest(pdevInfo, &reassocFrame->macHeader.srcAddr, REASON_CLASS2);
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeReassocApReceive -- Not Authenticated\n");
        }
#endif  /* DEBUG */
        pdevInfo->localSta->stats.DeAuthentications++;
        return MLME_REFUSED;
    }

    if (pSib->staState & STATE_ASSOC) {
        /* STA is already associated. Error. Should send a disassoc frame */
        /* 802.11 section 5.5 b) says that AP should send a disassoc frame to sender */
        wlanMlmeDisassocRequest(pdevInfo, &reassocFrame->macHeader.srcAddr, REASON_UNSPECIFIED);
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeReassocApReceive -- Already Associated\n");
        }
#endif  /* DEBUG */
        return MLME_REFUSED;
    }

    /* set the default value for statusCode */
    statusCode = WLAN_SUCCESS;

    /* compare assoc attributes */
    /* Need to compare various bits in capInfo field */
    if (reassocFrame->capInfo.cfPollable &&
        !(reassocFrame->capInfo.cfPollReq))
    {
        /* Spec section 7.3.1.4, table 16. We want to avoid the
         * case where STA is CF pollable and requesting to be
         * placed on the CF polling list.
         */
        statusCode = WLAN_UNSUPPORTED_CAP;
    }
#ifdef STRICT_PRIVACY
    else if (reassocFrame->capInfo.privacy != lSib->capInfo.privacy) {
        /* Privacy attributes don't match */
        statusCode = WLAN_UNSUPPORTED_CAP;
    }
#endif

    /* STA rate set compared with BSS basic rate set */
    pRate = (RATE_SET *)mgtElementGet(&reassocFrame->macHeader,
                                      len, ELE_SUPPORTED_RATES);
    if (basicRateSubsetCheck(&pdevInfo->bssDescr->bssRateSet,
                             pRate) == FALSE) {
        statusCode = WLAN_UNSUPPORTED_RATE;
    }
    /* Update the working rate set to be used for communicating
     * to this sta. It's an intersection of the two sets.
     */

    xSectionRateSet(&pdevInfo->bssDescr->bssRateSet,
                    pRate, &pSib->workingRateSet);

    RateCtrlSibTxRateInit(pSib);

    /* STA ucse set compared with BSS ucse set */
    /* make the element invalid first */
    /* need remove this when the mlmeElementCopy changed */
    pSib->ucseSet.length = 0;
    mlmeElementCopy((INFO_ELEMENT *)(mgtElementGet(&reassocFrame->macHeader,
                                                   len, ELE_UCSE_SET)),
                    (INFO_ELEMENT *)&pSib->ucseSet);

    if (VALID_UCSE_ELEMENT(&pSib->ucseSet)) {
        if (ucseSubsetCheck(&pdevInfo->bssDescr->bssUcseSet,
                            &pSib->ucseSet) == FALSE) {
            statusCode = WLAN_OUT_OF_SCOPE;
        } else {
            /* Update the working rate set to be used for communicating
             * to this sta. It's an intersection of the two sets.
             */

            xSectionUcseSet(&pdevInfo->bssDescr->bssUcseSet, &pSib->ucseSet,
                            &pSib->workingUcseSet);

            /* set MCSE */
            if (VALID_MCSE_ELEMENT(&pdevInfo->bssDescr->bssMcseSet)) {
                mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssMcseSet,
                                (INFO_ELEMENT *)&pSib->mcseSet);
            }

            if (pdevInfo->localSta->privacyInvoked &&
                (wlanKeyUpdate(pdevInfo, lSib, CONTEXT_NONE) != A_OK)) {
                wlanMlmeDeauthRequest(pdevInfo, &pSib->macAddr, REASON_UNSPECIFIED);
#ifdef  DEBUG
                if (mlmeDebug) {
                    uiPrintf ("mlmeReassocApReceive -- ReAssociate failed-> AES could not init\n");
                }
#endif
                pdevInfo->localSta->stats.DeAuthentications++;
                return MLME_REFUSED;
            }
        }
    }

    /* update privacy bits at the same time*/
    pSib->capInfo = reassocFrame->capInfo;

#ifdef BUILD_AP
    tagClr = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_NONE);
    tagWep = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_WEP);
    tagAes = getPrivacyVlan(pdevInfo, VLAN_PRIVACY_AES);

    /* Place in clear vlan until the first data packet is seen. */
    pSib->privacyInvoked = FALSE;
    pSib->vlanTag = tagClr;

    /*
    * Special workaround for legacy sta w/o csn, 
    * if this STAs privacy is known, place in that vlan.
    */
    if ((VLAN_TAG_INVALID == tagClr) && 
        (tagWep != tagClr) && 
        ((tagWep == tagAes) || (VLAN_TAG_INVALID == tagAes)))
    {
        #ifdef DEBUG
        extern int vlanDebug;
        if (vlanDebug & 0x100) {
            uiPrintf("\nFFV Wr %02x%02x%02x%02x%02x%02x, %d\n",
                pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                tagWep);
        }
        #endif
        pSib->privacyInvoked = TRUE;
        pSib->vlanTag = tagWep;
    }
    if ((VLAN_TAG_INVALID == tagClr) && 
        (tagAes != tagClr) && 
        (VLAN_TAG_INVALID == tagWep))
    {
        #ifdef DEBUG
        extern int vlanDebug;
        if (vlanDebug & 0x100) {
            uiPrintf("\nFFV Ar %02x%02x%02x%02x%02x%02x, %d\n",
                pSib->macAddr.octets[0], pSib->macAddr.octets[1],
                pSib->macAddr.octets[2], pSib->macAddr.octets[3],
                pSib->macAddr.octets[4], pSib->macAddr.octets[5],
                tagAes);
        }
        #endif
        pSib->privacyInvoked = TRUE;
        pSib->vlanTag = tagAes;
    }

    /* set up broadcast/multicast cipher */
    if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
        if ((pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB) ||
            (pSib->workingUcseSet.uCiphers[0] == CSE_AES_OCB_REVA)) {
            if (tagAes == tagClr) {
                pSib->mcseSet.mCipher[0] = CSE_NULL;
            } else if (tagAes != tagWep) {
                pSib->mcseSet.mCipher[0] = pSib->workingUcseSet.uCiphers[0];
            }
        }

        if (pSib->workingUcseSet.uCiphers[0] == CSE_BASIC_WEP) {
            if (tagWep == tagClr) {
                pSib->mcseSet.mCipher[0] = CSE_NULL;
            }
        }
    }
#endif

    /* Compute length of the desired frame */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) +
        sizeof(CAP_INFO) +
        sizeof(A_UINT16) + /* Association ID */
        sizeof(pFrame->statusCode) +
        INFO_ELEMENT_SIZE(pdevInfo->bssDescr->bssRateSet) +
        (VALID_UCSE_ELEMENT(&pSib->workingUcseSet) ?
         INFO_ELEMENT_SIZE(pSib->workingUcseSet) : 0) +
        (VALID_MCSE_ELEMENT(&pSib->mcseSet) ?
         INFO_ELEMENT_SIZE(pSib->mcseSet) : 0);
    /* Allocate reassoc response frame */
    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = (WLAN_FRAME_REASSOC_RESP *) pDesc->pBufferVirtPtr.reAssocResp;

    /* prepare assoc response frame */
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    &pSib->macAddr, SUBT_ASSOC_RESP);

    pFrame->capInfo = lSib->capInfo;    /* ANSI C required. Copying 16 bit structure */

    if (pSib->assocId != 0) {
        /* Release the association ID */
        mlmeAssocIdRelease(pSib->assocId);
        pSib->assocId = 0;
    }

    if (statusCode == WLAN_SUCCESS) {
        if ((pFrame->assocId = mlmeAssocIdNew()) == 0) {
            /* Failed to get a new assoc id. Table must be full */
            statusCode = WLAN_TOOMANY_ASSOC;
        }
    } else {
        pFrame->assocId = 0;
    }
    pSib->assocId = pFrame->assocId;
    pSib->stats.Associations++;
    pFrame->statusCode = statusCode;

    mlmeElementCopy((INFO_ELEMENT *) &pdevInfo->bssDescr->bssRateSet,
                    (INFO_ELEMENT *) &pFrame->rateSet);

    if (VALID_UCSE_ELEMENT(&pSib->workingUcseSet)) {
        p1 = mlmeElementCopy((INFO_ELEMENT *) &pSib->workingUcseSet,
                             (INFO_ELEMENT *) &pFrame->ucseSet);
    }
    if (VALID_MCSE_ELEMENT(&pSib->mcseSet)) {
        mlmeElementCopy((INFO_ELEMENT *) &pSib->mcseSet,
                        (INFO_ELEMENT *) p1);
    }

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("\nmlmeReassocApReceive response\n");
        if (mlmeDebug > 1) {
            displayMgtFrame(&pFrame->macHeader, frameLen);
        }
    }
#endif  /* DEBUG */

    /* send frame */
    pDesc->pDestSibEntry = pSib;
    pSib->transState = TSTATE_ASSOC_RESP_SENT;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        if (pSib->smePendDesc == NULL) {
            pSib->smePendDesc = (ATHEROS_DESC *) 0xffffffff;
        }
        return MLME_SUCCESS;
    } else {
        pSib->transState = TSTATE_QUIET;
        return MLME_DRIVER_ERROR;
    }

    /* return */
}

A_UINT16
mlmeReassocStaReceive(WLAN_DEV_INFO           *pdevInfo,
                      WLAN_FRAME_REASSOC_RESP *reassocFrame, /* Incoming frame pointer */
                      A_UINT16                len)           /* Total frame length */
{
    SIB_ENTRY    *lSib;

    lSib = pdevInfo->localSta;

    /* Ensure that it comes from desired AP (preauthenticate not supported yet) */
    if ((A_MACADDR_COMP(&(lSib->bssId), &(reassocFrame->macHeader.srcAddr))) != 0) {
        /* Got this frame from wrong AP !! */
        return MLME_REFUSED;
    }

    /* check the status and record the results */
    if (reassocFrame->statusCode == WLAN_SUCCESS) {
        /* Copy rate set, cipher set */
        mlmeElementCopy((INFO_ELEMENT *) &reassocFrame->rateSet,
                        (INFO_ELEMENT *) &lSib->workingRateSet);

        /* make the element invalid first */
        /* need remove this when the mlmeElementCopy changed */
        lSib->workingUcseSet.length = 0;
        mlmeElementCopy(mgtElementGet(&reassocFrame->macHeader,
                                      len, ELE_UCSE_SET),
                        (INFO_ELEMENT *) &lSib->workingUcseSet);
        /* make the element invalid first */
        /* need remove this when the mlmeElementCopy changed */
        lSib->mcseSet.length = 0;
        mlmeElementCopy(mgtElementGet(&reassocFrame->macHeader,
                                      len, ELE_MCSE_SET),
                        (INFO_ELEMENT *) &lSib->mcseSet);

        /* set key type */
        if (VALID_UCSE_ELEMENT(&lSib->workingUcseSet) &&
            pdevInfo->localSta->privacyInvoked) {
            if (wlanKeyUpdate(pdevInfo, lSib, CONTEXT_NONE) != A_OK) {
                wlanMlmeDeauthRequest(pdevInfo, &lSib->bssId, REASON_UNSPECIFIED);
#ifdef  DEBUG
                if (mlmeDebug) {
                    uiPrintf("mlmeReAssocStaReceive -- ReAssociate failed-> AES could not init\n");
                }
#endif
                pdevInfo->localSta->stats.DeAuthentications++;
                return MLME_REFUSED;
            } else {
                /* set up encryption */
                setupSwEncryption(pdevInfo);
            }
        }

        lSib->assocId = reassocFrame->assocId;
        /* Does capInfo represent BSS cap info? Should we keep a copy of it? */
        lSib->staState |= STATE_ASSOC;
#ifndef BUILD_AP
        osIndicateResourcesAvailable(pdevInfo);
#endif
        uiPrintf("mlmeReassocStaReceive: Successful Reassociate, assocId = %x\n",
                 lSib->assocId);
    } else {
        uiPrintf("mlmeReassocStaReceive: ERROR: Reassociate failed, code = %x\n",
                 reassocFrame->statusCode);
    }

    /* Based on the Assoc ID, tell the hardware, which TIM bit to look for.
     * Assumption here is that bssDescr already represents new BSS or its
     * timOffset is same as the previous AP.
     */

    halWriteAssocid(pdevInfo, &(lSib->bssId), lSib->assocId, pdevInfo->bssDescr->timOffset);

    lSib->transState = TSTATE_QUIET;

    A_UNTIMEOUT(&(pdevInfo->smeTimer));

    wlanMlmeReassocConfirm(pdevInfo, (MLME_RESULT) reassocFrame->statusCode);

    return MLME_SUCCESS;
}


/* wlanMlmeReassocConfirm - Confirm the result of reassoc. request
 *
 * this routine calls back reassocCallback() routine to deliver the result
 * of assoc request.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *            MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */

MLME_RESULT
wlanMlmeReassocConfirm(WLAN_DEV_INFO *pdevInfo, MLME_RESULT resultCode)
{
#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("wlanMlmeReassocConfirm: Successful Reassociate\n");
    }
#endif  /* DEBUG */

    /* If callback routine exists, call it */
    if (smeInfo.reassocCallback) {
        (*smeInfo.reassocCallback)(pdevInfo, resultCode);
    }
    return MLME_SUCCESS;
}

/* wlanMlmeReassocIndication - Inform upper layer of associtaion
 *
 * Currently unimplemented
 */

MLME_RESULT
wlanMlmeReassocIndication(WLAN_DEV_INFO *pdevInfo,
                          WLAN_MACADDR  *pPeerSTAAddress)   /* address of other STA */
{

    /* If callback routine exists, call it */
    if (smeInfo.reassocIndicationCallback) {
        (*smeInfo.reassocIndicationCallback)(pdevInfo, pPeerSTAAddress);
    }
    return MLME_SUCCESS;
}


/* basicRateSubsetCheck - Check if bss basic rates are part of sta rates
 *
 * DESCRIPTION
 * Spec section 7.3.2.2
 * STA can be associated only if the all bss basic rates are
 * supported by the sta. The basic rates in the RATE_SET element
 * are represented by msb being 1. This routine takes basic rates
 * (with msb = 1) in psub and compares its lower 7 bits with lower 7
 * bits of each rate value in the rate element psuper.
 */
A_BOOL
basicRateSubsetCheck(RATE_SET *psub,    /* expected subset */
                     RATE_SET *psuper)  /* set expected to be superset */
{
    int    i;
    int    j;
    A_BOOL    match;

    if (psub == NULL || psuper == NULL) {
        uiPrintf("basicRateSubsetCheck: Error: NULL basic rate element\n");
        return FALSE;
    }

    for (i = 0; i < psub->length; i++) {
        if ((psub->rates[i] & 0x80) == 0) {
            /* MSB is not 1. It is not a basic rate, skip it */
            continue;
        }
        match = FALSE;
        for (j = 0; j < psuper->length; j++) {
            if ((psub->rates[i] & 0x7f) == (psuper->rates[j] & 0x7f)) {
                /* lower 7 bits match. */
                match = TRUE;
                break;
            }
        }
        if (!match) {
            return match;
        }
    }
    /* It didn't fail any match. All basic rates in psub matched */
    return TRUE;
}

/* xSectionRateSet - Builds an intersection of two rate sets
 * into the third one. Used by AP and STA for building a working
 * rate set.
 */
void
xSectionRateSet(RATE_SET *pSrc1,        /* set 1 */
                RATE_SET *pSrc2,        /* set 2 */
                RATE_SET *pDest)        /* cross section set */
{
    int     i;
    int     j;
    A_UINT8 k;

    k = 0;
    for (i = 0; i < pSrc1->length; i++) {
        for (j = 0; j < pSrc2->length; j++) {
            if ((pSrc1->rates[i] & 0x7f) == (pSrc2->rates[j] & 0x7f)) {
                /* lower 7 bits match. Add it to dest */
                pDest->rates[k++] = pSrc1->rates[i] & 0x7f;
                break;
            }
        }
    }
    pDest->length = k;
    pDest->elementID = ELE_SUPPORTED_RATES;
}

/* ucseSubsetCheck - Check if sta ucse set are part of bss ucse set
 *
 * DESCRIPTION
 * Spec section 7.3.2.2
 * STA can be associated only if the one of the bss ucse is
 * supported by the sta.
 */
A_BOOL
ucseSubsetCheck(UCSE_SET *psub,         /* expected subset */
                UCSE_SET *psuper)       /* set expected to be superset */
{
    int     i;
    int     j;
    A_BOOL  match;

    if (psub == NULL || psuper == NULL) {
        uiPrintf("ucseSubsetCheck: Error: NULL ucse element\n");
        return FALSE;
    }

    match = FALSE;

    for (i = 0; i < psub->length/4; i++) {
        for (j = 0; j < psuper->length/4; j++) {
            if (psub->uCiphers[i] == psuper->uCiphers[j]) {
                /* ucse match. */
                match = TRUE;
                break;
            }
        }

    }

    return match;
}

/* xSectionUcseSet - Builds an intersection of two ucse sets
 * into the third one. Used by AP and STA for building a working
 * ucse set.
 */
void
xSectionUcseSet(UCSE_SET *pSrc1,        /* set 1 */
                UCSE_SET *pSrc2,        /* set 2 */
                UCSE_SET *pDest)        /* cross section set */
{
    int     i;
    int     j;
    A_UINT8 k;

    k = 0;
    for (i = 0; i < pSrc1->length/4; i++) {
        for (j = 0; j < pSrc2->length/4; j++) {
            if (pSrc1->uCiphers[i] == pSrc2->uCiphers[j]) {
                /* match. Add it to dest */
                pDest->uCiphers[k++] = pSrc1->uCiphers[i];
                break;
            }
        }
    }
    pDest->length    = k*4;
    pDest->elementID = ELE_UCSE_SET;
}

A_BOOL
mlmeBeaconTimCheck(WLAN_FRAME_BEACON *bFrame,
                   A_UINT16          frameLen,
                   A_UINT16          assocId,
                   A_BOOL            *pbrmc)
{
    TIM_ELEMENT *pele;
    A_UINT8     mask;
    A_UINT8     index;

    assocId &= 0x3fff; // Clear top 2 bits
    if (assocId < ASSOCID_MIN || assocId > ASSOCID_MAX) {
        return FALSE;
    }

    if (pbrmc != NULL) {
        *pbrmc = FALSE;
    }

    pele = (TIM_ELEMENT *) &bFrame->buffer;
    /* Step thru all elements until we find TIM */

    /* TIM is at least 6 bytes, + 4 for FCS */
    while ((A_UINT8 *)pele < ((A_UINT8 *)bFrame) + frameLen - 10) {
        if (pele->elementID == ELE_TIM) {
            /* Found a TIM ! */
            index = assocId >> 3;
            mask = 0x1 << (assocId - index * 8);
            /* See if our assocId is outside the bitmap range */
            if (index < (pele->bitmapControl & 0x7f)) {
                return FALSE;
            }
            index -= (pele->bitmapControl & 0x7f);
            if (pele->bitmapControl & 0x80) {
                *pbrmc = TRUE;
            }
            return (A_BOOL)(pele->bitmap[index] & mask);
        }
        pele = (TIM_ELEMENT *)((A_UINT8 *) pele + pele->length + 2);
    }

    return FALSE;
}

/* mlmeMgtTxCompletion - Process completion of mgt transmit descriptor
 *
 * DESCRIPTION:
 * This routine updates the station state in SIB based on success or
 * failure of mgt frame that was scheduled earlier. Certain MLME primitives
 * are not called completed until that frame is transmitted and ack received.
 *
 * INPUT PARAMETERS
 *    pDesc - Pointer to descriptor with valid tx status field.
 */

void
mlmeMgtTxCompletion(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_FRAME_ASSOC_RESP *pAssocRespFrame;
    WLAN_MGT_MAC_HEADER   *pFrame;
    SIB_ENTRY             *pSib;
    A_BOOL                result;

    pFrame = pDesc->pBufferVirtPtr.mgtHeader;

    result = (A_BOOL)pDesc->hwStatus.tx.pktTransmitOK;
    pSib = sibEntryFind(&pFrame->destAddr);

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("mlmeMgtTxCompletion result: %s. %s, ",
                result ? "ok" : "error",
                 mgtTypeStr[pFrame->frameControl.fSubtype]);
        printMacAddress(pFrame->destAddr);
        uiPrintf("\n");
    }
#endif  /* DEBUG */

    switch (pFrame->frameControl.fSubtype) {
    case SUBT_AUTH:
    {
        WLAN_FRAME_AUTH *p1 = (WLAN_FRAME_AUTH *)pFrame;
        A_UINT16        algNo;
        A_UINT16        transSeqNo;
        A_UINT16        statusCode;

        if (pFrame->frameControl.wep) {
            algNo      = ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->algNo;
            transSeqNo = ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->transSeqNo;
            statusCode = ((WLAN_FRAME_AUTH_ENCRYPT *)pFrame)->statusCode;
        } else {
            algNo      = ((WLAN_FRAME_AUTH *)pFrame)->algNo;
            transSeqNo = ((WLAN_FRAME_AUTH *)pFrame)->transSeqNo;
            statusCode = ((WLAN_FRAME_AUTH *)pFrame)->statusCode;
        }
        if ((!result) && (transSeqNo == 1 || transSeqNo == 3)) {
            /* We are a sta. Can't send out frames. Abort */
            pdevInfo->localSta->transState = TSTATE_QUIET;
            A_UNTIMEOUT(&(pdevInfo->smeTimer));
            wlanMlmeAuthConfirm (pdevInfo, &pFrame->destAddr,
                                 p1->algNo, MLME_TIMEOUT);
            pdevInfo->localSta->stats.txAuthenticateFailStatus = statusCode;
            A_MACADDR_COPY(&pFrame->destAddr,
                           &pdevInfo->localSta->stats.txAuthenticateFailSta);
        }
        else if ((transSeqNo == 2 && algNo == AUTH_OPEN_SYSTEM) ||
                 (transSeqNo == 4 && algNo == AUTH_SHARED_KEY)) {
            // we are on AP
            if (!result || (statusCode != WLAN_SUCCESS)) {
                // something failed, reset the transitional state,
                // change staState
                pSib->transState = TSTATE_QUIET;
                pSib->staState  &= ~(STATE_AUTH);

                pdevInfo->localSta->stats.txAuthenticateFailStatus = statusCode;
                A_MACADDR_COPY(&pFrame->destAddr,
                               &pdevInfo->localSta->stats.txAuthenticateFailSta);
            } else {
                // final successful frame has been transmitted ok.
                pSib->staState  |= STATE_AUTH;
                pSib->transState = TSTATE_QUIET;
                pSib->authenAlgo = p1->algNo;
                pdevInfo->localSta->authenAlgo = p1->algNo;
                wlanMlmeAuthIndication(pdevInfo, &pFrame->destAddr, algNo);
            }
        }

        // do nothing for successful tx of frame 2 in AUTH_SHARED_KEY
    }
    break;

    case SUBT_ASSOC_REQ:
        if (!result) {
            /* failed to generate first frame on sta.  Abort */
            A_UNTIMEOUT(&(pdevInfo->smeTimer));
            pdevInfo->localSta->transState = TSTATE_QUIET;
            wlanMlmeAssocConfirm(pdevInfo, MLME_TIMEOUT);
        }
        break;

    case SUBT_REASSOC_REQ:
        if (!result) {
            /* failed to generate first frame on sta. Abort */
            pdevInfo->localSta->transState = TSTATE_QUIET;
            wlanMlmeReassocConfirm(pdevInfo, MLME_TIMEOUT);
        }
        break;

    case SUBT_ASSOC_RESP:
        pAssocRespFrame  = (WLAN_FRAME_ASSOC_RESP *)pFrame;
        pSib->transState = TSTATE_QUIET;
        if (result && (pAssocRespFrame->statusCode == WLAN_SUCCESS)) {
            /* Successful transmission of response on AP */
            pSib->staState |= STATE_ASSOC;
            //TODO: Set rates for client sta
            // mlmeStaRateAssign();
            // RateCtrlSibTxRateInit(pSib,..);

            wlanMlmeAssocIndication(pdevInfo, &pFrame->destAddr);
        } else {
            pSib->staState &= ~(STATE_ASSOC);
            pSib->listenInterval = 0;
            // release the associd since we don't need it
            if (pSib->assocId) {
                sibPowerStateClean(pSib);
                mlmeAssocIdRelease(pSib->assocId);      /* Currently assigned ID */
            }
            sibCseStateClean(pSib);
            pSib->assocId = 0;
        }
        break;

    case SUBT_DEAUTH:
    {
        WLAN_FRAME_DEAUTH *p1 = (WLAN_FRAME_DEAUTH *)pFrame;
        // do not clear transState here. It is already cleared before
        // placing this frame in the tx queue. By now, SME may have started a join
        // operation for next connection. Don't interferce with it.
        pdevInfo->localSta->stats.txDeAuthReasonCode = p1->reasonCode;
        A_MACADDR_COPY(&pFrame->destAddr,
                       &pdevInfo->localSta->stats.txDeAuthenticationSta);
        wlanMlmeDeauthConfirm(pdevInfo,&pFrame->destAddr,MLME_SUCCESS);

        /* can't do anything on failed transmit of response on AP */
    }
    break;

    case SUBT_DISASSOC:
    {
#ifdef  DEBUG
        if (mlmeDebug) {
            uiPrintf("mlmeMgtTxCompletion -- diassoc result: %s\n",
                     result ? "ok" : "error");
        }
#endif  /* DEBUG */
        pdevInfo->localSta->stats.txDisAssocReasonCode = ((WLAN_FRAME_DISASSOC *)pFrame)->reasonCode;
        A_MACADDR_COPY(&pFrame->destAddr,
                       &pdevInfo->localSta->stats.txDisAssociationSta);
        // do not clear transState here. It is already cleared before
        // placing this frame in the tx queue. By now, SME may have started a join
        // operation for next connection. Don't interferce with it.
        wlanMlmeDisassocConfirm(pdevInfo, &pFrame->destAddr, MLME_SUCCESS);

        /* can't do anything on failed transmit of response on AP */
    }
    break;

    case SUBT_PROBE_REQ:
    case SUBT_PROBE_RESP:
    default:
        /* Ignore the success or failure of transmit*/
        break;
    }
}

void 
smePendDescProcess (WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSib)
{
    ATHEROS_DESC *pTmp;
    ATHEROS_DESC *pNext;
    if (pSib) {
        if (pSib->smePendDesc != NULL && pSib->smePendDesc != (ATHEROS_DESC *)0xffffffff) {
            // we have some mgt frames pending for this SIB. process them.
            pTmp = pSib->smePendDesc;
            pSib->smePendDesc = NULL;
            while (pTmp != NULL) {
                uiPrintf("smePendDescProcess: Cleaning up pending rx %p\n", pTmp);
                pNext = pTmp->pNextVirtPtr;
                mlmeMgtRxCompletion(pdevInfo, pTmp);
                pTmp = pNext;
            }
        }
        pSib->smePendDesc = NULL; // if it was 0xffffffff, change to 0
    }
}

/* mlmeMgtRxCompletion - Routine to process incoming mgt frame
 *
 * DESCRIPTION
 * It calls proper mlme routine based on incoming frame type and state of the station
 */
void
mlmeMgtRxCompletion(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_MGT_MAC_HEADER *pmacHdr = pDesc->pBufferVirtPtr.mgtHeader;
    A_UINT16    ret;
    A_UINT16    framelen;
    SIB_ENTRY   *pSib;

    if (A_MACADDR_COMP(&(pdevInfo->localSta->macAddr), &(pmacHdr->destAddr)) != 0 &&
        A_MACADDR_COMP(&beaconBroadcastAddr, &(pmacHdr->destAddr)) != 0)
    {
        // mgt frame not directed to us
        freeBuffandDesc(pdevInfo, pDesc);
        return;
    }

    /* The "wep" bit can only be on for auth frames. */
    if (pmacHdr->frameControl.wep &&
        (pmacHdr->frameControl.fSubtype != SUBT_AUTH))
    {
        freeBuffandDesc(pdevInfo, pDesc);
        return;
    }

    framelen = (A_UINT16)pDesc->hwStatus.rx.dataLength;
#ifdef  DEBUG
    if (probeDebug || (mlmeDebug && pmacHdr->frameControl.fSubtype != SUBT_PROBE_REQ)) {
        uiPrintf("\nmlmeMgtRxCompletion. %s, ",
                 mgtTypeStr[pmacHdr->frameControl.fSubtype]);
        printMacAddress(pmacHdr->srcAddr);
        uiPrintf("\n");
        if (mlmeDebug > 1) {
            displayMgtFrame(pDesc->pBufferVirtPtr.mgtHeader, framelen);
        }
    }
#endif  /* DEBUG */

    switch (pmacHdr->frameControl.fSubtype) {
        // in this switch, we just take care of non auth/assoc frames, beacons
        // first because they are most frequent.

    case SUBT_BEACON:
#if 0
        if (!(pdevInfo->localSta->transState & TSTATE_SCANNING)) {
            uiPrintf("mlmeMgtRxCompletion: Beacon from BSSID: ");
            printMacAddress(((WLAN_FRAME_BEACON *) pmacHdr)->macHeader.bssId);
            uiPrintf("\nWarning: Another AP on this channel?\n");
        }
#endif
        mlmeBeaconReceive(pdevInfo, pDesc);
        freeBuffandDesc(pdevInfo, pDesc);
        return;

    case SUBT_PROBE_REQ:
        mlmeProbeRequestReceive(pdevInfo, (WLAN_FRAME_PROBE_REQ *) pmacHdr,
                                framelen);
        freeBuffandDesc(pdevInfo, pDesc);
        return;

    case SUBT_PROBE_RESP:
        mlmeProbeRespReceive(pdevInfo, pDesc);
        freeBuffandDesc(pdevInfo, pDesc);
        return;

    default:
        break;
    }

    // Now see if we can process this rx descr.
    // if smePendDesc is NULL, there is no tx descr pending completion
    // if smePendDesc is 0xffffffff, some tx descr is pending completion, but no rx yet.
    // if smePendDesc is a valid pointer, we already have received frames and they are waiting
    // for tx frame to finish.
    if (pdevInfo->localSta->serviceType == WLAN_AP_SERVICE) {
        pSib = sibEntryFind(&pmacHdr->srcAddr);
    } else {
        // packet from AP, check local state
        pSib = pdevInfo->localSta;
    }

    if (pdevInfo->localSta->smePendDesc != NULL) {
        pDesc->pNextVirtPtr = NULL;
        if (pdevInfo->localSta->smePendDesc == (ATHEROS_DESC *) 0xffffffff) {
            // some Tx pending, but no rx pending as of yet
            // pDesc becomes first descriptor in the list
            pdevInfo->localSta->smePendDesc = pDesc;
            uiPrintf("mgtRxCompletion: first pending rx descriptor %p\n", pDesc);
        } else {
            // walk thru link list and tack it on at the end
            ATHEROS_DESC *pTmp = pdevInfo->localSta->smePendDesc;
            uiPrintf("mgtRxCompletion: additional pending rx descriptor %p\n", pDesc);
            while (pTmp->pNextVirtPtr != NULL) {
                pTmp = pTmp->pNextVirtPtr;
            }
            pTmp->pNextVirtPtr = pDesc;
        }
        // return from here, don't delete pDesc
        return;
    }

    switch (pmacHdr->frameControl.fSubtype) {
    case SUBT_ASSOC_REQ:
        pdevInfo->localSta->stats.Associations++;
        ret = mlmeAssocApReceive(pdevInfo, (WLAN_FRAME_ASSOC_REQ *) pmacHdr,
                                 framelen);
        break;

    case SUBT_ASSOC_RESP:
        ret = mlmeAssocStaReceive(pdevInfo, (WLAN_FRAME_ASSOC_RESP *) pmacHdr,
                                  framelen);
        break;

    case SUBT_REASSOC_REQ:
        pdevInfo->localSta->stats.Associations++;
        ret = mlmeReassocApReceive(pdevInfo, (WLAN_FRAME_REASSOC_REQ *) pmacHdr,
                                   framelen);
        break;

    case SUBT_REASSOC_RESP:
        ret = mlmeReassocStaReceive(pdevInfo, (WLAN_FRAME_REASSOC_RESP *) pmacHdr,
                                    framelen);
        break;

    case SUBT_DISASSOC:
        ret = (A_UINT16) mlmeDisassocReceive(pdevInfo, (WLAN_FRAME_DISASSOC *) pmacHdr,
                                             framelen);
        break;

    case SUBT_AUTH:
    {
        WLAN_FRAME_AUTH *authFrame = (WLAN_FRAME_AUTH *)pmacHdr;

        if (authFrame->macHeader.frameControl.wep) {
            /* WEP is set, must be frame #3 for the AP.  Call will
             * handle other sequence numbers with the wep bit on.
             */
            ret = mlmeAuthAp3Receive(pdevInfo,
                                     (WLAN_FRAME_AUTH_ENCRYPT *)pmacHdr,
                                     framelen,
                                     (A_BOOL)pDesc->decryptError);
            break;
        }

        switch (authFrame->transSeqNo) {
        case 1:
            /* frame 1 received on AP */
            pdevInfo->localSta->stats.Authentications++;
            ret = mlmeAuthAp1Receive(pdevInfo, (WLAN_FRAME_AUTH *)pmacHdr,
                                     framelen);
            break;
        case 2:
            /* frame 2 received on STA */
            ret = mlmeAuthSta2Receive(pdevInfo,
                                      (WLAN_FRAME_AUTH_CHLG *)pmacHdr,
                                      framelen);
            break;
        case 4:
            /* frame 4 received on STA */
            ret = mlmeAuthSta4Receive(pdevInfo, (WLAN_FRAME_AUTH *)pmacHdr,
                                      framelen);
            break;
        default:
            uiPrintf("mlmeMgtRxCompletion: ERROR: Bad authentication transaction sequence number: %d\n",
                     authFrame->transSeqNo);
            break;
        }
    }
    break;

    case SUBT_DEAUTH:
        ret = (A_UINT16) mlmeDeauthReceive(pdevInfo, (WLAN_FRAME_DEAUTH *) pmacHdr,
                                           framelen);
        break;

    default:
        uiPrintf("mlmeMgtRxCompletion: unexpected subtype %d\n",
                 pmacHdr->frameControl.fSubtype);
        break;
    }
    freeBuffandDesc(pdevInfo, pDesc);
    return;
}

void
drvBBChange(WLAN_DEV_INFO *pdevInfo, CHAN_VALUES *pChval)
{
    // We changed from OFDM to CCK? 5 to 2 GHz? or vice versa
    if (pChval->channelFlags & CHANNEL_CCK) {
        // cck i.e. 2.4 GHz 11b mode
        pdevInfo->staConfig.pRateTable = &globalRateTable11b;
        mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.cckRateSet,
                        (INFO_ELEMENT *)&pdevInfo->staConfig.operationalRateSet);
        mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.cckRateSet,
                        (INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet);
        pdevInfo->localSta->currentOperationalRateIndex = pdevInfo->staConfig.defaultCckRateIndex; // user configured value for 11b
        pdevInfo->localSta->defaultRateIndex = pdevInfo->staConfig.defaultCckRateIndex; // user configured value for 11b
    } else {
        // ofdm mode. can be 5 GHz or 2.4 GHz 'g' mode
        pdevInfo->staConfig.pRateTable = &globalRateTable11a;
        mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet,
                        (INFO_ELEMENT *)&pdevInfo->staConfig.operationalRateSet);
        mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->staConfig.ofdmRateSet,
                        (INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet);
        pdevInfo->localSta->currentOperationalRateIndex = pdevInfo->staConfig.defaultOfdmRateIndex; // user configured value for 11a
        pdevInfo->localSta->defaultRateIndex = pdevInfo->staConfig.defaultOfdmRateIndex; // user configured value for 11a
    }
    
    // Let rate control switch its tables too.
    RateCtrlBBChange(pdevInfo, pChval->channelFlags);

}
#ifndef BUILD_AP    /* Only used by STA, AP uses chipReset. TODO - commonize. */

/******************************************************************
 * drvHwReset - Reset the hardware and restore WLAN operation
 *
 * DESCRIPTION: Resets the hardware and restores WLAN operation without
 * affecting SME state, power management state, beacon operation etc.
 *
 * On Entry: Caller must stop tx/rx and disable interrupts
 * On Exit:  interrupts are disabled.
 */

A_STATUS
drvHwReset(WLAN_DEV_INFO *pdevInfo, CHAN_VALUES *pChval)
{
    A_STATUS status     = A_OK;
    A_BOOL   bbChange   = ((pChval->channelFlags & CHANNEL_CCK) ^
                            (pdevInfo->staConfig.phwChannel->channelFlags & CHANNEL_CCK)) || 0;

    /* Reset the hardware. Indicate that it is a channel change */
    status = halReset(pdevInfo, pChval, TRUE);
    if (status != A_OK) {
        return status;
    }

    if (status == A_OK && bbChange) {
        /* do the needful for rates, etc. after switching baseband */
        drvBBChange(pdevInfo, pChval);
    }

    pdevInfo->globISRReg = 0;

    /* Restore LEDs */
    halSetLedState(pdevInfo, pdevInfo->pOSHandle->connectionStatus);

    /* Restore HW encrypt engine setting */
    setupSwEncryption(pdevInfo);

    /* Restore power save. Must do this before PS-Polls can be re-enabled in halWriteAssocid() */
    powerHwResetNotify(pdevInfo);

    /* Restore SME related hardware state */
    if (pdevInfo->bssDescr->bsstype == INFRASTRUCTURE_BSS) {
        /* Resync beacon timers */
        if (pdevInfo->localSta->staState & STATE_JOINED) {
            WLAN_TIMESTAMP tsf;

            halGetTsf(pdevInfo, &tsf);
            halSetStaBeaconTimers(pdevInfo, (tsf.low >> 10) | (tsf.high << 22),
                pdevInfo->bssDescr->DTIMPeriod, pdevInfo->bssDescr->cfParamSet.cfpCount);
        }

        /* Set AID,  Re-enable PS-Polls */
        if (pdevInfo->localSta->staState & STATE_ASSOC) {
            halWriteAssocid(pdevInfo, &pdevInfo->localSta->bssId, pdevInfo->localSta->assocId,
                            pdevInfo->bssDescr->timOffset);
        }

    } else {
        /* Adhoc */
        halSetAdhocMode(pdevInfo);
    }

    /* Restore beacon generation (if currently enabled) */
    wlanBeaconHwResetNotify(pdevInfo);

    return status;
}

#endif /* #ifndef BUILD_AP */

/******************************************************************
 * drvChangeChannel - Change A2/D2 to a new channel
 *
 * DESCRIPTION: The routine does whatever is necessary to shut down
 * tx, rx, drain queues, change channel in the device, it should
 * reset the chip if required to change channel. Later, it restarts
 * the receive queue with the receive filter value of rxFilter.
 * This routine is primarily used during scan to switch thru a list
 * of channels and listening to the beacons.
 */

A_STATUS
drvChangeChannel(WLAN_DEV_INFO *pdevInfo, CHAN_VALUES *pChval, A_UINT32 rxFilter)
{
    A_STATUS status = A_OK;

#ifdef BUILD_AP

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("drvChangeChannel: Switching from channel %d to %d\n",
                 pdevInfo->staConfig.phwChannel->channel, pChval->channel);
    }
#endif
    chipReset(pdevInfo, FALSE, pChval);
    halRxFilter(pdevInfo, 0xffffffff, HAL_RX_FILTER_CLEAR); // clear all bits
    halRxFilter(pdevInfo, rxFilter, HAL_RX_FILTER_SET); // set given bits true

#else

    if (pdevInfo->NicResetInProgress ||
            pdevInfo->staConfig.phwChannel->channel == pChval->channel) {
        /* Don't channel change if a NIC reset is in progress. This avoids
         * recursion when StopReceive is called (SME may call DrvChangeChannel again)
         * But make sure the Rx filter is set.
         */
        halRxFilter(pdevInfo, 0xffffffff, HAL_RX_FILTER_CLEAR); 
        halRxFilter(pdevInfo, rxFilter, HAL_RX_FILTER_SET);     
        return A_OK;
    }

    /* Reject NDIS send around change channel */
    pdevInfo->invalidTxChannel = TRUE;

    halDisableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    StopSwReceive(pdevInfo);
    (void) StopTransmit(pdevInfo, RESET_TX_HW | RESET_TX_Q, FALSE, TRUE);
    StopReceive(pdevInfo);

    /* Reset the chip, changing the channel as well */
    status = drvHwReset(pdevInfo, pChval);

    halEnableInterrupts(pdevInfo, HAL_INT_GLOBAL);
    StartReceiveUnit(pdevInfo, rxFilter);
    RestartTransmit(pdevInfo);

    /* Allow NDIS send if now on configured channel */
    if (pdevInfo->staConfig.phwChannel == pdevInfo->staConfig.pChannel) {
        pdevInfo->invalidTxChannel = FALSE;
        osIndicateResourcesAvailable(pdevInfo);
    }

#endif  /* BUILD_AP */

    return status;
}

/***************************/
/* Beacon related routines */
/***************************/

void
mlmeBsssetCopy(BSSDESCR_SET *src, BSSDESCR_SET *dest)
{
    A_UINT16 sindex = 0;

    while (sindex < src->count && dest->count < BSSDESCR_SET_SIZE) {
        A_DRIVER_BCOPY(&src->bssDescrArray[sindex++],
                       &dest->bssDescrArray[dest->count++],
                       sizeof(BSSDESCR));
    }
}

void
mlmeBsssetMerge(BSSDESCR_SET *src, BSSDESCR_SET *dest)
{
    A_UINT16 sindex = -1;
    A_UINT16 dindex = 0;

    while (++sindex < src->count && dest->count < BSSDESCR_SET_SIZE) {
        for (dindex=0; dindex < dest->count; dindex ++) {
            if (A_MACADDR_COMP(&src->bssDescrArray[sindex].bssid,
                               &dest->bssDescrArray[dindex].bssid) == 0) {
                // BSSID already in destination set, skip it
                continue;
            }
        }
        A_DRIVER_BCOPY(&src->bssDescrArray[sindex], &dest->bssDescrArray[dest->count++],
                       sizeof(BSSDESCR));
    }
}

void
mlmeBsssetAdd(BSSDESCR *src, BSSDESCR_SET *dest)
{
    BSSDESCR *pOrig;

    if ((pOrig = mlmeBsssetFind(dest, &src->bssid, src->pChannel->channelFlags)) != NULL) {
        // Same bssid/channel type combination is already there. Replace it.
        // channel may change in the process
        if (isNullSsid(&src->ssid)) {
            /* If we already have real ssid for this null ssid, copy that to src */
            mlmeElementCopy((INFO_ELEMENT *)(&pOrig->ssid), (INFO_ELEMENT *)&src->ssid);
        }
        A_DRIVER_BCOPY(src, pOrig, sizeof(BSSDESCR));
        return;
    }

    if (dest->count < BSSDESCR_SET_SIZE) {
        A_DRIVER_BCOPY(src, &dest->bssDescrArray[dest->count++],
                   sizeof(BSSDESCR));
    }
}

void
mlmeBsssetAge(BSSDESCR_SET *src, A_UINT32 msAge)
{
    A_UINT16    di;
    A_UINT32    validTime;

    validTime = A_MS_TICKGET() - msAge;

    if (src == NULL || src->count <= 0) {
        return;
    }
    for (di = 0; di < src->count; di ++) {
        if (src->bssDescrArray[di].scanTime < validTime) {
            // too old, discard this bssdescr
            // move the last entry here, order is not important
            if (di < src->count - 1) {
                // no need to move last entry
                A_DRIVER_BCOPY(&src->bssDescrArray[src->count-1], &src->bssDescrArray[di],
                               sizeof(BSSDESCR));
            }
            src->count --;
        }
    }
}

BSSDESCR *
mlmeBsssetFind(BSSDESCR_SET *pSet, WLAN_MACADDR *pAddr, A_UINT16 channelFlags)
{
    int    i;

    if (pSet == NULL || pAddr == NULL) {
        return NULL;
    }
    for (i = 0; i < pSet->count; i++) {
        if (A_MACADDR_COMP(pAddr, &pSet->bssDescrArray[i].bssid) == 0 &&
            pSet->bssDescrArray[i].pChannel->channelFlags == channelFlags) {
            // found the matching BSSID
            return &pSet->bssDescrArray[i];
        }
    }
    return NULL;
}

void
mlmeBsssetDel(BSSDESCR_SET *pSet, WLAN_MACADDR *pAddr)
{
    int    i;

    if (pSet == NULL || pAddr == NULL) {
        return;
    }
    for (i = 0; i < pSet->count; i++) {
        if (A_MACADDR_COMP(pAddr, &pSet->bssDescrArray[i].bssid) == 0) {
            // found the matching BSSID
            // move the last entry here, order is not important
            if (i < pSet->count - 1) {
                // no need to move last entry
                A_DRIVER_BCOPY(&pSet->bssDescrArray[pSet->count-1], &pSet->bssDescrArray[i],
                               sizeof(BSSDESCR));
            }
            pSet->count--;
            return;
        }
    }
}

/* isNullSsid
 * A null ssid is either 0 bytes in length or non-zero length but all bytes
 * are 0. Strange but true. It's a common method used by vendors like Cisco.
 */
A_BOOL  isNullSsid(SSID *pSsid)
{
    A_UINT8 i;

    if (pSsid == NULL) {
        return TRUE;
    }
    i = pSsid->length;
    while (i > 0) {
        if (pSsid->ssid[--i] != 0) {
            return FALSE;
        }
    }
    return TRUE;
}

A_UINT32    maxTxRateCompute(RATE_SET *pRates, A_UINT16 channelFlags)
{
    int i;
    A_UINT8 maxrate;
    // Figure the highest rate supported by the current link.
    maxrate = 0;
    for (i=0; i < pRates->length; i++) {
                maxrate = max((pRates->rates[i] & 0x7f), maxrate);
    }
    if (channelFlags & CHANNEL_TURBO) {
        return (((A_UINT32) maxrate) * 500 * 2);
    } else {
        return (((A_UINT32) maxrate) * 500);
    }

}

A_INT8  mlmeGetControlRateIndex(RATE_SET *pRates, A_INT8 rateIndex)
{
    A_INT8  i = rateIndex;
    /* pRates is assumed to be in ascending order of rates */
    if (pRates == NULL || pRates->length <= rateIndex) {
        return 0; /* 0 is a valid rate index */
    }
    while (i >= 0) {
        if (pRates->rates[i] & 0x80) {
            return i;
        }
        i --;
    }
    return 0;
}
/* mlmeBeaconReceive - Parse and process received beacon frame
 *
 * Spec section 11.1.2.3
 * Beacons will be typically reach the MLME software only when
 * enabled for scan or join operations.
 */

void
mlmeBeaconReceive(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
#ifndef    BUILD_AP
    A_BOOL              brmc;           /* Broadcast/Multicast in beacon ? */
    CF_PARAM_SET        *cfp;
    SIB_ENTRY           *apSib;
#endif    /* BUILD_AP */
    WLAN_FRAME_BEACON   *bFrame;        /* Incoming frame */
    A_UINT16            len;            /* Total frame length */
    SSID                *pSsid;
    TIM_ELEMENT         *ptim;
    DS_PARAM_SET        *pDsparams;
#ifndef BUILD_AP
    A_BOOL               faulted;
#endif

    bFrame = pDesc->pBufferVirtPtr.beacon;
    len = (A_UINT16)pDesc->hwStatus.rx.dataLength;

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("mlmeBeaconReceive: Beacon from ");
        printMacAddress(bFrame->macHeader.bssId);
        uiPrintf(" on channel %d, transState: %x\n",
                 pdevInfo->staConfig.phwChannel->channel, pdevInfo->localSta->transState);
    }
#endif  /* DEBUG */

    // basic sanity check
    pSsid = (SSID *) mgtElementGet(&bFrame->macHeader, len, ELE_SSID);
    ptim  = (TIM_ELEMENT*)mgtElementGet(&bFrame->macHeader, len, ELE_TIM);
    pDsparams = (DS_PARAM_SET *)mgtElementGet(&bFrame->macHeader, len, ELE_DS_PARAM_SET);
    if (pSsid == NULL || pSsid->length == 0 ||
        (bFrame->capInfo.ess && ptim == NULL)) {
        // badly formatted beacon, invalid SSID or TIM
        return;
    }

    // check to see if we should accept this beacon
    if ((pDsparams != NULL) &&
        (pDsparams->currentChannel != wlanConvertGHztoCh(pdevInfo->staConfig.phwChannel->channel,
                                        pdevInfo->staConfig.phwChannel->channelFlags))) {
        // Picked up this beacon on a wrong channel. Can happen in 11b. Beacons can be
        // heard anywhere in the 25 MHz vicinity. Ignore it. Wait for true channel.
        return;
    }

#ifndef    BUILD_AP
    /* If we already have an entry for the sta of this AP, mark
     * its state as AP up and running. If we don't have an entry,
     * don't bother.
     */
    if ((apSib = sibEntryFind(&bFrame->macHeader.srcAddr)) != NULL) {
        /* May not be our AP, but update state & timestamp */
        apSib->staState |= STATE_AP_UP;
        apSib->lastBeaconTime = bFrame->timestamp;
    }

    /* Are we already in BSS ? */
    if (pdevInfo->localSta->staState & STATE_JOINED) {
        /* Is this beacon from our BSS ? */
        if ((A_MACADDR_COMP(&(pdevInfo->localSta->bssId),
                            &(bFrame->macHeader.bssId))) == 0) {
            /* We are in a live BSS */
            /* Copy timestamp */
            pdevInfo->localSta->lastBeaconTime = bFrame->timestamp;
        }    /* if A_MACADDR_COMP == 0 */
        else if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS &&
                 bFrame->capInfo.ibss &&
                 pSsid->length ==  pdevInfo->bssDescr->ssid.length &&
                 A_BCOMP(&pSsid->ssid, &pdevInfo->bssDescr->ssid.ssid, pSsid->length) == 0)
        {
            // It's ad-hoc network, ssid matches but bssid doesn't.
            // 802.11 protocol requires that if two different bssid's for the same ad-hoc
            // ssid are found in the network, they should merge. One with an earlier timestamp must prevail.
            UINT32 newTSFLow, newTSFHigh, tsf;
            WLAN_TIMESTAMP ts;

            halGetTsf(pdevInfo, &ts);
            newTSFLow  = ts.low;
            newTSFHigh = ts.high;

            if (newTSFHigh > bFrame->timestamp.high ||
                (newTSFHigh == bFrame->timestamp.high &&
                 newTSFLow >= bFrame->timestamp.low)) {
                // ours is a higher timestamp, we started early.
                return;
            }

            // now we need to change our bssid
            // BSSID will change here
            A_MACADDR_COPY(&(bFrame->macHeader.bssId), &pdevInfo->staConfig.bssId);
            A_MACADDR_COPY(&(bFrame->macHeader.bssId), &pdevInfo->localSta->bssId);
            A_MACADDR_COPY(&(bFrame->macHeader.bssId), &pdevInfo->bssDescr->bssid);
            /* Call the routine below just to set the BSS that we want to join */
            halWriteAssocid(pdevInfo, &(pdevInfo->bssDescr->bssid), 0, 0);
            pdevInfo->localSta->lastBeaconTime = bFrame->timestamp;
            tsf = (pdevInfo->localSta->lastBeaconTime.low >> 10) |
                  (pdevInfo->localSta->lastBeaconTime.high << 22);
            // TODO: ensure that beacons start going out with this call.
            // beaconInit may need to be called with/without enabling beacons
            wlanBeaconInit(pdevInfo, tsf);
        }
    } /* if STATE_JOINED */

    /* We are not in a BSS yet. */
    if (pdevInfo->localSta->transState & TSTATE_JOINING) {
        /* Is this beacon from our BSS ? */
        if ((A_MACADDR_COMP(&(pdevInfo->bssDescr->bssid),
                            &(bFrame->macHeader.bssId))) == 0) {
            if ((pdevInfo->bssDescr->bsstype == INFRASTRUCTURE_BSS && !(bFrame->capInfo.ess )) ||
               (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS     && !(bFrame->capInfo.ibss))) {
                // SSID matches, but type of network doesn't. This can happen when same SSID
                // is used by infrastructure network on one channel and ad-hoc network on another
                return;
            }
            /* Yes ! Got the beacon we're waiting for */
            /* Complete the join process. */
            /* Lock station */
            /* Update the local time to synchronize. How? */
            pdevInfo->localSta->staState  |= STATE_JOINED;
            pdevInfo->localSta->transState = TSTATE_QUIET;

            /* Copy timestamp and other elements from the beacon */
            pdevInfo->localSta->lastBeaconTime = bFrame->timestamp;
            pdevInfo->localSta->capInfo.ess    = bFrame->capInfo.ess;
            pdevInfo->localSta->capInfo.ibss   = bFrame->capInfo.ibss;
            pdevInfo->staConfig.pChannel       = pdevInfo->staConfig.phwChannel;
            pdevInfo->invalidTxChannel = FALSE;
            osIndicateResourcesAvailable(pdevInfo);

            /*
             * Fill out bss descr parameters that may not have been set, since
             * we may have decided to join this bss based on a probe response
             * instead of a beacon.
             */
            pdevInfo->bssDescr->DTIMPeriod = ptim ?
                ptim->dtimPeriod : 0;
            pdevInfo->bssDescr->timOffset = ptim ?
                ((A_UINT16)((A_INT8 *)ptim - (A_INT8 *)&bFrame->macHeader)) : 0;

            // BSSID and SSID can change here
            A_MACADDR_COPY(&(bFrame->macHeader.bssId), &pdevInfo->staConfig.bssId);
            A_MACADDR_COPY(&(bFrame->macHeader.bssId), &pdevInfo->localSta->bssId);
            if (!isNullSsid(pSsid)) {
                mlmeElementCopy((INFO_ELEMENT *)pSsid,(INFO_ELEMENT *)&pdevInfo->staConfig.ssid);
            }

            /* Don't change cfPoll bits for now. */
            // halRxFilter(pdevInfo, HAL_RX_BEACON, HAL_RX_FILTER_CLEAR);
            if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS) {
                // keep receiving beacons. Detect other BSS on the channel and merge.
                halRxFilter(pdevInfo, (HAL_RX_UCAST|HAL_RX_MCAST|HAL_RX_BCAST|HAL_RX_BEACON), HAL_RX_FILTER_SET);
            } else {
                // no need to keep beacons on.
                halRxFilter(pdevInfo, (HAL_RX_UCAST|HAL_RX_MCAST|HAL_RX_BCAST), HAL_RX_FILTER_SET);
            }
            /* May already be associated, but rejoining */
            if (pdevInfo->localSta->staState & STATE_ASSOC) {
                halWriteAssocid(pdevInfo, &(bFrame->macHeader.bssId),
                                pdevInfo->localSta->assocId,
                                pdevInfo->bssDescr->timOffset);
                halSetLedState(pdevInfo, TRUE);
            } else {
                /* Call the routine below just to set the BSS */
                halWriteAssocid(pdevInfo, &(bFrame->macHeader.bssId), 0, 0);
            }

            if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS) {
                A_UINT32 tsf;

                /* Set up privacy, if any */
                if (pdevInfo->localSta->privacyInvoked) {
                    initCipher(pdevInfo);

                    wlanKeyEnable(pdevInfo, pdevInfo->localSta, pdevInfo->staConfig.defaultKey);
                    wlanKeyCacheFault(pdevInfo, pdevInfo->localSta, &faulted);

                    if (wlanKeyUpdate(pdevInfo, pdevInfo->localSta, CONTEXT_NONE) != A_OK) {
                        return;
                    } else {
                        /* set up encryption */
                        setupSwEncryption(pdevInfo);
                    }
                }

                /* set adhoc mode and start generating beacons too! */
                // TODO: Is software retry enabled in adhoc now?
                // pdevInfo->staConfig.swretryEnabled = FALSE;
                halSetAdhocMode(pdevInfo);
                tsf = (pdevInfo->localSta->lastBeaconTime.low >> 10) |
                      (pdevInfo->localSta->lastBeaconTime.high << 22);
                // TODO: ensure that beacons start going out with this call.
                // beaconInit may need to be called with/without enabling beacons
                wlanBeaconInit(pdevInfo, tsf);
                halSetLedState(pdevInfo, TRUE);
            } else {
                /* Start the beacon related timers in the h/w after getting the
                 * updated dtim count and cfp count; the rest of the parameters
                 * are assumed to not change for now.. later on we may want/need
                 * to keep track of them and update the h/w whenever the parameters
                 * change
                 */
                A_UINT32 tsf;

                tsf = (pdevInfo->localSta->lastBeaconTime.low >> 10) |
                      (pdevInfo->localSta->lastBeaconTime.high << 22);

                halSetStaBeaconTimers(pdevInfo, tsf, ptim->dtimCount,
                    (cfp = (CF_PARAM_SET*)mgtElementGet(&bFrame->macHeader, len, ELE_CF_PARAM_SET))
                    ? cfp->cfpCount : 0);

                if (pdevInfo->staConfig.swRetryMaxRetries == 0) {
                    pdevInfo->staConfig.swretryEnabled = FALSE;
                } else {
                    pdevInfo->staConfig.swretryEnabled = TRUE;
                }
            }

            /* Update our local SIB parameters */
            A_MACADDR_COPY(&(bFrame->macHeader.bssId),
                           &(pdevInfo->localSta->bssId));
            /* We need to create an entry for the STA of this AP */
            if ((apSib = sibEntryFind(&bFrame->macHeader.srcAddr)) == NULL )
            {
                apSib = sibEntryAlloc(pdevInfo->localSta,pdevInfo->bssDescr,&bFrame->macHeader.srcAddr );
                apSib->devNo = pdevInfo->devno;
                /* Add sta to hash table */
                if (sibEntryAdd(apSib) != WLAN_OK) {
                    uiPrintf("mlmeBeaconReceive: ERROR: failed to add apSib to hash table\n");
                    /* @@@@ Return some error */
                    return;
                }
            }
            apSib->staState    = STATE_AP_UP;     /* more state bits later */
            apSib->transState  = TSTATE_QUIET;
            apSib->serviceType = WLAN_AP_SERVICE; /* @@ ad-hoc? */
            A_MACADDR_COPY(&bFrame->macHeader.bssId, &apSib->bssId);
            apSib->capInfo     = bFrame->capInfo;
            mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                          len, ELE_SUPPORTED_RATES),
                            (INFO_ELEMENT *)&apSib->operationalRateSet);
            pdevInfo->localSta->stats.MaxTxRateKb =
                maxTxRateCompute(&apSib->operationalRateSet, pdevInfo->staConfig.pChannel->channelFlags);


            /* Initialize the working rate set to be used for communicating
             * to this sta. It's a provisional setting. Association may change
             * it.
             */
            mlmeElementCopy((INFO_ELEMENT *)&apSib->operationalRateSet, 
                            (INFO_ELEMENT *)&apSib->workingRateSet);
            RateCtrlSibTxRateInit(apSib);

            /* make the element invalid first */
            /* need remove this when the mlmeElementCopy changed */
            apSib->ucseSet.length = 0;
            mlmeElementCopy(mgtElementGet(&bFrame->macHeader, len, ELE_UCSE_SET),
                            (INFO_ELEMENT *)&apSib->ucseSet);

            A_UNTIMEOUT(&(pdevInfo->smeTimer));
            // No need to send a frame to consume sequence number 0. We save and
            // restore seq no. in reset called during channel change.

            wlanMlmeJoinConfirm(pdevInfo, MLME_SUCCESS);
            /* Unlock */
        }
        /* Not our beacon */
    } /* if TSTATE_JOINING */
#endif    /* BUILD_AP */

    if (pdevInfo->localSta->transState & TSTATE_SCANNING) {
        BSSDESCR     tmpBssDescr;
        BSSDESCR     *pbssDescr = &tmpBssDescr;        /* BSS Description set */
        INFO_ELEMENT *p1;
        A_BOOL       newBeacon = TRUE;

        /* We are in the middle of scanning process */
        /* See if we have already received beacon from this AP */
        if (pdevInfo->pscanInfo == NULL) {
            /* something wrong. We are scanning but set is not allocated */
            uiPrintf("mlmeBeaconReceive: ERROR: scanning but no set allocated\n");
            return;
        }

#if 0
        {
            A_UINT16     index;

            /* We may already have some beacons */
            for (index = 0; index < pdevInfo->pscanInfo->bssSet.count; index++) {
                if (A_MACADDR_COMP(&pdevInfo->pscanInfo->bssSet.bssDescrArray[index].bssid,
                                   &(bFrame->macHeader.bssId)) == 0)
#ifndef    BUILD_AP
                    if (pdevInfo->pscanInfo->bssSet.bssDescrArray[index].pChannel ==
                        pdevInfo->staConfig.phwChannel)
#endif    /* BUILD_AP */
                {
                    /* We already have it on the same channel */
                    newBeacon = FALSE;
                    break;
                }
            }
        } else {
        }
#endif

        if (pdevInfo->pscanInfo->bssSet.count == BSSDESCR_SET_SIZE ) {
            /* We maxed out on scanned beacons. Punt. */
            /* I don't imagine a situation where one will
             * get beacons from like 32 different access points.
             */
            return;
        }

        /* Assume that space for entire set is already allocated
         * by whoever initiated the scan.
         */
        if (newBeacon) {

#ifdef  DEBUG
            if (mlmeDebug) {
                uiPrintf("New beacon found ssid: %x\n", (A_UINT32)pSsid);
            }
#endif  /* DEBUG */

            /* index is available */
            pbssDescr->beaconInterval = bFrame->beaconInterval;
            ptim = (TIM_ELEMENT *)mgtElementGet(&bFrame->macHeader, len, ELE_TIM);
            pbssDescr->DTIMPeriod = ptim ?  ptim->dtimPeriod : 0;
            pbssDescr->bsstype = (bFrame->capInfo.ibss == 1) ?
                INDEPENDENT_BSS : INFRASTRUCTURE_BSS;
            A_MACADDR_COPY(&bFrame->macHeader.bssId, &pbssDescr->bssid);
            pbssDescr->pChannel = pdevInfo->staConfig.phwChannel;

            // rssi can be -ve. Treat 8 bit value as signed. covert to abs
            pbssDescr->rssi = (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength) < 0 ?
                - (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength) :
                (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength);
            pbssDescr->timestamp = bFrame->timestamp;
            pbssDescr->capabilityInfo = bFrame->capInfo;
            pbssDescr->timOffset = ptim ?
                ((A_UINT16)((A_INT8 *)ptim - (A_INT8 *)&bFrame->macHeader)) : 0;
            /* What do we do about 'localTime' field? */
            /* What about PHY param set? */
            p1 = (INFO_ELEMENT *) &bFrame->buffer;
            p1 = mlmeElementCopy((INFO_ELEMENT *) pSsid, (INFO_ELEMENT *)&pbssDescr->ssid);
            /* Need to make a correction. Copy only the basic rate
             * values (with msb = 1) into the BSS description */
            p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                               len, ELE_SUPPORTED_RATES),
                                 (INFO_ELEMENT *)&pbssDescr->bssRateSet);
            /* make the element invalid first */
            /* need remove this when the mlmeElementCopy changed */
            pbssDescr->bssUcseSet.length = 0;
            p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                               len, ELE_UCSE_SET),
                                 (INFO_ELEMENT *)&pbssDescr->bssUcseSet);
            p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                               len, ELE_CF_PARAM_SET),
                                 (INFO_ELEMENT *)&pbssDescr->cfParamSet);
            p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                               len, ELE_IBSS_PARAM_SET),
                                 (INFO_ELEMENT *)&pbssDescr->ibssParamSet);
            pbssDescr->pChannel = pdevInfo->staConfig.phwChannel;
            pbssDescr->scanTime = A_MS_TICKGET(); // save the time of reception

            // add or replace the existing entry
            mlmeBsssetAdd(pbssDescr, &pdevInfo->pscanInfo->bssSet);
#ifdef  DEBUG
            if (mlmeDebug) {
                uiPrintf("New beacon found, channel %d, bssSet.count: %d\n",
                         pbssDescr->pChannel->channel,
                         pdevInfo->pscanInfo->bssSet.count);
            }
#endif  /* DEBUG */
        }

#ifndef    BUILD_AP
        // check for expiration of minChannelTime
        if ((pdevInfo->pscanInfo->scanStartTime +
            pdevInfo->pscanInfo->minChannelTime ) < A_MS_TICKGET()) {
            // We are done with this channel

            A_UNTIMEOUT(&(pdevInfo->smeTimer));
            pdevInfo->pscanInfo->probeRetry = 0;

            if ((pdevInfo->pscanInfo->curChannelIndex + 1) >=
                pdevInfo->pscanInfo->pScanList->listSize) {
                // We finished last channel on the list
                // Clear the state. Don't let any more bss's creep in
                pdevInfo->localSta->transState = TSTATE_QUIET;
                wlanMlmeScanConfirm(pdevInfo,
                                    &pdevInfo->pscanInfo->bssSet, MLME_SUCCESS);
            } else {
                pdevInfo->pscanInfo->curChannelIndex++;
                // Move over to the next channel, curChannelIndex is already incremented.
                mlmeScanFunction(pdevInfo);
            }
#ifdef  DEBUG
            if (mlmeDebug) {
                uiPrintf("done with channel %d\n", pdevInfo->staConfig.phwChannel->channel);
            }
#endif  /* DEBUG */
        }
#endif    /* BUILD_AP */
    }
}

#ifndef BUILD_AP

/*
 * wlanMlmeJoinRequest - Join a BSS
 *
 * Spec. 10.3.3.1, 11.1.2.4
 * JOIN primitive logically establishes a binding to the AP.
 * It doesn't need to send any packet to the AP. It will turn on
 * the beacon delivery to driver and latch on to the next beacon.
 * MAC time should be synchronized with that beacon.
 * It is not clear if it needs to send a probe request or not.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM
 */

MLME_RESULT
wlanMlmeJoinRequest(WLAN_DEV_INFO *pdevInfo,
                    BSSDESCR      *pbssDescr,           /* pointer to BSS description */
                    A_UINT32      joinFailureTimeout,   /* timeout in TU */
                    A_UINT32      probeDelay,           /* probe delay in usec */
                    RATE_SET      *operationalRateSet)  /* Set of data rates */
{
    /* Validate values in BSS Description. */
    /* This should be one returned by a previous scan */

    if (pbssDescr == NULL || operationalRateSet == NULL) {
        return MLME_INVALID_PARAM;
    }

#if 0
    // rates may not match at this point in time, we may change from ofdm to cck
    if (basicRateSubsetCheck(&pbssDescr->bssRateSet,
                             operationalRateSet) == FALSE)
    {
        /* BSS basic rate set is not a subset of operational rate
         * set, i.e. STA cannot support all BSS basic rates. Fail.
         */
        return MLME_INVALID_PARAM;
    }
#endif

    if (pdevInfo->staConfig.privacyInvoked &&
        pdevInfo->keyTable[pdevInfo->staConfig.defaultKey].keyLength == 0) {
        // asking for WEP enabled frames, but key is not set.
        // cannot start this SSID.
        return MLME_INVALID_PARAM;
    }

    if (!pdevInfo->staConfig.privacyInvoked && pbssDescr->capabilityInfo.privacy) {
        // AP is asking for WEP, but we don't have it enabled.
        // cannot join this SSID.
        return MLME_INVALID_PARAM;
    }

    // Figure out short/long preamble setting before trying to join.
    pdevInfo->localSta->capInfo.shortPreamble = pdevInfo->staConfig.shortPreamble &&
                                                pbssDescr->capabilityInfo.shortPreamble;

    uiPrintf("wlanMlmeJoinRequest: channel %d, %s\n",
             pbssDescr->pChannel->channel, &pbssDescr->ssid.ssid);
    /* Change the channel */
    // Need to change the channel, turn on beacon delivery
    // Currently, enable tx before actually joining, but it's ok since we
    // are not going to change channel.
    // we must call this routine to set rx filter (even for same channel)

    if (drvChangeChannel(pdevInfo, pbssDescr->pChannel, HAL_RX_BEACON) != A_OK) {
        // Noisy channel
        return MLME_INVALID_PARAM;
    }

    /* Call the routine below just to set the BSS that we want to join */
    /* hardware needs this BSSID to set its TSF */
    halWriteAssocid(pdevInfo, &(pbssDescr->bssid), 0, 0);

    // Reset our TSF so that its value is lower than the beacon that we are trying
    // to catch. Only then hw will update its TSF register with the new beacon.
    halResetTsf(pdevInfo);

    /* beacon reveice routine should catch the next beacon and
     * finish the join operation.
     */
    pdevInfo->localSta->transState = TSTATE_JOINING;

    /* Since this BSS descr can come from any outside context,
     * we need to make our own copy and store it.
     */
    if (pdevInfo->bssDescr == NULL) {
        pdevInfo->bssDescr = (BSSDESCR *)A_DRIVER_MALLOC(sizeof(BSSDESCR),0,0);
        if (pdevInfo->bssDescr == NULL) {
            return MLME_DRIVER_ERROR;
        }
    }

    A_DRIVER_BCOPY(pbssDescr, pdevInfo->bssDescr, sizeof(BSSDESCR));

    /* Set a timer to cancel after joinFailureTimeout */
    A_TIMEOUT(&pdevInfo->smeTimer, joinFailureTimeout, (A_UINT32) pdevInfo, MLME_JOIN_TIMEOUT);

    return MLME_SUCCESS;
}

void
wlanMlmeJoinConfirm(WLAN_DEV_INFO *pdevInfo, MLME_RESULT result)
{
    /* If callback routine exists, call it */
    if (smeInfo.joinCallback) {
        (*smeInfo.joinCallback)(pdevInfo, result);
    }
    return;

}

#endif    /* BUILD_AP */

/* mlmeScanFunction Real scan function under a separate thread.
 *
 * DESCRIPTION
 * Since the scan function needs to run in background, we are
 * forced to do it in a separate thread. wlanMlmeScanRequest()
 * validates the parameters and fires off a thread that executes
 * this function. It passes on all the parameters that were
 * supplied in wlanMlmeScanRequest().
 * ALGORITHM:
 * For each channel
 *        send MAC_CMD_CHANNEL command to MAC driver
 *        if ACTIVE_SCAN
 *        then
 *            send MAC_CMD_SEND_PROBE command to MAC driver to send probe packet
 *            wait for probe response packet
 *            if valid probe responese and quickScan
 *                terminate scan
 *        else
 *            enable beacon delivery from MAC to host
 *            wait for beacons on that channel
 *            if valid beacon and quickScan
 *                terminate scan
 *        end
 *    end
 *
 * Minimum and maximum waiting time condition is also observed on each channel.
 */
void
mlmeScanFunction(WLAN_DEV_INFO *pdevInfo)
{
    SCAN_INFO   *pscanInfo;
    A_UINT32    timeOut;
    A_STATUS    status = A_OK;
    SSID        dummySsid;
    A_UINT32    rxFilter = 0;

    /* What to do with the current AP and current activity?
     * Several factors must be considered.
     * 1. If we are not connected to any AP, just start scanning.
     * 2. If we are connected to AP, should we drop all auth/assoc?
     * 3. It's possible to say that we are logically connected to
     *    the AP as long as we come back to the original channel
     * 4. If we are changing channel, we must wait for tx to finish
     * 5. Should we automatically go into PS mode?
     * 6. What do we do in ad-hoc?
     */

    /* For now, we will just wait for tx to finish */
    /* The code for 'wait for tx to finish' needs to be written
     * when interfaces to other modules become clear.
     */

    pscanInfo           = pdevInfo->pscanInfo;
    timeOut             = pscanInfo->maxChannelTime;
    dummySsid.elementID = ELE_SSID;
    dummySsid.length    = 0;

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("mlmeScanFunction: Switching to channel %d\n",
                 pscanInfo->pScanList->pChanArray[pscanInfo->curChannelIndex]->channel);
    }
#endif   /* DEBUG */

    if (!pscanInfo->probeRetry) {
        /* Change Channel until there is no interference detected */
        while (pscanInfo->curChannelIndex < pscanInfo->pScanList->listSize) {
            if (pscanInfo->curChannelIndex >= pdevInfo->staConfig.pClist->listSize) {
                uiPrintf("Current Index into scan list %d\n", pscanInfo->curChannelIndex);
                uiPrintf("Current ScanList size %d\n", pscanInfo->pScanList->listSize);
                uiPrintf("Configed channel list size %d\n", pdevInfo->staConfig.pClist->listSize);
                ASSERT(0);
            }
            rxFilter = (pscanInfo->scantype == PASSIVE_SCAN) ? HAL_RX_BEACON : (HAL_RX_BEACON | HAL_RX_UCAST);
            if (pdevInfo->staConfig.radarDetectEnable) {
                rxFilter |= HAL_RX_RADAR;
            }
            status = drvChangeChannel(pdevInfo, pscanInfo->pScanList->pChanArray[pscanInfo->curChannelIndex],
                rxFilter);
            if (!(pscanInfo->pScanList->pChanArray[pscanInfo->curChannelIndex]->channelFlags & CHANNEL_CW_INT)) {
                break;
            }
            pscanInfo->curChannelIndex++;
        }
    }

    if (pscanInfo->curChannelIndex >= pscanInfo->pScanList->listSize) {
        pdevInfo->localSta->transState = TSTATE_QUIET;
        pdevInfo->smeTimer.operation = MLME_INVALID_TIMEOUT;
        wlanMlmeScanConfirm(pdevInfo,&pdevInfo->pscanInfo->bssSet, MLME_SUCCESS);
        return;
    }

    /*
     * Change state to scanning only after changing the channel, otherwise
     * some beacons from old channel may get recorded as on new channel
     */
    pdevInfo->localSta->transState |= TSTATE_SCANNING;
    pscanInfo->scanStartTime = A_MS_TICKGET(); // save the start time of scan

    /* Now we get into the scanning process */

    if (!(pdevInfo->staConfig.phwChannel->channelFlags & CHANNEL_PASSIVE)) {
        /* We are not sending a delayed probe request anymore, but sending two
         * probe requests if ssid is non-null.
         */
        /* Send out the probe request packet */
        if (wlanMlmeProbeRequest(pdevInfo,
                     &broadcastBssid,
                     &pscanInfo->ssid,
                     pscanInfo->probeDelay) != MLME_SUCCESS )
        {
            /* Something wrong with probe transmission. Skip */
            uiPrintf("wlanScanFunction: ERROR: failed probe request on channel %d\n",
                 pscanInfo->pScanList->pChanArray[pscanInfo->curChannelIndex]->channel);
        }

#if 0
        if (pscanInfo->probeRetry) {
            timeOut -= pscanInfo->minChannelTime;
        } else {
            timeOut = pscanInfo->minChannelTime;  // 50 ms - to get probe resp back if not send another at timerhandler
        }
#endif
        if (pscanInfo->ssid.length > 0) {
            /* Previous probe had actual ssid, send a null ssid now */
            if (wlanMlmeProbeRequest(pdevInfo,
                     &broadcastBssid,
                     &dummySsid,
                     pscanInfo->probeDelay) != MLME_SUCCESS ) {
                /* Something wrong with probe transmission. Skip */
                uiPrintf("mlmeScanFunction: ERROR: failed probe request on channel %x\n",
                     pscanInfo->pScanList->pChanArray[pscanInfo->curChannelIndex]->channel);
            }
        }
    }
    A_TIMEOUT(&pdevInfo->smeTimer, timeOut, (A_UINT32) pdevInfo, MLME_SCAN_TIMEOUT);
}

/*
 * wlanMlmeScanRequest - scan for potential BSSs.
 * MLME SCAN primitive
 *
 * Spec section 10.3.2.1
 *
 * PARAMETERS:
 *  bsstype        - INFRASTRUCURE_BSS or INDEPENDENT_BSS or both. It will accept
 *                     beacons from either or both.
 *  bssid          - A specific BSSID to scan for or broadcast MAC address for
 *                   accepting all of them.
 *  ssidStr        - Name of Service set. 32 bytes max. NULL value means accept any
 *                     SSID.
 *  scanType       - ACTIVE_SCAN or PASSIVE_SCAN
 *  probeDelay     - MAC will observe a specific algorithm before jumping in on
 *                    the WM. probeDelay is the maximum time that it will wait.
 *                    This parameter is in microseconds.
 *  channelList    - List of channels to scan.
 *  minChannelTime - Listen to each channel for at least this long.
 *  maxChannelTime - TU. Must switch to next channel after this time.
 *  quickScan      - Stop scanning after receiving one valid beacon or probe
 *                     response.
 *
 * RETURNS: MLME_SUCCESS or MLME_INVALID_PARAM
 */

MLME_RESULT
wlanMlmeScanRequest(WLAN_DEV_INFO     *pdevInfo,
                    BSSTYPE           bsstype,        /* Infra., indep. or any */
                    BSSID             *bssid,         /* Specific MAC or broadcast addr */
                    SSID              *ssid,          /* ESS identifier or NULL */
                    SCANTYPE          scantype,       /* active or passive */
                    A_UINT32          probeDelay,     /* max probe delay time in usec */
                    WLAN_SCAN_LIST    *pScanList,     /* array of chennel ptrs to scan */
                    A_INT32           minChannelTime, /* minimum waiting time */
                    A_INT32           maxChannelTime, /* max waiting time */
                    A_BOOL            quickScan)      /* if TRUE, stop scanning after */
                                                      /* first valid response */
{
    SCAN_INFO *pscanInfo;       /* place to hold scan information */

    ASSERT(pScanList);

#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("wlanMlmeScanRequest\n");
    }
#endif  /* DEBUG */

    if (pdevInfo->localSta->transState & TSTATE_SCANNING) {
        /* Already in scanning mode. Reject this request */
        return MLME_TOOMANY_REQ;
    }

    /* We need to hold the parameters in the form of expected
     * BSS Description. Allocate space and store it on the STA
     */

    if ( pdevInfo->pscanInfo == NULL ) {
        pdevInfo->pscanInfo = (SCAN_INFO *) A_DRIVER_MALLOC(sizeof(SCAN_INFO), 0, 0);
        if (!(pdevInfo->pscanInfo)) {
            uiPrintf("wlanMlmeScanRequest: Failed to allocate memory for the Scan Info\n");
            return MLME_REFUSED;
        }
        /* Only clear the channel list at allocation */
    } else {
        /* Free scanInfo scanList */
        wlanFreeScanList(pdevInfo->pscanInfo->pScanList);
        pdevInfo->pscanInfo->pScanList = NULL;
    }

    pscanInfo = pdevInfo->pscanInfo;
    pscanInfo->bssSet.count = 0; /* 0 will mark it empty */
    pscanInfo->bsstype = bsstype;
    pscanInfo->scantype = scantype;
    pscanInfo->maxChannelTime = maxChannelTime;
    pscanInfo->minChannelTime = minChannelTime;
    pscanInfo->probeRetry = 0;
    pscanInfo->pScanList = pScanList;
    ASSERT(pScanList->listSize <= pdevInfo->staConfig.pClist->listSize);
    pscanInfo->curChannelIndex = 0;

    A_MEM_ZERO((char *)&pscanInfo->bssSet.bssDescrArray, sizeof(pscanInfo->bssSet.bssDescrArray));

    A_MACADDR_COPY((WLAN_MACADDR *) bssid, &pscanInfo->bssid);
    if (ssid != NULL) {
        (void) mlmeElementCopy((INFO_ELEMENT *)ssid,
                               (INFO_ELEMENT *)&pscanInfo->ssid);
    } else {
        pscanInfo->ssid.elementID = ELE_SSID;
        pscanInfo->ssid.length    = 0;
    }

    /* Since we want the scan to be done in the background and
     * not have the calling program to wait for it to finish, the
     * scanning process must be done in a separate thread.
     */

    // Since we are changing the channel, deactivate the logical link
    pdevInfo->linkIsActive = FALSE;

    // Kick off the scan for first channel
    mlmeScanFunction(pdevInfo);
    return MLME_SUCCESS;
}

/*
 * wlanMlmeScanConfirm - returns result of scan
 *
 * It calls back scanCallback() routine to return the set of descriptions
 * for the BSS that were found on the WM.
 *
 * RETURNS: MLLE_SUCCESS, MLME_INVALID_PARAM
 *             pbssDescrSet points to the set of BSS descriptions found.
 *
 */

void wlanMlmeScanConfirm(WLAN_DEV_INFO *pdevInfo,
                         BSSDESCR_SET  *pbssDescrSet,   /* pointer to descriptor set */
                         MLME_RESULT   result)
{
    /* Turn off beacon delivery to the driver */
    halRxFilter(pdevInfo, HAL_RX_BEACON, HAL_RX_FILTER_CLEAR);

    pdevInfo->localSta->transState = TSTATE_QUIET;
#ifdef  DEBUG
    if (mlmeDebug) {
        uiPrintf("mlmeScanConfirm: Returning to channel %d\n", pdevInfo->staConfig.pChannel->channel);
    }
#endif   /* DEBUG */

    if (pdevInfo->staConfig.phwChannel != pdevInfo->staConfig.pChannel) {
        // Switch back to the original channel again.
        drvChangeChannel(pdevInfo, pdevInfo->staConfig.pChannel, pdevInfo->rxFilterReg);
    }

    // Turn on the data traffic again
    pdevInfo->linkIsActive = FALSE;

    /* If callback routine exists, call it */
    if (smeInfo.scanCallback) {
        (*smeInfo.scanCallback)(pdevInfo, pbssDescrSet, result);
    }
    return;
}


/* wlanMlmeProbeRequest
 *
 * This routine sends out a probe packet. Not required
 * by MLME, but used internally by algorithms.
 * This routine can be called for active scan even before we have a valid
 * ssid or bssid in the SIB of local station. Hence, those parameters are passed in.
 */

MLME_RESULT
wlanMlmeProbeRequest(WLAN_DEV_INFO *pdevInfo,
                     WLAN_MACADDR  *bssid,
                     SSID          *ssid,
                     A_UINT32      probeDelay)
{
    WLAN_FRAME_PROBE_REQ *pFrame;
    SIB_ENTRY    *pSib;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    INFO_ELEMENT *p1;           /* Temporary pointer */

#ifdef  DEBUG
    if (probeDebug || mlmeDebug) {
        uiPrintf("wlanMlmeProbeRequest -- channel %d\n",
                 pdevInfo->staConfig.phwChannel->channel);
    }
#endif   /* DEBUG */

    pSib = pdevInfo->localSta;
    /* Compute frame length. No challenge text */
    /* Don't use INFO_ELEMENT_SIZE macro for size of ssid, it can be null */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) +
        (ssid->length + sizeof(ssid->elementID) + sizeof(ssid->length)) +
        INFO_ELEMENT_SIZE(pdevInfo->localSta->operationalRateSet);

    if ((createBuffandDesc(pdevInfo, frameLen, (void **) &pDesc)) != 0) {
        /* Failed to allocate buffer for frame and descriptor */
        return MLME_DRIVER_ERROR;
    }
    pFrame = pDesc->pBufferVirtPtr.probe;
    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    &broadcastBssid, SUBT_PROBE_REQ);
    A_MACADDR_COPY(&broadcastBssid, &pFrame->macHeader.bssId);

    p1 = mlmeElementCopy((INFO_ELEMENT *)ssid, (INFO_ELEMENT *)&pFrame->ssid);
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->localSta->operationalRateSet, p1);

    /* displayMgtFrame(&pFrame->macHeader, frameLen); */
    pDesc->pDestSibEntry = pSib;
    if (powerFrameTransmit(pdevInfo, pDesc) == A_OK) {
        return MLME_SUCCESS;
    }
    return MLME_DRIVER_ERROR;
}


void
mlmeProbeRespReceive(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc)
{
    WLAN_FRAME_PROBE_RESP *bFrame;      /* Incoming frame */
    BSSDESCR     tmpBssDescr;
    BSSDESCR     *pbssDescr = &tmpBssDescr;        /* BSS Description set */
    INFO_ELEMENT *p1;
    SSID         *pSsid;
    A_UINT16     len;                  /* Total frame length */
    DS_PARAM_SET *pDsparams;


    /* Currently we process probe responses only while scanning. */
    if (!(pdevInfo->localSta->transState & TSTATE_SCANNING)) {
        // not scanning, ignore this probe resp
        return;
    }

    bFrame = pDesc->pBufferVirtPtr.probeResp;
    len    = (A_UINT16)pDesc->hwStatus.rx.dataLength;
    pDsparams = (DS_PARAM_SET *)mgtElementGet(&bFrame->macHeader, len, ELE_DS_PARAM_SET);

#ifdef  DEBUG
    if (probeDebug || mlmeDebug) {
        uiPrintf("mlmeProbeRespReceive -- ");
        printMacAddress (bFrame->macHeader.bssId);
        uiPrintf(" channel %d\n", pdevInfo->staConfig.phwChannel->channel);
        uiPrintf("\n");
    }
#endif  /* DEBUG */

    /* We are in the middle of scanning process */
    /* See if we have already received beacon from this AP */
    if (pdevInfo->pscanInfo == NULL) {
        /* something wrong. We are scanning but set is not allocated */
        uiPrintf("mlmeProbeRespReceive: ERROR: scanning but no set allocated\n");
        return;
    }

    if ((pDsparams != NULL) &&
        (pDsparams->currentChannel != wlanConvertGHztoCh(pdevInfo->staConfig.phwChannel->channel,
                                        pdevInfo->staConfig.phwChannel->channelFlags)))
    {
        // Picked up this beacon on a wrong channel. Can happen in 11b. Beacons can be
        // heard anywhere in the 25 MHz vicinity. Ignore it. Wait for true channel.
        return;
    }

    if (pdevInfo->pscanInfo->bssSet.count == BSSDESCR_SET_SIZE) {
        /* We maxed out on scanned beacons. Punt. */
        /* I don't imagine a situation where one will
         * get beacons from like 32 different access points.
         */
        return;
    }
    /* Assume that space for entire set is already allocated
     * by whoever initiated the scan.
     */

    pSsid = (SSID *) mgtElementGet(&bFrame->macHeader, len, ELE_SSID);
    if (pSsid == NULL) {
        return;
    }

    pbssDescr->beaconInterval = bFrame->beaconInterval;
    pbssDescr->bsstype = (bFrame->capInfo.ibss == 1) ?
        INDEPENDENT_BSS : INFRASTRUCTURE_BSS;
    /* punt if this is not the bss type that we're scanning for.. */
    if ((pdevInfo->pscanInfo->bsstype != ANY_BSS)
        && (pbssDescr->bsstype != pdevInfo->pscanInfo->bsstype)) {
        return;
    }
    A_MACADDR_COPY(&bFrame->macHeader.bssId, &pbssDescr->bssid);

    pbssDescr->pChannel = pdevInfo->staConfig.phwChannel;
    // rssi can be -ve. Treat 8 bit value as signed. convert to abs
    pbssDescr->rssi = (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength) < 0 ?
        - (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength) :
        (A_INT32)((A_INT8)pDesc->hwStatus.rx.rxSigStrength);
    pbssDescr->timestamp = bFrame->timestamp;
    pbssDescr->capabilityInfo = bFrame->capInfo;
    p1 = (INFO_ELEMENT *)&bFrame->buffer;
    p1 = mlmeElementCopy((INFO_ELEMENT *) pSsid,
                         (INFO_ELEMENT *)&pbssDescr->ssid);
    p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                       len, ELE_SUPPORTED_RATES),
                         (INFO_ELEMENT *)&pbssDescr->bssRateSet);
    /* make the element invalid first */
    /* need remove this when the mlmeElementCopy changed */
    pbssDescr->bssUcseSet.length = 0;
    p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                       len, ELE_UCSE_SET),
                         (INFO_ELEMENT *)&pbssDescr->bssUcseSet);
    p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                       len, ELE_CF_PARAM_SET),
                         (INFO_ELEMENT *)&pbssDescr->cfParamSet);
    p1 = mlmeElementCopy(mgtElementGet(&bFrame->macHeader,
                                       len, ELE_IBSS_PARAM_SET),
                         (INFO_ELEMENT *)&pbssDescr->ibssParamSet);
    pbssDescr->pChannel = pdevInfo->staConfig.phwChannel;
    pbssDescr->scanTime = A_MS_TICKGET(); // save the time of reception
    // add or replace the existing entry
    mlmeBsssetAdd(pbssDescr, &pdevInfo->pscanInfo->bssSet);

#ifndef BUILD_AP
    // check for expiration of minChannelTime
    if ((A_MS_TICKGET() - pdevInfo->pscanInfo->scanStartTime) >=
        pdevInfo->pscanInfo->minChannelTime) {
        // We are done with this channel

        A_UNTIMEOUT(&(pdevInfo->smeTimer));
        pdevInfo->pscanInfo->probeRetry = 0;
        if ((pdevInfo->pscanInfo->curChannelIndex + 1) >=
            pdevInfo->pscanInfo->pScanList->listSize)
        {
            // We finished last channel on the list
            // Clear the state. Don't let any more bss's creep in
            pdevInfo->localSta->transState = TSTATE_QUIET;
            wlanMlmeScanConfirm(pdevInfo,
                                &pdevInfo->pscanInfo->bssSet, MLME_SUCCESS);
        } else {
            pdevInfo->pscanInfo->curChannelIndex++;
            // Move over to the next channel, curChannelIndex is
            // already incremented.
            mlmeScanFunction(pdevInfo);
        }
    }
#endif /* BUILD_AP */
}

/* wlanMlmeStartRequest - Start the operation of an access point
 *
 * Tells the MAC entity to operate as a BSS (Access point)
 * It sets various BSS related parameters into the MAC, enables beacon
 * generation, DTIM generation, enable power save buffering, etc.
 *
 * This will be implemented as a synchronous operation.
 *
 * RETURNS:    MLME_SUCCESS, MLME_INVALID_PARAM, MLME_ALREADY_IN_BSS
 */

MLME_RESULT
wlanMlmeStartRequest(WLAN_DEV_INFO *pdevInfo, BSSDESCR *pbssDescr)
{
    int i;
    A_BOOL  faulted;

    if (!pdevInfo->staConfig.privacyInvoked && pbssDescr->capabilityInfo.privacy) {
        // IBSS is trying to use WEP, but we don't have it enabled.
        // cannot start this SSID.
        return MLME_INVALID_PARAM;
    }

    /* ideally, we shouldn't get anything which says ANY bss */
    if (pbssDescr->bsstype == ANY_BSS) {
        pbssDescr->bsstype = INDEPENDENT_BSS;
        pbssDescr->capabilityInfo.ess  = 0;
        pbssDescr->capabilityInfo.ibss = 1;
    }

    /* Copy BSS description into pdevInfo of the STA.
     */
    pdevInfo->bssDescr = (BSSDESCR *)A_DRIVER_MALLOC(sizeof(BSSDESCR),0,0);
    if (pdevInfo->bssDescr == NULL) {
        return MLME_DRIVER_ERROR;
    }

    A_DRIVER_BCOPY(pbssDescr, pdevInfo->bssDescr, sizeof(BSSDESCR));

    /* Clear all bits in association ID bit array */
    /* Association ID 0 is illegal. We mark it as allocated for ever */
    assocArray[0] = 0x1;
    for (i = 1; i < (ASSOCID_MAX+7)/8; i++) {
        assocArray[i] =0;
    }

    /* At this point, our MAC address becomes the BSSID */
    A_MACADDR_COPY(&pdevInfo->localSta->macAddr,
                   &pdevInfo->localSta->bssId);
    /* SSID from BSS description becomes our SSID */
    (void) mlmeElementCopy((INFO_ELEMENT *)&pbssDescr->ssid,
                           (INFO_ELEMENT *)&pdevInfo->staConfig.ssid);
    pdevInfo->bssDescr->dsParamSet.elementID = ELE_DS_PARAM_SET;
    pdevInfo->bssDescr->dsParamSet.length = 1;
    pdevInfo->bssDescr->dsParamSet.currentChannel =
        wlanConvertGHztoCh(pdevInfo->bssDescr->pChannel->channel, pdevInfo->bssDescr->pChannel->channelFlags);


    if (pbssDescr->bsstype == INFRASTRUCTURE_BSS) {
        pdevInfo->localSta->serviceType = WLAN_AP_SERVICE;
    }
    // enter the new channel into hw config.
    pdevInfo->staConfig.pChannel = pdevInfo->bssDescr->pChannel;

#ifdef BUILD_AP
    /* Call the Access Point routine that possibly goes thru hardware reset,
     * enables beacon transmission, sets correct channel
     */
#else
    if (pbssDescr->bsstype == INFRASTRUCTURE_BSS) {
        pdevInfo->localSta->staState |= STATE_AP_UP;
    }

    // move to the desired channel
    if (drvChangeChannel(pdevInfo, pbssDescr->pChannel,
        (pbssDescr->bsstype == INDEPENDENT_BSS) ? (HAL_RX_UCAST|
                                                   HAL_RX_MCAST|HAL_RX_BCAST|
                                                   HAL_RX_BEACON)
                                                : (HAL_RX_UCAST|HAL_RX_MCAST|
                                                   HAL_RX_BCAST)) != A_OK)
    {
        return MLME_INVALID_PARAM;
    }
#endif
    if (pbssDescr->bsstype == INDEPENDENT_BSS) {
        pdevInfo->localSta->staState |= STATE_JOINED;
        pdevInfo->localSta->transState = TSTATE_QUIET;

        pdevInfo->localSta->capInfo.ess  = pbssDescr->capabilityInfo.ess;
        pdevInfo->localSta->capInfo.ibss = pbssDescr->capabilityInfo.ibss;
        pdevInfo->localSta->capInfo.shortPreamble = pbssDescr->capabilityInfo.shortPreamble;
        // BSSID and SSID can change here
        A_MACADDR_COPY(&pbssDescr->bssid, &pdevInfo->staConfig.bssId);
        A_MACADDR_COPY(&pbssDescr->bssid, &pdevInfo->localSta->bssId);
        mlmeElementCopy((INFO_ELEMENT *)&pbssDescr->ssid,
                        (INFO_ELEMENT *)&pdevInfo->staConfig.ssid);

        pdevInfo->localSta->capInfo.privacy = pbssDescr->capabilityInfo.privacy;

        /* Don't change cfPoll bits for now. */

        /* Call the routine below just to set the BSS */
        halWriteAssocid(pdevInfo, &(pbssDescr->bssid), 0, 0);

        /* Set up privacy, if any */
        if (pdevInfo->localSta->privacyInvoked) {

            initCipher(pdevInfo);

            wlanKeyEnable(pdevInfo, pdevInfo->localSta, pdevInfo->staConfig.defaultKey);
            wlanKeyCacheFault(pdevInfo, pdevInfo->localSta, &faulted);

            if (wlanKeyUpdate(pdevInfo, pdevInfo->localSta, CONTEXT_NONE) != A_OK) {
                return MLME_REFUSED;
            } else {
                /* set up encryption */
                setupSwEncryption(pdevInfo);
            }
        }

        halSetAdhocMode(pdevInfo);

        /* Start generating beacons for adhoc mode */
        // TODO: ensure that beacons start going out with this call.
        // beaconInit may need to be called with/without enabling beacons
        wlanBeaconInit(pdevInfo, 0);
    }
    /* for infrastructure bss mode the AP beacon generation is
     * turned on in the station init code itself! it could ideally
     * be done here, but am not sure if mlmeStart gets called for
     * that case at all!
     */

    wlanMlmeStartConfirm(pdevInfo, MLME_SUCCESS);
    return MLME_SUCCESS;
}

/* wlanMlmeStartConfirm - Confirm the start of BSS
 * Since the request is implemented as a synchronous operation,
 * this is an empty routine.
 */

MLME_RESULT
wlanMlmeStartConfirm(WLAN_DEV_INFO *pdevInfo, MLME_RESULT resultCode)
{
    /* If callback routine exists, call it */
    if (smeInfo.startCallback) {
        (*smeInfo.startCallback)(pdevInfo, resultCode);
    }
    return resultCode;
}

#ifndef BUILD_AP

void
mlmePowerCallback(WLAN_DEV_INFO *pdevInfo, A_STATUS resultCode,
                  A_BOOL powerStatus)
{
    // got a callback from power mgt code. Cancel the timer.
    A_UNTIMEOUT(&(pdevInfo->smeTimer));
    wlanMlmePowermgtConfirm(pdevInfo, resultCode, powerStatus);
}

/*
 * wlanMlmePwrmgtRequest - Change PS mode of STA
 * MLME Power management primitive changes the curent power saving mode.
 * This call will change the setting and return. MLME module will
 * send a null frame to AP to indicate the state transition or use an
 * existing scheduled transmission. When going into PS mode, the packets
 * already scheduled to transmit will be transmitted before switching the
 * mode. If the transmission attempts by MAC hardware fail, then software will
 * suspend those packets instead of going into PS mode. When going into
 * active mode, if wakeup is TRUE, MAC is immediately
 * forced into active mode instead of waiting for beacons.
 *
 * RETURNS: MLME_SUCCESS, MLME_NOT_SUPPORTED or MLME_INVALID_PARAM
 */

MLME_RESULT
wlanMlmePwrmgtRequest(WLAN_DEV_INFO *pdevInfo,
                      A_UINT16      powermgtMode, /* PS_MODE_ACTIVE,PS_MODE_POWER_SAVE */
                      A_BOOL        wakeup,       /* TRUE for immediate active mode */
                      A_BOOL receiveDTIMs)        /* TRUE if STA should wake up for */
                                                  /* each DTIM */
{
    A_STATUS    pwStatus;

    if (pdevInfo->localSta->transState != TSTATE_QUIET) {
        // if tstate is not 0, mlme operation is in progress, cannot use
        // sme timer, cannot change power mode
        return MLME_REFUSED;
    }
    // sanity check
    switch (powermgtMode) {
    case PS_MODE_ACTIVE:
        // CALL NDIS driver power save routine to wake up
        pwStatus = powerSetMode(pdevInfo,
                                (A_UINT16)((wakeup) ? FORCE_AWAKE : NORMAL_AWAKE),
                                mlmePowerCallback);
        break;

    case PS_MODE_POWER_SAVE:
        // CALL NDIS driver power save routine to go to sleep
        // according to 802.11 spec, wakeup flag in POWER_SAVE mode has no
        // meaning. We will use that to signify PRETEND_SLEEP mode. In a way,
        // we are telling driver to do sleep semantics but wakeup immediately.
        if (wakeup) {
            // special case to take us to PRETEND_SLEEP mode
            pwStatus = powerSetMode(pdevInfo, BLOCK_SLEEP, mlmePowerCallback);
        } else {
            pwStatus = powerSetMode(pdevInfo, NORMAL_SLEEP, mlmePowerCallback);
        }
        break;
    }

    if (pwStatus == A_PENDING) {
        // power mgt frame is in the queue, hence it returned A_PENDING, set a timer
        // hardcoded value of 50 ms
        // mark the state as busy, callback will clear it.
        pdevInfo->localSta->transState = TSTATE_PWRMGT_PENDING;
        A_TIMEOUT(&pdevInfo->smeTimer, 50, (A_UINT32) pdevInfo,
                  MLME_PWRMGT_TIMEOUT);
        return MLME_SUCCESS;
    }
    return (pwStatus == A_OK )? MLME_SUCCESS : MLME_NOT_SUPPORTED;
}

/* wlanMlmePowermgtConfirm - Confirm the powermgt operation's result
 *
 * This routine is used to call pwrmgtCallback() to inform the operation
 * result. One possible reason for failure is that MAC fails to transmit
 * successfully a packet with PwrMgt bit set.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM
 */

void
wlanMlmePowermgtConfirm(WLAN_DEV_INFO *pdevInfo,
                        MLME_RESULT   result,
                        A_BOOL        powerStatus)
{
    pdevInfo->localSta->transState = TSTATE_QUIET;
    /* If callback routine exists, call it */
    if (smeInfo.pwrMgtCallback) {
        (*smeInfo.pwrMgtCallback)(pdevInfo, result, powerStatus);
    }
    return;
}

#endif /* BUILD_AP */

// Following routines are for test and debugging of SME, primarily for AP.
#ifdef DEBUG

extern SIB_ENTRY *sibAddr (int idx);

void
deauth(A_UINT16 idx, A_UINT16 reason)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;

    pdevInfo = globDrvInfo.pDevInfoArray[0];

    pSib = sibAddr(idx);
    if (!reason) {
        reason = 3;
    }
    pdevInfo->localSta->stats.DeAuthentications++;
    wlanMlmeDeauthRequest(pdevInfo, &pSib->macAddr, reason);
}


void
disassoc(A_UINT16 idx, A_UINT16 reason)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;

    pdevInfo = globDrvInfo.pDevInfoArray[0];

    pSib = sibAddr (idx);
    if (!reason) {
        reason = 5;
    }
    pdevInfo->localSta->stats.DisAssociations++;
    wlanMlmeDisassocRequest(pdevInfo, &pSib->macAddr, reason);
}


void
smeTestDisassoc(int staId)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;

    pdevInfo = globDrvInfo.pDevInfoArray[0];

    if (mlmeDebug) {
        uiPrintf("sta %d\n", staId);
    }
    pSib = sibTblFind (staId);
    if (!pSib) {
        uiPrintf("smeTestDisassoc -- can't find STA %d\n", staId);
        return;
    }
    pdevInfo->localSta->stats.DisAssociations++;
    wlanMlmeDisassocRequest(pdevInfo, &pSib->macAddr, 5);
}

void
smeTestDeauth(int staId)
{
    WLAN_DEV_INFO *pdevInfo;
    SIB_ENTRY     *pSib;

    pdevInfo = globDrvInfo.pDevInfoArray[0];

    pSib = sibTblFind (staId);
    if (!pSib) {
        uiPrintf("smeTestDeauth -- can't find STA %d\n", staId);
        return;
    }
    if (mlmeDebug) {
        uiPrintf("sta %d\n", staId);
    }
    pdevInfo->localSta->stats.DeAuthentications++;
    wlanMlmeDeauthRequest(pdevInfo, &pSib->macAddr, 3);
}
#endif /* DEBUG */

void
mlmeOperationCancel(WLAN_DEV_INFO *pdevInfo)
{
    // for now, just change the state. that will take care of failing
    // the operation when timeout routines are called.
    // We should really cancel the smetimer and call callback routines
    // to fail the operation, etc. It's little complex. Later.

#ifndef BUILD_AP
    if (A_UNTIMEOUT(&(pdevInfo->smeTimer))) {
        // timer operation was queued but we cancelled the timer.
        // hence call the function so that it will behave exactly like
        // an operation timeout that results in clearing the state
        // and calling confirm routines with timeout error codes
        // move scan channel index to the last channel so that cancelling
        // scan operation on this channel will bail out of scan request completely
        pdevInfo->pscanInfo->curChannelIndex =
            pdevInfo->pscanInfo->pScanList->listSize - 1;
        smeTimerHandler(NULL, &pdevInfo, NULL, NULL);
    }
#endif
    pdevInfo->localSta->transState = TSTATE_QUIET;    /* cancel state */
}

/* wlanResetRequest - Reset the MAC entity
 *
 * This call resets the MAC hardware and software structures that
 * are used to handle that specific MAC. There will be other routines
 * and commands to reset the hardware, driver, etc. but not done as
 * part of MLME operation.
 *
 * PARAMETERS:
 *    pLocalSTAAddress - MAC address to be set after the reset.
 *    setDefaultMIB     - If TRUE, set the default values into the MIB.
 *
 * RETURNS: MLME_SUCCESS, WLAN_UNKNOWN_FAILURE
 */

MLME_RESULT
wlanResetRequest(WLAN_DEV_INFO *pdevInfo,
                 WLAN_MACADDR  *pLocalSTAAddress,
                 A_BOOL        setDefaultMIB)
{
    // do a systematic mac shutdown.
#ifndef BUILD_AP
    // currently implemented only on station
    // first tell the AP that we are disconnecting
    if (pdevInfo->localSta->staState & STATE_ASSOC) {
        wlanMlmeDisassocRequest(pdevInfo, &pdevInfo->localSta->bssId,
                                REASON_ASSOC_LEAVING);
    }

    if (pdevInfo->localSta->staState & STATE_AUTH) {
        wlanMlmeDeauthRequest(pdevInfo, &pdevInfo->localSta->bssId,
                              REASON_AUTH_LEAVING);
    }
    pdevInfo->localSta->staState  &= ~(STATE_CONNECTED);
    pdevInfo->localSta->transState = 0;
    ResetOnError(pdevInfo, RESET_ALL_HW, TRUE);
    if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS) {
        // whether we are joined or not, free the adhoc beacon structures.
        wlanBeaconFree(pdevInfo);
    }
    wlanResetConfirm(pdevInfo);
#endif

    return MLME_SUCCESS;
}

/* wlanResetConfirm - Confirm the reset of MAC
 * This routine calls reseCallback() to deliver the Success/fail
 * result of reset operation.
 */

MLME_RESULT
wlanResetConfirm(WLAN_DEV_INFO *pdevInfo)
{
    /* If callback routine exists, call it */
    if (smeInfo.resetCallback) {
        (*smeInfo.resetCallback)(pdevInfo);
    }
    return MLME_SUCCESS;
}

/* clean up all cipher negotiation related stuff */
void
sibCseStateClean(SIB_ENTRY *pSib)
{
    if (pSib->pPriv != NULL) {
        if (pSib->pPriv->keyType == PRIV_KEY_TYPE_AES) {
            if (pSib->pPriv->initValue != NULL) {

                /* this is a placeholder for now to cleanup the software
                 * encrypt (aes) data structures when a link is no longer in use.
                 * a more appropriate place may be when you deauthenticate but
                 * that code may need to be reworked when 802.1X is implemented.
                 */
                ocb_done(pSib->pPriv->initValue);
                pSib->pPriv->initValue = NULL;
            }
            pSib->pPriv->keyType = PRIV_KEY_TYPE_WEP;
        }
    }

    A_MEM_ZERO(&pSib->workingUcseSet, (pSib->workingUcseSet.length + 2));
    A_MEM_ZERO(&pSib->mcseSet, (pSib->mcseSet.length + 2));

    /* clear sw encryption enable flag */
    pSib->swEncryptionEnabled = FALSE; 
}

/* clean up all power saving related stuff */
void
sibPowerStateClean(SIB_ENTRY *pSib)
{
// currently, this is used only on AP.
#ifdef BUILD_AP
    QUEUE_DETAILS *pQueue = NULL;
    WLAN_DEV_INFO *pdevInfo;

    if (pSib == NULL) {
        return;
    }

    pdevInfo = pSibTable->pdevInfo;

    /* update the power save state */
    if (pSib->psMode == PS_MODE_POWER_SAVE) {
        pSib->psMode = 0;
        pSib->pDevInfo->numSleepingClients--;
    }

    /* flush the power save queue */
    pQueue = &pSib->txPSQueue;
    A_SEM_LOCK(pQueue->qSem, WAIT_FOREVER);
    if (pQueue->pDescQueueHead != NULL) {
        freeBuffandDescChain(pdevInfo, pQueue->pDescQueueHead);
        pQueue->pDescQueueHead = NULL;
        pQueue->pDescQueueTail = NULL;
        pQueue->qFrameCount = 0;
    }

    A_SEM_UNLOCK(pQueue->qSem);

    /* update the TIM if needed */
    if (pSib->numPSTxPending == 0) {
        wlanTimUpdate(pdevInfo, pSib->assocId, FALSE);
    }

    return;
#endif
}

#endif // FADI

/*************************/
/* MLME support routines */
/*************************/

/* mgtFrameHdrFill - Fill management packet header fields
 *
 * DESCRIPTION
 * Copies information from device information, global info, etc. into
 * fields of management packet header.
 */

void
mgtFrameHdrFill(WLAN_DEV_INFO *pdevInfo, WLAN_MGT_MAC_HEADER *pFrame,
                WLAN_MACADDR  *destAddr, int subtype)
{
    /* Fill in Frame Control fields */
    /* All management frames have FromDS */
    /* and ToDS as 0                     */
    /* Most management frames not encrypted */
    A_MEM_ZERO((char *)&pFrame->frameControl, sizeof(pFrame->frameControl));  /* default initial value for all fields */
    pFrame->frameControl.fType    = FRAME_MGT;
    pFrame->frameControl.fSubtype = (A_UINT16) subtype;
    pFrame->frameControl.protoVer = PROTOCOL_VER_0;

    /* clear duration/ID field */
    pFrame->durationNav.clearBit = 0;
    pFrame->durationNav.duration = 0;

    /* BSSID */
    /* If it's a probe request, calling program should replace the value
     * of bssId anyway.
     */
    A_MACADDR_COPY(&pdevInfo->localSta->bssId, &pFrame->bssId);

    /* Source address */
    A_MACADDR_COPY(&pdevInfo->localSta->macAddr, &pFrame->srcAddr);

    /* Destination address */
    A_MACADDR_COPY(destAddr, &pFrame->destAddr);

    /* duration, seq control fields are not initialized here */
    pFrame->sequenceControl.fragNo = 0;
}

A_UINT16 mlmeCountryCodeElementLength (WLAN_DEV_INFO *pDevInfo) 
{
    if (pDevInfo->localSta->serviceType == WLAN_AP_SERVICE) {
        if (pDevInfo->staConfig.multiDomainCapEnabled) {
            COUNTRY_INFO_LIST       CountryInfo;
            COUNTRY_INFO_SET        *pInfoElement;

            wlanGetCountryInfoList(pDevInfo, &CountryInfo);
            return ((CountryInfo.len + sizeof(pInfoElement->elementID)) + 1) & ~0x1;
        }
    }
    return 0;
}

A_CHAR *mlmeCountryCodeElementFill (WLAN_DEV_INFO *pDevInfo, A_CHAR *pField) 
{
    if (pDevInfo->localSta->serviceType == WLAN_AP_SERVICE) {
        if (pDevInfo->staConfig.multiDomainCapEnabled) {
            COUNTRY_INFO_LIST   CountryInfo;
            COUNTRY_INFO_SET    *pInfoElement = (COUNTRY_INFO_SET *)pField;
            int                 elmsize;

            wlanGetCountryInfoList(pDevInfo, &CountryInfo);
            elmsize = mlmeCountryCodeElementLength (pDevInfo);
            pInfoElement->elementID = ELE_COUNTRY_INFO;
            A_DRIVER_BCOPY(&CountryInfo, (char *)&pInfoElement->length, elmsize);
            pField = (A_CHAR *)(((A_UINT8 *)pField) + elmsize);
        }
    }
    return pField;
}

INFO_ELEMENT *
mlmeElementCopy(INFO_ELEMENT *pSrc, INFO_ELEMENT *pDest)
{
    int isize;

    if (pSrc == NULL || pDest == NULL) {
        return NULL;
    }

    isize = pSrc->length + sizeof(pSrc->elementID) + sizeof(pSrc->length);

    A_DRIVER_BCOPY(pSrc, pDest, isize);
    return (INFO_ELEMENT *)(((A_UINT8 *)pDest) + isize);
}

/* wlanMlmeProbeRequestReceive
 *
 * This routine receives a probe request on AP. Not required
 * by MLME, but used internally by algorithms.
 */

void
mlmeProbeRequestReceive(WLAN_DEV_INFO        *pdevInfo,
                        WLAN_FRAME_PROBE_REQ *probeReqFrame,
                        A_UINT16             len)           /* Total frame length */
{
    WLAN_FRAME_PROBE_RESP *pFrame;
    A_UINT16     frameLen;
    ATHEROS_DESC *pDesc;
    SSID         *pSsid = NULL;
    INFO_ELEMENT *p1;           /* temporary pointer */

    //printk("TX: mlmeProbeRequestReceive...\n");

    // if a reset process is awake and doing it's thing while
    // we woke up, don't dequeue
    if (pwlanDevInfo->NicResetInProgress == TRUE) {
        printk("RX: NICRESETINPROGRESS....will try back later\n");
        halEnableInterrupts(pwlanDevInfo, HAL_INT_RX | HAL_INT_RXORN | HAL_INT_RXEOL);
        printk("RX: RE_ENABLING RX since they were disabled in ISR!!\n");
        return;
    }


    /* Check if our AP function is on */
    //if ((!(pdevInfo->localSta->staState & STATE_AP_UP)) &&
    if  (pdevInfo->bssDescr->bsstype != INFRASTRUCTURE_BSS) {
        printk("FAILED TO RESPOND..we're down in infra_struct mode\n");
        return;
    }

    /* Don't respond to mlmeProbeRequest when off channel */
#ifndef BUILD_AP
    if (pdevInfo->invalidTxChannel) {
        printk("NOT RESPONDING DUE TO INVALIDTxCHANNEL\n");
        return;
    }
#endif

    pSsid = (SSID *) mgtElementGet(&probeReqFrame->macHeader, len, ELE_SSID);
    if ( pSsid != NULL &&
         pSsid->length > 0 &&
         (pdevInfo->bssDescr->ssid.length != pSsid->length ||
          A_BCOMP(&pSsid->ssid, &pdevInfo->bssDescr->ssid.ssid, pSsid->length) != 0))
    {
        // requested SSID is not null but it doesn't match ours
        return;
    }

    /* Compute our beacon length */
    frameLen = sizeof(WLAN_MGT_MAC_HEADER) + sizeof(WLAN_TIMESTAMP)
        + sizeof(pFrame->beaconInterval) + sizeof(CAP_INFO)
        + INFO_ELEMENT_SIZE(pdevInfo->bssDescr->ssid)
        + INFO_ELEMENT_SIZE(pdevInfo->bssDescr->bssRateSet)
        //        + INFO_ELEMENT_SIZE(pdevInfo->bssDescr->fhParamSet)
        + INFO_ELEMENT_SIZE(pdevInfo->bssDescr->dsParamSet)
        + (VALID_CF_ELEMENT(&pdevInfo->bssDescr->cfParamSet)
           ? INFO_ELEMENT_SIZE(pdevInfo->bssDescr->cfParamSet)
           : 0);

    if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS) {
        frameLen += INFO_ELEMENT_SIZE(pdevInfo->bssDescr->ibssParamSet);
    } else {
        frameLen += (VALID_UCSE_ELEMENT(&pdevInfo->bssDescr->bssUcseSet) ?
           INFO_ELEMENT_SIZE(pdevInfo->bssDescr->bssUcseSet) : 0);
    }
    
    frameLen += mlmeCountryCodeElementLength (pdevInfo);
   // printk("TX: our computed len =%d\n", frameLen);

    if ((createBuffandDesc(pdevInfo, frameLen, (ATHEROS_DESC **) &pDesc)) != 0) {
        printk("TX: Failed to allocate buffer for frame and descriptor\n");
        return;
    }
    pFrame = pDesc->pBufferVirtPtr.probeResp;

    mgtFrameHdrFill(pdevInfo, &pFrame->macHeader,
                    &probeReqFrame->macHeader.srcAddr, SUBT_PROBE_RESP);
    
    halGetTsf(pdevInfo, &pFrame->timestamp);
    pFrame->timestamp.low += 100; // XXX approx 100 us adjustment
    if (pFrame->timestamp.low < 100) {
        // lower 32 bits rolled over, increment the higher 32 bit
        pFrame->timestamp.high ++;
    }
    pFrame->beaconInterval = pdevInfo->bssDescr->beaconInterval;
    pFrame->capInfo = pdevInfo->bssDescr->capabilityInfo;
    p1 = (INFO_ELEMENT *) &pFrame->buffer;
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->ssid, p1);
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssRateSet, p1);

//  p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->fhParamSet, p1);
    p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->dsParamSet, p1);
    if (VALID_CF_ELEMENT(&pdevInfo->bssDescr->cfParamSet)) {
        p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->cfParamSet, p1);
    }
    if (pdevInfo->bssDescr->bsstype == INDEPENDENT_BSS) {
        p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->ibssParamSet, p1);
    } else {
        if (VALID_UCSE_ELEMENT(&pdevInfo->bssDescr->bssUcseSet)) {
            p1 = mlmeElementCopy((INFO_ELEMENT *)&pdevInfo->bssDescr->bssUcseSet, p1);
        }
    }

    p1 = (INFO_ELEMENT *)mlmeCountryCodeElementFill (pdevInfo, (A_CHAR *)p1);
    
    pDesc->pDestSibEntry = NULL;
#ifdef DEBUG
    if (probeDebug) {
        displayMgtFrame(&pFrame->macHeader, frameLen);
    }
#endif
    // call the atheros_sap.c FrameTransmit version
    wlanFrameTransmit(pdevInfo, pDesc, FALSE);
    // no one will explicitly cleanup this created pDesc, so we do
    //freeBuffandDesc(pdevInfo, pDesc);

}

