/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/* sysAr5End.c - BSP interface to AR5 END network interface driver */

#include "vxWorks.h"
#include "stdio.h"
#include "config.h"
#include "drv/pci/pciIntLib.h"
#include "drv/pci/pciConfigLib.h"
#include "string.h"
#include "memLib.h"
#include "end.h"
#include "muxLib.h"
#include "m2Lib.h"
#include "wlandrv.h"
#include "ar5hwcEnd.h"
#include "halApi.h"

#ident "ACI: $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/atheros/drv/sysAr5Netif.c#1 $"

#define AR5_LOAD_STRING         "0x00:0x00:0:0:0:0"

STATUS
sysAr5PciInit(void)
{
    UINT32  busNo;
    UINT32  funcNo;
    UINT32  deviceNo;
    UINT8   irq;
    UINT32  ar5MemBase;
    UINT16  deviceId;
    UINT32  numDeviceIds = halGetNumDevid();
    int     i;

#ifdef INCLUDE_PCI_AUTOCONF
    extern END_TBL_ENTRY endDevTbl[];
#endif

    for (i = 0; i < numDeviceIds; i++) {
        deviceId = halGetDevids(i);
        if (pciFindDevice(ATHEROS_VENDOR_ID, deviceId, 0, &busNo, &deviceNo, &funcNo) == OK) {
             break;
        }
    }
    
    if (i >= numDeviceIds) {
        return A_ERROR;
    }

    /* read in values of iobase and irq */

#ifndef INCLUDE_PCI_AUTOCONF
    ar5MemBase = AR0_MEMBASE0;
    irq = AR0_INT_LVL;
    if (PCI_CFG_TYPE == PCI_CFG_FORCE) {

        pciConfigOutLong(busNo, deviceNo, funcNo,
                     PCI_CFG_BASE_ADDRESS_0, ar5MemBase);

        pciConfigOutByte(busNo, deviceNo, funcNo, PCI_CFG_DEV_INT_LINE, irq);
    }
#endif

    /*
     * Get memory base address and IO base address from the PCI
     * configuration space, as written by the PCI auto-config
     * code.  Note that this addresses must be CPU relative,
     * which may not be true if using a PCI to PCI bridge.
     */
    pciConfigInLong (busNo, deviceNo, funcNo, PCI_CFG_BASE_ADDRESS_0,
                     &ar5MemBase);
    ar5MemBase &= PCI_MEMBASE_MASK;
    pciConfigInByte (busNo, deviceNo, funcNo, PCI_CFG_DEV_INT_LINE, &irq);

    /* initialise arParam */
#if defined(PPC8240)
    arParam.ivec   = (int) irq + INT_NUM_IRQ0;
#else
#error ar5 ivec not programed for this platform!
#endif
    arParam.ilevel = (int) irq;
    arParam.pciMemBase = ar5MemBase;

    arParam.pciVendorID = ATHEROS_VENDOR_ID;
    arParam.pciDeviceID = deviceId;
    pciConfigInWord(busNo, deviceNo, funcNo, PCI_CFG_SUB_VENDER_ID,
        &arParam.pciSubVendorID);
    pciConfigInWord(busNo, deviceNo, funcNo, PCI_CFG_SUB_SYSTEM_ID,
        &arParam.pciSubDeviceID);
    pciConfigInByte(busNo, deviceNo, funcNo, PCI_CFG_REVISION,
        &arParam.pciRevisionID);

#ifndef INCLUDE_PCI_AUTOCONF
    pciConfigOutWord(busNo, deviceNo, funcNo,
             PCI_CFG_CACHE_LINE_SIZE,
                 (PCI_LAT_TIMER << 8) | (_CACHE_ALIGN_SIZE >> 2));
#endif

    /* enable mapped memory and IO addresses */
    pciConfigOutWord(busNo, deviceNo, funcNo,
             PCI_CFG_COMMAND, PCI_CMD_MEM_ENABLE|PCI_CMD_MASTER_ENABLE);

    /*
     * Change the PCI config reg to ignore many disconnects before
     * generating master cycle abort.
     */
    pciConfigOutWord(busNo, deviceNo, funcNo, 0x40, 0);


#ifdef INCLUDE_PCI_AUTOCONF
    /* Scan endDevTbl for free slot */
    for (i = 0; endDevTbl[i].unit != -1; i++) {
        if (endDevTbl[i].endLoadFunc == END_TBL_END)
            break;
    }
    if (endDevTbl[i].unit == -1) {
        return A_ERROR; /* table full */
    }

    endDevTbl[i].unit = 0;          /* we only support 1 device for now */
    endDevTbl[i].endLoadFunc = ar5hwcEndLoad;
    endDevTbl[i].endLoadString = AR5_LOAD_STRING;
    endDevTbl[i].endLoan = TRUE;
    endDevTbl[i].processed = FALSE;
#endif

    return A_OK;
}
