/* Copyright 1984-1997 Wind River Systems, Inc. */
#include "copyright_wrs.h"

/*
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * Atheros ar5311 Integrated AP BSP.  See target.nr for documentation.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/sysLib.c#1 $"

//#include "vxWorks.h"
#include "config.h"

#include "version.h"
#include "cacheLib.h"
#include "esf.h"                        /* exception stack frame */
#include "intLib.h"
#include "iv.h"
#include "memLib.h"
#include "sysLib.h"
#include "excLib.h"
#include "tickLib.h"
#include "wdLib.h"                      /* sysLedBlink() */
#include "taskLib.h"
#include "arch/mips/taskMipsLib.h"
#include "arch/mips/cacheMipsLib.h"
#include "mips4Kx.h"

#include "stdio.h"                      /* we do use printf */
#include "stdlib.h"
#include "string.h"
#include "assert.h"

#include "ar5311reg.h"
#include "wlanvers.h"

#ifdef INCLUDE_SYM_TBL
#include "symLib.h"
#include "sysSymTbl.h"
#endif

#if defined(BUILD_AP) || defined(INCLUDE_TRACE)
#include "wlantype.h"
#include "trace.h"
#endif

LOCAL UINT32 miscIMR, srIMR;            /* misc, sr IMR shadows */

#ifdef EMULATION                        /* support for emulation board */
int emulation = 0;                      /* default to no */
#endif

/* Include the serial port and timer driver code. */
#include "sysSerial.c"
#include "timer/mipsR4kTimer.c"

#ifdef INCLUDE_NETWORK
#include "sysNet.c"
#endif

void procErrorIntHook(int);             /* assembly hook to get eframe */

LOCAL void sysWatchdogUpdateHW(void);
LOCAL void sysWatchdogStart(void);

LOCAL void sysStrayInt(int);            /* ar5311 error handlers */
LOCAL void sysGpioInt(int);
LOCAL void dmaErrorInt(int);
LOCAL void sysFactoryInt(void);

LOCAL void sysBootLineInit(void);
LOCAL void sysBoardDataInit(void);

LOCAL STATUS ar5311ExcVecInitHook(void);

int   sysCpu = CPU;                     /* system CPU type (MIPS) */
char *sysBootLine = BOOT_LINE_ADRS;     /* address of boot line */
char *sysExcMsg = EXC_MSG_ADRS;         /* catastrophic message area */
int   sysProcNum;                       /* processor number of this CPU */
int   sysFlags;                         /* boot flags */
char  sysBootHost[BOOT_FIELD_LEN];      /* name of host from which we booted */
char  sysBootFile[BOOT_FIELD_LEN];      /* name of file from which we booted */

UINT32 sysTicksPerUsec = (40000000 / 1000000);  /* conservative to start */

static UINT32 mem0Size;                 /* primary (required) SDRAM bank */
static UINT32 mem1Size;                 /* secondary (optional) SDRAM bank */

#define LONG_WATCHDOG   800000000       /* conservative 10 seconds */
static int     sysWatchdogValue = LONG_WATCHDOG;
static int     sysWatchDelta;
static int     sysWatchTick;
static int     sysWatchdogsMissed;
static int     sysWatchdogUpdates;
static WDOG_ID wdTimeoutId;
static int     wdTimeout;

/* pull in cache library */
LOCAL STATUS sysCacheInit(CACHE_MODE instMode, CACHE_MODE dataMode);
FUNCPTR sysCacheLibInit = (FUNCPTR)sysCacheInit;

struct ar5311_boarddata sysBoardData;   /* Info on board stuffing */

/*
 * Since tying interrupt lines is board-dependent, sysHashOrder
 * is provided to select the prioritization of interrupt lines.
 *
 * Interrupt    bit      priority ffsMsbTbl
 * SW0          00000001 8        8
 * SW1          00000010 7        7
 * WLAN         00000100 4        6
 * ENET0        00001000 5        5
 * ENET1        00010000 6        4
 * UARTDMA      00100000 3        3
 * MISCIO       01000000 2        2
 * COMPARE      10000000 1        1
 */
static INT8 ffsAr5311Tbl[256] = {
/*  0000 0001 0010 0011 0100 0101 0110 0111 1000 1001 1010 1011 1100 1101 1110 1111 */
    0,   0,   1,   1,   2,   2,   2,   2,   3,   3,   3,   3,   2,   2,   2,   2, /* 0000____ */
    4,   4,   4,   4,   2,   2,   2,   2,   3,   3,   3,   3,   2,   2,   2,   2, /* 0001____ */
    5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5, /* 0010____ */
    5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5,   5, /* 0011____ */
    6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6, /* 0100____ */
    6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6, /* 0101____ */
    6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6, /* 0110____ */
    6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6,   6, /* 0111____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1000____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1001____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1010____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1011____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1100____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1101____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1110____ */
    7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7,   7, /* 1111____ */
};
UINT8 *sysHashOrder = ffsAr5311Tbl;

LOCAL int sysMiscIoVec(int);

#ifdef INCLUDE_AUX_CLK
LOCAL void sysAuxClkInt(int);
#endif

typedef struct {
    ULONG intCause;                     /* cause of interrupt */
    ULONG bsrTableOffset;               /* index to BSR table */
    ULONG intMask;                      /* interrupt mask */
    ULONG demux;                        /* demultiplex argument */
} PRIO_TABLE;

/*
 * intPrioTable is a board-dependent structure that aids in the
 * processing of the 8 interrupt conditions.
 * It is used by excLib to determine the pending interrupt and
 * to call the user attached handler.  The present search algorithm
 * relies on a one to one mapping of interrupt lines and
 * table entries.
 *
 * Each entry has 4 fields: first is the interrupt ID, second
 * is the vector number, third is the mask field, and last is the
 * demux field.  When an interrupt is received the handler
 * handler maps the pending line to its coresponding table
 * entry.  Next the multiplex field is read; if it is zero, field
 * two is taken as the interrupt vector number in the BSR table,
 * otherwise it is interpreted as a demultiplex function and called
 * with field 4 passed as its parameter.  The job of the demultiplex
 * routine is to calculate the correct interrupt vector number and pass
 * it back to the handler.  The handler can then call the routine the
 * user has installed in the BSR table with intVecSet() or intConnect().
 * The mask field is used to mask interrupt conditions for
 * the coresponding interrupt line.  A one in bits 8-15 denotes
 * an interrupt line will be ignored during the processing of
 * the current interrupt line.  Presently the values in the mask
 * field allow all interrupt conditions except the current to become
 * active during the interrupt handling.  This is generally used for
 * prioritizing interrupts.
 *
 * This table is critical to interrupt processing.  Do not alter
 * its contents until you understand the consequences.
 */
PRIO_TABLE intPrioTable[8] = {
    { CAUSE_SW1, IV_SWTRAP0_VEC,        SR_IBIT1, 0 },  /* SW trap 0 */
    { CAUSE_SW2, IV_SWTRAP1_VEC,        SR_IBIT2, 0 },  /* SW trap 1 */
    { CAUSE_IP3, IV_WLAN_VEC,           SR_IBIT3, 0 },  /* WLAN MAC */
    { CAUSE_IP4, IV_ENET0_VEC,          SR_IBIT4, 0 },  /* Enet 0 */
    { CAUSE_IP5, IV_ENET1_VEC,          SR_IBIT5, 0 },  /* Enet 1 */
    { CAUSE_IP6, IV_MISCDMA_VEC,        SR_IBIT6, 0 },  /* Uart DMA */
    { CAUSE_IP7, (ULONG)sysMiscIoVec,   SR_IBIT7, 1 },  /* Vectored IO */
    { CAUSE_IP8, IV_COMPARE_VEC,        SR_IBIT8, 0 }   /* count/compare */
};

/*
 * sysModel - return the model name of the CPU board string.
 */
char *
sysModel(void)
{
    return sysBoardData.boardName;
}

/*
 * sysBspRev - return the BSP version and revision number string.
 */
char *
sysBspRev(void)
{
    return ATHEROS_SW_VERSION BSP_REV;
}

/*
 * sysHwInit0 - Initialize before cache/exception init.
 */
void
sysHwInit0(void)
{
    extern FUNCPTR sysExcVecInitHook;   /* not in any header */

#ifdef LATE_CACHE_ENABLE
    sysCacheEnable();
#endif

    /* Need to hook exceptions 4Kp style (no tlb or cache error + IV) */
    sysExcVecInitHook = ar5311ExcVecInitHook;
}

/*
 * sysHwInit - Initialize the system hardware.
 *
 * It is the first board-specific C code executed, and runs with
 * interrupts masked in the processor.
 *
 * We must resets all devices to a quiescent (no DMA) state, typically
 * by calling driver initialization routines.
 *
 * NOTE: This routine should not be called directly by the user application.
 */
void
sysHwInit(void)
{
    int romStartedWatchdog = 1;

#ifdef EMULATION
    /* The emulator has a major # of zero.  First silicon is 1. */
    if ((sysRegRead(AR5311_REV) & REV_MAJ) == REV_EMULATION) {
        emulation = 1;
    }
#endif
    srIMR = AR5311_SR;
    taskSRInit(srIMR);          /* default task SR value */
    intSRSet(srIMR & ~SR_IE);   /* update SR, but leave intrs off */

    if (sysRegRead(AR5311_WDC) != 0) {
        sysWatchdogUpdateHW();  /* buy some time */
    } else {
        asm_puts("WARNING: bootrom did not start the watchdog.\r\n");
        romStartedWatchdog = 0;
    }

    /* Init the GPIOs */
    sysRegWrite(AR5311_GPIO_DO, 0);
    sysRegWrite(AR5311_GPIO_CR, 0);

    sysRegWrite(AR5311_RELOAD, 0xffffffff);

    /* Ensure all DMA is stopped. */
    (void)sysRegRead(AR5311_RESET);     /* read before */
    sysRegWrite(AR5311_RESET,
                RESET_WLAN | RESET_WARM_WLAN |
                RESET_ENET0 | RESET_EPHY0 |
                RESET_ENET1 | RESET_EPHY1 |
                RESET_DMA);
    sysUDelay(100);                     /* let reset settle a little */
    (void)sysRegRead(AR5311_RESET);     /* read after */
    /* Disable network interfaces and keep them in reset. */
    sysRegWrite(AR5311_ENABLE, ENABLE_APB);

    /* ensure these have not been stepped on */
    sysBootLine = (char *)BOOT_LINE_ADRS;
    sysExcMsg = (char *)EXC_MSG_ADRS;

    sysFlashConfigInit((void *)PHYS_TO_K1(AR5311_FLASH0),
                       NVRAM_SIZE,
                       sizeof(sysBoardData),
                       4096);
    sysBoardDataInit();         /* load boarddata from flash */
    sysLed(1);                  /* announce software is alive */

    /* Disable watchdog early -- should only be used for debugging */
    if (!romStartedWatchdog) {
        sysBoardData.config |= BD_DISWATCHDOG;
    }
    if (sysBoardData.config & BD_DISWATCHDOG) {
        sysWatchdogDisable();
    }

    /* Calibrate timers */
    sysTicksPerUsec = (CPU_CLOCK_RATE / 1000000);

#ifdef INCLUDE_NETWORK
    if (sysBoardData.config & BD_ENET0) {
        sysEnetInit(0, PHYS_TO_K1(AR5311_ENET0), IV_ENET0_VEC);
    }
#ifdef PRJ_BUILD         /* AP and MDK */
    if (sysBoardData.config & BD_ENET1) {
        sysEnetInit(1, PHYS_TO_K1(AR5311_ENET1), IV_ENET1_VEC);
    }
#endif
    if (sysBoardData.config & BD_WLAN) {
        sysWlanInit(PHYS_TO_K1(AR5311_WM2AHB), IV_WLAN_VEC);
    }
    sysNetDone();
#endif

#ifdef INCLUDE_SERIAL
    sysSerialHwInit();          /* serial devices */
#endif
    sysBootLineInit();          /* read boot line from flash */
}

/*
 * sysHwInit2 - initialize additional system hardware
 *
 * This routine connects system interrupt vectors and configures any
 * required features not configured by sysHwInit. It is called from usrRoot()
 * in usrConfig.c after the multitasking kernel has started.
 *
 * RETURNS: N/A
 */
void
sysHwInit2(void)
{
    static int initialized;     /* must protect against double call! */

    sysWatchdogUpdateHW();      /* buy some time */

    if (initialized) {
        return;
    }
    initialized = TRUE;

    /* MIPS count/compare is the main clock.  AR5311_TIMER is the Aux clock */
    (void)intConnect(INUM_TO_IVEC(IV_COMPARE_VEC), sysClkInt, 0);
#ifdef INCLUDE_AUX_CLK
    (void)intConnect(INUM_TO_IVEC(IV_TIMER_VEC), sysAuxClkInt, 0);
#else
    (void)intConnect(INUM_TO_IVEC(IV_TIMER_VEC), sysStrayInt, 0);
#endif

    /*
     * Connect all, and enable interrupts that the core system code
     * owns the interrupts for.
     */
    (void)intConnect(INUM_TO_IVEC(IV_WLAN_VEC), sysStrayInt, IV_WLAN_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_ENET0_VEC), sysStrayInt, IV_ENET0_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_ENET1_VEC), sysStrayInt, IV_ENET1_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_MISCDMA_VEC), sysStrayInt, IV_MISCDMA_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_AHBPROC_VEC), procErrorIntHook, 0);
    (void)intConnect(INUM_TO_IVEC(IV_AHBDMA_VEC), dmaErrorInt, 0);
    (void)intConnect(INUM_TO_IVEC(IV_GPIO_VEC), sysGpioInt, 0);
    (void)intConnect(INUM_TO_IVEC(IV_UART1_VEC), sysStrayInt, IV_UART1_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_UART0_VEC), sysStrayInt, IV_UART0_VEC);
    (void)intConnect(INUM_TO_IVEC(IV_WD_VEC), sysStrayInt, IV_WD_VEC);

    miscIMR = IMR_AHBPROC | IMR_AHBDMA;

    /* Enable reset to factory GPIO if enabled */
    if (sysBoardData.config & BD_RSTFACTORY) {
        sysGpioConnect(sysBoardData.resetConfigGpio, (FUNCPTR)sysFactoryInt, 0);
    }

#ifdef INCLUDE_SERIAL
    sysSerialHwInit2();         /* connect serial interrupts */
#endif

    /* Enable misc interrupts */
    sysRegWrite(AR5311_IMR, miscIMR);
    sysWbFlush();

    /* Update SR interrupt mask to our actual configuration. */
    taskSRInit(srIMR);
    intSRSet(intSRGet() | (srIMR & ~SR_IE));

    /* Make the 2nd bank of memory accessible via malloc(). */
    if (mem1Size) {
        memAddToPool((char *)(LOCAL_MEM_LOCAL_ADRS + AR5311_SDRAM1), mem1Size);
    }

    sysWatchdogStart();
    sysLedBlink();

#ifdef INCLUDE_TRACE
    eventLibInit();
#endif

#ifdef BUILD_AP
    if ((sysRegRead(AR5311_REV) & REV_CHIP) < 0x21) {
        /* early chip is subject to enet and wlan dma bugs */
        asm_puts("ar5311 is pre-production!\r\n");
    }
#endif
}

/*
 * sysCacheInit - cache initialization hook.
 */
LOCAL STATUS
sysCacheInit(CACHE_MODE instMode, CACHE_MODE dataMode)
{
    int isize, ilsize, dsize, dlsize;
    UINT32  c1 = sysConfig1Get();

    ilsize = 1 << (((c1 & CFG1_IL) >> CFG1_IL_S) + 1);
    isize = (((c1 & CFG1_IA) >> CFG1_IA_S) + 1) *
        (64 << ((c1 & CFG1_IS) >> CFG1_IS_S)) * ilsize;

    dlsize = 1 << (((c1 & CFG1_DL) >> CFG1_DL_S) + 1);
    dsize = (((c1 & CFG1_DA) >> CFG1_DA_S) + 1) *
        (64 << ((c1 & CFG1_IS) >> CFG1_DS_S)) * dlsize;

    return cache4kcLibInit(instMode, dataMode,
                           isize, ilsize, dsize, dlsize);
}

/*
 * sysPhysMemTop - get the address of the top of physical memory
 *
 * This routine returns the address of the first missing byte of
 * memory, which indicates the top of memory.  This is only for
 * the primary SDRAM bank (AR5311_SDRAM0).  The secondary SDRAM
 * bank (AR5311_SDRAM1) is added to the malloc() pool with
 * memAddToPool() if it present.
 *
 * Dynamic memory sizing is only mode.  LOCAL_MEM_SIZE is honored
 * by romInit.S for some memory configurations.
 *
 * SEE ALSO: sysMemTop()
 */
char *
sysPhysMemTop(void)
{
    if (mem0Size == 0) {
        UINT32  cfg1 = sysRegRead(AR5311_MEM_CFG1);
        int ac;

        if (cfg1 & MEM_CFG1_E0) {
            ac = (cfg1 & MEM_CFG1_AC0) >> MEM_CFG1_AC0_S;
            mem0Size = 0x00200000 << ac;
        }
#ifdef EMULATION                /* BUG #2276 should be fixed in real hw */
        else {
            mem0Size = 0x00200000 << MEM_CFG1_AC_8;
        }
#endif

        /*
         * CAP SDRAM size in software for testing various memory configs.
         * Do not worry about setting memory address checking to allow
         * bootrom to load vxWorks with reduced footprint.  Value is
         * checked during input.
         */
        if (sysBoardData.config & BD_MEMCAP) {
            UINT32 capSize = sysBoardData.memCap * 1024 * 1024;
            if (mem0Size > capSize) {
                mem0Size = capSize;
            }
        } else {
            /* skip bank1 if using a cap */
            if (cfg1 & MEM_CFG1_E1) {
                ac = (cfg1 & MEM_CFG1_AC1) >> MEM_CFG1_AC1_S;
                mem1Size = 0x00200000 << ac;
            } else {
                mem1Size = 0;
            }
        }

    }

    return (char *)LOCAL_MEM_LOCAL_ADRS + mem0Size;
}

/*
 * sysMemTop - Get the address of the top of VxWorks memory.  This is also
 * the start of USR_RESERVED_MEM (not managed by vxWorks), if non-zero.
 */
char *
sysMemTop(void)
{
    return sysPhysMemTop() - USER_RESERVED_MEM;
}

/*
 * sysToMonitor - transfer control to the ROM monitor
 *
 * This routine transfers control to the ROM monitor.  Normally, it is called
 * only by reboot()--which services ^X--and by bus errors at interrupt level.
 * However, in some circumstances, the user may wish to introduce a
 * <startType> to enable special boot ROM facilities.
 *
 * The entry point for a warm boot is defined by the macro ROM_WARM_ADRS
 * in config.h.  We do an absolute jump to this address to enter the
 * ROM code.
 *
 * RETURNS: Does not return.
 */
STATUS
sysToMonitor(int startType)
{
    FUNCPTR pRom = (FUNCPTR)ROM_WARM_ADRS;

    intLock();                  /* disable interrupts */

    /* Need some time to shutdown - add slop for old bootroms */
    sysWatchdogValue = LONG_WATCHDOG;

    sysLed(0);                  /* turn off system LED */
    sysHwInit();                /* initialize all HW to a quiet state */
    (*pRom)(startType);         /* jump to romInit.s */

    return OK;                  /* in case we continue from ROM monitor */
}

/*
 * sysProcNumGet - get the processor number.  This is always 0.
 */
int
sysProcNumGet(void)
{
    return sysProcNum;
}

/*
 * sysProcNumSet - set the processor number.  Should always be 0.
 */
void
sysProcNumSet(int procNum)      /* processor number */
{
    sysProcNum = procNum;
}

/*
 * ar5311ExcVecInitHook - Hook exceptions 4Kp style instead of R4000 style.
 */
LOCAL STATUS
ar5311ExcVecInitHook(void)
{
    extern void excNormVector(void), excIntIVVector(void);
    extern void excInvalidVector(void);
    extern void excTlbVec(void);

    /* Normal exception vector */
    bcopy((void *)excNormVector, (void *)PHYS_TO_K0(0x180), 0x80);
    cacheTextUpdate((void *)PHYS_TO_K0(0x180), 0x80);

    /* 4Kx IV mode interrupt vector */
    bcopy((void *)excIntIVVector, (void *)PHYS_TO_K0(0x200), 0x80);
    cacheTextUpdate((void *)PHYS_TO_K0(0x200), 0x80);

    /* TLB exception is invalid, but it still happens for badaddr */
    bcopy((void *)excTlbVec, (void *)PHYS_TO_K0(0x000), 0x80);
    cacheTextUpdate((void *)PHYS_TO_K0(0x000), 0x80);

    /* Other vectors are invalid (xtlb and cache error) */
    bcopy((void *)excInvalidVector, (void *)PHYS_TO_K0(0x080), 0x80);
    cacheTextUpdate((void *)PHYS_TO_K0(0x080), 0x80);
    bcopy((void *)excInvalidVector, (void *)PHYS_TO_K0(0x100), 0x80);
    cacheTextUpdate((void *)PHYS_TO_K0(0x100), 0x80);

    return OK;
}

/*
 * sysAutoAck - acknowledge the interrupt condition for specified intr.
 */
int
sysAutoAck(int vecNum)          /* vector num of active interrupt */
{
    switch (vecNum) {
    case IV_SWTRAP0_VEC:        /* software trap 0 */
        intCRSet(intCRGet() & ~CAUSE_SW1);
        break;

    case IV_SWTRAP1_VEC:        /* software trap 1 */
        intCRSet(intCRGet() & ~CAUSE_SW2);
        break;

    case IV_COMPARE_VEC:        /* Sometimes get a stray on boot */
        sysCompareSet(sysCompareGet());
        break;

    default:
        return -1;
    }

    return 0;
}

/*
 * sysMiscIoVec - multiplex code for IV_MISCIO_VEC.  Fan out is
 * prioritized as wd, errors, timer, uart, gpio.
 */
LOCAL int
sysMiscIoVec(int dummy)
{
    UINT32  reg = sysRegRead(AR5311_ISR) & miscIMR;

    /* Skip abnormal high priority things quickly */
    if (reg & (ISR_WD | ISR_AHBPROC | ISR_AHBDMA)) {
        /* All of these are fatal attempt to keep EPC */
        intLock();
        if (reg & ISR_WD) {
            return IV_WD_VEC;
        }
        if (reg & ISR_AHBPROC) {
            return IV_AHBPROC_VEC;
        }
        if (reg & ISR_AHBDMA) {
            return IV_AHBDMA_VEC;
        }
    }
    if (reg & ISR_TIMER) {
        return IV_TIMER_VEC;
    }
    if (reg & ISR_UART0) {              /* dma */
        return IV_UART0_VEC;
    }
    if (reg & ISR_UART1) {
        return IV_UART1_VEC;
    }
    if (reg & ISR_GPIO) {
        return IV_GPIO_VEC;
    }

    return -1;          /* phantom! */
}

/*
 * sysStrayInt - "handle" an unexpected interrupt.  HW watchdog will handle
 * a runaway interrupt.
 */
LOCAL void
sysStrayInt(int vec)
{
    asm_puts("sysStrayInt(");
    asm_puthex(vec);
    asm_puts(")\r\n");

    return;
}

/*
 * GPIO interrupt suppport
 */
struct {
    FUNCPTR func;
    UINT32 arg;
} gpioIntVec[AR5311_NUM_GPIO];
UINT32 gpioIntMask = 0;

void
sysGpioConnect(int gpio, FUNCPTR func, UINT32 arg)
{
    gpioIntVec[gpio].func = func;
    gpioIntVec[gpio].arg = arg;
    sysGpioCtrlIntr(sysBoardData.resetConfigGpio);
    gpioIntMask |= 1 << gpio;
    miscIMR |= IMR_GPIO;
}

LOCAL void
sysGpioInt(int dummy)
{
    UINT32 intr = sysRegRead(AR5311_GPIO_DI) & gpioIntMask;
    int i;

    /* simple interrupt fanout */
    for (i = 0; i < AR5311_NUM_GPIO; i++) {
        if ((intr & (1<<i)) && gpioIntVec[i].func) {
                gpioIntVec[i].func(gpioIntVec[i].arg);
        }
    }
}

/*
 * procErrorInt - "handle" AHB proc bus error.
 *
 * Use low level output to avoid dependance on being stable.
 *
 * Note this is an asynchronous interrupt.  The offending instruction
 * has past.  We get an eframe by using a hook to catch the stack
 * address that is used for Restore(), ie the exception frame.
 *
 * To make matters worse, this error interrupt is often blocked
 * so if it happens in an ISR the trail will lead back to a mtc0
 * C0_SR instruction.

 * We assume the system is more or less ok and try to dump a lot
 * of information.  If we access a bad pointer during the dump,
 * the interrupt is disabled as we are in an ISR so we should
 * not double panic.
 */
void
procErrorInt(ESFMIPS *eframe)
{
    UINT32   err, addr, epc;
    char     sz[2] = { '\0', '\0' };
    UINT32   *sp;
    char     *tname;
    int      tid;
    int      i;

#ifdef INCLUDE_SYM_TBL
    char     fn[MAX_SYS_SYM_LEN+1];
    SYM_TYPE sym;
    STATUS   rc;
    int      value;
#endif

    err  = sysRegRead(AR5311_PROCERR);
    addr = sysRegRead(AR5311_PROCADDR);         /* clears interrupt */

    /* dump this anyway on the off chance that there may be a console */
    asm_puts("\r\n\r\nInvalid CPU access: ");
    asm_puthex(err);
    asm_puts(" at phys addr ");
    asm_puthex(addr);
    asm_puts((err & PROCERR_HWRITE) ? " write" : " read");
    asm_puts(" size ");
    sz[0] = '0' + ((err & PROCERR_HSIZE) >> PROCERR_HSIZE_S);
    asm_puts(sz);
    asm_puts("\r\n");

    /*
     * Let error response continue to take an exception. Yes
     * it is odd to have an interrupt take priority over
     * an exception.
     */
    if ((err & PROCERR_HWRITE) == 0) {
        return;
    }

    /* only try to dump data if we have a console to listen */
    if (sysBoardData.config & BD_UART0) {

        asm_puts("task: ");
        tid = taskIdSelf();
        tname = taskName(tid);
        if (tname == NULL) {
            tname = "NONE";
        }
        asm_puts(tname);
        asm_puts("\r\n");

        asm_puts("exception frame: ");
        asm_puthex((int)eframe);
        asm_puts("\r\n");

        asm_puts("epc: ");
        epc = (UINT32)eframe->esfRegs.pc;
        asm_puthex(epc);
#ifdef INCLUDE_SYM_TBL
        rc = symFindByValue(sysSymTbl, epc, fn, &value, &sym);
        if (rc != OK) {
                strcpy(fn, "unknown");
                value = epc;
        }
        asm_puts(" ");
        asm_puts(fn);
        asm_puts("+");
        asm_puthex(epc - value);
#endif
        asm_puts("\r\n");

        asm_puts("ra:  ");
        asm_puthex(eframe->esfRegs.raReg);
#ifdef INCLUDE_SYM_TBL
        rc = symFindByValue(sysSymTbl, eframe->esfRegs.raReg, fn, &value, &sym);
        if (rc != OK) {
                strcpy(fn, "unknown");
                value = eframe->esfRegs.raReg;
        }
        asm_puts(" ");
        asm_puts(fn);
        asm_puts("+");
        asm_puthex(eframe->esfRegs.raReg - value);
#endif
        asm_puts("\r\n");

        /* rest of register dump */
        asm_puts("  sr: ");
        asm_puthex(eframe->esfRegs.sr);
        asm_puts("  ca: ");
        asm_puthex(eframe->esfRegs.cause);
        asm_puts("  gp: ");
        asm_puthex(eframe->esfRegs.gpReg);
        asm_puts("  at: ");
        asm_puthex(eframe->esfRegs.atReg);
        asm_puts("\r\n");

        asm_puts("  a0: ");
        asm_puthex(eframe->esfRegs.a0Reg);
        asm_puts("  a1: ");
        asm_puthex(eframe->esfRegs.a1Reg);
        asm_puts("  a2: ");
        asm_puthex(eframe->esfRegs.a2Reg);
        asm_puts("  a3: ");
        asm_puthex(eframe->esfRegs.a3Reg);
        asm_puts("\r\n");

        asm_puts("  t0: ");
        asm_puthex(eframe->esfRegs.t0Reg);
        asm_puts("  t1: ");
        asm_puthex(eframe->esfRegs.t1Reg);
        asm_puts("  t2: ");
        asm_puthex(eframe->esfRegs.t2Reg);
        asm_puts("  t3: ");
        asm_puthex(eframe->esfRegs.t3Reg);
        asm_puts("\r\n");

        asm_puts("  t4: ");
        asm_puthex(eframe->esfRegs.t4Reg);
        asm_puts("  t5: ");
        asm_puthex(eframe->esfRegs.t5Reg);
        asm_puts("  t6: ");
        asm_puthex(eframe->esfRegs.t6Reg);
        asm_puts("  t7: ");
        asm_puthex(eframe->esfRegs.t7Reg);
        asm_puts("\r\n");

        asm_puts("  t8: ");
        asm_puthex(eframe->esfRegs.t8Reg);
        asm_puts("  t9: ");
        asm_puthex(eframe->esfRegs.t9Reg);
        asm_puts("  v0: ");
        asm_puthex(eframe->esfRegs.v0Reg);
        asm_puts("  v1: ");
        asm_puthex(eframe->esfRegs.v1Reg);
        asm_puts("\r\n");

        /*
         * Callee saved registers may not be correct on the eframe as they
         * are saved as we go up the call stack.  The hook does try to fix
         * this.
         */
        asm_puts("  s0: ");
        asm_puthex(eframe->esfRegs.s0Reg);
        asm_puts("  s1: ");
        asm_puthex(eframe->esfRegs.s1Reg);
        asm_puts("  s2: ");
        asm_puthex(eframe->esfRegs.s2Reg);
        asm_puts("  s3: ");
        asm_puthex(eframe->esfRegs.s3Reg);
        asm_puts("\r\n");

        asm_puts("  s4: ");
        asm_puthex(eframe->esfRegs.s4Reg);
        asm_puts("  s5: ");
        asm_puthex(eframe->esfRegs.s5Reg);
        asm_puts("  s6: ");
        asm_puthex(eframe->esfRegs.s6Reg);
        asm_puts("  s7: ");
        asm_puthex(eframe->esfRegs.s7Reg);
        asm_puts("\r\n");

        /* dump users stack */
        sp = (UINT32 *)eframe->esfRegs.spReg;
        asm_puts("sp: ");
        asm_puthex((int)sp);
        for (i = 0;  i < (6*24); i++) {
            if ((i%6) == 0) {
                asm_puts("\r\n");
            }
            if (sp[i] & PHYS_TO_K0(0)) {
                asm_puts(" *");                 /* may be an address */
            } else {
            asm_puts("  ");
            }
            asm_puthex(sp[i]);
        }
        asm_puts("\r\n");
    }

#ifdef DEBUG
    /* Stop if the board has a console */
    if (sysBoardData.config & BD_UART0) {
        sysToMonitor(BOOT_NO_AUTOBOOT);
    }
#endif
    sysToMonitor(BOOT_QUICK_AUTOBOOT|BOOT_CLEAR);

    return;
}

/*
 * dmaErrorInt - "handle" AHB dma bus error.
 */
LOCAL void
dmaErrorInt(int unused)
{
        UINT32 err, addr;

        err = sysRegRead(AR5311_DMAERR);
        addr = sysRegRead(AR5311_DMAADDR);      /* clears interrupt */

        asm_puts("\r\n\nAHB DMA ERROR: ");      /* low level dump */
        asm_puthex(err);
        asm_puts(" at ");
        asm_puthex(addr);
        asm_puts("\r\n");

#ifdef DEBUG
        if (sysBoardData.config & BD_UART0) {
                sysToMonitor(BOOT_NO_AUTOBOOT);
        }
#endif
        sysToMonitor(BOOT_QUICK_AUTOBOOT|BOOT_CLEAR);

        return;
}

/*
 * sysLanEnetAddrGet - retrieve ethernet address for ethernet (not wlan).
 */
STATUS
sysLanEnetAddrGet(int unit, char *enetAdrs)
{
    int i;

    switch(unit) {
    case 0:
        for (i=0; i < 6; i++) {
            enetAdrs[i] = sysBoardData.enet0Mac[i];
        }
        return OK;
    case 1:
        for (i=0; i < 6; i++) {
            enetAdrs[i] = sysBoardData.enet1Mac[i];
        }
        return OK;
    }
    return ERROR;
}

/*
 * sysDelay - Ensure outstanding PIO operations are complete
 */
void
sysDelay(void)
{
    sysWbFlush();
}

/*
 * sysUDelay - Delay n microseconds via a spin loop.
 *
 * Uses C0_COUNT register, and depends on overflow of unsigned
 * 2s compliment math.  Limited to "reasonably sized" delays,
 * up to a handfull of seconds should be fine.
 */
void
sysUDelay(int us)
{
    UINT32 delta = sysTicksPerUsec * us;
    UINT32 st = sysCountGet();

    while ((((UINT32)sysCountGet()) - st) < delta)
        ;
}

/*
 * sysMsDelay - Delay n milliseconds via a spin loop.
 */
void
sysMsDelay(int ms)
{
    sysUDelay(ms*1000);
}

/*
 * sysNvRamGet - Get the contents of the nvram.  For bootline support.
 *               Implemented in the system flash.
 */
STATUS
sysNvRamGet(char *str, int len, int off)
{
    int i;

    off += NVRAM_BOOT_OFFSET;

    for (i=0; i < len; i++) {
        str[i] = sysFlashConfigRead(FLC_BOOTLINE, off+i);
    }
    return OK;
}


/*
 * sysNvRamSet - set the contents of the nvram.  For bootline support.
 *               Implemented in the system flash.
 */
STATUS
sysNvRamSet(char *str, int len, int off)
{
    sysFlashConfigWrite(FLC_BOOTLINE, off + NVRAM_BOOT_OFFSET, str, len);
    return OK;
}

/*
 * sysBootLineInit - read boot line from flash.
 */
LOCAL void
sysBootLineInit(void)
{
#if FORCE_DEFAULT_BOOTLINE
    strcpy(BOOT_LINE_ADRS, DEFAULT_BOOT_LINE);
#else
    char initValue[NVRAM_INIT_SIZE+2];

    /* Initialize Nvram with defaults if currently uninitialized */
    if (sysNvRamGet(initValue, NVRAM_INIT_SIZE, -NVRAM_BOOT_OFFSET) != ERROR) {
        /* determine if 'NVRAM_INITIALIZED' string is present */
        if (strncmp(initValue, NVRAM_INITIALIZED, NVRAM_INIT_SIZE) != 0) {
            /* write initBootLine and init string */
            char *initBootLine = DEFAULT_BOOT_LINE " s=factory";

            sysNvRamSet(initBootLine, strlen(initBootLine) + 1, 0);
            sysNvRamSet(NVRAM_INITIALIZED, NVRAM_INIT_SIZE, -NVRAM_BOOT_OFFSET);
        }
    }
    sysNvRamGet(sysBootLine, NVRAM_SIZE, 0);    /* save in low mem */
    return;
#endif
}

UINT16
sysBoardDataChecksum(struct ar5311_boarddata *bd)
{
    UINT16 cksum = 0;
    UINT16 save;
    UINT8 *p;
    int i;

    if (bd->rev == 1) {
        /* old ROMs have checksum of 0 */
        return 0;
    }

    save      = bd->cksum;
    bd->cksum = 0;
    p         = (UINT8 *)bd;
    for (i = 0; i < sizeof(struct ar5311_boarddata); i++) {
        cksum += p[i];
    }
    bd->cksum = save;

    return cksum;
}

/*
 * sysBoardDataInit - initialize board data structure from flash.
 */
LOCAL void
sysBoardDataInit(void)
{
    char *p;
    int i;

    p = (char *)&sysBoardData;
    sysBoardData.magic = 0;
    for (i=0; i < sizeof(sysBoardData); i++) {
        p[i] = sysFlashConfigRead(FLC_BOARDDATA, i);
    }

    /* validate GPIOs */
    sysBoardData.sysLedGpio      &= (1<<AR5311_NUM_GPIO) - 1;
    sysBoardData.resetConfigGpio &= (1<<AR5311_NUM_GPIO) - 1;

    if (sysBoardData.magic != BD_MAGIC) {
        /* Not intialized.  Use defaults */
        sysBoardData.magic = BD_MAGIC;
        sysBoardData.rev   = BD_REV;

        sysBoardData.wlanMac[0] = sysBoardData.enet0Mac[0] = 0x00;
        sysBoardData.wlanMac[1] = sysBoardData.enet0Mac[1] = 0x03;
        sysBoardData.wlanMac[2] = sysBoardData.enet0Mac[2] = 0x7f;
        sysBoardData.wlanMac[3] = sysBoardData.enet0Mac[3] = 0xfe;
        sysBoardData.wlanMac[4] = sysBoardData.enet0Mac[4] = 0x00;
        sysBoardData.wlanMac[5] = sysBoardData.enet0Mac[5] = 0x00;
        sysBoardData.wlanMac[5] = 0x00;
        sysBoardData.enet0Mac[5] = 0x01;

        sysBoardData.enet1Mac[0] = 0xff;
        sysBoardData.enet1Mac[1] = 0xff;
        sysBoardData.enet1Mac[2] = 0xff;
        sysBoardData.enet1Mac[3] = 0xff;
        sysBoardData.enet1Mac[4] = 0xff;
        sysBoardData.enet1Mac[5] = 0xff;

        sysBoardData.pciId = 0x011;

#ifdef EMULATION
        if (emulation) {
            strcpy(sysBoardData.boardName, "Atheros AR5311 emulation");
            sysBoardData.major = 0;
            sysBoardData.minor = 1;
            sysBoardData.config = BD_ENET0 |    /* only 1 enet */
                BD_UART0 | BD_UART1 |           /* both serial */
                BD_EXTUARTCLK |                 /* external clock */
                BD_SYSFREQ |
                BD_SYSLED;
            sysBoardData.sysFreq = 20000000;
            sysBoardData.sysLedGpio = 7;
            sysBoardData.resetConfigGpio = 0;
        } else
#endif
        {
            strcpy(sysBoardData.boardName, "Atheros AR5311 AP21 default");
            sysBoardData.major = 1;
            sysBoardData.minor = 0;
            sysBoardData.config = BD_ENET0 |    /* only 1 enet */
                BD_UART0 |                      /* console on dma port */
                BD_WLAN |
                BD_SYSLED |
                BD_RSTFACTORY;
            sysBoardData.sysLedGpio = 7;
            sysBoardData.resetConfigGpio = 6;
        }
    }

    /* This is just a check to detect any issues.  Probably better
     * to use what we have than the defaults.
     */
    if (sysBoardData.cksum != sysBoardDataChecksum(&sysBoardData)) {
        asm_puts("\r\nboardData checksum failed!\r\n");
    }

#ifndef BD_BRINGUP
    /* These are risky and potentially destructive to set! */
    sysBoardData.config &= ~(BD_CPUFREQ|BD_SYSFREQ|BD_EXTUARTCLK);
#endif

    if (((sysBoardData.config & BD_SYSFREQ) == 0) ||
        (sysBoardData.sysFreq <   1000000) ||   /* sanity test */
        (sysBoardData.sysFreq > 150000000)) {
        /* Real HW is fixed at (32.0MHz * 5) / 2 == 80MHz */
        sysBoardData.sysFreq = 80000000;
    }
    if (((sysBoardData.config & BD_CPUFREQ) == 0) ||
        (sysBoardData.cpuFreq <   1000000) ||   /* sanity test */
        (sysBoardData.cpuFreq > 250000000)) {
        UINT32 cpu_s, cpu_e, sys_e;

        sysRegWrite(AR5311_TIMER, 0x40000000);
        cpu_s = sysCountGet();
        sysWbFlush();
        while (((cpu_e = sysCountGet()) - cpu_s) < 100000)
            ;
        sys_e = sysRegRead(AR5311_TIMER);
        sysBoardData.cpuFreq = (cpu_e - cpu_s);
        sysBoardData.cpuFreq *=
            (sysBoardData.sysFreq + (0x40000000 - sys_e - 1)) /
            (0x40000000 - sys_e);
        sysBoardData.cpuFreq *= 2;      /* C0_count at 1/2 rate */

        /* round down to 0.1Mhz */
        sysBoardData.cpuFreq /= 100000;
        sysBoardData.cpuFreq *= 100000;
    }
    sysBoardData.cntFreq = sysBoardData.cpuFreq >> 1;
    return;
}

#if !defined(NDEBUG)
/*
 * bdShow - Show board data.
 */
int
bdShow(struct ar5311_boarddata *bd)
{
    if (bd == NULL) {
        bd = &sysBoardData;     /* vxworks shell case */
    }
    printf("name:    %s\n", bd->boardName);
    printf("magic:   %x\n", bd->magic);
    printf("cksum:   %x\n", bd->cksum);
    printf("rev:     %d\n", bd->rev);
    printf("major:   %d\n", bd->major);
    printf("minor:   %d\n", bd->minor);
    printf("pciid:   %04x\n", bd->pciId);
    printf("wlan:    %s %02x:%02x:%02x:%02x:%02x:%02x\n",
           (bd->config & BD_WLAN) ? "yes" : "no ",
           bd->wlanMac[0],
           bd->wlanMac[1],
           bd->wlanMac[2],
           bd->wlanMac[3],
           bd->wlanMac[4],
           bd->wlanMac[5]);
    printf("enet0:   %s %02x:%02x:%02x:%02x:%02x:%02x\n",
           (bd->config & BD_ENET0) ? "yes" : "no ",
           bd->enet0Mac[0],
           bd->enet0Mac[1],
           bd->enet0Mac[2],
           bd->enet0Mac[3],
           bd->enet0Mac[4],
           bd->enet0Mac[5]);
    printf("enet1:   %s %02x:%02x:%02x:%02x:%02x:%02x\n",
           (bd->config & BD_ENET1) ? "yes" : "no ",
           bd->enet1Mac[0],
           bd->enet1Mac[1],
           bd->enet1Mac[2],
           bd->enet1Mac[3],
           bd->enet1Mac[4],
           bd->enet1Mac[5]);
    printf("uart0:   %s\n",
           (bd->config & BD_UART0) ? "yes" : "no");
    printf("uart1:   %s\n",
           (bd->config & BD_UART1) ? "yes" : "no");
    printf("sysled:  %s, gpio %d\n",
           (bd->config & BD_SYSLED) ? "yes" : "no",
           bd->sysLedGpio);
    printf("factory: %s, gpio %d\n",
           (bd->config & BD_RSTFACTORY) ? "yes" : "no",
           bd->resetConfigGpio);
    printf("serclk:  %s\n",
           (bd->config & BD_EXTUARTCLK) ? "external" : "internal");
    printf("cpufreq: %s %d Hz\n",
           (bd->config & BD_CPUFREQ) ? "configured" : "calculated",
           bd->cpuFreq);
    printf("sysfreq: %s %d Hz\n",
           (bd->config & BD_SYSFREQ) ? "configured" : "calculated",
           bd->sysFreq);
    if (bd->config & BD_MEMCAP) {
        printf("memcap:  enabled %dMB\n", bd->memCap);
    } else {
        printf("memcap:  disabled\n");
    }
    printf("watchdg: %s\n", (bd->config & BD_DISWATCHDOG) ?
           "disabled (WARNING: for debugging only!)\n" : "enabled");
    return 0;
}

/*
 * bdChange - Configure board data.  Assumes trained fingers.  Should be
 *            read only once we leave the factory (ie, not in CLI).
 */
int
bdChange(void)
{
    struct ar5311_boarddata bd;
    UINT32 mac[6];                      /* use 32b items for sscanf() */
    char buf[160];

    bcopy((void *)&sysBoardData, (void *)&bd, sizeof(bd));
    bd.cntFreq = 0;
    bd.magic   = BD_MAGIC;
    bd.rev     = BD_REV;

    printf("Update board data (enter to keep, ^D to finish):\n");

    /* Things more likely to change during development are first */
    printf("name [%s] ", bd.boardName);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        strcpy(bd.boardName, buf);
    }

    printf("HW watchdog [%s] ", (bd.config & BD_DISWATCHDOG) ? "n" : "y");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'y') {
            bd.config &= ~BD_DISWATCHDOG;
        } else {
            bd.config |= BD_DISWATCHDOG;
        }
    }

    printf("memcap valid [%s] ", (bd.config & BD_MEMCAP) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'y') {
            bd.config |= BD_MEMCAP;
        } else {
            bd.config &= ~BD_MEMCAP;
        }
    }
    if (bd.config & BD_MEMCAP) {
        printf("memcap [%dMB] ", bd.memCap);
        if (gets(buf) == NULL) goto done;
        if (buf[0] != '\0') {
            bd.memCap = atoi(buf);
            /* Keep things sane. */
            if (bd.memCap != 4 && bd.memCap != 8) {
                printf("sdram cap must be 4 or 8!\n");
                bd.memCap = 0xffff;
                bd.config &= ~BD_MEMCAP;
            }
        }
    }

    printf("wlan mac address [%02x:%02x:%02x:%02x:%02x:%02x]: ",
           bd.wlanMac[0],
           bd.wlanMac[1],
           bd.wlanMac[2],
           bd.wlanMac[3],
           bd.wlanMac[4],
           bd.wlanMac[5]);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        mac[0] = bd.wlanMac[0];
        mac[1] = bd.wlanMac[1];
        mac[2] = bd.wlanMac[2];
        mac[3] = bd.wlanMac[3];
        mac[4] = bd.wlanMac[4];
        mac[5] = bd.wlanMac[5];
        sscanf(buf, "%x:%x:%x:%x:%x:%x",
               &mac[0], &mac[1], &mac[2],
               &mac[3], &mac[4], &mac[5]);
        bd.wlanMac[0] = mac[0];
        bd.wlanMac[1] = mac[1];
        bd.wlanMac[2] = mac[2];
        bd.wlanMac[3] = mac[3];
        bd.wlanMac[4] = mac[4];
        bd.wlanMac[5] = mac[5];
    }
    printf("enet0 mac address [%02x:%02x:%02x:%02x:%02x:%02x]: ",
           bd.enet0Mac[0],
           bd.enet0Mac[1],
           bd.enet0Mac[2],
           bd.enet0Mac[3],
           bd.enet0Mac[4],
           bd.enet0Mac[5]);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        mac[0] = bd.enet0Mac[0];
        mac[1] = bd.enet0Mac[1];
        mac[2] = bd.enet0Mac[2];
        mac[3] = bd.enet0Mac[3];
        mac[4] = bd.enet0Mac[4];
        mac[5] = bd.enet0Mac[5];
        sscanf(buf, "%x:%x:%x:%x:%x:%x",
               &mac[0], &mac[1], &mac[2],
               &mac[3], &mac[4], &mac[5]);
        bd.enet0Mac[0] = mac[0];
        bd.enet0Mac[1] = mac[1];
        bd.enet0Mac[2] = mac[2];
        bd.enet0Mac[3] = mac[3];
        bd.enet0Mac[4] = mac[4];
        bd.enet0Mac[5] = mac[5];
    }
    printf("enet1 mac address [%02x:%02x:%02x:%02x:%02x:%02x]: ",
           bd.enet1Mac[0],
           bd.enet1Mac[1],
           bd.enet1Mac[2],
           bd.enet1Mac[3],
           bd.enet1Mac[4],
           bd.enet1Mac[5]);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        mac[0] = bd.enet1Mac[0];
        mac[1] = bd.enet1Mac[1];
        mac[2] = bd.enet1Mac[2];
        mac[3] = bd.enet1Mac[3];
        mac[4] = bd.enet1Mac[4];
        mac[5] = bd.enet1Mac[5];
        sscanf(buf, "%x:%x:%x:%x:%x:%x",
               &mac[0], &mac[1], &mac[2],
               &mac[3], &mac[4], &mac[5]);
        bd.enet1Mac[0] = mac[0];
        bd.enet1Mac[1] = mac[1];
        bd.enet1Mac[2] = mac[2];
        bd.enet1Mac[3] = mac[3];
        bd.enet1Mac[4] = mac[4];
        bd.enet1Mac[5] = mac[5];
    }

    /* These usually are not modified...handle with care! */
#ifdef BD_BRINGUP
    printf("cpufreq valid [%s] ", (bd.config & BD_CPUFREQ) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_CPUFREQ;
        } else {
            bd.config |= BD_CPUFREQ;
        }
    }
    if (bd.config & BD_CPUFREQ) {
        printf("cpufreq [%d] ", bd.cpuFreq);
        if (gets(buf) == NULL) goto done;
        if (buf[0] != '\0') {
            bd.cpuFreq = atoi(buf);
        }
    } else {
        bd.cpuFreq = 0;
    }

    printf("sysfreq valid [%s] ", (bd.config & BD_SYSFREQ) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_SYSFREQ;
        } else {
            bd.config |= BD_SYSFREQ;
        }
    }
    if (bd.config & BD_SYSFREQ) {
        printf("sysfreq [%d] ", bd.sysFreq);
        if (gets(buf) == NULL) goto done;
        if (buf[0] != '\0') {
            bd.sysFreq = atoi(buf);
        }
    } else {
        bd.sysFreq = 0;
    }
#endif
    printf("major [%d]: ", bd.major);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        bd.major = atoi(buf);
    }
    printf("minor [%d]: ", bd.minor);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        bd.minor = atoi(buf);
    }
    printf("pciid [%04x]: ", bd.pciId);
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        bd.pciId = strtol(buf, NULL, 16);
    }
    printf("wlan enabled [%s]: ", (bd.config & BD_WLAN) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_WLAN;
        } else {
            bd.config |= BD_WLAN;
        }
    }
    printf("enet0 enabled [%s]: ", (bd.config & BD_ENET0) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_ENET0;
        } else {
            bd.config |= BD_ENET0;
        }
    }
    printf("enet1 enabled [%s]: ", (bd.config & BD_ENET1) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_ENET1;
        } else {
            bd.config |= BD_ENET1;
        }
    }
#ifdef BD_BRINGUP
    printf("external serial clock enabled [%s]: ",
           (bd.config & BD_EXTUARTCLK) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_EXTUARTCLK;
        } else {
            bd.config |= BD_EXTUARTCLK;
        }
    }
#endif
#if CONSOLE_ADDR == AR5311_UART0
    printf("uart0 (console) enabled [%s]: ",
           (bd.config & BD_UART0) ? "y" : "n");
#else
    printf("uart0 enabled [%s]: ",
           (bd.config & BD_UART0) ? "y" : "n");
#endif
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_UART0;
        } else {
            bd.config |= BD_UART0;
        }
    }
#if CONSOLE_ADDR == AR5311_UART0
    printf("uart1 enabled [%s]: ",
           (bd.config & BD_UART1) ? "y" : "n");
#else
    printf("uart1 (console) enabled [%s]: ",
           (bd.config & BD_UART1) ? "y" : "n");
#endif
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_UART1;
        } else {
            bd.config |= BD_UART1;
        }
    }
    printf("sysled enabled [%s] ", (bd.config & BD_SYSLED) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_SYSLED;
        } else {
            bd.config |= BD_SYSLED;
        }
    }
    if (bd.config & BD_SYSLED) {
        printf("sysled gpio [%d] ", bd.sysLedGpio);
        if (gets(buf) == NULL) goto done;
        if (buf[0] != '\0') {
            bd.sysLedGpio  = atoi(buf);
            bd.sysLedGpio &= (1<<AR5311_NUM_GPIO) - 1;
        }
    } else {
        bd.sysLedGpio = 0;
    }
    printf("factory reset enabled [%s] ",
           (bd.config & BD_RSTFACTORY) ? "y" : "n");
    if (gets(buf) == NULL) goto done;
    if (buf[0] != '\0') {
        if (buf[0] == 'n') {
            bd.config &= ~BD_RSTFACTORY;
        } else {
            bd.config |= BD_RSTFACTORY;
        }
    }
    if (bd.config & BD_RSTFACTORY) {
        printf("factory reset gpio [%d] ", bd.resetConfigGpio);
        if (gets(buf) == NULL) goto done;
        if (buf[0] != '\0') {
            bd.resetConfigGpio  = atoi(buf);
            bd.resetConfigGpio &= (1<<AR5311_NUM_GPIO) - 1;
        }
    } else {
        bd.resetConfigGpio = 0;
    }
done:
    bd.cksum = sysBoardDataChecksum(&bd);
    sysFlashConfigWrite(FLC_BOARDDATA, 0, (char *)&bd, sizeof(bd));
    printf("\n");
    bdShow(&bd);
    printf("done.  Reboot to activate\n");
    return 0;
}
#endif /* !NDEBUG */

/*
 * GPIO control routines.
 */
void
sysGpioCtrlOutput(int gpio)
{
    UINT32 reg;

    reg = sysRegRead(AR5311_GPIO_CR);
    reg &= ~(GPIO_CR_M(gpio) | GPIO_CR_UART(gpio) | GPIO_CR_INT(gpio));
    reg |= GPIO_CR_O(gpio);
    sysRegWrite(AR5311_GPIO_CR, reg);
}

void
sysGpioCtrlInput(int gpio)
{
    UINT32 reg;

    reg = sysRegRead(AR5311_GPIO_CR);
    reg &= ~(GPIO_CR_M(gpio) | GPIO_CR_UART(gpio) | GPIO_CR_INT(gpio));
    reg |= GPIO_CR_I(gpio);
    sysRegWrite(AR5311_GPIO_CR, reg);
}

void
sysGpioCtrlIntr(int gpio)
{
    UINT32 reg;

    reg = sysRegRead(AR5311_GPIO_CR);
    reg &= ~(GPIO_CR_M(gpio) | GPIO_CR_UART(gpio) | GPIO_CR_INT(gpio));
    reg |= GPIO_CR_I(gpio);
    reg |= GPIO_CR_INT(gpio);
    sysRegWrite(AR5311_GPIO_CR, reg);
}

void
sysGpioSet(int gpio, int val)
{
    UINT32 reg;

    reg = sysRegRead(AR5311_GPIO_DO);
    reg &= ~(1 << gpio);
    reg |= (val&1) << gpio;
    sysRegWrite(AR5311_GPIO_DO, reg);
}

int
sysGpioGet(int gpio)
{
    return (sysRegRead(AR5311_GPIO_DI) >> gpio) & 1;
}

void
sysLedSet(int on)
{
    sysGpioCtrlOutput(sysBoardData.sysLedGpio); /* handle reset */
    sysGpioSet(sysBoardData.sysLedGpio, on);
}

static WDOG_ID ledWD;
static int ledSt;

void
sysLed(int on)
{
    if ((sysBoardData.config & BD_SYSLED) == 0) {
        return;
    }

    if (ledWD != NULL) {
        wdCancel(ledWD);
    }
    sysLedSet(ledSt = on);
}

void
sysLedBlink(void)
{
    if ((sysBoardData.config & BD_SYSLED) == 0) {
        return;
    }

    ledSt ^= 1;
    sysLedSet(ledSt);

    if (ledWD == NULL) {
        ledWD = wdCreate();
        if (ledWD == NULL) {
            printf("ledWD creation failed!\n");
            return;
        }
    }

    wdStart(ledWD, sysClkRateGet() >> 1, (FUNCPTR)sysLedBlink, 0);
}

/*
 * sysReset - warm reset the system
 */
void
sysReset(void)
{
    while (1) {
        sysRegWrite(AR5311_RESET, RESET_SYSTEM);
        (void)sysRegRead(AR5311_RESET);     /* read after */
    }
    /*NOTREACHED*/
}

/*
 * sysToFactory - Set bootline to the default value and set factory
 *                tag to be picked up for APCFG refresh.
 */
void
sysToFactory(void)
{
    BOOT_PARAMS bp, origBp;
    char bl[160];

    bootStringToStruct(DEFAULT_BOOT_LINE " s=factory", &bp);
    bootStringToStruct(sysBootLine, &origBp);

    asm_puts("resetting to factory config.\r\n");

    /*
     * Keep current boot image if the flash fs if we
     * think it's good.  Since we are in interrupt
     * context this is harder than we would like.
     *
     * So we just check that it is mostly well formed.
     */
    if ((strncmp(origBp.bootFile, "/fl/apimg", 9) == 0) ||
        (strncmp(origBp.bootFile, "/fl/APIMG", 9) == 0))
    {
        strcpy(bp.bootFile, origBp.bootFile);   /* keep bootfile */
    }
    bootStructToString(bl, &bp);
    sysNvRamSet(bl, strlen(bl)+1, 0);
}

/*
 * sysFactoryInt - Interrupt handler for factory reset button.
 *
 * If the button is held down long, set-up to reset to the
 * defaults.  A short press will only reset the system.
 */
LOCAL void
sysFactoryInt(void)
{
    UINT32 st = sysCountGet();

    (void)intLock();

    while (sysRegRead(AR5311_GPIO_DI) & (1 << sysBoardData.resetConfigGpio)) {
        sysWatchdogUpdateHW();

        /* go ahead and reset configuration at four seconds */
        if ((((UINT32)sysCountGet() - st)) >= (4*sysBoardData.cntFreq)) {
                sysToFactory();
                break;
        }
    }
    sysReset();
}

/*
 * AR5311 hardware watchdog interfaces.
 *
 * Update Strategy:
 *      - Bootrom enables the watchdog with a long timeout.
 *      - Early C start-up pokes the watchdog to make sure we are ok.
 *      - sysWatchdogStart() starts a timer to update the hardware at
 *        the interrupt context.  This lets real hangs reset quickly.
 *      - sysWatchdogPeriodSet() sets the software timeout.  This
 *        is a longer timeout period to allow lazy scheduling of
 *        the low priority software watchdog task.
 *      - sysWatchdogUpdate() updates the software timeout value.  This
 *        is ment to be called from a low priority task to ensure
 *        tasks can run.
 *
 * Timeout Handling:
 *      - The bootrom will print a message to the console a watchdog
 *        occured.
 *      - Since the timeout causes an NMI the register set is dumped.
 *      - The system is then reset.
 *
 * Special cases:
 *      - sysToMonitor() resets the watchdog with a long timeout.
 *      - To debug the watchdog must be disabled from the CLI or
 *        in the boardData.
 *
 * sysWatchdogPeriodSet() - time in us until SW watchog expiries.
 * sysWatchdogUpdate()    - restart SW watchdog period.
 *
 * sysWatchdogUpdateHW()  - restart HW watchdog period.
 *
 * sysWatchdogEnable()    - enable watchdog (default case)
 * sysWatchdogDisable()   - disable watchog.
 *
 * sysWatchdogStart()     - start watchdog timer
 *
 */
void
sysWatchdogPeriodSet(int ticks)
{
    sysWatchDelta = ticks + (ticks/2);          /* include slop */
}

void
sysWatchdogUpdate(void)
{
    sysWatchTick = tickGet() + sysWatchDelta;
    if (sysWatchdogValue == 0) {
        /* re-enable watchdog in timer */
        sysWatchdogValue = sysBoardData.sysFreq;
    }
}

LOCAL void
sysWatchdogUpdateHW(void)
{
    if ((sysBoardData.config & BD_DISWATCHDOG) == 0) {
        sysRegWrite(AR5311_WD, sysWatchdogValue);
        sysRegWrite(AR5311_WDC, WDC_NMI);   /* always NMI for register dump */
    }
}

LOCAL void
sysWatchdogTimer(void)
{
    int tick;

    if (sysWatchdogValue) {
        if (sysWatchTick && ((tick = tickGet()) > sysWatchTick)) {
            sysWatchdogsMissed++;
            asm_puts("ar5311 watchdog: missed soft update ");
            asm_puthex(tick);
            asm_puts(" > ");
            asm_puthex(sysWatchTick);
            asm_puts("\r\n");
        } else {
            sysWatchdogUpdateHW();
            sysWatchdogUpdates++;
            sysWatchdogsMissed = 0;
        }
    }
    wdStart(wdTimeoutId, wdTimeout, (FUNCPTR)sysWatchdogTimer, 0);
}

/*
 * sysWatchdogFailed - Dump some more watchdog state.
 *
 * NOTE: I've tried to get apPanic() to work to drop into the
 *       shell.  Interrupts work ok, but tasking seems stuck.
 */
LOCAL void
sysWatchdogFailed(UINT32 epc, ESFMIPS *eframe)
{
#ifdef INCLUDE_SYM_TBL
    char     fn[MAX_SYS_SYM_LEN+1];
    SYM_TYPE sym;
    STATUS   rc;
#endif
    UINT32   *sp;
    char     *name;
    int      value;
    int      i;

    sysWatchdogValue = LONG_WATCHDOG;   /* give us some time */
    sysWatchdogUpdateHW();

    /* We are not comming back, disable interrupts */
    intSRSet((intSRGet() & ~(SR_NMI|SR_IE)) | SR_BEV);

    /* only try to dump data if we have a console to listen */
    if ((sysBoardData.config & BD_UART0) == 0) {
        return;
    }

    if (sysWatchdogsMissed == 0) {
        /* someone held interrupt context for too long */
        asm_puts("watchdog: tripped by an spl abuser\r\n");
    } else {
        asm_puts("watchdog: missed soft update\r\n");
    }

    asm_puts("intContext: ");
    asm_puts(intContext() ? "true\r\n" : "false\r\n");

    asm_puts("ErrorPC: ");
    asm_puthex(epc);
#ifdef INCLUDE_SYM_TBL
    rc = symFindByValue(sysSymTbl, epc, fn, &value, &sym);
    if (rc != OK) {
        strcpy(fn, "unknown");
        value = epc;
    }
    asm_puts(" ");
    asm_puts(fn);
    asm_puts("+");
    asm_puthex(epc - value);
#endif
    asm_puts("\r\n");

    asm_puts("Task Id: ");
    value = taskIdSelf();
    asm_puthex(value);
    if (value) {
        name = taskName(value);
        if (name == NULL) {
            name = "NONE";
        }
        asm_puts(" ");
        asm_puts(name);
    }
    asm_puts("\r\n");

    /* dump users stack */
    sp = (UINT32 *)eframe->esfRegs.spReg;
    asm_puts("stack dump: ");
    asm_puthex((int)sp);
    for (i = 0;  i < (6*24); i++) {
        if ((i%6) == 0) {
            asm_puts("\r\n");
        }
        if (sp[i] & PHYS_TO_K0(0)) {
            asm_puts(" *");                 /* may be an address */
        } else {
            asm_puts("  ");
        }
        asm_puthex(sp[i]);
    }
    asm_puts("\r\n");

    asm_puts("\r\n");

    return;                     /* reset the system */
}

LOCAL void
sysWatchdogStart(void)
{
    sysWatchdogsMissed = 0;
    *(volatile UINT32 *)SYS_NMI_CALLBACK = (UINT32)sysWatchdogFailed;

    if (wdTimeoutId == NULL) {
        wdTimeoutId = wdCreate();
        if (wdTimeoutId == NULL) {
            printf("wdTimeoutId creation failed!\n");
            return;
        }
    }
    /* Speed up to 1 second watchdog */
    sysWatchdogValue = sysBoardData.sysFreq;
    wdTimeout = sysClkRateGet() - 4;    /* 66.6ms window with HZ=60 */
    wdStart(wdTimeoutId, wdTimeout, (FUNCPTR)sysWatchdogTimer, 0);
}

void
sysWatchdogEnable(void)
{
    sysWatchdogValue = sysBoardData.sysFreq;
    sysWatchdogUpdateHW();
}

void
sysWatchdogDisable(void)
{
    sysWatchdogValue = 0;
    sysRegWrite(AR5311_WDC, 0);
}

#ifdef DEBUG
int
sysWatchdogShow(void)
{
    int now = tickGet();

    printf("sysWatchdog:\n");
    printf(" HW update every %d ticks\n", wdTimeout);
    printf(" HW timeout in %d cycles\n", sysWatchdogValue);
    printf(" HW updated %d times\n", sysWatchdogUpdates);
    printf(" HW ctrl %x count %08x\n",
           sysRegRead(AR5311_WDC), sysRegRead(AR5311_WD));
    printf(" SW timeout every %d ticks\n", sysWatchDelta);
    printf(" SW timeout at %d ticks now %d\n", sysWatchTick, now);

    return 0;
}

int
sysWatchdogKill(int context)
{
    switch (context) {
    case 1:
        /* software watchdog - the update may beat this */
        sysWatchTick = tickGet();
        break;
    case 2:
        /* hardware watchdog by holding spl */
        intLock();
        while (1) {
            asm_puts(".");
        }
        break;
    case 0:
    default:
        /* hardware watchdog */
        wdCancel(wdTimeoutId);
        break;
    }
    return 0;
}
#endif

#ifdef INCLUDE_AUX_CLK
/*
 * vxWorks aux clock implementation using ar5311 timer.
 */
static int     sysAuxClkTicksPerSecond = AUX_CLK_RATE_MIN;
static int     sysAuxClkConnected;
static int     sysAuxClkRunning;

static FUNCPTR sysAuxClkRoutine;
static int     sysAuxClkArg;

/*
 * sysAuxClkConnect - set aux clock interrupt callback.
 */
STATUS
sysAuxClkConnect(FUNCPTR routine, int arg)
{
    sysAuxClkRoutine   = routine;
    sysAuxClkArg       = arg;
    sysAuxClkConnected = TRUE;
    return OK;
}

/*
 * sysAuxClkInt - aux clock interrupt handler.
 */
LOCAL void
sysAuxClkInt(int dummy)
{
    (void)sysRegRead(AR5311_TIMER);             /* clear interrupt */

    if (sysAuxClkRoutine) {
        (*sysAuxClkRoutine)(sysAuxClkArg);
    }
}

/*
 * sysAuxClkDisable - disable aux clock.
 */
void
sysAuxClkDisable(void)
{
    if (sysAuxClkRunning == TRUE) {
        sysRegWrite(AR5311_RELOAD, 0xffffffff);
        miscIMR &= ~IMR_TIMER;
        sysRegWrite(AR5311_IMR, miscIMR);
        sysWbFlush();

        sysAuxClkRunning = FALSE;
    }
}

/*
 * sysAuxClkEnable - enable aux clock.
 */
void
sysAuxClkEnable(void)
{
    UINT32 reload;

    if (sysAuxClkRunning == FALSE) {
        reload = sysBoardData.sysFreq / sysAuxClkTicksPerSecond;
        sysRegWrite(AR5311_TIMER, reload);
        sysRegWrite(AR5311_RELOAD, reload);

        miscIMR |= IMR_TIMER;
        sysRegWrite(AR5311_IMR, miscIMR);
        sysWbFlush();

        sysAuxClkRunning = TRUE;
    }
}

/*
 * sysAuxClkRateGet - return aux clock rate in ticks per second.
 */
int
sysAuxClkRateGet(void)
{
    return sysAuxClkTicksPerSecond;
}

/*
 * sysAuxClkRateSet - set aux clock rate in ticks per second.
 */
STATUS
sysAuxClkRateSet(int tps)
{
    if ((tps < AUX_CLK_RATE_MIN) || (tps > AUX_CLK_RATE_MAX)) {
        return ERROR;
    }

    sysAuxClkTicksPerSecond = tps;

    if (sysAuxClkRunning) {
        sysAuxClkDisable();
        sysAuxClkEnable();
    }

    return OK;
}
#endif /* INCLUDE_AUX_CLK */

/*
 * Spirit performance counter support.  Only a few combinations
 * supported for simplicity.
 *
 * Assumes "perf debug enable" in AR5311_CONFIG is clear.
 */
unsigned long long sysPerfCnt0, sysPerfCnt1;
static int sysPerfMode;
void
sysPerfStart(int mode)
{
    UINT32 m0, m1;

    sysPerfMode = mode;

    switch(mode) {
    case SYSPERF_DC:
        m0 = PERF_DATAHIT;
        m1 = PERF_DATAMISS;
        break;
    case SYSPERF_IC:
        m0 = PERF_INSTHIT;
        m1 = PERF_INSTMISS;
        break;
    case SYSPERF_WB:
        m0 = PERF_WBHIT;
        m1 = PERF_WBMISS;
        break;
    case SYSPERF_CH:
        m0 = PERF_INSTHIT;
        m1 = PERF_DATAHIT;
        break;
    case SYSPERF_CM:
        m0 = PERF_INSTMISS;
        m1 = PERF_DATAMISS;
        break;
    default:
        m0 = m1 = 0;
        break;
    }

    sysRegWrite(AR5311_PERFMSK0, m0);
    sysRegWrite(AR5311_PERFMSK1, m1);
    (void)sysRegRead(AR5311_PERFCNT0);          /* read to clear */
    (void)sysRegRead(AR5311_PERFCNT1);
    sysPerfCnt0 = sysPerfCnt1 = 0;
}

void
sysPerfUpdate(void)
{
    UINT32 p0 = sysRegRead(AR5311_PERFCNT0);
    UINT32 p1 = sysRegRead(AR5311_PERFCNT1);

    sysPerfCnt0 += p0;
    sysPerfCnt1 += p1;
}

void
sysPerfShow(void)
{
    UINT32 p0 = sysRegRead(AR5311_PERFCNT0);
    UINT32 p1 = sysRegRead(AR5311_PERFCNT1);
    char *str;

    sysPerfCnt0 += p0;
    sysPerfCnt1 += p1;

    switch (sysPerfMode) {
    case SYSPERF_DC:
        str = "data cache: hits 0x%08x%08x (%x) misses 0x%08x%08x (%x)\n";
        break;
    case SYSPERF_IC:
        str = "instr cache: hits 0x%08x%08x (%x) misses 0x%08x%08x (%x)\n";
        break;
    case SYSPERF_WB:
        str = "write buffer: hits 0x%08x%08x (%x) misses 0x%08x%08x (%x)\n";
        break;
    case SYSPERF_CH:
        str = "cache hits: instr 0x%08x%08x (%x) data 0x%08x%08x (%x)\n";
        break;
    case SYSPERF_CM:
        str = "cache misses: instr 0x%08x%08x (%x) data 0x%08x%08x (%x)\n";
        break;
    default:
        str = "unknown mode!\n";
        break;
    }
    /* %llx still does not work */
    printf(str,
           (UINT32)(sysPerfCnt0 >> 32),
           (UINT32)(sysPerfCnt0 & 0xffffffff),
           p0,
           (UINT32)(sysPerfCnt1 >> 32),
           (UINT32)(sysPerfCnt1 & 0xffffffff),
           p1);
}

void
sysPerfStop(void)
{
    sysRegWrite(AR5311_PERFMSK0, 0);
    sysRegWrite(AR5311_PERFMSK1, 0);
}
