/*
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * Atheros ar5311 Integrated AP BSP network driver glue.
 *      - Handles spirit specific set-up before handing
 *        off to the more platform independent networking
 *        driver.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/sysNet.c#1 $"

#include "configNet.h"
#include "ae5311End.h"
#ifdef BUILD_AP			/* BSP only build only includes enet */
#include "wlandrv.h"
#include "ar5hwcEnd.h"
#include "ar5211/ar5211Reg.h"
#endif

END_TBL_ENTRY endDevTbl[IP_MAX_UNITS+1];
char sysEnetLoadStr[IP_MAX_UNITS][80];
int endDevIndex = 0;

/*
 * sysEnetInit - early init of ethernet.
 */
void
sysEnetInit(int unit, UINT32 base, int vec)
{
    UINT32 reg;

    /* Bring requested unit out of reset */
    reg = sysRegRead(AR5311_RESET);
    sysRegWrite(AR5311_RESET, reg | ((RESET_ENET0|RESET_EPHY0) << unit));
    sysUDelay(10);
    sysRegWrite(AR5311_RESET, reg & (~(RESET_ENET0|RESET_EPHY0) << unit));
    (void)sysRegRead(AR5311_RESET);     /* read after */

    /* Ensure DMA interface, and clocks are enabled */
    sysRegWrite(AR5311_ENABLE,
                sysRegRead(AR5311_ENABLE) | (ENABLE_ENET0 << unit));

    /* unmask interrupt */
    srIMR |= (AR5311_INTR_ENET0 << unit);

    /*
     * Create end load string for the driver, and plug
     *	"<devAdrs>:<inum>:<ivec>:<memBase>:<memSize>:<userFlags>"
     */
    /*	sprintf(sysEnetLoadStr[endDevIndex], "0x%08x:0x%x:0x%x:-1:0:0x0",
        base, vec, (int)INUM_TO_IVEC(vec));
     */
    /* FOO SJK FOR AE bringup. Resolve different initstring formats. */
    /*  "unit:dmaBufAdr:dmaBufSize:txDescCount:rxDescCount:clCount" */
    sprintf(sysEnetLoadStr[endDevIndex], AE_LOAD_STRING_BOOT);

    endDevTbl[endDevIndex].unit = unit;
    endDevTbl[endDevIndex].endLoadFunc = ae5311EndLoad;
    endDevTbl[endDevIndex].endLoadString = sysEnetLoadStr[endDevIndex];
    endDevTbl[endDevIndex].endLoan = TRUE;
    endDevTbl[endDevIndex].processed = FALSE;
    endDevIndex++;
}

/*
 * sysWlanInit - early init of wlan device.
 */
void
sysWlanInit(UINT32 base, int vec)
{
#ifdef BUILD_AP
    UINT32 reg;

    /* Enable global swapping so this looks like a normal BE system */
    reg = sysRegRead(AR5311_CONFIG);
    reg |= CONFIG_WLANSWAP;
    sysRegWrite(AR5311_CONFIG, reg);

    /* Bring wlan out of reset */
    reg = sysRegRead(AR5311_RESET);
    sysRegWrite(AR5311_RESET, reg | RESET_WLAN);
    (void)sysRegRead(AR5311_RESET);     /* read after */
    sysUDelay(10);
    sysRegWrite(AR5311_RESET, reg &
                ~(RESET_WLAN | RESET_WARM_WLAN | RESET_WARM_WBB));
    (void)sysRegRead(AR5311_RESET);     /* read after */

    /* Ensure DMA interface is enabled */
    sysRegWrite(AR5311_ENABLE, sysRegRead(AR5311_ENABLE) | ENABLE_WMAC);

    /* Unmask interrupt */
    srIMR |= AR5311_INTR_WLAN;

    /*
     * Create end load string for the driver.  Always use the defaults.
     */
    strcpy(sysEnetLoadStr[endDevIndex], "0x00:0x00:0:0:0:0");

    endDevTbl[endDevIndex].unit = 0;
    endDevTbl[endDevIndex].endLoadFunc = ar5hwcEndLoad;
    endDevTbl[endDevIndex].endLoadString = sysEnetLoadStr[endDevIndex];
    endDevTbl[endDevIndex].endLoan = TRUE;
    endDevTbl[endDevIndex].processed = FALSE;
    endDevIndex++;

    /*
     * Fill in arParam to the best of our ability, even though this
     * is also in the string.
     */
    arParam.pciMemBase = base;
    arParam.ilevel = AR5311_INTR_WLAN;
    arParam.ivec = vec;
    arParam.pciVendorID = ATHEROS_VENDOR_ID;
    arParam.pciDeviceID = sysBoardData.pciId;
    arParam.pciSubVendorID = ATHEROS_VENDOR_ID;
    arParam.pciSubDeviceID = sysBoardData.pciId;
    arParam.pciRevisionID = 1;
#endif /* BUILD_AP */
}

/*
 * sysNetDone - terminate endDevTbl[].
 */
void
sysNetDone(void)
{
    endDevTbl[endDevIndex].unit = -1;
    endDevTbl[endDevIndex].endLoadFunc = NULL;
    endDevTbl[endDevIndex].endLoadString = NULL;
    endDevTbl[endDevIndex].endLoan = FALSE;
    endDevTbl[endDevIndex].processed = FALSE;
}
