/* sysSerial.c - ar5311 BSP serial device initialization */

/* Copyright 1984-1997 Wind River Systems, Inc. */

/*
 *  Copyright  2001-2002 Atheros Communications, Inc.,  All Rights Reserved.
 */

/*
 * Serial port glue.
 */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/ap/os/vxworks/target/config/ar5311/sysSerial.c#1 $"

#include "vxWorks.h"
#include "config.h"
#include "intLib.h"
#include "iv.h"
#include "sysLib.h"
#include "ar5311reg.h"

#include "ar5311Sio.c"		/* Atheros flavor of National 16550 driver */

#if N_SIO_CHANNELS != 2
#error sysSerial supports 2 ports (2nd may not be connected).
#endif

LOCAL AR5311SIO_CHAN ar5311SioChan[N_SIO_CHANNELS];

/*
 * Array of pointers to all serial channels configured in system.
 * See sioChanGet(). It is this array that maps channel pointers
 * to standard device names.  The first entry will become "/tyCo/0",
 * the second "/tyCo/1", and so forth.
 */
LOCAL SIO_CHAN *sysSioChans[N_SIO_CHANNELS];

/*
 * sysSerialHwInit - initialize the BSP serial devices to a quiescent state
 *
 * This routine initializes the BSP serial device descriptors and puts the
 * devices in a quiescent state.  It is called from sysHwInit() with
 * interrupts locked.
 *
 * RETURNS: N/A
 *
 * SEE ALSO: sysHwInit()
 */
void
sysSerialHwInit(void)
{
    UINT32 reg;
    int xtal;

    reg = sysRegRead(AR5311_CLOCKCTL);
#ifdef BD_BRINGUP
    if (sysBoardData.config & BD_EXTUARTCLK) {
        /* Use external 14.7456Mhz crystal */
        xtal = 14745600;
        reg |= CLOCKCTL_UART0 | CLOCKCTL_UART1;
    } else
#endif
    {
        /* Use internal clock */
        xtal = sysBoardData.sysFreq;
        reg &= ~(CLOCKCTL_UART0 | CLOCKCTL_UART1);
    }
    sysRegWrite(AR5311_CLOCKCTL, reg);

    /* Ensure both UARTs are reset and ready */
    reg = sysRegRead(AR5311_RESET);
    sysRegWrite(AR5311_RESET, reg | RESET_UART0 | RESET_UART1);
    sysUDelay(10);
    reg &= ~(RESET_UART0|RESET_UART1);
    sysRegWrite(AR5311_RESET, reg);

    /*
     * Always init console port even if not stuffed so we have
     * a console drain.  If uart is clamed not to be stuffed
     * drop output until we see some input.  This keeps the
     * board driving unterminated wires that may cause
     * compliance problems.
     */
    sysSioChans[0] = &ar5311SioChan[0].sio;
#if CONSOLE_ADDR == AR5311_UART0
    ar5311SioChan[0].regs = (volatile UINT32 *)PHYS_TO_K1(AR5311_UART0);
    ar5311SioChan[0].dropOutput = (sysBoardData.config & BD_UART0) == 0;
    ar5311SioChan[0].dmaTx = TRUE;
#else
    ar5311SioChan[0].regs = (volatile UINT32 *)PHYS_TO_K1(AR5311_UART1);
    ar5311SioChan[0].dropOutput = (sysBoardData.config & BD_UART1) == 0;
    ar5311SioChan[0].dmaTx = FALSE;
#endif
    ar5311SioChan[0].dmaRx = FALSE;		/* no dma rx on console */
    ar5311SioChan[0].baudRate = 9600;
    ar5311SioChan[0].xtal = xtal;
    ar5311SioDevInit(&ar5311SioChan[0]);

    /*
     * If UART is disabled, still want to init as other code
     * expects to see sysSioChan[1] be valid.  Just disable
     * output and the DMA engine for this configuration.
     */
    sysSioChans[1] = &ar5311SioChan[1].sio;
#if CONSOLE_ADDR == AR5311_UART0
    ar5311SioChan[1].regs = (volatile UINT32 *)PHYS_TO_K1(AR5311_UART1);
    ar5311SioChan[1].dmaTx = FALSE;
    ar5311SioChan[1].dmaRx = FALSE;
    ar5311SioChan[1].dropOutput = (sysBoardData.config & BD_UART1) == 0;
#else
    ar5311SioChan[1].regs = (volatile UINT32 *)PHYS_TO_K1(AR5311_UART0);
    ar5311SioChan[1].dmaTx = (sysBoardData.config & BD_UART0) != 0;
    ar5311SioChan[1].dmaRx = (sysBoardData.config & BD_UART0) != 0;
    ar5311SioChan[1].dropOutput = (sysBoardData.config & BD_UART0) == 0;
#endif
    ar5311SioChan[1].baudRate = 9600;
    ar5311SioChan[1].xtal = xtal;
    ar5311SioDevInit(&ar5311SioChan[1]);
}

/*
 * sysSerialHwInit2 - connect BSP serial device interrupts
 *
 * This routine connects the BSP serial device interrupts.  It is called from
 * sysHwInit2().
 *
 * Serial device interrupts cannot be connected in sysSerialHwInit() because
 * the  kernel memory allocator is not initialized at that point and
 * intConnect() calls malloc().
 *
 * RETURNS: N/A
 *
 * SEE ALSO: sysHwInit2()
 */
void
sysSerialHwInit2(void)
{
    /*
     * Handle any stitching of vxworks serial port to interrupt vector
     */
    switch ((int)ar5311SioChan[0].regs) {
    case PHYS_TO_K1(AR5311_UART0):
        if (ar5311SioChan[0].dmaTx == TRUE) {
            ar5311SioDmaInit(&ar5311SioChan[0]);
            (void)intConnect(INUM_TO_IVEC(IV_MISCDMA_VEC),
                             ar5311SioDmaInt, (int)&ar5311SioChan[0]);
            srIMR |= AR5311_INTR_MISCDMA;
        }
        (void)intConnect(INUM_TO_IVEC(IV_UART0_VEC),
                         ar5311SioInt, (int)&ar5311SioChan[0]);
        miscIMR |= IMR_UART0;
        break;
    case PHYS_TO_K1(AR5311_UART1):
        (void)intConnect(INUM_TO_IVEC(IV_UART1_VEC),
                         ar5311SioInt, (int)&ar5311SioChan[0]);
        miscIMR |= IMR_UART1;
        break;
    }

    switch ((int)ar5311SioChan[1].regs) {
    case PHYS_TO_K1(AR5311_UART0):
        if (ar5311SioChan[1].dmaTx == TRUE) {
            ar5311SioDmaInit(&ar5311SioChan[1]);
            (void)intConnect(INUM_TO_IVEC(IV_MISCDMA_VEC),
                             ar5311SioDmaInt, (int)&ar5311SioChan[1]);
            srIMR |= AR5311_INTR_MISCDMA;
        }
        (void)intConnect(INUM_TO_IVEC(IV_UART0_VEC),
                         ar5311SioInt, (int)&ar5311SioChan[1]);
        miscIMR |= IMR_UART0;
        break;
    case PHYS_TO_K1(AR5311_UART1):
        (void)intConnect(INUM_TO_IVEC(IV_UART1_VEC),
                         ar5311SioInt, (int)&ar5311SioChan[1]);
        miscIMR |= IMR_UART1;
        break;
    }
}

/*
 * sysSerialChanGet - get the SIO_CHAN device associated with a serial channel
 *
 * This routine returns a pointer to the SIO_CHAN device associated
 * with a specified serial channel.  It is called by usrRoot() to obtain
 * pointers when creating the system serial devices, `/tyCo/x'.  It
 * is also used by the WDB agent to locate its serial channel.
 *
 * RETURNS: A pointer to the SIO_CHAN structure for the channel, or ERROR
 * if the channel is invalid.
 */
SIO_CHAN *
sysSerialChanGet(int channel)
{
    if ((channel >= 0) && (channel < N_SIO_CHANNELS) &&
        sysSioChans[channel]) {
        return sysSioChans[channel];
    }
    return NULL;
}
