/*
 *  Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * vxdrv.h - vxWorks specific function wrappers.
 *
 * "ACI $Id: //depot/sw/releases/2.0.1/src/include/vxdrv.h#1 $"
 */

#ifndef __INCvxdrvh
#define __INCvxdrvh

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include "wlantype.h"
#include "semLib.h"
#include "cacheLib.h"
#include "tickLib.h"
#include "string.h"
#include "assert.h"
#include "apdefs.h"

#ifdef DEBUG    /* allow static functions to be exported for stack backtrace */
#undef LOCAL
#define LOCAL
#endif

typedef END_OBJ OS_DEV_INFO;

#define A_MALLOC(a)                     (malloc(a))
#define A_FREE(a)                       (free(a))
#define A_DRIVER_MALLOC(a, b, c)        (malloc(a))
#define A_DRIVER_FREE(a, b, c)          (free(a))
#define A_MEM_SET(addr, value, size)    memset((char *)(addr), (int)(value), (int)(size))
#define A_MEM_ZERO(addr, size)          bzero((char *)(addr), (int)(size))
#define A_DRIVER_BCOPY(from, to, len)   bcopy((char *)(from), (char *)(to), (int)(len))
#define A_BCOPY(a, b, n)                bcopy((char *)(a), (char *)(b), (int)(n))
#define A_BCOMP(p1, p2, n)              bcmp((char *)(p1), (char *)(p2), (int)(n))

/*
 * Words (16b) should be aligned correctly on all platforms, so it
 * can be used even on platforms (mips) with strict alignment
 * requirements.
 */
#define A_MACADDR_COPY(from, to) \
    do { \
        (to)->words[0] = (from)->words[0]; \
        (to)->words[1] = (from)->words[1]; \
        (to)->words[2] = (from)->words[2]; \
    } while (0)

#define A_MACADDR_COMP(m1, m2) \
    ((((m1)->words[2] == (m2)->words[2]) && \
      ((m1)->words[1] == (m2)->words[1]) && \
      ((m1)->words[0] == (m2)->words[0])) == 0)

/*
 *  SIB entry and bridge aging period is 300 seconds
 */
#define DEFAULT_AGING_PERIOD            (60 * 5)

/*
 * Get the number of millisecond ticks since the system was started.
 *
 * For the AP, absolute accuracy is not all that important.  Use
 * an approximation of the conversion of ticks @ 60Hz to that does
 * not require any expensive math (multiply or divide).
 *
 * 60 ticks -> 997ms, or a 0.3% shortfall error, although there are
 * cases where the error is worse.
 *
 * vxAbsTicks.lower the same as getTick().
 */
#define A_MS_TICKGET()  \
    ({A_UINT32 ts = vxAbsTicks.lower;  (ts << 4) + (ts >> 1) + (ts >> 3);})

#define A_ROUNDUP(x, y)         ((((x)+((y)-1))/(y))*(y))

#define ASSERT assert

/*
 * Cache flushing and virtual address translation support.  vxWorks macros
 * handle things cleanly but with some overhead, so optimize for supported
 * platforms.
 *   - A_DATA_() routines are for frame data buffers.
 *   - A_DESC_() macros are for DMA descriptors.  Currently
 *     hardware descriptors are cached as they share memory
 *     with the software frame description.
 *
 * SWCOHERENCY   - descriptors and data are cached
 * SWCOHERENCYWB - SWCOHERENCY with a write back cache
 */
#if defined(AR5311) || (USER_D_CACHE_MODE & CACHE_WRITETHROUGH)
/*
 * AR5311 uses a write back cache, with it's own lower overhead
 * cache invalidation routine.  V != P, but upper bits are
 * ignored by the DMA bus so a null V2P/P2V macros are used.
 */
#define SWCOHERENCY
#include "ar5311.h"
#define A_DATA_CACHE_INVAL(addr, len) \
    cacheDataInvalidate(addr.ptr, (len))
#define A_DATA_CACHE_FLUSH(addr, len)
#define A_DATA_V2P(addr)                ((A_UINT32)(addr))
#define A_DATA_P2V(addr)                (addr)
#define A_DESC_CACHE_INVAL(addr) \
    cacheDescInvalidate((void *)(&(addr)->hwStatus))
#define A_DESC_CACHE_FLUSH(addr)
#define A_DESC_V2P(addr)                ((A_UINT32)(addr))
#define A_DESC_P2V(addr)                (addr)
#define A_PIPEFLUSH()                   sysWbFlush()

#elif defined(PPC8240) || (USER_D_CACHE_MODE & CACHE_SNOOP_ENABLE)
/*
 * MPC8245 is a fully coherent writeback cache with V==P.
 */
#define A_DATA_CACHE_INVAL(addr, len)
#define A_DATA_CACHE_FLUSH(addr, len)
#define A_DATA_V2P(addr)                ((A_UINT32)(addr))
#define A_DATA_P2V(addr)                (addr)
#define A_DESC_CACHE_INVAL(addr)
#define A_DESC_CACHE_FLUSH(addr)
#define A_DESC_V2P(addr)                ((A_UINT32)(addr))
#define A_DESC_P2V(addr)                (addr)
#define A_PIPEFLUSH()

#elif defined(CACHE_PROTO_COPYBACK) || (USER_D_CACHE_MODE & CACHE_COPYBACK)
/*
 * Sample writeback cache implementation keeping the desc/buffers in
 * the cached.  This code has not been tested, so it is not enabled
 * by default.
 */
#define SWCOHERENCY
#define SWCOHERENCYWB
#define A_DATA_CACHE_INVAL(addr, len)   CACHE_USER_INVALIDATE(addr, len)
#define A_DATA_CACHE_FLUSH(addr, len)   CACHE_USER_FLUSH(addr, len)
#define A_DATA_V2P(addr) \
    ((A_UINT32)CACHE_DRV_VIRT_TO_PHYS(&cacheUserFuncs, addr))
#define A_DATA_P2V(addr) \
    CACHE_DRV_PHYS_TO_VIRT(&cacheUserFuncs, addr)
#define A_DESC_CACHE_INVAL(addr) \
    A_DATA_CACHE_INVAL(addr, 32)
#define A_DESC_CACHE_FLUSH(addr) \
    A_DATA_CACHE_FLUSH(addr, 32)
#define A_DESC_V2P(addr)                A_DATA_V2P(addr)
#define A_DESC_P2V(addr)                A_DATA_P2V(addr)
#define A_PIPEFLUSH()                   CACHE_PIPE_FLUSH()

#else

/*
 * Sample code to use cacheDmaAlloc() which is uncached on non-coherent
 * platforms.  This is the default case for non-coherent copyback
 * caches.  This code has not been extensively tested, but should
 * be safe on most platforms.
 */
#define VX_CACHE_DMA_MALLOC
#define SWCOHERENCY
#define A_DATA_CACHE_INVAL(addr, len) \
    CACHE_DRV_INVAL(&cacheDmaFuncs, addr, len)
#define A_DATA_CACHE_FLUSH(addr, len) \
    CACHE_DRV_FLUSH(&cacheDmaFuncs, addr, len)
#define A_DATA_V2P(addr) \
    ((A_UINT32)CACHE_DRV_VIRT_TO_PHYS(&cacheDmaFuncs, addr))
#define A_DATA_P2V(addr) \
    CACHE_DRV_PHYS_TO_VIRT(&cacheDmaFuncs, addr)
#define A_DESC_CACHE_INVAL(addr) \
    A_DATA_CACHE_INVAL(addr, 32)
#define A_DESC_CACHE_FLUSH(addr) \
    A_DATA_CACHE_FLUSH(addr, 32)
#define A_DESC_V2P(addr)                A_DATA_V2P(addr)
#define A_DESC_P2V(addr)                A_DATA_P2V(addr)
#define A_PIPEFLUSH()                   CACHE_PIPE_FLUSH()

#endif /* PLATFORM */

/*
 * Locking macros
 */
#define CREATE_LOCKED   1
#define CREATE_UNLOCKED 0
#ifdef DEBUG
#define A_SEM_INIT(sem_val, sem_type, make_locked)      \
    do { \
        (sem_val) = semBCreate((sem_type), (make_locked) ? SEM_EMPTY : SEM_FULL); \
         if ((sem_val) == NULL) \
            apPanic("Could not allocate lock."); \
    } while (0)
#else
#define A_SEM_INIT(sem_val, sem_type, make_locked)      \
    do { \
        (sem_val) = semBCreate((sem_type), (make_locked) ? SEM_EMPTY : SEM_FULL); \
         if ((sem_val) == NULL) \
            uiPrintf("Could not allocate lock."); \
    } while (0)
#endif /* not DEBUG */

#define A_SEM_LOCK(sem, timeout)    semTake(sem, timeout)
#define A_SEM_UNLOCK(sem)           semGive(sem)
#define A_SEM_DELETE(sem)           semDelete(sem)

#define A_SIB_TAB_LOCK(sibTab)      semTake(sibTab->sibTableSem, WAIT_FOREVER)

#define A_SIB_TAB_UNLOCK(sibTab)    semGive(sibTab->sibTableSem)

#define A_SIB_ENTRY_LOCK(sib)       do { \
        semTake(sib->sibLock, WAIT_FOREVER); \
        ASSERT(sib->used & SIB_VALID); \
    } while (0)

#define A_SIB_ENTRY_UNLOCK(sib)     semGive(sib->sibLock)

#define A_TIMEOUT(ptimer, func, period, param)
#define A_UNTIMEOUT(handle)

#ifdef __cplusplus
}
#endif

#endif /* not __INCvxdrvh */
