/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * TODO: comment
 */

#ident  "ACI $Id: //depot/sw/releases/2.0.1/src/common/wlanbeacon.c#1 $"

/*
 * TODO: fix this DESCRIPTION
 *
 * This module implements the beacon related functions for AP only
 *
 * The following functions are needed for AP to generate beacons.
 *
 * 1) TIM update
 *
 * This function updates the TIM whenever the power saving mode changes
 * for any assocaited STAs detected
 *
 * 2) Beacon generator
 *
 * This functions is called at the AP system initialization stage and
 * from the software alert ISR. It takes a snapshot of the TIM and
 * create the beacon frame in Atheros descriptor, put it into
 * the corresponding TX queue.
 *
 */
#include "linuxdrv.h"
						   
#include "wlanbeacon.h"
#include "wlanframe.h"
#include "wlanproto.h"
#include "wlanext.h"
#include "wlansmeext.h"
#include "wlansme.h"
#include "wlanchannel.h"

#include "halApi.h"
#include "ui.h"
#include "display.h"

//#ifdef BUILD_AP
#include "apmem.h"      /* createBuffAndDesc() */
//#else
//fadi #include "hwext.h"
//fadi  #endif

#include "atheros_wlan.h"

extern struct pci_dev * gPciDev;

/* Forward declarations */
static A_UINT16
buildTim(BEACON_INFO *pBeaconInfo, A_BOOL cab);

/*******************************************************************************
* wlanApBeaconGen
*
* This routine generates the beacon frame based upon the snapshot of the
* current TIM element.  This routine is called only by the AP.
*
* Assumes: procedure can be called multiple times
*
* Returns: A_OK, A_EBUSY
*/
A_STATUS
wlanApBeaconGen(WLAN_DEV_INFO *pDev)
{
    ATHEROS_DESC    *pBeaconDesc = NULL, *pFirstCabDesc = NULL;
    QUEUE_DETAILS   *pCabQueue = NULL, *pPsQueue = NULL;
    BEACON_INFO     *pBeaconInfo = NULL;
    struct net_device *dev = NULL;

    ASSERT(pDev && pDev->pBeaconInfo);
    pBeaconInfo = pDev->pBeaconInfo;
    ASSERT(pBeaconInfo->beaconEnabled);

    /* See if the beacon & CAB queue(s) have stopped. */
    if (halWaitForBeaconDone(pDev) != A_OK) {
        return A_EBUSY;
    }

    pBeaconDesc   = pBeaconInfo->pBeaconDesc;
    pPsQueue      = &pDev->localSta->txPSQueue;
    pCabQueue     = &pDev->altQueue;
    pFirstCabDesc = pCabQueue->pDescQueueHead;

    /*
     * Walk through the CAB queue descriptors looking for
     * pending BC/MC packets that haven't gone out yet.
     */
    while (pFirstCabDesc != NULL) {
        A_DESC_CACHE_INVAL(pFirstCabDesc);
        if (!pFirstCabDesc->pTxLastDesc->hwStatus.tx.done) {
            break;
        }
        pFirstCabDesc = pFirstCabDesc->pTxLastDesc->pNextVirtPtr;
    }

    /* If this is a DTIM, send any buffered BC/MC frames */
    if (pBeaconInfo->pTim->dtimCount == 0) {

#ifdef PARTHA
        
        /* Take care of power saving multicast and broadcast packets first. */
        if (pPsQueue->pDescQueueHead != NULL) {

            // these descs should still be nulled by the chipreset() rtn
            printk("SHOULD NOT HAVE POWER SAVE MULTICAST PACKETS HERE !!!\n");

            if (pFirstCabDesc != NULL) {
                /*
                 * The last frame of still pending CAB traffic
                 * gets its more bit set.
                 */
                pCabQueue->pDescQueueTail->pTxFirstDesc->pBufferVirtPtr.headerLittle->frameControl.moreData = 1;
                A_DATA_CACHE_FLUSH(pAltQueue->pDescQueueTail->pTxFirstDesc->pBufferVirtPtr,
                                   sizeof(A_UINT16));
            } else {
                /* There are no pending PS frames in the CAB queue */
                pFirstCabDesc = pPsQueue->pDescQueueHead;
            }

            /* Update the CAB queue head and tail. */

            if (pCabQueue->pDescQueueHead == NULL) {
                /*
                 * All completed descriptors have been released so it
                 * is OK to set new head and tail.
                 */
                pCabQueue->pDescQueueHead = pPsQueue->pDescQueueHead;
                pCabQueue->pDescQueueTail = pPsQueue->pDescQueueTail;
            } else {
                /* Some DMA completed descriptors may not be released yet,
                 * so don't update the CAB queue head.  Only update the tail.
                 */
                pCabQueue->pDescQueueTail->pNextVirtPtr = pPsQueue->pDescQueueHead;
                pCabQueue->pDescQueueTail->nextPhysPtr  = pPsQueue->pDescQueueHead->thisPhysPtr;
                A_DESC_CACHE_FLUSH(pCabQueue->pDescQueueTail);
                pCabQueue->pDescQueueTail = pPsQueue->pDescQueueTail;
            }

            /* Flush the PS queue. */
            pPsQueue->pDescQueueHead = NULL;
            pPsQueue->pDescQueueTail = NULL;
            pPsQueue->qFrameCount    = 0;
        }
#else
        if (NULL != gPciDev) {
            dev = (struct net_device *)pci_get_drvdata(gPciDev);
            if (dev) {
                atheros_local_info_t *local = (atheros_local_info_t *)dev->priv;
                if (local && (local->func->asap.dtim_handler))
                    local->func->asap.dtim_handler(dev);
            }
        }
#endif
    }

    /* Rebuild the TIM element */
    pBeaconInfo->TIMLen = buildTim(pBeaconInfo, (A_BOOL)(pFirstCabDesc != NULL));

    /* Put the country code element after the TIM */
    mlmeCountryCodeElementFill (pDev, (A_CHAR *)pBeaconInfo->pTim + pBeaconInfo->TIMLen);

    /*
     * Recompute the beacon frame length
     * Update the beacon descriptor
     */
    pBeaconInfo->frameLen = pBeaconInfo->nonTIMLen +
                            pBeaconInfo->TIMLen + mlmeCountryCodeElementLength(pDev);

    halUpdateBeaconDesc(pDev, pBeaconDesc, pBeaconInfo->frameLen);

    halQueueBeaconAndCab(pDev, pBeaconDesc, pFirstCabDesc);

    pBeaconInfo->txBeaconCount++;

	//printk("...Done wlanApBeaconGen[frameLen=%d][beacon#%d]\n", pBeaconInfo->frameLen,
    //       pBeaconInfo->txBeaconCount);

    return A_OK;
}

/*******************************************************************************
* wlanBeaconInit
*
* This routine allocates the global beacon descriptor, buffers for beacon and
* TIM element (if AP) or IBSS element (if client).
*
* RETURNS: OK if succesful
*          ERROR if unsuccessful
*/
A_STATUS
wlanBeaconInit(WLAN_DEV_INFO *pDev, A_UINT32 tsf)
{
    ATHEROS_DESC        *pBeaconDesc;
    WLAN_FRAME_BEACON   *pBeaconBuf;
    BEACON_INFO         *pBeaconInfo;
    A_UINT16            bufLen, beaconLen;
    A_CHAR              *pField;
    A_BOOL              isAp, useCfp, isIbss, useCnp;
    BSSDESCR            *pBss;
    A_STATUS            status;

    ASSERT(pDev && pDev->bssDescr && pDev->localSta);

    pBss   = pDev->bssDescr;
    isAp   = (pDev->localSta->serviceType == WLAN_AP_SERVICE);
    useCfp = VALID_CF_ELEMENT(&pBss->cfParamSet);
    useCnp = VALID_UCSE_ELEMENT(&pBss->bssUcseSet);
    isIbss = pBss->bsstype == INDEPENDENT_BSS;

    /*
     * This is a temporary hack since we currently do not have the ability
     * to disable/enable beacons outside of wlanBeaconFree() and
     * wlanBeaconInit(). (see bug 4095)
     */
    if (pDev->pBeaconInfo != NULL) {
        wlanBeaconFree(pDev);
    }

    /* Allocate and zero-out the beacon information structure. */
    ASSERT(pDev->pBeaconInfo == NULL);
    pBeaconInfo = (BEACON_INFO *)A_DRIVER_MALLOC(sizeof(BEACON_INFO), 0, 0);
    if (pBeaconInfo == NULL) {
        uiPrintf("wlanBeaconInit: Error allocating beacon struct!\n");
        return A_ERROR;
    }
    A_MEM_ZERO(pBeaconInfo, sizeof(BEACON_INFO));
    pDev->pBeaconInfo = pBeaconInfo;

    /* Calculate the length of beacon frame. */
    bufLen = sizeof(WLAN_MGT_MAC_HEADER) +
                   sizeof(WLAN_TIMESTAMP) +
                   sizeof(A_UINT16) +               /* beacon interval */
                   sizeof(CAP_INFO) +
                   INFO_ELEMENT_SIZE(pBss->ssid) +
                   INFO_ELEMENT_SIZE(pBss->bssRateSet);

    bufLen += sizeof(DS_PARAM_SET);

    if (useCfp) {
        bufLen += sizeof(CF_PARAM_SET);
    }
    if (useCnp && isAp) {
        bufLen += INFO_ELEMENT_SIZE(pBss->bssUcseSet);
    }

    beaconLen = bufLen;   /* Beacon frame length, but w/o TIM element */

    if (isAp) {
        /* Add max TIM_ELEMENT size and country code element size */
        bufLen += sizeof(TIM_ELEMENT) + mlmeCountryCodeElementLength (pDev);
    } else {
        bufLen += sizeof(IBSS_PARAM_SET);
        beaconLen = bufLen;
    }

    pBeaconInfo->nonTIMLen = beaconLen;
    pBeaconInfo->TIMLen = 0;

    /* Allocate the beacon buffer and descriptor. */
    status = createBuffandDesc(pDev, bufLen, &pBeaconDesc);
    if (status != A_OK) {
        A_DRIVER_FREE(pBeaconInfo, sizeof(BEACON_INFO), 0);
        pDev->pBeaconInfo = NULL;
        uiPrintf("wlanBeaconInit: cannot allocate beacon!\n");
        return A_ERROR;
    }

    pBeaconBuf = pBeaconDesc->pBufferVirtPtr.beacon;
    pDev->pBeaconInfo->pBeaconDesc = pBeaconDesc;
	ASSERT (pBeaconBuf);
	ASSERT (pBeaconDesc);

    /* Initialize the beacon fields. */
    A_MEM_ZERO((char *)pBeaconBuf, bufLen);

    /* Fill in the MAC header. */
    mgtFrameHdrFill(pDev, (WLAN_MGT_MAC_HEADER *)pBeaconBuf,
                    &beaconBroadcastAddr, SUBT_BEACON);

    /* Fill in the frame body. */
    /* (timestamp value is supplied by HW) */
    pBeaconBuf->beaconInterval = pBss->beaconInterval;
    pBeaconBuf->capInfo        = pBss->capabilityInfo;

    /* Fill in the SSID and Supported Rates. */
    pField = (A_CHAR *)pBeaconBuf +
             sizeof(WLAN_MGT_MAC_HEADER) +
             sizeof(WLAN_TIMESTAMP) +
             sizeof(A_UINT16) +             /* beacon interval */
             sizeof(CAP_INFO);
    pField = (A_CHAR *)mlmeElementCopy((INFO_ELEMENT *)&pBss->ssid,
                                       (INFO_ELEMENT *)pField);
    pField = (A_CHAR *)mlmeElementCopy((INFO_ELEMENT *)&pBss->bssRateSet,
                                       (INFO_ELEMENT *)pField);

    /*
     * Build the DS parameter set and copy it into the beacon.
     * We'll do this for both 11b and 11a. Cisco uses it for 11a too.
     */
    pBss->dsParamSet.elementID = ELE_DS_PARAM_SET;
    pBss->dsParamSet.length = 1;
    pBss->dsParamSet.currentChannel =
        wlanConvertGHztoCh(pBss->pChannel->channel, pBss->pChannel->channelFlags);
    pField = (A_CHAR *)mlmeElementCopy((INFO_ELEMENT *)&pBss->dsParamSet,
                                       (INFO_ELEMENT *)pField);
    if (useCfp) {
        pBeaconInfo->pCfPtr = pField;
        pField += sizeof(CF_PARAM_SET);
        pBss->cfParamSet.cfpCount = pBss->cfParamSet.cfpPeriod - 1;
    }

    if (isIbss) {
        pField = (A_CHAR *)mlmeElementCopy((INFO_ELEMENT *)&pBss->ibssParamSet,
                                           (INFO_ELEMENT *)pField);
    }

    if (useCnp && isAp) {
        pField = (A_CHAR *)mlmeElementCopy((INFO_ELEMENT *)&pBss->bssUcseSet,
                                           (INFO_ELEMENT *) pField);
    }

    /* the cipher suites need to be swapped too */
    FRAME_CONVERT_CPUTOLE(pBeaconBuf, bufLen);

    if (isAp) {
        TIM_ELEMENT     *pTim;
        A_SEM_TYPE      timSemId;

        /* Assign a pointer to the TIM within the beacon for easy access later */
        pBeaconInfo->pTim = (TIM_ELEMENT *)((A_UCHAR *)pBeaconBuf + beaconLen);

/* Semaphores are not fully supported by NDIS yet, hence the conditional compile */
#ifdef BUILD_AP
        /* Create and init the semaphore for TIM */
        A_SEM_INIT(timSemId, 0, CREATE_UNLOCKED);
        if (timSemId == NULL ) {
            uiPrintf("wlanBeaconInit: cannot create semaphore!\n");
            freeBuffandDesc(pDev, pBeaconDesc);
            A_DRIVER_FREE(pBeaconInfo, sizeof(BEACON_INFO), 0);
            pDev->pBeaconInfo = NULL;
            return A_ERROR;
        }
        pBeaconInfo->timSemaphoreId = timSemId;
#endif
        /* Zero out the TIM bitmap. */
        A_MEM_ZERO(pBeaconInfo->timBitmap, TIM_BITMAP_LEN);

        /*
         * Initialize the TIM element in the beacon.
         *
         * Since the first beacon sent out will not be at TSF = 0,
         * dtimCount should be DTIMPeriod - 1
         */
        pTim = pBeaconInfo->pTim;

        pTim->elementID     = ELE_TIM;
        pTim->length        = 4;
        pTim->dtimPeriod    = pBss->DTIMPeriod;
        pTim->dtimCount     = pTim->dtimPeriod - 1;
        pTim->bitmapControl = 0;
        pTim->bitmap[0]     = 0;

        beaconLen += (ELEM_ID_LEN + pTim->length + mlmeCountryCodeElementLength (pDev));

        /* Initalize power save information. */
        pDev->localSta->txPSQueue.pDescQueueHead = NULL;
        pDev->localSta->txPSQueue.pDescQueueTail = NULL;
        pDev->numSleepingClients        = 0;
        pDev->pBeaconInfo->timBitsSet   = 0;

    } /* if (isAp) */

    /* Reset some counters. */
    pDev->pBeaconInfo->bssStartCount     = 0;
    pDev->pBeaconInfo->missedBeaconCount = 0;
    pDev->pBeaconInfo->txBeaconCount     = 0;

    /* Set the beacon HW descriptor fields. */
    halSetupBeaconDesc(pDev, pBeaconDesc, beaconLen, isAp);

    /* Initialize the HW to send beacons. */
    halBeaconInit(pDev, tsf, isAp);

    pDev->pBeaconInfo->beaconEnabled = TRUE;
    pDev->pBeaconInfo->bssStartCount++;

    return A_OK;
}

/*******************************************************************************
* wlanBeaconFree
*
* This routine frees all memory allocated in wlanBeaconInit().
*/
void
wlanBeaconFree(WLAN_DEV_INFO* pDev)
{
    ASSERT(pDev);

    if (pDev->pBeaconInfo) {
        if (pDev->pBeaconInfo->pBeaconDesc) {
            freeBuffandDesc(pDev, pDev->pBeaconInfo->pBeaconDesc);
        }
        A_DRIVER_FREE(pDev->pBeaconInfo, sizeof(BEACON_INFO), 0);
        pDev->pBeaconInfo = NULL;
    }
}

/*******************************************************************************
* wlanBeaconHwResetNotify
*
* This routine restores beacon operation after a hw reset
*/
void
wlanBeaconHwResetNotify(WLAN_DEV_INFO *pDev)
{
    A_BOOL          isAp = (pDev->localSta->serviceType == WLAN_AP_SERVICE);
    WLAN_TIMESTAMP  tsf;
    A_UINT32        tsfTu;  /* TSF in TUs */


    // Restore beacon generation only if back on our original channel
    if ((pDev->staConfig.phwChannel == pDev->staConfig.pChannel) &&
        pDev->pBeaconInfo && (pDev->pBeaconInfo->beaconEnabled == TRUE)) {
        halGetTsf(pDev, &tsf);
        tsfTu = (tsf.low >> 10) | (tsf.high << 22);

        /* For adhoc, force the next TBTT to be nextTBBT+beaconInterval. 
         * This avoids two problems:
         * 1. Restoring beacon generation during the SWBA/DBA timeframe
         * 2. Forces this STA to attempt to sync its TSF to another STA before
         *    generating its own beacons 
         */
        if (!isAp) {
            tsfTu += pDev->bssDescr->beaconInterval;
        }

        /* Initialize the HW to send beacons. */
        halBeaconInit(pDev, tsfTu , isAp);
    }
}

/*******************************************************************************
* wlanTimUpdate
*
* This routine updates the virtual bitmap of the TIM element
* to indicate buffered frames at the AP.
*
* Params:   set  TRUE if you want to set the bit in the bitmap, FALSE
*                if you want to clear it
*
*/
void
wlanTimUpdate(WLAN_DEV_INFO *pDev, A_UINT16 aid, A_BOOL set)
{
    A_UINT8         byteIndex, bitIndex;
    BEACON_INFO     *pBeaconInfo;
    A_UCHAR         *pBitmap;

    //    ASSERT(pDev && pDev->pBeaconInfo);

    pBeaconInfo = pDev->pBeaconInfo;
    pBitmap     = pBeaconInfo->timBitmap;

    aid &= 0x3fff;
    //    ASSERT((aid >= ASSOCID_MIN) && (aid <= ASSOCID_MAX));

    //    A_SEM_LOCK(pBeaconInfo->timSemaphoreId, WAIT_FOREVER);

    /* calculate the index for byte and bit */
    byteIndex = aid / 8;
    bitIndex  = 1 << (aid & 0x7);

    if ((pBitmap[byteIndex] & bitIndex) == (set ? bitIndex : 0)) {
        /* Bit is already in the right state. */
      //        A_SEM_UNLOCK(pBeaconInfo->timSemaphoreId);
        return;
    }

    if (set) {
        pBitmap[byteIndex] |= bitIndex;
        pBeaconInfo->timBitsSet++;
    } else {
        pBitmap[byteIndex] &= ~bitIndex;
        pBeaconInfo->timBitsSet--;
    }

    //    A_SEM_UNLOCK(pBeaconInfo->timSemaphoreId);
}

/*******************************************************************************
* buildTim
*
* This routine builds the TIM element of an infrastructure beacon.
*
* Returns: length of the TIM element in bytes
*/
static A_UINT16
buildTim(BEACON_INFO *pBeaconInfo, A_BOOL cab)
{
    A_UINT8     index, indexN1, indexN2;
    TIM_ELEMENT *pTim;
    A_UCHAR     *pBitmap;

    ASSERT(pBeaconInfo && pBeaconInfo->pTim);

    pBitmap = pBeaconInfo->timBitmap;
    pTim    = pBeaconInfo->pTim;

    indexN1 = indexN2 = 0;
    if (pBeaconInfo->timBitsSet) {
        /*
         * Find N1 equals to the largest even number such that
         * bits numbered 1 through (N1x8) - 1 in the bitmap are all 0
         */
        for (index = 0; index < ELE_BITMAP_SIZE; index ++) {
            if (pBitmap[index] != 0) {
                indexN1 = index & 0xfe;
                break;
            }
        }

        /*
         * Find N2 equals to the smallest number such that
         * bits numbered(N2+1)x8 through 2007 in the bitmap are all 0
         */
        for (index = ELE_BITMAP_SIZE - 1; index >= indexN1; index--) {
            if (pBitmap[index] != 0) {
                indexN2 = index;
                break;
            }
        }
    }

    ASSERT(indexN1 <= indexN2);
    
    /* Copy our partial virtual bitmap into the beacon's TIM element. */
    A_BCOPY(pBitmap + indexN1, pTim->bitmap, indexN2 - indexN1 + 1);

    /* Update the Bitmap Control field */
    pTim->bitmapControl = indexN1 | (cab ? 1 : 0);

    /* Update the length field */
    pTim->length = (indexN2 - indexN1) + 4;

    /* Decrement the DTIM Count modulo the DTIM Period */
    if (pTim->dtimCount == 0) {
        pTim->dtimCount = pTim->dtimPeriod - 1;
    } else {
        pTim->dtimCount--;
    }

    return ELEM_ID_LEN + pTim->length;
}


