/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * This file contains the function prototypes and data structure declarations
 * used for sending beacons.
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/wlanbeacon.h#1 $
 */

#ifndef _WLAN_BEACON_H_
#define _WLAN_BEACON_H_

#include "wlantype.h"
#include "wlandrv.h"
#include "wlanframe.h"
#include "wlanproto.h"

#ifdef __cplusplus
extern "C" {
#endif

#define ELEM_ID_LEN             2       /* Element ID length. Used to compute the
                                         * size of the element
                                         */
#define TIM_BITMAP_LEN          251

typedef struct BeaconInfo {
    A_BOOL          beaconEnabled;      /* enable or disable beacon transmission */
    ATHEROS_DESC    *pBeaconDesc;       /* pointer to beacon descriptor */
    TIM_ELEMENT     *pTim;              /* pointer to the start of TIM in the beacon */
    A_SEM_TYPE      timSemaphoreId;     /* semaphore for TIM */
    A_UINT16        nonTIMLen;          /* Beacon frame length, but w/o TIM */
    A_UINT16        TIMLen;             /* TIM length */
    A_UINT16        frameLen;           /* Total beacon frame length */
    A_UINT16        missedBeaconCount;  /* AP could not keep up with beacon generation */
    A_UINT32        txBeaconCount;      /* sent out beacon count */
    A_UINT32        bssStartCount;      /* how many times the BSS starts */
    A_UINT32        wdgDelayTicks;      /* beacon watchdog timer delay in ticks */
    A_UCHAR         *pCfPtr;            /* ptr to start of CF element in real beacon */
    A_INT32         timBitsSet;         /* number of STAs having buffered traffic */
    A_UCHAR         timBitmap[TIM_BITMAP_LEN]; /* for keeping track of the entire TIM bitmap */
} BEACON_INFO;

/* exportec function prototypes */

A_STATUS
wlanApBeaconGen(WLAN_DEV_INFO *pDev);

A_STATUS
wlanBeaconInit(WLAN_DEV_INFO *pDev, A_UINT32 tsf);

void
wlanBeaconFree(WLAN_DEV_INFO *pDev);

void
wlanBeaconHwResetNotify(WLAN_DEV_INFO *pDev);

void
wlanTimUpdate(WLAN_DEV_INFO *pDev, A_UINT16 aid, A_BOOL set);


#ifdef __cplusplus
}
#endif

#endif /* _WLAN_BEACON_H_ */
