/*
 * Copyright  2000-2002 Atheros Communications, Inc.,  All Rights Reserved.
 *
 * This file contains the function prototypes and data structure declarations
 * used by channel list creation and regulatory domain functions.
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/wlanchannel.h#1 $
 */

#ifndef _WLAN_CHANNEL_H_
#define _WLAN_CHANNEL_H_

#ifdef __cplusplus
extern "C" {
#endif

/* Extra information about channels detected during hardware operation. */
#define CHANNEL_CW_INT              0x0001   // CW interference detected on the channel
#define CHANNEL_RADAR_INT           0x0002   // Radar interference detected on the channel
#define CHANNEL_BUSY                0x0004   // Busy, occupied directly or overlap from BSS on adjoining channel
#define CHANNEL_TURBO               0x0100   // Turbo Channel
#define CHANNEL_CCK                 0x0200   // CCK channel
#define CHANNEL_OFDM                0x0400   // OFDM channel
#define CHANNEL_2GHZ                0x0800   // 2 GHz spectrum channel.
#define CHANNEL_5GHZ                0x1000   // 5 GHz spectrum channel
#define CHANNEL_PASSIVE             0x2000   // Only passive scan allowed in the channel

#define CHANNEL_A       (CHANNEL_5GHZ|CHANNEL_OFDM)
#define CHANNEL_B       (CHANNEL_2GHZ|CHANNEL_CCK)
#define CHANNEL_PUREG   (CHANNEL_2GHZ|CHANNEL_OFDM)
#define CHANNEL_T       (CHANNEL_5GHZ|CHANNEL_OFDM|CHANNEL_TURBO)
#define CHANNEL_ALL     (CHANNEL_OFDM|CHANNEL_CCK|CHANNEL_5GHZ|CHANNEL_2GHZ|CHANNEL_TURBO)
#define CHANNEL_DEFAULT (CHANNEL_5GHZ|CHANNEL_OFDM) // default is normal 11a mode
#define MAX_REG_DOM_NAME_SIZE       10       // Max string size of reg dom name inc NULL

#define IS_CHAN_TURBO(x) ((x) & (CHANNEL_OFDM | CHANNEL_TURBO))
#define IS_CHAN_BASE(x)  (((x) & CHANNEL_OFDM) && !((x) & CHANNEL_TURBO))
#define IS_CHAN_OFDM(x)  ((x) & CHANNEL_OFDM)
#define IS_CHAN_CCK(x)   ((x) & CHANNEL_CCK)

#define MODE_SELECT_11A   0x01
#define MODE_SELECT_TURBO 0x02
#define MODE_SELECT_11B   0x04
#define MODE_SELECT_PUREG 0x08
// This bit is not exactly a mode, but bit is set by NCP in same field. May move it elsewhere
#define MODE_SELECT_EXT_RANGE   0x100   // 11B using 1, 2, 5.5 and 11.

/* Channel Manipulation Macros */
/* 5 GHz range uses IEEE numbers */
#define MHZ2IEEE_CH(mhz)        ( ( (mhz) - 5000  ) / 5 )
#define IEEE_CH2MHZ(ch)         ( 5000 + ( (ch) << 2 ) + (ch) )

#define CTRY_DEFAULT  0
#define CTRY_DEBUG    0x1ff
#define CTRY_INVALID  0x2ff
#define CHAN_11A_DEFAULT 5280

/* Scan Types Allowed */
typedef enum {
    AP_SCAN_ALL_BSS,
    AP_SCAN_CLEAN_CHAN,
    AP_SCAN_SELECT_CHAN
} APSCAN_OP;


typedef struct countryInfoList {
    A_UINT8    len;            /* Total length of the list */
    A_UINT8     countryString[3];
    struct {
        A_UINT8 firstChannel;
        A_UINT8 numChannel;
        A_UINT8 maxTxPwr;
    } subBand[4];               /* SubBand List starts here... */
} __ATTRIB_PACK COUNTRY_INFO_LIST;


/* Channel List Initialization Functions */

A_STATUS
wlanInitChannelList(WLAN_DEV_INFO *pDevInfo, WLAN_CHANNEL_LIST **pCList, A_UINT16 regDmnEnum,
    A_UINT16 modeSelect, A_UINT16 enableOutdoor);

A_STATUS
wlanInitSingleChannelList(WLAN_DEV_INFO *pDevInfo, WLAN_CHANNEL_LIST **pCList, A_UINT16 channel, A_UINT16 channelFlags);

A_STATUS
asciiToClist(WLAN_DEV_INFO *pDevInfo, A_UCHAR *pStr, A_UINT16 alen, WLAN_CHANNEL_LIST **pList);

/* Channel List Manipulation Functions */

A_STATUS
wlanCopyChannelList(WLAN_CHANNEL_LIST *pCListFrom, WLAN_CHANNEL_LIST **pCListTo);

A_STATUS
wlanFreeChannelList(WLAN_CHANNEL_LIST *pCList);

A_STATUS
wlanInitScanList(WLAN_CHANNEL_LIST *pCListFrom, WLAN_SCAN_LIST **pSListTo);

A_STATUS
wlanInitSingleScanList(CHAN_VALUES *pChannelFrom, WLAN_SCAN_LIST **pSListTo);

A_STATUS
wlanFreeScanList(WLAN_SCAN_LIST *pSList);

CHAN_VALUES *
wlanGetChannelPtr(WLAN_CHANNEL_LIST *pCList, A_UINT16 channel, A_UINT16 modeSelect);

/* Channel List Accessor Functions */

A_BOOL
wlanIsValidChannel(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode, A_UINT16 modeSelect, A_UINT16 channel);

A_UINT8
wlanGetChannelPower(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel);

void
wlanPrintChannelList(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode,
    A_UINT16 wirelessMode, A_UINT16 enableOutdoor);

/* Country Code Manipulation Functions */

A_BOOL
wlanIsCountryCodeValid(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode);

A_UINT16
wlanGetCountryCodeTableSize(WLAN_DEV_INFO *pDevInfo);

const A_UCHAR *
wlanGetIndexedCountryCodeName(WLAN_DEV_INFO *pDevInfo, int index);

const A_UCHAR *
wlanGetIndexedCountryCodeIsoName(WLAN_DEV_INFO *pDevInfo, int index);

A_UINT16
wlanGetCountryCodeIndex(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode);

const A_UCHAR *
wlanGetCountryCodeName(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode);

A_UINT16
wlanGetCountryCodeByName(A_UCHAR *countryCodeName);

A_UINT16
wlanGetDefaultCountry(WLAN_DEV_INFO *pDevInfo);

void
wlanPrintCountryCodeList(WLAN_DEV_INFO *pDevInfo);

A_BOOL
wlanIsRegCcValid(WLAN_DEV_INFO *pDevInfo);

A_INT8
wlanGetAntennaReduction(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel, A_INT8 twiceAntennaGain);

A_UINT8
wlanGetCtl(WLAN_DEV_INFO *pDevInfo, CHAN_VALUES *pChannel);

void
wlanGetCountryInfoList(WLAN_DEV_INFO *pDevInfo, COUNTRY_INFO_LIST *pCountryInfo);

/* Regulatory Domain Table Manipulation Functions */

A_UINT16
wlanGetCountryCodeByRegDmnName(A_UCHAR *regName);

A_UINT16 
wlanGetRegDmnByName(A_UCHAR *regName);

void
wlanPrintRegDmnList(void);

/* Channel Conversion Functions */

A_UINT8
wlanConvertGHztoCh(A_UINT16 freq, A_UINT16 channelFlags);

A_UINT16
wlanConvertChtoGHz(A_UINT8 ch, A_UINT16 channelFlags);

/* Regulatory Domain checking Functions */

A_BOOL
wlanEnableDFS(WLAN_DEV_INFO *pDevInfo);

A_BOOL
wlanIsActiveScanAllowed(WLAN_DEV_INFO *pDevInfo);

A_BOOL
wlanIsNfCheckRequired(WLAN_DEV_INFO *pDevInfo);

/* Radio Frequency Specification Manipulation Functions */

A_UINT16
wlanGetWirelessMode(WLAN_DEV_INFO *pDevInfo, A_UINT16 countryCode);

/* Channel Scan Functions */

A_STATUS
apScan(WLAN_DEV_INFO *pDevInfo, APSCAN_OP scanOp, SCANTYPE scan_type);


/*
 *  Country/Region Codes from MS WINNLS.H
 *  Numbering from ISO 3166
 */
enum {
    CTRY_ALBANIA              = 8,       // Albania
    CTRY_ALGERIA              = 12,      // Algeria
    CTRY_ARGENTINA            = 32,      // Argentina
    CTRY_ARMENIA              = 51,      // Armenia
    CTRY_AUSTRALIA            = 36,      // Australia
    CTRY_AUSTRIA              = 40,      // Austria
    CTRY_AZERBAIJAN           = 31,      // Azerbaijan
    CTRY_BAHRAIN              = 48,      // Bahrain
    CTRY_BELARUS              = 112,     // Belarus
    CTRY_BELGIUM              = 56,      // Belgium
    CTRY_BELIZE               = 84,      // Belize
    CTRY_BOLIVIA              = 68,      // Bolivia
    CTRY_BRAZIL               = 76,      // Brazil
    CTRY_BRUNEI_DARUSSALAM    = 96,      // Brunei Darussalam
    CTRY_BULGARIA             = 100,     // Bulgaria
    CTRY_CANADA               = 124,     // Canada
    CTRY_CHILE                = 152,     // Chile
    CTRY_COLOMBIA             = 156,     // Colombia
    CTRY_COSTA_RICA           = 188,     // Costa Rica
    CTRY_CROATIA              = 191,     // Croatia
    CTRY_CZECH                = 203,     // Czech Republic
    CTRY_DENMARK              = 208,     // Denmark
    CTRY_DOMINICAN_REPUBLIC   = 214,     // Dominican Republic
    CTRY_ECUADOR              = 218,     // Ecuador
    CTRY_EGYPT                = 818,     // Egypt
    CTRY_EL_SALVADOR          = 222,     // El Salvador
    CTRY_ESTONIA              = 233,     // Estonia
    CTRY_FAEROE_ISLANDS       = 234,     // Faeroe Islands
    CTRY_FINLAND              = 246,     // Finland
    CTRY_FRANCE               = 250,     // France
    CTRY_GEORGIA              = 268,     // Georgia
    CTRY_GERMANY              = 276,     // Germany
    CTRY_GREECE               = 300,     // Greece
    CTRY_GUATEMALA            = 320,     // Guatemala
    CTRY_HONDURAS             = 340,     // Honduras
    CTRY_HONG_KONG            = 344,     // Hong Kong S.A.R., P.R.C.
    CTRY_HUNGARY              = 348,     // Hungary
    CTRY_ICELAND              = 352,     // Iceland
    CTRY_INDIA                = 356,     // India
    CTRY_INDONESIA            = 360,     // Indonesia
    CTRY_IRAN                 = 364,     // Iran
    CTRY_IRAQ                 = 368,     // Iraq
    CTRY_IRELAND              = 372,     // Ireland
    CTRY_ISRAEL               = 376,     // Israel
    CTRY_ITALY                = 380,     // Italy
    CTRY_JAMAICA              = 388,     // Jamaica
    CTRY_JAPAN                = 392,     // Japan
    CTRY_JORDAN               = 400,     // Jordan
    CTRY_KAZAKSTAN            = 398,     // Kazakstan
    CTRY_KENYA                = 404,     // Kenya
    CTRY_KUWAIT               = 414,     // Kuwait
    CTRY_LATVIA               = 428,     // Latvia
    CTRY_LEBANON              = 422,     // Lebanon
    CTRY_LIBYA                = 434,     // Libya
    CTRY_LIECHTENSTEIN        = 438,     // Liechtenstein
    CTRY_LITHUANIA            = 440,     // Lithuania
    CTRY_LUXEMBOURG           = 442,     // Luxembourg
    CTRY_MACAU                = 446,     // Macau
    CTRY_MACEDONIA            = 807,     // the Former Yugoslav Republic of Macedonia
    CTRY_MALAYSIA             = 458,     // Malaysia
    CTRY_MEXICO               = 484,     // Mexico
    CTRY_MONACO               = 492,     // Principality of Monaco
    CTRY_MOROCCO              = 504,     // Morocco
    CTRY_NETHERLANDS          = 528,     // Netherlands
    CTRY_NEW_ZEALAND          = 554,     // New Zealand
    CTRY_NICARAGUA            = 558,     // Nicaragua
    CTRY_NORWAY               = 578,     // Norway
    CTRY_OMAN                 = 512,     // Oman
    CTRY_PAKISTAN             = 586,     // Islamic Republic of Pakistan
    CTRY_PANAMA               = 591,     // Panama
    CTRY_PARAGUAY             = 600,     // Paraguay
    CTRY_PERU                 = 604,     // Peru
    CTRY_PHILIPPINES          = 608,     // Republic of the Philippines
    CTRY_POLAND               = 616,     // Poland
    CTRY_PORTUGAL             = 620,     // Portugal
    CTRY_PRCHINA              = 156,     // People's Republic of China
    CTRY_PUERTO_RICO          = 630,     // Puerto Rico
    CTRY_QATAR                = 634,     // Qatar
    CTRY_ROMANIA              = 642,     // Romania
    CTRY_RUSSIA               = 643,     // Russia
    CTRY_SAUDI_ARABIA         = 682,     // Saudi Arabia
    CTRY_SINGAPORE            = 702,     // Singapore
    CTRY_SLOVAK               = 703,     // Slovak Republic
    CTRY_SLOVENIA             = 705,     // Slovenia
    CTRY_SOUTH_AFRICA         = 710,     // South Africa
    CTRY_SOUTH_KOREA          = 410,     // South Korea
    CTRY_SPAIN                = 724,     // Spain
    CTRY_SWEDEN               = 752,     // Sweden
    CTRY_SWITZERLAND          = 756,     // Switzerland
    CTRY_SYRIA                = 760,     // Syria
    CTRY_TAIWAN               = 158,     // Taiwan
    CTRY_THAILAND             = 764,     // Thailand
    CTRY_TRINIDAD_Y_TOBAGO    = 780,     // Trinidad y Tobago
    CTRY_TUNISIA              = 788,     // Tunisia
    CTRY_TURKEY               = 792,     // Turkey
    CTRY_UAE                  = 784,     // U.A.E.
    CTRY_UKRAINE              = 804,     // Ukraine
    CTRY_UNITED_KINGDOM       = 826,     // United Kingdom
    CTRY_UNITED_STATES        = 840,     // United States
    CTRY_URUGUAY              = 858,     // Uruguay
    CTRY_UZBEKISTAN           = 860,     // Uzbekistan
    CTRY_VENEZUELA            = 862,     // Venezuela
    CTRY_VIET_NAM             = 704,     // Viet Nam
    CTRY_YEMEN                = 887,     // Yemen
    CTRY_ZIMBABWE             = 716      // Zimbabwe
};


#ifdef __cplusplus
}
#endif

#endif /* _WLAN_CHANNEL_H_ */
