/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * Definitions for core driver
 * This is a common header file for all platforms and operating systems.
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/wlandrv.h#6 $
 */

#ifndef _WLANDRV_H_
#define _WLANDRV_H_


#include <linux/pci.h>
#include "wlantype.h"
#include "wlanproto.h"
#include "wlansme.h"
#include "wlanos.h"
#include "queue.h"
#include "halDesc.h"
#include "wlanfragde.h"

#ifdef VXWORKS
#include "trace.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef  BUILD_AP
#include "endLib.h"
#endif  /* BUILD_AP */

#define WLAN_MAX_DEV            1       /* max # of devices per driver */

#define ATHEROS_VENDOR_ID       0x168C  /* vendor ID for our device */

#define MAX_KEY_LEN_BITS        128
#define MAX_KEY_LEN_BYTES       (MAX_KEY_LEN_BITS / 8)
#define MAX_SHARED_KEYS         4
#define UNIQ_KEY_INDEX          4
#define KEYINDEX_INVALID        0x7fff          /* key index does not exist */
#define KEYINDEX_NONE           0x3fff          /* implies no crypto in vlan support */
#define KEYINDEX_DYNAMIC        0x8000          /* dynamically allocate keyindex */
#define KEY_XOR                 0xaa

typedef struct WlanPrivStruct {
    A_UINT32    keyIV;                          /* IV generation control */
    A_UINT16    keyFlags;
    A_UINT16    keyLength;                      /* Length of key in bits */
    A_UINT16    keyType;                        /* type of key, WEP, WEP2 or AES */
    void*       initValue;                      /* Currently used to hold AES data struct only */
    A_UINT8     keyVal[MAX_KEY_LEN_BYTES];      /* Array for key */
    A_UINT16    aesKeyLength;                   /* Length of aes key in bits */
    A_UINT8     aesKeyVal[MAX_KEY_LEN_BYTES];   /* Array aes for key */
} WLAN_PRIV_RECORD;

typedef struct aesNonce {
    A_UINT32            keyIV;
    WLAN_MACADDR        srcAddr;
    WLAN_MACADDR        dstAddr;
} AES_NONCE;

/*
 * WLAN_PRIV_RECORD key types
 */
#define PRIV_KEY_TYPE_WEP       0x0
#define PRIV_KEY_TYPE_AES       0x2
#define PRIV_KEY_TYPE_NULL      0x7

/* WLAN_PRIV_RECORD keyFlags bit mappings */
#define PRIV_FRAMEKEY           0x03            /* Two bits of key in on air frame */
#define PRIV_UNIQUEKEY          0x04            /* Unique key */
#define PRIV_DYNAMIC            0x08            /* Key is generated by exchange protocol */

/* Encryption struct for key backing store */
typedef struct {
    unsigned long P[16 + 2];
    unsigned long S[4][256];
} BLOWFISH_CTX;

struct queueDetails;

/* Software descriptor structure */
typedef struct atherosDesc {
    A_UINT32    nextPhysPtr;        /* phys address of next descriptor */
    A_UINT32    bufferPhysPtr;      /* phys address of this buffer */
    union {                         /* harware control variables */
        HAL_DESC_TX_CONTROL tx;
        HAL_DESC_RX_CONTROL rx;
        A_UINT32 word[2];
    } hwControl;
    union {                         /* hardware status variables */
        HAL_DESC_TX_STATUS tx;
        HAL_DESC_RX_STATUS rx;
        volatile A_UINT32 word[2];
    } hwStatus;
    A_UINT32            pad8[2];            /* for SWCOHERENCY pad to 32B */
    /* SOFTWARE */
    A_UINT32            thisPhysPtr;        /* phys address of this descriptor */
    struct atherosDesc  *pNextVirtPtr;      /* virtual pointer to next descriptor */
    WLAN_FRAME_UNION    pBufferVirtPtr;     /* virtual pointer to buffer */
    void                *pOrigBufferVirtPtr;/* Original buffer virtual ptr, in case bufferVirtPtr changes */
    A_UINT32            pOrigbufferPhysPtr; /* Original buffer physical ptr, in case bufferPhysPtr changes */
    void                *pOSHandle;         /* OS specific */
    void                *pOSDescPtr;        /* OS specific */
    A_UINT32            OSspecificInfo;     /* platform-specific info, for whatever purpose */
    A_UINT32            OSbuffer;           /* pointer to OS-owned buffer, if any */
    struct sibEntry     *pSrcSibEntry;      /* source STA SIB entry */
    struct sibEntry     *pDestSibEntry;     /* destination STA SIB entry */
    struct atherosDesc  *pTxLastDesc;       /* pointer to the last descriptor of the frame in a transmit chain */
    struct atherosDesc  *pTxFirstDesc;      /* pointer to the first descriptor of the frame in a transmit chain */
    A_UINT32            timeStamp;          /* time stamp for PS aging in MS */
    struct queueDetails* freeQueue;         /* flag for which queue the descriptor, when free, belongs to */

    A_UINT16            memFlag;            /* flag for how buffer mem was allocated */
    A_UINT16            qFlag;              /* flag for which q the tx frame goes to */
    A_UINT32            staleFlag:1,        /* this descriptor has been processed already */
                                            /* but is being cached by the hardware and should not be freed */
                        swretryCount:4,     /* this desc is being swretried */
                        rxKeyIndex:8,       /* key from rx status */
                        fragSeqNumValid:1,  /* set if we've already saved the seqNum in a frag train */
                        psFlag:1,           /* set if this is a ps buffered unicast frame */
                        decryptError:1,     /* payload was not decrypted correctly */
                        unused:16;

    A_UINT32            padding32[7];       /* Padded to make it 32 dwords long, cache line aligned */
} ATHEROS_DESC;

/* Queue data structure */
typedef struct queueDetails {
    ATHEROS_DESC    *pDescQueueHead;    /* virt pointer to queue head */
    A_UINT32        physQueueHead;      /* physical pointer to queue head */
    ATHEROS_DESC    *pDescQueueTail;    /* virtual pointer to queue tail */
    A_SEM_TYPE      qSem;               /* Semaphore to manage this queue */
    A_INT32         qFrameCount;        /* frame count in this queue */
    A_BOOL          isTailSelfLinked;   /* should the tail descriptor be linked to itself? */
} QUEUE_DETAILS;

/*
 * Possible state bits for destState in the SIB.
 * The bits reflect current successful state of the station. e.g. STATE_AUTH
 * Authenticated bit won't be set until the frame exchange is completed with
 * successful status.
 */

#define STATE_QUIET         0x0     /* Initial empty state */
#define STATE_AUTH          0x1     /* Station is authenticated */
#define STATE_ASSOC         0x2     /* Station is associated */
#define STATE_PS            0x4     /* Station is in power save mode */
#define STATE_ADHOC         0x8     /* Station is in Ad-hoc networking mode */
#define STATE_JOINED        0x10    /* Station has joined an AP */
#define STATE_STA_LOCKED    0x20    /* SIB_ENTRY data struct is locked */
#define STATE_AP_DOWN       0x0100  /* This is an AP, down, or not started */
#define STATE_AP_STARTING   0x0200  /* This is an AP, in starting process */
#define STATE_AP_UP         0x0400  /* This is an AP, up and running */
#define STATE_CONNECTED     (STATE_JOINED|STATE_AUTH|STATE_ASSOC) /* All 3 bits set when connected */

/*
 * Possible state bits for transState in SIB_ENTRY
 * It is the current state of transactions going on for the station.
 * It will change when a request frame is sent or a response is received.
 * It will change when almost any communication is made with the station.
 * After going through a few states in transState, an operation like
 * Authentication is completed, then we change the corresponding state bit
 * in destState and clear the transState.
 */

#define TSTATE_QUIET                0       /* No transactions in progress */
#define TSTATE_AUTH_FRAME1_SENT     0x1     /* Frame 1 has been sent (on STA) */
#define TSTATE_AUTH_FRAME1_RCVD     0x2     /* Frame 1 is received (on AP) */
#define TSTATE_AUTH_FRAME2_SENT     0x4     /* Frame 2 has been sent (on AP) */
#define TSTATE_AUTH_FRAME2_RCVD     0x8     /* Frame 2 is received (on STA) */
#define TSTATE_AUTH_FRAME3_SENT     0x10    /* Frame 3 has been sent (on STA) */
#define TSTATE_AUTH_FRAME3_RCVD     0x20    /* Frame 3 is received (on AP) */
#define TSTATE_AUTH_FRAME4_SENT     0x40    /* Frame 4 has been sent (on AP) */
#define TSTATE_AUTH_FRAME4_RCVD     0x80    /* Frame 4 is received (on STA) */
#define TSTATE_AUTH_IN_PROGRESS     0xff    /* Any of above bits is set */
#define TSTATE_ASSOC_REQ_SENT       0x100   /* Assoc request is sent (on STA) */
#define TSTATE_ASSOC_REQ_RCVD       0x200   /* Assoc request is received (on AP) */
#define TSTATE_ASSOC_RESP_SENT      0x400   /* Assoc response is sent (on AP) */
#define TSTATE_ASSOC_RESP_RCVD      0x800   /* Assoc response received (on STA) */

#define TSTATE_SCANNING             0x1000  /* Scan in progress */
#define TSTATE_JOINING              0x2000  /* Join in progress */
#define TSTATE_DEAUTH_FRAME_SENT    0x4000  /* deauth in progress */
#define TSTATE_DISASSOC_FRAME_SENT  0x8000  /* Disassoc in progress */

#define TSTATE_REASSOC_REQ_SENT     0x10000 /* Reassoc request is sent (on STA) */
#define TSTATE_REASSOC_REQ_RCVD     0x20000 /* Reassoc request is received (on AP) */
#define TSTATE_REASSOC_RESP_SENT    0x40000 /* Reassoc response is sent (on AP) */
#define TSTATE_REASSOC_RESP_RCVD    0x80000 /* Reassoc response received (on STA) */
#define TSTATE_PWRMGT_PENDING       0x100000    /* Transmission of power mode packet pending */

/* Values for 'used' field in sibEntry.
 * sibEntryAlloc() will set the SIB_USED bit and reserve it.
 * sibEntryAdd() will make it SIB_VALID.
 * Someone may set the SIB_BUSY flag indicating it shouldn't be
 * deleted.
 */
#define SIB_FREE    0x0     /* SIB entry is free */
#define SIB_USED    0x1     /* SIB entry is taken */
#define SIB_VALID   0x2     /* Valid data in SIB entry */
#define SIB_BUSY    0x4     /* Someone is using this SIB, don't delete */

//
// Power modes specified by user
//
#define POWERMGT_CAM            0           // Continuously active mode
#define POWERMGT_SLEEP          1           // aggressive sleep mode
#define POWERMGT_PERF           2           // aggressive performance mode

//
// Power mode requests
//
#define FORCE_SLEEP             0
#define FORCE_AWAKE             1
#define NORMAL_SLEEP            2
#define NORMAL_AWAKE            3
#define BLOCK_SLEEP             4           // block tx requests and buffer at AP while we do something else
#define BLOCK_AWAKE             5           // clear blocking

//
// Power mode states
//
#define STATE_AWAKE                     0
#define STATE_PENDING_SLEEP             1
#define STATE_SLEEP                     2
#define STATE_PENDING_BLOCKING_SLEEP    3
#define STATE_BLOCKING_SLEEP            4

// Power timeouts
#define POWER_UP_TIME           2000        // usecs to wait for chip to come out of sleep state
#define SLEEP_DURATION_UNITS    0x00000008 // SLDUR units/TU


//================================================
// Definitions required for some routines
//================================================

#define RESET_CLEAR_ALL 0x00        /* Clear resets to all logic */
#define RESET_PCI_HW    0x01        /* Reset PCI interface. Need to reload driver */
#define RESET_TX_HW     0x02        /* Reset TX hardware interface */
#define RESET_RX_HW     0x04        /* Reset RX hardware interface */
#define RESET_PCU_HW    0x08        /* Reset PCU logic */
#define RESET_D2_HW     0x10        /* Reset D2 hardware */
#define RESET_TX_Q      0x20        /* Reject tx frames and empty tx Q */
#define RESET_ALL_HW    (RESET_PCI_HW|RESET_TX_HW|RESET_RX_HW|RESET_PCU_HW|RESET_D2_HW)

#define MAX_MULTICAST_ADDRESSES     32

#define NO_DOMAIN_NAME              "NONE"   // String Name of debug regulatory domain
/*
 * List of channels to be scanned. channel is to be used as a variable
 * size array.
 */

typedef struct channelValues {
    A_UINT16 channel;
    A_UINT16 channelFlags;
} CHAN_VALUES;

typedef struct wlanChannelList {
    A_UINT16    listSize;
    CHAN_VALUES *chanArray;
} WLAN_CHANNEL_LIST;

typedef struct wlanScanList {
    A_UINT16    listSize;
    CHAN_VALUES **pChanArray;
} WLAN_SCAN_LIST;

/*Constants used in tx service*/
#define ACK_BIT                     134 /*bit number in ack package*/
#define CTS_BIT                     134 /*bit number in cts package*/
#define SIFS                        16  /*SIFS time*/
#define SLOT_TIME                   9   /*slot time*/
#define PLCP_TIME                   20  /*PLCP time, this actually PLCP + PREAMBLE = 16 + 4*/


/* Default config information for STA. Config utilities
 * will fill in this structure. It will persist throuth
 * a reset, SIB entry will not.
 */
typedef struct StaConfig
{
    WLAN_MACADDR    macPermAddr;            // EEPROM's MAC Address - only used by driver attach and OID request
    WLAN_MACADDR    macAddr;                // MAC Address of device
    BSSID           bssId;                  // Desired BSSID, if possible
    BSSTYPE         bssType;                // Type of BSS - Infrastructure, Independent, or ANY
    CAP_INFO        capInfo;                // Capibilities info
    SSID            ssid;                   // Desired ESSID, domain
    struct RateTableStruct  *pRateTable;    // points to global rate table used (a or b)
    RATE_SET        cckRateSet;             // Rates to be used for CCK
    RATE_SET        ofdmRateSet;            // Rates to be used for OFDM
    RATE_SET        operationalRateSet;     // All supported operation rates by this station right now
    A_UINT8         defaultCckRateIndex;    // default rate for the station
    A_UINT8         defaultOfdmRateIndex;   // default rate for the station
    A_BOOL          privacyOptionImplemented;   // privacy option implemented
    A_BOOL          SmeEnable;              // Station Management Entity flag
    CHAN_VALUES     *pChannel;              // home channel, where it's AP is located
    CHAN_VALUES     *phwChannel;            // channel currently set in hardware, may be different
                                            // than Channel during scan
    struct wlanChannelList *pClist;         // The channel list created to use for selectable channels
    A_UINT16        adhocChannelFreq;       // adhoc channel for starting IBSS
    A_UINT16        adhocChannelFlags;      // adhoc channel type for starting IBSS
    A_UINT16        defaultFragThreshold;   // Default fragmentation threshold
    A_UINT16        userFragThreshold;      // User-selected fragmentation threshold
    A_UINT16        cfpDuration;            // CFP duration for NDIS based AP, if supported.
    A_UINT16        defaultDot11RTSThreshold;
    A_UINT16        userRTSThreshold;
    A_UINT32        maxTransmitMsduLifetime;//the elapsed time in TU, after the initial transmission of an MSDU, after which further
                                            // attempts to transmit the MSDU shall be terminated
    A_UINT32        maxReceiveLifeTime;     // the elapsed time in TU, after the initial reception of a fragmented MMPDU or MSDU,
                                            // after which further attempts to reassemble the MMPDU or
                                            // MSDU shall be terminated

    A_UINT32        NumberOfMcAddresses;    // Used for multicast address changes during resets
    WLAN_MACADDR    McTable[MAX_MULTICAST_ADDRESSES];

    A_UINT32        McastFilter0;           // Current selectively-set multicast filter bits
    A_UINT32        McastFilter1;

    A_UINT32        PacketFilter;           // The Current Global Packet Filter and look ahead size
    A_UINT32        NumberRxDesc;           // Number of receive descriptor/buffers to allocate
    A_UINT32        encryptionAlg;          // Encryption algorithm (AES/WEP)
    A_BOOL          privacyInvoked;         // Privacy enable status
    A_BOOL          enableACL;              // Enable MAC ACL checks
    A_BOOL          excludeUnlisted;        // Disallow unlisted MACs
    A_BOOL          authAllowOpen;          // Open authentication is ok
    A_BOOL          authAllowShared;        // Shared authentication is ok
    A_BOOL          authOpenFirst;          // STA first authentication request
    A_BOOL          rateCtrlEnable;         // Enable transmit rate control features
    A_BOOL          bkScanEnable;           // Enable background scanning features
    A_BOOL          radioEnable;            // Radio is enabled. Can be turned off by RFkill feature
    A_BOOL          filter8021x;            // 802.1X filtering. Don't encrypt .1x frames. Default 0.
    A_BOOL          multiDomainCapEnabled;  // MultiDomainCapability flag
    A_UINT16        countryCode;            // Country Code (MS numbering)
    A_UINT16        defaultKey;             // Default encryption key [0-4]
    A_UINT16        aesVlanKey;
    WLAN_PRIV_RECORD keys[MAX_SHARED_KEYS+1];   // array to hold configured keys
    A_UINT16        sleepMode;              // Awake, aggressive sleep or agressive performance
    A_UINT16        mbiTimeout;             // Missed beacon timeout, AP assumed to be lost after this
    A_UINT16        sleepTimePwrSave;       // station wakes after this many mS in max power save mode
    A_UINT16        sleepTimePerf;          // station wakes after this many mS in max performance mode
    A_UINT16        sleepDuration;          // one of the two above, converted to SCR units
    A_UINT16        listenInterval;         // station wakes after this TU period to listen for beacons
    A_UINT16        beaconInterval;         // beacon interval, applies to AP or ad-hoc only.
    A_UINT32        atimWindow;             // atim window, applies to ad-hoc only.
    A_UINT32        inactivityTimePwrSave;  // in max PS mode, how long (in mS) w/o Tx/Rx before going back to sleep
    A_UINT32        inactivityTimePerf;     // in max perf mode, how long (in mS) w/o Tx/Rx before going back to sleep
    A_UINT16        calibrationTime;        // how often to do calibrations, in seconds (0=off)
    A_UINT32        noBeaconTimeout;        // time in ms to wait while not receiving beacons before bailing
    A_UINT32        swRetryMaxRetries;      // max number of retries to be attempted by sw
                                            // for any given frame
    A_INT32         psQueueMaxDepth;        // Max depth of queue used while processing power saved frames
    A_UINT32        limitTxRetries;         // limit of num of times frame gets airborne
    A_UINT32        txTrigThresh;           // How many good tx's before decreasing tx trigger
    A_BOOL          swretryEnabled;         // feature enable flag
    A_BOOL          QosEnabled;             // If true, use the two timing values below
    A_UINT32        aifs;                   // the 802.11e AIFS parameter in terms of slots; default DIFS
    A_UINT32        cwmin;                  // 802.11e tunable cwmin - based on sta priority; default 15
    A_UINT16        tpScale;                // Scaling factor to be applied to max TP
    A_UINT16        RegDomain;              // Regulatory Domain
    A_UINT16        busConfig;              // PCI target retry and TRDY timeout
    A_UINT16        cacheLineSize;          // Saved from PCI config (set by BIOS)
                                            // Combination of phyType and phyFreq determines a, b or g.
    A_UINT16        NetBand;                // User's choice of a, b, turbo, g, etc. from registry
    A_UINT16        AdhocBand;              // User's choice of a, b, turbo, g, etc. to start adhoc
    A_BOOL          radarDetectEnable;      // radar detection feature
    A_BOOL          shortPreamble;           // Allow short and long preamble in 11b mode
    A_UINT32        scanType;               // active/passive scan, important for ETSI
    A_UINT32	    dot1xMode;	// Mode of 802.1x operation for AP
} WLAN_STA_CONFIG;

typedef struct WlanStats {
    // Frame counts
    A_UINT32                GoodTransmits;
    A_UINT32                GoodReceives;
    A_UINT32                MulticastReceives;
    A_UINT32                BeaconCount;

    // Count of transmit errors
    A_UINT32                TxExcessiveRetries;
    A_UINT32                TxDmaUnderrun;
    A_UINT32                TotalRetries;
    A_UINT32                AckRcvFailures;

    // Count of receive errors
    A_UINT32                RcvDecryptCrcErrors;
    A_UINT32                RcvDecipherErrors;
    A_UINT32                RcvKeyCacheMisses;
    A_UINT32                RcvCrcErrors;
    A_UINT32                RcvDmaOverrunErrors;
    A_UINT32                RcvCdtFrames;
    A_UINT32                RcvRuntErrors;
    A_UINT32                RcvPhyErrors;
    A_UINT32                RcvDupSeqErrors;
    A_UINT32                RtsFailCnt;
    A_UINT32                RtsSuccessCnt;
    A_UINT32                FcsFailCnt;
    A_UINT32                MultipleRxDuplicates;
    A_UINT32                RcvEOL;
    A_UINT32                RcvWEPExcludedCount;

    /* mibs */
    WLAN_MACADDR            txDisAssociationSta;
    A_UINT32                txDisAssocReasonCode;
    WLAN_MACADDR            txDeAuthenticationSta;
    A_UINT32                txDeAuthReasonCode;
    WLAN_MACADDR            txAuthenticateFailSta;
    A_UINT32                txAuthenticateFailStatus;

    // Atheros-defined info
    A_UINT32                TxFiltered;
    A_UINT32                RetryBins[16];              /* long frame bin */
    A_UINT32                shortFrameRetryBins[16];
    A_UINT32                swRetryTotalCnt;            /* sw retries for this connection */
    A_UINT32                swRetryMaxRetriesExceeded;  /* # dropped frames due to exceeding */
                                                        /* swRetryMaxRetries from staConfig */
    A_UINT32                swRetryHungCount;
    A_INT32                 CurrentAckSignalStrenth;
    A_UINT32                RcvRate;
    A_INT32                 RcvSignalStrength;
    A_UINT32                beaconRcvRate;
    A_INT32                 beaconRcvSignalStrength; /* current received beacon signal strength indicator */
    A_UINT32                MaxTxRateKb;        // Maximum link speed in kilobits per sec.

    A_UINT32                HungCount;      /* How many times tx got hung */
    A_UINT32                hwResetCount;   /* Keep track of how many times we reset hardware */
    A_UINT32                osResetCount;   // How many times OS reset us
    A_UINT32                NdisSendRequest; // Number of times NDIS requested us to send a frame.
    A_UINT32                NdisSendReRequest; // Number of times NDIS had to hold a frame for us.

    A_UINT32                RxInvalidFrames;
    A_UINT32                ReceiveErrors;
    A_UINT32                RxDiscardFrames;
    A_UINT32                RxMgmtFrames;
    A_UINT32                RxCtrlFrames;
    A_UINT32                RxDataFrames;
    A_UINT32                RxAntenna;
    A_UINT32                AntCnt[2];          /* more someday */
    A_UINT32                AntSwCnt;

    A_UINT32                TxMulticastFrames;
    A_UINT32                TxInvalidFrames;
    A_UINT32                TransmitErrors;
    A_UINT32                TxDiscardFrames;
    A_UINT32                TxMgmtFrames;
    A_UINT32                TxCtrlFrames;
    A_UINT32                TxDataFrames;
    A_UINT32                TxConsecNoUnder;    // Number of consecutive frames tx'd w/o underruns

    A_UINT32                psMulticastFrames;
    A_UINT32                psRequestFrames;
    A_UINT32                psDiscardFrames;
    A_UINT32                psAgedFrames;
    A_UINT32                psSavedFrames;

    A_UINT32                Associations;
    A_UINT32                DisAssociations;
    A_UINT32                DisAssocReasonCode;
    A_UINT32                Authentications;
    A_UINT32                DeAuthentications;
    A_UINT32                DeAuthReasonCode;

    A_UINT64                GoodTransmitBytes;  // byte counts
    A_UINT64                GoodReceiveBytes;
} WLAN_STATS;

// NIC registers
typedef struct _REGISTER_VAL {
    A_UINT32    Offset;
    A_UINT32    Value;
} REGISTER_VAL;

// Defines for tpScale levels
#define TP_SCALE_LOWEST 0
#define TP_SCALE_MAX 0
#define TP_SCALE_50  1
#define TP_SCALE_25  2
#define TP_SCALE_12  3
#define TP_SCALE_MIN 4
#define TP_SCALE_HIGHEST 4

struct wlanDevInfo;             /* Forward reference */

#define CIPHER_SELECTION_WEP    0
#define CIPHER_SELECTION_AES    1
#define CIPHER_SELECTION_AUTO   2

/*
 *  The followings are the Tx Rate Control Definitions
 *  and data structures
 */

#define NO_RETRY            1       /* No retry */
#define MAX_TX_RETRIES      16      /* Max number of Tx (hw) retries */
#define MAX_REL_RATE        3       /* Number of adjacent Tx rates under consideration */
#define MIN_TX_RATE         1       /* Lowest Tx rate table index */
#define MAX_TX_RATE         7       /* Highest Tx rate table index */
#define MAX_TX_RATE_TBL     8       /* Highest Tx rate table length */


struct TxRetry_s {
    struct bin_s {
      A_UINT16  retryBin[NO_RETRY+MAX_TX_RETRIES]; /* 16+1 because 0 is used to store no retry count */
      A_UINT32  total_time;             /* Total Tx time */
      A_UINT32  frameLoss;              /* Frame Loss */
    } tbl[MAX_REL_RATE];
    A_UINT16    rateIdx;                /* Rate table index for Tx frame */
    A_UINT32    throttle_count;
    A_UINT32    throttle_dncnt;
    A_UINT32    throttle_upcnt;
    A_UINT32    rateDnCnt;
    A_UINT32    rateUpCnt;
    A_UINT32    frameCnt;               /* Frame count */
    A_UINT32    txCnt;
    A_UINT32    rateCnt[MAX_TX_RATE_TBL];
    A_UINT8     optRateCnt;             /* Optimal rate counter.
                                         * Use exploratory rate when count maxs out.
                                         */
    A_UINT8     optRateMax;             /* Max optimal rate count */
    A_BOOL      xLoRate;                /* Use low exploratory rate */
    A_BOOL      xHiRate;                /* Use high exploratory rate */
};

/*
 * State structures for new rate adaptation code
 */
typedef struct TxRateCrtlStateCB_s {
	A_UINT8 rssiThres;	   /* required rssi for this rate */
	A_UINT8 cost;		   /* recent cost of using this rate */
} TxRateCtrlStateCB;

struct TxRateCtrlCB_s {
    TxRateCtrlStateCB state[MAX_TX_RATE_TBL];	/* state for each rate */
    A_UINT8 rssiLast;	   	   /* last ack rssi */
    A_UINT8 rssiLastLkup;	   /* last ack rssi used for lookup */
    A_UINT8 rssiLastPrev;	   /* previous last ack rssi */
    A_UINT8 rateMax;		   /* max rate that has recently worked */
    A_UINT8 probeRate;		   /* rate we are probing at */
    A_INT8  antFlipCnt;		   /* number of consec times retry=2,3 */
    /* no dummy bytes */
    A_UINT8 misc[16];		   /* miscellaneous state */
    A_UINT16 rssiSum;		   /* running sum of rssi for averaging */
    A_UINT8 rssiSumCnt;		   /* count of rssiSum for averaging */
    A_UINT8 rssiSumRate;	   /* rate that we are averaging */
    A_UINT8 rate;		   /* last rate we returned */
    A_UINT32 rssiTime;		   /* msec timestamp for last ack rssi */
    A_UINT32 rssiDownTime;	   /* msec timestamp for last down step */
    A_UINT32 probeTime;		   /* msec timestamp for last probe */
    A_UINT8 hwMaxRetryRate;	   /* rate of last max retry fail */
    A_UINT8 hwMaxRetryPktCnt;	   /* number of packets since we got
				      HW max retry error */
    A_UINT8 antTx;		   /* current transmit antenna */
    A_UINT8 antProbeCnt;	   /* number of packets since ant probe */
};


// TODO: Add full rate element and mark off each one valid/invalid.
struct TxRateCtrl_s {
    struct      TxRetry_s sf;       /* Tx retry bin for shorter length frame */
    struct      TxRetry_s lf;       /* Tx retry bin for longer length frame */
    struct {
      A_UINT8   curRate;
      A_UINT8   newRate;
      A_UINT16  enable;             /* Enable RTS/CTS */
      A_UINT32  count;              /* Number of times RTS/CTS is used */
      A_UINT32  thruput[MAX_REL_RATE];
      A_UINT32  frameLoss[MAX_REL_RATE];            /* Frame Loss */
      A_UINT32  aging;
      A_UINT32  ticks;
    } rtsCts;
    A_UINT8     minTxRateIdx;      /* Min Tx Rate index */
    A_UINT8     maxTxRateIdx;      /* Max Tx Rate index */
    A_UINT8     work2do;
    A_UINT8     inUse;              /* Rate adjustment in progress */
    struct	TxRateCtrlCB_s cb;  /* new rate adaptation state info */
};


/*
 * sibEntry is a data structure pivotal to operation of all modules.
 * It maintains the state, callback pointers, timeouts, etc.
 * We need to define and add more information fields to this structure
 * for storing information about rates, fragmentation, power management,
 * etc. Hence it is expected that this structure will grow significantly
 * as the development progresses.
 *
 * First field is the link pointer. Rest of the fields should be
 * optimized for packing the fields in the structure. First pointers,
 * 32 bit values, then 16 bit values then variable length and byte
 * fields.
 */

typedef struct  sibEntry {
    struct sibEntry  * pNext;       /* pointer to next in the list */
    struct sibTable  * pSibTab;     /* back ptr to enclosing sib table */
    A_SEM_TYPE         sibLock;     /* Lock for individual SIB */
    struct wlanDevInfo *pDevInfo;
    WLAN_PRIV_RECORD * pPriv;       /* WEP key record */
    CHLG_TEXT        * pChlg;       /* pointer to challenge text holding */
                                    /* place during authentication */
    QUEUE_DETAILS   txPSQueue;      /* queue to hold tx frames under power save mode */
    ATHEROS_DESC    *smePendDesc;   /* SME frames pending on tx completion */
    A_INT32         numTxPending;   /* number of frames pending in TX queue */
    A_INT32         numPSTxPending; /* number of ps buffered frames pending in TX queue */
    WLAN_TIMESTAMP  lastBeaconTime; /* timestamp of last beacon */
    A_UINT32        devNo;          /* device number, identifier */
    WLAN_MACADDR    macAddr;        /* MAC address of station */
    BSSID           bssId;          /* BSSID of the network */
    A_UINT16        staIndex;       /* index into station table */
    A_UINT16        staState;       /* Current state of station */
    A_UINT32        transState;     /* Current state of transaction */
    A_INT16         keyIndex;       /* key cache index in use (WEP | swretry) */
    CAP_INFO        capInfo;        /* Capability of station */
    A_UINT16        serviceType;    /* STA, AP, ad-hoc beacon, etc. */
    A_UINT16        assocId;        /* association ID */
    A_UINT16        listenInterval; /* listen interval in units of beacon interval */
    A_UINT16        used;           /* flag to reflect the status of use */

    struct {
        A_UINT32    needClearDest;      /* flag meaning need to set the clearDestMask on next frame */
        A_INT32     numFilteredPending; /* # frames which will be filtered in queue */
        QUEUE_DETAILS   retryQueue;     /* buffered frames waiting for filtered frames to */
                                        /* be processed */
        A_INT32     qFrames;            /* number of frames in the retryQueue */
    } swRetry;

    A_UINT32        staLifeStartTime;       /* the time of the first frame from
                                               the station (ms)*/
    A_UINT32        staLastActivityTime;    /* the time of the last frame from */
                                            /*  this station (ms) */
    A_UINT32        authenAlgo;             /* authentication algorithm used by */
                                            /* this station */
    A_UINT32        psMode;                 /* power saveing mode */
    A_INT32         rssi;                   /* current received signal strength indicator */
    WLAN_STATS      stats;                  /* Structure of statistics counters */
    A_UINT32        lastRxSeqNo;            /* Last frame's sequence number */
    A_UINT32        lastRxFragNo;           /* Last frame's fragment number */
    A_UINT32        oneBeforeLastRxSeqNo;   /* Memory of the sequence no. before last. */
    A_UINT16        Channel;                /* radio frequency */
    A_UINT16        seqNum;                 /* sequence number */

    struct          TxRateCtrl_s txRateCtrl;/* Tx Rate Control */

    RATE_SET        operationalRateSet;     /* All supported operation rate by this station */
    RATE_SET        basicRateSet;           /* only supported basic operation rate by this station */
    RATE_SET        workingRateSet;         /* Working rates to be used for current tx*/
    A_UINT8         defaultRateIndex;       /* index of default rate for the station */
    A_UINT8         currentOperationalRateIndex;    /* index of current TX rate */
    A_UINT8         basicRateIndex;         /* index of broadcast/multicast/beacon TX rate */
    A_UINT8         sfRateIndex;
    A_UINT8         lfRateIndex;
    A_UINT8         lastRateIndex;          /* Rate used for the last packet */
    UCSE_SET        ucseSet;                /* All supported unicast cipher by this station */
    UCSE_SET        workingUcseSet;         /* working unicast cipher by this station */
    MCSE_SET        mcseSet;                /* only supported broadcast/multicast cipher by this station */
    A_UINT8         antTx;                  /* Antenna used for transmission */
    A_UINT32        rateAdaptationOff;      /* Rate adaptation off count */
    TAILQ_ENTRY(sibEntry) lruField;         /* element of lruList */
    A_BOOL          privacyInvoked;         /* privacy enable status */
    A_BOOL          swEncryptionEnabled;    /* sw encryption enable status */
    A_UINT8         dot1xState;             /* result of 1x auth */
    A_BOOL          bridged;
#ifdef VXWORKS
    A_UINT16        vlanTag;                /* vlan the sta belongs to */
    TRACEBUF_DEFINE(Tx)
    TRACEBUF_DEFINE(Rx)
#endif
} SIB_ENTRY;

/* SIB Table */
#ifdef BUILD_AP
#define MAX_STA_NUM 256         /* "big" # for APs */
#else
#define MAX_STA_NUM 61          /* key $ size + localsta */
#endif

typedef struct sibTable {
    SIB_ENTRY   sta[MAX_STA_NUM];
    SIB_ENTRY * pFreeList;
    A_SEM_TYPE  sibTableSem;
    A_UINT32    currentEntryNum;
    struct wlanDevInfo *pdevInfo;  /* back ptr to device info struct */
} SIB_TABLE;

extern SIB_TABLE    sibTable, *pSibTable;

/* SIB Hash Table. Pointers to SIB entries */
#define MAX_STA_HASH_NUM    256 /* Max. possible hash values */

typedef struct sibHashTable {
    /* Points to first sibEntry for this hash value */
    SIB_ENTRY   *pFirst[MAX_STA_HASH_NUM];
} SIB_HASH_TABLE;

extern SIB_HASH_TABLE sibHashTable, *pSibHashTable;

/*
 * MAC ACL data structures.  Created by WEP configuration.  Managed
 * with the same hash table scheme as the SIB_ENTRYs.
 */
typedef struct macACL_entry {
    /* Points to first macACL_entry for this hash value */
    struct macACL_entry *pNext;
    WLAN_MACADDR    mac;
    A_UINT16    flags;
#define MACACL_NONE     0x0000
#define MACACL_PRIVACYINVOKED   0x0001
#define MACACL_DENY     0x0002
#define MACACL_DEFAULT      0x0004
#define MACACL_KEY      0x0008
    A_UINT16    key;
} MACACL_ENTRY;

typedef struct macACLHashTable {
    MACACL_ENTRY    *pFirst[MAX_STA_HASH_NUM];
} MACACL_HASH_TABLE;

extern MACACL_HASH_TABLE macACLHashTable;

typedef void *  OS_DEVICE_ID;

/* Settings for 802.1x operation */
enum {
    KEYSRC_LOCAL = 0,
    KEYSRC_SERVER,
    KEYSRC_MIXED,
};

/* Following information needs to be stored on the STA during
 * the process of scanning. It has the requested parameters
 * and received beacon/probe resp information too.
 */

typedef struct scanInfoStruct {
    BSSID               bssid;
    SSID                ssid;
    BSSTYPE             bsstype;
    SCANTYPE            scantype;       /* active or passive */
    A_UINT32            probeDelay;     /* max probe delay time in usec */
    WLAN_SCAN_LIST      *pScanList;     /* array of ptrs to the list of chennels to scan */
    A_UINT16            curChannelIndex;/* List index of current channel being scanned */
    A_UINT32            minChannelTime; /* minimum waiting time */
    A_UINT32            maxChannelTime; /* max waiting time */
    A_UINT32            scanStartTime;  /* Timestamp of when scan started on this channel */
    A_UINT32            scanEndTime;    /* Timestamp of when scan finished, used for aging */
    A_BOOL              quickScan;      /* if TRUE, stop scanning after */
                                        /* first valid response */
    A_BOOL              probeRetry;     /* if set probe is re-transmitted */
    struct BSSDescriptionSet bssSet;    /* Beacons received during scan */
} SCAN_INFO;

/* Generic timer structure to be used by driver code */
typedef struct athTimerStruct {
    struct wlanDevInfo  *pdevInfo;      /* standard devinfo pointer */
    void            *pFunc;             /* Function to be called */
    A_UINT32        param;              /* parameter */
    A_UINT32        operation;          /* operation that we are timing on */
    A_UINT32        timeVal;            /* timer value in milliseconds */
    void           *osTimerObj;         /* OS specific timer object */
    A_BOOL          repeat;             /* true if periodic timer */
    A_BOOL          state;              /* true if timer cancelled */
} A_TIMER;

/*
 * WLAN_DRIVER_INFO structure will hold the driver global information.
 */
typedef struct wlanDrvInfo {
    struct osDrvInfo       *pOSHandle;  /* OS specific driver info structure */
    struct smeInfoStruct   *smeInfo;    /* SME related info, callback pointers */
    A_SEM_TYPE              driverSem;  /* Highest level semaphore, entire driver */
    A_UINT32                devCount;   /* No. of currently connected devices */
    struct wlanDevInfo      *pDevInfoArray[WLAN_MAX_DEV]; /* Array of devinfo pointers */
    struct eventQueue      *triggeredQ; /* used by dk.  pointer to q of triggered events */
} WLAN_DRV_INFO;

struct HwFuncs;                         /* Fwd Decl see hal.h */
struct eepMap;                          /* Fwd Decl see ar52XX.h */
struct cservInfoStruct;                 /* Fwd Decl  see stacserv.h */

/*
 * WLAN_DEV_INFO structure will hold all kinds of device related information.
 * It will hold OS specific information about the device and a device number.
 * Most of the code doesn't need to know what's inside that structure.
 * The fields are very likely to change.
 */

typedef int (*ISR_CALLBACK)(void *, ATHEROS_DESC*);

//-------------------------------------------------------------------------
// PCI Cards found - returns hardware info after scanning for devices
//-------------------------------------------------------------------------
typedef struct PciCardInfo {
    A_UINT32    BaseIo;
    A_UINT16    VendorID;
    A_UINT16    DeviceID;
    A_UINT16    SubVendorID;
    A_UINT16    SubVendorDeviceID;
    A_UINT16    SlotNumber;         // NDIS slot number
    A_UINT32    MemPhysAddress;     // NIC physical address
    A_UINT32    MemLength;          // NIC footprint
    A_UCHAR     ChipRevision;
    A_UCHAR     Irq;
} PCI_CARD_INFO;


// routine to call (typically MLME) when power state changes
typedef void (*PowerRequestCallback)(void*);

typedef struct wlanDevInfo {
#ifdef BUILD_AP
    END_OBJ             *pOSHandle;         // Pointer to OS specific device data
#else
    struct osDevInfo    *pOSHandle;         // Pointer to OS specific device data
#endif
    struct dkDevInfo    *pdkInfo;           // pointer to structure containing info for dk
    WLAN_STA_CONFIG     defaultStaConfig;   // default values for station configuration info
    WLAN_STA_CONFIG     staConfig;          // current values for station configuration info
    PCI_CARD_INFO       pciInfo;            // Struct to hold PCI related information
    struct sibTable     *pSibTable;         // back ptr to SIB table assoc with this device
    struct BeaconInfo   *pBeaconInfo;

    A_UINT32    devno;          // Device number
    SIB_ENTRY   *localSta;      // Pointer to station structure for the STA of this device
    BSSDESCR    *bssDescr;      // BSS Description of current AP
    SCAN_INFO   *pscanInfo;     // scan related info
    struct cservInfoStruct *pCsInfo; // connection services data structure
    A_UINT32    baseAddress;    // virtual base address of hardware (mapped)
    A_UINT32    physAddress;    // physical base address of hardware
    A_BOOL      mapped;         // memory mapping has been done?
    A_UINT32    mapLength;      // length of mapped region

    // not so common resource pools
//fadi #ifdef BUILD_AP
    char *      pDmaBuf;        /* ptr to device  memory */
    int         dmaBufSize;     /* size of user specified memory */
    char *      pMclMem;        /* ptr to MBLK/CL_BLR memory */
    unsigned long  flags;
    int         ivec;           /* Device interrupt vector */
    A_UINT32    txFramesDropped;
    A_UINT32    rxFramesDropped;
    A_UINT32    psFramesDropped;
    QUEUE_DETAILS psBufferQueue;
    A_UINT16    psQueueDepth;
    A_UINT16    rxQueueDepth;
    A_UINT16    maxLoanedCount;
//fadi #else
#if FADI
    void        *pDescMemory;       // pointer to start of descriptor memory block
    A_UINT32    descPhysAddr;       // physical address of descriptor memory block
    void        *pLargebufMemory;   // pointer to start of rx buffer memory block (big)
    A_UINT32    largebufPhysAddr;   // Physical address of rx buffer memory block
    void        *pSmallbufMemory;   // pointer to start of smallbuf memory block
    A_UINT32    smallbufPhysAddr;   // Physical address of smallbuf memory block
    void       *pCryptBuf;          // pointer to start of cryption buffers

    QUEUE_DETAILS   smallBufDescQueue;  // queue for empty descr/small buff pairs
    QUEUE_DETAILS   largeBufDescQueue;  // queue for empty descr/large buff pairs
    A_UINT16        rxDescCount;        // Receive descriptor count
//fadi #endif
#endif /* FADI */

    A_UINT16        numDefragLanes;
    PARTIAL_DESC    defragLanes[DEFRAG_ENTRIES_MAX];

    QUEUE_DETAILS   emptyDescQueue;     // queue of free Atheros descriptors
    QUEUE_DETAILS   txHeaderQueue;      // queue of free Atheros descriptors
    A_UINT16        freeDescCount;      // Transmit descriptor count
    A_UINT16        txHeaderDescCount;  // Receive descriptor count
    QUEUE_DETAILS   txQueue;            // queue details for transmit queue
    QUEUE_DETAILS   altQueue;           // alternate queue details
    QUEUE_DETAILS   rxQueue;            // receive queue details
    A_UINT32        intEnabled;
    ATHEROS_DESC    *pQueueCurrent;     // needed to keep track of where are at in rx queue
    A_UINT32        receiveCounter;     // keep a count of how many frames have received
    ISR_CALLBACK    ptrCallback;        // callback function to be called from ISR

    // The following descriptor points to the first Tx descriptor in the Tx
    // queue that has not been serviced by mlmeMgmtTxCompletion yet
    ATHEROS_DESC    *lastTxFrameDesc;

    // Transmit descriptors, after completion, can't be re-used until 1) the s/w has
    // done post-processing on them, 2) the h/w has moved onto the next frame to send.
    // Since 1) often lags 2) we need to keep track of two descriptor pointers

    // Pointer to next descriptor to be freed as soon as both s/w and h/w are done with it.
    ATHEROS_DESC    *holdingTxDesc;
    ATHEROS_DESC    *holdingAltDesc;

    // Pointer to last descriptor on hardware queue onto which new tx desc. are added.
    ATHEROS_DESC    *TxTailDesc;
    ATHEROS_DESC    *altTailDesc;

    ATHEROS_DESC    *pPSPollDesc;       // PS-Poll descriptor for hardware that
                                        // does not generate PS-Polls automatically

    // Device status related fields
    A_BOOL          NicResetInProgress; // NIC hardware reset ongoing
    A_BOOL          NdisResetInProgress;// NDIS reset request    "
    A_BOOL          NdisHaltInProgress; // NDIS halt request    "
    A_BOOL          txResetInProgress;  // don't try to re-enable h/w when TRUE
    A_UINT32        hwResetCount;
    A_UINT32        txUrn;
    A_UINT32        txUrnBurst;
    A_UINT32        txUrnBurstMs;
    A_UINT32        rxEOLcount;         // num of RXEOLs suffered in this period
    A_UINT32        lastRXEOLms;        // ms since last RXEOL
    A_UINT32        MaskReg;            // value written to interrupt mask register
    A_UINT32        globISRReg;         // current ISR value
    A_UINT32        NdisFilterReg;      // value of filter requested by NDIS
    A_UINT32        rxFilterReg;        // value written to F2_RX_FILTER register
    A_UINT32        gpio0Bit;           // last value read from GPIO register
    A_UINT32        initDataIndex;      // index into ar5kInitData[]
    A_UINT16        macVersion;         // MAC silicon version (Crete, Oahu etc).
    A_UINT16        macRev;             // MAC silicon revision
    A_UINT16        phyRev;             // PHY baseband silicon revision
    A_UINT16        analog5GhzRev;      // analog silicon revision
    A_UINT16        analog2GhzRev;      // analog silicon revision
    A_UINT16        channelCap;         // Channel capabilities, CCK, 2.4 GHz, etc.
    A_UINT16        dataQueueNum;       // Maps the current single data Q to QCU/DCU
    A_UINT16        beaconQueueNum;     // Beacon Q number
    A_UINT16        cabQueueNum;        // Crap After Beacon Q number
    A_UINT16        psPollQueueNum;     // psPoll Q number
    A_UINT16        activeQueueMask;    // Mask of active Q nums - set in halAttach
    A_UINT32        rxOrnCnt;           // Count of RX overruns
    A_BOOL          bIQCalibration;     // IQ calibration started

    A_UINT32        NumPacketsQueued;

    A_UINT32        linkIsActive;

    A_UINT32        totalRxIndicated;   // how many we've sent up the stack
                                        // cannot unload driver unless count is 0
    A_BOOL          InsufficientRxDesc; // TRUE if we're out of Rx descriptors
    A_BOOL          partialRxDesc;      // first descr in rx q is not a full frame

    struct athTimerStruct smeTimer;     // timer object for auth, assoc, join sme functions
    struct athTimerStruct scanTimer;

    A_SEM_TYPE          keySem;             // protect key
    A_UINT16            keyTableSize;       // total entries in keyTable
    A_UINT16            keyTableStaticSize; // static entries in keyTable
    A_UINT32            keyCacheSize;       // size of hardware key cache
    WLAN_PRIV_RECORD    *keyTable;          // encryption key list
    WLAN_PRIV_RECORD    **keyCache;         // encryption keys in HW
    SIB_ENTRY           **keyCacheSib;      // SIB owner of keys in HW

    BLOWFISH_CTX        *Ctx;               // Control struct for WEP key encryption
    void                *virtEncryptBuf;    // pointer to start of encrypt buf
    A_UINT32            physEncryptBuf;     // Physical address of encrypt buf
    void                *virtDecryptBuf;    // pointer to start of decrypt
    A_UINT32            physDecryptBuf;     // Physical address of decrypt buf
    A_BOOL              TransmitBusy;       // Some frames still in tx queue
    A_BOOL              apAlive;            // TRUE if beacon received from AP
    A_UINT32            PowerState;         // Current power state: awake/asleep/pending change
    ATHEROS_DESC        *powerChangeTxDesc; // descriptor associated with pending power change frame
    //void                (*PowerRequestCallback)();// routine to call (typically MLME) when power state changes
    PowerRequestCallback powerReqCallBack;
    A_INT32             noActivityInterval; // power down if not activity for this time
    A_INT32             noActivityCur;      // current no activity time
    A_UINT32            tx6PowerInDbm;      // The power output level for 6Mb tx
    A_UINT16            keyType;            // type of key, WEP, WEP2 or AES, negotiated in most recent assoc.
    A_INT32             numTxFramesPending; // number of outstanding tx frames, used for flow control
    A_UINT32            invalidTxChannel;   // Set when driver is off channel and cannot accept TX's
    A_UINT32            numSleepingClients; // number of associated clients in Power Save
    A_BOOL              dropAllRxFrames;    // allows us to stop receive more easily in s/w
    const struct HwFuncs *pHwFunc;          // HW Funcs ptr
    struct HalInfo      *pHalInfo;
    A_UINT32            phyActivityFlag;
    A_UINT32            phyHwLastFcsFailCount;
    A_UINT32            phyHwLastRssi;
    A_UINT32            phyHwLastBeaconCount;
} WLAN_DEV_INFO;


// Macro definitions.

// Some macros are used to mask off environment specific implementation of an operation
// like A_DRIVER_MALLOC(). They will be in corresponding OS specific
// include file like ndisdrv.h
#define WLAN_NO_PRIV        0
#define WLAN_WEP_PRIV       1

// table of transmit rates and various constants associated with them

#define RATE_TABLE_SIZE     8

// OFDM Rates
#define RATE_6_INDEX        0
#define RATE_9_INDEX        1
#define RATE_12_INDEX       2
#define RATE_18_INDEX       3
#define RATE_24_INDEX       4
#define RATE_36_INDEX       5
#define RATE_48_INDEX       6
#define RATE_54_INDEX       7

// CCK Rates, indexes for short preamble by default
#define RATE_1_INDEX        0
#define RATE_2_INDEX        1
#define RATE_5_5_INDEX      2
#define RATE_11_INDEX       3
#define CCK_SHORT_PREAMBLE_ENABLE   0x4 // bit 2 in plcp rate

// Probe request and response is sent at lowest basic rate, and long preamble
// in CCK mode. Index 0 works for all these conditions.
#define PROBE_RATE_INDEX    0
#define PSPOLL_RATE_INDEX   0
#define LOWEST_RATE_INDEX   0

// Order of fields in info structure is important because hardcoded structures
// are initialized in wlanglobal.c
typedef struct RateTableStruct {
    A_UINT16    rateCount;
    A_UINT16    phyType;
    A_UINT16    sifs;               // in usec
    A_UINT16    slot;               // in usec
    A_UINT16    preambleBits;   // No. of bits in preamble
    A_UINT16    plcpBits;       // No. of bits in PLCP
    A_UINT16    ackBits;        // No. of bits in ACK frame
    A_UINT16    ctsBits;        // No. of bits in CTS frame
    A_UINT8 plcpRateToIndex[16];    // backward mapping
    struct {
        A_UINT16    rateKbps;       // Rate in Kbits per second
        A_UINT8     plcpRate:4;     // rate that goes into 4 bits of PLCP and tx descriptor
        A_UINT8     dot11Rate;      // Value that goes into supported rates info element of MLME
        A_UINT8     controlRate;    // Index of next lower basic rate, used for duration computation
        A_UINT16    ackDuration;    // Standard ack duration for this rate
        A_UINT16    symBits;        // Bits per symbol
        A_BOOL      used;
    } info[RATE_TABLE_SIZE];
} RATE_TABLE_STRUCT;

// External declarations for shared global variables from wlanglobal.c

extern WLAN_DRV_INFO        globDrvInfo;            // Global driver data structure
extern A_UINT8              assocArray[];           // Bit array for Association ID's
extern WLAN_STA_CONFIG      staConfigArray[];       // default Configurations, test only
extern WLAN_MACADDR         beaconBroadcastAddr;    // Broadcast Address
extern WLAN_MACADDR         broadcastBssid;         // Broadcast BSSID for probe
extern const WLAN_MACADDR   lanBroadcastAddr;       // Broadcast address on the DSM
extern const WLAN_MACADDR   nullMacAddr;
extern struct smeInfoStruct smeInfo;
#ifndef BUILD_AP
extern struct osDevInfo     defaultOSInfo;
#endif
extern RATE_TABLE_STRUCT    globalRateTable11a;
extern RATE_TABLE_STRUCT    globalRateTable11b;
extern A_UINT32             numDeviceIDs;

// Now include the OS specific header files for our drivers

// NT application environment for dk and jungo drivers
#ifdef NT_HW
#include "ntdrv.h"
#endif

// NDIS miniport driver environment
#ifdef NDIS_HW
#include "ndisdrv.h"
#endif

#ifdef VXWORKS
#include "vxdrv.h"
#endif

#if defined(Linux)
#include "linuxdrv.h"
#endif

#ifdef __cplusplus
}
#endif

#endif /* _WLANDRV_H_ */
