/*
 * Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * wlanext.h - Function prototype declarations for general driver routines
 *
 * $Id: //depot/sw/releases/2.0.1/src/include/wlanext.h#3 $
 */

#ifndef __INCwlanexth
#define __INCwlanexth

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <asm/byteorder.h>
/*
#ifdef __BIG_ENDIAN
#define BIG_ENDIAN __BIG_ENDIAN
#endif
*/
#include "wlansme.h"
#include "wlandrv.h"

INFO_ELEMENT *
mgtElementGet(WLAN_MGT_MAC_HEADER *pmacHdr,
              A_UINT16 frameSize,       /* Size of complete frame */
              A_UINT8  elementId);      /* element ID to search for */

A_UINT16 macHeaderLenGet(FRAME_CONTROL_FIELD *frameControl);

A_INT32 wlanMACHash(WLAN_MACADDR *pmacAddr);

/***********************************************************
 * This routine parses an ASCII string that looks like a standard
 * ethernet mac address like nnnnnnnnnnnn or nn:nn:nn:nn:nn:nn or
 * '.' or '-' instead of ':'. It is flexible about single digit
 * values too. It places resulting 6 bytes address in pmac.
 *
 * RETURNS: FALSE if conversion failed, TRUE if it succeeds
 */
A_BOOL
asciiToMacaddr(A_UCHAR *pstr,           // Input string, not null terminated
               A_UINT16 alen,           // input length
               WLAN_MACADDR *pmac);     // pointer to resulting mac address

/***********************************************************
 * This routine parses an ASCII string of hex digits
 * and converts them into a byte array of corresponding values.
 * It also updates key length field.
 *
 * RETURNS: FALSE if conversion failed, TRUE if it succeeds
 */
A_BOOL
asciiToKey(A_UCHAR *pstr,               // Input string, not null terminated
           A_UINT16 alen,               // length of input string
           WLAN_PRIV_RECORD *pKey,
           A_UINT16 keylen,             // length of desired hardware key in bytes
           A_UINT16 encryptedKeyLen);   // length of returned key

/* sibTableInit - Initialize the SIBtable at system initialization stage */
A_STATUS sibTableInit(WLAN_DEV_INFO *);

/* sibEntryFind - find the entry which has matched macAddres */
SIB_ENTRY *sibEntryFind(WLAN_MACADDR *macAddr);

/* sibEntryGet - get the entry which is indexed */
SIB_ENTRY *sibEntryGet(A_UINT32 entryIndex);

/* sibEntryDel - delete (invalidate) the sibEntry from the table */
A_STATUS sibEntryDel(SIB_ENTRY *sibEntry);

/* sibEntryAge*() - age out the SIB entry */
A_BOOL sibEntryAgeCheck(SIB_ENTRY *sibEntry, A_UINT32 aging_period_ms);
void sibEntryAge(SIB_ENTRY *sibEntry);

/* sibTableStateReset() -- reset state of all SIBs in SIB table */
void sibTableStateReset(WLAN_DEV_INFO *);

/* staRoamNotify() -- Notify SME code that sta has roamed away */
void staRoamNotify(WLAN_DEV_INFO *pdevInfo, L2_UPDATE_FRAME *pDestAddr);

/* sibEntryAlloc - Allocate a SIB entry from sibTable. */
SIB_ENTRY *
sibEntryAlloc(SIB_ENTRY *localSib, BSSDESCR *apDescr, WLAN_MACADDR *macAddr);

/* sibEntryAdd - add a entry to the SIBTable */
A_STATUS sibEntryAdd(SIB_ENTRY *sibEntry);

SIB_ENTRY *sibTblFind(A_UINT32  staId);
SIB_ENTRY *sibNextStaFind(SIB_ENTRY *pCurSib);

/*
 * macACLFind - Look-up priv entry for mac in mac ACL hash table.
 * macACLUpdate - Store WEP info about mac in mac ACL hash table.
 * macACLCheck - Return if mac address is allowed and updates pKey.
 */
MACACL_ENTRY *macACLFind(WLAN_MACADDR *mac);
void macACLUpdate(WLAN_MACADDR *mac, A_UINT32 flags, A_UINT32 key);
A_BOOL macACLCheck(WLAN_DEV_INFO *, WLAN_MACADDR *, SIB_ENTRY *);

/*
 * Privacy related routines
 *
 *   wlanKeyTableInit   - Initialize the encryption key table and cache.
 *   wlanKeyTableFree   - Free software resources allocated in wlanKeyTableInit.
 *   wlanKeyTableAdd    - Add or update an entry to the key table.
 *   wlanKeyTableRemove - Remove an entry from the key table (and cache).
 *   wlanKeyEnable      - Add instance of key @ keyTableIndex for pSib.
 *   wlanKeyFree        - Remove dynamic key table entry and clean-up key cache.
 *   wlanKeyCacheSet    - Add the key for pSib to the HW cache.
 *   wlanKeyCacheFault  - Ensure a SIB specific key is in the key cache.
 *   wlanKeyCacheFree   - Remove entry from sw cache and hw cache.
 *
 *   Blowfish_Init      - Initialize WEP key encryption structure
 *   Blowfish_Encrypt   - Encrypt WEP key
 *   Blowfish_Decrypt   - Decrypt WEP key
 *
 *   swretryDestMaskFault      - Fault occured on this dest mask index.
 *   swretryDestMaskVictimized - This dest mask index was stolden.
 */

A_STATUS wlanKeyTableInit(WLAN_DEV_INFO *pdevInfo, A_UINT16 nstatic,
                          A_UINT16 ndynamic);
void     wlanKeyTableFree(WLAN_DEV_INFO *pdevInfo);

WLAN_PRIV_RECORD *
wlanKeyTableAdd(WLAN_DEV_INFO   *pdevInfo,
                A_UINT16        keyIndex,
                A_UINT32        keyType,
                A_UINT32        keyLength,      /* in bits */
                A_UINT8         *keyValue);


A_BOOL
wlanKeyTableAddDynamic(WLAN_DEV_INFO    *pdevInfo,
                       SIB_ENTRY        *pSib,
                       A_UINT32         keyType,
                       A_UINT32         keyLength,      /* in bits */
                       A_UINT8          *keyValue);

void
wlanKeyEnable(WLAN_DEV_INFO *, SIB_ENTRY *, A_UINT16 keyTableIndex);
void
wlanKeyTableRemove(WLAN_DEV_INFO *pdevInfo, A_UINT16 keyIndex);
void
wlanKeyTableRemoveDynamic(WLAN_DEV_INFO *pdevInfo, WLAN_PRIV_RECORD *pKey);
void
wlanKeyFree(WLAN_DEV_INFO *, SIB_ENTRY *);

A_UINT16
wlanKeyCacheSet(WLAN_DEV_INFO   *pdevInfo,
                A_UINT16        keyIndex,       /* index of key entry */
                SIB_ENTRY       *pSib,
                WLAN_MACADDR    *pmacAddr,
                WLAN_PRIV_RECORD *pKey);

A_UINT16 wlanKeyCacheFault(WLAN_DEV_INFO *, SIB_ENTRY *, A_BOOL *faulted);
void     wlanKeyCacheFree(WLAN_DEV_INFO *, SIB_ENTRY *);
A_STATUS wlanKeyUpdate(WLAN_DEV_INFO *, SIB_ENTRY *, A_UINT32);
A_UINT16 wlanKeyMcIndexGet(WLAN_DEV_INFO *, SIB_ENTRY *, int *);

#define CONTEXT_NONE	0x00
#define CONTEXT_DOT1X	0x01

A_STATUS
aesKeyExtender( WLAN_DEV_INFO   *pdevInfo,
    SIB_ENTRY *pSib,
    WLAN_PRIV_RECORD *pKey,
    A_UCHAR *pKeyVal
    );
#ifndef BUILD_AP
//fadi A_STATUS processWEPKeys(WLAN_DEV_INFO *, NDIS_STRING *);

void Blowfish_Init(BLOWFISH_CTX *ctx, unsigned char *key, int keyLen);
void Blowfish_Encrypt(BLOWFISH_CTX *ctx, unsigned long *xl, unsigned long *xr);
void Blowfish_Decrypt(BLOWFISH_CTX *ctx, unsigned long *xl, unsigned long *xr);
#endif

void swretryDestMaskVictimized(SIB_ENTRY *pVictimSib);
void swretryDestMaskFault(SIB_ENTRY *, ATHEROS_DESC *);
void swretryQueueFlush(WLAN_DEV_INFO *, SIB_ENTRY *);
void descDurationSet(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc,
                     A_UINT16 rateIdx, A_BOOL swap, A_BOOL shortPreamble);

/*
 * Tx Connection services routines.
 */
int txRateFind(SIB_ENTRY *, int frameLen);
int txMbRateFind(int frameLen);
void RateCtrlTxRetryCntUpdate(int txRate, int excessive_retries, int retries,
                              int frameLen, SIB_ENTRY *, A_UINT8 rssiAck,
			      A_UINT8 antTx, A_UINT8 *flipReq);
void RateCtrlRxCntUpdate(SIB_ENTRY *, int);
void RateCtrlTask(void);
void RateCtrlSibInit(SIB_ENTRY *);
void RateCtrlSibTxRateInit(SIB_ENTRY *);
void swretryTxRateBackoff(ATHEROS_DESC *);
void swretryDefaultTxRateSet(ATHEROS_DESC *);
void RateCtrlBBChange(WLAN_DEV_INFO *pdevInfo, A_UINT16 channelFlags);
A_BOOL isRateCtrlEnabled(void);

/* Station environment initialize */
A_STATUS wlanStaInit(WLAN_DEV_INFO *pdevInfo, A_UINT16 serviceType);

void drvDescSend(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_UINT16 len);
void drvFrameSend(WLAN_DEV_INFO *pdevInfo, A_UINT8 *pBuf, A_UINT16 len);

A_STATUS setupEncrypt(WLAN_DEV_INFO *, ATHEROS_DESC *);
A_STATUS setupDecrypt(WLAN_DEV_INFO *, ATHEROS_DESC *);

#ifdef BUILD_AP
#define initCipher(x)
#else
void initCipher(WLAN_DEV_INFO *pDevInfo);
#endif
A_STATUS swEnCipher(WLAN_DEV_INFO *, WLAN_PRIV_RECORD *, ATHEROS_DESC *);
A_STATUS swDeCipher(WLAN_DEV_INFO *, WLAN_PRIV_RECORD *, ATHEROS_DESC *);

void setupSwEncryption(WLAN_DEV_INFO *pdevInfo);

void     descQInit(QUEUE_DETAILS *pQueue);
A_STATUS descQLock(QUEUE_DETAILS *pQueue);
A_STATUS descQUnlock(QUEUE_DETAILS *pQueue);

/* descQGetHead - Get pointer to first descriptor on the queue, don't remove it
 *
 * DESCRIPTION:
 * Returns a pointer to first descriptor on the queue. Doesn't remove it from
 * the queue. Used by the interrupt handler or polling routines to look at the
 * first descriptor on tx or rx queue. Those routines will call descQRemove
 * only if the descr is done.
 *
 * RETURNS: pointer to first descriptor on the queue
 */
ATHEROS_DESC *descQGetHead(QUEUE_DETAILS *pQueue);

/* descQGetTail - Get pointer to last descriptor on the queue, don't remove it
 *
 * DESCRIPTION:
 * Returns a pointer to last descriptor on the queue. Doesn't remove it from
 * the queue.
 * RETURNS: pointer to last descriptor on the queue
 */
ATHEROS_DESC *descQGetTail(QUEUE_DETAILS *pQueue);

/**************************************************************************
* descQremove - remove the descriptor from queue
*
* Removes the descriptor from queue.
* Sets the new head of the queue to the next descriptor
* Typically used for a queue descriptor that is somewhere in the middle
* RETURNS: N/A
*/
void descQRemove
    (
    QUEUE_DETAILS *pQueue,          /* pointer to queue details struct */
    ATHEROS_DESC  *pDesc
    );

/**************************************************************************
* descQPopHead - remove the descriptor from the start of queue
*
* Removes and then deletes the descriptor from the head of the queue.
* Sets the new head of the queue to the next descriptor
*
* RETURNS: pointer to descriptor at head of the queue
*/
ATHEROS_DESC *descQPopHead(QUEUE_DETAILS *pQueue);

/**************************************************************************
* descQPushHead - add the descriptor(s) to the start of queue struct
*
* Adds the given descriptor chain to start of the queue structure
*
* RETURNS: N/A
*/
void descQPushHead(ATHEROS_DESC *pDesc, QUEUE_DETAILS *pQueue);

/**************************************************************************
* descQPushTail - add the descriptor(s) to the tail of queue struct
*
* Adds the given descriptor chain to the tail of the queue structure
*
* RETURNS: N/A
*/
void descQPushTail(QUEUE_DETAILS *pQueue, ATHEROS_DESC *pDesc);

void removeRxField(ATHEROS_DESC *pDesc, A_UINT32 FieldLen);
void addTxField(ATHEROS_DESC *pDesc, A_UINT32 FieldLen);

/* Macros that call endian conversion routine
 * Atheros device and 802.11 frame contents are little endian.
 */
#ifdef BIG_ENDIAN
#define FRAME_CONVERT_CPUTOLE(pFrame, len)  frameEndianChange((WLAN_FRAME_HEADER *)(pFrame), ((A_UINT16)(len)), 1)
#define FRAME_CONVERT_LETOCPU(pFrame, len)  frameEndianChange((WLAN_FRAME_HEADER *)(pFrame), ((A_UINT16)(len)), 0)
#else
#define FRAME_CONVERT_CPUTOLE(pFrame, len)
#define FRAME_CONVERT_LETOCPU(pFrame, len)
#endif

/* frameEndianChange - Endian conversion of all kinds of 802.11 frames
 *
 * DESCRIPTION
 * Byte oriented fields do not need to be swapped, beccause they are
 * already in byte order. e.e. MAC addresses, INFO_ELEMENT, RATE_SET, etc.
 */
void
frameEndianChange(WLAN_FRAME_HEADER *pFrame, A_UINT16 len, int direction);

#ifdef BUILD_AP

void   apReboot(void);
void   apHalt(void);
A_BOOL isApShutdown(void);

int  strtoWlanMAC(char *macstr, WLAN_MACADDR *mac);
int  cliWrite(const char * format, ...) __ATTRIB_PRINTF;
void cliUptimeGet(char *p);

void sibSmeShow(void);
void sibStationShow(int);
void sibCipherShow (SIB_ENTRY *);

void chipReset(WLAN_DEV_INFO *, A_BOOL, CHAN_VALUES *);

char *httpPrintSibPrivacy(SIB_ENTRY *pSib);

STATUS beaconWdgCancel(void);

ATHEROS_DESC *
nullDataFrameCreate(WLAN_DEV_INFO *pdevInfo, SIB_ENTRY *pSibEntry);

#endif /* BUILD_AP */

/**************************************************************************
* Higher level functions
*/

A_STATUS PrepareForTransmit(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

A_STATUS txService(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pTxDesc, A_BOOL bypass);

A_STATUS wlanFrameTransmit(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc, A_BOOL bypass);

void wlanFrameReceive(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

A_STATUS powerInitialization(WLAN_DEV_INFO *pDevInfo);

A_STATUS powerSetMode(WLAN_DEV_INFO *pDevInfo, A_UINT16 ModeRequested,
                      void *callBack);

A_BOOL powerEnabled(WLAN_DEV_INFO *pDevInfo);

A_STATUS powerSendRequest(WLAN_DEV_INFO *pDevInfo);

A_STATUS powerFrameTransmit(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

void powerMgtPoll(WLAN_DEV_INFO *pDevInfo);

void powerReceiveIndication(WLAN_DEV_INFO *pDevInfo);

A_STATUS powerSendMgtFrame(WLAN_DEV_INFO *pDevInfo, A_BOOL PowerMgt);

void powerSendMgtCompletion(WLAN_DEV_INFO *pdevInfo, A_BOOL pktTransmitOK,
                            A_BOOL pwrMgt);

A_STATUS powerPSPollInit(WLAN_DEV_INFO *pDevInfo);
void     powerPSPollFree(WLAN_DEV_INFO *pDevInfo);
void     powerHwResetNotify(WLAN_DEV_INFO *pDevInfo);

void     sibPowerStateClean(SIB_ENTRY *pSib);

A_STATUS SetupTxHW(WLAN_DEV_INFO *pdevInfo, ATHEROS_DESC *pDesc);

void sibCseStateClean(SIB_ENTRY *pSIb);

/* Station environment initialize */
A_STATUS stationInitialize(WLAN_DEV_INFO *pdevInfo, A_UINT16 serviceType);

A_STATUS athDrvRadioEnable(WLAN_DEV_INFO *pDevInfo);

/* 802.1X support */
A_STATUS dot1xInit(void);
void    EAPportMsg(WLAN_DEV_INFO *, WLAN_MACADDR *, A_UINT8);
A_BOOL  isDot1xFrame(WLAN_DEV_INFO *, WLAN_DATA_MAC_HEADER3 *);
void    dot1xReceive(WLAN_DEV_INFO *, ATHEROS_DESC *);

A_BOOL isDynamicKey(WLAN_PRIV_RECORD *);

void deriveBroadcastKey(WLAN_DEV_INFO *pdevInfo, char *bcKeyBuf, int len);

A_STATUS athDrvRadioDisable(WLAN_DEV_INFO *pDevInfo);

int enterDebugger(void);

/* Now include the OS specific header files for our drivers */

/* NDIS miniport driver environment */
#ifdef NDIS_HW
#include "ndisdrvext.h"
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif // __INCwlanexth
