/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 */

/* wlanglobal.c - Global variable definitions. */

#ident "ACI $Id: //depot/sw/releases/2.0.1/src/common/wlanglobal.c#2 $"

#ifdef  VXWORKS
#include "vxworks.h"
#endif

#include "wlantype.h"
#include "wlansme.h"
#include "wlansmeext.h"

WLAN_DRV_INFO  globDrvInfo;                   /* Global driver info */

WLAN_MACADDR   beaconBroadcastAddr = {
    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};

const WLAN_MACADDR lanBroadcastAddr = {
    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};

WLAN_MACADDR    broadcastBssid = {
    { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }
};

const WLAN_MACADDR nullMacAddr = {
    { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};

struct smeInfoStruct    smeInfo;                /* Holds global info about SME */

A_UINT8        assocArray[(ASSOCID_MAX+7)/8];   /* Bit array for Association ID's */

/* Table of Station Information Blocks */
SIB_TABLE   *pSibTable;

/* Table of hash pointers to Station Information Blocks */
SIB_HASH_TABLE  sibHashTable;
SIB_HASH_TABLE *pSibHashTable= &sibHashTable;

/* Table of Duration values for transmit frame control duration field
 * Currently these values are tabulated, indexed on tx rate.
 * The values are in microseconds.
 * Duration = SIFS + Ack transmit time
 * where Ack transmit time =
 * | PLCP_TIME    |             frame bits                |
 * | 16 us | 4us  | 16 zeros + 112 ack bits + 6 tail bits |
 * | or 16 us for SHORT20 |       |      |     # of symbols * 4 us                                  |
 *
 * Frame bits tx time is # of symbols * 4us
 *
 * Bits per symbol are:
 * 6  Mb BPSK 1/2 = 24 bits per symbol
 * 9  Mb BPSK 3/4 = 36
 * 12 Mb QPSK 1/2 = 48
 * 18 Mb QPSK 3/4 = 72
 * 24 Mb 16QAM 1/2 = 96
 * 36 Mb 16QAM 3/4 = 144
 * 48 Mb 64QAM 2/3 = 192
 * 54 Mb 64QAM 3/4 = 216
 *
 * e.g. at 36 = 18 Mb X 2 for turbo, ACK always has 14 data bytes (112 bits), total 134 bits.
 * 18 Mb QPSK 3/4 is 72 bits per symbol, but in turbo, we get 144 bits per symbol. Hence, entire ack
 * needs 1 symbol.
 * duration = SIFS + (short 20) + long + sig + 1 symbol + SIFS + 2 * slot time
 *            16        16          8     4       4        16    18 (2 * 9 )
 * total = 82 us
 *
 * Ack can come back at the lower speed than transmitted packet. Hence we
 * should use the longest time anyway.
 */


/* Declare the rate table made up of all constants */
/* Bit 7 in dot11 field of basic rate is set to 1 */

// Rate table for 11a i.e. OFDM from 6 to 54 MBps
RATE_TABLE_STRUCT globalRateTable11a = {
    RATE_TABLE_SIZE,        /* count is maximum size */
    PHY_OFDM,
    16,                     /* SIFS time (usec) */
    9,                      /* slot time (usec) */
    144,                    /* preamble bits */
    48,                     /* plcp bits */
    134,                    /* ACK 112 bits + 16 zeros + 6 tail bits */
    134,                    /* CTS bits, same as ACK */
    {-1, -1, -1, -1, -1, -1, -1, -1, 6, 4, 2, 0, 7, 5, 3, 1}, /* plcpRateToIndex */
    {
    /*          Kbps        plcp    dot11           control ackdur symBits used */
    /* 6 Mb */  { 6000,     0xb,    (12 | 0x80),    0,      102,    24,     1 },
    /* 9 Mb */  { 9000,     0xf,    18,             0,      94,     36,     1 },
    /* 12 Mb */ { 12000,    0xa,    (24 | 0x80),    2,      90,     48,     1 },
    /* 18 Mb */ { 18000,    0xe,    36,             2,      86,     72,     1 },
    /* 24 Mb */ { 24000,    0x9,    (48 | 0x80),    4,      86,     96,     1 },
    /* 36 Mb */ { 36000,    0xd,    72,             4,      82,     144,    1 },
    /* 48 Mb */ { 48000,    0x8,    96,             4,      82,     192,    1 },
    /* 54 Mb */ { 54000,    0xc,    108,            4,      82,     216,    1 }
    }
};

// Rate table for 11b i.e. CCK
RATE_TABLE_STRUCT globalRateTable11b = {
    4,        /* count is only 4 rates */
    PHY_CCK,
    10,                     /* SIFS time (usec) */
    20,                     /* slot time (usec) */
    144,                    /* preamble bits */
    48,                     /* plcp bits */
    112,                    /* CCK ACK is just 112 data bits */
    112,                    /* CTS bits same as ACK */
    {-1, -1, -1, -1, -1, -1, -1, -1, 3, 2, 1, 0, 3, 2, 1, 0}, /* plcpRateToIndex */
    {
    /*            Kbps      plcp    dot11           control ackdur symBits used */
    /* 1 Mb */  { 1000,     0xf,    (2 | 0x80),     0,      104,    36,     1 },
    /* 2 Mb */  { 2000,     0xe,    (4 | 0x80),     1,      96,     72,     1 },
    /* 5.5 Mb */{ 5500,     0xd,    (11 | 0x80),    1,      92,     144,    1 },
    /* 11 Mb */ { 11000,    0xc,    (22 | 0x80),    1,      92,     216,    1 }
    }
};

