/*
 *  Copyright (c) 2000-2002 Atheros Communications, Inc., All Rights Reserved
 *
 * wlansme.h - MLME/PLME function prototypes and data structures
 *
 * $Header: //depot/sw/releases/2.0.1/src/include/wlansme.h#1 $
 */

#ifndef     __INCwlansmeh
#define     __INCwlansmeh

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#include "wlantype.h"
#include "wlanproto.h"

/* Defaults for some station parameters */
#define DEFAULT_ADHOC_BEACON_INTERVAL   100
#define DEFAULT_AP_BEACON_INTERVAL      100
#define DEFAULT_DTIM_PERIOD             1

/* Macros */

/* Following macro is used to determine effective size
 * of info element. Used for allocating space for element.
 */
#define INFO_ELEMENT_SIZE(e) (((e).length > 0)? (e).length+2:0)

/*
 * PHY Parameters as in Section 10.4.3.2 Semantics of the service primitive
 */

typedef struct PHYParamSet {
    A_UINT32    aSlotTime;
    A_UINT32    aSIFSTime;
    A_UINT32    aCCATime;
    A_UINT32    aRxTxTurnaroundTime;
    A_UINT32    aTxPLCPDelay;
    A_UINT32    aRxPLCPDelay;
    A_UINT32    aRxTxSwitchTime;
    A_UINT32    aTxRampOnTime;
    A_UINT32    aTxRampOffTime;
    A_UINT32    aTxRFDelay;
    A_UINT32    aRxRFDelay;
    A_UINT32    aAirPropagationTime;
    A_UINT32    aMACProcessingDelay;
    A_UINT32    aPreambleLength;
    A_UINT32    aPLCPHeaderLength;
    A_UINT32    aMPDUDurationFactor;
    A_UINT32    aMPDUMaxLength;
    A_UINT32    aCWmin;
    A_UINT32    aCWmax;
} PH_PARAM_SET;

typedef enum {
    MLME_SUCCESS,
    MLME_INVALID_PARAM,
    MLME_NOT_SUPPORTED,
    MLME_TOOMANY_REQ,
    MLME_TIMEOUT,
    MLME_REFUSED,
    MLME_ALREADY_IN_BSS,
    MLME_DRIVER_ERROR,      /* Errors not in the MLME RESULT CODE */
    MLME_OP_PENDING
} MLME_RESULT;

typedef enum {
    INFRASTRUCTURE_BSS = 1, /* MIB spec specifies this */
    INDEPENDENT_BSS,
    ANY_BSS
} BSSTYPE;

typedef enum {
    ACTIVE_SCAN,
    PASSIVE_SCAN,
    ANY_SCAN
} SCANTYPE;

typedef enum {
    MLME_INVALID_TIMEOUT,
    MLME_AUTH_TIMEOUT,
    MLME_ASSOC_TIMEOUT,
    MLME_JOIN_TIMEOUT,
    MLME_SCAN_TIMEOUT,
    MLME_BEACON_TIMEOUT,
    MLME_PWRMGT_TIMEOUT
} TIMEOUTTYPE;

/*
 * BSS description structure. Each set is a max size element.
 * This will be a fixed size data structure to accomodate all sets
 * at their max sizes. It is used in Scan, Start operations.
 * Length field in each set
 */
#define BAD_AP      0x01
#define AP_RETRY    0x10

typedef struct BSSDescription {
    BSSID           bssid;
    SSID            ssid;
    BSSTYPE         bsstype;
    A_UINT16        beaconInterval;
    A_UINT8         DTIMPeriod;
    WLAN_TIMESTAMP  timestamp;
    WLAN_TIMESTAMP  localTime;
    PH_PARAM_SET    phyParamSet;
    DS_PARAM_SET    dsParamSet;
    CF_PARAM_SET    cfParamSet;
    IBSS_PARAM_SET  ibssParamSet;
    CAP_INFO        capabilityInfo;
    RATE_SET        bssRateSet;
    RATE_SET        bssBasicRateSet;
    UCSE_SET        bssUcseSet;
    MCSE_SET        bssMcseSet;
    struct channelValues *pChannel;
    A_UINT16        timOffset;
    A_INT32         rssi;
    A_UINT32        scanTime; /* Time when beacon/probe received, used for aging */
    A_CHAR          apState;  /* valid values 0-3 */
} BSSDESCR;

/*
 * Set of BSS descriptions collected by SCAN.
 */

#define BSSDESCR_SET_SIZE   32  /* Max. scanned bss we can hold */
                                /* Do we expect more than 32 */
                                /* different AP's?  */
typedef struct BSSDescriptionSet {
    A_UINT16    count;                            /* No. of BSS description items in the set */
    A_CHAR      setIdx;                           /* index in the bss set */
    BSSDESCR    bssDescrArray[BSSDESCR_SET_SIZE]; /* Array of BSS description items */
} BSSDESCR_SET;

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __INCwlansmeh */
