/*
 *  Copyright (c) 2000-2001 Atheros Communications, Inc., All Rights Reserved
 */

/* wlansmeext.h - MLME/PLME function prototypes */

#ifndef     __INCwlansmeexth
#define     __INCwlansmeexth

#ident  "ACI $Header: //depot/sw/releases/2.0.1/src/include/wlansmeext.h#1 $"

/*
DESCRIPTION
This file contains the function prototypes specific to MLME

INCLUDE FILES: wlantype.h
*/

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */


#include "wlantype.h"
#include "wlanproto.h"
#include "wlansme.h"
#include "wlandrv.h"

/*
 * Most of the routines need a pointer to devInfo as the first parameter.
 * It is obtained during initialization with wlanMlmeBindDevice. It is
 * used to look up all the auxilliary information in addition to the
 * primitive parameters.
 */

 /* Global data structure maintained by MLME to hold its
 * static information.
 */

typedef struct smeInfoStruct
{
    A_UINT16   serviceType; /* WLAN_STA_SERVICE, WLAN_AP_SERVICE */
    void       (*authCallback)(WLAN_DEV_INFO *, WLAN_MACADDR *, A_UINT16, MLME_RESULT); /* Authentication confirm callback */
                                /* routine pointer */
    void       (*authIndicationCallback)(WLAN_DEV_INFO *, WLAN_MACADDR *, A_UINT16);    /* Authentication indication callback */
    void       (*deauthCallback)
                    (WLAN_DEV_INFO *, WLAN_MACADDR *, MLME_RESULT); /* Deauth confirm callback */
    void       (*deauthIndicationCallback)
                    (WLAN_DEV_INFO *, WLAN_MACADDR *, MLME_RESULT); /* Deauth indication callback */
    void       (*assocCallback)
                    (WLAN_DEV_INFO *, MLME_RESULT); /* Association confirm callback */
    void       (*assocIndicationCallback)
                    (WLAN_DEV_INFO *, WLAN_MACADDR *);  /* Association indication callback */
    void       (*reassocCallback)(WLAN_DEV_INFO *, MLME_RESULT);    /* Reassociate cb */
    void       (*reassocIndicationCallback)(WLAN_DEV_INFO *, WLAN_MACADDR *);   /* Reassociate cb */
    void       (*disassocCallback)
                    (WLAN_DEV_INFO *, WLAN_MACADDR *, MLME_RESULT); /* Disassociate confirm cb */
    void       (*disassocIndicationCallback)
                    (WLAN_DEV_INFO *, WLAN_MACADDR *, MLME_RESULT); /* Disassociate indication cb */
    void       (*pwrMgtCallback)(WLAN_DEV_INFO *, MLME_RESULT, A_BOOL); /* Power management cb */
    void       (*scanCallback)(WLAN_DEV_INFO *, BSSDESCR_SET *, MLME_RESULT);       /* Scan cb */
    void       (*joinCallback) (WLAN_DEV_INFO *, MLME_RESULT);  /* Join cb routine */
    void       (*startCallback) (WLAN_DEV_INFO *, MLME_RESULT); /* Start cb routine */
    void       (*resetCallback)(WLAN_DEV_INFO *);       /* Reset cb */
} SME_INFO_STRUCT;

/*
 * wlanMlmePwrmgtRequest - Change PS mode of STA
 * MLME Power management primitive changes the curent power saving mode.
 * This call will change the setting and return. MLME module will
 * send a null frame to AP to indicate the state transition or use an
 * existing scheduled transmission. When going into PS mode, the packets
 * already scheduled to transmit will be transmitted before switching the
 * mode. If the transmission attempts by MAC hardware fail, then software will
 * suspend those packets instead of going into PS mode. When going into
 * active mode, if wakeup is TRUE, MAC is immediately
 * forced into active mode instead of waiting for beacons.
 *
 * RETURNS: MLME_SUCCESS, MLME_NOT_SUPPORTED or MLME_INVALID_PARAM
 */

MLME_RESULT wlanMlmePwrmgtRequest
    (
    WLAN_DEV_INFO *pdevInfo,    /* device information structure */
    A_UINT16       powermgtMode,/* PS_MODE_ACTIVE,PS_MODE_POWER_SAVE */
    A_BOOL         wakeup,      /* TRUE for immediate active mode */
    A_BOOL         receiveDTIMs /* TRUE if STA should wake up for */
                                /* each DTIM */
    );

/* wlanMlmePowermgtConfirm - Confirm the powermgt operation's result
 *
 * This routine is used to call pwrmgtCallback() to inform the operation
 * result. One possible reason for failure is that MAC fails to transmit
 * successfully a packet with PwrMgt bit set.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM
 */
void wlanMlmePowermgtConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    MLME_RESULT    result,
    A_BOOL         powerStatus
    );

void mlmeScanFunction
    (
    WLAN_DEV_INFO *pdevInfo
    );

/*
 * wlanMlmeScanRequest - scan for potential BSSs.
 * MLME SCAN primitive
 *
 * PARAMETERS:
 *  bsstype - INFRASTRUCURE_BSS or INDEPENDENT_BSS or both. It will accept
 *            beacons from either or both.
 *  bssid   - A specific BSSID to scan for or broadcast MAC address for
 *            accepting all of them.
 *  ssidStr - Name of Service set. 32 bytes max. NULL value means accept any
 *            SSID.
 *  scanType - ACTIVE_SCAN or PASSIVE_SCAN
 *  probeDelay - MAC will observe a specific algorithm before jumping in on
 *               the WM. probeDelay is the maximum time that it will wait.
 *               This parameter is in microseconds.
 *  channelList - List of channels to scan.
 *  minChannelTime - Listen to each channel for at least this long.
 *  maxChannelTime - TU. Must switch to next channel after this time.
 *  quickScan - Stop scanning after receiving one valid beacon or probe
 *              response.
 *
 * ALGORITHM:
 * For each channel
 *      send MAC_CMD_CHANNEL command to MAC driver
 *      if ACTIVE_SCAN
 *      then
 *          send MAC_CMD_SEND_PROBE command to MAC driver to send probe packet
 *          wait for probe response packet
 *          if valid probe responese and quickScan
 *              terminate scan
 *      else
 *          enable beacon delivery from MAC to host
 *          wait for beacons on that channel
 *          if valid beacon and quickScan
 *              terminate scan
 *      end
 *  end
 *
 * Minimum and maximum waiting time condition is also observed on each channel.
 *
 * RETURNS: MLME_SUCCESS or MLME_INVALID_PARAM
 */
MLME_RESULT wlanMlmeScanRequest
    (
    WLAN_DEV_INFO  *pdevInfo,
    BSSTYPE         bsstype,        /* Infra., indep. or any */
    BSSID          *bssid,          /* Specific MAC or broadcast addr */
    SSID           *ssid,           /* ESS identifier or NULL */
    SCANTYPE        scantype,       /* active or passive */
    A_UINT32        probeDelay,     /* max probe delay time in usec */
    WLAN_SCAN_LIST *pScanList,      /* list of chennels to scan */
    A_INT32         minChannelTime, /* minimum waiting time */
    A_INT32         maxChannelTime, /* max waiting time */
    A_BOOL          quickScan       /* if TRUE, stop scanning after */
                                    /* first valid response */
    );

/*
 * wlanMlmeScanConfirm - returns result of scan
 *
 * It calls back scanCallback() routine to return the set of descriptions
 * for the BSS that were found on the WM.
 *
 * RETURNS: MLLE_SUCCESS, MLME_INVALID_PARAM
 *          pbssDescrSet points to the set of BSS descriptions found.
 *
 */
void wlanMlmeScanConfirm
    (
    WLAN_DEV_INFO   *pdevInfo,
    BSSDESCR_SET    *pbssDescrSet,      /* pointer to descriptor set */
    MLME_RESULT      result             /* mlme result code */
    );

/*
 * wlanMlmeJoinRequest - Join a BSS
 *
 * JOIN primitive logically establishes a binding to the AP.
 * It doesn't need to send any packet to the AP. It will turn on
 * the beacon delivery to driver and latch on to the next beacon.
 * MAC time should be synchronized with that beacon.
 * It is not clear if it needs to send a probe request or not.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM
 */

MLME_RESULT wlanMlmeJoinRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    BSSDESCR      *pbssDescr,           /* pointer to BSS description */
    A_UINT32       joinFailureTimeout,  /* timeout in TU */
    A_UINT32       probeDelay,          /* probe delay in usec */
    RATE_SET      *operationalRateSet   /* Set of data rates */
    );

/*
 * wlanMlmeJoinConfirm - Confirm the join operation
 *
 * This routine is called on successful JOIN operation. i.e. after receiving
 * a valid beacon from the AP and verifying that the BSS capabilities
 * match. It calls joinCallback() routine.
 *
 * RETURNS: MLME_SUCCESS, MLME_ALREADY_IN_BSS, MLME_INVALID_PARAM
 *          MLME_TIMEOUT
 */
void wlanMlmeJoinConfirm
    (
    WLAN_DEV_INFO   *pdevInfo,
    MLME_RESULT      result             /* Result good or timeout */
    );

/* wlanMlmeProbeRequest
 *
 * This routine sends out a probe packet. Not required
 * by MLME, but used internally by algorithms.
 */
MLME_RESULT wlanMlmeProbeRequest
    (
    WLAN_DEV_INFO   *pdevInfo,
    WLAN_MACADDR    *bssid,
    SSID            *ssid,
    A_UINT32         probeDelay
    );

/*
 * MLME Authentication primitives perform authentication with the
 * AP. There is no authentication in Ad-hoc.
 *
 * PARAMETERS: Parameters for all of the authentication routines are described
 *             here together.
 * pPeerSTAAddress - MAC address of the other STA to be auth'ed with. It is
 *                   the AP's STA address.
 * authType - OPEN_SYSTEM or SHARED_KEY.
 * authFailureTimeout - TU. Abort the operation if not completed within this
 *            time. Observed for the overall duration of frame exchange.
 *
 */

/* wlanMlmeAuthRequest - start authentication process
 *
 * Starts the authentication sequence with the peer STA. Returns
 * after initiating the operation.
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_NOT_SUPPORETD
 */
MLME_RESULT wlanMlmeAuthRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* Address of other STA */
    A_UINT16       authType,            /* open systems or shared key */
    A_UINT32       authFailureTimeout   /* timeout in TU */
    );

/* wlanMlmeAuthConfirm - confirm the result of AuthRequest operation.
 *
 * Calls the authCallback routine to deliver the result of authentication
 * operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_NOT_SUPPORTED,
 *          MLME_TOOMANY_REQ, MLME_TIMEOUT, MLME_REFUSED
 */
MLME_RESULT wlanMlmeAuthConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other MAC */
    A_UINT16       authType,            /* open systems or shared key */
    MLME_RESULT    resultCode           /* Result of operation */
    );

/*
 * wlanAuthIndication - Inform about a successful authentication.
 * it will be used to inform higher level entity of new authentication.
 * It is called by SME on the access point.
 */
MLME_RESULT wlanMlmeAuthIndication
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
    A_UINT16       authType
    );

/*
 * wlanMlmeDeauthRequest - terminates current authentication.
 *
 * It is a noticification. There is no notion of success/failure.
 * It is called by SME on either STA or AP.
 *
 * RETURNS: MLME_SUCCESS or MLME_INVALID_PARAM
 */
MLME_RESULT wlanMlmeDeauthRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other MAC */
    A_UINT16       reasonCode           /* one of the valid REASON_CODEs */
    );

/*
 * wlanMlmeDeauthForced - sends deauth frame whether authed or not
 *
 * It is a noticification. There is no notion of success/failure.
 * It is called by SME on either STA or AP.
 * RETURNS: MLME_SUCCESS or MLME_INVALID_PARAM
 */
MLME_RESULT wlanMlmeDeauthForced
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other MAC */
    A_UINT16       reasonCode           /* one of the valid REASON_CODEs */
    );

/* wlanMlmeDeauthConfirm - confirm a deauthentication operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_TOOMANY_REQ
 *
 */
MLME_RESULT wlanMlmeDeauthConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
    MLME_RESULT    resultCode           /* reason for disassoc */
    );

/* wlanMlmeDeauthIndication - Indicate deauth to higher layer
 *
 * Currently implemented as an empty routine
 */
MLME_RESULT wlanMlmeDeauthIndication
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
    MLME_RESULT    resultCode
    );

/*
 * MLME Association primitives perform association with the AP
 *
 * PARAMETERS: Parameters for all Association routines are listed here.
 *  pPeerSTAAddress - MAC address of AP's STA.
 *  assocFailureTimeout - TU. Timeout observed for overall duration
 *                        of frame exchange.
 *  capabilityInfo - Describes the capabilities that it wants for its MAC.
 *  listenInterval -    number of beacons that may pass before STA wakes up
 *                  listens for the next beacon.
 */

/* wlanMlmeAssocRequest - Request to associate with a AP
 *
 * This routine initiates the association operation with a AP.
 * It will fail if not authenticated with the requested AP or already
 * associated with some AP.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *          MLME_NOT_SUPPORTED
 */
MLME_RESULT wlanMlmeAssocRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
    A_UINT32       assocFailureTimeout, /* timeout in TU */
    CAP_INFO      *capabilityInfo,      /* describe capabilities */
    A_UINT16       listenInterval       /* sleep interval */
    );

/* wlanMlmeAssocConfirm - Confirm the result of assoc. request
 *
 * this routine calls back assocCallback() routine to deliver the result
 * of assoc request.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *          MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */
MLME_RESULT wlanMlmeAssocConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    MLME_RESULT    resultCode           /* result of operation */
    );

/* wlanMlmeAssocIndication - Inform upper layer of associtaion
 *
 * Currently unimplemented
 */
MLME_RESULT wlanMlmeAssocIndication
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress      /* address of other STA */
    );

MLME_RESULT wlanMlmeReassocRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of new AP's STA */
    A_UINT32       reassocFailureTimeout,/* timeout in TU */
    CAP_INFO      *capabilityInfo,      /* describe capabilities */
    A_UINT16       listenInterval       /* STA may sleep for this time */
    );

/* wlanMlmeReassocConfirm - Confirm the result of reassoc. request
 *
 * this routine calls back reassocCallback() routine to deliver the result
 * of reassoc request.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *          MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */
MLME_RESULT wlanMlmeReassocConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    MLME_RESULT    resultCode
    );

/* wlanMlmeReassocIndication - Inform upper layer of reassocitaion
 *
 * Currently unimplemented
 */
MLME_RESULT wlanMlmeReassocIndication
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress      /* address of other STA */
    );

/* wlanMlmeDisassocRequest - Disassociate (forcibly if necessary) a STA
 *
 * This routine is called by STA or AP SME to terminate a current association.
 * It is called when a STA wants to shut down its WLAN access, or AP wants
 * to cleanup or shutdown.
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM,
 *          MLME_NOT_SUPPORTED
 */
MLME_RESULT wlanMlmeDisassocRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other STA */
    A_UINT16       reasonCode           /* reason for disassoc */
    );

/* wlanMlmeDisassocConfirm - Confirm the disassoc operation.
 *
 * This routine calls back disassocCallback() to confirm the operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_REFUSED,
 *          MLME_TIMEOUT, MLME_NOT_SUPPORTED
 */
MLME_RESULT wlanMlmeDisassocConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pmacAddr,
    MLME_RESULT    resultCode           /* result of operation */
    );

/* wlanMlmeDisassocIndication - Inform upper layer of disassocitaion
 */
MLME_RESULT wlanMlmeDisassocIndication
    (
    WLAN_DEV_INFO * pdevInfo,
    WLAN_MACADDR  *     pPeerSTAAddress,
    MLME_RESULT         reasonCode
    );

MLME_RESULT wlanMlmeDisAssocForced
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pPeerSTAAddress,     /* address of other MAC */
    A_UINT16       reasonCode           /* one of the valid REASON_CODEs */
    );

/* mlmeAssocIdRelease - Make assocId available again
 *
 * Called by disassociate primitive. It will clear the corresponding
 * bit and make the assocId available for others.
 * Need to use locking mechanism to protect assocArray.
 */

A_UINT16 mlmeAssocIdNew(void);
void mlmeAssocIdRelease
    (
    A_UINT16 assocId            /* Currently assigned ID */
    );

/* wlanResetRequest - Reset the MAC entity
 *
 * This call resets the MAC hardware and software structures that
 * are used to handle that specific MAC. There will be other routines
 * and commands to reset the hardware, driver, etc. but not done as
 * part of MLME operation.
 *
 * PARAMETERS:
 *  pLocalSTAAddress - MAC address to be set after the reset.
 *  setDefaultMIB    - If TRUE, set the default values into the MIB.
 *
 * RETURNS: MLME_SUCCESS, WLAN_UNKNOWN_FAILURE
 */

MLME_RESULT wlanResetRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    WLAN_MACADDR  *pLocalSTAAddress,
    A_BOOL         setDefaultMIB
    );

/* wlanResetConfirm - Confirm the reset of MAC
 * This routine calls reseCallback() to deliver the Success/fail
 * result of reset operation.
 */
MLME_RESULT wlanResetConfirm
    (
    WLAN_DEV_INFO *pdevInfo
    );

/* wlanMlmeStartRequest - Start the operation of an access point
 *
 * Tells the MAC entity to operate as a BSS (Access point)
 * It sets various BSS related parameters into the MAC, enables beacon
 * generation, DTIM generation, enable power save buffering, etc.
 *
 * This will be implemented as a synchronous operation.
 *
 * RETURNS: MLME_SUCCESS, MLME_INVALID_PARAM, MLME_ALREADY_IN_BSS
 */
MLME_RESULT wlanMlmeStartRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    BSSDESCR      *pbssDescr
    );

/* wlanMlmeStartConfirm - Confirm the start of BSS
 * Since the request is implemented as a synchronous operation,
 * this is an empty routine.
 */
MLME_RESULT wlanMlmeStartConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    MLME_RESULT    resultCode
    );

MLME_RESULT wlanPlmeCharRequest
    (
    WLAN_DEV_INFO *pdevInfo,
    PH_PARAM_SET  *plmeCharSet
    );

MLME_RESULT wlanPlmeCharConfirm
    (
    WLAN_DEV_INFO *pdevInfo,
    PH_PARAM_SET  *plmeCharSet
    );

MLME_RESULT wlanMlmeInit
    (
    SME_INFO_STRUCT *psmeInfo   /* Structure with service type */
    );

MLME_RESULT wlanMlmeDeviceBind
    (
    OS_DEVICE_ID   deviceID,    /* OS dependent device type */
    WLAN_DEV_INFO *pdevInfo
    );

void mgtFrameHdrFill
    (
    WLAN_DEV_INFO       *pdevInfo,
    WLAN_MGT_MAC_HEADER *pFrame,
    WLAN_MACADDR        *destAddr,
    int                  subtype
    );

/* mlmeMgtTxCompletion - Process completion of mgt transmit descriptor
 *
 * DESCRIPTION:
 * This routine updates the station state in SIB based on success or
 * failure of mgt frame that was scheduled earlier. Certain MLME primitives
 * are not called completed until that frame is transmitted and ack received.
 * INPUT PARAMETERS
 * pDesc - Pointer to descriptor with valid tx status field.
 */
void mlmeMgtTxCompletion
    (
    WLAN_DEV_INFO *pdevInfo,
    ATHEROS_DESC  *pDesc        /* pointer to atheros descriptor */
    );

void smePendDescProcess 
    (
    WLAN_DEV_INFO *pdevInfo, 
    SIB_ENTRY *pSib
    );

/* mgtFrameDispatch - Routine to process incoming mgt frame
 *
 * DESCRIPTION
 * It calls proper mlme routine based on incoming frame type and state of the station
 */
void mlmeMgtRxCompletion
    (
    WLAN_DEV_INFO *pdevInfo,
    ATHEROS_DESC  *pDesc
    );

/*
 * wlanBSSDescriptionGet returns the BSS description of the specified
 * BSS from the list of beacons that STA has received.
 */

MLME_RESULT wlanBSSDescriptionGet
    (
    WLAN_DEV_INFO *pdevInfo,
    BSSID          bssid,
    BSSDESCR      *pbssDescr
    );

/* Non exported routines used by other pieces of our driver */

A_UINT16 mlmeAuthAp1Receive
    (
    WLAN_DEV_INFO   *pdevInfo,                  /* device context */
    WLAN_FRAME_AUTH *authFrame,                 /* Incoming frame pointer */
    A_UINT16         len                        /* Total frame length */
    );

A_UINT16 mlmeAuthAp3Receive
    (
    WLAN_DEV_INFO           *pdevInfo,
    WLAN_FRAME_AUTH_ENCRYPT *authFrame,         /* Incoming frame */
    A_UINT16                 len,               /* Total frame length */
    A_BOOL                   decryptError       /* decryption status */
    );

A_UINT16 mlmeAuthSta2Receive
    (
    WLAN_DEV_INFO        *pdevInfo,
    WLAN_FRAME_AUTH_CHLG *authFrame,            /* Incoming frame */
    A_UINT16              len                   /* Total frame length */
    );

A_UINT16 mlmeAuthSta4Receive
    (
    WLAN_DEV_INFO   *pdevInfo,
    WLAN_FRAME_AUTH *authFrame,                 /* Incoming frame pointer */
    A_UINT16         len                        /* Total frame length */
    );

MLME_RESULT mlmeDeauthApReceive
    (
    WLAN_DEV_INFO     *pdevInfo,
    WLAN_FRAME_DEAUTH *deauthFrame,             /* Incoming frame pointer */
    A_UINT16           len                      /* Total frame length */
    );

A_UINT16 mlmeAssocApReceive
    (
    WLAN_DEV_INFO         *pdevInfo,
    WLAN_FRAME_ASSOC_REQ  *assocFrame,          /* Incoming frame pointer */
    A_UINT16               len                  /* Total frame length */
    );

A_UINT16 mlmeAssocStaReceive
    (
    WLAN_DEV_INFO         *pdevInfo,
    WLAN_FRAME_ASSOC_RESP *assocFrame,          /* Incoming frame pointer */
    A_UINT16               len                  /* Total frame length */
    );

A_UINT16 mlmeReassocApReceive
    (
    WLAN_DEV_INFO          *pdevInfo,
    WLAN_FRAME_REASSOC_REQ *reassocFrame,       /* Incoming frame pointer */
    A_UINT16                len                 /* Total frame length */
    );

A_UINT16 mlmeReassocStaReceive
    (
    WLAN_DEV_INFO           *pdevInfo,
    WLAN_FRAME_REASSOC_RESP *reassocFrame,      /* Incoming frame pointer */
    A_UINT16                 len                /* Total frame length */
    );

MLME_RESULT mlmeDisassocApReceive
    (
    WLAN_DEV_INFO       *pdevInfo,
    WLAN_FRAME_DISASSOC *disassocFrame, /* Incoming frame pointer */
    A_UINT16             len            /* Total frame length */
    );

/* wlanMlmeProbeRequestReceive
 *
 * This routine receives a probe request on AP. Not required
 * by MLME, but used internally by algorithms.
 */

void mlmeProbeRequestReceive
    (
    WLAN_DEV_INFO        *pdevInfo,
    WLAN_FRAME_PROBE_REQ *probReqFrame,
    A_UINT16              len           /* Total frame length */
    );

void mlmeProbeRespReceive
    (
    WLAN_DEV_INFO *pdevInfo,
    ATHEROS_DESC  *pDesc
    );

void chlgTextFill
    (
    CHLG_TEXT *pChlg,           /* Pointer to challenge text element */
    A_UINT8    len              /* desired length of challenge text */
    );

A_BOOL   basicRateSubsetCheck
    (
    RATE_SET *psub,             /* expected subset */
    RATE_SET *psuper            /* set expected to be superset */
    );

/* xSectionRateSet - Builds an intersection of two rate sets
 * into the third one. Used by AP and STA for building a working
 * rate set.
 */
void xSectionRateSet
    (
    RATE_SET *pSrc1,            /* set 1 */
    RATE_SET *pSrc2,            /* set 2 */
    RATE_SET *pDest             /* cross section set */
    );

A_BOOL ucseSubsetCheck
    (
    UCSE_SET *psub,             /* expected subset */
    UCSE_SET *psuper            /* set expected to be superset */
    );

/* xSectionUcseSet - Builds an intersection of two ucse sets
 * into the third one. Used by AP and STA for building a working
 * ucse set.
 */
void xSectionUcseSet
    (
    UCSE_SET *pSrc1,            /* set 1 */
    UCSE_SET *pSrc2,            /* set 2 */
    UCSE_SET *pDest             /* cross section set */
    );

A_INT8  mlmeGetControlRateIndex(RATE_SET *pRates, A_INT8 rateIndex);

void mlmeBsssetCopy
    (
    BSSDESCR_SET *src,
    BSSDESCR_SET *dest
    );

void mlmeBsssetMerge
    (
    BSSDESCR_SET *src,
    BSSDESCR_SET *dest
    );

void mlmeBsssetAdd
    (
    BSSDESCR     *src,
    BSSDESCR_SET *dest
    );

void mlmeBsssetAge
    (
    BSSDESCR_SET *src,
    A_UINT32      msAge
    );

BSSDESCR *mlmeBsssetFind
    (
    BSSDESCR_SET *pSet,
    WLAN_MACADDR *pAddr,
    A_UINT16      channelFlags
    );

void mlmeBsssetDel
    (
    BSSDESCR_SET *pSet,
    WLAN_MACADDR *pAddr
    );

A_BOOL  isNullSsid
    (
    SSID *pSsid
    );

void mlmeBeaconReceive
    (
    WLAN_DEV_INFO *pdevInfo,
    ATHEROS_DESC  *pDesc        /* Incoming frame descr */
    );

INFO_ELEMENT * mlmeElementCopy
    (
    INFO_ELEMENT *pSrc,         /* pointer to element */
    INFO_ELEMENT *pDest
    );

void mlmeOperationCancel
    (
    WLAN_DEV_INFO *pdevInfo
    );

A_STATUS drvChangeChannel
    (
    WLAN_DEV_INFO *pdevInfo,
    CHAN_VALUES   *channel,
    A_UINT32       rxFilter
    );

void drvBBChange
    (
    WLAN_DEV_INFO *pdevInfo,
    CHAN_VALUES   *pChval
    );

A_UINT32 maxTxRateCompute
    (
    RATE_SET *pRates,
    A_UINT16 channelFlags
    );

A_UINT16 mlmeCountryCodeElementLength
    (
    WLAN_DEV_INFO *pDevInfo
    );

A_CHAR * mlmeCountryCodeElementFill
    (
    WLAN_DEV_INFO *pDevInfo,
    A_CHAR        *pField
    );
    
#ifdef DEBUG
extern int mlmeDebug;
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __INCwlansmeexth */

