#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/if_arp.h>
#include <linux/if_packet.h>
#include <net/ip.h>
#include <net/protocol.h>
#include <linux/interrupt.h>

#include <linux/delay.h>
#include <asm/io.h>
#include <linux/sched.h>

#include "prism2_wlan.h"
#include "wifi.h"
#include "sta_db.h"

static __u8 bcast_addr[6] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

static struct net_device *gre_dev = NULL;
static struct net_device *wlan_dev = NULL;
extern __u16 hfa384x_tx_ctrl_flags;

__u8 my_bssid[6];

static int sap_switch_tx_data_frame(struct sk_buff *, struct net_device *);
static int sap_switch_tx_mgmt_frame(struct sk_buff *, struct net_device *);
int aruba_rcv(struct sk_buff *, struct net_device *,
		     struct packet_type *pt);
extern int prism2_tx_kernel(struct sk_buff *, struct net_device *,
			    struct hfa384x_tx_frame *);

//#define DIRECT_CONNECT
#ifdef DIRECT_CONNECT
int direct_connect_tx(struct sk_buff *);
static struct net_device *eth_dev = NULL;
static __u8 mx_mac[6] = {0x00, 0x60, 0xcf, 0x12, 0x34, 0x56};
static int mx_mac_flag = 1;
#endif

struct sk_buff_head sap_readq; 


char *mac2str(__u8 *mac){
  static char s[20];
  sprintf(s, "%02x:%02x:%02x:%02x:%02x:%02x", mac[0], mac[1], mac[2],
	  mac[3], mac[4], mac[5]);
  return s;
}

static int sap_switch_tx_data_frame(struct sk_buff *skb,
				    struct net_device *dev){

  // dev is the wlan dev

  // assume laser beam packets will never come over the GRE tunnel

  struct sta_info *sta = NULL;
  struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
  struct hfa384x_tx_frame txdesc;
  int multicast_frame = 0;

  memset(&txdesc, 0, sizeof(txdesc));
  memcpy(&txdesc.frame_control, hdr, sizeof(*hdr));
  memcpy(txdesc.addr2, dev->dev_addr, 6);
  txdesc.frame_control |= cpu_to_le16(WLAN_FC_FROMDS);
  
  skb->dev = dev;
  
  if (!memcmp(hdr->addr1, bcast_addr, 6) ||
      (hdr->addr1[0] & 1)){
    // multicast frame
    multicast_frame = 1;
  } else {
    sta = sap_get_sta(hdr->addr1);
    if (sta == NULL){
      printk("sap_tx_frame: STA %s is unknown\n", mac2str(hdr->addr1));
      goto no_good;
    }

#if 0
    // have a good STA, check for power save, etc.
    if (sta->flags & WLAN_STA_PS)
      return handle_sta_ps(sta, skb);
#endif

  }

  txdesc.tx_rate = (multicast_frame) ? 10 : get_tx_rate(sta);
  txdesc.tx_control = cpu_to_le16(hfa384x_tx_ctrl_flags);
  txdesc.data_len = cpu_to_le16(skb->len - sizeof(struct ieee80211_hdr));

  skb_pull(skb, sizeof(struct ieee80211_hdr));

  prism2_tx_kernel(skb, dev, &txdesc);
  
  return 0;

 no_good:
  dev_kfree_skb(skb);
  return 0;
}

static int sap_switch_tx_mgmt_frame(struct sk_buff *skb,
				    struct net_device *dev){

  struct sta_info *sta;
  struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
  struct ieee80211_mgmt *m = (struct ieee80211_mgmt *)skb->data;
  struct hfa384x_tx_frame txdesc;
  u16 fc, type, stype;

  fc = cpu_to_le16(hdr->frame_control);
  type = WLAN_FC_GET_TYPE(fc);
  stype = WLAN_FC_GET_STYPE(fc);

  if (type != WLAN_FC_TYPE_MGMT)
    return 0;

  switch (stype){
  case WLAN_FC_STYPE_ASSOC_RESP:
  case WLAN_FC_STYPE_REASSOC_RESP:
    if (m->u.assoc_resp.status_code != WLAN_STATUS_SUCCESS)
      break;

    sta = sap_get_sta(m->da);
    if (sta == NULL){
      // create an entry for this STA
      sta = (struct sta_info *)kmalloc(sizeof(struct sta_info), GFP_ATOMIC);
      if (sta == NULL){
	printk("Unable to allocate mem for new STA\n");
	goto mgmt_failed;
      }
      memset(sta, 0, sizeof(struct sta_info));
      memcpy(sta->mac, m->da, 6);
      sap_add_sta(sta);
    }
    sta->aid = cpu_to_le16(m->u.assoc_resp.aid);
    sta->flags |= (WLAN_STA_ASSOC | WLAN_STA_AUTH);
    //    parse_sta_capabilities(sta, m);
    break;

  case WLAN_FC_STYPE_AUTH:
    // Auth response

    if (m->u.auth.status_code != WLAN_STATUS_SUCCESS)
      break;

    sta = sap_get_sta(m->da);
    if (sta == NULL){
      // create an entry for this STA
      sta = (struct sta_info *)kmalloc(sizeof(struct sta_info), GFP_ATOMIC);
      if (sta == NULL){
	printk("Unable to allocate mem for new STA\n");
	goto mgmt_failed;
      }
      memset(sta, 0, sizeof(struct sta_info));
      memcpy(sta->mac, m->da, 6);
      sap_add_sta(sta);
    }
     
    sta->flags |= WLAN_STA_AUTH;
    break;

  case WLAN_FC_STYPE_DEAUTH:
    sta = sap_get_sta(m->da);
    if (sta){
      sap_del_sta(sta);
    }
    break;
  default:
  }
    
  // now transmit the frame
  memset(&txdesc, 0, sizeof(txdesc));
  memcpy(&txdesc.frame_control, hdr, sizeof(*hdr));
  skb->dev = dev;
  
  txdesc.tx_rate = 10;
  txdesc.tx_control = cpu_to_le16(hfa384x_tx_ctrl_flags);
  txdesc.data_len = cpu_to_le16(skb->len - sizeof(struct ieee80211_hdr));
  skb_pull(skb, sizeof(struct ieee80211_hdr));
  
  prism2_tx_kernel(skb, dev, &txdesc);
  
  return 0;
  

 mgmt_failed:
  // dont need to transmit

  dev_kfree_skb(skb);
  return 0;
}



int aruba_rcv(struct sk_buff *skb, struct net_device *dev,
		     struct packet_type *pt){

  // assumes data will contain the wlan dev

#if 0
  struct net_device *wdev = pt->data;
#else
  struct net_device *wdev = wlan_dev;
#endif
  struct ieee80211_hdr *hdr = (struct ieee80211_hdr *)skb->data;
  u16 fc;

  if (!wdev){
    printk("aruba_rcv : WLAN device is not yet up\n");
    goto failed;
  }

  if (skb->len < sizeof(struct ieee80211_hdr))
    goto failed;

  // check if it is data, then transmit it immediately
  // if it is a mgmt frame, punt it to STA mgmt
  
  fc = cpu_to_le16(hdr->frame_control);

  switch (WLAN_FC_GET_TYPE(fc)){
  case WLAN_FC_TYPE_DATA:
    // transmit the frame

    sap_switch_tx_data_frame(skb, wdev);
    
    break;

  case WLAN_FC_TYPE_MGMT:
    // punt it to sta mgmt, which will then transmit

    sap_switch_tx_mgmt_frame(skb, wdev);

    break;

  }

  return 0;

 failed:

  dev_kfree_skb(skb);
  return 0; // for now

}

int sap_switch_setup_dev(int *ifindex){

  // ifindex contains the index number of the GRE device

  struct net_device *dev = dev_get_by_index(*ifindex);
  if (dev == NULL){
    printk("Cannot find device with index = %d\n", *ifindex);
    return 0;
  }

  gre_dev = dev;

  return 0;
}

int gre_tx(struct sk_buff *);

void sap_switch_uplink(unsigned long data){

  // called when a .11 frame is received from the radio interface

  struct sk_buff *skb;
  struct ieee80211_hdr *hdr;
  struct ieee80211_mgmt *m;
  __u16 fc, type;

  while ((skb = skb_dequeue(&sap_readq))){

#if 0
    if (!gre_dev){
      dev_kfree_skb(skb);
      return;
    }
#endif
    
    // filter unnecessary traffic
    
    hdr = (struct ieee80211_hdr *)skb->data;
    fc = cpu_to_le16(hdr->frame_control);
    type = WLAN_FC_GET_TYPE(fc);
    
    switch (type){
    case WLAN_FC_TYPE_MGMT:
      
      m = (struct ieee80211_mgmt *)skb->data;
      if (memcmp(m->bssid, skb->dev->dev_addr, 6))
	goto dont_transmit;
      break;
      
    case WLAN_FC_TYPE_DATA:
      if ((fc & WLAN_FC_TODS) && !(fc & WLAN_FC_FROMDS) && 
	  !memcmp(hdr->addr1, skb->dev->dev_addr, 6)){
	// transmit the frame
	break;
      } else {
	goto dont_transmit;
      }
      break;
      
    case WLAN_FC_TYPE_CTRL: // for now
    default:
      goto dont_transmit;
    }
    
    
    // assume the skb was created with enough headroom for GRE+IP headers
    
    skb->protocol = htons(ARUBA_80211_FRAME);
    

#ifdef DIRECT_CONNECT
    direct_connect_tx(skb);
#else
    gre_tx(skb);
#endif
    
    continue;
    
  dont_transmit:
    dev_kfree_skb(skb);
    continue;
  }

  return;
    
}

#if 0
static struct packet_type aruba_packet_type = {
  type:   __constant_htons(ARUBA_80211_FRAME),
  func:   aruba_rcv,
  data:   NULL // for now, should be initialized with the wlan dev later
};
#endif

#ifdef DIRECT_CONNECT
int register_wlan_pkt_handler(void (*handler)(struct sk_buff *));
void aruba_dc_rcv(struct sk_buff *skb){

  aruba_rcv(skb, NULL, NULL);

}

u32 eth_get_duplex(struct net_device *dev){

  // very specific to ne.c

  u16 ioaddr = dev->base_addr;
  u8 configreg, magicreg, omagicreg;
  u8 cmdreg = inb(ioaddr);
  

  outb(ioaddr, 0x22);
  configreg = inb(ioaddr+ 0xa);
  outb(ioaddr+0x0a, configreg | 0x80);
  inb(ioaddr+0x0a);
  omagicreg = magicreg = inb(ioaddr+0x19);
  
  printk("duplex_status = %02x\n", magicreg);

  magicreg |= 0x40;
  //  outb(ioaddr+0x19, magicreg);
  outb(ioaddr+ 0x0a,configreg);
  outb(ioaddr, cmdreg);

  return magicreg;


}

#endif // DIRECT_CONNECT


int sap_switch_init(struct net_device *dev){

#if 0
  aruba_packet_type.data = dev;
  dev_add_pack(&aruba_packet_type);
#else
  wlan_dev = dev;
#endif

#ifdef DIRECT_CONNECT
  eth_dev = dev_get_by_name("eth0");
  register_wlan_pkt_handler(&aruba_dc_rcv);
  //  eth_get_duplex(eth_dev);
#endif

  sap_hash_init();
  skb_queue_head_init(&sap_readq);

  return 0;
}

void sap_switch_cleanup(){
#if 0
  dev_remove_pack(&aruba_packet_type);
  aruba_packet_type.data = NULL;
#else
  wlan_dev = NULL;
#endif

#ifdef DIRECT_CONNECT
  if (eth_dev){
    dev_put(eth_dev);
    eth_dev = NULL;
  }
  mx_mac_flag = 0;
#endif

  sap_hash_cleanup();

  skb_queue_purge(&sap_readq);

  if (gre_dev){
    dev_put(gre_dev);
    gre_dev = NULL;
  }
}

#ifdef DIRECT_CONNECT

int direct_connect_setup(__u8 *mac){
  // sets up the M-switch MAC address

  memcpy(mx_mac, mac, 6);
  mx_mac_flag = 1;
  return 0;

}

int direct_connect_tx(struct sk_buff *skb){

  if (!eth_dev){
    dev_kfree_skb(skb);
    return 0;
  }

  
  skb_push(skb, 14);

  // set type to indicate .11 frame
  *(__u16 *)(skb->data+12) = htons(ARUBA_80211_FRAME);
  
  // copy the SA and DA
  memcpy(skb->data, mx_mac, 6);
  memcpy(skb->data+6, eth_dev->dev_addr, 6);

  skb->dev = eth_dev;

  // add 4 bytes for CRC
  *(__u32 *)(skb->data + skb->len) = 0;
  skb->len += 4;

  //  printk("eth_tx: %d\n", skb->len);

  if (eth_dev->hard_start_xmit){
    eth_dev->hard_start_xmit(skb, eth_dev);
  }

  return 0;
}


#endif
