#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/miscdevice.h>
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/wait.h>
#include <linux/ioctl.h>
#include <linux/mm.h>
#include <asm/io.h>
#include <linux/wrapper.h>

#include "prism2_wlan.h"

#include "chrdev.h"

#if 0
#ifdef PRISM2_PCI
#define HFA384X_INW(a) (u16) readw(dev->mem_start + (a))
#else
#define HFA384X_INW(a) inw(dev->base_addr + (a))
#endif /* PRISM2_PCI */
#endif // 0

struct softap_chr_struct {
  char *name;
  u32 flags;
  int attached;
  uid_t owner;
  
  wait_queue_head_t read_wait;
  int am_buf_ok;
  struct _am_buffer_struct am_buf;
  __u32 read_index;
  int to_read;

  spinlock_t sap_lock;
	
  struct net_device *dev;
  struct fasync_struct *fasync;
};

static struct net_device *wlan_dev = NULL;
static struct softap_chr_struct *wsap = NULL;
int prism2_ioctl(struct net_device *, struct ifreq *, int);


#define SAP_MINOR 201
#ifndef MIN
#define MIN(a, b) ((a) < (b)) ? (a) : (b)
#endif

static unsigned int sap_chr_poll(struct file *, poll_table *);
static int sap_chr_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
static ssize_t sap_chr_write(struct file *, const char *, size_t, loff_t *);
static ssize_t sap_chr_read(struct file *, char *, size_t, loff_t *);
static int sap_chr_fasync(int, struct file *, int);
static int sap_chr_open(struct inode *, struct file *);
static int sap_chr_close(struct inode *, struct file *);
static int sap_chr_mmap(struct file *, struct vm_area_struct *);

int prism2_tx_kernel(struct sk_buff *,
		     struct net_device *, 
		     struct hfa384x_tx_frame *);


#define AM_BUF_NUM 256
#define AM_BUF_SIZE 256


static unsigned int sap_chr_poll(struct file *file, poll_table *wait){
  
  struct softap_chr_struct *sap = 
    (struct softap_chr_struct *)file->private_data;
	
  unsigned int mask = POLLOUT | POLLWRNORM;
  
  if (!sap)
	return -EBADFD;
  
  poll_wait(file, &sap->read_wait, wait);

  //  if (skb_queue_len(&sap->readq))

  if (sap->to_read)
    mask |= POLLIN | POLLRDNORM;
  
  return mask;
}

static int sap_chr_ioctl(struct inode *inode, struct file *file, 
  unsigned int cmd, unsigned long arg){
	
  int ret = 0;
  struct softap_chr_struct *sap = 
    (struct softap_chr_struct *)file->private_data;
  if (!sap) return -EBADFD;

  switch (cmd){
  case AM_BUFFER_INIT:
    // no checks are done here, is this ok?
    copy_to_user((char *)arg, (char *)&sap->am_buf, sizeof(sap->am_buf));
    sap->am_buf_ok = 1;
    break;
	
  default:
    //    printk(KERN_INFO "sap_chr_ioctl: Unknown cmd %x\n", cmd);
    //    ret = -EINVAL;
    ret = prism2_ioctl(sap->dev, (struct ifreq *)arg, cmd);

    break;
  }
  return ret;
}

static __inline__ ssize_t sap_get_user(struct softap_chr_struct *sap, const char *buf, size_t count){
  
  static char src[3000];
  register const char *ptr = buf;
  register int len = count;
  char *s = src;

  copy_from_user(s, ptr, len);
  
  return count;
}

static ssize_t sap_chr_write(struct file *file, const char *buf,
  size_t count, loff_t *pos){
	
  struct softap_chr_struct *sap;
  struct sk_buff *skb;
  struct hfa384x_tx_frame *txdesc;

  sap = (struct softap_chr_struct *)file->private_data;
	
  if (!sap || !sap->dev)
    return -EBADFD;

  if (count < sizeof(*txdesc))
    return -EINVAL;

  //  printk(KERN_DEBUG "sap_chr_write buf %p, len %d\n", buf, count);
  
  if (verify_area(VERIFY_READ, buf, count))
    return -EFAULT;
  
  skb = dev_alloc_skb(count);
  if (skb != NULL)
    return -ENOMEM;

  skb_put(skb, count);
  copy_from_user(skb->data, buf, sizeof(txdesc));
  txdesc = (struct hfa384x_tx_frame *)skb->data;
  skb_pull(skb, sizeof(*txdesc));
  skb->dev = sap->dev;
  
  return prism2_tx_kernel(skb, sap->dev, txdesc);
}

static ssize_t sap_chr_read(struct file *file, char *buf, 
			    size_t count, loff_t *pos){
  
  struct softap_chr_struct *sap = 
    (struct softap_chr_struct *)file->private_data;

  DECLARE_WAITQUEUE(wait, current);
  ssize_t ret = 0;
  
  if (!sap) return -EBADFD;

  if (count < sizeof(__u32))
    return -EINVAL;

  add_wait_queue(&sap->read_wait, &wait);
  while (count){
    current->state = TASK_INTERRUPTIBLE;

    if (!sap->to_read){
#if 0
      if (file->f_flags & O_NONBLOCK){
	ret = -EAGAIN;
	break;
      }
      if (signal_pending(current)){
	ret = -ERESTARTSYS;
	break;
      }
#endif
      
      schedule();
      continue;
    }
    
    if (!verify_area(VERIFY_WRITE, buf, count)){
      int len = sizeof(__u32);
      copy_to_user(buf, (char *)&sap->read_index, len);
      ret = len;
      sap->to_read = 0;
      count -= len;
    } else
      ret = -EFAULT;
    
    break;
  }
  
  current->state = TASK_RUNNING;
  remove_wait_queue(&sap->read_wait, &wait);
  
  return ret;
}


static int sap_chr_fasync(int fd, struct file *file, int on){
  printk(KERN_INFO "sap_chr_fasync, not supported yet\n");
  return 0;
}

static int sap_chr_open(struct inode *inode, struct file *file){

  __u8 *am_buf_head = NULL;
  struct page *page;

  if (wsap != NULL)
    return -EFAULT;

  am_buf_head = kmalloc(AM_BUF_NUM*AM_BUF_SIZE, GFP_ATOMIC);
  if (am_buf_head == NULL)
    return -ENOMEM;
  
  wsap = kmalloc(sizeof(struct softap_chr_struct), GFP_ATOMIC);
  if (!wsap)
	return -ENOMEM;
  
  memset(wsap, 0, sizeof(struct softap_chr_struct));

  spin_lock_init(&wsap->sap_lock);
  spin_lock(&wsap->sap_lock);
  
  init_waitqueue_head(&wsap->read_wait);
  wsap->owner = -1;
  
  file->private_data = wsap;
  wsap->attached = 1;
  wsap->dev = wlan_dev;

  for(page = virt_to_page(am_buf_head); 
      page < virt_to_page(am_buf_head+AM_BUF_SIZE*AM_BUF_NUM);
      page++){
    mem_map_reserve(page);
  }


  wsap->am_buf.buf_head = am_buf_head;
  wsap->am_buf.buf_size = AM_BUF_SIZE;
  wsap->am_buf.num_bufs = AM_BUF_SIZE;

  spin_unlock(&wsap->sap_lock);

  //printk("sap_chr_open: buf_head %p, wsap %p\n", wsap->am_buf.buf_head, wsap);

  return 0;
}

static int sap_chr_close(struct inode *inode, struct file *file){
  
  struct softap_chr_struct *sap 
    = (struct softap_chr_struct *)file->private_data;

  if (!sap || (sap != wsap)) return 0;
	
  
  file->private_data = NULL;
  sap->attached = 0;
  

  kfree(sap->am_buf.buf_head);
  kfree(sap);

  wsap = NULL;

  
  return 0;
}


static int sap_chr_mmap(struct file *file, struct vm_area_struct *vma){

  unsigned long page,pos;
  struct softap_chr_struct *sap = 
    (struct softap_chr_struct *)file->private_data;
  unsigned long start = (unsigned long)vma->vm_start; 
  unsigned long size = (unsigned long)(vma->vm_end-vma->vm_start); 
  
  
  /* if userspace tries to mmap beyond end of our buffer, fail */ 
  if (size>AM_BUF_NUM*AM_BUF_SIZE)
    return -EINVAL;
  
  /* start off at the start of the buffer */ 
  pos=(unsigned long)sap->am_buf.buf_head;

  //printk(KERN_INFO "sap_chr_mmap called, pos %lx, start %lx, size %lx \n",
  //pos, start, size);
  
  /* loop through all the physical pages in the buffer */ 
	/* Remember this won't work for vmalloc()d memory ! */
  while (size > 0) {
    /* remap a single physical page to the process's vma */ 
    page = virt_to_phys((void *)pos);
    /* fourth argument is the protection of the map. you might
     * want to use vma->vm_page_prot instead.
     */
    if (remap_page_range(start, page, PAGE_SIZE, vma->vm_page_prot))
      return -EAGAIN;
    start+=PAGE_SIZE;
    pos+=PAGE_SIZE;
    size-=PAGE_SIZE;
  }
  return 0; 
  
}

static struct file_operations sap_fops = {
  owner:	THIS_MODULE,
  llseek:	no_llseek,
  read:		sap_chr_read,
  write:	sap_chr_write,
  poll:		sap_chr_poll,
  ioctl:	sap_chr_ioctl,
  open:		sap_chr_open,
  release:	sap_chr_close,
  fasync:	sap_chr_fasync,
  mmap:         sap_chr_mmap
};

static struct miscdevice sap_miscdevice = {
  SAP_MINOR,
  "net/sap",
  &sap_fops
};
  

int sap_chr_init(struct net_device *dev){
  printk(KERN_INFO "Registering SAP char device driver\n");
  if (misc_register(&sap_miscdevice)){
	printk(KERN_ERR "Error registering misc device %d\n", SAP_MINOR);
	return -EIO;
  }

  wlan_dev = dev;
  
  return 0;
}

void sap_chr_cleanup(void){

  if (wsap){
    printk("softap char dev is being closed, but someone has it open\n");
  }

  misc_deregister(&sap_miscdevice);
}

int process_am_frame(struct hfa384x_rx_frame *rxdesc, char *buf){

  struct softap_chr_struct *sap = wsap;
  int len;
  __u8 *pos;

  if ((sap == NULL) || (sap->am_buf_ok == 0))
    return 0;
  
  len = sap->am_buf.buf_size;
  pos = sap->am_buf.buf_head+sap->am_buf.buf_size*sap->read_index;

  memcpy(pos, rxdesc, sizeof(*rxdesc));
  if (rxdesc->data_len > 0){
    pos += sizeof(*rxdesc);
    len -= sizeof(*rxdesc);
    memcpy(pos, buf, MIN(len, rxdesc->data_len));
  }


  if (++sap->read_index >= sap->am_buf.num_bufs)
    sap->read_index = 0;

  sap->to_read = 1;
  wake_up_interruptible(&sap->read_wait);

  return 0;

}


int prism2_rx_monitor_mode(struct net_device *dev){

  intersil_local_info_t *local = (intersil_local_info_t *) dev->priv;
  struct softap_chr_struct *sap = wsap;
  int len;
  int res;
  u16 rxfid, status;
  int data_len;
  struct hfa384x_rx_frame *rxdesc;
  int monitor_flag = 0;
  u16 fc;
  __u8 *pos;
  static __u8 devnull[2500];


  if (!sap){
    pos = devnull;
    len = 2500;
  } else {
    pos = sap->am_buf.buf_head+sap->am_buf.buf_size*sap->read_index;
    len = sap->am_buf.buf_size;
  }
  rxdesc = (struct hfa384x_rx_frame *)pos;

  rxfid = HFA384X_INW(HFA384X_RXFID_OFF);
  
  /* Prism2 should support concurrent BAP0 (TX) and BAP1 (RX) access,
   * but it seems to bring some errors when TX path is interrupted by RX
   * event, so require baplock to be free before starting RX handling. */
  spin_lock(&local->baplock);
  res = hfa384x_setup_bap(dev, BAP1, rxfid, 0);
  if (!res)
    res = hfa384x_from_bap(dev, BAP1, pos, sizeof(*rxdesc));
  spin_unlock(&local->baplock);

  //  printk("rx_monitor: status %x, fc %x, data_len %x\n",
  //	 rxdesc->status, rxdesc->frame_control, rxdesc->data_len);


  if (res) {
    printk(KERN_DEBUG "rx_monitor: copy from BAP1 failed %d\n", res);
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    goto rx_dropped;
  }
  
  
  fc = le16_to_cpu(rxdesc->frame_control);
  if (WLAN_FC_GET_TYPE(fc) == WLAN_FC_TYPE_DATA)
    monitor_flag |= 1;
  if (WLAN_FC_GET_TYPE(fc) == WLAN_FC_TYPE_CTRL)
    rxdesc->data_len = 0;

  data_len = le16_to_cpu(rxdesc->data_len);
    
  if (data_len & 0x8000) {
    //		printk(KERN_DEBUG "%s: Received frame with invalid length "
    //		       "0x%04x\n", dev->name, le16_to_cpu(rxdesc->data_len));
    goto rx_dropped;
    
  }

  status = le16_to_cpu(rxdesc->status);

  
  /* FCS errors should not come this far, but let's make sure that frames
   * with errors will be dropped even in Host AP mode */
  if (status & HFA384X_RX_STATUS_FCSERR) {
    //    printk(KERN_DEBUG "%s: prism2_rx: dropped FCSErr frame "
    //	   "(status=%02X)\n", dev->name, status);
    goto rx_dropped;
  }

  pos += sizeof(*rxdesc);
  len -= sizeof(*rxdesc);

  if (monitor_flag & 1){
    if (data_len > 0){
      res = hfa384x_from_bap(dev, BAP1, devnull, data_len);
    }
  } else {
    // need to read some of the payload

    if (data_len > 0){
      if (len < data_len){
	res = hfa384x_from_bap(dev, BAP1, pos, len);
	res = hfa384x_from_bap(dev, BAP1, devnull, data_len-len);
      } else {
	res = hfa384x_from_bap(dev, BAP1, pos, data_len);
      }
    }
  }
    
  if (res) {
    printk(KERN_DEBUG "rx_monitor: copy from BAP1 failed %d\n", res);
    if (res == -ETIMEDOUT)
      prism2_hw_reset(dev);
    goto rx_dropped;
  }

  if (sap){

    if (++sap->read_index >= sap->am_buf.num_bufs)
      sap->read_index = 0;
    
    sap->to_read = 1;
    wake_up_interruptible(&sap->read_wait);
  }
  
  
 rx_dropped:
  return 0;

}




