#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/wireless.h>
#include <linux/wait.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>

#include <asm/types.h>
#include "wifi.h"
#include "sta_db.h"

#define HASH_SIZE 256
#define MAX_AID 256
#define MAX_STA_COUNT 256

extern char *mac2str(__u8 *);

#define STA_HASH(mac) (mac[5])

static struct sta_info * sta_hash[HASH_SIZE];
static struct sta_info * sta_aid[MAX_AID];
static int num_sta;


void sap_add_sta (struct sta_info *sta)
{
  sta->next = sta_hash[STA_HASH (sta->mac)];
  sta_hash[STA_HASH (sta->mac)] = sta;
}

void sap_del_sta (struct sta_info *sta)
{
  struct sta_info *s;

  s = sta_hash[STA_HASH (sta->mac)];
  if (s == NULL)
    return;
  if (memcmp (s->mac, sta->mac, 6) == 0) {
    sta_hash[STA_HASH (sta->mac)] = s->next;
    return;
  }

  while (s->next != NULL && memcmp (s->next->mac, sta->mac, 6) != 0)
    s = s->next;
  if (s->next != NULL)
    s->next = s->next->next;
  else
    printk(KERN_INFO "error removing STA %s from hash table\n",
	   mac2str (sta->mac));
}

struct sta_info * sap_create_sta(u8 *mac){
  
  struct sta_info *sta;
	
  if (num_sta >= MAX_STA_COUNT){
	printk(KERN_INFO "No room for any new STAs\n");
	return NULL;
  }
  
  sta = sap_get_sta(mac);
	
  if (sta != NULL){
	printk(KERN_INFO "STA %s exists in the table\n", mac2str(mac));
	return sta;
  }
  
  sta = (struct sta_info *)kmalloc(sizeof(struct sta_info), GFP_ATOMIC);
	
  if (sta == NULL) return sta;
	
  memset(sta, 0, sizeof(*sta));
  memcpy(sta->mac, mac, 6);
  sta->aid = 0;
  num_sta++;
  return sta;
}

void sap_free_sta (struct sta_info *sta)
{

  if (sta->aid > 0)
    sta_aid[sta->aid - 1] = NULL;

  num_sta--;
 
  kfree (sta);
}


struct sta_info *sap_get_sta (u8 * sta)
{
  struct sta_info *s;

  s = sta_hash[STA_HASH (sta)];
  while (s != NULL && memcmp (s->mac, sta, 6) != 0)
    s = s->next;
  return s;
}

int sap_hash_init(void){

  int ii;
  for(ii=0;ii<HASH_SIZE;ii++) sta_hash[ii] = NULL;
  for(ii=0;ii<MAX_AID;ii++) sta_aid[ii] = NULL;
  num_sta = 0;
  return 0;
}

void sap_hash_cleanup(void){

  int ii;

  for(ii=0;ii<HASH_SIZE;ii++){
    if (sta_hash[ii]){
      struct sta_info *sta = sta_hash[ii];
      while (sta){
	sta_hash[ii] = sta->next;
	kfree(sta);
	sta = sta_hash[ii];
      }
    }
  }

}

int handle_sta_ps(struct sta_info *sta, struct sk_buff *skb){
  return 0;
}

u8 get_tx_rate(struct sta_info *sta){
  return 110; // for now
}
