#ifndef STA_DB_H
#define STA_DB_H

#ifndef BIT
#define BIT(x) (1 << (x))
#endif

/* STA flags */
#define WLAN_STA_AUTH BIT(0)
#define WLAN_STA_ASSOC BIT(1)
#define WLAN_STA_PS BIT(2)

#define WLAN_RATE_1M BIT(0)
#define WLAN_RATE_2M BIT(1)
#define WLAN_RATE_5M5 BIT(2)
#define WLAN_RATE_11M BIT(3)
#define WLAN_RATE_COUNT 4

#define WLAN_TX_RATE_1M 10
#define WLAN_TX_RATE_2M 20
#define WLAN_TX_RATE_55M 55
#define WLAN_TX_RATE_11M 110

#define MAX_PS_BUFFERED_PKTS 32

struct sta_info {
  __u8 mac[6];
  __u16 aid;
  __u32 flags;
  __u16 listen_interval;
  //  __u8 supported_rates[8];
  
  __u8 tx_rate; // current_tx_rate
  
  __u8 last_rx_silence;
  __u8 last_rx_signal;
  __u8 last_rx_rate;
  __u8 last_rx_flow;
  __u32 last_rx_updated;

  __u32 tx_since_last_failure;

  __u8 tx_supp_rates;
  __u8 supported_rates[WLAN_RATE_COUNT];
  __u8 tx_max_rate;
  __u8 tx_rate_idx;

  __u32 tx_count[WLAN_RATE_COUNT];
  __u32 rx_count[WLAN_RATE_COUNT];
  
  struct sta_info *next; // for linked list
	
};

struct sap_stats{
  __u32 tx_frames, rx_frames;
  __u32 tx_bytes, rx_bytes;
  
  __u32 tx_data_frames, rx_data_frames;
  __u32 tx_mgmt_frames, rx_mgmt_frames;
  __u32 tx_ctrl_frames, rx_ctrl_frames;
  
  __u32 rx_beacons;
  __u32 rx_auth_requests;
  __u32 rx_assoc_requests;
  __u32 rx_probe_requests;
  __u32 rx_disassoc_requests;
  __u32 rx_deauth_requests;
  
  __u32 tx_auth_responses;
  __u32 tx_assoc_responses;
  __u32 tx_disassoc_responses;
  __u32 tx_deauth_responses;
  
};

#define MAX_ESSID_LEN 32

struct other_beacons {
  __u8 bssid[6];
  char essid[MAX_ESSID_LEN];
};


struct sta_info *sap_get_sta(__u8 *);
int handle_sta_ps(struct sta_info *, struct sk_buff *);
struct sta_info * sap_create_sta(__u8 *);
int sap_hash_init(void);
void sap_hash_cleanup(void);
void sap_free_sta (struct sta_info *);
void sap_add_sta (struct sta_info *);
void sap_del_sta (struct sta_info *);



u8 get_tx_rate(struct sta_info *);
void sap_check_tx_rates (struct sta_info *);

extern __u8 my_bssid[6];

#endif // STA_DB_H
