#ifndef _SIBYTE_H_
#define _SIBYTE_H_

#include <linux/config.h>

/*
 * Rings
 */
#define DSC_RING_ENTRIES_SIZE	(SIBYTE_DESCR_ENTRIES * sizeof(struct desc))
#define DSC_NEXT(idx)	        ((idx + 1) & (SIBYTE_DESCR_ENTRIES - 1))

static inline int tx_space (u32 csm, u32 prd)
{
	return (csm - prd - 1) & (SIBYTE_DESCR_ENTRIES - 1);
}

#if MAX_SKB_FRAGS
#define TX_RESERVED	(MAX_SKB_FRAGS+1) /* +1 for message header */
#define tx_ring_full(csm, prd)	(tx_space(csm, prd) <= TX_RESERVED)
#else
#define tx_ring_full		0
#endif

#define SIBYTE_DOWNLOAD		0
#define SIBYTE_FREE		1
#define SIBYTE_MBGET		2
#define SIBYTE_MBSET    	3
#define SIBYTE_PCI_RECONFIG	4
#define SIBYTE_PCI_DUMP  	5
#define SIBYTE_TEST_PANIC	6
#define SIBYTE_TEST_NULLPTR	7
#define SIBYTE_READ_DATA	8
#define SIBYTE_WRITE_DATA	9
#define SIBYTE_GET_VERSION	10
#define SIBYTE_TEST_HANG	11
#define SIBYTE_SYNC		12
#define SIBYTE_GET_STATE        13
#define SIBYTE_GET_STATS        14
#define SIBYTE_GET_BASE_VERSION	15


struct sibyte_cmd {
	u32	cmd;
        u32     address;      /* virtual address of image */
        u32     length;       /* size of image to download */
        u32     mailbox;      /* mailbox to get/set */
        u32     data[2];      /* contents of mailbox to read/write */
        union {
            struct {
                u32     cur_rx[SIBYTE_PRIOS];
                u32     tx_prd[SIBYTE_PRIOS];
                u32     tx_csm[SIBYTE_PRIOS];
            } dsc;
            struct net_device_stats stats;
        } u;
};

/*
 * Struct private for the Sibyte.
 *
 * Elements are grouped so variables used by the tx handling goes
 * together, and will go into the same cache lines etc. in order to
 * avoid cache line contention between the rx and tx handling on SMP.
 *
 * Frequently accessed variables are put at the beginning of the
 * struct to help the compiler generate better/shorter code.
 */
struct sibyte_private
{
        int			version;
        int			promisc;
        u32                     mb[2];

        /*
	 * RX and TX descriptors, must be adjacent
	 */
        sibyte_descr_t	        *rx_ring[SIBYTE_PRIOS];
        sibyte_descr_t	        *tx_ring[SIBYTE_PRIOS];

        struct sk_buff		**rx_skb[SIBYTE_PRIOS];
        struct sk_buff		**tx_skb[SIBYTE_PRIOS];

	/*
	 * RX elements
	 */
#ifndef SIBYTE_TARGET
        dma_addr_t		ring_base_dma;
#endif /* SIBYTE_TARGET */
        u32			rx_skbprd[SIBYTE_PRIOS];
	u32			cur_rx[SIBYTE_PRIOS];

	/*
	 * TX elements
	 */
	u32			tx_prd[SIBYTE_PRIOS];
        u32		        tx_csm[SIBYTE_PRIOS];

        /*
	 * Misc elements
	 */
	struct pci_dev		*pdev;
	struct net_device	*next;
	int			board_idx;
	u8			pci_latency;
	u16			pci_cmdstatus;
	int			base_addr0;
	int			base_addr2;
	int			base_addr3;
	char			name[48];
	struct net_device_stats stats;
	int		        fw_running;
        struct {
	    u32 address;
	    u32 length;
	    char *mapping;
	} desc;
};


/*
 * Prototypes
 */
static int  sibyte_init(struct net_device *dev);
#ifndef SIBYTE_TARGET
static void sibyte_load_rx_ring(struct sibyte_private *ap, u32 ring, int bufs);
#endif /* SIBYTE_TARGET */
static void sibyte_interrupt(int irq, void *dev_id, struct pt_regs *regs);
static int  sibyte_open(struct net_device *dev);
static int  sibyte_start_xmit(struct sk_buff *skb, struct net_device *dev);
static int  sibyte_close(struct net_device *dev);
#ifndef SIBYTE_TARGET
static void sibyte_tasklet(unsigned long dev);
#endif /* SIBYTE_TARGET */
static int  sibyte_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
static void sibyte_init_cleanup(struct net_device *dev);
static struct net_device_stats *sibyte_get_stats(struct net_device *dev);
#ifdef SIBYTE_LOAD_FIRMWARE
static int  sibyte_load_firmware(struct net_device *dev);
#endif /* SIBYTE_LOAD_FIRMWARE */

#endif /* _SIBYTE_H_ */
