/*
 * Aruba tunnel device w/ hard coded sibyte msg flags.
 *
 */


#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/netdevice.h>
#include <linux/poll.h>
#include <linux/ppp_defs.h>
#include <linux/filter.h>
#include <linux/if_ppp.h>
#include <linux/ppp_channel.h>
#include <linux/ppp-comp.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <linux/if_arp.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/spinlock.h>
#include <linux/smp_lock.h>
#include <linux/rwsem.h>
#include <linux/stddef.h>
#include <linux/sysctl.h>
#include <net/slhc_vj.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <net/sock.h>
#include <net/ip.h>

#if defined(CONFIG_MSWITCH) || defined(CONFIG_CABERNET) || defined(CONFIG_SYRAH) || defined(CONFIG_MALBEC)
#include "../../../cmn/util/sibyte_types.h"
#include "../../../cmn/util/sibyte_msg.h"
#include "../../../cmn/util/sibyte_l2.h"
#include "../../../cmn/util/stddefs.h"
#endif
#include "tsgw.h"

#define ETH1_LEN	5

static int sysctl_tsgw_tx_enable;
static int min_tsgw_tx_enable=0;
static int max_tsgw_tx_enable =1;
	
static int tsgw_tunnel_init(void);

static struct net_device *pdev;
static struct ctl_table_header *tsgw_table_header;

/* header needed by sibyte.c */
typedef struct aruba_hdr {
    sibyte_msg_t       sbmsg;
    unsigned char      dst[6], src[6]; /* MAC */
    //unsigned short     vlantag, vlan;  /* VLAN TAG */
    unsigned short      etype;
} aruba_hdr_t;

typedef struct tsgw_tunnel {
    struct tsgw_tunnel      *next;
    aruba_hdr_t              sbhdr;
    struct net_device       *eth1dev;
    struct net_device       *dev;
	struct net_device_stats	 stat;
} tsgw_tunnel_t;


static ctl_table tsgw_tx_table[] = {
		{CTL_TSGW_TX_ENABLE,"tsgw_tx_enable",  &sysctl_tsgw_tx_enable,  sizeof(int) , 0666,  NULL, 
		  &proc_dointvec_minmax, &sysctl_intvec, NULL, &min_tsgw_tx_enable, &max_tsgw_tx_enable},
		{0}
};


#if 0
static int tsgw_ctl_handler (ctl_table *table, int *name, int nlen,   
									void *oldval, size_t *oldlenp, 
									void *newval, size_t newlen,  void **context)    
{   

		printk("->tsgw_ctl_handler\n");
		if (oldval) 
		{
			printk("Reading!\n");            
			copy_to_user(oldval, "bonk!", (sizeof(char) * 6));           
			put_user((sizeof(char) * 6), oldlenp);        
		}     

		
		if ( newval && newlen) 
		{         
	           copy_from_user(table->data, newval, newlen);   
			printk("Wrinting '%s'!\n", table->data);        
			sysctl_tsgw_tx_enable = table->data;	
		}        	    
		
		return 1;    
}
#endif

/*
This function goes through all net devices and checks if packet is destined to one of switch 
interfaces subnet. If yes,checks arp table. If entry is not found , creates arp entry 
and sends arp request. If entry found then returns.
returns :	-1  error
			0    packet is not destined to our subnets
			1	arp entry is found
			2	creates arp entry and sends arp request
			3   found arp entry with 0 mac address
		
*/
static int check_arp_entry(struct sk_buff  *skb)
{
	struct net_device *dev=NULL;
	struct in_device *in_dev=NULL;
	struct in_ifaddr	*ifa_addr=NULL;
	struct neighbour *n=NULL;
	struct iphdr  *iph=skb->nh.iph;
	__u32 if_addr, if_mask, d_addr, s_addr;
	__u32 begin, end;
	char *eth1_dev="eth1.";
	char zero[MAX_ADDR_LEN];
	
	
	if (!iph )
	{
		return -1;
	}	
	d_addr = iph->daddr;
	s_addr = iph->saddr;
	memset(zero, 0, MAX_ADDR_LEN);
	
	//printk( "d_addr %0x\n",d_addr);		
	read_lock(&dev_base_lock);
	for (dev=dev_base; dev != NULL; dev=dev->next )
	{
		if ( !(dev->flags & IFF_UP) || (dev->flags & IFF_POINTOPOINT) )
			continue;
		if (memcmp(dev->name, eth1_dev, ETH1_LEN) )
			continue;

		in_dev = in_dev_get(dev);
		if(!in_dev)
			continue;
		read_lock(&in_dev->lock);
		
		
		//printk("Dev %s \n",dev->name);
		for(ifa_addr=in_dev->ifa_list; ifa_addr != NULL; ifa_addr=ifa_addr->ifa_next)
		{
				if_addr=ifa_addr->ifa_local;
				if_mask=ifa_addr->ifa_mask;
				/*Check if packet is destined to the same subnet*/
				begin =  (if_addr & if_mask) +(1 & ~if_mask);
				end = if_addr | ~if_mask;
				
				if ( (begin <=d_addr) && (d_addr<= end) )
				{
					//printk("Checks ARP table\n");
					//Check if there is ARP entry
					n=neigh_lookup(&arp_tbl, &d_addr, dev);
					if (n == NULL)
					{
						//Creates arp entry and sends arp
						//n=__neigh_lookup(&arp_tbl, &d_addr, dev,1);
						n = neigh_create(&arp_tbl, &d_addr, dev);
						if ( IS_ERR(n) )
						{
							read_unlock(&in_dev->lock);
							in_dev_put(in_dev);
							read_unlock(&dev_base_lock);
							return -1;			
						}							
						arp_send(ARPOP_REQUEST, ETH_P_ARP, d_addr, dev, s_addr,
						 NULL, dev->dev_addr,NULL);
						read_unlock(&in_dev->lock);
						in_dev_put(in_dev);
						read_unlock(&dev_base_lock);
						return 1;
					}
					else
					{
						//packet destined to our subnet has an arp entry
						//Check mac address
						if ( memcmp(n->ha, zero, dev->addr_len ) == 0)
						{
							//printk ("Mac address is zero\n");	
							read_unlock(&in_dev->lock);
							in_dev_put(in_dev);
							read_unlock(&dev_base_lock);
							return 3;
						}
						else
						{
							//printk ("Found\n");	
							read_unlock(&in_dev->lock);
							in_dev_put(in_dev);
							read_unlock(&dev_base_lock);
							return 2;
						}
					}

				}
		}	
		read_unlock(&in_dev->lock);
		in_dev_put(in_dev);
		
	}	
	read_unlock(&dev_base_lock);
	return 0;
}

/*
 * TX code.  Just make sure there's room and copy the packet hdr, then
 * transmit out of eth1.
 */
static int
tsgw_tunnel_xmit(struct sk_buff *skb, struct net_device *dev)
{
    struct tsgw_tunnel *tun = (struct tsgw_tunnel *)dev->priv;
	int status;
	 
//	printk ("->tsgw_tunnel_xmit sysctl_tsgw_tx_enable=%d\n", sysctl_tsgw_tx_enable);
	if (tun->eth1dev && sysctl_tsgw_tx_enable)
	{
		status=check_arp_entry(skb);
		if( status < 0 || status == 3)
		{
			tun->stat.tx_dropped++;
			dev_kfree_skb(skb);
			return 0;
		}
		
		skb->h.raw = skb->nh.raw;
		if (	skb_headroom(skb) < sizeof(struct aruba_hdr) || skb_cloned(skb) ||
	        	skb_shared(skb)) 
	      {
			struct sk_buff *new_skb = skb_realloc_headroom(skb, sizeof(struct aruba_hdr));
			if (!new_skb) {
	  			tun->stat.tx_dropped++;
				dev_kfree_skb(skb);
				return 0;
			}
			if (skb->sk)
				skb_set_owner_w(new_skb, skb->sk);
			dev_kfree_skb(skb);
			skb = new_skb;
		}

	    /* we don't actually use skb->h later since we send immediately, but just
	       in case we need it in the future */
		skb->nh.raw = skb_push(skb, sizeof(struct aruba_hdr));
	    	memcpy(skb->nh.raw, &tun->sbhdr, sizeof(struct aruba_hdr));
        	skb->protocol = SIBYTE_MSG_PROTO(SIBYTE_MSG_OPCODE_SESSIONPPP);

	    /* send out on eth1 immediately, no need to route any further since if
	       the tunnel is up, then a route in sibyte must already exist */
	     skb->dev = tun->eth1dev;	 		 
	    if( dev_queue_xmit(skb) < 0)
	    {
	           tun->stat.tx_dropped++;
			return 0;
	    }
		tun->stat.tx_packets++;
		tun->stat.tx_bytes += skb->len; 
		//printk ("<-tsgw_tunnel_xmit\n");
	}
	return 0;
}

static struct net_device_stats *
tsgw_tunnel_get_stats(struct net_device *dev)
{
	//printk ("tsgw_tunne_get_statsl\n");
	return &(((struct tsgw_tunnel*)dev->priv)->stat);
}

int tsgw_open (struct net_device *dev)
{
		//printk("tsgw_open\n");
		if (!capable(CAP_NET_ADMIN))
		{
			printk("tsgw is not capable for CAP_NET_ADMIN ");
			return -EPERM;
		}
		netif_start_queue (dev);
		return 0;
}

int tsgw_stop (struct net_device *dev)
{
		//printk ("tsgw_stop\n");
		netif_stop_queue(dev);
		return 0;
}

int tsgw_init (struct net_device *dev)
{
//		printk ("tsgw_init device initialized\n");
		sysctl_tsgw_tx_enable = TSGW_TX_DISABLE;
 		dev->mtu = PPP_MTU;
		dev->addr_len = 0;
		dev->type = ARPHRD_PPP;
		dev->flags = IFF_POINTOPOINT | IFF_NOARP | IFF_MULTICAST;
		dev->get_stats = tsgw_tunnel_get_stats;
		dev->open = tsgw_open;
		dev->stop = tsgw_stop;
		dev->hard_start_xmit = tsgw_tunnel_xmit;
		return 0;
}



static int tsgw_tunnel_init(void)
{
	int result;
	struct net_device *dev = NULL;
	struct tsgw_tunnel *tsgw=NULL;

//	printk ("tsgw_tunnel_init\n"); 
	dev = kmalloc(sizeof(struct net_device), GFP_KERNEL);
	if (dev == 0)
	{
		printk("tsgw: Could not allocate dev\n");
		return -1;
	}
	tsgw = kmalloc(sizeof(struct tsgw_tunnel), GFP_KERNEL);
	if (tsgw == 0)
	{
		kfree(dev);
		printk("tsgw: Could not allocate tsgw\n");
		return -1;
	}
	
	memset(dev, 0, sizeof(struct net_device));
	memset(tsgw, 0, sizeof(struct tsgw_tunnel));
	
	dev->init = tsgw_init;
	strcpy(dev->name, "tsgw");
	dev->priv = tsgw;

	tsgw->sbhdr.sbmsg.opcode  = SIBYTE_MSG_OPCODE_SESSIONPPP;
     	tsgw->sbhdr.sbmsg.ingress = SIBYTE_DEST_CP;
 	tsgw->sbhdr.sbmsg.vlan    = 1;
	tsgw->sbhdr.etype = ETH_P_IP;
	tsgw->eth1dev = dev_get_by_name("eth1");
	if (tsgw->eth1dev == NULL) {
        printk(KERN_ERR "tsgw: can't find eth1 device.\n");
        kfree(dev);
        kfree(tsgw);
    }
	result = register_netdev (dev);
	if (result)
	{
		printk ("tsgw: Error %d  in register_netdev",result);
		kfree(dev);
		kfree(tsgw);
		return result;
	}
	pdev = dev;
	 tsgw_table_header = register_sysctl_table(tsgw_tx_table, 0);
	 if (!tsgw_table_header )
	 {
		printk ("tsgw: Error  in register_sysctl_table");
		kfree(dev);
		kfree(tsgw);
	 	return -1;
	 }
	return 0;
}

void tsgw_cleanup (void)
{
		//printk ("tsgw_cleanup\n");
		
		if(pdev)
			unregister_netdev (pdev);
		
		if (tsgw_table_header)            
			unregister_sysctl_table(tsgw_table_header);
		
		if(pdev && pdev->priv)
			kfree(pdev->priv);
		if(pdev)
			kfree(pdev);
		return;
}

module_init (tsgw_tunnel_init);
module_exit(tsgw_cleanup);

