/*
 * BK Id: SCCS/s.immap_8245.h 1.8 07/18/01 15:46:50 trini
 */

/*
 * MPC8245 Internal Memory Map
 * Copyright (c) 1999 Dan Malek (dmalek@jlc.net)
 *
 * The Internal Memory Map of the 8245.  I don't know how generic
 * this will be, as I don't have any knowledge of the subsequent
 * parts at this time.  I copied this from the 8xx_immap.h.
 */
#ifdef __KERNEL__
#ifndef __IMMAP_824X__
#define __IMMAP_824X__

/* I2C
*/
typedef struct i2c {
	u_char	i2c_i2cadr;
	u_char	i2c_i2cfdr;
	u_char	i2c_i2ccr;
	u_char	i2c_i2csr;
	u_char	i2c_i2cdr;
} i2c8245_t;

#define MPC824X_I2C_ADR_REG 		((unsigned long)&(I2Cptr->i2cAddrReg))
#define MPC824X_I2C_FREQ_DIV_REG	((unsigned long)&(I2Cptr->i2cFreqDivReg))
#define MPC824X_I2C_CONTROL_REG		((unsigned long)&(I2Cptr->i2cControlReg))
#define MPC824X_I2C_STATUS_REG		((unsigned long)&(I2Cptr->i2cStatusReg))
#define MPC824X_I2C_DATA_REG		((unsigned long)&(I2Cptr->i2cDataReg))
 
/* Mpc824X Register masks */
#define MPC824X_I2C_ADDRESS_REG_MASK	0x000000FE
#define MPC824X_I2C_FREQ_DIV_REG_MASK	0x0000003F
#define MPC824X_I2C_CONTROL_REG_MASK	0x000000FC
#define MPC824X_I2C_STATUS_REG_MASK	0x000000F7
#define MPC824X_I2C_DATA_REG_MASK	0x000000FF
 
/* Mpc824X Control register values */
#define MPC824X_I2C_CONTROL_REG_MEN	(1<<7) /* module enable */
#define MPC824X_I2C_CONTROL_REG_MIEN	(1<<6) /* module interrupt enable */
#define MPC824X_I2C_CONTROL_REG_MSTA	(1<<5) /* master/slave mode */
#define MPC824X_I2C_CONTROL_REG_MTX	(1<<4) /* transmit/receiver mode */
#define MPC824X_I2C_CONTROL_REG_TXAK	(1<<3) /* transfer ack enable */
#define MPC824X_I2C_CONTROL_REG_RSTA	(1<<2) /* repeat start */
#define MPC824X_I2C_CONTROL_REG_PCII	(1<<1) /* PCI Interrupt enable */
#define MPC824X_I2C_CONTROL_REG_BCST	(1<<0) /* broadcast enable */
 
/* Mpc824X Status register values */
#define MPC824X_I2C_STATUS_REG_MCF	(1<<7) /* data transferring */
#define MPC824X_I2C_STATUS_REG_MAAS	(1<<6) /* addressed as a slave */
#define MPC824X_I2C_STATUS_REG_MBB	(1<<5) /* bus busy */
#define MPC824X_I2C_STATUS_REG_MAL	(1<<4) /* arbitration lost */
#define MPC824X_I2C_STATUS_REG_BCA	(1<<3) /* broadcast address detection */
#define MPC824X_I2C_STATUS_REG_SRW	(1<<2) /* slave read/write */
#define MPC824X_I2C_STATUS_REG_MIF	(1<<1) /* module interrupt */
#define MPC824X_I2C_STATUS_REG_RXAK	(1<<0) /* receive ack */

/* low level definitions, specific */
#define SYNC __asm__ volatile ("sync")

/* endian swap */
#define BYTE_SWAP_16_BIT(x) ( (((x) & 0x00ff) << 8) | ( (x) >> 8) )

#define LONGSWAP(x) ((((x) & 0x000000ff) << 24) | \
		     (((x) & 0x0000ff00) << 8 ) | \
		     (((x) & 0x00ff0000) >> 8 ) | \
		     (((x) & 0xff000000) >> 24) )

#define PCISWAP(x) LONGSWAP(x)



#endif /* __IMMAP_824X__ */
#endif /* __KERNEL__ */
