#ifndef ARUBA_SAP_H
#define ARUBA_SAP_H

/* different powerpc models */
#define ASAP_AP52   0   /* Accton AP */
#define ASAP_AP1200 1   /* Airespace AP */


#define SAP_CPLD_LED_BASE 0x70000000
#ifdef ARUBA_LINUX_BIN
// needed for initial linux image to boot from VxWorks bootloader
#define SAP_CPLD_WDG_BASE 0x78000000
#else
#define SAP_CPLD_WDG_BASE 0x70010000
#endif

/* kernel watchdog reset for Airespace AP1200 */
#define WATCHDOG_RESET()  do { (void) *(volatile char *)0xff000000; } while (0)
/* AP-1200 signatures */
#define AP1200_SIG "1200"
#define AP1200_ALCATEL_EXTANT_SIG "-E"
#define AP1200_ALCATEL_REAP_SIG   "1200R"
#define AP1200_SIG_ADDR (char *)0xffffa008

/* Airespace AP1200 GPIO defines.  GPIO is accessed over I2C bus */
#define ASAP_AP1200_GPIO_I2C_DEV   0x20
#define ASAP_AP1200_GPIO_ANT_REG   2
#define ASAP_AP1200_GPIO_ANT_2INT  0x0a
#define ASAP_AP1200_GPIO_ANT_2EXT  0x05
#define ASAP_AP1200_GPIO_ANT_5TOP  0x40
#define ASAP_AP1200_GPIO_ANT_5BOT  0x80

#define ASAP_AP1200_GPIO_LED_REG   3
#define ASAP_AP1200_GPIO_LED_ALARM 0x02
#define ASAP_AP1200_GPIO_LED_2GHZ  0x04
#define ASAP_AP1200_GPIO_LED_POWER 0x08
#define ASAP_AP1200_GPIO_LED_5GHZ  0x10
#define ASAP_AP1200_GPIO_LED_MASK  0x1e

extern int asap_model;

/* copied from <linux/pci.h> (too painful to include) */
#define SAP_PCI_SLOT(devfn) (((devfn) >> 3) & 0x1f)
#define AP1200_IS_PCI_DOTA_SLOT(devfn) (SAP_PCI_SLOT(devfn) == 0x1C)
#define AP1200_IS_PCI_DOTG_SLOT(devfn) (SAP_PCI_SLOT(devfn) == 0x1E)

extern int i2c_aruba_write(__u16 addr, __u8 reg, char data);
extern unsigned int i2c_aruba_read(__u16 addr, __u8 reg);

#endif
