#ifndef _LINUX_MSWITCH_H
#define _LINUX_MSWITCH_H

struct mswitch_monitor {
        unsigned long           int_count;
        unsigned long           char_count;
        unsigned long           char_max;
        unsigned long           char_last;
};

/*
 * These stats all reflect activity since the device was last initialized.
 * (i.e., since the port was opened with no other processes already having it
 * open)
 */
struct mswitch_idle_stats {
    time_t	   in_use;	/* Time device has been in use (secs) */
    time_t	   recv_idle;	/* Time since last char received (secs) */
    time_t	   xmit_idle;	/* Time since last char transmitted (secs) */
    unsigned long  recv_bytes;	/* Bytes received */
    unsigned long  xmit_bytes;	/* Bytes transmitted */
    unsigned long  overruns;	/* Input overruns */
    unsigned long  frame_errs;	/* Input framing errors */
    unsigned long  parity_errs;	/* Input parity errors */
};

#define MSWITCH_MAGIC           0x4359

/****************** ****************** *******************/

/* Per card data structure */
struct mswitch_card {
    unsigned long base_phys;
    unsigned long ctl_phys;
    unsigned long base_addr;
    unsigned long ctl_addr;
    int irq;
    int num_chips;	/* 0 if card absent, -1 if Z/PCI, else Y */
    int first_line;	/* minor number of first channel on card */
    int nports;		/* Number of ports in the card */
#ifdef __KERNEL__
    spinlock_t card_lock;
#else
    unsigned long filler;
#endif
};

#ifdef __KERNEL__

/*
 * Statistics counters
 */
struct mswitch_icount {
	__u32	cts, dsr, rng, dcd, tx, rx;
	__u32	frame, parity, overrun, brk;
	__u32	buf_overrun;
};

/*
 * This is our internal structure for each serial port's state.
 * 
 * Many fields are paralleled by the structure used by the serial_struct
 * structure.
 *
 * For definitions of the flags field, see tty.h
 */

struct mswitch_port {
	int                     magic;
	int			card;
	int			line;
	int			flags; 		/* defined in tty.h */
	int                     type;		/* UART type */
	struct tty_struct 	*tty;
	int			timeout;
	int			xmit_fifo_size;
	int			baud;
	int			custom_divisor;
	int			close_delay;
	unsigned short		closing_wait;
	unsigned long		event;
	unsigned long		last_active;
	int			count;	/* # of fd on device */
	int                     breakon;
	int                     breakoff;
	int			blocked_open; /* # of blocked opens */
	long			session; /* Session of opening process */
	long			pgrp; /* pgrp of opening process */
	unsigned char 		*xmit_buf;
	int			xmit_head;
	int			xmit_tail;
	int			xmit_cnt;
	unsigned long		jiffies[3];
	struct termios		normal_termios;
	struct mswitch_monitor	mon;
	struct mswitch_idle_stats	idle_stats;
	struct mswitch_icount	icount;
	struct tq_struct	tqueue;
	wait_queue_head_t       open_wait;
	wait_queue_head_t       close_wait;
	wait_queue_head_t       shutdown_wait;
};

/*
 * Events are used to schedule things to happen at timer-interrupt
 * time, instead of at cy interrupt time.
 */
#define Mx_EVENT_READ_PROCESS		0
#define Mx_EVENT_WRITE_WAKEUP		1
#define Mx_EVENT_HANGUP			2
#define Mx_EVENT_BREAK			3
#define Mx_EVENT_OPEN_WAKEUP		4
#define Mx_EVENT_SHUTDOWN_WAKEUP	5
#define	Mx_EVENT_DELTA_WAKEUP		6

#define	CLOSING_WAIT_DELAY	5*HZ
#define MX_CLOSING_WAIT_NONE	65535
#define MX_CLOSING_WAIT_INF	0
#define	MX_FIFO_SIZE	16

extern int __init mx_init(void);


/***************************************************************************/

#endif /* __KERNEL__ */
#endif /* _LINUX_MSWITCH_H */
