/*
 * prom.c - PROM console driver for IDT79S334 board
 *
 * Copyright (C) 2000 by Lineo, Inc.
 * Written by Quinn Jensen (jensenq@lineo.com)
 *
 */
#include <linux/config.h>
#include <linux/kdev_t.h>
#include <linux/major.h>
#include <linux/console.h>
#include <asm/bootinfo.h>
#include <asm/page.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/bootmem.h>
#include <linux/serial.h>
#include <linux/serialP.h>
#include <asm/serial.h>
#include <linux/ioport.h>
#include <asm/aruba/merlot.h>

char arcs_cmdline[COMMAND_LINE_SIZE];

/* Kernel Boot parameters */
// static unsigned char bootparm[]="console=ttyS00,9600";
static unsigned char bootparm[]="";

#ifndef CONFIG_BOOT_FROM_PROM

#define NVRAM_BASE_ADDR   0xac080000   

#define NVRAM_ENVSTART   32
#define NVRAM_ENVSIZE   4            

#define WAIT(x) { volatile int i=0; while (++i < (x)) ; }

unsigned char nvramRead( int offs )
{
   unsigned char* nvramDataPointer=(unsigned char *)(NVRAM_BASE_ADDR + offs);

   return ((unsigned char)(*nvramDataPointer)) ;
}
#endif /* CONFIG_BOOT_FROM_PROM */
  
#define PROM_ENTRY(x)   (0xbfc00000+((x)*8))

extern void cons_putc(char c);
extern void cons_puts(char *s);

void (*prom_puts)(char *str) = cons_puts;
void (*prom_putchar)(char c) = cons_putc;
int (*prom_getchar)(void) = (int (*)(void))PROM_ENTRY(11);

char ether_cmdline[32]; /* For Ethernet MAC address storage */
unsigned int idt_cpu_freq = 100;	/* default to 100 MHz */

#ifdef CONFIG_REMOTE_DEBUG
extern int remote_debug;
#endif

/* 
 * register serial ports at run time when we can compute baud_base
 */

#define ARUBA_BASE CONS_PORT

struct serial_struct idt_serial[2] = {
/*	{flags: STD_COM_FLAGS, iomem_base: (u8 *)(ARUBA_BASE), 
		iomem_reg_shift: 2, io_type: SERIAL_IO_MEM},
	{flags: STD_COM_FLAGS, iomem_base: (u8 *)(ARUBA_BASE + 0x20),
		iomem_reg_shift: 2, io_type: SERIAL_IO_MEM}, */
};
#ifndef CONFIG_BOOT_FROM_PROM
void prom_setup_cmdline(void)
{

    int byteCount = 0;
    int envCount = NVRAM_ENVSTART;

    char *cmdp = arcs_cmdline;
    char *ethp = ether_cmdline;

    int clen = COMMAND_LINE_SIZE - 1;

    int esize = (int) ( (nvramRead(NVRAM_ENVSIZE)<< 8) + 
			nvramRead(NVRAM_ENVSIZE + 1) );

    if(esize < 0 || esize > 512) return;	/* bad format */
    
    *cmdp = '\0';
    *ethp = '\0';

	/* copy relevant env vars from the nvram
	 *
	 * prom environment variables of the form "bootparm?=stuff"
	 * are combined in the order they appear in the nvram to
	 * to form the boot command line.  e.g.
	 *
	 *	bootaddr=1.2.3.4
	 *	bootparm1=root=/dev/nfs
	 *	bootparm2=nfsroot=/home/nfsroot,nolock
	 *
	 * becomes
	 *
	 *	root=/dev/nfs nfsroot=/home/nfsroot,nolock
	 */

    for(;;){

	/* first byte is length */
	unsigned int elen =
	    (unsigned int) (nvramRead(envCount) & 0xFF );
    
	char tmp_env[256], *env = tmp_env;
	
	if(elen == 0 || elen > esize) break;	/* bad format */
    
	for(byteCount=0; byteCount<elen; byteCount++)
	    env[byteCount] = nvramRead(envCount + byteCount);
	env[byteCount] = '\0';
	
	envCount += elen;
	esize -= elen;
    
	env++;
	elen--;

	if(strncmp(env, "bootparm", 8) == 0) {
	    unsigned int vlen;

	    /* skip beyond the '=' and trim to fit */
	    env += 10;
	    vlen = elen - 10;
	    if(clen < vlen + 2) vlen = clen;

	    /* copy it */

	    memcpy(cmdp, env, vlen);
	    cmdp += vlen; clen -= vlen;

	    /* add a blank and null terminate */
	    
	    *cmdp++ = ' '; clen--;
	    *cmdp = '\0';
	} 
	else if(strncmp(env,"ethaddress",10) == 0) {
#if 0
	    extern unsigned char Ether_Addr[6]; /* drivers/net/bancore.c */
	    unsigned int vlen;
	    unsigned int i =0;
	    unsigned int sn_enaddr[6];
	    char *p;

	    /* skip beyond the '=' and trim to fit */
	    env += 11;
	    vlen = elen - 11;
	    /* copy it */

	    memcpy(ethp, env, vlen);
	    ethp += vlen;
	    *ethp = '\0';

	    p = ether_cmdline;

	    for (i = 0; i < 6; i++){
		sn_enaddr[i] = simple_strtoul (p, &p, 0);
		if (i != 5) {
		    if (*p != '-' && *p != ':')
			{
			    break;
			}
		    p++;
		}
	    }
	    if (i != 6 || *p) {
		printk("Ethernet address %s is not valid. Using defaults.\n",
		       ether_cmdline);
		continue;
	    }
	    else
		for(i=0;i<6;i++)

		    Ether_Addr[i] = sn_enaddr[i];

	    printk("Kernel is compiled without ethernet support.\n");
#endif
	}

	else if(strncmp(env, "MHZ=", 4) == 0) {
	    idt_cpu_freq = simple_strtoul(env + 4, 0, 10);
	    printk("CPU Clock at %d MHz (from MHZ environment variable)\n", 
		   idt_cpu_freq);
	}
	
    
	/* all done? */
	
	if(clen <= 0) break;
	if(esize <= 0) break;
    }
    printk("\n");
}
#endif
void prom_console_write(struct console *con, const char *s, unsigned count)
{
	int i;

	for(i = 0; i < count; i++, s++) {
		if(*s == '\n') prom_putchar('\r');
		prom_putchar(*s);
	}
}

kdev_t prom_console_device(struct console *con)
{
	return MKDEV(TTY_MAJOR, 64 + con->index);
}

int prom_console_wait_key(struct console *con)
{
	return prom_getchar();
}

int prom_console_setup(struct console *con, char *options)
{
	return 0;
}

struct console prom_console_driver = {
	"ttyS",
	prom_console_write,		/* write */
	NULL,				/* read */
	prom_console_device,		/* device */
	prom_console_wait_key,		/* wait_key */
	NULL,				/* unblank */
	prom_console_setup,   		/* setup */
	CON_PRINTBUFFER,
	-1,
	0,
	NULL
};

#define PFN_UP(x)	(((x) + PAGE_SIZE-1) >> PAGE_SHIFT)
#define PFN_ALIGN(x)	(((unsigned long)(x) + (PAGE_SIZE - 1)) & PAGE_MASK)

/* ARUBA Merlot memory map -- we really should be auto sizing it */

#define RAM_FIRST       0x80000400  /* Leave room for interrupt vectors */
#define RAM_SIZE        (32*1024*1024) /* default ram size...filled in later */
#define RAM_END         (0x80000000 + RAM_SIZE)     

struct resource merlot_res_ram = {
	"RAM",
	0,
	RAM_SIZE,
	IORESOURCE_MEM
};
extern void _ftext, _end;
extern struct serial_state rs_table[];
extern volatile unsigned char *ports[2];

int getboardtype(void) {
	char *ptr;
	int done=0;

#if 0
	// DAMN!! DAMN!! SAMSUNG's boardname is not in
	// the same place as Merlot and Muscat. 
	// Find it.
	ptr = (char *)0xbfdfc000;
	while((!done) && ((int) ptr < 0xbfdfc0ff)) {
            if (*ptr == 'b') {
		if(strncmp(ptr, "boardname=Samsung", 17) == 0) {
			return MACH_ARUBA_SAMSUNG;
		}
            }
            ptr++;
	}

	// this crashes on SAMSUNG APs...hopefully we exit above.
#endif
	ptr = (char *)0xbfff8000;
	while((!done) && ((int) ptr < 0xbfff80ff)) {
            if (*ptr == 'b') {
		if(strncmp(ptr, "boardname=Merlot", 16) == 0) {
			return MACH_ARUBA_MERLOT;
		} else if(strncmp(ptr, "boardname=Muscat", 16) == 0) {
			return MACH_ARUBA_MUSCAT;
		} else if(strncmp(ptr, "boardname=Mataro", 16) == 0) {
			return MACH_ARUBA_MATARO;
		} else if(strncmp(ptr, "boardname=Alfresco", 18) == 0) {
			return MACH_ARUBA_ALFRESCO;
		} else if(strncmp(ptr, "boardname=WG102", 15) == 0) {
			return MACH_NETGEAR_WG102;
		} else if(strncmp(ptr, "boardname=Nespot", 16) == 0) {
			return MACH_NESPOT_MW1700;
		} else if(strncmp(ptr, "boardname=WiJackDuo", 19) == 0) {
			return MACH_ORTRONICS_WJD;
		} else if(strncmp(ptr, "boardname=WiDuo", 15) == 0) {
			return MACH_ORTRONICS_DUO;
		} else if(strncmp(ptr, "boardname=Delfresco", 19) == 0) {
			return MACH_ARUBA_DELFRESCO;
		} else if(strncmp(ptr, "boardname=WAG102", 16) == 0) {
			return MACH_NETGEAR_WAG102;
		}
            }
            ptr++;
	}

	ptr = (char *)0xbffc0000;
	while((!done) && ((int) ptr < 0xbffc00ff)) {
            if (*ptr == 'b') {
		if(strncmp(ptr, "boardname=RockBottom", 20) == 0) {
			return MACH_ARUBA_ROCKBOTTOM;
                }
		if(strncmp(ptr, "boardname=ReallyRockBottom", 26) == 0) {
			return MACH_ARUBA_REALLYROCKBOTTOM;
		}
            }
            ptr++;
	}

	// DAMN no environment variables
	// return something sane
	return MACH_ARUBA_MERLOT;
}

int __init prom_init(int argc, char **argv, char **envp)
{
	unsigned long configReg, dramSize=(16<<20);
        extern char *linux_model;
	/* turn on the console */

	/* set our arch type */
	mips_machgroup = MACH_GROUP_ARUBA;
	mips_machtype = getboardtype();
        switch (mips_machtype) {
        case MACH_ARUBA_ALFRESCO:
        case MACH_ARUBA_MATARO:
        case MACH_ARUBA_SAMSUNG:
        case MACH_NESPOT_MW1700:
        case MACH_ORTRONICS_WJD:
        case MACH_ORTRONICS_DUO:
        case MACH_NETGEAR_WAG102:
            mips_chip = MACH_ATHEROS_AR5312;  /* processor model */
            mips_bus_freq = 110;              /* used to compute baud, etc */
            break;
        case MACH_ARUBA_MERLOT:
        case MACH_NETGEAR_WG102:
        default:
            mips_chip = MACH_ATHEROS_AR2313;
            mips_bus_freq = 90;
            break;
        case MACH_ARUBA_MUSCAT:
            mips_chip = MACH_IDT_RC32434;
            mips_bus_freq = 133;
            break;
        case MACH_ARUBA_ROCKBOTTOM:
            mips_chip = MACH_ATHEROS_AR2313;
            mips_bus_freq = 90;
            break;
        case MACH_ARUBA_REALLYROCKBOTTOM:
            mips_chip = MACH_ATHEROS_AR2315;
            mips_bus_freq = 92;
            break;
        case MACH_ARUBA_DELFRESCO:
            mips_chip = MACH_IDT_RC32434;
            mips_bus_freq = 175;
            break;
        }

        switch (mips_chip) {
        case MACH_ATHEROS_AR5312:
        case MACH_ATHEROS_AR2313:
            mips_chip_vendor = MACH_VENDOR_ATHEROS;
            mips_local_ip = 6; /* STATUS int */
            mips_intr_status = 0xbc003010;
            mips_uart_irq = 4 + GROUP0_IRQ_BASE;
            mips_uart_base = AR5312_REG_BASE;
            break;
        case MACH_ATHEROS_AR2315:
            mips_chip_vendor = MACH_VENDOR_ATHEROS;
            mips_local_ip = 2; /* STATUS int */
            mips_intr_status = 0xb1000020;
            mips_uart_irq = 0 + GROUP0_IRQ_BASE;
            mips_uart_base = 0xB1100000;
            break;
        case MACH_IDT_RC32434:
            mips_chip_vendor = MACH_VENDOR_IDT;
            mips_uart_irq = GROUP3_IRQ_BASE;
            mips_uart_base = 0xA0000000 + RC32434_REG_BASE + 0x58000;
            break;
        }

        switch (mips_machtype) {
        case MACH_ARUBA_MERLOT:
            linux_model = "6x";
            break;
        case MACH_ARUBA_MUSCAT:
            linux_model = "70";
            break;
        case MACH_ARUBA_MATARO:
            linux_model = "65";
            break;
        case MACH_ARUBA_SAMSUNG:
            linux_model = "??";
            break;
        case MACH_ARUBA_ALFRESCO:
            linux_model = "80";
            break;
        case MACH_NETGEAR_WG102:
            linux_model = "WG102";
            break;
        case MACH_ARUBA_ROCKBOTTOM:
            linux_model = "41";
            break;
        case MACH_ARUBA_REALLYROCKBOTTOM:
            linux_model = "40";
            break;
        case MACH_NESPOT_MW1700:
            linux_model = "MW1700";
            break;
        case MACH_ORTRONICS_WJD:
            linux_model = "WiJackDuo";
            break;
        case MACH_ORTRONICS_DUO:
            linux_model = "Duo";
            break;
        case MACH_ARUBA_DELFRESCO:
            linux_model = "85";
            break;
        case MACH_NETGEAR_WAG102:
            linux_model = "WAG102";
            break;
        }

	rs_table[0].baud_base = MERLOT_BASE_BAUD;
	rs_table[0].iomem_base = (char *) KSEG1ADDR(MERLOT_UART_BASE);
	rs_table[0].irq = MERLOT_UART_IRQ;
	rs_table[0].iomem_reg_shift = 2;
	rs_table[0].io_type = SERIAL_IO_MEM;
	rs_table[0].flags = STD_COM_FLAGS;
	rs_table[0].type = 3;

	idt_serial[0].flags = STD_COM_FLAGS;
	idt_serial[0].iomem_base = (u8*)ARUBA_BASE;
	idt_serial[0].iomem_reg_shift = 2;
	idt_serial[0].io_type = SERIAL_IO_MEM;

	ports[0] = (volatile unsigned char *)CONS_PORT;
	ports[1] = (volatile unsigned char *)GDB_PORT;

#define CONFIG_IDT_PROM_CONSOLE
#ifdef CONFIG_IDT_PROM_CONSOLE

	register_console(&prom_console_driver);

#endif

	/* set up command line */

#ifdef CONFIG_BOOT_FROM_PROM
	sprintf(arcs_cmdline,"%s\n",bootparm);
#else
	prom_setup_cmdline();
#endif

	/*
	 * give all RAM to boot allocator,
	 * except where the kernel was loaded
	 */

        printk("%s\n", get_system_type());
#if ARUBA_DEBUG
	printk("prom.c:RAM_FIRST %x _ftext %x _end %x\n",
		PHYSADDR((unsigned long)RAM_FIRST),
		PHYSADDR((unsigned long)&_ftext),
		PHYSADDR((unsigned long)&_end));
#endif

        switch (mips_chip_vendor) {
        case MACH_VENDOR_ATHEROS:
        default:
            switch (mips_chip) {
            case MACH_ATHEROS_AR2313:
            case MACH_ATHEROS_AR5312:
		/* read the DRAM config register, set by apboot */
		configReg = *(unsigned long *)0xb8300004;
		switch((configReg>>8) & 0x7)
		{
			case 0x0: dramSize = 0x00200000; break; // 2   MB
			case 0x1: dramSize = 0x00400000; break; // 4   MB
			case 0x2: dramSize = 0x00800000; break; // 8   MB
			case 0x3: dramSize = 0x01000000; break; // 16  MB
			case 0x4: dramSize = 0x02000000; break; // 32  MB
			case 0x5: dramSize = 0x04000000; break; // 64  MB
			case 0x6: dramSize = 0x08000000; break; // 128 MB
			case 0x7: dramSize = 0x10000000; break; // 256 MB
			default:  dramSize = merlot_res_ram.end - merlot_res_ram.start; break;
		}
                break;
            default:
                dramSize = merlot_res_ram.end - merlot_res_ram.start;
                break;
            }
            break;
        case MACH_VENDOR_IDT:
            dramSize = merlot_res_ram.end - merlot_res_ram.start;
            break;
	}
	
#if ARUBA_DEBUG
	printk("DRAM SIZE = %d MB\n", (dramSize>>20));
#endif
  
	add_memory_region(0, dramSize, BOOT_MEM_RAM);

#ifdef CONFIG_REMOTE_DEBUG
	if(strstr(arcs_cmdline, "kgdb=on")) {
		remote_debug = 1;
	}
#endif

	return 0;
}

void prom_free_prom_memory(void)
{
#ifdef ARUBA_DEBUG
	printk("stubbed prom_free_prom_memory()\n");
#endif
}
