/*
 * reset.c 
 *
 *             Who                                      When                  What
 *
 * P. Sadik (Pallathu.Sadik@idt.com)                  July2803            Original form
 */
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/mipsregs.h>
#include <asm/pgtable.h>
#include <asm/reboot.h>
#include <asm/addrspace.h>     /* for KSEG1ADDR() */
#include <asm/aruba/merlot.h>
#include <asm/bootinfo.h>

/*--------------------------------------------------------------------------*/
/* For "reset" functionality. This does not belong here. 
/* From Atheros reference port. Move to appropriate place. Sigh. IFS.
/*--------------------------------------------------------------------------*/
#define AR5315_DSLBASE          0xB1000000      /* RESET CONTROL MMR - Atheros */
/* #define AR5315_DSLBASE          0x11000000      /* RESET CONTROL MMR */

#define AR5315_GPIO_DO          (AR5315_DSLBASE + 0x0090)
#define AR5315_GPIO_DIR         (AR5315_DSLBASE + 0x0098)

#define AR5315_RESET_GPIO       5

typedef unsigned int AR531X_REG;

#define PHYS_TO_K1(physaddr)    KSEG1ADDR(physaddr)

#define sysRegRead(phys)	\
	(*(volatile AR531X_REG *)PHYS_TO_K1(phys))

#define sysRegWrite(phys, val)	\
	((*(volatile AR531X_REG *)PHYS_TO_K1(phys)) = (val))
/*--------------------------------------------------------------------------*/

extern unsigned long int mips_machtype;

void idt_reset(void)
{

  switch (mips_chip_vendor) {
  case MACH_VENDOR_ATHEROS:
  default:
    set_cp0_status((ST0_BEV | ST0_ERL));
    set_cp0_config(CONF_CM_UNCACHED);
    flush_cache_all();
    write_32bit_cp0_register(CP0_WIRED, 0);

    switch (mips_chip) {
    case MACH_ATHEROS_AR2313:
    case MACH_ATHEROS_AR5312:
    default:
      /* Reset*/
      *((volatile u32 *)KSEG1ADDR(0x1c003020)) = 0x00080350; // reset everything in sight
      udelay(100);
      *((volatile u32 *)KSEG1ADDR(0x1c003020)) = 0; // reset everything in sight
      udelay(100);
      *((volatile u32 *)KSEG1ADDR(0x1c003020)) = 0x3; // cold reset the cpu & system
      break;
    case MACH_ATHEROS_AR2315:
        /* From Atheros reference port - IFS. */
        for(;;) 
        {
            /* 
            ** Cold reset does not work,work around is to use the GPIO reset bit.  
            */
            unsigned int reg;

            /* Make sure GPIO5 is set as output in the direction register. */
            reg = sysRegRead(AR5315_GPIO_DIR);
            reg |= (1 << AR5315_RESET_GPIO);
            sysRegWrite(AR5315_GPIO_DIR, reg);
            (void)sysRegRead(AR5315_GPIO_DIR); /* flush write to hardware */


            reg = sysRegRead(AR5315_GPIO_DO);
            reg &= ~(1 << AR5315_RESET_GPIO);
            sysRegWrite(AR5315_GPIO_DO, reg);
            (void)sysRegRead(AR5315_GPIO_DO); /* flush write to hardware */
            
        }
        /* Reset*/
/*--------------------------------------------------------------------------*/
/*       *((volatile u32 *)KSEG1ADDR(0x11000000)) = 0x6; */
/*       udelay(100);
/*--------------------------------------------------------------------------*/
      break;
    }
    break;

  case MACH_VENDOR_IDT: {
    int *ptr;
    /* warm reset */
    ptr = (int*)0xb8008000;
    *ptr = 0x80000001;
    break;
   }
  }
}


EXPORT_SYMBOL(idt_reset);

