/*
 * setup.c - boot time setup code for IDT 79S355/79EB355 (Banyan) board
 *
 * Copyright (C) 2000 by Lineo, Inc.
 * Written by Quinn Jensen (jensenq@lineo.com)
 * Ammended by Gray Girling (GGirling@uk.research.att.com) from IDT 79S334
 * (can still be used with 79S334 board)
 *
 */
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/blk.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <linux/ioport.h>
#include <asm/mipsregs.h>
#include <asm/pgtable.h>
#include <linux/mc146818rtc.h>	/* for rtc_ops, we fake the RTC */
#include <asm/reboot.h>
#include <asm/addrspace.h>     /* for KSEG1ADDR() */
#include <asm/aruba/merlot.h>
// #include <asm/rc32434/rc32434.h>
// #include <asm/rc32434/pci.h> 
// #include <asm/rc32434/pcikorina.h> 

extern char * __init prom_getcmdline(void);

extern void (*board_time_init)(void);
extern void (*board_timer_setup)(struct irqaction *irq);
extern void merlot_time_init(void);
extern void merlot_timer_setup(struct irqaction *irq);
extern void idt_reset(void);
void idt_disp_str(char *s);

#ifdef CONFIG_BLK_DEV_INITRD
extern unsigned long initrd_start, initrd_end;
extern void * __rd_start, * __rd_end;
#endif

#ifdef CONFIG_PCI
extern void __init rc32434_pcibridge_init(void);
#endif
extern struct resource rc32434_res_pci_mem1;
extern struct resource rc32434_res_pci_mem2;
extern struct resource rc32434_res_pci_mem3;
extern struct resource rc32434_res_pci_io1;

#define PCI_MEM1_START PCI_ADDR_START
#define PCI_MEM1_END   PCI_ADDR_START + CPUTOPCI_MEM_WIN - 1
#define PCI_MEM2_START PCI_ADDR_START + CPUTOPCI_MEM_WIN 
#define PCI_MEM2_END   PCI_ADDR_START + ( 2* CPUTOPCI_MEM_WIN)  - 1
#define PCI_IO1_START   PCI_ADDR_START + (2 * CPUTOPCI_MEM_WIN) 
#define PCI_IO1_END     PCI_ADDR_START + (2* CPUTOPCI_MEM_WIN) + CPUTOPCI_IO_WIN -1
#define PCI_IO2_START   PCI_ADDR_START + (2 * CPUTOPCI_MEM_WIN) + CPUTOPCI_IO_WIN
#define PCI_IO2_END     PCI_ADDR_START + (2* CPUTOPCI_MEM_WIN) + (2 * CPUTOPCI_IO_WIN) -1

#define DIG_CLEAR ((volatile unsigned char *)0xB9000004)
#define DIG0 ((volatile unsigned char *)0xB9000007)
#define DIG1 ((volatile unsigned char *)0xB9000006)
#define DIG2 ((volatile unsigned char *)0xB9000005)
#define DIG3 ((volatile unsigned char *)0xB9000004)
#define epldMask ((volatile unsigned char *)0xB900000d)

void idt_disp_char(int i, char c)
{
  switch (mips_chip_vendor) {
  case MACH_VENDOR_IDT:
    switch(i) {
    case 0: *DIG0 = c; break;
    case 1: *DIG1 = c; break;
    case 2: *DIG2 = c; break;
    case 3: *DIG3 = c; break;
    default: *DIG0 = '?'; break;
    }
    break;
  default:
    printk("%c", c);
    break;
  }
}

void idt_disp_str(char *s)
{
  switch (mips_chip_vendor) {
  case MACH_VENDOR_IDT: {
    char c;
    if(s == 0) {
      c = *DIG_CLEAR;
    } else {
      int i;
      for(i = 0; i < 4; i++) {
        if(s[i]) idt_disp_char(i, s[i]);
      }
    }
    break;
  }
  default:
    printk("%s\n",s);
    break;
  }
}

static void idt_machine_restart(char *command)
{
	idt_reset();
}

static void idt_machine_halt(void)
{
	for(;;) continue;
}

static void idt_machine_power_off(void)
{
	for(;;) continue;
}

void __init aruba_setup(void)
{
	extern int panic_timeout;
	unsigned int pciCntlVal;
        extern void aruba_panic_init();

	/* reboot on panic in 1 second */
	panic_timeout=1;

#ifdef ARUBA_DEBUG
	idt_disp_str("Unix");
#endif

	board_time_init = merlot_time_init;

	board_timer_setup = merlot_timer_setup;

	_machine_restart = idt_machine_restart;
	_machine_halt = idt_machine_halt;
	_machine_power_off = idt_machine_power_off;
	set_io_port_base(KSEG1);

	write_32bit_cp0_register(CP0_WIRED, 0);

	if ((mips_machtype == MACH_ARUBA_MUSCAT) ||
            (mips_machtype == MACH_ARUBA_DELFRESCO) ||
	    (mips_chip == MACH_IDT_RC32434)) {
		ioport_resource.start = rc32434_res_pci_io1.start;
		ioport_resource.end = rc32434_res_pci_io1.end;
		iomem_resource.start = rc32434_res_pci_mem1.start;
		iomem_resource.end = rc32434_res_pci_mem2.end;

		pciCntlVal=rc32434_pci->pcic;
		pciCntlVal &= 0xFFFFFF7;
		rc32434_pci->pcic = pciCntlVal;
#ifdef CONFIG_PCI
		/* Enable PCI interrupts in EPLD Mask register */	
		*epldMask = 0x0;		
		rc32434_pcibridge_init();
#endif
	}

#ifdef CONFIG_BLK_DEV_INITRD
	ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
	initrd_start = (unsigned long)&__rd_start;
	initrd_end = (unsigned long)&__rd_end;
#endif

        aruba_panic_init();

}

int page_is_ram(unsigned long pagenr)
{
	return 1;
}

const char *get_system_type(void)
{
  switch (mips_machtype) {
  case MACH_ARUBA_MERLOT:
    return "ARUBA MERLOT";
  case MACH_ARUBA_MUSCAT:
    return "ARUBA MUSCAT";
  case MACH_ARUBA_MATARO:
    return "ARUBA MATARO";
  case MACH_ARUBA_SAMSUNG:
    return "ARUBA SAMSUNG";
  case MACH_ARUBA_ALFRESCO:
    return "ARUBA ALFRESCO";
  case MACH_NETGEAR_WG102:
    return "NETGEAR WG102";
  case MACH_ARUBA_ROCKBOTTOM:
    return "ARUBA ROCKBOTTOM";
  case MACH_NESPOT_MW1700:
    return "NESPOT MW1700";
  case MACH_ARUBA_REALLYROCKBOTTOM:
    return "ARUBA REALLYROCKBOTTOM";
  case MACH_ORTRONICS_WJD:
    return "ORTRONICS WI-JACK DUO";
  case MACH_ORTRONICS_DUO:
    return "ORTRONICS DUO";
  case MACH_ARUBA_DELFRESCO:
    return "ARUBA DELFRESCO";
  case MACH_NETGEAR_WAG102:
    return "NETGEAR WAG102";
  default:
    return "ARUBA UNKNOWN";
  }
}














