/**************************************************************************
 * csu_idt.S
 *
 * Board initialization.
 **************************************************************************
 * Ported from IDT/Sim
 *
 *   28 Oct 2002
 * P. Sadik    Pallathu.Sadik@idt.com
 **************************************************************************/

#include <linux/config.h>
#include <linux/threads.h>

#include <asm/asm.h>
#include <asm/mipsregs.h>
#include <asm/offset.h>
#include <asm/cachectl.h>
#include "iregdef.h"
#include "idtcpu.h"
#include "idthdr.h"
	
#include "mhz.h"
#define  MHZ IDT_BUS_FREQ
	
#include "s365led.h"	
#include "s365ram.h"

#define IndexInvalidate_I       0x00
	
.extern _binary_zImage_bin_start,4
.extern _binary_zImage_bin_end,4

/*--------------------------------------------------------------
* prom entry point table
*-------------------------------------------------------------*/

FRAME(start,sp,0,ra)

	j	idtstart  /* begin monitor from start |00| */
	
idtstart:

	.set noreorder
	
	mtc0	zero, C0_CAUSE
	li	v0, 0x0
	or	v0, (SR_CU0 | SR_BEV | SR_DE)
	mtc0	v0, C0_SR

	mfc0	v1, C0_CONFIG
	li	v0, CFG_C_NCHRNT_WB  /* CFG_C_NCHRNT_WB  CFG_C_NCHRNT_WT_NWA CFG_C_UNCACHED */
	and	v1, ~(0x7)
	or	v1, v1, v0
	mtc0	v1, C0_CONFIG
	nop
	nop
	li	t0, ERR_CNTL_STATUS  
	sw	zero, 0(t0)

/* ------------------- Setup Device Controller ---------------------------- */
	li	t0, DEV_CTL_BASE      	/* load 2 base address registers' base    */

	lui	t2, 0xB800
	lw	t1, 0x8004(t2)        	/* get BCV                                */
	li	t2, 0x60              	/* check width of boot device 16/32 bit   */
	and	t1, t1, t2
	li	t2, 0x20
	bne	t1, t2, 1f
	nop
/* 16 bit device - boot from PROM - CS1 is FLASH                      	*/
	li	t1, DEV_PROM_CTRL     	/* device0 control parameter              */
	sw	t1, 0x8(t0)           	/* set the control register  CS0          */
	li	t1, DEV_PROM_TC       	/* device0 timing config parameter        */
	sw	t1, 0xC(t0)

	li	t1, DEV1_BASE         	/* set the device base register for CS1   */
	sw	t1, 0x10(t0)
	li	t1, DEV_FLASH_MASK    	/* set the device mask register for CS1   */
	sw	t1, 0x14(t0) 
	li	t1, DEV_FLASH_CTRL    	/* set the device control register for CS1*/
	sw	t1, 0x18(t0)
	li	t1, DEV_FLASH_TC      	/* set the device timing register for CS1 */
	sw	t1, 0x1C(t0) 
	b	2f                    
        nop
1:
/* 32 bit device - boot from FLASH - CS1 is PROM                      	*/
	li	t1, DEV_FLASH_CTRL    	/* device0 control parameter              */
	sw	t1, 0x8(t0)           	/* set the control register  CS0          */
	li	t1, DEV_FLASH_TC      	/* device0 timing config parameter        */
	sw	t1, 0xC(t0)
	li	t1, DEV1_BASE		/* set the device base register for CS1   */
	sw	t1, 0x10(t0)
	li	t1, DEV_PROM_MASK     	/* set the device mask register for CS1   */
	sw	t1, 0x14(t0) 
	li	t1, DEV_PROM_CTRL     	/* set the device control register for CS1*/
	sw	t1, 0x18(t0)
	li	t1, DEV_PROM_TC       	/* set the device timing register for CS1 */
	sw	t1, 0x1C(t0) 
2:
	li	t1, DEV2_BASE         	/* set the device base register for CS2   */
	sw	t1, 0x20(t0)
	li	t1, DEV2_MASK         	/* set the device mask register for CS2   */
	sw	t1, 0x24(t0) 
	li	t1, DEV2_CTRL         	/* set the device control register for CS2*/
	sw	t1, 0x28(t0) 
	li	t1, DEV2_TC           	/* set the device timing register for CS2 */
	sw	t1, 0x2C(t0) 

	li	t1, DEV3_BASE	        /* set the device base register for CS3   */
	sw	t1, 0x30(t0)
	li	t1, DEV3_MASK         	/* set the device mask register for CS3   */
	sw	t1, 0x34(t0) 
	li	t1, DEV3_CTRL         	/* set the device control register for CS3*/
	sw	t1, 0x38(t0)
	li	t1, DEV3_TC           	/* set the device timing register for CS3 */
	sw	t1, 0x3C(t0) 

	li	t1, DEV4_BASE         	/* set the device base register for CS4   */
	sw	t1, 0x40(t0)
	li	t1, DEV4_MASK         	/* set the device mask register for CS4   */
	sw	t1, 0x44(t0) 
	li	t1, DEV4_CTRL         	/* set the device control register for CS4*/
	sw	t1, 0x48(t0)
	li	t1, DEV4_TC           	/* set the device timing register for CS4 */
	sw	t1, 0x4C(t0) 

	li	t1, DEV5_BASE         	/* set the device base register for CS5   */
	sw	t1, 0x50(t0)
	li	t1, DEV5_MASK         	/* set the device mask register for CS5   */
	sw	t1, 0x54(t0) 
	li	t1, DEV5_CTRL         	/* set the device control register for CS5*/
	sw	t1, 0x58(t0)
	li	t1, DEV5_TC		/* set the device timing register for CS5 */
	sw	t1, 0x5C(t0)

/* ------------------ Initialize GPIO ------------------------------------- */
	li	t0, GPIO_BASE
	li	t1, GPIO_FUNC
	sw	t1, 0x0(t0)
  	li	t1, GPIO_CFG
	sw	t1, 0x4(t0)
  	li	t1, GPIO_OUT
	sw	t1, 0x8(t0)

/* ------------------- Assert PCI reset ----------------------------------- */
	li	t0, 0xb8068000
	lw	t1, (t0)
	ori	t2, t1, 0x1
	sw	t2, (t0)
        nop
        nop
	lui	t3, 0x2
rip:
	lw	t1, 4(t0)
	and	t1, t1, t3
	bnez	t1, rip
        nop

	DISPLAY('I','S','D','R')

	li	t0, SDRAM_BASE 		/* load t0 with internal registers' base address */

/*-------------- Initialize SDRAM Base and Mask Registers ----------*/

	li	t1, SDRAM_BNK0_BASE	/* load SDRAM bank 0 physical address in t1 */
	sw	t1, 0x0(t0)		/* set SDRAM bank 0 base */
	li	t1, SDRAM_BNK1_BASE	/* load SDRAM bank 1 physical address in t1 */
	sw	t1, 0x8(t0)		/* set SDRAM bank 1 base */
	li	t1, SDRAM_BNK0_ABASE	/* load Alternate SDRAM bank 0 physical address in t1 */
	sw	t1, 0x20(t0)		/* set Alternate SDRAM bank 0 base */

	li	t1, SDRAM_BNK0_MASK	/* load SDRAM bank 0 size in t1 */
	sw	t1, 0x4(t0)		/* set SDRAM bank 0 mask */
	li	t1, SDRAM_BNK1_MASK	/* load SDRAM bank 1 size in t1 */
	sw	t1, 0xC(t0)		/* set SDRAM bank 1 mask */
	li	t1, SDRAM_BNK0_AMASK	/* load Alternate SDRAM bank 0 size in t1 */
	sw	t1, 0x24(t0)		/* set Alternate SDRAM bank 0 mask */

/*-------------- Setup and Enable Refresh Timer --------------------*/
	li	t1, DISABLE_TIMER 	/* load diable timer bit settings into t1 */
	sw	t1, 0x1C(t0)    		/* disable timer */
        nop

/*-------------- Enable SDRAM Controller ---------------------------*/
	li	t1, SDRAM_CR_BS
	sw	t1, 0x10(t0)
        nop
          
/*-------------- Delay Loop ----------------------------------------*/
	li	v0, 10000 /* 8x256us */
1:  
	bne	v0, zero,1b
	subu	v0, 1   /* BDSLOT  */
        nop
        nop
/*-------------- Setup Precharge Command ---------------------------*/
	li	t2, 2
	li	t3, 0
1:  
	li	t1, SDRAM_PC_VAL
	sw	t1, 0x10(t0)
        nop
        nop
	lw	t7, 0x10(t0)
        nop
	li	t4, DATA_PATTERN
	li	t5, 0xA0000000 | SDRAM_BNK0_BASE
	sw	t4, 0x0(t5)
	addu	t3, 1
	bne	t3, t2, 1b
        nop
/*-------------- Setup Refresh Command -----------------------------*/
	li	t2, 8
	li	t3, 0
1:  
	li	t1, SDRAM_RFRSH_CMD
	sw	t1, 0x10(t0)
	sw	t4, 0x0(t5) 		/* note: t5 not disturbed */
	addu	t3, 1
	bne	t3, t2, 1b
        nop
/*-------------- Setup up to write to Mode Register ----------------*/
	li	t1, SDRAM_MODE_REG
	sw	t1, 0x10(t0)
	addu	t5, 0x80
	sw	t4, 0x0(t5)

/*-------------- Setup and Enable Refresh Timer --------------------*/

	li	t1, DISABLE_TIMER   	/* load diable timer bit settings into t1 */
	sw	t1, 0x1c(t0)        	/* disable timer */
	sw	zero, 0x14(t0)    	/* set SDRAM refresh timer count register bit settings */

	li	t1, SDRAM_RF_CMPR_SE_BS   /* load SDRAM refresh timer compare register bit settings in t1 */
	sw	t1, 0x18(t0)    		/* set SDRAM refresh timer compare register bit settings */

	li	t1, ENABLE_TIMER    	/* load timer enable bit */
	sw	t1, 0x1c(t0)    		/* enable refresh timer */

	li	t1, SDRAM_CR_BS
	sw	t1, 0x10(t0)

  /* the memory system may need some time to start up... */
	li	v0, 10000 		/* 8x256us */
1:  
	bne	v0, zero, 1b
	subu	v0, 1
        nop
        nop

/************************************************************************
** before doing anything
** initialize the section of memory used by cache initialization
** whenever you boot out of ROM or reset-vector
** This assumed to be 1MB.
** --Sugan (11-22-96)
************************************************************************/
	
	li	t0, 0xa0000000
	li	t1, 0xa0100000
1:
	sw	zero, 0x00(t0)
	sw	zero, 0x04(t0)
	sw	zero, 0x08(t0)
	sw	zero, 0x0c(t0)
	addiu	t0, 16
	nop
	blt	t0, t1, 1b
	nop
	nop
	nop
3:
	mfc0	t0, C0_SR
	nop
	nop
	and	t0, ~SR_BEV
	mtc0	t0, C0_SR
	nop
	nop

	DISPLAY('C','O','P','Y')

4:	
/*Copy ROM Into RAM and start execution */
	la      t0,_binary_zImage_bin_start
	la      t1,ZIMAGE_OFFSET
	la      t2,_binary_zImage_bin_end
	addu    t2,4
cpy:
	lw      t3,0(t0)
	sw      t3,0(t1)
	addu    t0,4
	addu    t1,4
	blt     t0,t2,cpy
	nop
	nop

	DISPLAY('U','N','Z','P')

	la      k0, ZIMAGE_OFFSET
	j       k0
	nop
	nop
			
ENDFRAME(start)
