/*
**  s365ram.h - this header files contains common definitions used by
**  Cedar SDRAM configuration.
*/


/*
*******************************************************************************
**  Copyright 2003 Integrated Device Technology, Inc.
**  All Rights Reserved
*******************************************************************************
*/

#ifndef __S365RAM__
#define __S365RAM__
/******************************** D E F I N E S *******************************/

#define MB16	1
#define MB32	2
#define MB64	3
#define MB128	4
#define MB32SO	5

#define DEV_CTL_BASE        PHYS_TO_K1(0x18010000)  /* device controller regs */
#define SDRAM_BASE          PHYS_TO_K1(0x18018000)  /* SDRAM controller regs */
#define GPIO_FUNC           0x0000609f		    /* SDRAM CLK; MEM[25:22]; UART */
#define GPIO_CFG	    0x00000020		    /* GPIO5 ouput for SPI CS */
#define GPIO_OUT	    0x00000020		    /* GPIO5 ouput 1 */
#define ERR_CNTL_STATUS     PHYS_TO_K1(0x18028030)

#define DEV1_BASE           0x08000000
#define DEV_PROM_MASK       0xFFE00000
#define DEV_PROM_CTRL       0x028A2205
#define DEV_PROM_TC         0x00000A44
#define DEV_FLASH_MASK      0xFF800000
#define DEV_FLASH_CTRL      0x028A2206
#define DEV_FLASH_TC        0x00000A44

#define DEV2_BASE           0x0C000000
#define DEV2_MASK           0xFF000000
#define DEV2_CTRL           0x04108324              /* 8-bit devices */
#define DEV2_TC             0x00000A44

#define DEV3_BASE           0x00000000
#define DEV3_MASK           0x00000000
#define DEV3_CTRL           0x0FFFFFF4              /* ?-bit devices */
#define DEV3_TC             0x00001FFF

#define DEV4_BASE           0x00000000
#define DEV4_MASK           0x00000000
#define DEV4_CTRL           0x0FFFFFF4              /* ?-bit devices */
#define DEV4_TC             0x00001FFF

#define DEV5_BASE           0x00000000
#define DEV5_MASK           0x00000000
#define DEV5_CTRL           0x0FFFFFF4              /* ?-bit devices */
#define DEV5_TC             0x00001FFF

#define DATA_PATTERN        0xA5A5A5A5

#if DRAMSZ == MB32SO

/* Address space allocations */
#define SDRAM_BNK0_BASE     0x00000000  /* 0 MB */
#define SDRAM_BNK0_MASK     0xFE000000
#define SDRAM_BNK1_BASE     0x02000000  /* 32 MB */
#define SDRAM_BNK1_MASK     0x00000000
#define SDRAM_BNK0_ABASE    0x00000000  /* 0 MB */
#define SDRAM_BNK0_AMASK    0x00000000

/* SDRAM Chip specific setup */
#if MHZ == 90
#define SDRAM_CR_BS 	    0x994330F0 /* SDRAM enable 	*/
#define SDRAM_DS_BS 	    0x19433080 /* SDRAM disable */
#define SDRAM_PC_VAL  	    0x194330c3 /* Precharge val */
#define SDRAM_RFRSH_CMD     0x19433093 /* refresh cmd 	*/
#define SDRAM_MODE_REG      0x19433083 /* mode register */
#else
#define SDRAM_CR_BS 	    0x984330F0 /* SDRAM enable 	*/
#define SDRAM_DS_BS 	    0x18433080 /* SDRAM disable */
#define SDRAM_PC_VAL  	    0x184330c3 /* Precharge val */
#define SDRAM_RFRSH_CMD     0x18433093 /* refresh cmd 	*/
#define SDRAM_MODE_REG      0x18433083 /* mode register */
#endif

#else

/* Address space allocations */
#define SDRAM_BNK0_BASE     0x00000000  /* 0 MB */
#define SDRAM_BNK0_MASK     0xFF000000
#define SDRAM_BNK1_BASE     0x01000000  /* 16 MB */
#define SDRAM_BNK1_MASK     0x00000000
#define SDRAM_BNK0_ABASE    0x00000000  /* 0 MB */
#define SDRAM_BNK0_AMASK    0x00000000

/* SDRAM Chip specific setup */
#if MHZ == 90
#define SDRAM_CR_BS 	    0x894370F0 /* SDRAM enable 	*/
#define SDRAM_DS_BS 	    0x09437080 /* SDRAM disable */
#define SDRAM_PC_VAL  	    0x094370c3 /* Precharge val */
#define SDRAM_RFRSH_CMD     0x09437093 /* refresh cmd 	*/
#define SDRAM_MODE_REG      0x09437083 /* mode register */
#else
#define SDRAM_CR_BS 	    0x884370F0 /* SDRAM enable 	*/
#define SDRAM_DS_BS 	    0x08437080 /* SDRAM disable */
#define SDRAM_PC_VAL  	    0x084370c3 /* Precharge val */
#define SDRAM_RFRSH_CMD     0x08437093 /* refresh cmd 	*/
#define SDRAM_MODE_REG      0x08437083 /* mode register */
#endif

#endif

/* Refresh timer */
#define SDRAM_RF_CNT_BS     0x00000000 /* refresh count reg */
#define SDRAM_RF_CMPR_BS    0x00000040 /* refresh comp reg  */
#if MHZ == 90
#define SDRAM_RF_CMPR_SE_BS 0x000002EE /* slow expiration   */
#else
#define SDRAM_RF_CMPR_SE_BS 0x00000271 /* slow expiration   */
#endif

#define DELAY_200USEC       30000      /* not exactly */
#define DISABLE_TIMER       0
#define ENABLE_TIMER        1

#endif
