/*
 * BRIEF MODULE DESCRIPTION
 *	IDT 79EB365 board setup.
 *
 * Copyright 2002 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *         	stevel@mvista.com or source@mvista.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/init.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/irq.h>
#include <linux/ioport.h>
#include <asm/bootinfo.h>
#include <asm/io.h>
#include <asm/mipsregs.h>
#include <asm/pgtable.h>
#include <asm/reboot.h>
#include <asm/rc32300/rc32300.h>
#include <asm/rc32300/pci.h>
#include <asm/rc32300/pcicedar.h>

extern void (*board_time_init)(void);
extern void (*board_timer_setup)(struct irqaction *irq);
extern void rc32300_time_init(void);
extern void rc32300_timer_setup(struct irqaction *irq);
extern char * __init prom_getcmdline(void);

extern void rc32300_restart(char *);
extern void rc32300_halt(void);
extern void rc32300_power_off(void);
#ifdef CONFIG_PCI
extern void __init rc32365_pcibridge_init(void);
#endif
struct resource rc32365_res_pci_mem1;
struct resource rc32365_res_pci_mem2;
struct resource rc32365_res_pci_mem3;

#define PCI_MEM1_START PCI_ADDR_START
#define PCI_MEM1_END   PCI_ADDR_START + CPUTOPCI_MEM_WIN - 1
#define PCI_MEM2_START PCI_ADDR_START + CPUTOPCI_MEM_WIN 
#define PCI_MEM2_END   PCI_ADDR_START + ( 2* CPUTOPCI_MEM_WIN)  - 1
#define PCI_IO1_START   PCI_ADDR_START + (2 * CPUTOPCI_MEM_WIN) 
#define PCI_IO1_END     PCI_ADDR_START + (2* CPUTOPCI_MEM_WIN) + CPUTOPCI_IO_WIN -1
#define PCI_IO2_START   PCI_ADDR_START + (2 * CPUTOPCI_MEM_WIN) + CPUTOPCI_IO_WIN
#define PCI_IO2_END     PCI_ADDR_START + (2* CPUTOPCI_MEM_WIN) + (2 * CPUTOPCI_IO_WIN) -1


struct resource rc32365_res_pci_mem1 = {
  "PCI Mem1",
  0x50000000,
  0x5FFFFFFF,
  IORESOURCE_MEM,
  &rc32365_res_pci_mem1,
  NULL,
  &rc32365_res_pci_mem2
};
struct resource rc32365_res_pci_mem2 = {
  "PCI Mem2",
  0x60000000,
  0x6FFFFFFF,
  IORESOURCE_MEM,
  &rc32365_res_pci_mem1,
  NULL,
  NULL
};
struct resource rc32365_res_pci_mem3 = {
  "PCI Mem3",
  0x18C00000,
  0x18FFFFFF,
  IORESOURCE_MEM,
  &rc32365_res_pci_mem1,
  NULL,
  NULL
};

struct resource rc32365_res_pci_io1 = {
  "PCI I/O1",
  0x18800000,
  0x188FFFFF,
  IORESOURCE_IO
};

#ifdef CONFIG_BLK_DEV_INITRD
extern unsigned long initrd_start, initrd_end;
extern void * __rd_start, * __rd_end;
#endif

#if 0
static void dump_dev(int devnum)
{
  printk("DEV%d_BASE   = %08x\n", devnum,
	 rc32300_readl(DEV0_BASE + devnum*DEV_REG_OFFSET));
  printk("DEV%d_MASK   = %08x\n", devnum,
	 rc32300_readl(DEV0_MASK + devnum*DEV_REG_OFFSET));
  printk("DEV%d_CNTL   = %08x\n", devnum,
	 rc32300_readl(DEV0_CNTL + devnum*DEV_REG_OFFSET));
  printk("DEV%d_TIMING = %08x\n", devnum,
	 rc32300_readl(DEV0_TIMING + devnum*DEV_REG_OFFSET));
}
#endif

#define DIG0 ((volatile unsigned char *)0xAC000003)
#define DIG1 ((volatile unsigned char *)0xAC000002)
#define DIG2 ((volatile unsigned char *)0xAC000001)
#define DIG3 ((volatile unsigned char *)0xAC000000)

void idt_disp_char(int i, char c)
{
  switch (i) {
  case 0: *DIG0 = c; break;
  case 1: *DIG1 = c; break;
  case 2: *DIG2 = c; break;
  case 3: *DIG3 = c; break;
  default: *DIG0 = '?'; break;
  }
}

void idt_disp_str(char *s)
{
  int i;
  
  if (s == 0) {
    idt_disp_char(0, 0x20);
    idt_disp_char(1, 0x20);
    idt_disp_char(2, 0x20);
    idt_disp_char(3, 0x20);
  } 
  else {
    for (i = 0; i < 4; i++) {
      if (s[i]) 
	idt_disp_char(i, s[i]);
    }
  }
}

void __init idt_setup(void)
{
  unsigned int pciCntlVal;
  char* argptr;
  idt_disp_str("Unix");
	
  argptr = prom_getcmdline();
#ifdef CONFIG_SERIAL_CONSOLE
  if ((argptr = strstr(argptr, "console=")) == NULL) {
    argptr = prom_getcmdline();
    strcat(argptr, " console=ttyS0,9600");
  }
#endif

  board_time_init = rc32300_time_init;
  board_timer_setup = rc32300_timer_setup;

  _machine_restart = rc32300_restart;
  _machine_halt = rc32300_halt;
  _machine_power_off = rc32300_power_off;

  set_io_port_base(KSEG1);

  // clear out any wired entries
  write_32bit_cp0_register(CP0_WIRED, 0);

  /* Disable Watchdog timer */
  rc32300_writel(0, 0xb802803c);

  ioport_resource.start = rc32365_res_pci_io1.start;
  ioport_resource.end = rc32365_res_pci_io1.end;
  iomem_resource.start = rc32365_res_pci_mem1.start;
  iomem_resource.end = rc32365_res_pci_mem2.end;
  /* Disable the IP bus error for PCI scaning */
  pciCntlVal=rc32365_pci->pcic;
  pciCntlVal &= 0xFFFFFF7;
  rc32365_pci->pcic = pciCntlVal;
#ifdef CONFIG_PCI
  rc32365_pcibridge_init();
#endif
  
#ifdef CONFIG_BLK_DEV_INITRD
  ROOT_DEV = MKDEV(RAMDISK_MAJOR, 0);
  initrd_start = (unsigned long)&__rd_start;
  initrd_end = (unsigned long)&__rd_end;
#endif
}

int page_is_ram(unsigned long pagenr)
{
  return 1;
}
