/*
 * Carsten Langgaard, carstenl@mips.com
 * Copyright (C) 1999,2000 MIPS Technologies, Inc.  All rights reserved.
 *
 * ########################################################################
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * ########################################################################
 *
 * Setting up the clock on the MIPS boards.
 *
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/mc146818rtc.h>
#include <linux/irq.h>
#include <linux/timex.h>

#include <asm/mipsregs.h>
#include <asm/ptrace.h>
#include <asm/debug.h>
#include <asm/rc32438/rc32438.h>
#include  <asm/rc32438/tim.h>
#include  <asm/rc32438/int.h>
static unsigned long r4k_offset; /* Amount to incr compare reg each time */
static unsigned long r4k_cur;    /* What counter should be at next timer irq */
extern void ll_timer_interrupt(int irq, struct pt_regs *regs);
extern unsigned int mips_counter_frequency;

#if defined(CONFIG_IDT_79EB438) && defined(CONFIG_MIPS_RTC)
extern void rtc_ds1553_init(void);
#endif

/* 
 * Figure out the r4k offset, the amount to increment the compare
 * register for each time tick. There is no RTC available.
 *
 * The RC32438 counts at half the CPU *core* speed.
 */
static unsigned long __init cal_r4koff(void)
{
	mips_counter_frequency = IDT_CLOCK_MULT * IDT_BUS_FREQ * 1000000 / 2;
	return (mips_counter_frequency / HZ);
}


void __init rc32438_time_init(void)
{
        unsigned int est_freq, flags;

	__save_and_cli(flags);

	printk("calculating r4koff... ");
	r4k_offset = cal_r4koff();
	printk("%08lx(%d)\n", r4k_offset, (int) r4k_offset);

	est_freq = 2*r4k_offset*HZ;	
	est_freq += 5000;    /* round */
	est_freq -= est_freq%10000;
	printk("CPU frequency %d.%02d MHz\n", est_freq/1000000, 
	       (est_freq%1000000)*100/1000000);
	__restore_flags(flags);


#if defined(CONFIG_IDT_79EB438) && defined(CONFIG_MIPS_RTC)
	rtc_ds1553_init();
#endif


}


void __init rc32438_timer_setup(struct irqaction *irq)
{
        /* we are using the cpu counter for timer interrupts */
#if 0
	setup_irq(MIPS_CPU_TIMER_IRQ, irq);
#endif
	setup_irq(8, irq);

#if 0
        interrupt->i[IPEND2].imask = 0xfffffffe;
        interrupt->i[IPEND5].imask = 0xfffffff6;
#endif


	timer->tim[0].compare = r4k_offset;
 	timer->tim[0].ctc = 0x1;
#if 0
	request_irq(8, ll_timer_interrupt, SA_INTERRUPT,
		    "RC32438 TIMER INTR", NULL);
#endif

#if 0
        /* to generate the first timer interrupt */
	r4k_cur = (read_32bit_cp0_register(CP0_COUNT) + r4k_offset);
	write_32bit_cp0_register(CP0_COMPARE, r4k_cur);
	/* this is the last board dependent code */
#endif
	db_run(board_init_done_flag = 1);

}




