#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/atmdev.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/capability.h>
#include <asm/system.h>
#include <asm/param.h>
#include <asm/uaccess.h>

#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>


#include "./h/ctrle.h"
#include "alcatel.c"

/*ame specific functions */
extern int ame_close(void);
extern void cs2_init(void);
extern int ame_get_opmode(void);
extern int adsl_ldrun(unsigned char *, int, unsigned char *, int);
extern int  ame_resp_nw( void );
extern int ame_monitor( unsigned char *, unsigned char * );
extern unsigned int rsctrle( char *, unsigned int, unsigned int ) ;
extern int  adsl_openco( unsigned char, unsigned char ) ;
extern const char  ADSL_CONNECTION_OPENED;
extern const unsigned int  CTEAD_MODEM_STATUS;
extern const unsigned int  C_CONF_DATA_ACK;

static unsigned int modemStatus = 0;
static int idtame_start(struct atm_dev *dev);
static void idtame_train(unsigned long);
/* For timer */
#define AME_TRAINING_TIMER_PERIOD 1 //secs
static const struct atmphy_ops idtame_ops = {
  idtame_start,
  NULL,
  NULL
};
static struct timer_list training_timer = {
  function:   &idtame_train
};
static int start_timer = 1;

int __init idtame_init(struct atm_dev *dev)
{
  MOD_INC_USE_COUNT;

  dev->phy = &idtame_ops;
  return 0;
}

static void idtame_train(unsigned long dummy)
{
  static unsigned char fstadr = 28, intadr = 29;
  static unsigned int adslOpencoStatus = 0;
  unsigned char xstatus[8];
  unsigned int  retval;
  static int trainingCount = 0;
  unsigned int flags;

  /* keep the line opened with a modem on the other side */
  /* open the modem connection */
  if(adslOpencoStatus == 0)
    {
      if( !adsl_openco(fstadr,intadr) ) 
	{
	  ame_close();
	  save_and_cli(flags);
	  if (!start_timer) mod_timer(&training_timer,jiffies+10*AME_TRAINING_TIMER_PERIOD*HZ);
	  restore_flags(flags);
	  return;
	}else
	  {
	    if( (fstadr == 28) && (intadr == 29))
	      printk("Operating in FAST mode.\n");
	    else 
	      if( (fstadr == 29) && (intadr == 28))
		printk("Operating in INTERLEAVE mode.\n");
	    adslOpencoStatus = 1;
	  }
    }
  
  if(modemStatus == 0)
    {
      unsigned int  respnd;
      respnd  =  ame_resp_nw();
      /* if no response, loop again */
      if( !respnd ){
	printk("Training... \n");
	save_and_cli(flags);
	if(trainingCount++ > 30){
	  ame_close();
	  adslOpencoStatus = 0;
	  trainingCount = 0;
	}
	if (!start_timer) mod_timer(&training_timer,jiffies+3*AME_TRAINING_TIMER_PERIOD*HZ);
	restore_flags(flags);
	return;
      }
      respnd  &=  0x0FF;
      if( respnd == ADSL_CONNECTION_OPENED )
	printk("SHOWTIME!\n");
    }

  rsctrle( xstatus, CTEAD_MODEM_STATUS, 1 );
  retval  =  xstatus[0];
  if( retval == 0x20 ) 
    {
      modemStatus = 1;
    } else {
      printk("Modem Down!\n");
      modemStatus = 0;
      ame_close();
    }

  if( !ame_monitor(&fstadr,&intadr) )
    {
      printk("Lost Connection.\n");
      ame_close();
      adslOpencoStatus = 0;
      modemStatus = 0;
    }
  save_and_cli(flags);
  if (!start_timer) mod_timer(&training_timer,jiffies+10*AME_TRAINING_TIMER_PERIOD*HZ);
  restore_flags(flags);
}

static int idtame_start(struct atm_dev *dev)
{
  unsigned long flags;
  save_and_cli(flags);

  if (!start_timer) restore_flags(flags);
  else 
    {
      start_timer = 0;
      restore_flags(flags);

      init_timer(&training_timer);
      training_timer.expires = jiffies+AME_TRAINING_TIMER_PERIOD;
      training_timer.function = idtame_train;
      add_timer(&training_timer);
    }
  return 0;
}
int idtame_stop(struct atm_dev *dev)
{
  ame_close();
  dev->phy = NULL;

  MOD_DEC_USE_COUNT;
  return 0;
}

extern struct proc_dir_entry *atm_proc_root;    /* from proc.c */
int idtame_init_module(void)
{
  unsigned int initsize = sizeof(initbin) - 4;
  unsigned int showsize = sizeof(showbin) - 4;
  unsigned char xstatus[8];
  volatile int  i;
  unsigned int  retval;

  struct proc_dir_entry *p;

  if ((p = create_proc_entry("ame", 0, atm_proc_root)) == NULL)
    return -ENOMEM;

  /* initialize the chip-select2 */
  cs2_init();
  for( ; ; ) 
    {
      /* download the binary only if the chipset is not in NORMAL mode */
      /* if failed, try again */
      if( !ame_get_opmode() ) 
	{
	  printk("Download Microcode");
	  if( !adsl_ldrun(initbin, initsize, showbin, showsize) ) 
	    {
	      printk("Failed. Trying again..\n");
	      continue;
	    } 
	  else 
	    {
	      printk("Done!\n");
	    } 
	}
    /* the modem status checking with delay */
      for(i=0;i<2;i++) mdelay(100);
      retval = rsctrle(xstatus, CTEAD_MODEM_STATUS, 1);
      idtame_start(NULL);
      printk("idtame module loaded.\n");
      return 0;
    }
}


void idtame_cleanup_module(void)
{
  ame_close();
}

module_init(idtame_init_module);
module_exit(idtame_cleanup_module);

