#include <linux/interrupt.h>
#include <linux/config.h>

#undef CTRLERW_DEBUG

#define  MTC_MP                 0xAC000000
#define  TXCOMAVREG             0x02
#define  RXCOMAVREG             0x03

unsigned int wrctrle( char *, unsigned int, unsigned int ) ;
unsigned int rdctrle( char *, unsigned int, unsigned int ) ;
unsigned int rsctrle( char *, unsigned int, unsigned int ) ;
unsigned char  ame_rd8bit( unsigned int ) ;
void  ame_wr8bit( unsigned int, unsigned char ) ;
unsigned int  adrswp( unsigned int ) ;
void  cp_m_to_e( unsigned int, unsigned int, int ) ;
void  cp_e_to_m( unsigned int, unsigned int, int ) ;

/*
 *
 */
unsigned int wrctrle( char *pval, unsigned int address, unsigned int number )
{

   unsigned char tx_comv = 0x01;

#ifdef CTRLERW_DEBUG
      printk( "WRCTRLE: PVAL = 0x%X \n\r", (unsigned int)pval );
      printk( "WRCTRLE: ADDRESS = 0x%X \n\r", address );
      printk( "WRCTRLE: NUMBER = 0x%X \n\r", number );
      printk( "WRCTRLE: MTC_MP = 0x%X \n\r", (unsigned int)MTC_MP );
#endif

   /* try repeatedly until TX_COM_AV is reset */
   for( ; tx_comv & 0x01; )
      /* tx_comv = *((unsigned char *)(MTC_MP+TXCOMAVREG)); */
      tx_comv = ame_rd8bit( MTC_MP+TXCOMAVREG );

 /* ame_memcpy( (((unsigned int)MTC_MP)+address),(unsigned int)pval,number ); */
   cp_m_to_e( (((unsigned int)MTC_MP)+address),(unsigned int)pval,number );
   return( number );
}



/*
 *
 */
unsigned int rdctrle( char *pval, unsigned int address, unsigned int number )
{
   unsigned char rx_comv = 0x0;

#ifdef CTRLERW_DEBUG
      printk( "RDCTRLE: PVAL = 0x%X \n\r", (unsigned int)pval );
      printk( "RDCTRLE: ADDRESS = 0x%X \n\r", address );
      printk( "RDCTRLE: NUMBER = 0x%X \n\r", number );
      printk( "RDCTRLE: MTC_MP = 0x%X \n\r", (unsigned int)MTC_MP );
#endif

   /* try repeatedly until RX_COM_AV is set */
   for( ; !(rx_comv & 0x01); )
      /* rx_comv = *((unsigned char *)(MTC_MP+RXCOMAVREG)); */
      rx_comv = ame_rd8bit( MTC_MP+RXCOMAVREG );

 /* ame_memcpy( (unsigned int)pval,(((unsigned int)MTC_MP)+address),number ); */
   cp_e_to_m( (unsigned int)pval,(((unsigned int)MTC_MP)+address),number );

   return( number );
}



/*
 *
 */
unsigned int rsctrle( char *pval, unsigned int address, unsigned int number )
{

   cp_e_to_m( (unsigned int)pval,(((unsigned int)MTC_MP)+address),number );

   return( number );
}



/*
 *
 */
unsigned char  ame_rd8bit( unsigned int adr8bit )
{
   volatile unsigned int  *xadr;
   volatile unsigned int  xret;
   unsigned long key;

   save_and_cli(key);
   xadr  =  (void *)adrswp(adr8bit);
   xret  =  *xadr;
   restore_flags(key);

#ifdef CTRLERW_DEBUG
      printk( "AME_RD8BIT: XADR = 0x%08X \n", (unsigned int)xadr );
      printk( "AME_RD8BIT: XRET = 0x%08X \n", xadr );
#endif

   return( (unsigned char)xret );
}



/*
 *
 */
void  ame_wr8bit( unsigned int adr8bit, unsigned char xdat )
{
   volatile unsigned int  *xadr;
   unsigned long key;

   save_and_cli(key);
   xadr  =  (void *)adrswp(adr8bit);
#ifdef CTRLERW_DEBUG
      printk( "AME_WR8BIT: XADR = 0x%X \n\r", (unsigned int)xadr );
#endif
   *xadr  =  (unsigned int)xdat;
   restore_flags(key);
}



/*
 * encode the address for CTRL-E access on RP35XREV00.
 */
#ifdef  CONFIG_IDT_79RP351  /* RP351 still REV00 */
unsigned int  adrswp( unsigned int adr8bit )
{
   unsigned int  xtmp;

   xtmp  =  adr8bit & 0x00000003;
   xtmp  <<=  12;

   return( (adr8bit&0xFFFFCFFC)|(xtmp&0x00003000) );
}
#else
unsigned int  adrswp( unsigned int adr8bit )
{
   unsigned int  xtmp;
   xtmp  =  adr8bit & 0x0000FFFF;
   xtmp  <<=  2;

   return( (adr8bit&0xFFFF0000)|(xtmp&0x0000FFFF) );
}
#endif


/*
 *
 */
void  cp_m_to_e( unsigned int dadr, unsigned int sadr, int xsize )
{
   int  i;
   volatile unsigned char  tmp;
   volatile unsigned char  *xsadr;
   unsigned long key;

   save_and_cli(key);
   for( i = 0; i < xsize; i++ ) {
#ifdef CTRLERW_DEBUG
         printk( "CP_M_TO_E: DADR = 0x%X, SADR = 0x%X \n\r",
                 dadr, sadr );
#endif
      xsadr  =  (void *)sadr;
      tmp = *xsadr;
      sadr++;
      ame_wr8bit( dadr, tmp );
      dadr++;
   }
   restore_flags(key);
}



/*
 *
 */
void  cp_e_to_m( unsigned int dadr, unsigned int sadr, int xsize )
{
   int  i;
   volatile unsigned char  tmp, *xdadr;
   unsigned long key;

   save_and_cli(key);
   for( i = 0; i < xsize; i++ ) {
#ifdef CTRLERW_DEBUG
         printk( "CP_E_TO_M: DADR = 0x%X, SADR = 0x%X \n\r",
                 dadr, sadr );
#endif
      tmp = ame_rd8bit( sadr );
      sadr++;
      xdadr  =  (void *)dadr;
      *xdadr  =  tmp;
      dadr++;
   }
   restore_flags(key);
}
