#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/string.h> 

#include "./h/ctrle.h"
#include "rldrun.h"


static unsigned int dispcount = 0;

struct T_adsl_defbm25s_set {
   unsigned char  near_end;
   unsigned char  far_end;
};

struct T_adsl_crc_near {
   unsigned char  fast[2];
   unsigned char  inlv[2];
};

struct T_adsl_crc_far {
   unsigned char  fast[2];
   unsigned char  inlv[2];
};

struct T_adsl_chl_data {
   unsigned char  int_near_lsb[2];
   unsigned char  int_near_msb[2];
   unsigned char  fast_near_lsb[2];
   unsigned char  fast_near_msb[2];
   unsigned char  int_far_lsb[2];
   unsigned char  int_far_msb[2];
   unsigned char  fast_far_lsb[2];
   unsigned char  fast_far_msb[2];
};



extern const char  BLCKDATAREG;
extern const char  MODEM_DUMP_BLOCK;
extern const char  TXCOMMANDREG;
extern const char  RXCOMMANDREG;
extern const char  RXCOMAVREG;
extern const char  MODEM_BLOCK_DUMP_ACK;
extern const char  MODEM_LOAD_BLOCK;
extern const char  MODEM_BLOCK_AVAILABLE;
extern const char  BLCKBADDRREG;
extern const char  GET_OPERATIONAL_MODE;
extern const char  ADSL_JUMP_ADDRESS;
extern const char  ADSL_CONNECTION_CLOSED;
extern const char  ADSL_CLOSE_REJECTED;
extern const char  ADSL_OPEN_REJECTED;
extern const char  ADSL_OPEN_FAILED;
extern const char  ADSL_OPEN_FAILED_ON_TIMEOUT;
extern const char  ADSL_CONNECTION_OPENED;
extern const char  ADSL_UNKNOWN_OPCODE;
extern const unsigned int  C_CONF_DATA_COMMPROBLEM;
extern const unsigned int  C_CONF_DATA_ALREADY_DOWN;
extern const unsigned int  C_CONF_DATA_ACK;
extern const unsigned int  C_CONF_DATA_REJ;
extern const unsigned int  CTEAD_MODEM_STATUS;
extern const unsigned int  CTEAD_NEAR_INT_CHANNELDATA_LSB;


extern unsigned char wrctrle( char *, unsigned int, unsigned int ) ;
extern unsigned char rdctrle( char *, unsigned int, unsigned int ) ;
extern unsigned char rsctrle( char *, unsigned int, unsigned int ) ;
extern unsigned int  ADSL_disableModemLine( void ) ;
extern unsigned int  LT_ADSL_enableModemLine_multimode( struct T_adsl_ConfigurationData * ) ;

/* jungo, the next two lines */
extern unsigned int  LT_ADSL_enableModemLine_ansi( struct T_adsl_ConfigurationData * ) ;
extern unsigned int  LT_ADSL_enableModemLine_gdmt( struct T_adsl_ConfigurationData * ) ;


extern int  xputstr( char * ) ;
extern void cs2_init( void ) ;

int  adsl_ldrun( unsigned char *, int, unsigned char *, int ) ;
int  adsl_init( unsigned char *, int ) ;
int  adsl_show( unsigned char *, int ) ;
int  banyan_dump( unsigned char *, unsigned int, unsigned int ) ;
int  banyan_load( unsigned char *, unsigned int, unsigned int ) ;
int  banyan_sndgo( unsigned int ) ;
int  write_addr_len( unsigned int, unsigned int ) ;
int  adsl_turnx80( void ) ;
int  adsl_openco( unsigned char, unsigned char ) ;
int  ame_monitor( unsigned char *, unsigned char * ) ;
int  ame_close( void ) ;
int  ame_defbm25s_poll( struct T_adsl_defbm25s_set * ) ;
int  ame_clear_dbm25s_near( void ) ;
int  ame_clear_dbm25s_far( void ) ;
int  ame_crcnear_poll( struct T_adsl_crc_near * ) ;
int  ame_crcfar_poll( struct T_adsl_crc_far * ) ;
int  ame_crc_monitor( unsigned short int, unsigned short int,
                      unsigned short int, unsigned short int, int ) ;
int  ame_get_opmode( void ) ;
volatile void ame_delay_1s( int num );
int  ame_fast_utopiadr( void ) ;
int  ame_chopdata_poll( struct T_adsl_chl_data * ) ;
int  ame_resp_nw( void );


/*
 *
 */
int  adsl_ldrun( unsigned char *initbin, int initsize,
                 unsigned char *showbin, int showsize )
{

   /* load and run the initial program */
   if( !adsl_init(initbin,initsize) )
      return( 0 );
#ifdef RLDR_DEBUG
      printk( "PASS ADSL_INIT .....\n\r" );
#endif

   /* polling for 0x80 at RX_COMMAND */
   if( !adsl_turnx80() )
      return( 0 );
#ifdef RLDR_DEBUG
      printk( "PASS ADSL_TURNX80 .....\n\r" );
#endif


   if( !adsl_show(showbin,showsize) )
      return( 0 );
#ifdef RLDR_DEBUG
      printk( "PASS ADSL_SHOW .....\n\r" );
#endif

   /* return TRUE if success */
   return( 1 );
}


/*
 *
 */
int  adsl_init( unsigned char *initbin, int initsize )
{
   unsigned char  tmpbin[512];
   int  i, blksize;

   for( i = 0; i < initsize; ) {
      if( (i+128) > initsize )
         blksize  =  initsize - i;
      else
         blksize  =  128;

      if( !(dispcount++ %50))
	printk(".");
#ifdef RLDR_DEBUG
         printk( "ADSL_INIT: BLKSIZE = 0x%X \n\r", blksize );
#endif

      /* return 0 if failed */
      if( !banyan_dump(&(initbin[i]),blksize,i) )
         return( 0 );
      if( !banyan_load(tmpbin,blksize,i) )
         return( 0 );

      /* verify the operation, repeat if failed */
      if( !strncmp(&(initbin[i]),tmpbin,blksize) )
         i += 128;
      else;
   }


   /* jump start the program, return 0 if failed */
   if( !banyan_sndgo( 0 ) )
      return( 0 );

#ifdef RLDR_DEBUG
      printk( "******** ADSL_INIT ********* \n\r" );
      printk( "INITSIZE = 0x%X \n\r", initsize );
      printk( "******** ADSL_INIT ********* \n\r" );
#endif
   
   return( 1 );
}



/*
 *
 */
int  adsl_show( unsigned char *showbin, int showsize )
{
   unsigned char  tmpbin[256];
   int   i, blksize;

   for( i = 0; i < showsize; ) {
      if( (i+128) > showsize )
         blksize  =  showsize - i;
      else
         blksize  =  128;

      if( !(dispcount++ %50))
	printk(".");
      /* return 0 if failed */
      if( !banyan_dump(&(showbin[i]),blksize,i) ) {
#ifdef RLDR_DEBUG
            printk( "ADSL_SHOW: banyan_dump failed... \n" );
#endif
         return( 0 );
      }
      if( !banyan_load(tmpbin,blksize,i) ) {
#ifdef RLDR_DEBUG
            printk( "ADSL_SHOW: banyan_load failed... \n" );
#endif
         return( 0 );
      }
      /* verify the operation, repeat if failed */
      if( !strncmp(&(showbin[i]),tmpbin,blksize) )
         i += 128;
      else;

      if( (i%1280) < 128 ) {
#ifdef RLDR_DEBUG
            printk( "ADSL_SHOW: BASE_ADDR = 0x%X \n\r", i );
#endif
      }
   }

   /* write the sub-function of ATU-C(150 ATU-C) */
   if( ADSL_write_functionality_code_1FD(0x06)==C_CONF_DATA_COMMPROBLEM )
      return( 0 );

   /* jump start the program, return 0 if failed */
   if( !banyan_sndgo( 0 ) )
      return( 0 );

#ifdef RLDR_DEBUG
      printk( "******** ADSL_SHOW ********* \n\r" );
      printk( "SHOWSIZE = 0x%X \n\r", showsize );
      printk( "******** ADSL_SHOW ********* \n\r" );
#endif
   
   return( 1 );
}



/*
 *
 */
int  banyan_dump( unsigned char *dmp, unsigned int blen, 
                  unsigned int baddr )
{
   unsigned char  embox[512];
   int  i;
#ifdef RLDR_DEBUG
      printk( "********* BANYAN_DUMP ************ \n\r" );
      printk( "DMP = 0x%X \n\r", (unsigned int)dmp );
      printk( "BLEN = 0x%X \n\r", blen );
      printk( "BADDR = 0x%X \n\r", baddr );
      printk( "********* BANYAN_DUMP ************ \n\r" );
#endif

   /* write the base address, and length of data on the controlE window */
   /* return 0 if failed */
   if( !write_addr_len( blen, baddr ) )
      return( 0 );

   /* dump the data into E-control window */
   for( i = 0; i < blen; i++ )
      embox[i]  =  dmp[i];
   /* return 0 if failed */
   if( !wrctrle(embox, BLCKDATAREG, blen) )
      return( 0 );

   /* write the Modem_Dump_Block opcode */
   embox[0]  =  (unsigned char)MODEM_DUMP_BLOCK;
   if( !wrctrle(embox, TXCOMMANDREG, 1) )
      return( 0 );

   /* return TRUE if success */
   rdctrle(embox, RXCOMMANDREG, 1);
   if( embox[0] == MODEM_BLOCK_DUMP_ACK ) {
#ifdef RLDR_DEBUG
         printk( "BANYAN_DUMP: RESPONSE = 0x%08X  \n", 
                 (unsigned int)embox[0] );
#endif
      return( 1 );
   } else {
#ifdef RLDR_DEBUG
         printk( "BANYAN_DUMP: RESPONSE = 0x%08X  \n", 
                 (unsigned int)embox[0] );
#endif
      return( 0 );
   }
}



/*
 *
 */
int  banyan_load( unsigned char *dld, unsigned int blen, 
                  unsigned int baddr )
{
   unsigned char  embox[512];

#ifdef RLDR_DEBUG
      printk( "********* BANYAN_LOAD ************ \n\r" );
      printk( "DLD = 0x%X \n\r", (unsigned int)dld );
      printk( "BLEN = 0x%X \n\r", blen );
      printk( "BADDR = 0x%X \n\r", baddr );
      printk( "********* BANYAN_LOAD ************ \n\r" );
#endif

   /* write the base address, and length of data on the controlE window */
   /* return 0 if failed */
   if( !write_addr_len( blen, baddr ) )
      return( 0 );

   /* write the Modem_Load_Block opcode */
   embox[0]  =  (unsigned char)MODEM_LOAD_BLOCK;
   if( !wrctrle(embox, TXCOMMANDREG, 1) )
      return( 0 );

   /* retreive the data from E-control window */
   if( !rdctrle(dld, BLCKDATAREG, blen) )
      return( 0 );

   /* return TRUE if success */
   rdctrle(embox, RXCOMMANDREG, 1);
   if( embox[0] == MODEM_BLOCK_AVAILABLE )
      return( 1 );
   else
      return( 0 );
}



/*
 *
 */
int  write_addr_len( unsigned int blen, unsigned int baddr )
{
   unsigned char  embox[6];
   unsigned int  xblen;

   /* assign the base address into E-control window */
   embox[0]  =  (unsigned char)(baddr%256);
   baddr  /=  256;
   embox[1]  =  (unsigned char)(baddr%256);
   baddr  /=  256;   
   embox[2]  =  (unsigned char)(baddr%256);
   baddr  /=  256;
   embox[3]  =  (unsigned char)(baddr%256);

   /* assign the length into E-control window */
   xblen  =  blen;
   embox[4]  =  (unsigned char)(xblen%256);
   xblen  /=  256;
   embox[5]  =  (unsigned char)(xblen%256);

   /* return 0 if failed */
   if( !wrctrle(embox, BLCKBADDRREG, 6) )
      return( 0 );

   /* return TRUE if success */
   return( 1 );
}



/*
 *
 */
int  banyan_sndgo( unsigned int jaddr )
{
   unsigned char  embox[4];

   embox[0]  =  (unsigned char)(jaddr%256);
   jaddr  /=  256;
   embox[1]  =  (unsigned char)(jaddr%256);
   jaddr  /=  256;
   embox[2]  =  (unsigned char)(jaddr%256);
   jaddr  /=  256;
   embox[3]  =  (unsigned char)(jaddr%256);

   /* prepare the go-address in the E-control window */
   if( !wrctrle(embox, BLCKBADDRREG, 4) )
      return( 0 );

   /* send the jump command via the E-control */
   embox[0]  =  ADSL_JUMP_ADDRESS;
   if( !wrctrle(embox, TXCOMMANDREG, 1) )
      return( 0 );

   /* return TRUE if success */
   return( 1 );
}



/*
 *
 */
int  adsl_turnx80( void )
{
   unsigned char  embox[16];

   for( embox[0] = 0; embox[0] != 0x80; ) {
      if( !(dispcount++ %60))
	printk(".");
      if( !rdctrle(embox,RXCOMMANDREG,1) )
         return( 0 );
   }

   return( 1 );
}



/*
 *
 */
int  adsl_openco( unsigned char fast, unsigned char interl )
{
   struct T_adsl_ConfigurationData  cfigco;
   unsigned int  retval;

   /* initialize the open parameters */
   /* cfigco.MinNoiseMarginDnstr  =  0;
   cfigco.MinNoiseMarginUpstr  =  0;
   cfigco.MaxAddNoiseMarginDnstr  =  0;
   cfigco.MaxAddNoiseMarginUpstr  =  0;
   cfigco.TargetNoiseMarginDnstr  =  0;
   cfigco.TargetNoiseMarginUpstr  =  0;
   cfigco.MaxPSDDnstr  =  0;
   cfigco.MaxAggrPwrLvlDnstr  =  0;
   cfigco.MaxAggrPwrLvlUpstr  =  0;
   strncpy( &(cfigco.CarrierMask),xcmask,32 );
   cfigco.RAModeDnstr  =  AUTOMATIC;
   cfigco.RAModeUpstr  =  AUTOMATIC;
   cfigco.RARatioDnstr  =  0;
   cfigco.RARatioUpstr  =  0;
   cfigco.RADownshiftMarginDnstr  =  0;
   cfigco.RADownshiftMarginUpstr  =  0;
   cfigco.RAUpshiftMarginDnstr  =  0;
   cfigco.RAUpshiftMarginUpstr  =  0;
   cfigco.RADownshiftIntervalDnstr  =  0;
   cfigco.RADownshiftIntervalUpstr  =  0;
   cfigco.RAUpshiftIntervalDnstr  =  0;
   cfigco.RAUpshiftIntervalUpstr  =  0;
   cfigco.InterlMinBitrateDnstr  =  0;
   cfigco.InterlMinBitrateUpstr  =  0;
   cfigco.InterlPlannedBitrateDnstr  =  0;
   cfigco.InterlPlannedBitrateUpstr  =  0;
   cfigco.InterlMaxBitrateDnstr  =  0;
   cfigco.InterlMaxBitrateUpstr  =  0;
   cfigco.InterlMaxDelayDnstr  =  0;
   cfigco.InterlMaxDelayUpstr  =  0;
   cfigco.FastMinBitrateDnstr  =  0;
   cfigco.FastMinBitrateUpstr  =  0;
   cfigco.FastPlannedBitrateDnstr  =  0;
   cfigco.FastPlannedBitrateUpstr  =  0;
   cfigco.FastMaxBitrateDnstr  =  0;
   cfigco.FastMaxBitrateUpstr  =  0;   */
   cfigco.UtopiaLevel  =  0x02;
   /* cfigco.UtopiaAddrFast  =  28; */
   cfigco.UtopiaAddrFast  =  fast;
   /* cfigco.UtopiaAddrInterl  =  29; */
   cfigco.UtopiaAddrInterl  =  interl;
   cfigco.FunctionCode1FD  =  0x15;
   cfigco.ExtraFunctionCode1FC  =  0;
   cfigco.itu_vendorid_countrycode  =  0;
   cfigco.itu_vendorid_reserved  =  0;
   cfigco.itu_vendorid_vendorcode_1  =  0;
   cfigco.itu_vendorid_vendorcode_2  =  0;
   cfigco.itu_vendorid_vendorcode_3  =  0;
   cfigco.itu_vendorid_vendorcode_4  =  0;
   cfigco.itu_vendorid_vendorspecific_1  =  0;
   cfigco.itu_vendorid_vendorspecific_2  =  0;

/* jungo, the next 7 lines */
#ifdef   AME_OPEN_ANSI
   retval  =  LT_ADSL_enableModemLine_ansi(&cfigco);
#else
#ifdef    AME_OPEN_GDMT
   retval  =  LT_ADSL_enableModemLine_gdmt(&cfigco);
#else
   retval  =  LT_ADSL_enableModemLine_multimode(&cfigco);
#endif
#endif

#if 0
   /* retval  =  LT_ADSL_enableModemLine_multimode(&cfigco); */
   retval  =  LT_ADSL_enableModemLine_ansi(&cfigco);
#endif

   if( retval == C_CONF_DATA_ACK )
      return( 1 );
   else
      return( 0 );
#if 0
   for( rdctrle(xstatus,RXCOMMANDREG,1); ; rdctrle(xstatus,RXCOMMANDREG,1) ) {
      if( (xstatus[0]==ADSL_OPEN_REJECTED) ||
          (xstatus[0]==ADSL_OPEN_FAILED) ||
          (xstatus[0]==ADSL_OPEN_FAILED_ON_TIMEOUT) ||
          (xstatus[0]==ADSL_CONNECTION_OPENED) ||
          (xstatus[0]==(unsigned int)0x70) ) {
         printk("THE OPENED RESPONSE = 0x%08X \n", (unsigned int)xstatus[0] );
         rsctrle(xstatus,CTEAD_MODEM_STATUS,1);
         printk("THE MODEM-STATUS = 0x%08X \n", (unsigned int)xstatus[0] );
         break;
      }
   }

   for( ; ; )
      printk( "THE OPEN-RESPONSE = 0x%08X \n", (unsigned int)xstatus[0] );

   if( xstatus[0] == ADSL_CONNECTION_OPENED )
      return( 1 );
   else
      return( 0 );
#endif
}



#define    FALSE           0
#define    TRUE            1
const char  ADSL_CLEARDEFBM25S_NEAR  =  (char)0x34;
const char  ADSL_CLEARDEFBM25S_FAR  =  (char)0x35;
const char  ADSL_CLEARDEFBM25S_NEAR_ACK  = (char)0x32;
const char  ADSL_CLEARDEFBM25S_FAR_ACK  = (char)0x33;
extern const char  MODEM_STATUS_UP;
const unsigned int  CTEAD_DEF_BITMAP_25S  =  (unsigned int)0x1E6;
const unsigned int  CTEAD_NEAR_BITMAP_25S  =  (unsigned int)0x1E6;
const unsigned int  CTEAD_FAR_BITMAP_25S  =  (unsigned int)0x1E7;
/*
 *
 */
int  ame_monitor( unsigned char *fstadr, unsigned char *intadr )
{
   int  nrcount[8] = {0,0,0,0,0,0,0,0};
   int  nrmax[8] = {5,5,3,15,15,60,0,0};   /*los,lof,lop,fcd,icd,lom*/
   int  frcount[8] = {0,0,0,0,0,0,0,0};
   int  frmax[8] = {6,6,3,17,17,60,0,0};   /*los,lof,lop,fcd,icd,lom*/
   int  i;
   struct T_adsl_defbm25s_set  bm25s;
   struct T_adsl_crc_near  crcnear;
   struct T_adsl_crc_far  crcfar;

   switch( ame_fast_utopiadr() ) {

   case 0:
     if( *fstadr == (unsigned char)28 ) {
       *fstadr = (unsigned char)29; *intadr = (unsigned char)28;
       return( 0 );
     } else 
       break;

   case 1:
     if( *fstadr == (unsigned char)28 )
       break;
     else {
       *fstadr = (unsigned char)28; *intadr = (unsigned char)29;
       return( 0 );
     }

      case 2:
         return( 0 );
   default:
     break;
   }
      /* case 8 */
   if( frcount[0] > 4 )
     if( nrcount[0] || nrcount[1] )
       return( 0 );
   /* case 11 */
   if( frcount[1] > 4 )
     if( nrcount[0] || nrcount[1] )
       return( 0 );

   if( ame_crcnear_poll(&crcnear) != C_CONF_DATA_ACK )
     return( 0 );

   if( ame_crcfar_poll(&crcfar) != C_CONF_DATA_ACK )
     return( 0 );
      /* case 14-17 */
#if 0
     if( !ame_crc_monitor( *(short int *)(crcnear.fast),
			   *(short int *)(crcnear.inlv),
			   *(short int *)(crcfar.fast),
			   *(short int *)(crcfar.inlv), newcount ) );
#else
   if( !ame_crc_monitor( *(short int *)(crcnear.fast),
			 *(short int *)(crcnear.inlv),
			 0, 0, 0 ) )
#endif
     return( 0 );
      

   if( ame_defbm25s_poll(&bm25s) != C_CONF_DATA_ACK )
     return( 0 );

   if( !ame_clear_dbm25s_near() )
     return( 0 );
   
   if( !ame_clear_dbm25s_far() )
     return( 0 );

   for( i = 2; i < 8; i++ ) {
     if( (bm25s.near_end>>i) & 0x01 )
       nrcount[7-i]  +=  1;
     else
       nrcount[7-i]  =  0;
     
     if( (bm25s.far_end>>i) & 0x01 )
       frcount[7-i]  +=  1;
     else
       frcount[7-i]  =  0;
   }

   /* case 1-4, 6-7, 9-10, 12-13 */
   for( i = 2; i < 8; i++ ) {
     if( nrmax[7-i] < nrcount[i] )
       return( 0 );
     if( frmax[7-i] < frcount[i] )
       return( 0 );
   }

   return 1;
}



/*
 *
 */
int  ame_close( void )
{
   unsigned int  retval;
   unsigned char  xstatus[4];

   retval  =  ADSL_disableModemLine();
   if( (retval!=C_CONF_DATA_ACK)&&(retval!=C_CONF_DATA_ALREADY_DOWN) ) {
      rsctrle( xstatus, CTEAD_MODEM_STATUS, 1 );
#ifdef RLDR_DEBUG
         printk( "AME_CLOSE: ModemStatus = 0x%02X \n",
                  (unsigned int)xstatus[0] );
#endif
      return( 0 );
   }

   if( retval == C_CONF_DATA_ALREADY_DOWN )
       return(1);

   for(rdctrle(xstatus,RXCOMMANDREG,1); ;rdctrle(xstatus,RXCOMMANDREG,1) ) {
      if( xstatus[0] == ADSL_CONNECTION_CLOSED )
         return( 1 );
      else if( xstatus[0] == ADSL_CLOSE_REJECTED ) {
#ifdef RLDR_DEBUG
            printk( "AME_CLOSE: CloseRejected.\n" );
#endif
         return( 0 );
      } else {
#ifdef RLDR_DEBUG
            printk( "AME_CLOSE: CommandResp = 0x%02X \n",
                    (unsigned int)xstatus[0] );
#endif
      }
   }
}



/*
 *
 */
int  ame_defbm25s_poll( struct T_adsl_defbm25s_set *bm25s )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* check if modem is in showtime */
   if( rsctrle(xstatus,CTEAD_MODEM_STATUS,1) ) {
      inshowtime  =  (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror  =  TRUE;

   /* read the 25s defect bitmap */
   if( !commerror ) {
     char  getdata[2];
     if( rsctrle(getdata,CTEAD_DEF_BITMAP_25S,2) ) {
        bm25s->near_end  =  (unsigned char)getdata[0];
        bm25s->far_end  =  (unsigned char)getdata[1];
     } else
        commerror  =  TRUE;
   }

   /* return C_CONF_DATA_ACK if all ok, C_CONF_DATA_REJ if comms ok */
   /* but modem is not in showtime, C_CONF_DATA_COMMPROBLEM if comms */
   /* not ok. */
   if( commerror )
      return( C_CONF_DATA_COMMPROBLEM );
   else {
      if( !inshowtime )
         return( C_CONF_DATA_REJ );
      else
         return( C_CONF_DATA_ACK );
   }

}



/*
 *
 */
int  ame_clear_dbm25s_near( void )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* verify if modem is in showtime */
   if( rsctrle(xstatus,CTEAD_MODEM_STATUS,1) ) {
      inshowtime = (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror  =  TRUE;

   /* write "ADSL_ClearDefBitmapDuring2.5sNear" command */
   if( !commerror ) {
      char  putdata[1];
      putdata[0]  =  ADSL_CLEARDEFBM25S_NEAR;
      if( !wrctrle(putdata,(unsigned int)TXCOMMANDREG,1) )
         commerror  =  TRUE;
   }

   /* return 1 if all ok, otherwise return 0 */
   if( commerror )
      return( 0 );
   else {
      if( !inshowtime )
         return( 0 );
      else {
         for( rdctrle(xstatus,RXCOMMANDREG,1); ;
              rdctrle(xstatus,RXCOMMANDREG,1) )
            if( xstatus[0] == ADSL_CLEARDEFBM25S_NEAR_ACK )
               return( 1 );
            else {
#ifdef RLDR_DEBUG
                  printk( "AME_CLEAR_NEAR: RESP = 0x%02X \n",
                          (unsigned int)xstatus[0] );
#endif
               /* return( 0 ); */
            }
      }
   }

}



/*
 *
 */
int  ame_clear_dbm25s_far( void )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* verify if modem is in showtime */
   if( rsctrle(xstatus,CTEAD_MODEM_STATUS,1) ) {
      inshowtime = (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror  =  TRUE;

   /* write "ADSL_ClearDefBitmapDuring2.5sFar" command */
   if( !commerror ) {
      char  putdata[1];
      putdata[0]  =  ADSL_CLEARDEFBM25S_FAR;
      if( !wrctrle(putdata,(unsigned int)TXCOMMANDREG,1) )
         commerror  =  TRUE;
   }

   /* return 1 if all ok, otherwise return 0 */
   if( commerror )
      return( 0 );
   else {
      if( !inshowtime )
         return( 0 );
      else {
         for( rdctrle(xstatus,RXCOMMANDREG,1); ;
              rdctrle(xstatus,RXCOMMANDREG,1) )
            if( xstatus[0] == ADSL_CLEARDEFBM25S_FAR_ACK )
               return( 1 );
            else {
#ifdef RLDR_DEBUG
                  printk( "AME_CLEAR_NEAR: RESP = 0x%02X \n",
                          (unsigned int)xstatus[0] );
#endif
               /* return( 0 ); */
            }

      }
   }

}



/*
 *
 */
int  ame_crcnear_poll( struct T_adsl_crc_near *crcn )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* check if modem is in showtime */
   if( rsctrle(xstatus,CTEAD_MODEM_STATUS,1) ) {
      inshowtime  =  (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror  =  TRUE;

   /* read the incorrect far CRC counter in fast/interleave */
   if( !commerror ) {
      unsigned char  getdata[4];
      if( rsctrle(getdata,0x1B4,4) ) {
         crcn->fast[1]  =  getdata[0];
         crcn->fast[0]  =  getdata[1];
         crcn->inlv[1]  =  getdata[2];
         crcn->inlv[0]  =  getdata[3];
      } else
         commerror  =  TRUE;
   }

   /* return C_CONF_DATA_ACK if all ok, C_CONF_DATA_REJ if comms ok */
   /* but modem is not in showtime, C_CONF_DATA_COMMPROBLEM if comms */
   /* not ok. */
   if( commerror )
      return( C_CONF_DATA_COMMPROBLEM );
   else {
      if( !inshowtime )
         return( C_CONF_DATA_REJ );
      else
         return( C_CONF_DATA_ACK );
   }
}



/*
 *
 */
int  ame_crcfar_poll( struct T_adsl_crc_far *crcf )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* check if modem is in showtime */
   if( rsctrle(xstatus,CTEAD_MODEM_STATUS,1) ) {
      inshowtime  =  (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror  =  TRUE;

   /* read the incorrect far CRC counter in fast/interleave */
   if( !commerror ) {
      unsigned char  getdata[4];
      if( rsctrle(getdata,0x1C0,4) ) {
         crcf->fast[1]  =  getdata[0];
         crcf->fast[0]  =  getdata[1];
         crcf->inlv[1]  =  getdata[2];
         crcf->inlv[0]  =  getdata[3];
      } else
         commerror  =  TRUE;
   }

   /* return C_CONF_DATA_ACK if all ok, C_CONF_DATA_REJ if comms ok */
   /* but modem is not in showtime, C_CONF_DATA_COMMPROBLEM if comms */
   /* not ok. */
   if( commerror )
      return( C_CONF_DATA_COMMPROBLEM );
   else {
      if( !inshowtime )
         return( C_CONF_DATA_REJ );
      else
         return( C_CONF_DATA_ACK );
   }
}



/*
 *
 */
int  ame_crc_monitor( unsigned short int crcnf, unsigned short int crcni,
                      unsigned short int crcff, unsigned short int crcfi,
                      int newcount )
{
   static unsigned short int  pcrcnf, pcrcni, pcrcff, pcrcfi;

   
   if( newcount )
      pcrcnf = pcrcni = pcrcff = pcrcfi = 0;

   if( crcnf < pcrcnf )
      pcrcnf  =  crcnf;
   if( crcni < pcrcni )
      pcrcni  =  crcni;
   if( crcff < pcrcff )
      pcrcff  =  crcff;
   if( crcfi < pcrcfi )
      pcrcfi  =  crcfi;

   if( (crcnf-pcrcnf) > 20 )
      return( 0 );
   if( (crcni-pcrcni) > 20 )
      return( 0 );
   if( (crcff-pcrcff) > 20 )
      return( 0 );
   if( (crcfi-pcrcfi) > 20 )
      return( 0 );

   return( 1 );

}


volatile void ame_delay_1s( int num )
{    
#if 0
  int  i;   for( i = 0; i < (num/10+1); i++ )
    {
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout (1*HZ);
    }
#else
  int i;
  for(i=0;i<1000;i++)
    mdelay(num);
#endif
}  

/*
 *
 */
int  ame_resp_nw( void )
{
   unsigned char  xstatus[8];

   rsctrle( xstatus, RXCOMAVREG, 1 );
   if( xstatus[0] & 0x01 ) {
      rsctrle( xstatus, RXCOMMANDREG, 1 );
      return( ((unsigned int)xstatus[0]&0x0FF)|0x0F00 );
   } else
      return( 0 );
}



/*
 *
 */
int  ame_get_opmode( void )
{
   unsigned char  embox[4];

#ifdef RLDR_DEBUG
      printk( "********* AME_GET_OPMODE ********* \n\r" );
#endif

   /* issue the ADSL_Get_Operational Mode command */
   /* this command is only available in NORMAL mode */
   embox[0]  =  (unsigned char)GET_OPERATIONAL_MODE;
   if( !wrctrle(embox,TXCOMMANDREG,1) )
      return( 0 );

   /* return TRUE if the chipset is in NORMAL mode */
   embox[0]  =  0;
   rdctrle( embox, RXCOMMANDREG, 1 );
#ifdef RLDR_DEBUG
      printk( "AME_GET_OPMODE: EMBOX[0] = 0x%08X \n\r", (unsigned int)embox[0] );
#endif
   if( embox[0] == ADSL_UNKNOWN_OPCODE )
      return( 0 );
   else
      return( 1 );
}



/*
 *
 */
int  ame_fast_utopiadr( void )
{
   int  lcount;
   struct T_adsl_chl_data  chl_data;

   for( lcount = 0; lcount < 3; lcount++ ) {
      if( ame_chopdata_poll(&chl_data) == C_CONF_DATA_ACK ) {
#ifdef RLDR_DEBUG
	    printk( "INT_NEAR_LSB[0/1] = 0x%02X/0x%02X \n",
                    chl_data.int_near_lsb[0], chl_data.int_near_lsb[1] );
            printk( "INT_NEAR_MSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.int_near_msb[0], chl_data.int_near_msb[1] );
            printk( "FAST_NEAR_LSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.fast_near_lsb[0], chl_data.fast_near_lsb[1] );
            printk( "FAST_NEAR_MSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.fast_near_msb[0], chl_data.fast_near_msb[1] );
            printk( "INT_FAR_LSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.int_far_lsb[0], chl_data.int_far_lsb[1] );
            printk( "INT_FAR_MSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.int_far_msb[0], chl_data.int_far_msb[1] );
            printk( "FAST_FAR_LSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.fast_far_lsb[0], chl_data.fast_far_lsb[1] );
            printk( "FAST_FAR_MSB[0/1] = 0x%02X/0x%02X \n", 
                    chl_data.fast_far_msb[0], chl_data.fast_far_msb[1] );
#endif
         if( chl_data.fast_near_lsb[0] | chl_data.fast_near_lsb[1] |
             chl_data.fast_near_msb[0] | chl_data.fast_near_msb[1] ) {
            return( 1 );
	 } else {
            return( 0 );
	 }
      }
   }

   return( 2 );
}



/*
 *
 */
int  ame_chopdata_poll( struct T_adsl_chl_data *chl_data )
{
   char  xstatus[1];
   unsigned char  commerror = FALSE;
   unsigned char  inshowtime = 0;

   /* check if modem is in showtime */
   if( rsctrle(xstatus, CTEAD_MODEM_STATUS,1) ) {
      inshowtime = (xstatus[0]==MODEM_STATUS_UP);
   } else
      commerror = TRUE;

   /* read the byte 0 of the channel operational data */
   if( !commerror ) {
      unsigned char  getdata[8];
      if( rsctrle(getdata,0x01A,8) ) {
         chl_data->int_near_lsb[0]  =  getdata[0];
         chl_data->int_near_msb[0]  =  getdata[1];
         chl_data->fast_near_lsb[0]  =  getdata[2];
         chl_data->fast_near_msb[0]  =  getdata[3];
         chl_data->int_far_lsb[0]  =  getdata[4];
         chl_data->int_far_lsb[0]  =  getdata[5];
         chl_data->fast_far_lsb[0]  =  getdata[6];
         chl_data->fast_far_lsb[0]  =  getdata[7];
      } else
         commerror  =  TRUE;
      if( rsctrle(getdata,CTEAD_NEAR_INT_CHANNELDATA_LSB,8) ) {
         chl_data->int_near_lsb[1]  =  getdata[0];
         chl_data->int_near_msb[1]  =  getdata[1];
         chl_data->fast_near_lsb[1]  =  getdata[2];
         chl_data->fast_near_msb[1]  =  getdata[3];
         chl_data->int_far_lsb[1]  =  getdata[4];
         chl_data->int_far_lsb[1]  =  getdata[5];
         chl_data->fast_far_lsb[1]  =  getdata[6];
         chl_data->fast_far_lsb[1]  =  getdata[7];
      } else
         commerror  =  TRUE;
   }

   /* return C_CONF_DATA_ACK if all OK, C_CONF_DATA_REJ if comms OK */
   /* but modem is not in SHOWTIME, C_CONF_DATA_COMMPROBLEM if comms */
   /* not OK. */
   if( commerror )
      return( C_CONF_DATA_COMMPROBLEM );
   else {
      if( !inshowtime )
         return( C_CONF_DATA_REJ );
      else
         return( C_CONF_DATA_ACK );
   }
}
EXPORT_SYMBOL(ame_resp_nw);
