/*
*******************************************************************************
**	Copyright 1999, 2000 Integrated Device Technology, Inc.
**	All Rights Reserved
**
*******************************************************************************
*/
/*
******************************************************************************
** banyan.h -- data definitions for Banyan interfaces.
**
**	Portions of this header file were taken from our partner AT&T in the
**	Banyan project, and has been left as is or has been modified.
**
**	For pusposes of reference following is the banner from the AT&T header
**	file banyan.h and has been reproduced exactly as I found it.
**						-- hak, IDT	06-15-2000.

******************************************************************************
*/
/*****************************************************************************
 *     $Id: banyan.h,v 1.19 2000/05/18 15:56:32 licwang Exp $
 *
 *     Description:  Banyan C/C++ Header File
 *     Author:       Peter Z. Onufryk
 *     Company:      AT&T
 *
 *     $Log: banyan.h,v $
 *     Revision 1.19  2000/05/18 15:56:32  licwang
 *     correct crc10 to crc10c.(From spec 15-19)
 *
 *     Revision 1.18  2000/04/19 15:10:13  pzo
 *     Changed address of WTCOUNT, WTCOMPARE, and WTC reigsters to remove
 *     gap caused the by address change for the BTADDR register.
 *
 *     Revision 1.17  2000/04/17 22:49:09  licwang
 *     Change vccfe's name
 *
 *     Revision 1.16  2000/04/14 14:36:08  pzo
 *     Updated header file to reflect the new address for the BTADDR register.
 *
 *     Revision 1.15  2000/04/06 14:47:33  pzo
 *     Added GPIONMIEN register to GPIO interface.
 *
 *     Revision 1.14  2000/03/14 20:14:39  pzo
 *     Fixed problem with UART register map identified by IDT (Dan).
 *
 *     Revision 1.13  2000/02/16 21:13:30  pzo
 *     Corrected typo in UART registers (UART2RB to UART1RB).
 *
 *     Revision 1.12  2000/02/04 19:49:31  srinivas
 *     Updated for ETHCFSA[0|1|2] registers.
 *
 *     Revision 1.11  2000/02/03 20:43:52  pzo
 *     Added UART0RR and UART1RR register definitions.
 *
 *     Revision 1.10  2000/02/02 18:22:22  raju
 *     Removed rcompare, rtc and rcount registers from sdram structure.
 *
 *     Revision 1.9  2000/01/10 21:55:27  raju
 *     usbepbd has been changed to usbepd, the data register name.
 *
 *     Revision 1.8  2000/01/07 15:31:51  srinivas
 *     Fixed the repeat definitions for REFRESH counters in SDRAM and
 *     endif statement in system integrity sections
 *
 *     Revision 1.7  2000/01/06 22:45:44  pzo
 *     Updated system integrity and counter timers to refleced updated spec
 *     register map.
 *
 *     Revision 1.6  2000/01/03 20:51:18  pzo
 *     Fixed typo in adding I2CDEBUG register.
 *
 *     Revision 1.5  2000/01/03 20:45:05  pzo
 *     Added I2CDEBUG register to I2C register map. This is an internal register
 *     for hardware debug and is NOT documented in the Banyan spec.
 *
 *     Revision 1.4  1999/12/11 00:42:24  pzo
 *     Ethernet registers were updated by Jiann.
 *
 *     Revision 1.3  1999/11/02 21:48:39  srinivas
 *     Added I2C input/output data registers
 *
 *     Revision 1.2  1999/09/22 17:20:19  licwang
 *     Corrected gpiolevel to gpioilevel in banyan.h
 *
 *     Revision 1.1.1.1  1999/08/03 21:44:25  pzo
 *     Banyan simulator imported into CVS
 *
 *
 *****************************************************************************/

#ifndef __banyan__
#define __banyan__

/********************************************************************
 * Banyan register map - see architecture spec                  
 * all addresses are virtual addresses in kseg1                         
 ********************************************************************/

/********************************************************************
 * system identification 
 ********************************************************************/
#define SYSID_BASE        0xB8000018
#define SYSID             0xB8000018

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int sysid;
} SYSID_STRUCT;

#define systemid (*((volatile SYSID_STRUCT *) SYSID_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * reset handler 
 ********************************************************************/
#define RESET_BASE        0xB8008000
#define RESET             0xB8008000

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int reset;
} RESET_STRUCT;

#define reset_handler (*((volatile RESET_STRUCT *) RESET_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * device controller 
 ********************************************************************/
#define DEVICE_BASE       0xB8010000
#define DEV0BASE          0xB8010000
#define DEV0MASK          0xB8010004
#define DEV0C             0xB8010008
#define DEV0TC            0xB801000C
#define DEV1BASE          0xB8010010
#define DEV1MASK          0xB8010014
#define DEV1C             0xB8010018
#define DEV1TC            0xB801001C
#define DEV2BASE          0xB8010020
#define DEV2MASK          0xB8010024
#define DEV2C             0xB8010028
#define DEV2TC            0xB801002C
#define DEV3BASE          0xB8010030
#define DEV3MASK          0xB8010034
#define DEV3C             0xB8010038
#define DEV3TC            0xB801003C
#define DEV4BASE          0xB8010040
#define DEV4MASK          0xB8010044
#define DEV4C             0xB8010048
#define DEV4TC            0xB801004C
#define DEV5BASE          0xB8010050
#define DEV5MASK          0xB8010054
#define DEV5C             0xB8010058
#define DEV5TC            0xB801005C

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int devbase;
  unsigned int devmask;
  unsigned int devc;
  unsigned int devtc;
} DEVICE_NUM;

typedef struct {
 DEVICE_NUM dev0;
 DEVICE_NUM dev1;
 DEVICE_NUM dev2;
 DEVICE_NUM dev3;
 DEVICE_NUM dev4;
 DEVICE_NUM dev5;
} DEVICE;

#define device (*((volatile DEVICE *) DEVICE_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * SDRAM controller 
 ********************************************************************/
#define SDRAM_BASE    	  0xB8018000
#define SDRAM0BASE        0xB8018000
#define SDRAM0MASK        0xB8018004
#define SDRAM1BASE        0xB8018008
#define SDRAM1MASK        0xB801800C
#define SDRAMC            0xB8018010

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int sdram0base;
  unsigned int sdram0mask;
  unsigned int sdram1base;
  unsigned int sdram1mask;
  unsigned int sdramc;
} SDRAM;

#define sdram (*((volatile SDRAM *) SDRAM_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * bus arbiter 
 ********************************************************************/
#define ARBITER_BASE      0xB8020000
#define ARBC0             0xB8020000
#define ARBC1             0xB8020004

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int arbc0;
  unsigned int arbc1;
} ARBITER;

#define arbiter (*((volatile ARBITER *) ARBITER_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * counter timers 
 ********************************************************************/
#define TIMER_BASE        0xB8028000
#define COUNT0            0xB8028000
#define COMPARE0          0xB8028004
#define CTC0              0xB8028008
#define COUNT1            0xB802800C
#define COMPARE1          0xB8028010
#define CTC1              0xB8028014
#define COUNT2            0xB8028018
#define COMPARE2          0xB802801C
#define CTC2              0xB8028020
#define RCOUNT            0xB8028024
#define RCOMPARE          0xB8028028
#define RTC               0xB802802C

#define CNTR_TMR1_MASK    0x02 

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int count;
  unsigned int compare;
  unsigned int ctc;
} CT;

typedef struct {
  CT ct0;
  CT ct1;
  CT ct2;
  unsigned int rcount;
  unsigned int rcompare;
  unsigned int rtc;
} TIMER;

#define timer (*((volatile TIMER *) TIMER_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * system integrity features 
 ********************************************************************/
#define SYSINTEG_BASE     0xB8028030
#define ERRCS             0xB8028030
#define BTCOMPARE         0xB8028034
#define WTCOUNT           0xB8028038
#define WTCOMPARE         0xB802803C
#define WTC               0xB8028040
#define BTADDR            0xB800001C

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int errcs;
  unsigned int btcompare;
  unsigned int wtcount;
  unsigned int wtcompare;
  unsigned int wtc;
} SYSINTEG;

#define sysinteg (*((volatile SYSINTEG *) SYSINTEG_BASE))
#define btaddr   (*((volatile int *) BTADDR))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * interrupt controller 
 ********************************************************************/
#define INTERRUPT_BASE    0xB8030000
#define IPEND2            0xB8030000
#define IMASK2            0xB8030004
#define IPEND3            0xB8030008
#define IMASK3            0xB803000C
#define IPEND4            0xB8030010
#define IMASK4            0xB8030014
#define IPEND5            0xB8030018
#define IMASK5            0xB803001C
#define IPEND6            0xB8030020
#define IMASK6            0xB8030024

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int ipend;
  unsigned int imask;
} INTR;

typedef struct {
  INTR i2; 
  INTR i3; 
  INTR i4; 
  INTR i5; 
  INTR i6; 
} INTERRUPT;

#define interrupt (*((volatile INTERRUPT *) INTERRUPT_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * DMA controller
 ********************************************************************/
#define DMA_BASE          0xB8038000
#define DMA0C             0xB8038000
#define DMA0S             0xB8038004
#define DMA0SM            0xB8038008
#define DMA0DPTR          0xB803800C
#define DMA0NDPTR         0xB8038010
#define DMA1C             0xB8038014
#define DMA1S             0xB8038018
#define DMA1SM            0xB803801C
#define DMA1DPTR          0xB8038020
#define DMA1NDPTR         0xB8038024
#define DMA2C             0xB8038028
#define DMA2S             0xB803802C
#define DMA2SM            0xB8038030
#define DMA2DPTR          0xB8038034
#define DMA2NDPTR         0xB8038038
#define DMA3C             0xB803803C
#define DMA3S             0xB8038040
#define DMA3SM            0xB8038044
#define DMA3DPTR          0xB8038048
#define DMA3NDPTR         0xB803804C
#define DMA4C             0xB8038050
#define DMA4S             0xB8038054
#define DMA4SM            0xB8038058
#define DMA4DPTR          0xB803805C
#define DMA4NDPTR         0xB8038060
#define DMA5C             0xB8038064
#define DMA5S             0xB8038068
#define DMA5SM            0xB803806C
#define DMA5DPTR          0xB8038070
#define DMA5NDPTR         0xB8038074
#define DMA6C             0xB8038078
#define DMA6S             0xB803807C
#define DMA6SM            0xB8038080
#define DMA6DPTR          0xB8038084
#define DMA6NDPTR         0xB8038088
#define DMA7C             0xB803808C
#define DMA7S             0xB8038090
#define DMA7SM            0xB8038094
#define DMA7DPTR          0xB8038098
#define DMA7NDPTR         0xB803809C
#define DMA8C             0xB80380A0
#define DMA8S             0xB80380A4
#define DMA8SM            0xB80380A8
#define DMA8DPTR          0xB80380AC
#define DMA8NDPTR         0xB80380B0
#define DMA9C             0xB80380B4
#define DMA9S             0xB80380B8
#define DMA9SM            0xB80380BC
#define DMA9DPTR          0xB80380C0
#define DMA9NDPTR         0xB80380C4
#define DMA10C            0xB80380C8
#define DMA10S            0xB80380CC
#define DMA10SM           0xB80380D0
#define DMA10DPTR         0xB80380D4
#define DMA10NDPTR        0xB80380D8
#define DMA11C            0xB80380DC
#define DMA11S            0xB80380E0
#define DMA11SM           0xB80380E4
#define DMA11DPTR         0xB80380E8
#define DMA11NDPTR        0xB80380EC
#define DMA12C            0xB80380F0
#define DMA12S            0xB80380F4
#define DMA12SM           0xB80380F8
#define DMA12DPTR         0xB80380FC
#define DMA12NDPTR        0xB8038100
#define DMA13C            0xB8038104
#define DMA13S            0xB8038108
#define DMA13SM           0xB803810C
#define DMA13DPTR         0xB8038110
#define DMA13NDPTR        0xB8038114
#define DMA14C            0xB8038118
#define DMA14S            0xB803811C
#define DMA14SM           0xB8038120
#define DMA14DPTR         0xB8038124
#define DMA14NDPTR        0xB8038128
#define DMA15C            0xB803812C
#define DMA15S            0xB8038130
#define DMA15SM           0xB8038134
#define DMA15DPTR         0xB8038138
#define DMA15NDPTR        0xB803813C

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int dmac;
  unsigned int dmas;
  unsigned int dmasm;
  unsigned int dmadptr;
  unsigned int dmandptr;
} DMA_CHAN;

typedef struct {
 DMA_CHAN ch0;
 DMA_CHAN ch1;
 DMA_CHAN ch2;
 DMA_CHAN ch3;
 DMA_CHAN ch4;
 DMA_CHAN ch5;
 DMA_CHAN ch6;
 DMA_CHAN ch7;
 DMA_CHAN ch8;
 DMA_CHAN ch9;
 DMA_CHAN ch10;
 DMA_CHAN ch11;
 DMA_CHAN ch12;
 DMA_CHAN ch13;
 DMA_CHAN ch14;
 DMA_CHAN ch15;
} DMA;

#define dma (*((volatile DMA *) DMA_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * GPIO
 ********************************************************************/
#define GPIO_BASE         0xB8040000
#define GPIOFUNC          0xB8040000
#define GPIOCFG           0xB8040004
#define GPIOD             0xB8040008
#define GPIOILEVEL        0xB804000C
#define GPIOISTAT         0xB8040010
#define GPIONMIEN         0xB8040014

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int gpiofunc;
  unsigned int gpiocfg;
  unsigned int gpiod;
  unsigned int gpioilevel;
  unsigned int gpioistat;
  unsigned int gpionmien;
} GPIO;

#define gpio (*((volatile GPIO *) GPIO_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * TDM bus interface 
 ********************************************************************/
#define TDM_BASE          0xB8048000
#define TDMC              0xB8048000
#define TDMS              0xB8048004
#define TDMSM             0xB8048008
#define TDMPC             0xB804800C
#define TDMOSS0           0xB8048010
#define TDMOSS1           0xB8048014
#define TDMOSS2           0xB8048018
#define TDMOSS3           0xB804801C
#define TDMISS0           0xB8048020
#define TDMISS1           0xB8048024
#define TDMISS2           0xB8048028
#define TDMISS3           0xB804802C
#define TDMID             0xB8048030
#define TDMOD             0xB8048040

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int tdmc;
  unsigned int tdms;
  unsigned int tdmsm;
  unsigned int tdmpc;
  unsigned int tdmoss0;
  unsigned int tdmoss1;
  unsigned int tdmoss2;
  unsigned int tdmoss3;
  unsigned int tdmiss0;
  unsigned int tdmiss1;
  unsigned int tdmiss2;
  unsigned int tdmiss3;
  unsigned int tdmid;
  unsigned int _u0;
  unsigned int _u1;
  unsigned int _u2;
  unsigned int tdmod;
  unsigned int _u3;
  unsigned int _u4;
  unsigned int _u5;
} TDM;

#define tdm (*((volatile TDM *) TDM_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * UART
 ********************************************************************/
#define UART_BASE         0xB8050000
#define UART0RB           0xB8050000
#define UART0IE           0xB8050004
#define UART0II           0xB8050008
#define UART0LC           0xB805000C
#define UART0MC           0xB8050010
#define UART0LS           0xB8050014
#define UART0MS           0xB8050018
#define UART0S            0xB805001C
#define UART1RB           0xB8050020
#define UART1IE           0xB8050024
#define UART1II           0xB8050028
#define UART1LC           0xB805002C
#define UART1MC           0xB8050030
#define UART1LS           0xB8050034
#define UART1MS           0xB8050038
#define UART1S            0xB805003C
#define UART0RR           0xB8050040
#define UART1RR           0xB8050044

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int uart0rb;
  unsigned int uart0ie;
  unsigned int uart0ii;
  unsigned int uart0lc;
  unsigned int uart0mc;
  unsigned int uart0ls;
  unsigned int uart0ms;
  unsigned int uart0s;
  unsigned int uart1rb;
  unsigned int uart1ie;
  unsigned int uart1ii;
  unsigned int uart1lc;
  unsigned int uart1mc;
  unsigned int uart1ls;
  unsigned int uart1ms;
  unsigned int uart1s;
  unsigned int uart0rr;
  unsigned int uart1rr;
} UART;

#define uart (*((volatile UART *) UART_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * ATM interface
 ********************************************************************/
#define ATM_BASE          0xB8058000
#define ATMINTFC          0xB8058000
#define CRC32VAL          0xB8058004
#define CRC10C            0xB8058008
#define CRC10VAL          0xB805800C
#define ATMCP             0xB8058010
#define ATMU2PC           0xB8058014
#define ATM0C             0xB8058020
#define ATM0S             0xB8058024
#define ATM0IH            0xB8058028
#define ATM0IEH           0xB805802C
#define ATM0ID            0xB8058030
#define ATM0OD            0xB8058040
#define ATM0ODH           0xB8058050
#define ATM1C             0xB8058060
#define ATM1S             0xB8058064
#define ATM1IH            0xB8058068
#define ATM1IEH           0xB805806C
#define ATM1ID            0xB8058070
#define ATM1OD            0xB8058080
#define ATM1ODH           0xB8058090
#define ATMVCCH1          0xB80580A0
#define ATMVCCH2          0xB80580A4
#define ATMVCCH3          0xB80580A8
#define ATMVCCH4          0xB80580AC
#define ATMVCCH5          0xB80580B0
#define ATMVCCH6          0xB80580B4
#define ATMVCCH7          0xB80580B8
#define ATMVCCH8          0xB80580BC
#define ATMVCCD           0xB80580C0
#define ATMVCCS           0xB80580C4
#define ATMVCCA           0xB80580C8
#define ATMVCCC           0xB80580CC
#define ATMVCFE           0xB80580D0
#define ATM0OH            0xB80580D4
#define ATM0OEH           0xB80580D8
#define ATM1OH            0xB80580DC
#define ATM1OEH           0xB80580E0

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int atmc;
  unsigned int atms;
  unsigned int atmih;
  unsigned int atmieh;
  unsigned int atmid;
  unsigned int _u0;
  unsigned int _u1;
  unsigned int _u2;
  unsigned int atmod;
  unsigned int _u3;
  unsigned int _u4;
  unsigned int _u5;
  unsigned int atmodh;
  unsigned int _u6;
  unsigned int _u7;
  unsigned int _u8;
} ATM_CHAN;

typedef struct {
  unsigned int atmintfc;
  unsigned int crc32val;
  unsigned int crc10c;
  unsigned int crc10val;
  unsigned int atmcp;
  unsigned int atmu2pc;
  unsigned int _u0;
  unsigned int _u1;
  ATM_CHAN ch0;
  ATM_CHAN ch1;
  unsigned int atmvcch1;
  unsigned int atmvcch2;
  unsigned int atmvcch3;
  unsigned int atmvcch4;
  unsigned int atmvcch5;
  unsigned int atmvcch6;
  unsigned int atmvcch7;
  unsigned int atmvcch8;
  unsigned int atmvccd;
  unsigned int atmvccs;
  unsigned int atmvcca;
  unsigned int atmvccc;
  unsigned int atmvccfe;
  unsigned int atm0oh;
  unsigned int atm0oeh;
  unsigned int atm1oh;
  unsigned int atm1oeh;
} ATM;
 
#define atm (*((volatile ATM *) ATM_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * ethernet interface
 ********************************************************************/

#define ETHERNET_BASE     0xB8060000
#define ETHINTFC          0xB8060000
#define ETHFIFOTT         0xB8060004
#define ETHARC            0xB8060008
#define ETHHASH0          0xB806000C
#define ETHHASH1          0xB8060010
#define ETHFIFOST         0xB8060014
#define ETHFIFOS          0xB8060018
#define ETHODEOPS         0xB806001C
#define ETHIS             0xB8060020
#define ETHOS             0xB8060024
#define ETHMCP            0xB8060028
#define ETHID             0xB8060030
#define ETHOD             0xB8060040
#define ETHODEOP          0xB8060050
#define ETHSAL0           0xB8060100
#define ETHSAH0           0xB8060104
#define ETHSAL1           0xB8060108
#define ETHSAH1           0xB806010C
#define ETHSAL2           0xB8060110
#define ETHSAH2           0xB8060114
#define ETHSAL3           0xB8060118
#define ETHSAH3           0xB806011C 
#define ETHRBC            0xB8060120
#define ETHRPC            0xB8060124
#define ETHRUPC           0xB8060128
#define ETHRFC            0xB806012C
#define ETHTBC            0xB8060130
#define ETHGPF            0xB8060134
#define ETHMAC1           0xB8060200
#define ETHMAC2           0xB8060204
#define ETHIPGT           0xB8060208
#define ETHIPGR           0xB806020C
#define ETHCLRT           0xB8060210
#define ETHMAXF           0xB8060214
#define ETHMTEST          0xB806021C
#define MIIMCFG           0xB8060220
#define MIIMCMD           0xB8060224
#define MIIMADDR          0xB8060228
#define MIIMWTD           0xB806022C
#define MIIMRDD           0xB8060230
#define MIIMIND           0xB8060234
#define ETHCFSA0          0xB8060240
#define ETHCFSA1          0xB8060244
#define ETHCFSA2          0xB8060248

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int ethintfc;
  unsigned int ethfifott;
  unsigned int etharc;
  unsigned int ethhash0;
  unsigned int ethhash1;
  unsigned int ethfifost;
  unsigned int ethfifos;
  unsigned int ethodeops;
  unsigned int ethis;
  unsigned int ethos;
  unsigned int ethmcp;
  unsigned int _u1;
  unsigned int ethid;
  unsigned int _u2;
  unsigned int _u3;
  unsigned int _u4;
  unsigned int ethod;
  unsigned int _u5;
  unsigned int _u6;
  unsigned int _u7;
  unsigned int ethodeop;
  unsigned int _u8[43]; 
  unsigned int ethsal0;
  unsigned int ethsah0;
  unsigned int ethsal1;
  unsigned int ethsah1;
  unsigned int ethsal2;
  unsigned int ethsah2;
  unsigned int ethsal3;
  unsigned int ethsah3;
  unsigned int ethrbc;
  unsigned int ethrpc;
  unsigned int ethrupc;
  unsigned int ethrfc;
  unsigned int ethtbc;
  unsigned int ethgpf;
  unsigned int _u9[50];
  unsigned int ethmac1;
  unsigned int ethmac2;
  unsigned int ethipgt;
  unsigned int ethipgr;
  unsigned int ethclrt;
  unsigned int ethmaxf;
  unsigned int _u10;
  unsigned int ethmtest;
  unsigned int miimcfg;
  unsigned int miimcmd;
  unsigned int miimaddr;
  unsigned int miimwtd;
  unsigned int miimrdd;
  unsigned int miimind;
  unsigned int _u11;
  unsigned int _u12;
  unsigned int ethcfsa0;
  unsigned int ethcfsa1;
  unsigned int ethcfsa2;
} ETHERNET;
 
#define ethernet (*((volatile ETHERNET *) ETHERNET_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * USB interface
 ********************************************************************/
#define USB_BASE          0xB8068000
#define USBC              0xB8068000
#define USBS              0xB8068004
#define USBSM             0xB8068008
#define USBADDR           0xB806800C
#define USBFC             0xB8068010
#define USBEPAC           0xB8068020
#define USBEPAS           0xB8068024
#define USBEPASM          0xB8068028
#define USBEPATBS         0xB806802C
#define USBEPATD          0xB8068030
#define USBEPARD          0xB8068040
#define USBEPARBS         0xB8068050
#define USBEPBC           0xB8068060
#define USBEPBS           0xB8068064
#define USBEPBSM          0xB8068068
#define USBEPBBS          0xB806806C
#define USBEPBD           0xB8068070
#define USBEPCC           0xB8068080
#define USBEPCS           0xB8068084
#define USBEPCSM          0xB8068088
#define USBEPCBS          0xB806808C
#define USBEPCD           0xB8068090
#define USBEPDC           0xB80680A0
#define USBEPDS           0xB80680A4
#define USBEPDSM          0xB80680A8
#define USBEPDBS          0xB80680AC
#define USBEPDD           0xB80680B0
#define USBEPEC           0xB80680C0
#define USBEPES           0xB80680C4
#define USBEPESM          0xB80680C8
#define USBEPEBS          0xB80680CC
#define USBEPED           0xB80680D0
#define USBEPFC           0xB80680E0
#define USBEPFS           0xB80680E4
#define USBEPFSM          0xB80680E8
#define USBEPFBS          0xB80680EC
#define USBEPFD           0xB80680F0
#define USBEPGC           0xB8068100
#define USBEPGS           0xB8068104
#define USBEPGSM          0xB8068108
#define USBEPGBS          0xB806810C
#define USBEPGD           0xB8068110

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int usbepc;
  unsigned int usbeps;
  unsigned int usbepsm;
  unsigned int usbeptbs;
  unsigned int usbeptd;
  unsigned int _u0;
  unsigned int _u1;
  unsigned int _u2;
  unsigned int usbeprd;
  unsigned int _u3;
  unsigned int _u4;
  unsigned int _u5;
  unsigned int usbeprbs;
  unsigned int _u6;
  unsigned int _u7;
  unsigned int _u8;
} USB_EPA;

typedef struct {
  unsigned int usbepc;
  unsigned int usbeps;
  unsigned int usbepsm;
  unsigned int usbepbs;
  unsigned int usbepd;
  unsigned int _u0;
  unsigned int _u1;
  unsigned int _u2;
} USB_EP;

typedef struct {
  unsigned int usbc;
  unsigned int usbs;
  unsigned int usbsm;
  unsigned int usbaddr;
  unsigned int usbfc;
  unsigned int _u0;
  unsigned int _u1;
  unsigned int _u2;
  USB_EPA epa;
  USB_EP  epb;
  USB_EP  epc;
  USB_EP  epd;
  USB_EP  epe;
  USB_EP  epf;
  USB_EP  epg;
} USB;
 
#define usb (*((volatile USB *) USB_BASE))

#endif /*_ASMLANGUAGE*/

/********************************************************************
 * I2C interface
 ********************************************************************/
#define I2C_BASE          0xB8070000
#define I2CC              0xB8070000
#define I2CDI             0xB8070004
#define I2CDO             0xB8070008
#define I2CCP             0xB807000C
#define I2CMCMD           0xB8070010
#define I2CMS             0xB8070014
#define I2CMSM            0xB8070018
#define I2CSS             0xB807001C
#define I2CSSM            0xB8070020
#define I2CSADDR          0xB8070024
#define I2CSACK           0xB8070028
#define I2CDEBUG          0xB807002C

#if !defined(_ASMLANGUAGE)

typedef struct {
  unsigned int i2cc;
  unsigned int i2cdi;
  unsigned int i2cdo;
  unsigned int i2ccp;
  unsigned int i2cmcmd;
  unsigned int i2cms;
  unsigned int i2cmsm;
  unsigned int i2css;
  unsigned int i2cssm;
  unsigned int i2csaddr;
  unsigned int i2csack;
  unsigned int i2cdebug;
} I2C;
 
#define i2c (*((volatile I2C *) I2C_BASE))

#endif /*_ASMLANGUAGE*/


#endif /* __banyan__ */
