#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/vmalloc.h> /* for vmalloc */
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/errno.h>
#include <linux/atm.h>
#include <linux/atmdev.h>
#include <linux/sonet.h>
#include <linux/skbuff.h>
#include <linux/time.h>
#include <linux/timex.h>
#include <linux/sched.h> /* for xtime */
#include <linux/delay.h>
#include <linux/uio.h>
#include <linux/init.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/string.h>
#include <asm/byteorder.h>

#include <linux/netdevice.h>
#include <linux/ioport.h>
#include <linux/atm.h>

#include "sys_ss_def.h" 
#include "sar_if.h" 
#include "sardrv.h" 
#include "sar_linux_specific.h" 
#include "timer.h"
#include "banyan.h"

//extern void       softsar_detect(void);
extern void       appl_tx_hdlr_1(pATM_VC_st pVC, M_BLK_ID m);

extern void       s_init(UInt32 type);
extern void       s_vp_setup(UInt32 sp, UInt32 vpi, UInt32 mask, UInt32 bit_pos) ;
extern pATM_VC_st s_open_vc(UInt32 sp, UByte vcch, UByte fm, struct atm_vcc *vcc);
extern void s_close_vc(UInt32 sp, UInt32 vpi, UInt32 vci);
extern int        add_timer_entry(pATM_VC_st , UInt32 );

static pATM_VC_st pVC;

static int  banyansar_open(struct atm_vcc *vcc,short vpi,int vci);
static void banyansar_close(struct atm_vcc *vcc);
static int  banyansar_getsockopt(struct atm_vcc *vcc,int level,int optname, void *optval,int optlen);
static int  banyansar_setsockopt(struct atm_vcc *vcc,int level,int optname, void *optval,int optlen);
static int  banyansar_send(struct atm_vcc *vcc,struct sk_buff *skb);
static int  banyansar_sg_send(struct atm_vcc *vcc,unsigned long start, unsigned long size);
static int  banyansar_ioctl(struct atm_dev *dev, unsigned int cmd,void *arg);

/* operations registered by the atm devices */
const struct atmdev_ops banyansar_ops =
  {
    open:        banyansar_open,
    close:       banyansar_close,
    getsockopt:  banyansar_getsockopt,
    setsockopt:  banyansar_setsockopt,
    send:        banyansar_send,
    //    sg_send:     banyansar_sg_send,
    ioctl:       banyansar_ioctl,
  };

static int  banyansar_open(struct atm_vcc *vcc,short vpi,int vci)
{
  int result;

  /* Make sure we are opening an AAL5 connection */
  if ((vcc->qos.aal != ATM_AAL5))
    {   
      printk(KERN_WARNING "Invalid AAL\n");
      return -EINVAL;
    }


  /* Fill in any VPI/VCI wildcards and check VPI/VCI is not in use */
  if ((result = atm_find_ci(vcc, &vpi, &vci)))
    {   
      printk(KERN_WARNING "Error in atm_find_ci()\n");
      return result;
    }

  // Address is in use
  set_bit(ATM_VF_ADDR, &vcc->flags.bits);

  /* Assign VPI and VCI in both atm_vcc and cpm_channel_t */
  vcc->vpi = vpi;
  vcc->vci = vci;
  pVC = s_open_vc(0,1,1,vcc);
  vcc->dev_data = pVC;

  /* Connection is now ready to receive data */
  set_bit(ATM_VF_READY, &vcc->flags);

  return 0;
}

static void banyansar_close(struct atm_vcc *vcc)
{
  s_close_vc(0, vcc->vpi, vcc->vci);
  // Indicate channel closed
  clear_bit(ATM_VF_READY, &vcc->flags);

}

static int banyansar_getsockopt(struct atm_vcc *vcc,int level,int optname,
				void *optval,int optlen)
{
  printk(__FUNCTION__ "\n");
  return -EINVAL;
}

static int banyansar_setsockopt(struct atm_vcc *vcc,int level,int optname,
				void *optval,int optlen)
{
  printk(__FUNCTION__ "\n");
  return -EINVAL;
}

static int banyansar_send(struct atm_vcc *vcc,struct sk_buff *skb)
{
  
  ATM_SKB(skb)->vcc = vcc;
  banyan_atm_ioctl(vcc->dev_data,SARIO_XMIT, (struct sk_buff *)skb, 
		   vcc->qos.aal, (void *)appl_tx_hdlr_1, 0, NULL);
  return(0);
}


static int banyansar_sg_send(struct atm_vcc *vcc,unsigned long start, unsigned long size)
{

  printk(__FUNCTION__ "\n");
  return vcc->qos.aal == ATM_AAL5 && !((start | size) & 3);
  /* don't tolerate misalignment */
}
static int  banyansar_ioctl(struct atm_dev *dev, unsigned int cmd,void *arg)
{
  return -EINVAL;
}

int softsar_detect()
{
  struct atm_dev *banyansar_dev;
#ifdef CONFIG_IDT_79EB355
  s_init(1);
#else
  s_init(4);
#endif
  s_vp_setup(0,0,0x3f,0);
  s_vp_setup(0,8,0x3f,0);
  banyansar_dev = atm_dev_register("banyansar",&banyansar_ops,-1,0);
  banyansar_dev->ci_range.vpi_bits = 8;
  banyansar_dev->ci_range.vci_bits = 10;
  return 0;
}
