/*
*******************************************************************************
**	Copyright 1999, 2000 Integrated Device Technology, Inc.
**	All Rights Reserved
**
*******************************************************************************
*/
/*
******************************************************************************
** dma.h -- data definitions for Banyan dma interface.
**
**	Portions of this header file were taken from our partner AT&T in the
**	Banyan project, and has been left as is or has been modified.
**
**	For pusposes of reference following is the banner from the AT&T header
**	file dma.h and has been reproduced exactly as I found it.
**						-- hak, IDT	06-15-2000.

******************************************************************************
*/
/*****************************************************************************
 *     $Id: dma.h,v 1.6 2000/03/22 21:05:33 pzo Exp $
 *
 *     Description:  Banyan DMA Header File
 *     Author:       Peter Z. Onufryk
 *     Company:      AT&T
 *
 *     $Log: dma.h,v $
 *     Revision 1.6  2000/03/22 21:05:33  pzo
 *     Added DMA channel device select definitions.
 *
 *     Revision 1.5  1999/10/26 14:21:46  pzo
 *     Made DMA_DESCRIPTOR volatile so that C optimizer will not remove
 *     reads and writes.
 *
 *     Revision 1.4  1999/09/27 21:55:40  pzo
 *     added macro to convert from virtual to physical address
 *
 *     Revision 1.3  1999/09/27 15:41:07  raju
 *     Corrected dma.h, the comment for Mode field of DMAXC register, before it was DM bit.
 *
 *     Revision 1.2  1999/09/27 15:27:19  pzo
 *     added definitions for DMA mode field in DMAxC register
 *
 *     Revision 1.1  1999/09/24 19:44:14  pzo
 *     added header files to cvs
 *
 *
 *****************************************************************************/

#ifndef __dma__
#define __dma__

/******************************************************************************
 * macro to convert from virtual to physical address  
 *****************************************************************************/
#define PHYS_ADDR(x)      (x & 0x1fffffff) 

/******************************************************************************
 * DMA descriptor
 *****************************************************************************/

#define DMA_FINISHED      (1<<31)             /* F bit                       */
#define DMA_DONE          (1<<30)             /* D bit                       */
#define DMA_TERMINATED    (1<<29)             /* T bit                       */

#define DMA_IOD           (1<<28)             /* IOD bit                     */
#define DMA_IOF           (1<<27)             /* IOF bit                     */
#define DMA_COD           (1<<26)             /* COD bit                     */
#define DMA_COF           (1<<25)             /* COF bit                     */

#define DMA_DEVCMD(v)     ((v&0x7)<<22)       /* DEVCMD field                */
#define DMA_DS(v)         ((v&0x3)<<20)       /* DS field                    */
#define DMA_COUNT(v)      (v&0x3ffff)         /* COUNT field                 */

typedef volatile struct dma_descriptor {
  unsigned int control;
  unsigned int ca;
  unsigned int devcs;
  volatile struct dma_descriptor *link;
} DMA_DESCRIPTOR;

/******************************************************************************
 * DMA descriptor device selects
 *****************************************************************************/

/* channel 0 */
#define DS_ATM_0_INPUT         0
#define DS_ATM_1_INPUT         1
#define DS_ATM_0_OUTPUT        2
#define DS_ATM_1_OUTPUT        3

/* channel 1 */
#define DS_ATM_VCC_1           0

/* channel 2 */
#define DS_ATM_VCC_2           0

/* channel 3 */
#define DS_ATM_VCC_3           0

/* channel 4 */
#define DS_ATM_VCC_4           0

/* channel 5 */
#define DS_ATM_VCC_5           0

/* channel 6 */
#define DS_ATM_VCC_6           0
#define DS_MEM2FIFO            1

/* channel 7 */
#define DS_ATM_VCC_7           0
#define DS_ATM_FMC0            1

/* channel 8 */
#define DS_ATM_VCC_8           0
#define DS_ATM_FMC1            1

/* channel 9 */
#define DS_ETHERNET_INPUT      0

/* channel 10 */
#define DS_ETHERNET_OUTPUT     0

/* channel 11 */
#define DS_TDM_INPUT           0

/* channel 12 */
#define DS_TDM_OUTPUT          0

/* channel 13 */
#define DS_USB_INPUT           0

/* channel 14 */
#define DS_USB_OUTPUT          0

/* channel 15 */
#define DS_EXT_DMA             0
#define DS_FIFO2MEM            1

/******************************************************************************
 * DMAxC register
 *****************************************************************************/

#define DMAxC_RUN         (1<<0)              /* RUN bit                     */
#define DMAxC_DM          (1<<1)              /* DM bit                      */
#define DMAxC_MODE(v)     ((v&0x3)<<2)        /* MODE field                  */

/* DMA modes */
#define DMA_AUTO_REQUEST     0
#define DMA_BURST_REQUEST    1
#define DMA_TRANSFER_REQUEST 2

/******************************************************************************
 * DMAxS register
 *****************************************************************************/

#define DMAxS_F           (1<<0)              /* F bit                       */
#define DMAxS_D           (1<<1)              /* D bit                       */
#define DMAxS_C           (1<<2)              /* C bit                       */
#define DMAxS_E           (1<<3)              /* E bit                       */
#define DMAxS_H           (1<<4)              /* H bit                       */

/******************************************************************************
 * dma_init_blk - Dma channel initialization block.
 *****************************************************************************/
typedef struct	dma_init_blk {
	unsigned chnl_id;	/* dma channel id */
	unsigned mode;
	unsigned dm;
	unsigned h_flag;
	unsigned e_flag;
	unsigned c_flag;
	unsigned d_flag;
	unsigned f_flag;
} dma_init_blk;
/******************************************************************************
 * dma_descr_info - Dma Descriptor initialization information.
 *****************************************************************************/
typedef struct	dma_descr_info {
	unsigned iod;
	unsigned iof;
	unsigned cod;
	unsigned cof;
	unsigned devcmd;
	unsigned ds;
	unsigned count;
	unsigned ca;
	unsigned devcs;
	unsigned link;
} dma_descr_info;

#endif /* __dma__ */
/* -- eof -- */
