/*****************************************************************************
 *     $Id: interrupt.h,v 1.4 2000/03/16 20:28:31 pzo Exp $
 *
 *     Description:  Banyan Interrupt Controller Header File
 *     Author:       Peter Z. Onufryk
 *     Company:      AT&T
 *
 *     $Log: interrupt.h,v $
 *     Revision 1.4  2000/03/16 20:28:31  pzo
 *     Added support for interrupt bit masks from interrupt vectors.
 *
 *     Revision 1.3  2000/01/14 21:31:54  pzo
 *     Updated header file to support the three interrupts that were added from
 *     the UART and moved the USB interrutps.
 *
 *     Revision 1.2  2000/01/12 22:55:25  licwang
 *     Correct errors to add underflow0 and underflow1
 *
 *     Revision 1.1  1999/09/30 21:24:21  pzo
 *     Added macros to support interrupts and CP0 operations
 *
 *
 *****************************************************************************/

#ifndef _interrupt
#define _interrupt


/* indices for interupt vectors */

#define COUNTER_TIMER_0            (32*0 + 0)
#define COUNTER_TIMER_1            (32*0 + 1)
#define COUNTER_TIMER_2            (32*0 + 2)
#define REFRESH_TIMER              (32*0 + 3)
#define WATCHDOG_TIMER             (32*0 + 4)
#define UNDECODED_CPU_WRITE        (32*0 + 5)

#define DMA_CHANNEL_0              (32*1 + 0)
#define DMA_CHANNEL_1              (32*1 + 1)
#define DMA_CHANNEL_2              (32*1 + 2)
#define DMA_CHANNEL_3              (32*1 + 3)
#define DMA_CHANNEL_4              (32*1 + 4)
#define DMA_CHANNEL_5              (32*1 + 5)
#define DMA_CHANNEL_6              (32*1 + 6)
#define DMA_CHANNEL_7              (32*1 + 7)
#define DMA_CHANNEL_8              (32*1 + 8)
#define DMA_CHANNEL_9              (32*1 + 9)
#define DMA_CHANNEL_10             (32*1 + 10)
#define DMA_CHANNEL_11             (32*1 + 11)
#define DMA_CHANNEL_12             (32*1 + 12)
#define DMA_CHANNEL_13             (32*1 + 13)
#define DMA_CHANNEL_14             (32*1 + 14)
#define DMA_CHANNEL_15             (32*1 + 15)

#define ATM_0_INPUT                (32*2 + 0)
#define ATM_1_INPUT                (32*2 + 1)
#define ATM_0_OUTPUT               (32*2 + 2)
#define ATM_1_OUTPUT               (32*2 + 3)
#define ATM_0_OVERFLOW             (32*2 + 4)
#define ATM_1_OVERFLOW             (32*2 + 5)
#define ATM_0_UNDERFLOW            (32*2 + 6)
#define ATM_1_UNDERFLOW            (32*2 + 7)
#define ATM_0_DISCARD              (32*2 + 8)
#define ATM_1_DISCARD              (32*2 + 9)

#define TDM_INPUT                  (32*3 + 0)
#define TDM_OUTPUT                 (32*3 + 1)
#define TDM_STATUS                 (32*3 + 2)
#define ETHERNET_INPUT             (32*3 + 3)
#define ETHERNET_OUTPUT            (32*3 + 4)
#define PAUSE_FRAME_DONE           (32*3 + 5)
#define USB_ENDPOINT_A             (32*3 + 6)
#define USB_ENDPOINT_B             (32*3 + 7)
#define USB_ENDPOINT_C             (32*3 + 8)
#define USB_ENDPOINT_D             (32*3 + 9)
#define USB_ENDPOINT_E             (32*3 + 10)
#define USB_ENDPOINT_F             (32*3 + 11)
#define USB_ENDPOINT_G             (32*3 + 12)
#define USB_STATUS                 (32*3 + 13)
#define UART_0_GENERAL             (32*3 + 14)
#define UART_0_TXRDY               (32*3 + 15)
#define UART_0_RXRDY               (32*3 + 16)
#define UART_1_GENERAL             (32*3 + 17)
#define UART_1_TXRDY               (32*3 + 18)
#define UART_1_RXRDY               (32*3 + 19)
#define I2C_MASTER                 (32*3 + 20)
#define I2C_SLAVE                  (32*3 + 21)

#define GPIO_0_INTERRUPT           (32*4 + 0)
#define GPIO_1_INTERRUPT           (32*4 + 1)
#define GPIO_2_INTERRUPT           (32*4 + 2)
#define GPIO_3_INTERRUPT           (32*4 + 3)
#define GPIO_4_INTERRUPT           (32*4 + 4)
#define GPIO_5_INTERRUPT           (32*4 + 5)
#define GPIO_6_INTERRUPT           (32*4 + 6)
#define GPIO_7_INTERRUPT           (32*4 + 7)
#define GPIO_8_INTERRUPT           (32*4 + 8)
#define GPIO_9_INTERRUPT           (32*4 + 9)
#define GPIO_10_INTERRUPT          (32*4 + 10)
#define GPIO_11_INTERRUPT          (32*4 + 11)
#define GPIO_12_INTERRUPT          (32*4 + 12)
#define GPIO_13_INTERRUPT          (32*4 + 13)
#define GPIO_14_INTERRUPT          (32*4 + 14)
#define GPIO_15_INTERRUPT          (32*4 + 15)
#define GPIO_16_INTERRUPT          (32*4 + 16)
#define GPIO_17_INTERRUPT          (32*4 + 17)
#define GPIO_18_INTERRUPT          (32*4 + 18)
#define GPIO_19_INTERRUPT          (32*4 + 19)
#define GPIO_20_INTERRUPT          (32*4 + 20)
#define GPIO_21_INTERRUPT          (32*4 + 21)
#define GPIO_22_INTERRUPT          (32*4 + 22)
#define GPIO_23_INTERRUPT          (32*4 + 23)
#define GPIO_24_INTERRUPT          (32*4 + 24)
#define GPIO_25_INTERRUPT          (32*4 + 25)
#define GPIO_26_INTERRUPT          (32*4 + 26)
#define GPIO_27_INTERRUPT          (32*4 + 27)
#define GPIO_28_INTERRUPT          (32*4 + 28)
#define GPIO_29_INTERRUPT          (32*4 + 29)
#define GPIO_30_INTERRUPT          (32*4 + 30)
#define GPIO_31_INTERRUPT          (32*4 + 31)


#define interrupt_bit(x)           (1<<(x&0x1f))

#endif /* _interrupt */


