/*****************************************************************
**  File : phy_77v1053.c
**
**		API to access the IDT's 77V1053 3-port ATM PHY.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  10/31/2000   srinivas voleti           Original
**
******************************************************************/
#define S355
#include "sys_ss_def.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"
#include "sar_if.h"
#include "sardrv.h"
#include "phy_77v1053.h"

/******************************************************************
** FUNCTION : init_atm_phy
**
** DESCRIPTION:
**  Initialize specified ATM PHY register.
**
** NOTE:
**
******************************************************************/
void init_atm_phy()
{
#ifdef S355

	/* Initialize the device controller registers for the ATM PHY */
	device.dev4.devbase = 0x14000000;
	device.dev4.devmask = 0xffff0000;
	device.dev4.devc    = 0x0ffffff4;
	device.dev4.devtc   = 0x00001fff;
#endif /* S355 */

} /* init_atm_phy */

/******************************************************************
** FUNCTION : get_atm_phy
**
** DESCRIPTION:
**  Read the contents of the specified ATM PHY register.
**
**  Input Parameters:
**      sp      : subport number (0,1 or 2)
**      offset  : Register offset
**
**  Return Value : Value of the register.
**
** NOTE:
**
******************************************************************/
UInt16 get_atm_phy(UByte sp, UByte offset)
{
#ifdef S355
	UByte   *addr;

	if (offset == ATM_PHY_REF_CTRL_REG)
		addr = (UByte *)((UInt32 )ATM_PHY_BASE + ATM_PHY_REF_CTRL_REG);
	else
		addr = (UByte *)((UInt32 )ATM_PHY_BASE + (sp << 4) + offset);

	return (UInt16 )*addr;
#else
	return 0;
#endif /* S355 */

} /* get_atm_phy */

/******************************************************************
** FUNCTION : set_atm_phy
**
** DESCRIPTION:
**  Write a value to the specified ATM PHY register.
**
**  Input Parameters:
**      sp      : subport (0, 1 or 2)
**      offset  : Register offset
**      val     : Value to be written
**
** NOTE:
**
******************************************************************/
UInt16 set_atm_phy(UByte sp, UByte offset, UByte val)
{
#ifdef S355
	UByte   *addr;

	if (offset == ATM_PHY_REF_CTRL_REG)
		addr = (UByte *)((UInt32 )ATM_PHY_BASE + ATM_PHY_REF_CTRL_REG);
	else
		addr = (UByte *)((UInt32 )ATM_PHY_BASE + (sp << 4) + offset);

	*addr = val;

#endif /* S355 */
	return TRUE;
} /* set_atm_phy */

