/*****************************************************************
**  File : sar_api.c
**
**		ATM SOFTSAR API.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  08/16/2000   srinivas voleti           Original
**  01/05/2001	 srinivas voleti           Changed some function prototypes.
**    09/2001    P. Sadik                 Ported to linux
******************************************************************/

#include <linux/string.h>
#include <linux/kernel.h>

#include "sys_ss_def.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"
#include "sar_if.h"
#include "sardrv.h"

void oam_cell_gen();
void oam_cell_rx();
void init_atm_phy(void );
int get_atm_phy();
int set_atm_phy();

/* Handlers used for the SAR ioctl functions. This table is indexed by the 
** command value passed to the sar_ioctl API. The number of entries in the
** table is computed at the time of initialization.
**
** Note:
** 		The order of the functions is important. The order is in accordance
**		with the command value for the function i.e. 
**		Command 0 handler is sar_vpt_block_setup & hence the first entry
**		Command 1 handler is sar_vpt_block_purge & hence the second entry
**		and so on....
*/
int sar_vpt_setup(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_vpt_purge(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_conn_setup(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_conn_terminate(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_purge_all_vcs(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_set_if_admin_state(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_get_if_oper_state(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_get_per_vc_stats(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_get_per_if_vc_stats(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_get_if_err_stats(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_clr_if_stats(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_clr_vc_stats(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_read_reg(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_write_reg(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_read_phy(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_write_phy(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_dump_atmif_struct(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_dump_vc_struct(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_dump_status_struct(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_xmit(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);
int sar_recv(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);

int get_vc_handle(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);

int sar_sched_conn(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);

int sar_desched_conn(void *handle, void *ptr1, UInt32 val1, void *ptr2, UInt32 val2, UInt32 *var);    

static int	(*sar_ioctl_hdlr[])() = 
{
  sar_vpt_setup, 		/* SARIO_VPT_SETUP */
  sar_vpt_purge, 		/* SARIO_VPT_PURGE */
  sar_conn_setup, 		/* SARIO_VC_SETUP */
  sar_conn_terminate, 		/* SARIO_VC_TEARDOWN */
  sar_purge_all_vcs, 		/* SARIO_PURGE_VCS */
  sar_set_if_admin_state, 	/* SARIO_SET_STATE */
  sar_get_if_oper_state, 	/* SARIO_GET_STATE */
  sar_get_per_vc_stats, 	/* SARIO_VC_VC_STATS */
  sar_get_per_if_vc_stats, 	/* SARIO_IF_VC_STATS */
  sar_get_if_err_stats, 	/* SARIO_ERR_STATS */
  sar_clr_if_stats, 		/* SARIO_IF_CLR_STATS */
  sar_clr_vc_stats, 		/* SARIO_VC_CLR_STATS */
  sar_read_reg, 		/* SARIO_REG_READ */
  sar_write_reg, 		/* SARIO_REG_WRITE */
  sar_read_phy, 		/* SARIO_PHY_READ */
  sar_write_phy, 		/* SARIO_PHY_WRITE */
  sar_dump_atmif_struct, 	/* SARIO_GET_ATMIF_ST */
  sar_dump_vc_struct, 		/* SARIO_GET_VC_ST */
  sar_dump_status_struct, 	/* SARIO_GET_STATUS_ST */
  sar_xmit,			/* SARIO_XMIT */
  sar_recv,			/* SARIO_RECV */
  get_vc_handle,		/* SARIO_GET_VC_HANDLE */

  sar_sched_conn,               /* SARIO_SCHED_CONN */  
  sar_desched_conn,             /* SARIO_DESCHED_CONN */
  0				/* This MUST be the last item in the table. DO NOT DELETE. */
};

static	UInt32	ioctl_num_hdlrs = 0;

extern UInt32 atm_config_type;
extern ATMIF_CONFIG_st if_config[];

/******************************************************************
** FUNCTION : atm_register
**
** DESCRIPTION:
**	Initialize the SOFTSAR device driver.
**
** NOTE:
**	A "handle" is returned to the calling routine. This handler is 
**	passed for all the API calls for using the SOFTSAR.
**
******************************************************************/
void *atm_register()
{
  void *handle;

  handle = (void *)sardrv_init(&if_config[atm_config_type]);
  if (handle == NULL)
    {
      printk("ATM driver installation failed \n");
      return handle;
    }

  /* Compute the number of entries in the ioctl handler table */
  while (sar_ioctl_hdlr[ioctl_num_hdlrs]) 
    ioctl_num_hdlrs++;

  return handle;

} /* atm_register */

/******************************************************************
** FUNCTION : atm_unregister
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
void atm_unregister(void *handle)
{
	sardrv_shutdown(handle);

} /* atm_unregister */

/******************************************************************
** FUNCTION : atm_open
**
** DESCRIPTION:
**	NULL function.
**
** NOTE:
**
******************************************************************/
void atm_open()
{

} /* atm_open */

/******************************************************************
** FUNCTION : atm_close
**
** DESCRIPTION:
**	NULL function.
**
** NOTE:
**
******************************************************************/
void atm_close()
{

} /* atm_close */

/******************************************************************
** FUNCTION : atm_send
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
int atm_send()
{

  return OK;

} /* atm_send */

/******************************************************************
** FUNCTION : atm_receive
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
int atm_receive()
{
  return OK;

} /* atm_receive */

/******************************************************************
** FUNCTION : banyan_atm_ioctl
**
** DESCRIPTION:
**	The command is used to index into the ioctl handler table. The
**	handler is then invoked and the parameters are passed.
**
** NOTE:
**
******************************************************************/
int banyan_atm_ioctl(void *handle, UInt32 cmd, void *ptr1, UInt32 val1, \
	      void *ptr2, UInt32 val2, UInt32 *var)
{
  int     status;   
	if (cmd >= ioctl_num_hdlrs)
		return ERROR;

	status = (*sar_ioctl_hdlr[cmd])(handle, ptr1, val1, ptr2, val2, var);  

	return status;
} /* banyan_atm_ioctl */

