/***********************************************************
** FILE NAME: sar_conn.c
**
**   Software SAR - Connection Management utilities
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
************************************************************************
** DATE      DEVELOPER                 REASON
************************************************************************
** 08/16/00  srinivas voleti           original
** 12/21/00  srinivas voleti	       Modified to run the code cached
** 09/01  P. Sadik                     Ported to linux
************************************************************************/

#include <linux/string.h>
#include <linux/slab.h>
#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"

extern UNALIGNED_MEM_st    umem; 
extern void taskDelay(unsigned long);
/******************************************************************
** FUNCTION : sar_conn_setup()
**
** DESCRIPTION:
**	This function sets up a VC. 
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Pointer to the VC configuration structure.
**		ptr2	: Connection handle (updated by this function).
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.
**
** NOTE:
**
******************************************************************/
int sar_conn_setup(void *handle, void *ptr1, UInt32 val1, \
						void *ptr2, UInt32 val2, UInt32 *var)
{
  pATMIF_st		pATM = (pATMIF_st )handle;
  pVC_CONFIG_st	pconf = (pVC_CONFIG_st )ptr1;
  pATM_VC_st		pVC;
#if defined (TWO_LEVEL_LOOKUP)
  pL1_LKUP_st		pVPT;
  UInt32			sp_bits;
#endif /* TWO_LEVEL_LOOKUP */
  pDMA_st			pdesc = (pDMA_st )NULL;
  DMA_CHAN		*pdma_channel;
  UInt32			*patm_channel, dummy_read;
  UInt32			index, cell_hdr, channel = 0;
  unsigned long key;

  /* Initialize the return pointer */
  *(pATM_VC_st *)ptr2 = 0;

  /* Free VC head pointer will be non-NULL if the number of active
     VCCs is less than the maximum number of VCCs.
	*/
  if (pATM->free_vc_list_head)
    {

      /* Build cell header */
      cell_hdr = BLD_CELL_HDR(pconf->subport, pconf->vpi, \
			      pconf->vci, 0, 0, 0);

#if defined (INDEXED_LOOKUP)
      /* "Indexed Lookup Mechanism" is being used */

      /* Get the offset in the VC table */
      GET_VPCI(cell_hdr, pATM->vpi_mask, pATM->vci_mask, \
	       pATM->vpi_shft, &index);

      if (index > pATM->max_VCCs)
	{
	  printk("VC value excceded maximum range\n");
	  return ERROR;
	}

      /* Validate the connection */
      if (pATM->vc_table[index])
	{
	  printk("Connection entry already set up\n");
	  return ERROR;
	}
#else
      /* "2-Level Lookup Mechanism" is being used */
      sp_bits = pATM->status & ATMIF_SP_BITS_MASK;
      index = ((pconf->vpi << sp_bits) | pconf->subport);
      pVPT = &(pATM->vp_table[index]);
      if (pVPT->control & VPLKUP_VALID)
	{
	  /* Get the offset in the VC table */
	  index = pVPT->vc_mask & pconf->vci;
	  if (pVPT->pL2_lkup[index])
	    {
	      printk("Connection entry already set up\n");
	      return ERROR;
	    }
	}
      else
	{
	  /* VP Table not configured */
	  printk("VP table entry not configured\n");
	  return ERROR;
	}
#endif /* INDEXED_LOOKUP */
	
      /* Configure the VC entry */
      pVC		= pATM->free_vc_list_head;
      pVC->subport	= pconf->subport;
      pVC->vpi		= pconf->vpi;
      pVC->vci		= pconf->vci;
      pVC->cell_hdr	= cell_hdr;
      pVC->handle	= pVC;
      pVC->tx_hdlr	= pconf->tx_hdlr;
      pVC->rx_hdlr	= pconf->rx_hdlr;
      pVC->pif		= pATM;
      pVC->proto	= pconf->proto;

      pVC->tx_pdu_desc = 0;
      pVC->dma0_txpend = 0; 

      pVC->control = (pconf->control & ~CHAN_MASK);
      if (pconf->control & CHAN_MASK)
	{
	  /* Fetch a free VC Cache entry */
	  channel = atm.atmvccfe;
	  pVC->control |= (channel << CHAN_BIT_POS);
	}

      /* If VC Cache channels are used for receive flow */

      if (channel)
	{
	  save_and_cli(key);
	  pdesc = get_2k_buf_desc_list(pATM, pVC, NUM_BDS_PER_LIST);
	  restore_flags(key);

	  if (pdesc == (pDMA_st )NULL)
	    {
	      printk("No descriptor-buffer resources\n");
	      return ERROR;
	    }
	  else
	    {
	      pATM->curr_desc[channel] = pdesc;
	      pATM->pVC_array[channel] = pVC; 
	    }

	  /* Set up the VC cache entry */
	  patm_channel = (UInt32 *)(&atm.atmvcch1 + channel - 1);
	  *patm_channel = pVC->cell_hdr;
	}
      else
	{
	  save_and_cli(key);
	  if ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE)
	    {
	      pVC->rx_data_desc = get_2k_buf_desc_list(pATM, pVC, 1);
	    }
	  else
	    {
	      pVC->rx_data_desc = get_64_buf_desc_list(pATM, pVC);
	    }
	  restore_flags(key);

	  if (pVC->subport)
	    {
	      pVC->control |= (DS_ATM_1_INPUT << RCV_DEVCS_POS);
	    }
	  else
	    {
	      pVC->control |= (DS_ATM_0_INPUT << RCV_DEVCS_POS);
	    }  
	}

      save_and_cli(key);
      pVC->rx_oam_desc = get_64_buf_desc_list(pATM, pVC);
      restore_flags(key);

      if (pVC->subport)
	{
	  if (pVC->control & FRAME_MODE)
	    {
	      pVC->control |= (DS_ATM_FMC0 << XMIT_DEVCS_SHFT);
	      pVC->pad_buffer = pATM->pad_buffer2;
	    }
	  else
	    {
	      pVC->control |= (DS_ATM_1_OUTPUT << XMIT_DEVCS_SHFT);
	      pVC->pad_buffer = pATM->pad_buffer0;
	    }
	}
      else
	{
	  if (pVC->control & FRAME_MODE)
	    {
	      pVC->control |= (DS_ATM_FMC0 << XMIT_DEVCS_SHFT);
	      pVC->pad_buffer = pATM->pad_buffer1;
	    }
	  else
	    {
	      pVC->control |= (DS_ATM_0_OUTPUT << XMIT_DEVCS_SHFT);
	      pVC->pad_buffer = pATM->pad_buffer0;
	    }
	}

      pVC->state			|= VPCI_OPER_UP;
      pATM->active_VCCs++;

      /* Update the return pointer with the connection handle */
      *(pATM_VC_st *)ptr2 = pVC;

      /* Update the free VC list */
      pATM->free_vc_list_head = pVC->next;
      if (pATM->free_vc_list_head == (pATM_VC_st )NULL)
	pATM->free_vc_list_tail = (pATM_VC_st )NULL;
      pVC->next			= 0;

      save_and_cli(key);
#if defined (INDEXED_LOOKUP)
      /* Update the VC table pointer table */
      pATM->vc_table[index] = pVC;
      pATM->registered_VCCs++;
#else
      /* Update the L2 lookup entry table */
      pVPT->pL2_lkup[index] = pVC;
#endif /* INDEXED_LOOKUP */
      if (channel)
	{
	  /* Enable the DMA channel */
	  pdma_channel = (DMA_CHAN *)((UInt32 )&dma.ch0 + \
				      channel * sizeof(DMA_CHAN));
	  dummy_read = pdma_channel->dmandptr;
	  pdma_channel->dmandptr = VIRT_TO_PHYS((UInt32 )pdesc);
	}
      restore_flags(key);

      return OK;
    }
  else
    {
      printk("ERROR: Exceeded maximum number of VCs\n");
      return ERROR;
    }

} /* sar_conn_setup */

/******************************************************************
** FUNCTION : sar_conn_terminate()
**
** DESCRIPTION:
**	Close the VC.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Pointer to connection handle.
**		ptr2	: Not used.
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.
**
** NOTE:
**
******************************************************************/
int sar_conn_terminate (void *handle, void *ptr1, UInt32 val1, \
                        void *ptr2, UInt32 val2, UInt32 *var)
{
	volatile pATMIF_st	pATM = (pATMIF_st )handle;
	pATM_VC_st			pVC = (pATM_VC_st )ptr1;
	pDMA_st				pdesc, pnext;
	volatile DMA_CHAN	*pdma_channel;
	UInt32				chan, evt_mask, index;
	UInt32				*patm_channel, dummy_read;
#if defined (TWO_LEVEL_LOOKUP)
	pL1_LKUP_st			pVPT;
	UInt32				sp_bits, vci;
#endif /* TWO_LEVEL_LOOKUP */
	unsigned long key;

	/* Reset the administrative status */
	pVC->control &= ~VPCI_ADMIN_UP;

	chan = (pVC->control & CHAN_MASK) >> CHAN_BIT_POS;

	/* Does it use VCC entry */
	if (chan)
	{
		/* VC Cache entry is used */

		/* Halt the DMA channel */
		pdma_channel = (DMA_CHAN *)((UInt32 )&dma.ch0 + \
									chan * sizeof(DMA_CHAN));

		save_and_cli(key);
		pdma_channel->dmac &= ~DMAxC_RUN;

		/* Wait for the HALT bit in the DMA status register to be set */
		while(!(pdma_channel->dmas & DMAxS_H));

		/* Disable the VC Cache entry so that it is not re-allocated
		   while maintenance is in progress */
		atm.atmvccd |= (1 << (chan - 1));
		restore_flags(key);

		dummy_read = pdma_channel->dmandptr;
		pdma_channel->dmandptr = 0;

		/* Wait for the status task to finish processing its "DONE" 
		   descriptors. This is indicated by the setting of the appropriate
		   bit in the DMA EVENT status bit of the ATM interface data 
		   structure. */
		evt_mask = (1 << (chan + ATM_DMA_EVT_SHFT));
		while (!(pATM->status & evt_mask))
			taskDelay(1);

		pATM->pVC_array[chan] = 0;   
		/* Free the buffer-descriptor pair to the free 2K BD pool. The first
		   descriptor pointer is present in the curr_desc field of the ATM
		   interface data structure. 
		*/
		pdesc = pATM->curr_desc[chan];

		while((VIRT_TO_PHYS((UInt32 )pdesc)))
		{
			if (pdesc->link)
				pnext = (pDMA_st )GET_NEXT_DESC(pdesc);
			else
				pnext = pdesc->chain_desc;

			save_and_cli(key);
			free_2k_buf_desc(pATM, pdesc);
			restore_flags(key);
			pdesc = pnext;
		}

		/* Clear the cell header in the VC cache entry */
		patm_channel = (UInt32 *)(&atm.atmvcch1 + (chan - 1));
		*patm_channel = 0;

		/* Re-enable the VC Cache entry for future use */
		atm.atmvccd &= ~(1 << (chan - 1));

	}
	else
	{
		/* Non-VC Cache entry */
		if ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE)
		{
			save_and_cli(key);
			free_2k_buf_desc(pATM, pVC->rx_data_desc);
			restore_flags(key);
		}
		else
		{
			save_and_cli(key);
			free_64_buf_desc(pATM, pVC->rx_data_desc);
			restore_flags(key);
		}
	}

	save_and_cli(key);
	free_64_buf_desc(pATM, pVC->rx_oam_desc);
	restore_flags(key);

#if defined (INDEXED_LOOKUP)
	/* Get the offset in the VC table */
	GET_VPCI(pVC->cell_hdr, pATM->vpi_mask, pATM->vci_mask, \
										pATM->vpi_shft, &index);
#else
	sp_bits = pATM->status & ATMIF_SP_BITS_MASK;
	index = ((pVC->vpi << sp_bits) | pVC->subport);
	pVPT = &(pATM->vp_table[index]);
	vci = pVC->vci;
#endif /* INDEXED_LOOKUP */
	
	/* Clear the VC structure */
	bzero((char *)pVC, sizeof(ATM_VC_st ));

	/* Update the free VC list */
	if (pATM->free_vc_list_tail)
		pATM->free_vc_list_tail->next = pVC;
	else
		pATM->free_vc_list_head = pVC;
	pATM->free_vc_list_tail = pVC;

	/* Update the active VCCs count */
	pATM->active_VCCs--;

#if defined (INDEXED_LOOKUP)
	/* Remove entry from the VC table */
	pATM->vc_table[index] = 0;
	pATM->registered_VCCs--;
#else
	/* Remove entry from the Level 2 Lookup table */
	pVPT->pL2_lkup[vci] = 0;
#endif /* INDEXED_LOOKUP */

	return OK;
	
} /* sar_conn_terminate */

/******************************************************************
** FUNCTION : sar_vpt_setup()
**
** DESCRIPTION:
**	This function is used to setup the VPI address lookup table.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Pointer to the VP configuration structure.
**		ptr2	: Not used.
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.

** NOTE:
**	If the connection management uses the 2-level lookup mechanism 
**	it is mandatory for the user to invoke this function before 
**	setting up any connections.
**
******************************************************************/
int sar_vpt_setup(void *handle, void *ptr1, UInt32 val1, \
		  void *ptr2, UInt32 val2, UInt32 *var)
{
  pATMIF_st		pATM = (pATMIF_st)handle;
  pL1_LKUP_st	pVPT;
  pVPT_CONFIG_st	pconf = (pVPT_CONFIG_st )ptr1;
  UInt32			num_vcs, val, sp_bits;
  unsigned long key;
	
  /* Return ERROR if "Indexed Lookup Mechanism" is being used */
  if (pATM->control & INDEXED_LKUP_EN)
    {
      printk("ERROR : Indexed look up mechanism being used\n");
      return ERROR;
    }

  /* We are here, so "2-Level Lookup Mechanism" is being used */
  sp_bits = pATM->status & ATMIF_SP_BITS_MASK;


  val = ((pconf->vpi << sp_bits) | pconf->subport);

  pVPT = &(pATM->vp_table[val]);

  if (pVPT->control & VPLKUP_VALID)
    {
      printk("ERROR: VP already configured \n");
      return ERROR;
    }

  /* Compute number of VCs */
  num_vcs = (pconf->mask >> pconf->bit_pos) + 1;
  if (num_vcs <= (pATM->max_VCCs - pATM->registered_VCCs))
    {
      val = ((pconf->vpi << sp_bits) | pconf->subport);
      umem.L2_table[val] = kmalloc( ( (num_vcs * sizeof(UInt32) ) + ALIGN_4),
				    GFP_KERNEL);

      pVPT->pL2_lkup = (pATM_VC_st *)((UInt32 )(umem.L2_table[val] + 
						ALIGN_4 - 1) & ~(ALIGN_4 - 1));

      if (pVPT->pL2_lkup == NULL)
	{
	  printk("ERROR: Memory not available for VC table\n");
	  return ERROR;
	}
      bzero((char *)pVPT->pL2_lkup, (num_vcs * sizeof (UInt32)));
		
      /* Set up the VP table entry */
      pVPT->vc_mask = pconf->mask;
      pVPT->vc_bit_pos = pconf->bit_pos;
      pATM->registered_VCCs += num_vcs;
      save_and_cli(key);
      pVPT->control = VPLKUP_VALID;
      restore_flags(key);

      return	OK;
    }

  else
    {
      printk("ERROR: Exceeded required number of connections \n");
      return ERROR;
    }
	
} /* sar_vpt_setup */

/******************************************************************
** FUNCTION : sar_vpt_purge()
**
** DESCRIPTION:
**	This function is used to purge all the entries for the specified
**	combination of the sub-port number and VPI by traversing through
**	the level 1 & level 2 tables.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Pointer to the VP configuration structure.
**		ptr2	: Not used.
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.
**
** NOTE:
**	This function is required only if the connection management uses
**	the 2-level lookup mechanism.
**
******************************************************************/
int sar_vpt_purge(void *handle, void *ptr1, UInt32 val1, \
		  void *ptr2, UInt32 val2, UInt32 *var)
{
  pATMIF_st		pATM = (pATMIF_st )handle;
  pL1_LKUP_st		pVPT;
  pVPT_CONFIG_st	pconf = (pVPT_CONFIG_st )ptr1;
  pATM_VC_st		pVC;
  UInt32			num_vcs, val, sp_bits;
	
  /* Return ERROR if "Indexed Lookup Mechanism" is being used */
  if (pATM->control & INDEXED_LKUP_EN)
    {
      printk("ERROR : Indexed look up mechanism being used\n");
      return ERROR;
    }

  /* We are here, so "2-Level Lookup Mechanism" is being used */
  sp_bits = pATM->status & ATMIF_SP_BITS_MASK;
  val = ((pconf->vpi << sp_bits) | pconf->subport);
  pVPT = &(pATM->vp_table[val]);

  if (pVPT->control & VPLKUP_VALID)
    {
      /* Compute number of VCs */
      val = pconf->mask >> pconf->bit_pos;
      num_vcs = 1;
      while (val & 0x01)
	{
	  num_vcs <<= 1;
	  val >>= 1;
	}
      num_vcs -= 1;
      /* Free up the resources allocated for the connection table entries */
      for (val = 0; val < num_vcs;val++)
	{
	  pVC = pVPT->pL2_lkup[val];
	  if (pVC)
	    {
	      sar_conn_terminate(handle, pVC, 0, NULL, 0, NULL);
	    }
	}

      pVPT->vc_mask = 0;
      pVPT->vc_bit_pos = 0;
      pVPT->control = 0;

      val = ((pconf->vpi << sp_bits) | pconf->subport);
      kfree(umem.L2_table[val]); 

      pVPT->pL2_lkup = 0;
      pATM->registered_VCCs -= num_vcs;
      return OK;
    }
  else
    {
      printk("ERROR: No VP table entry\n");
      return ERROR;
    }

} /* sar_vpt_purge */

/******************************************************************
** FUNCTION : sar_purge_all_vcs()
**
** DESCRIPTION:
**	This function deletes all the VCs configured for the interface.
**	It traverses the VP & VC tables and purges all the entries.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		ptr2	: Not used.
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.
**
** NOTE:
**
******************************************************************/
int sar_purge_all_vcs(void *handle, void *ptr1, UInt32 val1,
		      void *ptr2, UInt32 val2, UInt32 *var)
{
  pATMIF_st		pATM = (pATMIF_st )handle;
  pATM_VC_st		pVC;
  UInt32			i;
#if defined (TWO_LEVEL_LOOKUP)
  pL1_LKUP_st		pVPT;
  UInt32			num_vcs, val;
#endif /* TWO_LEVEL_LOOKUP */

#if defined (INDEXED_LOOKUP)
  for (i = 0; i < pATM->max_VCCs; i++)
    {
      if ((pVC = pATM->vc_table[i]) == 0)
	continue;
      sar_conn_terminate(handle, pVC, 0, NULL, 0, NULL);
    }
#else
  for (i = 0; i < MAX_VPI; i++)
    {
      pVPT = &(pATM->vp_table[i]);

      if (pVPT->control & VPLKUP_VALID)
	{
	  /* Compute number of VCs */
	  val = pVPT->vc_mask >> pVPT->vc_bit_pos;
	  num_vcs = 1;
	  while (val & 0x01)
	    {
	      num_vcs <<= 1;
	      val >>= 1;
	    }
	  num_vcs -= 1;
	  /* Free up the resources allocated for the connection table entries */
	  for (val = 0; val < num_vcs;val++)
	    {
	      pVC = pVPT->pL2_lkup[val];
	      if (pVC)
		{
		  sar_conn_terminate(handle, pVC, 0, NULL, 0, NULL);
		}
	    }

	  pVPT->vc_mask = 0;
	  pVPT->vc_bit_pos = 0;
	  pVPT->control = 0;
	  kfree(umem.L2_table[i]); 
	  pVPT->pL2_lkup = 0;
	  pATM->registered_VCCs -= num_vcs;
	}
    }
#endif /* INDEXED_LOOKUP */
  return OK;
} /* sar_purge_all_vcs */

/******************************************************************
** FUNCTION : sar_set_conn_state()
**
** DESCRIPTION:
**	Set the VC connection state.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Connection handle
#if defined (INDEXED_LOOKUP
**		ptr2	: Not used.
**		val1	: VC state - 0 for DOWN state, 1 for UP state.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.

** NOTE:
**
******************************************************************/
int sar_set_conn_state(void *handle, void *ptr1, UInt32 val1, \
								void *ptr2, UInt32 val2, UInt32 *var)
{
	pATM_VC_st		pVC = (pATM_VC_st )ptr1;

	if (val1)
	{
		if (pVC->state & VPCI_OPER_UP)
		{
			/* Do nothing */
		}
		else
		{
			/* If using VC Cache channel, enable the VC cache entry */
			
			pVC->state |= VPCI_OPER_UP;
		}
	}
	else
	{
		if (pVC->state & VPCI_OPER_UP)
		{
			/* If using VC Cache channel, disable the VC cache entry */
			
			pVC->state &= ~VPCI_OPER_UP;
		}
		else
		{
			/* Do nothing */
		}
	}

	return OK;
} /* sar_set_conn_state */

/******************************************************************
** FUNCTION : get_vc_handle()
**
** DESCRIPTION:
**	Fetch the connection handle, given the VPI & VCI values.
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Pointer to the VC configuration structure.
**		ptr2	: Connection handle (updated by this function).
**		val1	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK			: if successful.
**		Error code	: if failed.
**
** NOTE:
**
******************************************************************/
int get_vc_handle(void *handle, void *ptr1, UInt32 val1, \
								void *ptr2, UInt32 val2, UInt32 *var)
{
	pATMIF_st		pATM = (pATMIF_st )handle;
	pVC_CONFIG_st	pconf = (pVC_CONFIG_st )ptr1;
#if defined (TWO_LEVEL_LOOKUP)
	pL1_LKUP_st		pVPT;
	UInt32			sp_bits;
#endif /* TWO_LEVEL_LOOKUP */
	UInt32			index, cell_hdr;
	
	/* Build cell header */
	cell_hdr = BLD_CELL_HDR(pconf->subport, pconf->vpi, \
											pconf->vci, 0, 0, 0);

	*(pATM_VC_st *)ptr2 = 0;
#if defined (INDEXED_LOOKUP)
	/* "Indexed Lookup Mechanism" is being used */

	/* Get the offset in the VC table */
	GET_VPCI(cell_hdr, pATM->vpi_mask, pATM->vci_mask, \
									pATM->vpi_shft, &index);
	*(pATM_VC_st *)ptr2 = (pATM_VC_st )pATM->vc_table[index];
#else
		/* "2-Level Lookup Mechanism" is being used */

		sp_bits = pATM->status & ATMIF_SP_BITS_MASK;
		index = ((pconf->vpi << sp_bits) | pconf->subport);
		pVPT = &(pATM->vp_table[index]);
		if (pVPT->control & VPLKUP_VALID)
		{
			/* Get the offset in the VC table */
			index = pVPT->vc_mask & pconf->vci;
			*(pATM_VC_st *)ptr2 = (pATM_VC_st )pVPT->pL2_lkup[index];
		}
		else
		{
			/* VP Table not configured */
			printk("VP table entry not configured\n");
			return ERROR;
		}
#endif /* INDEXED_LOOKUP */
		if (*(pATM_VC_st *)ptr2)
			return OK;
		else
			return ERROR;
} /* get_vc_handle */

/******************************************************************
** FUNCTION : fetch_vc_handle()
**
** DESCRIPTION:
**	Fetch the connection handle, given the VPI & VCI values.
**
**	Input parameters:
**		pATM	: Pointer to the ATM interface specific pointer.
**		cell_hdr: ATM cell header.
**
**	Return Value:
**		pVC			: VC handle.
**		NULL		: if failed.
**
** NOTE:
**
******************************************************************/
pATM_VC_st fetch_vc_handle(pATMIF_st pATM, UInt32 cell_hdr)
{
#if defined (TWO_LEVEL_LOOKUP)
  pL1_LKUP_st	pVPT;
  UInt32			vci;
  UInt32			max_vcs;
#endif /* TWO_LEVEL_LOOKUP */
  pATM_VC_st		pVC = NULL;
  UInt32			index;
	
#if defined (INDEXED_LOOKUP)
  /* "Indexed Lookup Mechanism" is being used */

  /* Get the offset in the VC table */
  GET_VPCI(cell_hdr, pATM->vpi_mask, pATM->vci_mask, \
	   pATM->vpi_shft, &index);
  if (index > pATM->max_VCCs)
    {
      printk("Invalid index\n");
      return NULL;
    }
  else
    pVC = (pATM_VC_st )pATM->vc_table[index];

#else
  /* "2-Level Lookup Mechanism" is being used */
  index = GET_VPI_SP(cell_hdr);
  if (index > (MAX_VPI * ((pATM->status & 0x3) + 1)))
    {
      printk("Invalid VP table index <hdr = 0x%x>\n", (UInt32)cell_hdr);
      return NULL;
    }
  pVPT = &(pATM->vp_table[index]);

  if (pVPT->control & VPLKUP_VALID)
    {
      /* Compute maximum number of VCs */
      max_vcs = (pVPT->vc_mask >> pVPT->vc_bit_pos) + 1;
      
      /* Get the offset in the VC table */
      vci = GET_VCI(cell_hdr);
      index = (pVPT->vc_mask & vci) >> pVPT->vc_bit_pos;
      
      if (index > max_vcs)
	{
	  printk("Invalid VC table index\n");
	  return NULL;
	}
      
      pVC = (pATM_VC_st )pVPT->pL2_lkup[index];
    }
#endif /* INDEXED_LOOKUP */
  if ((pVC) && (pVC->control & VPCI_ADMIN_UP))
    return pVC;
  else
    return NULL;
} /* fetch_vc_handle */

