/*****************************************************************
**  File : sar_ctrl.c
**
**      ATM SOFTSAR I/O control functions.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  09/11/2000   srinivas voleti           Original
**  02/10/2001   srinivas voleti           Remove reference to recycle 
**                                         buffer & descriptor
**     10/2001   P. Sadik                  Ported to linux
******************************************************************/
#include <linux/string.h>
#include "sys_ss_def.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"

int sar_enqueue_raw(pATM_VC_st , void *, void (*)() ,UInt32 );
int sar_enqueue_oam(pATM_VC_st , void *, void (*)() ,UInt32 );
int sar_xmit_aal5(pATM_VC_st , void *,  void (*)(),UInt32 );

int add_timer_entry(pATM_VC_st , UInt32 );   
int delete_timer_entry(pATM_VC_st ); 

int (*func[])() = {sar_enqueue_raw, NULL, NULL, NULL, \
                                NULL, sar_xmit_aal5, sar_enqueue_oam};

/******************************************************************
** FUNCTION : sar_set_if_admin_state
**
** DESCRIPTION:
**	This function administratively enables or disables the ATM 
**	interface. 
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: TRUE or FALSE.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_set_if_admin_state(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_set_if_admin_state */

/******************************************************************
** FUNCTION : sar_get_if_oper_state
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: TRUE or FALSE.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_get_if_oper_state(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATMIF_st	pATM = (pATMIF_st )handle;

	*var = (pATM->status & ATMIF_OPER_UP) ? TRUE: FALSE;

	return (OK);
} /* sar_get_if_oper_state */

/******************************************************************
** FUNCTION : sar_get_per_vc_stats
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Connection handle.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Pointer to the VC statistics data structure.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_get_per_vc_stats(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATM_VC_st pVC = (pATM_VC_st )handle;

	ptr2 = &pVC->stats;

	printk("VC Statistics <sp = 0x%x  vpi = 0x%x  vci = 0x%x>\n", \
			pVC->subport, pVC->vpi, pVC->vci);
	printk("Tx. Cells          = %d\n", pVC->stats.tx_cells);
	printk("Tx. PDUs           = %d\n", pVC->stats.tx_pdus);
	printk("Rx. Cells          = %d\n", pVC->stats.rx_cells);
	printk("Rx. PDUs           = %d\n", pVC->stats.rx_pdus);
	printk("Tx. OAM            = %d\n", pVC->stats.tx_oam);
	printk("Rx. OAM            = %d\n", pVC->stats.rx_oam);
	printk("Tx. Raw Cells      = %d\n", pVC->stats.tx_raw_cells);
	printk("Rx. Raw Cells      = %d\n", pVC->stats.rx_raw_cells);
	return (OK);
} /* sar_get_per_vc_stats */

/******************************************************************
** FUNCTION : sar_get_per_if_vc_stats 
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Pointer to the interface specific statistics.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_get_per_if_vc_stats(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATMIF_st	pATM = (pATMIF_st )handle;

	ptr2 = &pATM->vc_stats;

	printk("Tx. Cells          = %d\n", pATM->vc_stats.tx_cells);
	printk("Tx. PDUs           = %d\n", pATM->vc_stats.tx_pdus);
	printk("Rx. Cells          = %d\n", pATM->vc_stats.rx_cells);
	printk("Rx. PDUs           = %d\n", pATM->vc_stats.rx_pdus);
	printk("Tx. OAM            = %d\n", pATM->vc_stats.tx_oam);
	printk("Rx. OAM            = %d\n", pATM->vc_stats.rx_oam);
	printk("Tx. Raw Cells      = %d\n", pATM->vc_stats.tx_raw_cells);
	printk("Rx. Raw Cells      = %d\n", pATM->vc_stats.rx_raw_cells);
	return (OK);
} /* sar_get_per_if_vc_stats */

/******************************************************************
** FUNCTION : sar_get_if_err_stats
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Pointer to the error statistics.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_get_if_err_stats(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
  pATMIF_st       pATM = (pATMIF_st )handle;

  ptr2 = &pATM->err_stats;

  printk("Unknown Tx. Connections = %d\n", pATM->err_stats.tx_unk_conn_err);
  printk("Unknown Rx. Connections = %d\n", pATM->err_stats.rx_unk_conn_err);
  printk("Invalid AAL Type        = %d\n", pATM->err_stats.inv_aal_err);
  printk("Insuff. Tx. Desc        = %d\n", pATM->err_stats.insuff_tx_desc_err);
  printk("Insuff. Rx. Desc        = %d\n", pATM->err_stats.insuff_rx_desc_err);
  printk("Insuff Buffers          = %d\n", pATM->err_stats.insuff_bufs_err);
  printk("Cell Drops              = %d\n", pATM->err_stats.cell_drop_err);
  printk("Tx Queue Exceeded       = %d\n", pATM->err_stats.txq_exceeded_err);
  printk("CRC10 Errors            = %d\n", pATM->err_stats.CRC10_err);
  printk("CRC32 Errors            = %d\n", pATM->err_stats.CRC32_err);
  printk("Overflow Errors         = %d\n", pATM->err_stats.overflow_err);
  printk("VCC Congestion State    = %d\n", (*(UInt32 *)ATMVCCC & 0xFF));
  printk("Inv. Tx PDU size Errors = %d\n", pATM->err_stats.inv_txsize_err);
  printk("Inv. Rx PDU size Errors = %d\n", pATM->err_stats.inv_rxsize_err);
  printk("DMA Halt Errors         = %d\n", pATM->err_stats.dma_halt_err);

  printk("DMA Errors              = %d\n", pATM->err_stats.dma_err);   

  return (OK);
} /* sar_get_if_err_stats */

/******************************************************************
** FUNCTION : sar_clr_if_stats
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_clr_if_stats(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATMIF_st	pATM = (pATMIF_st )handle;

	bzero((char *)&pATM->vc_stats, sizeof(VC_STATS_st));
	bzero((char *)&pATM->err_stats, sizeof(SAR_ERR_st));
	return (OK);
} /* sar_clr_if_stats */

/******************************************************************
** FUNCTION : sar_clr_vc_stats
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the VC handle.
**		ptr1	: Not used.
**		val1	: ADMIN_STATE_UP or ADMIN_STATE_DOWN.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_clr_vc_stats(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATM_VC_st pVC = (pATM_VC_st )handle;

	bzero((char *)&pVC->stats, sizeof(VC_STATS_st));
	return (OK);
} /* sar_clr_vc_stats */

/******************************************************************
** FUNCTION : sar_read_reg
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: ADMIN_STATE_UP or ADMIN_STATE_DOWN.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_read_reg(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_read_reg */

/******************************************************************
** FUNCTION : sar_write_reg
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: ADMIN_STATE_UP or ADMIN_STATE_DOWN.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_write_reg(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_write_reg */

/******************************************************************
** FUNCTION : sar_read_phy
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: ADMIN_STATE_UP or ADMIN_STATE_DOWN.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_read_phy(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_read_phy */

/******************************************************************
** FUNCTION : sar_write_phy
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: ADMIN_STATE_UP or ADMIN_STATE_DOWN.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_write_phy(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_write_phy */

/******************************************************************
** FUNCTION : sar_dump_atmif_struct
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_dump_atmif_struct(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATMIF_st	pATM = (pATMIF_st )handle;

	printk("ATM interface data structure\n");
	printk("Switch\t: 0x%x\n", (int )pATM->switch_no);
	printk("Port\t\t: 0x%x\n", (int )pATM->port_no);
	printk("Interface\t: 0x%x\n", (int )pATM->if_no);
	printk("Control\t\t: 0x%x\n", (int )pATM->control);
	printk("Status\t\t: 0x%x\n", (int )pATM->status);
	printk("Maximum VCCs\t: 0x%x\n", (int )pATM->max_VCCs);
	printk("Registered VCCs: 0x%x\n", (int )pATM->registered_VCCs);
	printk("Active VCCs\t: 0x%x\n", (int )pATM->active_VCCs);
	printk("VP table pointer: 0x%x\n", (int )pATM->vp_table);
	printk("VC table pointer: 0x%x\n", (int )pATM->vc_table);
	printk("OS specific pointer: 0x%x\n", (int )pATM->OS_specific);

	printk("OAM transmit handler: 0x%x\n", (int )pATM->oam_tx_hdlr);
	printk("OAM receive handler: 0x%x\n", (int )pATM->oam_rx_hdlr);
	if (pATM->get_phy)
		printk("PHY read handler: 0x%x\n", (int )pATM->get_phy);
	if (pATM->set_phy)
		printk("PHY write handler: 0x%x\n", (int )pATM->set_phy);
	
	printk("Status base pointer: 0x%x\n", (int )pATM->status_base);
	printk("Status head pointer: 0x%x\n", (int )pATM->status_head);
	printk("Status tail pointer: 0x%x\n", (int )pATM->status_tail);
	printk("Status last pointer: 0x%x\n", (int )pATM->status_last);
	printk("VC base pointer: 0x%x\n", (int )pATM->free_vc_base);
	printk("VC head pointer: 0x%x\n", (int )pATM->free_vc_list_head);
	printk("VC tail pointer: 0x%x\n", (int )pATM->free_vc_list_tail);
	printk("Active VC list pointer: 0x%x\n", (int )pATM->active_vc_list);
	printk("Free desc list base pointer: 0x%x\n", (int )pATM->free_dlist_base);
	printk("Free desc list head pointer: 0x%x\n", (int )pATM->free_dlist_head);
	printk("Free desc list tail pointer: 0x%x\n", (int )pATM->free_dlist_tail);
	printk("Number of free descriptors: 0x%x\n", (int )pATM->num_free_descs);
	printk("BD 2k head pointer: 0x%x\n", (int )pATM->free_2k_bd_head);
	printk("BD 2k tail pointer: 0x%x\n", (int )pATM->free_2k_bd_tail);
	printk("Number of 2k BDs: 0x%x\n", (int )pATM->free_2k_bds);
	printk("BD 64 head pointer: 0x%x\n", (int )pATM->free_64_bd_head);
	printk("BD 64 tail pointer: 0x%x\n", (int )pATM->free_64_bd_tail);
	printk("Number of 64 BDs: 0x%x\n", (int )pATM->free_64_bds);
	return (OK);
} /* sar_dump_atmif_struct */

/******************************************************************
** FUNCTION : sar_dump_vc_struct 
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Connection handle.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_dump_vc_struct(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
	pATM_VC_st pVC = (pATM_VC_st )handle;

	printk("VC structure\n");
	printk("Control = %x \n", (int )pVC->control);
	printk("State = %x \n", (int )pVC->state);
	printk("Sub-port = %x \n", (int )pVC->subport);
	printk("VPI = %x \n", (int )pVC->vpi);
	printk("VCI = %x \n", (int )pVC->vci);
	printk("Cell header = %x \n", (int )pVC->cell_hdr);
	printk("Transmit PDU descriptor = %x \n", (int )pVC->tx_pdu_desc);
	printk("Current receive data descriptor = %x \n", (int )pVC->rx_data_desc);
	printk("Current receive OAM descriptor = %x \n", (int )pVC->rx_oam_desc);
	printk("Interface back pointer = %x \n", (int )pVC->pif);
	printk("Protocol specific pointer = %x \n", (int )pVC->proto);
	return (OK);
} /* sar_dump_vc_struct */

/******************************************************************
** FUNCTION : sar_dump_status_struct
**
** DESCRIPTION:
**
**	Input parameters:
**		handle	: Pointer to the ATM interface specific pointer.
**		ptr1	: Not used.
**		val1	: Not used.
**		ptr2	: Not used.
**		val2	: Not used.
**		var		: Not used.
**
**	Return Value:
**		OK
**	
** NOTE:
**
******************************************************************/
int sar_dump_status_struct(void *handle, void *ptr1, UInt32 val1, \
                                void *ptr2, UInt32 val2, UInt32 *var)
{
#if 0
	pATMIF_st	pATM = (pATMIF_st )handle;
#endif

	return (OK);
} /* sar_dump_status_struct */

/******************************************************************
** FUNCTION : sar_xmit()
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
int sar_xmit(void *handle, void *ptr1, UInt32 type, \
					void *ptr2, UInt32 val, UInt32 *var)
{
	pATM_VC_st  pVC = (pATM_VC_st )handle;
	pATMIF_st   pATM;
	int         retval;

	pATM = pVC->pif;
	if (type != OAM_TYPE)
		type = pVC->control & AAL_TYPE_MASK;

	/* Check the AAL type & call appropriate functions */
	retval = (*func[type])(pVC, ptr1, ptr2, val);

	return(retval);
} /* sar_xmit */

/******************************************************************
** FUNCTION : sar_recv()
**
** DESCRIPTION:
**
** NOTE:
**	The calling function should provide a handler via hdlr. This function
**	will call the handler & passes the mbuf pointer along with "val2"
**	and "var" as parameters.
**
******************************************************************/
int sar_recv(void *handle, void *ptr1, UInt32 type, \
					void *hdlr(), UInt32 val2, UInt32 *var)
{
  return OK;
} /* sar_recv */


/******************************************************************
** FUNCTION : sar_sched_conn()
**
** DESCRIPTION:
**      Add the connection to the scheduler list.
**
** NOTE:
**
******************************************************************/
int sar_sched_conn(void *handle, void *ptr1, UInt32 val1, 
		   void *ptr2, UInt32 val2, UInt32 *var)
{
        int     status;

        status = add_timer_entry((pATM_VC_st )handle, val1);

        return status;

} /* sar_sched_conn */

/******************************************************************
** FUNCTION : sar_desched_conn()
**
** DESCRIPTION:
**      Delete the connection to the scheduler list.
**
** NOTE:
**
******************************************************************/
int sar_desched_conn(void *handle, void *ptr1, UInt32 val1, 
		     void *ptr2, UInt32 val2, UInt32 *var)
{
        int     status;

        status = delete_timer_entry((pATM_VC_st )handle);

        return status;
} /* sar_desched_conn */  
