/*****************************************************************
**  File : sar_hw.c
**
**		Hardware specific code.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  08/21/2000   srinivas voleti           Original
**  09/22/2001   sadik                     ported to linux
******************************************************************/

#include <linux/string.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/interrupt.h>
#include <linux/tqueue.h>   
#include <linux/skbuff.h>

#include "banyan.h"
#include "sys_ss_def.h"
#include "atm.h"
#include "arbiter.h"
#include "interrupt.h"
#include "dma.h"
#include "gpio.h"
#include "timer.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"

#define	DFLT_DMA_ARB_PRIOR 0xaaaaaaab
#define	DFLT_IM2	0	
#define	DFLT_IM3	0x1fe		/* DMA channels 1 - 8 */
#define	DFLT_IM4	0x03
#define	DFLT_IM5	0
#define	DFLT_IM6	0

#define	TMR_500US	37500	/* No. of ticks to get 0.5ms with a 75MHz 
				   system clock (75e+6 * 0.5e-3 = 37500) */

#define	DEVCMD_MASK (DMA_DEVCMD(7))
#define	DEVCS_MASK	(DMA_DS(3))

extern pATM_VC_st  crunch_pvc_array[];
extern UInt32      atm_crunch_state;

extern void init_dma_intr_vector_tbl(void);
extern void set_dma_intr_vector(void (*fptr)(), unsigned int index);
extern void sar_schedule_cell(unsigned long);
DECLARE_TASKLET(timer_tasklet, sar_schedule_cell, 0);  

extern void *gATM;
/******************************************************************
** FUNCTION : hw_atm_init
**
** DESCRIPTION:
**	Initialize the ATM interface of the banyan processor.
**
** NOTE:
**
******************************************************************/
void hw_atm_init(pATMIF_st pATM)
{
  UInt32	control_0 = 0, control_1 = 0;

  /* Initialize the ATM interface */
  switch (pATM->control & ATMIF_TYPE_MASK) {

  case ATM_UTOPIA1:
    control_0 = (ATMC_IE | ATMC_OE);

    if (pATM->control & ATMIF_CMS_OUTPUT)
      {
	atm.atmcp = (pATM->control & ATMIF_CLK_PRESCALAR_MASK) >> ATMIF_CLK_PRESCALAR_SHFT;
	control_0 |= ATMC_CMS;
      }
    /* Set the ATM VC Cache disable register - clear the appropriate 
       bits to enable the usage of the appropriate VC cache entries */
    if (pATM->control & ATMIF_FRAME_MODE)
      atm.atmvccd = 0x40;

    break;

  case ATM_UTOPIA2:
    control_0 = (ATMC_IE | ATMC_OE);
    control_1 = (ATMC_IE | ATMC_OE);

    if (pATM->control & ATMIF_CMS_OUTPUT)
      {
	atm.atmcp = (pATM->control & ATMIF_CLK_PRESCALAR_MASK) >> ATMIF_CLK_PRESCALAR_SHFT;
	control_0 |= ATMC_CMS;
	control_1 |= ATMC_CMS;
      }

    /* Utopia 2 PHY enables */
    atm.atmu2pc = 0;
    if (pATM->control & ATMIF_PHY28)
      atm.atmu2pc |= ATMU2PC_P28;
    if (pATM->control & ATMIF_PHY29)
      atm.atmu2pc |= ATMU2PC_P29;
    if (pATM->control & ATMIF_PHY30)
      atm.atmu2pc |= ATMU2PC_P30;

    /* Set the ATM VC Cache disable register - clear the appropriate 
       bits to enable the usage of the appropriate VC cache entries */
    if (pATM->control & ATMIF_FRAME_MODE)
      atm.atmvccd = 0xc0;

    break;

  default:
    break;
  };
	
  /* Clear the status registers */
  atm.ch0.atms = 0;
  atm.ch1.atms = 0;

  /* Set the control registers */
  atm.ch0.atmc = control_0;
  atm.ch1.atmc = control_1;

} /* hw_atm_init */

/******************************************************************
** FUNCTION : sar_hw_module_init
**
** DESCRIPTION:
**	Initialize the banyan processor for use with SOFTSAR.
**
** NOTE:
**
******************************************************************/
void sar_hw_module_init(pATMIF_st pATM)
{
  UInt32		i, dummy_read;
  DMA_CHAN	*pchan;

  /* Disable the ATM Interface */
  atm.atmintfc = ATM_DISABLE; 
  /* Initialize the ATM PHY */
  if (pATM->init_phy)
    (*pATM->init_phy)();
  install_sar_int_hdlr(pATM);

  /* ATM interface initialization */
  hw_atm_init(pATM);

  /* DMA channel initialization */
  /* NUM_ATM_DMA_CHANNELS = 9 */
  for (i = 0, pchan = (DMA_CHAN *)DMA_BASE; i < NUM_ATM_DMA_CHANNELS; i++, pchan++)
    {
      pchan->dmac  = (DMAxC_MODE(DMA_TRANSFER_REQUEST));
      if (i == 0)
	{
	  pchan->dmasm = (DMAxS_F | DMAxS_D | DMAxS_C | DMAxS_E | DMAxS_H);
	}
      else
	pchan->dmasm = (DMAxS_F | DMAxS_D |          DMAxS_E | DMAxS_H);

      if (i == 7)
	{
	  if (pATM->control & ATMIF_FRAME_MODE)
	    {
	      pchan->dmasm = (DMAxS_F | DMAxS_D | DMAxS_C | DMAxS_E | DMAxS_H);
	    }
	}

      if (i == 8)
	{
	  if (pATM->control & ATMIF_FRAME_MODE)
	    {
	      if (((pATM->control & ATMIF_TYPE_MASK) == ATM_UTOPIA2) || \
		  ((pATM->control & ATMIF_TYPE_MASK) == ATM_DPI2)) 
		{
		  pchan->dmasm = (DMAxS_F | DMAxS_D | DMAxS_C | DMAxS_E | DMAxS_H);
		}
	    }
	}

      dummy_read = pchan->dmadptr;
      pchan->dmadptr = 0;
      dummy_read = pchan->dmandptr;
      pchan->dmandptr = 0;
    }

	
  /* Set up Timer 1 for 0.5ms */
  timer.ct1.count = 0;
  timer.ct1.ctc = 0;
  timer.ct1.compare = TMR_500US;
  interrupt.i2.imask &= ~0x2;

} /* sar_hw_module_init */
  
void display_dma_channel(unsigned int dmaChannel)
{
  UInt32  dummy_read;
  DMA_CHAN *pchan = (DMA_CHAN *)DMA_BASE;
  pchan += dmaChannel;

  printk("Channel %d Control 0x%08x Status 0x%08x StatusMask 0x%08x DPTR 0x%08x NDPTR 0x%08x\n",
	 dmaChannel,pchan->dmac,pchan->dmas,pchan->dmasm,pchan->dmadptr,pchan->dmandptr);

}

  
  
/******************************************************************
** FUNCTION : sar_hw_module_reset
**
** DESCRIPTION:
**	Reset the banyan processor modules used for SOFTSAR.
**
** NOTE:
**
******************************************************************/
void sar_hw_module_reset(pATMIF_st pATM)
{
  /* Disable the ATM interface */
  atm.atmintfc = ATM_DISABLE; 

  /* Disable the DMA channels, if appropriate */

  /* Disable the clocks or counter registers, if appropriate */
	
} /* sar_hw_module_reset */


/******************************************************************
** FUNCTION : sar_timer_isr()
**
** DESCRIPTION:
**	Timer interrupt to periodically process the ATM DMA channels.
**
** NOTE:
**
******************************************************************/
void sar_timer_isr(int irq, void *dev_id, struct pt_regs * regs)
{
  register UInt32       ipend,imask;
  register pATMIF_st	pATM = (pATMIF_st )gATM;
  register pSTATUS_st	pstatus;
  pSTATUS_st		pnext;
  UInt32	      	evt_mask;

  unsigned long key;

      save_and_cli(key);
      timer.ct1.ctc &= ~CTC_TO; /* Enable timer interrupt */
      tasklet_schedule(&timer_tasklet); /* run sar_schedule_cell */

      evt_mask = (1 << ATM_DMA_EVT_SHFT);

      if (pATM->status & evt_mask)
	{
	  pstatus = pATM->status_tail;
	  pnext = pstatus;
	  pnext++;
	  if (pnext> pATM->status_last)
	    pnext = pATM->status_base;
      
	  if (pnext == pATM->status_head)
	    {
	      /* Status queue is full; Drop the PDU */
	    }
	  else
	    {
	      pstatus->state = RXINT_STATE;
	      /* Build the status message & enqueue it */
	      pATM->status_tail = pnext;
	  
	      pATM->status &= ~(evt_mask);
	  
	      post_status(pATM->OS_specific);
	    }
	}
      restore_flags(key);
      //    }
} /* sar_timer_isr */

/******************************************************************
** FUNCTION : atm_ir_status_intr()
**
** DESCRIPTION:
**	ATM IR status interrupt processing.
**
** NOTE:
**
******************************************************************/
UByte atm_ir_status_intr(pATMIF_st pATM, volatile ATM_CHAN *atmch)
{
  pATM_VC_st			pVC = 0;
  volatile pDMA_st	pdesc;
  UInt32      		cell_hdr, state = 0;
  UInt32				done = 0, ds;
  UByte				retval = 0;
  while ( atmch->atms & ATMS_IR)
    {
      /* Read the cell header & fetch its handle */
      cell_hdr = atmch->atmih;
      pVC = fetch_vc_handle(pATM, cell_hdr);

      if (pVC)
	{
	  ds = (pVC->control & RCV_DEVCS_MASK) << RCV_DEVCS_SHFT;

	  /* If the connection is configured, is an OAM cell */
	  if (cell_hdr & OAM_CELL_DETECT)
	    {
	      /* Receive an OAM cell */
	      pdesc = pVC->rx_oam_desc;
	      if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		{
		  pdesc = get_64_buf_desc_list(pATM, pVC);
		  if ((VIRT_TO_PHYS((UInt32 )pdesc)))
		    pVC->rx_oam_desc = pdesc;
		  else
		    {
		      /* Drop the cell & update error statistics */
		      DROP_ATM_CELL(&atmch->atmid);
		      pATM->err_stats.cell_drop_err++;
		      done = 0;
		      goto done_chk;
		    }

		}

	      pdesc->control &= ~(DMA_DONE | DEVCS_MASK);
	      pdesc->control |= ((DMA_DEVCMD(CRC10)) |ds);

	      /* Check that the DMA channel 0 is not running */
	      while((dma.ch0.dmac & DMAxC_RUN));

	      dma.ch0.dmas &= ~(DMAxS_H); 
	      dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
	      /* Wait for DMA to be done */
	      while(!(dma.ch0.dmas & (DMAxS_H)));
	      dma.ch0.dmas &= ~(DMAxS_H);

	      /* Allocate new buffer-descriptor pair for future use */
	      pVC->rx_oam_desc = get_64_buf_desc_list(pATM, pVC);
	      state = RX_OAM_STATE;
	      done = 1;

	    }
	  else if ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE)
	    {
	      /* Non-VC CACHE AAL5 connection or a VC cache entry
		 which is disabled due to halting of the corresponding
		 DMA channel */ 

	      /* If the PDU is marked to be dropped, just drop it */
	      if (pVC->state & RX_DROP_PDU)                                     
		{                                  
		  /* Drop the cell */        
		  DROP_ATM_CELL(&atmch->atmid);
		  
		  /* If EPDU is seen, prepare to receive the next PDU */
		  if (cell_hdr & EPDU_DETECT)                           
		    pVC->state &= ~RX_DROP_PDU;                   
                                                                                              
		  done = 0;                                             
		  goto done_chk;                                        
		}                                                             

	      /* NOTE: Current implementation doesn't support connections
		 on disabled VCC entries. */
	      if (pVC->control & CHAN_MASK)
		{
		  /* Drop the cell & update error statistics */
		  DROP_ATM_CELL(&atmch->atmid);
		  pATM->err_stats.cell_drop_err++;
		  done = 0;
		}
	      else
		{
		  pdesc = pVC->rx_data_desc;

		  if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		    {
		      pdesc = get_2k_buf_desc_list(pATM, pVC, 1);
		      if ((VIRT_TO_PHYS((UInt32 )pdesc)))
			pVC->rx_data_desc = pdesc;
		      else
			{
			  /* Drop the cell & update error statistics */
			  DROP_ATM_CELL(&atmch->atmid);
			  pATM->err_stats.cell_drop_err++;
			  done = 0;
			  goto done_chk;
			}
		    }

		  /* If the buffer doesn't have enough space to take a
		     cell, probably the received PDU is > 2016 bytes. In
		     this case drop the cell */
		  if (DMA_COUNT(pdesc->control) < MAX_CELL_PAYLOAD)
		    {
		      /* Drop the cell & update error statistics */
		      DROP_ATM_CELL(&atmch->atmid);
		      pATM->err_stats.inv_rxsize_err++;
		      
		      /* Note that all the cells associated with that PDU
			 MUST BE DROPPED */
		      pVC->state |= RX_DROP_PDU;
		      
		      /* Re-set the descriptor in use for future use */
		      pdesc->curr_addr = VIRT_TO_PHYS(pdesc->buf_start);
		      pdesc->status = 0xffffffff;
		      pdesc->control = DFLT_AAL5_RX_DESC_CTRL;
		      pdesc->link = 0;
		      
		      done = 0;
		      goto done_chk;
		    }  

		  /*
		    NOTE:
		    If the DEVCS for the descriptor is initialized to
		    0xffffffff then the DEVCMD could take just two values
		    i.e. either CRC32 or CRC32_LD.
		  */
		  pdesc->control &= ~(DMA_DONE | DEVCMD_MASK | DEVCS_MASK);
		  if (cell_hdr & EPDU_DETECT)
		    {
		      pdesc->control |= (DMA_DEVCMD(CRC32_LD) | ds);
		    }
		  else
		    pdesc->control |= (DMA_DEVCMD(CRC32) | ds);

		  /* Check that the DMA channel 0 is not running */
		  while((dma.ch0.dmac & DMAxC_RUN));

		  dma.ch0.dmas &= ~(DMAxS_H);  
		  dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
		  /* Wait for DMA to be done */
		  while(!(dma.ch0.dmas & (DMAxS_H)));

		  dma.ch0.dmas &= ~(DMAxS_H);

		  /* Is it EPDU */
		  if (cell_hdr & EPDU_DETECT)
		    {
		      /* Allocate new buffer-descriptor pair for future use */
		      pVC->rx_data_desc = get_2k_buf_desc_list(pATM, pVC, 1);
		      state = RX_AAL5_DMA0_STATE;
		      done = 1;
		    }
		  else
		    done = 0;
		}
	    }
	  else /* if NOT ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE) */
	    {
	      pdesc = pVC->rx_data_desc;

	      if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		{
		  pdesc = get_64_buf_desc_list(pATM, pVC);
		  if ((VIRT_TO_PHYS((UInt32 )pdesc)))
		    pVC->rx_data_desc = pdesc;
		  else
		    {
		      /* Drop the cell & update error statistics */
		      DROP_ATM_CELL(&atmch->atmid);
		      pATM->err_stats.cell_drop_err++;
		      done = 0;
		      goto done_chk;
		    }

		}

	      pdesc->control &= ~(DMA_DONE | DEVCS_MASK);
	      pdesc->control |= ds;

	      /* Check that the DMA channel 0 is not running */
	      while((dma.ch0.dmac & DMAxC_RUN));

	      dma.ch0.dmas &= ~(DMAxS_H); 
	      dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
	      /* Wait for DMA to be done */
	      while(!(dma.ch0.dmas & (DMAxS_H)));
	      dma.ch0.dmas &= ~(DMAxS_H);

	      /* Allocate new buffer-descriptor pair for future use */
	      pVC->rx_data_desc = get_64_buf_desc_list(pATM, pVC);
	      state = RAW_CELL_STATE;
	      done = 1;
	    }
	  retval = 1;
	}
      else /* if !pVC */
	{
	  /* Drop the cell & update error statistics */
	  DROP_ATM_CELL(&atmch->atmid);
	  pATM->err_stats.cell_drop_err++;
	  pATM->err_stats.rx_unk_conn_err++; 
	  done = 0;
	}

      /* Usage of the current buffer-descriptor done ?? */
    done_chk:
      if (done)
	{
	  ADD_STATUS_ENTRY(pATM, state, pdesc);
	  state = 0;
	}
    }
  return done;
} /* atm_ir_status_intr */

/******************************************************************
** FUNCTION : atm_ir_status_poll()
**
** DESCRIPTION:
**	ATM IR status processing by polling.
**
** NOTE:
**
******************************************************************/
UByte atm_ir_status_poll(pATMIF_st pATM, volatile ATM_CHAN *atmch)
{
  pATM_VC_st			pVC = 0;
  volatile pDMA_st	pdesc;
  UInt32      		cell_hdr, state = 0;
  UInt32				done = 0, ds;
  UByte				retval = 0;
  unsigned long key;

  while ( atmch->atms & ATMS_IR)
    {
      /* Read the cell header & fetch its handle */
      cell_hdr = atmch->atmih;
      pVC = fetch_vc_handle(pATM, cell_hdr);

      if (pVC)
	{
	  ds = (pVC->control & RCV_DEVCS_MASK) << RCV_DEVCS_SHFT;

	  /* If the connection is configured, is an OAM cell */
	  if (cell_hdr & OAM_CELL_DETECT)
	    {
	      /* Receive an OAM cell */
	      pdesc = pVC->rx_oam_desc;
	      if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		{
		  save_and_cli(key);
		  pdesc = get_64_buf_desc_list(pATM, pVC);
		  restore_flags(key);
		  if ((VIRT_TO_PHYS((UInt32 )pdesc)))
		    pVC->rx_oam_desc = pdesc;
		  else
		    {
		      /* Drop the cell & update error statistics */
		      save_and_cli(key);
		      DROP_ATM_CELL(&atmch->atmid);
		      restore_flags(key);
		      pATM->err_stats.cell_drop_err++;
		      done = 0;
		      goto done_chk;
		    }

		}

	      pdesc->control &= ~(DMA_DONE | DEVCS_MASK);
	      pdesc->control |= ds;
	      /* Check that the DMA channel 0 is not running */
	      save_and_cli(key);
	      while((dma.ch0.dmac & DMAxC_RUN));
	      dma.ch0.dmas &= ~(DMAxS_H); 
	      dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
	      /* Wait for DMA to be done */
	      while(!(dma.ch0.dmas & (DMAxS_H)));
	      dma.ch0.dmas &= ~(DMAxS_H);

	      /* Allocate new buffer-descriptor pair for future use */
	      pVC->rx_oam_desc = get_64_buf_desc_list(pATM, pVC);
	      restore_flags(key);
	      state = RX_OAM_STATE;
	      done = 1;

	    }
	  else if ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE)
	    {
	      /* Non-VC CACHE AAL5 connection or a VC cache entry
		 which is disabled due to halting of the corresponding
		 DMA channel */ 

	      /* If the PDU is marked to be dropped, just drop it */
	      if (pVC->state & RX_DROP_PDU)
		{
		  /* Drop the cell */
		  save_and_cli(key);
		  DROP_ATM_CELL(&atmch->atmid);

		  /* If EPDU is seen, prepare to receive the next PDU */
		  if (cell_hdr & EPDU_DETECT)
		    pVC->state &= ~RX_DROP_PDU;
		  restore_flags(key);

		  done = 0;
		  goto done_chk;
		}

	      /* NOTE: Current implementation doesn't support connections
		 on disabled VCC entries. This would be supported 
		 later */
	      if (pVC->control & CHAN_MASK)
		{
		  /* Drop the cell & update error statistics */
		  save_and_cli(key);
		  DROP_ATM_CELL(&atmch->atmid);
		  restore_flags(key);
		  pATM->err_stats.cell_drop_err++;
		  done = 0;
		}
	      else
		{
		  pdesc = pVC->rx_data_desc;

		  if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		    {
		      save_and_cli(key);
		      pdesc = get_2k_buf_desc_list(pATM, pVC, 1);
		      restore_flags(key);
		      if ((VIRT_TO_PHYS((UInt32 )pdesc)))
			pVC->rx_data_desc = pdesc;
		      else
			{
			  /* Drop the cell & update error statistics */
			  save_and_cli(key);
			  DROP_ATM_CELL(&atmch->atmid);
			  restore_flags(key);
			  pATM->err_stats.cell_drop_err++;
			  done = 0;
			  goto done_chk;
			}
		    }

		  /* If the buffer doesn't have enough space to take a
		     cell, probably the received PDU is > 2016 bytes. In
		     this case drop the cell */
		  if (DMA_COUNT(pdesc->control) < MAX_CELL_PAYLOAD)
		    {
		      save_and_cli(key);
		      /* Drop the cell & update error statistics */
		      DROP_ATM_CELL(&atmch->atmid);
		      pATM->err_stats.inv_rxsize_err++;

		      /* Note that all the cells associated with that PDU
			 MUST BE DROPPED */
		      pVC->state |= RX_DROP_PDU;

		      /* Re-set the descriptor in use for future use */
		      pdesc->curr_addr = VIRT_TO_PHYS(pdesc->buf_start);
		      pdesc->status = 0xffffffff;
		      pdesc->control = DFLT_AAL5_RX_DESC_CTRL;
		      pdesc->link = 0;
		      restore_flags(key);

		      done = 0;
		      goto done_chk;
		    }   

		  /*
		    NOTE:
		    If the DEVCS for the descriptor is initialized to
		    0xffffffff then the DEVCMD could take just two values
		    i.e. either CRC32 or CRC32_LD.
		  */
		  pdesc->control &= ~(DMA_DONE | DEVCMD_MASK | DEVCS_MASK);
		  if (cell_hdr & EPDU_DETECT)
		    pdesc->control |= (DMA_DEVCMD(CRC32_LD) | ds);
		  else
		    pdesc->control |= (DMA_DEVCMD(CRC32) | ds);


		  /* Check that the DMA channel 0 is not running */
		  save_and_cli(key);
		  while((dma.ch0.dmac & DMAxC_RUN));
		  dma.ch0.dmas &= ~(DMAxS_H);  
		  dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
		  /* Wait for DMA to be done */
		  while(!(dma.ch0.dmas & (DMAxS_H)));
		  dma.ch0.dmas &= ~(DMAxS_H);
		  restore_flags(key);

		  /* Is it EPDU */
		  if (cell_hdr & EPDU_DETECT)
		    {
		      /* Allocate new buffer-descriptor pair for future use */
		      save_and_cli(key);
		      pVC->rx_data_desc = get_2k_buf_desc_list(pATM, pVC, 1);
		      restore_flags(key);
		      state = RX_AAL5_DMA0_STATE;
		      done = 1;
		    }
		  else
		    done = 0;
		}
	    }
	  else
	    {
	      /* Process an instream cell (used in DPI mode only with
		 a switch supporting Simple Device Protocol (SDP) */
	      pdesc = pVC->rx_data_desc;

	      if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
		{
		  save_and_cli(key);
		  pdesc = get_64_buf_desc_list(pATM, pVC);
		  restore_flags(key);
		  if ((VIRT_TO_PHYS((UInt32 )pdesc)))
		    pVC->rx_data_desc = pdesc;
		  else
		    {
		      /* Drop the cell & update error statistics */
		      save_and_cli(key);
		      DROP_ATM_CELL(&atmch->atmid);
		      restore_flags(key);
		      pATM->err_stats.cell_drop_err++;
		      done = 0;
		      goto done_chk;
		    }

		}

	      pdesc->control &= ~(DMA_DONE | DEVCS_MASK);
	      pdesc->control |= ds;

	      /* Check that the DMA channel 0 is not running */
	      save_and_cli(key);
	      while((dma.ch0.dmac & DMAxC_RUN));
	      dma.ch0.dmas &= ~(DMAxS_H); 
	      dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc);
	      /* Wait for DMA to be done */
	      while(!(dma.ch0.dmas & (DMAxS_H)));
	      dma.ch0.dmas &= ~(DMAxS_H);

	      /* Allocate new buffer-descriptor pair for future use */
	      pVC->rx_data_desc = get_64_buf_desc_list(pATM, pVC);
	      restore_flags(key);
	      state = RAW_CELL_STATE;
	      done = 1;
	    }
	  retval = 1;
	}
      else
	{
	  /* Drop the cell & update error statistics */
	  save_and_cli(key);
	  DROP_ATM_CELL(&atmch->atmid);
	  restore_flags(key);
	  pATM->err_stats.cell_drop_err++;
	  pATM->err_stats.rx_unk_conn_err++; 
	  done = 0;
	}

      /* Usage of the current buffer-descriptor done ?? */
    done_chk:
      if (done)
	{
	  save_and_cli(key);
	  ADD_STATUS_ENTRY(pATM, state, pdesc);
	  restore_flags(key);
	  state = 0;
	}
    }
  return done;
} /* atm_ir_status_poll */

/******************************************************************
** FUNCTION : sar_atm_isr()
**
** DESCRIPTION:
**	ATM Interface Interrupt Service Routine.
**
** NOTE: This routine is called when IR bit is set in ATM[0|1]S register.
**       Refer page 19-22 and 19-5
**
******************************************************************/
void sar_atm_isr(int irq, void *dev_id, struct pt_regs * regs)
{
  pATMIF_st	pATM = (pATMIF_st )gATM;

  volatile UInt32		ipend;
  UInt32		ir_set = 0;
  unsigned long key;

  static int atmIrqCount = 0;

  save_and_cli(key);

  ipend = interrupt.i4.ipend;
  if (ipend & 0x01)
    {

      /* Process the IR interrupt */
      if ((atm_ir_status_intr(pATM, &atm.ch0)))
	  ir_set = 1;
    }

  if (ipend & 0x02)
    {
      /* Process the IR interrupt */
      if ((atm_ir_status_intr(pATM, &atm.ch1)))
	  ir_set = 1;
    }
  if (ir_set)
    {
      /* Inhibit ATM IR interrupts */
      interrupt.i4.imask |= 0x03;

      /* Post an event to the status task */
      pATM->status &= ~(ATM_IR_INT_EN);
      post_status(pATM->OS_specific);
    }

  atmIrqCount++;

  restore_flags(key);

} /* sar_atm_isr */

/******************************************************************
** FUNCTION : final_poll_atm_ir()
**
** DESCRIPTION:
**	Poll ATM IR status bit. If the bit is not set, the IR interrupt 
**	is enabled. Else, the IR interrupt is processed & a semaphore is
**	given to invoke the status task.
**
** NOTE:
**	IR bit should be polled only when IR interrupt is turned off.
**
******************************************************************/
void final_poll_atm_ir(pATMIF_st pATM)
{
  UInt32		ipend;
  UInt32		if0_done, if1_done;
  unsigned long key;

  /* Continue only if the IR interrupt was previously disabled */
  if (!(pATM->status & ATM_IR_INT_EN))
    {
      ipend = interrupt.i4.ipend;

      if (ipend & 0x01)
	{
	  if ((atm_ir_status_poll(pATM, &atm.ch0)))
	    {
	      if0_done = 0;
	    }
	  else
	    if0_done = 1;
	}
      else
	if0_done = 1;

      if (ipend & 0x02)
	{
	  if ((atm_ir_status_poll(pATM, &atm.ch1)))
	    {
	      if1_done = 0;
	    }
	  else
	    if1_done = 1;
	}
      else
	if1_done = 1;

      save_and_cli(key);
      /* Is valid data is received on DMA channel 0? */
      if ((if0_done) && (if1_done))
	{
	  /* Check if any of the other ATM DMA interrupts is enabled */
	  if ((pATM->status & ATM_DMA_EVT_MASK) == ATM_DMA_EVT_MASK)
	    {
	      /* Enable the ATM IR interrupts */
	      pATM->status |= ATM_IR_INT_EN;
	      interrupt.i4.imask &= ~(0x03);
	    }
	}
      else
	{
	  post_status(pATM->OS_specific);
	}
      restore_flags(key);
    }

} /* final_poll_atm_ir */

/******************************************************************
** FUNCTION : poll_atm_ir()
**
** DESCRIPTION:
**	Poll the ATM IR periodically in the status task.
**
** NOTE:
**	IR bit should be polled only when IR interrupt is turned off.
**
******************************************************************/
void poll_atm_ir(pATMIF_st pATM)
{
  UInt32		ipend;

  if (!(pATM->status & ATM_IR_INT_EN))
    {
      ipend = interrupt.i4.ipend;

      if (ipend & 0x01)
	{
	  atm_ir_status_poll(pATM, &atm.ch0);	
	}

      if (ipend & 0x02)
	{
	  atm_ir_status_poll(pATM, &atm.ch1);	
	}
    }

} /* poll_atm_ir */

/******************************************************************
** FUNCTION : send_aal5_cell
**
** DESCRIPTION:
**	Transmit one cell out of the AAL5 PDU for the specified VC.
**
** NOTE:
**
******************************************************************/
void send_aal5_cell(pATM_VC_st pVC, UInt32 state)
{
  pATMIF_st	pATM = pVC->pif;
  pDMA_st		pdesc, ptemp;
  UInt32		cell_hdr, cell_cnt, size;
  UInt32		*plocal;
  void (* clbk)();

  struct sk_buff	*mbuf;

  cell_hdr = pVC->cell_hdr;
  cell_cnt = state & TX_CELL_CNT_MASK;

  if (cell_cnt)
    pdesc = pVC->tx_buf_desc;
  else
    {
      pdesc = pVC->tx_pdu_desc;
      cell_cnt = pdesc->cell_cnt;
    }

  cell_cnt--;

  if (cell_cnt == 0)
    {
      /* Insert the PDU length info in the last buffer */
      mbuf = pVC->tx_pdu_desc->mbuf_ptr;
      pVC->pad_buffer[12] = mbuf->len;
    }

  ptemp = pdesc;
  size = 0;

  /* Accumulate a cell worth of data */
  while(1)
    {
      size += ptemp->control & 0x3ffff;
      if (size < MAX_CELL_PAYLOAD)
	{
	  ptemp->link = ptemp->link_copy;
	  ptemp = (pDMA_st )PHYS_to_K1((UInt32 )ptemp->link_copy);
	}
      else
	{
	  ptemp->link = 0;
	  break;
	}
    }

  if (pVC->subport)
    {


      /* Wait for the right time to xmit a cell */
      while((dma.ch0.dmac & DMAxC_RUN) || (!(atm.ch1.atms & ATMS_OR))); 
                                                                        
      /* Set the cell header */
      if (cell_cnt)
	atm.ch1.atmodh = cell_hdr;
      else
	atm.ch1.atmodh = (cell_hdr | EPDU_DETECT);

    }
  else
    {
      /* Wait for the right time to xmit a cell */                                   
      while((dma.ch0.dmac & DMAxC_RUN) || (!(atm.ch0.atms & ATMS_OR)));  

      /* Set the cell header */
      if (cell_cnt)
	atm.ch0.atmodh = cell_hdr;
      else
	atm.ch0.atmodh = (cell_hdr | EPDU_DETECT);
    }

  dma.ch0.dmas &= ~(DMAxS_H); 
  /* Initiate DMA transfer */
  dma.ch0.dmadptr = VIRT_TO_PHYS((UInt32 )pdesc); 

  /* Wait for the transfer to complete */
  while(!(dma.ch0.dmas & (DMAxS_H)));
  dma.ch0.dmas &= ~(DMAxS_H);
		

  pVC->state &= ~TX_CELL_CNT_MASK;
  ptemp->link = ptemp->link_copy;
  if (cell_cnt == 0)
    {
      pdesc = pVC->tx_pdu_desc;
      clbk = pdesc->clbk;
      mbuf = pdesc->mbuf_ptr;
      pdesc->clbk = 0;

      /* Fetch the next PDU to be transmitted for the connection */
      if ((pVC->tx_pdu_desc = ptemp->chain_desc) == NULL)
	pVC->state &= ~DATA_PRESENT;

      do {
	ptemp = pdesc;
	pdesc = (pDMA_st )GET_NEXT_DESC(pdesc);

	plocal = (UInt32 *)ptemp;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
	*plocal++ = 0;
			
	FREE_DMA_DESC(pATM, ptemp);
	flush_cache_all();
      } while((VIRT_TO_PHYS((UInt32 )pdesc)));

      /* Call the transmit callback function */
      (*clbk)(pVC, mbuf);
      pVC->dma0_txpend--;    
    }
  else
    {
      if (ptemp->control & DMA_FINISHED)
	{
	  pdesc = (pDMA_st )GET_NEXT_DESC(ptemp);
	  pdesc->status = ptemp->status;
	  pVC->tx_buf_desc = pdesc;
	}
      else
	pVC->tx_buf_desc = ptemp;

      pVC->state |= cell_cnt;
    }	

} /* send_aal5_cell */

/******************************************************************
** FUNCTION : sar_xmit_dma0
**
** DESCRIPTION:
**  Transmit data (OAM cells, raw cells & AAL5 PDU) on channel 0.
**
** NOTE:
**
******************************************************************/
void sar_xmit_dma0(pATM_VC_st pVC)
{
  UInt32		state, cell_hdr = 0;
  UInt32		*plocal = 0;
  pDMA_st		pdesc = 0, ptemp;
  pATMIF_st	pATM = pVC->pif;
  void (* clbk)();

  struct sk_buff *mbuf;

  /* Is there any data enqueued for this connection? */
  if ((state = pVC->state) & DATA_PRESENT)
    {
      pdesc = pVC->tx_pdu_desc;
      if (pdesc->control_1)
	{
	  cell_hdr = (pVC->cell_hdr | (pdesc->control_1 << CH_PTI_SHIFT));

	  /* Increment statistics */
	  pVC->stats.tx_oam++;
	  pATM->vc_stats.tx_oam++;
	}

      /* AAL5 connection? */
      else if ((pVC->control & AAL_TYPE_MASK) == AAL5_TYPE)
	{
	  send_aal5_cell(pVC, state);
	  return;
	}
      else
	{
	  /* Raw cell */
	  cell_hdr = pVC->cell_hdr;
	}

      if (pVC->subport)
	{                
	  /* Wait for the right time to initiate transmission */
	  while((dma.ch0.dmac & DMAxC_RUN) || (!(atm.ch1.atms & ATMS_OR)));
	  
	  /* Set the cell header */ 
	  atm.ch1.atmodh = cell_hdr; 
	}
      else
	{
	  /* Wait for the right time to initiate transmission */
	  while((dma.ch0.dmac & DMAxC_RUN) || (!(atm.ch0.atms & ATMS_OR)));

	  /* Set the cell header */  
	  atm.ch0.atmodh = cell_hdr;  
	}
      dma.ch0.dmas &= ~(DMAxS_H); 

      /* Initiate DMA transfer */
      dma.ch0.dmandptr = VIRT_TO_PHYS((UInt32 )pdesc);

      /* Wait for the transfer to complete */
      while(!(dma.ch0.dmas & (DMAxS_H)));
      dma.ch0.dmas &= ~(DMAxS_H);

      /* Traverse the descriptor list to find last descriptor */
      clbk = pdesc->clbk;
      mbuf = pdesc->mbuf_ptr;
      while (pdesc->link)
	{
	  /* Free up intermediate descriptors */
	  ptemp = pdesc;
	  pdesc = (pDMA_st )GET_NEXT_DESC(pdesc);

	  plocal = (UInt32 *)ptemp;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;
	  *plocal++ = 0;

	  FREE_DMA_DESC(pATM, ptemp);
	}

      if ((pVC->tx_pdu_desc = pdesc->chain_desc) == NULL)
	state &= ~DATA_PRESENT;

      pVC->state = state;

      plocal = (UInt32 *)pdesc;	

      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
      *plocal++ = 0;
	
      FREE_DMA_DESC(pATM, pdesc);

      /* Call the transmit handler */
      (*clbk)(pVC, mbuf);
      pVC->dma0_txpend--; 
    }

} /* sar_xmit_dma0 */
