/***************************************************************
** File : sar_if.h
**
**    Definitions and data structures to interface to SOFTSAR
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
****************************************************************
**
**  Date         Developer                 Comment
****************************************************************
**  08/17/2000   srinivas voleti           Original
**     10/2001   P. Sadik                  Ported to linux
***************************************************************/

#ifndef __SAR_IF_H__
#define	__SAR_IF_H__


/* Two types of connection management mechanisms are supported.
** 1. 2-level lookup mechanism
** 2. Indexed lookup mechanism 
**
** This selection is done at compile time by defining TWO_LEVEL_LOOKUP or
** INDEXED_LOOKUP but not both. If one option is defined, then "undef" the
** other
*/
#define	TWO_LEVEL_LOOKUP
#undef	INDEXED_LOOKUP	

/* IO control commands */
#define	SARIO_VPT_SETUP			0	/* VP table setup */
#define	SARIO_VPT_PURGE			1	/* Purge VCs within for a VP table entry */
#define	SARIO_VC_SETUP			2	/* Connection set up */
#define	SARIO_VC_TEARDOWN		3	/* Connection teardown */
#define	SARIO_PURGE_VCS			4	/* Purge all VCs for the interface */
#define	SARIO_SET_STATE			5	/* Set the admin state of the interface */
#define	SARIO_GET_STATE			6	/* Read the oper state of the interface */
#define	SARIO_VC_VC_STATS		7	/* Get per conection VC statistics */
#define	SARIO_IF_VC_STATS		8	/* Get per interface VC  statistics */
#define	SARIO_ERR_STATS			9	/* Get error statistics */
#define	SARIO_IF_CLR_STATS		10	/* Clear interface specific statistics */
#define	SARIO_VC_CLR_STATS		11	/* Clear per VC statistics */
#define	SARIO_REG_READ			12	/* Read Banyan's internal register */
#define	SARIO_REG_WRITE			13	/* Write to Banyan internal register */
#define	SARIO_PHY_READ			14	/* Read from PHY register */
#define	SARIO_PHY_WRITE			15	/* Write to PHY register */
#define	SARIO_GET_ATMIF_ST		16	/* Read ATMIF data structure */
#define	SARIO_GET_VC_ST			17	/* Read VC specific data structure */
#define	SARIO_GET_STATUS_ST		18	/* Read head of the status queue */
#define	SARIO_XMIT				19	/* Transmit on a connection */
#define	SARIO_RECV				20	/* Receive on a connection */
#define	SARIO_GET_VC_HANDLE		21	/* Get the VC handle for a connection */
#define SARIO_SCHED_CONN                22      /* Schedule a connection */     
#define SARIO_DESCHED_CONN              23      /* De-schedule a connection */  

#define	PROTOCOL_PLACEHOLDER_SIZE	14

/* Definitions used for Indexed VC lookup mechanism */
#if defined (INDEXED_LOOKUP)
#define	DFLT_VP_MASK	0x000c0000		/* Only subport id & vpi  = 0 */
#define	DFLT_VC_MASK	0x0000001f		/* Maximum of 128 vcs */
#define	DFLT_VP_SHFT	13
#if 0
#define GET_VPCI(hdr) (((hdr & VP_MASK) >> VP_SHFT) | \
						((hdr & VC_MASK) >> VC_SHFT))
#endif
#else
#define	DFLT_VP_MASK	0
#define	DFLT_VC_MASK	0
#define	DFLT_VP_SHFT	0
#endif /* INDEXED_LOOKUP */
#define	VPI_MASK	DFLT_VP_MASK
#define	VCI_MASK	DFLT_VC_MASK
#define	VPI_SHFT	DFLT_VP_SHFT
#define	MAX_VPI		256

/***************************************************************************
** Configuration definitions for ATM IF 
****************************************************************************/

/* Administrative control enable (bit 31): 0 - disable, 1 - enable */
#define	ATMIF_ADMIN_UP				0x80000000
#define	ATMIF_ADMIN_DOWN			0x00000000

/* Cell format (bit 30): 0 - 53 byte cell format */
#define	ATMIF_NOR_CELL_FORMAT		0x00000000

/* Frame mode operation enable (bit 29): 0 - disabled, 1 - enabled */
#define FRAME_MODE          0x20000000
#define	NO_FRAME_MODE		0x00000000
#define	ATMIF_FRAME_MODE	FRAME_MODE

/* Clock mode select for the ATM interface (bit 28). 
** 	In Utopia mode, direction of transmit clock: 0 - input, 1 - output 
*/
#define	ATMIF_CMS_OUTPUT			0x10000000
#define	ATMIF_CMS_INPUT				0x00000000
#define	ATMIF_CLK_PRESCALAR_MASK	0x00ff0000
#define	ATMIF_CLK_PRESCALAR_SHFT	16

#define INDEXED_LKUP_EN		0x08000000
#define TWO_LEVEL_LKUP_EN	0x00000000
#if defined (INDEXED_LOOKUP)
/* Connection Management scheme (bit 27): 
**       0 - 2 level lookup, 1 - Indexed lookup
*/
#define LKUP_EN		INDEXED_LKUP_EN
#else
#define LKUP_EN		TWO_LEVEL_LKUP_EN
#endif /* INDEXED_LOOKUP */

/* Settings to select the active PHY interfaces */
#define	ATMIF_PHYS_MASK				0x000000f0
#define	ATMIF_PHY28					0x00000010
#define	ATMIF_PHY29					0x00000020
#define	ATMIF_PHY30					0x00000040

/* ATM Interface type setting */
#define	ATMIF_TYPE_MASK			0x0000000f

/* Hardware Types */
#define	HW_UTOPIA_1_TYPE			0
#define	HW_UTOPIA_2_TYPE			1
#define	HW_UTOPIA_2_SLAVE_TYPE		2
#define	HW_UTOPIA_1_SLAVE_TYPE		3
#define	HW_UTOPIA_2_ALC_TYPE		4
#define	HW_UTOPIA_2_RP351_TYPE		5

/* BOARD SPECIFIC SETTINGS */

#define	DFLT_PRESCALAR_VAL 8

#define	DFLT_SWITCH		0
#define	DFLT_PORT		0
#define	DFLT_IF			0
#define	PRESCALAR_SETTING	(DFLT_PRESCALAR_VAL << ATMIF_CLK_PRESCALAR_SHFT)
#define	MAX_VCCS		128

#define	CFG_UTO1 \
			( \
			 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_OUTPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATM_UTOPIA1 \
			)

#define	CFG_UTO2 \
			( \
			 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_OUTPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATMIF_PHY28 | \
			 ATMIF_PHY29 | ATMIF_PHY30 | ATM_UTOPIA2 \
			)

#define	CFG_UTO2_SLAVE \
			( \
			 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_INPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATMIF_PHY28 | \
			 ATMIF_PHY29 | ATMIF_PHY30 | ATM_UTOPIA2 \
			)

#define	CFG_UTO1_SLAVE \
			( \
			 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_INPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATM_UTOPIA1 \
			)

#define CFG_UTO2_ALC_SLAVE \
			( \
		   	 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_INPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATMIF_PHY28 |ATMIF_PHY29 |\
			 ATM_UTOPIA2 \
			)
#define CFG_UTO2_RP351 \
			( \
			 ATMIF_ADMIN_UP | ATMIF_NOR_CELL_FORMAT | \
			 ATMIF_FRAME_MODE | ATMIF_CMS_OUTPUT | LKUP_EN | \
			 PRESCALAR_SETTING | ATMIF_PHY28 | \
			 ATMIF_PHY29 | ATM_UTOPIA2 \
			)

/***********************************************************************
** Status definitions for ATM IF 
***********************************************************************/
/* Operational control state (bit 31): 0 - disable, 1 - enable */
#define	ATMIF_OPER_UP			0x80000000
#define	ATMIF_OPER_DOWN			0x00000000
#define	ATMIF_SP_BITS_MASK		0x00000003

/* configuration data structure */
typedef struct _t_ATMIF_CONFIG_st {
  UInt32		switch_no;
  UInt32		port_no;
  UInt32		if_no;
  UInt32		config;
  UInt32		max_VCCs;
  UInt32		vpi_mask;		/* Used with Indexed lookup only */
  UInt32		vci_mask;		/* Used with Indexed lookup only */
  UInt32		vpi_shft;		/* Used with Indexed lookup only */	
#if 0
  VOIDFUNCPTR     oam_tx_hdlr;
  VOIDFUNCPTR	oam_rx_hdlr;
  VOIDFUNCPTR	init_phy;
  FUNCPTR		get_phy;
  FUNCPTR		set_phy;
#endif
  void (*oam_tx_hdlr)(void *, void *);
  void (*oam_rx_hdlr)(void *, void *);
  void (*init_phy)(void);
  int  (*get_phy)();
  int  (*set_phy)();
  
} ATMIF_CONFIG_st, *pATMIF_CONFIG_st;

/* VP Table configuration for 2-level lookup mechanism */
typedef struct _t_VPT_CONFIG_st {
	UInt32	subport;
	UInt32	vpi;
	UInt32	mask;
	UInt32	bit_pos;
} VPT_CONFIG_st, *pVPT_CONFIG_st;

/* Definitions used for connection set up */
/* Bits 0-3 used for AAL type */
#define AAL_TYPE_MASK   0x0f
#define RAW_TYPE           	0
#define AAL1_TYPE           1
#define AAL2_TYPE           2
#define AAL5_TYPE           5
#define	OAM_TYPE			6

/* Bits 4-7 used for channel number */
#define CHAN_MASK           0xf0

/* Bit 31 for Connection admin state */
#define VPCI_ADMIN_UP		0x80000000
#define VPCI_ADMIN_DOWN		0x00000000

/* Bit 29 VC type : SVC or PVC */
#define PVC_TYPE    0
#define SVC_TYPE    1

/* Bits 27-28 Service type */
#define SERV_CLASS_MASK 0x18000000
#define UBR_TYPE    0
#define CBR_TYPE    1
#define VBR_TYPE    2
#define ABR_TYPE    3

/* Bits 8-9 used for xmit devcs */
#define	XMIT_DEVCS_MASK		0x00030000
#define	XMIT_DEVCS_SHFT		16

/* Bits 10-11 used for receive devcs */
#define	RCV_DEVCS_MASK		0x000c0000
#define	RCV_DEVCS_POS		18
#define	RCV_DEVCS_SHFT		2

/* VC configuration data structure */
typedef struct _t_VC_CONFIG_st {
  UInt32	subport;			/* Sub port */
  UInt32	vpi;				/* VPI */
  UInt32	vci;				/* VCI */
  void (*tx_hdlr)();
  void (*rx_hdlr)();

  void	*proto;				/* Parameter to be passed to the application */
  UInt32	control;
} VC_CONFIG_st, *pVC_CONFIG_st;

/* Cluster sizes */
#define	CL_SIZE_64		64
#define	CL_SIZE_128		128
#define	CL_SIZE_256		256
#define	CL_SIZE_512		512
#define	CL_SIZE_1024	1024
#define	CL_SIZE_2048	2048

//#define	ATM_SAR_MTU     1536  /* 32 cells = 1536 bytes */
#define	ATM_SAR_MTU     1920
//#define	ATM_SAR_MTU     2048

/* Error Codes */
#define	ERR_MAX_ACTIVE_VCS_EXCEEDED		1
#define	ERR_INV_LKUP_SCHEME				2
#define	ERR_ENTRY_EXISTS				3
#define	ERR_NULL_MEM					4
#define	ERR_MAX_VCCS_EXCEEDED			5
#define	ERR_NO_LEV1_ENTRY				6

/* GPIO pins */
#define GPIO_00 0x00000001	/* Not Usable */
#define GPIO_01 0x00000002	/* Not Usable */
#define GPIO_02 0x00000004	/* Not Usable */
#define GPIO_03 0x00000008	/* Not Usable */
#define GPIO_04 0x00000010	/* Usable - UART0 DTR */
#define GPIO_05 0x00000020	/* Not Usable */
#define GPIO_06 0x00000040	/* Usable - UART0 RTS */
#define GPIO_07 0x00000080	/* Not Usable */
#define GPIO_08 0x00000100	/* Not Usable */
#define GPIO_09 0x00000200	/* Not Usable */
#define GPIO_10 0x00000400	/* Not Usable */
#define GPIO_11 0x00000800	/* Not Usable */
#define GPIO_12 0x00001000	/* Not Usable */
#define GPIO_13 0x00002000	/* Not Usable */
#define GPIO_14 0x00004000	/* Not Usable */
#define GPIO_15 0x00008000	/* Not Usable */
#define GPIO_16 0x00010000	/* Not Usable */
#define GPIO_17 0x00020000	/* Not Usable */
#define GPIO_18 0x00040000	/* Usable - DMAREQ */
#define GPIO_19 0x00080000	/* Usable - DMADONE */
#define GPIO_20 0x00100000	/* Usable - USB SOF */
#define GPIO_21 0x00200000	/* Not Usable */
#define GPIO_22 0x00400000	/* Not Usable */
#define GPIO_23 0x00800000	/* Not Usable */
#define GPIO_24 0x01000000	/* Not Usable */
#define GPIO_25 0x02000000	/* Not Usable */
#define GPIO_26 0x04000000	/* Not Usable */
#define GPIO_27 0x08000000	/* Not Usable */
#define GPIO_28 0x10000000	/* Usable */
#define GPIO_29 0x20000000	/* Usable */
#define GPIO_30 0x40000000	/* Usable - INTREQ */
#define GPIO_31 0x80000000	/* Not Usable */

/* Macros to set GPIO pins - when used as outputs */
#define SET_GPIO(val) { \
		gpio.gpiod |= val; \
	}

/* Macros to clear GPIO pins - when used as outputs */
#define CLR_GPIO(val) { \
		gpio.gpiod &= ~val; \
	}

/* Function prototypes for the SOFTSAR API */
void *atm_register(void );
void atm_unregister(void *);
void atm_open(void );
void atm_close(void );
int atm_send(void);
int atm_receive(void);
int banyan_atm_ioctl(void *, UInt32, void *, UInt32, void *, UInt32, UInt32 *);
#endif /* !__SAR_IF_H__ */
