/***********************************************************
** FILE NAME: sar_oam.c
**
**    File contains the utilities to support OAM over the softSAR.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
***********************************************************
** DATE      DEVELOPER                 REASON
***********************************************************
** 10/12/00  srinivas voleti           original
** 11/1/01   P. Sadik                  Ported to linux.
***********************************************************/
/* Include files */
#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include <linux/string.h>
#include "sar_oam.h"
#include "sar_linux_specific.h"

extern pATMIF_st gATM;

/******************************************************************
** FUNCTION : oam_tx_hdlr()
**
** DESCRIPTION:
**    OAM cell generation.
**
** NOTE:
**	Only F5 flows are supported by the SOFTSAR.
**
******************************************************************/
void oam_tx_hdlr(pATM_VC_st pVC, struct sk_buff *mbuf)
{
  dev_kfree_skb(mbuf);

} /* oam_tx_hdlr */

/******************************************************************
** FUNCTION : oam_cell_gen()
**
** DESCRIPTION:
**    OAM cell generation.
**
** NOTE:
**	Only F5 flows are supported by the SOFTSAR.
**
******************************************************************/
void oam_cell_gen(UInt32 subport, UInt32 vpi, UInt32 vci, UInt32 pti, 
		  UByte oam_type, UByte func_type)
{
  pATMIF_st	pATM = gATM;
  pATM_VC_st	pVC;
  M_BLK_ID	pmbuf;
  UInt32		cell_hdr;
  char		*buffer;
  Byte		result;

  cell_hdr = BLD_CELL_HDR(subport, vpi, vci, 0, 0, 0);
  
  pVC = fetch_vc_handle(pATM, cell_hdr);

  if ((pti != OAM_F5_SEG_PTI) && (pti != OAM_F5_END_2_END_PTI))
    {
      printk("Invalid PTI field<0x%x> specified for F5 flow\n",(int)pti);	
      return;
    }

  if( (pmbuf = dev_alloc_skb( (unsigned int)CL_SIZE_64 + 2)) == NULL)
    {
      pATM->err_stats.insuff_bufs_err++;
      printk("Failed to allocate mbuf for OAM cell\n");
      return;
    }

  /* Initialize the mbuf structure */
  skb_put(pmbuf,MAX_CELL_PAYLOAD);

  *buffer = JOIN_OAM_FUNC_FIELDS(oam_type, func_type);

  /* Call the send routine */
  result = banyan_atm_ioctl(pVC, SARIO_XMIT, pmbuf, OAM_TYPE, \
		     (void *)oam_tx_hdlr, pti, 0);

  if (result != OK)
    {
      printk("Send PDU failed \n");
      return;
    }

  /* Wait at the TSQ for acknowledgement */

} /* oam_cell_gen */

/******************************************************************
** FUNCTION : oam_cell_rx()
**
** DESCRIPTION:
**    Process the received OAM cell.
**
** NOTE:
**
******************************************************************/
void oam_cell_rx(pATM_VC_st pVC, M_BLK_ID m)
{
  UByte	*data, oam_type, func_type;
  data = (UByte *)m->data;

  oam_type = GET_OAM_TYPE(*data);
  func_type = GET_FUNC_TYPE(*data);

  /* Process the OAM cell */
  printk("Received OAM cell\n");
  printk("Subport = %x  VPI = %x  VCI = %x\n", 
	 (int )pVC->subport, (int )pVC->vpi, (int )pVC->vci);
  printk("OAM type = %x  Function type = %x\n", (int )oam_type, \
	 (int )func_type);

  printk("Description: ");

  /* OAM Cell handling */
  switch(oam_type)
    {
    case OAM_FLT_MGMT_TYPE:
      switch(func_type) {
      case OAM_FLT_AIS:
	printk("Fault Management(AIS)\n");
	break;

      case OAM_FLT_RDI:
	printk("Fault Management(RDI)\n");
	break;

      case OAM_FLT_CONT_CHK:
	printk("Fault Management(Continuity Check)\n");
	break;

      case OAM_FLT_LPBK:
	printk("Fault Management(Loopback)\n");
	break;

      default:
	printk("Fault Management(Unknown)\n");
	break;
      };
      break;

    case OAM_PERF_MGMT_TYPE:
      switch(func_type) {
      case OAM_PERF_FWD_MON:
	printk("Performance Management(Forward Monitoring)\n");
	break;

      case OAM_PERF_BCK_REP:
	printk("Performance Management(Backward Monitoring)\n");
	break;

      default:
	printk("Performance Management(Unknown)\n");
	break;
      };
      break;

    case OAM_ACT_DEACT_TYPE:
      switch(func_type) {
      case OAM_ACT_DEACT_PERF_MON:
	printk("Activation-Deactivation(Performance Monitoring)\n");
	break;

      case OAM_ACT_DEACT_CONT_CHK:
	printk("Activation-Deactivation(Continuity Check)\n");
	break;

      default:
	printk("Activation-Deactivation(Unknown)\n");
	break;
      };
      break;

    case OAM_SYS_MGMT_TYPE:
      printk("System Management(Unknown)\n");
      break;

    default:
      printk("Unknown OAM Type\n");
      break;
    };

  dev_kfree_skb(m);
} /* oam_cell_rx */

