/*****************************************************************
**  File : sar_process.c
**
**		SOFTSAR tasks viz. the scheduler & status processing tasks.
**		Additionally, it includes the transmit function.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  09/13/2000   srinivas voleti           Original
**  10/10/2001   P. Sadik                  Ported to Linux
******************************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/slab.h> 
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <asm/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <linux/string.h>
#include <linux/config.h>
#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"
#include "ame/rldrun.h"

#define FM_XMIT_DONE    (DMA_FINISHED | DMA_DEVCMD(CRC32_LD))
#define FM_XMIT_MASK    (DMA_FINISHED | DMA_DEVCMD(CRC32_FLD))

extern SIZE_2_CELL_CONVERTER_st size_2_cell_table[];
extern void final_poll_atm_ir(pATMIF_st );
extern void poll_atm_ir(pATMIF_st );

extern void *gATM;

UInt32 fix_enable = 1;
/******************************************************************
** FUNCTION : process_rx_descs()
**
** DESCRIPTION:
**	Process the receive descriptors for AAL5 traffic 
**
** NOTE:
**
******************************************************************/
pATM_VC_st	crunch_pvc_array[8];
UInt32		atm_crunch_state = 0;

void process_rx_descs(pATMIF_st pATM, pDMA_st pdesc, UInt32 chan)
{
  volatile pDMA_st	pnext;
  M_BLK_ID		m;
  pATM_VC_st		pVC;
  UInt32		len, cells, *ptemp;
  UByte			*plen;
  unsigned long key;

	/* Process all the used descriptors */
  do
    {
	//	printk("CRC error. Dropping PDU.\n");
      /* CRC32 error ? */
      if (pdesc->status != CRC32_MAGIC_CODE){
	/* Drop the PDU */
	pATM->err_stats.CRC32_err++;

	/* Get next descriptor pointer */
	if (pdesc->link)
	  pnext = (pDMA_st )GET_NEXT_DESC(pdesc);
	else
	  pnext = pdesc->chain_desc;
	
	save_and_cli(key);
	pATM->curr_desc[chan] = pnext;
	restore_flags(key);

	save_and_cli(key);
	free_2k_buf_desc(pATM, pdesc);
	restore_flags(key);

	pdesc = pnext;
      }
      else
	{
	  /* Process the descriptor-buffer */
	  pVC = pdesc->ptr;
	  plen = (UByte *)  K0_TO_K1((PHYS_TO_VIRT(((UInt32 )pdesc->curr_addr - 6))));
	  len = ((*plen << 8) | *(++plen));  

	  if (len > (ATM_SAR_MTU - 8))
	    {
	      /* Invalid Size for Rx PDU */
	      pATM->err_stats.inv_rxsize_err++;

	      /* Get next descriptor pointer */
	      if (pdesc->link)
		pnext = (pDMA_st )GET_NEXT_DESC(pdesc);
	      else
		pnext = pdesc->chain_desc;
	    
	      save_and_cli(key);
	      pATM->curr_desc[chan] = pnext;
	      restore_flags(key);

	      save_and_cli(key);
	      free_2k_buf_desc(pATM, pdesc);
	      restore_flags(key);

	      pdesc = pnext;
	    }
	  else
	    {  
	      cells = pdesc->buf_size - (pdesc->control & 0x3ffff);
	      cells /= MAX_CELL_PAYLOAD;
	    
	      /* Build an mbuf from the descriptor list */
	      m = (M_BLK_ID )pdesc->mbuf_ptr;
	      dma_cache_inv((unsigned long)m, len);
	      skb_put(m,len);

	      /* Get next descriptor pointer */
	      if (pdesc->link)
		pnext = (pDMA_st )GET_NEXT_DESC(pdesc);
	      else
		pnext = pdesc->chain_desc;

	      save_and_cli(key);
	      pATM->curr_desc[chan] = pnext;
	      restore_flags(key);
	      /* Free up descriptor to free list */
	      ptemp = (UInt32 *)pdesc;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	      *ptemp++= 0;
	
	      save_and_cli(key);
	      FREE_DMA_DESC(pATM, pdesc);
	      restore_flags(key);
	
	      /* Update statistics */
	      pATM->vc_stats.rx_pdus++;
	      pATM->vc_stats.rx_cells += cells;
	      pVC->stats.rx_pdus++;
	      pVC->stats.rx_cells += cells;
	    
	      /* Call receive handler */
	      if(*pVC->rx_hdlr){
		(*pVC->rx_hdlr)(pVC, m, len);
	      }
	      else
		{
		    printk("process_rx_descs: no rx_hdlr.\n");
		  /* Can use the queues !! */
		}

	      pdesc = pnext;
	    }

	  /* Process the ATM IR status bit & initiate transfer over 
	   DMA channel 0 */

	  poll_atm_ir(pATM);
	}
    } while ((pdesc) && (pdesc->control & DMA_DONE));

} /* process_rx_descs */

/******************************************************************
** FUNCTION : process_dma0_rx_aal5_descs()
**
** DESCRIPTION:
**	Process the receive descriptors on DMA channel 0 for AAL5 traffic 
**
** NOTE:
**
******************************************************************/
void process_dma0_rx_aal5_descs(pATMIF_st pATM, pDMA_st pdesc)
{
  M_BLK_ID			m;
  pATM_VC_st			pVC;
  UInt32				len, cells, *ptemp;
  UByte				*plen;
  unsigned long key;

  /* CRC32 error ? */
  if (pdesc->status != CRC32_MAGIC_CODE)
    {
      /* Drop the PDU */
      pATM->err_stats.CRC32_err++;
      save_and_cli(key);
      free_2k_buf_desc(pATM, pdesc);
      restore_flags(key);
    }
  else
    {
      /* Process the descriptor-buffer */
      pVC = pdesc->ptr;

      plen = (UByte *)\
	K0_TO_K1((PHYS_TO_VIRT(((UInt32 )pdesc->curr_addr - 6))));
      len = ((*plen << 8) | *(++plen));
      cells = pdesc->buf_size - (pdesc->control & 0x3ffff);
      cells /= MAX_CELL_PAYLOAD;

      /* Build an mbuf from the descriptor list */
      m = (M_BLK_ID )pdesc->mbuf_ptr;
      skb_put(m,len);
      /* Free up descriptor to free list */
      ptemp = (UInt32 *)pdesc;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;

      save_and_cli(key);
      FREE_DMA_DESC(pATM, pdesc);
      restore_flags(key);

      /* Update statistics */
      pATM->vc_stats.rx_pdus++;
      pATM->vc_stats.rx_cells += cells;
      pVC->stats.rx_pdus++;
      pVC->stats.rx_cells += cells;

      /* Call receive handler */
      if(*pVC->rx_hdlr)
	(*pVC->rx_hdlr)(pVC, m, len);
      else
	{
	  /* Can use the queues !! */
	}
    }

} /* process_dma0_rx_aal5_descs */

/******************************************************************
** FUNCTION : process_rx_oam_descs()
**
** DESCRIPTION:
**	Process the receive descriptors on DMA channel 0 for OAM traffic 
**
** NOTE:
**
******************************************************************/
void process_rx_oam_descs(pATMIF_st pATM, pDMA_st pdesc)
{
  M_BLK_ID			m;
  pATM_VC_st			pVC;
  UInt32				*ptemp;
  unsigned long key;

  /* CRC10 error ? */
  if (pdesc->status != CRC10_MAGIC_CODE)
    {
      /* Drop the PDU */
      pATM->err_stats.CRC10_err++;
      save_and_cli(key);
      free_64_buf_desc(pATM, pdesc);
      restore_flags(key);
    }
  else
    {
      /* Process the descriptor-buffer */
      pVC = pdesc->ptr;

      /* Build an mbuf from the descriptor list */
      m = (M_BLK_ID )pdesc->mbuf_ptr;
      skb_put(m,MAX_CELL_PAYLOAD);
      /* Free up descriptor to free list */
      ptemp = (UInt32 *)pdesc;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      *ptemp++= 0;
      save_and_cli(key);
      FREE_DMA_DESC(pATM, pdesc);
      restore_flags(key);

      /* Call receive handler */
      pATM->vc_stats.rx_oam++;
      pVC->stats.rx_oam++;
      (*pATM->oam_rx_hdlr)(pVC, m);
    }
	
} /* process_rx_oam_descs */

/******************************************************************
** FUNCTION : process_rx_raw_cell_descs()
**
** DESCRIPTION:
**	Process the raw cell receive descriptors.
**
** NOTE:
**
******************************************************************/
void process_rx_raw_cell_descs(pATMIF_st pATM, pDMA_st pdesc)
{
	M_BLK_ID			m;
	pATM_VC_st			pVC;
	UInt32				*ptemp;
	unsigned long key;

	/* Process the descriptor-buffer */
	pVC = pdesc->ptr;

	/* Build an mbuf from the descriptor list */
	m = (M_BLK_ID )pdesc->mbuf_ptr;
	skb_put(m,MAX_CELL_PAYLOAD);

	/* Free up descriptor to free list */
	ptemp = (UInt32 *)pdesc;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;
	*ptemp++= 0;

	save_and_cli(key);
	FREE_DMA_DESC(pATM, pdesc);
	restore_flags(key);

	/* Call receive handler */
	if(*pVC->rx_hdlr)
		(*pVC->rx_hdlr)(pVC, m, MAX_CELL_PAYLOAD);

	pATM->vc_stats.rx_raw_cells++;
	pVC->stats.rx_raw_cells++;

} /* process_rx_raw_cell_descs */

/******************************************************************
** FUNCTION : sar_status_task()
**
** DESCRIPTION:
**	AAL5 Status Queue Task
**
** NOTE:
**
******************************************************************/
void sar_status_task(pATMIF_st pATM)
{
  volatile pSTATUS_st	ps;
  UInt32		chan;
  UInt32		evt_mask;
  pDMA_st		pdesc, pnext;
  volatile DMA_CHAN     *pdma_channel;
  UInt32		state, if0_state, if1_state;
  pATM_VC_st		pVC;
  unsigned long         key;

  ps = pATM->status_head;

  while (ps != pATM->status_tail)
    {
      switch (ps->state & STATE_MASK) {

      case RXINT_STATE:
	for (chan = 1; chan < 9; chan++)
	  {
	    pdesc = pATM->curr_desc[chan];

	    /* Discard PDU > 2016 bytes */
	    while ((pdesc) && (pdesc->control & DMA_FINISHED))
	      {
		if (pdesc->control & DMA_DONE)
		  pATM->err_stats.inv_rxsize_err++;
		
		if (pdesc->link)
		  pnext = (pDMA_st )GET_NEXT_DESC(pdesc);
		else
		  pnext = pdesc->chain_desc;
		
		save_and_cli(key);
		free_2k_buf_desc(pATM, pdesc);
		restore_flags(key);
		pdesc = pnext;
		pATM->curr_desc[chan] = pdesc;
	      }   

	    if ((pdesc) && (pdesc->control & DMA_DONE))
	      {
		process_rx_descs(pATM, pdesc, chan);
	      }
	  }

	    /* Queue in more free descriptor-buffer pairs, if needed */
	    if (pATM->free_2k_bds <= BD2K_WATER_MARK)
	      add_free_2k_bd_list(pATM, (UInt32 )250);

	if (atm_crunch_state)
	  {
	    interrupt.i4.imask |= 0x03;
	    state = atm_crunch_state;
	    if0_state = 0;
	    if1_state = 0;
	    chan = 0;
	    while(state)
	      {
		if (state & 0x1)
		  {
		    pdma_channel = (DMA_CHAN *)((UInt32 )&dma.ch1 + chan * sizeof(DMA_CHAN));

		    save_and_cli(key);
		    pVC = crunch_pvc_array[chan];
		    pdesc = get_2k_buf_desc_list(pATM, pVC, NUM_BDS_PER_LIST);

		    if (pATM->curr_desc[chan + 1] == NULL)
		      pATM->curr_desc[chan + 1] = pdesc;

		    atm_crunch_state &= ~(1 << chan);
		    
		    if (!(pdma_channel->dmac & DMAxC_RUN))
		      {
			if (fix_enable)
			  {
			    if (pVC->subport)
			      if1_state |= (1 << chan);
			    else
			      if0_state |= (1 << chan);	

			    atm.atmvccd |= (1 << chan);
			  }
			while(!(pdma_channel->dmas & DMAxS_H));
		      }
		    pdma_channel->dmandptr = VIRT_TO_PHYS((UInt32 )pdesc);
		    restore_flags(key);
		  }
		state >>= 1;
		chan++;
	      }

	    if (fix_enable)
	      {
		if (if0_state)					
		  {
		    while (!(atm.ch0.atms & ATMS_IR));
		    atm.atmvccd &= ~if0_state;
		    DROP_ATM_CELL(&atm.ch0.atmid);
		  }
		if (if1_state)
		  {
		    while (!(atm.ch1.atms & ATMS_IR));
		    atm.atmvccd &= ~if1_state;
		    DROP_ATM_CELL(&atm.ch1.atmid);
		  }
	      }
	    interrupt.i4.imask &= ~0x03;
	  }
	break;

      case RX_AAL5_DMA0_STATE:
	process_dma0_rx_aal5_descs(pATM, ps->desc);

	/* Queue in more free descriptor-buffer pairs,
	   if needed */
	if (pATM->free_2k_bds <= BD2K_WATER_MARK)
	  {
	    add_free_2k_bd_list(pATM, (UInt32 )250);
	  }
	break;

      case RAW_CELL_STATE:
	process_rx_raw_cell_descs(pATM, ps->desc);

	/* Queue in more free descriptor-buffer pairs,
	   if needed */
	if (pATM->free_64_bds <= BD64_WATER_MARK)
	  add_free_64_bd_list(pATM, (UInt32 )125);
	break;

      case RX_OAM_STATE:
	process_rx_oam_descs(pATM, ps->desc);

	/* Queue in more free descriptor-buffer pairs,
	   if needed */
	if (pATM->free_64_bds <= BD64_WATER_MARK)
	  add_free_64_bd_list(pATM, (UInt32 )125);
	break;

      default:
	break;
      };

      /* Next status entry */
      ps++;
      if (ps > pATM->status_last)
	ps = pATM->status_base;
      pATM->status_head = ps;
    }
  /* Do a final check on ATM IR status & take necessary action */
  final_poll_atm_ir(pATM);
  save_and_cli(key);
  evt_mask = (1 << ATM_DMA_EVT_SHFT);
  pATM->status |= evt_mask;
  restore_flags(key);

} /* sar_status_task */

/******************************************************************
** FUNCTION : sar_enqueue_raw()
**
** DESCRIPTION:
**  Build and enqueue a raw cell.
**
** NOTE:
**
******************************************************************/
int sar_enqueue_raw(pATM_VC_st pVC, void *ptr1, void (* clbk)(), UInt32 parm)
{
  struct sk_buff	*m = (struct sk_buff *)ptr1, *m1;
  pATMIF_st	pATM = pVC->pif;
  pDMA_st		pdesc, ptemp;
  UInt32		val, ds;
  unsigned long key;

  /* If the length is not one cell, return ERROR */

  if (m->len != MAX_CELL_PAYLOAD)
    return ERROR;

  /* Compute the number of descriptors needed */
  for (m1 = m, val = 1; m1->next; m1 = m1->next)
    val++;

  /* Build the descriptor list for the mbuf */
  save_and_cli(key);
  GET_DMA_DESC_LIST(pATM, pdesc, val);
  restore_flags(key);

  if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
    {
      pATM->err_stats.insuff_tx_desc_err++; 
    return ERROR;
    }

  pVC->dma0_txpend++; 

  if (pVC->subport)
    ds = DS_ATM_1_OUTPUT;
  else
    ds = DS_ATM_0_OUTPUT;

  ptemp = pdesc;
	
  ptemp->control = SET_DMA_DESC_CTRL(NOCRC, ds, m->len);
  ptemp->status     = 0xffffffff;
  ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
  ptemp->clbk = clbk;
  ptemp->mbuf_ptr = m;
  ptemp->chain_desc = 0;
	
  m = m->next;

  while(m)
    {
      ptemp = (pDMA_st )GET_NEXT_DESC(ptemp);
      ptemp->control = SET_DMA_DESC_CTRL(NOCRC, ds, m->len);
      ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
      ptemp->chain_desc = 0;
      m = m->next;
    }

  ptemp->link = 0;

  /* Update statistics */
  pATM->vc_stats.tx_raw_cells++;
  pVC->stats.tx_raw_cells++;

  /* Chain the PDU to the list */
  pdesc->cell_cnt = 1;
  save_and_cli(key);
  if (pVC->tx_pdu_desc == 0)
    pVC->tx_pdu_desc = pdesc;
  else
    pVC->tx_last_pdu_desc->chain_desc = pdesc;

  pVC->tx_last_pdu_desc = pdesc;

  pVC->state |= DATA_PRESENT;
  restore_flags(key);
	
  return OK;
} /* sar_enqueue_raw */

/******************************************************************
** FUNCTION : sar_enqueue_oam()
**
** DESCRIPTION:
**  Build and enqueue a raw cell.
**
** NOTE:
**
******************************************************************/
int sar_enqueue_oam(pATM_VC_st pVC, void *ptr1, void (* clbk)(), UInt32 oam_type)
{
  struct sk_buff  *m = (struct sk_buff *)ptr1, *m1;
  pATMIF_st	pATM = pVC->pif;
  pDMA_st		pdesc, ptemp;
  UInt32		val, ds;
  unsigned long key;

  /* If the length is not one cell, return ERROR */
  if (m->len != MAX_CELL_PAYLOAD)
    {
      pATM->err_stats.inv_txsize_err++; 
      return ERROR;
    }

  /* Compute the number of descriptors needed */
  for (m1 = m, val = 1; m1->next; m1 = m1->next)
    val++;

  /* Build the descriptor list for the mbuf*/
  save_and_cli(key);
  GET_DMA_DESC_LIST(pATM, pdesc, val);
  restore_flags(key);

  if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
    {
      pATM->err_stats.insuff_tx_desc_err++;  
      return ERROR;
    }

  if (pVC->subport)
    ds = DS_ATM_1_OUTPUT;
  else
    ds = DS_ATM_0_OUTPUT;

  ptemp = pdesc;
	
  ptemp->control = SET_DMA_DESC_CTRL(CRC10, ds, m->len);
  ptemp->status     = 0xffffffff;
  ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
  ptemp->control_1 = oam_type;
  ptemp->clbk = clbk;
  ptemp->mbuf_ptr = m;
  ptemp->chain_desc = 0;
	
  m = m->next;

  while(m)
    {
      ptemp = (pDMA_st )GET_NEXT_DESC(ptemp);
      ptemp->control = SET_DMA_DESC_CTRL(CRC10, ds, m->len);
      ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
      ptemp->chain_desc = 0;
      m = m->next;
    }

  ptemp->link = 0;

  /* Update statistics */
  pATM->vc_stats.tx_oam++;
  pVC->stats.tx_oam++;

  /* Chain the PDU to the list */
  pdesc->cell_cnt = 1;
  save_and_cli(key);
  if (pVC->tx_pdu_desc == 0)
    pVC->tx_pdu_desc = pdesc;
  else
    pVC->tx_last_pdu_desc->chain_desc = pdesc;

  pVC->tx_last_pdu_desc = pdesc;

  pVC->state |= DATA_PRESENT;
  restore_flags(key);
	
  return OK;
} /* sar_enqueue_oam */

/******************************************************************
** FUNCTION : sar_enqueue_aal5()
**
** DESCRIPTION:
**  Build and enqueue a AAL5 PDU.
**
** NOTE:
**
******************************************************************/
int sar_enqueue_aal5(pATM_VC_st pVC, void *ptr1, UInt32 plen,void (* clbk)())
{
  struct sk_buff			*m = (struct sk_buff *)ptr1, *m1;
  pATMIF_st			pATM = pVC->pif;
  pDMA_st				pdesc, ptemp;
  pSIZE_2_CELL_CONVERTER_st	psize;
  UByte				*p;
  UInt32				val, ds;
  unsigned long key;


  if (pVC->dma0_txpend >= MAX_DMA0_TXPEND_VAL) 
    {                                      
      pATM->err_stats.txq_exceeded_err++;
      return ERROR;                      
    }                                          

  /* Determine number of bytes for padding */
  psize = &size_2_cell_table[plen];

  /* Compute the number of descriptors needed for the data buffer */
  for (m1 = m, val = 2; m1->next; m1 = m1->next)
    val++;	

  /* Build the descriptor list for the mbuf */
  save_and_cli(key);
  GET_DMA_DESC_LIST(pATM, pdesc, val);
  restore_flags(key);

  if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
    {
      pATM->err_stats.insuff_tx_desc_err++;    
      return ERROR;
    }

  pVC->dma0_txpend++; 

  if (pVC->subport)
    ds = DS_ATM_1_OUTPUT;
  else
    ds = DS_ATM_0_OUTPUT;

  ptemp = pdesc;
  ptemp->control = SET_DMA_DESC_CTRL(CRC32, ds, m->len);  

  ptemp->status     = 0xffffffff;
  ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
  ptemp->link_copy = ptemp->link;
  ptemp->clbk = clbk;
  ptemp->mbuf_ptr = m;
  ptemp->chain_desc = 0;
  ptemp->ptr = pVC;

  ptemp = (pDMA_st )GET_NEXT_DESC(ptemp);
  m = m->next;
  while(m)
    {
      ptemp->control = SET_DMA_DESC_CTRL(CRC32, ds, m->len);
      ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )m->data);
      ptemp->link_copy = ptemp->link;
      ptemp->chain_desc = 0;
      ptemp->ptr = pVC;
      m = m->next;
      ptemp = (pDMA_st )GET_NEXT_DESC(ptemp);
    }

  ptemp->chain_desc = 0;
  ptemp->link = 0;
  p = (UByte *)pVC->pad_buffer;
  p += (PAD_BUFFER_SIZE - psize->pad_size);
  ptemp->control    = SET_DMA_DESC_CTRL(CRC32_LD, ds, psize->pad_size);
  ptemp->curr_addr  = VIRT_TO_PHYS((UInt32 )p);
  ptemp->ptr = pVC;
  ptemp->link_copy = 0;

  /* Update statistics */
  pATM->vc_stats.tx_pdus++;
  pVC->stats.tx_pdus++;
  pATM->vc_stats.tx_cells += psize->cells;
  pVC->stats.tx_cells += psize->cells;

  /* Chain the PDU to the list */
  pdesc->cell_cnt = psize->cells;
  save_and_cli(key);
  if (pVC->tx_pdu_desc == 0)
    {
      pVC->tx_pdu_desc = pdesc;
    }
  else
    {
      pVC->tx_last_pdu_desc->chain_desc = pdesc;
    }

  pVC->tx_last_pdu_desc = pdesc;

  pVC->state |= DATA_PRESENT;
  restore_flags(key);

  return OK;

} /* sar_enqueue_aal5 */

/******************************************************************
** FUNCTION : sar_xmit_aal5_if0()
**
** DESCRIPTION:
**	Transmit AAL5 PDU on interface 0 using frame mode transfer.
**
** NOTE: At present, we poll till the dma is finished. This can be
** improved by using a descriptor queue. However, for RP355, the
** uplink speed is anyway limited. Hence, we use one descriptor
** for the dma and not a queue.
******************************************************************/
int sar_xmit_aal5_if0(pATM_VC_st pVC, struct sk_buff *skb, UInt32 plen, void (* clbk)())
{
  pATMIF_st	            pATM  = pVC->pif;
  pDMA_st	            pdesc = pATM->if0_tx_desc;
  pDMA_st	            ptemp;
  pSIZE_2_CELL_CONVERTER_st psize;
  UByte		            *p;
  volatile UInt32           *k1_ptr;
  unsigned long             key;
  unsigned long             dummyRead;
  static unsigned int       count  = 0;
  unsigned int              retVal = 0;

  save_and_cli(key);
  /* make sure payload gets written to memory */
  dma_cache_wback_inv((unsigned long)skb->data, skb->len);

  /* Determine number of bytes for padding */
  psize               = &size_2_cell_table[plen];
  ptemp               = pdesc;
  ptemp->control      = SET_DMA_DESC_CTRL(CRC32_FD, DS_ATM_FMC0, skb->len);
  ptemp->status       = 0xffffffff;
  ptemp->curr_addr    = VIRT_TO_PHYS((UInt32 )skb->data);	

  ptemp               = (pDMA_st )GET_NEXT_DESC(ptemp);
  ptemp->link         = 0;
  plen                = htonl(plen);
  pVC->pad_buffer[12] = plen;
  p                   = (UByte *)pVC->pad_buffer;
  p                  += (PAD_BUFFER_SIZE - psize->pad_size);

  dma_cache_wback_inv((unsigned long)p, psize->pad_size);

  ptemp->control      = SET_DMA_DESC_CTRL(CRC32_LD, DS_ATM_FMC0, psize->pad_size);
  ptemp->curr_addr    = VIRT_TO_PHYS((UInt32 )p);

  atm.atm0oh          = pVC->cell_hdr;

  /* Set up the DMA channel & enable transmission */
  dma.ch7.dmandptr    = VIRT_TO_PHYS((UInt32 )pdesc);

  /* Wait for the transfer to be complete */
  /* Should be polling in the memory & NOT CACHE!! - so get the K1 seg address */
  k1_ptr              = &ptemp->control;

#ifdef CONFIG_IDT_79EB355
  while(!(*k1_ptr & DMA_FINISHED));
#else
  /* On RP355 board, when the DSL link is down, the dma engine cannot transfer
     the data. So, we drop the packets here till the link is up again. */

  for(;;)
    {
      if( (*k1_ptr & DMA_FINISHED))
        {
	  count  = 0;
	  retVal = OK;
          break;
        }
      else if(count++>0xFFFFF)
        {
	  dma.ch7.dmac    &= ~DMAxC_RUN;
	  while(!(dma.ch7.dmas & DMAxS_H));
	  dummyRead        = dma.ch7.dmadptr;
	  dma.ch7.dmadptr  = 0x0;
	  dummyRead        = dma.ch7.dmandptr;
	  dma.ch7.dmandptr = 0x0;
	  count            = 0;
	  retVal           = 1;
	  break;
        }
    }
#endif

  /* Call the transmit callback function */
  (*clbk)(pVC, skb);
  restore_flags(key);

  if(retVal == OK)
    {
  /* Update statistics */
      pATM->vc_stats.tx_pdus++;
      pVC->stats.tx_pdus++;
      pATM->vc_stats.tx_cells += psize->cells;
      pVC->stats.tx_cells += psize->cells;
    }

  return retVal;
} /* sar_xmit_aal5_if0 */

/******************************************************************
** FUNCTION : sar_xmit_aal5_if1()
**
** DESCRIPTION:
**	Transmit AAL5 PDU on interface 1 using frame mode transfer.
**
** NOTE:
**	
******************************************************************/
int sar_xmit_aal5_if1(pATM_VC_st pVC, struct sk_buff *skb, UInt32 plen, void(* clbk)())
{
  pATMIF_st		    pATM = pVC->pif;
  pDMA_st		    pdesc = pATM->if1_tx_desc;
  pDMA_st		    ptemp;
  pSIZE_2_CELL_CONVERTER_st psize;
  UByte			    *p;
  volatile UInt32	    *k1_ptr;
  unsigned long             key;
  unsigned long             dummyRead;
  static unsigned int       count  = 0;
  unsigned int              retVal = 0;
  save_and_cli(key);

  /* make sure payload gets written to memory */
  dma_cache_wback_inv((unsigned long)skb->data, skb->len);

  /* Determine number of bytes for padding */
  psize = &size_2_cell_table[plen];
  ptemp = pdesc;
  ptemp->control      = SET_DMA_DESC_CTRL(CRC32_FD, DS_ATM_FMC1, skb->len);
  ptemp->status       = 0xffffffff;
  ptemp->curr_addr    = VIRT_TO_PHYS((UInt32 )skb->data);	

  ptemp               = (pDMA_st )GET_NEXT_DESC(ptemp);
  ptemp->link         = 0;
  plen                = htonl(plen);
  pVC->pad_buffer[12] = plen;
  p                   = (UByte *)pVC->pad_buffer;
  p                  += (PAD_BUFFER_SIZE - psize->pad_size);

  dma_cache_wback_inv((unsigned long)p, psize->pad_size);

  ptemp->control      = SET_DMA_DESC_CTRL(CRC32_LD, DS_ATM_FMC1, psize->pad_size);
  ptemp->curr_addr    = VIRT_TO_PHYS((UInt32 )p);

  atm.atm1oh = pVC->cell_hdr;

  /* Set up the DMA channel & enable transmission */
  dma.ch8.dmandptr    = VIRT_TO_PHYS((UInt32 )pdesc);

  /* Wait for the transfer to be complete */
  /* Should be polling in the memory & NOT CACHE!! - so get the K1 seg
     address */
  k1_ptr              = &ptemp->control;

#ifdef CONFIG_IDT_79EB355
  while(!(*k1_ptr & DMA_FINISHED));
#else
  /* On RP355 board, when the DSL link is down, the dma engine cannot transfer
     the data. So, we drop the packets here till the link is up again. */

  for(;;)
    {
      if( (*k1_ptr & DMA_FINISHED))
        {
          count  = 0;
	  retVal = OK;
          break;
        }
      else if(count++>0xFFFFF)
        {
          dma.ch8.dmac    &= ~DMAxC_RUN;
          while(!(dma.ch8.dmas & DMAxS_H));
          dummyRead        = dma.ch8.dmadptr;
          dma.ch8.dmadptr  = 0x0;
          dummyRead        = dma.ch8.dmandptr;
          dma.ch8.dmandptr = 0x0;
          count            = 0;
	  retVal = 1;
          break;
        }
    }
#endif

  /* Call the transmit callback function */
  (*clbk)(pVC, skb);

  restore_flags(key);

  if(retVal == OK)
    {
  /* Update statistics */
      pATM->vc_stats.tx_pdus++;
      pVC->stats.tx_pdus++;
      pATM->vc_stats.tx_cells += psize->cells;
      pVC->stats.tx_cells += psize->cells;
    }

  return retVal;
} /* sar_xmit_aal5_if1 */

/******************************************************************
** FUNCTION : sar_xmit_aal5()
**
** DESCRIPTION:
**	Transmit or enqueue an AAL5 PDU. If frame mode transfer is used
**	the PDU is transmitted on the appropriate interfaces. Otherwise,
**	it is enqueued for transmission. The scheduler would transmit
**	cells that form this PDU at appropriate times.
**
** NOTE:
**	
******************************************************************/
int sar_xmit_aal5(pATM_VC_st pVC, void *ptr1, void (* clbk)(), UInt32 parm)
{
  pATMIF_st       pATM = pVC->pif; 
  struct sk_buff	*skb = (struct sk_buff *)ptr1;
  UInt32		plen, control;
  int status;

  if (!((control = pVC->control) & VPCI_ADMIN_UP))
    {
      pATM->err_stats.tx_unk_conn_err++;
      printk("sar_xmit_aal5: Admin state is down for connection\n");  
      return ERROR;
    }
  plen = skb->len;
  //  printk(__FUNCTION__" %d\n",plen);

  if (plen > (ATM_SAR_MTU - 8))
    {
      pATM->err_stats.inv_txsize_err++;  
      printk("sar_xmit_aal5: Exceeded MTU <%d>\n", (int )plen); 
      return ERROR;
    }

  if (control & FRAME_MODE)
    {
      if (pVC->subport)
	status = sar_xmit_aal5_if1(pVC, skb, plen, clbk);
      else
	status = sar_xmit_aal5_if0(pVC, skb, plen, clbk);
    }
  else
    {
      status = sar_enqueue_aal5(pVC, skb, plen, clbk);
    }

  return status;
} /* sar_xmit_aal5 */

