/*****************************************************************
**  File : sar_scheduler.c
**
**	Cell based scheduler code.
**
** Copyright 2001, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  07/18/2001   srinivas voleti           Original
**  10/01/2001   P. Sadik                  Ported to linux
******************************************************************/
#include <asm/system.h>
#include <linux/string.h>
#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"

#define	NUM_CONNECTIONS		8

typedef struct _t_TIMER_LIST_st {
  UInt32      timer_cnt;
  UInt32      init_cnt;
  pATM_VC_st  pVC;
  UInt32      reserved;
} TIMER_st, *pTIMER_st;

TIMER_st timer_list[NUM_CONNECTIONS] = \
  { {0,0,0,0}, {0,0,0,0}, \
    {0,0,0,0}, {0,0,0,0}, \
    {0,0,0,0}, {0,0,0,0}, \
    {0,0,0,0}, {0,0,0,0}};

UInt32	active_timer_entries = 0;

extern void sar_xmit_dma0(pATM_VC_st pVC);

/******************************************************************
** FUNCTION : add_timer_entry()
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
int add_timer_entry(pATM_VC_st pVC, UInt32 ticks)
{
  unsigned long key;

  if (active_timer_entries < (NUM_CONNECTIONS - 1))
    {
      save_and_cli(key);
      timer_list[active_timer_entries].timer_cnt = ticks;
      timer_list[active_timer_entries].init_cnt  = ticks;
      timer_list[active_timer_entries].pVC       = pVC;
      restore_flags(key);
      active_timer_entries++;
      printk("softsar: add_timer_entry() Done!\n");
      return OK;
    }
  else
      {
	  printk("softsar: add_timer_entry() Failed!\n");
	  return ERROR;
      }
} /* add_timer_entry */

/******************************************************************
** FUNCTION : delete_timer_entry()
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
int delete_timer_entry(pATM_VC_st pVC)
{
  pTIMER_st	p;
  UInt32		i;
  unsigned long key;

  for (i = 0, p = timer_list; i < active_timer_entries; i++, p++){
    if (p->pVC == pVC){
      save_and_cli(key);
      p->timer_cnt = timer_list[active_timer_entries].timer_cnt;
      p->init_cnt  = timer_list[active_timer_entries].init_cnt;
      p->pVC       = timer_list[active_timer_entries].pVC;
      active_timer_entries--;	
      restore_flags(key);
      break;
    }
  }
  if (i == active_timer_entries)
    return ERROR;
  else
    return OK;
} /* delete_timer_entry */

/******************************************************************
** FUNCTION : sar_schedule_cell()
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
void sar_schedule_cell(unsigned long val)
{
  pTIMER_st p;
  UInt32    i;

  for (i = 0, p = timer_list; i < active_timer_entries; i++, p++){
    p->timer_cnt--;
    if (p->timer_cnt == 0){
      p->timer_cnt = p->init_cnt;
      sar_xmit_dma0(p->pVC);
    }
  }
} /* sar_schedule_cell */
