/*****************************************************************
**  File : sardrv.c
**
**		Device driver for the ATM SOFTSAR.
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
** 02/10/2001   srinivas voleti           Removed references to recycle
**                                         buffer & descriptor
** 08/16/2000   srinivas voleti           Original
** 10/14/2001   P. Sadik                  Ported to linux, incorporated
                                          changes.
******************************************************************/
#include <asm/system.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/delay.h>

#include "sys_ss_def.h"
#include "sar_if.h"
#include "sardrv.h"
#include "sar_linux_specific.h"
#include "banyan.h"
#include "atm.h"
#include "dma.h"
#include "timer.h"

UInt32  get_num_subports[] = {0, 0, 2, 0, 1};

/* Table to hold the cell cnt & pad size for a specified PDU size for 
   AAL5 PDUs */
SIZE_2_CELL_CONVERTER_st size_2_cell_table[ATM_SAR_MTU + 1];

int sar_vpt_purge(void *handle, void *ptr1, UInt32 val1, 
		  void *ptr2, UInt32 val2, UInt32 *var);

UNALIGNED_MEM_st umem;  
void *gATM;
/******************************************************************
** FUNCTION : build_size2cell_table()
**
** DESCRIPTION:
**	Build a table that contains the cell count & the pad size for a 
**	given AAL5 PDU size. This table can be readily used for any PDU
**	size (upto the ATM_MTU) & avoid the computation at real-time.
**
** NOTE:
**
******************************************************************/
void build_size2cell_table(void)
{
  UInt32						i, cell_cnt, size;
  pSIZE_2_CELL_CONVERTER_st	p;
  
  for (i = 0, p = &size_2_cell_table[0]; i < (ATM_SAR_MTU + 1); i++, p++)
    {
      cell_cnt = i/MAX_CELL_PAYLOAD;
      cell_cnt++;
      size = cell_cnt * MAX_CELL_PAYLOAD;
      if ((size - i) < 8)
	{
	  size += MAX_CELL_PAYLOAD;
	  cell_cnt++;
	}
      size -= i;
      
      p->cells = cell_cnt;
      p->pad_size = size;
    }
} /* build_size2cell_table */

/******************************************************************
** FUNCTION : add_free_2k_bd_list()
**
** DESCRIPTION:
**	Replenish the free buffer-descriptor list. This list uses 2K byte
**	buffers.
**	
** NOTE:
**
******************************************************************/
int add_free_2k_bd_list(pATMIF_st pATM, UInt32 num)
{
  pDMA_st pdesc = (pDMA_st )0, prev = (pDMA_st )0;
  pDMA_st pfirst = (pDMA_st )0;
  struct sk_buff *skb;

  unsigned int 	i;
  unsigned long key;
  int status;

  save_and_cli(key);

  for (i = 0; i < num; i++){
    GET_DMA_DESC(pATM, pdesc);
    if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
      {
	printk("sardrv: insufficient rx desc.\n");
	pATM->err_stats.insuff_rx_desc_err++;
	if (prev)
	  prev->link = 0;
	break;
      }
    else
      {
	if(pfirst == (pDMA_st )0)
	  pfirst = pdesc;
	else
	  prev->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc);

	if( (skb = dev_alloc_skb(ATM_SAR_MTU+2)) == NULL)
	  {
	    pATM->err_stats.insuff_bufs_err++;
	    FREE_DMA_DESC(pATM, pdesc);
	    if (prev)
	      prev->link = 0;
	    break;
	  }
	else
	  {
	    skb_reserve(skb,2);
	  pdesc->mbuf_ptr = (struct sk_buff *)skb;
	  pdesc->curr_addr = VIRT_TO_PHYS((UInt32 )skb->data);
	  pdesc->status = 0xffffffff;
	  pdesc->buf_start = (UInt32 )skb->data;
	  pdesc->buf_size = ATM_SAR_MTU;
	  pdesc->control = DFLT_AAL5_RX_DESC_CTRL;
	  pdesc->chain_desc = 0;
	  }
	prev = pdesc;
      }
  }

  if(i != 0)	
    {
      pATM->free_2k_bds += i;
      if ((VIRT_TO_PHYS((UInt32 )pATM->free_2k_bd_head)) == 0)
	pATM->free_2k_bd_head = pfirst;
      else
	pATM->free_2k_bd_tail->link = 	(pDMA_st )VIRT_TO_PHYS((UInt32 )pfirst);
      pATM->free_2k_bd_tail = prev;
      status = OK;
    }
  else
    {
      status = ERROR;
    }
  restore_flags(key);
    
  return status;
} /* add_free_2k_bd_list */

/******************************************************************
** FUNCTION : get_2k_buf_desc_list()
**
** DESCRIPTION:
**	Build buffer descriptor list for the specified connection.
**	.
** NOTE:
**
******************************************************************/
pDMA_st get_2k_buf_desc_list(pATMIF_st pATM, pATM_VC_st pVC, UInt32 num)
{
  pDMA_st p = (pDMA_st )0;
  pDMA_st pfirst = (pDMA_st )NULL;
  UInt32	cnt = num;

  if (pATM->free_2k_bds < (num + 1))
    {
      pfirst = (pDMA_st )NULL;
    }
  else
    {
      while(cnt)
	{
	  if (pfirst == (pDMA_st )NULL)
	    {
	      pfirst = pATM->free_2k_bd_head;
	      p = pfirst;
	    }
	  else
	    {
	      p = (pDMA_st )GET_NEXT_DESC(p);
	    }
	  p->ptr = pVC;
	  p->chain_desc = 0;
	  
	  cnt--;
	}
      
      pATM->free_2k_bd_head = (pDMA_st )GET_NEXT_DESC(p);
      pATM->free_2k_bds -= num;
      p->link = 0;
      if (pVC == 0)
	{
	  printk("zero pvc\n");
	}
      
      if (pVC->last_rx_desc)
	{
	  pVC->last_rx_desc->chain_desc = pfirst;
	}
      pVC->last_rx_desc = p;
    }
  
  return (pfirst);
} /* get_2k_buf_desc_list */

/******************************************************************
** FUNCTION : add_free_64_bd_list()
**
** DESCRIPTION:
**	Replenish the free buffer-descriptor list. This list uses 64 byte
**	buffers.
**	.
** NOTE:
**
******************************************************************/
int add_free_64_bd_list(pATMIF_st pATM, UInt32 num)
{
  pDMA_st 	pdesc = (pDMA_st )0, prev = (pDMA_st )0;
  pDMA_st pfirst = (pDMA_st )0;
  struct sk_buff *skb;

  UInt32  	i;
  unsigned long key;
  int status;

  save_and_cli(key);
  for (i = 0; i < num; i++)
    {
      GET_DMA_DESC(pATM, pdesc);
      if ((VIRT_TO_PHYS((UInt32 )pdesc)) == 0)
	{
	  pATM->err_stats.insuff_rx_desc_err++;
	  if (prev)
	    prev->link = 0;
	  break;
	}
      else
	{
	if(pfirst == (pDMA_st )0)
	  pfirst = pdesc;
	else
	  prev->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc);

	if( (skb = dev_alloc_skb(64+2)) == NULL)
	    {
	      pATM->err_stats.insuff_bufs_err++;
	      FREE_DMA_DESC(pATM, pdesc);
	      if (prev)
		prev->link = 0;
	      break;
	    }
	  else
	    {
	      skb_reserve(skb,2);
	      pdesc->curr_addr = VIRT_TO_PHYS((UInt32 )skb->data);
	      pdesc->status = 0xffffffff;
	      pdesc->buf_start = (UInt32 )skb->data;
	      pdesc->buf_size = 64;
	      pdesc->control = DFLT_RAW_RX_DESC_CTRL;
	      pdesc->chain_desc = 0;
	    }
	  prev = pdesc;
	}
    }

  if (i != 0)
    {
      pATM->free_64_bds += i;
      if ((VIRT_TO_PHYS((UInt32 )pATM->free_64_bd_head)) == 0)
	pATM->free_64_bd_head = pfirst;
      else
	pATM->free_64_bd_tail->link =  (pDMA_st )VIRT_TO_PHYS((UInt32 )pfirst);
      pATM->free_64_bd_tail = prev;
      status = OK;
  }
  else
    {
      status = ERROR;
    }
  restore_flags(key);
    
  return status;
} /* add_free_64_bd_list */

/******************************************************************
** FUNCTION : get_64_buf_desc_list()
**
** DESCRIPTION:
**	Fetch a 64 byte buffer descriptor for the specified connection.
**	.
** NOTE:
**
******************************************************************/
pDMA_st get_64_buf_desc_list(pATMIF_st pATM, pATM_VC_st pVC)
{
  pDMA_st pdesc;

  if (pATM->free_64_bds < 2)
    pdesc = (pDMA_st )NULL;
  else
    {
      pdesc = pATM->free_64_bd_head;
      pdesc->ptr = pVC;

      pATM->free_64_bd_head = (pDMA_st )GET_NEXT_DESC(pdesc);
      pATM->free_64_bds--;
      
      pdesc->link = 0;
      pdesc->chain_desc = 0;
    }

  return pdesc;
} /* get_64_buf_desc_list */

/******************************************************************
** FUNCTION : free_2k_buf_desc()
**
** DESCRIPTION:
**	Move the buffer descriptor pair to the free list.
**	.
** NOTE:
**
******************************************************************/
void free_2k_buf_desc(pATMIF_st pATM, pDMA_st pdesc)
{
  pdesc->curr_addr = VIRT_TO_PHYS(pdesc->buf_start);
  pdesc->status = 0xffffffff;
  pdesc->control = DFLT_AAL5_RX_DESC_CTRL;
  pdesc->ptr = 0;
  pdesc->chain_desc = 0;
  pdesc->link = 0;
  if ((VIRT_TO_PHYS((UInt32 )pATM->free_2k_bd_head)) == 0)
    pATM->free_2k_bd_head = pdesc;
  else
    pATM->free_2k_bd_tail->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc);
  pATM->free_2k_bd_tail = pdesc;
  pATM->free_2k_bds++;
} /* free_2k_buf_desc */

/******************************************************************
** FUNCTION : free_64_buf_desc()
**
** DESCRIPTION:
**	Move the buffer descriptor pair to the free list.
**	.
** NOTE:
**
******************************************************************/
void free_64_buf_desc(pATMIF_st pATM, pDMA_st pdesc)
{
  pdesc->curr_addr = VIRT_TO_PHYS(pdesc->buf_start);
  pdesc->status = 0xffffffff;
  pdesc->control = (DMA_IOD | DMA_COUNT(64));
  pdesc->ptr = 0;
  pdesc->link = 0;
  pdesc->chain_desc = 0;
  if ((VIRT_TO_PHYS((UInt32 )pATM->free_64_bd_head)) == 0)
    pATM->free_64_bd_head = pdesc;
  else
    pATM->free_64_bd_tail->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc);
  pATM->free_64_bd_tail = pdesc;
  pATM->free_64_bds++;
} /* free_64_buf_desc */

/******************************************************************
** FUNCTION : free_resources()
**
** DESCRIPTION:
**   Free up system resources used by the softSAR
**
** NOTE:
**
******************************************************************/
void free_resources(pATMIF_st pATM)
{

  if (pATM->free_dlist_base)	
    free_pages((unsigned long)umem.free_dlist_base, 6);

  if (pATM->status_base)
    kfree(umem.status_base);

  if (pATM->free_vc_base)
    kfree(umem.free_vc_base);

  if (pATM->vp_table)
    kfree(umem.vp_table);

  if (pATM->vc_table)
    kfree(umem.vc_table);

  if (pATM->pad_buffer0)
    {
      kfree(umem.pad_buffer0);
    }

  if (pATM->pad_buffer1)
    {
      kfree(umem.pad_buffer1);
    }

  if (pATM->pad_buffer2)
    {
      kfree(umem.pad_buffer2);
    }

  free_atmif_rtos_resources(pATM->OS_specific);
  
  kfree(umem.pATM);

} /* free_resources */

/******************************************************************
** FUNCTION : sardrv_init
**
** DESCRIPTION:
**	The SOFTSAR initialization routine.
**
** NOTE:
**
******************************************************************/
pATMIF_st sardrv_init(pATMIF_CONFIG_st pconf)
{
  pATMIF_st	pATM;
  pATM_VC_st	pVC, next_vc;
  pDMA_st		plist, next;
  UInt32		i;

  /* Allocate memory for the SOFTSAR ATM IF data structure */

  umem.pATM = kmalloc( (sizeof(ATMIF_st ) + ALIGN_16),GFP_KERNEL);
  if (umem.pATM == NULL)
    {
      printk("kalloc (pATM) failed \n");
      return ((pATMIF_st )NULL);
    }

  pATM = (pATMIF_st )((UInt32 )(umem.pATM + ALIGN_16 - 1) & ~(ALIGN_16 - 1)); 
  bzero((char *)pATM, sizeof(ATMIF_st ));

  /* Initialize the fields */
  pATM->switch_no        = pconf->switch_no;
  pATM->port_no          = pconf->port_no;
  pATM->if_no            = pconf->if_no;
  pATM->max_VCCs         = pconf->max_VCCs;
  pATM->registered_VCCs  = 0;
  pATM->active_VCCs      = 0;
  pATM->oam_tx_hdlr      = pconf->oam_tx_hdlr;
  pATM->oam_rx_hdlr      = pconf->oam_rx_hdlr;
  if (pconf->init_phy)
    pATM->init_phy         = pconf->init_phy;
  if (pconf->get_phy)
    pATM->get_phy          = pconf->get_phy;
  if (pconf->set_phy)
    pATM->set_phy          = pconf->set_phy;
  pATM->control          = pconf->config;

#if defined (INDEXED_LOOKUP)
  /* Create the VC lookup table (indexed lookup table mechanism) */

  umem.vc_table = 
    kmalloc( ((pATM->max_VCCs * sizeof(UInt32 )) + ALIGN_16),GFP_KERNEL);

  if (umem.vc_table == NULL)
    {
      printk("kmalloc (level1) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }

  pATM->vc_table = (pATM_VC_st *)((UInt32 )(umem.vc_table + ALIGN_16 -1) & ~(ALIGN_16 - 1));  
  bzero((char *)pATM->vc_table,(pATM->max_VCCs * sizeof(UInt32 )));

#else /* INDEXED_LOOKUP */

  /* Compute the number of bits in the header used for subport */
  pATM->status = get_num_subports[pATM->control & ATMIF_TYPE_MASK];

  /* Create the Level 1 table (2-level lookup mechanism) */
  umem.vp_table = 
    kmalloc( (ALIGN_16 + ((MAX_VPI * (pATM->status + 1)) * sizeof (L1_LKUP_st ))),
			   GFP_KERNEL);

  if (umem.vp_table == NULL)
    {
      printk("kmalloc (vp_table) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->vp_table = (pL1_LKUP_st)((UInt32 )(umem.vp_table + ALIGN_16 -1) & ~(ALIGN_16 - 1));  
  
  bzero((char *)pATM->vp_table,((MAX_VPI * (pATM->status + 1)) * sizeof (L1_LKUP_st )));

#endif /* INDEXED_LOOKUP */

  /* Build a free list to hold the connection table entries */

  umem.free_vc_base = 
    kmalloc( (ALIGN_16 +(pATM->max_VCCs * sizeof (ATM_VC_st))),GFP_KERNEL);
  if (umem.free_vc_base == NULL)
    {
      printk("kmalloc (free_vc_base) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->free_vc_base = 
    (pATM_VC_st )((UInt32 )(umem.free_vc_base + ALIGN_16 -1) & ~(ALIGN_16 - 1));  

  bzero((char *)pATM->free_vc_base, (pATM->max_VCCs * sizeof (ATM_VC_st)));

  pATM->free_vc_list_head = pATM->free_vc_base;
	
  for (i = 0, pVC = pATM->free_vc_base, next_vc = pATM->free_vc_base; i < pATM->max_VCCs; i++)
    {
      if ( i < (pATM->max_VCCs - 1))
	{
	  next_vc++;
	  pVC->next = next_vc;
	  pVC++;
	}

      else
	{
	  pVC->next = NULL;
	  pATM->free_vc_list_tail = pVC;
	}
    }

  /* Create status queue */

  umem.status_base =  
    kmalloc( (ALIGN_16+(MAX_STATUS_QUEUE_SIZE * sizeof (STATUS_st ))), GFP_KERNEL);

  if (umem.status_base == NULL)
    {
      printk("Malloc (status_base) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->status_base =  
    (pSTATUS_st )((UInt32 )(umem.status_base + ALIGN_16 -1) & ~(ALIGN_16 - 1));  
  bzero((char *)pATM->status_base, (MAX_STATUS_QUEUE_SIZE * sizeof (STATUS_st )));

  pATM->status_head = pATM->status_base;
  pATM->status_tail = (pSTATUS_st )pATM->status_base;
  pATM->status_last = (pSTATUS_st )((UInt32 )pATM->status_base + 
				    ((MAX_STATUS_QUEUE_SIZE - 1) * sizeof(STATUS_st)));
	

  /* Build a free list to hold the DMA descriptors */
  
/*********************************************************************************
  free_dlist requires 160K of memory, which kmalloc cannot allocate.
  Hence, get_free_pages is called. The 'order' is hardcoded as 6,
  which is 64 pages = 64*4096 = 262144 bytes.
*********************************************************************************/
 
  umem.free_dlist_base = (char *)__get_free_pages(GFP_KERNEL,6);

  if (umem.free_dlist_base == NULL){
    printk("kalloc (free_dlist_base) failed\n");
    free_resources(pATM);
    return ((pATMIF_st )NULL);
  }

  pATM->free_dlist_base = 
    (pDMA_st )((UInt32 )(umem.free_dlist_base + ALIGN_16 - 1) & ~(ALIGN_16 - 1)); 

  pATM->free_dlist_base = (pDMA_st )K0_TO_K1((UInt32 )pATM->free_dlist_base);
  bzero((char *)pATM->free_dlist_base,(NUM_DMA_DESCS * sizeof (DMA_st)));

  pATM->free_dlist_head = pATM->free_dlist_base;
	
  for (i = 0, plist = pATM->free_dlist_base, next = pATM->free_dlist_base; i < NUM_DMA_DESCS; i++)
    {
      if ( i < (NUM_DMA_DESCS - 1))
	{
	  next++;
	  plist->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )next);
	  plist++;
	}

      else
	{
	  plist->link = NULL;
	  pATM->free_dlist_tail = plist;
	}
    }

  pATM->num_free_descs = NUM_DMA_DESCS;

  /* Setup OS specific resources & update data structures. Also attach the
     interrupt handlers */


  pATM->OS_specific = rtos_dependent_init((pATMIF_st) pATM);

  /* Initialize the buffer-descriptor list */
  pATM->free_2k_bd_head = 0;
  pATM->free_2k_bd_tail = 0;
  pATM->free_2k_bds = 0;
  add_free_2k_bd_list(pATM, MAX_NUM_2K_BDS);


  pATM->free_64_bd_head = 0;
  pATM->free_64_bd_tail = 0;
  pATM->free_64_bds = 0;
  add_free_64_bd_list(pATM, MAX_NUM_64_BDS);

  /* Allocate 56 byte buffers for holding AAL5 trailer information in 
     DMA channel 0 transfer */

  umem.pad_buffer0 = kmalloc( (ALIGN_4+ PAD_BUFFER_SIZE),GFP_KERNEL);
  if (umem.pad_buffer0 == NULL)
    {
      printk("Memory allocation (pad_buffer0) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->pad_buffer0 = (UInt32 *)((UInt32 )(umem.pad_buffer0 + ALIGN_4 - 1) & ~(ALIGN_4 - 1));    
  pATM->pad_buffer0 = (UInt32 *)K0_TO_K1((UInt32 )pATM->pad_buffer0);
  bzero((char *)pATM->pad_buffer0,PAD_BUFFER_SIZE);

  /* Allocate 56 byte buffers for holding AAL5 trailer information in 
	   frame mode transfer */
  umem.pad_buffer1 = kmalloc( (ALIGN_4+ PAD_BUFFER_SIZE),GFP_KERNEL);
  if (umem.pad_buffer1 == NULL)
    {
      printk("Memory allocation (pad_buffer1) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->pad_buffer1 = (UInt32 *)((UInt32 )(umem.pad_buffer1 + ALIGN_4 - 1) & ~(ALIGN_4 - 1));    

  pATM->pad_buffer1 = (UInt32 *)K0_TO_K1((UInt32 )pATM->pad_buffer1);
  bzero((char *)pATM->pad_buffer1,PAD_BUFFER_SIZE);

  umem.pad_buffer2 = kmalloc( (ALIGN_4+ PAD_BUFFER_SIZE),GFP_KERNEL);
  if (umem.pad_buffer2 == NULL)
    {
      printk("Memory allocation (pad_buffer2) failed\n");
      free_resources(pATM);
      return ((pATMIF_st )NULL);
    }
  pATM->pad_buffer2 = (UInt32 *)((UInt32 )(umem.pad_buffer2 + ALIGN_4 - 1) & ~(ALIGN_4 - 1));    

  pATM->pad_buffer2 = (UInt32 *)K0_TO_K1((UInt32 )pATM->pad_buffer2);
  bzero((char *)pATM->pad_buffer2,PAD_BUFFER_SIZE);

  /* Pre-allocate descriptors for use by the frame mode transmit 
	   operations */
  GET_DMA_DESC_LIST(pATM, plist, MAX_FRM_MODE_TX_DESCS);
  pATM->if0_tx_desc = plist;
  for (i = 0; i < MAX_FRM_MODE_TX_DESCS; i++)
    {
      plist->link_copy = plist->link;
      plist = (pDMA_st )GET_NEXT_DESC(plist);
    }
  pATM->if0_tx_desc = (pDMA_st )K0_TO_K1((UInt32 )pATM->if0_tx_desc);

  GET_DMA_DESC_LIST(pATM, plist, MAX_FRM_MODE_TX_DESCS);
  pATM->if1_tx_desc = plist;
  for (i = 0; i < MAX_FRM_MODE_TX_DESCS; i++)
    {
      plist->link_copy = plist->link;
      plist = (pDMA_st )GET_NEXT_DESC(plist);
    }
  pATM->if1_tx_desc = (pDMA_st )K0_TO_K1((UInt32 )pATM->if1_tx_desc);

  /* Build the PDU size to cell size converter table (this is to avoid
     computation of the cell count & the pad size at run time)
  */
  build_size2cell_table();

  /* Initialize the hardware used for SOFTSAR */
  sar_hw_module_init(pATM);

  /* Enable the ATM interface & mark the ATM interface as active */
  atm.atmintfc = pATM->control & ATMIF_TYPE_MASK;

  pATM->status |= (ATMIF_OPER_UP | ATM_DMA_EVT_MASK | ATM_IR_INT_EN);

  gATM =pATM;

  timer.ct1.ctc = CTC_EN;

  return (pATM);	
} /* sardrv_init */

/******************************************************************
** FUNCTION : sardrv_shutdown
**
** DESCRIPTION:
**
** NOTE:
**
******************************************************************/
void sardrv_shutdown(void *handle)
{
  pATMIF_st	pATM = (pATMIF_st )handle;

  /* Reset the relevant hardware modules used by SOFTSAR */
  sar_hw_module_reset(pATM);

  /* Purge all the open connections */
  sar_purge_all_vcs(handle, NULL, 0, NULL, 0, NULL);

  /* Free up resources */
  free_resources(pATM);

} /* sardrv_shutdown */

