/*****************************************************************
**	File : sardrv.h
**
**		Definitions & data structures of the ATM driver (SOFTSAR)
**		for IDT's RC32355 (Banyan) processor
**
** Copyright 2000, Integrated Device Technology, Inc.
** Company Confidential of Integrated Device Technology, Inc.
**
*******************************************************************
**
**  Date         Developer                 Comment
*******************************************************************
**  08/16/2000   srinivas voleti           Original
**  02/10/2001   srinivas voleti           Removed recycle specific
**                                         fields from ATMIF_st data
**                                         structure
**     10/2001   P. Sadik                  Ported to linux
******************************************************************/

#ifndef __SARDRV_H__
#define	__SARDRV_H__

/* Global Macros used by SOFTSAR */
#define	NUM_ATM_DMA_CHANNELS	9
#define	ATM_DMA_INTS			0x1ff
#define	OAM_CELL_DETECT			0x08
#define	EPDU_DETECT				0x02
#define CH_GFC_MASK				0xF0000000  /* bits 31-28 */
#define CH_GFC_SHIFT			28
#define CH_VPI_MASK				0x0FF00000  /* bits 27-20 */
#define CH_VPI_SHIFT			20
#define CH_SP_MASK				0x000F0000  /* bits 19-16 */
#define CH_SP_SHIFT				18
#define CH_VCI_MASK				0x0000FFF0  /* bits 19- 4 */
#define CH_VCI_SHIFT			4
#define CH_PTI_MASK				0x0000000E  /* bits 3- 1 */
#define CH_PTI_SHIFT			1
#define CH_CLP_MASK				0x00000001  /* bit 0    */

#define CH_VPI_SP_MASK			0x0FFF0000

/* Cell header to field converter */
#define GET_GFC(cellhdr)	((cellhdr & CH_GFC_MASK) >> CH_GFC_SHIFT)
#define GET_VPI(cellhdr)	((cellhdr & CH_VPI_MASK) >> CH_VPI_SHIFT)
#define GET_SP(cellhdr)		((cellhdr & CH_SP_MASK)  >> CH_SP_SHIFT)
#define GET_VCI(cellhdr)	((cellhdr & CH_VCI_MASK) >> CH_VCI_SHIFT)
#define GET_PTI(cellhdr)	((cellhdr & CH_PTI_MASK) >> CH_PTI_SHIFT)
#define GET_CLP(cellhdr)	(cellhdr & CH_CLP_MASK)
#define GET_VPI_SP(cellhdr)	((cellhdr & CH_VPI_SP_MASK) >> CH_SP_SHIFT)

/* Compute the VPCI index from the cell header (used by the indexed lookup
   mechanism only) */
#define GET_VPCI(hdr, vpi_mask, vci_mask, vpi_shft, vpci) \
		{ \
			UInt32	vpi, vci; \
			vpi = GET_VPI(hdr); \
			vci = GET_VCI(hdr); \
			*vpci = (((vpi & vpi_mask) << vpi_shft) | (vci & vci_mask)); \
		}

/* Subport IDs used with Utopia-2 interface */
#define	SP_0		0
#define	SP_1		(1 << CH_SP_SHIFT)
#define	SP_2		(2 << CH_SP_SHIFT)
#define	SP_0_1_ID	0
#define	SP2_ID		3

/* Field to cell header converter */
#define BLD_CELL_HDR(sp, vpi, vci, pti, gfc, clp) \
                     (((UInt32)vci << CH_VCI_SHIFT) | \
                      ((UInt32)sp  << CH_SP_SHIFT)  | \
                      ((UInt32)vpi << CH_VPI_SHIFT) | \
                      ((UInt32)gfc << CH_GFC_SHIFT) | \
                      ((UInt32)pti << CH_PTI_SHIFT) | clp)

/* Drop an ATM cell: "addr" is the address of ATM Input Data Register */
#define DROP_ATM_CELL(addr) \
		{ \
			volatile UInt32 val; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
			val = *addr; \
		}

/* Status state definitions */
#define	STATE_MASK				0xf0000000
#define	STATUS_CHAN_MASK		0x0000000f
#define	RXINT_STATE				0x10000000
#define	TXINT_STATE				0x20000000
#define	RXPEND_STATE			0x30000000
#define	TXPEND_STATE			0x40000000
#define	RX_AAL5_DMA0_STATE		0x50000000
#define	RAW_CELL_STATE			0x60000000
#define	RX_OAM_STATE			0x70000000

/* CRC32 magic value - this is the value that the status field of the 
   descriptor should contain after the transfer of the PDU */
#define	CRC32_MAGIC_CODE	0xc704dd7b

/* CRC10 magic value - this is the value that the status field of the 
   descriptor should contain after the transfer of the PDU */
#define	CRC10_MAGIC_CODE	0

/* Structure holding the cell count & the pad_size for a specified PDU size */
typedef struct _t_SIZE_2_CELL_CONVERTER_st {
	UInt32  cells;
	UInt32  pad_size;
} SIZE_2_CELL_CONVERTER_st, *pSIZE_2_CELL_CONVERTER_st;


/* SOFTSAR DMA descriptor structure */
typedef volatile struct _t_DMA_st {
	UInt32			control;	/* Control field of the DMA descriptor */
	UInt32			curr_addr;	/* Current address of the buffer */
	UInt32			status;		/* DEVCS */
	volatile struct _t_DMA_st	*link;		/* Link to the next descriptor */
	UInt32			buf_start;	/* Buffer start address */
	UInt32			buf_size;	/* Size of the buffer */
	void			*ptr;		/* Pointer to driver specific data structure */
	UInt32			control_1;	/* Driver specific control field */
	volatile struct _t_DMA_st	*chain_desc;/* Next descriptor list to be chained */ 
	void			*mbuf_ptr;	/* Pointer to an MBUF (used in RTK only) */
	UInt16			cell_cnt;	/* Cell count field - set in the first 
								   descriptor of the PDU */
	UInt16			tx_tag;		/* User defined tag field - set in the first
								   descriptor of the PDU */
	volatile struct _t_DMA_st *link_copy; /* Copy of the link field */
	void	(*clbk)();		/* Call back function for the buffer */
	UInt32			res[3];		/* Reserved & so desc is on cache boundary */
} DMA_st, *pDMA_st;

#define GET_NEXT_DESC(pdesc) (PHYS_to_K1((UInt32 )pdesc->link))

#define SET_DMA_DESC_CTRL(devcmd, ds, cnt) \
				(DMA_DEVCMD(devcmd) | DMA_DS(ds) | DMA_COUNT(cnt))

#define DFLT_AAL5_RX_DESC_CTRL \
                (DMA_IOD | DMA_DEVCMD(CRC32_LD) | DMA_COUNT(ATM_SAR_MTU))

#define DFLT_AAL0_RX_DESC_CTRL \
                (DMA_IOD | DMA_DEVCMD(CRC10) | DMA_COUNT(64))

#define DFLT_RAW_RX_DESC_CTRL \
                (DMA_IOD | DMA_DEVCMD(NOCRC) | DMA_COUNT(64))

#define	PDU_SIZE_MASK		0xfff
#define	CTRL1_TX_DESC_TYPE	0x80000000
#define	CTRL1_RX_DESC_TYPE	0
#define OAM_CELL_FLAG		0x40000000

/* Macro to fetch free DMA descriptor list from the free descriptor pool */
#define GET_DMA_DESC_LIST(pATM, pdesc, num_descs) \
		{ \
			pDMA_st	ptemp; \
			UInt32	i; \
			if (pATM->num_free_descs < (num_descs + 1)) \
				pdesc = NULL; \
			else \
			{ \
				pdesc = pATM->free_dlist_head; \
				for (i = 0, ptemp = NULL; i < num_descs; i++) \
				{ \
					if (ptemp == NULL) \
						ptemp = pdesc; \
					else \
						ptemp = (pDMA_st )GET_NEXT_DESC(ptemp); \
				} \
				pATM->free_dlist_head = (pDMA_st )GET_NEXT_DESC(ptemp); \
				ptemp->link = 0; \
				pATM->num_free_descs -= num_descs; \
			} \
		}

/* Macro to free DMA descriptor list to the free descriptor pool */
#define FREE_DMA_DESC_LIST(pATM, pdesc, num_descs) \
		{ \
			pDMA_st	ptemp; \
			UInt32	i; \
			pATM->free_dlist_tail->link = (pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc); \
			for (i = 0, ptemp = (pDMA_st )NULL; i < num_descs; i++)	\
			{ \
				if (ptemp == (pDMA_st )NULL) \
					ptemp = pdesc; \
				else \
					ptemp = (pDMA_st )GET_NEXT_DESC(ptemp); \
			} \
			pATM->free_dlist_tail = ptemp; \
			pATM->num_free_descs += num_descs; \
		}

/* Macro to fetch one free DMA descriptor from the free descriptor pool */
#define GET_DMA_DESC(pATM, pdesc) \
		{ \
			if (pATM->num_free_descs < 2) \
				pdesc = (pDMA_st )NULL; \
			else \
			{ \
				pdesc = pATM->free_dlist_head; \
				pATM->free_dlist_head = (pDMA_st )GET_NEXT_DESC(pdesc); \
				pdesc->link = (pDMA_st )0; \
				pATM->num_free_descs--; \
			} \
		}

/* Macro to free one DMA descriptor to the free descriptor pool */
#define FREE_DMA_DESC(pATM, pdesc) \
		{ \
			pATM->free_dlist_tail->link = \
								(pDMA_st )VIRT_TO_PHYS((UInt32 )pdesc); \
			pATM->free_dlist_tail = pdesc; \
			pATM->num_free_descs++; \
		}

/* Status Queue */
#define	MAX_STATUS_QUEUE_SIZE	8000
typedef struct _t_STATUS_st {
	pDMA_st				desc;		/* Descriptor pointer */
	UInt32				state;		/* Number of descriptors for PDU */
} STATUS_st, *pSTATUS_st;

/* VC Statistics */
typedef struct	_t_VC_STATS_st {
	UInt32	tx_cells;	/* Transmit cell count */
	UInt32	tx_pdus;	/* Transmit PDU count */
	UInt32	rx_cells;	/* Receive cell count */
	UInt32	rx_pdus;	/* Receive PDU count */
	UInt32	tx_oam;		/* Transmit OAM cell count */
	UInt32	rx_oam;		/* Receive OAM cell count */
	UInt32	tx_raw_cells; /* Transmit Raw cell count */
	UInt32	rx_raw_cells; /* Receive Raw cell count */
} VC_STATS_st, *pVC_STATS_st;

/* SAR Error Statistics */
typedef struct _t_SAR_ERR_st {
        UInt32  rx_unk_conn_err;        /* Received PDU on unknown connection */
        UInt32  tx_unk_conn_err;        /* Transmit PDU on unknown connection */
        UInt32  inv_aal_err;            /* Invalid AAL type */
        UInt32  insuff_tx_desc_err;     /* Insufficient transmit descriptors */
        UInt32  insuff_rx_desc_err;     /* Insufficient receive descriptors */
        UInt32  insuff_bufs_err;        /* Insufficient buffers */
        UInt32  cell_drop_err;          /* Cell drop counter */
        UInt32  txq_exceeded_err;       /* Tx Queue exceeded counter */
        UInt32  CRC10_err;                      /* CRC10 errors */
        UInt32  CRC32_err;                      /* CRC32 errors */
        UInt32  overflow_err;           /* Overflow errors */
        UInt32  inv_txsize_err;         /* Invalid Tx PDU Size error */
        UInt32  inv_rxsize_err;         /* Invalid Rx PDU Size error */
        UInt32  dma_halt_err;           /* DMA halted errors */
        UInt32  dma_err;                        /* DMA errors */
} SAR_ERR_st, *pSAR_ERR_st;
 
#define MAX_DMA0_TXPEND_VAL     10  /* Maximum number of PDUs that can be queued
				       on a connection (doesn't include OAM cells) */  

#define	PAD_BUFFER_SIZE		56
#define AAL_TYPE_MASK   0x0f
#define RAW_TYPE           	0
#define AAL1_TYPE           1
#define AAL2_TYPE           2
#define AAL5_TYPE           5
#define	OAM_TYPE			6
#define CHAN_MASK           0xf0
#define	CHAN_BIT_POS		4
#define FRAME_MODE          0x20000000

/* VC type : SVC or PVC */
#define PVC_TYPE    0
#define SVC_TYPE    1

/* Service type */
#define UBR_TYPE    0
#define CBR_TYPE    1
#define VBR_TYPE    2
#define ABR_TYPE    3

/* State field */
#define	VPCI_OPER_UP		0x80000000
#define VPCI_OPER_DOWN		0x00000000
#define RX_IN_PROGRESS  	0x40000000
#define TX_IN_PROGRESS  	0x20000000
#define RX_DROP_PDU     	0x10000000
#define TX_DROP_PDU     	0x08000000
#define	LIST_A				0x04000000
#define	DATA_PRESENT		0x02000000
#define	CHAIN_PENDING		0x01000000
#define	IFACE0				0x00000000
#define	IFACE1				0x00001000
#define	IFACE_MASK			0x00001000
#define	TX_CELL_CNT_MASK	0x00000fff

/* Connection Table entry */
typedef struct _t_ATM_VC_st {
  struct _t_ATM_VC_st	*next;
  UInt32				control;		/* Control field */
  UInt32				state;			/* Connection state at run-time */
  UInt16				subport;		/* Sub-port number */
  UInt16				vpi;			/* VPI */
  UInt32				vci;			/* VCI */
  UInt32				cell_hdr;		/* ATM cell header */
  struct _t_ATM_VC_st	*handle;		/* Connection handle */
  void (*tx_hdlr)();
  void (*rx_hdlr)();
  VC_STATS_st			stats;			/* Per connection statistics */
  pDMA_st				tx_pdu_desc;	/* Current PDU descriptor */
  pDMA_st				tx_buf_desc;	/* Current transmit buffer descriptor 
							   used with DMA channel 0 only */
  pDMA_st				tx_last_pdu_desc; /* Last PDU descriptor */
  pDMA_st				rx_data_desc;	/* Receive descriptor used with DMA 
							   channel 0 only */
  pDMA_st             rx_oam_desc;    /* Receive descriptor for OAM rx */
  pDMA_st				last_rx_desc;	/* Last Rx descriptor in current list */
  UInt32				*pad_buffer;	/* Padding buffer used in frame mode */
  struct _t_ATMIF_st	*pif;			/* Device back pointer */
  void				*proto;			/* Protocol specific pointer */
  UInt32                          dma0_txpend;    /* Number of PDUs enqueued for xmit */  
} ATM_VC_st, *pATM_VC_st;

/*
  Constants used for the control field in the VPI look-up table entry
  -------------------------------------------------------------------
   Bit 0-30 : Reserved
   Bit 31   : Entry is valid
  --------------------------------------------------------------------
*/
#define	VPLKUP_VALID	0x80000000

/* Level 1 look-up table (VP table) for 2-level lookup connection management
   mechanism */
typedef struct _t_L1_LKUP_st {
	UInt32		control;	/* L1 table control */
	pATM_VC_st	*pL2_lkup;	/* L2 base pointer */
	UInt32		vc_mask;	/* Mask to be used with the VCI field */
	UInt32	vc_bit_pos;		/* First non-zero bit position in the VC field */
} L1_LKUP_st, *pL1_LKUP_st;

#define MAX_VPI_BITS
#define MAX_VCI_BITS
#define UNUSED_BITS_MASK
#define BUILD_INDEX(hdr)

/* Number of data buffers of each size */
#if 0
#define ATM_NUM_64          500 /* Original 1000 */
#endif
#define ATM_NUM_64          1000

#define ATM_NUM_128         0
#define ATM_NUM_256         0
#define ATM_NUM_512         0
#define ATM_NUM_1024        0

#if 0
#define ATM_NUM_2048        1000 /* Original 1500 */
#endif
#define ATM_NUM_2048        1500 

#define ATM_NUM_CL_BLKS     (ATM_NUM_64 + ATM_NUM_128 + ATM_NUM_256 + \
                             ATM_NUM_512 + ATM_NUM_1024 + ATM_NUM_2048)

#define	NUM_DMA_DESCS		ATM_NUM_CL_BLKS

#if 0
#define MAX_NUM_2K_BDS      500 /* Original 1000 */
#endif
#define MAX_NUM_2K_BDS      1000

#define	BD2K_WATER_MARK		(MAX_NUM_2K_BDS >> 1)

#if 0
#define MAX_NUM_64_BDS      300 /* Original 500 */
#endif
#define MAX_NUM_64_BDS      500

#define	BD64_WATER_MARK		(MAX_NUM_64_BDS >> 1)
#define NUM_BDS_PER_LIST    30
#define	NUM_DESCS_PER_CONN	10

#define MAX_FRM_MODE_TX_DESCS	10

/* Status field */
#define	ATM_DMA_EVT_MASK	0x00001FF0
#define	ATM_DMA_EVT_SHFT	4
#define	ATM_IR_INT_EN		0x00002000

/* ATM interface specific data structure */
typedef struct _t_ATMIF_st {
  UInt32		switch_no;			/* Switch number */
  UInt32		port_no;			/* Port number */
  UInt32		if_no;				/* Interface number */
  UInt32		control;			/* Control information */
  UInt32		status;				/* Status */
  UInt32		max_VCCs;			/*	Maximum number of connections */
  UInt32		registered_VCCs;	/* Registered number of connections */
  UInt32		active_VCCs;		/* Active VCCs */
  UInt32		vpi_mask;			/* Used with Indexed lookup only */
  UInt32		vci_mask;			/* Used with Indexed lookup only */
  UInt32		vpi_shft;			/* Used with Indexed lookup only */
  pL1_LKUP_st	vp_table;			/* VP table pointer */
  pATM_VC_st	*vc_table;			/* VC pointer table */
  void		*OS_specific;		/* OS dependent data structure */
  void (* oam_tx_hdlr)();
  void (* oam_rx_hdlr)();
  void (* init_phy   )();
  int  (* get_phy)();
  int  (* set_phy)();

  VC_STATS_st	vc_stats;			/* VC statistics */
  SAR_ERR_st	err_stats;			/* SAR error statistics */
  pDMA_st		curr_desc[9];		/* Array of current descriptor pointers */
  pATM_VC_st      pVC_array[9];           /* Array of VC handles for VCC entries */  
  pSTATUS_st	status_base;		/* Status queue base pointer */
  pSTATUS_st	status_head;		/* Status queue head pointer */
  pSTATUS_st	status_tail;		/* Status queue tail pointer */
  pSTATUS_st	status_last;		/* Status queue last pointer */
  pATM_VC_st	free_vc_base;
  pATM_VC_st	free_vc_list_head;
  pATM_VC_st	free_vc_list_tail;
  pATM_VC_st	active_vc_list;
  pDMA_st		free_dlist_base;
  pDMA_st		free_dlist_head;
  pDMA_st		free_dlist_tail;
  UInt32		num_free_descs;
  pDMA_st		free_2k_bd_head;
  pDMA_st		free_2k_bd_tail;
  UInt32		free_2k_bds;
  pDMA_st		free_64_bd_head;
  pDMA_st		free_64_bd_tail;
  UInt32		free_64_bds;
  UInt32		*pad_buffer0;
  UInt32		*pad_buffer1;
  UInt32		*pad_buffer2;
  pDMA_st		if0_tx_desc;
  pDMA_st		if1_tx_desc;
} ATMIF_st, *pATMIF_st;

/* Enqueue status entry */
#define	ADD_STATUS_ENTRY(pATM, state, pdesc) \
		{ \
			pSTATUS_st pstatus, pnext; \
			pstatus = pATM->status_tail; \
			pnext = pstatus; \
			pnext++; \
			if (pnext > pATM->status_last) \
				pnext = pATM->status_base; \
			if (pnext == pATM->status_head) \
			{ \
				printk("add_status_entry : Status queue full\n"); \
			} \
			else \
			{ \
				pstatus->state = state; \
				pstatus->desc = pdesc; \
				pATM->status_tail = pnext; \
			} \
		}

				
#define	ALIGN_16	16
#define	ALIGN_4		4
/* Function prototypes */
pATMIF_st sardrv_init(pATMIF_CONFIG_st );
void sardrv_shutdown(void *);
int sar_conn_setup(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
int sar_conn_terminate (void *, void *, UInt32 , void *, UInt32 , UInt32 *);
int sar_vpt_setup(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
int sar_vpt_purge(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
int sar_purge_all_vcs(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
pATM_VC_st fetch_vc_handle(pATMIF_st , UInt32);
int sar_set_conn_state(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
int get_vc_handle(void *, void *, UInt32 , void *, UInt32 , UInt32 *);
char *get_buffer(void *, UInt32 , UInt32 *);
void free_buffer(void *, UByte *);
void sar_hw_module_init(pATMIF_st );
void sar_hw_module_reset(pATMIF_st );
int add_free_2k_bd_list(pATMIF_st , UInt32 );
int add_free_64_bd_list(pATMIF_st , UInt32 );
pDMA_st get_2k_buf_desc_list(pATMIF_st , pATM_VC_st , UInt32 );
pDMA_st get_64_buf_desc_list(pATMIF_st , pATM_VC_st );
void free_2k_buf_desc(pATMIF_st , pDMA_st );
void free_64_buf_desc(pATMIF_st , pDMA_st );


#endif /* __SARDRV__ */
