#define	Z_WAKE

/*
 *  linux/drivers/char/cabernet.c
 *
 * This file contains the driver for the cabernet switch
 * that contain serial ports on ethernet ports.
 *
 * Based upon cyclades.c
 *
 * Much of the design and some of the code came from serial.c
 * which was copyright (C) 1991, 1992  Linus Torvalds.  It was
 * extensively rewritten by Theodore Ts'o, 8/16/92 -- 9/14/92,
 * and then fixed as suggested by Michael K. Johnson 12/12/92.
 */

/* If you need to install more boards than NR_CARDS, change the constant
   in the definition below. No other change is necessary to support up to
   eight boards. Beyond that you'll have to extend cab_isa_addresses. */

#define NR_CARDS        1 /* not to be confused with line cards */
                          /* there is only one FPGA, and hence 1 card */

/*
   If the total number of ports is larger than NR_PORTS, change this
   constant in the definition below. No other change is necessary to
   support more boards/ports. */
#define NR_PORTS        1

/*
 * Include section 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/bootinfo.h>

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/version.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include "../../../cmn/util/mxc_ioctl.h"
#include "../../../cmn/util/aruba_image.h"

unsigned char sector64[(64*1024)];

#define PHY_TIMEOUT    10000 


#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef SERIAL_XMIT_SIZE
#define	SERIAL_XMIT_SIZE	(MIN(PAGE_SIZE, 4096))
#endif
#define WAKEUP_CHARS		256
#define STD_COM_FLAGS (0)

#define	JIFFIES_DIFF(n, j)	((j) - (n))

#define IMAGE_SECT_NUM(x) (63-((0xbfffffff-x)/0x10000))
#define SAM_IMAGE_SECT_NUM(x) ((x-0xbfc00000)/(0x1000))

#include "../../../sap/flash/ap60_flash_common.h"
#include "../../../sap/flash/samsung_flash.h"

extern unsigned long mips_machtype;
static struct tty_driver ap_flash_driver;
static int  ap_flash_refcount;

#define IS_SAMSUNG(x) ((mips_machtype == MACH_ARUBA_SAMSUNG)?1:0)

typedef  unsigned char  flash_type_t;
typedef  unsigned short sam_flash_type_t;

/* This is the per-channel data structure containing pointers, flags
 and variables for the port. This driver supports a maximum of NR_PORTS.
*/
static struct tty_struct *serial_table[NR_PORTS];
static struct termios *serial_termios[NR_PORTS];
static struct termios *serial_termios_locked[NR_PORTS];


void lock_ap60_flash(void);
void unlock_ap60_flash(void);
void ap_flash_erase_sector(volatile unsigned char * flash_addr);
static int ap_flash_open(struct tty_struct *tty, struct file * filp);
static void ap_flash_close(struct tty_struct *tty, struct file *filp);
static int cab_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count);
static void ap_flash_put_char(struct tty_struct *tty, unsigned char ch);
static void ap_flash_flush_chars(struct tty_struct *tty);
static int ap_flash_write_room(struct tty_struct *tty);
static int ap_flash_chars_in_buffer(struct tty_struct *tty);
static void ap_flash_flush_buffer(struct tty_struct *tty);
static void ap_flash_throttle(struct tty_struct * tty);
static void ap_flash_unthrottle(struct tty_struct * tty);
static void ap_flash_set_termios(struct tty_struct *tty, struct termios *old_termios);
static void ap_flash_stop(struct tty_struct *tty);
static void ap_flash_start(struct tty_struct *tty);
static void ap_flash_hangup(struct tty_struct *tty);
static void ap_flash_break(struct tty_struct *tty, int break_state);
static void  ap_flash_wait_until_sent(struct tty_struct *tty, int timeout);
static int cabernet_get_proc_info(char *buf, char **start, off_t offset, int length,
int *eof, void *data);

extern unsigned long mips_machtype;

int isAp70(void)
{
    switch (mips_machtype) {
    case MACH_ARUBA_MUSCAT:
    case MACH_ARUBA_DELFRESCO:
        return 1;
    case MACH_IDT_RC32434: 
        return 1;
    default:
        return 0;
    }
}


/*
 * ------------------------------------------------------------
 * ap_flash_ioctl() and friends
 * ------------------------------------------------------------
 */


#ifdef ARUBA_DEBUG		
static void 
ap_flash_print_sector(volatile char *buffer, unsigned int size, char * flash_addr)
{
	/* limit printing to 256 chars max */
	size = (size > 256)?256:size;
	
	int i = 0, j = 0;
	for(i=0; i<size; i++) {
	    if((i%16)==0) 
	    {
		if (i != 0) 
		{
			printk(" ");
			for (j=(i-16); j < i; j++)
				printk("%c", buffer[j]);
		}
	    	printk("\n0x%08x: ", (unsigned int)(flash_addr+(16*(i/16))));		
	    }
	    printk("%02x ", (unsigned char)buffer[i]);
	}

	printk("\n");
}
#endif

/*******************************************************************************
 *     ap_flash_read
 */
void ap_flash_read(mxc_nvram_t * arg_param)
{
    volatile unsigned char *flash_addr  = (unsigned char*)((unsigned long)(AP60_FLASH_START) + arg_param->offset);
    
    if ( AP60_FLASH_END < (unsigned long)flash_addr + arg_param->length)
    {
        printk("%s: start 0x%x len 0x%x End of Flash (0x%x)\n", __FUNCTION__,
                (unsigned int)flash_addr, (unsigned int)arg_param->length, (unsigned int)AP60_FLASH_END);
        return;
    }

    copy_to_user(arg_param->data,flash_addr,arg_param->length);

    return;
}

#define AP70_IDT_FLASH_WP           (1<<3)             

/*******************************************************************************
 *     lock_ap60_flash
 */
void lock_ap60_flash(void)
{
	if (isAp70())
	{
                volatile unsigned long * pLock = (volatile unsigned long *)AP70_FLASH_LOCK_ADR;
                *pLock |= AP70_FLASH_LOCK_PIN;
	}
	else
	{
		volatile unsigned long * pLock = (volatile unsigned long *)ATHEROS_FLASH_LOCK_ADR;
		*pLock |= ATHEROS_FLASH_LOCK_PIN;	
	}
}

/*******************************************************************************
 *     unlock_ap60_flash
 */
void unlock_ap60_flash(void)
{
	if (isAp70())
        {
                volatile unsigned long * pLock = (volatile unsigned long *)AP70_FLASH_LOCK_ADR;
                *pLock &= ~AP70_FLASH_LOCK_PIN;
        }
        else
        {
		volatile unsigned long * pLock = (volatile unsigned long *)ATHEROS_FLASH_LOCK_ADR;
		*pLock &= ~ATHEROS_FLASH_LOCK_PIN;	
	}
}

/*******************************************************************************
 *     ap_flash_erase_sector
 */
void ap_flash_erase_sector(volatile unsigned char * flash_addr)
{
	unsigned long sector_size = 0, timeout = 0;
	
    volatile unsigned char *flash_sect	 = NULL;
    volatile unsigned char* pFlashWriter = NULL; 
    volatile char 	   	  * flash_base	 = (char *)AP60_FLASH_START;       	
    volatile flash_type_t *sd 			 = flash_addr;


	if ( (OVERWRITE_IDTBOOT((unsigned long)flash_addr)) ||
	     (OVERWRITE_LAST_4((unsigned long)flash_addr)) )
	{
		printk("Erasing IDTBoot not allowed![addr=%p][boot_end=%lx]\n", flash_addr, 
		(unsigned long)AP60_FLASH_IDTBOOT_END);
		return;
	}

	if (((AP60_FLASH_START)+(SIZE_64K_SECTORS)) > (unsigned long)(flash_addr))
		sector_size = AP60_FLASH_SECTOR1_SIZE; /* 64KB */
	else
		sector_size = AP60_FLASH_SECTOR2_SIZE; /* 8KB */		
	flash_sect 	= (unsigned char *)((long)flash_addr & ~(sector_size-1));

//	printk("Erasing sector 0x%p, size=0x%lx...", flash_sect, sector_size);

	unlock_ap60_flash();
	pFlashWriter  = flash_base + AP60_FLASH_FIRST_OFFSET;
	*pFlashWriter = 0xAA;
	
	pFlashWriter  = flash_base + AP60_FLASH_SECND_OFFSET;
	*pFlashWriter = 0x55;  		

	pFlashWriter  = flash_base + AP60_FLASH_FIRST_OFFSET;
	*pFlashWriter = 0x80;  
		
	pFlashWriter  = flash_base + AP60_FLASH_FIRST_OFFSET;
	*pFlashWriter = 0xAA;  
		
	pFlashWriter  = flash_base + AP60_FLASH_SECND_OFFSET;
	*pFlashWriter = 0x55;  
		
	*sd = 0x30;  		
		
	while (*sd != (flash_type_t)0xffff)
	{
		timeout++;
		if (timeout > 1000000)
		{
			printk("Flash must be locked! Could not erase!\n");
			return;
		}
		udelay(10);
		schedule() ;
	}		
	
	lock_ap60_flash();

//	printk("Done\n");		
	return;		  
}

/*******************************************************************************
 *     ap_flash_write
 */
void ap_flash_write(mxc_nvram_t * arg_param)
{
    int   i=0, j=0;
    int   len	      		= arg_param->length;
    int   copy_len 	    	= 0;    
	unsigned long sector_size = 0;
    unsigned char *buffer;

    unsigned char 		   *mem_addr     = arg_param->data; 
    volatile unsigned char *pFlashWriter = NULL;    
    volatile char 	   	   *flash_base 	= (char *)AP60_FLASH_START;    
    volatile unsigned char *flash_addr  = (unsigned char*)((unsigned long)(AP60_FLASH_START) + arg_param->offset);	    
    volatile unsigned char *flash_sect; 

	if (((AP60_FLASH_START)+(SIZE_64K_SECTORS)) > (unsigned long)(flash_addr))
		sector_size = AP60_FLASH_SECTOR1_SIZE; /* 64KB */
	else
		sector_size = AP60_FLASH_SECTOR2_SIZE; /* 8KB */		
	flash_sect 	= (unsigned char *)((long)flash_addr & ~(sector_size-1));	

#ifdef ARUBA_DEBUG		
	printk("ap_flash_write: BASE = %p, flash_addr = %p, len = %d\n", 
			flash_base, flash_addr, len);    
#endif

	if ( AP60_FLASH_END < (unsigned long)flash_addr)
	{
		printk("%s: Flash addr (0x%p) invalid > End of Flash (0x%lx)\n", __FUNCTION__,
				flash_addr, (unsigned long)AP60_FLASH_END);
		return;
	}

	buffer=(unsigned char*)&sector64[0];
	if (NULL == buffer)
	{
		printk("NOT ENOUGH RAM buffer memory to write to flash!\n");
		return;
	}
	
	do {
		unsigned space_left = (sizeof(sector64) - (flash_addr - flash_sect));
		copy_len = MIN(len, sector_size);
		copy_len = MIN(copy_len, space_left);

#ifdef ARUBA_DEBUG
		printk("Writing: flash_addr = %x, flash_sect=%x(#%d), copy_len=%d, len=%d\n", 
		flash_addr, flash_sect, IMAGE_SECT_NUM((unsigned long)flash_sect), copy_len, len);		
#endif
		/* read entire sector */
#if 1
		memcpy(buffer, flash_sect, sector_size);
#else
		for(i=0; i<sector_size; i++) {
			buffer[i] = flash_sect[i];
		}
#endif

		/* might need to remove if we're going to update idtboot dynamically */
		if ( (OVERWRITE_IDTBOOT((unsigned long)flash_addr)) ||
      	             (OVERWRITE_LAST_4((unsigned long)flash_addr)) )
		{
			printk("Overwriting critical sector in FL not allowed![addr=%p][sect adr=%p]\n", flash_addr, flash_sect);
			lock_ap60_flash();
			return;
		}

		/* have to erase sector before writing into it..can only program 1->0s */
		ap_flash_erase_sector(flash_addr);		

   	        /* update buffer with write data */
		if ((flash_addr - flash_sect) > (sizeof(sector64) - copy_len)) { 
			panic("flash_addr 0x%x, flash_sect 0x%x, copy_len 0x%x\n", 
				flash_addr, flash_sect, copy_len);
		}

		copy_from_user(&buffer[flash_addr - flash_sect], mem_addr, copy_len);
		unlock_ap60_flash();

		/* program the sector */
		for(i=0; i<sector_size; i++) 
		{
			/* tell flash we want to program one byte/word */
			pFlashWriter  = flash_base + AP60_FLASH_FIRST_OFFSET;
			*pFlashWriter = 0xAA;
			pFlashWriter  = flash_base + AP60_FLASH_SECND_OFFSET;
			*pFlashWriter = 0x55;		
			pFlashWriter  = flash_base + AP60_FLASH_FIRST_OFFSET;				
			*pFlashWriter = 0xA0;	
		
			flash_sect[i]=buffer[i];
			
			j=0; 
			while((buffer[i]^flash_sect[i]) & 0x80) {
				if (j++ > 1000000) {
					printk("Flash_W: TIMED-OUT %02x\n", buffer[i]);
					lock_ap60_flash();
					return;
				}
				if (!(j % 1000)) schedule() ;

			} /* while, timeout */
		} 

#ifdef ARUBA_DEBUG
		if (flash_addr == 0xbfc80000)
		{
			aruba_image_t *flsh_img_hdr = (aruba_image_t *)buffer;
			printk("\tsize		: %d\n", flsh_img_hdr->size);
			printk("\tmagic str : %s\n", flsh_img_hdr->magic);	 
			flsh_img_hdr = (aruba_image_t *)flash_addr;
			printk("\t(flsh@%x)size		: %d\n", flash_addr, flsh_img_hdr->size);
			printk("\t(flsh@%x)magic str : %s\n", flash_addr, flsh_img_hdr->magic);	
		}
#endif /* ARUBA_DEBUG */		
        /* advance ptrs for additional sectors if any */
		len 		-= copy_len;
		flash_addr 	+= copy_len;
		mem_addr 	+= copy_len;

#if 1
		flash_sect = (unsigned char *)((long)flash_addr & ~(sector_size-1));
#else
		if ((flash_addr - flash_sect) > sector_size)
			flash_sect += sector_size;
#endif
		/* make sure to count for 64KB->8KB sector jump */	
		if (((AP60_FLASH_START)+(SIZE_64K_SECTORS)) > 
			(unsigned long)(flash_addr)) 
			sector_size = AP60_FLASH_SECTOR1_SIZE; /* 64KB */
		else
			sector_size = AP60_FLASH_SECTOR2_SIZE; /* 8KB */	
	
	} while (len > 0);

	lock_ap60_flash();

	return;	
}

/*******************************************************************************
 *     samsung_flash_read
 */
void samsung_flash_read(mxc_nvram_t * arg_param)
{
    int i;
    volatile unsigned char *flash_addr  = (unsigned char*)((unsigned long)(SAM_FLASH_START) + arg_param->offset);
    unsigned char *mem_addr = (unsigned char *)arg_param->data; 
    int   len	      		= arg_param->length;
    int   read_len	  		= 0;
	unsigned long sector_size = SAM_FLASH_SECTORS_SIZE1;
	volatile unsigned char *flash_sect;
    
	flash_sect 	= (volatile unsigned char *)((long)flash_addr & ~(sector_size-1));	

#ifdef ARUBA_DEBUG		
		printk("sam_flash_addr = %p, len = %d, flash_sect=0x%p, offset=%x\n", 
				flash_addr, len, flash_sect, arg_param->offset);
#endif

	if ( SAM_FLASH_END < (unsigned long)flash_addr)
	{
		printk("%s: Flash addr (0x%p) invalid > End of Flash (0x%lx)\n",  __FUNCTION__,
			flash_addr, (unsigned long)SAM_FLASH_END);
		return;
	}

    do 
	{
		read_len = MIN(len, sector_size - (flash_addr - flash_sect));
#ifdef ARUBA_DEBUG		
		printk("\tReading %d bytes from flash_sect_offset (0x%x)\n", 
				read_len, (flash_addr - flash_sect)); 
#endif
		if (read_len <= sector_size) 
		{
		    for(i = 0; i < read_len; i++) 
				mem_addr[i]=flash_addr[i];
		}

#ifdef ARUBA_DEBUG_1
		 printk("AFTER READ VALUES:\n"); 
		 ap_flash_print_sector(mem_addr, read_len, flash_addr);
#endif

        /* advance ptrs for additional sectors if any */
		len 		-= read_len;
		flash_addr 	+= read_len;
		flash_sect 	+= sector_size;
		mem_addr 	+= read_len;
		
	} while (len);
}

/*******************************************************************************
 *     samsung_flash_erase_sector
 */
void samsung_flash_erase_sector(volatile unsigned char * flash_addr)
{
	unsigned long sector_size = SAM_FLASH_SECTORS_SIZE1, timeout = 0;
	
    volatile unsigned short *flash_sect	 = NULL;
    volatile unsigned short* pFlashWriter = NULL; 
    volatile char 	   	  * flash_base	 = (char *)SAM_FLASH_START;       	
    volatile sam_flash_type_t *sd 		 = (sam_flash_type_t *)flash_addr;


	if ( (SAM_OVERWRITE_IDTBOOT((unsigned long)flash_addr)) ||
	     (SAM_OVERWRITE_LAST_4((unsigned long)flash_addr)) )
	{
		printk("Erasing IDTBoot not allowed![addr=%p][boot_end=%lx]\n", flash_addr, 
		(unsigned long)SAM_FLASH_IDTBOOT_END);
		return;
	}

	flash_sect 	= (unsigned short *)((long)flash_addr & ~(sector_size-1));
#ifdef ARUBA_DEBUG
	printk("Erasing sector 0x%p, size=0x%lx...", flash_sect, sector_size);
#endif /* ARUBA_DEBUG */

	unlock_ap60_flash();

	pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_FIRST_OFFSET);
	*pFlashWriter = 0xAAAA;
	
	pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_SECND_OFFSET);
	*pFlashWriter = 0x5555;  		

	pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_FIRST_OFFSET);
	*pFlashWriter = 0x8080;  
		
	pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_FIRST_OFFSET);
	*pFlashWriter = 0xAAAA;  
		
	pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_SECND_OFFSET);
	*pFlashWriter = 0x5555;  
		
	*sd = 0x3030;  		
		
	while (*sd != (sam_flash_type_t)0xffff)
	{
		timeout++;
		if (timeout > 1000000)
		{
			printk("Flash must be locked! Could not erase!\n");
			return;
		}
		udelay(10);
	}		
	
	lock_ap60_flash();

//	printk("Done\n");		
	return;		  
}

/*******************************************************************************
 *     samsung_flash_write
 */
void samsung_flash_write(mxc_nvram_t * arg_param)
{
    int   i=0, j=0;
    int   len	      		= arg_param->length;
    int   copy_len 	    	= 0;    
    unsigned long sector_size = SAM_FLASH_SECTORS_SIZE1;

    unsigned char 		    *mem_addr     = arg_param->data; 
    volatile char 	   	    *flash_base   = (char *)SAM_FLASH_START;    	
    volatile unsigned short *pFlashWriter = NULL;    
    volatile unsigned short *flash_addr   = (unsigned short*)((unsigned long)(SAM_FLASH_START) + arg_param->offset);	    
    volatile unsigned short *flash_sect   = (unsigned short *)((long)flash_addr & ~(sector_size-1));
	volatile unsigned char  *pTempRdr     = (unsigned char*)flash_addr;
	volatile unsigned char  *pTempSect    = (unsigned char*)flash_sect;	
    unsigned short 			*buffer 	  = NULL;

#ifdef ARUBA_DEBUG		
	printk("samsung_flash_write: BASE = %p, flash_addr = %p, len = %d\n", 
			flash_base, flash_addr, len);    
#endif

	if ( SAM_FLASH_END < (unsigned long)flash_addr)
	{
		printk("%s: Flash addr (0x%p) invalid > End of Flash (0x%lx)\n",  __FUNCTION__,
				flash_addr, (unsigned long)SAM_FLASH_END);
		return;
	}

	buffer=sector64;
	if (NULL == buffer)
	{
		printk("NOT ENOUGH RAM buffer memory to write to flash!\n");
		return;
	}
	
	do {
	
		copy_len = MIN(len, sector_size);
#ifdef ARUBA_DEBUG
		printk("Writing: flash_addr = %x, flash_sect=%x(#%d), copy_len=%d, len=%d\n", 
		flash_addr, flash_sect, SAM_IMAGE_SECT_NUM((unsigned long)flash_sect), copy_len, len);		
#endif
		/* read entire sector */
	    for(i=0; i<sector_size; i++) 
			buffer[i] = flash_sect[i];

		/* might need to remove if we're going to update idtboot dynamically */
		if ( (SAM_OVERWRITE_IDTBOOT((unsigned long)flash_addr)) ||
     	     (SAM_OVERWRITE_LAST_4((unsigned long)flash_addr)) )
		{
			printk("Overwriting critical sector in FL not allowed![addr=%p][sect adr=%p]\n", flash_addr, flash_sect);
			lock_ap60_flash();
			return;
		}

		/* have to erase sector before writing into it..can only program 1->0s */
		samsung_flash_erase_sector(flash_addr);		

	    /* update buffer with write data */
		memcpy(&buffer[flash_addr - flash_sect], mem_addr, copy_len);

		unlock_ap60_flash();

		/* program the sector */
		for(i=0; i<sector_size; i++) 
		{
			/* tell flash we want to program one byte/word */
			pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_FIRST_OFFSET);
			*pFlashWriter = 0xAAAA;
			pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_SECND_OFFSET);
			*pFlashWriter = 0x5555;		
			pFlashWriter  = (unsigned short*)(flash_base + SAM_FLASH_FIRST_OFFSET);				
			*pFlashWriter = 0xA0A0;	
		
			flash_sect[i]=buffer[i];
			
			j=0; 
			while((buffer[i]^flash_sect[i]) & 0x8080) {
				if (j++ > 1000000) {
					printk("Flash_W: TIMED-OUT %02x\n", buffer[i]);
					lock_ap60_flash();
					return;
				}
			} /* while, timeout */
		} 

        /* advance ptrs for additional sectors if any */
		len 		-= copy_len;
		pTempRdr	+= copy_len;
		flash_addr 	 = (unsigned short*)pTempRdr;
		mem_addr 	+= copy_len;
				
		if ((flash_addr - flash_sect) > sector_size)
		{
			pTempSect += sector_size;
			flash_sect = (unsigned short*)pTempSect;
		}
	
	} while (len);

	lock_ap60_flash(); 

	return;	
}

/************************************************************
 * This routine allows the tty driver to implement device-
 * specific ioctl's.  If the ioctl number passed in cmd is
 * not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int
ap_flash_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
    int ret_val = 0;
    volatile unsigned char *flash_addr  = NULL;
    mxc_nvram_t * arg_param =  (mxc_nvram_t *)arg;
    if (IS_SAMSUNG(mips_machtype))
    {
	flash_addr  = (unsigned char*)((unsigned long)(SAM_FLASH_START) + arg_param->offset);	    	
    }
    else
    {
	flash_addr  = (unsigned char*)((unsigned long)(AP60_FLASH_START) + arg_param->offset);	    	
    }

    switch (cmd) {
	/* add more commands later, possibly AP H/W read/write utils */
	
    case MXFLASHREAD:
		if (IS_SAMSUNG(mips_machtype))
			samsung_flash_read((mxc_nvram_t *) arg);					
		else
			ap_flash_read((mxc_nvram_t *) arg);
    	break;
    case MXFLASHWRITE:
		if (IS_SAMSUNG(mips_machtype))
			samsung_flash_write((mxc_nvram_t *) arg);					
		else	
			ap_flash_write((mxc_nvram_t *) arg);
    	break;
	/* no external API provided for this call intentionally */		
    case MXFLASHERASE: 
		if (IS_SAMSUNG(mips_machtype))
			samsung_flash_erase_sector(flash_addr);					
		else	
	    	ap_flash_erase_sector(flash_addr);
    	break;		
		
    default:
        ret_val = -ENOIOCTLCMD;
		break;
    }

    return ret_val;
} /* ap_flash_ioctl */


#ifndef FTL_MAJOR
#define FTL_MAJOR	229
#endif
int __init
ap_flash_init(void)
{
    memset(&ap_flash_driver, 0, sizeof(struct tty_driver));
    ap_flash_driver.magic = TTY_DRIVER_MAGIC;
    ap_flash_driver.driver_name = "ap_flash";
    ap_flash_driver.name = "mtd0";
    ap_flash_driver.major = FTL_MAJOR;
    ap_flash_driver.minor_start = 0;
    ap_flash_driver.num = 1;
    ap_flash_driver.type = TTY_DRIVER_TYPE_SERIAL;
    ap_flash_driver.subtype = SERIAL_TYPE_NORMAL;
    ap_flash_driver.init_termios = tty_std_termios;
    ap_flash_driver.init_termios.c_cflag =
            B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    ap_flash_driver.flags = TTY_DRIVER_REAL_RAW;
    ap_flash_driver.refcount = &ap_flash_refcount;
    ap_flash_driver.table = serial_table;
    ap_flash_driver.termios = serial_termios;
    ap_flash_driver.termios_locked = serial_termios_locked;

    ap_flash_driver.open = ap_flash_open;
    ap_flash_driver.close = ap_flash_close;
    ap_flash_driver.write = cab_write;
    ap_flash_driver.put_char = ap_flash_put_char;
    ap_flash_driver.flush_chars = ap_flash_flush_chars;
    ap_flash_driver.write_room = ap_flash_write_room;
    ap_flash_driver.chars_in_buffer = ap_flash_chars_in_buffer;
    ap_flash_driver.flush_buffer = ap_flash_flush_buffer;
    ap_flash_driver.ioctl = ap_flash_ioctl;
    ap_flash_driver.throttle = ap_flash_throttle;
    ap_flash_driver.unthrottle = ap_flash_unthrottle;
    ap_flash_driver.set_termios = ap_flash_set_termios;
    ap_flash_driver.stop = ap_flash_stop;
    ap_flash_driver.start = ap_flash_start;
    ap_flash_driver.hangup = ap_flash_hangup;
    ap_flash_driver.break_ctl = ap_flash_break;
    ap_flash_driver.wait_until_sent = ap_flash_wait_until_sent;
    ap_flash_driver.read_proc = cabernet_get_proc_info;

    if (tty_register_driver(&ap_flash_driver))
            panic("Couldn't register AP Flash driver\n");
    else {
#ifdef ARUBA_DEBUG
      printk("Successfully registered TTY driver for mtd0\n");
#endif
    }
#ifdef ARUBA_DEBUG
    printk("mtd0: AP Flash Control Driver built %s, %s\n", __DATE__, __TIME__);
#endif

    return 0;
} /* ap_flash_init */

#define MODULE
#ifdef MODULE
void
ap_flash_cleanup_module(void)
{
    int e;
    unsigned long flags;

    save_flags(flags); cli();
    if ((e = tty_unregister_driver(&ap_flash_driver)))
            printk("mtd0: failed to unregister Cabernet driver(%d)\n",
		   e);

    restore_flags(flags);

} /* ap_flash_cleanup_module */

#if 0
/* Module entry-points */
module_init(ap_flash_init);
module_exit(ap_flash_cleanup_module);
#endif
#else /* MODULE */
/* called by linux/init/main.c to parse command line options */
void
ap_flash_setup(char *str, int *ints)
{
} /* ap_flash_setup */
#endif /* MODULE */

MODULE_LICENSE("GPL");


static int ap_flash_open(struct tty_struct *tty, struct file * filp)
{
	return 0;
}

static void ap_flash_close(struct tty_struct *tty, struct file *filp)
{
	return;
}

static int cab_write(struct tty_struct * tty, int from_user,
	 const unsigned char *buf, int count)
{
	return 0;	 
}

static void ap_flash_put_char(struct tty_struct *tty, unsigned char ch)
{
	return;
}

static void ap_flash_flush_chars(struct tty_struct *tty)
{
	return;
}

static int ap_flash_write_room(struct tty_struct *tty)
{
	return 0;
}

static int ap_flash_chars_in_buffer(struct tty_struct *tty)
{
	return 0;
}

static void ap_flash_flush_buffer(struct tty_struct *tty)
{

} 

static void ap_flash_throttle(struct tty_struct * tty)
{

} 

static void ap_flash_unthrottle(struct tty_struct * tty)
{

} 

static void ap_flash_set_termios(struct tty_struct *tty, struct termios * old_termios)
{

}

static void ap_flash_stop(struct tty_struct *tty)
{

}

static void ap_flash_start(struct tty_struct *tty)
{

} 

static void ap_flash_hangup(struct tty_struct *tty)
{

}

static void ap_flash_break(struct tty_struct *tty, int break_state)
{

}

static void  ap_flash_wait_until_sent(struct tty_struct *tty, int timeout)
{

}

static int cabernet_get_proc_info(char *buf, char **start, off_t offset, int length,
		       int *eof, void *data)
{
	return 0;
}

