/*******************************************************************************
 *
 * rng_regs.h
 *
 * rc32365 RNG registers
 *
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef	_RNG_REGS_H_
#define	_RNG_REGS_H_

#include <linux/config.h>
#include <asm/rc32300/types.h> 
#include <asm/rc32300/rc32365_dma_v.h>

enum {
	RNG_PhysicalAddress	= 0x18080000,	
	RNG_VirtualAddress	= 0xb8080000,
};

typedef struct RNG_REG_ {
	U32 rngc;		/* RNGC */
	U32 rngd;		/* RNGD */
	U32 rngs;		/* RNGS */
	U32 rngsm;		/* RNGSM */
	U32 rngseed0;		/* RNGSEED0 */
	U32 rngseed1;		/* RNGSEED1 */
	U32 rngocfg;		/* RNGOCFG */		
	U32 rngcfg;		/* RNGCFG */
} RNG_REG;

enum {
	RNGC_en_b		= 0,
	RNGC_en_m		= 0x00000001,
	RNGC_f_b		= 1,
	RNGC_f_m		= 0x00000002,
	RNGC_roc_b		= 2,
	RNGC_roc_m		= 0x00000004,

	RNGS_r_b		= 0,
	RNGS_r_m		= 0x00000001,
	RNGS_br_b		= 1,
	RNGS_br_m		= 0x00000002,
	RNGS_td_b		= 2,
	RNGS_td_m		= 0x00000004,
	RNGS_mf_b		= 3,
	RNGS_mf_m		= 0x00000008,
	RNGS_rf_b		= 4,
	RNGS_rf_m		= 0x00000010,
	RNGS_lrf_b		= 5,
	RNGS_lfr_m		= 0x00000020,
	RNGS_brf_b		= 6,
	RNGS_brf_m		= 0x00000040,

	RNGSM_r_b		= 0,
	RNGSM_r_m		= 0x00000001,
	RNGSM_br_b		= 1,
	RNGSM_br_m		= 0x00000002,
	RNGSM_td_b		= 2,
	RNGSM_td_m		= 0x00000004,
	RNGSM_mf_b		= 3,
	RNGSM_mf_m		= 0x00000008,
	RNGSM_rf_b		= 4,
	RNGSM_rf_m		= 0x00000010,
	RNGSM_lrf_b		= 5,
	RNGSM_lfr_m		= 0x00000020,
	RNGSM_brf_b		= 6,
	RNGSM_brf_m		= 0x00000040,

	RNGOCFG_oscd_b		= 0,
	RNGOCFG_oscd_m		= 0x0000001f,
	RNGOCFG_rocfg_b		= 5,
	RNGOCFG_rocfg_m		= 0x000000e0,
	RNGOCFG_rocen_b		= 8,
	RNGOCFG_rocen_m		= 0x00000f00,

	RNGCFG_ert_b		= 0,
	RNGCFG_ert_m		= 0x00000001,
	RNGCFG_rbgos_b		= 1,
	RNGCFG_rbgos_m		= 0x0000001e,
	RNGCFG_bbc_b		= 5,
	RNGCFG_bbc_m		= 0x00000020,
	RNGCFG_dl_b		= 6,
	RNGCFG_dl_m		= 0x00000040,
	RNGCFG_unlock_b		= 11,
	RNGCFG_unlock_m		= 0xfffff800,
};

/*
 * RNG device data structure
 */
typedef struct RNG_DEVICE_ {
	volatile RNG_REG *rng_regs;	
	DMA_Chan_t dma_regs;
 	DMAD_t dma_desc;	/* RNG descriptor ring */ 
	int dma_irq;		/* RNG DMA irq */ 
	int irq;		/* RNG irq */
	u8* dma_buf;
	spinlock_t lock; 
} RNG_DEVICE;

#define RNG_DESC_NUM		1
#define RNG_DESC_RING_SIZE	(RNG_DESC_NUM * sizeof(struct DMAD_s))
#define RNG_DESC_BUF_SIZE	2500

#define RNG_SEED0		0xdeadbeef
#define RNG_SEED1		0x12345678
#define RNG_OCFG		0x00000f61	
#define RNG_CFG			0x00000000	

#endif
