/*
 * Include section 
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/mm.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/spinlock.h>
#include <linux/compile.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/bootinfo.h>

#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/version.h>

#include <linux/stat.h>
#include <linux/proc_fs.h>

#include "../../../cmn/util/mxc_ioctl.h"


extern unsigned long mips_machtype;
static struct tty_driver sam_driver;


/* This is the per-channel data structure containing pointers, flags
 and variables for the port. 
*/
static struct tty_struct *sam_serial_table[1];
static struct termios *sam_serial_termios[1];
static struct termios *sam_serial_termios_locked[1];


static void  sam_refcount(struct tty_struct *tty, int timeout);
static int sam_open(struct tty_struct *tty, struct file * filp);
static void sam_close(struct tty_struct *tty, struct file *filp);
static int sam_cab_write(struct tty_struct * tty, int from_user, const unsigned char *buf, int count);
static void sam_put_char(struct tty_struct *tty, unsigned char ch);
static void sam_flush_chars(struct tty_struct *tty);
static int sam_write_room(struct tty_struct *tty);
static int sam_chars_in_buffer(struct tty_struct *tty);
static void sam_flush_buffer(struct tty_struct *tty);
static void sam_throttle(struct tty_struct * tty);
static void sam_unthrottle(struct tty_struct * tty);
static void sam_set_termios(struct tty_struct *tty, struct termios *old_termios);
static void sam_stop(struct tty_struct *tty);
static void sam_start(struct tty_struct *tty);
static void sam_hangup(struct tty_struct *tty);
static void sam_break(struct tty_struct *tty, int break_state);
static void  sam_wait_until_sent(struct tty_struct *tty, int timeout);
static int sam_cabernet_get_proc_info(char *buf, char **start, off_t offset, int length,
int *eof, void *data);

/*
 * ------------------------------------------------------------
 * sam_ioctl() and friends
 * ------------------------------------------------------------
 */


/*******************************************************************************
 *     sam_read
 */
void sam_read(mxc_nvram_t * arg_param)
{
    unsigned char *machtype = (unsigned char *)arg_param->data; 

    if (mips_machtype == MACH_ARUBA_SAMSUNG)
    	*machtype = 0x1;
    else
    	*machtype = 0x0;    

#ifdef ARUBA_DEBUG		
	printk("sam_read: machine type is SAM? [%s]", (machtype)?"TRUE":"FALSE");
#endif
}

/************************************************************
 * This routine allows the tty driver to implement device-
 * specific ioctl's.  If the ioctl number passed in cmd is
 * not recognized by the driver, it should return ENOIOCTLCMD.
 */
static int
sam_ioctl(struct tty_struct *tty, struct file * file,
	 unsigned int cmd, unsigned long arg)
{
    int ret_val = 0;

    switch (cmd) {
	/* add more commands later, possibly AP H/W read/write utils */
	case SAMMACHTYPEREAD:
		sam_read((mxc_nvram_t *) arg);
	    	break;
/*		
	case SAMMACHTYPERELEASE:
		sam_cleanup_module(void);
	    	break;		
*/		
	default:
        	ret_val = -ENOIOCTLCMD;
		break;
    }

    return ret_val;
} /* sam_ioctl */


#ifndef SAM_MAJOR
#define SAM_MAJOR	222
#endif
/************************************************************
 * sam_init
 *
 */ 
int __init sam_init(void)
{


    memset(&sam_driver, 0, sizeof(struct tty_driver));
    sam_driver.magic = TTY_DRIVER_MAGIC;
    sam_driver.driver_name = "sam";
    sam_driver.name = "sam0";
    sam_driver.major = SAM_MAJOR;
    sam_driver.minor_start = 0;
    sam_driver.num = 1;
    sam_driver.type = TTY_DRIVER_TYPE_SERIAL;
    sam_driver.subtype = SERIAL_TYPE_NORMAL;
    sam_driver.init_termios = tty_std_termios;
    sam_driver.init_termios.c_cflag =
            B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    sam_driver.flags = TTY_DRIVER_REAL_RAW;
    sam_driver.refcount = sam_refcount;
    sam_driver.table = sam_serial_table;
    sam_driver.termios = sam_serial_termios;
    sam_driver.termios_locked = sam_serial_termios_locked;

    sam_driver.open = sam_open;
    sam_driver.close = sam_close;
    sam_driver.write = sam_cab_write;
    sam_driver.put_char = sam_put_char;
    sam_driver.flush_chars = sam_flush_chars;
    sam_driver.write_room = sam_write_room;
    sam_driver.chars_in_buffer = sam_chars_in_buffer;
    sam_driver.flush_buffer = sam_flush_buffer;
    sam_driver.ioctl = sam_ioctl;
    sam_driver.throttle = sam_throttle;
    sam_driver.unthrottle = sam_unthrottle;
    sam_driver.set_termios = sam_set_termios;
    sam_driver.stop = sam_stop;
    sam_driver.start = sam_start;
    sam_driver.hangup = sam_hangup;
    sam_driver.break_ctl = sam_break;
    sam_driver.wait_until_sent = sam_wait_until_sent;
    sam_driver.read_proc = sam_cabernet_get_proc_info;

    if (tty_register_driver(&sam_driver))
            panic("Couldn't register SAM machine type driver\n");
    else {
#ifdef ARUBA_DEBUG
      printk("Successfully registered TTY driver for sam0\n");
#endif
    }
#ifdef ARUBA_DEBUG
    printk("sam0: SAM mach type built %s, %s\n", __DATE__, __TIME__);
#endif

    return 0;
} /* sam_init */

#define MODULE
#ifdef MODULE
void
sam_cleanup_module(void)
{
    int e;
    unsigned long flags;

    save_flags(flags); 
    cli();
    if ((e = tty_unregister_driver(&sam_driver)))
            printk("sam0: failed to unregister driver(%d)\n", e);

    restore_flags(flags);

} /* sam_cleanup_module */

/* Module entry-points */
module_init(sam_init);
module_exit(sam_cleanup_module);

#else /* MODULE */
/* called by linux/init/main.c to parse command line options */
void
sam_setup(char *str, int *ints)
{
} /* sam_setup */
#endif /* MODULE */

MODULE_LICENSE("GPL");


static int sam_open(struct tty_struct *tty, struct file * filp)
{
	return 0;
}

static void sam_close(struct tty_struct *tty, struct file *filp)
{
	return;
}

static int sam_cab_write(struct tty_struct * tty, int from_user,
	 const unsigned char *buf, int count)
{
	return 0;	 
}

static void sam_put_char(struct tty_struct *tty, unsigned char ch)
{
	return;
}

static void sam_flush_chars(struct tty_struct *tty)
{
	return;
}

static int sam_write_room(struct tty_struct *tty)
{
	return 0;
}

static int sam_chars_in_buffer(struct tty_struct *tty)
{
	return 0;
}

static void sam_flush_buffer(struct tty_struct *tty)
{

} 

static void sam_throttle(struct tty_struct * tty)
{

} 

static void sam_unthrottle(struct tty_struct * tty)
{

} 

static void sam_set_termios(struct tty_struct *tty, struct termios * old_termios)
{

}

static void sam_stop(struct tty_struct *tty)
{

}

static void sam_start(struct tty_struct *tty)
{

} 

static void sam_hangup(struct tty_struct *tty)
{

}

static void sam_break(struct tty_struct *tty, int break_state)
{

}

static void  sam_wait_until_sent(struct tty_struct *tty, int timeout)
{

}

static int sam_cabernet_get_proc_info(char *buf, char **start, off_t offset, int length,
		       int *eof, void *data)
{
	return 0;
}

static void  sam_refcount(struct tty_struct *tty, int timeout)
{

}
