#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/smp_lock.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>
#include <asm/bootinfo.h>

// The timer needs to be reset within 16 seconds

#define SAP_WDT 203

#define WDT_TIMER HZ

// refresh the watchdog timer for this many seconds in kernel
// before letting the watchdog process take over.
#define WDT_COUNT (10*60)

static struct timer_list wdt_timer;
#define FOR_NOW
#ifdef FOR_NOW
static int wdt_count = WDT_COUNT;
#endif

extern unsigned long mips_machtype;
static void wdt_merlot_refresh(void){
  volatile __u32 *wdt;
  switch (mips_chip) {
  case MACH_ATHEROS_AR2313:
  case MACH_ATHEROS_AR5312:
    wdt = (__u32 *)0xbc00300c;
    *wdt = 0x40000000;
    break;
  case MACH_ATHEROS_AR2315:
    wdt = (__u32 *)0xb1000038;
    *wdt = 0x40000000;
    break;
  case MACH_IDT_RC32434:
    wdt = (__u32 *)0xb8030034;
    *wdt = 0x10000000;
    break;
  }
}

static void wdt_merlot_disable()
{
  volatile __u32 *wdt_errcs;
  volatile __u32 *wdt_wtc;
  volatile __u32 *wdt_ctl;
  volatile __u32 val ;

  switch (mips_chip) {
  case MACH_ATHEROS_AR2313:
  case MACH_ATHEROS_AR5312:
    wdt_ctl = (__u32 *)0xbc003008;
    *wdt_ctl = 0 ;
    break;
  case MACH_ATHEROS_AR2315:
    wdt_ctl = (__u32 *)0xb100003c;
    *wdt_ctl = 0 ;
    break;
  case MACH_IDT_RC32434:
    wdt_errcs = (__u32 *)0xb8030030 ;
    wdt_wtc   = (__u32 *)0xb803003c ;
    val = *wdt_errcs ;
    val &= ~0x201 ;
    *wdt_errcs = val ;
    val = *wdt_wtc ;
    val &= ~0x1 ;
    *wdt_wtc = val ;
    break;
  default:
    printk("%s: unsupported mips_chip=0x%x\n",
           __FUNCTION__, mips_chip) ;
  }
}

#if 1
// until the user process is ready, keep clearing the timer within the kernel

static void wdt_merlot_timer_fn(unsigned long data){

  wdt_merlot_refresh();
#ifdef FOR_NOW
  if (--wdt_count >= 0)
#endif
    mod_timer(&wdt_timer, jiffies+WDT_TIMER);
}

static int wdt_merlot_setup_timer(void){

  init_timer(&wdt_timer);
  wdt_timer.function = wdt_merlot_timer_fn;
  wdt_timer.data = 0;
  wdt_timer.expires = jiffies + HZ;
  add_timer(&wdt_timer);
  return 0;
}
#endif


static ssize_t wdt_merlot_read(struct file *file, char *buf, 
			       size_t count, loff_t *pos){

  return 0;

}

static ssize_t wdt_merlot_write(struct file *file, const char *buf, 
				size_t count, loff_t *pos){

  wdt_merlot_refresh();
  return count;

}

static int wdt_merlot_open(struct inode *inode, struct file *file){

  // no need to delete the timer since we will stop after 60 seconds
  //  del_timer(&wdt_timer);
  return 0;

}

static int wdt_merlot_close(struct inode *inode, struct file *file){

  return 0;
}

static struct file_operations wdt_ops = {
  owner: THIS_MODULE,
  read:  wdt_merlot_read,
  write: wdt_merlot_write,
  open:  wdt_merlot_open,
  release: wdt_merlot_close
};

static struct miscdevice wdt_merlot_dev = {
  SAP_WDT,
  "net/wdt",
  &wdt_ops
};


int __init wdt_merlot_init(void){
#ifdef ARUBA_DEBUG
  printk("Merlot WDT initialization\n");
#endif
  if (misc_register(&wdt_merlot_dev)){
    printk("Error registering wdt device\n");
    return -1;
  }

#if 1
  wdt_merlot_refresh();
#ifdef FOR_NOW
  wdt_merlot_setup_timer();
#endif
#else
  printk("Disabling Watchdog\n") ;
  wdt_merlot_disable() ;
#endif
  return 0;

}

void wdt_merlot_cleanup(void){
  misc_deregister(&wdt_merlot_dev);
}


