/*******************************************************************************
 *
 * se.h
 *
 * rc32365 Security Engine defines and data structures. 
 * 
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef _SE_H_
#define	_SE_H_

#include <linux/config.h>
#include <asm/rc32300/types.h> 
#include <asm/rc32300/rc32365_dma_v.h>
#include "se_regs.h"

/*
 * Buffer scheme options:
 * zero-memory-copy (USE_IO_BUF) is the default ("USE_IO_BUF" defined in linux/idt_crypto_api.h).
 * if USE_IO_BUF not defined, "USE_SK_BUF" allows new sk buffer be allocated.
 * otherwise, memory copy is used in driver. 
 */ 
/* #define USE_SK_BUF */	/* driver allocates new sk buffer (experimental)	*/

/*
 * DMA scheme options:
 * If USE_IO_BUF not defined. DMA chaining can be enabled. 
 */
//#define DMA_CHAINING	/* use dma chaining	*/

#define SE_PROC_DEBUG
//#undef SE_PROC_DEBUG

#if defined(USE_IO_BUF)
#undef DMA_CHAINING
#undef USE_SK_BUF
#endif

/* use 0 for production, 1 for verification, >2 for debug */
//#define SE_DEBUG 6
#define SE_DEBUG 0
#ifdef SE_DEBUG
#define dbg(lvl, format, arg...) \
    if (SE_DEBUG > lvl) \
        printk(KERN_INFO "se: " format, ## arg)
#else
#define dbg(lvl, format, arg...) do {} while (0)
#endif
#define err(format, arg...) \
		printk(KERN_ERR "se: " format, ## arg)
#define info(format, arg...) \
		printk(KERN_INFO "se: " format, ## arg)
#define warn(format, arg...) \
		printk(KERN_WARNING "se: " format, ## arg)

/* interrupt */
#define SE_DMA_I_IRQ	GROUP1_IRQ_BASE + 7	
#define SE_DMA_O_IRQ	GROUP1_IRQ_BASE + 6
#define SE_IRQ		GROUP2_IRQ_BASE + 1

/* descriptors */
#define SE_NUM_I	32		/* number of input descriptors */
#define SE_NUM_O	SE_NUM_I	/* number of output descrptors */
#define SE_I_MASK	(SE_NUM_I - 1)
#define SE_O_MASK	SE_I_MASK
#define SE_S_MASK	SE_I_MASK
#define SE_I_RING_SIZE	(SE_NUM_I * sizeof(struct DMAD_s))
#define SE_O_RING_SIZE	SE_I_RING_SIZE
#define SE_I_DATA_BUF_SIZE	1600	
#ifdef USE_SK_BUF
#define SE_I_DATA_HDR_SIZE	80
#endif

/* SE descriptor shadow structure */
typedef struct SE_DMAD_S_ {
	CRYPTO_REQ *req;
	int type;	/* data or command */
	int mode;	/* encode or decode */
	int cmd;	/* command type */	
	u32 *buf;	/* generic, cast to struct sk_buff *skb */
	int bufLen;	/* buffer length */
} SE_DMAD_S;

/*
 * SE device statistics 
 */
typedef struct SE_DEVICE_STATS_ {
	unsigned long	i_packets;		/* total packets inputed	*/
	unsigned long	o_packets;		/* total packets outputed	*/
	unsigned long	i_bytes;		/* total bytes inputed 	*/
	unsigned long	o_bytes;		/* total bytes outputed	*/
	unsigned long	i_errors;		/* bad packets inputed	*/
	unsigned long	o_errors;		/* packet outpu problems	*/
	unsigned long	i_dropped;		/* no space in linux buffers	*/
} SE_DEVICE_STATS;

/*
 * SE device data structure
 */
typedef struct SE_DEVICE_ {
	volatile SE_REG *se_regs;	
	DMA_Chan_t i_dma_regs;
	DMA_Chan_t o_dma_regs;
 	DMAD_t o_ring;		/* output descriptor ring */ 
 	DMAD_t i_ring;		/* input descriptor ring */ 
#ifdef USE_SK_BUF
	struct sk_buff *i_skb[SE_NUM_I]; /* skbuffs for input dma */
#endif
	SE_DMAD_S *s_ring;	/* shadow ring */
	u8 *i_data_buf;		/* input DMA data buffer area */  
	int i_tail;		/* next to be added to input ring */ 
	int i_head;		/* last received from device	*/ 
#ifdef DMA_CHAINING
	int o_chain;		/* next output chain */
	int i_chain;		/* next input chain */
	int i_chain_empty;
#endif
	int o_tail;		/* next to be added to output ring */
	int o_head;		/* last processed by device */
	int s_tail;		/* next to be added to sync ring */
	int s_head;		/* next taken out from sync ring */
	int o_count;		/* outstanding packets in the out ring */
	int o_full;		/* output descriptor ring full flag */
	int i_irq;		/* input ring DAM irq */
	int o_irq;		/* output ring DMA irq */
	int irq;		/* SE irq */

	SE_DEVICE_STATS stats;
	spinlock_t lock; 
	/* debug /proc entry */
	struct proc_dir_entry *ps;
	int dma_full_cnt;
	int cntx_swap_cnt;
	int pkt_cnt;
	int cmd_cnt;

	QHEAD crypto_done_q;	
} SE_DEVICE;

#define COMMAND		0x1
#define DATA		0x2

#define INVALID_ID	0xFFFF
typedef struct CONTEXT_CACHE_ {
	unsigned int contextId;
	SECURITY_RAM_CONTEXT securityContext;
	byte pb;
} CONTEXT_CACHE;

#define SE_CONTEXT_PB_COMMAND_LOAD	0x0041 

#endif 
