/*******************************************************************************
 *
 * se_regs.h
 *
 * rc32365 Security Engine registers
 * 
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef	_SE_REGS_H_
#define	_SE_REGS_H_
#include  <asm/rc32300/types.h> 

enum {
	SE_PhysicalAddress	= 0x18078000,	
	SE_VirtualAddress	= 0xb8078000,
};

typedef struct SE_REG_ {
	U32 sec;		/* SEC */
	U32 ses;		/* SES */
	U32 sesm;		/* SESM	*/
	U32 serc;		/* SERC	*/
	U32 sespare;
	U32 sejunk[58];		
	U32 serlc;		/* SERLC */
} SE_REG;

enum {
	SESEC_en_b		= 0,
	SESEC_en_m		= 0x00000001,
	SESEC_vc_b		= 1,
	SESEC_vc_m		= 0x00000002,

	SESES_mids_b		= 0,
	SESES_mids_m		= 0x00000001,
	SESES_rme_b		= 1,
	SESES_rme_m		= 0x00000002,
	SESES_dme_b		= 2,
	SESES_dme_m		= 0x00000004,
	SESES_cde_b		= 3,
	SESES_cde_m		= 0x00000008,
	
	SESESM_mids_b		= 0,
	SESESM_mids_m		= 0x00000001,
	SESESM_rme_b		= 1,
	SESESM_rme_m		= 0x00000002,
	SESESM_dme_b		= 2,
	SESESM_dme_m		= 0x00000004,
	SESESM_cde_b		= 3,
	SESESM_cde_m		= 0x00000008,

	SESERC_rl_b		= 0,
	SESERC_rl_m		= 0x000003ff,
	SESERC_crl_b		= 16,
	SESERC_crl_m		= 0x03ff0000,

	SESERLC_serl_b		= 0,
	SESERLC_serl_m		= 0x000003ff
};

/*
 * DEVCS field for SE DMA operations.
 */
enum {
	/* command */
	SEDEVCSCMD_cntxt_b	= 0,
	SEDEVCSCMD_cntxt_m	= 0x000001ff,
	SEDEVCSCMD_ce_b		= 10,
	SEDEVCSCMD_ce_m		= 0x00000400,
	SEDEVCSCMD_e_b		= 11,
	SEDEVCSCMD_e_m		= 0x00000800,
	SEDEVCSCMD_cmd_b	= 12,
	SEDEVCSCMD_cmd_m	= 0x000ff000,
		SEDEVCSCMD_cmd_ld_cntxt_v	= 0,	/* load security context */
		SEDEVCSCMD_cmd_st_cntxt_v	= 1,	/* store security context */
		SEDEVCSCMD_cmd_rp_cntxt_v	= 2,	/* replace security context */
		SEDEVCSCMD_cmd_ld_dssf_v	= 3,	/* load data stream segmentation fields */
		SEDEVCSCMD_cmd_ld_iv_v		= 4,	/* load iv */
		SEDEVCSCMD_cmd_st_iv_v		= 5, 	/* store iv */
		SEDEVCSCMD_cmd_ld_pb_v		= 6,	/* load pad byte */
		SEDEVCSCMD_cmd_init_hmac_md5_v	= 7,	/* initialize hash unit for hmac-md5 operation */
		SEDEVCSCMD_cmd_init_hmac_sha1_v	= 8,	/* initizlize hash unit for hmac-sha1 operation */
	SEDEVCSCMD_imm_b	= 20,
	SEDEVCSCMD_imm_m	= 0x0ff00000,

	/* data stream (memory to SE) */
	SEDEVCSOUT_cntxt_b	= 0,
	SEDEVCSOUT_cntxt_m	= 0x000001ff,
	SEDEVCSOUT_dm_b		= 9,
	SEDEVCSOUT_dm_m		= 0x00000200,
	SEDEVCSOUT_ovr_b	= 10,
	SEDEVCSOUT_ovr_m	= 0x00000400,
	SEDEVCSOUT_f_b		= 11,
	SEDEVCSOUT_f_m		= 0x0001f800,
	SEDEVCSOUT_l_b		= 17,
	SEDEVCSOUT_l_m		= 0xfffe0000, 

	/* data stream (SE to memory) */
	SEDEVCSIN_cntxt_b	= 0,
	SEDEVCSIN_cntxt_m	= 0x000001ff,
	SEDEVCSIN_ce_b		= 10,
	SEDEVCSIN_ce_m		= 0x00000400,
	SEDEVCSIN_e_b		= 11,
	SEDEVCSIN_e_m		= 0x00000800,
	SEDEVCSIN_eme_b		= 12,
	SEDEVCSIN_eme_m		= 0x00001000,
	SEDEVCSIN_ebe_b		= 13,
	SEDEVCSIN_ebe_m		= 0x00002000,
	SEDEVCSIN_hme_b		= 14,
	SEDEVCSIN_hme_m		= 0x00004000,
	SEDEVCSIN_htm_b		= 15,
	SEDEVCSIN_htm_m		= 0x00008000,
	SEDEVCSIN_pme_b		= 16,
	SEDEVCSIN_pme_m		= 0x00010000,
	SEDEVCSIN_pe_b		= 17,
	SEDEVCSIN_pe_m		= 0x00020000,
	SEDEVCSIN_nhe_b		= 18,
	SEDEVCSIN_nhe_m		= 0x00040000,
	SEDEVCSIN_ahm_b		= 19,
	SEDEVCSIN_ahm_m		= 0x00080000,
	SEDEVCSIN_plength_b	= 24,
	SEDEVCSIN_plength_m	= 0x1f000000,
	
	SEDEVCSIN_error_b	= 10,
	SEDEVCSIN_error_m	= 0x000ffc00, 
};

/*
 * DEVCMD field for SE DMA operations. 
 */
enum {
	SEDMAD_devcmd_cmd_v		= 0,
	SEDMAD_devcmd_middle_data_v	= 4,
	SEDMAD_devcmd_first_data_v	= 5,
	SEDMAD_devcmd_last_data_v	= 6,
	SEDMAD_devcmd_first_last_data_v	= 7
};

#define IS_CMD_CXT_ERR(x)	(((x) & SEDEVCSCMD_ce_m) >> SEDEVCSCMD_ce_b)	/* command context */
#define IS_CMD_PROC_ERR(x)	(((x) & SEDEVCSCMD_e_m) >> SEDEVCSCMD_e_b)	/* command process */	

#define IS_DATA_CXT_ERR(x)	(((x) & SEDEVCSIN_ce_m) >> SEDEVCSIN_ce_b)	/* context */ 
#define IS_DATA_ERR(x)		(((x) & SEDEVCSIN_e_m) >> SEDEVCSIN_e_b)	/* data */ 
#define IS_DATA_EME_ERR(x)	(((x) & SEDEVCSIN_eme_m) >> SEDEVCSIN_eme_b)	/* encrypt unit misconfig */ 
#define IS_DATA_EBE_ERR(x)	(((x) & SEDEVCSIN_ebe_m) >> SEDEVCSIN_ebe_b)	/* encrypt unit block */ 
#define IS_DATA_HME_ERR(x)	(((x) & SEDEVCSIN_hme_m) >> SEDEVCSIN_hme_b)	/* hash unit misconfig */ 
#define IS_DATA_HTM_ERR(x)	(((x) & SEDEVCSIN_htm_m) >> SEDEVCSIN_htm_b)	/* hash trailer mismatch */ 
#define IS_DATA_PME_ERR(x)	(((x) & SEDEVCSIN_pme_m) >> SEDEVCSIN_pme_b)	/* pad unit misconfig */ 
#define IS_DATA_PE_ERR(x)	(((x) & SEDEVCSIN_pe_m) >> SEDEVCSIN_pe_b)	/* pad */ 
#define IS_DATA_NHE_ERR(x)	(((x) & SEDEVCSIN_nhe_m) >> SEDEVCSIN_nhe_b)	/* next header */ 
#define IS_DATA_AHM_ERR(x)	(((x) & SEDEVCSIN_ahm_m) >> SEDEVCSIN_ahm_b)	/* AH hash mismatch */ 

#define DATA_PAD_LENGTH(x)	(((x) & SEDEVCSIN_plength_m) >> SEDEVCSIN_plength_b)	/* pad length */ 
#define DATA_CONTEXT(x)		(((x) & SEDEVCSIN_cntxt_m) >> SEDEVCSIN_cntxt_b)	/* security context */ 
#define CMD_CONTEXT(x)		(((x) & SEDEVCSCMD_cntxt_m) >> SEDEVCSCMD_cntxt_b)	/* security context */ 
#define IS_SE_DATA(x)		(((x) & DMAD_devcmd_m) >> DMAD_devcmd_b)		/* data or command */
#define IS_SESES_MIDS_ERR(x)	((x) & SESES_mids_m)		/* Malformed input data stream */
#define IS_SESES_RME_ERR(x)	((x) & SESES_rme_m) 		/* Reserved mode error */
#define IS_SESES_DME_ERR(x)	((x) & SESES_dme_m) 		/* Disabled mode error */
#define IS_SESES_CDE_ERR(x)	((x) & SESES_cde_m) 		/* Command data error */
#define IS_ANY_ERROR(x)		((x) & SEDEVCSIN_error_m)	/* any error */

#endif	/* _SE_REGS_H_ */

