/*  i2c-rc32438.h - Linux i2c driver for IDT 79RC32438
 
   	Author: Integrated Device Technology, Inc.

modification history
--------------------
March 6 2003	hfkou	create



+-------------------------------------------------------------------------------+
|   DESCRIPTION                                                                 								|
|      This module contains the low level driver .H for 79RC32438 			    				|
|      																			|
+-------------------------------------------------------------------------------+
*/

#ifndef _I2C_RC32438_
#define _I2C_RC32438_
#ifdef __cplusplus
extern "C" {
#endif



struct i2c_rc32438_iface
{
	struct i2c_adapter	adapter;

	unsigned long		base;
	int				irq;
	char				read_write;
	u8*				data;
	unsigned			datalen;
	int				state;
	int				result;
};

#define I2CMCMD_CMD(i)      ((i&0xffff)<<0)

// I2CM Commands
#define NOP     0
#define START   1
#define STOP    2
#define RD      4
#define RDACK   5
#define WD      6
#define WDACK   7

#define TRUE    1
#define FALSE   0 

//	NOTE:!!!
//	Please assign your I2C device's Slave Adress here, otherwise it will use the default values
#define SLAVE_ADDR      0x50        

#define PRESCALER_VAL   110         

typedef struct {
  unsigned int transfer_type;
  unsigned int data_len;
  unsigned char *data_ptr;
  unsigned char dest_addr[1];
  unsigned int stop_when_done;
  unsigned int state;
} I2C_MASTER_TYPE;
// Global Variables
I2C_MASTER_TYPE rc32438_master;

#define MASTER_WRITE 0
#define MASTER_READ  1

// Master State machine states...
#define MASTER_IDLE        0
#define MASTER_START       1
#define MASTER_ADDR        2
#define MASTER_READ_DATA   3
#define MASTER_WRITE_DATA  4
#define MASTER_STOP        5


/********************************************************************
 * I2C interface
 ********************************************************************/
#define I2C_REGION          0x7FFF
#define I2C_BASE          0xB8070000


typedef struct {
  unsigned int i2cc;
  unsigned int i2cdi;
  unsigned int i2cdo;
  unsigned int i2ccp;
  unsigned int i2cmcmd;
  unsigned int i2cms;
  unsigned int i2cmsm;
  unsigned int i2css;
  unsigned int i2cssm;
  unsigned int i2csaddr;
  unsigned int i2csack;
} I2C;
 
#define i2c (*((volatile I2C *) I2C_BASE))

// I2CC register
#define I2CC_MEN            (1<<0)              // MEN bit
#define I2CC_SEN            (1<<1)              // SEN bit
#define I2CC_IOM            (1<<2)              // IOM bit

// I2CMS register
#define I2CMS_D             (1<<0)              // D bit
#define I2CMS_NA            (1<<1)              // No ACK bit
#define I2CMS_LA            (1<<2)              // LA bit
#define I2CMS_ERR           (1<<3)              // ERR bit

// I2CSS register
#define I2CSS_RR            (1<<0)              // RR bit
#define I2CSS_WR            (1<<1)              // WR bit
#define I2CSS_SA            (1<<2)              // SA bit
#define I2CSS_TF            (1<<3)              // TF bit
#define I2CSS_GC            (1<<4)              // GC bit
#define I2CSS_NA            (1<<5)              // No ACK bit
#define I2CSS_ERR           (1<<6)              // ERR bit

// I2CSADDR register
#define I2CSADDR_ADDR(i)    ((i&0x3ff)<<0)      // ADDR field
#define I2CSADDR_GC         (1<<10)             // GC bit
#define I2CSADDR_A10        (1<<11)             // A10 bit

//I2CSACK register
#define I2CSACK_ACK         (1<<0)              // ACK bit

//I2CCP register
#define I2CCP_DIV(i)        ((i&0xffff)<<0)     // DIV field

//I2CDI register
#define I2CDI_DATA(i)       ((i&0xffff)<<0)     // DATA field

//I2CDO register
#define I2CDO_DATA(i)       ((i&0xffff)<<0)     // DATA field



#define I2CDELAY 	400000
 

/*24LC64 i2c device commands */

#define I2C_WRITE	0xAE	/*Write operation selected */
#define I2C_READ	0xAF	/*Read operation selected */

#define ADDR_MASK       0xFF
#define OK 0
#define ERROR 1
#define I2C_MEM_SIZE    8192
#define EOS             '\0'


#endif /* _I2C_RC32438_ */

