/*
 *
 * BRIEF MODULE DESCRIPTION
 *	Include file for cedar.c
 *
 * Author: MontaVista Software, Inc.
 *		stevel@mvista.com or source@mvista.com
 *
 * Heavily modified from original version by:
 *
 *
 * ########################################################################
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ########################################################################
 */

#ifndef CEDAR_H
#define CEDAR_H

#include  <asm/rc32300/rc32365_eth.h> 
#include  <asm/rc32300/rc32365_dma.h> 
#include  <asm/rc32300/rc32365.h> 
#include  <asm/rc32300/79eb365.h> 


//#define ASSERT(boolExp) do{}while(0)
#define ASSERT(boolExp) do{if(!(boolExp))err("ASSERT %s(%d): %s\n",__FILE__, __LINE__, #boolExp);}while(0)
 

#ifdef CEDAR_DEBUG
#define dbg(lvl, format, arg...) \
    if (cedar_debug > lvl) \
	printk(KERN_INFO "%s: " format, dev->name , ## arg)
#else
#define dbg(lvl, format, arg...) do {} while (0)
#endif
#define err(format, arg...) \
    printk(KERN_ERR "%s: " format, dev->name , ## arg)
#define info(format, arg...) \
    printk(KERN_INFO "%s: " format, dev->name , ## arg)
#define warn(format, arg...) \
    printk(KERN_WARNING "%s: " format, dev->name , ## arg)

#define ETH0_DMA_RX_IRQ   (GROUP1_IRQ_BASE + 0)
#define ETH0_DMA_TX_IRQ   (GROUP1_IRQ_BASE + 1)
#define ETH0_RX_OVR_IRQ   (GROUP3_IRQ_BASE + 4)
#define ETH0_TX_UND_IRQ   (GROUP3_IRQ_BASE + 5)
#define ETH1_DMA_RX_IRQ   (GROUP1_IRQ_BASE + 2)
#define ETH1_DMA_TX_IRQ   (GROUP1_IRQ_BASE + 3)
#define ETH1_RX_OVR_IRQ   (GROUP3_IRQ_BASE + 7)
#define ETH1_TX_UND_IRQ   (GROUP3_IRQ_BASE + 8)

/* Index to functions, as function prototypes. */

static int cedar_open(struct net_device *dev);
static int cedar_send_packet(struct sk_buff *skb, struct net_device *dev);
static void cedar_ovr_interrupt(int irq, void *dev_id,
			      struct pt_regs * regs);
static void cedar_und_interrupt(int irq, void *dev_id,
			      struct pt_regs * regs);
static void cedar_rx_dma_interrupt(int irq, void *dev_id,
				 struct pt_regs * regs);
static void cedar_tx_dma_interrupt(int irq, void *dev_id,
				 struct pt_regs * regs);
static void cedar_rx(struct net_device *dev);
static void cedar_tx(struct net_device *dev);
static int  cedar_close(struct net_device *dev);
static struct net_device_stats *cedar_get_stats(struct net_device *dev);
static void cedar_multicast_list(struct net_device *dev);
static int  cedar_init(struct net_device *dev);
static void cedar_tx_timeout(struct net_device *dev);

#define CEDAR_NUM_RDS	 128    /* number of receive descriptors */
#define CEDAR_NUM_TDS	 128    /* number of transmit descriptors */

#define CEDAR_RBSIZE	 1536  /* size of one resource buffer = Ether MTU */
#define CEDAR_RDS_MASK	 (CEDAR_NUM_RDS-1)
#define CEDAR_TDS_MASK	 (CEDAR_NUM_TDS-1)
#define RD_RING_SIZE	 (CEDAR_NUM_RDS * sizeof(struct DMAD_s))
#define TD_RING_SIZE	 (CEDAR_NUM_TDS * sizeof(struct DMAD_s))

#define CEDAR_TX_TIMEOUT HZ/4

#define rc32365_eth0_regs ((ETH_t)(ETH0_VirtualAddress);
#define rc32365_eth1_regs ((ETH_t)(ETH1_VirtualAddress);

enum status	{ filled,	empty};
#define IS_DMA_USED(X) (((X) & (DMAD_f_m | DMAD_d_m | DMAD_t_m)) != 0)
#define IS_DMA_FINISHED(X)   (((X) & (DMAD_f_m)) != 0)
#define IS_DMA_DONE(X)   (((X) & (DMAD_d_m)) != 0)

/* Information that need to be kept for each board. */
struct cedar_local {
	ETH_t  eth_regs;
	DMA_Chan_t  rx_dma_regs;
	DMA_Chan_t  tx_dma_regs;
	volatile DMAD_t   td_ring;  					/* transmit descriptor ring */ 
	volatile DMAD_t   rd_ring;  					/* receive descriptor ring  */

	struct sk_buff* tx_skb[CEDAR_NUM_TDS]; 	/* skbuffs for pkt to trans */
	struct sk_buff* rx_skb[CEDAR_NUM_RDS]; 	/* skbuffs for pkt to trans */


	int	rx_next_done;
	int	rx_chain_head;
	int	rx_chain_tail;
	enum status	rx_chain_status;
	
	int	tx_next_done;
	int	tx_chain_head;
	int	tx_chain_tail;
	enum status	tx_chain_status;
	int	tx_count;	   						/* current # of pkts waiting to be sent */
	int	tx_full;

	int	rx_irq;
	int	tx_irq;
	int	ovr_irq;
	int	und_irq;

	struct net_device_stats stats;
	spinlock_t lock; 							/* Serialise access to device */

	/* debug /proc entry */
	struct proc_dir_entry *ps;
	int dma_halt_cnt;    u32 halt_tx_count;
	int dma_collide_cnt; u32 collide_tx_count;
	int dma_run_cnt;     u32 run_tx_count;
	int dma_race_cnt;    u32 race_tx_count;
};


#define IS_TX_TOK(X)	     (((X) & (1<<ETHTX_tok_b)) >> ETHTX_tok_b )   /* Transmit Okay    */
#define IS_TX_MP(X)	     (((X) & (1<<ETHTX_mp_b))  >> ETHTX_mp_b )	  /* Multicast	      */
#define IS_TX_BP(X)	     (((X) & (1<<ETHTX_bp_b))  >> ETHTX_bp_b )	  /* Broadcast	      */
#define IS_TX_UND_ERR(X)     (((X) & (1<<ETHTX_und_b)) >> ETHTX_und_b )   /* Transmit FIFO Underflow */
#define IS_TX_OF_ERR(X)      (((X) & (1<<ETHTX_of_b))  >> ETHTX_of_b )	  /* Oversized frame  */
#define IS_TX_ED_ERR(X)      (((X) & (1<<ETHTX_ed_b))  >> ETHTX_ed_b )	  /* Excessive deferral  */
#define IS_TX_EC_ERR(X)      (((X) & (1<<ETHTX_ec_b))  >> ETHTX_ec_b)	  /* Excessive collisions  */
#define IS_TX_LC_ERR(X)      (((X) & (1<<ETHTX_lc_b))  >> ETHTX_lc_b )	  /* Late Collision   */
#define IS_TX_TD_ERR(X)      (((X) & (1<<ETHTX_td_b))  >> ETHTX_td_b )	  /* Transmit deferred*/
#define IS_TX_CRC_ERR(X)     (((X) & (1<<ETHTX_crc_b)) >> ETHTX_crc_b )   /* CRC Error	      */
#define IS_TX_LE_ERR(X)      (((X) & (1<<ETHTX_le_b))  >>  ETHTX_le_b )    /* Length Error     */

#define TX_COLLISION_COUNT(X) (((X) & ETHTX_cc_m)>>ETHTX_cc_b)	/* Collision Count  */

#define IS_RCV_ROK(X)	     (((X) & (1<<ETHRX_rok_b)) >> ETHRX_rok_b)	  /* Receive Okay     */
#define IS_RCV_FM(X)	     (((X) & (1<<ETHRX_fm_b))  >> ETHRX_fm_b)	  /* Is Filter Match  */
#define IS_RCV_MP(X)	     (((X) & (1<<ETHRX_mp_b))  >> ETHRX_mp_b)	  /* Is it MP	      */
#define IS_RCV_BP(X)	     (((X) & (1<<ETHRX_bp_b))  >> ETHRX_bp_b)	  /* Is it BP	      */
#define IS_RCV_VLT(X)	     (((X) & (1<<ETHRX_vlt_b)) >> ETHRX_vlt_b)	  /* VLAN Tag Detect  */
#define IS_RCV_CF(X)	     (((X) & (1<<ETHRX_cf_b))  >> ETHRX_cf_b)	  /* Control Frame    */
#define IS_RCV_OVR_ERR(X)    (((X) & (1<<ETHRX_ovr_b)) >> ETHRX_ovr_b)	  /* Receive Overflow */
#define IS_RCV_CRC_ERR(X)    (((X) & (1<<ETHRX_crc_b)) >> ETHRX_crc_b)	  /* CRC Error	      */
#define IS_RCV_CV_ERR(X)     (((X) & (1<<ETHRX_cv_b))  >> ETHRX_cv_b)	  /* Code Violation   */
#define IS_RCV_DB_ERR(X)     (((X) & (1<<ETHRX_db_b))  >> ETHRX_db_b)	  /* Dribble Bits     */
#define IS_RCV_LE_ERR(X)     (((X) & (1<<ETHRX_le_b))  >> ETHRX_le_b)	  /* Length error     */
#define IS_RCV_LOR_ERR(X)    (((X) & (1<<ETHRX_lor_b)) >> ETHRX_lor_b)	  /* Length Out of Range */
#define IS_RCV_CES_ERR(X)    (((X) & (1<<ETHRX_ces_b)) >> ETHRX_ces_b)	/* Preamble error   */
#define RCVPKT_LENGTH(X)     (((X) & ETHRX_length_m) >> ETHRX_length_b)   /* Length of the received packet */


#define DMA_COUNT(count)   \
  ((count) & DMAD_count_m)
#define DMA_HALT_TIMEOUT 500

static inline int rc32365_halt_dma(DMA_Chan_t ch)
{
	int timeout=1;
	if (local_readl(&ch->dmac) & DMAC_run_m) {
		local_writel(0, &ch->dmac); 
		for (timeout = DMA_HALT_TIMEOUT; timeout > 0; timeout--) {
			if (local_readl(&ch->dmas) & DMAS_h_m) {
				local_writel(0, &ch->dmas);  
				break;
			}
		}
	}

	return timeout ? 0 : 1;
}

static inline void rc32365_start_dma(DMA_Chan_t ch, u32 dma_addr)
{
	local_writel(0, &ch->dmandptr); 
	local_writel(dma_addr, &ch->dmadptr);
}

static inline void rc32365_chain_dma(DMA_Chan_t ch, u32 dma_addr)
{
	local_writel(dma_addr, &ch->dmandptr);
}


#endif /* CEDAR_H */













