/*
   	atp870u.c


   	Author: Integrated Device Technology, Inc.
  
	1997	Wu Ching Chen
	2.1.x update 1998  Krzysztof G. Baranowski
	Marcelo Tosatti <marcelo@conectiva.com.br> : SMP fixes
   	$Id: atp870u.c,v 1.0 1997/05/07 15:22:00 root Exp root $
    
   	A driver for the Acard ATP870 SCSI controller.
   
   	This driver is modified from Acard:
  
  
   	Haofeng Kou:			Design modify and maintain the code. 
      						The author may be reached at hkou@idt.com or
						IDT-IPD, 2975 Stender Way, Santa Clara, CA 95054 


	Changelog:

	Version 1.0			Wu Ching Chen		NULL pointer fixes  2000/06/02
	 					: support atp876 chip
						: enable 32 bit fifo transfer
						: support cdrom & remove device run ultra speed
						: fix disconnect bug  2000/12/21
						: support atp880 chip lvd u160 2001/05/15 (7.1)

					  	Haofeng Kou			June 8, 2003
                    				: Ported to IDT RC32XXX.
                    				: Add in code to support Memory Map IO.
                    				: Fix certain bugs for the PRD Table handling.

   ########################################################################
  
    This program is free software; you can redistribute  it and/or modify it
    under  the terms of  the GNU General  Public License as published by the
    Free Software Foundation;  either version 2 of the  License, or (at your
    option) any later version.
  
    THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
    WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
    NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
    USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
    ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  
    You should have received a copy of the  GNU General Public License along
    with this program; if not, write  to the Free Software Foundation, Inc.,
    675 Mass Ave, Cambridge, MA 02139, USA.
  
   ########################################################################
 */



#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/spinlock.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/blk.h>
#include<linux/stat.h>
#include <linux/in.h>

#include "scsi.h"
#include "hosts.h"
#include "atp870u.h"



#define MMIO
//#define _DEUBG_ATP__ 

#ifndef _DEUBG_ATP__
#define FUNC_IN() 
#define FUNC_STR(s) 
#define FUNC_VAR(a) 
#define FUNC_OUT() 
#else
#define FUNC_IN() printk("Entering %s\n",__FUNCTION__);
#define FUNC_STR(s) printk("    IN THE %s	%s\n",__FUNCTION__, s);
#define FUNC_VAR(a) printk("    IN THE %s	%x\n",__FUNCTION__, a);
#define FUNC_OUT() printk("Leaving %s\n",__FUNCTION__);
#endif

#define PRD_SIZE		1024

#ifdef MMIO
#define atp870u_inb(a)		readb(a)
#define atp870u_inw(a)		readw(a)
#define atp870u_inl(a)		readl(a)
#define atp870u_outb(v, a)	writeb(v, a)
#define atp870u_outw(v, a)	writew(v, a)
#define atp870u_outl(v, a)	writel(v, a)
#else
#define atp870u_inb(a)		inb(a)
#define atp870u_inw(a)		inw(a)
#define atp870u_inl(a)		inl(a)
#define atp870u_outb(v, a)	outb(v, a)
#define atp870u_outw(v, a)	outw(v, a)
#define atp870u_outl(v, a)	outl(v, a)
#endif



void mydlyu(unsigned int);

/*
 *   static const char RCSid[] = "$Header: /usr/src/linux/kernel/blk_drv/scsi/RCS/atp870u.c,v 1.0 1997/05/07 15:22:00 root Exp root $";
 */

static unsigned char admaxu = 1;
static unsigned short int sync_idu;

static unsigned int irqnumu[2] = {0, 0};

struct atp_unit
{
	unsigned long ioport;
	unsigned long irq;
	unsigned long pciport;
	unsigned char last_cmd;
	unsigned char in_snd;
	unsigned char in_int;
	unsigned char quhdu;
	unsigned char quendu;
	unsigned char scam_on;
	unsigned char global_map;
	unsigned char chip_veru;
	unsigned char host_idu;
	int working;
	unsigned short wide_idu;
	unsigned short active_idu;
	unsigned short ultra_map;
	unsigned short async;
	unsigned short deviceid;
	unsigned char ata_cdbu[16];
	unsigned char sp[16];
	Scsi_Cmnd *querequ[qcnt];
	struct atp_id
	{
		unsigned char dirctu;
		unsigned char devspu;
		unsigned char devtypeu;
		unsigned long prdaddru;
		unsigned long tran_lenu;
		unsigned long last_lenu;
		unsigned char *prd_posu;
		unsigned char *prd_tableu;
		Scsi_Cmnd *curr_req;
	} id[16];
};

static struct Scsi_Host *atp_host[2] = {NULL, NULL};
static struct atp_unit atp_unit[2];


static void atp870u_intr_handle(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned long flags;
//	unsigned short int tmpcip, id;	
	unsigned int tmpcip, id;
	unsigned char i, j, h, target_id, lun;
	unsigned char *prd;
	Scsi_Cmnd *workrequ;
	unsigned int workportu, tmport;
	unsigned long adrcntu, k;
	int errstus;
	struct atp_unit *dev = dev_id;

	FUNC_IN();
	for (h = 0; h < 2; h++) {
		if (irq == irqnumu[h]) {
			goto irq_numok;
		}
	}
	FUNC_STR("		1");
	return;
irq_numok:
	dev->in_int = 1;
	workportu = dev->ioport;
	tmport = workportu;

	if (dev->working != 0)
	{
		tmport += 0x1f;
		j = atp870u_inb(tmport);
		if ((j & 0x80) == 0)
		{
			dev->in_int = 0;
			FUNC_STR("		irq_numok");			
			return;
		}

		tmpcip = dev->pciport;
		if ((atp870u_inb(tmpcip) & 0x08) != 0)
		{
			tmpcip += 0x2;
			for (k=0; k < 1000; k++)
			{
				if ((atp870u_inb(tmpcip) & 0x08) == 0)
				{
					goto stop_dma;
				}
				if ((atp870u_inb(tmpcip) & 0x01) == 0)
				{
					goto stop_dma;
				}
			}
		}
stop_dma:
		tmpcip = dev->pciport;
		atp870u_outb(0x00, tmpcip);
		tmport -= 0x08;

		i = atp870u_inb(tmport);

		FUNC_VAR(i);
		
		tmport -= 0x02;
		target_id = atp870u_inb(tmport);
		tmport += 0x02;

		/*
		 *	Remap wide devices onto id numbers
		 */

		if ((target_id & 0x40) != 0) {
			target_id = (target_id & 0x07) | 0x08;
		} else {
			target_id &= 0x07;
		}

		if ((j & 0x40) != 0)
		{
		     if (dev->last_cmd == 0xff)
		     {
			dev->last_cmd = target_id;
		     }
		     dev->last_cmd |= 0x40;
		}

		if (i == 0x85)
		{
			if ((dev->last_cmd & 0xf0) != 0x40)
			{
			   dev->last_cmd = 0xff;
			}
			/*
			 *	Flip wide
			 */
			if (dev->wide_idu != 0)
			{
				tmport = workportu + 0x1b;
				atp870u_outb(0x01,tmport);
				while ((atp870u_inb(tmport) & 0x01) != 0x01)
				{
				   atp870u_outb(0x01,tmport);
				}
			}
			/*
			 *	Issue more commands
			 */
			if (((dev->quhdu != dev->quendu) || (dev->last_cmd != 0xff)) &&
			    (dev->in_snd == 0))
			{
				send_s870(h);
			}
			/*
			 *	Done
			 */
			dev->in_int = 0;
			FUNC_STR("		stop_dma	85");						
			return;
		}

		if (i == 0x40)
		{
		     dev->last_cmd |= 0x40;
		     dev->in_int = 0;
			FUNC_STR("		stop_dma	40");					 
		     return;
		}

		if (i == 0x21)
		{
			if ((dev->last_cmd & 0xf0) != 0x40)
			{
			   dev->last_cmd = 0xff;
			}
			tmport -= 0x05;
			adrcntu = 0;
			((unsigned char *) &adrcntu)[2] = atp870u_inb(tmport++);
			((unsigned char *) &adrcntu)[1] = atp870u_inb(tmport++);
			((unsigned char *) &adrcntu)[0] = atp870u_inb(tmport);
			k = dev->id[target_id].last_lenu;
			k -= adrcntu;
			dev->id[target_id].tran_lenu = k;
			dev->id[target_id].last_lenu = adrcntu;
			tmport -= 0x04;
			atp870u_outb(0x41, tmport);
			tmport += 0x08;
			atp870u_outb(0x08, tmport);
			dev->in_int = 0;
			FUNC_STR("		stop_dma	21");				
			return;
		}
		if ((i == 0x80) || (i == 0x8f))
		{
			lun = 0;
			tmport -= 0x07;
			j = atp870u_inb(tmport);
			FUNC_VAR(j);
			
			if (j == 0x44 || i==0x80) {
				tmport += 0x0d;
				lun = atp870u_inb(tmport) & 0x07;
			} else {
				if ((dev->last_cmd & 0xf0) != 0x40)
				{
				   dev->last_cmd = 0xff;
				}
				if (j == 0x41)
				{
					tmport += 0x02;
					adrcntu = 0;
					((unsigned char *) &adrcntu)[2] = atp870u_inb(tmport++);
					((unsigned char *) &adrcntu)[1] = atp870u_inb(tmport++);
					((unsigned char *) &adrcntu)[0] = atp870u_inb(tmport);
					k = dev->id[target_id].last_lenu;
					k -= adrcntu;
					dev->id[target_id].tran_lenu = k;
					dev->id[target_id].last_lenu = adrcntu;
					tmport += 0x04;
					atp870u_outb(0x08, tmport);
					dev->in_int = 0;
					FUNC_STR("		stop_dma	80 8f 41");									
					return;
				}
				else
				{
					atp870u_outb(0x46, tmport);
					dev->id[target_id].dirctu = 0x00;
					tmport += 0x02;
					atp870u_outb(0x00, tmport++);
					atp870u_outb(0x00, tmport++);
					atp870u_outb(0x00, tmport++);
					tmport += 0x03;
					atp870u_outb(0x08, tmport);
					dev->in_int = 0;
					FUNC_STR("		stop_dma	80 8f 41else");														
					return;
				}
			}

			if (dev->last_cmd != 0xff)
			{
			   dev->last_cmd |= 0x40;
			}
			tmport = workportu + 0x10;
			atp870u_outb(0x45, tmport);
			tmport += 0x06;
			target_id = atp870u_inb(tmport);
			
			/*
			 *	Remap wide identifiers
			 */
			if ((target_id & 0x10) != 0)
			{
				target_id = (target_id & 0x07) | 0x08;
			} else {
				target_id &= 0x07;
			}
			workrequ = dev->id[target_id].curr_req;
			tmport = workportu + 0x0f;
			atp870u_outb(lun, tmport);
			tmport += 0x02;
			atp870u_outb(dev->id[target_id].devspu, tmport++);
			adrcntu = dev->id[target_id].tran_lenu;
			k = dev->id[target_id].last_lenu;
			atp870u_outb(((unsigned char *) &k)[2], tmport++);
			atp870u_outb(((unsigned char *) &k)[1], tmport++);
			atp870u_outb(((unsigned char *) &k)[0], tmport++);
			/* Remap wide */
			j = target_id;
			if (target_id > 7) {
				j = (j & 0x07) | 0x40;
			}
			/* Add direction */
			j |= dev->id[target_id].dirctu;
			atp870u_outb(j, tmport++);
			atp870u_outb(0x80, tmport);

			/* enable 32 bit fifo transfer */
			if (dev->deviceid != 0x8081)
			{
			   tmport = workportu + 0x3a;
			   if ((dev->ata_cdbu[0] == 0x08) || (dev->ata_cdbu[0] == 0x28) ||
			       (dev->ata_cdbu[0] == 0x0a) || (dev->ata_cdbu[0] == 0x2a))
			   {
			      atp870u_outb((unsigned char)((atp870u_inb(tmport) & 0xf3) | 0x08),tmport);
			   }
			   else
			   {
			      atp870u_outb((unsigned char)(atp870u_inb(tmport) & 0xf3),tmport);
			   }
			}
			else
			{
			   tmport = workportu - 0x05;
			   if ((dev->ata_cdbu[0] == 0x08) || (dev->ata_cdbu[0] == 0x28) ||
			       (dev->ata_cdbu[0] == 0x0a) || (dev->ata_cdbu[0] == 0x2a))
			   {
			      atp870u_outb((unsigned char)((atp870u_inb(tmport) & 0x3f) | 0xc0),tmport);
			   }
			   else
			   {
			      atp870u_outb((unsigned char)(atp870u_inb(tmport) & 0x3f),tmport);
			   }
			}

			tmport = workportu + 0x1b;
			j = 0;
			id = 1;
			id = id << target_id;
			/*
			 *	Is this a wide device
			 */
			if ((id & dev->wide_idu) != 0) {
				j |= 0x01;
			}
			atp870u_outb(j, tmport);
			while ((atp870u_inb(tmport) & 0x01) != j)
			{
			   atp870u_outb(j,tmport);
			}

			if (dev->id[target_id].last_lenu == 0) {
				tmport = workportu + 0x18;
				atp870u_outb(0x08, tmport);
				dev->in_int = 0;
				FUNC_STR("		stop_dma	wide device");	
				return;
			}
			prd = dev->id[target_id].prd_posu;
			while (adrcntu != 0)
			{
				id = ((unsigned short int *) (prd))[2];
				if (id == 0) {
					k = 0x10000;
				} else {
					k = id;
				}
				if (k > adrcntu) {
					((unsigned short int *) (prd))[2] = (unsigned short int)
					    (k - adrcntu);
					((unsigned long *) (prd))[0] += adrcntu;
					adrcntu = 0;
					dev->id[target_id].prd_posu = prd;
				} else {
					adrcntu -= k;
					dev->id[target_id].prdaddru += 0x08;
					prd += 0x08;
					if (adrcntu == 0) {
						dev->id[target_id].prd_posu = prd;
					}
				}
			}
			tmpcip = dev->pciport + 0x04;
			
			dma_cache_wback_inv((unsigned long)dev->id[target_id].prd_tableu, PRD_SIZE);							
			atp870u_outl(dev->id[target_id].prdaddru, tmpcip);
			
			tmpcip -= 0x02;
			atp870u_outb(0x06, tmpcip);
			atp870u_outb(0x00, tmpcip);
			tmpcip -= 0x02;
			tmport = workportu + 0x18;
			/*
			 *	Check transfer direction
			 */
			if (dev->id[target_id].dirctu != 0) {
				atp870u_outb(0x08, tmport);
				atp870u_outb(0x01, tmpcip);
				dev->in_int = 0;
				FUNC_STR("		stop_dma	transfer direction");					
				return;
			}
			atp870u_outb(0x08, tmport);
			atp870u_outb(0x09, tmpcip);
			dev->in_int = 0;
			FUNC_STR("		stop_dma	transfer direction after");		
			return;
		}

		/*
		 *	Current scsi request on this target
		 */

		workrequ = dev->id[target_id].curr_req;

		if (i == 0x42) {
			if ((dev->last_cmd & 0xf0) != 0x40)
			{
			   dev->last_cmd = 0xff;
			}
			errstus = 0x02;
			workrequ->result = errstus;
			goto go_42;
		}
		if (i == 0x16)
		{
			if ((dev->last_cmd & 0xf0) != 0x40)
			{
			   dev->last_cmd = 0xff;
			}
			errstus = 0;
			tmport -= 0x08;
			errstus = atp870u_inb(tmport);
			workrequ->result = errstus;
go_42:
			/*
			 *	Complete the command
			 */
			spin_lock_irqsave(&io_request_lock, flags);
			(*workrequ->scsi_done) (workrequ);

			/*
			 *	Clear it off the queue
			 */
			dev->id[target_id].curr_req = 0;
			dev->working--;
			spin_unlock_irqrestore(&io_request_lock, flags);
			/*
			 *	Take it back wide
			 */
			if (dev->wide_idu != 0) {
				tmport = workportu + 0x1b;
				atp870u_outb(0x01,tmport);
				while ((atp870u_inb(tmport) & 0x01) != 0x01)
				{
				   atp870u_outb(0x01,tmport);
				}
			}
			/*
			 *	If there is stuff to send and nothing going then send it
			 */
			if (((dev->last_cmd != 0xff) || (dev->quhdu != dev->quendu)) &&
			    (dev->in_snd == 0))
			{
			   send_s870(h);
			}
			dev->in_int = 0;
			FUNC_STR("		go_42");					
			return;
		}
		if ((dev->last_cmd & 0xf0) != 0x40)
		{
		   dev->last_cmd = 0xff;
		}
		if (i == 0x4f) {
			i = 0x89;
		}
		i &= 0x0f;
		if (i == 0x09) {
			tmpcip = tmpcip + 4;
			
			dma_cache_wback_inv((unsigned long)dev->id[target_id].prd_tableu, PRD_SIZE);				
			atp870u_outl(dev->id[target_id].prdaddru, tmpcip);
			
			tmpcip = tmpcip - 2;
			atp870u_outb(0x06, tmpcip);
			atp870u_outb(0x00, tmpcip);
			tmpcip = tmpcip - 2;
			tmport = workportu + 0x10;
			atp870u_outb(0x41, tmport);
			dev->id[target_id].dirctu = 0x00;
			tmport += 0x08;
			atp870u_outb(0x08, tmport);
			atp870u_outb(0x09, tmpcip);
			dev->in_int = 0;
			FUNC_STR("		go_42	09");					
			return;
		}
		if (i == 0x08) {
			tmpcip = tmpcip + 4;
			dma_cache_wback_inv((unsigned long)dev->id[target_id].prd_tableu, PRD_SIZE);				
			atp870u_outl(dev->id[target_id].prdaddru, tmpcip);
			tmpcip = tmpcip - 2;
			atp870u_outb(0x06, tmpcip);
			atp870u_outb(0x00, tmpcip);
			tmpcip = tmpcip - 2;
			tmport = workportu + 0x10;
			atp870u_outb(0x41, tmport);
			tmport += 0x05;
			atp870u_outb((unsigned char) (atp870u_inb(tmport) | 0x20), tmport);
			dev->id[target_id].dirctu = 0x20;
			tmport += 0x03;
			atp870u_outb(0x08, tmport);
			atp870u_outb(0x01, tmpcip);
			dev->in_int = 0;
			FUNC_STR("		go_42	08");								
			return;
		}
		tmport -= 0x07;
		if (i == 0x0a) {
			atp870u_outb(0x30, tmport);
		} else {
			atp870u_outb(0x46, tmport);
		}
		dev->id[target_id].dirctu = 0x00;
		tmport += 0x02;
		atp870u_outb(0x00, tmport++);
		atp870u_outb(0x00, tmport++);
		atp870u_outb(0x00, tmport++);
		tmport += 0x03;
		atp870u_outb(0x08, tmport);
		dev->in_int = 0;
		FUNC_STR("		dev");										
		return;
	} else {
//		tmport = workportu + 0x17;
//		atp870u_inb(tmport);
//		dev->working = 0;
		dev->in_int = 0;
		FUNC_STR("		dev 0");										
		return;
	}
}

int atp870u_queuecommand(Scsi_Cmnd * req_p, void (*done) (Scsi_Cmnd *))
{
	unsigned char h;
	unsigned long flags;
	unsigned short int m;
	unsigned int tmport;
	struct atp_unit *dev;

	FUNC_IN();

	for (h = 0; h <= admaxu; h++) {
		if (req_p->host == atp_host[h]) {
			goto host_ok;
		}
	}
	FUNC_STR("		admaxu");		
	return 0;
host_ok:
	if (req_p->channel != 0) {
		req_p->result = 0x00040000;
		done(req_p);
		FUNC_STR("		host_ok");		
		
		return 0;
	}
	dev = &atp_unit[h];
	m = 1;
	m = m << req_p->target;

	/*
	 *	Fake a timeout for missing targets
	 */

	if ((m & dev->active_idu) == 0) {
		req_p->result = 0x00040000;
		done(req_p);
		FUNC_STR("		timeout");				
		return 0;
	}
	if (done) {
		req_p->scsi_done = done;
	} else {
		printk(KERN_WARNING ": done can't be NULL\n");
		req_p->result = 0;
		done(req_p);
		FUNC_STR("		not done");		
		return 0;
	}
	/*
	 *	Count new command
	 */
	save_flags(flags);
	cli();
	dev->quendu++;
	if (dev->quendu >= qcnt) {
		dev->quendu = 0;
	}
	/*
	 *	Check queue state
	 */
	if (dev->quhdu == dev->quendu) {
		if (dev->quendu == 0) {
			dev->quendu = qcnt;
		}
		dev->quendu--;
		req_p->result = 0x00020000;
		done(req_p);
		restore_flags(flags);
		FUNC_STR("		queue");			
		return 0;
	}
	dev->querequ[dev->quendu] = req_p;
	tmport = dev->ioport + 0x1c;
	restore_flags(flags);
	if ((atp870u_inb(tmport) == 0) && (dev->in_int == 0) && (dev->in_snd == 0)) {
		send_s870(h);
	}
	
	FUNC_STR("		Leave");			
	
	return 0;
}

void mydlyu(unsigned int dlycnt)
{
	unsigned int i;


	for (i = 0; i < dlycnt; i++) {
		udelay(5);
		//atp870u_inb(0x80);
	}

	
}

void send_s870(unsigned char h)
{
	unsigned int tmport;
	Scsi_Cmnd *workrequ;
	unsigned long flags;
	unsigned int i;
	unsigned char j, target_id;
	unsigned char *prd;
//	unsigned short int tmpcip, w;
	unsigned int tmpcip, w;
	unsigned long l, bttl;
	unsigned int workportu;
	struct scatterlist *sgpnt;
	struct atp_unit *dev = &atp_unit[h];

	FUNC_IN();
	FUNC_VAR(h);

	save_flags(flags);
	cli();
	if (dev->in_snd != 0) {
		restore_flags(flags);
		FUNC_STR("		in_snd");			
		return;
	}
	dev->in_snd = 1;

	FUNC_STR("		last_cmd	0");			
	if ((dev->last_cmd != 0xff) && ((dev->last_cmd & 0x40) != 0)) {
		dev->last_cmd &= 0x0f;
		workrequ = dev->id[dev->last_cmd].curr_req;
		if (workrequ != NULL)	     /* check NULL pointer */
		{
		   goto cmd_subp;
		}
		dev->last_cmd = 0xff;
		if (dev->quhdu == dev->quendu)
		{
		   dev->in_snd = 0;
		   restore_flags(flags);
			FUNC_STR("		quhdu");					   
		   return ;
		}
	}

	FUNC_STR("		last_cmd	1");				
	if ((dev->last_cmd != 0xff) && (dev->working != 0))
	{
	     dev->in_snd = 0;
	     restore_flags(flags);
		FUNC_STR("		last_cmd");		   
	     return ;
	}
	dev->working++;
	j = dev->quhdu;
	dev->quhdu++;
	if (dev->quhdu >= qcnt) {
		dev->quhdu = 0;
	}

	FUNC_STR("		querequ");					
	workrequ = dev->querequ[dev->quhdu];
	if (dev->id[workrequ->target].curr_req == 0) {
		dev->id[workrequ->target].curr_req = workrequ;
		dev->last_cmd = workrequ->target;
		goto cmd_subp;
	}
	dev->quhdu = j;
	dev->working--;
	dev->in_snd = 0;
	restore_flags(flags);
	FUNC_STR("		restore_flags");		   
	
	return;
cmd_subp:
	FUNC_STR("		cmd_subp	0");					
	
	workportu = dev->ioport;
	tmport = workportu + 0x1f;
	if ((atp870u_inb(tmport) & 0xb0) != 0) {
		goto abortsnd;
	}
	FUNC_STR("		cmd_subp	1");					
	
	tmport = workportu + 0x1c;
	if (atp870u_inb(tmport) == 0) {
		goto oktosend;
	}
abortsnd:
	
	dev->last_cmd |= 0x40;
	dev->in_snd = 0;
	restore_flags(flags);
	FUNC_STR("		abortsnd");		   
	
	return;
oktosend:
	FUNC_STR("		oktosend");					
	
	memcpy(&dev->ata_cdbu[0], &workrequ->cmnd[0], workrequ->cmd_len);
	if (dev->ata_cdbu[0] == READ_CAPACITY) {
		if (workrequ->request_bufflen > 8) {
			workrequ->request_bufflen = 0x08;
		}
	}
	if (dev->ata_cdbu[0] == 0x00) {
		workrequ->request_bufflen = 0;
	}

	tmport = workportu + 0x1b;
	j = 0;
	target_id = workrequ->target;

	/*
	 *	Wide ?
	 */
	FUNC_STR("		Wide ?");					
	w = 1;
	w = w << target_id;
	if ((w & dev->wide_idu) != 0) {
		j |= 0x01;
	}
	atp870u_outb(j, tmport);
	while ((atp870u_inb(tmport) & 0x01) != j)
	{
	   atp870u_outb(j,tmport);
	}

	/*
	 *	Write the command
	 */
	FUNC_STR("		Write the command");					

	tmport = workportu;
	atp870u_outb(workrequ->cmd_len, tmport++);
	atp870u_outb(0x2c, tmport++);
	atp870u_outb(0xcf, tmport++);
	for (i = 0; i < workrequ->cmd_len; i++) {
		atp870u_outb(dev->ata_cdbu[i], tmport++);
	}
	tmport = workportu + 0x0f;
	atp870u_outb(workrequ->lun, tmport);
	tmport += 0x02;
	/*
	 *	Write the target
	 */
	atp870u_outb(dev->id[target_id].devspu, tmport++);

	/*
	 *	Figure out the transfer size
	 */
	FUNC_STR("		Figure out the transfer size");					
	 
	if (workrequ->use_sg)
	{
		l = 0;
		sgpnt = (struct scatterlist *) workrequ->request_buffer;
		for (i = 0; i < workrequ->use_sg; i++)
		{
			if (sgpnt[i].length == 0 || workrequ->use_sg > ATP870U_SCATTER)
			{
				panic("Foooooooood fight!");
			}
			l += sgpnt[i].length;
		}
	} else {
		l = workrequ->request_bufflen;
	}
	/*
	 *	Write transfer size
	 */
	FUNC_STR("		Write transfer size");					
	 
	atp870u_outb((unsigned char) (((unsigned char *) (&l))[2]), tmport++);
	atp870u_outb((unsigned char) (((unsigned char *) (&l))[1]), tmport++);
	atp870u_outb((unsigned char) (((unsigned char *) (&l))[0]), tmport++);
	j = target_id;
	dev->id[j].last_lenu = l;
	dev->id[j].tran_lenu = 0;
	/*
	 *	Flip the wide bits
	 */
	if ((j & 0x08) != 0) {
		j = (j & 0x07) | 0x40;
	}
	/*
	 *	Check transfer direction
	 */
	FUNC_STR("		Check transfer direction");					
	 
	if ((dev->ata_cdbu[0] == WRITE_6) || (dev->ata_cdbu[0] == WRITE_10) ||
	    (dev->ata_cdbu[0] == WRITE_12) || (dev->ata_cdbu[0] == MODE_SELECT)) {
		atp870u_outb((unsigned char) (j | 0x20), tmport++);
	} else {
		atp870u_outb(j, tmport++);
	}
	atp870u_outb((unsigned char)(atp870u_inb(tmport) | 0x80),tmport);
	atp870u_outb(0x80, tmport);
	tmport = workportu + 0x1c;
	dev->id[target_id].dirctu = 0;
	if (l == 0) {
		if (atp870u_inb(tmport) == 0) {
			tmport = workportu + 0x18;
			atp870u_outb(0x08, tmport);
		} else {
			dev->last_cmd |= 0x40;
		}
		dev->in_snd = 0;
		restore_flags(flags);
		FUNC_STR("		l != 0");		   
		
		return;
	}
	tmpcip = dev->pciport;
	FUNC_STR("dev->pciport	tmpcip:");
	FUNC_VAR(dev->pciport);
	FUNC_VAR(tmpcip);
	
	prd = dev->id[target_id].prd_tableu;
	dev->id[target_id].prd_posu = prd;

	/*
	 *	Now write the request list. Either as scatter/gather or as
	 *	a linear chain.
	 */
	FUNC_STR("		write the request list");					

	if (workrequ->use_sg)
	{

		sgpnt = (struct scatterlist *) workrequ->request_buffer;
		i = 0;
		for (j = 0; j < workrequ->use_sg; j++) {
			(unsigned long) (((unsigned long *) (prd))[i >> 1]) = virt_to_bus(sgpnt[j].address);
			(unsigned short int) (((unsigned short int *) (prd))[i + 2]) = sgpnt[j].length;
			(unsigned short int) (((unsigned short int *) (prd))[i + 3]) = 0;
			i += 0x04;
		}
		
		(unsigned short int) (((unsigned short int *) (prd))[i - 1]) = 0x8000;
		
	} else {
		/*
		 *	For a linear request write a chain of blocks
		 */
		 
		bttl = virt_to_bus(workrequ->request_buffer);
		
		l = workrequ->request_bufflen;
		i = 0;
		while (l > 0x10000) {
			(unsigned short int) (((unsigned short int *) (prd))[i + 3]) = 0x0000;
			(unsigned short int) (((unsigned short int *) (prd))[i + 2]) = 0x0000;
			(unsigned long) (((unsigned long *) (prd))[i >> 1]) = bttl;
			l -= 0x10000;
			bttl += 0x10000;
			i += 0x04;
		}
		
		(unsigned short int) (((unsigned short int *) (prd))[i + 3]) = 0x8000;
		(unsigned short int) (((unsigned short int *) (prd))[i + 2]) = l;
		(unsigned long) (((unsigned long *) (prd))[i >> 1]) = bttl;
		
	}
	tmpcip = tmpcip + 4;

	FUNC_STR("		prd_tableu	tmpcip:");					
	FUNC_VAR(tmpcip);					

	dev->id[target_id].prdaddru = virt_to_bus(dev->id[target_id].prd_tableu);
	dma_cache_wback_inv((unsigned long)dev->id[target_id].prd_tableu, PRD_SIZE);	
	
	atp870u_outl(dev->id[target_id].prdaddru, tmpcip);
	tmpcip = tmpcip - 2;
	atp870u_outb(0x06, tmpcip);
	atp870u_outb(0x00, tmpcip);
	tmpcip = tmpcip - 2;

	FUNC_STR("		dev->deviceid ? 0x8081");					
	FUNC_VAR(dev->deviceid);					

	if (dev->deviceid != 0x8081)
	{
	   tmport = workportu + 0x3a;
	   if ((dev->ata_cdbu[0] == 0x08) || (dev->ata_cdbu[0] == 0x28) ||
	       (dev->ata_cdbu[0] == 0x0a) || (dev->ata_cdbu[0] == 0x2a))
	   {
	      atp870u_outb((unsigned char)((atp870u_inb(tmport) & 0xf3) | 0x08),tmport);
	   }
	   else
	   {
	      atp870u_outb((unsigned char)(atp870u_inb(tmport) & 0xf3),tmport);
	   }
	}
	else
	{
	   tmport = workportu - 0x05;
	   if ((dev->ata_cdbu[0] == 0x08) || (dev->ata_cdbu[0] == 0x28) ||
	       (dev->ata_cdbu[0] == 0x0a) || (dev->ata_cdbu[0] == 0x2a))
	   {
	      atp870u_outb((unsigned char)((atp870u_inb(tmport) & 0x3f) | 0xc0),tmport);
	   }
	   else
	   {
	      atp870u_outb((unsigned char)(atp870u_inb(tmport) & 0x3f),tmport);
	   }
	}
	tmport = workportu + 0x1c;

	FUNC_STR("		WRITE_6_10_12");					

	if ((dev->ata_cdbu[0] == WRITE_6) || (dev->ata_cdbu[0] == WRITE_10) ||
	    (dev->ata_cdbu[0] == WRITE_12) || (dev->ata_cdbu[0] == MODE_SELECT))
	{
		dev->id[target_id].dirctu = 0x20;
		if (atp870u_inb(tmport) == 0) {
			tmport = workportu + 0x18;
			atp870u_outb(0x08, tmport);
			atp870u_outb(0x01, tmpcip);
		} else {
			dev->last_cmd |= 0x40;
		}
		dev->in_snd = 0;
		restore_flags(flags);
		FUNC_STR("		ata_cdbu");		   
		return;
	}
	if (atp870u_inb(tmport) == 0)
	{
		tmport = workportu + 0x18;
		atp870u_outb(0x08, tmport);
		atp870u_outb(0x09, tmpcip);
	} else {
		dev->last_cmd |= 0x40;
	}
	dev->in_snd = 0;
	restore_flags(flags);
	FUNC_STR("		Leave");		   

	return;

}

static void internal_done(Scsi_Cmnd * SCpnt)
{
	FUNC_IN();
	SCpnt->SCp.Status++;
	FUNC_OUT();

}

int atp870u_command(Scsi_Cmnd * SCpnt)
{
	FUNC_IN();

	atp870u_queuecommand(SCpnt, internal_done);

	SCpnt->SCp.Status = 0;
	while (!SCpnt->SCp.Status)
		barrier();

	FUNC_OUT();		   
	
	return SCpnt->result;
}

unsigned char fun_scam(struct atp_unit *dev, unsigned short int *val)
{
	unsigned int tmport;
	unsigned short int i, k;
	unsigned char j;

	FUNC_IN();

	tmport = dev->ioport + 0x1c;
	atp870u_outw(*val, tmport);
FUN_D7:
	for (i = 0; i < 10; i++) {	/* stable >= bus settle delay(400 ns)  */
		k = atp870u_inw(tmport);
		j = (unsigned char) (k >> 8);
		if ((k & 0x8000) != 0) {	/* DB7 all release?    */
			goto FUN_D7;
		}
	}
	*val |= 0x4000; 	/* assert DB6		*/
	atp870u_outw(*val, tmport);
	*val &= 0xdfff; 	/* assert DB5		*/
	atp870u_outw(*val, tmport);
FUN_D5:
	for (i = 0; i < 10; i++) {	/* stable >= bus settle delay(400 ns) */
		if ((atp870u_inw(tmport) & 0x2000) != 0) {	/* DB5 all release?	  */
			goto FUN_D5;
		}
	}
	*val |= 0x8000; 	/* no DB4-0, assert DB7    */
	*val &= 0xe0ff;
	atp870u_outw(*val, tmport);
	*val &= 0xbfff; 	/* release DB6		   */
	atp870u_outw(*val, tmport);
      FUN_D6:
	for (i = 0; i < 10; i++) {	/* stable >= bus settle delay(400 ns)  */
		if ((atp870u_inw(tmport) & 0x4000) != 0) {	/* DB6 all release?  */
			goto FUN_D6;
		}
	}
	FUNC_OUT();		   

	return j;
}

void tscam(unsigned char host)
{

	unsigned int tmport;
	unsigned char i, j, k;
	unsigned long n;
	unsigned short int m, assignid_map, val;
	unsigned char mbuf[33], quintet[2];
	struct atp_unit *dev = &atp_unit[host];
	static unsigned char g2q_tab[8] = {
		0x38, 0x31, 0x32, 0x2b, 0x34, 0x2d, 0x2e, 0x27
	};

	FUNC_IN();

	for (i = 0; i < 0x10; i++) {
		mydlyu(0xffff);
	}

	tmport = dev->ioport + 1;
	atp870u_outb(0x08, tmport++);
	atp870u_outb(0x7f, tmport);
	tmport = dev->ioport + 0x11;
	atp870u_outb(0x20, tmport);

	if ((dev->scam_on & 0x40) == 0) {
		FUNC_STR("		scam_on");		   		
		return;
	}
	m = 1;
	m <<= dev->host_idu;
	j = 16;
	if (dev->chip_veru < 4) {
		m |= 0xff00;
		j = 8;
	}
	assignid_map = m;
	tmport = dev->ioport + 0x02;
	atp870u_outb(0x02, tmport++);	/* 2*2=4ms,3EH 2/32*3E=3.9ms */
	atp870u_outb(0, tmport++);
	atp870u_outb(0, tmport++);
	atp870u_outb(0, tmport++);
	atp870u_outb(0, tmport++);
	atp870u_outb(0, tmport++);
	atp870u_outb(0, tmport++);

	for (i = 0; i < j; i++) {
		m = 1;
		m = m << i;
		if ((m & assignid_map) != 0) {
			continue;
		}
		tmport = dev->ioport + 0x0f;
		atp870u_outb(0, tmport++);
		tmport += 0x02;
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		if (i > 7) {
			k = (i & 0x07) | 0x40;
		} else {
			k = i;
		}
		atp870u_outb(k, tmport++);
		tmport = dev->ioport + 0x1b;
		if (dev->chip_veru == 4) {
			atp870u_outb(0x01, tmport);
		} else {
			atp870u_outb(0x00, tmport);
		}
wait_rdyok:
		tmport = dev->ioport + 0x18;
		atp870u_outb(0x09, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		k = atp870u_inb(tmport);
		if (k != 0x16) {
			if ((k == 0x85) || (k == 0x42)) {
				continue;
			}
			tmport = dev->ioport + 0x10;
			atp870u_outb(0x41, tmport);
			goto wait_rdyok;
		}
		assignid_map |= m;

	}
	tmport = dev->ioport + 0x02;
	atp870u_outb(0x7f, tmport);
	tmport = dev->ioport + 0x1b;
	atp870u_outb(0x02, tmport);

	atp870u_outb(0, 0x80);

	val = 0x0080;		/* bsy	*/
	tmport = dev->ioport + 0x1c;
	atp870u_outw(val, tmport);
	val |= 0x0040;		/* sel	*/
	atp870u_outw(val, tmport);
	val |= 0x0004;		/* msg	*/
	atp870u_outw(val, tmport);
	//atp870u_inb(0x80);		/* 2 deskew delay(45ns*2=90ns) */
	udelay(10000);
	val &= 0x007f;		/* no bsy  */
	atp870u_outw(val, tmport);
	mydlyu(0xffff); 	/* recommanded SCAM selection response time */
	mydlyu(0xffff);
	val &= 0x00fb;		/* after 1ms no msg */
	atp870u_outw(val, tmport);
wait_nomsg:
	if ((atp870u_inb(tmport) & 0x04) != 0) {
		goto wait_nomsg;
	}
	atp870u_outb(1, 0x80);
	mydlyu(100);
	for (n = 0; n < 0x30000; n++) {
		if ((atp870u_inb(tmport) & 0x80) != 0) {	/* bsy ? */
			goto wait_io;
		}
	}
	goto TCM_SYNC;
wait_io:
	for (n = 0; n < 0x30000; n++) {
		if ((atp870u_inb(tmport) & 0x81) == 0x0081) {
			goto wait_io1;
		}
	}
	goto TCM_SYNC;
wait_io1:
	//atp870u_inb(0x80);
	udelay(10000);
	
	val |= 0x8003;		/* io,cd,db7  */
	atp870u_outw(val, tmport);
	//atp870u_inb(0x80);
	udelay(10000);
	
	val &= 0x00bf;		/* no sel     */
	atp870u_outw(val, tmport);
	atp870u_outb(2, 0x80);
TCM_SYNC:
	mydlyu(0x800);
	if ((atp870u_inb(tmport) & 0x80) == 0x00) {	/* bsy ? */
		atp870u_outw(0, tmport--);
		atp870u_outb(0, tmport);
		tmport = dev->ioport + 0x15;
		atp870u_outb(0, tmport);
		tmport += 0x03;
		atp870u_outb(0x09, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0);
		tmport -= 0x08;
		atp870u_inb(tmport);
		FUNC_STR("		TCM_SYNC");		   		
		
		return;
	}
	val &= 0x00ff;		/* synchronization  */
	val |= 0x3f00;
	fun_scam(dev, &val);
	atp870u_outb(3, 0x80);
	val &= 0x00ff;		/* isolation	    */
	val |= 0x2000;
	fun_scam(dev, &val);
	atp870u_outb(4, 0x80);
	i = 8;
	j = 0;
TCM_ID:
	if ((atp870u_inw(tmport) & 0x2000) == 0) {
		goto TCM_ID;
	}
	atp870u_outb(5, 0x80);
	val &= 0x00ff;		/* get ID_STRING */
	val |= 0x2000;
	k = fun_scam(dev, &val);
	if ((k & 0x03) == 0) {
		goto TCM_5;
	}
	mbuf[j] <<= 0x01;
	mbuf[j] &= 0xfe;
	if ((k & 0x02) != 0) {
		mbuf[j] |= 0x01;
	}
	i--;
	if (i > 0) {
		goto TCM_ID;
	}
	j++;
	i = 8;
	goto TCM_ID;

TCM_5:			/* isolation complete..  */
/*    mbuf[32]=0;
	printk(" \n%x %x %x %s\n ",assignid_map,mbuf[0],mbuf[1],&mbuf[2]); */
	i = 15;
	j = mbuf[0];
	if ((j & 0x20) != 0) {	/* bit5=1:ID upto 7	 */
		i = 7;
	}
	if ((j & 0x06) == 0) {	/* IDvalid?		*/
		goto G2Q5;
	}
	k = mbuf[1];
small_id:
	m = 1;
	m <<= k;
	if ((m & assignid_map) == 0) {
		goto G2Q_QUIN;
	}
	if (k > 0) {
		k--;
		goto small_id;
	}
G2Q5:				/* srch from max acceptable ID#  */
	k = i;			/* max acceptable ID#		 */
G2Q_LP:
	m = 1;
	m <<= k;
	if ((m & assignid_map) == 0) {
		goto G2Q_QUIN;
	}
	if (k > 0) {
		k--;
		goto G2Q_LP;
	}
G2Q_QUIN:		/* k=binID#,	   */
	assignid_map |= m;
	if (k < 8) {
		quintet[0] = 0x38;	/* 1st dft ID<8    */
	} else {
		quintet[0] = 0x31;	/* 1st	ID>=8	   */
	}
	k &= 0x07;
	quintet[1] = g2q_tab[k];

	val &= 0x00ff;		/* AssignID 1stQuintet,AH=001xxxxx  */
	m = quintet[0] << 8;
	val |= m;
	fun_scam(dev, &val);
	val &= 0x00ff;		/* AssignID 2ndQuintet,AH=001xxxxx */
	m = quintet[1] << 8;
	val |= m;
	fun_scam(dev, &val);

	goto TCM_SYNC;

}

void is870(unsigned long host, unsigned int wkport)
{
	unsigned int tmport;
	unsigned char i, j, k, rmb, n;
	unsigned short int m;
	static unsigned char mbuf[512];
	static unsigned char satn[9] =	{0, 0, 0, 0, 0, 0, 0, 6, 6};
	static unsigned char inqd[9] =	{0x12, 0, 0, 0, 0x24, 0, 0, 0x24, 6};
	static unsigned char synn[6] =	{0x80, 1, 3, 1, 0x19, 0x0e};
	static unsigned char synu[6] =	{0x80, 1, 3, 1, 0x0c, 0x0e};
	static unsigned char synw[6] =	{0x80, 1, 3, 1, 0x0c, 0x07};
	static unsigned char wide[6] =	{0x80, 1, 2, 3, 1, 0};
	struct atp_unit *dev = &atp_unit[host];

	FUNC_IN();

	sync_idu = 0;
	tmport = wkport + 0x3a;
	FUNC_VAR(wkport);
	atp870u_outb((unsigned char) (atp870u_inb(tmport) | 0x10), tmport);

	for (i = 0; i < 16; i++) {
		if ((dev->chip_veru != 4) && (i > 7)) {
			break;
		}
		m = 1;
		m = m << i;
		if ((m & dev->active_idu) != 0) {
			continue;
		}
		if (i == dev->host_idu) {
			printk(KERN_INFO "         ID-870: %2d  Host Adapter\n", dev->host_idu);
			continue;
		}
		tmport = wkport + 0x1b;
		if (dev->chip_veru == 4) {
		   atp870u_outb(0x01, tmport);
		}
		else
		{
		   atp870u_outb(0x00, tmport);
		}
		tmport = wkport + 1;
		atp870u_outb(0x08, tmport++);
		atp870u_outb(0x7f, tmport++);
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		j = i;
		if ((j & 0x08) != 0) {
			j = (j & 0x07) | 0x40;
		}
		atp870u_outb(j, tmport);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
		dev->active_idu |= m;

		tmport = wkport + 0x10;
		atp870u_outb(0x30, tmport);
		tmport = wkport + 0x04;
		atp870u_outb(0x00, tmport);

phase_cmd:
		tmport = wkport + 0x18;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			tmport = wkport + 0x10;
			atp870u_outb(0x41, tmport);
			goto phase_cmd;
		}
sel_ok:
		tmport = wkport + 3;
		atp870u_outb(inqd[0], tmport++);
		atp870u_outb(inqd[1], tmport++);
		atp870u_outb(inqd[2], tmport++);
		atp870u_outb(inqd[3], tmport++);
		atp870u_outb(inqd[4], tmport++);
		atp870u_outb(inqd[5], tmport);
		tmport += 0x07;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(inqd[6], tmport++);
		atp870u_outb(inqd[7], tmport++);
		tmport += 0x03;
		atp870u_outb(inqd[8], tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
		tmport = wkport + 0x1b;
		if (dev->chip_veru == 4) {
			atp870u_outb(0x00, tmport);
		}
		tmport = wkport + 0x18;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		j = 0;
rd_inq_data:
		k = atp870u_inb(tmport);
		if ((k & 0x01) != 0) {
			tmport -= 0x06;
			mbuf[j++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto rd_inq_data;
		}
		if ((k & 0x80) == 0) {
			goto rd_inq_data;
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j == 0x16) {
			goto inq_ok;
		}
		tmport = wkport + 0x10;
		atp870u_outb(0x46, tmport);
		tmport += 0x02;
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		tmport += 0x03;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if (atp870u_inb(tmport) != 0x16) {
			goto sel_ok;
		}
inq_ok:
		mbuf[36] = 0;
		printk(KERN_INFO "         ID-870: %2d  %s\n", i, &mbuf[8]);
		dev->id[i].devtypeu = mbuf[0];
		rmb = mbuf[1];
		n = mbuf[7];
		if (dev->chip_veru != 4) {
			goto not_wide;
		}
		if ((mbuf[7] & 0x60) == 0) {
			goto not_wide;
		}
		if ((dev->global_map & 0x20) == 0) {
			goto not_wide;
		}
		tmport = wkport + 0x1b;
		atp870u_outb(0x01, tmport);
		tmport = wkport + 3;
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
try_wide:
		j = 0;
		tmport = wkport + 0x14;
		atp870u_outb(0x05, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(wide[j++], tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto try_wide;
		}
		continue;
widep_out:
		tmport = wkport + 0x18;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(0, tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto widep_out;
		}
		continue;
widep_in:
		tmport = wkport + 0x14;
		atp870u_outb(0xff, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		k = 0;
widep_in1:
		j = atp870u_inb(tmport);
		if ((j & 0x01) != 0) {
			tmport -= 0x06;
			mbuf[k++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto widep_in1;
		}
		if ((j & 0x80) == 0x00) {
			goto widep_in1;
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto widep_out;
		}
		continue;
widep_cmd:
		tmport = wkport + 0x10;
		atp870u_outb(0x30, tmport);
		tmport = wkport + 0x14;
		atp870u_outb(0x00, tmport);
		tmport += 0x04;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			if (j == 0x4e) {
				goto widep_out;
			}
			continue;
		}
		if (mbuf[0] != 0x01) {
			goto not_wide;
		}
		if (mbuf[1] != 0x02) {
			goto not_wide;
		}
		if (mbuf[2] != 0x03) {
			goto not_wide;
		}
		if (mbuf[3] != 0x01) {
			goto not_wide;
		}
		m = 1;
		m = m << i;
		dev->wide_idu |= m;
not_wide:
		if ((dev->id[i].devtypeu == 0x00) || (dev->id[i].devtypeu == 0x07) ||
		    ((dev->id[i].devtypeu == 0x05) && ((n & 0x10) != 0)))
		{
			goto set_sync;
		}
		continue;
set_sync:
		tmport = wkport + 0x1b;
		j = 0;
		if ((m & dev->wide_idu) != 0) {
			j |= 0x01;
		}
		atp870u_outb(j, tmport);
		tmport = wkport + 3;
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
try_sync:
		j = 0;
		tmport = wkport + 0x14;
		atp870u_outb(0x06, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				if ((m & dev->wide_idu) != 0) {
					atp870u_outb(synw[j++], tmport);
				} else {
					if ((m & dev->ultra_map) != 0) {
						atp870u_outb(synu[j++], tmport);
					} else {
						atp870u_outb(synn[j++], tmport);
					}
				}
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto try_sync;
		}
		continue;
phase_outs:
		tmport = wkport + 0x18;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00) {
			if ((atp870u_inb(tmport) & 0x01) != 0x00) {
				tmport -= 0x06;
				atp870u_outb(0x00, tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j == 0x85) {
			goto tar_dcons;
		}
		j &= 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto phase_outs;
		}
		continue;
phase_ins:
		tmport = wkport + 0x14;
		atp870u_outb(0xff, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		k = 0;
phase_ins1:
		j = atp870u_inb(tmport);
		if ((j & 0x01) != 0x00) {
			tmport -= 0x06;
			mbuf[k++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto phase_ins1;
		}
		if ((j & 0x80) == 0x00) {
			goto phase_ins1;
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport);
		if (j == 0x85) {
			goto tar_dcons;
		}
		j &= 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto phase_outs;
		}
		continue;
phase_cmds:
		tmport = wkport + 0x10;
		atp870u_outb(0x30, tmport);
tar_dcons:
		tmport = wkport + 0x14;
		atp870u_outb(0x00, tmport);
		tmport += 0x04;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			continue;
		}
		if (mbuf[0] != 0x01) {
			continue;
		}
		if (mbuf[1] != 0x03) {
			continue;
		}
		if (mbuf[4] == 0x00) {
			continue;
		}
		if (mbuf[3] > 0x64) {
			continue;
		}
		if (mbuf[4] > 0x0c) {
			mbuf[4] = 0x0c;
		}
		dev->id[i].devspu = mbuf[4];
		if ((mbuf[3] < 0x0d) && (rmb == 0)) {
			j = 0xa0;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x1a) {
			j = 0x20;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x33) {
			j = 0x40;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x4c) {
			j = 0x50;
			goto set_syn_ok;
		}
		j = 0x60;
	      set_syn_ok:
		dev->id[i].devspu = (dev->id[i].devspu & 0x0f) | j;
	}
	tmport = wkport + 0x3a;
	atp870u_outb((unsigned char) (atp870u_inb(tmport) & 0xef), tmport);
}

void is880(unsigned long host, unsigned int wkport)
{
	unsigned int tmport;
	unsigned char i, j, k, rmb, n, lvdmode;
	unsigned short int m;
	static unsigned char mbuf[512];
	static unsigned char satn[9] =	{0, 0, 0, 0, 0, 0, 0, 6, 6};
	static unsigned char inqd[9] =	{0x12, 0, 0, 0, 0x24, 0, 0, 0x24, 6};
	static unsigned char synn[6] =	{0x80, 1, 3, 1, 0x19, 0x0e};
	unsigned char synu[6] =  {0x80, 1, 3, 1, 0x0a, 0x0e};
	static unsigned char synw[6] =	{0x80, 1, 3, 1, 0x19, 0x0e};
	unsigned char synuw[6] =  {0x80, 1, 3, 1, 0x0a, 0x0e};
	static unsigned char wide[6] =	{0x80, 1, 2, 3, 1, 0};
	static unsigned char u3[9] = { 0x80,1,6,4,0x09,00,0x0e,0x01,0x02 };
	struct atp_unit *dev = &atp_unit[host];

	FUNC_IN();

	sync_idu = 0;
	lvdmode=atp870u_inb(wkport + 0x3f) & 0x40;

	for (i = 0; i < 16; i++) {
		m = 1;
		m = m << i;
		if ((m & dev->active_idu) != 0) {
			continue;
		}
		if (i == dev->host_idu) {
			printk(KERN_INFO "         ID-880: %2d  Host Adapter\n", dev->host_idu);
			continue;
		}
		tmport = wkport + 0x5b;
		atp870u_outb(0x01, tmport);
		tmport = wkport + 0x41;
		atp870u_outb(0x08, tmport++);
		atp870u_outb(0x7f, tmport++);
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		j = i;
		if ((j & 0x08) != 0) {
			j = (j & 0x07) | 0x40;
		}
		atp870u_outb(j, tmport);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
		dev->active_idu |= m;

		tmport = wkport + 0x50;
		atp870u_outb(0x30, tmport);
		tmport = wkport + 0x54;
		atp870u_outb(0x00, tmport);

phase_cmd:
		tmport = wkport + 0x58;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			tmport = wkport + 0x50;
			atp870u_outb(0x41, tmport);
			goto phase_cmd;
		}
sel_ok:
		tmport = wkport + 0x43;
		atp870u_outb(inqd[0], tmport++);
		atp870u_outb(inqd[1], tmport++);
		atp870u_outb(inqd[2], tmport++);
		atp870u_outb(inqd[3], tmport++);
		atp870u_outb(inqd[4], tmport++);
		atp870u_outb(inqd[5], tmport);
		tmport += 0x07;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(inqd[6], tmport++);
		atp870u_outb(inqd[7], tmport++);
		tmport += 0x03;
		atp870u_outb(inqd[8], tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
		tmport = wkport + 0x5b;
		atp870u_outb(0x00, tmport);
		tmport = wkport + 0x58;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		j = 0;
rd_inq_data:
		k = atp870u_inb(tmport);
		if ((k & 0x01) != 0) {
			tmport -= 0x06;
			mbuf[j++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto rd_inq_data;
		}
		if ((k & 0x80) == 0) {
			goto rd_inq_data;
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j == 0x16) {
			goto inq_ok;
		}
		tmport = wkport + 0x50;
		atp870u_outb(0x46, tmport);
		tmport += 0x02;
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(0, tmport++);
		tmport += 0x03;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if (atp870u_inb(tmport) != 0x16) {
			goto sel_ok;
		}
inq_ok:
		mbuf[36] = 0;
		printk(KERN_INFO "         ID-880: %2d  %s\n", i, &mbuf[8]);
		dev->id[i].devtypeu = mbuf[0];
		rmb = mbuf[1];
		n = mbuf[7];
		if ((mbuf[7] & 0x60) == 0) {
			goto not_wide;
		}
		if ((i < 8) && ((dev->global_map & 0x20) == 0)) {
			goto not_wide;
		}
		if (lvdmode == 0)
		{
		   goto chg_wide;
		}
		if (dev->sp[i] != 0x04) 	 // force u2
		{
		   goto chg_wide;
		}

		tmport = wkport + 0x5b;
		atp870u_outb(0x01, tmport);
		tmport = wkport + 0x43;
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
try_u3:
		j = 0;
		tmport = wkport + 0x54;
		atp870u_outb(0x09, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(u3[j++], tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto u3p_in;
		}
		if (j == 0x0a) {
			goto u3p_cmd;
		}
		if (j == 0x0e) {
			goto try_u3;
		}
		continue;
u3p_out:
		tmport = wkport + 0x58;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(0, tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto u3p_in;
		}
		if (j == 0x0a) {
			goto u3p_cmd;
		}
		if (j == 0x0e) {
			goto u3p_out;
		}
		continue;
u3p_in:
		tmport = wkport + 0x54;
		atp870u_outb(0x09, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		k = 0;
u3p_in1:
		j = atp870u_inb(tmport);
		if ((j & 0x01) != 0) {
			tmport -= 0x06;
			mbuf[k++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto u3p_in1;
		}
		if ((j & 0x80) == 0x00) {
			goto u3p_in1;
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto u3p_in;
		}
		if (j == 0x0a) {
			goto u3p_cmd;
		}
		if (j == 0x0e) {
			goto u3p_out;
		}
		continue;
u3p_cmd:
		tmport = wkport + 0x50;
		atp870u_outb(0x30, tmport);
		tmport = wkport + 0x54;
		atp870u_outb(0x00, tmport);
		tmport += 0x04;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			if (j == 0x4e) {
				goto u3p_out;
			}
			continue;
		}
		if (mbuf[0] != 0x01) {
			goto chg_wide;
		}
		if (mbuf[1] != 0x06) {
			goto chg_wide;
		}
		if (mbuf[2] != 0x04) {
			goto chg_wide;
		}
		if (mbuf[3] == 0x09) {
		   m = 1;
		   m = m << i;
		   dev->wide_idu |= m;
		   dev->id[i].devspu = 0xce;
		   continue;
		}
chg_wide:
		tmport = wkport + 0x5b;
		atp870u_outb(0x01, tmport);
		tmport = wkport + 0x43;
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
try_wide:
		j = 0;
		tmport = wkport + 0x54;
		atp870u_outb(0x05, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(wide[j++], tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto try_wide;
		}
		continue;
widep_out:
		tmport = wkport + 0x58;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				atp870u_outb(0, tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto widep_out;
		}
		continue;
widep_in:
		tmport = wkport + 0x54;
		atp870u_outb(0xff, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		k = 0;
widep_in1:
		j = atp870u_inb(tmport);
		if ((j & 0x01) != 0) {
			tmport -= 0x06;
			mbuf[k++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto widep_in1;
		}
		if ((j & 0x80) == 0x00) {
			goto widep_in1;
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto widep_in;
		}
		if (j == 0x0a) {
			goto widep_cmd;
		}
		if (j == 0x0e) {
			goto widep_out;
		}
		continue;
widep_cmd:
		tmport = wkport + 0x50;
		atp870u_outb(0x30, tmport);
		tmport = wkport + 0x54;
		atp870u_outb(0x00, tmport);
		tmport += 0x04;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			if (j == 0x4e) {
				goto widep_out;
			}
			continue;
		}
		if (mbuf[0] != 0x01) {
			goto not_wide;
		}
		if (mbuf[1] != 0x02) {
			goto not_wide;
		}
		if (mbuf[2] != 0x03) {
			goto not_wide;
		}
		if (mbuf[3] != 0x01) {
			goto not_wide;
		}
		m = 1;
		m = m << i;
		dev->wide_idu |= m;
not_wide:
		if ((dev->id[i].devtypeu == 0x00) || (dev->id[i].devtypeu == 0x07) ||
		    ((dev->id[i].devtypeu == 0x05) && ((n & 0x10) != 0)))
		{
			m = 1;
			m = m << i;
			if ((dev->async & m) != 0)
			{
			   goto set_sync;
			}
		}
		continue;
set_sync:
		if (dev->sp[i] == 0x02)
		{
		   synu[4]=0x0c;
		   synuw[4]=0x0c;
		}
		else
		{
		   if (dev->sp[i] >= 0x03)
		   {
		      synu[4]=0x0a;
		      synuw[4]=0x0a;
		   }
		}
		tmport = wkport + 0x5b;
		j = 0;
		if ((m & dev->wide_idu) != 0) {
			j |= 0x01;
		}
		atp870u_outb(j, tmport);
		tmport = wkport + 0x43;
		atp870u_outb(satn[0], tmport++);
		atp870u_outb(satn[1], tmport++);
		atp870u_outb(satn[2], tmport++);
		atp870u_outb(satn[3], tmport++);
		atp870u_outb(satn[4], tmport++);
		atp870u_outb(satn[5], tmport++);
		tmport += 0x06;
		atp870u_outb(0, tmport);
		tmport += 0x02;
		atp870u_outb(dev->id[i].devspu, tmport++);
		atp870u_outb(0, tmport++);
		atp870u_outb(satn[6], tmport++);
		atp870u_outb(satn[7], tmport++);
		tmport += 0x03;
		atp870u_outb(satn[8], tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		if ((atp870u_inb(tmport) != 0x11) && (atp870u_inb(tmport) != 0x8e)) {
			continue;
		}
		while (atp870u_inb(tmport) != 0x8e);
try_sync:
		j = 0;
		tmport = wkport + 0x54;
		atp870u_outb(0x06, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;

		while ((atp870u_inb(tmport) & 0x80) == 0) {
			if ((atp870u_inb(tmport) & 0x01) != 0) {
				tmport -= 0x06;
				if ((m & dev->wide_idu) != 0) {
					if ((m & dev->ultra_map) != 0) {
						atp870u_outb(synuw[j++], tmport);
					} else {
						atp870u_outb(synw[j++], tmport);
					}
				} else {
					if ((m & dev->ultra_map) != 0) {
						atp870u_outb(synu[j++], tmport);
					} else {
						atp870u_outb(synn[j++], tmport);
					}
				}
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport) & 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto try_sync;
		}
		continue;
phase_outs:
		tmport = wkport + 0x58;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00) {
			if ((atp870u_inb(tmport) & 0x01) != 0x00) {
				tmport -= 0x06;
				atp870u_outb(0x00, tmport);
				tmport += 0x06;
			}
		}
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j == 0x85) {
			goto tar_dcons;
		}
		j &= 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto phase_outs;
		}
		continue;
phase_ins:
		tmport = wkport + 0x54;
		atp870u_outb(0x06, tmport);
		tmport += 0x04;
		atp870u_outb(0x20, tmport);
		tmport += 0x07;
		k = 0;
phase_ins1:
		j = atp870u_inb(tmport);
		if ((j & 0x01) != 0x00) {
			tmport -= 0x06;
			mbuf[k++] = atp870u_inb(tmport);
			tmport += 0x06;
			goto phase_ins1;
		}
		if ((j & 0x80) == 0x00) {
			goto phase_ins1;
		}
		tmport -= 0x08;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		j = atp870u_inb(tmport);
		if (j == 0x85) {
			goto tar_dcons;
		}
		j &= 0x0f;
		if (j == 0x0f) {
			goto phase_ins;
		}
		if (j == 0x0a) {
			goto phase_cmds;
		}
		if (j == 0x0e) {
			goto phase_outs;
		}
		continue;
phase_cmds:
		tmport = wkport + 0x50;
		atp870u_outb(0x30, tmport);
tar_dcons:
		tmport = wkport + 0x54;
		atp870u_outb(0x00, tmport);
		tmport += 0x04;
		atp870u_outb(0x08, tmport);
		tmport += 0x07;
		while ((atp870u_inb(tmport) & 0x80) == 0x00);
		tmport -= 0x08;
		j = atp870u_inb(tmport);
		if (j != 0x16) {
			continue;
		}
		if (mbuf[0] != 0x01) {
			continue;
		}
		if (mbuf[1] != 0x03) {
			continue;
		}
		if (mbuf[4] == 0x00) {
			continue;
		}
		if (mbuf[3] > 0x64) {
			continue;
		}
		if (mbuf[4] > 0x0e) {
			mbuf[4] = 0x0e;
		}
		dev->id[i].devspu = mbuf[4];
		if (mbuf[3] < 0x0c){
			j = 0xb0;
			goto set_syn_ok;
		}
		if ((mbuf[3] < 0x0d) && (rmb == 0)) {
			j = 0xa0;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x1a) {
			j = 0x20;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x33) {
			j = 0x40;
			goto set_syn_ok;
		}
		if (mbuf[3] < 0x4c) {
			j = 0x50;
			goto set_syn_ok;
		}
		j = 0x60;
	      set_syn_ok:
		dev->id[i].devspu = (dev->id[i].devspu & 0x0f) | j;
	}
}

/* return non-zero on detection */
int atp870u_detect(Scsi_Host_Template * tpnt)
{
	unsigned char irq, h, k, m;
	unsigned long flags;
	unsigned int base_io, error, tmport;
	unsigned short index = 0;
	struct pci_dev *pdev[3];
	unsigned char chip_ver[3], host_id;
	unsigned short dev_id[3], n;
	struct Scsi_Host *shpnt = NULL;
	int tmpcnt = 0;
	int count = 0;

	static unsigned short devid[9] = {
	  0x8081, 0x8002, 0x8010, 0x8020, 0x8030, 0x8040, 0x8050, 0x8060, 0
	};

	FUNC_IN();

	if (!pci_present()) {
		printk(KERN_INFO"   NO PCI SUPPORT.\n");
		return count;
	}
	tpnt->proc_name = "atp870u";

	for (h = 0; h < 2; h++) {
		struct atp_unit *dev = &atp_unit[h];
		for(k=0;k<16;k++)
		{
			dev->id[k].prd_tableu = kmalloc(PRD_SIZE, GFP_KERNEL);
			dev->id[k].devspu=0x20;
			dev->id[k].devtypeu = 0;
			dev->id[k].curr_req = NULL;
		}
		dev->active_idu = 0;
		dev->wide_idu = 0;
		dev->host_idu = 0x07;
		dev->quhdu = 0;
		dev->quendu = 0;
		pdev[h]=NULL;
		pdev[2]=NULL;
		dev->chip_veru = 0;
		dev->last_cmd = 0xff;
		dev->in_snd = 0;
		dev->in_int = 0;
		for (k = 0; k < qcnt; k++) {
			dev->querequ[k] = 0;
		}
		for (k = 0; k < 16; k++) {
			dev->id[k].curr_req = 0;
			dev->sp[k] = 0x04;
		}
	}
	h = 0;
	while (devid[h] != 0) {
		pdev[2] = pci_find_device(0x1191, devid[h], pdev[2]);
		if (pdev[2] == NULL || pci_enable_device(pdev[2])) {
			h++;
			index = 0;
			continue;
		}
		chip_ver[2] = 0;
		dev_id[2] = devid[h];

		if (devid[h] == 0x8002) {
			error = pci_read_config_byte(pdev[2], 0x08, &chip_ver[2]);
			if (chip_ver[2] < 2) {
				goto nxt_devfn;
			}
		}
		if (devid[h] == 0x8010 || devid[h] == 0x8081 || devid[h] == 0x8050)
		{
			chip_ver[2] = 0x04;
		}
		pdev[tmpcnt] = pdev[2];
		chip_ver[tmpcnt] = chip_ver[2];
		dev_id[tmpcnt] = dev_id[2];
		tmpcnt++;
	      nxt_devfn:
		index++;
		if (index > 3) {
			index = 0;
			h++;
		}
		if(tmpcnt>1)
			break;
	}
	for (h = 0; h < 2; h++) {
		struct atp_unit *dev=&atp_unit[h];
		if (pdev[h]==NULL) {
			FUNC_STR("atp_unit ");
			return count;
		}

		/* Found an atp870u/w. */
		base_io = pci_resource_start(pdev[h], 0);
		irq = pdev[h]->irq;
		
		if (dev_id[h] != 0x8081)
		{

			FUNC_VAR(dev_id[h]);

		   error = pci_read_config_byte(pdev[h],0x49,&host_id);

		   base_io &= 0xfffffff8;

		   if (check_region(base_io,0x40) != 0)
		   {
   			FUNC_STR("check_region ");
			   return 0;
		   }

#ifdef	MMIO
		   base_io = (unsigned int)ioremap_nocache(base_io, 0x40);
#endif

		   printk(KERN_INFO "   ACARD AEC-671X PCI Ultra/W SCSI-3 Host Adapter: %d    IO:%x, IRQ:%d.\n", h, base_io, irq);
		   
		   dev->ioport = base_io;
		   dev->pciport = base_io + 0x20;
		   dev->deviceid = dev_id[h];
		   irqnumu[h] = irq;
		   host_id &= 0x07;
		   dev->host_idu = host_id;
		   dev->chip_veru = chip_ver[h];

		   tmport = base_io + 0x22;

		   FUNC_STR("before atp870u_inb");
	   
		   FUNC_VAR(atp870u_inb(tmport));
		   FUNC_STR("before atp870u_inb 1");
		   
		   dev->scam_on = atp870u_inb(tmport);
		   FUNC_STR("before atp870u_inb 2");
		   
		   FUNC_VAR(dev->scam_on);		   
		   
		   tmport += 0x0b;
		   dev->global_map = atp870u_inb(tmport++);
		   dev->ultra_map = atp870u_inw(tmport);
		   if (dev->ultra_map == 0) {
			   dev->scam_on = 0x00;
			   dev->global_map = 0x20;
			   dev->ultra_map = 0xffff;
		   }
		   shpnt = scsi_register(tpnt, 4);
		   if(shpnt==NULL)
			{
				FUNC_STR("scsi_register ");
				   return count;
		   	}				  

		   save_flags(flags);
		   cli();
		   if (request_irq(irq, atp870u_intr_handle, SA_SHIRQ, "atp870u", dev)) {
			   printk(KERN_ERR "Unable to allocate IRQ for Acard controller.\n");
			   goto unregister;
		   }

		   if (chip_ver[h] > 0x07)	     /* check if atp876 chip   */
		   {				     /* then enable terminator */
		      tmport = base_io + 0x3e;
		      atp870u_outb(0x00, tmport);
		   }

		   tmport = base_io + 0x3a;
		   k = (atp870u_inb(tmport) & 0xf3) | 0x10;
		   atp870u_outb(k, tmport);
		   atp870u_outb((k & 0xdf), tmport);
		   mydlyu(0x8000);
		   atp870u_outb(k, tmport);
		   mydlyu(0x8000);
		   tmport = base_io;
		   atp870u_outb((host_id | 0x08), tmport);
		   tmport += 0x18;
		   atp870u_outb(0, tmport);						// 18
		   tmport += 0x07;
		   while ((atp870u_inb(tmport) & 0x80) == 0);		// 1F
		   tmport -= 0x08;
		   atp870u_inb(tmport);						// 17
		   tmport = base_io + 1;
		   atp870u_outb(8, tmport++);					// PIO
		   atp870u_outb(0x7f, tmport);					// Timeout Reg
		   tmport = base_io + 0x11;
		   atp870u_outb(0x20, tmport);					// Sync Reg

		   tscam(h);
		   is870(h, base_io);
		   tmport = base_io + 0x3a;
		   atp870u_outb((atp870u_inb(tmport) & 0xef), tmport);
		   tmport++;
		   atp870u_outb((atp870u_inb(tmport) | 0x20),tmport);
		}
		else
		{
		   base_io &= 0xfffffff8;

		   if (check_region(base_io,0x60) != 0)
		   {
			FUNC_STR("check_region ");
			   return 0;
		   }
		   host_id = atp870u_inb(base_io + 0x39);
		   host_id >>= 0x04;

		   printk(KERN_INFO "   ACARD AEC-67160 PCI Ultra160 LVD/SE SCSI Adapter: %d    IO:%x, IRQ:%d.\n"
			  ,h, base_io, irq);
		   dev->ioport = base_io + 0x40;
		   dev->pciport = base_io + 0x28;
		   dev->deviceid = dev_id[h];
		   irqnumu[h] = irq;
		   dev->host_idu = host_id;
		   dev->chip_veru = chip_ver[h];

		   tmport = base_io + 0x22;
		   dev->scam_on = atp870u_inb(tmport);
		   tmport += 0x13;
		   dev->global_map = atp870u_inb(tmport);
		   tmport += 0x07;
		   dev->ultra_map = atp870u_inw(tmport);

		   n=0x3f09;
next_fblk:
		   if (n >= 0x4000)
		   {
		      goto flash_ok;
		   }
		   m=0;
		   atp870u_outw(n,base_io + 0x34);
		   n += 0x0002;
		   if (atp870u_inb(base_io + 0x30) == 0xff)
		   {
		      goto flash_ok;
		   }
		   dev->sp[m++]=atp870u_inb(base_io + 0x30);
		   dev->sp[m++]=atp870u_inb(base_io + 0x31);
		   dev->sp[m++]=atp870u_inb(base_io + 0x32);
		   dev->sp[m++]=atp870u_inb(base_io + 0x33);
		   atp870u_outw(n,base_io + 0x34);
		   n += 0x0002;
		   dev->sp[m++]=atp870u_inb(base_io + 0x30);
		   dev->sp[m++]=atp870u_inb(base_io + 0x31);
		   dev->sp[m++]=atp870u_inb(base_io + 0x32);
		   dev->sp[m++]=atp870u_inb(base_io + 0x33);
		   atp870u_outw(n,base_io + 0x34);
		   n += 0x0002;
		   dev->sp[m++]=atp870u_inb(base_io + 0x30);
		   dev->sp[m++]=atp870u_inb(base_io + 0x31);
		   dev->sp[m++]=atp870u_inb(base_io + 0x32);
		   dev->sp[m++]=atp870u_inb(base_io + 0x33);
		   atp870u_outw(n,base_io + 0x34);
		   n += 0x0002;
		   dev->sp[m++]=atp870u_inb(base_io + 0x30);
		   dev->sp[m++]=atp870u_inb(base_io + 0x31);
		   dev->sp[m++]=atp870u_inb(base_io + 0x32);
		   dev->sp[m++]=atp870u_inb(base_io + 0x33);
		   n += 0x0018;
		   goto next_fblk;
flash_ok:
		   atp870u_outw(0,base_io + 0x34);
		   dev->ultra_map=0;
		   dev->async = 0;
		   for (k=0; k < 16; k++)
		   {
		       n=1;
		       n = n << k;
		       if (dev->sp[k] > 1)
		       {
			  dev->ultra_map |= n;
		       }
		       else
		       {
			  if (dev->sp[k] == 0)
			  {
			     dev->async |= n;
			  }
		       }
		   }
		   dev->async = ~(dev->async);
		   atp870u_outb(dev->global_map,base_io + 0x35);

		   shpnt = scsi_register(tpnt, 4);
		   if(shpnt==NULL)
		   	{
   				FUNC_STR("scsi_register ");
				   return count;
		   	}			

		   save_flags(flags);
		   cli();
		   if (request_irq(irq, atp870u_intr_handle, SA_SHIRQ, "atp870u", dev)) {
			   printk(KERN_ERR "Unable to allocate IRQ for Acard controller.\n");
			   goto unregister;
		   }

		   tmport = base_io + 0x38;
		   k = atp870u_inb(tmport) & 0x80;
		   atp870u_outb(k, tmport);
		   tmport += 0x03;
		   atp870u_outb(0x20, tmport);
		   mydlyu(0x8000);
		   atp870u_outb(0, tmport);
		   mydlyu(0x8000);
		   tmport = base_io + 0x5b;
		   atp870u_inb(tmport);
		   tmport -= 0x04;
		   atp870u_inb(tmport);
		   tmport = base_io + 0x40;
		   atp870u_outb((host_id | 0x08), tmport);
		   tmport += 0x18;
		   atp870u_outb(0, tmport);
		   tmport += 0x07;
		   while ((atp870u_inb(tmport) & 0x80) == 0);
		   tmport -= 0x08;
		   atp870u_inb(tmport);
		   tmport = base_io + 0x41;
		   atp870u_outb(8, tmport++);
		   atp870u_outb(0x7f, tmport);
		   tmport = base_io + 0x51;
		   atp870u_outb(0x20, tmport);

		   tscam(h);
		   is880(h, base_io);
		   tmport = base_io + 0x38;
		   atp870u_outb(0xb0, tmport);
		}

		atp_host[h] = shpnt;
		if (dev->chip_veru == 4) {
			shpnt->max_id = 16;
		}
		shpnt->this_id = host_id;
		shpnt->unique_id = base_io;
		shpnt->io_port = base_io;
		if (dev_id[h] == 0x8081)
		{
		   shpnt->n_io_port = 0x60;	   /* Number of bytes of I/O space used */
		}
		else
		{
		   shpnt->n_io_port = 0x40;	   /* Number of bytes of I/O space used */
		}
		shpnt->irq = irq;
		restore_flags(flags);
		if (dev_id[h] == 0x8081)
		{
		   request_region(base_io, 0x60, "atp870u");       /* Register the IO ports that we use */
		}
		else
		{
		   request_region(base_io, 0x40, "atp870u");       /* Register the IO ports that we use */
		}
		count++;
		index++;
		continue;
unregister:
		scsi_unregister(shpnt);
		restore_flags(flags);
		index++;
		continue;
	}

	FUNC_OUT();
	return count;
}

/* The abort command does not leave the device in a clean state where
   it is available to be used again.  Until this gets worked out, we will
   leave it commented out.  */

int atp870u_abort(Scsi_Cmnd * SCpnt)
{
	unsigned char h, j, k;
	Scsi_Cmnd *workrequ;
	unsigned int tmport;
	struct atp_unit *dev;

	FUNC_IN();
	
	for (h = 0; h <= admaxu; h++) {
		if (SCpnt->host == atp_host[h]) {
			goto find_adp;
		}
	}
	panic("Abort host not found !");
find_adp:
	dev=&atp_unit[h];
	printk(KERN_DEBUG "working=%x last_cmd=%x ", dev->working, dev->last_cmd);
	printk(" quhdu=%x quendu=%x ", dev->quhdu, dev->quendu);
	tmport = dev->ioport;
	for (j = 0; j < 0x17; j++) {
		printk(" r%2x=%2x", j, atp870u_inb(tmport++));
	}
	tmport += 0x05;
	printk(" r1c=%2x", atp870u_inb(tmport));
	tmport += 0x03;
	printk(" r1f=%2x in_snd=%2x ", atp870u_inb(tmport), dev->in_snd);
	tmport= dev->pciport;
	printk(" r20=%2x", atp870u_inb(tmport));
	tmport += 0x02;
	printk(" r22=%2x", atp870u_inb(tmport));
	tmport = dev->ioport + 0x3a;
	printk(" r3a=%2x \n",atp870u_inb(tmport));
	tmport = dev->ioport + 0x3b;
	printk(" r3b=%2x \n",atp870u_inb(tmport));
	for(j=0;j<16;j++)
	{
	   if (dev->id[j].curr_req != NULL)
	   {
		workrequ = dev->id[j].curr_req;
		printk("\n que cdb= ");
		for (k=0; k < workrequ->cmd_len; k++)
		{
		    printk(" %2x ",workrequ->cmnd[k]);
		}
		printk(" last_lenu= %lx ",dev->id[j].last_lenu);
	   }
	}

	FUNC_OUT();
	
	return (SCSI_ABORT_SNOOZE);
}

int atp870u_reset(Scsi_Cmnd * SCpnt, unsigned int reset_flags)
{
	unsigned char h;
	struct atp_unit *dev;


	FUNC_IN();
	
	/*
	 * See if a bus reset was suggested.
	 */
	for (h = 0; h <= admaxu; h++) {
		if (SCpnt->host == atp_host[h]) {
			goto find_host;
		}
	}
	panic("Reset bus host not found !");
find_host:
	dev=&atp_unit[h];
/*	SCpnt->result = 0x00080000;
	SCpnt->scsi_done(SCpnt);
	dev->working=0;
	dev->quhdu=0;
	dev->quendu=0;
	return (SCSI_RESET_SUCCESS | SCSI_RESET_BUS_RESET);  */

	FUNC_OUT();

	return (SCSI_RESET_SNOOZE);
}

const char *atp870u_info(struct Scsi_Host *notused)
{
	static char buffer[128];

	FUNC_IN();

	strcpy(buffer, "ACARD AEC-6710/6712/67160 PCI Ultra/W/LVD SCSI-3 Adapter Driver V2.5+ac ");
	FUNC_OUT();

	return buffer;
}

int atp870u_set_info(char *buffer, int length, struct Scsi_Host *HBAptr)
{

	FUNC_IN();

	return -ENOSYS; 	/* Currently this is a no-op */
	FUNC_OUT();
	
}

#define BLS buffer + len + size
int atp870u_proc_info(char *buffer, char **start, off_t offset, int length,
		      int hostno, int inout)
{
	struct Scsi_Host *HBAptr;
	static u8 buff[512];
	int i;
	int size = 0;
	int len = 0;
	off_t begin = 0;
	off_t pos = 0;

	FUNC_IN();

	HBAptr = NULL;
	for (i = 0; i < 2; i++) {
		if ((HBAptr = atp_host[i]) != NULL) {
			if (HBAptr->host_no == hostno) {
				break;
			}
			HBAptr = NULL;
		}
	}

	if (HBAptr == NULL) {
		size += sprintf(BLS, "Can't find adapter for host number %d\n", hostno);
		len += size;
		pos = begin + len;
		size = 0;
		goto stop_output;
	}
	if (inout == TRUE) {	/* Has data been written to the file? */
		FUNC_STR("		inout");
		
		return (atp870u_set_info(buffer, length, HBAptr));
	}
	if (offset == 0) {
		memset(buff, 0, sizeof(buff));
	}
	size += sprintf(BLS, "ACARD AEC-671X Driver Version: 2.5+ac\n");
	len += size;
	pos = begin + len;
	size = 0;

	size += sprintf(BLS, "\n");
	size += sprintf(BLS, "Adapter Configuration:\n");
	size += sprintf(BLS, "               Base IO: %#.4lx\n", HBAptr->io_port);
	size += sprintf(BLS, "                   IRQ: %d\n", HBAptr->irq);
	len += size;
	pos = begin + len;
	size = 0;

stop_output:
	*start = buffer + (offset - begin);	/* Start of wanted data */
	len -= (offset - begin);	/* Start slop */
	if (len > length) {
		len = length;	/* Ending slop */
	}

	FUNC_STR("		Leave");
	
	return (len);
}

#include "sd.h"

int atp870u_biosparam(Scsi_Disk * disk, kdev_t dev, int *ip)
{
	int heads, sectors, cylinders;

	FUNC_IN();

	heads = 64;
	sectors = 32;
	cylinders = disk->capacity / (heads * sectors);

	if (cylinders > 1024) {
		heads = 255;
		sectors = 63;
		cylinders = disk->capacity / (heads * sectors);
	}
	ip[0] = heads;
	ip[1] = sectors;
	ip[2] = cylinders;

	FUNC_OUT();

	return 0;
}


int atp870u_release (struct Scsi_Host *pshost)
{
	int h;

	FUNC_IN();
	
	for (h = 0; h <= admaxu; h++)
	{
		if (pshost == atp_host[h]) {
			int k;
			free_irq (pshost->irq, &atp_unit[h]);
			release_region (pshost->io_port, pshost->n_io_port);
			scsi_unregister(pshost);
			for(k=0;k<16;k++)
				kfree(atp_unit[h].id[k].prd_tableu);

			FUNC_OUT();
			return 0;
		}
	}
	panic("atp870u: bad scsi host passed.\n");

}
MODULE_LICENSE("GPL");

static Scsi_Host_Template driver_template = ATP870U;
#include "scsi_module.c"
