#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ctype.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/ptrace.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/proc_fs.h>
#include <linux/in.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/errno.h>
#include <asm/bootinfo.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/pgtable.h>
#include <asm/segment.h>
#include <asm/io.h>
#include <asm/dma.h>

#include "usb_slave_common.h"
#include "usb_slave.h"
#include "asm/rc32300/rc32355.h"

/* USB registers addresses */
#define USBC (volatile __u32 *)0xB8068000
#define USBS (volatile __u32 *)0xB8068004
#define USBSM (volatile __u32 *)0xB8068008
#define USBADDR (volatile __u32 *)0xB806800C
#define USBFC (volatile __u32 *)0xB8068010
#define USBEPAC (volatile __u32 *)0xB8068020
#define USBEPAS (volatile __u32 *)0xB8068024
#define USBEPASM (volatile __u32 *)0xB8068028
#define USBEPATBS (volatile __u32 *)0xB806802C
#define USBEPATD (volatile __u32 *)0xB8068030
#define USBEPARD (volatile __u32 *)0xB8068040
#define USBEPARBS (volatile __u32 *)0xB8068050
#define USBEPBC (volatile __u32 *)0xB8068060
#define USBEPBS (volatile __u32 *)0xB8068064
#define USBEPBSM (volatile __u32 *)0xB8068068
#define USBEPBBS (volatile __u32 *)0xB806806C
#define USBEPBD (volatile __u32 *)0xB8068070
#define USBEPCC (volatile __u32 *)0xB8068080
#define USBEPCS (volatile __u32 *)0xB8068084
#define USBEPCSM (volatile __u32 *)0xB8068088
#define USBEPCBS (volatile __u32 *)0xB806808C
#define USBEPCD (volatile __u32 *)0xB8068090
#define USBEPDC (volatile __u32 *)0xB80680A0
#define USBEPDS (volatile __u32 *)0xB80680A4
#define USBEPDSM (volatile __u32 *)0xB80680A8
#define USBEPDBS (volatile __u32 *)0xB80680AC
#define USBEPDD (volatile __u32 *)0xB80680B0
#define USBEPEC (volatile __u32 *)0xB80680C0
#define USBEPES (volatile __u32 *)0xB80680C4
#define USBEPESM (volatile __u32 *)0xB80680C8
#define USBEPEBS (volatile __u32 *)0xB80680CC
#define USBEPED (volatile __u32 *)0xB80680D0
#define USBEPFC (volatile __u32 *)0xB80680E0
#define USBEPFS (volatile __u32 *)0xB80680E4
#define USBEPFSM (volatile __u32 *)0xB80680E8
#define USBEPFBS (volatile __u32 *)0xB80680EC
#define USBEPFD (volatile __u32 *)0xB80680F0
#define USBEPGC (volatile __u32 *)0xB8068100
#define USBEPGS (volatile __u32 *)0xB8068104
#define USBEPGSM (volatile __u32 *)0xB8068108
#define USBEPGBS (volatile __u32 *)0xB806810C
#define USBEPGD (volatile __u32 *)0xB8068110

/* USBC register bits */
#define USBC_EN (1)		 
#define USBC_SP (1<<1)		 
#define USBC_RSM (1<<2)
#define USBC_DFG (1<<3)
#define USBC_RES (1<<4)

/* USBSM register bits */
#define USBSM_EPA  (1<<0)	
#define USBSM_EPB  (1<<1)
#define USBSM_EPC  (1<<2)
#define USBSM_EPD  (1<<3)
#define USBSM_EPE  (1<<4)
#define USBSM_EPF  (1<<5)
#define USBSM_EPG  (1<<6)
#define USBSM_FSL  (1<<7)
#define USBSM_SOF  (1<<8)
#define USBSM_LGF  (1<<9)
#define USBSM_SUS  (1<<10)
#define USBSM_RSM  (1<<11)
#define USBSM_RST  (1<<12)
#define USBSM_FSE  (1<<13)

 /* USBS register bits */
#define USBS_EPA (1<<0)
#define USBS_EPB (1<<1)
#define USBS_EPC (1<<2)
#define USBS_EPD (1<<3)	
#define USBS_EPE (1<<4)
#define USBS_EPF (1<<5)
#define USBS_EPG (1<<6)
#define USBS_FSL (1<<7)
#define USBS_SOF (1<<8)
#define USBS_LGF (1<<9)
#define USBS_SUS (1<<10)
#define USBS_RSM (1<<11)
#define USBS_RST (1<<12)
#define USBS_FSE (1<<13)

 /* USBEPAC register bits */
#define USBEPAC_MODE(v) (v&0x3)<<0
#define USBEPAC_FSH 1<<8          

 /* USBASM register bits */
#define USBEPASM_RBR (1<<0)
#define USBEPASM_TBR (1<<1)
#define USBEPASM_SUP (1<<3)
#define USBEPASM_S (1<<6)
#define USBEPASM_ACK (1<<7)
#define USBEPASM_ERR (1<<8)

/* USBAS register bits */
#define USBEPAS_RBR (1<<0)
#define USBEPAS_TBR (1<<1)
#define USBEPAS_SUP (1<<3)
#define USBEPAS_S (1<<6)
#define USBEPAS_ACK (1<<7)
#define USBEPAS_ERR (1<<8)

/* USBEP[B|C|D|E|F|G]SM register bits */
#define USBEPxSM_BR (1<<0)
#define USBEPxSM_BE (1<<2)
#define USBEPxSM_MIP (1<<4)
#define USBEPxSM_IRE (1<<5)
#define USBEPxSM_S (1<<6)
#define USBEPxSM_ACK (1<<7)
#define USBEPxSM_ERR (1<<8)
#define USBEPxSM_BRS (1<<9)

/* USBEP[B|C|D|E|F|G]S register bits */
#define USBEPxS_BR (1<<0)
#define USBEPxS_BE (1<<2)
#define USBEPxS_MIP (1<<4)
#define USBEPxS_IRE (1<<5)
#define USBEPxS_S (1<<6)
#define USBEPxS_ACK (1<<7)
#define USBEPxS_ERR (1<<8)
#define USBEPxS_BRS (1<<9)

/* USBEP[B|C|D|E|F|G]C register bits */
#define USBEPxC_MODE(v) ((v&0x3)<<0) 
#define USBEPxC_TYPE(v) ((v&0x3)<<2)
#define USBEPxC_EPNUM(v) ((v&0xf)<<4)
#define USBEPxC_FSH (1<<8)
#define USBEPxC_SFM (1<<9)
#define USBEPxC_ITS (1<<10)
#define USBEPxC_MPS(v) ((v&0x3ff)<<16)

/* USB[F|G]DEVCS - DMA section */
#define USB_DEVCS_FD (1<<0)
#define USB_DEVCS_LD (1<<1)
#define USB_DEVCS_BE (1<<2)
#define USB_DEVCS_MIP (1<<3)
#define USB_DEVCS_IRE (1<<4)
#define USB_DEVCS_FRMNUM(v) ((v&0x7ff)<<5)

/* DMAxC register */
#define DMAxC_RUN (1<<0)
#define DMAxC_DM (1<<1)
#define DMAxC_MODE(v) ((v&0x3)<<2)    

/* DMA modes */
#define DMA_AUTO_REQUEST     0
#define DMA_BURST_REQUEST    1
#define DMA_TRANSFER_REQUEST 2

 /* DMAxS register */
#define DMAxS_F (1<<0)  
#define DMAxS_D (1<<1) 
#define DMAxS_C (1<<2)
#define DMAxS_E (1<<3)
#define DMAxS_H (1<<4)

/* Transfers information */
#define SET_NAK_MODE_HW 1
#define SET_ACK_MODE_HW 2
#define SET_NAK_MODE_SW 3
#define SET_ACK_MODE_SW 4

/* Specific vendor control request */
#define SEND_VEND_DATA  5
#define GET_VEND_DATA 6
#define FILL_EP_DATA 7
#define SEND_EP_SIZE_DATA 8
#define SEND_NORMAL_DATA 9
#define ACK_MODE 0
#define NAK_MODE 1

#define INTR_USBM_EPA (1<<6)
#define INTR_USBM_EPB (1<<7)
#define INTR_USBM_EPC (1<<8)
#define INTR_USBM_EPD (1<<9)
#define INTR_USBM_EPE (1<<10)
#define INTR_USBM_EPF (1<<11)
#define INTR_USBM_EPG (1<<12)
#define INTR_USBM_USBS (1<<13)

#define MISC_IRQ 5
#define DMA_IRQ 3

#define USB_DMA_BUFF_SIZE 64

/* Interrupts registers base */
#define USB_BASE          0xB8068000
#define DMA_BASE          0xB8038000
#define INTERRUPT_BASE    0xB8030000
#define GPIO_BASE         0xB8040000
#define PHYS_ADDR(x)      (x & 0x1fffffff) 

 /* DMA descriptor fields */
#define DMA_FINISHED (1<<31)
#define DMA_DONE (1<<30)
#define DMA_TERMINATED (1<<29)
#define DMA_IOD (1<<28)
#define DMA_IOF (1<<27)
#define DMA_COD (1<<26)
#define DMA_COF (1<<25)
#define DMA_DEVCMD(v) ((v&0x7)<<22)
#define DMA_DS(v) ((v&0x3)<<20)
#define DMA_COUNT(v) (v&0x3ffff) 

#ifdef CONFIG_USB_SIMPLE_DMA
#define NUM_DMA_DESCS 256/* DMA buffer size is 64 */
#else
#define NUM_DMA_DESCS 30 /* DMA buffer size is 1024 */
#endif

#define MASK_ALL_EP_INTERRUPTS USBEPxSM_BR | USBEPxSM_BE | USBEPxSM_MIP | \
    USBEPxSM_IRE | USBEPxSM_S | USBEPxSM_ACK | USBEPxSM_ERR | USBEPxSM_BRS

/* Endpoint hardware data */
/* Max packet size */
#define EP_BC_MAX_PKT_SIZE 16 
#define EP_DE_MAX_PKT_SIZE 32
#define EP_FG_MAX_PKT_SIZE 64

/* Serials numbers */
#define EP_A 0
#define EP_B 1
#define EP_C 2
#define EP_D 3
#define EP_E 4
#define EP_F 5
#define EP_G 6

#define OFF 0
#define ON 1

#define UDC_MAX_EPS_NUM 6

typedef volatile struct dma_descriptor {
    unsigned int control;
    unsigned int ca;
    unsigned int devcs;
    volatile struct dma_descriptor *link;
} DMA_DESCRIPTOR;

typedef struct {
    unsigned int ipend;
    unsigned int imask;
} INTR;

typedef struct {
    INTR i2; 
    INTR i3; 
    INTR i4; 
    INTR i5; 
    INTR i6; 
} interrupt_regs_t;

#define interrupt (*((volatile interrupt_regs_t *) INTERRUPT_BASE))

typedef struct {
    unsigned int dmac;
    unsigned int dmas;
    unsigned int dmasm;
    unsigned int dmadptr;
    unsigned int dmandptr;
} DMA_CHAN;

typedef struct {
    DMA_CHAN ch0;
    DMA_CHAN ch1;
    DMA_CHAN ch2;
    DMA_CHAN ch3;
    DMA_CHAN ch4;
    DMA_CHAN ch5;
    DMA_CHAN ch6;
    DMA_CHAN ch7;
    DMA_CHAN ch8;
    DMA_CHAN ch9;
    DMA_CHAN ch10;
    DMA_CHAN ch11;
    DMA_CHAN ch12;
    DMA_CHAN ch13;
    DMA_CHAN ch14;
    DMA_CHAN ch15;
} DMA;

#define dma (*((volatile DMA *) DMA_BASE))

typedef struct {
    unsigned int gpiofunc;
    unsigned int gpiocfg;
    unsigned int gpiod;
    unsigned int gpioilevel;
    unsigned int gpioistat;
    unsigned int gpionmien;
} GPIO;

#define gpio (*((volatile GPIO *) GPIO_BASE))

typedef struct {
    unsigned int usbepc;
    unsigned int usbeps;
    unsigned int usbepsm;
    unsigned int usbeptbs;
    unsigned int usbeptd;
    unsigned int _u0;
    unsigned int _u1;
    unsigned int _u2;
    unsigned int usbeprd;
    unsigned int _u3;
    unsigned int _u4;
    unsigned int _u5;
    unsigned int usbeprbs;
    unsigned int _u6;
    unsigned int _u7;
    unsigned int _u8;
} USB_EPA;

typedef struct {
    unsigned int usbepc;
    unsigned int usbeps;
    unsigned int usbepsm;
    unsigned int usbepbs;
    unsigned int usbepd;
    unsigned int _u0;
    unsigned int _u1;
    unsigned int _u2;
} USB_EP;

typedef struct {
    unsigned int usbc;
    unsigned int usbs;
    unsigned int usbsm;
    unsigned int usbaddr;
    unsigned int usbfc;
    unsigned int _u0;
    unsigned int _u1;
    unsigned int _u2;
    USB_EPA epa;
    USB_EP  epb;
    USB_EP  epc;
    USB_EP  epd;
    USB_EP  epe;
    USB_EP  epf;
    USB_EP  epg;
} USB;

#define usb (*((volatile USB *) USB_BASE))

typedef struct {
    volatile __u32 *control_reg;
    volatile __u32 *status_reg;
    volatile __u32 *status_mask_reg;
    volatile __u32 *buffer_status_reg;
    volatile __u32 *data_reg;
    int serial_num;
    int is_taken;
    int is_dma;
    int direction;
    int max_pkt_size;
} ep_phys_info_t;

typedef struct {
    volatile __u32 *control_reg;
    volatile __u32 *status_reg;
    volatile __u32 *status_mask_reg;
    volatile __u32 *tx_buffer_status_reg;
    volatile __u32 *tx_data_reg;
    volatile __u32 *rx_data_reg;
    volatile __u32 *rx_buffer_status_reg;
} ctrl_ep_phys_info_t;  

typedef struct {
    ctrl_ep_phys_info_t ep_a;
    ep_phys_info_t ep_b;
    ep_phys_info_t ep_c;
    ep_phys_info_t ep_d;
    ep_phys_info_t ep_e;
    ep_phys_info_t ep_f;
    ep_phys_info_t ep_g;
} eps_setup_t;

/* Local function and call backs declaretion */
static void (*usb_slave_generic_interrupt_handler)(__u32 status, 
    int ep_addr);
static void usb_slave_misc_interrupt_handler(int irq, void *dev_id, 
    struct pt_regs * regs);
static void usb_dma_ep_g_interrupt_handler(endpoint_t *ep);
static void usb_slave_dma_interrupt_handler(int irq, void *dev_id, 
    struct pt_regs *regs);
static void init_device_endpoints(void);
static int setup_phys_ep(endpoint_t *ep);
static void setup_dummy_eps(void);

/* global DMA variables */
#ifdef CONFIG_USB_SIMPLE_DMA
static DMA_DESCRIPTOR *first_descriptor;
static DMA_DESCRIPTOR *dma_descriptors_list;
static __u8 *dma_buffers_list;
static int desc_count;
static int usb_dma_current_transfer_state;
#endif

eps_setup_t eps_pool = {
  {USBEPAC, USBEPAS, USBEPASM, USBEPATBS, USBEPATD, USBEPARD, USBEPARBS},
  {USBEPBC, USBEPBS, USBEPBSM, USBEPBBS, USBEPBD, EP_B, 0, 0, RG_USB_OUT, EP_BC_MAX_PKT_SIZE},
  {USBEPCC, USBEPCS, USBEPCSM, USBEPCBS, USBEPCD, EP_C, 0, 0, RG_USB_IN, EP_BC_MAX_PKT_SIZE},
  {USBEPDC, USBEPDS, USBEPDSM, USBEPDBS, USBEPDD, EP_D, 0, 0, RG_USB_OUT, EP_DE_MAX_PKT_SIZE},
  {USBEPEC, USBEPES, USBEPESM, USBEPEBS, USBEPED, EP_E, 0, 0, RG_USB_IN, EP_DE_MAX_PKT_SIZE},
  {USBEPFC, USBEPFS, USBEPFSM, USBEPFBS, USBEPFD, EP_F, 0, 1, RG_USB_OUT, EP_FG_MAX_PKT_SIZE},
  {USBEPGC, USBEPGS, USBEPGSM, USBEPGBS, USBEPGD, EP_G, 0, 1, RG_USB_IN, EP_FG_MAX_PKT_SIZE}
};
static endpoint_t *idt_tx_dma_ep;
static int is_initialized;
static int interrupt_handler_dev_id = 0;

void init_dma_tx_ep_channel(void)
{
    dma.ch14.dmac = 0;
    dma.ch14.dmasm = 0x1f; /* Mask all interrupts, for now */
    dma.ch14.dmas = 0;
    dma.ch14.dmadptr = 0;
    dma.ch14.dmandptr = 0;

    /* Mandatory for internal peripherals */
    dma.ch14.dmac = DMAxC_MODE(DMA_TRANSFER_REQUEST);
}

static void usb_slave_misc_interrupt_handler(int irq, void *dev_id, 
    struct pt_regs *regs)
{
    int idt_status = usb.usbs;
    __u32 gen_status = 0;
    
    /* This is not our interrupt */
    if (!(idt_status & ~(usb.usbsm)))
	return;
    if (idt_status & USBS_EPA)
    {
	int control_ep_status = usb.epa.usbeps;

	USB_DPRINTK(__FUNCTION__": Got EPA interrupt\n");
	gen_status |= INTERRUPT_ENDPOINT;
	if (control_ep_status & USBEPAS_S)
	{
	    int control_ep_control = usb.epa.usbepc;
	    
	    if (control_ep_control & 0x3)
	    {
		control_ep_control |= 0x3;
                usb.epa.usbepc = control_ep_control;
	    }
	    usb.epa.usbeps &= ~(USBEPAS_S);
	}
	if (control_ep_status & (USBEPAS_SUP))
	    gen_status |= INTERRUPT_CTRL_EP_SETUP;
	usb_slave_generic_interrupt_handler(gen_status, 0);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPA);
    }
    if (idt_status & USBS_EPB)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPB interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_b.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPB);
    }
    if (idt_status & USBS_EPC)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPC interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_c.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPC);
    }
    if (idt_status & USBS_EPD)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPD interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_d.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPD);
    }
    if (idt_status & USBS_EPE)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPE interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_e.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPE);
    }
    if (idt_status & USBS_EPF)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPF interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_f.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPF);
    }
    if (idt_status & USBS_EPG)
    { 
	USB_DPRINTK(__FUNCTION__": Got EPG interrupt\n");
	usb_slave_generic_interrupt_handler(INTERRUPT_ENDPOINT,
	   ((*eps_pool.ep_g.control_reg) & 0xf0) >> 4);
	gen_status = 0;
	usb.usbs &= ~(USBS_EPG);
    }
    if (idt_status & USBS_FSL)
    {
	USB_DPRINTK(__FUNCTION__": Got FSL interrupt\n");
	
	/* TODO: increase counters */
	
	usb.usbs &= ~(USBS_FSL);
    }
    if (idt_status & USBS_SOF)
    {
	USB_DPRINTK(__FUNCTION__": Got SOF interrupt\n");
	
	/* TODO: increase counters */
	
	usb.usbs &= ~(USBS_SOF);
	gen_status |= INTERRUPT_SOF;
    }	
    if (idt_status & USBS_LGF)
    {
	USB_DPRINTK(__FUNCTION__": Got LGF interrupt\n");
	
	/* TODO: increase counters */
	
	usb.usbs &= ~(USBS_LGF);
    }	
    if (idt_status & USBS_SUS)
    {
	/* TODO: increase counters */
	
	usb.usbs &= ~(USBS_SUS);
	gen_status |= INTERRUPT_SUSPEND;
    }	
    if (idt_status & USBS_RSM)
    {
	/* TODO: increase counters */

	usb.usbs &= ~(USBS_RSM);
	gen_status |= INTERRUPT_RESUME;
    }	
    if (idt_status & USBS_RST)
    {
	/* TODO: increase counters */

	usb.usbs &= ~(USBS_RST);
	gen_status |= INTERRUPT_RESET;
    }	
    if (idt_status & USBS_FSE)
    {
	USB_DPRINTK(__FUNCTION__": Got FSE interrupt\n");
	
        /* TODO: increase counters */

	usb.usbs &= ~(USBS_FSE);
    }	
    if (gen_status)
	usb_slave_generic_interrupt_handler(gen_status, 0);
}

static void usb_slave_dma_interrupt_handler(int irq, void *dev_id, 
    struct pt_regs *regs)
{
    unsigned long flag;

    save_and_cli(flag);
    usb_dma_ep_g_interrupt_handler(idt_tx_dma_ep);	  
    restore_flags(flag);
}

static int init_dma_memory(void)
{
    dma_descriptors_list = (DMA_DESCRIPTOR *) kmalloc(sizeof(DMA_DESCRIPTOR) * NUM_DMA_DESCS, GFP_KERNEL);
    if (!dma_descriptors_list)
      { 
	printk(KERN_ERR __FUNCTION__":Can't allocate memory for DMA "
	       "descriptors\n");
	return -1;
      }
    dma_cache_wback_inv((unsigned long)(dma_descriptors_list), sizeof(DMA_DESCRIPTOR) * NUM_DMA_DESCS);

    dma_descriptors_list = (DMA_DESCRIPTOR *) KSEG1ADDR(dma_descriptors_list);

    dma_buffers_list = (__u8 *)kmalloc(USB_DMA_BUFF_SIZE * NUM_DMA_DESCS, GFP_KERNEL);
    if (!dma_buffers_list)
      { 
	printk(KERN_ERR __FUNCTION__":Can't allocate memory for DMA buffers\n");
	kfree((void *)KSEG0ADDR(dma_descriptors_list));
	dma_descriptors_list = NULL;
	return -1;
      }
    dma_cache_wback_inv((unsigned long)(dma_buffers_list), USB_DMA_BUFF_SIZE * NUM_DMA_DESCS);

    dma_buffers_list = (__u8 *)KSEG1ADDR(dma_buffers_list);
    return 0;
}

int usb_slave_phys_init(void (*misc_interrupt_handler)(__u32 status, 
    int ep_addr))
{
    __u32 *gpiod;
    
    printk(KERN_INFO __FUNCTION__": Starting IDT's USB initialization\n");
    usb_slave_generic_interrupt_handler = misc_interrupt_handler;
    if (!is_initialized)
    {
	request_region(USB_BASE,  sizeof usb ,"Jungo USB");
	request_region(DMA_BASE, sizeof dma, "Jungo DMA");
	request_region(INTERRUPT_BASE, sizeof interrupt,"Jungo INTERRUPT");
	request_region(GPIO_BASE, sizeof gpio, "Jungo GPIO");
    }
    
    /* GPIOD register is at 0xb8040008  - set bit 26 to 1*/
    gpiod = (__u32*)(0xb8040008);
    *gpiod |= 0x04000000;

    /* Disable the USB Interface in the USB Control Register */
    usb.usbc &= ~(USBC_EN);

    /* Enable the USB Interface in the USB Control Register */
    usb.usbc |= (USBC_EN);

    /* Endpoint A Settings */
    /* Enable control endpoint */
    usb.epa.usbepc = USBEPAC_MODE(USB_ENABLED);   

    /* Disable control endpoint functionality */
    usb.epa.usbepsm = (USBEPASM_RBR | USBEPASM_TBR |
	USBEPASM_SUP | USBEPASM_S | USBEPASM_ACK | USBEPASM_ERR);

    /* Initial global DMA descriptor and buffers for transmit */
    if (init_dma_memory() < 0)
	return -1;
    init_dma_tx_ep_channel();

    /* Enable USB DMA transmit interrupt */
    request_irq(GROUP1_IRQ_BASE + 14, usb_slave_dma_interrupt_handler, SA_SHIRQ, 
	"USB_Slave_DMA", &interrupt_handler_dev_id);
    
    /* Disable epA to epG interrupts */
    usb.usbsm = (USBSM_EPA | USBSM_EPB | USBSM_EPC |
	USBSM_EPD | USBSM_EPE | USBSM_EPF | USBSM_EPG | USBSM_FSL |
	USBSM_SOF | USBSM_LGF | USBSM_SUS | USBSM_RSM | USBSM_RST | USBSM_FSE);

    /* Enable reset interrupts */
    usb.usbsm &= ~USBSM_RST;

    /* Enable control endpoint functionality */
    usb.epa.usbepsm &= ~USBEPASM_RBR;
    usb.epa.usbepsm &= ~USBEPASM_SUP;
    usb.epa.usbepsm &= ~USBEPASM_TBR;
    usb.epa.usbepsm &= ~USBEPASM_S;
   
    /* Enable control endpoint interrupt */
    usb.usbsm &= ~USBSM_EPA;

    /* Enable USB MISC interrupts */
    request_irq(GROUP3_IRQ_BASE + 13, usb_slave_misc_interrupt_handler, SA_SHIRQ, 
	"USB_Slave_MISC", &interrupt_handler_dev_id);
    is_initialized = 1;
    return 0;
}

int usb_slave_phys_uninit(void)
{
    /* TODO : 
     * usb_slave_disable_misc_interrupt(INTR_USBM_USBS) 
     * usb_slave_free_dma_memory(); 
     * close_dma_tx_ep_channel(); 
     * disable registers
     */
    interrupt.i3.imask &= (1<<14);
    is_initialized = 0;
    return 0;
}

void set_usb_ep_interrupt_status(int ep_serial_num, int flag, int is_mask)
{
    int status_bit = (1<<ep_serial_num);
   
    if (!flag)
    {
	usb.usbs &= ~status_bit; /* clear interrupt */
	if (is_mask)
	    usb.usbsm |= status_bit; /* mask interrupt */
	return;
    }
    usb.usbsm &= ~status_bit; /* un mask interrupt */
}

static int setup_phys_ep(endpoint_t *ep)
{
    ep_phys_info_t *phys_info = &eps_pool.ep_b;
    int i;
   
    /* Find in the pool physical endpoint with suitable information */
    for (phys_info = &eps_pool.ep_b, i = 0; i< UDC_MAX_EPS_NUM; i++, 
	phys_info++)
    {
	if (phys_info->is_taken || 
	    ep->abs_info->direction != phys_info->direction || 
	    ep->abs_info->max_pkt_size > phys_info->max_pkt_size)
	{
	    continue;
	}
	if (!ep->abs_info->is_dma || phys_info->is_dma)
	    break;
    }
    if (i == UDC_MAX_EPS_NUM)
    {
	printk(KERN_ERR __FUNCTION__": no suitable phys endpoint found\n");
	return -1;
    }
    
    if (!ep->abs_info->max_pkt_size)
        ep->abs_info->max_pkt_size = phys_info->max_pkt_size;
     phys_info->max_pkt_size = ep->abs_info->max_pkt_size;
    
    /* Setup endpoint configuration in the endpoint control register 
     * with the abstruct data, disable endpoint for now.
     */
    *(phys_info->control_reg) = USBEPxC_MODE(USB_DISABLED) |  
	USBEPxC_TYPE(ep->abs_info->type) | 
	USBEPxC_EPNUM(ep->abs_info->addr) | 
	USBEPxC_MPS(ep->abs_info->max_pkt_size);
    
    /* Enable endpoint interrupt in the USB status register */
    if (!ep->abs_info->is_dma)
        set_usb_ep_interrupt_status(phys_info->serial_num, ON, 1);
    
    /* Enable endpoint "Buffer Ready" interrupt at the endpoint status mask
     * register
     */
    if (ep->abs_info->direction == RG_USB_OUT)
       *(phys_info->status_mask_reg) &= ~(USBEPxSM_BR);
    phys_info->is_taken = 1;
    ep->phys_info = (void *)phys_info;
    
    /* Only transmit endpoint transfer is done by DMA */
    if (ep->abs_info->direction == RG_USB_IN && ep->abs_info->is_dma)
    {
	/* Disable  "Buffer Ready"- the transfer is not interrupt driven */
        *(phys_info->status_mask_reg) |= (USBEPxSM_BR);
	ep->abs_info->is_interrupt = USB_NO_INTERRUPT;
	idt_tx_dma_ep = ep;
    }

    /* All other endpoints transfers are interrupt driven */
    else 
    {
	if (ep->abs_info->direction == RG_USB_IN)
	    ep->abs_info->is_interrupt = USB_INTERRUPT_ON_TX_START_PKT;
	else
	    ep->abs_info->is_interrupt = USB_INTERRUPT_ON_RX_START_PKT;
    }
    return 0;
}

/* This function will be called between changing configuration to inititialize
 * all endpoints except for the control endpoint.
 */
static void init_device_endpoints(void)
{
    ep_phys_info_t *ep;
    int i;
    
    for (i = 0, ep = &eps_pool.ep_b; i< UDC_MAX_EPS_NUM && ep; ep++, i++)
    {
	ep->is_taken = 0;	
        (*ep->status_mask_reg) = MASK_ALL_EP_INTERRUPTS; 
	
	/* Disable endpoint */
        (*ep->control_reg) &= ~(USBEPxC_MODE(USB_ENABLED)); 
	(*ep->control_reg) |= USBEPxC_MODE(USB_DISABLED);
    }
}

/* All hardware endpoints should be configured before the device is configured,
 * So after getting from logical layer the needed endpoints setup and setting 
 * it in the hardware, we need to set the rest of the endpoints with dummy
 * values
 */
static void setup_dummy_eps(void)
{
    ep_phys_info_t *ep;
    int i;
    __u8 curr_max_ep_num = 1;
    
    /* Find the higher address number that was given to an active endpoint */
    for (i = 0, ep = &eps_pool.ep_b; i< UDC_MAX_EPS_NUM && ep; ep++, i++)
    {
	if (ep->is_taken)
	{  
	    if (curr_max_ep_num < (((*(ep->control_reg)) & 0xf0) >> 4))
		curr_max_ep_num = (((*(ep->control_reg)) & 0xf0) >> 4) + 1;
	}
    }

    /* For every endpoint that is should not be active in the current
     * configuration, configure the hardware with dummy values (uniq address 
     * number) and disable the endpoint.
     */
    for (i = 0, ep = &eps_pool.ep_b; i< UDC_MAX_EPS_NUM && ep; ep++, i++)
    {
	if (ep->is_taken)
	    continue;
        
	/* Setup endpoint configuration in the endpoint control register */
	*(ep->control_reg) = USBEPxC_MODE(USB_DISABLED) | 
	    USBEPxC_TYPE(RG_USB_BULK) | USBEPxC_EPNUM(curr_max_ep_num++) | 
	    USBEPxC_MPS(16);

	/* Disable endpoint interrupt in the USB status register */
	set_usb_ep_interrupt_status(ep->serial_num, OFF, 1);
    }
}

int usb_slave_init_eps_phys_info(endpoint_t *ctrl_ep, endpoint_t *ep_array,
    int array_size)
{
    int i;
    endpoint_t *ep = NULL;
   
    /* Only on first call */
    if (ctrl_ep)
    {
	if (!ctrl_ep->abs_info)
	{
	    printk(KERN_ERR __FUNCTION__":control endpoint software info is empty\n");
	    return -1;
	}
	ctrl_ep->abs_info->is_interrupt = USB_INTERRUPT_ON_TX_START_PKT | 
	    USB_INTERRUPT_ON_RX_START_PKT;
	ctrl_ep->phys_info = (void *)(&eps_pool.ep_a);
    }
    init_device_endpoints();
    for (i = 0, ep = ep_array; i< array_size; i++, ep++)
    {
	if (!ep || !ep->abs_info)
	{
	    printk(KERN_ERR __FUNCTION__": invalid endpoint, i = %d\n", i); 
	    return -1;
	}
	if (setup_phys_ep(ep) < 0)
	{
	    printk(KERN_WARNING "failed getting phys endpoint, ep num = %d\n", 
		ep->abs_info->addr);
	    return -1;
	}
    }
    setup_dummy_eps();
    return 0;
}

int usb_slave_enable_ep(endpoint_t *ep)
{
    ep_phys_info_t *phys_info;
   
    if (!ep || !ep->phys_info)
	return -1;
    phys_info = (ep_phys_info_t *)ep->phys_info;
    *(phys_info->control_reg) &= ~(USBEPxC_MODE(USB_DISABLED));
    *(phys_info->control_reg) |= USBEPxC_MODE(USB_ENABLED);
    return 0;
}

int usb_slave_disable_ep(endpoint_t *ep)
{
    ep_phys_info_t *phys_info;
    
    if (!ep || !ep->phys_info)
	return - 1;
    phys_info = (ep_phys_info_t *)ep->phys_info;
    *(phys_info->control_reg) &= ~(USBEPxC_MODE(USB_ENABLED)); 
    *(phys_info->control_reg) |= USBEPxC_MODE(USB_DISABLED);
    ep->abs_info->mode = USB_DISABLED;
    return 0;
}

/* Stall the endpoint */
void usb_slave_stall_ep(endpoint_t *ep)
{
    ep_phys_info_t *ep_phys;
    
    if (!ep || !ep->abs_info || !ep->phys_info)
	return;
    ep_phys = (ep_phys_info_t *)ep->phys_info;
    *(ep_phys->status_reg) &= USBEPxS_S;
    ep->abs_info->transfer_mode = USB_STALL;
}


/* usb_dma_start_transfer: 
 * Enable DMA relevant interrupts by unmasking there bits in the DMA status 
 * (mask) register.
 * Starts descriptor proccesing by setting DPTR (Descriptor Pointer) register
 * to the first descriptor in the descriptors list and by setting NDPTR (Next
 * Descriptor Pointer) to 0.
 */
static int usb_dma_start_transfer(void)
{

    /* Setting STATUS MASK register to unmask the needed interrupts */
    dma.ch14.dmasm &= ~DMAxS_H; /* halt */
    dma.ch14.dmasm &= ~DMAxS_E; /* error */

    /* Check CONTROL register RUN bit to insure that the DMA controller is 
     * not allready running
     */
    if ((dma.ch14.dmac) & DMAxC_RUN)
    {
	printk(__FUNCTION__": Can't send new packet, DMA controller"
	    "is already running, DMA channel 14, control register content "
	    "= 0x%08x\n", dma.ch14.dmac);
	return -1;
    }
    usb_dma_current_transfer_state = USB_TRANSFER_STARTED;

    flush_cache_all();
    /* Setting NDPTR and DPTR register to start DMA descriptor processing of
     * the current logic descriptor.
     */
    dma.ch14.dmandptr = 0;  
    dma.ch14.dmadptr = PHYS_ADDR(((unsigned int)first_descriptor));
    return 0;
}

/* usb_on_dma_transfer_end:
 * Called from interrupt handler when DMA haltded (Stop running)
 * Call to usb_dma_send_data_end in the upper layer.
 */
static void usb_on_dma_transfer_end(endpoint_t *ep)
{   
    usb_dma_current_transfer_state = USB_TRANSFER_ENDED;
    first_descriptor = NULL;
    desc_count = 0;
    ep->abs_info->is_busy = 0;
    ep->abs_info->end_func(ep);
}

/* usb_dma_ep_g_interrupt_handler: 
 * DMA channel 14 (USB tx) interrupt handler.
 */
static void usb_dma_ep_g_interrupt_handler(endpoint_t *ep)
{
    __u32 status = dma.ch14.dmas;

    /* Halt interrupt: */
    if ((status & DMAxS_H) == DMAxS_H)
    {
	USB_DPRINTK(__FUNCTION__": Got USB DMA Halt interrupt\n");
	if (usb_dma_current_transfer_state == USB_TRANSFER_STARTED)
	    usb_on_dma_transfer_end(ep);

	/* clear halt bit in status register */
	dma.ch14.dmas &= ~DMAxS_H;

	/* Mask halt interrupt until the next packet is being sent */ 
	dma.ch14.dmasm |= DMAxS_H;
    }

    /* Error interrupt: */
    if ((status & DMAxS_E) == DMAxS_E)
    {
	USB_DPRINTK(__FUNCTION__": Got USB DMA Error interrupt\n");
	ep->abs_info->end_arg1 = USB_TRANSFER_ERROR;

	/* Stop current descriptor proccesing */
	dma.ch14.dmac &= ~DMAxC_RUN;

	/* clear Error bit in status register */
	dma.ch14.dmas &= ~DMAxS_E;
    }

    /* Finish interrupt :*/
    if ((status & DMAxS_F) == DMAxS_F)
    {
	/* clear finished interrupt on status register */
	dma.ch14.dmas &= ~DMAxS_F;
	
	/* Last descriptor in the list finished */
	if (!((dma.ch14.dmac) & DMAxC_RUN) || !(dma.ch14.dmadptr))
	{
	  printk(__FILE__":"__FUNCTION__": %d\n",__LINE__);
	    ep->abs_info->end_arg1 = USB_TRANSFER_SUCCESS;

	    /* Stop current descriptor proccesing */
	    dma.ch14.dmac &= ~DMAxC_RUN;

	    /* Mask finished interrupt until the next packet is being sent */ 
	    dma.ch14.dmasm |= DMAxS_F; 
	}
    }

    /* Chain interrupt: */ 
    if ((status & DMAxS_C) == DMAxS_C)
    {
	dma.ch14.dmas &= ~(DMAxS_C);
	dma.ch14.dmasm |= DMAxS_C; 
    }

    /* Done interrupt: */
    if ((status & DMAxS_D) == DMAxS_D)
    {
	dma.ch14.dmas &= ~DMAxS_D;
	dma.ch14.dmasm |= DMAxS_D; 
    }
}

/* add_next_dma_pkt:
 * copy packet to next buffer in the DMA buffers list;
 * Insert this buffer phisical address to the next DMA descriptor CA (Current
 * Address) field; Link this descriptor to the privious one in the list.
 */
static DMA_DESCRIPTOR *add_next_dma_pkt(__u8 *buff, int size, 
    DMA_DESCRIPTOR *previous_desc)
{
    __u8 *data;
    DMA_DESCRIPTOR *new;
	
    if (desc_count >= NUM_DMA_DESCS)
    {
	printk(KERN_WARNING __FUNCTION__": Too many descriptors has been "
	    "used\n");
	return NULL;
    }
    if (!dma_descriptors_list)
    {
	printk(KERN_WARNING __FUNCTION__": DMA descriptor was not "
	    "initialized\n");
	return NULL;
    }
    if (!dma_buffers_list)
    {
	printk(KERN_WARNING __FUNCTION__": DMA buffers memory was not "
	    "initialized\n");
	return NULL;
    }
    new = dma_descriptors_list + desc_count;
    data = dma_buffers_list + (USB_DMA_BUFF_SIZE * desc_count); 
    desc_count++;
    memcpy(data, buff, size);

    dma_cache_wback_inv((unsigned long)data,size);

    new->ca = PHYS_ADDR((unsigned int)data);
    new->control = DMA_COUNT(size);
    new->devcs = USB_DEVCS_FD | USB_DEVCS_LD;
    new->link = 0;
    if (previous_desc)
    {
	previous_desc->link = (DMA_DESCRIPTOR *) 
	    PHYS_ADDR(((unsigned int)new));
    }
    if (!first_descriptor)
	first_descriptor = new;
    return new;
}

static int fill_dma_descriptor_and_buffers(__u8 *buff, int len, 
    int max_pkt_size)
{
    int pkt_size = 0;
    int i = 0; 
    DMA_DESCRIPTOR *current_desc = NULL; 

    /* Fragmnet the data and fill a descriptor buffer */
    while (i < len)
    {
	pkt_size = (len - i < max_pkt_size) ? len - i : max_pkt_size;
	current_desc = add_next_dma_pkt(buff + i, pkt_size, current_desc);
	if (!current_desc)
	    return -1;
	i += pkt_size;	
    }

    /* Send one byte zero message to indicate the host end of transfer.
     * (end of transfer indication is a pkt of length less than max_pkt_size)
     */
    if (pkt_size == max_pkt_size) 
    {
	current_desc = add_next_dma_pkt("0", 1, current_desc);
	if (!current_desc)
	    return -1;
    }
    return 0;
}

int usb_slave_start_dma_transfer(endpoint_t *ep)
{
    if (fill_dma_descriptor_and_buffers(ep->abs_info->buff + 
	ep->abs_info->buff_pos, ep->abs_info->buff_size, 
	ep->abs_info->max_pkt_size) < 0)
    {
	return -1;
    }
    return usb_dma_start_transfer();
}

void usb_slave_terminate_current_transfers(void)
{
    idt_tx_dma_ep->abs_info->end_arg1 = USB_TRANSFER_ERROR;
    dma.ch14.dmac &= ~DMAxC_RUN;
}

int usb_slave_on_reset_cmd(void)
{
    /* flush all hardware fifos */
    usb.epb.usbepc |= USBEPxC_FSH; 
    usb.epc.usbepc |= USBEPxC_FSH; 
    usb.epd.usbepc |= USBEPxC_FSH; 
    usb.epe.usbepc |= USBEPxC_FSH; 
    usb.epf.usbepc |= USBEPxC_FSH; 
    usb.epg.usbepc |= USBEPxC_FSH;

    /* disable all endpoints */
    usb.epb.usbepc &= ~(USBEPxC_MODE(USB_ENABLED)); 
    usb.epb.usbepc |= USBEPxC_MODE(USB_DISABLED); 
    usb.epc.usbepc &= ~(USBEPxC_MODE(USB_ENABLED)); 
    usb.epc.usbepc |= USBEPxC_MODE(USB_DISABLED); 
    usb.epd.usbepc &= ~(USBEPxC_MODE(USB_ENABLED)); 
    usb.epd.usbepc |= USBEPxC_MODE(USB_DISABLED); 
    usb.epe.usbepc &= ~(USBEPxC_MODE(USB_ENABLED));; 
    usb.epe.usbepc |= USBEPxC_MODE(USB_DISABLED); 
    usb.epf.usbepc &= ~(USBEPxC_MODE(USB_ENABLED)); 
    usb.epf.usbepc |=  USBEPxC_MODE(USB_DISABLED); 
    usb.epg.usbepc &= ~(USBEPxC_MODE(USB_ENABLED)); 
    usb.epg.usbepc |= (USBEPxC_MODE(USB_DISABLED)); 

    /* Mask Buffer Ready interrupt */
    usb.epb.usbepsm |= (USBEPxSM_BR);
    usb.epc.usbepsm |= (USBEPxSM_BR);
    usb.epd.usbepsm |= (USBEPxSM_BR);
    usb.epe.usbepsm |= (USBEPxSM_BR);
    usb.epf.usbepsm |= (USBEPxSM_BR);
    usb.epg.usbepsm |= (USBEPxSM_BR);
    return 0;
}

int usb_slave_handle_non_control_ep(endpoint_t *ep, __u8 *buff_p, int size)
{
    __u32 status;
    int i = 0; 
    ep_phys_info_t *phys_ep;
   
    if (!ep || !ep->abs_info || !ep->phys_info || ep->abs_info->is_dma || 
	(ep->abs_info->type == RG_USB_ISO))
    {
	printk(KERN_ERR __FUNCTION__": called with invalid endpoint\n");
	return -1;
    }
    phys_ep = (ep_phys_info_t *)(ep->phys_info);
    if (!buff_p || size > ep->abs_info->max_pkt_size || 
	(ep->abs_info->buff_size < size))
    {
	printk(KERN_ERR __FUNCTION__": called with invalid buffer \n"
	    "information ep num = %d\n", ep->abs_info->addr);

	/* Clear BR interrupt */
       *(phys_ep->status_reg) &= ~(USBEPxS_BR);
       if (ep->abs_info->direction == RG_USB_IN)
       {
	   /* Mask Endpoint interrupt for now */
           set_usb_ep_interrupt_status(phys_ep->serial_num, OFF, 1);
       }
       size = 0;
       goto Exit;
    }
    status = *(phys_ep->status_reg);
    if (ep->abs_info->transfer_mode == USB_NAK)
    {  
	printk(KERN_INFO __FUNCTION__": NAK mode\n");
	return 0;
    }
    
    /* Endpoint is not in "Buffer ready" state */
    if (!(status & USBEPxS_BR))
    {
	printk(KERN_WARNING __FUNCTION__": endpoint is not in buffer ready "
	    "state\n");
	size = 0;
	goto Exit;
    }

    /* Transmit (IN) endpoint */
    if (ep->abs_info->direction == RG_USB_IN)
    {
	if (!size)
	{
	    /* Clear BR interrupt, Mask endpoint interrupt for now */
            *(phys_ep->status_reg) &= ~(USBEPxS_BR);
	    usb_slave_disable_transfer(ep);
	    return 0;
	}
	/* Send data */
	do
	{
	    *(phys_ep->data_reg) = buff_p[i++];
	} while (*(phys_ep->buffer_status_reg) < size);

	/* Clear interrupt */
	*(phys_ep->status_reg) &= ~(USBEPxS_BR);
	goto Exit;
    }

    /* Receive (OUT) endpoint */
    size = *(phys_ep->buffer_status_reg);
    
    /* Data received */
    if (size > 0)
    {
	i = 0;
	do
	{
	    buff_p[i++] = *(phys_ep->data_reg);
	} while (i < size);
    }
     
    /* Clear BR interrupt */
    *(phys_ep->status_reg) &= ~(USBEPxS_BR);
Exit:
	ep->abs_info->buff_pos += size;
	ep->abs_info->buff_size -= size;
	ep->abs_info->ep_operation((void *)ep);
	return size;
}

int usb_slave_handle_control_ep(endpoint_t *ep, usb_ep_dir_t dir, 
    control_phase_t req_phase)
{
    int size = 0;
    __u32 status = 0;
    int i = 0;
    int mask_tbr = 0;
    int rc = 0;

    if (!ep || !ep->abs_info || !ep->phys_info || ep->abs_info->addr ||
	(ep->abs_info->type != RG_USB_CONTROL) || !ep->abs_info->buff)
    {
	USB_DPRINTK(KERN_INFO __FUNCTION__": endpoint information "
	    " is not initialized yet\n");
	mask_tbr = 1;
	rc = -1;
    }
    status = usb.epa.usbeps;

    /* Receive buffer is ready */ 
    if ((status & USBEPAS_RBR))
    {
	if ((status & USBEPAS_SUP))
	{
	    usb.epa.usbeps &= ~USBEPAS_SUP;
	}
	if (req_phase ==  DATA_RECEIVE || req_phase == SETUP_TOKEN)
	{
	    size = usb.epa.usbeprbs;
	    if (size < 0 || size > ep->abs_info->buff_size) 
	    {
		printk(KERN_ERR __FUNCTION__":Invalid RBR size:\nsize = %d, "
		    "expected left size =  %d\n", size, 
		    ep->abs_info->buff_size );
		rc = -1;
		goto Exit;
	    }
	    if (size > 0)
	    {
		do
		{
		    ep->abs_info->buff[ep->abs_info->buff_pos + i] = 
			usb.epa.usbeprd;
		    i++;
		} while (i < size);
	    }
	    ep->abs_info->buff_pos += size;
	    ep->abs_info->buff_size -= size;
	}
	else if (req_phase != SETUP_TOKEN && dir == RG_USB_IN)
	{
	    size = usb.epa.usbeprbs;

	    /* host indicate end of transfer - ACK */
	    if (!size)
		ep->abs_info->buff_size = 0;
	}
    }

    /* transmit buffer is ready */
    if (status & USBEPAS_TBR)
    {
	if (req_phase != DATA_TRANSMIT && req_phase != DATA_SEND_ZERO_LEN_PKT)
	{
	    mask_tbr = 1;
	    goto Exit;
	}

	/* Send answer to an IN request */
	if (dir == RG_USB_IN)
	{
	    if (ep->abs_info->buff_size > 0)
	    {
		/* fragment data into USB control pipe packets */ 
		size = (ep->abs_info->buff_size > ep->abs_info->max_pkt_size) ? 
		    ep->abs_info->max_pkt_size :  ep->abs_info->buff_size;


		/* Copy current packet into control endpoint FIFO */
		do
		{
		    usb.epa.usbeptd = 
			ep->abs_info->buff[ep->abs_info->buff_pos++];
		}while (usb.epa.usbeptbs < size);

		/* Update "left to write" size */  
		ep->abs_info->buff_size -= size;

		/* Clear TBR interrupt */
		usb.epa.usbeps &= ~USBEPAS_TBR;
		goto Exit;
	    }  

	    /* finished sending current message so mask interrupt until 
	     * next one
	     */
	    if (req_phase != DATA_SEND_ZERO_LEN_PKT)
	    {
		mask_tbr = 1;
		usb.epa.usbeps &= ~USBEPAS_TBR;
		goto Exit;
	    }
	    usb.epa.usbeps &= ~USBEPAS_TBR;
	}
	/* Send zero length packet to acknowledg to an OUT request */
	if (dir == RG_USB_OUT)
	{
	    if (req_phase != DATA_SEND_ZERO_LEN_PKT)
		mask_tbr = 1;
	    usb.epa.usbeps &= ~USBEPAS_TBR;	
	}
    }
Exit:
    if (mask_tbr)
	usb.epa.usbepsm |= USBEPASM_TBR;
    ep->abs_info->ep_operation((void *)ep);
    if (status & USBEPAS_RBR)
	usb.epa.usbeps &= ~USBEPAS_RBR;
    return rc;
}

void usb_slave_enable_transfer(endpoint_t *ep)
{
    ep_phys_info_t *phys_info;
  
    if (!ep || !ep->abs_info || !ep->abs_info->is_interrupt)
    {
	printk(KERN_ERR __FUNCTION__": invalid endpoint information\n");
	return;
    }
    
    /* Control endpoint */
    if (!ep->abs_info->addr)
    {
	usb.epa.usbepsm &= ~(USBEPASM_TBR);
	return;
    }

    phys_info = (ep_phys_info_t *)ep->phys_info;
    if (!phys_info)
    {
	printk(KERN_ERR __FUNCTION__": invalid phys endpoint information\n");
	return;
    }
    /* Non-control endpoint */
    set_usb_ep_interrupt_status(phys_info->serial_num, ON, 1);
    *(phys_info->status_mask_reg) &= ~(USBEPxSM_BR);
}

void usb_slave_disable_transfer(endpoint_t *ep)
{
    ep_phys_info_t *phys_info;
    
    if (!ep || !ep->abs_info || !ep->abs_info->is_interrupt)
	return;
    phys_info = (ep_phys_info_t *)ep->phys_info;
    *(phys_info->status_mask_reg) |= USBEPxSM_BR;
    set_usb_ep_interrupt_status(phys_info->serial_num, OFF, 1);
}

void usb_slave_handle_get_status_req(__u8 recipient, endpoint_t *ep, 
    __u8 **status_resp_p)
{
    return;
}

int usb_slave_handle_clear_feature_req(__u8 feature_selector, endpoint_t *ep)
{
    printk(KERN_WARNING __FUNCTION__": stub should never be called\n");
    return -1;
}

int usb_slave_handle_set_feature_req(__u8 feature_selector, endpoint_t *ep)
{
    printk(KERN_WARNING __FUNCTION__": stub should never be called\n");
    return -1;
}

int usb_slave_handle_set_address_req(__u32 addr)
{
    printk(KERN_WARNING __FUNCTION__": stub should never be called\n");
    return -1;
}

