/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#include <linux/sched.h>
#include <asm/current.h>
#include <asm/byteorder.h>
#include "rndis.h"
#include "../usb_slave_vendor_info.h"

/* RNDIS control messages types */
/* get */
#define REMOTE_NDIS_INITIALIZE_MSG 0x00000002
#define REMOTE_NDIS_HALT_MSG 0x00000003
#define REMOTE_NDIS_QUERY_MSG 0x00000004
#define REMOTE_NDIS_SET_MSG 0x00000005
#define REMOTE_NDIS_RESET_MSG 0x00000006
#define REMOTE_NDIS_INDICATE_STATUS_MSG 0x00000007
#define REMOTE_NDIS_KEEPALIVE_MSG 0x00000008

/* return */
#define REMOTE_NDIS_INITIALIZE_CMPLT 0x80000002
#define REMOTE_NDIS_QUERY_CMPLT 0x80000004
#define REMOTE_NDIS_SET_CMPLT 0x80000005
#define REMOTE_NDIS_RESET_CMPLT 0x80000006
#define REMOTE_NDIS_KEEPALIVE_CMPLT 0x80000008
#define RESPONSE_AVAILABLE 0x00000001

/* RNDIS control messages sizes */
#define REMOTE_NDIS_INIT_MSG_SIZE 0x00000018
#define REMOTE_NDIS_INIT_CMPLT_SIZE 0x00000034
#define REMOTE_NDIS_HALT_MSG_SIZE 0x0000000C 
#define REMOTE_NDIS_QUERY_MSG_SIZE  0x0000001C
#define REMOTE_NDIS_QUERY_CMPLT_SIZE 0x00000018
#define REMOTE_NDIS_SET_MSG_SIZE 0x0000001C
#define REMOTE_NDIS_SET_CMPLT_SIZE 0x00000010
#define REMOTE_NDIS_RESET_MSG_SIZE 0x0000000C
#define REMOTE_NDIS_RESET_CMPLT_SIZE 0x00000010
#define REMOTE_NDIS_STATUS_MSG_MAX_SIZE 
#define REMOTE_NDIS_KEEPALIVE_MSG_SIZE 0x0000000C
#define REMOTE_NDIS_KEEPALIVE_CMPLT_SIZE 0x00000010
#define REMOTE_NDIS_INDICATE_STATUS_MSG_SIZE 0x00000014

/* RNDIS control messages details */
/* response */
#define RNDIS_RESPONSE_BLOCK_MAX_SIZE 1024
#define RNDIS_RESPONSE_BLOCK_MIN_SIZE 16
#define RNDIS_RESPONSE_DATA_MAX_SIZE (RNDIS_RESPONSE_BLOCK_MAX_SIZE -  \
	RNDIS_RESPONSE_BLOCK_MIN_SIZE)
#define RNDIS_RESPONSE_QUERY_MSG_DATA_MIN_OFFSET 16
#define RNDIS_INDICATE_STATUS_INFO_MIN_OFFSET 12

/* request */
#define RNDIS_REQUEST_BLOCK_MAX_SIZE MAX_CTRL_DATA_PKT_LEN 
#define RNDIS_REQUEST_BLOCK_MIN_SIZE 12
#define RNDIS_REQUEST_DATA_MAX_SIZE (RNDIS_REQUEST_BLOCK_MAX_SIZE -  \
	RNDIS_REQUEST_BLOCK_MIN_SIZE)

/* RNDIS version (part of initialization messages) */
#define RNDIS_DEVICE_MAJOR_VERSION 0x0001
#define RNDIS_DEVICE_MINOR_VERSION 0x0000

/* RNDIS other initialization details */
#define RNDIS_DF_CONNECTIONLESS 0x00000001 
#define RNDIS_DF_CONNECTION_ORIENTED 0x00000002
#define RNDIS_DF  RNDIS_DF_CONNECTIONLESS 
#define MAX_PACKETS_PER_TRANSFER 0x00000001

/* RNDIS data messages details */
#define RNDIS_PACKET_MSG 0x1
#define RNDIS_PACKET_MSG_MSG_LEN_OFFSET 4
#define RNDIS_PACKET_MSG_HDR_OFFSET 8
#define RNDIS_PACKET_MSG_DATA_LEN_OFFSET 12
#define RNDIS_PACKET_MSG_DATA_OFFSET_OFFSET 8
#define RNDIS_PACKET_MSG_MIN_DATA_OFFSET 36

#define RNDIS_ERR_NO_MEMORY 1

/* RNDIS Oid detailes */
#define RNDIS_MAX_FRAME_SIZE 1520
#define RNDIS_MAX_TOTAL_SIZE 1558
#define RNDIS_DATA_BLOCK_SIZE 1514
#define RNDIS_CURRENT_SUPPORTED_OIDS_NUM 35
#define RNDIS_CURRENT_MEDIA_SUPPORTED_NUM 1

/* RNDIS Packet Filter Bits (OID_GEN_CURRENT_PACKET_FILTER) */
#define NDIS_PACKET_TYPE_DIRECTED 0x00000001
#define NDIS_PACKET_TYPE_MULTICAST 0x00000002
#define NDIS_PACKET_TYPE_ALL_MULTICAST 0x00000004
#define NDIS_PACKET_TYPE_BROADCAST 0x00000008
#define NDIS_PACKET_TYPE_SOURCE_ROUTING	0x00000010
#define NDIS_PACKET_TYPE_PROMISCUOUS 0x00000020
#define NDIS_PACKET_TYPE_SMT 0x00000040
#define NDIS_PACKET_TYPE_ALL_LOCAL 0x00000080
#define NDIS_PACKET_TYPE_GROUP 0x00001000
#define NDIS_PACKET_TYPE_ALL_FUNCTIONAL	0x00002000
#define NDIS_PACKET_TYPE_FUNCTIONAL 0x00004000
#define NDIS_PACKET_TYPE_MAC_FRAME 0x00008000

/* RNDIS protocol option bits (OID_GEN_PROTOCOL_OPTIONS) */
#define NDIS_PROT_OPTION_ESTIMATED_LENGTH 0x00000001
#define NDIS_PROT_OPTION_NO_LOOPBACK 0x00000002
#define NDIS_PROT_OPTION_NO_RSVD_ON_RCVPKT 0x00000004

/* RNDIS MAC option bits (OID_GEN_MAC_OPTIONS) */
#define NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA 0x00000001
#define NDIS_MAC_OPTION_RECEIVE_SERIALIZED 0x00000002
#define NDIS_MAC_OPTION_TRANSFERS_NOT_PEND 0x00000004
#define NDIS_MAC_OPTION_NO_LOOPBACK 0x00000008
#define NDIS_MAC_OPTION_FULL_DUPLEX 0x00000010
#define	NDIS_MAC_OPTION_8021P_PRIORITY 0x00000040

/* connection states for monitor */
#define  RNDIS_USB_CONNECTED 0
#define  RNDIS_SENT_KEEPALIVE_MSG_3 3
#define RNDIS_USB_DISCONNECTED 7

#define MAX_RETRY 3
#define CFG_DESC_TOTAL_LEN 0x43
#define RNDIS_ENDPOINTS_NUM 3
#define RNDIS_NOTIFICATION_EP_MAX_SIZE 0x08
#define RNDIS_DATA_EP_MAX_SIZE 0x40
#define USB_DEV_VENDOR_VERSION 0x0100

/* Specific rndis descriptors sizes */
#define SIZEOF_COMM_HDR_DESC 0x05
#define SIZEOF_COMM_MNG_DESC 0x05
#define SIZEOF_COMM_ABS_DESC 0x04
#define SIZEOF_COMM_FUNC_DESC 0x05

/* RNDIS status types */
typedef enum {
    RNDIS_STATUS_SUCCESS = 0x00000000,
    RNDIS_STATUS_FAILURE = 0xC0000001,
    RNDIS_STATUS_INVALID_DATA = 0xC0000015,
    RNDIS_STATUS_NOT_SUPPORTED = 0xC00000BB,
    RNDIS_STATUS_MEDIA_CONNECT = 0x4001000B,
    RNDIS_STATUS_MEDIA_DISCONNECT = 0x4001000C
} rndis_status_t;

/* Local structures*/
/* RNDIS hardware status types */
typedef enum {
    RNDIS_HARDWARE_STATUS_READY = 0x00000000,
    RNDIS_HARDWARE_STATUS_INITIALIZING = 0x00000001,
    RNDIS_HARDWARE_STATUS_RESET = 0x00000002,
    RNDIS_HARDWARE_STATUS_CLOSING = 0x00000003,
    RNDIS_HARDWARE_STATUS_NOT_READY = 0x00000004
} hardware_status_t;

/* RNDIS mandatory OIDs list */
typedef enum {
    /* Required OIDs */
    OID_GEN_SUPPORTED_LIST = 0x00010101,
    OID_GEN_HARDWARE_STATUS = 0x00010102,
    OID_GEN_MEDIA_SUPPORTED = 0x00010103,
    OID_GEN_MEDIA_IN_USE = 0x00010104,
    OID_GEN_MAXIMUM_LOOKAHEAD = 0x00010105,
    OID_GEN_MAXIMUM_FRAME_SIZE = 0x00010106,
    OID_GEN_LINK_SPEED = 0x00010107,
    OID_GEN_TRANSMIT_BUFFER_SPACE = 0x00010108,
    OID_GEN_RECEIVE_BUFFER_SPACE = 0x00010109,
    OID_GEN_TRANSMIT_BLOCK_SIZE = 0x0001010A,
    OID_GEN_RECEIVE_BLOCK_SIZE = 0x0001010B,
    OID_GEN_VENDOR_ID = 0x0001010C,
    OID_GEN_VENDOR_DESCRIPTION = 0x0001010D,
    OID_GEN_CURRENT_PACKET_FILTER = 0x0001010E,
    OID_GEN_CURRENT_LOOKAHEAD = 0x0001010F,
    OID_GEN_DRIVER_VERSION = 0x00010110,
    OID_GEN_MAXIMUM_TOTAL_SIZE = 0x00010111,
    OID_GEN_PROTOCOL_OPTIONS = 0x00010112,
    OID_GEN_MAC_OPTIONS = 0x00010113,
    OID_GEN_MEDIA_CONNECT_STATUS = 0x00010114,
    OID_GEN_MAXIMUM_SEND_PACKETS = 0x00010115,
    OID_GEN_VENDOR_DRIVER_VERSION = 0x00010116,
    
    /* Required statistics */
    OID_GEN_XMIT_OK = 0x00020101,
    OID_GEN_RCV_OK = 0x00020102,
    OID_GEN_XMIT_ERROR = 0x00020103,
    OID_GEN_RCV_ERROR = 0x00020104,
    OID_GEN_RCV_NO_BUFFER = 0x00020105,
   
    /* 802.3 Objects (Ethernet) */
    OID_802_3_PERMANENT_ADDRESS = 0x01010101,
    OID_802_3_CURRENT_ADDRESS = 0x01010102,
    OID_802_3_MULTICAST_LIST = 0x01010103,
    OID_802_3_MAXIMUM_LIST_SIZE = 0x01010104,
    OID_802_3_MAC_OPTIONS = 0x01010105,
    OID_802_3_RCV_ERROR_ALIGNMENT = 0x01020101,
    OID_802_3_XMIT_ONE_COLLISION  = 0x01020102,
    OID_802_3_XMIT_MORE_COLLISIONS = 0x01020103
} oid_t;

/* RNDIS medium list */
typedef enum {
    RNDIS_MEDIUM_802_3 = 0x00000000,
    RNDIS_MEDIUM_802_5 = 0x00000001,
    RNDIS_MEDIUM_FDDI = 0x00000002,
    RNDIS_MEDIUM_WAN = 0x00000003,
    RNDIS_MEDIUM_LOCAL_TALK = 0x00000004,
    RNDIS_MEDIUM_DIX = 0x00000005,
    RNDIS_MEDIUM_ARCENT_RAW = 0x00000006,
    RNDIS_MEDIUM_ARCENT_878_2 = 0x00000007,
    RNDIS_MEDIUM_ATM = 0x00000008,
    RNDIS_MEDIUM_WIRELESS_WAN = 0x00000009,
    RNDIS_MEDIUM_IRDA = 0x0000000A,
    RNDIS_MEDIUM_BPC = 0x0000000B,
    RNDIS_MEDIUM_CO_WAN = 0x0000000C,
    RNDIS_MEDIUM_1394 = 0x0000000D,
    RNDIS_MEDIUM_MAX /* Not a real medium, defined as an upper-bound */
} medium_t; 

#ifdef CONFIG_USB_CONN_MONITOR
typedef struct {
    char name[IFNAMSIZ];
    int kernel_thread_id;
    void (*start_entry)(void);
    struct task_struct *tsk;
    struct semaphore sem;
} task_data_t;
#endif

/* RNDIS encapsulation for data message */
typedef struct {
    __u32 MessageType;
    __u32 MessageLength;
    __u32 DataOffset;
    __u32 DataLength;
    __u32 OOBDataOffset;
    __u32 OOBDataLength;
    __u32 NumOOBDataElements;
    __u32 PerPacketInfoOffset;
    __u32 PerPacketInfoLength;
    __u32 VcHandle;
    __u32 Reserved;
} remote_ndis_pkt_msg_t;

/* RNDIS encapsulation for control request */
typedef struct {
    __u32 req_type;
    __u32 req_length;
    __u32 req_id;
} rndis_req_hdr_t;

/* RNDIS encapsulation for control response */
typedef struct {
    __u32 resp_type;
    __u32 resp_length;
    __u32 resp_id;
    rndis_status_t resp_status;
    __u32 resp_data[RNDIS_RESPONSE_DATA_MAX_SIZE/4];
} rndis_resp_t;

/* RNDIS diagnostic for initiate indicate status */
typedef struct {
    rndis_status_t DiagStatus;
    __u32 ErrorOffset;
} rndis_diagnostic_info_t;

/* RNDIS indicate status message structure */
typedef struct {
    __u32 MessageType;
    __u32 MessageLength;
    rndis_status_t Status;
    __u32 StatusBufferLength;
    __u32 StatusBufferOffset;
    rndis_diagnostic_info_t info;
    __u8 StatusBuffer[RNDIS_REQUEST_BLOCK_MAX_SIZE];
} rndis_status_msg_t;

/* Local functions declaretion */
static __u32 rndis_handle_control_request(__u8 *d_buff, __u8 *s_buff, 
   int req_size);
static int rndis_receive_pkt(__u8 *rndis_data, int size_limit);
#ifdef FIX_ZERO_LEN_PKT_BUG
static int rndis_get_data_msg_len(__u8 *msg_hdr, int hdr_size);
#endif
static void rndis_receive_data(void *context);
static int rndis_create_query_msg_response(__u32 o_id, 
    __u32 *resp_data_block, __u8 *input_data, int input_data_size);
static int rndis_set_object_id(__u32 o_id, __u8 *data, __u32 data_size);
static void rndis_initiate_msg(int msg_size, __u32 msg_type);
static void rndis_initiate_halt_message(void);
static void rndis_send_keep_alive_message(void);
static void rndis_initiate_status_indicate_msg(rndis_status_t status, 
    rndis_diagnostic_info_t *info, __u8 *buffer, __u32 buffer_length);
static int rndis_init_descriptors(__u8 *dev_desc_p, int dev_desc_size, 
    __u8 *cfg_desc_p, int cfg_desc_size, string_desc_t *prod_desc, 
    string_desc_t *mfg_desc, string_desc_t *lang_desc);
static void rndis_usb_config(__u8 cfg_num, __u32 alt_if_num, 
    endpoint_t **cfg, int *eps_num);
static void rndis_uninitialized(void);
static void rndis_handle_send_encapsulated_command_msg(__u8 *req_block, 
    __u32 req_size);
static void rndis_handle_get_encapsulated_response_msg(__u8 *resp_buff,
    __u32 *resp_size_p);
static void rndis_send_response_available_notification(void *context);
static int rndis_usb_ep_send_stub(__u8 *buff, int len, endpoint_t *ep, 
    void *context1, void *context2);
static void rndis_swap_list(__u32 *list, int num_of_elements);
static void rndis_usb_terminate_transfers_stub(void);
static void rndis_on_receive_usb_data_pkt_stub(void *context);
#ifdef CONFIG_RG_USB_SLAVE_MODULE    
static void on_rndis_closed_stub(void);
#endif

#ifdef CONFIG_USB_CONN_MONITOR
/* Task and monitor functions */
static void rndis_monitor_usb_conn_state(void);
static int rndis_on_usb_conn_state_check(int *data);
static int task_start_cb(void *context);
static int create_and_start_task(char *name, void (*task_ptr)(void));
static void task_delay(unsigned long timeout);
#endif

/* Globals variables */
/* should be exschanged between the host and device in RNDIS initialization */
int host_maxTransferSize;
int host_major_version;
int host_minor_version;

/* RNDIS Oids global variables */
static oid_t supported_oids_list[RNDIS_CURRENT_SUPPORTED_OIDS_NUM] = 
{   OID_GEN_SUPPORTED_LIST, OID_GEN_HARDWARE_STATUS, OID_GEN_MEDIA_SUPPORTED,
    OID_GEN_MEDIA_IN_USE, OID_GEN_MAXIMUM_LOOKAHEAD, OID_GEN_MAXIMUM_FRAME_SIZE,
    OID_GEN_LINK_SPEED, OID_GEN_TRANSMIT_BUFFER_SPACE, 
    OID_GEN_RECEIVE_BUFFER_SPACE,OID_GEN_TRANSMIT_BLOCK_SIZE,
    OID_GEN_RECEIVE_BLOCK_SIZE, OID_GEN_VENDOR_ID, OID_GEN_VENDOR_DESCRIPTION,
    OID_GEN_CURRENT_PACKET_FILTER, OID_GEN_CURRENT_LOOKAHEAD, 
    OID_GEN_DRIVER_VERSION, OID_GEN_MAXIMUM_TOTAL_SIZE, 
    OID_GEN_PROTOCOL_OPTIONS,
    OID_GEN_MAC_OPTIONS, OID_GEN_MEDIA_CONNECT_STATUS,
    OID_GEN_MAXIMUM_SEND_PACKETS, OID_GEN_VENDOR_DRIVER_VERSION, 
    OID_GEN_XMIT_OK, OID_GEN_RCV_OK, OID_GEN_XMIT_ERROR, OID_GEN_RCV_ERROR,
    OID_GEN_RCV_NO_BUFFER, OID_802_3_PERMANENT_ADDRESS, 
    OID_802_3_CURRENT_ADDRESS, OID_802_3_MULTICAST_LIST, 
    OID_802_3_MAXIMUM_LIST_SIZE, OID_802_3_MAC_OPTIONS, 
    OID_802_3_RCV_ERROR_ALIGNMENT, OID_802_3_XMIT_ONE_COLLISION,
    OID_802_3_XMIT_MORE_COLLISIONS };
static hardware_status_t current_hardware_status =  
    RNDIS_HARDWARE_STATUS_INITIALIZING;
static medium_t media_supported[RNDIS_CURRENT_MEDIA_SUPPORTED_NUM] = 
    { RNDIS_MEDIUM_802_3 };
static medium_t media_in_use[RNDIS_MEDIUM_MAX]= { RNDIS_MEDIUM_802_3, 
    RNDIS_MEDIUM_802_5, RNDIS_MEDIUM_FDDI, RNDIS_MEDIUM_WAN,
    RNDIS_MEDIUM_LOCAL_TALK, RNDIS_MEDIUM_DIX, RNDIS_MEDIUM_ARCENT_RAW,
    RNDIS_MEDIUM_ARCENT_878_2, RNDIS_MEDIUM_ATM, RNDIS_MEDIUM_WIRELESS_WAN,
    RNDIS_MEDIUM_IRDA, RNDIS_MEDIUM_BPC, RNDIS_MEDIUM_CO_WAN, RNDIS_MEDIUM_1394 
};
static __u32 current_packet_filter;
static __u32 current_lookahead;
extern __u8 rndis_device_mac_str[6]; 
extern __u8 rndis_host_mac_str[6]; 
static __u32 mac_options = NDIS_MAC_OPTION_COPY_LOOKAHEAD_DATA |
    NDIS_MAC_OPTION_RECEIVE_SERIALIZED | NDIS_MAC_OPTION_TRANSFERS_NOT_PEND|
    NDIS_MAC_OPTION_NO_LOOPBACK; 

/* Main structure */
rndis_device_t rndis_device;

/* Remote NDIS API as a USB device */
usb_device_ops rndis_ops = 
{
#ifdef FIX_ZERO_LEN_PKT_BUG
    rndis_get_data_msg_len,
#endif    
    rndis_init_descriptors, /* init_descriptors */
    rndis_usb_config, /* get_configuration */
    
    /* handle_send_encapsulated_cmd */
    rndis_handle_send_encapsulated_command_msg, 
    
    /* handle_get_encapsulated_resp */
    rndis_handle_get_encapsulated_response_msg, 
    rndis_uninitialized, /* device_class_uninitialized */
    rndis_usb_ep_send_stub, /* ep_send */
    rndis_usb_terminate_transfers_stub, /* terminate_current_transfers */
    rndis_on_receive_usb_data_pkt_stub, /* on_receive_usb_data_pkt */
    NULL, /* on_send_usb_data_pkt */
#ifdef CONFIG_RG_USB_SLAVE_MODULE    
    on_rndis_closed_stub, /* on_device_class_closed */
#endif    
};

#ifdef CONFIG_USB_CONN_MONITOR
/* task_start_cb:
 * Start running a new kernel thread.
 */ 
static int task_start_cb(void *context)
{
    task_data_t *task = (task_data_t *)context;
    current->context = context;
    current->exit_signal = SIGCHLD;
    memcpy(current->comm, task->name, sizeof(current->comm));
    task->start_entry();
    up(&task->sem);
    kfree((void *)task);
    return 0;
}

/* create_and_start_task:
 * Create new kernel thread from task_ptr 
 */
static int create_and_start_task(char *name, void (*task_ptr)(void))
{
    task_data_t *task = (task_data_t *)kmalloc(sizeof(task_data_t), 
	GFP_KERNEL);
    if (!task)
    {
	printk(__FUNCTION__": Can't alloc memmory");
	return -1;
    }
    memset(task, 0, sizeof(task_data_t));
    memcpy(task->name, name, sizeof(task->name));
    task->start_entry = task_ptr;
    task->sem = MUTEX;
    task->kernel_thread_id = kernel_thread(task_start_cb, (void *)task, 
	CLONE_SIGHAND);
    if (task->kernel_thread_id < 0)
    {
	printk(__FUNCTION__": failed creating kernel_thread\n");
	kfree(task);
	return -1;
    }
    down(&task->sem);
    return  task->kernel_thread_id;
}

static void task_delay(unsigned long timeout)
{
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(timeout);
}

/* rndis_monitor_usb_conn_state:
 * Kernel thread for calling rndis_on_usb_conn_state_check() every 1 second.
 * Starts to run at init_rndis()
 */
static void rndis_monitor_usb_conn_state(void)
{
    int is_connected = 0;
    int retry = 0;

    for (;;)
    {
	if (rndis_on_usb_conn_state_check(&is_connected) < 0)
	{
	    printk(__FUNCTION__":Can't get connection status\n");
	    if (retry < MAX_RETRY)
	    {
	        retry ++;
	        continue;
	    }
	    break;
	}
	if (is_connected && !rndis_device.is_usb_connected)
	{
	    printk("New connection has established between USB host and "
		"Remote-NDIS device\n");
	}
	if (!is_connected && rndis_device.is_usb_connected)
	    printk("USB Remote-NDIS device has Lost connection with host\n"); 
	rndis_device.is_usb_connected = is_connected;
	task_delay(1*HZ);
    }
}

/* rndis_on_usb_conn_state_check:
 * Called every 1 second by the monitor to check current connection state.
 */
static int rndis_on_usb_conn_state_check(int *is_conn)
{
    if ((rndis_device.current_state != RNDIS_DATA_INITIALIZED))
    {
	*is_conn = 0;
	rndis_device.last_update_connection_state = USB_UNKNOWN;
	return 0;
    }
    if (rndis_device.last_update_connection_state)
    {
	*is_conn = 1;
	rndis_device.last_update_connection_state = USB_UNKNOWN;
	rndis_device.usb_connection_state = RNDIS_USB_CONNECTED;
	return 0;
    }
    if (rndis_device.usb_connection_state < 0 || 
	rndis_device.usb_connection_state > RNDIS_USB_DISCONNECTED)
    {
	printk(KERN_ERR __FUNCTION__": This connection state is not "
	    "supported\n");
	return -1;
    }
    if (rndis_device.usb_connection_state == RNDIS_USB_DISCONNECTED)
    {
	*is_conn = 0;
	rndis_device.ops->terminate_current_transfers();
	return 0;
    }
    *is_conn = 1;
    if (rndis_device.usb_connection_state <= RNDIS_SENT_KEEPALIVE_MSG_3)
	rndis_send_keep_alive_message();
    rndis_device.usb_connection_state++;
    return 0;
}
#endif

#ifdef FIX_ZERO_LEN_PKT_BUG
static int rndis_get_data_msg_len(__u8 *msg_hdr, int hdr_size)
{
    remote_ndis_pkt_msg_t *hdr = NULL;

    if (hdr_size < sizeof(remote_ndis_pkt_msg_t))
    {
      printk(__FUNCTION__" hdr size too small, hdr size = %d rndis hdr "
	     "size = %d\n", hdr_size, sizeof(remote_ndis_pkt_msg_t));
	return -1;
    }
    hdr = (remote_ndis_pkt_msg_t *)msg_hdr;
    return __le32_to_cpu(hdr->MessageLength);
}
#endif

/*
 * rndis_receive_pkt:
 * recieve ethernet packet and transfer it to ethernet layer
 * return the size of the rndis message that was handled or -1 on error.
 */
static int rndis_receive_pkt(__u8 *rndis_data, int size_limit)
{
    __u8 *data;
    __u32 data_length, message_length, data_offset, tmp;

    if (!rndis_data)
	return -1;
    memcpy(&tmp, rndis_data + RNDIS_PACKET_MSG_DATA_LEN_OFFSET,
	sizeof(__u32));
    data_length = __le32_to_cpu(tmp);
    memcpy(&tmp, rndis_data + RNDIS_PACKET_MSG_MSG_LEN_OFFSET,
	sizeof(__u32));
    message_length = __le32_to_cpu(tmp);
    memcpy(&tmp, rndis_data + RNDIS_PACKET_MSG_DATA_OFFSET_OFFSET,
	sizeof(__u32));
    data_offset = __le32_to_cpu(tmp);
    if (RNDIS_PACKET_MSG_HDR_OFFSET + data_offset + data_length > size_limit)
    {
	printk(KERN_ERR __FUNCTION__": corrupted ethernet packet data " 
	    "offset = %d, data size = %d size limit = %d\n",
	    data_offset + RNDIS_PACKET_MSG_HDR_OFFSET, data_length, size_limit);
	return -1;
    }
    data = rndis_data + RNDIS_PACKET_MSG_HDR_OFFSET + data_offset;
#ifdef USB_SLAVE_DEBUG
    usb_slave_dump_buffer(data, data_length);
#endif    
    if (!(rndis_device.eth_dev && netif_running(rndis_device.eth_dev)))
    {
	USB_DPRINTK(__FUNCTION__": dropped ethernet packet- interface is "
	    "down\n");	 
	return -1;
    }
    rndis_eth_rx(rndis_device.eth_dev, data_length, data);
    return message_length;
}

/* 
 * rndis_receive_data:
 * receive RNDIS packet, remove RNDIS header and pass the eth packet to 
 * rndis_receive_pkt
 */
static void rndis_receive_data(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    int buff_size = ep->abs_info->buff_pos;
    __u8 *buff = ep->abs_info->buff;

    while (buff_size > sizeof(remote_ndis_pkt_msg_t)) 
    {
	int rndis_msg_size = rndis_receive_pkt(buff, buff_size);
	
	if (rndis_msg_size < 0 || rndis_msg_size > buff_size)
	{
	    USB_DPRINTK(__FUNCTION__": Bad USB transfer received on RNDIS "
		"driver\n");
	    return;
	}
	buff_size -= rndis_msg_size;
	buff += rndis_msg_size;
    }
}

/*
 * rndis_send_data:
 * called from upper layer to send data through the usb addapter
 * add to the sk_buff header the rndis pkt msg header 
 * if there is no space in the beginning of sk_buff realloc sk_buff
 * header 
 */
int rndis_send_data(struct sk_buff **skb, void *context)
{
    remote_ndis_pkt_msg_t rndis_pkt_msg_hdr;
    int data_size = (*skb)->len;
    
    if ((rndis_device.current_state != RNDIS_DATA_INITIALIZED) || 
#ifdef CONFIG_USB_CONN_MONITOR	
	(rndis_device.usb_connection_state == RNDIS_USB_DISCONNECTED)
#else
	(rndis_device.last_update_connection_state == USB_DISCONNECTED)
#endif
       )
    {
	USB_DPRINTK(__FUNCTION__": Can't send data to USB, RNDIS device is "
	    "disconnected\n");
	return RNDIS_ERROR_DISCONNECTED;
    }
  
    /* FIXME: instead of returnning error, split the packet */
    if ((*skb)->len + sizeof(remote_ndis_pkt_msg_t) > host_maxTransferSize)
    {
	printk(__FUNCTION__": Can't send data to USB, size of packet is too "
	    "big for one transfer\n");	
	return RNDIS_ERROR_TX_FAILED;
    }
    
    /* There is no space in skb header */
    if (skb_headroom(*skb) < sizeof(remote_ndis_pkt_msg_t))
    {
	struct sk_buff *new_skb;
       
	USB_DPRINTK("No space in skb head - reallocating memory\n");
	new_skb = skb_realloc_headroom(*skb, sizeof(remote_ndis_pkt_msg_t));
        if (!new_skb)
	{
	    printk(KERN_ERR ": Can't reallocate memory for sk_buff head,"
		"No memory left\n");
	    return RNDIS_ERROR_TX_FAILED;
	}
	dev_kfree_skb_any(*skb);
	*skb = new_skb;	
    }
    
    /* Increase (*skb)->len, decrease (*skb)->data */
    skb_push(*skb, sizeof(remote_ndis_pkt_msg_t));

    /* Fill remote NDIS packet msg header fields */
    memset(&rndis_pkt_msg_hdr, 0 , sizeof(remote_ndis_pkt_msg_t));
    rndis_pkt_msg_hdr.MessageType = __cpu_to_le32(RNDIS_PACKET_MSG);
    rndis_pkt_msg_hdr.MessageLength = __cpu_to_le32((*skb)->len);
    rndis_pkt_msg_hdr.DataOffset = __cpu_to_le32(RNDIS_PACKET_MSG_MIN_DATA_OFFSET);
    rndis_pkt_msg_hdr.DataLength = __cpu_to_le32(data_size);
    
    /* Put in the beginning of skb data Remote NDIS header */
    memcpy((*skb)->data, &rndis_pkt_msg_hdr, sizeof(remote_ndis_pkt_msg_t));
    
    /* If packet size devided into transfer endpoint max pkt size, we need to
     * send a zero packet to end RNDIS transfer.
     */
    if (!((*skb)->len % rndis_device.tx_data_ep.abs_info->max_pkt_size))
    {
	__u8 *ptr;
	
	if (skb_tailroom(*skb) < 1)
	{
	    USB_DPRINTK("No space in skb tail - can't add zero pending to "
		"Remote NDIS packet\n");
	    goto Error;
	}
	ptr = skb_put(*skb, 1);
        *ptr = '0';
    }
    
    /* Transfer data to USB level */
    if (rndis_device.ops->ep_send((char *)((*skb)->data), (*skb)->len,
	&rndis_device.tx_data_ep, context, (void *)(*skb)) < 0)
    {
	goto Error;
    }
    return 0;
Error:
    /* Decrease back (*skb)->len, increase back (*skb)->data */
    skb_pull(*skb, sizeof(remote_ndis_pkt_msg_t));
    return RNDIS_ERROR_TX_FAILED;
}

void rndis_flush_queues(void)
{
    /* Flush control queue */
    if (rndis_device.control_response_q.is_init)
        usb_queue_free(&rndis_device.control_response_q);
    
    /* Flush notification queue */
    rndis_device.notification_ep_q_size = 0;
}

/*
 * rndis_disconnected:
 * called when USB device is physicly disconnected to host
 * or when got RNDIS_HALT_MSG
 */
static void rndis_uninitialized(void)
{
    if (rndis_device.current_state == RNDIS_UNINITIALIZED)
	return;

    /* change RNDIS state to uninitialized */
    rndis_device.current_state = RNDIS_UNINITIALIZED;
    
    /* flush queues */
    rndis_flush_queues();
    rndis_device.ops->terminate_current_transfers();
}

static void rndis_swap_list(__u32 *list, int num_of_elements)
{
    int i;
    
    for(i = 0; i < num_of_elements ; i++)
        list[i] = __cpu_to_le32(list[i]);
}

/* 
 * rndis_create_query_msg_response:
 * for every object id add to the response the relevant values 
 */
int rndis_create_query_msg_response(__u32 o_id, __u32 *resp_data_block, 
	__u8 *input_data, int input_data_size)
{
      __u32 size = 0;
      resp_data_block[1] = 
	  __cpu_to_le32(RNDIS_RESPONSE_QUERY_MSG_DATA_MIN_OFFSET); 

      switch(o_id)
      {
      case OID_GEN_SUPPORTED_LIST:
	  USB_DPRINTK("Got QUERY for OID_GEN_SUPPORTED_LIST\n");
	  size = 4 * RNDIS_CURRENT_SUPPORTED_OIDS_NUM;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  memcpy(resp_data_block + 2, supported_oids_list, size);
	  break;
      case OID_GEN_HARDWARE_STATUS:
	  USB_DPRINTK("Got QUERY for OID_GEN_HARDWARE_STATUS\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] =  __cpu_to_le32(current_hardware_status);
	  break;
      case OID_GEN_MEDIA_SUPPORTED:
	  USB_DPRINTK("Got QUERY for OID_GEN_MEDIA_SUPPORTED\n");
	  size = 4 * RNDIS_CURRENT_MEDIA_SUPPORTED_NUM;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  memcpy(resp_data_block + 2, media_supported, size);
	  break;
      case OID_GEN_MEDIA_IN_USE:
	  USB_DPRINTK("Got QUERY for OID_GEN_MEDIA_IN_USE\n");
	  size = 4 * RNDIS_MEDIUM_MAX;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  memcpy(resp_data_block + 2, media_in_use, size);
	  break;
      case OID_GEN_MAXIMUM_LOOKAHEAD:
	  USB_DPRINTK("Got QUERY for OID_GEN_MAXIMUM_LOOKAHEAD\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_MAX_FRAME_SIZE);
	  break;
      case OID_GEN_MAXIMUM_FRAME_SIZE:
	  USB_DPRINTK("Got QUERY for OID_GEN_MAXIMUM_FRAME_SIZE\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_MAX_FRAME_SIZE);
	  break;
      case OID_GEN_LINK_SPEED:
	  USB_DPRINTK("Got QUERY for OID_GEN_LINK_SPEED\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(120000); /* 12 Mbps */
	  break;
      case OID_GEN_TRANSMIT_BUFFER_SPACE:
	  USB_DPRINTK("Got QUERY for OID_GEN_TRANSMIT_BUFFER_SPACE "
	      "\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_MAX_FRAME_SIZE);
	  break;
      case OID_GEN_RECEIVE_BUFFER_SPACE:
	  USB_DPRINTK("Got QUERY for OID_GEN_RECEIVE_BUFFER_SPACE"
	      "\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_MAX_FRAME_SIZE);
	  break;
      case OID_GEN_TRANSMIT_BLOCK_SIZE:
	  USB_DPRINTK("Got QUERY for OID_GEN_TRANSMIT_BLOCK_SIZE\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_DATA_BLOCK_SIZE);
	  break;
      case OID_GEN_RECEIVE_BLOCK_SIZE:
	  USB_DPRINTK("Got QUERY for OID_GEN_RECEIVE_BLOCK_SIZE\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_DATA_BLOCK_SIZE);
	  break;
      case OID_GEN_VENDOR_ID:
	  USB_DPRINTK("Got QUERY for OID_GEN_VENDOR_ID\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(USB_DEV_VENDOR_ID);
	  break;
      case OID_GEN_VENDOR_DESCRIPTION:
	  USB_DPRINTK("Got QUERY for OID_GEN_VENDOR_DESCRIPTION\n");
	  size = strlen(USB_DEV_VENDOR_DESCRIPTION);
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  strcpy((char *)(resp_data_block + 2) ,USB_DEV_VENDOR_DESCRIPTION);
	  break;
      case OID_GEN_CURRENT_PACKET_FILTER:
	  USB_DPRINTK("Got QUERY for OID_GEN_CURRENT_PACKET_FILTER\n"
	      );
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(current_packet_filter);
	  break;
      case OID_GEN_CURRENT_LOOKAHEAD:
	  USB_DPRINTK("Got QUERY for OID_GEN_CURRENT_LOOKAHEAD\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(current_lookahead);
	  break;
      case OID_GEN_DRIVER_VERSION:
	  USB_DPRINTK("Got QUERY for OID_GEN_DRIVER_VERSION\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  *((__u16 *)resp_data_block + 4) = 
	      __cpu_to_le16(RNDIS_DEVICE_MAJOR_VERSION);
	  *((__u16 *)resp_data_block + 5) = 
	      __cpu_to_le16(RNDIS_DEVICE_MINOR_VERSION);
	  break;
      case OID_GEN_MAXIMUM_TOTAL_SIZE:
	  USB_DPRINTK("Got QUERY for OID_GEN_MAXIMUM_TOTAL_SIZE\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(RNDIS_MAX_TOTAL_SIZE);
	  break;
      case OID_GEN_MAC_OPTIONS:
	  USB_DPRINTK("Got QUERY for OID_GEN_MAC_OPTIONS\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(mac_options);
	  break;
      case OID_GEN_MEDIA_CONNECT_STATUS:
	  {
	      __u32 media_connect_status = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_MEDIA_CONNECT_STATUS\n");
	      if (rndis_device.current_state!=RNDIS_DATA_INITIALIZED ||
		  !rndis_device.eth_dev || 
		  !netif_running(rndis_device.eth_dev))
	      {
		  media_connect_status = 1;
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(media_connect_status);

	  }
	  break;
      case OID_GEN_MAXIMUM_SEND_PACKETS:
	  USB_DPRINTK("Got QUERY for OID_GEN_MAXIMUM_SEND_PACKETS\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(MAX_PACKETS_PER_TRANSFER);

	  break;
      case OID_GEN_VENDOR_DRIVER_VERSION:
	  USB_DPRINTK("Got QUERY for OID_GEN_VENDOR_DRIVER_VERSION\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(USB_DEV_VENDOR_VERSION);
	  break;
      case OID_GEN_XMIT_OK:
	  {
	      __u32 tx_ok = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_XMIT_OK\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);
                  if (rndis_stat)
		  {
		      tx_ok = rndis_stat->tx_packets - (rndis_stat->tx_errors + 
			  rndis_stat->tx_dropped);
		  }
		  if (tx_ok < 0)
		  {
		      printk(KERN_ERR ": Can't find valid statistic of "
			  "tx_ok\n");
		      resp_data_block[1] = 0;
		      goto Error;
		  }
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(tx_ok);

	  }
	  break;
      case OID_GEN_RCV_OK:
	  {
	      __u32 rx_ok = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_RCV_OK\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);

                  if (rndis_stat)
		  {
		      rx_ok = rndis_stat->rx_packets - (rndis_stat->rx_errors +
		      rndis_stat->rx_dropped);
		  }
		  if (rx_ok < 0)
		  {
		      printk(KERN_ERR ": Can't find valid statistic of rx_ok\n");
		      resp_data_block[1] = 0;
		      goto Error;
		  }
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(rx_ok);

	  }
	  break;
      case OID_GEN_XMIT_ERROR:
	  {
	      __u32 tx_errors = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_XMIT_ERROR\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);
                  if (rndis_stat)
		      tx_errors = rndis_stat->tx_errors;
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(tx_errors);
	  }
	  break;
      case OID_GEN_RCV_ERROR:
	  {
	      __u32 rx_errors = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_RCV_ERROR\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);

                  if (rndis_stat)
		      rx_errors = rndis_stat->rx_errors;
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(rx_errors); 

	  }
	  break;
      case OID_GEN_RCV_NO_BUFFER:
	  {
	      __u32 rx_dropped = 0;

	      USB_DPRINTK("Got QUERY for OID_GEN_RCV_NO_BUFFER\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);

                  if (rndis_stat)
		      rx_dropped =  rndis_stat->rx_dropped;
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(rx_dropped); 
	  }
	  break;
      case OID_802_3_PERMANENT_ADDRESS:
	  USB_DPRINTK("Got QUERY for OID_802_3_PERMANENT_ADDRESS\n");
	  size = 6;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  memcpy (resp_data_block + 2, rndis_host_mac_str, 6);
	  break;
      case OID_802_3_CURRENT_ADDRESS:
	  USB_DPRINTK("Got QUERY for OID_802_3_CURRENT_ADDRESS\n");
	  size = 6;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  memcpy (resp_data_block + 2, rndis_host_mac_str, 6);
	  break;
      case OID_802_3_MULTICAST_LIST:

	  /* We don't support this */
	  USB_DPRINTK("Got QUERY for OID_802_3_MULTICAST_LIST\n");
	  break;
      case OID_802_3_MAXIMUM_LIST_SIZE:
	  USB_DPRINTK("Got QUERY for OID_802_3_MAXIMUM_LIST_SIZE\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = 0;
	  break;
      case OID_802_3_MAC_OPTIONS:
	  USB_DPRINTK("Got QUERY for OID_802_3_MAC_OPTIONS\n");
	  size = 4;
	  resp_data_block[0] = __cpu_to_le32(size); /* data length */
	  resp_data_block[2] = __cpu_to_le32(mac_options);
	  break;
      case OID_802_3_RCV_ERROR_ALIGNMENT:
	  {
	      __u32 rx_frame_errors = 0;

	      USB_DPRINTK("Got QUERY for OID_802_3_RCV_ERROR_ALIGNMENT"
		  "\n");
	      if (rndis_device.eth_dev)
	      {
		  struct net_device_stats *rndis_stat = 
		      rndis_eth_stats(rndis_device.eth_dev);

                  if (rndis_stat)
		      rx_frame_errors =  rndis_stat->rx_frame_errors;
	      }
	      size = 4;
	      resp_data_block[0] = __cpu_to_le32(size); /* data length */
	      resp_data_block[2] = __cpu_to_le32(rx_frame_errors);

	  }
	  break;
      case OID_802_3_XMIT_ONE_COLLISION:
	  USB_DPRINTK("Got QUERY for OID_802_3_XMIT_ONE_COLLISION"
	      "\n");

	  /* The number of frames successfully transmited after axacly
	   * one collision
	   */
	  break;
      case OID_802_3_XMIT_MORE_COLLISIONS:
	  USB_DPRINTK("Got QUERY for OID_802_3_XMIT_MORE_COLLISIONS"
	      "\n");

	  /* The number of frames successfully transmited after more
	   * than one collision.
	   */
	  break;
      default:goto Error;
      }
      return  size;
Error:
    USB_DPRINTK("Can't handle this OID query : 0x%08x\n", o_id);
    return -1;
}

/* 
 * rndis_set_if_flags:
 * translate windows ndis flags to linux network device flags and set the 
 * current flags we got from the host to our device
 */
void rndis_set_if_flags(void)
{
    if (current_packet_filter & NDIS_PACKET_TYPE_ALL_MULTICAST)
	rndis_device.eth_dev->flags |= IFF_MULTICAST;
    if (current_packet_filter & NDIS_PACKET_TYPE_BROADCAST)
	rndis_device.eth_dev->flags |= IFF_BROADCAST;
    if (current_packet_filter & NDIS_PACKET_TYPE_MULTICAST)
	rndis_device.eth_dev->flags |= IFF_ALLMULTI;
    if (current_packet_filter & NDIS_PACKET_TYPE_PROMISCUOUS)
	rndis_device.eth_dev->flags |= IFF_PROMISC;
}

/*
 * rndis_set_object_id:
 * for every mandatory object id, update the relevant values in the device 
 */
int rndis_set_object_id(__u32 o_id, __u8 *data, __u32 data_size)
{
      switch(o_id)
      {
      case OID_GEN_CURRENT_PACKET_FILTER:
	  USB_DPRINTK("Got SET message for OID_GEN_CURRENT_PACKET_FILTER\n");
	  if (data_size != 4)
	      goto Error;
	  current_packet_filter = *(__u32 *)data;
	  if (*(__u32 *)data)
	  {
	      if (rndis_device.eth_dev)
		  rndis_set_if_flags();
	      rndis_device.current_state = RNDIS_DATA_INITIALIZED;
	      USB_DPRINTK("*** entering  RNDIS_DATA_INITIALIZED state ***\n");
	  }
	  else
	  {
	      rndis_device.current_state = RNDIS_INITIALIZED;
	  }
	  break;
      case OID_GEN_CURRENT_LOOKAHEAD:
	  USB_DPRINTK("Got SET message for OID_GEN_CURRENT_LOOKAHEAD\n");
	  if (data_size != 4)
	      goto Error;
	  current_lookahead = *(int*)data;
	  break;
      case OID_GEN_PROTOCOL_OPTIONS:
	  USB_DPRINTK("Got SET message for OID_GEN_PROTOCOL_OPTIONS\n");

	  /* TODO */
	  /* specific a bitmask that defines optional properties of the
	   * protocol driver.A protocol informs NDIS of its
	   * proporties, which can optionally take advantage of them.
	   * If the protocol driver does not set its flags on binding,
	   * NDIS assumes they are all clear.
	   * The following flags are currently defined:
	   * NDIS_PROT_OPTIONS_ESTIMATED_LENGTH:
	   * packets can be indicate at the worst-case estimate of
	   * packets size, instead of an axact value, to this
	   * protocol.
	   * NDIS_PROT_OPTION_NO_LOOPBACK:
	   * The protocol does not requier loopback support on the
	   * binding
	   */
	  break;
      case OID_802_3_MULTICAST_LIST:
	  USB_DPRINTK("Got SET message for OID_802_3_MULTICAST_LIST\n");
	  if (data_size)
	  {
	      printk(KERN_ERR ": host set request for "
		  "OID_802_3_MULTICAST_LIST :try to set an empty list\n");
	      goto Error;
	  }
	  break;
      default: goto Error;
      }
      return 0;
Error:
      USB_DPRINTK(__FUNCTION__" : Wrong Oid type or wrong data input\n");
      return -EINVAL;
}

/* 
 * rndis_handle_control_request:
 * handle the rndis specific host control request 
 * called after a "SEND_ENCAPSULATED_COMMAND" of the host
 */
static __u32 rndis_handle_control_request(__u8 *rndis_req_block, 
    __u8 *rndis_resp_block, int req_size)
{   
    rndis_resp_t *rndis_resp = (rndis_resp_t *)rndis_resp_block;
    rndis_req_hdr_t *rndis_req; 
    __u32 *req_data;
    __u32 size = 0;
    __u32 req_length;  
    
    if (req_size < RNDIS_REQUEST_BLOCK_MIN_SIZE || 
        req_size > RNDIS_REQUEST_BLOCK_MAX_SIZE )
    {
	printk(KERN_ERR __FUNCTION__": corrupted request\n");
	goto Error;
    }
    rndis_req = (rndis_req_hdr_t *)rndis_req_block;
    req_length = __le32_to_cpu(rndis_req->req_length);
    req_data = (__u32 *)(rndis_req_block + sizeof(rndis_req_hdr_t)); 

    /* put response general details in response,
     * initial the status to be failure until it changed 
     */
    size = RNDIS_RESPONSE_BLOCK_MIN_SIZE;
    rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_FAILURE); 
    rndis_resp->resp_id = rndis_req->req_id; 

    /* do the specific request and put response specific details in the
     * response
     */
    switch (__le32_to_cpu(rndis_req->req_type))
    {
    case REMOTE_NDIS_INITIALIZE_MSG:
	USB_DPRINTK("Got REMOTE_NDIS_INITIALIZE_MSG\n");
	rndis_resp->resp_type = __cpu_to_le32(REMOTE_NDIS_INITIALIZE_CMPLT);
	rndis_resp->resp_length = __cpu_to_le32(REMOTE_NDIS_INIT_CMPLT_SIZE);

	/* if we can't do the initialization, return empty response */
	memset(rndis_resp->resp_data, 0, (REMOTE_NDIS_INIT_CMPLT_SIZE - 
	    RNDIS_RESPONSE_BLOCK_MIN_SIZE));
	size = REMOTE_NDIS_INIT_CMPLT_SIZE;

	/* check request data */
	if (req_size != REMOTE_NDIS_INIT_MSG_SIZE ||
	    req_length != REMOTE_NDIS_INIT_MSG_SIZE)
	{
	    printk(KERN_ERR __FUNCTION__": corrupted remote ndis control "
		"message\n");
	    rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_INVALID_DATA); 
	    break;
	}

	/* do the initialization: use request data */
	host_major_version = __le32_to_cpu(*req_data);
	host_minor_version = __le32_to_cpu(*(req_data + 1));
	host_maxTransferSize = __le32_to_cpu(*(req_data + 2));

	/* save response data */
	rndis_resp->resp_data[0] = __cpu_to_le32(RNDIS_DEVICE_MAJOR_VERSION);
	rndis_resp->resp_data[1] = __cpu_to_le32(RNDIS_DEVICE_MINOR_VERSION);
	rndis_resp->resp_data[2] = __cpu_to_le32(RNDIS_DF);
	rndis_resp->resp_data[3] = __cpu_to_le32(RNDIS_MEDIUM_802_3);
	rndis_resp->resp_data[4] = __cpu_to_le32(MAX_PACKETS_PER_TRANSFER);
	rndis_resp->resp_data[5] = __cpu_to_le32(DEVICE_MAX_TRANSFER_SIZE);	
	rndis_resp->resp_data[6] = __cpu_to_le32(0x00000003);  /* packet alignment factor = 
		                				* 8 (2^3) 
								*/
	rndis_resp->resp_data[7] = 0x00000000; /* reserved */
	rndis_resp->resp_data[8] = 0x00000000; /* reserved */
	rndis_device.current_state = RNDIS_INITIALIZED;
	rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_SUCCESS);
	break;
    case REMOTE_NDIS_HALT_MSG:
	USB_DPRINTK("Got REMOTE_NDIS_HALT_MSG\n");
	if (rndis_device.current_state != RNDIS_UNINITIALIZED)
	    rndis_uninitialized();
	size = 0;
	break;
    case REMOTE_NDIS_QUERY_MSG:
	{
	    __u32 Oid;
	    __u32 req_InformationBufferLength ; 
	    __u32 req_InformationBufferOffset ;
	    __u8 *req_InformationBuffer;

	    USB_DPRINTK("Got REMOTE_NDIS_QUERY_MSG\n");
	    rndis_resp->resp_type = __cpu_to_le32(REMOTE_NDIS_QUERY_CMPLT);
	    rndis_resp->resp_length = __cpu_to_le32(REMOTE_NDIS_QUERY_CMPLT_SIZE);

	    /* if we can't do the query, return empty response */
	    memset(rndis_resp->resp_data, 0 , (REMOTE_NDIS_QUERY_CMPLT_SIZE 
		- RNDIS_RESPONSE_BLOCK_MIN_SIZE));
	    size = REMOTE_NDIS_QUERY_CMPLT_SIZE;

	    /* check request data*/
	    if (req_size < REMOTE_NDIS_QUERY_MSG_SIZE ||
		req_length < REMOTE_NDIS_QUERY_MSG_SIZE)
	    {
		printk(KERN_ERR __FUNCTION__": Corrupted remote ndis control "
		    "message\n");
		rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_INVALID_DATA); 
		break;
	    }

	    if (rndis_device.current_state == RNDIS_UNINITIALIZED)
	    {
		printk(KERN_ERR __FUNCTION__":Got a query message without "
		    "initializing the device\n"); 
		break;
	    }
	    Oid = __le32_to_cpu(*req_data);
	    req_InformationBufferLength = 
		__le32_to_cpu(*(req_data + 1)); 
	    req_InformationBufferOffset = 
		__le32_to_cpu(*(req_data + 2));
	    req_InformationBuffer = ((__u8 *)&rndis_req->req_id 
		+ req_InformationBufferOffset);
	    size += rndis_create_query_msg_response(Oid, 
		&rndis_resp->resp_data[0], req_InformationBuffer, 
		req_InformationBufferLength); 
	    if (size >= REMOTE_NDIS_QUERY_CMPLT_SIZE)
	    {
		rndis_resp->resp_length = __cpu_to_le32(size);
		rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_SUCCESS);
	    }
	}
	break;
    case REMOTE_NDIS_SET_MSG:
	{
	    __u32 Oid;
	    __u32 req_InformationBufferLength ; 
	    __u32 req_InformationBufferOffset ;
	    __u8 *req_InformationBuffer;

	    USB_DPRINTK("Got REMOTE_NDIS_SET_MSG\n");
	    rndis_resp->resp_type = __cpu_to_le32(REMOTE_NDIS_SET_CMPLT);
	    rndis_resp->resp_length = __cpu_to_le32(REMOTE_NDIS_SET_CMPLT_SIZE);

	    /* check request data */
	    if (req_size < REMOTE_NDIS_SET_MSG_SIZE ||
		req_length < REMOTE_NDIS_SET_MSG_SIZE)
	    {
		printk(KERN_ERR __FUNCTION__":Corrupted remote ndis control "
		    "message\n");
		rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_INVALID_DATA); 
		break;
	    }
	    if (rndis_device.current_state == RNDIS_UNINITIALIZED)
	    {
		printk(KERN_ERR __FUNCTION__"Got a set message without "
		    "initializing the device\n"); 
		break;
	    }
	    Oid = __le32_to_cpu(*req_data);
	    req_InformationBufferLength = __le32_to_cpu(*(req_data + 1)); 
	    req_InformationBufferOffset = __le32_to_cpu(*(req_data + 2));
	    req_InformationBuffer = ((__u8 *)&rndis_req->req_id +  
		req_InformationBufferOffset);
	    if (rndis_set_object_id(Oid, req_InformationBuffer, 
		req_InformationBufferLength) < 0)
	    {
		printk(KERN_ERR __FUNCTION__ "Can't set the object value\n");
		break;
	    }
	    rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_SUCCESS); 
	}
	break;
    case REMOTE_NDIS_RESET_MSG:
	USB_DPRINTK("Got REMOTE_NDIS_RESET_MSG\n");
	rndis_resp->resp_type = __cpu_to_le32(REMOTE_NDIS_RESET_CMPLT);
	rndis_resp->resp_length = __cpu_to_le32(REMOTE_NDIS_RESET_CMPLT_SIZE);
	rndis_resp->resp_id = __cpu_to_le32(RNDIS_STATUS_FAILURE); /* status */
	rndis_resp->resp_status = 0; /* AddressingReset */
	if (rndis_device.current_state == RNDIS_UNINITIALIZED)
	{
	    printk(KERN_ERR __FUNCTION__ "Got a reset message without "
		"initializing the device\n"); 
	    break;
	}
	/* flush queues */
	rndis_flush_queues();
	rndis_device.ops->terminate_current_transfers();
	rndis_resp->resp_id = __cpu_to_le32(RNDIS_STATUS_SUCCESS); /* status */
	break;
    case REMOTE_NDIS_KEEPALIVE_MSG:
	USB_DPRINTK("Got REMOTE_NDIS_KEEPALIVE_MSG\n");
	rndis_resp->resp_type = __cpu_to_le32(REMOTE_NDIS_KEEPALIVE_CMPLT);
	rndis_resp->resp_length = __cpu_to_le32(REMOTE_NDIS_KEEPALIVE_CMPLT_SIZE);
	rndis_resp->resp_status = __cpu_to_le32(RNDIS_STATUS_SUCCESS);
	rndis_device.last_update_connection_state = USB_CONNECTED;
	break;

    case REMOTE_NDIS_KEEPALIVE_CMPLT:
	USB_DPRINTK("Got REMOTE_NDIS_KEEPALIVE_CMPLT\n");
	size = 0;
	rndis_device.last_update_connection_state = USB_CONNECTED;
	break;      
    default:
	{
	    USB_DPRINTK("rndis_handle_control_request: wrong host control "
		"request type number\n");
	    goto Error;
	};
    }
    if (size > 0)
    {
	/* we have one more to transmit so increase the number of
	 * notificatin transmit of the interrupt endpoint
	 */  
	rndis_device.notification_ep_q_size++;
	rndis_send_response_available_notification(
            (void *)(&rndis_device.notification_ep));
    }
    return size;  

/* we didn't succsess in creating even a "failure status" response, so we
 * need to initiate indicate status message
 */
Error:
    {
	rndis_diagnostic_info_t info;
	
	info.DiagStatus = __cpu_to_le32(RNDIS_STATUS_NOT_SUPPORTED);
	info.ErrorOffset = 0;
	rndis_initiate_status_indicate_msg(RNDIS_STATUS_INVALID_DATA, 
	    &info, rndis_req_block, req_size);
	return -EINVAL;
    }
}

/*
 * rndis_initiate_status_indicate_msg:
 * in some drastic situation the device will want to initiate 
 * this message. 
 * the function create the message and add it to the control response Q
 * it also send a notification in the interrupt channel
 */
static void rndis_initiate_status_indicate_msg(rndis_status_t status, 
    rndis_diagnostic_info_t *info, __u8 *buffer, __u32 buffer_length)
{
    usb_hdr2 *resp_node = NULL;
    int req_size = REMOTE_NDIS_INDICATE_STATUS_MSG_SIZE;
    rndis_status_msg_t *msg;
    int rc = 0;
    
    if (buffer_length > RNDIS_REQUEST_BLOCK_MAX_SIZE)
    {
	rc = -1;
	goto Error;
    }
    if (rndis_device.current_state == RNDIS_UNINITIALIZED)
    {
        printk(KERN_WARNING __FUNCTION__": device is in RNDIS_UNINITIALIZED "
	    "state\n");
	return;
    }
    resp_node = (usb_hdr2 *)kmalloc(sizeof(usb_hdr2) + 
	sizeof(rndis_status_msg_t) ,GFP_ATOMIC);
    if (!resp_node)
    {
	printk(KERN_WARNING __FUNCTION__": Can't allocate memory\n");
	return;
    }
    USB_DPRINTK("Initiating REMOTE_NDIS_INDICATE_STATUS_MSG status = 0x%08x\n",
	status);
    memset(resp_node, 0, sizeof(usb_hdr2) + sizeof(rndis_status_msg_t));
    msg = (rndis_status_msg_t *)(resp_node->buff) ;
    msg->StatusBufferLength = 0;
    msg->StatusBufferOffset = 0;
    if (info)    
    {
	if (!buffer || buffer_length <= 0)
	{
	    rc = -1;
	    goto Error;
	}
	req_size += (buffer_length + sizeof(rndis_diagnostic_info_t));
	msg->StatusBufferLength = __cpu_to_le32(buffer_length + 
	    sizeof(rndis_diagnostic_info_t));
	msg->StatusBufferOffset = 
	    __cpu_to_le32(RNDIS_INDICATE_STATUS_INFO_MIN_OFFSET);
	msg->info.DiagStatus = info->DiagStatus;
	msg->info.ErrorOffset = info->ErrorOffset;
	memcpy(msg->StatusBuffer, buffer, buffer_length);
    }
    msg->MessageType = __cpu_to_le32(REMOTE_NDIS_INDICATE_STATUS_MSG);
    msg->Status = __cpu_to_le32(status);
    msg->MessageLength = __cpu_to_le32(req_size);
    resp_node->size = req_size;
    
    /* Add element to the notification queue, and start transfer */
    rndis_device.notification_ep_q_size++;
    rndis_send_response_available_notification(
	(void *)(&rndis_device.notification_ep));

    /* Add the message to the response queue */
    usb_queue_tail(&rndis_device.control_response_q, resp_node);
Error:
    if (rc < 0)
    {
	printk(KERN_ERR __FUNCTION__": Size of input buffer for indicate "
	    "status msg is invalid\n");
    }
}
void rndis_send_media_disconnected_msg(void)
{
    rndis_initiate_status_indicate_msg(RNDIS_STATUS_MEDIA_DISCONNECT, NULL, 
	NULL, 0);
}

void rndis_send_media_connected_msg(void)
{
    rndis_initiate_status_indicate_msg(RNDIS_STATUS_MEDIA_CONNECT, NULL, NULL,
        0);
}

static void rndis_initiate_msg(int msg_size, __u32 msg_type)
{
    usb_hdr2 *new = NULL;
    
    /* Create new node and put inside it the relevant msg */
    new = (usb_hdr2 *)kmalloc(sizeof(usb_hdr2) + msg_size, GFP_KERNEL);
    if (!new)
    {
	printk(KERN_ERR __FUNCTION__": Can't allocate memory\n");
	return;
    }
    memset(new, 0, sizeof(usb_hdr2) + msg_size);
    *(__u32 *)new->buff = msg_type;
    *((__u32 *)new->buff + 1) = msg_size;
    *((__u32 *)new->buff + 2) = 
	rndis_device.current_req_id;
    rndis_device.current_req_id ++;
    if (rndis_device.current_req_id == 0xffffffff)
	rndis_device.current_req_id = 0;
    new->size = msg_size;
    rndis_device.notification_ep_q_size++;
    rndis_send_response_available_notification(
        (void *)(&rndis_device.notification_ep));
    usb_queue_tail(&rndis_device.control_response_q, new);
}

/*
 * rndis_send_keep_alive_message:
 * If we have no other indication for connection we send this message to the 
 * host and wait for a respons.
 * the function create the KEEPALIVE message and add it to the control 
 * response queue, it also send a notification in the interrupt channel.
 */
static void rndis_send_keep_alive_message(void)
{
    USB_DPRINTK(__FUNCTION__": Sending REMOTE_NDIS_KEEPALIVE_MSG\n");
    rndis_initiate_msg(REMOTE_NDIS_KEEPALIVE_MSG_SIZE, 
        REMOTE_NDIS_KEEPALIVE_MSG);	    
}

/*
 * rndis_initiate_halt_msg:
 * In some situation the device will want to halt, so it send halt message 
 * to the host. 
 * the function create the halt message and add it to the control response 
 * queue, it also send a notification in the interrupt channel.
 */
static void rndis_initiate_halt_message(void)
{
    USB_DPRINTK(__FUNCTION__": Sending REMOTE_NDIS_HALT_MSG\n");
    rndis_initiate_msg(REMOTE_NDIS_HALT_MSG_SIZE, REMOTE_NDIS_HALT_MSG);	    
    rndis_device.current_state = RNDIS_UNINITIALIZED;
}

/* 
 * rndis_send_response_available_notification:
 * called in: UdcEpCHandler to see if there is new data to send.
 * In every call to the function: check if the "queue" is not empty,
 * if not, fill the endpoint data buffer with new 
 * notificatin messege and decrease the number of notificatin in the "queue" 
 * in 1.
 */
static void rndis_send_response_available_notification(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    static __u32 notification_pkt[2];

    notification_pkt[0] = __cpu_to_le32(RESPONSE_AVAILABLE);
    ep->abs_info->is_busy = 0;
    
    /* There are no messages in the "queue" */
    if (!rndis_device.notification_ep_q_size) 
	return;

    /* Insert the notification to the software structure, it will be taken on
     * the endpoint interrupt and decrease the number of notification that 
     * left to be sent (in the "queue").
     */
    rndis_device.notification_ep_q_size--;
    rndis_device.ops->ep_send((__u8 *)notification_pkt, RNDIS_NOTIFICATION_EP_MAX_SIZE, 
	ep, NULL, NULL);
}

/*
 * rndis_handle_get_encapsulated_response_msg:
 * called in: usb_slave_handle_control_request for GET_ENCAPSULATED_RESPONSE
 * request.
 * In every call to this function take out the first response from the response 
 * queue and put it in resp_buff
 */
static void rndis_handle_get_encapsulated_response_msg(__u8 *resp_buff,
    __u32 *resp_size_p)
{
     unsigned long flags; 
     usb_hdr2 *next = NULL;

    /* take out the first element in the control_ep response queue;
     * put its size in resp_size and its data in resp_buff
     */
    spin_lock_irqsave(&(rndis_device.control_response_q.lock), flags);
    if (rndis_device.control_response_q.qlen > 0)
    {
	next =  rndis_device.control_response_q.next;
	usb_unlink(next);
	memcpy(resp_buff, next->buff, next->size);
	*resp_size_p = next->size;
	kfree(next);
    }
    spin_unlock_irqrestore(&(rndis_device.control_response_q.lock), flags);
}

/* 
 * rndis_handle_send_encapsulated_command_msg:
 * called in: usb_slave_handle_control_request for SEND_ENCAPSULATED_COMMAND 
 * request.
 * Create RNDIS response queue if not exist.
 * In every call to this function add response to the queue
 */
static void rndis_handle_send_encapsulated_command_msg(__u8 *req_block, 
    __u32 req_size)
{
     usb_hdr2 *resp_node = NULL;
    
     /* only if hedear of request is full, we can use the request block */
     if (req_size < RNDIS_REQUEST_BLOCK_MIN_SIZE)
     {
	 printk(KERN_ERR __FUNCTION__": Bad SEND_ENCAPSULATED_CMD, header "
	     "size = %d\n", req_size);
	 return;
     }
     /* create new node and put inside it the relevant response */
     resp_node = (usb_hdr2 *)kmalloc(sizeof(usb_hdr2) + 
	 RNDIS_RESPONSE_BLOCK_MAX_SIZE, GFP_ATOMIC);
     if (!resp_node)
     {
	 printk(KERN_ERR __FUNCTION__": Can't allocate memory\n");
	 return;
     }
     memset(resp_node, 0, sizeof(usb_hdr2) + 
	 RNDIS_RESPONSE_BLOCK_MAX_SIZE);
     resp_node->size = rndis_handle_control_request(req_block, 
	 (__u8 *)resp_node->buff, req_size);

     /* if there is a response, add to the control_ep_response queue the 
      * new response 
      */
     if (resp_node->size > 0)
	 usb_queue_tail(&rndis_device.control_response_q, resp_node);

     /* if there is no response free the node */
     else
     {
	 kfree(resp_node);
	 resp_node = NULL;
     }
}

static int rndis_init_ep(endpoint_t *ep, int ep_dir, int ep_type)
{
    ep->abs_info = (abstract_ep_t *)kmalloc(sizeof(abstract_ep_t), GFP_KERNEL);
    if (!ep->abs_info)
    {
	printk(KERN_ERR __FUNCTION__": failed allocating memory\n");
	return -1;
    }
    memset(ep->abs_info, 0, sizeof(abstract_ep_t));
    ep->abs_info->type = ep_type;
    ep->abs_info->direction = ep_dir;
    ep->abs_info->addr = rndis_device.current_max_ep_addr;
    rndis_device.current_max_ep_addr++;
    ep->abs_info->mode = USB_DISABLED;
    ep->abs_info->transfer_mode = USB_NAK;
    ep->abs_info->is_busy = 0;
    ep->abs_info->end_context1 = NULL;
    ep->abs_info->end_context2 = NULL;
    ep->abs_info->end_arg1 = 0;
    ep->abs_info->end_arg2 = 0;
    ep->abs_info->buff = NULL;
    ep->abs_info->buff_pos = 0;
    if (ep_type == RG_USB_INTERRUPT)
    {
	ep->abs_info->interface_num = 0;
	ep->abs_info->max_pkt_size = RNDIS_NOTIFICATION_EP_MAX_SIZE;
        ep->abs_info->is_dma = 0;
        ep->abs_info->ep_operation = 
	    rndis_send_response_available_notification;
	ep->abs_info->end_func = rndis_send_response_available_notification;
	return 0;
    }
    ep->abs_info->interface_num = 1;
    ep->abs_info->max_pkt_size = RNDIS_DATA_EP_MAX_SIZE;
    if (ep_dir == RG_USB_IN)
    {	
        ep->abs_info->is_dma = 1;
        ep->abs_info->ep_operation = NULL; /* using DMA, so all proccecing 
					    * done by the 
				            * hardware (DMA controler)
				            */
	ep->abs_info->end_func = rndis_eth_tx_end;
	return 0;
    }
#ifdef CONFIG_USB_SIMPLE_DMA    
    ep->abs_info->is_dma = 0;
#else    
    ep->abs_info->is_dma = 1;
#endif    
    ep->abs_info->ep_operation = rndis_device.ops->on_receive_usb_data_pkt;
    ep->abs_info->end_func = rndis_receive_data;
    ep->abs_info->buff = rndis_device.receive_buffer;
    ep->abs_info->buff_size = DEVICE_MAX_TRANSFER_SIZE;
    return 0;
}

static int rndis_init_all_eps(void)
{
    if (rndis_init_ep(&rndis_device.notification_ep, RG_USB_IN, 
	RG_USB_INTERRUPT) < 0)
    {
	printk(KERN_ERR __FUNCTION__": faild initializing interrupt ep\n");
	return -1;
    }
    if (rndis_init_ep(&rndis_device.tx_data_ep, RG_USB_IN, RG_USB_BULK) < 0)
    {
	printk(KERN_ERR __FUNCTION__": faild initializing bulk in ep\n");
	return -1;
    }
    if (rndis_init_ep(&rndis_device.rx_data_ep, RG_USB_OUT, RG_USB_BULK) < 0)
    {
	printk(KERN_ERR __FUNCTION__": faild initializing bulk out ep\n");
	return -1;
    }
    return 0;
}

/* initializing the device before getting frome the host initializing
 * message
 */
int rndis_init(struct net_device *dev)
{
    USB_DPRINTK("rndis_init called\n");
    rndis_device.ops = &rndis_ops;
    rndis_device.eth_dev = dev;
    rndis_device.notification_ep_q_size = 0;
    rndis_device.current_req_id = 0;
    rndis_device.current_max_ep_addr = 1;
    rndis_device.current_state = RNDIS_UNINITIALIZED;

    /* Oids that should be initialized */ 
    current_hardware_status = RNDIS_HARDWARE_STATUS_NOT_READY;
    current_lookahead = RNDIS_MAX_FRAME_SIZE;
#ifdef __BIG_ENDIAN    
    rndis_swap_list(supported_oids_list, RNDIS_CURRENT_SUPPORTED_OIDS_NUM);
    rndis_swap_list(media_supported, RNDIS_CURRENT_MEDIA_SUPPORTED_NUM);
    rndis_swap_list(media_in_use, RNDIS_MEDIUM_MAX);
#endif    
    memcpy(dev->dev_addr, rndis_device_mac_str, 6);
   

    /* if control_ep queue is not initialized */
    if (!rndis_device.control_response_q.is_init)
    {
	usb_queue_head_init(&rndis_device.control_response_q);
	rndis_device.control_response_q.is_init = 1;
    }
#ifdef CONFIG_USB_CONN_MONITOR    
    if (create_and_start_task("usb_monitor", rndis_monitor_usb_conn_state)
	< 0)
    {
	printk(KERN_ERR __FUNCTION__": Can't start USB connection monitor, "
	    "exiting\n");
	return -1;
    }
#endif   
    if (rndis_init_all_eps() < 0)
    {
	printk(KERN_ERR __FUNCTION__": failed initializing endpoints "
	    "structures\n");
	return -1;
    }
    if (usb_slave_on_device_class_initialized(rndis_device.ops, 1,
	CFG_DESC_TOTAL_LEN, &rndis_device.last_update_connection_state) < 0)
    {
	printk(KERN_ERR __FUNCTION__": failed initializing USB\n");
	return -1;
    }
    rndis_device.rx_data_ep.abs_info->ep_operation = 
	rndis_device.ops->on_receive_usb_data_pkt;
    current_hardware_status = RNDIS_HARDWARE_STATUS_READY;
    return 0;
}

/* Initialized the values of the USB descriptors */ 
/* FIXME: this way of insertion of the values to the descriptors will not
 * work for a little endian arch.
 */
static int rndis_init_descriptors(__u8 *dev_desc_p, int dev_desc_size, 
    __u8 *cfg_desc_p, int cfg_desc_size, string_desc_t *prod_desc, 
    string_desc_t *mfg_desc, string_desc_t *lang_desc) 
{
    __u8 temp_str[50], *p_str;
    __u32 i = 0;

    if (!dev_desc_p || dev_desc_size != 18)
    {
	printk(KERN_ERR __FUNCTION__": device descriptor pointer is invalid\n");
	return -1;
    }
    if (!cfg_desc_p || cfg_desc_size != CFG_DESC_TOTAL_LEN)
    {
	printk(KERN_ERR __FUNCTION__": configuration descriptor pointer is "
	    "invalid\n");
	return -1;
    }
     
    /* The following array contains all the descriptors information.
     * The specific fields of every descriptor, are documented in:
     * General descriptor: USB spec 1.1 chapter 9
     * RNDIS specific descriptors: Remote-NDIS 1.0 Appendix B
     */
    /* 1. Initialize Device Descriptor */
    dev_desc_p[0] = SIZEOF_DEV_DESC;
    dev_desc_p[1] = DEVICE_DESC; 
    dev_desc_p[2] = 0x00; /* usb rev LSB */
    dev_desc_p[3] = 0x01; /* usb rev MSB */
    dev_desc_p[4] = 0x02; /* communication device class */
    dev_desc_p[5] = 0x00; /* No sub-class */
    dev_desc_p[6] = 0x00; /* No device protocol */
    dev_desc_p[7] = USB_DEV_EP_0_MAX_PKT_SIZE; 
    dev_desc_p[8] = (__u8)USB_DEV_VENDOR_ID;
    dev_desc_p[9] = (__u8)(USB_DEV_VENDOR_ID >> 8);
    dev_desc_p[10] = (__u8)USB_DEV_PRODUCT_ID;
    dev_desc_p[11] = (__u8)(USB_DEV_PRODUCT_ID >> 8);
    dev_desc_p[12] = 0x00; /* RNDIS release version  LSB */
    dev_desc_p[13] = 0x01; /* RNDIS release version  MSB */
    dev_desc_p[14] = 0x01; /* index of mfg string desc */
    dev_desc_p[15] = 0x02; /* index of product string desc */
    dev_desc_p[16] = 0x00; /* We don't have serial number */
    dev_desc_p[17] = 0x01; /* one configuration */		

    /* 2. Initialize Configuration Descriptor */
    cfg_desc_p[0] = SIZEOF_CFG_DESC;
    cfg_desc_p[1] = CONFIGURATION_DESC; /* configuration descriptor */
    cfg_desc_p[2] = CFG_DESC_TOTAL_LEN; /* Total lengt LSB */
    cfg_desc_p[3] = 0x00; /* Total lengt MSB */
    cfg_desc_p[4] = 0x02; /* Two interfaces */
    cfg_desc_p[5] = 0x01; /* Only one configuration with number = 1 */
    cfg_desc_p[6] = 0x00; /* No string descriptor for this configuration */
    cfg_desc_p[7] = USB_DEV_CONFIG_POWER_ATTRIBUTES;
    cfg_desc_p[8] = USB_DEV_CONFIG_MAX_POWER_CONSUME;

    /* Communication interface */   
    cfg_desc_p[9] = SIZEOF_IF_DESC;
    cfg_desc_p[10] = INTERFACE_DESC;/* if descriptor */
    cfg_desc_p[11] = 0x00; /* The first interface in this configuration */
    cfg_desc_p[12] = 0x00; /* No alternate setting */
    cfg_desc_p[13] = 0x01; /* one endpoint */
    cfg_desc_p[14] = 0x02; /* communicatin class */
    cfg_desc_p[15] = 0x02; /* abstract control sub-class */
    cfg_desc_p[16] = 0xFF; /* Protocol code */
    cfg_desc_p[17] = 0x00; /* No string descriptor for this interface */

    /* Communication header */
    cfg_desc_p[18] = SIZEOF_COMM_HDR_DESC;
    cfg_desc_p[29] = COMMUNICATION_CLASS; /* Communication Class */ 
    cfg_desc_p[20] = 0x00; /*  */
    cfg_desc_p[21] = 0x10; /* rev of CDC = 1.1 */
    cfg_desc_p[22] = 0x01; /* rev of CDC = 1.1 */

    /* Call-management */
    cfg_desc_p[23] = SIZEOF_COMM_MNG_DESC;
    cfg_desc_p[24] = COMMUNICATION_CLASS; /* Communication Class */ 
    cfg_desc_p[25] = 0x01; /* Call-management */
    cfg_desc_p[26] = 0x00; 
    cfg_desc_p[27] = 0x01; /* data ifaceNumber */

    /* Abstruct control */
    cfg_desc_p[28] = SIZEOF_COMM_ABS_DESC;
    cfg_desc_p[29] = COMMUNICATION_CLASS;
    cfg_desc_p[30] = 0x02; /* Abstruct control */
    cfg_desc_p[31] = 0x00;

    /* Communication unit function */
    cfg_desc_p[32] = SIZEOF_COMM_FUNC_DESC;
    cfg_desc_p[33] = COMMUNICATION_CLASS;
    cfg_desc_p[34] = 0x03; /* Unit function */
    cfg_desc_p[35] = 0x00;
    cfg_desc_p[36] = 0x01;

    /* Interrupt endpoint */
    cfg_desc_p[37] = SIZEOF_EP_DESC;
    cfg_desc_p[38] = ENDPOINT_DESC; /* endpoint descriptor */
    cfg_desc_p[39] = GET_ENDPOINT_ADDR(rndis_device.notification_ep);
    cfg_desc_p[40] = RG_USB_INTERRUPT; /* Interrupt */
    cfg_desc_p[41] = RNDIS_NOTIFICATION_EP_MAX_SIZE; /* max_pkt LSB */
    cfg_desc_p[42] =  0x00; /* max_pkt MSB */
    cfg_desc_p[43] = 0x01; /* polling interval */

    /* Data interface */
    cfg_desc_p[44] = SIZEOF_IF_DESC;
    cfg_desc_p[45] = INTERFACE_DESC;/* if descriptor */
    cfg_desc_p[46] = 0x01; /* Second interface in this configuration */
    cfg_desc_p[47] = 0x00; /* No alternate setting */
    cfg_desc_p[48] = 0x02; /* two endpoints */
    cfg_desc_p[49] = 0x0A; /* data class */
    cfg_desc_p[50] = 0x00; /* No sub-calss */
    cfg_desc_p[51] = 0x00; 
    cfg_desc_p[52] = 0x00; /* No string descriptor for this interface */

    /* Data out endpoint */
    cfg_desc_p[53] = SIZEOF_EP_DESC;
    cfg_desc_p[54] = ENDPOINT_DESC; 
    cfg_desc_p[55] = GET_ENDPOINT_ADDR(rndis_device.rx_data_ep);
    cfg_desc_p[56] = RG_USB_BULK ;
    cfg_desc_p[57] = RNDIS_DATA_EP_MAX_SIZE; /* max_pkt LSB */
    cfg_desc_p[58] = 0x00; /* max_pkt MSB */
    cfg_desc_p[59] = 0x00; /* Ignored for bulk endpoints */

    /* Data in endpoint */
    cfg_desc_p[60] = SIZEOF_EP_DESC;
    cfg_desc_p[61] = ENDPOINT_DESC; 
    cfg_desc_p[62] = GET_ENDPOINT_ADDR(rndis_device.tx_data_ep);
    cfg_desc_p[63] = RG_USB_BULK ;
    cfg_desc_p[64] = RNDIS_DATA_EP_MAX_SIZE; /* max_pkt LSB */
    cfg_desc_p[65] = 0x00; /* max_pkt MSB */
    cfg_desc_p[66] = 0x00; /* Ignored for bulk endpoints */

    /* 3.initialize strings descriptors */ 
    lang_desc->length = 0x04;
    lang_desc->type = STRING_DESC; 
    lang_desc->str[0] = 0x09; 
    lang_desc->str[1] = 0x04;
    mfg_desc->type = STRING_DESC; 
    strcpy(temp_str, USB_DEV_VENDOR_DESCRIPTION);
    for (i = 0, p_str = temp_str; i < strlen(temp_str); i++)
    {
	mfg_desc->str[2 * i] = p_str[i];
	mfg_desc->str[2 * i + 1] = 0;

    }
    mfg_desc->length = strlen(temp_str) * 2 + 2; 
    memset(temp_str, 0, 50);
    prod_desc->type = STRING_DESC; 
    strcpy(temp_str, USB_DEV_PRODUCT_DESCRIPTION);
    for (p_str = temp_str, i = 0; i < strlen(temp_str); i++)
    {
	prod_desc->str[2 * i] = p_str[i];
	prod_desc->str[2 * i + 1] = 0;
    }

    /* size of this string descriptor */
    prod_desc->length = strlen(temp_str) * 2 + 2;
    return 0;
}

static void rndis_usb_config(__u8 cfg_num, __u32 alt_if_num, endpoint_t **cfg,
    int *eps_num)
{
    *eps_num = RNDIS_ENDPOINTS_NUM;
    *cfg = &rndis_device.tx_data_ep;
}

/* Stubs for call-backs that being registered at USB layer */
static int rndis_usb_ep_send_stub(__u8 *buff, int len, endpoint_t *ep, 
    void *context1, void *context2)
{	
    printk(__FUNCTION__": ep_send was not initialized\n");
    return -1;
}

static void rndis_usb_terminate_transfers_stub(void)
{
    printk(__FUNCTION__": usb_slave_terminate_current_transfers was not "
	"initialized\n");
}

#ifdef CONFIG_RG_USB_SLAVE_MODULE    
static void on_rndis_closed_stub(void)
{
    printk(__FUNCTION__": usb_slave_upper_device_closed was not "
	"initialized\n");
}   
#endif    

static void rndis_on_receive_usb_data_pkt_stub(void *context)
{
    printk(__FUNCTION__": usb_slave_on_receive_usb_data_pkt was not "
	"initialized\n");
}
