/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#ifndef RNDIS_H
#define RNDIS_H
#include <linux/netdevice.h>
#include <linux/skbuff.h>
#include "../usb_slave_common.h"
#include "usb_queue.h" /* queue defenitions */
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#if LINUX_VERSION_CODE <  KERNEL_VERSION(2,4,0)
#include <linux/netdev_22_24.h>
#endif

#define DEVICE_MAX_TRANSFER_SIZE MAX_ALLOWED_TRANSFER_SIZE
#define RNDIS_ERROR_TX_FAILED -1
#define RNDIS_ERROR_DISCONNECTED -2

/* RNDIS states */
#define RNDIS_UNINITIALIZED 0
#define RNDIS_INITIALIZED 1
#define RNDIS_DATA_INITIALIZED 2

typedef struct {
    endpoint_t tx_data_ep;
    endpoint_t rx_data_ep;
    endpoint_t notification_ep;
    int current_state;
    int current_req_id;
    struct usb_head control_response_q;
    int notification_ep_q_size;
    struct net_device *eth_dev;
    volatile int last_update_connection_state; 
    volatile int is_usb_connected;
    volatile int usb_connection_state; 
    __u8 receive_buffer[DEVICE_MAX_TRANSFER_SIZE];
    int current_max_ep_addr;
    usb_device_ops *ops;
} rndis_device_t;

/* RNDIS functions */
int rndis_send_data(struct sk_buff **skb, void *context);
int rndis_init(struct net_device *dev);
void rndis_send_media_disconnected_msg(void);
void rndis_send_media_connected_msg(void);

/* USB ETHERNET DEVICE functions */
int rndis_eth_probe(struct net_device *dev);
void rndis_eth_rx(struct net_device *dev, int len, unsigned char *buf);
struct net_device_stats *rndis_eth_stats(struct net_device *dev);
void rndis_eth_tx_end(void *context);
#endif
