/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#include "rndis.h"
#include <linux/kernel.h> 
#include <linux/module.h> 
#include <linux/sockios.h>
#include <asm/uaccess.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/etherdevice.h>
#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif
#if LINUX_VERSION_CODE <  KERNEL_VERSION(2,4,0)
#define netif_set_tx_timeout(dev, func, deltajiffs)   do {; } while (0)
#define netif_pause_tx_queue(dev) (test_and_set_bit(0, (void*)&(dev)->tbusy))
#define netif_unpause_tx_queue(dev) do { clear_bit(0, (void*)&(dev)->tbusy); } while (0)
#else
/* Only used in transmit path.  No function in 2.4. */
#define netif_pause_tx_queue(dev)  netif_stop_queue(dev)
//#define netif_unpause_tx_queue(dev) do {; } while (0)
#define netif_unpause_tx_queue(dev) netif_wake_queue(dev)
#define netif_set_tx_timeout(dev, tf, tm) \
    do { \
	(dev)->tx_timeout = (tf); \
	    (dev)->watchdog_timeo = (tm); \
    } while(0)

#endif
#define GET_USB_CONN_STAT_IOCTL SIOCDEVPRIVATE
static int rndis_eth_open(struct net_device *dev);
static int rndis_eth_stop(struct net_device *dev);
static int rndis_eth_tx(struct sk_buff *skb, struct net_device *dev);
static int rndis_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd);
static int rndis_eth_set_mac_addr(struct net_device *dev, void *p);

char rndis_device_mac_str[6] = {0x00, 0x00, 0x00, 0x02, 0x00, 0x00};
char rndis_host_mac_str[6] = {0x00, 0xDE, 0xAD, 0xBE, 0xEF, 0x00};

struct rndis_priv {
    struct net_device_stats stats;
};

extern rndis_device_t rndis_device;

struct net_device usb_eth_dev = 
{
    "usb0", /* FIXME: only temporery until we find a way to get it from device 
	     * list
	     */ 
    0, 0, 0, 0,
    0x000,
    0,
    0, 0, 0,
    NULL,
    rndis_eth_probe,
};

int remote_ndis_init(void)
{
    
    if (register_netdev(&usb_eth_dev))
    {
	printk(__FUNCTION__": error registrating netdev\n");
	return -1;
    }
    return 0;

}

#ifdef MODULE

int init_module(void)
{
    return remote_ndis_init();
}

void cleanup_module(void)
{
    rndis_device.ops->device_class_uninitialized(); 
    
    /* free all rndis resorces */
    
#ifdef CONFIG_RG_USB_SLAVE_MODULE       
    rndis_device.ops->on_device_class_closed();    
#endif    
    kfree(usb_eth_dev.priv);
    unregister_netdev(&usb_eth_dev);
}
#endif

static void rndis_eth_timeout(struct net_device *dev)
{
    /* TODO: define what to do on timeout in the USB level */

    netif_wake_queue(dev);
}

int rndis_eth_probe(struct net_device *dev)
{
    ether_setup(dev); 
    dev->open = rndis_eth_open;
    dev->stop = rndis_eth_stop;
    dev->hard_start_xmit = rndis_eth_tx;
    dev->do_ioctl = rndis_eth_ioctl;
    dev->get_stats = rndis_eth_stats;
    dev->set_mac_address = rndis_eth_set_mac_addr;
    dev->tx_timeout = rndis_eth_timeout;
    dev->watchdog_timeo = 5*HZ;

    dev->priv = kmalloc(sizeof(struct rndis_priv), GFP_KERNEL); 
    if (!dev->priv)
	return -ENOMEM;
    memset(dev->priv, 0, sizeof(struct rndis_priv));
    return rndis_init(dev);
}

static int rndis_eth_open(struct net_device *dev)
{
    netif_device_attach(dev);
    rndis_send_media_connected_msg();
#ifdef MODULE    
    MOD_INC_USE_COUNT;
#endif    
    return 0;
}

static int rndis_eth_stop(struct net_device *dev)
{
    netif_stop_queue(dev);
#ifdef MODULE    
    MOD_DEC_USE_COUNT;
#endif    
    return 0;
}

void rndis_eth_tx_end(void *context)
{
    endpoint_t *ep = (endpoint_t *)context;
    struct net_device *dev = (struct net_device *)(ep->abs_info->end_context1);
    struct sk_buff *skb = (struct sk_buff *)(ep->abs_info->end_context2);
    struct rndis_priv *privp = NULL;
    
    if (!dev || !(privp = (struct rndis_priv *)dev->priv))
	goto Exit;
    ep->abs_info->buff_pos = 0;
    ep->abs_info->buff_size = 0;
    ep->abs_info->is_busy = 0;
    
    /* Error in sending */
    if (ep->abs_info->end_arg1 < 0)
    {
	privp->stats.tx_errors++;
	goto Exit;
    }
    rndis_device.last_update_connection_state = USB_CONNECTED;

    privp->stats.tx_packets++;
    privp->stats.tx_bytes += ep->abs_info->end_arg2;
Exit:
    dev_kfree_skb_any(skb);
    if (dev)
    {
      netif_wake_queue(dev);
    }
}

static int rndis_eth_tx(struct sk_buff *skb, struct net_device *dev)
{
    struct rndis_priv *privp = NULL;
   
    if (!dev || !(privp = (struct rndis_priv *)dev->priv))
	return 0;
    if (!skb)
        return 0;

    netif_stop_queue(dev);
    /* Send packet via rndis/usb */
    if (rndis_send_data(&skb,(void *)dev) < 0)
    {
	privp->stats.tx_dropped++;
	dev_kfree_skb_any(skb);
	netif_wake_queue(dev);
    }
    
    dev->trans_start = jiffies; /* save the timestamp */
    
    return 0;
}

void rndis_eth_rx(struct net_device *dev, int len, unsigned char *buf)
{
    struct sk_buff *skb;
    struct rndis_priv *privp = NULL;
    __u8 *tmp; 
    
    if (!dev || !(privp = (struct rndis_priv *)dev->priv))
	return;
    skb = dev_alloc_skb(len + 2);
    if (!skb)
    {
	printk("rndis_rx: low on mem, dropping packet\n");
	privp->stats.rx_dropped++;
	return;
    }
    
    skb->dev = dev;
    skb_reserve(skb, 2);/* Align IP on 16 byte boundaries */
    tmp = skb_put(skb, len);
    memcpy(tmp, buf, len);
    dev->last_rx = jiffies;
    skb->protocol = eth_type_trans(skb, dev);
    privp->stats.rx_packets++;
    privp->stats.rx_bytes += len;
#ifdef USB_SLAVE_DEBUG
    usb_slave_dump_buffer(skb->data, skb->len);
#endif    
    netif_rx(skb);
}

static u8 *rndis_eth_mac_increment(u8 *mac)
{
    int byte, overflow;
#ifdef CONFIG_RG_MAC_COMPANY_ID
    overflow = 2;
#else
    overflow = 0;
#endif

    for (byte = 5; byte > overflow ; byte--)
    {
	mac[byte]++;
	if (mac[byte])
	    return mac; /* No overflow */
    }
    /* Overflow - increment the first byte */
    mac[5] = 1;

    return mac;
}

static int rndis_eth_set_mac_addr(struct net_device *dev, void *p)
{
    struct sockaddr *addr = (struct sockaddr *)p;
    
    if (netif_running(dev))
	return -EBUSY;
    memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);
    memcpy(rndis_device_mac_str, addr->sa_data, 6);
    memcpy(rndis_host_mac_str, addr->sa_data, 6);
    rndis_eth_mac_increment((u8 *)rndis_host_mac_str);

    return 0;
}

struct net_device_stats *rndis_eth_stats(struct net_device *dev)
{
    struct rndis_priv *privp ;
   
    if (!dev || !(privp = (struct rndis_priv *)dev->priv))
	return NULL;
    return &privp->stats;
}

static int rndis_eth_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
    switch (cmd)
    {
    case GET_USB_CONN_STAT_IOCTL:
#ifndef CONFIG_USB_CONN_MONITOR
        rndis_device.is_usb_connected = 
	    (rndis_device.last_update_connection_state == USB_CONNECTED) &&
	    (rndis_device.current_state == RNDIS_DATA_INITIALIZED) ? 1 :
	    0;
#endif	
	copy_to_user(ifr->ifr_data, (void *)&rndis_device.is_usb_connected, 
	    sizeof(int));
	break;
    default: return -EOPNOTSUPP;
    }
    return 0;
}
