/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#ifndef USB_QUEUE_H 
#define USB_QUEUE_H

struct usb_head {
    struct usb_hdr *next, *prev;
    
    /* Must be same length as a pointer for using debugging */
    __u32 qlen;					   
    int is_init;
    spinlock_t lock;
};

typedef struct usb_hdr {
    struct usb_hdr *next, *prev;
    struct usb_head *list;
    int size;
    __u8 buff[1];
} usb_hdr2;

extern void usb_queue_head_init(struct usb_head *list);
extern void usb_queue_tail(struct usb_head *list, struct usb_hdr *buf);
extern void usb_unlink(struct usb_hdr *buf);

extern __inline__ void usb_queue_head_init(struct usb_head *list)
{
    list->prev = (struct usb_hdr *)list;
    list->next = (struct usb_hdr *)list;
    list->qlen = 0;
}

extern __inline__ void __usb_queue_tail(struct usb_head *list, 
    struct usb_hdr *newsk)
{
    struct usb_hdr *prev, *next;

    newsk->list = list;
    list->qlen++;
    next = (struct usb_hdr *)list;
    prev = next->prev;
    newsk->next = next;
    newsk->prev = prev;
    next->prev = newsk;
    prev->next = newsk;
}

extern __inline__ void usb_queue_tail(struct usb_head *list, 
    struct usb_hdr *newsk)
{
    unsigned long flags;

    spin_lock_irqsave(&(list->lock), flags);
    __usb_queue_tail(list, newsk);
    spin_unlock_irqrestore(&(list->lock), flags);
}

extern __inline__ void __usb_unlink(struct usb_hdr *item, 
    struct usb_head *list)
{
    struct usb_hdr * next, * prev;

    list->qlen--;
    next = item->next;
    prev = item->prev;
    item->next = NULL;
    item->prev = NULL;
    item->list = NULL;
    next->prev = prev;
    prev->next = next;
}

extern __inline__ void usb_unlink(struct usb_hdr *item)
{
    unsigned long flags;

    if(item->list)
    {
	spin_lock_irqsave(&((item->list)->lock), flags);
	__usb_unlink(item, item->list);
	spin_unlock_irqrestore(&((item->list)->lock), flags);
    }
}

extern __inline__ void usb_queue_free(struct usb_head *list)
{
    struct usb_hdr *temp;
    unsigned long flags;

    spin_lock_irqsave(&(list->lock), flags);
    for (temp = list->next ; list->qlen > 0; temp = list->next)
    {
	usb_unlink(temp);
	kfree(temp);
    }
    spin_unlock_irqrestore(&(list->lock), flags);

}
#endif



