/*****************************************************************************
 * 
 *  USB Slave Support Software 
 *  Linux Version
 *
 *  Developed by: 
 *  Jungo Software Technologies - Residential Gateway Software Division
 *  www.jungo.com/openrg
 *  info@jungo.com
 *  
 *  
 *  Jungo Software Technologies (Jungo) and Integarated Device Technology (IDT) 
 *  jointly own all rights in the USB Slave Support Software images and source 
 *  code. 
 *  You may not, without express written permission from either Jungo or IDT,
 *  a. Distribute, sell, reproduce or copy the source code in any way.
 *  b. Rent or lease the USB Slave Support Software or otherwise transfer or 
 *  assign the right to use the USB Slave Support Software.
 *
 *
 *  
 ****************************************************************************/


#ifndef USB_SLAVE_H
#define USB_SLAVE_H
#include "usb_slave_common.h"

/* Transfers status */
#define USB_TRANSFER_SUCCESS 0
#define USB_TRANSFER_ERROR -1
#define USB_TRANSFER_NOT_STARTED 0
#define USB_TRANSFER_STARTED 1 
#define USB_TRANSFER_ENDED 2

/* Recipient */
#define RCPT_DEV	0x0
#define RCPT_IF		0x1
#define RCPT_EP		0x2
#define RCPT_OTHER	0x3

typedef enum {
    INTERRUPT_CONNECTED = 0x0000001, 
    INTERRUPT_DISCONNECTED = 0x0000002, 
    INTERRUPT_POWERED = 0x0000004, 
    INTERRUPT_ADDRESSED = 0x0000008, 
    INTERRUPT_CONFIGURED = 0x0000010, 
    INTERRUPT_RESET = 0x00000020, 
    INTERRUPT_SUSPEND = 0x00000040, 
    INTERRUPT_RESUME = 0x00000080, 
    INTERRUPT_SOF = 0x00000100, 
    INTERRUPT_ENDPOINT = 0x00000200, 
    INTERRUPT_CTRL_EP_SETUP = 0x00000800, 
} interrupt_status_t;

typedef enum {
    RG_ENDPOINT_HALT = 0x0,
    RG_DEVICE_REMOTE_WAKEUP = 0x1,
} feature_selector_t;

int usb_slave_phys_init(void (* misc_interrupt_handler)(__u32 status, 
    int ep_addr));

int usb_slave_phys_uninit(void);

int usb_slave_on_reset_cmd(void);

int usb_slave_init_eps_phys_info(endpoint_t *ctrl_ep, endpoint_t *ep_array,
    int array_size);

int usb_slave_handle_set_address_req(__u32 addr);

int usb_slave_handle_clear_feature_req(__u8 feature_selector, endpoint_t *ep);

int usb_slave_handle_set_feature_req(__u8 feature_selector, endpoint_t *ep);

void usb_slave_handle_get_status_req(__u8 recipient, endpoint_t *ep, 
    __u8 **status_resp_p);

int usb_slave_enable_ep(endpoint_t *ep);

int usb_slave_disable_ep(endpoint_t *ep);

void usb_slave_stall_ep(endpoint_t *ep);

int usb_slave_handle_non_control_ep(endpoint_t *ep, __u8 *buff_p, int size);

int usb_slave_start_dma_transfer(endpoint_t *ep);

int usb_slave_handle_control_ep(endpoint_t *ep, usb_ep_dir_t direction, 
    control_phase_t phase);

void usb_slave_terminate_current_transfers(void);

void usb_slave_enable_transfer(endpoint_t *ep);

void usb_slave_disable_transfer(endpoint_t *ep);

#endif

