/*
 *
 * BRIEF MODULE DESCRIPTION
 *	Definitions for IDT RC32355 CPU.
 *
 * Copyright 2002 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *         	stevel@mvista.com or source@mvista.com
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _MERLOT_H_
#define _MERLOT_H_

#include <linux/delay.h>
#include <asm/io.h>
#include <asm/bootinfo.h>
#include "merlot/gpio.h"

#define RC32434_REG_BASE  0x18000000
#define AR5312_REG_BASE   0xbc000000

#define timer     ((volatile TIM_t)  TIM0_VirtualAddress)

// only for muscat

enum
{
	PCI0_PhysicalAddress	= 0x18080000,
	PCI_PhysicalAddress	= PCI0_PhysicalAddress,

	PCI0_VirtualAddress	= 0xb8080000,
	PCI_VirtualAddress	= PCI0_VirtualAddress,
} ;

enum
{
	PCI_LbaCount	= 4,		// Local base addresses.
} ;

typedef struct
{
	U32	a ;		// Address.
	U32	c ;		// Control.
	U32	m ;		// mapping.
} PCI_Map_s ;
typedef struct
{
	U32		pcic ;
	U32		pcis ;
	U32		pcism ;
	U32		pcicfga ;
	U32		pcicfgd ;
	PCI_Map_s	pcilba [PCI_LbaCount] ;
	U32		pcidac ;
	U32		pcidas ;
	U32		pcidasm ;
	U32		pcidad ;
	U32		pcidma4c ;
	U32		pcidma5c ;
	U32		pcitc ;
} volatile *PCI_t ;

#define gpio	  ((volatile GPIO_t) GPIO0_VirtualAddress_muscat) 
#define rc32434_pci ((volatile PCI_t) PCI0_VirtualAddress)
// done only for muscat


#define IDT_BUS_FREQ mips_bus_freq
#define IDT_CLOCK_MULT 2

/* Interrupt Controller */
#define NUM_INTR_GROUPS    5

/* 16550 UARTs */

#define GROUP0_IRQ_BASE 8		       /* GRP2 IRQ numbers start here */
#define GROUP1_IRQ_BASE (GROUP0_IRQ_BASE + 32) /* GRP3 IRQ numbers start here */
#define GROUP2_IRQ_BASE (GROUP1_IRQ_BASE + 32) /* GRP4 IRQ numbers start here */
#define GROUP3_IRQ_BASE (GROUP2_IRQ_BASE + 32) /* GRP5 IRQ numbers start here */
#define GROUP4_IRQ_BASE (GROUP3_IRQ_BASE + 32)

#define TIMER_RELOAD_REG (AR5312_REG_BASE + 0x03004)

#define MERLOT_WLAN1_IRQ   2		// bit 10 in CP0_status register
#define MERLOT_ENET_IRQ    3            // bit 11 in CP0_status register
#define MERLOT_WLAN_IRQ    5		// bit 13 in CP0_status register
#define MERLOT_MISC_IRQ	   6		// bit 14 in CP0_status register = GROUP 0
#define MIPS_CPU_TIMER_IRQ 7		// bit 15 in CP0_status register

#ifdef __MIPSEB__
#define MERLOT_UART_BASE (mips_uart_base + 3)
#else
#define MERLOT_UART_BASE mips_uart_base
#endif
#define MERLOT_BASE_BAUD  (IDT_BUS_FREQ * 1000 * 1000 / 16)

#define MERLOT_NR_IRQS      168

#define MERLOT_TIMER_IRQ    8		// group base 8 irq bit 0, bit 14 in CP0_status set
#define MERLOT_AHBPROC_IRQ  9		// group base 8 irq bit 1, bit 14 in CP0_status set
#define MERLOT_AHBDMA_IRQ   10		// group base 8 irq bit 2, bit 14 in CP0_status set
#define MERLOT_GPIO_IRQ     11		// group base 8 irq bit 3, bit 14 in CP0_status set

#define MERLOT_UART_IRQ     mips_uart_irq
#define MERLOT_UARTDMA_IRQ  13		// group base 8 irq bit 5, bit 14 in CP0_status set
#define MERLOT_WATCHDOG_IRQ 14		// group base 8 irq bit 6, bit 14 in CP0_status set
#define MERLOT_LOCALBUS_IRQ 15		// group base 8 irq bit 7, bit 14 in CP0_status set



#define local_readl(addr) __raw_readl(addr)
#define local_writel(l,addr) __raw_writel(l,addr)

/* cpu pipeline flush */
static inline void merlot_sync(void)
{
        __asm__ volatile ("sync");
}

static inline void merlot_sync_udelay(int us)
{
        __asm__ volatile ("sync");
        udelay(us);
}

static inline void merlot_sync_delay(int ms)
{
        __asm__ volatile ("sync");
        mdelay(ms);
}
static inline u8 merlot_readb(unsigned long pa)
{
	return *((volatile u8 *)KSEG1ADDR(pa));
}
static inline u16 merlot_readw(unsigned long pa)
{
	return *((volatile u16 *)KSEG1ADDR(pa));
}
static inline u32 merlot_readl(unsigned long pa)
{
	return *((volatile u32 *)KSEG1ADDR(pa));
}
static inline void merlot_writeb(u8 val, unsigned long pa)
{
	*((volatile u8 *)KSEG1ADDR(pa)) = val;
}
static inline void merlot_writew(u16 val, unsigned long pa)
{
	*((volatile u16 *)KSEG1ADDR(pa)) = val;
}
static inline void merlot_writel(u32 val, unsigned long pa)
{
	*((volatile u32 *)KSEG1ADDR(pa)) = val;
}

/*
 * C access to CLZ and CLO instructions
 * (count leading zeroes/ones).
 */
static inline int merlot_clz(unsigned long val)
{
	int ret;
        __asm__ volatile (
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips32\n\t"
		"clz\t%0,%1\n\t"
                ".set\tmips0\n\t"
                ".set\tat\n\t"
                ".set\treorder"
                : "=r" (ret)
		: "r" (val));

	return ret;
}
static inline int merlot_clo(unsigned long val)
{
	int ret;
        __asm__ volatile (
		".set\tnoreorder\n\t"
		".set\tnoat\n\t"
		".set\tmips32\n\t"
		"clo\t%0,%1\n\t"
                ".set\tmips0\n\t"
                ".set\tat\n\t"
                ".set\treorder"
                : "=r" (ret)
		: "r" (val));

	return ret;
}
#endif /* _MERLOT_H_ */
