#ifndef __IDT_GPIO_H__
#define __IDT_GPIO_H__

/*******************************************************************************
 *
 * Copyright 2002 Integrated Device Technology, Inc.
 *		All rights reserved.
 *
 * GPIO register definition.
 *
 * File   : $Id: gpio.h,v 1.2 2002/06/06 18:34:04 astichte Exp $
 *
 * Author : ryan.holmQVist@idt.com
 * Date   : 20011005
 * Update :
 *	    $Log: gpio.h,v $
 *	    Revision 1.2  2002/06/06 18:34:04  astichte
 *	    Added XXX_PhysicalAddress and XXX_VirtualAddress
 *	
 *	    Revision 1.1  2002/05/29 17:33:22  sysarch
 *	    jba File moved from vcode/include/idt/acacia
 *	
 *
 ******************************************************************************/

#include  <asm/rc32300/types.h> 
enum
{
  GPIO0_PhysicalAddress	= 0x18048000,
  GPIO_PhysicalAddress	= GPIO0_PhysicalAddress,	// Default
  
  GPIO0_VirtualAddress	= 0xb8048000,
  GPIO_VirtualAddress	= GPIO0_VirtualAddress,		// Default
} ;

typedef struct
{
  U32   gpiofunc;   /* GPIO Function Register
		     * gpiofunc[x]==0 bit = gpio
		     * func[x]==1  bit = altfunc
		     */
  U32   gpiocfg;    /* GPIO Configuration Register
		     * gpiocfg[x]==0 bit = input
		     * gpiocfg[x]==1 bit = output
		     */
  U32   gpiod;	    /* GPIO Data Register
		     * gpiod[x] read/write gpio pinX status
		     */
  U32   gpioilevel; /* GPIO Interrupt Status Register
		     * interrupt level (see gpioistat)
		     */
  U32   gpioistat;  /* Gpio Interrupt Status Register
		     * istat[x] = (gpiod[x] == level[x])
		     * cleared in ISR (STICKY bits)
		     */
  U32   gpionmien;  /* GPIO Non-maskable Interrupt Enable Register */
} volatile * GPIO_t ;

typedef enum
  {
    GPIO_gpio_v		= 0,		// gpiofunc use pin as GPIO.
    GPIO_alt_v		= 1,		// gpiofunc use pin as alt.
    GPIO_input_v	= 0,		// gpiocfg use pin as input.
    GPIO_output_v	= 1,		// gpiocfg use pin as output.
    GPIO_pin0_b		= 0,
    GPIO_pin0_m		= 0x00000001,
    GPIO_pin1_b		= 1,
    GPIO_pin1_m		= 0x00000002,
    GPIO_pin2_b		= 2,
    GPIO_pin2_m		= 0x00000004,
    GPIO_pin3_b		= 3,
    GPIO_pin3_m		= 0x00000008,
    GPIO_pin4_b		= 4,
    GPIO_pin4_m		= 0x00000010,
    GPIO_pin5_b		= 5,
    GPIO_pin5_m		= 0x00000020,
    GPIO_pin6_b		= 6,
    GPIO_pin6_m		= 0x00000040,
    GPIO_pin7_b		= 7,
    GPIO_pin7_m		= 0x00000080,
    GPIO_pin8_b		= 8,
    GPIO_pin8_m		= 0x00000100,
    GPIO_pin9_b		= 9,
    GPIO_pin9_m		= 0x00000200,
    GPIO_pin10_b	= 10,
    GPIO_pin10_m	= 0x00000400,
    GPIO_pin11_b	= 11,
    GPIO_pin11_m	= 0x00000800,
    GPIO_pin12_b	= 12,
    GPIO_pin12_m	= 0x00001000,
    GPIO_pin13_b	= 13,
    GPIO_pin13_m	= 0x00002000,
    GPIO_pin14_b	= 14,
    GPIO_pin14_m	= 0x00004000,
    GPIO_pin15_b	= 15,
    GPIO_pin15_m	= 0x00008000,

// Alternate function pins.  Corrsponding gpiofunc bit set to GPIO_alt_v.

    GPIO_u0sout_b	= GPIO_pin0_b,		// UART 0 serial out.
    GPIO_u0sout_m	= GPIO_pin0_m,
    GPIO_u0sout_cfg_v	= GPIO_output_v,

    GPIO_u0sinp_b	= GPIO_pin1_b,			// UART 0 serial in.
    GPIO_u0sinp_m	= GPIO_pin1_m,
    GPIO_u0sinp_cfg_v	= GPIO_input_v,

    GPIO_maddr22_b	= GPIO_pin2_b, 	// M&P bus bit 22.
    GPIO_maddr22_m	= GPIO_pin2_m,
    GPIO_maddr22_cfg_v	= GPIO_output_v,

    GPIO_maddr23_b	= GPIO_pin3_b, 	// M&P bus bit 23.
    GPIO_maddr23_m	= GPIO_pin3_m,
    GPIO_maddr23_cfg_v	= GPIO_output_v,
   
    GPIO_maddr24_b	= GPIO_pin4_b, 	// M&P bus bit 24.
    GPIO_maddr24_m	= GPIO_pin4_m,
    GPIO_maddr24_cfg_v	= GPIO_output_v,
    
    GPIO_maddr25_b	= GPIO_pin5_b, 	// M&P bus bit 25.
    GPIO_maddr25_m	= GPIO_pin5_m,
    GPIO_maddr25_cfg_v	= GPIO_output_v,
    
    GPIO_rngclk_b 	= GPIO_pin6_b, 	// reserved.
    GPIO_rngclk_m 	= GPIO_pin6_m,
    GPIO_rngclk_cfg_v	= GPIO_input_v,

    GPIO_sdckenp_b 	= GPIO_pin7_b, 	// reserved.
    GPIO_sdckenp_m 	= GPIO_pin7_m,
    GPIO_sdckenp_cfg_v	= GPIO_output_v,

    GPIO_cen1_b 	= GPIO_pin8_b, 	// reserved.
    GPIO_cen1_m 	= GPIO_pin8_m,
    GPIO_cen1_cfg_v	= GPIO_output_v,

    GPIO_cen2_b 	= GPIO_pin9_b, 	// reserved.
    GPIO_cen2_m 	= GPIO_pin9_m,
    GPIO_cen2_cfg_v	= GPIO_output_v,

    GPIO_regn_b 	= GPIO_pin10_b, 	// reserved.
    GPIO_regn_m 	= GPIO_pin10_m,
    GPIO_regn_cfg_v	= GPIO_output_v,

    GPIO_iordn_b 	= GPIO_pin11_b, 	// reserved.
    GPIO_iordn_m 	= GPIO_pin11_m,
    GPIO_iordn_cfg_v	= GPIO_output_v,

    GPIO_iowrn_b 	= GPIO_pin12_b, 	// reserved.
    GPIO_iowrn_m 	= GPIO_pin12_m,
    GPIO_iowrn_cfg_v	= GPIO_output_v,
    
    GPIO_pcireqn2_b	= GPIO_pin13_b, 	// PCI messaging int.
    GPIO_pcireqn2_m	= GPIO_pin13_m,
    GPIO_pcireqn2_cfg_v	= GPIO_input_v,

    GPIO_pcigntn2_b	= GPIO_pin14_b, 	// PCI messaging int.
    GPIO_pcigntn2_m	= GPIO_pin14_m,
    GPIO_pcigntn2_cfg_v	= GPIO_output_v,

    GPIO_pcimuintn_b	= GPIO_pin15_b, 	// PCI messaging int.
    GPIO_pcimuintn_m	= GPIO_pin15_m,
    GPIO_pcimuintn_cfg_v	= GPIO_output_v,
    
  } GPIO_DEFS_t;

#endif	// __IDT_GPIO_H__

