/*******************************************************************************
 *
 * idt_cryto_api.h 
 *
 * rc32365 Crypto API header file to be included by applications.
 *
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef _IDT_CRYPTO_API_H_
#define _IDT_CRYPTO_API_H_

/* 
 * OS specific 
 */
#define IDT_RC32365_LINUX
#ifdef IDT_RC32365_LINUX
#include <linux/skbuff.h>
#define USE_IO_BUF		/* zero-memory-copy (default buffer scheme) */
#endif

#define PKT_API_BUF_INCLUDED	/* user allocates crypto request buffer used by API for a bit of efficient packet processing */ 
#ifdef PKT_API_BUF_INCLUDED
#define PKT_API_BUF_SIZE 76	/* buffer size appended at the end of PACKET_SECURITY_CONTEXT struc */
#endif
/*
 * Data Types
 */
typedef unsigned char byte;
typedef unsigned short uint16;
typedef unsigned long uint32;
typedef unsigned int SECURITY_CONTEXT_ID;

#define MAX_SECURITY_CONTEXT	512	/* max. number of security context */

/*
 * API Return status
 */
#define	idt_RET_OK			0	/* Successful	*/									
#define	idt_RET_ERROR			1	/* Error		*/ 
#define idt_RET_RESOURCE		2	/* System resource allocation error */
#define	idt_RET_NULL_CB			3	/* NULL callback function pointer	*/
#define idt_RET_TEST_SYNC_MODE_USED	4	/* In SYNC test mode, only one request allowed	*/
#define idt_RET_BUF			5	/* Buffer error (null pointer or invalid length) */
#define idt_RET_SYNC_CALL_ONLY		6	/* Only synchronous all is supported */
#define idt_RET_DEVICE_QUEUE_FULL	7	/* Crypto device queue full, request dropped	*/	
#define idt_RET_PKT_API_PARM_ERR	8	/* Packet API parameter error	*/	

#define idt_RET_CMD_ERROR		10	/* Command error (e.g, nonexist, conflict)	*/
#define idt_RET_CMD_GEN_OP		11	/* Invalid general operation	*/
#define idt_RET_CMD_GEN_MODE		12	/* Invalid general operation mode	*/
#define idt_RET_CMD_GEN_PROT		13	/* Not Supported protocol	*/
#define idt_RET_CMD_LOAD_PARAM		14	/* Invalid parameter in LOAD context command	*/
#define idt_RET_CMD_STORE_PARAM		15	/* Invalid parameter in STORE context command	*/
#define idt_RET_CMD_REPLACE_PARAM	16	/* Invalid parameter in REPLACE context command	*/
#define idt_RET_CONTEXT_RAM_ENTRY_UNINITED	17	/* Uninitialized context RAM entry	*/
#define idt_RET_CONTEXT_ID_INVALID	18	/* Invalid context id	*/

#define idt_RET_CMD_ENC_MODE		20	/* Invalid encryption mode	*/
#define idt_RET_CMD_ENC_KEY		21	/* Invalid encryption key	*/
#define idt_RET_CMD_HV			22	/* Invalid hash value		*/
#define idt_RET_CMD_IV			23	/* Invalid IV			*/
#define idt_RET_CMD_UIV			24	/* Invalid UIV - not CBC mode	*/
#define idt_RET_CMD_IVP			25	/* Invalid IVP - not CBC mode	*/
#define idt_RET_CMD_IV_PROCESSING	26

#define idt_RET_CMD_PAD_BLOCK_SIZE	30	/* Invalid or missing pad unit block size	*/

#define idt_RET_CMD_HMODE		40	/* Invalid Hash Mode	*/
#define idt_RET_CMD_HC			40	/* Invalid Hash Check	*/

#define idt_RET_RNG_PARAM_ERROR		50	/* Invalid RNG API parameter	*/
#define idt_RET_INIT_ERROR		51	/* Device initialization failed 	*/
#define idt_RET_DEVICE_ERROR		52	/* Crypto device not initialized or not present */

/* Encryption key size */
#define CONTEXT_ENC_DES_KEY_SIZE	8	/* 64 bits	*/
#define CONTEXT_ENC_AES_128_KEY_SIZE	16	/* 128 bits	*/
#define CONTEXT_ENC_TDES_KEY_SIZE	24	/* 192 bits	*/
#define CONTEXT_ENC_AES_192_KEY_SIZE	24	/* 192 bits	*/

/* Initialization vector size */
#define CONTEXT_ENC_DES_IV_SIZE		8	/* 64 bits	*/
#define CONTEXT_ENC_TDES_IV_SIZE	8	/* 64 bits	*/
#define CONTEXT_ENC_AES_IV_SIZE		16	/* 128 bits	*/

/* Encryption block size */
#define ENC_DES_BLOCK_SIZE		8	/* 8 bytes	*/
#define ENC_TDES_BLOCK_SIZE		8	/* 8 bytes	*/
#define ENC_AES_BLOCK_SIZE		16	/* 16 bytes	*/

/* Hash value size */
#define HASHLEN_HMAC_MD5_96		12	/* 96 bits	*/
#define HASHLEN_HMAC_SHA1_96		12	/* 96 bits	*/

/* HMAC key sizes */
#define MD5HMAC96_KEYSIZE       	16	/* 128 bits	*/
#define SHA1HMAC96_KEYSIZE      	20	/* 160 bits	*/


#define CONTEXT_RAM_HV1_SIZE		20
#define CONTEXT_RAM_HV2_SIZE		20
#define CONTEXT_RAM_IV_SIZE		16	
#define CONTEXT_RAM_EKEY_SIZE		24

#define MAX_ENC_KEY_SIZE	CONTEXT_RAM_EKEY_SIZE
#define MAX_HASH_KEY_SIZE	CONTEXT_RAM_HV1_SIZE
#define MAX_ENC_IV_SIZE		CONTEXT_RAM_IV_SIZE


/* Security Context Operation parameters */ 
typedef struct SECURITY_CONTEXT_OPERATION_PARAM_ {
	/* encryption */
	uint32 eKey[6];	
	uint32 eKeyLength;
	uint32 iv[4];                
	uint32 ivLength;
	uint16 eLength;            
	uint16 eOffset;		
	/* pad */
	byte padByte;
	/* hash */
	uint32 hKey[5];		/* authentication key (unalgned string)	*/
	uint32 hKeyLength;
	uint16 hTruncation;
	uint32 hv1[5];		/* Initialized hash value (aligned)	*/		               
	uint32 hv2[5]; 		/* Initialized hash value (aligned)	*/ 
	uint16 hLength;	
	uint16 hOffset;		
} SECURITY_CONTEXT_OPERATION_PARAM;

/* Context controlword 0 */  
typedef struct CONTEXT_CONTEXT_WORD0_ {
	uint32 hc:2;
	uint32 htm:2;
	uint32 reserved:1;
	uint32 htrunc:5;
	uint32 hmode:4;
	uint32 pblk:2;
	uint32 rp:1;
	uint32 palg:3;
	uint32 ivp:1;
	uint32 uiv:1;
	uint32 emode:4;
	uint32 dsd:1;
	uint32 prot:1;
	uint32 mode:2;
	uint32 op:2;
} SECURITY_CONTEXT_WORD0;

/* Context controlword 1 */ 
typedef struct SECURITY_CONTEXT_WORD1_ {
	uint32 reserved:21;
	uint32 mfp:1;
	uint32 om:2;
	uint32 pb:8;
} SECURITY_CONTEXT_WORD1;

enum {
	WORD1_reserved_b	= 0,
	WORD1_reserved_m	= 0x001fffff,
	WORD1_mfp_b		= 21,
	WORD1_mfp_m		= 0x00200000,
	WORD1_om_b		= 22,
	WORD1_om_m		= 0x00c00000,
	WORD1_pb_b		= 24,
	WORD1_pb_m		= 0xff000000
};

enum {
	WORD0_hc_b		= 0,
	WORD0_hc_m		= 0x00000003,
	WORD0_htm_b		= 2,
	WORD0_htm_m		= 0x0000000c,
	WORD0_reserved_b	= 4,
	WORD0_reserved_m	= 0x00000010,
	WORD0_htrunc_b		= 5,
	WORD0_htrunc_m		= 0x000003E0,
	WORD0_hmode_b		= 10,
	WORD0_hmode_m		= 0x00003C00,
	WORD0_pblk_b		= 14,
	WORD0_pblk_m		= 0x0000C000,
	WORD0_rp_b		= 16,
	WORD0_rp_m		= 0x00010000,
	WORD0_palg_b		= 17,
	WORD0_palg_m		= 0x000E0000,
	WORD0_ivp_b		= 20,
	WORD0_ivp_m		= 0x00100000,
	WORD0_uiv_b		= 21,
	WORD0_uiv_m		= 0x00200000,
	WORD0_emode_b		= 22,
	WORD0_emode_m		= 0x03C00000,
	WORD0_dsd_b		= 26,
	WORD0_dsd_m		= 0x04000000,
	WORD0_prot_b		= 27,
	WORD0_prot_m		= 0x08000000,
	WORD0_mode_b		= 28,
	WORD0_mode_m		= 0x30000000,
	WORD0_op_b		= 30,
	WORD0_op_m		= 0xC0000000
};

typedef struct SECURITY_RAM_CONTEXT_ {
	SECURITY_CONTEXT_WORD0 word0;               
	SECURITY_CONTEXT_WORD1 word1;               
	uint16 hOffset;
	uint16 hLength;
	uint16 eOffset;
	uint16 eLength;             
	uint32 hv1[5];               
	uint32 hv2[5];               
	uint32 iv[4];                
	uint32 ekey[6];
} SECURITY_RAM_CONTEXT;


typedef enum PACKET_OP_FLAG_ {
	OVR_MODE	= 0x0001,	/* Enable override mode, use per-packet length and offset  */
	DIS_MFP		= 0x0002,	/* Disable Mutable Field Processor */
	HASH_TRAILER	= 0x0004,	/* copy hash trailer to buffer */
	PAD_BYTE	= 0x0008,	/* update pad byte */
	INTERNAL_TEST	= 0x0080	//test		
} PACKET_OP_FLAG;

/* Packet-specific Security Context */
typedef struct PACKET_SECURITY_CONTEXT_ {
	int operation;			/* encode (outbound) or decode (inbound) */
	byte *pInputPacketBuffer;
	byte *pOutputPacketBuffer;
	int inputBufferLength;
	int outputBufferLength;
	byte *iphdr;			/* IP header for ESP payload */
	byte padByte;			/* update next header for esp transport mode (outbound) */
	byte offset;			/* per-packet offset (F) */
	unsigned short length;		/* per-packet length (L) */
	unsigned short opFlag;		/* operation flag (see above) */ 
	char hashTrailer[12];		/* hash trailer output */
	int hashTrailerSize;		/* hash trailer output size */
	byte *externalSecurityRAM;
	void *cb;			/* caller's opaque control block to be returned through callback */
	void  (*pCallback)(
		SECURITY_CONTEXT_ID contextID,
		void *cb,		/* caller's opaque control block */
		void *newbuf,		/* driver allocated new sk buffer */
		int callHandle,
		int result);
	int handle;			/* user provided opaque parameter */
#ifdef PKT_API_BUF_INCLUDED
	char pktApiBuf[PKT_API_BUF_SIZE];	/* buffer used by packet API */
#endif
} PACKET_SECURITY_CONTEXT;


/* Security Engine Security Context RAM Command */
typedef enum SE_SECURITY_CONTEXT_COMMAND_ {
	SE_CONTEXT_COMMAND_LOAD			= 0x0001,
	SE_CONTEXT_COMMAND_STORE		= 0x0002,
	SE_CONTEXT_COMMAND_REPLACE		= 0x0004,
	SE_CONTEXT_COMMAND_LOAD_SEGMENT		= 0x0008,
	SE_CONTEXT_COMMAND_LOAD_IV		= 0x0010,
	SE_CONTEXT_COMMAND_STORE_IV		= 0x0020,
	SE_CONTEXT_COMMAND_LOAD_PB		= 0x0040,
	SE_CONTEXT_COMMAND_INIT_HMAC_MD5	= 0x0080,
	SE_CONTEXT_COMMAND_INIT_HMAC_SHA1	= 0x0100,
} SE_SECURITY_CONTEXT_COMMAND;

/* Security Context Opertion Mode - General */
typedef enum SECURITY_CONTEXT_OP_MODE_GENERAL_ {
	/* General */
	SE_SECURITY_CONTEXT_OP_DISABLED			= 0x00000001,  
	SE_SECURITY_CONTEXT_OP_ENCODE			= 0x00000002,
	SE_SECURITY_CONTEXT_OP_DECODE			= 0x00000004,

	SE_SECURITY_CONTEXT_MODE_HASH			= 0x00000008,
	SE_SECURITY_CONTEXT_MODE_ENCRYPT		= 0x00000010,
	SE_SECURITY_CONTEXT_MODE_ENCRYPT_BEFORE_HASH	= 0x00000020,
	SE_SECURITY_CONTEXT_MODE_HASH_BEFORE_ENCRYPT	= 0x00000040,

	SE_SECURITY_CONTEXT_PROT_IPV4			= 0x00000080,
	SE_SECURITY_CONTEXT_PROT_IPV6			= 0x00000100,

	SE_SECURITY_CONTEXT_DSD_DISCARD			= 0x00000200,

	SE_SECURITY_CONTEXT_OVERRIDE_MODE_0		= 0x00000400,
	SE_SECURITY_CONTEXT_OVERRIDE_MODE_1		= 0x00000800,
	SE_SECURITY_CONTEXT_OVERRIDE_MODE_2		= 0x00001000,
	SE_SECURITY_CONTEXT_OVERRIDE_MODE_3		= 0x00002000,
} SECURITY_CONTEXT_OP_MODE_GENERAL;

/* Security Context Opertion Mode - Encryption */
typedef enum SECURITY_CONTEXT_OP_MODE_ENCRYPTION_ {
	/* Encryption */
	SE_SECURITY_CONTEXT_EMODE_FLOWTHROUGH		= 0x00000001,
	SE_SECURITY_CONTEXT_EMODE_DES_ECB		= 0x00000002,
	SE_SECURITY_CONTEXT_EMODE_DES_CBC		= 0x00000004,
	SE_SECURITY_CONTEXT_EMODE_TDES_ECB		= 0x00000008,
	SE_SECURITY_CONTEXT_EMODE_TDES_CBC		= 0x00000010,
	SE_SECURITY_CONTEXT_EMODE_AES_128_ECB		= 0x00000020,
	SE_SECURITY_CONTEXT_EMODE_AES_128_CBC		= 0x00000040,
	SE_SECURITY_CONTEXT_EMODE_AES_192_ECB		= 0x00000080,
	SE_SECURITY_CONTEXT_EMODE_AES_192_CBC		= 0x00000100,

	SE_SECURITY_CONTEXT_UPDATE_IV			= 0x00000800,
	SE_SECURITY_CONTEXT_IV_PROCESSING		= 0x00001000,

	/* Pad */
	SE_SECURITY_CONTEXT_PALG_NO_PADDING		= 0x00002000,
	SE_SECURITY_CONTEXT_PALG_CONSTANT		= 0x00004000,
	SE_SECURITY_CONTEXT_PALG_OPTIONAL_LENGTH	= 0x00008000,
	SE_SECURITY_CONTEXT_PALG_LENGTH			= 0x00010000,
	SE_SECURITY_CONTEXT_PALG_LENGTH_1		= 0x00020000,
	SE_SECURITY_CONTEXT_PALG_COUNT			= 0x00040000,
	SE_SECURITY_CONTEXT_PALG_OPTIONAL_COUNNT_LEN	= 0x00080000,
	SE_SECURITY_CONTEXT_PALG_IPSEC			= 0x00100000,

	SE_SECURITY_CONTEXT_REMOVE_PADDING		= 0x00200000,

	SE_SECURITY_CONTEXT_PAD_BLOCK_SIZE_4		= 0x00400000,
	SE_SECURITY_CONTEXT_PAD_BLOCK_SIZE_8		= 0x00800000,
	SE_SECURITY_CONTEXT_PAD_BLOCK_SIZE_16		= 0x01000000,
} SECURITY_CONTEXT_OP_MODE_ENCRYPTION;

/* Security Context Opertion Mode - Hash */
typedef enum SECURITY_CONTEXT_OP_MODE_HASH_ {
    /* Hash */
	SE_SECURITY_CONTEXT_HMODE_DISABLED		= 0x00000001,
	SE_SECURITY_CONTEXT_HMODE_MD5			= 0x00000002,
	SE_SECURITY_CONTEXT_HMODE_SHA_1			= 0x00000004,
	SE_SECURITY_CONTEXT_HMODE_HMAC_MD5		= 0x00000010,
	SE_SECURITY_CONTEXT_HMODE_HMAC_SHA_1		= 0x00000020,
	SE_SECURITY_CONTEXT_HTM_MODE_0			= 0x00000400,
	SE_SECURITY_CONTEXT_HTM_MODE_1			= 0x00000800,
	SE_SECURITY_CONTEXT_HTM_MODE_2			= 0x00001000,
	SE_SECURITY_CONTEXT_HTM_MODE_3			= 0x00002000,

	SE_SECURITY_CONTEXT_CHT_DISABLE_CHECK		= 0x00004000,
	SE_SECURITY_CONTEXT_CHT_ENABLE_CHECK		= 0x00008000,

	SE_SECURITY_CONTEXT_AH_DISALBE_CHECK		= 0x00010000,
	SE_SECURITY_CONTEXT_AH_ENABLE_CHECK		= 0x00020000,

	SE_SECURITY_CONTEXT_MFP_ENABLE			= 0x00040000,
} SECURITY_CONTEXT_OP_MODE_HASH;


/* Security Context Opertion Mode/Option */
typedef struct SECURITY_CONTEXT_OPERATION_MODE_ {
	SECURITY_CONTEXT_OP_MODE_GENERAL securityContextOpModeGeneral;
	SECURITY_CONTEXT_OP_MODE_ENCRYPTION securityContextOpModeEncryption;
	SECURITY_CONTEXT_OP_MODE_HASH securityContextOpModeHash;
} SECURITY_CONTEXT_OPERATION_MODE;

/* This is for context load command */
typedef enum LOAD_PARAM_MODE {
	LOAD_USE_PARAM_LIST	= 1,		/* Parameter is specified in the list */
	LOAD_USE_INPUT_BUFFER	= 2,		/* Parameter is in input buffer */
} LOAD_PARAM_MODE;

/* SE Security Context Command Load Parameter */
typedef struct SE_CONTEXT_COMMAND_LOAD_PARAM_ {
	LOAD_PARAM_MODE paramMode;
	/* One way to specify */
	SECURITY_CONTEXT_OPERATION_MODE contextMode;
	SECURITY_CONTEXT_OPERATION_PARAM *pContextParam;
	/*or another */
	SECURITY_RAM_CONTEXT *pInBuffer;
} SE_CONTEXT_COMMAND_LOAD_PARAM;

/* SE Security Context Command Store Parameter */
typedef struct SE_CONTEXT_COMMAND_STORE_PARAM_ {
	SECURITY_RAM_CONTEXT *pOutBuffer;
} SE_CONTEXT_COMMAND_STORE_PARAM;

/* This is for context replace command  */
typedef enum REPLACE_PARAM_MODE {
	REPLACE_USE_PARAM_LIST		= 1,	/* Paramter is specified in list */
	REPLACE_USE_IO_BUFFER		= 2,	/* Parameter is in input buffer, and buffer is used for output */
} REPLACE_PARAM_MODE;

/* SE Security Context Command Replace Parameter */
typedef struct SE_CONTEXT_COMMAND_REPLACE_PARAM_ { 
	REPLACE_PARAM_MODE paramMode;
	/* One way to specify the security context */
	SECURITY_CONTEXT_OPERATION_MODE contextMode;
	SECURITY_CONTEXT_OPERATION_PARAM *pContextParam;
	/* or another way */
	SECURITY_RAM_CONTEXT *pInBuffer;
	/* either way requires an output buffer to store returned context */
	SECURITY_RAM_CONTEXT *pOutBuffer;
} SE_CONTEXT_COMMAND_REPLACE_PARAM;

/* SE Security Context Command Load Segment Parameter */
typedef struct SE_CONTEXT_COMMAND_LOAD_SEGMENT_PARAM_ {
	uint16 eOffset;
	uint16 eLength;
	uint16 hOffset;
	uint16 hLength;
} SE_CONTEXT_COMMAND_LOAD_SEGMENT_PARAM;

/* SE Security Context Command Load IV Parameter */
typedef struct SE_CONTEXT_COMMAND_LOAD_IV_PARAM_ {
	byte iv[MAX_ENC_IV_SIZE];	/* unaligned string */
	int ivSize;	
} SE_CONTEXT_COMMAND_LOAD_IV_PARAM;

/* SE Security Context Command Store IV Parameter */
typedef struct SE_CONTEXT_COMMAND_STORE_IV_PARAM_ {
	uint32 iv[4];
} SE_CONTEXT_COMMAND_STORE_IV_PARAM;

/* SE Security Context Command Load PB Parameter */
typedef struct SE_CONTEXT_COMMAND_LOAD_PB_PARAM_ {
	byte padByte;
} SE_CONTEXT_COMMAND_LOAD_PB_PARAM;

/* SE Security Context Command Init Hmac Md5 Parameter */
typedef struct SE_CONTEXT_COMMAND_INIT_HMAC_MD5_PARAM_ {
	byte key[MD5HMAC96_KEYSIZE];               
} SE_CONTEXT_COMMAND_INIT_HMAC_MD5_PARAM;

/* SE Security Context Command Init Hmac Sha1 Parameter */
typedef struct SE_CONTEXT_COMMAND_INIT_HMAC_SHA1_PARAM_ {
	byte key[SHA1HMAC96_KEYSIZE];               
} SE_CONTEXT_COMMAND_INIT_HMAC_SHA1_PARAM;

/* Security Engine Security Context Command Parameter */
typedef struct SE_SECURITY_CONTEXT_COMMAND_PARAM_ {
	int operation;		/* encode or decode */
	union context_ {
		SE_CONTEXT_COMMAND_LOAD_PARAM  seContextCommandLoadParam;
		SE_CONTEXT_COMMAND_STORE_PARAM  seContextCommandStoreParam;
		SE_CONTEXT_COMMAND_REPLACE_PARAM  seContextCommandReplaceParam;
	} context;
	SE_CONTEXT_COMMAND_LOAD_SEGMENT_PARAM  seContextCommandLoadSegmentParam;
	union iv_ {
		SE_CONTEXT_COMMAND_LOAD_IV_PARAM  seContextCommandLoadIVparam;
		SE_CONTEXT_COMMAND_STORE_IV_PARAM  seContextCommandStoreIVParam;
	} iv;
	SE_CONTEXT_COMMAND_LOAD_PB_PARAM seContextCommandLoadPBParam;
	union key_ {
		SE_CONTEXT_COMMAND_INIT_HMAC_MD5_PARAM  seContextCommandInitHmacMd5Param;
		SE_CONTEXT_COMMAND_INIT_HMAC_SHA1_PARAM  seContextCommandInitHmacSha1Param;
	} key;
} SE_SECURITY_CONTEXT_COMMAND_PARAM;

typedef enum CRYPO_UNIT_TYPE_ {
	SE	= 1,	
	RNG	= 2,
} CRYPTO_UNIT_TYPE;

/* SE init parameter */
typedef struct SECURITY_ENGINE_UNIT_PARAM_ {
	/* none defined */
} SECURITY_ENGINE_UNIT_PARAM;

/* RNG init parameter */
typedef struct RNG_UNIT_PARAM_ {
	/* none defined */
} RNG_UNIT_PARAM;

/* RNG Command Type */
typedef enum RNG_COMMAND_TYPE_ {
	RNG_RANDOM_RAW		= 1,
} RNG_COMMAND_TYPE;

/* RNG RANDOM GEN command parameter */
typedef struct RNG_RANDOM_GEN_PARAM_ {
	int random_count;		/* number of 32bit random number requested */
	uint32 *random_buffer;		/* output buffer */
} RNG_RANDOM_GEN_PARAM;

/* RNG Command Parameter */
typedef union RNG_COMMAND_PARAM_ {
	RNG_RANDOM_GEN_PARAM rngRandomGenParam;
} RNG_COMMAND_PARAM;

/*
 * Generic API Function Prototypes
 */
int
idt_Init (
	CRYPTO_UNIT_TYPE  cryptoUnit,
	SECURITY_ENGINE_UNIT_PARAM *pSecurityEngineUnitParam,
	RNG_UNIT_PARAM *pRngUnitParam,
	void  (*pCallback)(int callHandle,
		int result),
	int callHandle);


int
idt_Security_Context_RAM_API (
	SECURITY_CONTEXT_ID contextID,
	SE_SECURITY_CONTEXT_COMMAND securityContextCommand,
	SE_SECURITY_CONTEXT_COMMAND_PARAM *pSecurityContextCommandParam,
	void  (*pCmdCallback)(SECURITY_CONTEXT_ID contextID,
				void *cb,
				int callHandle,
				int result),
	void *cb,
	int callHandle);

int
idt_Packet_API (
	SECURITY_CONTEXT_ID contextID,
	PACKET_SECURITY_CONTEXT *pContext);

int
idt_RNG_API (
	RNG_COMMAND_TYPE  randomCommand,
	RNG_COMMAND_PARAM *randomParam,
	void  (*pCallback)(int callHandle,
				int result),
	int callHandle);

#endif /* _IDT_CRYPTO_API_H_ */
