/*
 * idt_crypto_queue.h
 * 
 * crypto FIFO queue primitives.
 */ 
#ifndef _IDT_CRYPTO_QUEUE_H_
#define	_IDT_CRYPTO_QUEUE_H_

#include <asm/system.h>
#include <linux/spinlock.h>

typedef struct QTYPE_ {
	struct QTYPE_ *prev;
	struct QTYPE_ *next;
} QTYPE;

typedef struct QHEAD_ {
	struct QTYPE_ *prev;
	struct QTYPE_ *next;
	unsigned int qSize;
	spinlock_t lock;
} QHEAD;

static inline void idt_en_queue(QHEAD *,QTYPE *);
static inline QTYPE *idt_de_queue(QHEAD *);
static inline void idt_re_queue(QHEAD *,QTYPE *);
static inline int idt_queue_size(QHEAD *q);
static inline void idt_queue_init(QHEAD *q);

/*
 * Add entry to queue
 */
static inline void idt_en_queue(QHEAD *q, QTYPE *entry)
{
	unsigned long flags;

	local_irq_save(flags);
	q->qSize++;
	entry->prev = (QTYPE *)q;
	entry->next = q->next;
	q->next->prev = entry;
	q->next = entry;
	local_irq_restore(flags);
	return;
}

/*
 * Remove entry from queue
 */
static inline QTYPE *idt_de_queue(QHEAD *q)
{
	unsigned long flags;
	QTYPE *entry;

	if(q->prev == (QTYPE *)q)
		return NULL;

	local_irq_save(flags);
	q->qSize--;
	entry = q->prev;
	q->prev = entry->prev;
	entry->prev->next = entry->next;
	local_irq_restore(flags);
	entry->next = entry->prev = NULL;
	return entry;
}

/*
 * Re-add the entry to queue
 */
static inline void idt_re_queue(QHEAD *q, QTYPE *entry)
{
	idt_en_queue((QHEAD *)q->prev,entry);
	return;
}


static inline int idt_queue_size(QHEAD *q)
{
	return(q->qSize);
}

static inline void idt_queue_init(QHEAD *q)
{
	spin_lock_init(&q->lock);
	q->prev = (QTYPE *)q;
	q->next = (QTYPE *)q;
	q->qSize = 0;
}

#endif



