/*******************************************************************************
 *
 * idt_rng_interface.h
 *
 * rc32365 RNG device interface header file 
 *
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef _IDT_RNG_INTERFACE_H_
#define _IDT_RNG_INTERFACE_H_

#include <linux/idt_crypto_api.h>
#include <linux/idt_crypto_queue.h>
#include <linux/idt_crypto_if_common.h>


typedef struct idt_RNG_DEVICE_ {
	char *name;
	int (*open)(void *, void *);	/* open device */
	int (*close)(void *, void *);	/* close device */
	int (*reset)(void);		/* reset device */
	int state; 			/* device state */
	int rnFlag;			/* 1:rnbuf full 0:rnbuf empty */
	char *rnBuf;			/* random number buffer (2500 bytes) */
	int (*start) (void);		/* resume device */
} idt_RNG_DEVICE;

#define RNG_POOL_BYTE_SIZE		2500	/* byte unit */
#define RNG_POOL_SIZE			625	/* 32byte unit */

/*
 *  RNG device control block
 */
typedef struct RNG_DCB_	{
	idt_DEVICE_STATE state;
	idt_RNG_DEVICE *dev;	
	ulong pool[RNG_POOL_SIZE];
	ushort pool_head;
	ushort pool_count;
} RNG_DCB; 

/*
 * RNG API control block
 */
typedef struct RNG_API_CB_ {
	idt_DEVICE_STATE state;
	RNG_DCB *dcb;
	int call_count;
} RNG_API_CB;

extern int register_rng_device(idt_RNG_DEVICE *dev);
extern int unregister_rng_device(idt_RNG_DEVICE *dev);

#endif /*_IDT_RNG_INTERFACE_H_*/
