/*******************************************************************************
 *
 * idt_se_interface.h
 *
 * rc32365 Security Engine device interface header file 
 *
 * Author: calvin.young@idt.com
 * Date: 06/30/2003
 * Update:
 *
 ******************************************************************************/
#ifndef _IDT_SE_IF_H_
#define	_IDT_SE_IF_H_

#include <linux/idt_crypto_api.h>
#include <linux/idt_crypto_queue.h>
#include <linux/idt_crypto_if_common.h>

#ifndef TRUE
#define TRUE 1
#endif
/*
 * Context RAM specific
 */
/* General */	
#define CONTEXT_GEN_OP_DISABLE			0x0
#define CONTEXT_GEN_OP_ENCODE			0x1
#define CONTEXT_GEN_OP_DECODE			0x2
#define CONTEXT_GEN_OP_RESERVED			0x3

#define CONTEXT_GEN_MODE_HASH_ONLY		0x0
#define CONTEXT_GEN_MODE_ENCRYPT_ONLY		0x1
#define CONTEXT_GEN_MODE_ENCRYPT_BF_HASH	0x2
#define CONTEXT_GEN_MODE_HASH_BF_ENCRYPT	0x3

#define CONTEXT_GEN_PROT_IPV4			0x0
#define CONTEXT_GEN_PROT_IPV6			0x1

#define CONTEXT_GEN_DSD				0x1
	
#define CONTEXT_GEN_OM_0			0x0
#define CONTEXT_GEN_OM_1			0x1
#define CONTEXT_GEN_OM_2			0x2
#define CONTEXT_GEN_OM_3			0x3

/* Encryption */
#define CONTEXT_EMODE_FLOW_THROUGH		0x0
#define CONTEXT_EMODE_DES_ECB			0x2
#define CONTEXT_EMODE_DES_CBC			0x3
#define CONTEXT_EMODE_TDES_ECB			0x4
#define CONTEXT_EMODE_TDES_CBC			0x5
#define CONTEXT_EMODE_AES_128_ECB		0x6
#define CONTEXT_EMODE_AES_128_CBC		0x7
#define CONTEXT_EMODE_AES_192_ECB		0x8
#define CONTEXT_EMODE_AES_192_CBC		0x9

/* Padding */
#define	CONTEXT_PAD_PALG_NO_PADDING		0x0          
#define	CONTEXT_PAD_PALG_CONSTANT		0x1
#define	CONTEXT_PAD_PALG_OPTIONAL_LENGTH	0x2 
#define	CONTEXT_PAD_PALG_LENGTH			0x3 
#define	CONTEXT_PAD_PALG_LENGTH_1		0x4 
#define	CONTEXT_PAD_PALG_COUNT			0x5 
#define	CONTEXT_PAD_PALG_OPTIONAL_COUNNT_LEN	0x6 
#define CONTEXT_PAD_PALG_IPSEC			0x7 

#define	CONTEXT_PAD_BLOCK_SIZE_4		0x0
#define CONTEXT_PAD_BLOCK_SIZE_8		0x1 
#define	CONTEXT_PAD_BLOCK_SIZE_16		0x2 

/* Hash */
#define	CONTEXT_HASH_HMODE_DISABLED		0x0
#define CONTEXT_HASH_HMODE_MD5			0x1 
#define	CONTEXT_HASH_HMODE_SHA_1		0x2 
#define CONTEXT_HASH_HMODE_HMAC_MD5		0x5 
#define CONTEXT_HASH_HMODE_HMAC_SHA_1		0x6 

#define	CONTEXT_HASH_HTM_MODE_0			0x0 
#define CONTEXT_HASH_HTM_MODE_1			0x1 
#define	CONTEXT_HASH_HTM_MODE_2			0x2 
#define CONTEXT_HASH_HTM_MODE_3			0x3 

#define CONTEXT_HASH_CHECK_DISABLED		0x0
#define CONTEXT_HASH_CHECK_CHT			0x1
#define CONTEXT_HASH_CHECK_CAH			0x2
#define CONTEXT_HASH_CHECK_RESERVED		0x3

/*
 * Security engine crypto request type
 */
typedef enum {
	SE_COMMAND	= 0x1,	/* SE context RAM command	*/
	SE_DATA_STREAM	= 0x2,	/* SE data stream processing	*/
} CRYPTO_REQUEST_TYPE;

/*
 * Security engine crypto request flags
 */
typedef enum {
	CONTIGUOUS_BUF	= 0x01,	/* contiguous buffer		*/
	CHAINED_BUF	= 0x02,	/* chained buffer		*/
	IO_SAME_BUF	= 0x04,	/* IO use the same buffer	*/
	SYNC_CALL	= 0x08,	/* Synchronous call		*/
	CALLBACK_REG	= 0x10,	/* Register packet callback	*/
} REQ_FLAGS;

typedef struct COMMAND_SECURITY_CONTEXT_ {
	SECURITY_RAM_CONTEXT *context;	/* se security ram context	*/
	union {
		struct {
			uint16 hOffset;
			uint16 hLength;
			uint16 eOffset;
			uint16 eLength;
		} segmentField;
		uint32 iv[4];		/* IV		*/             
		uint32 hKey[8];		/* hash key	*/	               
		byte padByte;		/* pad byte	*/
	} param;
} COMMAND_SECURITY_CONTEXT;

/*
 * Crypto request data structure
 */
typedef struct CRYPTO_REQ_ {
	QTYPE link;			/* linkaage */
	SECURITY_CONTEXT_ID contextID;	/* security context ID */
	CRYPTO_REQUEST_TYPE reqType;	/* se request type, command or packet */
	SE_SECURITY_CONTEXT_COMMAND cmd;/* se context ram command type */
	void *cb;			/* caller's opaque control data */
	void  (*pCallback)(		/* callback routine */
		SECURITY_CONTEXT_ID contextID,
		void *cb,		/* caller's opaque control block */
		void *newbuf,		/* crypto driver returned new buffer */
		int callHandle,
		int result);
	int callHandle;			/* command opaque parameter */
	/* Per-packet processing specific */
	int op;				/* operation */
	byte *inBuf;			/* Input buffer */
	byte *outBuf;			/* Output buffer */
	int inLen;			/* Length of input buffer */
	int outLen;			/* On input, max size. On Output, real size of output */
	COMMAND_SECURITY_CONTEXT *commandContext;
	PACKET_SECURITY_CONTEXT *packetContext;	/* per-packet processing input parameters and output hashtrailer */
	void *newbuf;			/* data buffer allocated/used by driver */
	int result;			/* result */
} CRYPTO_REQ, *pCRYPTO_REQ;

/*
 * Hardware Crypto Device Registration 
 */
typedef struct idt_CRYPTO_DEVICE_ {
	char *name;
	int (*open)(SECURITY_ENGINE_UNIT_PARAM *param);
	int (*close)(void);
	int (*process)(CRYPTO_REQ *req);	
	int state;
} idt_CRYPTO_DEVICE;

extern int register_crypto_device(idt_CRYPTO_DEVICE *dev);
extern int unregister_crypto_device(idt_CRYPTO_DEVICE *dev);

/*
 * Security engine device control block
 */
typedef struct SE_DCB_ {
	idt_DEVICE_STATE state;		/* device state */
	QHEAD crypto_req_q;		/* crypto request queue */
	int se_in_use;			/* in use count	*/
	idt_CRYPTO_DEVICE *dev;
} SE_DCB;

/*
 * Security engine API control block
 */
typedef struct SE_API_CB_ {
	idt_DEVICE_STATE state;		/* API state */
	SE_DCB *dcb;			/* security engine control block */
	kmem_cache_t *crop_cache;	/* crypto operation memory allocated */
	int call_count;			/* call count */
} SE_API_CB;


static inline void crypto_stop_queue(idt_CRYPTO_DEVICE *dev) {
	set_bit(0, &dev->state);
}

static inline void crypto_wake_queue(idt_CRYPTO_DEVICE *dev) {
	clear_bit(0, &dev->state);
}

static inline int crypto_queue_stopped(idt_CRYPTO_DEVICE *dev) {
	return test_bit(0, &dev->state);
}

extern inline int idt_run_crypto_request(void);
extern inline void se_free_crypto_request(CRYPTO_REQ *);
extern inline int idt_requeue_crypto_request(CRYPTO_REQ *req);


/* #define BYPASS_REQ_QUEUEING	1	*/	/* test */
#define REQ_QUEUE_MAX_SIZE	100

#endif /*_IDT_SE_IF_H_*/
