/*
 * (C) Copyright 2001
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <watchdog.h>
#include <mpc8xx.h>

/* AMD AM29LV320DT, 4 MB, sect size 0x10000 */

#if defined(CFG_ENV_IS_IN_FLASH)
# ifndef  CFG_ENV_ADDR
#  define CFG_ENV_ADDR	(CFG_FLASH_BASE + CFG_ENV_OFFSET)
# endif
# ifndef  CFG_ENV_SIZE
#  define CFG_ENV_SIZE	CFG_ENV_SECT_SIZE
# endif
# ifndef  CFG_ENV_SECT_SIZE
#  define CFG_ENV_SECT_SIZE  CFG_ENV_SIZE
# endif
#endif

/*---------------------------------------------------------------------*/
#undef DEBUG_FLASH
//#define DEBUG_FLASH

#ifdef DEBUG_FLASH
#define DEBUGF(fmt,args...) printf(fmt ,##args)
#else
#define DEBUGF(fmt,args...)
#endif
/*---------------------------------------------------------------------*/

flash_info_t	flash_info[CFG_MAX_FLASH_BANKS]; /* info for FLASH chips	*/

/*-----------------------------------------------------------------------
 * Functions
 */
static ulong flash_get_size (vu_long *addr, flash_info_t *info);
static int write_data (flash_info_t *info, ulong dest, unsigned char data);
static void flash_get_offsets (ulong base, flash_info_t *info);

/*-----------------------------------------------------------------------
 */

unsigned long flash_init (void)
{
    unsigned long size = 0;
    int i;

    /* Init: no FLASHes known */
    for (i=0; i<CFG_MAX_FLASH_BANKS; ++i) {
        flash_info[i].flash_id = FLASH_UNKNOWN;
    }

    size = flash_get_size((vu_long *)CFG_FLASH_BASE, &flash_info[0]);
    if (size) {
        flash_get_offsets (CFG_FLASH_BASE, &flash_info[0]);
        flash_info[0].size = size;
    }

    return (size);
}

/*-----------------------------------------------------------------------
 */
static void flash_get_offsets (ulong base, flash_info_t *info)
{
    int i;

    if (info->flash_id == FLASH_UNKNOWN) {
        return;
    }

    switch (info->flash_id & FLASH_VENDMASK) {
    case FLASH_MAN_INTEL:
    case FLASH_MAN_STM:
        for (i = 0; i < info->sector_count; i++) {
	    info->start[i] = base;
	    base += 0x00020000;		    /* 128k * 1 chip per bank */
	}
	return;
    case FLASH_MAN_AMD:
        for (i = 0; i < info->sector_count; i++) {
	    info->start[i] = base;
            if (i < 63) {
                base += 0x00010000;
            } else {
                base += 0x00002000;
            }
	}
	return;

    default:
        printf ("Don't know sector ofsets for flash type 0x%lx\n",
		info->flash_id);          
	return;
    }
}

/*-----------------------------------------------------------------------
 */
void flash_print_info  (flash_info_t *info)
{
    int i;

    if (info->flash_id == FLASH_UNKNOWN) {
        printf ("missing or unknown FLASH type\n");
	return;
    }

    switch (info->flash_id & FLASH_VENDMASK) {
    case FLASH_MAN_AMD:	  printf ("AMD ");		break;
    case FLASH_MAN_FUJ:	  printf ("Fujitsu ");		break;
    case FLASH_MAN_SST:	  printf ("SST ");		break;
    case FLASH_MAN_STM:	  printf ("STM ");		break;
    case FLASH_MAN_INTEL: printf ("Intel ");		break;
    case FLASH_MAN_MT:	  printf ("MT ");		break;
    case FLASH_MAN_ATMEL: printf ("Atmel ");		break;
    default:		  printf ("Unknown Vendor ");	break;
    }

    switch (info->flash_id & FLASH_TYPEMASK) {
    case FLASH_28F320J3A:
        printf ("28F320J3A (32Mbit = 128K x 32)\n");
	break;
    case FLASH_28F640J3A:
        printf ("28F640J3A (64Mbit = 128K x 64)\n");
	break;
    case FLASH_28F128J3A:
        printf ("28F128J3A (128Mbit = 128K x 128)\n");
	break;
    case FLASH_AM320T:
        printf ("29LV320T (32Mbit = 128K x 32)\n");
        break;
    default:
        printf ("Unknown Chip Type\n");
	break;
    }

    if (info->size >= (1 << 20)) {
        i = 20;
    } else {
        i = 10;
    }
    printf ("  Size: %ld %cB in %d Sectors\n",
	    info->size >> i,
	    (i == 20) ? 'M' : 'k',
	    info->sector_count);

    printf ("  Sector Start Addresses:");
    for (i=0; i<info->sector_count; ++i) {
        if ((i % 5) == 0)
	    printf ("\n   ");
	printf (" %08lX%s",
		info->start[i],
		info->protect[i] ? " (RO)" : "     "
		);
    }
    printf ("\n");
    return;
}

/*-----------------------------------------------------------------------
 */


/*-----------------------------------------------------------------------
 */

/*            
 * The following code cannot be run from FLASH!
 */

static ulong flash_get_size (vu_long *addr, flash_info_t *info)
{
    volatile ulong value;
    
    
    /* Read Manufacturer ID */

    volatile unsigned char* cAddr  = (unsigned char*)addr;
    volatile unsigned char  cValue = 0;

    cAddr[0x0AAA] = 0xAA;
    cAddr[0x0555] = 0x55;
    cAddr[0x0AAA] = 0x90;
    cValue = cAddr[0];
    value = (ulong)cValue;
    DEBUGF("Manuf. ID @ 0x%p: 0x%x\n", cAddr, cValue);

    switch (value|value<<16) {
    case AMD_MANUFACT:
        info->flash_id = FLASH_MAN_AMD;
	break;
    case FUJ_MANUFACT:
        info->flash_id = FLASH_MAN_FUJ;
	break;
    case SST_MANUFACT:
        info->flash_id = FLASH_MAN_SST;
	break;
    case STM_MANUFACT:
        info->flash_id = FLASH_MAN_STM;
	break;
    case INTEL_MANUFACT:
        info->flash_id = FLASH_MAN_INTEL;
        break;
    default:
        info->flash_id = FLASH_UNKNOWN;
	info->sector_count = 0;
	info->size = 0;
	*cAddr = 0xFF; 
	return (0);			/* no or unknown flash	*/
    }

    cValue = cAddr[2];
    value = (ulong)cValue;
    DEBUGF("\nDevice ID @ %p: 0x%x\n", cAddr, cValue);

#if 1
    value = AMD_ID_LV320T & 0xffff;
#endif

    switch (value|value<<16) {
    case INTEL_ID_28F320J3A:
        info->flash_id += FLASH_28F320J3A;
	info->sector_count = 32;
	info->size = 0x00400000;
	break;				/* =>  4 MB		*/

    case INTEL_ID_28F640J3A:
        info->flash_id += FLASH_28F640J3A;
        info->sector_count = 64;
        info->size = 0x00800000;
	break;				/* => 8 MB		*/

    case INTEL_ID_28F128J3A:
        info->flash_id += FLASH_28F128J3A;
	info->sector_count = 128;
	info->size = 0x01000000 * 2;
	break;				/* => 32 MB		*/

    case AMD_ID_LV320T:
        info->flash_id += FLASH_AM320T;
	info->sector_count = 71;
	info->size = 0x00400000;
        break;

    default:
        info->flash_id = FLASH_UNKNOWN;
	*cAddr = 0xFF; 
	return (0);			/* => no or unknown flash */
    }

    if (info->sector_count > CFG_MAX_FLASH_SECT) {
        printf ("** ERROR: sector count %d > max (%d) **\n",
		info->sector_count, CFG_MAX_FLASH_SECT);
	info->sector_count = CFG_MAX_FLASH_SECT;
    }

    *cAddr = 0xFF; 

    return (info->size);
}


/*-----------------------------------------------------------------------
 */

int flash_erase (flash_info_t *info, int s_first, int s_last)
{
	vu_char *addr = (vu_char*)(info->start[0]);
	int flag, prot, sect, l_sect;
	ulong start, now, last;

	if ((s_first < 0) || (s_first > s_last)) {
		if (info->flash_id == FLASH_UNKNOWN) {
			printf ("- missing\n");
		} else {
			printf ("- no sectors to erase\n");
		}
		return 1;
	}

	if ((info->flash_id == FLASH_UNKNOWN) ||
	    (info->flash_id > FLASH_AMD_COMP)) {
		printf ("Can't erase unknown flash type %08lx - aborted\n",
			info->flash_id);
		return 1;
	}

	prot = 0;
	for (sect=s_first; sect<=s_last; ++sect) {
		if (info->protect[sect]) {
			prot++;
		}
	}

	if (prot) {
		printf ("- Warning: %d protected sectors will not be erased!\n",
			prot);
	} else {
		printf ("\n");
	}

	l_sect = -1;

	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	addr[0x0AAA] = 0xAA;
	addr[0x0555] = 0x55;
	addr[0x0AAA] = 0x80;
	addr[0x0AAA] = 0xAA;
	addr[0x0555] = 0x55;

	/* Start erase on unprotected sectors */
	for (sect = s_first; sect<=s_last; sect++) {
		if (info->protect[sect] == 0) {	/* not protected */
			addr = (vu_char*)(info->start[sect]);
			addr[0] = 0x30;
			l_sect = sect;
		}
	}

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* wait at least 80us - let's wait 1 ms */
	udelay (1000);

	/*
	 * We wait for the last triggered sector
	 */
	if (l_sect < 0)
		goto DONE;

	start = get_timer (0);
	last  = start;
	addr = (vu_char*)(info->start[l_sect]);
	while ((addr[0] & 0x80) != 0x80) {
                WATCHDOG_RESET();
		if ((now = get_timer(start)) > CFG_FLASH_ERASE_TOUT) {
			printf ("Timeout\n");
			return 1;
		}
		/* show that we're waiting */
		if ((now - last) > 1000) {	/* every second */
			putc ('.');
			last = now;
		}
	}

DONE:
	/* reset to read mode */
	addr = (vu_char *)info->start[0];
	addr[0] = 0xF0;	/* reset bank */

	printf (" done\n");
	return 0;
}

/*-----------------------------------------------------------------------
 * Copy memory to flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */

int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt)
{
	int rc;

	while (cnt > 0) {
		if ((rc = write_byte(info, addr++, *src++)) != 0) {
			return (rc);
		}
		--cnt;
	}

	return (0);
}

/*-----------------------------------------------------------------------
 * Write a word to Flash, returns:
 * 0 - OK
 * 1 - write timeout
 * 2 - Flash not erased
 */
static int write_byte (flash_info_t *info, ulong dest, uchar data)
{
	vu_char *addr = (vu_char*)(info->start[0]);
	ulong start;
	int flag;

	/* Check if Flash is (sufficiently) erased */
	if ((*((vu_char *)dest) & data) != data) {
		return (ERR_NOT_ERASED);
	}
	/* Disable interrupts which might cause a timeout here */
	flag = disable_interrupts();

	addr[0x0AAA] = 0xAA;
	addr[0x0555] = 0x55;
	addr[0x0AAA] = 0xA0;

	*((vu_char *)dest) = data;

	/* re-enable interrupts if necessary */
	if (flag)
		enable_interrupts();

	/* data polling for D7 */
	start = get_timer (0);
	while ((*((vu_char *)dest) & 0x80) != (data & 0x80)) {
                WATCHDOG_RESET();
		if (get_timer(start) > CFG_FLASH_WRITE_TOUT) {
			return (ERR_TIMOUT);
		}
	}
	return (ERR_OK);
}

/*-----------------------------------------------------------------------
 */
