#include <ppcboot.h>
#include <command.h>
#include <net.h>
#include <watchdog.h>
#include "cmd_cabernet.h"
#include "cmd_elf.h"
#include "cmd_boot.h"
#include "ns16550.h"
#include "../disk/part_dos.h"
#include "flash.h"

#include "../../../cmn/util/cab_cpld.h"
#include "../../../cmn/util/cab_fpga.h"
#include "../../../cmn/util/robo.h"
#include "../../../cmn/util/aruba_image.h"

#define TIMEOUT	       2		/* Seconds before failing */
#define PKTSIZEMIN     64
#define PHY_REGS       32
#define PHY_DEVS       64
#define PHY_PREAMBLE   32

#define CHAR(c)       (((c >= 0x20) && (c <= 0x7e))? (char)c: '.')

extern struct NS16550 *COM_PORTS[];
static u_long loopback_length;
static u_long loopback_count;
static u_long loopback_vlan;
static u_long loopback_incr;
static u_long loopback_received;
static int loopback_rcv;
void LoopbackRequest(void);
static unsigned long adjustSectors(unsigned long lba512);

typedef struct {
  u16 pvid;
  u8  mac;
  u8  phy;
} port_map_t;

port_map_t port_map[ROBO_PORTS] = {
    { 0x08, 0x07, 0x07 }, { 0x07, 0x06, 0x06 },
    { 0x06, 0x05, 0x05 }, { 0x05, 0x04, 0x04 },
    { 0x04, 0x03, 0x03 }, { 0x03, 0x02, 0x02 },
    { 0x02, 0x01, 0x01 }, { 0x01, 0x00, 0x00 },
    { 0x0c, 0x08, 0x08 }, { 0x0b, 0x09, 0x09 },
    { 0x09, 0x0b, 0x0b }, { 0x0a, 0x0c, 0x0c },
};

typedef struct {
  __u8 addr;
  __u8 length;
  char *str;
} robo_sstat_t;

robo_sstat_t robo_stats[] = {
  {0x00, 8, "TxOctets"},
  {0x08, 4, "TxDropPkts"},
  {0x0c, 4, "TxQoSQPkt"},
  {0x10, 4, "TxBroadcastPkts"},
  {0x14, 4, "TxMulticastPkts"},
  {0x18, 4, "TxUnicastPkts"},
  {0x1c, 4, "TxCollisions"},
  {0x20, 4, "TxSingleCollisions"},
  {0x24, 4, "TxMultipleCollisio"},
  {0x28, 4, "TxDeferredTransmit"},
  {0x2c, 4, "TxLateCollisions"},
  {0x30, 4, "TxExcessiveCollisi"},
  {0x34, 4, "TxFrameInDisc?"},
  {0x38, 4, "TxPausePkts"},
  {0x3c, 8, "TxQosQOctets"},
  {0x44, 8, "RxOctets"},
  {0x4c, 4, "RxUndersizePkts"},
  {0x50, 4, "RxPausePkts"},
  {0x54, 4, "RxPkts64Octets"},
  {0x58, 4, "RxPkts65to128Octet"},
  {0x5c, 4, "RxPkts128to255Octe"},
  {0x60, 4, "RxPkts256to511Octe"},
  {0x64, 4, "RxPkts512to1023Oct"},
  {0x68, 4, "RxPkts1024to1522Oc"},
  {0x6c, 4, "RxOversizePkts"},
  {0x70, 4, "RxJabbers"},
  {0x74, 4, "RxAlignmentErrors"},
  {0x78, 4, "RxFCSErrors"},
  {0x7c, 8, "RxGoodOctets"},
  {0x84, 4, "RxDropPkts"},
  {0x88, 4, "RxUnicastPkts"},
  {0x8c, 4, "RxMulticastPkts"},
  {0x90, 4, "RxBroadcastPkts"},
  {0x94, 4, "RxSAChanges"},
  {0x98, 4, "RxFragments"},
  {0x9c, 4, "RxExcessSizeDiscar"},
  {0xa0, 4, "RxSymbolErrors"},
  {0xa4, 4, "RxQoSQPkts"},
  {0xa8, 8, "RxQoSQOctets"},
  {0x00, 0, NULL},
};

static ulong robo_write(ulong reg, uchar *data)
{
    ulong error;
    ROBO_REG_WRITE(CAB_BASE_ADDR, reg, data, error);
    if (error) {
        printf("robo_write %02lx,%02lx,%lx: failed %08lx\n",
	       ROBO_PAGE(reg), ROBO_ADDR(reg), ROBO_LEN(reg), error);
    }
    return error;
}

static ulong robo_read(ulong reg, uchar *data)
{
    ulong error;
    ROBO_REG_READ(CAB_BASE_ADDR, reg, data, error);
    return error;
}

#define CAB_CLOCK_PHY(Val) { \
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_OPT_CARD, (Val)); \
    udelay(1); \
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_OPT_CARD, (Val) | CAB_OPT_CARD_CLOCK); \
    udelay(1); \
}

static ulong cab_write_phy(ulong device, ulong reg, ulong val)
{
    ulong i;
    /* preamble */
    for (i = 0; i < PHY_PREAMBLE; i++) {
        CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    }
    /* Start of Frame (ST) */
    CAB_CLOCK_PHY(0);
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    /* Operation Code (Write) */
    CAB_CLOCK_PHY(0);
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    /* device */
    for (i = 0; i < 5; i++) {
        CAB_CLOCK_PHY((device & (0x1 << (4-i)))? CAB_OPT_CARD_DOUT: 0);
    }
    /* register */
    for (i = 0; i < 5; i++) {
        CAB_CLOCK_PHY((reg & (0x1 << (4-i)))? CAB_OPT_CARD_DOUT: 0);
    }
    /* Turnaround (TA) */
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    CAB_CLOCK_PHY(0);
    /* value */
    for (i = 0; i < 16; i++) {
        CAB_CLOCK_PHY((val & (0x1 << (15-i)))? CAB_OPT_CARD_DOUT: 0);
    }
    return 0;
}

static ulong cab_read_phy(ulong device, ulong reg, ulong *val)
{
    ulong i, data;
    /* preamble */
    for (i = 0; i < PHY_PREAMBLE; i++) {
        CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    }
    /* Start of Frame (ST) */
    CAB_CLOCK_PHY(0);
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    /* Operation Code (read) */
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    CAB_CLOCK_PHY(0);
    /* device */
    for (i = 0; i < 5; i++) {
        CAB_CLOCK_PHY((device & (0x1 << (4-i)))? CAB_OPT_CARD_DOUT: 0);
    }
    /* register */
    for (i = 0; i < 5; i++) {
        CAB_CLOCK_PHY((reg & (0x1 << (4-i)))? CAB_OPT_CARD_DOUT: 0);
    }
    /* Turnaround (TA) */
    CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
    /* value */
    *val = 0;
    for (i = 0; i < 16; i++) {
        CAB_CLOCK_PHY(CAB_OPT_CARD_DOUT);
	CAB_REG_READ(CAB_BASE_ADDR, CAB_OPT_CARD, data);
	*val |= ((data & CAB_OPT_CARD_DIN)? (0x1 << (15-i)) : 0);
    }
    return 0;
}

static void
LoopbackTimeout (void)
{
    puts ("Timeout expired\n");
    NetState = NETLOOP_FAIL;
}

static void
LoopbackHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
    int i;

    if (len != loopback_length) {
        printf("Error: length mismatch %d != %ld\n",
	       len, loopback_length);
	goto error;
    }
    for (i = 0; i < len; i++, pkt++) {
        if (*pkt != NetTxPacket[i]) {
	    printf("Error: byte %d mismatch: %02x != %02x\n",
		   i, *pkt, NetTxPacket[i]);
	    goto error;
	}
    }
    loopback_received++;
    
    if (loopback_incr) {
        if (loopback_length++ >= 1514) {
	    loopback_length = PKTSIZEMIN;
	}
    }

    if (serial_tstc() || 
	(loopback_count && (loopback_count-- == 1))) {
        NetState = NETLOOP_SUCCESS;
    } else {
        NetSendPacket(NetTxPacket, loopback_length);
	NetSetTimeout (TIMEOUT * CFG_HZ, LoopbackTimeout);
    }
    return;

 error:
    NetState = NETLOOP_FAIL;
}

void LoopbackRequest(void)
{
    /* called when a packet is received during ether test */
    int i;
    volatile uchar *pkt;

    pkt = NetTxPacket;

    NetSetEther(pkt, NetBcastAddr, PROT_TAG);

    pkt += ETHER_HDR_SIZE;
    *((ushort *)pkt) = loopback_vlan; /* VLAN ID */
    pkt += sizeof(ushort);

    for (i = 0; i < PKTSIZE - ETHER_HDR_SIZE - sizeof(ushort); i++) {
        pkt[i] = i & 0xff;
    }

    if (!loopback_rcv) {
        while (loopback_count--) {
	    NetSendPacket(NetTxPacket, loopback_length);
	}
        NetState = NETLOOP_SUCCESS;
    } else {

        NetSetHandler(LoopbackHandler);
	NetSetTimeout(TIMEOUT * CFG_HZ, LoopbackTimeout);

	NetSendPacket(NetTxPacket, loopback_length);
    }
}

int do_cabernet_ether(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong length = PKTSIZEMIN, count = 0, incr = 0, vlan = 1;

    if ((argc < 2) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (eth_set(argv[1])) {
        printf("unknown device %s\n", argv[1]);
	return 1;
    }

    if (argc > 2) {
        if (argv[2][0] == 'i') {
  	    incr = 1;
	} else {
	    length = simple_strtoul(argv[2], NULL, 16);
	    if ((length < PKTSIZEMIN) || (length > 1514)) {
	        printf("length is invalid: %ld\n", length);
		return 1;
	    }
	}
    }

    if (argc > 3) {
        count = simple_strtoul(argv[3], NULL, 16);
	if (count == 0) {
	    printf("count is invalid: %ld\n", count);
	    return 1;
	}
    }

    if (argc > 4) {
        vlan = simple_strtoul(argv[4], NULL, 16);
	if (vlan == 0 || vlan > 4095) {
	    printf("vlan is invalid: %ld\n", vlan);
	    return 1;
	}
    }

    if (argc > 5) {
        /* do bother receiving */
        loopback_rcv = 0;
    } else {
        loopback_rcv = 1;
    }

    loopback_length = length;
    loopback_incr = incr;
    loopback_count = count;
    loopback_vlan = vlan;
    loopback_received = 0;

    NetLoop(bd, LOOPBACK);

    printf("%lx packets successfully received\n", loopback_received);

    return 0;
}

int do_cabernet_cab(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong reg, val;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 2) || (argc > 5)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {

	/* reset SOE RX, SOE CPLD, and optical card, if any */
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_RESET, CAB_SOE_RESET_RX);
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_OPT_CARD, CAB_OPT_CARD_RESET);
	udelay(1000000);
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_RESET, 0);
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_OPT_CARD, 0);

        /* set status LED to green */
      	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_LED_STATUS,
		      CAB_LED_STAT_COLOR_SET(CAB_LED_COLOR_GREEN));

    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    if (argc < 4) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	    reg = simple_strtoul(argv[2], NULL, 16);
	    val = simple_strtoul(argv[3], NULL, 16);

	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    while (times--) {
	        CAB_REG_WRITE(CAB_BASE_ADDR, reg, val);
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    /* Print the lines.
	     */
	    while (times--) {
	        reg = simple_strtoul(argv[2], NULL, 16);
	        if (argc > 3) {
		    length = simple_strtoul(argv[3], NULL, 16);
		}
	        do {
		    CAB_REG_READ(CAB_BASE_ADDR, reg, val);
		    if (!times) {
		        printf("%04lx: %08lx\n", reg, val);
		    }
		    reg += 1;
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return (rc);	
}

static void robo_set_vlan(__u16 vlan, __u32 entry)
{
    __u16 index;
    __u32 addr;
    ulong error;
    int   trys;
    __u8  data[ROBO_LEN(ROBO_REG_MEMORY_DATA)];

    /* form bit-swapped index */
    index = ROBO_VLAN_INDEX(vlan);

    addr = ((ROBO_QVLAN_TABLE_BASE + (index >> 1)) |
	    ROBO_REG_MEMORY_READ |
	    ROBO_REG_MEMORY_REQUEST);
    robo_write(ROBO_REG_MEMORY_ADDRESS, (__u8 *)&addr);
    
    /* wait for request to complete */
    trys = 0;
    do {
        if (trys++ > 10000) {
	    printf("Error: Memory read %x timed-out\n", addr);
	    return;
	}
	error = robo_read(ROBO_REG_MEMORY_ADDRESS, (__u8 *)&addr);
	if (error) {
	    printf("Error: Memory read %x failed %lx\n", addr, error);
	}
    } while (addr & ROBO_REG_MEMORY_REQUEST);

    error = robo_read(ROBO_REG_MEMORY_DATA, (__u8 *)data);
    if (error) {
        printf("Error: Memory Access %x failed %lx\n", addr, error);
    }
    
    if (index & 0x1) {
        /* update "left" entry */
        memcpy(&data[sizeof(data) - 2*sizeof(entry)], &entry, sizeof(entry));
    } else {
        /* update "right" entry */
        memcpy(&data[sizeof(data) - sizeof(entry)], &entry, sizeof(entry));
    }

    robo_write(ROBO_REG_MEMORY_DATA, (__u8 *)data);
    addr = ((ROBO_QVLAN_TABLE_BASE + (index >> 1)) |
	    ROBO_REG_MEMORY_REQUEST);
    robo_write(ROBO_REG_MEMORY_ADDRESS, (__u8 *)&addr);

    /* wait for request to complete */
    trys = 0;
    do {
        if (trys++ > 10000) {
	    printf("Error: Memory write %x timed-out\n", addr);
	    return;
	}
	error = robo_read(ROBO_REG_MEMORY_ADDRESS, (__u8 *)&addr);
	if (error) {
	    printf("Error: Memory write %x failed %lx\n", addr, error);
	}
    } while (addr & ROBO_REG_MEMORY_REQUEST);
}

static void robo_init(int fbm, int bridge)
{
    __u16 data16;
    __u8 data8;
    int i;

    /* 
     * Initialize RoboSwitch
     *
     * most of this was taken from "EEPROM code" errata
     */
    /* reset robo */
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE,
		       CAB_CPLD_RESET_ENABLE_VALUE);
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET,
		       CAB_CPLD_RESET_SWITCH);
    udelay(1000);
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET, 0);
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE, 0);

    /* Enable HOL */
    data16 = 0x03d4;
    if (robo_write(ROBO_REG(0x0a, 0x9c, 2), (__u8 *)&data16)) {
        /* check first write for error */
        return;
    }

    /* Adjust broadcast and multicast thresholds */
    data16 = 0x550c;
    robo_write(ROBO_REG(0x0a, 0x08, 2), (__u8 *)&data16);
    data16 = 0x3f0c;
    robo_write(ROBO_REG(0x0a, 0x04, 2), (__u8 *)&data16);
    data16 = 0x0030;
    robo_write(ROBO_REG(0x0a, 0x88, 2), (__u8 *)&data16);

    /* Shore up half-duplex (workaround for lock-up) */
    data16 = 0x0003;
    robo_write(ROBO_REG(0x0a, 0x9e, 2), (__u8 *)&data16);
    data16 = 0x0015;
    robo_write(ROBO_REG(0x0a, 0x30, 2), (__u8 *)&data16);
    data16 = 0xb00a;
    robo_write(ROBO_REG(0x0a, 0x4a, 2), (__u8 *)&data16);
    data16 = 0xbe80;
    robo_write(ROBO_REG(0x0a, 0x10, 2), (__u8 *)&data16);
    
    /* Set External GE PHYs to advertise 10/100 full/half and FC capable */
    data16 = 0x0de1;
    robo_write(ROBO_REG_GMII_ADVER(ROBO_EXT_GMAC-ROBO_FIRST_GMAC),
	       (__u8 *)&data16);
    /* Enhance GPHY0 cable length */
    data16 = 0x0c00;
    robo_write(ROBO_REG(0x1a, 0x30, 2), (__u8 *)&data16);
    data16 = 0x000a;
    robo_write(ROBO_REG(0x1a, 0x2e, 2), (__u8 *)&data16);
    data16 = 0x032a;
    robo_write(ROBO_REG(0x1a, 0x2a, 2), (__u8 *)&data16);
    /* Restart auto-negotiation */
    data16 = 0x1240;
    robo_write(ROBO_REG_GMII_CONTROL(ROBO_EXT_GMAC-ROBO_FIRST_GMAC),
	       (__u8 *)&data16);

    if (bridge) {
        /* configure all ports to be in vlan 1, and no tagging */
        __u16 port_mask;
        /* disable .1Q */
        data8 = 0;
	robo_write(ROBO_REG_QVLAN_CONTROL_0, (__u8 *)&data8);
	/* configure port VLAN to include all other ports */
	port_mask = 0;
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_INT_PORT(i) || ROBO_EXT_PORT(i)) {
	        port_mask |= (0x1 << i);
	    }
	}
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_INT_PORT(i) || ROBO_EXT_PORT(i)) {
	        data16 = port_mask ^ (0x1 << i);
		robo_write(ROBO_REG_VLAN_CONTROL(i), (__u8 *)&data16);
	    }
	}
    } else if (fbm) {
        /* configure all ports to be in vlan 1, and no tagging */
        /* disable .1Q */
        data8 = 0;
	robo_write(ROBO_REG_QVLAN_CONTROL_0, (__u8 *)&data8);
	/* configure internal GMAC vlan to include all other ports */
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_EXT_PORT(i)) {
	        data16 |= (0x1 << i);
	    }
	}
	robo_write(ROBO_REG_VLAN_CONTROL(ROBO_INT_GMAC), (__u8 *)&data16);
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_EXT_PORT(i)) {
	        if (ROBO_GE_PORT(i)) {
		    /* loop itself */
		    data16 = 0x1 << i;
		} else {
		    /* loop to adjacent port */
		    data16 = 0x1 << (i ^ 0x1);
		}
	        robo_write(ROBO_REG_VLAN_CONTROL(i), (__u8 *)&data16);
	    }
	}
    } else {
	/* configure port VLANs to only include internal GMAC */
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_EXT_PORT(i)) {
	        data16 = 0x1 << ROBO_INT_GMAC;
	    } else {
	        /* configure internal GMAC to include ALL ports */
	        data16 = (0x1 << ROBO_PORTS) - 1;
	    }
	    robo_write(ROBO_REG_VLAN_CONTROL(i), (__u8 *)&data16);
	}

        /* Configure each port to be in a different .1Q VLAN */
        for (i = 0; i < ROBO_PORTS; i++) {
	    if (ROBO_EXT_PORT(i)) {
	        __u32 vlan_entry;
	        /* configure VLAN */
		data16 = port_map[i].pvid;
	        if (ROBO_GE_PORT(i)) {
		    robo_write(ROBO_REG_QVLAN_PVID(i-1), (__u8 *)&data16);
		    vlan_entry = (ROBO_QVLAN_VALID |
				  ((0x1 << (ROBO_EXT_GMAC+1)) << ROBO_QVLAN_UNTAG_SHIFT) |
				  (((0x1 << (ROBO_EXT_GMAC+1)) |
				    (0x1 << (ROBO_INT_GMAC+1))) << ROBO_QVLAN_FWD_SHIFT));
		} else {
		    robo_write(ROBO_REG_QVLAN_PVID(i), (__u8 *)&data16);
		    vlan_entry = (ROBO_QVLAN_VALID |
				  ((0x1 << i) << ROBO_QVLAN_UNTAG_SHIFT) |
				  (((0x1 << i) |
				    (0x1 << (ROBO_INT_GMAC+1))) << ROBO_QVLAN_FWD_SHIFT));
		}
		robo_set_vlan(data16, vlan_entry);
	    }
	}

	/* Enable .1Q */
        data8 = (ROBO_REG_QVLAN_CONTROL_0_HSH |
		 ROBO_REG_QVLAN_CONTROL_0_CHK |
		 ROBO_REG_QVLAN_CONTROL_0_ENA);
	robo_write(ROBO_REG_QVLAN_CONTROL_0, (__u8 *)&data8);
	/* drop VLAN ID violations */
        data8 = (ROBO_REG_QVLAN_CONTROL_4_VIO |
		 ROBO_REG_QVLAN_CONTROL_4_G1);
	robo_write(ROBO_REG_QVLAN_CONTROL_4, (__u8 *)&data8);
	/* ignore ether FCS and BCM CRC from mgmt port */
	data8 = (ROBO_REG_QVLAN_CONTROL_5_IG |
		 ROBO_REG_QVLAN_CONTROL_5_FCS);
	robo_write(ROBO_REG_QVLAN_CONTROL_5, (__u8 *)&data8);
    }

    /* set port states to forwarding */
    data8 = ROBO_REG_FE_STP_FORWARDING << ROBO_REG_FE_STP_STATE_SHIFT;
    for (i = 0; i < ROBO_PORTS; i++) {
        if (ROBO_FE_PORT(i)) {
	    robo_write(ROBO_REG_FE_CONTROL(i), (__u8 *)&data8);
	} else if (ROBO_GE_PORT(i)) {
	    robo_write(ROBO_REG_GE_CONTROL(i-ROBO_FIRST_GMAC), (__u8 *)&data8);
	}
    }

    /* hardcode gig port 1 (uplink to sibyte) */
    data8 = (ROBO_REG_GMII_OVERRIDE_ENA  |
	     ROBO_REG_GMII_OVERRIDE_1000 |
	     ROBO_REG_GMII_OVERRIDE_FD   |
	     ROBO_REG_GMII_OVERRIDE_LINK);
    robo_write(ROBO_REG_GMII_OVERRIDE, (__u8 *)&data8);

    /* enable switching */
    data8 = ROBO_REG_SWITCH_MANAGED | ROBO_REG_SWITCH_FORWARD;
    robo_write(ROBO_REG_SWITCH_MODE, (__u8 *)&data8);
}

static int robo_port_stats(ulong port)
{
    robo_sstat_t *stat;
    __u8 data[CAB_SPI_MAX_LENGTH];
    ulong error;
    int i;
    int mac;
    int first;
    int bank = 0;

    mac = port_map[port].mac;
    if (mac == 9) return 0;
    stat = robo_stats;
    while (stat->length && stat->str) {
        error = robo_read(ROBO_REG(ROBO_PAGE_MIB(mac), stat->addr, stat->length),
			  data);
	if (error) {
	    printf("robo_read %02x,%02x,%x: failed %08lx\n",
		   ROBO_PAGE_MIB(mac), stat->addr, stat->length,
		   error);
	    return 1;
	}
	for (i = 0; i < stat->length; i++) {
	    if (data[i]) break;
	}

	if (i < stat->length) {
	    if (!bank) printf("port %lx:\n", port);
	    printf("  %-18s ", stat->str);
	    for (i = 8; i >= stat->length; i--) {
	        printf("  ");
	    }
	    first = 0;
	    for (i = 0; i < stat->length; i++) {
	        if (first) {
		    printf("%02x", data[i]);
	        } else if (data[i]) {
		    first = 1;
		    printf("%2x", data[i]);
		} else {
		    printf("  ");
		}
	    }
	    if (bank++ & 0x1) printf("\n");
	}
	stat++;
    }
    if (bank & 0x1) printf("\n");
    return 0;
}

static int robo_port_stats_clear(ulong port)
{
    robo_sstat_t *stat;
    __u8 data[CAB_SPI_MAX_LENGTH];
    int mac;

    mac = port_map[port].mac;
    if (mac == 9) return 0;
    memset(data, 0, sizeof(data));
    stat = robo_stats;
    while (stat->length && stat->str) {
        robo_write(ROBO_REG(ROBO_PAGE_MIB(mac), stat->addr, stat->length),
		   data);
	stat++;
    }
    return 0;
}

int do_cabernet_robo(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    int rc = 0;
    ulong error;
    ulong addr;
    ulong port;
    ulong page, length;
    __u8 data[CAB_SPI_MAX_LENGTH];
    int times = 1;
    int i;

    if ((argc < 2) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {
	int fbm = 0;
	int bridge = 0;
        if (argc > 2) {
	    if (argv[2][0] == 'f') {
	        /* fast loopback test */
	        fbm = 1;
	    } else if (argv[2][0] == 'b') {
	        /* init simple bridge */
	        bridge = 1;
	    }
	}

	/* initialize roboswitch state */
	robo_init(fbm, bridge);

    } else if (argv[1][0] == 's') {
        if (argc < 3) {
	    /* display stats from ALL ports */
	    for (port = 0; port < ROBO_PORTS; port++) {
	       if (robo_port_stats(port)) {
		   return 1;
	       }
	    }
	} else {
	    if (argv[2][0] == 'c') {
	        /* clear stats on all ports */
	        for (port = 0; port < ROBO_PORTS; port++) {
		    if (robo_port_stats_clear(port)) {
		        return 1;
		    }
		}
		printf("cleared\n");
	    } else {
	        /* display stats for specific port */
	        port = simple_strtoul(argv[2], NULL, 16);
		if (port >= ROBO_PORTS) {
		    printf ("Illegal port %lx\n", port);
		    return 1;
		}
		robo_port_stats(port);
	    }
	}
      
    } else {
        if (argc < 5) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	page = simple_strtoul(argv[2], NULL, 16);
	addr = simple_strtoul(argv[3], NULL, 16);
	length = simple_strtoul(argv[4], NULL, 16);

	if (page > 0xff) {
	    printf("Error: invalid page %lx\n", page);
	    return 1;
	}
	if (addr > 0xff) {
	    printf("Error: invalid addr %lx\n", addr);
	    return 1;
	}
	if (length > CAB_SPI_MAX_LENGTH) {
	    printf("Error: invalid length %lx\n", length);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    int j;
	    if (argc < 6) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }

	    for (i = 0; i < length; i++) {
	        data[i] = 0;
	    }
	    /* convert string nibbles into bytes */
	    for (j = 2*length - 1, i = strlen(argv[5]) - 1;
		 i >= 0 && j >= 0;
		 i--, j--) {
	        char c = argv[5][i];
		__u8 v;
		if (c >= '0' && c <= '9') {
		    v = c - '0';
		} else if (c >= 'a' && c <= 'f') {
		    v = c - 'a' + 0xa;
		} else 	if (c >= 'A' && c <= 'F') {
		    v = c - 'A' + 0xA;
		} else {
		    printf("Error: invalid data %s\n", argv[5]);
		    return 1;
		}
	        if (j & 0x1) {
		  data[j/2] = v;
		} else {
		  data[j/2] |= (v << 4);
		}
	    }
	    
	    printf("%02lx,%02lx,%lx: ", page, addr, length);
	    for (i = 0; i < length; i++) {
	        printf("%02x", data[i]);
	    }
	    printf("\n");

	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }

	    while (times--) {
	        robo_write(ROBO_REG(page, addr, length), data);
	    }
	} else {
	    if (argc > 5) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }

	    while (times--) {
	        /* Print the lines.
		 */
		error = robo_read(ROBO_REG(page, addr, length), data);
		if (error) {
		    printf("%02lx,%02lx,%lx: failed %08lx\n",
			   page, addr, length, error);
		    return 1;
		} else if (!times) {
		    int i;
		    printf("%02lx,%02lx,%lx: ", page, addr, length);
		    for (i = 0; i < length; i++) {
		        printf("%02x", data[i]);
		    }
		    printf("\n");
		}
		if (ctrlc()) {
		    return 1;
		}
	    }
	}
    }
    return (rc);	
}

static void
uart_putc(int uart, const char c)
{
    if (c == '\n') {
        NS16550_putc(COM_PORTS[uart], '\r');
    }
    NS16550_putc(COM_PORTS[uart], c);
}

static void
uart_puts (int uart, const char *s)
{
    while (*s) {
        uart_putc (uart, *s++);
    }
}

static int
uart_getc(int uart)
{
    return NS16550_getc(COM_PORTS[uart]);
}

static int
uart_tstc(int uart)
{
    return NS16550_tstc(COM_PORTS[uart]);
}

static int
uart_space(int uart)
{
    return NS16550_space(COM_PORTS[uart]);
}

static int soe_loop = 0;
static int soe_cr = 0;

static void
soe_init(int port, int loop, int cr)
{
    ulong data;

    soe_loop = loop;
    soe_cr = cr;

    /* reset fifos */
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_RX_FIFO_RESET, ~0); /* all RX */
    /* set LOOPBACK bit */
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_CONTROL,
		  (soe_loop? CAB_SOE_CONTROL_LOOPBACK: 0));
    /* wait some time */
    udelay(20000);
    /* bring out of reset */
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_RX_FIFO_RESET, 0);

    /* enable flow control */
    CAB_REG_READ(CAB_BASE_ADDR, CAB_SOE_FC_CONTROL, data);
    data |= CAB_SOE_PORT(port);
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_FC_CONTROL, data);
}

static void
soe_putc(int port, const char c)
{
    ulong data;

    if ((c == '\n') && soe_cr) {
        soe_putc(port, '\r');
    }

    /* wait for TX to be ready for another character */
    do {
        CAB_REG_READ(CAB_BASE_ADDR, CAB_SOE_TX_FIFO_NOTFULL, data);
    } while (!(data & CAB_SOE_PORT(port)));

    /* write character */
    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_CONTROL,
		  ((port << CAB_SOE_CONTROL_ADDR_SHIFT) |
		   (soe_loop? CAB_SOE_CONTROL_LOOPBACK: 0) |
		   c));
}

static void
soe_puts (int port, const char *s)
{
    while (*s) {
        soe_putc (port, *s++);
    }
}

static int
soe_getc(int port)
{
    ulong data;

    do {
        CAB_REG_READ(CAB_BASE_ADDR, CAB_SOE_RX_FIFO_NOTEMPTY, data);
    } while (!(data & CAB_SOE_PORT(port)));

    CAB_REG_WRITE(CAB_BASE_ADDR, CAB_SOE_CONTROL, 
		  (CAB_SOE_CONTROL_RX |
		   (soe_loop? CAB_SOE_CONTROL_LOOPBACK: 0) |
		   (port << CAB_SOE_CONTROL_ADDR_SHIFT)));
    do {
        CAB_REG_READ(CAB_BASE_ADDR, CAB_SOE_CONTROL, data);
    } while (!(data & CAB_SOE_CONTROL_RD_DONE));

    return (int) (data & CAB_SOE_CONTROL_DATA_MASK);
}

static int
soe_tstc(int port)
{
    ulong data;

    CAB_REG_READ(CAB_BASE_ADDR, CAB_SOE_RX_FIFO_NOTEMPTY, data);

    return (data & CAB_SOE_PORT(port))? 1: 0;
}

int do_cabernet_soe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    if (argc < 4) {
        printf ("Usage %d:\n%s\n", argc, cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'c') {
	int i, loop = 0, cr = 0;
        ulong iport;
        ulong eport;
	char c;

        eport = simple_strtoul(argv[2], NULL, 16);
	if (eport > 24) {
	    printf("%04lx: eport invalid\n", eport);
	    return 1;
	}

	if (argv[3][0] == 'l') {
	    loop = 1;
	    iport = eport;
        } else if (argv[3][0] == 'c') {
            cr = 1;
	    iport = eport;
	} else {
	    iport = simple_strtoul(argv[3], NULL, 16);
	    if (iport > 8) { /* XXX */
	        printf("%04lx: iport invalid\n", iport);
		return 1;
	    }
	}

	soe_init(iport, loop, cr);
	if (eport != iport) {
	    soe_init(eport, loop, cr);
	}
	if (argc > 5) {
	    for (i = 5; i < argc; i++) {
	        soe_puts(eport, argv[i]);
		if (i < argc - 1) {
		    soe_putc(eport, ' ');
		} else {
		    soe_putc(eport, '\n');
		}
	    }
	}

	printf("Connecting UART A to SOE UART %lx->%lx...\n"
	       "Press <Control-y> to stop\n",
	       eport, iport);
	while (1) {
	    WATCHDOG_RESET();
	    if (uart_tstc(0)) {
	        c = uart_getc(0);
		if (c == 0x19) break; /* control-y */
		if (c == '\r') c = '\n';
		soe_putc(eport, c);
	    }
	    if (soe_tstc(iport)) {
	        c = soe_getc(iport);
		uart_putc(0, c);
	    }
	}
	printf("\n");

    } else {
    }

    return 0;
}

#include "tps2383.h"

int do_cabernet_poe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong port;
    ulong data;
    char nTps = 0;
    ulong ms = 16;

    if (argc < 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {
        /* initialize POE chip */
        tps_init(nTps);

    } else if (argv[1][0] == 'u') {
        /* ramp up POE port */
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        port = simple_strtoul(argv[2], NULL, 16);
	if (argc > 3) {
	    ms = simple_strtoul(argv[3], NULL, 16);
	}
	
	/* jog to specific port */
	if (tps_jog(port, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_ramp_up(port, nTps, ms);

	/* turn POE LED on */
	CAB_REG_READ(CAB_BASE_ADDR, CAB_LED_POE, data);
	data |= (CAB_LED_COLOR_GREEN << (port*2));
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_LED_POE, data);

    } else if (argv[1][0] == 'd') {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	/* jog to specific port */
	if (tps_jog(port, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_ramp_down(port, nTps);

	/* turn POE LED off */
	CAB_REG_READ(CAB_BASE_ADDR, CAB_LED_POE, data);
	data &= ~(CAB_LED_COLOR_GREEN << (port*2));
	CAB_REG_WRITE(CAB_BASE_ADDR, CAB_LED_POE, data);

    } else if (argv[1][0] == 'p') {
        int present;
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	/* jog to specific port */
	if (tps_jog(port, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	if (argv[3][0] == 'r') {
	    present = tps_discover_resistor(port, nTps);
	} else {
	    present = tps_discover_capacitor(port, nTps);
	}
	printf("%d\n", present);

    } else if (argv[1][0] == 's') {
        int value;
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	/* jog to specific port */
	if (tps_jog(port, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	value = tps_sample(port, nTps, (argv[3][0] == 'v'));
	printf("%d m%c\n", value, (argv[3][0] == 'v')? 'V':'A');

    } else if (argv[1][0] == 'c') {

        /* calibrate TPS */
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	/* jog to specific port */
	if (tps_jog(port, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_calibrate(port, nTps);

    } else if (argv[1][0] == 'r') {

        tps_reg_dump(nTps);

    } else {
        printf("%s: command invalid\n", argv[1]);
	return 1;
    }

    return (0);
}


#define BUF_SIZE 2048
#define UART_NEXT(I) (((I) + 1) & (BUF_SIZE-1))
char uart_buffer[UARTS][BUF_SIZE];


int do_cabernet_uartb(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    char c;
    int i;
    int prod[UARTS] = {0};
    int cons[UARTS] = {0};

    NS16550_init(1, CONFIG_BUS_CLK_FREQ / 16 / 115200);

    if (argc > 1) {
        for (i = 1; i < argc; i++) {
          uart_puts(1, argv[i]);
          uart_putc(1, ' ');
        }
	uart_putc(1, '\r');
        return 0;
    }

    printf("Connecting UART A to B...\nPress 'Q' to quit\n");
    while (1) {
        WATCHDOG_RESET();
	for (i = 0; i < UARTS; i++) {
	    if (uart_tstc(i) &&
		(UART_NEXT(prod[i]) != cons[i])) {
	        c = uart_getc(i);
		if ((c == 'Q') && (i == 0)) return 0;
		uart_buffer[i][prod[i]] = c;
		prod[i] = UART_NEXT(prod[i]);
	    }
	    if ((prod[i] != cons[i]) &&
		uart_space(i^1)) {
	        uart_putc(i^1, uart_buffer[i][cons[i]]);
		cons[i] = UART_NEXT(cons[i]);
	    }
	}
    }

    return 0;
}

int do_cabernet_phy(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong device, reg, val, error;
    int length = 1;
    int times = 1;
    int rc = 0;
    
    if ((argc < 3) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'l') {
        for (device = 0; device < PHY_DEVS; device++) {
	    cab_read_phy(device, 0x1, &val);
	    cab_read_phy(device, 0x1, &val);
	    if (val & 0x4) {
	        printf("link on %lx\n", device);
	    }
	}
    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        device = simple_strtoul(argv[2], NULL, 16);
	if (device >= PHY_DEVS) {
	    printf("%lx: device invalid\n", device);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    if (argc < 5) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	
	    reg = simple_strtoul(argv[3], NULL, 16);
	    if (reg >= PHY_REGS) {
	        printf("%04lx: register invalid\n", reg);
		return 1;
	    } 
	    val = simple_strtoul(argv[4], NULL, 16);
	    if (val >= 65536) {
	        printf("%02lx: value invalid\n", val);
		return 1;
	    }

	    if (argc > 6) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }

	    while (times--) {
	        error = cab_write_phy(device, reg, val);
		if (error) {
		    printf("cab_phy_write %lx: failed %08lx\n",
			   reg, error);
		    return 1;
		}
	    }

	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 5) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }
	    
	    while (times--) {
	        if (argc > 3) {
		    reg = simple_strtoul(argv[3], NULL, 16);
		    if (reg >= PHY_REGS) {
		        printf("%04lx: register invalid\n", reg);
			return 1;
		    } 
		} else {
		    reg = 0;
		    length = 0x20;
		}
	       if (argc > 4) {
		   length = simple_strtoul(argv[4], NULL, 16);
	       }
    
	       /* Print the lines.
		*/
	       do {
		   error = cab_read_phy(device, reg, &val);
		   if (error) {
		       printf("%lx,%02lx: failed %lx\n",
			      device, reg, error);
		       return 1;
		   } else if (!times) {
		       printf("%lx,%02lx: %04lx\n",
			      device, reg, val & 0xffff);
		   }
		   reg += 1;
		   if (reg >= PHY_REGS) {
		       break;
		   }
		   if (ctrlc()) {
		       return 1;
		   }
	       } while (--length > 0);
	    }
	}
    }

    return (rc);	
}

#if (CONFIG_COMMANDS & CFG_CMD_IDE)
extern block_dev_desc_t * ide_get_dev(int dev);

static int le32_to_int(unsigned char *le32)
{
    return ((le32[3] << 24) +
	    (le32[2] << 16) +
	    (le32[1] << 8) +
	     le32[0]
	   );
}

static int upgrade_device;
static int upgrade_block;
static int upgrade_end;
static int upgrade_cksum;
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

int ide_tftp_write(ulong offset, u8 *src, unsigned len)
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    int i, *ptr;

    ptr = (int *) src;
    for (i = 0; i < len/sizeof(int); i++, ptr++) {
        upgrade_cksum += *ptr;
    }

    dev = ide_get_dev(upgrade_device);
    for (; len && (upgrade_block < upgrade_end); upgrade_block++) {
        if (dev->block_write(dev->dev, upgrade_block, 1, (ulong*)src) != 1) {
	    printf("%d write failed\n", upgrade_block);
	}
	if (len < dev->blksz) {
	    len = 0;
	} else {
	    len -= dev->blksz;
	}
	src += dev->blksz;
    }
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
    return len? ERR_INVAL: ERR_OK;
}

#define DOS_PARTITIONS 4

int do_cabernet_upgrade(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    int block, count;
    unsigned device;
    ulong lba512;
    unsigned partition;
    char c;

    if (argc == 2) {
        if (argv[1][0] == 'i') {
            run_command("fpga p", bd, 0);
            run_command("cab i", bd, 0);
            run_command("robo i b", bd, 0);
            run_command("uartb eforw eth1 eth2", bd, 0);
            return 0;
        }
        return 1;
    }

    if (argc != 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][1] == ':') {
        device = simple_strtoul(argv[1], NULL, 16);
        partition = simple_strtoul(&argv[1][2], NULL, 16);
    } else {
        device = 0;
        partition = simple_strtoul(argv[1], NULL, 16);
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("%x: Device invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("%x: Partition invalid\n", partition);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("%x: Device invalid\n", device);
	return 1;
    }

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("%x: Device not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Partition type (if any) unrecognized\n");
	return 1;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d: Read failed\n", device);
	return 1;
    }

    /* The first two partitions can contain images */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        printf("%x: Partition not created\n", partition);
    }

    block = le32_to_int (pt->start4);
    count = block + le32_to_int (pt->size4);

    /* save state */
    upgrade_device = device;
    upgrade_block = block;
    upgrade_end = count;
    upgrade_cksum = 0;
    copy_filename (BootFile, argv[2], sizeof(BootFile));
    load_addr = CFG_ATA_BASE_ADDR;

    /* fire off TFTP */
    eth_set("i82559#0");
    if ((ar->size = NetLoop(bd, TFTP)) == 0) {
        return 1;
    }

    memcpy(buffer, (u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET, CFG_ENV_SECT_SIZE);
    if (buffer[0] != (device<<4|partition)) {
        printf("Try to boot this partition first ([y]/n)? ");
	if ((c = uart_getc(0)) != 'n') {
	    buffer[0] = device<<4|partition;
	    flash_write(buffer,
			CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET,
			CFG_ENV_SECT_SIZE);
	}
	printf("%c\n", c);
    }

#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
    return (0);
}

#define FLASH_PARTS 4
#define FLASH_MARGIN 136
dos_partition_t part[FLASH_PARTS] = {
  { 0x00, 0x01, 0x01, 0x00, 0x83, 0x07, 0x60, 0x7f,
    { 0x20, 0x00, 0x00, 0x00 }, { 0xe0, 0x7f, 0x01, 0x00 } },
  { 0x00, 0x00, 0x41, 0x80, 0x83, 0x07, 0xe0, 0x00,
    { 0x00, 0x80, 0x01, 0x00 }, { 0x00, 0x81, 0x01, 0x00 } },
  { 0x00, 0x00, 0xc1, 0x01, 0x83, 0x07, 0xe0, 0xd1,
    { 0x00, 0x01, 0x03, 0x00 }, { 0x00, 0xd1, 0x00, 0x00 } },
  { 0x00 },
};

static unsigned long adjustSectors(unsigned long lba512)
{
    // printf("old lba512 = %d\n", (int) lba512);
    if((lba512 > 500000) && ( lba512 < 512000)) {
        //
        // 256 MB flash
        //
        // AVED comes in with 500400
        // WINTEC comes in with 500736
        //
        // Some web sites mention there are two types of 256 MB flashes
        // One that has 1954 blocks and one that has 1956 blocks.
        // To be safe and not trust the numbers passed in by the device 
        // we go with the smaller number of blocks i.e. 1954.
        //
        // 1954 did not work well with WINTEC flashes we try 1950
        // number of sectors = number of blocks(1950) * sectors(32) * heads(8)
        //                   = 499200
        //
        if(lba512 > 500224) {
                lba512 = 499200; // 1950 * 32 * 8;
        }
    } else {
        // 128 MB flash and 
        // all other sizes except 256MB 
        lba512 -= FLASH_MARGIN; /* SAMEER - 2/24/04 - remove sectors as margin */
    }
    // printf("new lba512 = %d\n", (int) lba512);
    return lba512;
}

int do_cabernet_part(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    ulong lba512;
    ulong size;
    unsigned device;
    int j;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        device = simple_strtoul(argv[1], NULL, 16);
    } else {
        device = 0;
    }
    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    for (j = 0; j < FLASH_PARTS; j++) {
        if (j < 2) {
            size = (part[j].size4[0] |
                    (part[j].size4[1] << 8) |
                    (part[j].size4[2] << 16) |
                    (part[j].size4[3] << 24));
            lba512 -= size;
        } else if (j == 2) {
            part[j].size4[0] = lba512 & 0xff;
            part[j].size4[1] = (lba512 >>  8) & 0xff;
            part[j].size4[2] = (lba512 >> 16) & 0xff;
            part[j].size4[3] = (lba512 >> 24) & 0xff;
        }

        set_dos_partition(dev, &part[j], j);
    }
    print_part_dos(dev);
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return (0);
}

#if (CONFIG_COMMANDS & CFG_CMD_IDE)
static int verify_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    int i;
    char rbuf[512];

    if (dev->block_read(dev->dev, sector, 1, (ulong*)rbuf) != 1) {
        printf("%d read failed\n", sector);
        return 0;
    }
    for (i = 0; i < 512; i++) {
        if (wbuf[i] != rbuf[i]) {
            printf("%d read data mismatch\n", sector);
            return 0;
        }
    }
    return 1;
}

static int format_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    if (dev->block_write(dev->dev, sector, 1, (ulong*)wbuf) != 1) {
        printf("%d write failed\n", sector);
        return 0;
    }
    return 1;
}

static void format_partition(block_dev_desc_t *dev, int start, int end,
                             char *wbuf, int verify)
{
#define PUTC_MASK 0x7ff
    int i;

    for(i = start; i < end; i++){
        format_sector(dev, i, wbuf);
        if (!(i & PUTC_MASK)) {
            serial_putc('.');
        }
    }
    if (verify) {
        for(i = start; i < end; i++){
            if (!verify_sector(dev, i, wbuf)) {
                return;
            }
            if (!(i & PUTC_MASK)) {
                serial_putc(',');
            }
        }
    }
}
#endif

int do_cabernet_format(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    block_dev_desc_t *dev;
    ulong lba512;
    unsigned int device,partition;
    char c,src[512];
    char formatPart0=0;
    char formatPart1=0;
    char formatPart2=0;
    int verify = 0;
    int start, end;
    dos_partition_t *pt;

    if (argc > 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);

	    switch(partition){
		    case 0: formatPart0=1; break;
		    case 1: formatPart1=1; break;
		    case 2: formatPart2=1; break;
		    default:
			printf("Invalid device %s\n", argv[1]);
			return 1;
	    }
    	    printf("Format will erase everything on the %s flash's partition %d.\n", (device?"external":"internal"), partition);

	} else {
	    device = simple_strtoul(argv[1], NULL, 16);
	    formatPart0=1;
	    formatPart1=1;
	    formatPart2=1;
    	    printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
	} 

        if (argc > 2) {
            verify = 1;
        }

    } else {
	/* default format device 0 all three partitions. */
	formatPart0=1;
	formatPart1=1;
	formatPart2=1;
        device = 0;
    	printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    printf("Are you sure you want to continue (y/n)?");
    c=getc();
    if(c!='y' && c!='Y') {
	    printf("\nFormat Cancelled\n");
	    return 1;
    }
    printf("%c\nFormatting", c);

    memset(src,0xFF,sizeof(src));

    if(formatPart0) {
        pt = &part[0];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart1) {
        pt = &part[1];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart2) {
        pt = &part[2];
        start = le32_to_int (pt->start4);
        end = lba512;
        format_partition(dev, start, end, src, verify);
    } 
    printf("\ndone\n");

    /* partition it */
    return (do_cabernet_part(cmdtp, bd, flag, 1, argv));
#else /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return (0);
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */
}

static u8 *load_partition(bd_t *bd, unsigned device, unsigned partition,
			  u32 *len)
{
    u8 *addr = NULL;
#if (CONFIG_COMMANDS & CFG_CMD_IDE)
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    ulong lba512;
    int block, count;
    int cksum, i, *ptr;

    dev = ide_get_dev(device);
    if (!dev) return NULL;

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) return NULL;
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Device %d partition type (if any) unrecognized\n", device);
	return NULL;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, 0);
	return NULL;
    }

    /* valid partition? */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        return NULL;
    }

    block = le32_to_int (pt->start4);
    if (dev->block_read(dev->dev, block, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, block);
	return NULL;
    }

    /* First block of partition contains aruba image header */
    ar = (aruba_image_t *) buffer;
    if (strcmp(ARUBA_IMAGE_MAGIC, ar->magic)) {
        printf("%d:%d Image not present\n", device, partition);
	return NULL;
    }
    if (!ar->valid) {
        printf("%d:%d Image not valid\n", device, partition);
	return NULL;
    }
    if (ar->machineType != ARUBA_IMAGE_CABERNET) {
        printf("%d:%d Image not for this platform\n", device, partition);
	return NULL;
    }

    /* remaining blocks contain actual image */
    addr = (u8 *) ((CFG_MEMTEST_END - ar->size) & 0xfff00000);
    count = (ar->size + dev->blksz - 1) / dev->blksz;
    /* read image into memory */
    printf("Loading image %d:%d", device, partition);
    if (dev->block_read(dev->dev, block + 1, count,
			(ulong *)addr) != count) {
        printf("%d:%d:%d read failed\n", device, partition, block+1);
	return NULL;
    }
    if (ctrlc())
        return NULL;

    if (ar->chksum) {
        /* verify checksum */
        printf("\nVerifying checksum... ");

	cksum = 0;
	ptr = (int *) buffer;
	for (i = 0; i < sizeof(buffer)/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	ptr = (int *) addr;
	for (i = 0; i < ar->size/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	if (cksum) {
	    /* invalid cksum */
	    printf("failed!\n");
	    return NULL;
	}

    }

    if (len) {
        *len = ar->size;
    }
#endif /* (CONFIG_COMMANDS & CFG_CMD_IDE) */

    return addr;
}

static void boot_partition(bd_t *bd, unsigned device, unsigned partition)
{
    u8 *addr;

    /* disable interrupts while accessing compact FLASH */
    disable_interrupts();
    addr = load_partition(bd, device, partition, NULL);
    enable_interrupts();
    if (addr) {
        printf("\nHalting active components...");
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE,
                           CAB_CPLD_RESET_ENABLE_VALUE);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET,
                           CAB_CPLD_RESET_CFLASH);
        udelay (10 * 1000);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET, 0);
        CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_RESET_ENABLE, 0);
        printf("\nBooting image...\n");
	/* set load addr */
	load_addr = (ulong)addr;
	/* boot it (DOES NOT RETURN if successful) */
	do_bootelf(NULL, bd, 0, 0, NULL);
    }
}

int do_cabernet_bootf(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    unsigned partition, part;
    unsigned device, dev;
    int search;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        search = 0;
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);
	} else {
	    device = 0;
	    partition = simple_strtoul(argv[1], NULL, 16);
	}
    } else {
        device = *((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET);
	partition = device & 0xf;
	device >>= 4;
	if ((device >= CFG_IDE_MAXDEVICE) ||
	    (partition >= DOS_PARTITIONS)) {
	    device = 0;
	    partition = 0;
	}
        search = 1;
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("Partition %x invalid\n", partition);
	return 1;
    }

    if (!search) {
        boot_partition(bd, device, partition);
	printf("Failed\n");
	return 1;
    }

    /* try to boot primary partition */
    boot_partition(bd, device, partition);

    /* failed, try to boot from any partition */
    for (dev = 0; dev < CFG_IDE_MAXDEVICE; dev++) {
        for (part = 0; part < DOS_PARTITIONS; part++) {
	    if ((part != partition) || (dev != device)) {
	        boot_partition(bd, dev, part);
	    }
	}
    }

    return 1;
}

#define JTAG_WRITE(Val) { \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val)); \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val) | CAB_CPLD_JTAG_OPERATION_CLK); \
}

#define JTAG_WRITE_READ(Val) { \
    int Tmp; \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val)); \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val) | CAB_CPLD_JTAG_OPERATION_CLK); \
    CAB_CPLD_REG_READ(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		     Tmp); \
    printf("%d", (Tmp & CAB_CPLD_JTAG_OPERATION_TDO)? 1:0); \
}

#define JTAG_READ(Val, Rc) { \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val) | CAB_CPLD_JTAG_OPERATION_CLK); \
    CAB_CPLD_REG_READ(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		     (Rc)); \
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_JTAG_OPERATION, \
		      (Val)); \
}

static int fpga_erase(void)
{
    int i;
    ulong status;

    /* Erase FPGA */
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_FPGA_INTERFACE,
		       CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_FPGA_INTERFACE,
		       CAB_CPLD_FPGA_INTERFACE_PRO|
		       CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    udelay(1); /* wait at least one us */
    CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_FPGA_INTERFACE,
		       CAB_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    i = 0;
    while (1) {
        udelay(1); /* wait at least one us */
	CAB_CPLD_REG_READ(CAB_CPLD_BASE_ADDR, CAB_CPLD_FPGA_INTERFACE,
			  status);
	if (status & CAB_CPLD_FPGA_INTERFACE_INIT_L) break;
	if (i++ > 100000) {
	    printf("Erase failed\n");
	    return 1;
	}
    }

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 10; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Enter Run-Test-Idle */
    JTAG_WRITE(0);

    /* 3. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 4. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    /* Load bypass to other FPGA */
    for (i = 0; i < 5; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
    }
    /* 5. Load CFG-IN */
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
    JTAG_WRITE(0);
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
    JTAG_WRITE(0);
    /* 6. Load last bit of CFG-IN */
    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    /* 7. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
    }
    /* 8. Enter SHIFT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }

    return 0;
}

int fpga_tftp_write(ulong offset, u8 *src, unsigned len)
{
    int bit, ch;
    u8 val;

    if (offset == 0) {
        if (fpga_erase()) {
	    return ERR_INVAL;
	}
    }

    for (ch = 0; ch < len-1; ch++, src++) {
        for (bit = 7; bit >= 0; bit--) {
	    val = ((*src >> bit) & 0x1) << 1;
	    JTAG_WRITE(val);
	}
    }
    /* last byte */
    for (bit = 7; bit > 0; bit--) {
        val = ((*src >> bit) & 0x1) << 1;
	JTAG_WRITE(val);
    }
    /* last bit */
    val = (*src & 0x1) << 1;
    if (len != 512) {
        printf("*");
	val |= CAB_CPLD_JTAG_OPERATION_TMS;
    }
    JTAG_WRITE(val);

    return ERR_OK;
}

#include "fpga.c"

static int fpga_flash_write(bd_t *bd)
{
    if (fpga_image) {
        printf("\nWriting image...");
        /* write data to FPGA */
        fpga_tftp_write(0, fpga_image, sizeof(fpga_image));
	printf("\n");

	/* success */
	return 0;
    }

    return 1;
}

int do_cabernet_fpga(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    int rc = 0;
    int i;

    if ((argc < 2) || (argc > 3)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'p') {
        char *filename;

	filename = argv[argc-1];

	if (argc < 3) {
	    /* erase FPGA and start initial programming */
	    fpga_erase();
	    /* Program FPGA from boot flash */
	    if (fpga_flash_write(bd)) {
	        return 1;
	    }
	} else {
	    /* TFTP data into FPGA */
	    load_addr = CFG_FPGA_BASE_ADDR;
	    copy_filename (BootFile, filename, sizeof(BootFile));
	    eth_set("sibyte#0");
	    if (NetLoop(bd, TFTP) == 0) {
	        return 1;
	    }
	}

	/* 11. Enter UPDATE-DR */
	JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	/* 12. Enter SELECT-IR */
	for (i = 0; i < 2; i++) {
	   JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	}
	/* 13. Enter SHIFT-IR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(0);
	}
	/* Load bypass to other FPGA */
	for (i = 0; i < 5; i++) {
	    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
	}
	/* 14. Load JSTART */
	JTAG_WRITE(0);
	JTAG_WRITE(0);
	JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TDI);
	/* 15. Load last bit of JSTART */
	JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	/* 16. Enter SELECT-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	}
	/* 17. Enter SHIFT-DR and clock STARTUP */
	for (i = 0; i < 20; i++) {
	    JTAG_WRITE(0);
	}
	/* 18. Enter UPDATE-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	}
	/* Enter Run-Test-Idle */
	JTAG_WRITE(0);

	/* Enter Test-Logic-Reset */
	for (i = 0; i < 10; i++) {
	    JTAG_WRITE(CAB_CPLD_JTAG_OPERATION_TMS);
	}

	/* disable JTAG programming */
	CAB_CPLD_REG_WRITE(CAB_CPLD_BASE_ADDR, CAB_CPLD_FPGA_INTERFACE, 0);

    } else {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    return (rc);
}
