#include <ppcboot.h>
#include <command.h>
#include <net.h>
#include <watchdog.h>
#include "cmd_mswitch.h"
#include "cmd_elf.h"
#include "cmd_boot.h"
#include "ns16550.h"
#include "../disk/part_dos.h"
#include "mswitch.h"

#include "../../../cmn/util/scci.h"
#include "../../../cmn/util/lcci.h"
#include "../../../cmn/util/cxe.h"
#include "../../../cmn/util/gbic.h"
#include "../../../cmn/util/sc28.h"
#include "../../../cmn/util/pd7124.h"
#include "../../../cmn/util/sc_cpld.h"
#include "../../../cmn/util/aruba_image.h"

#define TIMEOUT	       2		/* Seconds before failing */
#define PKTSIZEMIN     64
#define PHY_TIMEOUT    10000
#define PHY_REGS       32
#define PHY_DEVS       28

#define CHAR(c)       (((c >= 0x20) && (c <= 0x7e))? (char)c: '.')

extern struct NS16550 *COM_PORTS[];
static u_long loopback_length;
static u_long loopback_count;
static u_long loopback_incr;
static u_long loopback_received;
static int loopback_rcv;
void LoopbackRequest(void);
static unsigned long adjustSectors(unsigned long lba512);

typedef struct {
  u16 pvid;
  u8  mac;
  u8  phy;
} port_map_t;

port_map_t port_map[CXE_PORTS] = {
    { 0x01, 0x0b, 0x0d }, { 0x02, 0x1b, 0x0e },
    { 0x03, 0x0a, 0x0b }, { 0x04, 0x1a, 0x0c },
    { 0x05, 0x09, 0x09 }, { 0x06, 0x19, 0x0a },
    { 0x07, 0x08, 0x07 }, { 0x08, 0x18, 0x08 },
    { 0x09, 0x07, 0x15 }, { 0x0a, 0x17, 0x16 },
    { 0x0b, 0x06, 0x13 }, { 0x0c, 0x16, 0x14 },
    { 0x0d, 0x05, 0x11 }, { 0x0e, 0x15, 0x12 },
    { 0x0f, 0x04, 0x0f }, { 0x10, 0x14, 0x10 },
    { 0x11, 0x03, 0x1d }, { 0x12, 0x13, 0x1e },
    { 0x13, 0x02, 0x1b }, { 0x14, 0x12, 0x1c },
    { 0x15, 0x01, 0x19 }, { 0x16, 0x11, 0x1a },
    { 0x17, 0x00, 0x17 }, { 0x18, 0x10, 0x18 },
    { 0x19, 0x0e, 0x02 }, { 0x1a, 0x0f, 0x03 },
    { 0x02, 0x0d, 0x01 }, { 0x01, 0x0c, 0x00 }
};

#define CXE_MAC(port) ((0xb - ((port) >> 1)) | (((port) & 0x1) << 4))

char *cxe_stats_str[] = {
  "TxUcast",
  "TxMcast",
  "TxBcast",
  "TxDeferred",
  "TxColisions",
  "Tx1Colision",
  "TxNColisions",
  "TxXColisions",
  "TxLColisions",
  "TxLen64",
  "TxLen65",
  "TxLen128",
  "TxLen256",
  "TxLen512",
  "TxLen1024",
  "TxLen1519",
  "TxPause",
  "TxOctets",
  "TxOctetsBad",
  "RxOctets",
  "RxOctetsBad",
  "RxRunt",
  "RxShort",
  "RxShortCrc",
  "RxUcast",
  "RxMcast",
  "RxBcast",
  "RxCrc",
  "RxAlignment",
  "RxLong",
  "RxLongCrc",
  "RxLen64",
  "RxLen65",
  "RxLen128",
  "RxLen256",
  "RxLen512",
  "RxLen1024",
  "RxLen1519",
  "RxPause",
};

char *cxe_qe_discard_str[] = {
  "InqOverflow",
  "FifoOverflow",
  "BufMemOverflow",
  "RliDiscard",
  "RleDiscard",
  "McLimitDiscard",
  "TtlScoping",
  "WFHBD",
  "MacError",
};

char *cxe_qe_rl_str[] = {
  "Port 0",
  "Port 1",
  "Port 2",
  "Port 3",
  "Port 4",
  "Port 5",
  "Port 6",
  "Port 7",
  "Port 8",
  "Port 9",
  "Port a",
  "Port b",
  "Port c",
  "Port d",
  "Port e",
  "Port f",
  "Port 10",
  "Port 11",
  "Port 12",
  "Port 13",
  "Port 14",
  "Port 15",
  "Port 16",
  "Port 17",
  "Port 18",
  "Port 19",
  "Port 1a",
  "Port 1b",
  "CPU port",
  "Multicast",
  NULL, NULL,
  "FFA",
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  "Link",
};

static ulong lcci_write(ulong lc, ulong reg, ulong val)
{
    ulong error;
    LCCI_REG_WRITE(SCCI_BASE_ADDR, lc, reg, val, error);
    return error;
}

static ulong lcci_read(ulong lc, ulong reg, ulong *data)
{
    ulong error, val;
    LCCI_REG_READ(SCCI_BASE_ADDR, lc, reg, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong lcci_write_spoe(ulong lc, ulong dev, ulong reg, ulong val)
{
    ulong error;
    LCCI_SPOE_REG_WRITE(SCCI_BASE_ADDR, lc, dev, reg, val, error);
    return error;
}

static ulong lcci_read_spoe(ulong lc, ulong dev, ulong reg, ulong *data)
{
    ulong error, val;
    LCCI_SPOE_REG_READ(SCCI_BASE_ADDR, lc, dev, reg, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static void cxe_write(ulong lc, ulong addr, ulong data)
{
    ulong error;
    CXE_REG_WRITE(SCCI_BASE_ADDR, lc, 0, addr, data, error);
    if (error) {
        printf("cxe_write %lx,%04lx: failed %08lx\n",
	       lc, addr, error);
    }
}

static ulong cxe_read(ulong lc, ulong addr, ulong *data)
{
    ulong val, error;
    CXE_REG_READ(SCCI_BASE_ADDR, lc, 0, addr, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong cxe_write_phy(ulong lc, ulong device, ulong reg, ulong val)
{
    ulong tries, error;
    cxe_write(lc, CXE_GC_MDIO_ACCESS,
	      (CXE_GC_MDIO_ACCESS_WRITE |
	       (device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT) |
	       (val << CXE_GC_MDIO_ACCESS_DATA_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(lc, CXE_GC_MDIO_ACCESS, &val);
	if (error) {
	    return error;
	}
	if (val & CXE_GC_MDIO_ACCESS_DONE) {
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

static ulong cxe_read_phy(ulong lc, ulong device, ulong reg, ulong *val)
{
    ulong tries, error;
    cxe_write(lc, CXE_GC_MDIO_ACCESS,
	      ((device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(lc, CXE_GC_MDIO_ACCESS, val);
	if (error) {
	    break;
	}
	if (*val & CXE_GC_MDIO_ACCESS_DONE) {
	    *val &= 0xffff;
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

static void
LoopbackTimeout (void)
{
    puts ("Timeout expired\n");
    NetState = NETLOOP_FAIL;
}

static void
LoopbackHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
    int i;

    if (len != loopback_length) {
        printf("Error: length mismatch %d != %ld\n",
	       len, loopback_length);
	goto error;
    }
    for (i = 0; i < len; i++, pkt++) {
        if (*pkt != NetTxPacket[i]) {
	    printf("Error: byte %d mismatch: %02x != %02x\n",
		   i, *pkt, NetTxPacket[i]);
	    goto error;
	}
    }
    loopback_received++;
    
    if (loopback_incr) {
        if (loopback_length++ >= 1514) {
	    loopback_length = PKTSIZEMIN;
	}
    }

    if (serial_tstc() || 
	(loopback_count && (loopback_count-- == 1))) {
        NetState = NETLOOP_SUCCESS;
    } else {
        NetSendPacket(NetTxPacket, loopback_length);
	NetSetTimeout (TIMEOUT * CFG_HZ, LoopbackTimeout);
    }
    return;

 error:
    NetState = NETLOOP_FAIL;
}

void LoopbackRequest(void)
{
    /* called when a packet is received during ether test */
    int i;
    volatile uchar *pkt;

    pkt = NetTxPacket;

    NetSetEther(pkt, NetBcastAddr, PROT_TAG);

    pkt += ETHER_HDR_SIZE;
    *((ushort *)pkt) = 0x0001; /* VLAN ID */
    pkt += sizeof(ushort);

    for (i = 0; i < PKTSIZE - ETHER_HDR_SIZE - sizeof(ushort); i++) {
        pkt[i] = i & 0xff;
    }

    if (!loopback_rcv) {
        while (loopback_count--) {
	    NetSendPacket(NetTxPacket, loopback_length);
	}
        NetState = NETLOOP_SUCCESS;
    } else {

        NetSetHandler(LoopbackHandler);
	NetSetTimeout(TIMEOUT * CFG_HZ, LoopbackTimeout);

	NetSendPacket(NetTxPacket, loopback_length);
    }
}

int do_mswitch_ether(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong length = PKTSIZEMIN, count = 0, incr = 0;

    if ((argc < 2) || (argc > 5)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (eth_set(argv[1])) {
        printf("unknown device %s\n", argv[1]);
	return 1;
    }

    if (argc > 2) {
        if (argv[2][0] == 'i') {
  	    incr = 1;
	} else {
	    length = simple_strtoul(argv[2], NULL, 16);
	    if ((length < PKTSIZEMIN) || (length > 1514)) {
	        printf("length is invalid: %ld\n", length);
		return 1;
	    }
	}
    }

    if (argc > 3) {
        count = simple_strtoul(argv[3], NULL, 16);
	if (count == 0) {
	    printf("count is invalid: %ld\n", count);
	    return 1;
	}
    }

    if (argc > 4) {
        /* do bother receiving */
        loopback_rcv = 0;
    } else {
        loopback_rcv = 1;
    }

    loopback_length = length;
    loopback_incr = incr;
    loopback_count = count;
    loopback_received = 0;

    NetLoop(bd, LOOPBACK);

    printf("%lx packets successfully received\n", loopback_received);

    return 0;
}

int do_mswitch_scci(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong reg, val;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 2) || (argc > 5)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {

      	SCCI_REG_WRITE(SCCI_BASE_ADDR, SCCI_LED_CONTROL,
		       (SCCI_LED_POWER_STATE_SET(SCCI_LED_STATE_SOLID) |
			SCCI_LED_POWER_COLOR_SET(SCCI_LED_COLOR_GREEN) |
			SCCI_LED_ACTIVE_STATE_SET(SCCI_LED_STATE_SOLID) |
			SCCI_LED_ACTIVE_COLOR_SET(SCCI_LED_COLOR_GREEN) |
			SCCI_LED_STAT_STATE_SET(SCCI_LED_STATE_SOLID) |
			SCCI_LED_STAT_COLOR_SET(SCCI_LED_COLOR_GREEN)));
      	SCCI_REG_WRITE(SCCI_BASE_ADDR, SCCI_SF_CONTROL, 0x1f2);
	SCCI_REG_WRITE(SCCI_BASE_ADDR, SCCI_SF_CONTROL, 0x072);

    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    if (argc < 4) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	    reg = simple_strtoul(argv[2], NULL, 16);
	    val = simple_strtoul(argv[3], NULL, 16);

	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    while (times--) {
	        SCCI_REG_WRITE(SCCI_BASE_ADDR, reg, val);
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    /* Print the lines.
	     */
	    while (times--) {
	        reg = simple_strtoul(argv[2], NULL, 16);
	        if (argc > 3) {
		    length = simple_strtoul(argv[3], NULL, 16);
		}
	        do {
		    SCCI_REG_READ(SCCI_BASE_ADDR, reg, val);
		    if (!times) {
		        printf("%04lx: %08lx\n", reg, val);
		    }
		    reg += 1;
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return (rc);	
}

int do_mswitch_lcci(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc, reg, val, error;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 3) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, val);
    if (!(val & SCCI_MODULE_LC_PRES(lc))) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (argv[1][0] == 'i') {

        /* read FPGA id before resetting line, see Jerry for details */
        static int fpga_id(int dev, int magic);
        (void) fpga_id(SC_CPLD_JTAG_SELECTION_LC0+lc, 0x55);

        /* Reset linecard */
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE,
			  SC_CPLD_RESET_ENABLE_VALUE);
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_REMOTE_RESET,
			  SC_CPLD_REMOTE_RESET_LC(lc));
	udelay(10000);
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_REMOTE_RESET, 0);

	/* activate LC accesses (mark this SC as active for this LC) */
	SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_VERSION, val);
	if (val > 0x0b) {
	    /* actives were inverted prior to rev 0x0c */
	    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_LC_CONTROL, val);
	    val |= SC_CPLD_LC_CONTROL_ACTIVE(lc);
	    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LC_CONTROL, val);
	}

        /* enable LC on SCCI */
        SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_SCB_CONTROL, val);
	val |= SCCI_SCBC_LC_ENA(lc);
	SCCI_REG_WRITE(SCCI_BASE_ADDR, SCCI_SCB_CONTROL, val);

	/* bring LM80 out of reset (yes bit is in i2c control reg) */
        lcci_write(lc, LCCI_I2C_CONTROL, 0);

	/* reset GBICs */
	lcci_write(lc, LCCI_GBIC_CONTROL, (LCCI_GBIC_CONTROL_GBIC_DIS(0) |
					   LCCI_GBIC_CONTROL_GBIC_DIS(1)));
	udelay(40);
	lcci_write(lc, LCCI_GBIC_CONTROL, 0);

	/* bring CXE and phys out of reset */
	lcci_write(lc, LCCI_CXE_CONTROL, 0);
	lcci_write(lc, LCCI_PHY_CONTROL, 0);
	lcci_write(lc, LCCI_SF_CONTROL, 0x1f2);
	lcci_write(lc, LCCI_SF_CONTROL, 0x072);

	/* clear LEDS */
	lcci_write(lc, LCCI_FE_LED_0, 0x88888888);
	lcci_write(lc, LCCI_FE_LED_1, 0x88888888);
	lcci_write(lc, LCCI_FE_LED_2, 0x88888888);

        /* configure serdes, enable PCS */
	cxe_read_phy(lc, 0x0, 0x11, &val); /* regs 0x10 and greater are global */
	cxe_write_phy(lc, 0x0, 0x11, val | 0x8004);

	/* reset POE circuitry */
	lcci_read(lc, LCCI_MODULE_STATUS, &val);
	if (val & LCCI_SPOE_PRES) {
	    /* bring UARTS out of reset */
	    lcci_write(lc, LCCI_SPOE_CONTROL, 0);
	    /* wait some time */
	    udelay(10000);
	    /* For POE, enable selection of GPIO 1, 0 based upon GPOR bits 1,0 */
	    lcci_write_spoe(lc, 3, 0x87, 0xEF);
	    /* reset POE controller */
	    lcci_write_spoe(lc, 3, 0x97, 0x03);
	    /* bring POE controller out of reset */
	    lcci_write_spoe(lc, 3, 0x97, 0x00);
	}

	/* set power and status LEDs to green */
	lcci_write(lc, LCCI_LED_CONTROL,
		   ((LCCI_LED_COLOR_GREEN << LCCI_LED_STAT_COLOR_SHIFT) |
		    (LCCI_LED_STATE_SOLID << LCCI_LED_STAT_STATE_SHIFT) |
		    (LCCI_LED_COLOR_GREEN << LCCI_LED_POWER_COLOR_SHIFT) |
		    (LCCI_LED_STATE_SOLID << LCCI_LED_POWER_STATE_SHIFT)));

    } else if (argv[1][0] == 'w') {
        if (argc < 5) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
	if (argc > 5) {
	  times = simple_strtoul(argv[5], NULL, 16);
	}

	reg = simple_strtoul(argv[3], NULL, 16);
	val = simple_strtoul(argv[4], NULL, 16);
	while (times--) {
	    error = lcci_write(lc, reg, val);
	    if (error) {
	        printf("%lx,%04lx: failed %08lx\n",
		       lc, reg, error);
		return 1;
	    }
	}

    } else {

        /* If another parameter, it is the length to display.
	 * Length is the number of objects, not number of bytes.
	 */
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	if (argc > 5) {
	    times = simple_strtoul(argv[5], NULL, 16);
	}

	while (times--) {
	    reg = simple_strtoul(argv[3], NULL, 16);
	    if (argc > 4) {
	        length = simple_strtoul(argv[4], NULL, 16);
	    }
	    /* Print the lines.
	     */
	    do {
	        error = lcci_read(lc, reg, &val);
		if (error) {
		    printf("%lx,%04lx: failed %08lx\n",
			   lc, reg, error);
		    return 1;
		} else if (!times) {
		    printf("%lx,%04lx: %08lx\n",
			   lc, reg, val);
		}
		reg++;
		if (ctrlc()) {
		    return 1;
		}
	    } while (--length > 0);
	}
    }

    return (rc);	
}

static void cxe_init(ulong lc, ulong mask, int fbm, int bridge)
{
    cxe_vlan_bc_t *vbc;
    cxe_u128_t u128;
    ulong stat, mac;
    ulong error;
    ulong val;
    ulong rev;
    int mtu = CXE_MTU;

    /* Port and limited protocol-based VLANS, no VRP (route) table, etc */
    cxe_write(lc, CXE_AL_CONFIG, 0x0001c061);

    if (!bridge) {
        /* disable L2 learning on ports */
        cxe_write(lc, CXE_DONT_LEARN_PORTS, 0x0fffffff);
    }

    /* clear interrupts */
    error = cxe_read(lc, CXE_GC_INT_STATUS, &val);
    if (error) {
        printf("cxe_read %lx,%02x: failed %08lx\n",
	       lc, CXE_GC_INT_STATUS, error);
	return;
    }

    /* Set LED stretch to 64 ms */
    cxe_write(lc, CXE_LED_DISP_MODE, 0x00000005);
    /* Set LED event selection to 0:RX act, 1:TX act, 2:link status */
    for (mac = 0; mac < CXE_PORTS; mac++) {
        cxe_write(lc, CXE_LED_EVENT_SEL(mac), 0x00000345);
    }

    cxe_read(lc, CXE_GC_INT_STATUS, &rev);
    if (CXE_CG_DEVICE_ID_REV(rev) == CXE_CG_DEVICE_ID_REV_G) {
        /* Queue Engine XXX ER_9_1*/
        cxe_write(lc, CXE_QE_CONFIG, 0x10);

        /* Queue Engine Resource Limit configuration */
        val = CXE_INT_CELLS;
        stat = ((mtu + (CXE_CELL_SIZE-1))/CXE_CELL_SIZE) * 2; /* 2 * mtu size */
        cxe_write(lc, CXE_QE_RL_DATA, stat);
        for (mac = 0; mac < CXE_PORTS; mac++) {
            cxe_write(lc, CXE_QE_RL_INSTR, CXE_QE_RL_INSTR_WRITE | mac);
            val -= stat;
        }
        /* reserve same for mcast, even though there should be no mcast */
        cxe_write(lc, CXE_QE_RL_DATA, stat);
        cxe_write(lc, CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
                                        CXE_QE_RL_INSTR_ID_MCAST));
        val -= stat;
        /* rest is Free For All (FFA) */
        cxe_write(lc, CXE_QE_RL_DATA, val);
        for (mac = 0; mac < CXE_PRIOS; mac++) {
            cxe_write(lc, CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
                                            CXE_QE_RL_INSTR_ID_FFA_PRIO(mac)));
        }
        cxe_write(lc, CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
                                        CXE_QE_RL_INSTR_ID_FFA_MCAST));
    } else {
        cxe_write(lc, CXE_GC_DLL_CTRL, 0x0007c00c);
    }
    

    /* MAC configuration */
    for (mac = 0; mac < CXE_PORTS; mac++) {
        if (mac < 16) {
	    /* 0-11 configure two FE ports (don't ask) */
	    if (CXE_FE_PORT(mac)) {
	        /* PAUSE, IPG, RMII, 100FD */
	        cxe_write(lc, CXE_MAC_PORT_CONFIG(mac), 0x000821f2);
	    } else {
	        /* PAUSE, IPG, FD, GMII */
	        cxe_write(lc, CXE_MAC_PORT_CONFIG(mac), 0x00082144);
	    }
	}
	/* Set max pkt len */
	if (CXE_INT_PORT(mac)) {
	    cxe_write(lc, CXE_MAC_MAX_PKT_LEN(mac), mtu+4); /* for Double .Q */
	} else {
	    cxe_write(lc, CXE_MAC_MAX_PKT_LEN(mac), mtu);
	}
    }

    /* Clear TCAM */
    cxe_write(lc, CXE_INT_AL_CAM_0, 0);
    cxe_write(lc, CXE_INT_AL_CAM_1, 0);
    cxe_write(lc, CXE_INT_AL_CAM_2, 0);
    cxe_write(lc, CXE_INT_AL_CAM_3, 0);
    for (mac = 0; mac < CXE_INT_AL_CAM_ENTRIES; mac++) {
        cxe_write(lc, CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
		   mac));
	cxe_write(lc, CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
		   mac));
    }

    /* VLAN configuration */
    vbc = (cxe_vlan_bc_t *) &u128;
    memset(vbc, 0, sizeof(*vbc));
    vbc->TYPE = 1;
    vbc->PERMANENT = 1;
    vbc->VALID = 1;
    if (bridge || fbm) {
        /* Add all ports to VLAN 1 */
        ulong pvid = 1;
        for (mac = 0; mac < CXE_PORTS; mac++) {
	    /* Set PVID of each port to port num + 1 */
	    cxe_write(lc, CXE_PORT_VID_1(mac), pvid<<16|pvid);
	    cxe_write(lc, CXE_PORT_VID_2(mac), pvid<<16|pvid);
	    if (CXE_EXT_PORT(mac)) {
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_2 |= (0x1 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 |= (0x1 << mac);
		}
	    }
	}
	vbc->VID = pvid;
	vbc->HIT_COUNTER = mac;
	if (!fbm) {
	    if (lc == 0) {
	        vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_0);
	    } else {
	        vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_1);
	    }
	}
	cxe_write(lc, CXE_INT_AL_CAM_0, u128.ulBits0_31);
	cxe_write(lc, CXE_INT_AL_CAM_1, u128.ulBits32_63);
	cxe_write(lc, CXE_INT_AL_CAM_2, u128.ulBits64_95);
	cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
	cxe_write(lc, CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
		   mac));

	vbc->VID = 0xfff;
	cxe_write(lc, CXE_INT_AL_CAM_0, 0);
	cxe_write(lc, CXE_INT_AL_CAM_1, 0x80000000);
	cxe_write(lc, CXE_INT_AL_CAM_2, 0);
	cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
	cxe_write(lc, CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
		   mac));
    }
    if (fbm) {
        for (mac = 0; mac < CXE_PORTS; mac += 2) {
	    /* Set PVID of each port to port num + 1 */
	    ulong pvid = CXE_PVID(mac) + 1;
	    if (CXE_EXT_PORT(mac)) {
	        /* loop frames to peer port */
	        cxe_write(lc, CXE_PORT_VID_1(mac), pvid<<16|pvid);
		cxe_write(lc, CXE_PORT_VID_2(mac), pvid<<16|pvid);
	        cxe_write(lc, CXE_PORT_VID_1(mac+1), pvid<<16|pvid);
		cxe_write(lc, CXE_PORT_VID_2(mac+1), pvid<<16|pvid);
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_1 = 0;
		    vbc->OUTPORTS_MASK_PART_2 = (0x3 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 = (0x3 << mac);
		    vbc->OUTPORTS_MASK_PART_2 = 0;
		}
		vbc->VID = pvid;
		vbc->HIT_COUNTER = mac;

		cxe_write(lc, CXE_INT_AL_CAM_0, u128.ulBits0_31);
		cxe_write(lc, CXE_INT_AL_CAM_1, u128.ulBits32_63);
		cxe_write(lc, CXE_INT_AL_CAM_2, u128.ulBits64_95);
		cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(lc, CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
			   mac));

		vbc->VID = 0xfff;
		cxe_write(lc, CXE_INT_AL_CAM_0, 0);
		cxe_write(lc, CXE_INT_AL_CAM_1, 0x80000000);
		cxe_write(lc, CXE_INT_AL_CAM_2, 0);
		cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(lc, CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
			   mac));
	    }
	}
    }
    if (!bridge && !fbm) {
        for (mac = 0; mac < CXE_PORTS; mac++) {
	    ulong pvid = CXE_PVID(mac);
	    /* Set PVID of each port to port num + 1 */
	    cxe_write(lc, CXE_PORT_VID_1(mac), pvid<<16|pvid);
	    cxe_write(lc, CXE_PORT_VID_2(mac), pvid<<16|pvid);
	    if (CXE_EXT_PORT(mac)) {
	        /* Add VLAN entry that includes internal port(s) */
	        vbc->VID = pvid;
		vbc->HIT_COUNTER = pvid;
		if (mask) {
		    vbc->TAGGED_OUTPORTS_MASK_PART_2 = 0;
		    vbc->OUTPORTS_MASK_PART_1 = 0;
		    if (mask & 0x1) {
		        vbc->TAGGED_OUTPORTS_MASK_PART_2 |= (0x1 << (CXE_INT_PORT_0-4));
			vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_0);
		    }
		    if (mask & 0x2) {
		        vbc->TAGGED_OUTPORTS_MASK_PART_2 |= (0x1 << (CXE_INT_PORT_1-4));
			vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_1);
		    }
		} else if (lc == 0) {
		    vbc->TAGGED_OUTPORTS_MASK_PART_2 = (0x1 << (CXE_INT_PORT_0-4));
		    vbc->OUTPORTS_MASK_PART_1 = (0x1 << CXE_INT_PORT_0);
		} else {
		    vbc->TAGGED_OUTPORTS_MASK_PART_2 = (0x1 << (CXE_INT_PORT_1-4));
		    vbc->OUTPORTS_MASK_PART_1 = (0x1 << CXE_INT_PORT_1);
		}
		/* add external port */
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_2 = (0x1 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 |= (0x1 << mac);
		    vbc->OUTPORTS_MASK_PART_2 = 0;
		}
		cxe_write(lc, CXE_INT_AL_CAM_0, u128.ulBits0_31);
		cxe_write(lc, CXE_INT_AL_CAM_1, u128.ulBits32_63);
		cxe_write(lc, CXE_INT_AL_CAM_2, u128.ulBits64_95);
		cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(lc, CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
			   mac));

		vbc->VID = 0xfff;
		cxe_write(lc, CXE_INT_AL_CAM_0, 0);
		cxe_write(lc, CXE_INT_AL_CAM_1, 0x80000000);
		cxe_write(lc, CXE_INT_AL_CAM_2, 0);
		cxe_write(lc, CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(lc, CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
			   mac));
	    }
	}
    }

    for (mac = 0; mac < CXE_PORTS; mac++) {
        if (CXE_EXT_PORT(mac)) {
	    /* Enable DoubleQ Tagging on external ports */
	    cxe_write(lc, CXE_PD_CONFIG(mac), 0x00000110);
	}

	/* clear MAC stats */
	for (stat = 0; stat < CXE_MAC_STATS; stat++) {
	    cxe_write(lc, CXE_MAC_STAT_INSTR(mac), stat);
	}
    }

    /* clear Address Lookup stats */
    for (stat = CXE_CNT_VLAN_DOESNT_EXIST;
	 stat <= CXE_CNT_IPX_FORW_DATAGRAMS; stat++) {
        cxe_read(lc, stat, &val);
    }

    /* clear Queue Engine stats */
    for (stat = 0; stat < CXE_QE_DISCARD_STATS; stat++) {
        cxe_write(lc, CXE_QE_DISCARD_INSTR, stat);
    }

    /* restart Queue Engine */
    cxe_write(lc, CXE_QE_RESTART_COMMAND, 1);
}

int do_mswitch_cxe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc, bank, reg, val, error, port, stat, mac;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 3) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, val);
    if (!(val & SCCI_MODULE_LC_PRES(lc))) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (argv[1][0] == 'i') {
        ulong mask = 0;
	int fbm = 0;
	int bridge = 0;
        if (argc > 3) {
	    if (argv[3][0] == 'f') {
	        /* fast loopback test */
	        fbm = 1;
	    } else if (argv[3][0] == 'b') {
	        /* init simple bridge */
	        bridge = 1;
	    } else {
	        mask = simple_strtoul(argv[3], NULL, 16);
		if (mask > 3) {
		    printf("%lx: mask invalid\n", mask);
		    return 1;
		}
	    }
	}

	/* initialize CXE state */
	cxe_init(lc, mask, fbm, bridge);

    } else if (argv[1][0] == 't') {
	for (mac = 0; mac < CXE_PORTS+1; mac++) {
	    cxe_write(lc, CXE_INT_AL_CAM_INSTR,
		      ((CXEPHYS_CAM_ACTION_READ << 20) |
		       mac));
	    cxe_read(lc, CXE_INT_AL_CAM_3, &val);
	    printf("%02lx: %08lx", mac, val);
	    cxe_read(lc, CXE_INT_AL_CAM_2, &val);
	    printf(" %08lx", val);
	    cxe_read(lc, CXE_INT_AL_CAM_1, &val);
	    printf(" %08lx", val);
	    cxe_read(lc, CXE_INT_AL_CAM_0, &val);
	    printf(" %08lx\n", val);
	}
    } else if (argv[1][0] == 's') {
        if (argc < 4) {
	    for (port = 0; port < CXE_PORTS; port++) {
	        mac = port_map[port].mac;
	        bank = 0;
		for (stat = 0; stat < CXE_MAC_STATS; stat++) {
		    cxe_write(lc, CXE_MAC_STAT_INSTR(mac), stat);
		    error = cxe_read(lc, CXE_MAC_STAT_DATA_L(mac), &val);
		    if (error) {
		        printf("cxe_read %lx,%04lx: failed %08lx\n",
			       lc, CXE_MAC_STAT_DATA_L(mac), error);
			return 1;
		    } 
		    if (val) {
		        if (!bank) printf("port %lx:\n", port);
		        printf("  %-15s %08lx", cxe_stats_str[stat], val);
			if (bank++ & 0x1) printf("\n");
		    }
		}
		if (bank & 0x1) {
		    printf("\n");
		}
	    }
	} else {
	    port = simple_strtoul(argv[3], NULL, 16);
	    if (port >= CXE_PORTS) {
	        printf ("Illegal port %lx\n", port);
		return 1;
	    }
	    mac = port_map[port].mac;

	    bank = 0;
	    for (stat = 0; stat < CXE_MAC_STATS; stat++) {
	        cxe_write(lc, CXE_MAC_STAT_INSTR(mac), stat);
		error = cxe_read(lc, CXE_MAC_STAT_DATA_L(mac), &val);
		if (error) {
		    printf("cxe_read %lx,%04lx: failed %08lx\n",
			   lc, CXE_MAC_STAT_DATA_L(mac), error);
		    return 1;
		} 
		if (val) {
		    printf("  %-15s %08lx", cxe_stats_str[stat], val);
		    if (bank++ & 0x1) printf("\n");
		}
	    }
	    printf("\n");
	}
      
    } else if (argv[1][0] == 'q') {

        printf("Discard stats:\n");
        for (stat = 0; stat < CXE_QE_DISCARD_STATS; stat++) {
	    cxe_write(lc, CXE_QE_DISCARD_INSTR, stat);
	    error = cxe_read(lc, CXE_QE_DISCARD_DATA, &val);
	    if (error) {
	        printf("cxe_read %lx,%04x: failed %08lx\n",
		       lc, CXE_QE_DISCARD_DATA, error);
		return 1;
	    } 
	    if (val) {
		printf("  %-15s %08lx\n", cxe_qe_discard_str[stat], val);
	    }
	}
        printf("Resource Limit queue lengths:\n");
	for (stat = 0; stat < CXE_QE_RL_INSTR_ID_LINK_MCAST; stat++) {
	    if (cxe_qe_rl_str[stat] == NULL) continue;
	    cxe_write(lc, CXE_QE_RL_INSTR, CXE_QE_RL_INSTR_COUNT | stat);
	    error = cxe_read(lc, CXE_QE_RL_DATA, &val);
	    if (error) {
	        printf("cxe_read %lx,%04x: failed %08lx\n",
		       lc, CXE_QE_RL_DATA, error);
		return 1;
	    } 
	    if (val) {
		printf("  %-15s %08lx\n", cxe_qe_rl_str[stat], val);
	    }
	}

    } else {
        if (argc < 5) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        bank = simple_strtoul(argv[3], NULL, 16);
	if (bank > CXE_BANKS) {
	    printf("%lx: bank invalid\n", bank);
	    return 1;
	}


	if (argv[1][0] == 'w') {
	    if (argc < 6) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }
	    reg = simple_strtoul(argv[4], NULL, 16);
	    val = simple_strtoul(argv[5], NULL, 16);

	    while (times--) {
	        CXE_REG_WRITE(SCCI_BASE_ADDR, lc, bank, reg, val, error);
		if (error) {
		    printf("%lx,%lx,%02lx: failed %08lx\n",
			   lc, bank, reg, error);
		    return 1;
		}
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }

	    while (times--) {
	        /* Print the lines.
		 */
	        reg = simple_strtoul(argv[4], NULL, 16);
		if (argc > 5) {
	            length = simple_strtoul(argv[5], NULL, 16);
		}
		do {
	            CXE_REG_READ(SCCI_BASE_ADDR, lc, bank, reg, val, error);
		    if (error) {
		        printf("%lx,%lx,%02lx: failed %08lx\n",
			       lc, bank, reg, error);
			return 1;
		    } else if (!times) {
		        printf("%lx,%lx,%02lx: %08lx\n",
			       lc, bank, reg, val);
		    }
		    reg += 1;
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return (rc);	
}

int do_mswitch_li2c(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc, device, reg, val, error;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 5) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, val);
    if (!(val & SCCI_MODULE_LC_PRES(lc))) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    device = simple_strtoul(argv[3], NULL, 16);
    if (device >= LCCI_I2C_DEVICES) {
        printf("%lx: device invalid\n", device);
	return 1;
    }

    reg = simple_strtoul(argv[4], NULL, 16);
    if (reg >= LCCI_I2C_REGS) {
        printf("%02lx: register invalid\n", reg);
	return 1;
    }

    if (argv[1][0] == 'w') {
        if (argc != 6) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	val = simple_strtoul(argv[5], NULL, 16);
	if (val >= 256) {
	    printf("%02lx: value invalid\n", val);
	    return 1;
	}
	if (argc > 6) {
	    times = simple_strtoul(argv[6], NULL, 16);
	}

	while (times--) {
	    LCCI_I2C_REG_WRITE(SCCI_BASE_ADDR, lc, device, reg, val, error);
	    if (error) {
	        printf("%lx,%lx,%04lx: failed %08lx\n",
		       lc, device, reg, error);
		return 1;
	    }
	}
    } else {
        /* If another parameter, it is the length to display.
	 * Length is the number of objects, not number of bytes.
	 */
        if (argc > 6) {
	    times = simple_strtoul(argv[6], NULL, 16);
	}
	while (times--) {
	    reg = simple_strtoul(argv[4], NULL, 16);
	    if (argc > 5) {
	        length = simple_strtoul(argv[5], NULL, 16);
	    }
    
	    /* Print the lines.
	     */
	    do {
	        LCCI_I2C_REG_READ(SCCI_BASE_ADDR, lc, device, reg, val, error);
		if (error) {
		    printf("%lx,%lx,%04lx: failed %08lx\n",
			   lc, device, reg, error);
		    return 1;
		} else if (!times) {
		    printf("%lx,%lx,%04lx: %02lx  %c\n",
			   lc, device, reg, val, CHAR(val));
		}
		reg++;
		if (reg >= LCCI_I2C_REGS) {
		    break;
		}
		if (ctrlc()) {
		    return 1;
		}
	    } while (--length > 0);
	}
    }

    return (rc);	
}

static void
uart_putc(int uart, const char c)
{
    if (c == '\n') {
        NS16550_putc(COM_PORTS[uart], '\r');
    }
    NS16550_putc(COM_PORTS[uart], c);
}

static void
uart_puts (int uart, const char *s)
{
    while (*s) {
        uart_putc (uart, *s++);
    }
}

static int
uart_getc(int uart)
{
    return NS16550_getc(COM_PORTS[uart]);
}

static int
uart_tstc(int uart)
{
    return NS16550_tstc(COM_PORTS[uart]);
}

static int
uart_space(int uart)
{
    return NS16550_space(COM_PORTS[uart]);
}

static int luart_cr = 0;

static void
luart_init(int lc, int port, int speed, int loop, int ports, int cr)
{
    ulong chan, chip;

    luart_cr = cr;

    chip = SC28_LC_CHIP(port);
    chan = SC28_LC_CHAN(port);

    lcci_write_spoe(lc, chip, SC28_GCCR, 0);
    lcci_write_spoe(lc, chip, SC28_MR0(chan), 0);
    lcci_write_spoe(lc, chip, SC28_MR1(chan), SC28_MR1_8N);
    lcci_write_spoe(lc, chip, SC28_MR2(chan), loop? SC28_MR2_LOCALLOOP:0);
    lcci_write_spoe(lc, chip, SC28_TxCSR(chan), speed);
    lcci_write_spoe(lc, chip, SC28_RxCSR(chan), speed);

    lcci_write_spoe(lc, chip, SC28_CR(chan), SC28_CR_CMD_RxRST);
    lcci_write_spoe(lc, chip, SC28_CR(chan), SC28_CR_CMD_TxRST);
    lcci_write_spoe(lc, chip, SC28_CR(chan), SC28_CR_RxEN|SC28_CR_TxEN);
}

static void 
soe_ssr_init (int lc) // initialize the ssrs
{
  lcci_write_spoe(lc, 0, 0x97, 0x0);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 1, 0x97, 0x2);// bit 1 is EN_L,  bit 0 is SHIFT_IN
  lcci_write_spoe(lc, 2, 0x97, 0x0);// bit 1 is NC,    bit 0 is CLR_L 
  // tell the UARTs to use 0x97 contents to output to the GOUT pins
  lcci_write_spoe(lc, 0, 0x87, 0xef);
  lcci_write_spoe(lc, 1, 0x87, 0xef);
  lcci_write_spoe(lc, 2, 0x87, 0xef);
  // send a few clocks first
  lcci_write_spoe(lc, 0, 0x97, 0x3);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 0, 0x97, 0x0);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 2, 0x97, 0x1);// bit 1 is NC,    bit 0 is CLR_L 
  lcci_write_spoe(lc, 0, 0x97, 0x3);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 0, 0x97, 0x0);// bit 1 is SCK,   bit 0 is RCK
  // soe_enable(lc,0x000000);// set all ports to SoE off
}

static void 
soe_enable (int lc, ulong z) { // bit 23 of z is port 23 etc...
  int high_port;
  int i;
  soe_ssr_init(lc);
  for (i = 0; i < 24; i++) {
    high_port = (z >>(23- i)) & 0x1;
    // printf("the value is 0x%lx and the 0x%lx\n",z,high_port);
    //    z = z >> 1;
    lcci_write_spoe(lc, 1, 0x97, 0x2 | high_port);
    // bit 1 is EN_L,  bit 0 is SHIFT_IN
    lcci_write_spoe(lc, 0, 0x97, 0x2);// bit 1 is SCK,   bit 0 is RCK
    lcci_write_spoe(lc, 0, 0x97, 0x0);// bit 1 is SCK,   bit 0 is RCK
  }
  lcci_write_spoe(lc, 0, 0x97, 0x1);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 0, 0x97, 0x0);// bit 1 is SCK,   bit 0 is RCK
  lcci_write_spoe(lc, 1, 0x97, 0x0);// bit 1 is EN_L,  bit 0 is SHIFT_IN

}


static void
luart_putc(int lc, int port, const char c)
{
    ulong data, chan, chip;

    if ((c == '\n') && luart_cr) {
        luart_putc(lc, port, '\r');
    }
    chip = SC28_LC_CHIP(port);
    chan = SC28_LC_CHAN(port);

    /* wait for TX to be ready for another character */
    do {
        if (lcci_read_spoe(lc, chip, SC28_SR(chan), &data)) break;
    } while (!(data & SC28_SR_TxRDY));

    /* write character */
    lcci_write_spoe(lc, chip, SC28_TxFIFO(chan), c);
}

static void
luart_puts (int lc, int port, const char *s)
{
    while (*s) {
        luart_putc (lc, port, *s++);
    }
}

static int
luart_getc(int lc, int port)
{
    ulong data, chan, chip;

    chip = SC28_LC_CHIP(port);
    chan = SC28_LC_CHAN(port);

    do {
        if (lcci_read_spoe(lc, chip, SC28_SR(chan), &data)) break;
    } while (!(data & SC28_SR_RxRDY));

    lcci_read_spoe(lc, chip, SC28_RxFIFO(chan), &data);

    return (int) data;
}

static int
luart_tstc(int lc, int port)
{
    ulong data, chan, chip;

    chip = SC28_LC_CHIP(port);
    chan = SC28_LC_CHAN(port);

    lcci_read_spoe(lc, chip, SC28_SR(chan), &data);

    return (data & SC28_SR_RxRDY)? 1: 0;
}

int do_mswitch_luart(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc;
    ulong status;
    ulong z;

    int times = 1;

    if (argc < 4) {
        printf ("Usage %d:\n%s\n", argc, cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    if (lcci_read(lc, LCCI_MODULE_STATUS, &status)) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (!(status & LCCI_SPOE_PRES)) {
        printf("%lx: SPOE card not present\n", lc);
	return 1;
    }
    
    if (argv[1][0] == 's') {
      z = simple_strtoul(argv[3],NULL,16);
      soe_enable(lc,z);
      return 1;
    }


    if (argv[1][0] == 'c') {
	int i, loop = 0, cr = 0;
        ulong iport;
        ulong eport;
	char c;

        eport = simple_strtoul(argv[3], NULL, 16);
	if (eport > 24) {
	    printf("%04lx: eport invalid\n", eport);
	    return 1;
	}

	if (argv[4][0] == 'l') {
	    loop = 1;
	    iport = eport;
        } else if (argv[4][0] == 'c') {
            cr = 1;
	    iport = eport;
	} else {
	    iport = simple_strtoul(argv[4], NULL, 16);
	    if (iport > 24) {
	        printf("%04lx: iport invalid\n", iport);
		return 1;
	    }
	}

	luart_init(lc, iport, SC28_CSR_9600, loop, 8, cr);
	if (eport != iport) {
	    luart_init(lc, eport, SC28_CSR_9600, loop, 8, cr);
	}
	if (argc > 5) {
	    for (i = 5; i < argc; i++) {
	        luart_puts(lc, eport, argv[i]);
		if (i < argc - 1) {
		    luart_putc(lc, eport, ' ');
		} else {
		    luart_putc(lc, eport, '\n');
		}
	    }
	}

	printf("Connecting UART A to LC UART %lx/%lx->%lx...\n"
	       "Press <Control-c|y> to stop\n",
	       lc, eport, iport);
	while (1) {
	    WATCHDOG_RESET();
	    if (uart_tstc(0)) {
	        c = uart_getc(0);
		if ((c == 0x03) || (c == 0x19)) break; /* control-c/y */
		if (c == '\r') c = '\n';
		luart_putc(lc, eport, c);
	    }
	    if (luart_tstc(lc, iport)) {
	        c = luart_getc(lc, iport);
		uart_putc(0, c);
	    }
	}
	printf("\n");

    } else {
        ulong device, reg, val, error;
	int length = 1;

	device = simple_strtoul(argv[3], NULL, 16);
	if (device >= LCCI_SPOE_DEVICES) {
	    printf("%lx: device invalid\n", device);
	    return 1;
	}

	reg = simple_strtoul(argv[4], NULL, 16);
	if (reg >= LCCI_SPOE_REGS) {
	    printf("%04lx: register invalid\n", reg);
	    return 1;
	}

	if (argv[1][0] == 'w') {

	    if (argc < 6) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }

	    val = simple_strtoul(argv[5], NULL, 16);
	    if (val >= 256) {
	        printf("%02lx: value invalid\n", val);
		return 1;
	    }

	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }

	    while (times--) {
	        error = lcci_write_spoe(lc, device, reg, val);
		if (error) {
		    printf("%lx,%lx,%02lx: failed %08lx\n",
			   lc, device, reg, error);
		    return 1;
		}
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }
	    
	    while (times--) {
	        reg = simple_strtoul(argv[4], NULL, 16);
	        if (argc > 5) {
		    length = simple_strtoul(argv[5], NULL, 16);
		}
    
		/* Print the lines.
		 */
		do {
		    error = lcci_read_spoe(lc, device, reg, &val);
		    if (error) {
		        printf("%lx,%lx,%02lx: failed %08lx\n",
			       lc, device, reg, error);
			return 1;
		    } else if (!times) {
		        printf("%lx,%lx,%02lx: %02lx  %c\n",
			       lc, device, reg, val, CHAR(val));
		    }
		    reg += 1;
		    if (reg >= LCCI_SPOE_REGS) {
		        break;
		    }
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return 0;
}

static int poe_recv(int lc, char *buffer, int len, int timeout)
{
    /* wait for response */
    char *s = (char *) buffer;
    int i, tries;

    for (i = 0; i < len; i++, s++) {
        for (tries = 0; tries < timeout; tries++) {
	    if (luart_tstc(lc, PD7124_LC_PORT)) {
	        *s = luart_getc(lc, PD7124_LC_PORT);
		break;
	    }
	}
	if (tries >= timeout) {
	    break;
	}
	if (ctrlc()) {
	    break;
	}
    }
    return i;
}

#define POE_TIMEOUT 100000

static void poe_send(int lc, char *buffer, int len)
{
    char *s;
    int i;

    /* send message */
    luart_init(lc, PD7124_LC_PORT, SC28_CSR_19200, 0, 4, 0);
    s = buffer;
    for (i = 0; i < len; i++, s++) {
        luart_putc(lc, PD7124_LC_PORT, *s);
    }
}

int do_mswitch_poe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    static unsigned char seq = 0;
    pd7124_msg_t msg;
    ulong status;
    ulong lc;
    char *s;
    int i;
    
    if (argc < 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    if (lcci_read(lc, LCCI_MODULE_STATUS, &status)) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (!(status & LCCI_SPOE_PRES)) {
        printf("%lx: SPOE card not present\n", lc);
	return 1;
    }

    if (argv[1][0] == 's') {
        /* send message */
        s = (char *) &msg;
	/* construct message */
	for (i = 0; i < argc-3; i++, s++) {
	    *s = (__u8) simple_strtoul(argv[i+3], NULL, 16);
	    if (i == 0) {
	        /* skip checksum */
	        s++;
		*s = seq;
		seq = PD7124_SEQNEXT(seq);
	    }
	}
	for (; i < PD7124_MSG_LEN-2; i++, s++) {
	    *s = PD7124_SPACE; /* 'N' */
	}
	PD7124_CKSUM(&msg); /* checksum */

	/* display message */
	printf("send: ");
	s = (char *) &msg;
	for (i = 0; i < PD7124_MSG_LEN; i++, s++) {
	    printf("%02x ", *s);
	}
	printf("\n");

	/* send message and wait for reply */
	poe_send(lc, (char *)&msg, PD7124_MSG_LEN);
	i = poe_recv(lc, (char *)&msg, PD7124_MSG_LEN, POE_TIMEOUT);

	/* display response */
	printf("recv: ");
	if (i < PD7124_MSG_LEN) {
	    printf("%d timed-out", i);
	} else {
	    s = (char *) &msg;
	    for (i = 0; i < PD7124_MSG_LEN; i++, s++) {
	        printf("%02x ", *s);
	    }
	}
	printf("\n");

    } else if (argv[1][0] == 'p') {
        /* program POE (download microcode) */
        int line = 0;
	int last;

        if (argc != 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	/* TFTP file into memory */
	copy_filename (BootFile, argv[3], sizeof(BootFile));
	load_addr = 0x200000;
	if (NetLoop(bd, TFTP) == 0) {
	    return 1;
	}

	/* TFTP complete, file at load_addr, program POE */
	/* send program command */
	msg.key = PD7124_KEY_PROGRAM;
	msg.num = seq;
	seq = PD7124_SEQNEXT(seq);
	msg.subject[0] = PD7124_SUBJECT_FLASH;
	msg.subject[1] = 0x99;
	msg.data[0] = 0x15;
	msg.data[1] = 0x16;
	msg.data[2] = 0x16;
	s = &msg.data[3];
	i = s - (char *)&msg;
	for (; i < PD7124_MSG_LEN-2; i++, s++) {
	    *s = PD7124_SPACE; /* 'N' */
	}
	PD7124_CKSUM(&msg); /* checksum */
	poe_send(lc, (char *)&msg, PD7124_MSG_LEN);
	i = poe_recv(lc, (char *)&msg, PD7124_MSG_LEN, POE_TIMEOUT);

	if (i < PD7124_MSG_LEN) {
	    printf("%d timed-out -6", i);
	    return (1);
	}

	if ((msg.key == PD7124_KEY_TELEMETRY)  &&
	    (msg.subject[0] & 0x2)) {
	    /* needs programming, send "ENTR" to start download */
	    poe_send(lc, "E", 1);
	    udelay(10000);
	    poe_send(lc, "N", 1);
	    udelay(10000);
	    poe_send(lc, "T", 1);
	    udelay(10000);
	    poe_send(lc, "R", 1);
	} else if ((msg.key != PD7124_KEY_REPORT) ||
		   (msg.subject[0] != 0) ||
		   (msg.subject[1] != 0)) {
	    printf("Report failed\n");
	    return (1);
	}

	/* Wait for "TPE\r\n" */
	i = poe_recv(lc, (char *)&msg, 5, POE_TIMEOUT);
	if (i < 5) {
	    printf("%d timed-out -5", i);
	    return (1);
	}
	/* Send "E" for erase */
	poe_send(lc, "E", 1);
	/* Wait for "TOE\r\n" */
	i = poe_recv(lc, (char *)&msg, 5, POE_TIMEOUT);
	if (i < 5) {
	    printf("%d timed-out -4", i);
	    return (1);
	}
	/* Wait for "TE\r\n" which indicates flash erased (5 seconds) */
	i = poe_recv(lc, (char *)&msg, 4, 5*POE_TIMEOUT);
	if (i < 4) {
	    printf("%d timed-out -3", i);
	    return (1);
	}
	/* Wait for "TPE\r\n" */
	i = poe_recv(lc, (char *)&msg, 5, POE_TIMEOUT);
	if (i < 5) {
	    printf("%d timed-out -2", i);
	    return (1);
	}
	/* Send "P" for program */
	poe_send(lc, "P", 1);
	/* Wait for "TOP\r\n" */
	i = poe_recv(lc, (char *)&msg, 5, POE_TIMEOUT);
	if (i < 5) {
	    printf("%d timed-out -1", i);
	    return (1);
	}

	/* Send each line from file */
	s = (char *)load_addr;
	last = 0;
	do {
	    char *n;

	    /* do not send "S0" lines for some reason */
	    if ((s[0] == 'S') && (s[1] == '0')) {
	        while (*s++ != 0x0a);
		continue;
	    }
	    /* stop sending when end of file is seen */
	    if ((s[0] != 'S') || (s[1] == '8')) {
	        last = 1;
	    }
	    /* send line */
	    do {
	        luart_putc(lc, PD7124_LC_PORT, *s);
	    } while (*s++ != 0x0a);
	    
	    /* Wait for "T*\r\n" */
	    i = poe_recv(lc, (char *)&msg, 4, POE_TIMEOUT);
	    if (i < 4) {
	        printf("%d timed-out %d", i, line);
		return (1);
	    }
	    line++;
	    s = n;
	} while (!last);

	/* Wait 400 ms */
	udelay(400*1000);

	/* Send "RST" */
	poe_send(lc, "RST", 3);

    } else {
        printf("%s: command invalid\n", argv[1]);
	return 1;
    }

    return (0);
}

int do_mswitch_gbic(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc, device, reg, val, error;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 4) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, val);
    if (!(val & SCCI_MODULE_LC_PRES(lc))) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (lcci_read(lc, LCCI_MODULE_STATUS, &val)) {
        printf("%lx: linecard read failed\n", lc);
	return 1;
    }

    if (!(val & LCCI_MEZZ_PRES)) {
        printf("%lx: mezzanine card not present\n", lc);
	return 1;
    }

    device = simple_strtoul(argv[3], NULL, 16);
    if (device >= LCCI_GBIC_DEVICES) {
        printf("%lx: device invalid\n", device);
	return 1;
    }

    if (argv[1][0] == 'd') {
	if (argc > 4) {
	    times = simple_strtoul(argv[4], NULL, 16);
	}
    } else {
	if (argc > 6) {
	    times = simple_strtoul(argv[6], NULL, 16);
	}
    }

    /* Print the lines.
     */
    while (times--) {
        if (argv[1][0] == 'd') {
	    reg = 0;
	    length = 96;
	} else {
	    reg = simple_strtoul(argv[4], NULL, 16);
	    if (reg >= LCCI_GBIC_REGS) {
	        printf("%02lx: address invalid\n", reg);
		return 1;
	    }

	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 5) {
	        length = simple_strtoul(argv[5], NULL, 16);
	    }
	}
        do {
	    GBIC_EEPROM_READ(SCCI_BASE_ADDR, lc, device, reg, val, error);
	    if (error) {
	        printf("%lx,%lx,%02lx: failed %08lx\n",
		       lc, device, reg, error);
		return 1;
	    } else if (!times) {
	        printf("%lx,%lx,%02lx: %02lx  %c\n",
		       lc, device, reg, val, CHAR(val));
	    }
	    reg += 1;
	    if (reg >= LCCI_GBIC_REGS) {
		break;
	    }
	    if (ctrlc()) {
	        return 1;
	    }
	} while (--length > 0);
    }

    return (rc);	
}

#define BUF_SIZE 2048
#define UART_NEXT(I) (((I) + 1) & (BUF_SIZE-1))
char uart_buffer[UARTS][BUF_SIZE];


int do_mswitch_uartb(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    char c;
    int i;
    int prod[UARTS] = {0};
    int cons[UARTS] = {0};

    NS16550_init(1, CONFIG_BUS_CLK_FREQ / 16 / 115200);

    if (argc > 1) {
        for (i = 1; i < argc; i++) {
          uart_puts(1, argv[i]);
          uart_putc(1, ' ');
        }
	uart_putc(1, '\r');
        return 0;
    }

    printf("Connecting UART A to B...\nPress 'Q' to quit\n");
    while (1) {
        WATCHDOG_RESET();
	for (i = 0; i < UARTS; i++) {
	    if (uart_tstc(i) &&
		(UART_NEXT(prod[i]) != cons[i])) {
	        c = uart_getc(i);
		if ((c == 'Q') && (i == 0)) return 0;
		uart_buffer[i][prod[i]] = c;
		prod[i] = UART_NEXT(prod[i]);
	    }
	    if ((prod[i] != cons[i]) &&
		uart_space(i^1)) {
	        uart_putc(i^1, uart_buffer[i][cons[i]]);
		cons[i] = UART_NEXT(cons[i]);
	    }
	}
    }

    return 0;
}

int do_mswitch_phy(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong lc, device, reg, val, error, port;
    int length = 1;
    int times = 1;
    int rc = 0;
    
    if ((argc < 3) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    lc = simple_strtoul(argv[2], NULL, 16);
    if (lc >= SCCI_LCS) {
        printf("%lx: linecard invalid\n", lc);
	return 1;
    }

    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, val);
    if (!(val & SCCI_MODULE_LC_PRES(lc))) {
        printf("%lx: linecard not present\n", lc);
	return 1;
    }

    if (argv[1][0] == 'l') {
        for (port = 0; port < CXE_PORTS; port++) {
	    device = port_map[port].phy;
	    cxe_read_phy(lc, device, 0x1, &val);
	    cxe_read_phy(lc, device, 0x1, &val);
	    if (val & 0x4) {
	        printf("link on %lx\n", port);
	    }
	}
    } else {
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        device = simple_strtoul(argv[3], NULL, 16);
	if (device >= PHY_DEVS) {
	    printf("%lx: device invalid\n", device);
	    return 1;
	}

	/* map to real phy */
	device = port_map[device].phy;

	if (argv[1][0] == 'w') {
	    if (argc < 6) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	
	    reg = simple_strtoul(argv[4], NULL, 16);
	    if (reg >= PHY_REGS) {
	        printf("%04lx: register invalid\n", reg);
		return 1;
	    } 
	    val = simple_strtoul(argv[5], NULL, 16);
	    if (val >= 65536) {
	        printf("%02lx: value invalid\n", val);
		return 1;
	    }

	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }

	    while (times--) {
	        error = cxe_write_phy(lc, device, reg, val);
		if (error) {
		    printf("cxe_phy_write %lx: failed %08lx\n",
			   lc, error);
		    return 1;
		}
	    }

	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 6) {
	        times = simple_strtoul(argv[6], NULL, 16);
	    }
	    
	    while (times--) {
	        if (argc > 4) {
		    reg = simple_strtoul(argv[4], NULL, 16);
		    if (reg >= PHY_REGS) {
		        printf("%04lx: register invalid\n", reg);
			return 1;
		    } 
		} else {
		    reg = 0;
		    length = 0x20;
		}
	       if (argc > 5) {
		   length = simple_strtoul(argv[5], NULL, 16);
	       }
    
	       /* Print the lines.
		*/
	       do {
		   error = cxe_read_phy(lc, device, reg, &val);
		   if (error) {
		       printf("%lx,%lx,%02lx: failed %lx\n",
			      lc, device, reg, error);
		       return 1;
		   } else if (!times) {
		       printf("%lx,%lx,%02lx: %02lx\n",
			      lc, device, reg, val & 0xffff);
		   }
		   reg += 1;
		   if (reg >= PHY_REGS) {
		       break;
		   }
		   if (ctrlc()) {
		       return 1;
		   }
	       } while (--length > 0);
	    }
	}
    }

    return (rc);	
}

int do_mswitch_ieepro(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    int eepro100_mdi_read(struct eth_device* dev, int reg);
    void eepro100_mdi_write(struct eth_device* dev, int reg, int value);
    void eepro100_eeprom_program(struct eth_device *dev, u8 *mac);
    void eepro100_eeprom_dump(struct eth_device *dev);

    struct eth_device *dev;
    ulong reg, val;
    int length = 1;
    int rc = 0;
    
    if ((argc < 2) || (argc > 4)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    dev = eth_get("i82559#0");
    if (!dev) {
        printf("i82559#0 not found\n");
	return 1;
    }

    if (argv[1][0] == 'p') {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        /* program MAC address */
        eepro100_eeprom_program(dev, argv[2]);

    } else if (argv[1][0] == 'd') {

        /* program MAC address */
        eepro100_eeprom_dump(dev);

    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        reg = simple_strtoul(argv[2], NULL, 16);
	if (reg >= PHY_REGS) {
	    printf("%04lx: register invalid\n", reg);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    if (argc != 4) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	
	    val = simple_strtoul(argv[3], NULL, 16);
	    if (val >= 65536) {
	        printf("%02lx: value invalid\n", val);
		return 1;
	    }

	    eepro100_mdi_write(dev, reg, val);
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 3) {
	        length = simple_strtoul(argv[3], NULL, 16);
	    }
    
	    /* Print the lines.
	     */
	    do {
	        val = eepro100_mdi_read(dev, reg);
		printf("%02lx: %04lx\n", reg, val);
		reg += 1;
		if (reg >= PHY_REGS) {
		    break;
		}
		if (ctrlc()) {
		    rc = 1;
		    break;
		}
	    } while (--length > 0);
	}
    }

    return (rc);	
}

extern block_dev_desc_t * ide_get_dev(int dev);

static int le32_to_int(unsigned char *le32)
{
    return ((le32[3] << 24) +
	    (le32[2] << 16) +
	    (le32[1] << 8) +
	     le32[0]
	   );
}

static int upgrade_device;
static int upgrade_block;
static int upgrade_end;
static int upgrade_cksum;

int ide_tftp_write(ulong offset, u8 *src, unsigned len)
{
    block_dev_desc_t *dev;
    int i, *ptr;

    ptr = (int *) src;
    for (i = 0; i < len/sizeof(int); i++, ptr++) {
        upgrade_cksum += *ptr;
    }

    dev = ide_get_dev(upgrade_device);
    for (; len && (upgrade_block < upgrade_end); upgrade_block++) {
        if (dev->block_write(dev->dev, upgrade_block, 1, (ulong*)src) != 1) {
	    printf("%d write failed\n", upgrade_block);
	}
	if (len < dev->blksz) {
	    len = 0;
	} else {
	    len -= dev->blksz;
	}
	src += dev->blksz;
    }
    return len? ERR_INVAL: ERR_OK;
}

#define DOS_PARTITIONS 4

int do_mswitch_upgrade(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    extern block_dev_desc_t * ide_get_dev(int dev);
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    int block, count;
    unsigned device;
    ulong lba512;
    unsigned partition;
    char c;

    if (argc != 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][1] == ':') {
        device = simple_strtoul(argv[1], NULL, 16);
        partition = simple_strtoul(&argv[1][2], NULL, 16);
    } else {
        device = 0;
        partition = simple_strtoul(argv[1], NULL, 16);
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("%x: Device invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("%x: Partition invalid\n", partition);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("%x: Device invalid\n", device);
	return 1;
    }

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("%x: Device not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Partition type (if any) unrecognized\n");
	return 1;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d: Read failed\n", device);
	return 1;
    }

    /* The first two partitions can contain images */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        printf("%x: Partition not created\n", partition);
    }

    block = le32_to_int (pt->start4);
    count = block + le32_to_int (pt->size4);

    /* save state */
    upgrade_device = device;
    upgrade_block = block;
    upgrade_end = count;
    upgrade_cksum = 0;
    copy_filename (BootFile, argv[2], sizeof(BootFile));
    load_addr = CFG_ATA_BASE_ADDR;

    /* fire off TFTP */
    eth_set("i82559#0");
    if ((ar->size = NetLoop(bd, TFTP)) == 0) {
        return 1;
    }

    if (*((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET) !=
	(device<<4|partition)) {
        printf("Try to boot this partition first ([y]/n)? ");
	if ((c = uart_getc(0)) != 'n') {
	    *((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET) = device<<4|partition;
	}
	printf("%c\n", c);
    }

    return (0);
}

#define FLASH_PARTS 4
#define FLASH_MARGIN 136
dos_partition_t part[FLASH_PARTS] = {
  { 0x00, 0x01, 0x01, 0x00, 0x83, 0x07, 0x60, 0x7f,
    { 0x20, 0x00, 0x00, 0x00 }, { 0xe0, 0x7f, 0x01, 0x00 } },
  { 0x00, 0x00, 0x41, 0x80, 0x83, 0x07, 0xe0, 0x00,
    { 0x00, 0x80, 0x01, 0x00 }, { 0x00, 0x81, 0x01, 0x00 } },
  { 0x00, 0x00, 0xc1, 0x01, 0x83, 0x07, 0xe0, 0xd1,
    { 0x00, 0x01, 0x03, 0x00 }, { 0x00, 0xd1, 0x00, 0x00 } },
  { 0x00 },
};

static unsigned long adjustSectors(unsigned long lba512)
{
    // printf("old lba512 = %d\n", (int) lba512);
    if((lba512 > 500000) && ( lba512 < 512000)) {
        //
        // 256 MB flash
        //
        // AVED comes in with 500400
        // WINTEC comes in with 500736
        //
        // Some web sites mention there are two types of 256 MB flashes
        // One that has 1954 blocks and one that has 1956 blocks.
        // To be safe and not trust the numbers passed in by the device 
        // we go with the smaller number of blocks i.e. 1954.
        //
        // 1954 did not work well with WINTEC flashes we try 1950
        // number of sectors = number of blocks(1950) * sectors(32) * heads(8)
        //                   = 499200
        //
        if(lba512 > 500224) {
                lba512 = 499200; // 1950 * 32 * 8;
        }
    } else {
        // 128 MB flash and 
        // all other sizes except 256MB 
        lba512 -= FLASH_MARGIN; /* SAMEER - 2/24/04 - remove sectors as margin */
    }
    // printf("new lba512 = %d\n", (int) lba512);
    return lba512;
}

int do_mswitch_part(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    block_dev_desc_t *dev;
    ulong lba512;
    ulong size;
    unsigned device;
    int j;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        device = simple_strtoul(argv[1], NULL, 16);
    } else {
        device = 0;
    }
    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    for (j = 0; j < FLASH_PARTS; j++) {
        if (j < 2) {
            size = (part[j].size4[0] |
                    (part[j].size4[1] << 8) |
                    (part[j].size4[2] << 16) |
                    (part[j].size4[3] << 24));
            lba512 -= size;
        } else if (j == 2) {
            part[j].size4[0] = lba512 & 0xff;
            part[j].size4[1] = (lba512 >>  8) & 0xff;
            part[j].size4[2] = (lba512 >> 16) & 0xff;
            part[j].size4[3] = (lba512 >> 24) & 0xff;
        }
                  
        set_dos_partition(dev, &part[j], j);
    }
    print_part_dos(dev);

    return (0);
}

static int verify_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    int i;
    char rbuf[512];

    if (dev->block_read(dev->dev, sector, 1, (ulong*)rbuf) != 1) {
        printf("%d read failed\n", sector);
        return 0;
    }
    for (i = 0; i < 512; i++) {
        if (wbuf[i] != rbuf[i]) {
            printf("%d read data mismatch\n", sector);
            return 0;
        }
    }
    return 1;
}

static int format_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    if (dev->block_write(dev->dev, sector, 1, (ulong*)wbuf) != 1) {
        printf("%d write failed\n", sector);
        return 0;
    }
    return 1;
}

static void format_partition(block_dev_desc_t *dev, int start, int end,
                             char *wbuf, int verify)
{
#define PUTC_MASK 0x7ff
    int i;

    for(i = start; i < end; i++){
        format_sector(dev, i, wbuf);
        if (!(i & PUTC_MASK)) {
            serial_putc('.');
        }
    }
    if (verify) {
        for(i = start; i < end; i++){
            if (!verify_sector(dev, i, wbuf)) {
                return;
            }
            if (!(i & PUTC_MASK)) {
                serial_putc(',');
            }
        }
    }
}

int do_mswitch_format(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    block_dev_desc_t *dev;
    ulong lba512;
    unsigned int device,partition;
    char c,src[512];
    char formatPart0=0;
    char formatPart1=0;
    char formatPart2=0;
    int verify = 0;
    int start, end;
    dos_partition_t *pt;

    if (argc > 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);

	    switch(partition){
		    case 0: formatPart0=1; break;
		    case 1: formatPart1=1; break;
		    case 2: formatPart2=1; break;
		    default:
			printf("Invalid device %s\n", argv[1]);
			return 1;
	    }
    	    printf("Format will erase everything on the %s flash's partition %d.\n", (device?"external":"internal"), partition);

	} else {
	    device = simple_strtoul(argv[1], NULL, 16);
	    formatPart0=1;
	    formatPart1=1;
	    formatPart2=1;
    	    printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
	} 
        if (argc > 2) {
            verify = 1;
        }

    } else {
	/* default format device 0 all three partitions. */
	formatPart0=1;
	formatPart1=1;
	formatPart2=1;
        device = 0;
    	printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    printf("Are you sure you want to continue (y/n)?");
    c=getc();
    if(c!='y' && c!='Y') {
	    printf("\nFormat Cancelled\n");
	    return 1;
    }
    printf("%c\nFormatting", c);

    memset(src,0xFF,sizeof(src));

    if(formatPart0) {
        pt = &part[0];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart1) {
        pt = &part[1];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart2) {
        pt = &part[2];
        start = le32_to_int (pt->start4);
        end = lba512;
        format_partition(dev, start, end, src, verify);
    } 
    printf("\ndone\n");

    /* partition it */
    return (do_mswitch_part(cmdtp, bd, flag, 1, argv));
}

static u8 *load_partition(bd_t *bd, unsigned device, unsigned partition,
			  u32 *len)
{

    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    ulong lba512;
    u8 *addr;
    int block, count;
    int cksum, i, *ptr;

    dev = ide_get_dev(device);
    if (!dev) return NULL;

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) return NULL;
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Device %d partition type (if any) unrecognized\n", device);
	return NULL;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, 0);
	return NULL;
    }

    /* valid partition? */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        return NULL;
    }

    block = le32_to_int (pt->start4);
    if (dev->block_read(dev->dev, block, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, block);
	return NULL;
    }

    /* First block of partition contains aruba image header */
    ar = (aruba_image_t *) buffer;
    if (strcmp(ARUBA_IMAGE_MAGIC, ar->magic)) {
        printf("%d:%d Image not present\n", device, partition);
	return NULL;
    }
    if (!ar->valid) {
        printf("%d:%d Image not valid\n", device, partition);
	return NULL;
    }

    if (ar->machineType != ARUBA_IMAGE_MSWITCH) {
        printf("%d:%d Image not for this platform\n", device, partition);
	return NULL;
    }

    /* remaining blocks contain actual image */
    addr = (u8 *) ((CFG_MEMTEST_END - ar->size) & 0xfff00000);
    count = (ar->size + dev->blksz - 1) / dev->blksz;
    /* read image into memory */
    printf("Loading image %d:%d", device, partition);
    if (dev->block_read(dev->dev, block + 1, count,
			(ulong *)addr) != count) {
        printf("%d:%d:%d read failed\n", device, partition, block+1);
	return NULL;
    }
    if (ctrlc())
        return NULL;

    if (ar->chksum) {
        /* verify checksum */
        printf("\nVerifying checksum... ");

	cksum = 0;
	ptr = (int *) buffer;
	for (i = 0; i < sizeof(buffer)/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	ptr = (int *) addr;
	for (i = 0; i < ar->size/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	if (cksum) {
	    /* invalid cksum */
	    printf("failed!\n");
	    return NULL;
	}

    }

    if (len) {
        *len = ar->size;
    }

    return addr;
}

static void boot_partition(bd_t *bd, unsigned device, unsigned partition)
{
    u8 *addr;

    /* disable interrupts while accessing compact FLASH */
    disable_interrupts();
    addr = load_partition(bd, device, partition, NULL);
    enable_interrupts();
    if (addr) {
        printf("\nHalting active components...");
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE,
                          SC_CPLD_RESET_ENABLE_VALUE);
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LOCAL_RESET,
                          SC_CPLD_LOCAL_RESET_PCI1520);
        udelay (10 * 1000);
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LOCAL_RESET, 0);
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE, 0);
        printf("\nBooting image...\n");
	/* set load addr */
	load_addr = (ulong)addr;
	/* boot it (DOES NOT RETURN if successful) */
	do_bootelf(NULL, bd, 0, 0, NULL);
    }
}

int do_mswitch_bootf(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    unsigned partition, part;
    unsigned device, dev;
    int search;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        search = 0;
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);
	} else {
	    device = 0;
	    partition = simple_strtoul(argv[1], NULL, 16);
	}
    } else {
        device = *((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET);
	partition = device & 0xf;
	device >>= 4;
	if ((device >= CFG_IDE_MAXDEVICE) ||
	    (partition >= DOS_PARTITIONS)) {
	    device = 0;
	    partition = 0;
	}
        search = 1;
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("Partition %x invalid\n", partition);
	return 1;
    }

    if (!search) {
        boot_partition(bd, device, partition);
	printf("Failed\n");
	return 1;
    }

    /* try to boot primary partition */
    boot_partition(bd, device, partition);

    /* failed, try to boot from any partition */
    for (dev = 0; dev < CFG_IDE_MAXDEVICE; dev++) {
        for (part = 0; part < DOS_PARTITIONS; part++) {
	    if ((part != partition) || (dev != device)) {
	        boot_partition(bd, dev, part);
	    }
	}
    }

    return 1;
}

#define JTAG_WRITE(Val) { \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val)); \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val) | SC_CPLD_JTAG_OPERATION_CLK); \
}

#define JTAG_WRITE_READ(Val) { \
    int Tmp; \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val)); \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val) | SC_CPLD_JTAG_OPERATION_CLK); \
    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		     Tmp); \
    printf("%d", (Tmp & SC_CPLD_JTAG_OPERATION_TDO)? 1:0); \
}

#define JTAG_READ(Val, Rc) { \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val) | SC_CPLD_JTAG_OPERATION_CLK); \
    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		     (Rc)); \
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_OPERATION, \
		      (Val)); \
}

static int fpga_crypto = 0;
static int fpga_dev = 0;

static int fpga_id(int dev, int magic)
{
    int i, status, val;

    /* enable JTAG programming */
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_SELECTION,
		      dev | SC_CPLD_JTAG_SELECTION_ENABLE);

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 12; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Send 12 1's */
    for (i = 0; i < 12; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS|SC_CPLD_JTAG_OPERATION_TDI);
    }
    /* 3. Send magic in reverse order */
    for (i = 0; i < 8; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS|
		   ((magic & (1 << i))? SC_CPLD_JTAG_OPERATION_TDI:0));
    }
    /* 4. Read 8 bits */
    status = 0;
    for (i = 0; i < 8; i++) {
        JTAG_READ(SC_CPLD_JTAG_OPERATION_TMS, val);
	status |= ((val & SC_CPLD_JTAG_OPERATION_TDO)? 1:0) << i;
    }

    /* disable JTAG programming */
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_SELECTION, 0);

    return status;
}

static int fpga_erase(int dev)
{
    int i;
    ulong status;

    if (dev != SC_CPLD_JTAG_SELECTION_LOCAL) {
        /* erase FPGA */
        (void) fpga_id(dev, 0xa8);
    } else {
        /* Erase FPGA */
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_FPGA_INTERFACE,
			  (fpga_crypto? SC_CPLD_FPGA_INTERFACE_CRYPTO_PRO:
			   SC_CPLD_FPGA_INTERFACE_PRO));
	i = 0;
	while (1) {
	    udelay(1); /* wait at least one us */
	    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_FPGA_INTERFACE,
			     status);
	    if (status & SC_CPLD_FPGA_INTERFACE_INIT_L) break;
	    if (i++ > 100000) {
	        printf("Erase failed\n");
		return 1;
	    }
	}
    }

    /* enable JTAG programming */
    SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_SELECTION,
		      dev | SC_CPLD_JTAG_SELECTION_ENABLE);

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 10; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Enter Run-Test-Idle */
    JTAG_WRITE(0);

    /* 3. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 4. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    if (dev != SC_CPLD_JTAG_SELECTION_LOCAL) { /* LCCI */
        /* 5. Load CFG-IN */
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	/* 6. Load last bit of CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    } else if (fpga_crypto) { /* Crypto */
        /* 5. Load CFG-IN */
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(0);
	/* Load bypass to other FPGA */
	for (i = 0; i < 4; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 6. Load last bit of bypass */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
    } else { /* SCCI */
        /* Load bypass to other FPGA */
        for (i = 0; i < 5; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	}
	/* 5. Load CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(0);
	/* 6. Load last bit of CFG-IN */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 7. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
    }
    /* 8. Enter SHIFT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }

    if (fpga_crypto) {
        /* must shift in 31 zeros (32 - device) first */
        for (i = 0; i < 31; i++) {
	    JTAG_WRITE(0);
	}
    }
    return 0;
}

int fpga_tftp_write(ulong offset, u8 *src, unsigned len)
{
    int bit, ch;
    u8 val;

    if (offset == 0) {
        if (fpga_erase(fpga_dev)) {
	    return ERR_INVAL;
	}
    }

    for (ch = 0; ch < len-1; ch++, src++) {
        for (bit = 7; bit >= 0; bit--) {
	    val = ((*src >> bit) & 0x1) << 1;
	    JTAG_WRITE(val);
	}
    }
    /* last byte */
    for (bit = 7; bit > 0; bit--) {
        val = ((*src >> bit) & 0x1) << 1;
	JTAG_WRITE(val);
    }
    /* last bit */
    val = (*src & 0x1) << 1;
    if (len != 512) {
	if (fpga_crypto) {
	    printf("$");
	} else {
	    printf("*");
	    val |= SC_CPLD_JTAG_OPERATION_TMS;
	}
    }
    JTAG_WRITE(val);

    return ERR_OK;
}

#undef BOOTFLASH_SCCI
#if defined(BOOTFLASH_SCCI)
#include "fpga.c"
#endif /* defined(BOOTFLASH_SCCI) */

static int fpga_flash_write(bd_t *bd, char *str)
{
    unsigned device, partition;
    u32 len;
    u8 *addr;

    /* determine device and partition */
    if (str[1] == ':') {
        device = simple_strtoul(str, NULL, 16);
	partition = simple_strtoul(&str[2], NULL, 16);
    } else {
        device = 0;
	partition = simple_strtoul(str, NULL, 16);
    }

    /* need to disable interrupts during FLASH accesses */
#if defined(BOOTFLASH_SCCI)
    if (device >= CFG_IDE_MAXDEVICE) {
        addr = fpga_image;
	len = sizeof(fpga_image);
    } else
#endif /* defined(BOOTFLASH_SCCI) */
    {
        disable_interrupts();
	addr = load_partition(bd, device, partition, &len);
	enable_interrupts();
    }
    if (addr) {
        printf("\nWriting image...");
        /* write data to FPGA */
        fpga_tftp_write(0, addr, len);
	printf("\n");
    }
    
    /* success */
    return 0;
}

int do_mswitch_fpga(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong dev, status;
    int rc = 0;
    int i;

    if ((argc < 2) || (argc > 4)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

#if CFG_BANK1_ENABLE
    if (argv[1][0] == 'm') {
        set_bat2_mem();
        return rc;
    } else if (argv[1][0] == 'f') {
        set_bat2_fpga();
        return rc;
    } else
#endif

    if (argv[1][0] == 'p') {
        char *filename;

	fpga_crypto = 0;

#if defined(BOOTFLASH_SCCI)
	if (argc < 3) {
	    dev = SC_CPLD_JTAG_SELECTION_LOCAL;
	    filename = "2:0";
	} else
#endif /* defined(BOOTFLASH_SCCI) */
	if ((argc == 4) && (argv[2][0] != 'c')) {
	    dev = simple_strtoul(argv[2], NULL, 16);
	    if (dev >= SCCI_LCS) {
	        printf("%lx: linecard invalid\n", dev);
		return 1;
	    }

	    SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, status);
	    if (!(status & SCCI_MODULE_LC_PRES(dev))) {
	        printf("%lx: linecard not present\n", dev);
		return 1;
	    }

	    /* convert lc to jtag device */
	    dev += SC_CPLD_JTAG_SELECTION_LC0;
	    filename = argv[3];

	} else {
	    dev = SC_CPLD_JTAG_SELECTION_LOCAL;
	    filename = argv[argc-1];

	    if (argc == 4) {
	        fpga_crypto = 1;
	    }
	}

	fpga_dev = dev;

	if (strlen(filename) < 4) {
	    /* erase FPGA and start initial programming */
	    fpga_erase(dev);
	    /* Program FPGA from flash partition */
	    if (fpga_flash_write(bd, filename)) {
	        return 1;
	    }
	} else {
	    /* TFTP data into FPGA */
	    load_addr = CFG_FPGA_BASE_ADDR;
	    copy_filename (BootFile, filename, sizeof(BootFile));
	    eth_set("i82559#0");
	    if (NetLoop(bd, TFTP) == 0) {
	        return 1;
	    }
	}

	if (fpga_crypto) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
	}

	/* 11. Enter UPDATE-DR */
	JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	/* 12. Enter SELECT-IR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	}
	/* 13. Enter SHIFT-IR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(0);
	}
	if (dev != SC_CPLD_JTAG_SELECTION_LOCAL) { /* LCCI */
	    /* 14. Load JSTART */
	    JTAG_WRITE(0);
	    JTAG_WRITE(0);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    /* 15. Load last bit of JSTART */
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	} else if (fpga_crypto) { /* Crypto */
	    /* 14. Load JSTART */
	    JTAG_WRITE(0);
	    JTAG_WRITE(0);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    JTAG_WRITE(0);
	    /* Load bypass to other FPGA */
	    for (i = 0; i < 4; i++) {
	        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    }
	    /* 15. Load last bit of bypass */
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI|SC_CPLD_JTAG_OPERATION_TMS);
	} else { /* SCCI */
	    /* Load bypass to other FPGA */
	    for (i = 0; i < 5; i++) {
	        JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    }
	    /* 14. Load JSTART */
	    JTAG_WRITE(0);
	    JTAG_WRITE(0);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TDI);
	    /* 15. Load last bit of JSTART */
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	}
	/* 16. Enter SELECT-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	}
	/* 17. Enter SHIFT-DR and clock STARTUP */
	for (i = 0; i < 20; i++) {
	    JTAG_WRITE(0);
	}
	/* 18. Enter UPDATE-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	}
	/* Enter Run-Test-Idle */
	JTAG_WRITE(0);

	/* Enter Test-Logic-Reset */
	for (i = 0; i < 10; i++) {
	    JTAG_WRITE(SC_CPLD_JTAG_OPERATION_TMS);
	}

	/* disable JTAG programming */
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_JTAG_SELECTION, 0);
    } else if (argv[1][0] == 'i') {
        if (argc != 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	dev = simple_strtoul(argv[2], NULL, 16);
	if (dev >= SCCI_LCS) {
	    printf("%lx: linecard invalid\n", dev);
	    return 1;
	}

	SCCI_REG_READ(SCCI_BASE_ADDR, SCCI_MODULE_STATUS, status);
	if (!(status & SCCI_MODULE_LC_PRES(dev))) {
	    printf("%lx: linecard not present\n", dev);
	    return 1;
	}

	/* convert lc to jtag device */
	dev += SC_CPLD_JTAG_SELECTION_LC0;

	printf("%02x\n", fpga_id(dev, 0x55));

    } else {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    return (rc);
}
