/*
 * (C) Copyright 2000
 * Rob Taylor, Flying Pig Systems. robt@flyingpig.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc824x.h>
#include <watchdog.h>
#include <command.h>
#include <cmd_mswitch.h>

#include "../../../cmn/util/sc_cpld.h"
#include "mswitch.h"

int board_pre_init(ulong bootflag)
{
    /* if WARM boot, reset board */
    if (bootflag & BOOTFLAG_WARM) {
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE, 0x55);
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LOCAL_RESET,
			  SC_CPLD_LOCAL_RESET_MASTER);
    }
    return 0;
}

int checkboard(void)
{
    ulong busfreq  = get_bus_freq(0);
    char  buf[16];

    printf("M-Switch Local Bus at %s MHz\n", strmhz(buf, busfreq));

    return 0;
}

static long int initbank(int bank, volatile uchar *base)
{
    int		 i, cnt;
    volatile ulong * addr;
    ulong		 save[32];
    ulong		 val, ret  = 0;

    for (i=0, cnt=(CFG_MAX_RAM_SIZE / sizeof(long)) >> 1; cnt > 0; cnt >>= 1) {
        addr = (volatile ulong *)base + cnt;
	save[i++] = *addr;
	*addr = ~cnt;
    }

    addr = (volatile ulong *)base;
    save[i] = *addr;
    *addr = 0;

    if (*addr != 0) {
        *addr = save[i];
	goto Done;
    }

    for (cnt = 1; cnt < CFG_MAX_RAM_SIZE / sizeof(long); cnt <<= 1) {
        addr = (volatile ulong *)base + cnt;
	val = *addr;
	*addr = save[--i];
	if (val != ~cnt) {
	    ulong new_bank_end = cnt * sizeof(long) - 1;
	    ulong mear1  = mpc824x_mpc107_getreg(MEAR1);
	    ulong emear1 = mpc824x_mpc107_getreg(EMEAR1);
            ulong mask = ~(0xFF << (8*bank));
            mear1 &= mask;
            mear1 |= ((new_bank_end & MICR_ADDR_MASK) >> MICR_ADDR_SHIFT) << (8*bank);
            emear1 &= mask;
            emear1 |= ((new_bank_end & MICR_ADDR_MASK) >> MICR_EADDR_SHIFT) << (8*bank);
#if 0
	    mear1 =  (mear1  & 0xFFFFFF00) |
	      ((new_bank_end & MICR_ADDR_MASK) >> MICR_ADDR_SHIFT);
	    emear1 = (emear1 & 0xFFFFFF00) |
	      
#endif
	    mpc824x_mpc107_setreg(MEAR1,  mear1);
	    mpc824x_mpc107_setreg(EMEAR1, emear1);
	    
	    ret = cnt * sizeof(long);
	    goto Done;
	}
    }

    ret = CFG_MAX_RAM_SIZE;

Done:
    return ret;
}

long int initdram(int board_type)
{
    ulong ret;

    /* initialize bank 0 */
    ret = initbank(0, (volatile uchar *)CFG_SDRAM_BASE);

#if CFG_BANK1_ENABLE
    { 
        /* look for second bank, on back of DIMM */
        ulong ret1;
        /* set BAT2 to mat second bank */
        set_bat2_mem();
        ret1 = initbank(1, (volatile uchar *)CFG_MAX_RAM_SIZE);
        if (ret1 < ret) {
            /* doesn't exist, disable second bank */
            ulong mber  = mpc824x_mpc107_getreg(MBER);
            mber &= ~(0x2);
            mpc824x_mpc107_setreg(MBER,  mber);
            ret1 = 0;
            /* revert BAT2 to FPGA/CPLD map */
            set_bat2_fpga();
        }
        ret += ret1;
    }
#endif

    return ret;
}

static int dram_test(int test, ulong *base, ulong step, ulong length,
		     ulong pattern)
{
    ulong *ptr = base;
    ulong i;
    printf("mem %d ", test);
    for (i = 0; i < length; i += step, ptr += step) {
        *ptr = pattern ^ (ulong) ptr;
    }
    ptr = base;
    for (i = 0; i < length; i += step, ptr += step) {
        if (*ptr != (pattern ^ (ulong) ptr)) {
	    printf("(Mismatch at %08lx: Expected %08lx got %08lx) ",
		   (ulong) ptr, pattern ^ (ulong) ptr, *ptr);
	    return 1;
	}
    }
    printf("\b\b\b\b\b\b");
    return 0;
}

int testdram(ulong dram_size)
{
    int test = 0;
    ulong *ptr;
    ulong data;

    printf("POST:  ");

    /* test entire last 1MB, where ppcboot is copied */
    ptr = (ulong *)(dram_size - 0x100000);
    if (dram_test(test++, ptr, 16, 0x100000 / sizeof(*ptr),
		  0x55555555)) {
        goto failure;
    }
    if (dram_test(test++, ptr, 16, 0x100000 / sizeof(*ptr),
		  0xaaaaaaaa)) {
        goto failure;
    }
    /* test each 128KB region */
    ptr = (ulong *)CFG_SDRAM_BASE;
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0xff00ff00)) {
        goto failure;
    }
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0x00ff00ff)) {
        goto failure;
    }

    printf("passed\n");

#if CFG_BANK1_ENABLE
    if (dram_size > (256<<20)) {
        set_bat2_fpga();
    }
#endif
    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_VERSION, data);
    printf("CPLD:  revision %lx\n", data);
#if CFG_BANK1_ENABLE
    if (dram_size > (256<<20)) {
        set_bat2_mem();
    }
#endif

    return 0;

 failure:
    printf("FAILED!\n");
    return 0;
}
/*
 * Initialize PCI Devices, report devices found.
 */
void pci_init(bd_t *board)
{
    extern int pci_dev_init(int busNo);
    extern int pci1520_initialize(bd_t * bis);

    /* initialize PCI devices on bus 0, the only bus */
    pci_dev_init(0);

    /* initialize card bus bridge */
    printf("IDE:   ");
    if (!pci1520_initialize(board)) {
        printf("none");
    }
    printf("\n");
}

#ifdef CFG_NVRAM_ACCESS_ROUTINE
void *nvram_read(void *dest, const long src, size_t count);
void nvram_write(long dest, const void *src, size_t count);

void *nvram_read(void *dest, const long src, size_t count)
{
    char *d = (char *) dest, *s = (char *) src;
    while (count--) {
        *d++ = *s++;
    }
    return dest;
}

void nvram_write(long dest, const void *src, size_t count)
{
    char *d = (char *) dest, *s = (char *) src;
    while (count--) {
        *d++ = *s++;
    }
}
#endif /* CFG_NVRAM_ACCESS_ROUTINE */
