/*
 * (C) Copyright 2000
 * Rob Taylor, Flying Pig Systems. robt@flyingpig.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc85xx.h>
#include <watchdog.h>
#include <command.h>
#include <cmd_nebbiolo.h>

#include "../../../cmn/util/sc_cpld.h"

int board_pre_init(ulong bootflag)
{
    /* if WARM boot, reset board */
    if (bootflag & BOOTFLAG_WARM) {
        SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_RESET_ENABLE, 0x55);
	SC_CPLD_REG_WRITE(SC_CPLD_BASE_ADDR, SC_CPLD_LOCAL_RESET,
			  SC_CPLD_LOCAL_RESET_MASTER);
    }
    return 0;
}

int checkboard(void)
{
    ulong busfreq  = get_bus_freq(0);
    char  buf[16];

    printf("Nebbiolo Local Bus at %s MHz\n", strmhz(buf, busfreq));

    return 0;
}

long int initdram(int board_type)
{
    int ret = CFG_MAX_RAM_SIZE;
Done:
    return ret;
}

static int dram_test(int test, ulong *base, ulong step, ulong length,
		     ulong pattern)
{
    ulong *ptr = base;
    ulong i;
    printf("mem %d ", test);
    for (i = 0; i < length; i += step, ptr += step) {
        *ptr = pattern ^ (ulong) ptr;
    }
    ptr = base;
    for (i = 0; i < length; i += step, ptr += step) {
        if (*ptr != (pattern ^ (ulong) ptr)) {
	    printf("Mismatch at %08lx: Expected %08lx got %08lx",
		   (ulong) ptr, pattern ^ (ulong) ptr, *ptr);
	    return 1;
	}
    }
    printf("\b\b\b\b\b\b");
    return 0;
}

int testdram(ulong dram_size)
{
    int test = 0;
    ulong *ptr;
    ulong data;

    printf("POST:  ");

    /* test entire last 1MB, where ppcboot is copied */
    ptr = (ulong *)(dram_size - 0x100000);
    if (dram_test(test++, ptr, 16, 0x100000 / sizeof(*ptr),
		  0x55555555)) {
        goto failure;
    }
    if (dram_test(test++, ptr, 16, 0x100000 / sizeof(*ptr),
		  0xaaaaaaaa)) {
        goto failure;
    }
    /* test each 128KB region */
    ptr = (ulong *)CFG_SDRAM_BASE;
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0xff00ff00)) {
        goto failure;
    }
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0x00ff00ff)) {
        goto failure;
    }

    printf("passed\n");

    SC_CPLD_REG_READ(SC_CPLD_BASE_ADDR, SC_CPLD_VERSION, data);
    printf("CPLD:  Revision %lx\n", data);
    
    return 0;

 failure:
    printf("FAILED!!!!!!!\n");
    return 1;
}
/*
 * Initialize PCI Devices, report devices found.
 */
void pci_init(bd_t *board)
{
#ifdef CONFIG_PCI
    extern int pci_dev_init(int busNo);
    extern int pci1520_initialize(bd_t * bis);

    /* initialize PCI devices on bus 0, the only bus */
    pci_dev_init(0);

    /* initialize card bus bridge */
    printf("IDE:   ");
    if (!pci1520_initialize(board)) {
        printf("none");
    }
    printf("\n");
#endif
}

#ifdef CFG_NVRAM_ACCESS_ROUTINE
void *nvram_read(void *dest, const long src, size_t count);
void nvram_write(long dest, const void *src, size_t count);

void *nvram_read(void *dest, const long src, size_t count)
{
    char *d = (char *) dest, *s = (char *) src;
    while (count--) {
        *d++ = *s++;
    }
    return dest;
}

void nvram_write(long dest, const void *src, size_t count)
{
    char *d = (char *) dest, *s = (char *) src;
    while (count--) {
        *d++ = *s++;
    }
}
#endif /* CFG_NVRAM_ACCESS_ROUTINE */
