#include <ppcboot.h>
#include <malloc.h>
#include <mpc85xx.h>
#include <part.h>
#include <ide.h>
#include <ata.h>

#include "pci.h"

#undef DEBUG

#if (CONFIG_COMMANDS & CFG_CMD_NEBBIOLO) && defined(CONFIG_PCI)

#define PCI_VENDOR_ID_TI      0x104c
#define PCI_DEVICE_ID_PCI1420 0xac51
#define PCI_DEVICE_ID_PCI1520 0xac55

#define pci_mem_to_phys(x) (x)
#define pci_ram_to_mem(x) (x)

static inline int INL (struct block_dev_desc* dev, u_long addr)
{
    return le32_to_cpu (*(volatile u32 *) (addr + dev->iobase));
}

static inline int INW (struct block_dev_desc* dev, u_long addr)
{
    return le16_to_cpu (*(volatile u16 *) (addr + dev->iobase));
}

static inline int INB (struct block_dev_desc* dev, u_long addr)
{
    return *(volatile u8 *) (addr + dev->iobase);
}

static inline void OUTB (struct block_dev_desc* dev, int command, u_long addr)
{
    *(volatile u8 *) (addr + dev->iobase) = command;
}

static inline void OUTW (struct block_dev_desc* dev, int command, u_long addr)
{
    *(volatile u16 *) (addr + dev->iobase) = cpu_to_le16 (command);
}

static inline void OUTL (struct block_dev_desc* dev, int command, u_long addr)
{
    *(volatile u32 *) (addr + dev->iobase) = cpu_to_le32 (command);
}

static inline int inb (u_long addr)
{
    return *(volatile u8 *) (CFG_ATA_BASE_ADDR + addr);
}

static inline void outb (u_long addr, u8 data)
{
    *(volatile u8 *) (CFG_ATA_BASE_ADDR + addr) = data;
}

extern block_dev_desc_t * ide_get_dev(int dev);

int pci1520_initialize (bd_t * bis)
{
    int card_number, i, j, status = 0;
    int bus, devno, func;
    struct block_dev_desc *dev;
    uint32 iobase;
    uint32 atabase;
    static int supported[] = { 
        PCI_DEVICE_ID_PCI1420,
	PCI_DEVICE_ID_PCI1520,
	-1 
    }, *devid;

    for (card_number = i = 0, devid = supported; *devid != -1; devid++) {

        for (; ; i++) {
	    /* Find PCI device
	     */
	    if (pci_dev_find (PCI_VENDOR_ID_TI, *devid,
			      i, &bus, &devno, &func) != TRUE) {
	        break;
	    }

	    pci_config_inl (bus, devno, func, PCI_CFG_BASE_ADDRESS_0, &iobase);
	    iobase &= PCI_MEMBASE_MASK;

	    /* Configure memory window for card accesses */
	    if (card_number) {
	        atabase = CFG_ATA_BASE_ADDR + CFG_ATA_IDE1_OFFSET;
	    } else {
	        atabase = CFG_ATA_BASE_ADDR + CFG_ATA_IDE0_OFFSET;
	    }

#ifdef DEBUG
	    printf ("\npci1520: TI PCI CardBus Bridge @0x%X "
		    "(bus=%d, device=%d, func=%d)\n",
		    iobase, bus, devno, func);
#endif

	    pci_config_outl (bus, devno, func,
			     PCI_CFG_COMMAND,
			     PCI_CMD_MEM_ENABLE | PCI_CMD_MASTER_ENABLE);

	    /* Check if I/O accesses and Bus Mastering are enabled.
	     */
	    pci_config_inl (bus, devno, func, PCI_CFG_COMMAND, &status);
	    if (!(status & PCI_CMD_MEM_ENABLE)) {
	        puts ("Error: Can not enable MEM access\n");
		continue;
	    }
	    if (!(status & PCI_CMD_MASTER_ENABLE)) {
	        puts ("Error: Can not enable Bus Mastering\n");
		continue;
	    }

	    dev = ide_get_dev(i);
	    dev->iobase = iobase;

	    /* Set the latency timer */
	    pci_config_outb (bus, devno, func,
			     PCI_CFG_LATENCY_TIMER, PCI_LATENCY_TIMER);
	    pci_config_outb (bus, devno, func,
			     PCI_CFG_SEC_LATENCY, 0x40);

	    if (i == 0) { /* global registers */
	        /* Enable clock for power switch */
	        pci_config_inl (bus, devno, func, 0x80, &status);
		status |= 0x08000000;
		pci_config_outl (bus, devno, func, 0x80, status);
		
		/* enable LED */
		pci_config_outl (bus, devno, func, 0x8C, 0xd0);
	    }

	    /* Configure ExCA registers */
	    OUTB(dev, 0x40, 0x803); /* de-assert reset */

	    /* Configure 16bit Memory Windows
	     * (0 is memory/ata, 1 is attribute) */
	    for (j = 0; j < 2; j++, atabase += 0x1000) {
	        /* memory window page (bits 31-24) */
	        OUTB(dev, (atabase >> 24) & 0xff, 0x840+j);
		/* memory window start address (bits 23-12) */
		OUTB(dev, (atabase >> 12) & 0xff, 0x810+8*j);
		OUTB(dev, 0x80 | ((atabase >> 20) & 0x0f), 0x811+8*j);
		/* memory window end address (bits 23-12, and wait state) */
		OUTB(dev, (atabase >> 12) & 0xff, 0x812+8*j);
		OUTB(dev, ((atabase >> 20) & 0x0f), 0x813+8*j);
		/* memory window offset (bits 23-12, and attrib) */
		OUTB(dev, 0, 0x814+8*j);
		OUTB(dev, (j? 0xC0: 0x00), 0x815+8*j);
	    }

	    /* enable memory windows */
	    OUTB(dev, 0x03, 0x806);
	    /* wait 10ms before enabling card */
	    udelay (10 * 1000);
	    /* enable card */
	    OUTB(dev, 0x99, 0x802);

#ifdef DEBUG
	    { 
		printf("Sock %08x ", INL(dev, 0x8));
		printf("ExCA %02x\n", INB(dev, 0x801));
		for (j = 0; j < 2; j++) {
		    printf("Window %d:%02x%01x%02x000-%02x%01x%02x000 %02x%02x\n", j,
			   INB(dev, 0x840+j),
			   INB(dev, 0x811+8*j) & 0xf,
			   INB(dev, 0x810+8*j),
			   INB(dev, 0x840+j),
			   INB(dev, 0x813+8*j) & 0xf,
			   INB(dev, 0x812+8*j),
			   INB(dev, 0x815+8*j),
			   INB(dev, 0x814+8*j));
		}
	    }
#endif

	    printf("%spci1520#%d", (i? ", ": ""), i);
	    card_number++;

	    if (i >= CFG_IDE_MAXDEVICE) 
	        return card_number;

	}
    }

    return card_number;
}

#endif
