/*
 * (C) Copyright 2000
 * Rob Taylor, Flying Pig Systems. robt@flyingpig.com.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <mpc824x.h>
#include <watchdog.h>
#include "../../sap/flash/fl_common.h"

void watchdog_reset(void);
void printManufacHwInfo (struct manufact_hw_info_t* mInfo);
void toggle_leds(void);

/* flash stuff */
extern flash_info_t flash_info[];
int write_buff (flash_info_t *info, uchar *src, ulong addr, ulong cnt);
int flash_erase (flash_info_t *info, int s_first, int s_last);
char manuf_flash_sign[MAX_SIGN] = {0xd, 0xe, 0xa, 0xd, 0xb, 0xe, 0xa, 0xd};

int board_pre_init(ulong bootflag)
{
    return 0;
}

int checkboard(void)
{
    ulong busfreq  = get_bus_freq(0);
    char  buf[16];

    printf("Rivet Local Bus at %s MHz\n", strmhz(buf, busfreq));

    return 0;
}

long int initdram(int board_type)
{
    volatile int		 i, cnt;
    volatile uchar * base= CFG_SDRAM_BASE;
    volatile ulong * addr;
    volatile ulong		 save[32];
    volatile ulong		 val, ret  = 0;

    for (i=0, cnt=(CFG_MAX_RAM_SIZE / sizeof(long)) >> 1; cnt > 0; cnt >>= 1) {
        addr = (volatile ulong *)base + cnt;
        save[i++] = *addr;
        *addr = ~cnt;
    }

    addr = (volatile ulong *)base;
    save[i] = *addr;
    *addr = 0;

    if (*addr != 0) {
        *addr = save[i];
        goto Done;
    }

    for (cnt = 1; cnt < CFG_MAX_RAM_SIZE / sizeof(long); cnt <<= 1) {
        addr = (volatile ulong *)base + cnt;
        val = *addr;
        *addr = save[--i];
        if (val != ~cnt) {
            ulong new_bank0_end = cnt * sizeof(long) - 1;
            ulong mear1  = mpc824x_mpc107_getreg(MEAR1);
            ulong emear1 = mpc824x_mpc107_getreg(EMEAR1);

            mear1 =  (mear1  & 0xFFFFFF00) |
                ((new_bank0_end & MICR_ADDR_MASK) >> MICR_ADDR_SHIFT);
            emear1 = (emear1 & 0xFFFFFF00) |
                ((new_bank0_end & MICR_ADDR_MASK) >> MICR_EADDR_SHIFT);
            mpc824x_mpc107_setreg(MEAR1,  mear1);
            mpc824x_mpc107_setreg(EMEAR1, emear1);
            
            ret = cnt * sizeof(long);

            goto Done;
        }
    }

    ret = CFG_MAX_RAM_SIZE;

Done:

    return ret;
}

static int dram_test(int test, ulong *base, ulong step, ulong length,
		     ulong pattern)
{
    ulong *ptr = base;
    ulong i;
    printf("mem %d ", test);
    for (i = 0; i < length; i += step, ptr += step) {
        *ptr = pattern ^ (ulong) ptr;
    }
    ptr = base;
    for (i = 0; i < length; i += step, ptr += step) {
        if (*ptr != (pattern ^ (ulong) ptr)) {
	    printf("Mismatch at %08lx: Expected %08lx got %08lx",
		   (ulong) ptr, pattern ^ (ulong) ptr, *ptr);
	    return 1;
	}
    }
    printf("\b\b\b\b\b\b");
    return 0;
}

int testdram(ulong dram_size)
{
    volatile unsigned char* pLed    = (volatile unsigned char*)0x70000000;
    int test;
    ulong *ptr;

    /* turn the READY LED on initially, to indicate power-on */
    *pLed = 0x80;

    printf("POST:  ");

    ptr = (ulong *)CFG_SDRAM_BASE;
    test = 0;
    /* test entire first 128KB */
    if (dram_test(test++, ptr, 1, (128 * 1024) / sizeof(*ptr),
		  0x55555555)) {
        goto failure;
    }
    if (dram_test(test++, ptr, 1, (128 * 1024) / sizeof(*ptr),
		  0xaaaaaaaa)) {
        goto failure;
    }
    /* test each 128KB region */
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0xff00ff00)) {
        goto failure;
    }
    if (dram_test(test++, ptr, (128 * 1024) / sizeof(*ptr), dram_size/sizeof(*ptr),
		  0x00ff00ff)) {
        goto failure;
    }

    printf("passed\n");

    return 0;

 failure:
    printf("FAILED!!!!!!!\n");
    return 1;
}
/*
 * Initialize PCI Devices, report devices found.
 */
void pci_init(bd_t *board)
{
    extern int pci_dev_init(int busNo);

    /* initialize PCI devices on bus 0, the only bus */
    pci_dev_init(0);
}

int last_stage_init(bd_t *bd)
{
    struct manufact_hw_info_t * mInfo;
    char ethaddr[18];
    
    if (getenv("ethaddr")) {
        /* manually configured */
        return 0;
    }

    mInfo = (struct manufact_hw_info_t *)OLD_MANUFAC_DATA_LOC;

#if NOT_SURE_OF_FUTURE_PROD_ID
    if (strncmp(mInfo->productName, "WA6101B", (sizeof(mInfo->productName)-3))) {
        /* nothing found */
        return 0;
    }
#endif

    sprintf(ethaddr, "%02x:%02x:%02x:%02x:%02x:%02x",
	    mInfo->macAddress[0],
	    mInfo->macAddress[1],
	    mInfo->macAddress[2],
	    mInfo->macAddress[3],
	    mInfo->macAddress[4],
	    mInfo->macAddress[5]);
    setenv ("ethaddr", ethaddr);

    return 0;
}

void watchdog_reset(void)
{
    /* these values should match the CS2/CS3 values concluded from
       ERCR1,3/2,4 in config_RIVET.h */
    static int toggle = 1;
    volatile u8 *wdp   = (volatile u8 *)0x70010000;

    /* addressing the CPLDs watchdog using extendEprom, 
       RCS3 cpu reg, DH0 for watchdog enable, DH1 for watchdog trigger */

    /* we need to strobe the accton board's CPLD watchdog
     * so it won't reboot us in 5 secs: write high then low */
    if (toggle) {
        *wdp = 0x7f;
    } else {
        *wdp = 0xff;
    }
    toggle = !toggle;

    {
      /* to cut down noise on the board, add a few nop's */
      int i;
      for (i = 0; i < 100; i++) {
	asm volatile ("nop");
      }
    }

    return;
}

static int led_ready = 0;

void toggle_leds(void)
{
    /* these values should match the CS2/CS3 values concluded from
       ERCR1,3/2,4 in config_RIVET.h */
    volatile unsigned char* pLed    = (volatile unsigned char*)0x70000000;
    
    if (0==led_ready) {
        *pLed = 0x00;
        led_ready = 1;
    }
    else {
        *pLed = 0xb0;
        led_ready = 0;
    }

}

void printManufacHwInfo (struct manufact_hw_info_t* mInfo)
{
    int i=0;
    printf("Hardware Manufacturing Info as stored in NV: \n\n");
    printf("\tMAC ADDRESS       : ");
    for(i=0; i< 6; i++) printf("%x ", mInfo->macAddress[i]);printf("\n");
    printf("\tPRODUCT ID        : ");
    for (i=0; i<10;i++) printf("%c", mInfo->productId[i]); printf("\n");
    printf("\tCOUNTRY CODE      : %c%c\n", 
           mInfo->countrycode[0], mInfo->countrycode[1]);
    printf("\tMANUFACTURING DATE: ");
    for (i=0; i<6; i++) printf("%c", mInfo->manufacDate[i]);printf("\n");
    printf("\tHW REVISION       : ");
    for (i=0; i < 4; i++) printf("%c", mInfo->hwRevision[i]);printf("\n");
    printf("\tPRODUCT NAME      : ");
    for (i=0; i < 10; i++) printf("%c", mInfo->productName[i]);printf("\n");
    printf("\tHW ID             : %d\n\n", mInfo->hwId[0]);
    return;
}
