#include <ppcboot.h>
#include <command.h>
#include <net.h>
#include <watchdog.h>
#include "cmd_syrah.h"
#include "cmd_elf.h"
#include "cmd_boot.h"
#include "ns16550.h"
#include "../disk/part_dos.h"

#include "../../../cmn/util/syr_cpld.h"
#include "../../../cmn/util/syr_fpga.h"
#include "../../../cmn/util/cxe.h"
#include "../../../cmn/util/gbic.h"
#include "../../../cmn/util/aruba_image.h"

#define TIMEOUT	       2		/* Seconds before failing */
#define PKTSIZEMIN     64
#define PHY_TIMEOUT    10000
#define PHY_REGS       32
#define PHY_DEVS       28

#define CHAR(c)       (((c >= 0x20) && (c <= 0x7e))? (char)c: '.')

extern struct NS16550 *COM_PORTS[];
static u_long loopback_length;
static u_long loopback_count;
static u_long loopback_vlan;
static u_long loopback_incr;
static u_long loopback_received;
static int loopback_rcv;
void LoopbackRequest(void);
static unsigned long adjustSectors(unsigned long lba512);

typedef struct {
  u16 pvid;
  u8  mac;
  u8  phy;
} port_map_t;

port_map_t port_map[CXE_PORTS] = {
    { 0x01, 0x0b, 0x0e }, { 0x02, 0x1b, 0x0f },
    { 0x03, 0x0a, 0x0c }, { 0x04, 0x1a, 0x0d },
    { 0x05, 0x09, 0x0a }, { 0x06, 0x19, 0x0b },
    { 0x07, 0x08, 0x08 }, { 0x08, 0x18, 0x09 },
    { 0x09, 0x07, 0x16 }, { 0x0a, 0x17, 0x17 },
    { 0x0b, 0x06, 0x14 }, { 0x0c, 0x16, 0x15 },
    { 0x0d, 0x05, 0x12 }, { 0x0e, 0x15, 0x13 },
    { 0x0f, 0x04, 0x10 }, { 0x10, 0x14, 0x11 },
    { 0x11, 0x03, 0x1e }, { 0x12, 0x13, 0x1f },
    { 0x13, 0x02, 0x1c }, { 0x14, 0x12, 0x1d },
    { 0x15, 0x01, 0x1a }, { 0x16, 0x11, 0x1b },
    { 0x17, 0x00, 0x18 }, { 0x18, 0x10, 0x19 },
    { 0x19, 0x0e, 0x02 }, { 0x1a, 0x0f, 0x03 },
    { 0x02, 0x0d, 0x01 }, { 0x01, 0x0c, 0x00 }
};

#define CXE_MAC(port) ((0xb - ((port) >> 1)) | (((port) & 0x1) << 4))

char *cxe_stats_str[] = {
  "TxUcast",
  "TxMcast",
  "TxBcast",
  "TxDeferred",
  "TxColisions",
  "Tx1Colision",
  "TxNColisions",
  "TxXColisions",
  "TxLColisions",
  "TxLen64",
  "TxLen65",
  "TxLen128",
  "TxLen256",
  "TxLen512",
  "TxLen1024",
  "TxLen1519",
  "TxPause",
  "TxOctets",
  "TxOctetsBad",
  "RxOctets",
  "RxOctetsBad",
  "RxRunt",
  "RxShort",
  "RxShortCrc",
  "RxUcast",
  "RxMcast",
  "RxBcast",
  "RxCrc",
  "RxAlignment",
  "RxLong",
  "RxLongCrc",
  "RxLen64",
  "RxLen65",
  "RxLen128",
  "RxLen256",
  "RxLen512",
  "RxLen1024",
  "RxLen1519",
  "RxPause",
};

char *cxe_qe_discard_str[] = {
  "InqOverflow",
  "FifoOverflow",
  "BufMemOverflow",
  "RliDiscard",
  "RleDiscard",
  "McLimitDiscard",
  "TtlScoping",
  "WFHBD",
  "MacError",
};

char *cxe_qe_rl_str[] = {
  "Port 0",
  "Port 1",
  "Port 2",
  "Port 3",
  "Port 4",
  "Port 5",
  "Port 6",
  "Port 7",
  "Port 8",
  "Port 9",
  "Port a",
  "Port b",
  "Port c",
  "Port d",
  "Port e",
  "Port f",
  "Port 10",
  "Port 11",
  "Port 12",
  "Port 13",
  "Port 14",
  "Port 15",
  "Port 16",
  "Port 17",
  "Port 18",
  "Port 19",
  "Port 1a",
  "Port 1b",
  "CPU port",
  "Multicast",
  NULL, NULL,
  "FFA",
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  "Link",
};

static void cxe_write(ulong addr, ulong data)
{
    ulong error;
    CXE_REG_WRITE(CXE_BASE_ADDR, addr, data, error);
    if (error) {
        printf("cxe_write %04lx: failed %08lx\n",
	       addr, error);
    }
}

static ulong cxe_read(ulong addr, ulong *data)
{
    ulong val, error;
    CXE_REG_READ(CXE_BASE_ADDR, addr, val, error);
    if (!error) {
        *data = val;
    }
    return error;
}

static ulong cxe_write_phy(ulong device, ulong reg, ulong val)
{
    ulong tries, error;
    cxe_write(CXE_GC_MDIO_ACCESS,
	      (CXE_GC_MDIO_ACCESS_WRITE |
	       (device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT) |
	       (val << CXE_GC_MDIO_ACCESS_DATA_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(CXE_GC_MDIO_ACCESS, &val);
	if (error) {
	    return error;
	}
	if (val & CXE_GC_MDIO_ACCESS_DONE) {
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

static ulong cxe_read_phy(ulong device, ulong reg, ulong *val)
{
    ulong tries, error;
    cxe_write(CXE_GC_MDIO_ACCESS,
	      ((device << CXE_GC_MDIO_ACCESS_DEV_SHIFT) |
	       (reg << CXE_GC_MDIO_ACCESS_REG_SHIFT)));
    for (tries = 0; tries < PHY_TIMEOUT; tries++) {
        error = cxe_read(CXE_GC_MDIO_ACCESS, val);
	if (error) {
	    break;
	}
	if (*val & CXE_GC_MDIO_ACCESS_DONE) {
	    *val &= 0xffff;
	    break;
	}
    }
    if (tries >= PHY_TIMEOUT) {
        error = tries;
    }
    return error;
}

static void
LoopbackTimeout (void)
{
    puts ("Timeout expired\n");
    NetState = NETLOOP_FAIL;
}

static void
LoopbackHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
    int i;

    if (len != loopback_length) {
        printf("Error: length mismatch %d != %ld\n",
	       len, loopback_length);
	goto error;
    }
    for (i = 0; i < len; i++, pkt++) {
        if (*pkt != NetTxPacket[i]) {
	    printf("Error: byte %d mismatch: %02x != %02x\n",
		   i, *pkt, NetTxPacket[i]);
	    goto error;
	}
    }
    loopback_received++;
    
    if (loopback_incr) {
        if (loopback_length++ >= 1514) {
	    loopback_length = PKTSIZEMIN;
	}
    }

    if (serial_tstc() || 
	(loopback_count && (loopback_count-- == 1))) {
        NetState = NETLOOP_SUCCESS;
    } else {
        NetSendPacket(NetTxPacket, loopback_length);
	NetSetTimeout (TIMEOUT * CFG_HZ, LoopbackTimeout);
    }
    return;

 error:
    NetState = NETLOOP_FAIL;
}

void LoopbackRequest(void)
{
    /* called when a packet is received during ether test */
    int i;
    volatile uchar *pkt;

    pkt = NetTxPacket;

    NetSetEther(pkt, NetBcastAddr, PROT_TAG);

    pkt += ETHER_HDR_SIZE;
    *((ushort *)pkt) = loopback_vlan; /* VLAN ID */
    pkt += sizeof(ushort);

    for (i = 0; i < PKTSIZE - ETHER_HDR_SIZE - sizeof(ushort); i++) {
        pkt[i] = i & 0xff;
    }

    if (!loopback_rcv) {
        while (loopback_count--) {
	    NetSendPacket(NetTxPacket, loopback_length);
	}
        NetState = NETLOOP_SUCCESS;
    } else {

        NetSetHandler(LoopbackHandler);
	NetSetTimeout(TIMEOUT * CFG_HZ, LoopbackTimeout);

	NetSendPacket(NetTxPacket, loopback_length);
    }
}

int do_syrah_ether(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong length = PKTSIZEMIN, count = 0, incr = 0, vlan = 1;

    if ((argc < 2) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (eth_set(argv[1])) {
        printf("unknown device %s\n", argv[1]);
	return 1;
    }

    if (argc > 2) {
        if (argv[2][0] == 'i') {
  	    incr = 1;
	} else {
	    length = simple_strtoul(argv[2], NULL, 16);
	    if ((length < PKTSIZEMIN) || (length > 1514)) {
	        printf("length is invalid: %ld\n", length);
		return 1;
	    }
	}
    }

    if (argc > 3) {
        count = simple_strtoul(argv[3], NULL, 16);
	if (count == 0) {
	    printf("count is invalid: %ld\n", count);
	    return 1;
	}
    }

    if (argc > 4) {
        vlan = simple_strtoul(argv[4], NULL, 16);
	if (vlan == 0 || vlan > 4095) {
	    printf("vlan is invalid: %ld\n", vlan);
	    return 1;
	}
    }

    if (argc > 5) {
        /* do bother receiving */
        loopback_rcv = 0;
    } else {
        loopback_rcv = 1;
    }

    loopback_length = length;
    loopback_incr = incr;
    loopback_count = count;
    loopback_vlan = vlan;
    loopback_received = 0;

    NetLoop(bd, LOOPBACK);

    printf("%lx packets successfully received\n", loopback_received);

    return 0;
}

int do_syrah_syrah(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong reg, val;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 2) || (argc > 5)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {

        /* reset all components */
        SYR_REG_WRITE(SYR_BASE_ADDR, SYR_RESET,
		      SYR_RESET_PHY(0) |
		      SYR_RESET_PHY(1) |
		      SYR_RESET_PHY(2) |
		      SYR_RESET_SWC |
		      SYR_RESET_POE |
		      SYR_RESET_SD);
	udelay(1000);
        SYR_REG_WRITE(SYR_BASE_ADDR, SYR_RESET, 0);

	/* reset GBICs */
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_GBIC_CONTROL,
		      (SYR_GBIC_CONTROL_GBIC_DIS(0) |
		       SYR_GBIC_CONTROL_GBIC_DIS(1)));
	udelay(40);
	cxe_write_phy(port_map[24].phy, 0x11, 0xa364); /* enable serdes clocks */
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_GBIC_CONTROL, 0);

	/* reset SOE RX and CPLD */
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_RESET, SYR_SOE_RESET_RX);
	udelay(1000000);
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_RESET, 0);

        /* set status LED to green */
      	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_LED_STATUS,
		      SYR_LED_STAT_COLOR_SET(SYR_LED_COLOR_GREEN));

    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

	if (argv[1][0] == 'w') {
	    if (argc < 4) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	    reg = simple_strtoul(argv[2], NULL, 16);
	    val = simple_strtoul(argv[3], NULL, 16);

	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    while (times--) {
	        SYR_REG_WRITE(SYR_BASE_ADDR, reg, val);
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 4) {
	        times = simple_strtoul(argv[4], NULL, 16);
	    }

	    /* Print the lines.
	     */
	    while (times--) {
	        reg = simple_strtoul(argv[2], NULL, 16);
	        if (argc > 3) {
		    length = simple_strtoul(argv[3], NULL, 16);
		}
	        do {
		    SYR_REG_READ(SYR_BASE_ADDR, reg, val);
		    if (!times) {
		        printf("%04lx: %08lx\n", reg, val);
		    }
		    reg += 1;
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return (rc);	
}

static void cxe_init(ulong mask, int fbm, int bridge)
{
    cxe_vlan_bc_t *vbc;
    cxe_u128_t u128;
    ulong stat, mac;
    ulong error;
    ulong val;
    ulong rev;
    int mtu = CXE_MTU;

    /* Port and limited protocol-based VLANS, no VRP (route) table, etc */
    cxe_write(CXE_AL_CONFIG, 0x0001c061);

    if (!bridge) {
        /* disable L2 learning on ports */
        cxe_write(CXE_DONT_LEARN_PORTS, 0x0fffffff);
    }

    /* clear interrupts */
    error = cxe_read(CXE_GC_INT_STATUS, &val);
    if (error) {
        printf("cxe_read %02x: failed %08lx\n",
	       CXE_GC_INT_STATUS, error);
	return;
    }

    /* Set LED stretch to 64 ms */
    cxe_write(CXE_LED_DISP_MODE, 0x00000005);
    /* Set LED event selection to 0:RX act, 1:TX act, 2:link status */
    for (mac = 0; mac < CXE_PORTS; mac++) {
        cxe_write(CXE_LED_EVENT_SEL(mac), 0x00000345);
    }

    cxe_read(CXE_GC_INT_STATUS, &rev);
    if (CXE_CG_DEVICE_ID_REV(rev) == CXE_CG_DEVICE_ID_REV_G) {
        /* Queue Engine XXX ER_9_1*/
        cxe_write(CXE_QE_CONFIG, 0x10);

        /* Queue Engine Resource Limit configuration */
        val = CXE_INT_CELLS;
        stat = ((mtu + (CXE_CELL_SIZE-1))/CXE_CELL_SIZE) * 2; /* 2 * mtu size */
        cxe_write(CXE_QE_RL_DATA, stat);
        for (mac = 0; mac < CXE_PORTS; mac++) {
            cxe_write(CXE_QE_RL_INSTR, CXE_QE_RL_INSTR_WRITE | mac);
            val -= stat;
        }
        /* reserve same for mcast, even though there should be no mcast */
        cxe_write(CXE_QE_RL_DATA, stat);
        cxe_write(CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
				    CXE_QE_RL_INSTR_ID_MCAST));
        val -= stat;
        /* rest is Free For All (FFA) */
        cxe_write(CXE_QE_RL_DATA, val);
        for (mac = 0; mac < CXE_PRIOS; mac++) {
            cxe_write(CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
					CXE_QE_RL_INSTR_ID_FFA_PRIO(mac)));
        }
        cxe_write(CXE_QE_RL_INSTR, (CXE_QE_RL_INSTR_WRITE |
				    CXE_QE_RL_INSTR_ID_FFA_MCAST));
    } else {
        cxe_write(CXE_GC_DLL_CTRL, 0x0007c00c);
    }
    

    /* MAC configuration */
    for (mac = 0; mac < CXE_PORTS; mac++) {
        if (mac < 16) {
	    /* 0-11 configure two FE ports (don't ask) */
	    if (CXE_FE_PORT(mac)) {
	        /* PAUSE, IPG, RMII, 100FD */
	        cxe_write(CXE_MAC_PORT_CONFIG(mac), 0x000821f2);
	    } else {
	        /* PAUSE, IPG, FD, GMII */
	        cxe_write(CXE_MAC_PORT_CONFIG(mac), 0x00082144);
	    }
	}
	/* Set max pkt len */
	if (CXE_INT_PORT(mac)) {
	    cxe_write(CXE_MAC_MAX_PKT_LEN(mac), mtu+4); /* for Double .Q */
	} else {
	    cxe_write(CXE_MAC_MAX_PKT_LEN(mac), mtu);
	}
    }

    /* Clear TCAM */
    cxe_write(CXE_INT_AL_CAM_0, 0);
    cxe_write(CXE_INT_AL_CAM_1, 0);
    cxe_write(CXE_INT_AL_CAM_2, 0);
    cxe_write(CXE_INT_AL_CAM_3, 0);
    for (mac = 0; mac < CXE_INT_AL_CAM_ENTRIES; mac++) {
        cxe_write(CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
		   mac));
	cxe_write(CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
		   mac));
    }

    /* VLAN configuration */
    vbc = (cxe_vlan_bc_t *) &u128;
    memset(vbc, 0, sizeof(*vbc));
    vbc->TYPE = 1;
    vbc->PERMANENT = 1;
    vbc->VALID = 1;
    if (bridge || fbm) {
        /* Add all ports to VLAN 1 */
        ulong pvid = 1;
        for (mac = 0; mac < CXE_PORTS; mac++) {
	    /* Set PVID of each port to port num + 1 */
	    cxe_write(CXE_PORT_VID_1(mac), pvid<<16|pvid);
	    cxe_write(CXE_PORT_VID_2(mac), pvid<<16|pvid);
	    if (CXE_EXT_PORT(mac)) {
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_2 |= (0x1 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 |= (0x1 << mac);
		}
	    }
	}
	vbc->VID = pvid;
	vbc->HIT_COUNTER = mac;
	if (!fbm) {
	    vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_0);
	}
	cxe_write(CXE_INT_AL_CAM_0, u128.ulBits0_31);
	cxe_write(CXE_INT_AL_CAM_1, u128.ulBits32_63);
	cxe_write(CXE_INT_AL_CAM_2, u128.ulBits64_95);
	cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
	cxe_write(CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
		   mac));

	vbc->VID = 0xfff;
	cxe_write(CXE_INT_AL_CAM_0, 0);
	cxe_write(CXE_INT_AL_CAM_1, 0x80000000);
	cxe_write(CXE_INT_AL_CAM_2, 0);
	cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
	cxe_write(CXE_INT_AL_CAM_INSTR,
		  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
		   mac));
    }
    if (fbm) {
        for (mac = 0; mac < CXE_PORTS; mac += 2) {
	    /* Set PVID of each port to port num + 1 */
	    ulong pvid = CXE_PVID(mac) + 1;
	    if (CXE_EXT_PORT(mac)) {
	        /* loop frames to peer port */
	        cxe_write(CXE_PORT_VID_1(mac), pvid<<16|pvid);
		cxe_write(CXE_PORT_VID_2(mac), pvid<<16|pvid);
	        cxe_write(CXE_PORT_VID_1(mac+1), pvid<<16|pvid);
		cxe_write(CXE_PORT_VID_2(mac+1), pvid<<16|pvid);
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_1 = 0;
		    vbc->OUTPORTS_MASK_PART_2 = (0x3 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 = (0x3 << mac);
		    vbc->OUTPORTS_MASK_PART_2 = 0;
		}
		vbc->VID = pvid;
		vbc->HIT_COUNTER = mac;

		cxe_write(CXE_INT_AL_CAM_0, u128.ulBits0_31);
		cxe_write(CXE_INT_AL_CAM_1, u128.ulBits32_63);
		cxe_write(CXE_INT_AL_CAM_2, u128.ulBits64_95);
		cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
			   mac));

		vbc->VID = 0xfff;
		cxe_write(CXE_INT_AL_CAM_0, 0);
		cxe_write(CXE_INT_AL_CAM_1, 0x80000000);
		cxe_write(CXE_INT_AL_CAM_2, 0);
		cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
			   mac));
	    }
	}
    }
    if (!bridge && !fbm) {
        for (mac = 0; mac < CXE_PORTS; mac++) {
	    ulong pvid = CXE_PVID(mac);
	    /* Set PVID of each port to port num + 1 */
	    cxe_write(CXE_PORT_VID_1(mac), pvid<<16|pvid);
	    cxe_write(CXE_PORT_VID_2(mac), pvid<<16|pvid);
	    if (CXE_EXT_PORT(mac)) {
	        /* Add VLAN entry that includes internal port(s) */
	        vbc->VID = pvid;
		vbc->HIT_COUNTER = pvid;
		if (mask) {
		    vbc->TAGGED_OUTPORTS_MASK_PART_2 = 0;
		    vbc->OUTPORTS_MASK_PART_1 = 0;
		    if (mask & 0x1) {
		        vbc->TAGGED_OUTPORTS_MASK_PART_2 |= (0x1 << (CXE_INT_PORT_0-4));
			vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_0);
		    }
		    if (mask & 0x2) {
		        vbc->TAGGED_OUTPORTS_MASK_PART_2 |= (0x1 << (CXE_INT_PORT_1-4));
			vbc->OUTPORTS_MASK_PART_1 |= (0x1 << CXE_INT_PORT_1);
		    }
		} else {
		    vbc->TAGGED_OUTPORTS_MASK_PART_2 = (0x1 << (CXE_INT_PORT_0-4));
		    vbc->OUTPORTS_MASK_PART_1 = (0x1 << CXE_INT_PORT_0);
		}
		/* add external port */
		if (mac > 15) {
		    vbc->OUTPORTS_MASK_PART_2 = (0x1 << (mac - 16));
		} else {
		    vbc->OUTPORTS_MASK_PART_1 |= (0x1 << mac);
		    vbc->OUTPORTS_MASK_PART_2 = 0;
		}
		cxe_write(CXE_INT_AL_CAM_0, u128.ulBits0_31);
		cxe_write(CXE_INT_AL_CAM_1, u128.ulBits32_63);
		cxe_write(CXE_INT_AL_CAM_2, u128.ulBits64_95);
		cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE << 20) |
			   mac));

		vbc->VID = 0xfff;
		cxe_write(CXE_INT_AL_CAM_0, 0);
		cxe_write(CXE_INT_AL_CAM_1, 0x80000000);
		cxe_write(CXE_INT_AL_CAM_2, 0);
		cxe_write(CXE_INT_AL_CAM_3, u128.ulBits96_127);
		cxe_write(CXE_INT_AL_CAM_INSTR,
			  ((CXEPHYS_CAM_ACTION_WRITE_LOCAL_MASK << 20) |
			   mac));
	    }
	}
    }

    for (mac = 0; mac < CXE_PORTS; mac++) {
        if (CXE_EXT_PORT(mac)) {
	    /* Enable DoubleQ Tagging on external ports */
	    cxe_write(CXE_PD_CONFIG(mac), 0x00000110);
	}

	/* clear MAC stats */
	for (stat = 0; stat < CXE_MAC_STATS; stat++) {
	    cxe_write(CXE_MAC_STAT_INSTR(mac), stat);
	}
    }

    /* clear Address Lookup stats */
    for (stat = CXE_CNT_VLAN_DOESNT_EXIST;
	 stat <= CXE_CNT_IPX_FORW_DATAGRAMS; stat++) {
        cxe_read(stat, &val);
    }

    /* clear Queue Engine stats */
    for (stat = 0; stat < CXE_QE_DISCARD_STATS; stat++) {
        cxe_write(CXE_QE_DISCARD_INSTR, stat);
    }

    /* restart Queue Engine */
    cxe_write(CXE_QE_RESTART_COMMAND, 1);
}

int do_syrah_cxe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong bank, reg, val, error, port, stat, mac;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 2) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {
        ulong mask = 0;
	int fbm = 0;
	int bridge = 0;
        if (argc > 2) {
	    if (argv[2][0] == 'f') {
	        /* fast loopback test */
	        fbm = 1;
	    } else if (argv[2][0] == 'b') {
	        /* init simple bridge */
	        bridge = 1;
	    } else {
	        mask = simple_strtoul(argv[2], NULL, 16);
		if (mask > 3) {
		    printf("%lx: mask invalid\n", mask);
		    return 1;
		}
	    }
	}

	/* initialize CXE state */
	cxe_init(mask, fbm, bridge);

    } else if (argv[1][0] == 't') {
	for (mac = 0; mac < CXE_PORTS+1; mac++) {
	    cxe_write(CXE_INT_AL_CAM_INSTR,
		      ((CXEPHYS_CAM_ACTION_READ << 20) |
		       mac));
	    cxe_read(CXE_INT_AL_CAM_3, &val);
	    printf("%02lx: %08lx", mac, val);
	    cxe_read(CXE_INT_AL_CAM_2, &val);
	    printf(" %08lx", val);
	    cxe_read(CXE_INT_AL_CAM_1, &val);
	    printf(" %08lx", val);
	    cxe_read(CXE_INT_AL_CAM_0, &val);
	    printf(" %08lx\n", val);
	}
    } else if (argv[1][0] == 's') {
        if (argc < 3) {
	    for (port = 0; port < CXE_PORTS; port++) {
	        mac = port_map[port].mac;
	        bank = 0;
		for (stat = 0; stat < CXE_MAC_STATS; stat++) {
		    cxe_write(CXE_MAC_STAT_INSTR(mac), stat);
		    error = cxe_read(CXE_MAC_STAT_DATA_L(mac), &val);
		    if (error) {
		        printf("cxe_read %04lx: failed %08lx\n",
			       CXE_MAC_STAT_DATA_L(mac), error);
			return 1;
		    } 
		    if (val) {
		        if (!bank) printf("port %lx:\n", port);
		        printf("  %-15s %08lx", cxe_stats_str[stat], val);
			if (bank++ & 0x1) printf("\n");
		    }
		}
		if (bank & 0x1) {
		    printf("\n");
		}
	    }
	} else {
	    port = simple_strtoul(argv[2], NULL, 16);
	    if (port >= CXE_PORTS) {
	        printf ("Illegal port %lx\n", port);
		return 1;
	    }
	    mac = port_map[port].mac;

	    bank = 0;
	    for (stat = 0; stat < CXE_MAC_STATS; stat++) {
	        cxe_write(CXE_MAC_STAT_INSTR(mac), stat);
		error = cxe_read(CXE_MAC_STAT_DATA_L(mac), &val);
		if (error) {
		    printf("cxe_read %04lx: failed %08lx\n",
			   CXE_MAC_STAT_DATA_L(mac), error);
		    return 1;
		} 
		if (val) {
		    printf("  %-15s %08lx", cxe_stats_str[stat], val);
		    if (bank++ & 0x1) printf("\n");
		}
	    }
	    printf("\n");
	}
      
    } else if (argv[1][0] == 'q') {

        printf("Discard stats:\n");
        for (stat = 0; stat < CXE_QE_DISCARD_STATS; stat++) {
	    cxe_write(CXE_QE_DISCARD_INSTR, stat);
	    error = cxe_read(CXE_QE_DISCARD_DATA, &val);
	    if (error) {
	        printf("cxe_read %04x: failed %08lx\n",
		       CXE_QE_DISCARD_DATA, error);
		return 1;
	    } 
	    if (val) {
		printf("  %-15s %08lx\n", cxe_qe_discard_str[stat], val);
	    }
	}
        printf("Resource Limit queue lengths:\n");
	for (stat = 0; stat < CXE_QE_RL_INSTR_ID_LINK_MCAST; stat++) {
	    if (cxe_qe_rl_str[stat] == NULL) continue;
	    cxe_write(CXE_QE_RL_INSTR, CXE_QE_RL_INSTR_COUNT | stat);
	    error = cxe_read(CXE_QE_RL_DATA, &val);
	    if (error) {
	        printf("cxe_read %04x: failed %08lx\n",
		       CXE_QE_RL_DATA, error);
		return 1;
	    } 
	    if (val) {
		printf("  %-15s %08lx\n", cxe_qe_rl_str[stat], val);
	    }
	}

    } else {
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        bank = simple_strtoul(argv[2], NULL, 16);
	if (bank > CXE_BANKS) {
	    printf("%lx: bank invalid\n", bank);
	    return 1;
	}


	if (argv[1][0] == 'w') {
	    if (argc < 5) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	    if (argc > 5) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }
	    reg = simple_strtoul(argv[3], NULL, 16);
	    val = simple_strtoul(argv[4], NULL, 16);

	    while (times--) {
	        CXE_REG_WRITE(CXE_BASE_ADDR, reg, val, error);
		if (error) {
		    printf("%lx,%02lx: failed %08lx\n",
			   bank, reg, error);
		    return 1;
		}
	    }
	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 5) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }

	    while (times--) {
	        /* Print the lines.
		 */
	        reg = simple_strtoul(argv[3], NULL, 16);
		if (argc > 4) {
	            length = simple_strtoul(argv[4], NULL, 16);
		}
		do {
	            CXE_REG_READ(CXE_BASE_ADDR, reg, val, error);
		    if (error) {
		        printf("%lx,%02lx: failed %08lx\n",
			       bank, reg, error);
			return 1;
		    } else if (!times) {
		        printf("%lx,%02lx: %08lx\n",
			       bank, reg, val);
		    }
		    reg += 1;
		    if (ctrlc()) {
		        return 1;
		    }
		} while (--length > 0);
	    }
	}
    }

    return (rc);	
}

static void
uart_putc(int uart, const char c)
{
    if (c == '\n') {
        NS16550_putc(COM_PORTS[uart], '\r');
    }
    NS16550_putc(COM_PORTS[uart], c);
}

static void
uart_puts (int uart, const char *s)
{
    while (*s) {
        uart_putc (uart, *s++);
    }
}

static int
uart_getc(int uart)
{
    return NS16550_getc(COM_PORTS[uart]);
}

static int
uart_tstc(int uart)
{
    return NS16550_tstc(COM_PORTS[uart]);
}

static int
uart_space(int uart)
{
    return NS16550_space(COM_PORTS[uart]);
}

static int soe_loop = 0;
static int soe_cr = 0;

static void
soe_init(int port, int loop, int cr)
{
    ulong data;

    soe_loop = loop;
    soe_cr = cr;

    /* reset fifos */
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_RX_FIFO_RESET, ~0); /* all RX */
    /* set LOOPBACK bit */
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_CONTROL,
		  (soe_loop? SYR_SOE_CONTROL_LOOPBACK: 0));
    /* wait some time */
    udelay(20000);
    /* bring out of reset */
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_RX_FIFO_RESET, 0);

    /* enable flow control */
    SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_FC_CONTROL, data);
    data |= SYR_SOE_PORT(port);
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_FC_CONTROL, data);
}

static void
soe_close(int port)
{
    ulong data;

    /* disable flow control */
    SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_FC_CONTROL, data);
    data &= ~SYR_SOE_PORT(port);
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_FC_CONTROL, data);
}

static void
soe_putc(int port, const char c)
{
    ulong data;

    if ((c == '\n') && soe_cr) {
        soe_putc(port, '\r');
    }

    /* wait for TX to be ready for another character */
    do {
        SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_TX_FIFO_NOTFULL, data);
    } while (!(data & SYR_SOE_PORT(port)));

    /* write character */
    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_CONTROL,
		  ((port << SYR_SOE_CONTROL_ADDR_SHIFT) |
		   (soe_loop? SYR_SOE_CONTROL_LOOPBACK: 0) |
		   c));
}

static void
soe_puts (int port, const char *s)
{
    while (*s) {
        soe_putc (port, *s++);
    }
}

static int
soe_getc(int port)
{
    ulong data;

    do {
        SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_RX_FIFO_NOTEMPTY, data);
    } while (!(data & SYR_SOE_PORT(port)));

    SYR_REG_WRITE(SYR_BASE_ADDR, SYR_SOE_CONTROL, 
		  (SYR_SOE_CONTROL_RX |
		   (soe_loop? SYR_SOE_CONTROL_LOOPBACK: 0) |
		   (port << SYR_SOE_CONTROL_ADDR_SHIFT)));
    do {
        SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_CONTROL, data);
    } while (!(data & SYR_SOE_CONTROL_RD_DONE));

    return (int) (data & SYR_SOE_CONTROL_DATA_MASK);
}

static int
soe_tstc(int port)
{
    ulong data;

    SYR_REG_READ(SYR_BASE_ADDR, SYR_SOE_RX_FIFO_NOTEMPTY, data);

    return (data & SYR_SOE_PORT(port))? 1: 0;
}

int do_syrah_soe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    if (argc < 4) {
        printf ("Usage %d:\n%s\n", argc, cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'c') {
	int i, loop = 0, cr = 0;
        ulong iport;
        ulong eport;
	char c;

        eport = simple_strtoul(argv[2], NULL, 16);
	if (eport > MXC_MAX_APSPERLC) {
	    printf("%04lx: eport invalid\n", eport);
	    return 1;
	}

	if (argv[3][0] == 'l') {
	    loop = 1;
	    iport = eport;
	} else if (argv[3][0] == 'c') {
	    cr = 1;
	    iport = eport;
	} else {
	    iport = simple_strtoul(argv[3], NULL, 16);
	    if (iport > MXC_MAX_APSPERLC) {
	        printf("%04lx: iport invalid\n", iport);
		return 1;
	    }
	}

	soe_init(iport, loop, cr);
	if (eport != iport) {
	    soe_init(eport, loop, cr);
	}
	if (argc > 5) {
	    for (i = 5; i < argc; i++) {
	        soe_puts(eport, argv[i]);
		if (i < argc - 1) {
		    soe_putc(eport, ' ');
		} else {
		    soe_putc(eport, '\n');
		}
	    }
	}

	printf("Connecting UART A to SOE UART %lx->%lx...\n"
	       "Press <Control-y> to stop\n",
	       eport, iport);
	while (1) {
	    WATCHDOG_RESET();
	    if (uart_tstc(0)) {
	        c = uart_getc(0);
		if (c == 0x19) break; /* control-y */
		if (c == '\r') c = '\n';
		soe_putc(eport, c);
	    }
	    if (soe_tstc(iport)) {
	        c = soe_getc(iport);
		uart_putc(0, c);
	    }
	}
	soe_close(iport);
	printf("\n");

    } else {
    }

    return 0;
}

#include "tps2383.h"

int do_syrah_poe(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong port;
    ulong subport;
    ulong data;
    char nTps;
    ulong ramps = 16;
    ulong ms = 1;

    if (argc < 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'i') {
        int i;
        /* initialize POE chip */
        for (i = 0; i < 3; i++) {
	    tps_init(i);
	}

    } else if (argv[1][0] == 'u') {
        /* ramp up POE port */
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        port = simple_strtoul(argv[2], NULL, 16);
	if (argc > 3) {
            ramps = simple_strtoul(argv[3], NULL, 16);
	}
	if (argc > 4) {
	    ms = simple_strtoul(argv[4], NULL, 16);
	}

	nTps = port / 8;
	subport = port % 8;
	
	/* jog to specific port */
	if (tps_jog(subport, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_ramp_up(subport, nTps, ramps, ms);

	/* turn POE LED on */
	SYR_REG_READ(SYR_BASE_ADDR, SYR_LED_POE(port/16), data);
	data |= (SYR_LED_COLOR_GREEN << ((port%16)*2));
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_LED_POE(port/16), data);

    } else if (argv[1][0] == 'd') {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	nTps = port / 8;
	subport = port % 8;

	/* jog to specific port */
	if (tps_jog(subport, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_ramp_down(subport, nTps);
      
	/* turn POE LED off */
	SYR_REG_READ(SYR_BASE_ADDR, SYR_LED_POE(port/16), data);
	data &= ~(SYR_LED_COLOR_GREEN << ((port%16)*2));
	SYR_REG_WRITE(SYR_BASE_ADDR, SYR_LED_POE(port/16), data);

    } else if (argv[1][0] == 'p') {
        int present;
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	nTps = port / 8;
	subport = port % 8;

	/* jog to specific port */
	if (tps_jog(subport, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	if (argv[3][0] == 'r') {
	    present = tps_discover_resistor(subport, nTps);
	} else {
	    present = tps_discover_capacitor(subport, nTps);
	}
	printf("%d\n", present);

    } else if (argv[1][0] == 's') {
        int value;
        if (argc < 4) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	nTps = port / 8;
	subport = port % 8;

	/* jog to specific port */
	if (tps_jog(subport, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	value = tps_sample(subport, nTps, (argv[3][0] == 'v'));
	printf("%d m%c\n", value, (argv[3][0] == 'v')? 'V':'A');

    } else if (argv[1][0] == 'c') {

        /* calibrate TPS */
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}
        port = simple_strtoul(argv[2], NULL, 16);

	nTps = port / 8;
	subport = port % 8;

	/* jog to specific port */
	if (tps_jog(subport, nTps)) {
	    printf("jog failed\n");
	    return 1;
	}

	tps_calibrate(subport, nTps);

    } else {
        printf("%s: command invalid\n", argv[1]);
	return 1;
    }

    return (0);
}

int do_syrah_gbic(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong device, reg, val, error;
    int length = 1;
    int times = 1;
    int rc = 0;

    if ((argc < 3) || (argc > 6)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    device = simple_strtoul(argv[2], NULL, 16);
    if (device >= SYR_GBIC_DEVICES) {
        printf("%lx: device invalid\n", device);
	return 1;
    }

    if (argv[1][0] == 'd') {
	if (argc > 3) {
	    times = simple_strtoul(argv[3], NULL, 16);
	}
    } else {
	if (argc > 5) {
	    times = simple_strtoul(argv[5], NULL, 16);
	}
    }

    /* Print the lines.
     */
    while (times--) {
        if (argv[1][0] == 'd') {
	    reg = 0;
	    length = 96;
	} else {
	    reg = simple_strtoul(argv[3], NULL, 16);
	    if (reg >= SYR_GBIC_REGS) {
	        printf("%02lx: address invalid\n", reg);
		return 1;
	    }

	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 4) {
	        length = simple_strtoul(argv[4], NULL, 16);
	    }
	}
        do {
	    GBIC_EEPROM_READ(SYR_BASE_ADDR, device, reg, val, error);
	    if (error) {
	        printf("%lx,%02lx: failed %08lx\n",
		       device, reg, error);
		return 1;
	    } else if (!times) {
	        printf("%lx,%02lx: %02lx  %c\n",
		       device, reg, val, CHAR(val));
	    }
	    reg += 1;
	    if (reg >= SYR_GBIC_REGS) {
		break;
	    }
	    if (ctrlc()) {
	        return 1;
	    }
	} while (--length > 0);
    }

    return (rc);	
}

#define BUF_SIZE 2048
#define UART_NEXT(I) (((I) + 1) & (BUF_SIZE-1))
char uart_buffer[UARTS][BUF_SIZE];


int do_syrah_uartb(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    char c;
    int i;
    int prod[UARTS] = {0};
    int cons[UARTS] = {0};

    NS16550_init(1, CONFIG_BUS_CLK_FREQ / 16 / 115200);

    if (argc > 1) {
        for (i = 1; i < argc; i++) {
          uart_puts(1, argv[i]);
          uart_putc(1, ' ');
        }
	uart_putc(1, '\r');
        return 0;
    }

    printf("Connecting UART A to B...\nPress 'Q' to quit\n");
    while (1) {
        WATCHDOG_RESET();
	for (i = 0; i < UARTS; i++) {
	    if (uart_tstc(i) &&
		(UART_NEXT(prod[i]) != cons[i])) {
	        c = uart_getc(i);
		if ((c == 'Q') && (i == 0)) return 0;
		uart_buffer[i][prod[i]] = c;
		prod[i] = UART_NEXT(prod[i]);
	    }
	    if ((prod[i] != cons[i]) &&
		uart_space(i^1)) {
	        uart_putc(i^1, uart_buffer[i][cons[i]]);
		cons[i] = UART_NEXT(cons[i]);
	    }
	}
    }

    return 0;
}

int do_syrah_phy(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    ulong device, reg, val, error, port;
    int length = 1;
    int times = 1;
    int rc = 0;
    
    if ((argc < 3) || (argc > 7)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'l') {
        for (port = 0; port < CXE_PORTS; port++) {
	    device = port_map[port].phy;
	    cxe_read_phy(device, 0x1, &val);
	    cxe_read_phy(device, 0x1, &val);
	    if (val & 0x4) {
	        printf("link on %lx\n", port);
	    }
	}
    } else {
        if (argc < 3) {
	    printf ("Usage:\n%s\n", cmdtp->help);
	    return 1;
	}

        device = simple_strtoul(argv[2], NULL, 16);
	if (device >= PHY_DEVS) {
	    printf("%lx: device invalid\n", device);
	    return 1;
	}

	/* map to real phy */
	device = port_map[device].phy;

	if (argv[1][0] == 'w') {
	    if (argc < 5) {
	        printf ("Usage:\n%s\n", cmdtp->help);
		return 1;
	    }
	
	    reg = simple_strtoul(argv[3], NULL, 16);
	    if (reg >= PHY_REGS) {
	        printf("%04lx: register invalid\n", reg);
		return 1;
	    } 
	    val = simple_strtoul(argv[4], NULL, 16);
	    if (val >= 65536) {
	        printf("%02lx: value invalid\n", val);
		return 1;
	    }

	    if (argc > 6) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }

	    while (times--) {
	        error = cxe_write_phy(device, reg, val);
		if (error) {
		    printf("cxe_phy_write %lx: failed %08lx\n",
			   reg, error);
		    return 1;
		}
	    }

	} else {
	    /* If another parameter, it is the length to display.
	     * Length is the number of objects, not number of bytes.
	     */
	    if (argc > 5) {
	        times = simple_strtoul(argv[5], NULL, 16);
	    }
	    
	    while (times--) {
	        if (argc > 3) {
		    reg = simple_strtoul(argv[3], NULL, 16);
		    if (reg >= PHY_REGS) {
		        printf("%04lx: register invalid\n", reg);
			return 1;
		    } 
		} else {
		    reg = 0;
		    length = 0x20;
		}
	       if (argc > 4) {
		   length = simple_strtoul(argv[4], NULL, 16);
	       }
    
	       /* Print the lines.
		*/
	       do {
		   error = cxe_read_phy(device, reg, &val);
		   if (error) {
		       printf("%lx,%02lx: failed %lx\n",
			      device, reg, error);
		       return 1;
		   } else if (!times) {
		       printf("%lx,%02lx: %02lx\n",
			      device, reg, val & 0xffff);
		   }
		   reg += 1;
		   if (reg >= PHY_REGS) {
		       break;
		   }
		   if (ctrlc()) {
		       return 1;
		   }
	       } while (--length > 0);
	    }
	}
    }

    return (rc);	
}

extern block_dev_desc_t * ide_get_dev(int dev);

static int le32_to_int(unsigned char *le32)
{
    return ((le32[3] << 24) +
	    (le32[2] << 16) +
	    (le32[1] << 8) +
	     le32[0]
	   );
}

static int upgrade_device;
static int upgrade_block;
static int upgrade_end;
static int upgrade_cksum;

int ide_tftp_write(ulong offset, u8 *src, unsigned len)
{
    block_dev_desc_t *dev;
    int i, *ptr;

    ptr = (int *) src;
    for (i = 0; i < len/sizeof(int); i++, ptr++) {
        upgrade_cksum += *ptr;
    }

    dev = ide_get_dev(upgrade_device);
    for (; len && (upgrade_block < upgrade_end); upgrade_block++) {
        if (dev->block_write(dev->dev, upgrade_block, 1, (ulong*)src) != 1) {
	    printf("%d write failed\n", upgrade_block);
	}
	if (len < dev->blksz) {
	    len = 0;
	} else {
	    len -= dev->blksz;
	}
	src += dev->blksz;
    }
    return len? ERR_INVAL: ERR_OK;
}

#define DOS_PARTITIONS 4

int do_syrah_upgrade(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    extern block_dev_desc_t * ide_get_dev(int dev);
    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    int block, count;
    unsigned device;
    ulong lba512;
    unsigned partition;
    char c;

    if (argc == 2) {
        if (argv[1][0] == 'i') {
            run_command("fpga p", bd, 0);
            run_command("syr i", bd, 0);
            run_command("cxe i b", bd, 0);
            run_command("uartb eforw eth0 eth2", bd, 0);
            return 0;
        }
        return 1;
    }

    if (argc != 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][1] == ':') {
        device = simple_strtoul(argv[1], NULL, 16);
        partition = simple_strtoul(&argv[1][2], NULL, 16);
    } else {
        device = 0;
        partition = simple_strtoul(argv[1], NULL, 16);
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("%x: Device invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("%x: Partition invalid\n", partition);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("%x: Device invalid\n", device);
	return 1;
    }

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("%x: Device not present\n", device);
	return 1;
    }

    // adjust the number of sectors
    lba512 = adjustSectors(lba512);

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Partition type (if any) unrecognized\n");
	return 1;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d: Read failed\n", device);
	return 1;
    }

    /* The first two partitions can contain images */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        printf("%x: Partition not created\n", partition);
    }

    block = le32_to_int (pt->start4);
    count = block + le32_to_int (pt->size4);

    /* save state */
    upgrade_device = device;
    upgrade_block = block;
    upgrade_end = count;
    upgrade_cksum = 0;
    copy_filename (BootFile, argv[2], sizeof(BootFile));
    load_addr = CFG_ATA_BASE_ADDR;

    /* fire off TFTP */
    eth_set("i82559#0");
    if ((ar->size = NetLoop(bd, TFTP)) == 0) {
        return 1;
    }

    memcpy(buffer, (u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET, CFG_ENV_SECT_SIZE);
    if (buffer[0] != (device<<4|partition)) {
        printf("Try to boot this partition first ([y]/n)? ");
	if ((c = uart_getc(0)) != 'n') {
	    buffer[0] = device<<4|partition;
	    flash_write(buffer,
			CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET,
			CFG_ENV_SECT_SIZE);
	}
	printf("%c\n", c);
    }

    return (0);
}

#define FLASH_PARTS 4
#define FLASH_MARGIN 136
dos_partition_t part[FLASH_PARTS] = {
  { 0x00, 0x01, 0x01, 0x00, 0x83, 0x07, 0x60, 0x7f,
    { 0x20, 0x00, 0x00, 0x00 }, { 0xe0, 0x7f, 0x01, 0x00 } },
  { 0x00, 0x00, 0x41, 0x80, 0x83, 0x07, 0xe0, 0x00,
    { 0x00, 0x80, 0x01, 0x00 }, { 0x00, 0x81, 0x01, 0x00 } },
  { 0x00, 0x00, 0xc1, 0x01, 0x83, 0x07, 0xe0, 0xd1,
    { 0x00, 0x01, 0x03, 0x00 }, { 0x00, 0xd1, 0x00, 0x00 } },
  { 0x00 },
};

static unsigned long adjustSectors(unsigned long lba512)
{
    // printf("old lba512 = %d\n", (int) lba512);
    if((lba512 > 500000) && ( lba512 < 512000)) {
        //
        // 256 MB flash
        //
        // AVED comes in with 500400
        // WINTEC comes in with 500736
        //
        // Some web sites mention there are two types of 256 MB flashes
        // One that has 1954 blocks and one that has 1956 blocks.
        // To be safe and not trust the numbers passed in by the device 
        // we go with the smaller number of blocks i.e. 1954.
        //
        // 1954 did not work well with WINTEC flashes we try 1950
        // number of sectors = number of blocks(1950) * sectors(32) * heads(8)
        //                   = 499200
        //
        if(lba512 > 500224) {
                lba512 = 499200; // 1950 * 32 * 8;
        }
    } else {
        // 128 MB flash and 
        // all other sizes except 256MB 
        lba512 -= FLASH_MARGIN; /* SAMEER - 2/24/04 - remove sectors as margin */
    }
    // printf("new lba512 = %d\n", (int) lba512);
    return lba512;
}

int do_syrah_part(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    block_dev_desc_t *dev;
    ulong lba512;
    ulong size;
    unsigned device;
    int j;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        device = simple_strtoul(argv[1], NULL, 16);
    } else {
        device = 0;
    }
    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }

    // adjust the number of sectors
    lba512 = adjustSectors(lba512);

    for (j = 0; j < FLASH_PARTS; j++) {
        if (j < 2) {
            size = (part[j].size4[0] |
                    (part[j].size4[1] << 8) |
                    (part[j].size4[2] << 16) |
                    (part[j].size4[3] << 24));
            lba512 -= size;
        } else if (j == 2) {
            part[j].size4[0] = lba512 & 0xff;
            part[j].size4[1] = (lba512 >>  8) & 0xff;
            part[j].size4[2] = (lba512 >> 16) & 0xff;
            part[j].size4[3] = (lba512 >> 24) & 0xff;
        }
                  
        set_dos_partition(dev, &part[j], j);
    }
    print_part_dos(dev);

    return (0);
}

static int verify_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    int i;
    char rbuf[512];

    if (dev->block_read(dev->dev, sector, 1, (ulong*)rbuf) != 1) {
        printf("%d read failed\n", sector);
        return 0;
    }
    for (i = 0; i < 512; i++) {
        if (wbuf[i] != rbuf[i]) {
            printf("%d read data mismatch\n", sector);
            return 0;
        }
    }
    return 1;
}

static int format_sector(block_dev_desc_t *dev, int sector, char *wbuf)
{
    if (dev->block_write(dev->dev, sector, 1, (ulong*)wbuf) != 1) {
        printf("%d write failed\n", sector);
        return 0;
    }
    return 1;
}

static void format_partition(block_dev_desc_t *dev, int start, int end,
                             char *wbuf, int verify)
{
#define PUTC_MASK 0x7ff
    int i;

    for(i = start; i < end; i++){
        format_sector(dev, i, wbuf);
        if (!(i & PUTC_MASK)) {
            serial_putc('.');
        }
    }
    if (verify) {
        for(i = start; i < end; i++){
            if (!verify_sector(dev, i, wbuf)) {
                return;
            }
            if (!(i & PUTC_MASK)) {
                serial_putc(',');
            }
        }
    }
}

int do_syrah_format(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    block_dev_desc_t *dev;
    ulong lba512;
    unsigned int device,partition;
    char c,src[512];
    char formatPart0=0;
    char formatPart1=0;
    char formatPart2=0;
    int verify = 0;
    int start, end;
    dos_partition_t *pt;

    if (argc > 3) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);

	    switch(partition){
		    case 0: formatPart0=1; break;
		    case 1: formatPart1=1; break;
		    case 2: formatPart2=1; break;
		    default:
			printf("Invalid device %s\n", argv[1]);
			return 1;
	    }
    	    printf("Format will erase everything on the %s flash's partition %d.\n", (device?"external":"internal"), partition);

	} else {
	    device = simple_strtoul(argv[1], NULL, 16);
	    formatPart0=1;
	    formatPart1=1;
	    formatPart2=1;
    	    printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
	} 
        if (argc > 2) {
            verify = 1;
        }

    } else {
	/* default format device 0 all three partitions. */
	formatPart0=1;
	formatPart1=1;
	formatPart2=1;
        device = 0;
    	printf("Format will erase the entire %s flash.\n", (device?"external":"internal"));
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    dev = ide_get_dev(device);
    if (!dev) {
        printf("Device %x invalid\n", device);
	return 1;
    }

    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) {
        printf("Device %x not present\n", device);
	return 1;
    }
    lba512 = adjustSectors(lba512);

    printf("Are you sure you want to continue (y/n)?");
    c=getc();
    if(c!='y' && c!='Y') {
	    printf("\nFormat Cancelled\n");
	    return 1;
    }
    printf("%c\nFormatting", c);

    memset(src,0xFF,sizeof(src));

    if(formatPart0) {
        pt = &part[0];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart1) {
        pt = &part[1];
        start = le32_to_int (pt->start4);
        end = start + le32_to_int (pt->size4);
        format_partition(dev, start, end, src, verify);
    } 
    if(formatPart2) {
        pt = &part[2];
        start = le32_to_int (pt->start4);
        end = lba512;
        format_partition(dev, start, end, src, verify);
    } 
    printf("\ndone\n");

    /* partition it */
    return (do_syrah_part(cmdtp, bd, flag, 1, argv));
}

static u8 *load_partition(bd_t *bd, unsigned device, unsigned partition,
			  u32 *len)
{

    unsigned char buffer[DEFAULT_SECTOR_SIZE];
    block_dev_desc_t *dev;
    dos_partition_t *pt;
    aruba_image_t *ar;
    ulong lba512;
    u8 *addr;
    int block, count;
    int cksum, i, *ptr;

    dev = ide_get_dev(device);
    if (!dev) return NULL;

    /* valid device? */
    lba512 = (dev->lba * (dev->blksz/512));
    if (!lba512) return NULL;

    /* valid partition? */
    if (dev->part_type != PART_TYPE_DOS) {
        printf("Device %d partition type (if any) unrecognized\n", device);
	return NULL;
    }

    /* Read partition information */
    if (dev->block_read(dev->dev, 0, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, 0);
	return NULL;
    }

    /* valid partition? */
    pt = (dos_partition_t *) (buffer + DOS_PART_TBL_OFFSET) + partition;
    if (pt->sys_ind == 0) {
        return NULL;
    }

    block = le32_to_int (pt->start4);
    if (dev->block_read(dev->dev, block, 1, (ulong *)buffer) != 1) {
        printf("%d:%d:%d Read failed\n", device, partition, block);
	return NULL;
    }

    /* First block of partition contains aruba image header */
    ar = (aruba_image_t *) buffer;
    if (strcmp(ARUBA_IMAGE_MAGIC, ar->magic)) {
        printf("%d:%d Image not present\n", device, partition);
	return NULL;
    }
    if (!ar->valid) {
        printf("%d:%d Image not valid\n", device, partition);
	return NULL;
    }
    if (ar->machineType != ARUBA_IMAGE_SYRAH) {
        printf("%d:%d Image not for this platform\n", device, partition);
	return NULL;
    }

    /* remaining blocks contain actual image */
    addr = (u8 *) ((CFG_MEMTEST_END - ar->size) & 0xfff00000);
    count = (ar->size + dev->blksz - 1) / dev->blksz;
    /* read image into memory */
    printf("Loading image %d:%d", device, partition);
    if (dev->block_read(dev->dev, block + 1, count,
			(ulong *)addr) != count) {
        printf("%d:%d:%d read failed\n", device, partition, block+1);
	return NULL;
    }
    if (ctrlc())
        return NULL;

    if (ar->chksum) {
        /* verify checksum */
        printf("\nVerifying checksum... ");

	cksum = 0;
	ptr = (int *) buffer;
	for (i = 0; i < sizeof(buffer)/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	ptr = (int *) addr;
	for (i = 0; i < ar->size/sizeof(int); i++, ptr++) {
	    cksum += *ptr;
	}

	if (cksum) {
	    /* invalid cksum */
	    printf("failed!\n");
	    return NULL;
	}

    }

    if (len) {
        *len = ar->size;
    }

    return addr;
}

static void boot_partition(bd_t *bd, unsigned device, unsigned partition)
{
    u8 *addr;

    /* disable interrupts while accessing compact FLASH */
    disable_interrupts();
    addr = load_partition(bd, device, partition, NULL);
    enable_interrupts();
    if (addr) {
        printf("\nHalting active components...");
        SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_RESET_ENABLE,
                           SYR_CPLD_RESET_ENABLE_VALUE);
        SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_RESET,
                           SYR_CPLD_RESET_CFLASH);
        udelay (10 * 1000);
        SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_RESET, 0);
        SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_RESET_ENABLE, 0);
        printf("\nBooting image...\n");
	/* set load addr */
	load_addr = (ulong)addr;
	/* boot it (DOES NOT RETURN if successful) */
	do_bootelf(NULL, bd, 0, 0, NULL);
    }
}

int do_syrah_bootf(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    unsigned partition, part;
    unsigned device, dev;
    int search;

    if (argc > 2) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argc > 1) {
        search = 0;
        if (argv[1][1] == ':') {
	    device = simple_strtoul(argv[1], NULL, 16);
	    partition = simple_strtoul(&argv[1][2], NULL, 16);
	} else {
	    device = 0;
	    partition = simple_strtoul(argv[1], NULL, 16);
	}
    } else {
        device = *((u8 *)CFG_ENV_ADDR + ARUBA_IMAGE_NVRAM_OFFSET);
	partition = device & 0xf;
	device >>= 4;
	if ((device >= CFG_IDE_MAXDEVICE) ||
	    (partition >= DOS_PARTITIONS)) {
	    device = 0;
	    partition = 0;
	}
        search = 1;
    }

    if (device >= CFG_IDE_MAXDEVICE) {
        printf("Device %x invalid\n", device);
	return 1;
    }
    if (partition >= DOS_PARTITIONS) { /* max 4 partitions */
        printf("Partition %x invalid\n", partition);
	return 1;
    }

    if (!search) {
        boot_partition(bd, device, partition);
	printf("Failed\n");
	return 1;
    }

    /* try to boot primary partition */
    boot_partition(bd, device, partition);

    /* failed, try to boot from any partition */
    for (dev = 0; dev < CFG_IDE_MAXDEVICE; dev++) {
        for (part = 0; part < DOS_PARTITIONS; part++) {
	    if ((part != partition) || (dev != device)) {
	        boot_partition(bd, dev, part);
	    }
	}
    }

    return 1;
}

#define JTAG_WRITE(Val) { \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val)); \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val) | SYR_CPLD_JTAG_OPERATION_CLK); \
}

#define JTAG_WRITE_READ(Val) { \
    int Tmp; \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val)); \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val) | SYR_CPLD_JTAG_OPERATION_CLK); \
    SYR_CPLD_REG_READ(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		      Tmp); \
    printf("%d", (Tmp & SYR_CPLD_JTAG_OPERATION_TDO)? 1:0); \
}

#define JTAG_READ(Val, Rc) { \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val) | SYR_CPLD_JTAG_OPERATION_CLK); \
    SYR_CPLD_REG_READ(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		      (Rc)); \
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_JTAG_OPERATION, \
		       (Val)); \
}

static int fpga_erase(void)
{
    int i;
    ulong status;

    /* Erase FPGA */
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_FPGA_INTERFACE,
		       SYR_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_FPGA_INTERFACE,
		       SYR_CPLD_FPGA_INTERFACE_PRO|
		       SYR_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    udelay(1); /* wait at least one us */
    SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_FPGA_INTERFACE,
		       SYR_CPLD_FPGA_INTERFACE_PRO_ENABLE);
    i = 0;
    while (1) {
        udelay(1); /* wait at least one us */
	SYR_CPLD_REG_READ(SYR_CPLD_BASE_ADDR, SYR_CPLD_FPGA_INTERFACE,
			  status);
	if (status & SYR_CPLD_FPGA_INTERFACE_INIT_L) break;
	if (i++ > 100000) {
	    printf("Erase failed\n");
	    return 1;
	}
    }

    /* 1. Enter Test-Logic-Reset */
    for (i = 0; i < 10; i++) {
        JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
    }
    /* 2. Enter Run-Test-Idle */
    JTAG_WRITE(0);

    /* 3. Enter SELECT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
    }
    /* 4. Enter SHIFT-IR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }
    /* Load bypass to other FPGA */
    for (i = 0; i < 5; i++) {
        JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
    }
    /* 5. Load CFG-IN */
    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
    JTAG_WRITE(0);
    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
    JTAG_WRITE(0);
    /* 6. Load last bit of CFG-IN */
    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
    /* 7. Enter SELECT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
    }
    /* 8. Enter SHIFT-DR */
    for (i = 0; i < 2; i++) {
        JTAG_WRITE(0);
    }

    return 0;
}

int fpga_tftp_write(ulong offset, u8 *src, unsigned len)
{
    int bit, ch;
    u8 val;

    if (offset == 0) {
        if (fpga_erase()) {
	    return ERR_INVAL;
	}
    }

    for (ch = 0; ch < len-1; ch++, src++) {
        for (bit = 7; bit >= 0; bit--) {
	    val = ((*src >> bit) & 0x1) << 1;
	    JTAG_WRITE(val);
	}
    }
    /* last byte */
    for (bit = 7; bit > 0; bit--) {
        val = ((*src >> bit) & 0x1) << 1;
	JTAG_WRITE(val);
    }
    /* last bit */
    val = (*src & 0x1) << 1;
    if (len != 512) {
        printf("*");
	val |= SYR_CPLD_JTAG_OPERATION_TMS;
    }
    JTAG_WRITE(val);

    return ERR_OK;
}

#include "fpga.c"

static int fpga_flash_write(bd_t *bd)
{
    if (fpga_image) {
        printf("\nWriting image...");
        /* write data to FPGA */
        fpga_tftp_write(0, fpga_image, sizeof(fpga_image));
	printf("\n");

	/* success */
	return 0;
    }

    return 1;
}

int do_syrah_fpga(cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
    int rc = 0;
    int i;

    if ((argc < 2) || (argc > 3)) {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    if (argv[1][0] == 'p') {
        char *filename;

	filename = argv[argc-1];

	if (argc < 3) {
	    /* erase FPGA and start initial programming */
	    fpga_erase();
	    /* Program FPGA from boot flash */
	    if (fpga_flash_write(bd)) {
	        return 1;
	    }
	} else {
	    /* TFTP data into FPGA */
	    load_addr = CFG_FPGA_BASE_ADDR;
	    copy_filename (BootFile, filename, sizeof(BootFile));
	    eth_set("sibyte#0");

	    if (NetLoop(bd, TFTP) == 0) {
	        return 1;
	    }
	}

	/* 11. Enter UPDATE-DR */
	JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	/* 12. Enter SELECT-IR */
	for (i = 0; i < 2; i++) {
	   JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	}
	/* 13. Enter SHIFT-IR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(0);
	}
	/* Load bypass to other FPGA */
	for (i = 0; i < 5; i++) {
	    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
	}
	/* 14. Load JSTART */
	JTAG_WRITE(0);
	JTAG_WRITE(0);
	JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
	JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TDI);
	/* 15. Load last bit of JSTART */
	JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	/* 16. Enter SELECT-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	}
	/* 17. Enter SHIFT-DR and clock STARTUP */
	for (i = 0; i < 20; i++) {
	    JTAG_WRITE(0);
	}
	/* 18. Enter UPDATE-DR */
	for (i = 0; i < 2; i++) {
	    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	}
	/* Enter Run-Test-Idle */
	JTAG_WRITE(0);

	/* Enter Test-Logic-Reset */
	for (i = 0; i < 10; i++) {
	    JTAG_WRITE(SYR_CPLD_JTAG_OPERATION_TMS);
	}

	/* disable JTAG programming */
	SYR_CPLD_REG_WRITE(SYR_CPLD_BASE_ADDR, SYR_CPLD_FPGA_INTERFACE, 0);

	/* wait at least 120 us before accessing FPGA */
	udelay(120);

    } else {
        printf ("Usage:\n%s\n", cmdtp->help);
	return 1;
    }

    return (rc);
}
