/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/*
 * CFG register Functions
 *
 */

#include <ppcboot.h>
#include <asm/processor.h>
#include <command.h>
#include <cmd_cfg.h>

#if (CONFIG_COMMANDS & CFG_CMD_CFG)
#include <mpc824x.h>


static	ulong	base_address = 0x80000000;

/* Memory Display
 *
 * Syntax:
 *	cd{.b, .w, .l} {addr} {len}
 */
#define DISP_LINE_LEN	16
int do_cfg_cd    (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	ulong	addr, size = 4, length = 1;
	ulong	i, nbytes, linebytes;
	int rc = 0;

	/* We use the last specified parameters, unless new ones are
	 * entered.
	 */
	if (argc < 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	if ((flag & CMD_FLAG_REPEAT) == 0) {
		/* New command specified.  Check for a size specification.
		 * Defaults to long if no or incorrect specification.
		 */
		size = 4;
		if (argv[0][2] == '.') {
			if (argv[0][3] == 'b') {
				size = 1;
			} else if (argv[0][3] == 'w') {
				size = 2;
			}
		}

		/* Address is specified since argc > 1
		*/
		addr = simple_strtoul(argv[1], NULL, 16);
		addr += base_address;

		/* If another parameter, it is the length to display.
		 * Length is the number of objects, not number of bytes.
		 */
		if (argc > 2)
			length = simple_strtoul(argv[2], NULL, 16);
	}

	/* Print the lines.
	 *
	 * We buffer all read data, so we can make sure data is read only
	 * once, and all accesses are with the specified bus width.
	 */
	nbytes = length * size;
	do {
		uint    val;

		printf("%08lx:", addr);
		linebytes = (nbytes>DISP_LINE_LEN)?DISP_LINE_LEN:nbytes;
		for (i=0; i<linebytes; i+= size) {
			if (size == 4) {
			        CONFIG_READ_WORD(addr, val);
				printf(" %08x", val);
			} else if (size == 2) {
			        CONFIG_READ_HALFWORD(addr, val);
				printf(" %04x", val);
#if 0
			} else {
			        CONFIG_READ_BYTE(addr, val);
				printf(" %02x", val);
#endif
			}
			addr += size;
		}
		printf("\n");
		nbytes -= linebytes;
		if (ctrlc()) {
			rc = 1;
			break;
		}
	} while (nbytes > 0);

	return (rc);	
}

int do_cfg_cw    (cmd_tbl_t *cmdtp, bd_t *bd, int flag, int argc, char *argv[])
{
	ulong	addr, size, writeval, count;

	if ((argc < 3) || (argc > 4)) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	/* Check for size specification.
	*/
	size = 4;
	if (argv[0][2] == '.') {
		if (argv[0][3] == 'b') {
			size = 1;
		} else if (argv[0][3] == 'w') {
			size = 2;
		}
	}

	/* Address is specified since argc > 1
	*/
	addr = simple_strtoul(argv[1], NULL, 16);
	addr += base_address;

	/* Get the value to write.
	*/
	writeval = simple_strtoul(argv[2], NULL, 16);

	/* Count ? */
	if (argc == 4) {
		count = simple_strtoul(argv[3], NULL, 16);
	} else {
		count = 1;
	}

	while (count-- > 0) {
	        if (size == 4) {
		        CONFIG_WRITE_WORD(addr, (ulong)writeval);
		} else if (size == 2) {
		        CONFIG_WRITE_HALFWORD(addr, (ushort)writeval);
#if 0
		} else {
		        CONFIG_WRITE_BYTE(addr, (u_char)writeval);
#endif
		}
		addr += size;
	}
	return 0;
}

#endif	/* CFG_CMD_CFG */
