/*
 * (C) Copyright 2001, 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* ------------------------------------------------------------------------- */

/*
 * board/config.h - configuration options, board specific
 */

#ifndef __CONFIG_H
#define __CONFIG_H

/*
 * High Level Configuration Options
 * (easy to change)
 */

#define	PPCBOOT_VERSION	        "APBoot 1.2.3"

#define CONFIG_MPC8260		1
#define CONFIG_MPC8272		1
#define CONFIG_MPC8248		1
#define CONFIG_FEC_ENET         1
#define CONFIG_ETHER_ON_FCC		/* define if ether on FCC   */
#define CONFIG_ASAP     	1
#define CONFIG_BARBERA	        1

#define CONFIG_CONS_ON_SCC		/* define if console on SCC */
#define CONFIG_CONS_INDEX	1
#define CONFIG_BAUDRATE		9600
#undef  CONFIG_WATCHDOG

#define CONFIG_BOARD_PRE_INIT

/*
 * select ethernet configuration
 *
 * if either CONFIG_ETHER_ON_SCC or CONFIG_ETHER_ON_FCC is selected, then
 * CONFIG_ETHER_INDEX must be set to the channel number (1-4 for SCC, 1-3
 * for FCC)
 *
 * if CONFIG_ETHER_NONE is defined, then either the ethernet routines must be
 * defined elsewhere (as for the console), or CFG_CMD_NET must be removed
 * from CONFIG_COMMANDS to remove support for networking.
 */
#undef	CONFIG_ETHER_ON_SCC		/* define if ether on SCC   */
#define CONFIG_ETHER_ON_FCC		/* define if ether on FCC   */
#undef	CONFIG_ETHER_NONE		/* define if ether on something else */

#ifdef CONFIG_ETHER_ON_FCC

#define CONFIG_ETHER_INDEX	1	/* which SCC/FCC channel for ethernet */

#if   CONFIG_ETHER_INDEX == 1

# define CFG_PHY_ADDR		0
# define CFG_CMXFCR_VALUE	(CMXFCR_RF1CS_CLK11 | CMXFCR_TF1CS_CLK10)
# define CFG_CMXFCR_MASK	(CMXFCR_FC1 | CMXFCR_RF1CS_MSK | CMXFCR_TF1CS_MSK)

#elif CONFIG_ETHER_INDEX == 2

# define CFG_PHY_ADDR		3
# define CFG_CMXFCR_VALUE	(CMXFCR_RF2CS_CLK15 | CMXFCR_TF2CS_CLK16)
# define CFG_CMXFCR_MASK	(CMXFCR_FC2 | CMXFCR_RF2CS_MSK | CMXFCR_TF2CS_MSK)

#endif	/* CONFIG_ETHER_INDEX */

#define CFG_CPMFCR_RAMTYPE	0		/* BDs and buffers on 60x bus */
#define CFG_FCC_PSMR		(FCC_PSMR_FDE | FCC_PSMR_LPB)  /* Full duplex */

#define CONFIG_MII			/* MII PHY management		*/
#define CONFIG_BITBANGMII		/* bit-bang MII PHY management	*/
/*
 * GPIO pins used for bit-banged MII communications
 */
#define MDIO_PORT	2		/* Port C */

#define CFG_MDIO_PIN	0x00002000	/* PC18 */
#define CFG_MDC_PIN	0x00001000	/* PC19 */

#define MDIO_ACTIVE	(iop->pdir |=  CFG_MDIO_PIN)
#define MDIO_TRISTATE	(iop->pdir &= ~CFG_MDIO_PIN)
#define MDIO_READ	((iop->pdat &  CFG_MDIO_PIN) != 0)

#define MDIO(bit)	if(bit) iop->pdat |=  CFG_MDIO_PIN; \
			else	iop->pdat &= ~CFG_MDIO_PIN

#define MDC(bit)	if(bit) iop->pdat |=  CFG_MDC_PIN; \
			else	iop->pdat &= ~CFG_MDC_PIN

#define MIIDELAY	udelay(1)

#endif /* CONFIG_ETHER_ON_FCC */

#define CONFIG_COMMANDS  	(CONFIG_CMD_DFL | \
		                 CFG_CMD_ELF    | \
                                 CFG_CMD_DHCP   | \
                                 CFG_CMD_ADP    | \
                                 CFG_CMD_NET) & ~( \
				 CFG_CMD_CFG)

#define CONFIG_BOOTP_MASK       (CONFIG_BOOTP_DNS | \
                                 CONFIG_BOOTP_SUBNETMASK | \
                                 CONFIG_BOOTP_DOMAINNAME | \
                                 CONFIG_BOOTP_VENDORSPEC | \
                                 CONFIG_BOOTP_VENDORCLASS | \
                                 CONFIG_BOOTP_GATEWAY)

#define CONFIG_BOOTP_VENDOR     "ArubaAP"

#define CONFIG_BOOTCOMMAND      "tftpboot"
#define CONFIG_BOOTDELAY        2
#define CONFIG_BOOT_RETRY_TIME  120
#define CONFIG_RESET_TO_RETRY
#define CONFIG_AUTOSTART        "yes"
#define CONFIG_SERVERNAME       "aruba-master"
#define CONFIG_BOOTFILE         "sap.bin"
#define CONFIG_LAST_STAGE_INIT
#define CONFIG_I2C

/* this must be included AFTER the definition of CONFIG_COMMANDS (if any)	*/
#include <cmd_confdefs.h>

/* this is stuff came out of the Motorola docs */
#define CFG_DEFAULT_IMMR	0x0F010000

#define CFG_IMMR		0xF0000000
#define CFG_BCSR		0xF8000000
#define CFG_SDRAM_BASE		0x00000000

#define RS232EN_1		0x02000002
#define RS232EN_2		0x01000001
#define FETHIEN1		0x08000008
#define FETH1_RST		0x04000004
#define FETHIEN2		0x10000000
#define FETH2_RST		0x08000000
#define BCSR_PCI_MODE		0x01000000

/*
 * Miscellaneous configurable options
 */
#define CFG_LONGHELP	        1   		/* undef to save memory		*/
#define CFG_PROMPT		"apboot> "	/* Monitor Command Prompt	*/
#define CFG_CBSIZE		256	    	/* Console I/O Buffer Size	*/
#define CFG_CONSOLE_INFO_QUIET

/* Print Buffer Size
 */
#define CFG_PBSIZE      (CFG_CBSIZE+sizeof(CFG_PROMPT)+16)	/* Print Buffer Size	*/
#define CFG_MAXARGS		16	/* max number of command args	*/
#define CFG_BARGSIZE	CFG_CBSIZE	/* Boot Argument Buffer Size	*/
#define CFG_LOAD_ADDR	0x00100000	/* default load address		*/

#define CFG_HZ			1000	/* decrementer freq: 1 ms ticks */

#define CONFIG_NET_MULTI		/* Multi ethernet cards support */


/*-----------------------------------------------------------------------
 * Start addresses for the final memory configuration
 * (Set up by the startup code)
 * Please note that CFG_SDRAM_BASE _must_ start at 0
 */
#define CFG_DRAM_TEST
#define CFG_SDRAM_BASE		0x00000000
#define CFG_MAX_RAM_SIZE	0x01000000

#ifdef CONFIG_START_IN_RAM
#define CFG_RESET_ADDRESS	0xFFF00100
#else
#define CFG_RESET_ADDRESS	0xFFF00110
#endif

#undef	CFG_RAMBOOT
#define CFG_MONITOR_LEN		0x00040000
#define CFG_MONITOR_BASE	TEXT_BASE

#define CFG_FLASH_BASE		0xFFC00000     /* SMC was ff800000 */

#define CFG_MALLOC_LEN		(256 << 10)	/* Reserve 256 kB for malloc()	*/

#define CFG_MEMTEST_START	0x00010000	/* memtest works on		*/
#define CFG_MEMTEST_END		(bd->bi_memsize - 0x100000)

#define CFG_EUMB_ADDR		0xFC000000      /* match linux */

#define CFG_BAUDRATE_TABLE	{ 9600, 19200, 38400, 57600, 115200 }
#define CFG_FLASH_BANKS		{ CFG_FLASH_BASE }

/*-----------------------------------------------------------------------
 * Definitions for initial stack pointer and data area (in DPRAM)
 */

/*
 * Low Level Configuration Settings
 * (address mappings, register initial values, etc.)
 * You should know what you are doing if you make changes here.
 */

#define CONFIG_8260_CLKIN    66666666	/* in Hz */

#define CFG_HID0_INIT		0
#define CFG_HID0_FINAL		(HID0_ICE | HID0_IFEM | HID0_ABE )

#define CFG_HID2		0

#define CFG_SYPCR		0xFFFFFFC3
#define CFG_BCR			0x100C0000

#define CFG_SIUMCR		0x0D20C000
#define CFG_SCCR		SCCR_DFBRG01
#define CFG_BR0_PRELIM		CFG_FLASH_BASE | 0x00001801
#define CFG_OR0_PRELIM		0xFE000876
#define CFG_BR1_PRELIM		CFG_BCSR | 0x00001801
#define CFG_OR1_PRELIM		0xFFFF8010

#define CFG_BR3_PRELIM		0xF8201801   /* PCI int controller */
#define CFG_OR3_PRELIM		0xFFFF8010
#define CFG_BR4_PRELIM		0xF8300801   /* EEPROM */
#define CFG_OR4_PRELIM		0xFFFF8866
#define CFG_BR5_PRELIM		0xF8100801   /* PM5384 */
#define CFG_OR5_PRELIM		0xFFFF8E56
#define CFG_BR6_PRELIM		0xF8401801   /* CodeTEST */
#define CFG_OR6_PRELIM		0xFFFF8010

#define CFG_RMR			RMR_CSRE
#define CFG_TMCNTSC		(TMCNTSC_SEC|TMCNTSC_ALR|TMCNTSC_TCF|TMCNTSC_TCE)
#define CFG_PISCR		(PISCR_PS|PISCR_PTF|PISCR_PTE)
#define CFG_RCCR		0

#define CFG_OR2			0xFC002CC0
#define CFG_PSDMR		0x834E24A3
#define CFG_PSRT		0x13
#define CFG_MPTPR		0x2800

#if 0
/* 8272 PCI init */
#define CONFIG_PCI
#define CONFIG_PCI_PNP
#define CONFIG_PCI_BOOTDELAY 0
#undef  CONFIG_PCI_SCAN_SHOW


#define CFG_PCI_SLV_MEM_LOCAL    CFG_SDRAM_BASE /* Local base */
#define CFG_PCI_SLV_MEM_BUS      0x00000000      /* PCI base */
#define CFG_PICMR0_MASK_ATTRIB    (PICMR_MASK_512MB | PICMR_ENABLE | \
                          	 PICMR_PREFETCH_EN)

/* 
 * These are the windows that allow the CPU to access PCI address space.
 * All three PCI master windows, which allow the CPU to access PCI 
 * prefetch, non prefetch, and IO space (see below), must all fit within 
 * these windows.
 */

/* PCIBR0 */
#define CFG_PCI_MSTR0_LOCAL		0x80000000		/* Local base */
#define CFG_PCIMSK0_MASK		PCIMSK_1GB		/* Size of window */
/* PCIBR1 */
#define CFG_PCI_MSTR1_LOCAL		0xF4000000		/* Local base */
#define CFG_PCIMSK1_MASK		PCIMSK_64MB		/* Size of window */

/* 
 * Master window that allows the CPU to access PCI Memory (prefetch).
 * This window will be setup with the first set of Outbound ATU registers
 * in the bridge.
 */

#define CFG_PCI_MSTR_MEM_LOCAL	0x80000000          /* Local base */
#define CFG_PCI_MSTR_MEM_BUS	0x80000000          /* PCI base   */
#define CFG_CPU_PCI_MEM_START	PCI_MSTR_MEM_LOCAL
#define CFG_PCI_MSTR_MEM_SIZE	0x20000000          /* 512MB */
#define CFG_POCMR0_MASK_ATTRIB	(POCMR_MASK_512MB | POCMR_ENABLE | POCMR_PREFETCH_EN)

/* 
 * Master window that allows the CPU to access PCI Memory (non-prefetch).
 * This window will be setup with the second set of Outbound ATU registers
 * in the bridge.
 */

#define CFG_PCI_MSTR_MEMIO_LOCAL    0xA0000000          /* Local base */
#define CFG_PCI_MSTR_MEMIO_BUS      0xA0000000          /* PCI base   */
#define CFG_CPU_PCI_MEMIO_START     PCI_MSTR_MEMIO_LOCAL
#define CFG_PCI_MSTR_MEMIO_SIZE     0x20000000          /* 512MB */
#define CFG_POCMR1_MASK_ATTRIB      (POCMR_MASK_512MB | POCMR_ENABLE)

/* 
 * Master window that allows the CPU to access PCI IO space.
 * This window will be setup with the third set of Outbound ATU registers
 * in the bridge.
 */

#define CFG_PCI_MSTR_IO_LOCAL       0xF4000000          /* Local base */
#define CFG_PCI_MSTR_IO_BUS         0xF4000000          /* PCI base   */
#define CFG_CPU_PCI_IO_START        PCI_MSTR_IO_LOCAL
#define CFG_PCI_MSTR_IO_SIZE        0x04000000          /* 64MB */
#define CFG_POCMR2_MASK_ATTRIB      (POCMR_MASK_64MB | POCMR_ENABLE | POCMR_PCI_IO)
#endif /* CONFIG_ADSTYPE == CFG_8272ADS */

#define CFG_INIT_RAM_ADDR	CFG_IMMR
#define CFG_INIT_RAM_END	0x4000	/* End of used area in DPRAM	*/
#define CFG_INIT_DATA_SIZE	128	/* size in bytes reserved for initial data */
#define CFG_INIT_DATA_OFFSET	(CFG_INIT_RAM_END - CFG_INIT_DATA_SIZE)
#define CFG_INIT_SP_OFFSET	CFG_INIT_DATA_OFFSET

/* 0x0EA28205 */
#define CFG_HRCW_MASTER (   ( HRCW_BPS11 | HRCW_CIP )			    |\
			    ( HRCW_L2CPC10 | HRCW_DPPC10 | HRCW_ISB010 )    |\
			    ( HRCW_BMS | HRCW_APPC10 )			    |\
			    ( HRCW_MODCK_H0101 )			     \
			)

/* no slaves */
#define CFG_HRCW_SLAVE1 0
#define CFG_HRCW_SLAVE2 0
#define CFG_HRCW_SLAVE3 0
#define CFG_HRCW_SLAVE4 0
#define CFG_HRCW_SLAVE5 0
#define CFG_HRCW_SLAVE6 0
#define CFG_HRCW_SLAVE7 0

/*
 * For booting Linux, the board info and command line data
 * have to be in the first 8 MB of memory, since this is
 * the maximum mapped by the Linux kernel during initialization.
 */
#define CFG_BOOTMAPSZ		(8 << 20)	/* Initial Memory map for Linux */

/*-----------------------------------------------------------------------
 * FLASH organization
 */
#define CFG_MAX_FLASH_BANKS	1	/* max number of memory banks		*/
#define CFG_MAX_FLASH_SECT	128	/* max number of sectors on one chip	*/
#define CFG_DIRECT_FLASH_TFTP   1       /* Allow TFTP to program FLASH          */
#define CFG_FLASH_ERASE_TOUT	1000	/* Timeout for Flash Erase (in ms)	*/
#define CFG_FLASH_WRITE_TOUT	500	/* Timeout for Flash Write (in ms)	*/

/* Warining: environment is not EMBEDDED in the ppcboot code.
 * It's stored in NVRAM separately.
 */
#define CFG_ENV_IS_IN_FLASH	1
#define CFG_ENV_ADDR            0xffffc000
#define CFG_ENV_SIZE		0x00002000	/* Total Size of Environment Sector */
#define CONFIG_ENV_OVERWRITE

/*-----------------------------------------------------------------------
 * Cache Configuration
 */
#define CFG_CACHELINE_SIZE	32	/* For MPC8240 CPU		*/
#if (CONFIG_COMMANDS & CFG_CMD_KGDB)
#  define CFG_CACHELINE_SHIFT	5	/* log base 2 of the above value */
#endif

/*
 * Internal Definitions
 *
 * Boot Flags
 */
#define BOOTFLAG_COLD		0x01	/* Normal Power-On: Boot from FLASH	*/
#define BOOTFLAG_WARM		0x02	/* Software reboot			*/

#define FLASH_START             0xffc00000    /* 0xFFFFFFFF - 4MBs */
#define FLASH_SIZE             (4*1024*1024)  /* 4MB */
#define FLASH_BLOCK_SIZE        0x10000       /* 128KB */
#define OLD_MANUFAC_DATA_LOC    0xffffa000

#define SCRATCH_PAD_LOC        (BOOTLOADER_LOC + (BOOTLOADER_RES_SIZE))
#define SCRATCH_PAD_SIZE       (FLASH_BLOCK_SIZE)   /*128KB for scratch pads*/

#endif	/* __CONFIG_H */
