/*
 * (C) Copyright 2001, 2002
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

/* ------------------------------------------------------------------------- */

/*
 * board/config.h - configuration options, board specific
 */

#ifndef __CONFIG_H
#define __CONFIG_H

/*
 * High Level Configuration Options
 * (easy to change)
 */

#define	PPCBOOT_VERSION	        "CPBoot 1.2.11"

#define CONFIG_MPC824X		1
#define CONFIG_MPC8245		1
#define CONFIG_MALBEC  	1

#define CONFIG_CONS_INDEX	1
#define CONFIG_BAUDRATE		9600

#define CONFIG_BOARD_PRE_INIT
#define CONFIG_ENV_OVERWRITE

#define CONFIG_COMMANDS		( CONFIG_CMD_DFL | \
				  CFG_CMD_ELF    | \
				  CFG_CMD_I2C 	 | \
				  CFG_CMD_NET 	 | \
				  CFG_CMD_IDE 	 | \
				  CFG_CMD_MALBEC )

#define CONFIG_BOOTCOMMAND      "boot flash"
#define CONFIG_BOOTDELAY        2
#define CONFIG_BOOT_RETRY_TIME  120
#define CONFIG_RESET_TO_RETRY

/* this must be included AFTER the definition of CONFIG_COMMANDS (if any)	*/
#include <cmd_confdefs.h>

/*
 * Miscellaneous configurable options
 */
#define CFG_LONGHELP		1		/* undef to save memory		*/
#define CFG_PROMPT		"cpboot> "	/* Monitor Command Prompt	*/
#define CFG_CBSIZE		256		/* Console I/O Buffer Size	*/
#define CFG_CONSOLE_INFO_QUIET

/* Print Buffer Size
 */
#define CFG_PBSIZE (CFG_CBSIZE+sizeof(CFG_PROMPT)+16)	/* Print Buffer Size	*/
#define CFG_MAXARGS		16		/* max number of command args	*/
#define CFG_BARGSIZE		CFG_CBSIZE	/* Boot Argument Buffer Size	*/
#define CFG_LOAD_ADDR		0x08000000	/* default load address		*/

#define CFG_HZ			1000		/* decrementer freq: 1 ms ticks */

/*-----------------------------------------------------------------------
 * PCI stuff
 *-----------------------------------------------------------------------
 */
#define CONFIG_PCI				/* include pci support		*/

#define PCI_BASE_IOADDR	        0x00000000
#define PCI_BASE_MEMADDR	0xF0000000

#define CONFIG_NET_MULTI			/* Multi ethernet cards support */
#define CONFIG_SIBYTE


/*-----------------------------------------------------------------------
 * Start addresses for the final memory configuration
 * (Set up by the startup code)
 * Please note that CFG_SDRAM_BASE _must_ start at 0
 */
#define CFG_DRAM_TEST
#define CFG_SDRAM_BASE		0x00000000
#define CFG_MAX_RAM_SIZE	0x10000000

#define CFG_RESET_ADDRESS	0xFFF00110

#undef	CFG_RAMBOOT
#define CFG_MONITOR_LEN		0x0007f000
#define CFG_MONITOR_BASE	TEXT_BASE
#define CFG_BSS_MULTIPLE        4

#define CFG_INIT_DATA_SIZE	128
#define CFG_INIT_RAM_ADDR       0x40000000
#define CFG_INIT_RAM_END        0x1000
#define CFG_INIT_DATA_OFFSET    (CFG_INIT_RAM_END - CFG_INIT_DATA_SIZE)

#define CFG_MALLOC_LEN		(256 << 10)	/* Reserve 256 kB for malloc() */

#define CFG_MEMTEST_START	0x00010000	/* memtest works on */
#define CFG_MEMTEST_END		(bd->bi_memsize - (0x100000*CFG_BSS_MULTIPLE))

#define CFG_EUMB_ADDR		0xFC000000      /* match linux */

#define CONFIG_I2C		1		/* To enable I2C support */
#define CFG_I2C_MULTI_EEPROMS   1
#ifndef CFG_I2C_MULTI_EEPROMS
#define CFG_I2C_EEPROM_ADDR	0x50		/* EEPROM */
#endif
#define CFG_EEPROM_PAGE_WRITE_BITS	3
#define CFG_EEPROM_PAGE_WRITE_DELAY_MS	10	/* and takes up to 10 msec */

#define CFG_BAUDRATE_TABLE	{ 9600, 19200, 38400, 57600, 115200 }

#define CFG_FLASH_BASE		0xFF800000
#define CFG_FLASH_BANKS		{ CFG_FLASH_BASE }

/*-----------------------------------------------------------------------
 * Definitions for initial stack pointer and data area (in DPRAM)
 */

/*
 * Low Level Configuration Settings
 * (address mappings, register initial values, etc.)
 * You should know what you are doing if you make changes here.
 */

#define CONFIG_SYS_CLK_FREQ     66666666 /* external frequency to pll */
#define CONFIG_PLL_PCI_TO_MEM_MULTIPLIER 1.5
#define CONFIG_BUS_CLK_FREQ \
    (CONFIG_PLL_PCI_TO_MEM_MULTIPLIER*CONFIG_SYS_CLK_FREQ)

#define CFG_ROMNAL		0xc	/* rom/flash next access time		*/
#define CFG_ROMFAL		0x14	/* rom/flash access time		*/

#define CFG_MCCR2               0x2ce006b8

/* the following are for SDRAM only*/
#define CFG_BSTOPRE	        224	/* Burst To Precharge,set open page int */
#define CFG_REFREC		8	/* Refresh to activate interval		*/
#define CFG_RDLAT		4	/* data latency from read command	*/
#define CFG_PRETOACT		3	/* Precharge to activate interval	*/
#define CFG_ACTTOPRE		5	/* Activate to Precharge interval	*/
#define CFG_ACTORW		3	/* Activate to R/W			*/
#define CFG_SDMODE_CAS_LAT	3	/* SDMODE CAS latency			*/
#define CFG_SDMODE_WRAP		0	/* SDMODE wrap type			*/

#define CFG_REGISTERD_TYPE_BUFFER   1
#define CFG_REGDIMM             0


/* memory bank settings*/
/*
 * only bits 20-29 are actually used from these vales to set the
 * start/end address the upper two bits will be 0, and the lower 20
 * bits will be set to 0x00000 for a start address, or 0xfffff for an
 * end address
 */
#define CFG_BANK_ROW            0xa /* 13 x N x 4 for Banks 0 & 1 */
#define CFG_BANK0_START		0x00000000
#define CFG_BANK0_END		(CFG_MAX_RAM_SIZE/2 - 1)
#define CFG_BANK0_ENABLE	1
#define CFG_BANK1_START		(CFG_MAX_RAM_SIZE/2)
#define CFG_BANK1_END		(CFG_MAX_RAM_SIZE - 1)
#define CFG_BANK1_ENABLE	1
#define CFG_BANK2_START		0x3ff00000
#define CFG_BANK2_END		0x3fffffff
#define CFG_BANK2_ENABLE	0
#define CFG_BANK3_START		0x3ff00000
#define CFG_BANK3_END		0x3fffffff
#define CFG_BANK3_ENABLE	0
#define CFG_BANK4_START		0x00000000
#define CFG_BANK4_END		0x00000000
#define CFG_BANK4_ENABLE	0
#define CFG_BANK5_START		0x00000000
#define CFG_BANK5_END		0x00000000
#define CFG_BANK5_ENABLE	0
#define CFG_BANK6_START		0x00000000
#define CFG_BANK6_END		0x00000000
#define CFG_BANK6_ENABLE	0
#define CFG_BANK7_START		0x00000000
#define CFG_BANK7_END		0x00000000
#define CFG_BANK7_ENABLE	0

#define CFG_ODCR		0x94	/* configures line driver impedances	*/
#define CFG_PGMAX		0x32	/* how long the 8245 retains a page	*/
#define CFG_CLKDCR              0x7c00  /* disable PCI_CLKx                     */
#define CFG_DBUS_SIZE2          1       /* set MCCR4 DBUS_SIZE2 bit */
#define CFG_MIOCR2              0x30    /* TapDelay 3 */

/* SDRAM 0 - 256MB */
#define CFG_IBAT0L	(CFG_SDRAM_BASE | BATL_PP_10 | BATL_MEMCOHERENCE)
#define CFG_IBAT0U	(CFG_SDRAM_BASE | BATU_BL_256M | BATU_VS | BATU_VP)

/* stack in DCACHE @ 1GB (no backing mem) */
#define CFG_IBAT1L	(CFG_INIT_RAM_ADDR | BATL_PP_10 | BATL_MEMCOHERENCE)
#define CFG_IBAT1U	(CFG_INIT_RAM_ADDR | BATU_BL_128K | BATU_VS | BATU_VP)

/* Compact FLASH */
#define CFG_IBAT2L	(0x70000000 | BATL_PP_10 | BATL_CACHEINHIBIT)
#define CFG_IBAT2U	(0x70000000 | BATU_BL_128K | BATU_VS | BATU_VP)

/* Flash, CPLD, PCI, config addrs, etc */
#define CFG_IBAT3L	(0xF0000000 | BATL_PP_10 | BATL_CACHEINHIBIT)
#define CFG_IBAT3U	(0xF0000000 | BATU_BL_256M | BATU_VS | BATU_VP)

#define CFG_DBAT0L	CFG_IBAT0L
#define CFG_DBAT0U	CFG_IBAT0U
#define CFG_DBAT1L	CFG_IBAT1L
#define CFG_DBAT1U	CFG_IBAT1U
#define CFG_DBAT2L	CFG_IBAT2L
#define CFG_DBAT2U	CFG_IBAT2U
#define CFG_DBAT3L	CFG_IBAT3L
#define CFG_DBAT3U	CFG_IBAT3U

/* Extended ROM Configuration Registers for Compact FLASH */
#define CFG_EXTROM      1
#define CFG_ERCR1       0x9dff9cef /* CFLASH: ena, 16 bit, port-X */
#define CFG_ERCR2       0xa1460000 /* FPGA: ena, 32 bit, timings */
#define CFG_ERCR3       0x00000000
#define CFG_ERCR4       0x00010000

/*
 * For booting Linux, the board info and command line data
 * have to be in the first 8 MB of memory, since this is
 * the maximum mapped by the Linux kernel during initialization.
 */
#define CFG_BOOTMAPSZ		(8 << 20)	/* Initial Memory map for Linux */

/*-----------------------------------------------------------------------
 * FLASH organization
 */
#define CFG_MAX_FLASH_BANKS	1	/* max number of memory banks		*/
#define CFG_MAX_FLASH_SECT	2048	/* max number of sectors on one chip	*/
#define CFG_DIRECT_FLASH_TFTP   1       /* Allow TFTP to program FLASH          */
#define CFG_DIRECT_IDE_TFTP     1       /* Allow TFTP to program IDE            */
#define CFG_FLASH_ERASE_TOUT	120000	/* Timeout for Flash Erase (in ms)	*/
#define CFG_FLASH_WRITE_TOUT	500	/* Timeout for Flash Write (in ms)	*/

/* Warining: environment is not EMBEDDED in the ppcboot code.
 * It's stored in NVRAM separately.
 */
#define CFG_ENV_IS_IN_FLASH	1
#define CFG_ENV_ADDR            (CFG_FLASH_BASE+CFG_MONITOR_LEN)
#define CFG_ENV_SIZE		0x00000200	/* Total Size of Environment Sector */
#define CFG_ENV_SECT_SIZE       256

/*-----------------------------------------------------------------------
 * Cache Configuration
 */
#define CFG_CACHELINE_SIZE	32	/* For MPC8240 CPU		*/
#if (CONFIG_COMMANDS & CFG_CMD_KGDB)
#  define CFG_CACHELINE_SHIFT	5	/* log base 2 of the above value */
#endif

/*-----------------------------------------------------------------------
 * PCMCIA stuff
 *
 */
#define CFG_PCMCIA_MEM_ADDR	0x70000000
#define CFG_PCMCIA_MEM_SIZE	(64 << 20)

/*----------------------------------------------------------------------
 * IDE/ATA Configuration
 */
#define CFG_IDE_MAXBUS		1   /* max IDE busses */
#define CFG_IDE_MAXDEVICE	CFG_IDE_MAXBUS /* max 1 drive per IDE bus */
#define	CFG_IDE_BUS(dev)	(dev)
#define CFG_IDE_WIDTH           2   /* bytes */
#define CFG_ATA_DEVICE(x)       0   /* always 0 */

#define CFG_ATA_BASE_ADDR       CFG_PCMCIA_MEM_ADDR
#define CFG_ATA_IDE0_OFFSET	(0x01F0*CFG_IDE_WIDTH) /* default ide0 offset */
#define CFG_ATA_DATA_OFFSET	0x0000	/* data reg offset */
#define CFG_ATA_REG_OFFSET	0x0000	/* reg offset */
#define CFG_ATA_ALT_OFFSET	(0x0200*CFG_IDE_WIDTH) /* alternate register offset */

#define CONFIG_DOS_PARTITION

/*
 * Internal Definitions
 *
 * Boot Flags
 */
#define BOOTFLAG_COLD		0x01	/* Normal Power-On: Boot from FLASH	*/
#define BOOTFLAG_WARM		0x02	/* Software reboot			*/

#endif	/* __CONFIG_H */
