/*
 * (C) Copyright 2000
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <ppcboot.h>
#include <command.h>
#include "net.h"
#include "bootp.h"
#include "tftp.h"
#include "arp.h"
#include "dns.h"

#if (CONFIG_COMMANDS & CFG_CMD_NET)
#if 0
#define DEBUG
#endif

#define TIMEOUT		2		/* Seconds before trying ARP again */
#define TIMEOUT_COUNT	5		/* # of timeouts before giving up  */

#define DNS_HDR_SIZE   16

static void DnsHandler(uchar *pkt, unsigned dest, unsigned src, unsigned len);
static void DnsTimeout(void);

int	DnsTry = 0;
int	DnsRetry = 0;
int     DnsLen;

/*
 *	Handle a Dns received packet.
 */
static void
DnsHandler(uchar *pkt, unsigned dest, unsigned src, unsigned len)
{
    u16 flags;
    int namelen;
    uchar *p = pkt;
    int i;
    int answers;

    if (dest != 53) {
        return;
    }

#ifdef DEBUG
    printf("DNS receive:");
    for (i = 0; i < len; i++) {
	if (!(i & 0x7)) printf("\n");
        printf(" %02x", pkt[i]);
    }
    printf("\n");
#endif
    namelen = strlen(NetServerName) + strlen(NetOurDomainName);
    if (len < DNS_HDR_SIZE + namelen + 2 + 16) {
        return;
    }
    if (*((ushort *)pkt)++ != DnsRetry) { /* ID */
        return;
    }
    flags = *((ushort *)pkt)++;
    if ((flags & 0xFB0F) != 0x8100) { /* flags */
        return;
    }

    if ((*((ushort *)pkt)++ != 0x0001) ||             /* questions */
	((answers = *((ushort *)pkt)++) < 0x0001)) {  /* answers */
        /* do not force 0 Authority or Additional RRs */
        return;
    }

    /* skip over Authority & Additional RRs */
    ((ushort *)pkt)++; /* authority  */
    ((ushort *)pkt)++; /* additional */


    /* parse question */
    for (i = 0; i < 10; i++) {
        if (!*pkt) break;
	pkt += *pkt + 1; /* skip name */
    }
    if (i == 10) {
        /* something very wrong */
        return;
    }
    pkt++; /* skip 0 */
    /* verify host inet */
    if ((*((ushort *)pkt)++ != 0x0001) || /* host addr */
	(*((ushort *)pkt)++ != 0x0001)) { /* inet */
        return;
    }

    /* parse answers */
    for (i = 0; i < answers; i++) {
        int len;
        pkt += 2; /* name */
	if (*((ushort *)pkt)++ == 0x0001) {
	    /* Host address (A record) */
	    pkt += 8; /* class + ttl + len */
	    NetServerIP = *((ulong *)pkt);
	    break; /* we're done */
	} else {
	    /* CNAME, etc */
	    pkt += 6; /* class + ttl */
	    len = *((ushort *)pkt)++;
	    pkt += len;
	}
    }
    if (i == answers) {
        /* no address :( */
        return;
    }

    /* copy DHCP response to 64K */    
    memcpy((char *)0x10000, p, len);

    ArpTry = 0;
    ArpRequest(TFTP);
}


/*
 *	Timeout on ARP request.
 */
static void
DnsTimeout(void)
{
	if (DnsRetry >= TIMEOUT_COUNT) {
		puts ("\nRetry count exceeded; starting again\n");
		NetStartAgain ();
	} else {
		puts ("T ");
		NetSetTimeout (TIMEOUT * CFG_HZ, DnsTimeout);
		DnsSend ();
	}
}

static int DnsLength(char *str)
{
    int len = 0;
    while (*str && *str++ != '.') {
        len++;
    }
    return len;
}


void
DnsStart (void)
{
	int len;
#ifdef DEBUG
	int i;
#endif
	uchar *pkt, *p;
	char *name;

	DnsTry = 0;
	DnsRetry = 0;

	printf("DNS request %d for %s.%s to %ld.%ld.%ld.%ld\n", ++DnsTry,
	       NetServerName, NetOurDomainName,
	       NetOurDNSIP>>24&0xff, NetOurDNSIP>>16&0xff,
	       NetOurDNSIP>>8&0xff, NetOurDNSIP&0xff);

	pkt = (uchar *)NetTxPacket + ETHER_HDR_SIZE + IP_HDR_SIZE;
	len = strlen(NetServerName);
	DnsLen = DNS_HDR_SIZE;

	p = (char *)pkt;

	*((ushort *)pkt)++ = SWAP16c(0x0000); /* id */
	*((ushort *)pkt)++ = SWAP16c(0x0100); /* flags */
	*((ushort *)pkt)++ = SWAP16c(0x0001); /* questions */
	*((ushort *)pkt)++ = SWAP16c(0x0000); /* answers */
	*((ushort *)pkt)++ = SWAP16c(0x0000); /* authority */
	*((ushort *)pkt)++ = SWAP16c(0x0000); /* additional */
	name = NetServerName;
	while ((len = DnsLength(name))) {
	    *((uchar *)pkt)++  = len;         /* query length */
	    strncpy((char *)pkt, name, len);  /* name */
	    name += len + 1;                  /* skip '.' */
	    pkt += len;
	    DnsLen += len + 1;
	}
	name = NetOurDomainName;
	while ((len = DnsLength(name))) {
	    *((uchar *)pkt)++  = len;         /* query length */
	    strncpy((char *)pkt, name, len);  /* name */
	    name += len + 1;                  /* skip '.' */
	    pkt += len;
	    DnsLen += len + 1;
	}
	*((uchar *)pkt)++  = 0;             /* query length */
	DnsLen++;
	*((ushort *)pkt)++ = SWAP16c(0x0001); /* host address */
	*((ushort *)pkt)++ = SWAP16c(0x0001); /* inet */

#ifdef DEBUG
	printf("DNS send:");
	for (i = 0; i < DnsLen; i++) {
	    if (!(i & 0x7)) printf("\n");
	    printf(" %02x", p[i]);
	}
	printf("\n");
#endif

	NetSetTimeout(TIMEOUT * CFG_HZ, DnsTimeout);
	NetSetHandler(DnsHandler);

	NetSetEther (NetTxPacket, NetServerEther, PROT_IP);
	NetSetIP (NetTxPacket + ETHER_HDR_SIZE, NetOurDNSIP,
		  53, 53, DnsLen);

	DnsSend();
}

void
DnsSend (void)
{
	uchar *pkt;
	pkt = (uchar *)NetTxPacket + ETHER_HDR_SIZE + IP_HDR_SIZE;
	*(ushort *)pkt = SWAP16c(++DnsRetry); /* id */

	NetSendPacket(NetTxPacket, ETHER_HDR_SIZE + IP_HDR_SIZE + DnsLen);
}
#endif /* CFG_CMD_NET */
