/*
 *	Copyright 1994, 1995, 2000 Neil Russell.
 *	(See License)
 *	Copyright 2000, 2001 DENX Software Engineering, Wolfgang Denk, wd@denx.de
 */

#include <ppcboot.h>
#include <command.h>
#include "net.h"
#include "ping.h"
#include "bootp.h"

#undef	ET_DEBUG

#if (CONFIG_COMMANDS & CFG_CMD_NET)

#define TIMEOUT		1		/* Seconds to timeout for a lost pkt	*/
#define PING_COUNT	5		/* # of timeouts before giving up	*/
					/* (for checking the image size)	*/
static int	PingSequence;		/* The UDP port at their end		*/
static int	PingCount;

IPaddr_t	NetPingIP;

static void PingSend (void);
static void PingTimeout (void);

/**********************************************************************/

static void
PingSend (void)
{
	int			len;
	volatile IP_t  *        ip;
	volatile ICMP_t*        icmp;

	len = sizeof(ICMP_t);

	/*
	 *	We will always be sending some sort of packet, so
	 *	cobble together the packet headers now.
	 */

	NetSetEther (NetTxPacket, NetServerEther, PROT_IP);

	/*
	 *	Construct an IP and ICMP header.
	 */
	ip = (IP_t *)(NetTxPacket + ETHER_HDR_SIZE);
	ip->ip_hl_v  = 0x45;		/* IP_HDR_SIZE / 4 (not including UDP) */
	ip->ip_tos   = 0;
	ip->ip_len   = SWAP16(IP_HDR_SIZE_NO_UDP + len);
	ip->ip_id    = SWAP16(NetIPID++);
	ip->ip_off   = SWAP16c(0x4000);	/* No fragmentation */
	ip->ip_ttl   = 255;
	ip->ip_p     = 1;		/* ICMP */
	ip->ip_sum   = 0;
	ip->ip_src   = NetOurIP;
	ip->ip_dst   = NetPingIP;
	ip->ip_sum   = ~NetCksum((uchar *)ip, IP_HDR_SIZE_NO_UDP / 2);

	PingSequence = 1024 + (get_timer(0) % 3072);

	icmp = (ICMP_t *)(NetTxPacket + ETHER_HDR_SIZE + IP_HDR_SIZE_NO_UDP);
	icmp->type = ICMP_ECHO_REQUEST;
	icmp->code = 0;
	icmp->un.echo.id = 0;
	icmp->un.echo.sequence = PingSequence;
	icmp->checksum = 0;
	icmp->checksum = ~NetCksum((uchar *)icmp, sizeof(ICMP_t) / 2);

	NetSendPacket (NetTxPacket, ETHER_HDR_SIZE + IP_HDR_SIZE_NO_UDP + len);
}


static void
PingHandler (uchar * pkt, unsigned dest, unsigned src, unsigned len)
{
        ICMP_t *icmp;
	IP_t *ip;
	if (src != (ICMP_ECHO_REPLY << 8)) {
		return;
	}
	ip = (IP_t *)(pkt - IP_HDR_SIZE);

	puts("Echo reply received from ");
	print_IPaddr(ip->ip_src);
	puts("\n");
	
	if (ip->ip_src != NetPingIP) {
	        puts("Source mismatch\n");
		return;
	}

	icmp = (ICMP_t *)((uchar *)ip + IP_HDR_SIZE_NO_UDP);
	if (icmp->un.echo.sequence != PingSequence) {
	        puts("Sequence mismatch\n");
		return;
	}

	if (++PingCount >= PING_COUNT) {
	        NetState = NETLOOP_SUCCESS;
	} else {
	        PingSend();
	}
}


static void
PingTimeout (void)
{
	if (++PingCount >= PING_COUNT) {
	        NetState = NETLOOP_FAIL;
	} else {
		puts ("Timedout\n");
		NetSetTimeout (TIMEOUT * CFG_HZ, PingTimeout);
		PingSend ();
	}
}


void
PingStart (void)
{
	puts ("Ping ");	print_IPaddr (NetPingIP);
	puts ("; our address is ");	print_IPaddr (NetOurIP);

	// Check if we need to send across this subnet
	if (NetOurGatewayIP && NetOurSubnetMask) {
	    IPaddr_t OurNet 	= NetOurIP  & NetOurSubnetMask;
	    IPaddr_t ServerNet 	= NetPingIP & NetOurSubnetMask;

	    if (OurNet != ServerNet) {
		puts ("; sending through ");
		print_IPaddr (NetOurGatewayIP) ;
	    }
	}
	putc ('\n');

	NetSetTimeout (TIMEOUT * CFG_HZ, PingTimeout);
	NetSetHandler (PingHandler);

	PingCount = 0;

	PingSend ();
}

#endif /* CFG_CMD_NET */
