#include "pppoe.h"

#include <unistd.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <asm/types.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <util/sibyte_types.h>
#include <util/sibyte_l2.h>
#include <mgmtcommon.h>

#include "if_papi.h"

UINT16_t Eth_PPPOE_Discovery = ETH_PPPOE_DISCOVERY;
UINT16_t Eth_PPPOE_Session   = ETH_PPPOE_SESSION;

static uint8_t switch_mac_addr[ETH_ALEN];
static Opaque papi_hdl;
static DISPATCHER_ID dispatcher;
uint32_t pppoe_vlan = 0;
int num_lcp_echo_failures = 0;

extern PPPoEConnection *conn;
extern uint32_t ppp_get_peer_magicnum(int unit);
extern void lcp_close(int unit, char *reason);

UINT16_t
etherType(PPPoEPacket *packet)
{
    UINT16_t type = (UINT16_t) ntohs(packet->ethHdr.h_proto);
    if (type != Eth_PPPOE_Discovery && type != Eth_PPPOE_Session) {
        arubaLog(LOG_ERR, "Invalid ether type 0x%x", type);
    }
    return type;
}

int
pppoe_get_switch_mac(void)
{
struct ifreq ifr;
int sock_fd = -1, rc = 0;
  
    memset((char*)&ifr, 0, sizeof(struct ifreq));
    strncpy(ifr.ifr_name, PPPOE_INTF_NAME, sizeof(ifr.ifr_name));
  
    if ((sock_fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1) {
        goto err;
    }
  
    if ((rc = ioctl(sock_fd, SIOCGIFHWADDR, &ifr)) == -1) {
        goto err;
    }
    memcpy(switch_mac_addr, ifr.ifr_hwaddr.sa_data, ETH_ALEN);

err:
    arubaLog(LOG_DEBUG, "Unable to get switch MAC address: %s\n",
                        strerror(errno));
    if (sock_fd >= 0) {
        close(sock_fd);
    }
    return rc;
}

int
openInterface(char const *ifname, UINT16_t type, unsigned char *hwaddr)
{
PAPI_Info_Pvt *pi = (PAPI_Info_Pvt *)papi_hdl;

    memcpy(hwaddr, switch_mac_addr, ETH_ALEN);
    return (pi->sibyte[SIBYTE_MSG_OPCODE_PPPOE].sibyteSocket);
}

static int
sibyte_send_message(uint8_t *pkt, int size, uint16_t opcode)
{
char *papi_buf = PAPI_Alloc(papi_hdl, sizeof (msgHndl_Hdr) +
                                      sizeof(sibyte_msg_t) + size);
msgHndl_Hdr *msg_hdr = (msgHndl_Hdr *)papi_buf;
sibyte_msg_t *sb_hdr = (sibyte_msg_t *)((uint8_t *)msg_hdr +
                                        sizeof(msgHndl_Hdr));
uint8_t *msg_body = (uint8_t *)sb_hdr + sizeof(sibyte_msg_t);

    memset(sb_hdr, 0, sizeof(sibyte_msg_t));

    sb_hdr->opcode = opcode;
    sb_hdr->flags = SIBYTE_MSG_FLAG_HIGH_PRIO | SIBYTE_MSG_FLAG_CKSUM;
    sb_hdr->ingress = SIBYTE_DEST_CP;
    sb_hdr->vlan = pppoe_vlan;
    sb_hdr->egress = pppoe_vlan;

    memcpy(msg_body, pkt, size);

    msg_hdr->SrcPortNum = htons(PPPOE_DAEMON);
    msg_hdr->DestIpAddr = htonl(INADDR_LOOPBACK);
    msg_hdr->DestPortNum = htons(SIBYTE_RAW);
    msg_hdr->MessageCode = htons(opcode);

    size += sizeof(msgHndl_Hdr) + sizeof(sibyte_msg_t);
    PAPI_Send(papi_hdl, NULL, papi_buf, size, NON_BLOCKING_WO_ACK, NULL);
    PAPI_Free(papi_hdl, papi_buf);
    return (0);
}

void
pppoe_send_packet(uint8_t *payload, int size)
{
PPPoEPacket pkt;

    memcpy(pkt.ethHdr.h_dest, conn->peerEth, ETH_ALEN);
    memcpy(pkt.ethHdr.h_source, conn->myEth, ETH_ALEN);
    pkt.ethHdr.h_proto = htons(Eth_PPPOE_Session);
    pkt.ver = 1;
    pkt.type = 1;
    pkt.code = CODE_SESS;
    pkt.session = conn->session;
    pkt.length = htons(size);
    memcpy(pkt.payload, payload, size);
    sibyte_send_message((uint8_t *)&pkt,
                        size + sizeof(PPPoEPacket) - sizeof(pkt.payload),
                        SIBYTE_MSG_OPCODE_FORWARD);
}

int
sendPacket(PPPoEConnection *c, int sock, PPPoEPacket *pkt, int size)
{
    return (sibyte_send_message((uint8_t *)pkt, size,
                                SIBYTE_MSG_OPCODE_FORWARD));
}

int
receivePacket(int sock, PPPoEPacket *pkt, int *size)
{
uint8_t buf[sizeof(sibyte_msg_t) + sizeof(PPPoEPacket)];
struct sockaddr_in From;
PPPoEPacket *buf_pkt;
int fromLen;

    memset((char*)&From,0,sizeof(struct sockaddr_ll));
    fromLen = sizeof(From);
    if ((*size = recvfrom(sock, buf, sizeof(buf), MSG_DONTWAIT,
                          (struct sockaddr *)&From, &fromLen)) < 0) {
        // sysErr("recv (receivePacket)");
        return -1;
    }
    *size -= sizeof(sibyte_msg_t);
    buf_pkt = (PPPoEPacket *)(buf + sizeof(sibyte_msg_t));

    if (!FRAME_MACCMP(buf_pkt->ethHdr.h_source, switch_mac_addr)) {
        /*
         * Drop the packets that are generated by us.
         */
        return (-1);
    }

    if (buf_pkt->ethHdr.h_proto == FRAME_ETYPE_TAG) {
        /*
         * Remove VLAN tag
         */
        memcpy(pkt, buf_pkt, 12);
        *size -= 4;
        memcpy(&pkt->ethHdr.h_proto,
               ((uint8_t *)&buf_pkt->ethHdr.h_proto) + 4, *size - 12);
    } else {
        memcpy(pkt, buf_pkt, *size);
    }

    return (0);
}

static void
pppoe_papi_send(uint8_t *buffer, uint32_t len, uint32_t ipaddr,
                uint16_t port, uint16_t message_code, uint16_t flags)
{
char *papi_buf = PAPI_Alloc(papi_hdl, sizeof (msgHndl_Hdr) + len);
char *msg_body = papi_buf + sizeof (msgHndl_Hdr);
msgHndl_Hdr *msg_hdr = (msgHndl_Hdr *)papi_buf;

    memcpy(msg_body, buffer, len);
    if (ipaddr == 0) {
        msg_hdr->DestIpAddr = htonl(INADDR_LOOPBACK);
    } else {
        msg_hdr->DestIpAddr = htonl(ipaddr);
    }
    msg_hdr->DestPortNum = htons(port);
    msg_hdr->MessageCode = htons(message_code);
    PAPI_Send(papi_hdl, NULL, papi_buf, sizeof(msgHndl_Hdr) + len, flags, NULL);
    if (flags & NON_BLOCKING_WO_ACK) {
        PAPI_Free(papi_hdl, papi_buf);
    }
}

static void
lcp_echo_task(Opaque param, TIMER_ID timer)
{
uint8_t lcp_echo[16];
static uint8_t lcp_id = 0;

    if (num_lcp_echo_failures >= PPPOE_LCP_MAX_ECHO_FAILURES) {
        arubaLog(LOG_ERR, "Max LCP ECHO failures reached. Exiting");
        lcp_close(0, "Session terminated");
        sendPADT(conn, "PPPoE: Session terminated");
        exit(0);
    }

    if (num_lcp_echo_failures) {
        arubaLog(LOG_INFO, "No response for LCP ECHO Request. Retrying");
    }

    /*
     * Send LCP echo request
     */
    *((uint16_t *)&lcp_echo[0]) = 0xc021;
    lcp_echo[2] = LCP_ECHOREQ;
    lcp_echo[3] = lcp_id++;
    *((uint16_t *)&lcp_echo[4]) = 8; // length
    *((uint32_t *)&lcp_echo[6]) = ppp_get_peer_magicnum(0);
    pppoe_send_packet(lcp_echo, 10);
    arubaLog(LOG_DEBUG, "Sending LCP ECHOREQ");

    num_lcp_echo_failures++;
}

void
pppoe_ipcp_complete(uint32_t client_ip, uint32_t netmask,
                    uint32_t server_ip, uint32_t *dns_ip, uint32_t *wins_ip)
{
fpappsMsgDynamicIP_t pppoe_params;

    memset(&pppoe_params, 0, sizeof(pppoe_params));
    pppoe_params.owner = IP_DYNAMIC_IP_OWNER_PPPOE;
    pppoe_params.ipaddr = client_ip;
    pppoe_params.ipmask = netmask;
    pppoe_params.session_id = conn->session;
    memcpy(pppoe_params.server_mac, conn->peerEth, ETH_ALEN);

    pppoe_params.server_ip = server_ip;
    pppoe_params.router_ip[0] = server_ip;
    if (dns_ip) {
        pppoe_params.dns_ip[0] = dns_ip[0];
        if (dns_ip[0] != dns_ip[1]) {
            pppoe_params.dns_ip[1] = dns_ip[1];
        }
    }

    if (wins_ip) {
        pppoe_params.netbios_ip[0] = wins_ip[0];
        if (wins_ip[0] != wins_ip[1]) {
            pppoe_params.netbios_ip[1] = wins_ip[1];
        }
    }

    pppoe_params.msgType = FPAPPS_PPPOE_CONN_COMPLETE;
    pppoe_papi_send((uint8_t *)&pppoe_params, sizeof(pppoe_params), 0,
                    FASTPATH_SERVER, FPAPPS_PPPOE_CONN_COMPLETE,
                    NON_BLOCKING_WO_ACK);
    arubaLog(LOG_INFO, "Sending connection params to FPAPPS");

    DispAddTimer(dispatcher, PPPOE_LCP_ECHO_INTERVAL,
                 (APP_HANDLER)lcp_echo_task, NULL, TRUE);

    DispProcess(dispatcher, FALSE);
}

int
pppoe_recv_packet(uint8_t *pkt, int size)
{
PAPI_Info_Pvt *pi = (PAPI_Info_Pvt *)papi_hdl;
uint8_t buf[sizeof(sibyte_msg_t) + sizeof(PPPoEPacket)];
struct sockaddr_in From;
int fromLen, len = 0;
PPPoEPacket *pppoe_pkt;
msgHndl_Hdr *papi_hdr = (msgHndl_Hdr *)pkt;

    if ((papi_hdr->SrcPortNum == FASTPATH_SERVER) &&
        (papi_hdr->MessageCode == FPAPPS_PPPOED_TERMINATE)) {
        arubaLog(LOG_ERR, "Terminate request received from FPAPPS. Exiting");
        lcp_close(0, "User request");
        sendPADT(conn, "PPPoE: Session killed manually");
        exit(0);
    }

    memset((char*)&From,0,sizeof(struct sockaddr_ll));
    fromLen = sizeof(From);
    if ((len = recvfrom(pi->sibyte[SIBYTE_MSG_OPCODE_PPPOE].sibyteSocket,
                        buf, size, MSG_DONTWAIT,
                        (struct sockaddr *)&From, &fromLen)) < 0) {
        // sysErr("recv (receivePacket)");
        return -1;
    }
    if (len == 0) {
        return (0);
    }
    len -= sizeof(sibyte_msg_t);
    pppoe_pkt = (PPPoEPacket *)(buf + sizeof(sibyte_msg_t));

    if (pppoe_pkt->ethHdr.h_proto == FRAME_ETYPE_TAG) {
        /*
         * Remove VLAN tag
         */
        pppoe_pkt = (PPPoEPacket *)(buf + sizeof(sibyte_msg_t) + 4);
    }
    len = pppoe_pkt->length;

    memcpy(pkt, pppoe_pkt->payload, len);
    return (len);
}

static int
pppoe_papi_callback(char *buffer, int len)
{
msgHndl_Hdr *papi_hdr = (msgHndl_Hdr *)buffer;

    if ((papi_hdr->SrcPortNum == FASTPATH_SERVER) &&
        (papi_hdr->MessageCode == FPAPPS_PPPOED_TERMINATE)) {
        arubaLog(LOG_ERR, "Terminate request received from FPAPPS. Exiting");
        lcp_close(0, "User request");
        sendPADT(conn, "PPPoE: Session killed manually");
        exit(0);
    }
    return 0;
}

static int
pppoe_sibyte_receive(char *msg, int len)
{
PPPoEPacket *pppoe_pkt;

    pppoe_pkt = (PPPoEPacket *)(msg + sizeof(sibyte_msg_t));

    if (pppoe_pkt->ethHdr.h_proto == FRAME_ETYPE_TAG) {
        /*
         * Remove VLAN tag
         */
        pppoe_pkt = (PPPoEPacket *)(msg + sizeof(sibyte_msg_t) + 4);
    }
    len = pppoe_pkt->length;

    if (pppoe_pkt->ethHdr.h_proto == ETH_PPPOE_DISCOVERY) {
        if (pppoe_pkt->code == CODE_PADT) {
            arubaLog(LOG_ERR, "PADT received from peer. Exiting");
            sendPADT(conn, "PPPoE: Session terminated");
            exit(0);
        } else {
            arubaLog(LOG_DEBUG, "Ignoring packet with PPPoE Code: 0x%x",
                                pppoe_pkt->code);
        }
        return (0);
    } else if (pppoe_pkt->ethHdr.h_proto != ETH_PPPOE_SESSION) {
        arubaLog(LOG_INFO, "Invalid Ether Type: 0x%x",
                           pppoe_pkt->ethHdr.h_proto);
        return (0);
    }

    switch (*(uint16_t *)pppoe_pkt->payload) {
    case 0x8021: // IPCP
        if (pppoe_pkt->payload[2] == IPCP_CONFREQ) {
            /*
             * Send IPCP CONFACK
             */
            arubaLog(LOG_DEBUG, "Received IPCP CONFREQ. Sending CONFACK");
            pppoe_pkt->payload[2] = IPCP_CONFACK;
            pppoe_send_packet(pppoe_pkt->payload, len);
        } if (pppoe_pkt->payload[2] == IPCP_TERMREQ) {
            /*
             * Send IPCP terminate ack
             */
            arubaLog(LOG_ERR, "Received IPCP TERMREQ. Connection Terminated");
            pppoe_pkt->payload[2] = IPCP_TERMACK;
            pppoe_send_packet(pppoe_pkt->payload, len);
            lcp_close(0, "Session terminated");
            sendPADT(conn, "PPPoE: Session terminated");
            exit(0);
        }
        break;
    case 0xc021: // LCP
        if (pppoe_pkt->payload[2] == LCP_ECHOREQ) {
            /*
             * Send LCP echo reply
             */
            pppoe_pkt->payload[2] = LCP_ECHOREP;
            *(uint32_t *)(pppoe_pkt->payload + 6) = ppp_get_peer_magicnum(0);
            pppoe_send_packet(pppoe_pkt->payload, len);
            arubaLog(LOG_DEBUG, "Received LCP ECHOREQ. Sending ECHOREP");
        } else if (pppoe_pkt->payload[2] == LCP_TERMREQ) {
            /*
             * Send LCP terminate ack
             */
            arubaLog(LOG_ERR, "Received LCP TERMREQ. Connection Terminated");
            pppoe_pkt->payload[2] = LCP_TERMACK;
            *(uint32_t *)(pppoe_pkt->payload + 6) = ppp_get_peer_magicnum(0);
            pppoe_send_packet(pppoe_pkt->payload, len);
            sendPADT(conn, "PPPoE: Session terminated");
            exit(0);
        } else if (pppoe_pkt->payload[2] == LCP_ECHOREP) {
            num_lcp_echo_failures = 0;
            arubaLog(LOG_DEBUG, "Received LCP ECHOREP");
        }
        break;
    default:
        arubaLog(LOG_INFO, "PPPoED: invalid ppp proto: 0x%04x\n",
                         *(uint16_t *)pppoe_pkt->payload);
        break;
    }
    return (0);
}

static void
pppoe_papi_send_sts_callback(char *buffer, int error, Opaque args)
{
msgHndl_Hdr *papi_hdr = (msgHndl_Hdr *)buffer;

    if (error != 0) {
        arubaLog(LOG_WARNING, "PAPI_Send failed. Dest: %d",
                              papi_hdr->DestPortNum);
    }

    PAPI_Free(papi_hdl, buffer);
}

int
pppoe_papi_init(void)
{
PAPI_Info pinfo;

    if (pppoe_get_switch_mac() < 0) {
        arubaLog(LOG_ERR, "Cannot switch MAC address");
        return (-1);
    }

    dispatcher = DispatcherCreate();

    memset(&pinfo, 0, sizeof(PAPI_Info));

    pinfo.my_service = htons(PPPOE_DAEMON);
    pinfo.args.timeout = PPPOE_PAPI_TIMEOUT;
    pinfo.args.retryCount = PPPOE_PAPI_RETRIES;
    pinfo.PAPI_SendStatusCallback = pppoe_papi_send_sts_callback;
    pinfo.PAPI_ReceiveCallback = pppoe_papi_callback;
    pinfo.PacketSize = PPPOE_PAPI_MESSAGE_LEN;
    pinfo.NumPackets = PPPOE_PAPI_NUM_OUTSTANDING_MSGS;
  
    if ((papi_hdl = PAPI_Init(&pinfo, dispatcher)) == NULL) {
        arubaLog(LOG_ERR, "Cannot initialize PAPI (%s)", __FUNCTION__);
        return (-1);
    }

    /*
     * Register with SiByte for PPPOE packets
     */
    PAPI_AddSibyteOpcode(papi_hdl, SIBYTE_MSG_OPCODE_PPPOE,
                         &pppoe_sibyte_receive);

    return (0);
}

void
pppoe_papi_close()
{
    DispatcherDestroy(dispatcher);
    PAPI_Close(papi_hdl);
}
