/*
 * compat.c
 *
 * Compatibility functions for different OSes
 *
 * $Id: compat.c,v 1.2 2003/02/06 16:25:48 fenix_nl Exp $
 */

#if HAVE_CONFIG_H
#include "config.h"
#endif
#include "stdio.h"
#include "stdlib.h"
#include "string.h"
// mlord
// #include "stropts.h"
#include "grp.h"
#include "sys/ioctl.h"
#include "our_syslog.h"

/* not defined in our stdlib.h for some reason??? */
extern char *ptsname(int fd);
extern int grantpt (int fd);
extern int unlockpt (int fd);

#ifndef HAVE_STRLCPY
#include "compat.h"
#include <string.h>
#include <stdio.h>



void strlcpy(char *dst, const char *src, size_t size)
{
	strncpy(dst, src, size - 1);
	dst[size - 1] = '\0';
}
#endif

#ifndef HAVE_MEMMOVE
void *memmove(void *dst, const void *src, size_t size)
{
	bcopy(src, dst, size);
	return dst;
}
#endif

//#ifndef HAVE_OPENPTY
#if 1
/*
 * Finds a free PTY/TTY pair.
 *
 * This is derived from C.S. Ananian's pty.c that was with his pptp client.
 *
 *************************************************************************
 * pty.c - find a free pty/tty pair.
 *         inspired by the xterm source.
 *         NOTE: This is very likely to be highly non-portable.
 *         C. Scott Ananian <cananian@alumni.princeton.edu>
 *
 * Heavily modified to chage from getpseudopty() to openpty().
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#if HAVE_SYSLOG_H
#include <syslog.h>
#else
#include "our_syslog.h"
#endif

int my_openpty(int *master, int *slave, char *name, void *unused1, void *unused2)
{
#if 0
	int devindex = 0, letter = 0;
	int fd1, fd2;
	char ttydev[PTYMAX], ptydev[TTYMAX];

	syslog(LOG_DEBUG, "CTRL: Allocating pty/tty pair");
	strcpy(ttydev, TTYDEV);
	strcpy(ptydev, PTYDEV);
	while (PTYCHAR1[letter]) {
		ttydev[TTYMAX - 3] = ptydev[PTYMAX - 3] = PTYCHAR1[letter];
		while (PTYCHAR2[devindex]) {
			ttydev[TTYMAX - 2] = ptydev[PTYMAX - 2] = PTYCHAR2[devindex];
			if ((fd1 = open(ptydev, O_RDWR)) >= 0) {
				if ((fd2 = open(ttydev, O_RDWR)) >= 0) {
					goto out;
				} else {
					close(fd1);
				}
			}
			devindex++;
		}
		devindex = 0;
		letter++;
	}
	syslog(LOG_ERR, "CTRL: Failed to allocate pty");
	return -1;		/* Unable to allocate pty */

      out:
	syslog(LOG_INFO, "CTRL: Allocated pty/tty pair (%s,%s)", ptydev, ttydev);
	if (master)
		*master = fd1;
	if (slave)
		*slave = fd2;
	if (name)
		strcpy(name, ttydev);	/* no way to bounds check */
	return 0;
#else
     int fd_master, fd_slave;
     char slave_name[32];
     int ptyno;
     slave_name[0] = '\0';
     
     fd_master = open("/dev/ptmx", O_RDWR);
     if (fd_master < 0) {
         //syslog(LOG_ERR, "CTRL: Failed to allocate pty");
         return -1;
     }

     if (ioctl(fd_master, TIOCGPTN, &ptyno) == 0) {
         sprintf(slave_name, "/dev/pts/%d", ptyno);
     } else {
         //syslog_perror("ioctl");
     }
     
     //syslog(LOG_ERR, "CTRL: pty name is %s", slave_name);
     grantpt(fd_master);
     unlockpt(fd_master);
     if (name) {
         strcpy(name, slave_name);
     }
     fd_slave = open(slave_name, O_RDWR);
     if (fd_slave < 0) {
         //syslog(LOG_ERR, "CTRL: Failed to allocate slave pty %s, master pty is %d", slave_name, fd_master);
         return -1;
     }

//     ioctl(fd_slave, I_PUSH, "ptem");
//     ioctl(fd_slave, I_PUSH, "ldterm");
     if (master)
         *master = fd_master;
     if (slave)
         *slave = fd_slave;
#if 0
     syslog(LOG_INFO, "CTRL: Allocated pty/tty pair (%d,%d)",
            fd_master, fd_slave);
#endif
     return 0;
#endif
}
#endif

#ifndef HAVE_STRERROR
char *strerror(int errnum) {
	static char buf[16];
	sprintf(buf, "Error %d", errnum);
	return buf;
}
#endif

#ifndef HAVE_SETPROCTITLE
//#include "inststr.h"
/* Makefile killling me... */
#include "inststr.h"
#include "compat.h"
#include <string.h>

void inststr(int argc, char **argv, char *src)
{
	if (strlen(src) <= strlen(argv[0])) {
		char *ptr, **pptr;

		for (ptr = argv[0]; *ptr; *(ptr++) = '\0')
			;
		strcpy(argv[0], src);
		for (pptr = argv + 1; *pptr; pptr++)
			for (ptr = *pptr; *ptr; *(ptr++) = '\0')
				;
	} else {
		/* Originally from the source to perl 4.036 (assigning to $0) */
		char *ptr, *ptr2;
		int count;

		ptr = argv[0] + strlen(argv[0]);
		for (count = 1; count < argc; count++) {
			if (argv[count] == ptr + 1) {
				ptr++;
				ptr += strlen(ptr);
			}
		}
		count = 0;
		for (ptr2 = argv[0]; ptr2 <= ptr; ptr2++) {
			*ptr2 = '\0';
			count++;
		}
		strlcpy(argv[0], src, count);
	}
}
#endif

#define __USE_BSD 1
#include <stdarg.h>
#include <stdio.h>

void my_setproctitle(int argc, char **argv, const char *format, ...) {
       char proctitle[64];
       va_list parms;
       va_start(parms, format);
       vsnprintf(proctitle, sizeof(proctitle), format, parms);

#ifndef HAVE_SETPROCTITLE
       inststr(argc, argv, proctitle);
#else
       setproctitle(proctitle);
#endif
       va_end(parms);
}
