/* ifrename was added to enable fallback to DHCP if PPPoE setup
 * failed. When PPPoE is provisioned, the interface name starts
 * of as "ether0" but if we can't set it up, we have to change
 * it back to "bond0" so DHCP can work
 */
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <linux/sockios.h>      /* SIOCSIFNAME */

#define ETHER_IF	"ether0"
#define BOND_IF		"bond0"

/* argv[1] - old name ex. "ether0" */
/* argv[2] - new name ex. "bond0" */

int
main (int argc, char *argv[])
{
  struct ifreq  ifr;
  int skfd ;
  unsigned int saved_flags ;

  if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("socket") ;
    exit (0) ;
  }
  bzero(&ifr, sizeof(struct ifreq));
  if (argc > 1) 
    strncpy(ifr.ifr_name, argv[1], IFNAMSIZ);
  else
    strncpy(ifr.ifr_name, ETHER_IF, IFNAMSIZ);
  if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0) {
    perror ("GIFFLAGS") ;
    exit (1) ;
  }
  saved_flags = ifr.ifr_flags &= ~IFF_UP ;
  if (argc > 1) 
    strncpy(ifr.ifr_name, argv[1], IFNAMSIZ);
  else
    strncpy(ifr.ifr_name, ETHER_IF, IFNAMSIZ);
  if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
    perror ("SIFFLAGS") ;
    exit (3) ;
  }

  bzero(&ifr, sizeof(struct ifreq));
  if (argc > 1) 
    strcpy(ifr.ifr_name, argv[1]);
  else
    strncpy(ifr.ifr_name, ETHER_IF, IFNAMSIZ);
  if (argc > 2) 
    strcpy(ifr.ifr_newname, argv[2]);
  else
  strncpy(ifr.ifr_newname, BOND_IF, IFNAMSIZ);
  if (ioctl(skfd, SIOCSIFNAME, &ifr) < 0) {
    perror ("SIFNAME") ;
    exit (4) ;
  }
  ifr.ifr_flags = (saved_flags | IFF_UP) ;
  if (argc > 2) 
    strncpy(ifr.ifr_name, argv[2], IFNAMSIZ);
  else
    strncpy(ifr.ifr_name, BOND_IF, IFNAMSIZ);
  if (ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0) {
    perror ("SIFFLAGS") ;
    exit (5) ;
  }
  exit (0) ;
}
